# NNScore 2.02 is released under the GNU General Public License (see
# http://www.gnu.org/licenses/gpl.html). If you have any questions, comments,
# or suggestions, please don't hesitate to contact me, Jacob Durrant, at
# durrantj [at] pitt [dot] edu. If you use NNScore 2.02 in your work, please
# cite [REFERENCE HERE].


import __future__

import textwrap
import math
import os
import sys
import glob
import pickle


################################## MODIFY THIS VARIABLE TO POINT TO THE AUTODOCK VINA EXECUTABLE ##################################
vina_executable = "vina"
#vina_executable = "/PATH/TO/VINA_1_1_2/vina"
# Example: vina_executable = "/gpfs/autodock_vina_1_1_2_linux_x86/bin/vina"
###################################################################################################################################

# The ffnet class was derived from the ffnet python package developed by Marek Wojciechowski (http://ffnet.sourceforge.net/).
# As Mr. Wojciechowski's program was released under the GPL license, NNScore is likewise GPL licensed.
class ffnet:

    def load(self, net_array):

        self.outno = net_array['outno']
        self.eni = net_array['eni']
        self.weights = net_array['weights']
        self.conec = net_array['conec']
        self.deo = net_array['deo']
        self.inno = net_array['inno']
        self.units = {}

        self.output = {}

    def normcall(self, input):
        #Takes single input pattern and returns network output
        #This have the same functionality as recall but now input and
        #output are normalized inside the function.
        #eni = [ai, bi], eno = [ao, bo] - parameters of linear mapping

        self.input = input

        #set input units
        self.setin()

        #propagate signals
        self.prop()

        #get output
        self.getout()

        return self.output[1]

    def setin(self):
        #normalize and set input units

        for k in range(1,len(self.inno) + 1):
            self.units[self.inno[k]] = self.eni[k][1] * self.input[k-1] + self.eni[k][2] # because self.input is a python list, and the others were inputted from a fortran-format file

    def prop(self):
        #Gets conec and units with input already set
        #and calculates all activations.
        #Identity input and sigmoid activation function for other units
        #is assumed

        #propagate signals with sigmoid activation function
        if len(self.conec) > 0:
            ctrg = self.conec[1][2]
            self.units[ctrg] = 0.0
            for xn in range(1,len(self.conec) + 1):
                src = self.conec[xn][1]
                trg = self.conec[xn][2]
                # if next unit
                if trg != ctrg:
                    self.units[ctrg] = 1.0/(1.0+math.exp(-self.units[ctrg]))
                    ctrg = trg
                    if src == 0: # handle bias
                        self.units[ctrg] = self.weights[xn]
                    else:
                        self.units[ctrg] = self.units[src] * self.weights[xn]
                else:
                    if src == 0: # handle bias
                        self.units[ctrg] = self.units[ctrg] + self.weights[xn]
                    else:
                        self.units[ctrg] = self.units[ctrg] + self.units[src] * self.weights[xn]
            self.units[ctrg] = 1.0/(1.0+math.exp(-self.units[ctrg])) # for last unit

    def getout(self):
        #get and denormalize output units

        for k in range(1,len(self.outno)+1):
            self.output[k] = self.deo[k][1] * self.units[self.outno[k]] + self.deo[k][2]

class point:
    x=99999.0
    y=99999.0
    z=99999.0

    def __init__ (self, x, y ,z):
        self.x = x
        self.y = y
        self.z = z

    def copy_of(self):
        return point(self.x, self.y, self.z)

    def dist_to(self,apoint):
        return math.sqrt(math.pow(self.x - apoint.x,2) + math.pow(self.y - apoint.y,2) + math.pow(self.z - apoint.z,2))

    def Magnitude(self):
        return self.dist_to(point(0,0,0))

    def CreatePDBLine(self, index):

        output = "ATOM "
        output = output + str(index).rjust(6) + "X".rjust(5) + "XXX".rjust(4)
        output = output + ("%.3f" % self.x).rjust(18)
        output = output + ("%.3f" % self.y).rjust(8)
        output = output + ("%.3f" % self.z).rjust(8)
        output = output + "X".rjust(24)
        return output

class atom:

    def __init__ (self):
        self.atomname = ""
        self.residue = ""
        self.coordinates = point(99999, 99999, 99999)
        self.element = ""
        self.PDBIndex = ""
        self.line=""
        self.atomtype=""
        self.IndeciesOfAtomsConnecting=[]
        self.charge = 0
        self.resid = 0
        self.chain = ""
        self.structure = ""
        self.comment = ""

    def copy_of(self):
        theatom = atom()
        theatom.atomname = self.atomname
        theatom.residue = self.residue
        theatom.coordinates = self.coordinates.copy_of()
        theatom.element = self.element
        theatom.PDBIndex = self.PDBIndex
        theatom.line= self.line
        theatom.atomtype= self.atomtype
        theatom.IndeciesOfAtomsConnecting = self.IndeciesOfAtomsConnecting[:]
        theatom.charge = self.charge
        theatom.resid = self.resid
        theatom.chain = self.chain
        theatom.structure = self.structure
        theatom.comment = self.comment

        return theatom

    def CreatePDBLine(self, index):

        output = "ATOM "
        output = output + str(index).rjust(6) + self.atomname.rjust(5) + self.residue.rjust(4)
        output = output + ("%.3f" % self.coordinates.x).rjust(18)
        output = output + ("%.3f" % self.coordinates.y).rjust(8)
        output = output + ("%.3f" % self.coordinates.z).rjust(8)
        output = output + self.element.rjust(24)
        return output

    def NumberOfNeighbors(self):
        return len(self.IndeciesOfAtomsConnecting)

    def AddNeighborAtomIndex(self, index):
        if not (index in self.IndeciesOfAtomsConnecting):
            self.IndeciesOfAtomsConnecting.append(index)

    def SideChainOrBackBone(self): # only really applies to proteins, assuming standard atom names
        if self.atomname.strip() == "CA" or self.atomname.strip() == "C" or self.atomname.strip() == "O" or self.atomname.strip() == "N":
            return "BACKBONE"
        else:
            return "SIDECHAIN"

    def ReadPDBLine(self, Line):
        self.line = Line
        self.atomname = Line[11:16].strip()

        if len(self.atomname)==1:
            self.atomname = self.atomname + "  "
        elif len(self.atomname)==2:
            self.atomname = self.atomname + " "
        elif len(self.atomname)==3:
            self.atomname = self.atomname + " " # This line is necessary to work, though many PDBs in the PDB would have this line commented out

        self.coordinates = point(float(Line[30:38]), float(Line[38:46]), float(Line[46:54]))

        # now atom type (for pdbqt)
        self.atomtype = Line[77:79].strip().upper()

        if Line[69:76].strip() != "":
            self.charge = float(Line[69:76])
        else:
            self.charge = 0.0

        if self.element == "": # try to guess at element from name
            two_letters = self.atomname[0:2].strip().upper()
            if two_letters=='BR':
                self.element='BR'
            elif two_letters=='CL':
                self.element='CL'
            elif two_letters=='BI':
                self.element='BI'
            elif two_letters=='AS':
                self.element='AS'
            elif two_letters=='AG':
                self.element='AG'
            elif two_letters=='LI':
                self.element='LI'
            #elif two_letters=='HG':
            # self.element='HG'
            elif two_letters=='MG':
                self.element='MG'
            elif two_letters=='MN':
                self.element='MN'
            elif two_letters=='RH':
                self.element='RH'
            elif two_letters=='ZN':
                self.element='ZN'
            elif two_letters=='FE':
                self.element='FE'
            else: #So, just assume it's the first letter.
                # Any number needs to be removed from the element name
                self.element = self.atomname
                self.element = self.element.replace('0','')
                self.element = self.element.replace('1','')
                self.element = self.element.replace('2','')
                self.element = self.element.replace('3','')
                self.element = self.element.replace('4','')
                self.element = self.element.replace('5','')
                self.element = self.element.replace('6','')
                self.element = self.element.replace('7','')
                self.element = self.element.replace('8','')
                self.element = self.element.replace('9','')
                self.element = self.element.replace('@','')

                self.element = self.element[0:1].strip().upper()

        self.PDBIndex = Line[6:12].strip()
        self.residue = Line[16:20]
        self.residue = " " + self.residue[-3:] # this only uses the rightmost three characters, essentially removing unique rotamer identification

        if Line[23:26].strip() != "": self.resid = int(Line[23:26])
        else: self.resid = 1

        self.chain = Line[21:22]
        if self.residue.strip() == "": self.residue = " MOL"

class PDB:

    def __init__ (self):
        self.AllAtoms={}
        self.NonProteinAtoms = {}
        self.max_x = -9999.99
        self.min_x = 9999.99
        self.max_y = -9999.99
        self.min_y = 9999.99
        self.max_z = -9999.99
        self.min_z = 9999.99
        self.rotateable_bonds_count = 0
        self.functions = MathFunctions()
        self.protein_resnames = ["ALA", "ARG", "ASN", "ASP", "ASH", "ASX", "CYS", "CYM", "CYX", "GLN", "GLU", "GLH", "GLX", "GLY", "HIS", "HID", "HIE", "HIP", "ILE", "LEU", "LYS", "LYN", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL"]
        self.aromatic_rings = []
        self.charges = [] # a list of points
        self.OrigFileName = ""

    def LoadPDB_from_file(self, FileName, line_header=""):

        self.line_header=line_header

        # Now load the file into a list
        file = open(FileName,"r")
        lines = file.readlines()
        file.close()
        self.LoadPDB_from_list(lines, self.line_header)

    def LoadPDB_from_list(self, lines, line_header=""):

        self.line_header=line_header

        autoindex = 1

        self.__init__()

        atom_already_loaded = [] # going to keep track of atomname_resid_chain pairs, to make sure redundants aren't loaded. This basically
                                 # gets rid of rotomers, I think.

        for t in range(0,len(lines)):
            line=lines[t]

            if "between atoms" in line and " A " in line:
                    self.rotateable_bonds_count = self.rotateable_bonds_count + 1

            if len(line) >= 7:
                if line[0:4]=="ATOM" or line[0:6]=="HETATM": # Load atom data (coordinates, etc.)
                    TempAtom = atom()
                    TempAtom.ReadPDBLine(line)

                    key = TempAtom.atomname.strip() + "_" + str(TempAtom.resid) + "_" + TempAtom.residue.strip() + "_" + TempAtom.chain.strip() # this string unique identifies each atom

                    if key in atom_already_loaded and TempAtom.residue.strip() in self.protein_resnames: # so this is a receptor atom that has already been loaded once
                        print((self.line_header + "WARNING: Duplicate receptor atom detected: \"" + TempAtom.line.strip()+ "\". Not loading this duplicate."))

                    if not key in atom_already_loaded or not TempAtom.residue.strip() in self.protein_resnames: # so either the atom hasn't been loaded, or else it's a non-receptor atom
                                                                                                        # so note that non-receptor atoms can have redundant names, but receptor atoms cannot.
                                                                                                        # This is because protein residues often contain rotamers
                        atom_already_loaded.append(key) # so each atom can only be loaded once. No rotamers.
                        self.AllAtoms[autoindex] = TempAtom # So you're actually reindexing everything here.
                        if not TempAtom.residue[-3:] in self.protein_resnames: self.NonProteinAtoms[autoindex] = TempAtom

                        autoindex = autoindex + 1

        self.CheckProteinFormat()

        self.CreateBondsByDistance() # only for the ligand, because bonds can be inferred based on atomnames from PDB
        self.assign_aromatic_rings()
        self.assign_charges()

    def printout(self, thestring):
        lines = textwrap.wrap(thestring, 80)
        for line in lines:
            print(line)

    def SavePDB(self, filename):
        f = open(filename, 'w')
        towrite = self.SavePDBString()
        if towrite.strip() == "": towrite = "ATOM      1  X   XXX             0.000   0.000   0.000                       X" # just so no PDB is empty, VMD will load them all
        f.write(towrite)
        f.close()

    def SavePDBString(self):

        ToOutput = ""

        # write coordinates
        for atomindex in self.AllAtoms:
            ToOutput = ToOutput + self.AllAtoms[atomindex].CreatePDBLine(atomindex) + "\n"

        return ToOutput

    def AddNewAtom(self, atom):

        # first get available index
        t = 1
        while t in list(self.AllAtoms.keys()):
            t = t + 1

        # now add atom
        self.AllAtoms[t] = atom

    def connected_atoms_of_given_element(self, index, connected_atom_element):
        atom = self.AllAtoms[index]
        connected_atoms = []
        for index2 in atom.IndeciesOfAtomsConnecting:
            atom2 = self.AllAtoms[index2]
            if atom2.element == connected_atom_element:
                connected_atoms.append(index2)
        return connected_atoms

    def connected_heavy_atoms(self, index):
        atom = self.AllAtoms[index]
        connected_atoms = []
        for index2 in atom.IndeciesOfAtomsConnecting:
            atom2 = self.AllAtoms[index2]
            if atom2.element != "H": connected_atoms.append(index2)
        return connected_atoms

    def CheckProteinFormat(self):
        curr_res = ""
        first = True
        residue = []

        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]

            key = atom.residue + "_" + str(atom.resid) + "_" + atom.chain

            if first is True:
                curr_res = key
                first = False

            if key != curr_res:

                self.CheckProteinFormat_process_residue(residue, last_key)

                residue = []
                curr_res = key

            residue.append(atom.atomname.strip())
            last_key = key

        self.CheckProteinFormat_process_residue(residue, last_key)


    def CheckProteinFormat_process_residue(self, residue, last_key):
        temp = last_key.strip().split("_")
        resname = temp[0]
        real_resname = resname[-3:]
        resid = temp[1]
        chain = temp[2]

        if real_resname in self.protein_resnames: # so it's a protein residue

            if not "N" in residue:
                self.printout('WARNING: There is no atom named "N" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine secondary structure. If this residue is far from the active site, this warning may not affect the NNScore.')
                print("")
            if not "C" in residue:
                self.printout('WARNING: There is no atom named "C" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine secondary structure. If this residue is far from the active site, this warning may not affect the NNScore.')
                print("")
            if not "CA" in residue:
                self.printout('WARNING: There is no atom named "CA" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine secondary structure. If this residue is far from the active site, this warning may not affect the NNScore.')
                print("")

            if real_resname == "GLU" or real_resname == "GLH" or real_resname == "GLX":
                if not "OE1" in residue:
                    self.printout('WARNING: There is no atom named "OE1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "OE2" in residue:
                    self.printout('WARNING: There is no atom named "OE2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "ASP" or real_resname == "ASH" or real_resname == "ASX":
                if not "OD1" in residue:
                    self.printout('WARNING: There is no atom named "OD1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "OD2" in residue:
                    self.printout('WARNING: There is no atom named "OD2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "LYS" or real_resname == "LYN":
                if not "NZ" in residue:
                    self.printout('WARNING: There is no atom named "NZ" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-cation and salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "ARG":
                if not "NH1" in residue:
                    self.printout('WARNING: There is no atom named "NH1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-cation and salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "NH2" in residue:
                    self.printout('WARNING: There is no atom named "NH2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-cation and salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "HIS" or real_resname == "HID" or real_resname == "HIE" or real_resname == "HIP":
                if not "NE2" in residue:
                    self.printout('WARNING: There is no atom named "NE2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-cation and salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "ND1" in residue:
                    self.printout('WARNING: There is no atom named "ND1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-cation and salt-bridge interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "PHE":
                if not "CG" in residue:
                    self.printout('WARNING: There is no atom named "CG" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CD1" in residue:
                    self.printout('WARNING: There is no atom named "CD1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CD2" in residue:
                    self.printout('WARNING: There is no atom named "CD2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CE1" in residue:
                    self.printout('WARNING: There is no atom named "CE1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CE2" in residue:
                    self.printout('WARNING: There is no atom named "CE2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CZ" in residue:
                    self.printout('WARNING: There is no atom named "CZ" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "TYR":
                if not "CG" in residue:
                    self.printout('WARNING: There is no atom named "CG" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CD1" in residue:
                    self.printout('WARNING: There is no atom named "CD1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CD2" in residue:
                    self.printout('WARNING: There is no atom named "CD2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CE1" in residue:
                    self.printout('WARNING: There is no atom named "CE1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CE2" in residue:
                    self.printout('WARNING: There is no atom named "CE2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CZ" in residue:
                    self.printout('WARNING: There is no atom named "CZ" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "TRP":
                if not "CG" in residue:
                    self.printout('WARNING: There is no atom named "CG" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CD1" in residue:
                    self.printout('WARNING: There is no atom named "CD1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CD2" in residue:
                    self.printout('WARNING: There is no atom named "CD2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "NE1" in residue:
                    self.printout('WARNING: There is no atom named "NE1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CE2" in residue:
                    self.printout('WARNING: There is no atom named "CE2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CE3" in residue:
                    self.printout('WARNING: There is no atom named "CE3" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CZ2" in residue:
                    self.printout('WARNING: There is no atom named "CZ2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CZ3" in residue:
                    self.printout('WARNING: There is no atom named "CZ3" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CH2" in residue:
                    self.printout('WARNING: There is no atom named "CH2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")

            if real_resname == "HIS" or real_resname == "HID" or real_resname == "HIE" or real_resname == "HIP":
                if not "CG" in residue:
                    self.printout('WARNING: There is no atom named "CG" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "ND1" in residue:
                    self.printout('WARNING: There is no atom named "ND1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CD2" in residue:
                    self.printout('WARNING: There is no atom named "CD2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "CE1" in residue:
                    self.printout('WARNING: There is no atom named "CE1" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")
                if not "NE2" in residue:
                    self.printout('WARNING: There is no atom named "NE2" in the protein residue ' + last_key + '. Please use standard naming conventions for all protein residues. This atom is needed to determine pi-pi and pi-cation interactions. If this residue is far from the active site, this warning may not affect the NNScore.')
                    print("")


    # Functions to determine the bond connectivity based on distance
    # ==============================================================

    def CreateBondsByDistance(self):
        for AtomIndex1 in self.NonProteinAtoms:
            atom1 = self.NonProteinAtoms[AtomIndex1]
            if not atom1.residue[-3:] in self.protein_resnames: # so it's not a protein residue
                for AtomIndex2 in self.NonProteinAtoms:
                    if AtomIndex1 != AtomIndex2:
                        atom2 = self.NonProteinAtoms[AtomIndex2]
                        if not atom2.residue[-3:] in self.protein_resnames: # so it's not a protein residue
                            dist = self.functions.distance(atom1.coordinates, atom2.coordinates)

                            if (dist < self.BondLength(atom1.element, atom2.element) * 1.2):
                                atom1.AddNeighborAtomIndex(AtomIndex2)
                                atom2.AddNeighborAtomIndex(AtomIndex1)

    def BondLength(self, element1, element2):

        '''Bond lengths taken from Handbook of Chemistry and Physics. The information provided there was very specific,
        so I tried to pick representative examples and used the bond lengths from those. Sitautions could arise where these
        lengths would be incorrect, probably slight errors (<0.06) in the hundreds.'''

        distance = 0.0
        if element1 == "C" and element2 == "C": distance = 1.53
        if element1 == "N" and element2 == "N": distance = 1.425
        if element1 == "O" and element2 == "O": distance = 1.469
        if element1 == "S" and element2 == "S": distance = 2.048
        if (element1 == "C" and element2 == "H") or (element1 == "H" and element2 == "C"): distance = 1.059
        if (element1 == "C" and element2 == "N") or (element1 == "N" and element2 == "C"): distance = 1.469
        if (element1 == "C" and element2 == "O") or (element1 == "O" and element2 == "C"): distance = 1.413
        if (element1 == "C" and element2 == "S") or (element1 == "S" and element2 == "C"): distance = 1.819
        if (element1 == "N" and element2 == "H") or (element1 == "H" and element2 == "N"): distance = 1.009
        if (element1 == "N" and element2 == "O") or (element1 == "O" and element2 == "N"): distance = 1.463
        if (element1 == "O" and element2 == "S") or (element1 == "S" and element2 == "O"): distance = 1.577
        if (element1 == "O" and element2 == "H") or (element1 == "H" and element2 == "O"): distance = 0.967
        if (element1 == "S" and element2 == "H") or (element1 == "H" and element2 == "S"): distance = 2.025/1.5 # This one not from source sited above. Not sure where it's from, but it wouldn't ever be used in the current context ("AutoGrow")
        if (element1 == "S" and element2 == "N") or (element1 == "N" and element2 == "S"): distance = 1.633

        if (element1 == "C" and element2 == "F") or (element1 == "F" and element2 == "C"): distance = 1.399
        if (element1 == "C" and element2 == "CL") or (element1 == "CL" and element2 == "C"): distance = 1.790
        if (element1 == "C" and element2 == "BR") or (element1 == "BR" and element2 == "C"): distance = 1.910
        if (element1 == "C" and element2 == "I") or (element1 == "I" and element2 == "C"): distance=2.162

        if (element1 == "S" and element2 == "BR") or (element1 == "BR" and element2 == "S"): distance = 2.321
        if (element1 == "S" and element2 == "CL") or (element1 == "CL" and element2 == "S"): distance = 2.283
        if (element1 == "S" and element2 == "F") or (element1 == "F" and element2 == "S"): distance = 1.640
        if (element1 == "S" and element2 == "I") or (element1 == "I" and element2 == "S"): distance= 2.687

        if (element1 == "P" and element2 == "BR") or (element1 == "BR" and element2 == "P"): distance = 2.366
        if (element1 == "P" and element2 == "CL") or (element1 == "CL" and element2 == "P"): distance = 2.008
        if (element1 == "P" and element2 == "F") or (element1 == "F" and element2 == "P"): distance = 1.495
        if (element1 == "P" and element2 == "I") or (element1 == "I" and element2 == "P"): distance= 2.490
        if (element1 == "P" and element2 == "O") or (element1 == "O" and element2 == "P"): distance= 1.6 # estimate based on eye balling Handbook of Chemistry and Physics

        if (element1 == "N" and element2 == "BR") or (element1 == "BR" and element2 == "N"): distance = 1.843
        if (element1 == "N" and element2 == "CL") or (element1 == "CL" and element2 == "N"): distance = 1.743
        if (element1 == "N" and element2 == "F") or (element1 == "F" and element2 == "N"): distance = 1.406
        if (element1 == "N" and element2 == "I") or (element1 == "I" and element2 == "N"): distance= 2.2

        if (element1 == "SI" and element2 == "BR") or (element1 == "BR" and element2 == "SI"): distance = 2.284
        if (element1 == "SI" and element2 == "CL") or (element1 == "CL" and element2 == "SI"): distance = 2.072
        if (element1 == "SI" and element2 == "F") or (element1 == "F" and element2 == "SI"): distance = 1.636
        if (element1 == "SI" and element2 == "P") or (element1 == "P" and element2 == "SI"): distance= 2.264
        if (element1 == "SI" and element2 == "S") or (element1 == "S" and element2 == "SI"): distance= 2.145
        if (element1 == "SI" and element2 == "SI") or (element1 == "SI" and element2 == "SI"): distance= 2.359
        if (element1 == "SI" and element2 == "C") or (element1 == "C" and element2 == "SI"): distance= 1.888
        if (element1 == "SI" and element2 == "N") or (element1 == "N" and element2 == "SI"): distance= 1.743
        if (element1 == "SI" and element2 == "O") or (element1 == "O" and element2 == "SI"): distance= 1.631

        return distance;

    # Functions to identify positive charges
    # ======================================

    def assign_charges(self):
        # Get all the quartinary amines on non-protein residues (these are the only non-protein groups that will be identified as positively charged)
        AllCharged = []
        for atom_index in self.NonProteinAtoms:
            atom = self.NonProteinAtoms[atom_index]
            if atom.element == "MG" or atom.element == "MN" or atom.element == "RH" or atom.element == "ZN" or atom.element == "FE" or atom.element == "BI" or atom.element == "AS" or atom.element == "AG":
                    chrg = self.charged(atom.coordinates, [atom_index], True)
                    self.charges.append(chrg)

            if atom.element == "N":
                if atom.NumberOfNeighbors() == 4: # a quartinary amine, so it's easy
                    indexes = [atom_index]
                    indexes.extend(atom.IndeciesOfAtomsConnecting)
                    chrg = self.charged(atom.coordinates, indexes, True) # so the indicies stored is just the index of the nitrogen and any attached atoms
                    self.charges.append(chrg)
                elif atom.NumberOfNeighbors() == 3: # maybe you only have two hydrogen's added, by they're sp3 hybridized. Just count this as a quartinary amine, since I think the positive charge would be stabalized.
                    nitrogen = atom
                    atom1 = self.AllAtoms[atom.IndeciesOfAtomsConnecting[0]]
                    atom2 = self.AllAtoms[atom.IndeciesOfAtomsConnecting[1]]
                    atom3 = self.AllAtoms[atom.IndeciesOfAtomsConnecting[2]]
                    angle1 = self.functions.angle_between_three_points(atom1.coordinates, nitrogen.coordinates, atom2.coordinates) * 180.0 / math.pi
                    angle2 = self.functions.angle_between_three_points(atom1.coordinates, nitrogen.coordinates, atom3.coordinates) * 180.0 / math.pi
                    angle3 = self.functions.angle_between_three_points(atom2.coordinates, nitrogen.coordinates, atom3.coordinates) * 180.0 / math.pi
                    average_angle = (angle1 + angle2 + angle3) / 3
                    if math.fabs(average_angle - 109.0) < 5.0:
                        indexes = [atom_index]
                        indexes.extend(atom.IndeciesOfAtomsConnecting)
                        chrg = self.charged(nitrogen.coordinates, indexes, True) # so indexes added are the nitrogen and any attached atoms.
                        self.charges.append(chrg)

            if atom.element == "C": # let's check for guanidino-like groups (actually H2N-C-NH2, where not CN3.)
                if atom.NumberOfNeighbors() == 3: # the carbon has only three atoms connected to it
                    nitrogens = self.connected_atoms_of_given_element(atom_index,"N")
                    if len(nitrogens) >= 2: # so carbon is connected to at least two nitrogens
                        # now we need to count the number of nitrogens that are only connected to one heavy atom (the carbon)
                        nitrogens_to_use = []
                        all_connected = atom.IndeciesOfAtomsConnecting[:]
                        not_isolated = -1

                        for atmindex in nitrogens:
                            if len(self.connected_heavy_atoms(atmindex)) == 1:
                                nitrogens_to_use.append(atmindex)
                                all_connected.remove(atmindex)

                        if len(all_connected) > 0: not_isolated = all_connected[0] # get the atom that connects this charged group to the rest of the molecule, ultimately to make sure it's sp3 hybridized

                        if len(nitrogens_to_use) == 2 and not_isolated != -1: # so there are at two nitrogens that are only connected to the carbon (and probably some hydrogens)

                            # now you need to make sure not_isolated atom is sp3 hybridized
                            not_isolated_atom = self.AllAtoms[not_isolated]
                            if (not_isolated_atom.element == "C" and not_isolated_atom.NumberOfNeighbors()==4) or (not_isolated_atom.element == "O" and not_isolated_atom.NumberOfNeighbors()==2) or not_isolated_atom.element == "N" or not_isolated_atom.element == "S" or not_isolated_atom.element == "P":

                                pt = self.AllAtoms[nitrogens_to_use[0]].coordinates.copy_of()
                                pt.x = pt.x + self.AllAtoms[nitrogens_to_use[1]].coordinates.x
                                pt.y = pt.y + self.AllAtoms[nitrogens_to_use[1]].coordinates.y
                                pt.z = pt.z + self.AllAtoms[nitrogens_to_use[1]].coordinates.z
                                pt.x = pt.x / 2.0
                                pt.y = pt.y / 2.0
                                pt.z = pt.z / 2.0

                                indexes = [atom_index]
                                indexes.extend(nitrogens_to_use)
                                indexes.extend(self.connected_atoms_of_given_element(nitrogens_to_use[0],"H"))
                                indexes.extend(self.connected_atoms_of_given_element(nitrogens_to_use[1],"H"))

                                chrg = self.charged(pt, indexes, True) # True because it's positive
                                self.charges.append(chrg)

            if atom.element == "C": # let's check for a carboxylate
                if atom.NumberOfNeighbors() == 3: # a carboxylate carbon will have three items connected to it.
                    oxygens = self.connected_atoms_of_given_element(atom_index,"O")
                    if len(oxygens) == 2: # a carboxylate will have two oxygens connected to it.
                        # now, each of the oxygens should be connected to only one heavy atom (so if it's connected to a hydrogen, that's okay)
                        if len(self.connected_heavy_atoms(oxygens[0])) == 1 and len(self.connected_heavy_atoms(oxygens[1])) == 1:
                            # so it's a carboxylate! Add a negative charge.
                            pt = self.AllAtoms[oxygens[0]].coordinates.copy_of()
                            pt.x = pt.x + self.AllAtoms[oxygens[1]].coordinates.x
                            pt.y = pt.y + self.AllAtoms[oxygens[1]].coordinates.y
                            pt.z = pt.z + self.AllAtoms[oxygens[1]].coordinates.z
                            pt.x = pt.x / 2.0
                            pt.y = pt.y / 2.0
                            pt.z = pt.z / 2.0
                            chrg = self.charged(pt, [oxygens[0], atom_index, oxygens[1]], False)
                            self.charges.append(chrg)

            if atom.element == "P": # let's check for a phosphate or anything where a phosphorus is bound to two oxygens where both oxygens are bound to only one heavy atom (the phosphorus). I think this will get several phosphorus substances.
                oxygens = self.connected_atoms_of_given_element(atom_index,"O")
                if len(oxygens) >=2: # the phosphorus is bound to at least two oxygens
                    # now count the number of oxygens that are only bound to the phosphorus
                    count = 0
                    for oxygen_index in oxygens:
                        if len(self.connected_heavy_atoms(oxygen_index)) == 1: count = count + 1
                    if count >=2: # so there are at least two oxygens that are only bound to the phosphorus
                        indexes = [atom_index]
                        indexes.extend(oxygens)
                        chrg = self.charged(atom.coordinates, indexes, False)
                        self.charges.append(chrg)

            if atom.element == "S": # let's check for a sulfonate or anything where a sulfur is bound to at least three oxygens and at least three are bound to only the sulfur (or the sulfur and a hydrogen).
                oxygens = self.connected_atoms_of_given_element(atom_index,"O")
                if len(oxygens) >=3: # the sulfur is bound to at least three oxygens
                    # now count the number of oxygens that are only bound to the sulfur
                    count = 0
                    for oxygen_index in oxygens:
                        if len(self.connected_heavy_atoms(oxygen_index)) == 1: count = count + 1
                    if count >=3: # so there are at least three oxygens that are only bound to the sulfur
                        indexes = [atom_index]
                        indexes.extend(oxygens)
                        chrg = self.charged(atom.coordinates, indexes, False)
                        self.charges.append(chrg)

        # Now that you've found all the positive charges in non-protein residues, it's time to look for aromatic rings in protein residues
        curr_res = ""
        first = True
        residue = []

        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]

            key = atom.residue + "_" + str(atom.resid) + "_" + atom.chain

            if first is True:
                curr_res = key
                first = False

            if key != curr_res:

                self.assign_charged_from_protein_process_residue(residue, last_key)

                residue = []
                curr_res = key

            residue.append(atom_index)
            last_key = key

        self.assign_charged_from_protein_process_residue(residue, last_key)

    def assign_charged_from_protein_process_residue(self, residue, last_key):
        temp = last_key.strip().split("_")
        resname = temp[0]
        real_resname = resname[-3:]
        resid = temp[1]
        chain = temp[2]

        if real_resname == "LYS" or real_resname == "LYN": # regardless of protonation state, assume it's charged.
            for index in residue:
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "NZ":

                    # quickly go through the residue and get the hydrogens attached to this nitrogen to include in the index list
                    indexes = [index]
                    for index2 in residue:
                        atom2 = self.AllAtoms[index2]
                        if atom2.atomname.strip() == "HZ1": indexes.append(index2)
                        if atom2.atomname.strip() == "HZ2": indexes.append(index2)
                        if atom2.atomname.strip() == "HZ3": indexes.append(index2)

                    chrg = self.charged(atom.coordinates, indexes, True)
                    self.charges.append(chrg)
                break

        if real_resname == "ARG":
            charge_pt = point(0.0,0.0,0.0)
            count = 0.0
            indices = []
            for index in residue:
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "NH1":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "NH2":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "2HH2": indices.append(index)
                if atom.atomname.strip() == "1HH2": indices.append(index)
                if atom.atomname.strip() == "CZ": indices.append(index)
                if atom.atomname.strip() == "2HH1": indices.append(index)
                if atom.atomname.strip() == "1HH1": indices.append(index)

            if count != 0.0:

                charge_pt.x = charge_pt.x / count
                charge_pt.y = charge_pt.y / count
                charge_pt.z = charge_pt.z / count

                if charge_pt.x != 0.0 or charge_pt.y != 0.0 or charge_pt.z != 0.0:
                    chrg = self.charged(charge_pt, indices, True)
                    self.charges.append(chrg)

        if real_resname == "HIS" or real_resname == "HID" or real_resname == "HIE" or real_resname == "HIP":      # regardless of protonation state, assume it's charged. This based on "The Cation-Pi Interaction," which suggests protonated state would be stabalized. But let's not consider HIS when doing salt bridges.
            charge_pt = point(0.0,0.0,0.0)
            count = 0.0
            indices = []
            for index in residue:
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "NE2":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "ND1":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "HE2": indices.append(index)
                if atom.atomname.strip() == "HD1": indices.append(index)
                if atom.atomname.strip() == "CE1": indices.append(index)
                if atom.atomname.strip() == "CD2": indices.append(index)
                if atom.atomname.strip() == "CG": indices.append(index)

            if count != 0.0:
                charge_pt.x = charge_pt.x / count
                charge_pt.y = charge_pt.y / count
                charge_pt.z = charge_pt.z / count
                if charge_pt.x != 0.0 or charge_pt.y != 0.0 or charge_pt.z != 0.0:
                    chrg = self.charged(charge_pt, indices, True)
                    self.charges.append(chrg)

        if real_resname == "GLU" or real_resname == "GLH" or real_resname == "GLX": # regardless of protonation state, assume it's charged. This based on "The Cation-Pi Interaction," which suggests protonated state would be stabalized.
            charge_pt = point(0.0,0.0,0.0)
            count = 0.0
            indices = []
            for index in residue:
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "OE1":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "OE2":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "CD": indices.append(index)

            if count != 0.0:
                charge_pt.x = charge_pt.x / count
                charge_pt.y = charge_pt.y / count
                charge_pt.z = charge_pt.z / count
                if charge_pt.x != 0.0 or charge_pt.y != 0.0 or charge_pt.z != 0.0:
                    chrg = self.charged(charge_pt, indices, False) # False because it's a negative charge
                    self.charges.append(chrg)

        if real_resname == "ASP" or real_resname == "ASH" or real_resname == "ASX": # regardless of protonation state, assume it's charged. This based on "The Cation-Pi Interaction," which suggests protonated state would be stabalized.
            charge_pt = point(0.0,0.0,0.0)
            count = 0.0
            indices = []
            for index in residue:
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "OD1":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "OD2":
                    charge_pt.x = charge_pt.x + atom.coordinates.x
                    charge_pt.y = charge_pt.y + atom.coordinates.y
                    charge_pt.z = charge_pt.z + atom.coordinates.z
                    indices.append(index)
                    count = count + 1.0
                if atom.atomname.strip() == "CG": indices.append(index)

            if count != 0.0:
                charge_pt.x = charge_pt.x / count
                charge_pt.y = charge_pt.y / count
                charge_pt.z = charge_pt.z / count
                if charge_pt.x != 0.0 or charge_pt.y != 0.0 or charge_pt.z != 0.0:
                    chrg = self.charged(charge_pt, indices, False) # False because it's a negative charge
                    self.charges.append(chrg)

    class charged():
        def __init__(self, coordinates, indices, positive):
            self.coordinates = coordinates
            self.indices = indices
            self.positive = positive # true or false to specifiy if positive or negative charge

    # Functions to identify aromatic rings
    # ====================================

    def add_aromatic_marker(self, indicies_of_ring):
        # first identify the center point
        points_list = []
        total = len(indicies_of_ring)
        x_sum = 0.0
        y_sum = 0.0
        z_sum = 0.0

        for index in indicies_of_ring:
            atom = self.AllAtoms[index]
            points_list.append(atom.coordinates)
            x_sum = x_sum + atom.coordinates.x
            y_sum = y_sum + atom.coordinates.y
            z_sum = z_sum + atom.coordinates.z

        if total == 0: return # to prevent errors in some cases

        center = point(x_sum / total, y_sum / total, z_sum / total)

        # now get the radius of the aromatic ring
        radius = 0.0
        for index in indicies_of_ring:
            atom = self.AllAtoms[index]
            dist = center.dist_to(atom.coordinates)
            if dist > radius: radius = dist

        # now get the plane that defines this ring
        if len(indicies_of_ring) < 3:
            return # to prevent an error in some cases. If there aren't three point, you can't define a plane
        elif len(indicies_of_ring) == 3:
            A = self.AllAtoms[indicies_of_ring[0]].coordinates
            B = self.AllAtoms[indicies_of_ring[1]].coordinates
            C = self.AllAtoms[indicies_of_ring[2]].coordinates
        elif len(indicies_of_ring) == 4:
            A = self.AllAtoms[indicies_of_ring[0]].coordinates
            B = self.AllAtoms[indicies_of_ring[1]].coordinates
            C = self.AllAtoms[indicies_of_ring[3]].coordinates
        else: # best, for 5 and 6 member rings
            A = self.AllAtoms[indicies_of_ring[0]].coordinates
            B = self.AllAtoms[indicies_of_ring[2]].coordinates
            C = self.AllAtoms[indicies_of_ring[4]].coordinates

        AB = self.functions.vector_subtraction(B,A)
        AC = self.functions.vector_subtraction(C,A)
        ABXAC = self.functions.CrossProduct(AB,AC)

        # formula for plane will be ax + by + cz = d
        x1 = self.AllAtoms[indicies_of_ring[0]].coordinates.x
        y1 = self.AllAtoms[indicies_of_ring[0]].coordinates.y
        z1 = self.AllAtoms[indicies_of_ring[0]].coordinates.z

        a = ABXAC.x
        b = ABXAC.y
        c = ABXAC.z
        d = a*x1 + b*y1 + c*z1

        ar_ring = self.aromatic_ring(center, indicies_of_ring, [a,b,c,d], radius)
        self.aromatic_rings.append(ar_ring)

    class aromatic_ring():
        def __init__(self, center, indices, plane_coeff, radius):
            self.center = center
            self.indices = indices
            self.plane_coeff = plane_coeff # a*x + b*y + c*z = dI think that
            self.radius = radius

    def assign_aromatic_rings(self):
        # Get all the rings containing each of the atoms in the ligand
        AllRings = []
        for atom_index in self.NonProteinAtoms:
            AllRings.extend(self.all_rings_containing_atom(atom_index))

        for ring_index_1 in range(len(AllRings)):
            ring1 = AllRings[ring_index_1]
            if len(ring1) != 0:
                for ring_index_2 in range(len(AllRings)):
                    if ring_index_1 != ring_index_2:
                        ring2 = AllRings[ring_index_2]
                        if len(ring2) != 0:
                            if self.set1_is_subset_of_set2(ring1, ring2) is True: AllRings[ring_index_2] = []

        while [] in AllRings: AllRings.remove([])

        # Now we need to figure out which of these ligands are aromatic (planar)

        for ring_index in range(len(AllRings)):
            ring = AllRings[ring_index]
            is_flat = True
            for t in range(-3, len(ring)-3):
                pt1 = self.NonProteinAtoms[ring[t]].coordinates
                pt2 = self.NonProteinAtoms[ring[t+1]].coordinates
                pt3 = self.NonProteinAtoms[ring[t+2]].coordinates
                pt4 = self.NonProteinAtoms[ring[t+3]].coordinates

                # first, let's see if the last atom in this ring is a carbon connected to four atoms. That would be a quick way of telling this is not an aromatic ring
                cur_atom = self.NonProteinAtoms[ring[t+3]]
                if cur_atom.element == "C" and cur_atom.NumberOfNeighbors() == 4:
                    is_flat = False
                    break

                # now check the dihedral between the ring atoms to see if it's flat
                angle = self.functions.dihedral(pt1, pt2, pt3, pt4) * 180 / math.pi
                if (angle > -165 and angle < -15) or (angle > 15 and angle < 165): # 15 degress is the cutoff #, ring[t], ring[t+1], ring[t+2], ring[t+3] # range of this function is -pi to pi
                    is_flat = False
                    break

                # now check the dihedral between the ring atoms and an atom connected to the current atom to see if that's flat too.
                for substituent_atom_index in cur_atom.IndeciesOfAtomsConnecting:
                    pt_sub = self.NonProteinAtoms[substituent_atom_index].coordinates
                    angle = self.functions.dihedral(pt2, pt3, pt4, pt_sub) * 180 / math.pi
                    if (angle > -165 and angle < -15) or (angle > 15 and angle < 165): # 15 degress is the cutoff #, ring[t], ring[t+1], ring[t+2], ring[t+3] # range of this function is -pi to pi
                        is_flat = False
                        break

            if is_flat is False: AllRings[ring_index] = []
            if len(ring) < 5: AllRings[ring_index] = [] # While I'm at it, three and four member rings are not aromatic
            if len(ring) > 6: AllRings[ring_index] = [] # While I'm at it, if the ring has more than 6, also throw it out. So only 5 and 6 member rings are allowed.



        while [] in AllRings: AllRings.remove([])

        for ring in AllRings:
            self.add_aromatic_marker(ring)

        # Now that you've found all the rings in non-protein residues, it's time to look for aromatic rings in protein residues
        curr_res = ""
        first = True
        residue = []

        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]

            key = atom.residue + "_" + str(atom.resid) + "_" + atom.chain

            if first is True:
                curr_res = key
                first = False

            if key != curr_res:

                self.assign_aromatic_rings_from_protein_process_residue(residue, last_key)

                residue = []
                curr_res = key

            residue.append(atom_index)
            last_key = key

        self.assign_aromatic_rings_from_protein_process_residue(residue, last_key)

    def assign_aromatic_rings_from_protein_process_residue(self, residue, last_key):
        temp = last_key.strip().split("_")
        resname = temp[0]
        real_resname = resname[-3:]
        resid = temp[1]
        chain = temp[2]

        if real_resname == "PHE":
            indicies_of_ring = []

            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CG": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CZ": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE2": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD2": indicies_of_ring.append(index)

            self.add_aromatic_marker(indicies_of_ring)

        if real_resname == "TYR":
            indicies_of_ring = []

            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CG": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CZ": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE2": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD2": indicies_of_ring.append(index)

            self.add_aromatic_marker(indicies_of_ring)

        if real_resname == "HIS" or real_resname == "HID" or real_resname == "HIE" or real_resname == "HIP":
            indicies_of_ring = []

            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CG": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "ND1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "NE2": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD2": indicies_of_ring.append(index)

            self.add_aromatic_marker(indicies_of_ring)

        if real_resname == "TRP":
            indicies_of_ring = []

            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CG": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "NE1": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE2": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD2": indicies_of_ring.append(index)

            self.add_aromatic_marker(indicies_of_ring)

            indicies_of_ring = []

            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE2": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CD2": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CE3": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CZ3": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CH2": indicies_of_ring.append(index)
            for index in residue: # written this way because order is important
                atom = self.AllAtoms[index]
                if atom.atomname.strip() == "CZ2": indicies_of_ring.append(index)

            self.add_aromatic_marker(indicies_of_ring)

    def set1_is_subset_of_set2(self, set1, set2):
        is_subset = True
        for item in set1:
            if not item in set2:
                is_subset = False
                break
        return is_subset

    def all_rings_containing_atom(self, index):

        AllRings = []

        atom = self.AllAtoms[index]
        for conneceted_atom in atom.IndeciesOfAtomsConnecting:
            self.ring_recursive(conneceted_atom, [index], index, AllRings)

        return AllRings

    def ring_recursive(self, index, AlreadyCrossed, orig_atom, AllRings):

        if len(AlreadyCrossed) > 6: return # since you're only considering aromatic rings containing 5 or 6 members anyway, save yourself some time.

        atom = self.AllAtoms[index]

        temp = AlreadyCrossed[:]
        temp.append(index)

        for conneceted_atom in atom.IndeciesOfAtomsConnecting:
            if not conneceted_atom in AlreadyCrossed:
                self.ring_recursive(conneceted_atom, temp, orig_atom, AllRings)
            if conneceted_atom == orig_atom and orig_atom != AlreadyCrossed[-1]:
                AllRings.append(temp)

    # Functions to assign secondary structure to protein residues
    # ===========================================================

    def assign_secondary_structure(self):
        # first, we need to know what resid's are available
        resids = []
        last_key = "-99999_Z"
        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]
            key = str(atom.resid) + "_" + atom.chain
            if key != last_key:
                last_key = key
                resids.append(last_key)

        structure = {}
        for resid in resids:
            structure[resid] = "OTHER"

        atoms = []

        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]
            if atom.SideChainOrBackBone() == "BACKBONE":
                if len(atoms) < 8:
                    atoms.append(atom)
                else:
                    atoms.pop(0)
                    atoms.append(atom)

                    # now make sure the first four all have the same resid and the last four all have the same resid
                    if atoms[0].resid == atoms[1].resid and atoms[0].resid == atoms[2].resid and atoms[0].resid == atoms[3].resid and atoms[0] != atoms[4].resid and atoms[4].resid == atoms[5].resid and atoms[4].resid == atoms[6].resid and atoms[4].resid == atoms[7].resid and atoms[0].resid + 1 == atoms[7].resid and atoms[0].chain == atoms[7].chain:
                        resid1 = atoms[0].resid
                        resid2 = atoms[7].resid

                        # Now give easier to use names to the atoms
                        for atom in atoms:
                            if atom.resid == resid1 and atom.atomname.strip() == "N": first_N = atom
                            if atom.resid == resid1 and atom.atomname.strip() == "C": first_C = atom
                            if atom.resid == resid1 and atom.atomname.strip() == "CA": first_CA = atom

                            if atom.resid == resid2 and atom.atomname.strip() == "N": second_N = atom
                            if atom.resid == resid2 and atom.atomname.strip() == "C": second_C = atom
                            if atom.resid == resid2 and atom.atomname.strip() == "CA": second_CA = atom

                        # Now compute the phi and psi dihedral angles
                        phi = self.functions.dihedral(first_C.coordinates, second_N.coordinates, second_CA.coordinates, second_C.coordinates) * 180.0 / math.pi
                        psi = self.functions.dihedral(first_N.coordinates, first_CA.coordinates, first_C.coordinates, second_N.coordinates) * 180.0 / math.pi

                        # Now use those angles to determine if it's alpha or beta
                        if phi > -145 and phi < -35 and psi > -70 and psi < 50:
                            key1 = str(first_C.resid) + "_" + first_C.chain
                            key2 = str(second_C.resid) + "_" + second_C.chain
                            structure[key1] = "ALPHA"
                            structure[key2] = "ALPHA"
                        if (phi >= -180 and phi < -40 and psi <= 180 and psi > 90) or (phi >= -180 and phi < -70 and psi <= -165): # beta. This gets some loops (by my eye), but it's the best I could do.
                            key1 = str(first_C.resid) + "_" + first_C.chain
                            key2 = str(second_C.resid) + "_" + second_C.chain
                            structure[key1] = "BETA"
                            structure[key2] = "BETA"

        # Now update each of the atoms with this structural information
        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]
            key = str(atom.resid) + "_" + atom.chain
            atom.structure = structure[key]

        # Some more post processing.
        CA_list = [] # first build a list of the indices of all the alpha carbons
        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]
            if atom.residue.strip() in self.protein_resnames and atom.atomname.strip() == "CA": CA_list.append(atom_index)

        # some more post processing.
        change = True
        while change is True:
            change = False

            # A residue of index i is only going to be in an alpha helix its CA is within 6 A of the CA of the residue i + 3
            for CA_atom_index in CA_list:
                CA_atom = self.AllAtoms[CA_atom_index]
                if CA_atom.structure == "ALPHA": # so it's in an alpha helix
                    another_alpha_is_close = False
                    for other_CA_atom_index in CA_list: # so now compare that CA to all the other CA's
                        other_CA_atom = self.AllAtoms[other_CA_atom_index]
                        if other_CA_atom.structure == "ALPHA": # so it's also in an alpha helix
                            if other_CA_atom.resid - 3 == CA_atom.resid or other_CA_atom.resid + 3 == CA_atom.resid: # so this CA atom is one of the ones the first atom might hydrogen bond with
                                if other_CA_atom.coordinates.dist_to(CA_atom.coordinates) < 6.0: # so these two CA atoms are close enough together that their residues are probably hydrogen bonded
                                    another_alpha_is_close = True
                                    break
                    if another_alpha_is_close is False:
                        self.set_structure_of_residue(CA_atom.chain, CA_atom.resid, "OTHER")
                        change = True

            # Alpha helices are only alpha helices if they span at least 4 residues (to wrap around and hydrogen bond). I'm going to require them to span at least 5 residues, based on examination of many structures.
            for index_in_list in range(len(CA_list)-5):

                index_in_pdb1 = CA_list[index_in_list]
                index_in_pdb2 = CA_list[index_in_list+1]
                index_in_pdb3 = CA_list[index_in_list+2]
                index_in_pdb4 = CA_list[index_in_list+3]
                index_in_pdb5 = CA_list[index_in_list+4]
                index_in_pdb6 = CA_list[index_in_list+5]

                atom1 = self.AllAtoms[index_in_pdb1]
                atom2 = self.AllAtoms[index_in_pdb2]
                atom3 = self.AllAtoms[index_in_pdb3]
                atom4 = self.AllAtoms[index_in_pdb4]
                atom5 = self.AllAtoms[index_in_pdb5]
                atom6 = self.AllAtoms[index_in_pdb6]

                if atom1.resid + 1 == atom2.resid and atom2.resid + 1 == atom3.resid and atom3.resid + 1 == atom4.resid and atom4.resid + 1 == atom5.resid and atom5.resid + 1 == atom6.resid: # so they are sequential

                    if atom1.structure != "ALPHA" and atom2.structure == "ALPHA" and atom3.structure != "ALPHA":
                        self.set_structure_of_residue(atom2.chain, atom2.resid, "OTHER")
                        change = True
                    if atom2.structure != "ALPHA" and atom3.structure == "ALPHA" and atom4.structure != "ALPHA":
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        change = True
                    if atom3.structure != "ALPHA" and atom4.structure == "ALPHA" and atom5.structure != "ALPHA":
                        self.set_structure_of_residue(atom4.chain, atom4.resid, "OTHER")
                        change = True
                    if atom4.structure != "ALPHA" and atom5.structure == "ALPHA" and atom6.structure != "ALPHA":
                        self.set_structure_of_residue(atom5.chain, atom5.resid, "OTHER")
                        change = True

                    if atom1.structure != "ALPHA" and atom2.structure == "ALPHA" and atom3.structure == "ALPHA" and atom4.structure != "ALPHA":
                        self.set_structure_of_residue(atom2.chain, atom2.resid, "OTHER")
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        change = True
                    if atom2.structure != "ALPHA" and atom3.structure == "ALPHA" and atom4.structure == "ALPHA" and atom5.structure != "ALPHA":
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        self.set_structure_of_residue(atom4.chain, atom4.resid, "OTHER")
                        change = True
                    if atom3.structure != "ALPHA" and atom4.structure == "ALPHA" and atom5.structure == "ALPHA" and atom6.structure != "ALPHA":
                        self.set_structure_of_residue(atom4.chain, atom4.resid, "OTHER")
                        self.set_structure_of_residue(atom5.chain, atom5.resid, "OTHER")
                        change = True

                    if atom1.structure != "ALPHA" and atom2.structure == "ALPHA" and atom3.structure == "ALPHA" and atom4.structure == "ALPHA" and atom5.structure != "ALPHA":
                        self.set_structure_of_residue(atom2.chain, atom2.resid, "OTHER")
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        self.set_structure_of_residue(atom4.chain, atom4.resid, "OTHER")
                        change = True
                    if atom2.structure != "ALPHA" and atom3.structure == "ALPHA" and atom4.structure == "ALPHA" and atom5.structure == "ALPHA" and atom6.structure != "ALPHA":
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        self.set_structure_of_residue(atom4.chain, atom4.resid, "OTHER")
                        self.set_structure_of_residue(atom5.chain, atom5.resid, "OTHER")
                        change = True

                    if atom1.structure != "ALPHA" and atom2.structure == "ALPHA" and atom3.structure == "ALPHA" and atom4.structure == "ALPHA" and atom5.structure == "ALPHA" and atom6.structure != "ALPHA":
                        self.set_structure_of_residue(atom2.chain, atom2.resid, "OTHER")
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        self.set_structure_of_residue(atom4.chain, atom4.resid, "OTHER")
                        self.set_structure_of_residue(atom5.chain, atom5.resid, "OTHER")
                        change = True

            # now go through each of the BETA CA atoms. A residue is only going to be called a beta sheet if CA atom is within 6.0 A of another CA beta, same chain, but index difference > 2.
            for CA_atom_index in CA_list:
                CA_atom = self.AllAtoms[CA_atom_index]
                if CA_atom.structure == "BETA": # so it's in a beta sheet
                    another_beta_is_close = False
                    for other_CA_atom_index in CA_list:
                        if other_CA_atom_index != CA_atom_index: # so not comparing an atom to itself
                            other_CA_atom = self.AllAtoms[other_CA_atom_index]
                            if other_CA_atom.structure == "BETA": # so you're comparing it only to other BETA-sheet atoms
                                if other_CA_atom.chain == CA_atom.chain: # so require them to be on the same chain. needed to indecies can be fairly compared
                                    if math.fabs(other_CA_atom.resid - CA_atom.resid) > 2: # so the two residues are not simply adjacent to each other on the chain
                                        if CA_atom.coordinates.dist_to(other_CA_atom.coordinates) < 6.0: # so these to atoms are close to each other
                                            another_beta_is_close = True
                                            break
                    if another_beta_is_close is False:
                        self.set_structure_of_residue(CA_atom.chain, CA_atom.resid, "OTHER")
                        change = True

            # Now some more post-processing needs to be done. Do this again to clear up mess that may have just been created (single residue beta strand, for example)
            # Beta sheets are usually at least 3 residues long

            for index_in_list in range(len(CA_list)-3):

                index_in_pdb1 = CA_list[index_in_list]
                index_in_pdb2 = CA_list[index_in_list+1]
                index_in_pdb3 = CA_list[index_in_list+2]
                index_in_pdb4 = CA_list[index_in_list+3]

                atom1 = self.AllAtoms[index_in_pdb1]
                atom2 = self.AllAtoms[index_in_pdb2]
                atom3 = self.AllAtoms[index_in_pdb3]
                atom4 = self.AllAtoms[index_in_pdb4]

                if atom1.resid + 1 == atom2.resid and atom2.resid + 1 == atom3.resid and atom3.resid + 1 == atom4.resid: # so they are sequential

                    if atom1.structure != "BETA" and atom2.structure == "BETA" and atom3.structure != "BETA":
                        self.set_structure_of_residue(atom2.chain, atom2.resid, "OTHER")
                        change = True
                    if atom2.structure != "BETA" and atom3.structure == "BETA" and atom4.structure != "BETA":
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        change = True
                    if atom1.structure != "BETA" and atom2.structure == "BETA" and atom3.structure == "BETA" and atom4.structure != "BETA":
                        self.set_structure_of_residue(atom2.chain, atom2.resid, "OTHER")
                        self.set_structure_of_residue(atom3.chain, atom3.resid, "OTHER")
                        change = True

    def set_structure_of_residue(self, chain, resid, structure):
        for atom_index in self.AllAtoms:
            atom = self.AllAtoms[atom_index]
            if atom.chain == chain and atom.resid == resid:
                atom.structure = structure

class MathFunctions:

    def vector_subtraction(self, vector1, vector2): # vector1 - vector2
        return point(vector1.x - vector2.x, vector1.y - vector2.y, vector1.z - vector2.z)

    def CrossProduct(self, Pt1, Pt2): # never tested
        Response = point(0,0,0)

        Response.x = Pt1.y * Pt2.z - Pt1.z * Pt2.y
        Response.y = Pt1.z * Pt2.x - Pt1.x * Pt2.z
        Response.z = Pt1.x * Pt2.y - Pt1.y * Pt2.x

        return Response;

    def vector_scalar_multiply(self, vector, scalar):
        return point(vector.x * scalar, vector.y * scalar, vector.z * scalar)

    def dot_product(self, point1, point2):
        return point1.x * point2.x + point1.y * point2.y + point1.z * point2.z

    def dihedral(self, point1, point2, point3, point4): # never tested

        b1 = self.vector_subtraction(point2, point1)
        b2 = self.vector_subtraction(point3, point2)
        b3 = self.vector_subtraction(point4, point3)

        b2Xb3 = self.CrossProduct(b2,b3)
        b1Xb2 = self.CrossProduct(b1,b2)

        b1XMagb2 = self.vector_scalar_multiply(b1,b2.Magnitude())
        radians = math.atan2(self.dot_product(b1XMagb2,b2Xb3), self.dot_product(b1Xb2,b2Xb3))
        return radians

    def angle_between_three_points(self, point1, point2, point3): # As in three connected atoms
        vector1 = self.vector_subtraction(point1, point2)
        vector2 = self.vector_subtraction(point3, point2)
        return self.angle_between_points(vector1, vector2)

    def angle_between_points(self, point1, point2):
        new_point1 = self.return_normalized_vector(point1)
        new_point2 = self.return_normalized_vector(point2)
        dot_prod = self.dot_product(new_point1, new_point2)
        if dot_prod > 1.0: dot_prod = 1.0 # to prevent errors that can rarely occur
        if dot_prod < -1.0: dot_prod = -1.0
        return math.acos(dot_prod)

    def return_normalized_vector(self, vector):
        dist = self.distance(point(0,0,0), vector)
        return point(vector.x/dist, vector.y/dist, vector.z/dist)

    def distance(self, point1, point2):
        deltax = point1.x - point2.x
        deltay = point1.y - point2.y
        deltaz = point1.z - point2.z

        return math.sqrt(math.pow(deltax,2) + math.pow(deltay,2) + math.pow(deltaz,2))

    def project_point_onto_plane(self, apoint, plane_coefficients): # essentially finds the point on the plane that is closest to the specified point
        # the plane_coefficients are [a,b,c,d], where the plane is ax + by + cz = d

        # First, define a plane using cooeficients a, b, c, d such that ax + by + cz = d
        a = plane_coefficients[0]
        b = plane_coefficients[1]
        c = plane_coefficients[2]
        d = plane_coefficients[3]

        # Now, define a point in space (s,u,v)
        s = apoint.x
        u = apoint.y
        v = apoint.z

        # the formula of a line perpendicular to the plan passing through (s,u,v) is:
        #x = s + at
        #y = u + bt
        #z = v + ct

        t = (d - a*s - b*u - c*v) / (a*a + b*b + c*c)

        # here's the point closest on the plane
        x = s + a*t
        y = u + b*t
        z = v + c*t

        return point(x,y,z)

def getCommandOutput2(command):
    child = os.popen(command)
    data = child.read()
    err = child.close()
    if err:
        raise RuntimeError('%s failed w/ exit code %d' % (command, err))
    return data

class binana:

    functions = MathFunctions()

    # supporting functions
    def list_alphebetize_and_combine(self, list):
        list.sort()
        return '_'.join(list)

    def hashtable_entry_add_one(self, hashtable, key, toadd = 1): # note that dictionaries (hashtables) are passed by reference in python
        if key in hashtable:
            hashtable[key] = hashtable[key] + toadd
        else:
            hashtable[key] = toadd

    def extend_list_by_dictionary(self, input_list, dictionary):
        # first, sort the dictionary by the key


        keys = dictionary.keys()

        keys = list(keys)
        keys.sort()

        # now make a list of the values
        vals = []
        for key in keys:
            vals.append(dictionary[key])

        # now append vals to the list
        newlist = input_list[:]
        newlist.extend(vals)

        # return the extended list
        return newlist

    def center(self, string, length):
        while len(string) < length:
            string = " " + string
            if len(string) < length:
                string = string + " "
        return string

    # The meat of the class
    #def __init__(self, ligand_pdbqt_filename, receptor_pdbqt_filename, parameters, line_header, actual_filename_if_ligand_is_list="", actual_filename_if_receptor_is_list=""): # must be a more elegant way of doing this
    def __init__(self, ligand_pdbqt_filename, receptor, parameters, line_header, actual_filename_if_ligand_is_list="", actual_filename_if_receptor_is_list=""): # must be a more elegant way of doing this

        receptor_pdbqt_filename = receptor.OrigFileName

        ligand = PDB()
        if actual_filename_if_ligand_is_list!="": # so a list was passed as the ligand
            ligand.LoadPDB_from_list(ligand_pdbqt_filename, line_header)

            # now write the file so when VINA is run it has a ligand file for input
            f = open(actual_filename_if_ligand_is_list,'w')
            for line in ligand_pdbqt_filename:
                if not "MODEL" in line and not "ENDMDL" in line: f.write(line)
            f.close()

            ligand_pdbqt_filename = actual_filename_if_ligand_is_list
        else: # so a filename was passed as the ligand
            ligand.LoadPDB_from_file(ligand_pdbqt_filename, line_header)


        #receptor = PDB()
    #if actual_filename_if_ligand_is_list=="": # receptor is a filename instead of a list
        # receptor.LoadPDB_from_file(receptor_pdbqt_filename, line_header)
    #else: # so it's a list that was passed, not a filename
        # receptor.LoadPDB_from_list(receptor_pdbqt_filename, line_header)
        # receptor_pdbqt_filename = actual_filename_if_receptor_is_list

        receptor.assign_secondary_structure()

        # Get distance measurements between protein and ligand atom types, as well as some other measurements

        ligand_receptor_atom_type_pairs_less_than_two_half = {}
        ligand_receptor_atom_type_pairs_less_than_four = {}
        ligand_receptor_atom_type_pairs_electrostatic = {}
        active_site_flexibility = {}
        hbonds = {}
        hydrophobics = {}
        ligand.rotateable_bonds_count
        functions = MathFunctions()

        pdb_close_contacts = PDB()
        pdb_contacts = PDB()
        pdb_contacts_alpha_helix = PDB()
        pdb_contacts_beta_sheet = PDB()
        pdb_contacts_other_2nd_structure = PDB()
        pdb_side_chain = PDB()
        pdb_back_bone = PDB()
        pdb_hydrophobic = PDB()
        pdb_hbonds = PDB()

        for ligand_atom_index in ligand.AllAtoms:
            for receptor_atom_index in receptor.AllAtoms:
                ligand_atom = ligand.AllAtoms[ligand_atom_index]
                receptor_atom = receptor.AllAtoms[receptor_atom_index]

                dist = ligand_atom.coordinates.dist_to(receptor_atom.coordinates)
                if dist < 2.5: # less than 2.5 A
                    list = [ligand_atom.atomtype, receptor_atom.atomtype]
                    self.hashtable_entry_add_one(ligand_receptor_atom_type_pairs_less_than_two_half, self.list_alphebetize_and_combine(list))
                    pdb_close_contacts.AddNewAtom(ligand_atom.copy_of())
                    pdb_close_contacts.AddNewAtom(receptor_atom.copy_of())
                elif dist < 4.0: # less than 4 A
                    list = [ligand_atom.atomtype, receptor_atom.atomtype]
                    self.hashtable_entry_add_one(ligand_receptor_atom_type_pairs_less_than_four, self.list_alphebetize_and_combine(list))
                    pdb_contacts.AddNewAtom(ligand_atom.copy_of())
                    pdb_contacts.AddNewAtom(receptor_atom.copy_of())

                if dist < 4.0:
                    # calculate electrostatic energies for all less than 4 A
                    ligand_charge = ligand_atom.charge
                    receptor_charge = receptor_atom.charge
                    coulomb_energy = (ligand_charge * receptor_charge / dist) * 138.94238460104697e4 # to convert into J/mol # might be nice to double check this
                    list = [ligand_atom.atomtype, receptor_atom.atomtype]
                    self.hashtable_entry_add_one(ligand_receptor_atom_type_pairs_electrostatic, self.list_alphebetize_and_combine(list), coulomb_energy)

                if dist < 4.0:
                    # Now get statistics to judge active-site flexibility
                    flexibility_key = receptor_atom.SideChainOrBackBone() + "_" + receptor_atom.structure # first can be sidechain or backbone, second back be alpha, beta, or other, so six catagories
                    if receptor_atom.structure == "ALPHA": pdb_contacts_alpha_helix.AddNewAtom(receptor_atom.copy_of())
                    elif receptor_atom.structure == "BETA": pdb_contacts_beta_sheet.AddNewAtom(receptor_atom.copy_of())
                    elif receptor_atom.structure == "OTHER": pdb_contacts_other_2nd_structure.AddNewAtom(receptor_atom.copy_of())

                    if receptor_atom.SideChainOrBackBone() == "BACKBONE": pdb_back_bone.AddNewAtom(receptor_atom.copy_of())
                    elif receptor_atom.SideChainOrBackBone() == "SIDECHAIN": pdb_side_chain.AddNewAtom(receptor_atom.copy_of())

                    self.hashtable_entry_add_one(active_site_flexibility, flexibility_key)

                if dist < 4.0:
                    # Now see if there's hydrophobic contacts (C-C contacts)
                    if ligand_atom.element == "C" and receptor_atom.element == "C":
                        hydrophobic_key = receptor_atom.SideChainOrBackBone() + "_" + receptor_atom.structure
                        pdb_hydrophobic.AddNewAtom(ligand_atom.copy_of())
                        pdb_hydrophobic.AddNewAtom(receptor_atom.copy_of())

                        self.hashtable_entry_add_one(hydrophobics, hydrophobic_key)

                if dist < 4.0:
                    # Now see if there's some sort of hydrogen bond between these two atoms. distance cutoff = 4, angle cutoff = 40. Note that this is liberal.
                    if (ligand_atom.element == "O" or ligand_atom.element == "N") and (receptor_atom.element == "O" or receptor_atom.element == "N"):

                        # now build a list of all the hydrogens close to these atoms
                        hydrogens = []

                        for atm_index in ligand.AllAtoms:
                            if ligand.AllAtoms[atm_index].element == "H": # so it's a hydrogen
                                if ligand.AllAtoms[atm_index].coordinates.dist_to(ligand_atom.coordinates) < 1.3: # O-H distance is 0.96 A, N-H is 1.01 A. See http://www.science.uwaterloo.ca/~cchieh/cact/c120/bondel.html
                                    ligand.AllAtoms[atm_index].comment = "LIGAND"
                                    hydrogens.append(ligand.AllAtoms[atm_index])

                        for atm_index in receptor.AllAtoms:
                            if receptor.AllAtoms[atm_index].element == "H": # so it's a hydrogen
                                if receptor.AllAtoms[atm_index].coordinates.dist_to(receptor_atom.coordinates) < 1.3: # O-H distance is 0.96 A, N-H is 1.01 A. See http://www.science.uwaterloo.ca/~cchieh/cact/c120/bondel.html
                                    receptor.AllAtoms[atm_index].comment = "RECEPTOR"
                                    hydrogens.append(receptor.AllAtoms[atm_index])

                        # now we need to check the angles
                        for hydrogen in hydrogens:
                            if math.fabs(180 - functions.angle_between_three_points(ligand_atom.coordinates, hydrogen.coordinates, receptor_atom.coordinates) * 180.0 / math.pi) <= 40.0:
                                hbonds_key = "HDONOR_" + hydrogen.comment + "_" + receptor_atom.SideChainOrBackBone() + "_" + receptor_atom.structure
                                pdb_hbonds.AddNewAtom(ligand_atom.copy_of())
                                pdb_hbonds.AddNewAtom(hydrogen.copy_of())
                                pdb_hbonds.AddNewAtom(receptor_atom.copy_of())
                                self.hashtable_entry_add_one(hbonds, hbonds_key)

        # Get the total number of each atom type in the ligand
        ligand_atom_types = {}
        for ligand_atom_index in ligand.AllAtoms:
            ligand_atom = ligand.AllAtoms[ligand_atom_index]
            self.hashtable_entry_add_one(ligand_atom_types, ligand_atom.atomtype)

        pi_padding = 0.75 # This is perhaps controversial. I noticed that often a pi-cation interaction or other pi interaction was only slightly off, but looking at the structure, it was clearly supposed to be a
        # pi-cation interaction. I've decided then to artificially expand the radius of each pi ring. Think of this as adding in a VDW radius, or accounting for poor crystal-structure resolution, or whatever you want
        # to justify it.

        # Count pi-pi stacking and pi-T stacking interactions
        PI_interactions = {}
        pdb_pistack = PDB()
        pdb_pi_T = PDB()
        # "PI-Stacking Interactions ALIVE AND WELL IN PROTEINS" says distance of 7.5 A is good cutoff. This seems really big to me, except that pi-pi interactions (parallel) are actuall usually off centered. Interesting paper.
        # Note that adenine and tryptophan count as two aromatic rings. So, for example, an interaction between these two, if positioned correctly, could count for 4 pi-pi interactions.

        for aromatic1 in ligand.aromatic_rings:
            for aromatic2 in receptor.aromatic_rings:
                dist = aromatic1.center.dist_to(aromatic2.center)
                if dist < 7.5: # so there could be some pi-pi interactions.
                    # first, let's check for stacking interactions. Are the two pi's roughly parallel?
                    aromatic1_norm_vector = point(aromatic1.plane_coeff[0], aromatic1.plane_coeff[1], aromatic1.plane_coeff[2])
                    aromatic2_norm_vector = point(aromatic2.plane_coeff[0], aromatic2.plane_coeff[1], aromatic2.plane_coeff[2])
                    angle_between_planes = self.functions.angle_between_points(aromatic1_norm_vector, aromatic2_norm_vector) * 180.0/math.pi

                    if math.fabs(angle_between_planes-0) < 30.0 or math.fabs(angle_between_planes-180) < 30.0: # so they're more or less parallel, it's probably pi-pi stackingoutput_dir
                        # now, pi-pi are not usually right on top of each other. They're often staggared. So I don't want to just look at the centers of the rings and compare. Let's look at each of the atoms.
                        # do atom of the atoms of one ring, when projected onto the plane of the other, fall within that other ring?

                        pi_pi = False # start by assuming it's not a pi-pi stacking interaction
                        for ligand_ring_index in aromatic1.indices:
                            # project the ligand atom onto the plane of the receptor ring
                            pt_on_receptor_plane = self.functions.project_point_onto_plane(ligand.AllAtoms[ligand_ring_index].coordinates, aromatic2.plane_coeff)
                            if pt_on_receptor_plane.dist_to(aromatic2.center) <= aromatic2.radius + pi_padding:
                                pi_pi = True
                                break

                        if pi_pi is False: # if you've already determined it's a pi-pi stacking interaction, no need to keep trying
                            for receptor_ring_index in aromatic2.indices:
                                # project the ligand atom onto the plane of the receptor ring
                                pt_on_ligand_plane = self.functions.project_point_onto_plane(receptor.AllAtoms[receptor_ring_index].coordinates, aromatic1.plane_coeff)
                                if pt_on_ligand_plane.dist_to(aromatic1.center) <= aromatic1.radius + pi_padding:
                                    pi_pi = True
                                    break

                        if pi_pi is True:
                            structure = receptor.AllAtoms[aromatic2.indices[0]].structure
                            if structure == "": structure = "OTHER" # since it could be interacting with a cofactor or something
                            key = "STACKING_" + structure

                            for index in aromatic1.indices: pdb_pistack.AddNewAtom(ligand.AllAtoms[index].copy_of())
                            for index in aromatic2.indices: pdb_pistack.AddNewAtom(receptor.AllAtoms[index].copy_of())

                            self.hashtable_entry_add_one(PI_interactions, key)

                    elif math.fabs(angle_between_planes-90) < 30.0 or math.fabs(angle_between_planes-270) < 30.0: # so they're more or less perpendicular, it's probably a pi-edge interaction
                        # having looked at many structures, I noticed the algorithm was identifying T-pi reactions when the two rings were in fact quite distant, often with other atoms
                        # in between. Eye-balling it, requiring that at their closest they be at least 5 A apart seems to separate the good T's from the bad
                        min_dist = 100.0
                        for ligand_ind in aromatic1.indices:
                            ligand_at = ligand.AllAtoms[ligand_ind]
                            for receptor_ind in aromatic2.indices:
                                receptor_at = receptor.AllAtoms[receptor_ind]
                                dist = ligand_at.coordinates.dist_to(receptor_at.coordinates)
                                if dist < min_dist: min_dist = dist

                        if min_dist <= 5.0: # so at their closest points, the two rings come within 5 A of each other.

                            # okay, is the ligand pi pointing into the receptor pi, or the other way around?
                            # first, project the center of the ligand pi onto the plane of the receptor pi, and vs. versa

                            # This could be directional somehow, like a hydrogen bond.

                            pt_on_receptor_plane = self.functions.project_point_onto_plane(aromatic1.center, aromatic2.plane_coeff)
                            pt_on_lignad_plane = self.functions.project_point_onto_plane(aromatic2.center, aromatic1.plane_coeff)

                            # now, if it's a true pi-T interaction, this projected point should fall within the ring whose plane it's been projected into.
                            if (pt_on_receptor_plane.dist_to(aromatic2.center) <= aromatic2.radius + pi_padding) or (pt_on_lignad_plane.dist_to(aromatic1.center) <= aromatic1.radius + pi_padding): # so it is in the ring on the projected plane.
                                structure = receptor.AllAtoms[aromatic2.indices[0]].structure
                                if structure == "": structure = "OTHER" # since it could be interacting with a cofactor or something
                                key = "T-SHAPED_" + structure

                                for index in aromatic1.indices: pdb_pi_T.AddNewAtom(ligand.AllAtoms[index].copy_of())
                                for index in aromatic2.indices: pdb_pi_T.AddNewAtom(receptor.AllAtoms[index].copy_of())

                                self.hashtable_entry_add_one(PI_interactions, key)

        # Now identify pi-cation interactions
        pdb_pi_cat = PDB()

        for aromatic in receptor.aromatic_rings:
            for charged in ligand.charges:
                if charged.positive is True: # so only consider positive charges
                    if charged.coordinates.dist_to(aromatic.center) < 6.0: # distance cutoff based on "Cation-pi interactions in structural biology."
                        # project the charged onto the plane of the aromatic
                        charge_projected = self.functions.project_point_onto_plane(charged.coordinates,aromatic.plane_coeff)
                        if charge_projected.dist_to(aromatic.center) < aromatic.radius + pi_padding:
                            structure = receptor.AllAtoms[aromatic.indices[0]].structure
                            if structure == "": structure = "OTHER" # since it could be interacting with a cofactor or something
                            key = "PI-CATION_LIGAND-CHARGED_" + structure

                            for index in aromatic.indices: pdb_pi_cat.AddNewAtom(receptor.AllAtoms[index].copy_of())
                            for index in charged.indices: pdb_pi_cat.AddNewAtom(ligand.AllAtoms[index].copy_of())

                            self.hashtable_entry_add_one(PI_interactions, key)

        for aromatic in ligand.aromatic_rings: # now it's the ligand that has the aromatic group
            for charged in receptor.charges:
                if charged.positive is True: # so only consider positive charges
                    if charged.coordinates.dist_to(aromatic.center) < 6.0: # distance cutoff based on "Cation-pi interactions in structural biology."
                        # project the charged onto the plane of the aromatic
                        charge_projected = self.functions.project_point_onto_plane(charged.coordinates,aromatic.plane_coeff)
                        if charge_projected.dist_to(aromatic.center) < aromatic.radius + pi_padding:
                            structure = receptor.AllAtoms[charged.indices[0]].structure
                            if structure == "": structure = "OTHER" # since it could be interacting with a cofactor or something
                            key = "PI-CATION_RECEPTOR-CHARGED_" + structure

                            for index in aromatic.indices: pdb_pi_cat.AddNewAtom(ligand.AllAtoms[index].copy_of())
                            for index in charged.indices: pdb_pi_cat.AddNewAtom(receptor.AllAtoms[index].copy_of())

                            self.hashtable_entry_add_one(PI_interactions, key)

        # now count the number of salt bridges
        pdb_salt_bridges = PDB()
        salt_bridges = {}
        for receptor_charge in receptor.charges:
            for ligand_charge in ligand.charges:
                if ligand_charge.positive != receptor_charge.positive: # so they have oppositve charges
                    if ligand_charge.coordinates.dist_to(receptor_charge.coordinates) < 5.5: # 4  is good cutoff for salt bridges according to "Close-Range Electrostatic Interactions in Proteins", but looking at complexes, I decided to go with 5.5 A
                        structure = receptor.AllAtoms[receptor_charge.indices[0]].structure
                        if structure == "": structure = "OTHER" # since it could be interacting with a cofactor or something
                        key = "SALT-BRIDGE_" + structure

                        for index in receptor_charge.indices: pdb_salt_bridges.AddNewAtom(receptor.AllAtoms[index].copy_of())
                        for index in ligand_charge.indices: pdb_salt_bridges.AddNewAtom(ligand.AllAtoms[index].copy_of())

                        self.hashtable_entry_add_one(salt_bridges, key)

        # Now save the files
        preface ="REMARK "

        # if an output directory is specified, and it doesn't exist, create it
        #if parameters.params['output_dir'] != "":
        # if not os.path.exists(parameters.params['output_dir']):
        #  os.mkdir(parameters.params['output_dir'])

        # Now get vina
        vina_output = getCommandOutput2('"' + parameters.params['vina_executable'] + '"' + ' --score_only --receptor ' + '"' + receptor_pdbqt_filename + '"' + ' --ligand ' + '"' + ligand_pdbqt_filename + '"')

        print(vina_output)
        vina_output = vina_output.split("\n")
        vina_affinity = 0.0
        vina_gauss_1 = 0.0
        vina_gauss_2 = 0.0
        vina_repulsion = 0.0
        vina_hydrophobic = 0.0
        vina_hydrogen = 0.0
        for item in vina_output:
            item = item.strip()
            if "Affinity" in item: vina_affinity = float(item.replace("Affinity: ","").replace(" (kcal/mol)",""))
            if "gauss 1" in item: vina_gauss_1 = float(item.replace("gauss 1     : ",""))
            if "gauss 2" in item: vina_gauss_2 = float(item.replace("gauss 2     : ",""))
            if "repulsion" in item: vina_repulsion = float(item.replace("repulsion   : ",""))
            if "hydrophobic" in item: vina_hydrophobic = float(item.replace("hydrophobic : ",""))
            if "Hydrogen" in item: vina_hydrogen = float(item.replace("Hydrogen    : ",""))

        vina_output = [vina_affinity, vina_gauss_1, vina_gauss_2, vina_repulsion, vina_hydrophobic, vina_hydrogen]

        stacking = []
        t_shaped = []
        pi_cation = []
        for key in PI_interactions:
            value = PI_interactions[key]
            together = key + "_" + str(value) # not that this is put together strangely!!!
            if "STACKING" in together: stacking.append(together)
            if "CATION" in together: pi_cation.append(together)
            if "SHAPED" in together: t_shaped.append(together)

        # now create a single descriptor object
        data = {}
        data['vina_output'] = vina_output
        data['ligand_receptor_atom_type_pairs_less_than_two_half'] = ligand_receptor_atom_type_pairs_less_than_two_half
        data['ligand_receptor_atom_type_pairs_less_than_four'] = ligand_receptor_atom_type_pairs_less_than_four
        data['ligand_atom_types'] = ligand_atom_types
        data['ligand_receptor_atom_type_pairs_electrostatic'] = ligand_receptor_atom_type_pairs_electrostatic
        data['rotateable_bonds_count'] = ligand.rotateable_bonds_count
        data['active_site_flexibility'] = active_site_flexibility
        data['hbonds'] = hbonds
        data['hydrophobics'] = hydrophobics
        data['stacking'] = stacking
        data['pi_cation'] = pi_cation
        data['t_shaped'] = t_shaped
        data['salt_bridges'] = salt_bridges

        self.vina_output = data['vina_output']

        self.rotateable_bonds_count = {'rot_bonds':data['rotateable_bonds_count']}

        self.ligand_receptor_atom_type_pairs_less_than_two_half = {"A_A": 0, "A_C": 0, "A_CL": 0, "A_F": 0, "A_FE": 0, "A_MG": 0, "A_MN": 0, "A_NA": 0, "A_SA": 0, "BR_C": 0, "BR_OA": 0, "C_CL": 0, "CD_OA": 0, "CL_FE": 0, "CL_MG": 0, "CL_N": 0, "CL_OA": 0, "CL_ZN": 0, "C_MN": 0, "C_NA": 0, "F_N": 0, "F_SA": 0, "F_ZN": 0, "HD_MN": 0, "MN_N": 0, "NA_SA": 0, "N_SA": 0, "A_HD": 0, "A_N": 0, "A_OA": 0, "A_ZN": 0, "BR_HD": 0, "C_C": 0, "C_F": 0, "C_HD": 0, "CL_HD": 0, "C_MG": 0, "C_N": 0, "C_OA": 0, "C_SA": 0, "C_ZN": 0, "FE_HD": 0, "FE_N": 0, "FE_OA": 0, "F_HD": 0, "F_OA": 0, "HD_HD": 0, "HD_I": 0, "HD_MG": 0, "HD_N": 0, "HD_NA": 0, "HD_OA": 0, "HD_P": 0, "HD_S": 0, "HD_SA": 0, "HD_ZN": 0, "MG_NA": 0, "MG_OA": 0, "MN_OA": 0, "NA_OA": 0, "NA_ZN": 0, "N_N": 0, "N_NA": 0, "N_OA": 0, "N_ZN": 0, "OA_OA": 0, "OA_SA": 0, "OA_ZN": 0, "SA_ZN": 0, "S_ZN": 0}
        for key in data['ligand_receptor_atom_type_pairs_less_than_two_half']:
            if not key in self.ligand_receptor_atom_type_pairs_less_than_two_half:
                  print(("\tWARNING: Atoms of types " + key.replace("_"," and ") + " come within 2.5 angstroms of each other."))
                  print("\t  The neural networks were not trained to deal with this juxtaposition,")
                  print("\t  so it will be ignored.")
                  self.error = True
            else:
                  self.ligand_receptor_atom_type_pairs_less_than_two_half[key] = data['ligand_receptor_atom_type_pairs_less_than_two_half'][key]

        self.ligand_receptor_atom_type_pairs_less_than_four = {"A_CU": 0, "A_MG": 0, "A_MN": 0, "BR_SA": 0, "C_CD": 0, "CL_FE": 0, "CL_MG": 0, "CL_MN": 0, "CL_NA": 0, "CL_P": 0, "CL_S": 0, "CL_ZN": 0, "CU_HD": 0, "CU_N": 0, "FE_NA": 0, "FE_SA": 0, "MG_N": 0, "MG_S": 0, "MG_SA": 0, "MN_NA": 0, "MN_S": 0, "MN_SA": 0, "NA_P": 0, "P_S": 0, "P_SA": 0, "S_SA": 0, "A_A": 0, "A_BR": 0, "A_C": 0, "A_CL": 0, "A_F": 0, "A_FE": 0, "A_HD": 0, "A_I": 0, "A_N": 0, "A_NA": 0, "A_OA": 0, "A_P": 0, "A_S": 0, "A_SA": 0, "A_ZN": 0, "BR_C": 0, "BR_HD": 0, "BR_N": 0, "BR_OA": 0, "C_C": 0, "C_CL": 0, "C_F": 0, "C_FE": 0, "C_HD": 0, "C_I": 0, "CL_HD": 0, "CL_N": 0, "CL_OA": 0, "CL_SA": 0, "C_MG": 0, "C_MN": 0, "C_N": 0, "C_NA": 0, "C_OA": 0, "C_P": 0, "C_S": 0, "C_SA": 0, "C_ZN": 0, "FE_HD": 0, "FE_N": 0, "FE_OA": 0, "F_HD": 0, "F_N": 0, "F_OA": 0, "F_SA": 0, "HD_HD": 0, "HD_I": 0, "HD_MG": 0, "HD_MN": 0, "HD_N": 0, "HD_NA": 0, "HD_OA": 0, "HD_P": 0, "HD_S": 0, "HD_SA": 0, "HD_ZN": 0, "I_N": 0, "I_OA": 0, "MG_NA": 0, "MG_OA": 0, "MG_P": 0, "MN_N": 0, "MN_OA": 0, "MN_P": 0, "NA_OA": 0, "NA_S": 0, "NA_SA": 0, "NA_ZN": 0, "N_N": 0, "N_NA": 0, "N_OA": 0, "N_P": 0, "N_S": 0, "N_SA": 0, "N_ZN": 0, "OA_OA": 0, "OA_P": 0, "OA_S": 0, "OA_SA": 0, "OA_ZN": 0, "P_ZN": 0, "SA_SA": 0, "SA_ZN": 0, "S_ZN": 0}
        for key in data['ligand_receptor_atom_type_pairs_less_than_four']:
            if not key in self.ligand_receptor_atom_type_pairs_less_than_four:
                  print(("\tWARNING: Atoms of types " + key.replace("_"," and ") + " come within 4 angstroms of each other."))
                  print("\t  The neural networks were not trained to deal with this juxtaposition,")
                  print("\t  so it will be ignored.")

                  self.error = True
            else:
                  self.ligand_receptor_atom_type_pairs_less_than_four[key] = data['ligand_receptor_atom_type_pairs_less_than_four'][key]

        self.ligand_atom_types = {'A': 0, 'BR': 0, 'C': 0, 'CL': 0, 'F': 0, 'HD': 0, 'I': 0, 'N': 0, 'NA': 0, 'OA': 0, 'P': 0, 'S': 0, 'SA': 0}
        for key in data['ligand_atom_types']:
            if not key in self.ligand_atom_types:
                  print(("\tWARNING: The ligand contains an atoms of type " + key + ". The neural networks"))
                  print("\t  were not trained to deal with this ligand atom type, so it will be ignored.")

                  self.error = True
            else:
                  self.ligand_atom_types[key] = data['ligand_atom_types'][key]

        self.ligand_receptor_atom_type_pairs_electrostatic = {"A_MG": 0, "A_MN": 0, "BR_SA": 0, "CL_FE": 0, "CL_MG": 0, "CL_MN": 0, "CL_NA": 0, "CL_P": 0, "CL_S": 0, "CL_ZN": 0, "CU_HD": 0, "CU_N": 0, "FE_NA": 0, "FE_SA": 0, "MG_N": 0, "MG_S": 0, "MG_SA": 0, "MN_NA": 0, "MN_S": 0, "MN_SA": 0, "NA_P": 0, "P_S": 0, "P_SA": 0, "S_SA": 0, "A_A": 0.0, "A_BR": 0.0, "A_C": 0.0, "A_CL": 0.0, "A_F": 0.0, "A_FE": 0.0, "A_HD": 0.0, "A_I": 0.0, "A_N": 0.0, "A_NA": 0.0, "A_OA": 0.0, "A_P": 0.0, "A_S": 0.0, "A_SA": 0.0, "A_ZN": 0.0, "BR_C": 0.0, "BR_HD": 0.0, "BR_N": 0.0, "BR_OA": 0.0, "C_C": 0.0, "C_CL": 0.0, "C_F": 0.0, "C_FE": 0.0, "C_HD": 0.0, "C_I": 0.0, "CL_HD": 0.0, "CL_N": 0.0, "CL_OA": 0.0, "CL_SA": 0.0, "C_MG": 0.0, "C_MN": 0.0, "C_N": 0.0, "C_NA": 0.0, "C_OA": 0.0, "C_P": 0.0, "C_S": 0.0, "C_SA": 0.0, "C_ZN": 0.0, "FE_HD": 0.0, "FE_N": 0.0, "FE_OA": 0.0, "F_HD": 0.0, "F_N": 0.0, "F_OA": 0.0, "F_SA": 0.0, "HD_HD": 0.0, "HD_I": 0.0, "HD_MG": 0.0, "HD_MN": 0.0, "HD_N": 0.0, "HD_NA": 0.0, "HD_OA": 0.0, "HD_P": 0.0, "HD_S": 0.0, "HD_SA": 0.0, "HD_ZN": 0.0, "I_N": 0.0, "I_OA": 0.0, "MG_NA": 0.0, "MG_OA": 0.0, "MG_P": 0.0, "MN_N": 0.0, "MN_OA": 0.0, "MN_P": 0.0, "NA_OA": 0.0, "NA_S": 0.0, "NA_SA": 0.0, "NA_ZN": 0.0, "N_N": 0.0, "N_NA": 0.0, "N_OA": 0.0, "N_P": 0.0, "N_S": 0.0, "N_SA": 0.0, "N_ZN": 0.0, "OA_OA": 0.0, "OA_P": 0.0, "OA_S": 0.0, "OA_SA": 0.0, "OA_ZN": 0.0, "P_ZN": 0.0, "SA_SA": 0.0, "SA_ZN": 0.0, "S_ZN": 0, "F_ZN": 0}

        for key in data['ligand_receptor_atom_type_pairs_electrostatic']:
            if not key in self.ligand_receptor_atom_type_pairs_electrostatic:
                 print(("\tWARNING: Atoms of types " + key.replace("_"," and ") + ", which come within 4 angstroms of each"))
                 print("\t  other, may interact electrostatically. However, the neural networks")
                 print("\t  were not trained to deal with electrostatic interactions between atoms")
                 print("\t  of these types, so they will be ignored.")

                 self.error = True
            else:
                 self.ligand_receptor_atom_type_pairs_electrostatic[key] = data['ligand_receptor_atom_type_pairs_electrostatic'][key]

        self.active_site_flexibility = {'BACKBONE_ALPHA': 0, 'BACKBONE_BETA': 0, 'BACKBONE_OTHER': 0, 'SIDECHAIN_ALPHA': 0, 'SIDECHAIN_BETA': 0, 'SIDECHAIN_OTHER': 0}
        for key in data['active_site_flexibility']:
            self.active_site_flexibility[key] = data['active_site_flexibility'][key]

        alpha_tmp = self.active_site_flexibility['BACKBONE_ALPHA'] + self.active_site_flexibility['SIDECHAIN_ALPHA']
        beta_tmp = self.active_site_flexibility['BACKBONE_BETA'] + self.active_site_flexibility['SIDECHAIN_BETA']
        other_tmp = self.active_site_flexibility['BACKBONE_OTHER'] + self.active_site_flexibility['SIDECHAIN_OTHER']
        self.active_site_flexibility_by_structure = {'ALPHA':alpha_tmp, 'BETA':beta_tmp, 'OTHER':other_tmp}

        backbone_tmp = self.active_site_flexibility['BACKBONE_ALPHA'] + self.active_site_flexibility['BACKBONE_BETA'] + self.active_site_flexibility['BACKBONE_OTHER']
        sidechain_tmp = self.active_site_flexibility['SIDECHAIN_ALPHA'] + self.active_site_flexibility['SIDECHAIN_BETA'] + self.active_site_flexibility['SIDECHAIN_OTHER']
        self.active_site_flexibility_by_backbone_or_sidechain = {'BACKBONE':backbone_tmp, 'SIDECHAIN':sidechain_tmp}

        all = self.active_site_flexibility['BACKBONE_ALPHA'] + self.active_site_flexibility['BACKBONE_BETA'] + self.active_site_flexibility['BACKBONE_OTHER'] + self.active_site_flexibility['SIDECHAIN_ALPHA'] + self.active_site_flexibility['SIDECHAIN_BETA'] + self.active_site_flexibility['SIDECHAIN_OTHER']
        self.active_site_flexibility_all = {'all': all}

        self.hbonds = {'HDONOR-LIGAND_BACKBONE_ALPHA': 0, 'HDONOR-LIGAND_BACKBONE_BETA': 0, 'HDONOR-LIGAND_BACKBONE_OTHER': 0, 'HDONOR-LIGAND_SIDECHAIN_ALPHA': 0, 'HDONOR-LIGAND_SIDECHAIN_BETA': 0, 'HDONOR-LIGAND_SIDECHAIN_OTHER': 0, 'HDONOR-RECEPTOR_BACKBONE_ALPHA': 0, 'HDONOR-RECEPTOR_BACKBONE_BETA': 0, 'HDONOR-RECEPTOR_BACKBONE_OTHER': 0, 'HDONOR-RECEPTOR_SIDECHAIN_ALPHA': 0, 'HDONOR-RECEPTOR_SIDECHAIN_BETA': 0, 'HDONOR-RECEPTOR_SIDECHAIN_OTHER': 0}
        for key in data['hbonds']:
            key2 = key.replace("HDONOR_","HDONOR-")
            self.hbonds[key2] = data['hbonds'][key]

        hdonor_ligand = self.hbonds['HDONOR-LIGAND_BACKBONE_ALPHA'] + self.hbonds['HDONOR-LIGAND_BACKBONE_BETA'] + self.hbonds['HDONOR-LIGAND_BACKBONE_OTHER'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_ALPHA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_BETA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_OTHER']
        hdonor_receptor = self.hbonds['HDONOR-RECEPTOR_BACKBONE_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_BETA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_OTHER'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_BETA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_OTHER']
        self.hbonds_by_location_of_hdonor = {'LIGAND':hdonor_ligand, 'RECEPTOR':hdonor_receptor}

        hbond_backbone = self.hbonds['HDONOR-LIGAND_BACKBONE_ALPHA'] + self.hbonds['HDONOR-LIGAND_BACKBONE_BETA'] + self.hbonds['HDONOR-LIGAND_BACKBONE_OTHER'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_BETA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_OTHER']
        hbond_sidechain = self.hbonds['HDONOR-LIGAND_SIDECHAIN_ALPHA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_BETA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_OTHER'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_BETA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_OTHER']
        self.hbonds_by_backbone_or_sidechain = {'BACKBONE':hbond_backbone, 'SIDECHAIN':hbond_sidechain}

        hbond_alpha = self.hbonds['HDONOR-LIGAND_BACKBONE_ALPHA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_ALPHA']
        hbond_beta = self.hbonds['HDONOR-LIGAND_BACKBONE_BETA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_BETA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_BETA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_BETA']
        hbond_other = self.hbonds['HDONOR-LIGAND_BACKBONE_OTHER'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_OTHER'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_OTHER'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_OTHER']
        self.hbonds_by_structure = {'ALPHA':hbond_alpha, 'BETA':hbond_beta, 'OTHER':hbond_other}

        all = self.hbonds['HDONOR-LIGAND_BACKBONE_ALPHA'] + self.hbonds['HDONOR-LIGAND_BACKBONE_BETA'] + self.hbonds['HDONOR-LIGAND_BACKBONE_OTHER'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_ALPHA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_BETA'] + self.hbonds['HDONOR-LIGAND_SIDECHAIN_OTHER'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_BETA'] + self.hbonds['HDONOR-RECEPTOR_BACKBONE_OTHER'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_ALPHA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_BETA'] + self.hbonds['HDONOR-RECEPTOR_SIDECHAIN_OTHER']
        self.hbonds_all = {'all':all}

        self.hydrophobics = {'BACKBONE_ALPHA': 0, 'BACKBONE_BETA': 0, 'BACKBONE_OTHER': 0, 'SIDECHAIN_ALPHA': 0, 'SIDECHAIN_BETA': 0, 'SIDECHAIN_OTHER': 0}
        for key in data['hydrophobics']:
            self.hydrophobics[key] = data['hydrophobics'][key]

        alpha_tmp = self.hydrophobics['BACKBONE_ALPHA'] + self.hydrophobics['SIDECHAIN_ALPHA']
        beta_tmp = self.hydrophobics['BACKBONE_BETA'] + self.hydrophobics['SIDECHAIN_BETA']
        other_tmp = self.hydrophobics['BACKBONE_OTHER'] + self.hydrophobics['SIDECHAIN_OTHER']
        self.hydrophobics_by_structure = {'ALPHA':alpha_tmp, 'BETA':beta_tmp, 'OTHER':other_tmp}

        backbone_tmp = self.hydrophobics['BACKBONE_ALPHA'] + self.hydrophobics['BACKBONE_BETA'] + self.hydrophobics['BACKBONE_OTHER']
        sidechain_tmp = self.hydrophobics['SIDECHAIN_ALPHA'] + self.hydrophobics['SIDECHAIN_BETA'] + self.hydrophobics['SIDECHAIN_OTHER']
        self.hydrophobics_by_backbone_or_sidechain = {'BACKBONE':backbone_tmp, 'SIDECHAIN':sidechain_tmp}

        all = self.hydrophobics['BACKBONE_ALPHA'] + self.hydrophobics['BACKBONE_BETA'] + self.hydrophobics['BACKBONE_OTHER'] + self.hydrophobics['SIDECHAIN_ALPHA'] + self.hydrophobics['SIDECHAIN_BETA'] + self.hydrophobics['SIDECHAIN_OTHER']
        self.hydrophobics_all = {'all':all}

        stacking_tmp = {}
        for item in data['stacking']:
            item = item.split("_")
            stacking_tmp[item[1]] = int(item[2])
        self.stacking = {'ALPHA': 0, 'BETA': 0, 'OTHER': 0}
        for key in stacking_tmp:
            self.stacking[key] = stacking_tmp[key]

        all = self.stacking['ALPHA'] + self.stacking['BETA'] + self.stacking['OTHER']
        self.stacking_all = {'all': all}

        pi_cation_tmp = {}
        for item in data['pi_cation']:
            item = item.split("_")
            pi_cation_tmp[item[1] + "_" + item[2]] = int(item[3])
        self.pi_cation = {'LIGAND-CHARGED_ALPHA': 0, 'LIGAND-CHARGED_BETA': 0, 'LIGAND-CHARGED_OTHER': 0, 'RECEPTOR-CHARGED_ALPHA': 0, 'RECEPTOR-CHARGED_BETA': 0, 'RECEPTOR-CHARGED_OTHER': 0}
        for key in pi_cation_tmp:
            self.pi_cation[key] = pi_cation_tmp[key]

        pi_cation_ligand_charged = self.pi_cation['LIGAND-CHARGED_ALPHA'] + self.pi_cation['LIGAND-CHARGED_BETA'] + self.pi_cation['LIGAND-CHARGED_OTHER']
        pi_cation_receptor_charged = self.pi_cation['RECEPTOR-CHARGED_ALPHA'] + self.pi_cation['RECEPTOR-CHARGED_BETA'] + self.pi_cation['RECEPTOR-CHARGED_OTHER']
        self.pi_cation_charge_location = {'LIGAND':pi_cation_ligand_charged, 'RECEPTOR':pi_cation_receptor_charged}

        pi_cation_alpha = self.pi_cation['LIGAND-CHARGED_ALPHA'] + self.pi_cation['RECEPTOR-CHARGED_ALPHA']
        pi_cation_beta = self.pi_cation['LIGAND-CHARGED_BETA'] + self.pi_cation['RECEPTOR-CHARGED_BETA']
        pi_cation_other = self.pi_cation['LIGAND-CHARGED_OTHER'] + self.pi_cation['RECEPTOR-CHARGED_OTHER']
        self.pi_cation_by_structure = {'ALPHA':pi_cation_alpha, 'BETA':pi_cation_beta, "OTHER":pi_cation_other}

        all = self.pi_cation['LIGAND-CHARGED_ALPHA'] + self.pi_cation['LIGAND-CHARGED_BETA'] + self.pi_cation['LIGAND-CHARGED_OTHER'] + self.pi_cation['RECEPTOR-CHARGED_ALPHA'] + self.pi_cation['RECEPTOR-CHARGED_BETA'] + self.pi_cation['RECEPTOR-CHARGED_OTHER']
        self.pi_cation_all = {'all': all}

        t_shaped_tmp = {}
        for item in data['t_shaped']:
            item = item.split("_")
            t_shaped_tmp[item[1]] = int(item[2])
        self.t_shaped = {'ALPHA': 0, 'BETA': 0, 'OTHER': 0}
        for key in t_shaped_tmp:
            self.t_shaped[key] = t_shaped_tmp[key]

        all = self.t_shaped['ALPHA'] + self.t_shaped['BETA'] + self.t_shaped['OTHER']
        self.t_shaped_all = {'all': all}

        self.salt_bridges = {'ALPHA': 0, 'BETA': 0, 'OTHER': 0}
        for key in data['salt_bridges']:
            key2 = key.replace("SALT-BRIDGE_","")
            self.salt_bridges[key2] = data['salt_bridges'][key]

        all = self.salt_bridges['ALPHA'] + self.salt_bridges['BETA'] + self.salt_bridges['OTHER']
        self.salt_bridges_all = {'all': all}

        self.input_vector = []
        self.input_vector.extend(self.vina_output) # a list
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.ligand_receptor_atom_type_pairs_less_than_four)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.ligand_receptor_atom_type_pairs_electrostatic)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.ligand_atom_types)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.ligand_receptor_atom_type_pairs_less_than_two_half)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.hbonds)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.hydrophobics)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.stacking)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.pi_cation)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.t_shaped)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.active_site_flexibility)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.salt_bridges)
        self.input_vector = self.extend_list_by_dictionary(self.input_vector, self.rotateable_bonds_count)


class command_line_parameters:

    params = {}

    def __init__(self, parameters):

        global vina_executable

        # first, set defaults
        self.params['receptor'] = ''
        self.params['ligand'] = ''
        self.params['vina_executable'] = vina_executable
        self.params['check_vina_version'] = "FALSE" # TRUE by default, but setting to false will speed up execution. Good when rescoring many poses.

        # now get user inputed values

        for index in range(len(parameters)):
            item = parameters[index]
            if item[:1] == '-': # so it's a parameter key value
                key = item.replace('-','').lower()

                if key == "help":
                    print("INTRODUCTION")
                    print("============\n")
                    print((textwrap.fill("NNScore 2.02 (NNScore2s.py) is a python script for predicting the binding affinity of receptor-ligand complexes. It is particularly well suited for rescoring ligands that have been docked using AutoDock Vina.") + "\n"))
                    print("REQUIREMENTS")
                    print("============\n")
                    print((textwrap.fill("Python: NNScore 2.02 has been tested using Python versions 2.6.5, 2.6.1, and 2.5.2 on Ubuntu 10.04.1 LTS, Mac OS X 10.6.8, and Windows XP Professional, respectively. A copy of the Python interpreter can be downloaded from http://www.python.org/getit/.") + "\n"))
                    print((textwrap.fill("AutoDock Vina 1.1.2: NNScore 2.02 uses AutoDock Vina 1.1.2 to obtain some information about the receptor-ligand complex. Note that previous versions of AutoDock Vina are not suitble. AutoDock Vina 1.1.2 can be downloaded from http://vina.scripps.edu/download.html.") + "\n"))
                    print((textwrap.fill("MGLTools: As receptor and ligand inputs, NNScore 2.02 accepts models in the PDBQT format. Files in the more common PDB format can be converted to the PDBQT format using scripts included in MGLTools (prepare_receptor4.py and prepare_ligand4.py). MGLTools can be obtained from http://mgltools.scripps.edu/downloads.") + "\n"))
                    print("COMMAND-LINE PARAMETERS")
                    print("=======================\n")
                    print((textwrap.fill("-receptor: File name of the receptor PDBQT file.") + "\n"))
                    print((textwrap.fill("-ligand: File name of the ligand PDBQT file. AutoDock Vina output files, typically containing multiple poses, are also permitted.") + "\n"))
                    print((textwrap.fill("-vina_executable: The location of the AutoDock Vina 1.1.2 executable. If you don't wish to specify the location of this file every time you use NNScore 2.02, simply edit the vina_executable variable defined near the begining of the NNScore2.py script.") + "\n"))
                    print("PROGRAM OUTPUT")
                    print("==============\n")
                    print((textwrap.fill("NNScore 2.02 evaluates each of the ligand poses contained in the file specified by the -ligand tag using 20 distinct neural-network scoring functions. The program then seeks to identify which of the poses has the highest predicted affinity using several metrics:") + "\n"))
                    print((textwrap.fill("1) Each of the 20 networks are considered separately. The poses are ranked in 20 different ways by the scores assigned by each network.") + "\n"))
                    print((textwrap.fill("2) The poses are ranked by the best score given by any of the 20 networks.") + "\n"))
                    print((textwrap.fill("3) The poses are ranked by the average of the scores given by the 20 networks. This is the recommended metric.") + "\n"))


                    sys.exit(0)

                value = parameters[index+1]
                if key in list(self.params.keys()):
                    self.params[key] = value
                    parameters[index] = ""
                    parameters[index + 1] = ""

        if self.okay_to_proceed() is True:
            print("Command-line parameters used:")
            print(("\tReceptor:        " + self.params["receptor"]))
            print(("\tLigand:          " + self.params["ligand"]))
            print(("\tVina executable: " + self.params["vina_executable"] + "\n"))

        # make a list of all the command-line parameters not used
        error = ""
        for index in range(1,len(parameters)):
            item = parameters[index]
            if item != "": error = error + item + " "

        if error != "":
            print("WARNING: The following command-line parameters were not used:")
            print(("\t" + error + "\n"))


        # do a check to make sure the autodock vina is version 1.1.2
        if self.params["check_vina_version"] == "TRUE":
            vina_version_output = ""
            if not os.path.exists(self.params['vina_executable']):
                vina_version_output = ""
            else:
                try:
                    vina_version_output = getCommandOutput2('"' + self.params['vina_executable'] + '"' + ' --version')
                except:
                    vina_version_output = ""

            if not " 1.1.2 " in vina_version_output:
                print("ERROR: NNScore 2.02 is designed to work with AutoDock Vina 1.1.2.\nOther versions of AutoDock may not work properly. Please download\nAutoDock Vina 1.1.2 from http://vina.scripps.edu/download.html.\n")
                print("Once this executable is downloaded, you can use the -vina_executable\ntag to indicate its location. Alternatively, you can simply modify\nthe vina_executable variable defined near the beginning of\nthe NNScore2.py file.\n")
                sys.exit(0)


    def okay_to_proceed(self):
        if self.params['receptor'] != '' and self.params['ligand'] != '' and self.params['vina_executable'] != '':
            return True
        else: return False

def score_to_kd(score):
    kd = math.pow(10,-score)
    if score <= 0: return "Kd = " + str(round(kd,2)) + " M"
    temp = kd * pow(10,3)
    if temp >=1 and temp <1000: return "Kd = " + str(round(temp,2)) + " mM"
    temp = temp * pow(10,3)
    if temp >=1 and temp <1000: return "Kd = " + str(round(temp,2)) + " uM"
    temp = temp * pow(10,3)
    if temp >=1 and temp <1000: return "Kd = " + str(round(temp,2)) + " nM"
    temp = temp * pow(10,3)
    if temp >=1 and temp <1000: return "Kd = " + str(round(temp,2)) + " pM"
    temp = temp * pow(10,3)
    #if temp >=1 and temp <1000:
    return "Kd = " + str(round(temp,2)) + " fM"
    #return "Kd = " + str(kd) + " M"

def networks():
    nets = []
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.0365550106805, 2: -7.70497088454, 3: 4.03521786068, 4: -17.7780869406, 5: 0.312436881124, 6: -0.604422116938, 7: -11.3684936384, 8: 6.40139241738, 9: -0.781349952423, 10: -1.33651172566, 11: 2.20875884439, 12: -47.9323010122, 13: -13.5034062799, 14: -7.54110904968, 15: -7.0681928446, 16: 7.39488788317, 17: -6.58289705202, 18: -13.7078006997, 19: -11.0656030895, 20: 3.15712571843, 21: 6.40764916301, 22: 7.55704960887, 23: 1.41769609969, 24: 1.57413087701, 25: 1.94108429672, 26: -3.23019234928, 27: 9.50072482917, 28: -7.05323364739, 29: 6.81959181299, 30: 19.7716739577, 31: 19.1419147149, 32: 12.7745568234, 33: -1.55930876711, 34: 8.16963302198, 35: 4.40491492155, 36: 8.39971617503, 37: 1.16540288772, 38: 5.38405200922, 39: 3.43304434508, 40: 3.57744627532, 41: -3.74830905053, 42: 1.56805048008, 43: 0.861173561675, 44: -7.10135720693, 45: 1.59086241013, 46: -0.29198295757, 47: -7.97377357354, 48: 1.57350330459, 49: 2.33625781798, 50: -3.96043071847, 51: 3.12992421996, 52: -12.5005211681, 53: 1.32889168261, 54: 0.60933352974, 55: 2.68018360277, 56: 1.12581251902, 57: 4.29872334072, 58: 4.87615232627, 59: 8.63315225586, 60: 9.33278363721, 61: 3.97724693909, 62: -23.9413909979, 63: -3.15927498771, 64: -5.06642985755, 65: -20.7879487762, 66: 1.06555235472, 67: -10.728658535, 68: 1.42433965576, 69: -2.97460071552, 70: 1.26783045057, 71: 1.33884893361, 72: -5.29417474381, 73: 1.26000386861, 74: -7.8103156031, 75: 0.623704853506, 76: 1.16714615833, 77: 1.84301896217, 78: 2.960822791, 79: 16.1950046047, 80: -1.87805939291, 81: 0.0984598097621, 82: -3.92122784858, 83: 1.14726778327, 84: -0.793967875049, 85: 2.63827912739, 86: 1.43699628039, 87: -2.84998335937, 88: -12.1700951163, 89: -7.01140138997, 90: -10.0314666574, 91: 1.13789951418, 92: 1.14014124171, 93: -15.8311178938, 94: 1.08335400182, 95: 1.37474780076, 96: 1.24277667453, 97: 1.39716499165, 98: 1.08696756323, 99: 0.555743336234, 100: 1.28010875896, 101: -6.34583696691, 102: 1.00575217624, 103: -17.4169099494, 104: 1.08348055603, 105: 0.481890814105, 106: 4.89577252197, 107: 1.08471434151, 108: 1.32831647638, 109: -7.12462127059, 110: -0.153544487389, 111: -5.03941434167, 112: -6.87357843089, 113: -42.3719578012, 114: -1.21007707402, 115: 3.46040178329, 116: 7.19819085356, 117: -19.6734372783, 118: -2.4954112808, 119: 22.9256332602, 120: 1.83774077142, 121: 19.4646869092, 122: 4.08346528081, 123: -5.31043248935, 124: 18.7959918839, 125: 0.185968169005, 126: 0.265880947693, 127: -0.176325941207, 128: 3.17049039324, 129: 7.64811560701, 130: 0.173048563114, 131: 1.26078443681, 132: -0.343065726164, 133: -15.2738403862, 134: -0.0529705215103, 135: 0.952529722019, 136: -3.35342381513, 137: -1.6489837641, 138: 12.0510230413, 139: -1.33645178657, 140: -2.48485427004, 141: -3.72004274233, 142: 20.3768588462, 143: 1.4069806984, 144: 0.0391194121979, 145: 4.85668623684, 146: 0.314409364312, 147: 0.20036444266, 148: 19.1318137367, 149: 0.148422515127, 150: 1.19146442279, 151: -2.11759528826, 152: 0.479799140544, 153: 0.0757619871662, 154: 0.295357624706, 155: 0.140765703741, 156: 1.28874520381, 157: -1.70552204436, 158: -0.582466515171, 159: 0.333906984874, 160: 0.137451204434, 161: -6.28344269498, 162: 0.340837920963, 163: -9.51482245726, 164: 3.91894301531, 165: 3.66643138455, 166: 28.0152652667, 167: -0.380679246146, 168: 3.76322117038, 169: -10.246689862, 170: -0.563568136985, 171: 1.38587710386, 172: 0.285954781022, 173: 7.17914456615, 174: 0.259691583737, 175: 6.44256106411, 176: 1.57554350861, 177: -10.793589222, 178: 1.26602293155, 179: 0.170968080358, 180: 0.284413440334, 181: 0.268913721915, 182: 0.116682470486, 183: 0.776551222061, 184: 4.90515398825, 185: 3.08165260626, 186: -0.549138426907, 187: -25.6055014524, 188: 4.55462835653, 189: 9.2965195351, 190: 2.73402221222, 191: 1.01132279009, 192: -8.27958047034, 193: 0.0957224232405, 194: -3.60140618331, 195: 1.55709903416, 196: 0.341138757997, 197: 0.783888279583, 198: 2.45176084705, 199: 0.0726514278223, 200: 1.15670405132, 201: 0.347363991551, 202: -5.71516043964, 203: 0.255294993254, 204: 0.125839304469, 205: -19.5649294135, 206: 1.15586705186, 207: -7.50148301249, 208: 0.200588094116, 209: 0.13545352421, 210: 1.20634492183, 211: 0.108317873677, 212: 1.27871146189, 213: 1.28522117023, 214: -5.39149199374, 215: 0.142757015556, 216: 0.375279802694, 217: 2.43389232468, 218: 0.964011498815, 219: 21.5190535872, 220: 0.253077201474, 221: -2.14065687056, 222: -2.02839214755, 223: -9.65336385869, 224: 1.25046310619, 225: 0.865302347185, 226: 0.49252333162, 227: 0.41402521696, 228: 0.857418025416, 229: 1.06703697235, 230: 0.459834534498, 231: 5.81896869265, 232: 0.583359831542, 233: 5.90605702972, 234: -0.388068063217, 235: 1.19918753068, 236: -3.39417556395, 237: 18.0695585521, 238: 8.0801721136, 239: 2.12499481712, 240: 1.19725406997, 241: 10.2887510375, 242: -10.2592676487, 243: -10.6263098781, 244: -7.95164045816, 245: 2.9502046878, 246: 9.09549663252, 247: 3.06207451392, 248: 0.305859691321, 249: -10.434667962, 250: 7.66185180816, 251: -3.73750046917, 252: 6.08069342316, 253: 3.65556036565, 254: 0.13759869877, 255: -3.0977630425, 256: 1.13984995541, 257: -4.49107104232, 258: 1.12147824882, 259: 1.19914929769, 260: -2.40704481053, 261: -6.41700247298, 262: 1.28929002933, 263: 2.32668820314, 264: 0.478400630724, 265: 1.32779980714, 266: 13.6428046327, 267: -18.0148725118, 268: 2.50324675053, 269: 2.04905430164, 270: -5.06440553131, 271: -1.10028889448, 272: -19.6822036891, 273: -5.17393652708, 274: -3.75310065106, 275: -11.7775820378, 276: -2.67441187366, 277: -17.8082464253, 278: 6.69330654063, 279: -1.54104841445, 280: 15.562179556, 281: -2.79670629983, 282: -9.14684710056, 283: 1.51527223903, 284: 17.4687684732, 285: -13.6360886057, 286: 2.17702818522, 287: 1.1162244013, 288: 0.924402035402, 289: 1.1727016077, 290: 2.06525003462, 291: 0.222030336575, 292: -0.31037890408, 293: -9.53982334782, 294: -1.86703942593, 295: 31.3546735098, 296: -3.39161498144, 297: -14.4640996768, 298: 0.885198932898, 299: 1.12021366964, 300: -0.0726362603648, 301: -0.209723959823, 302: 2.3545996294, 303: 1.29637795309, 304: 4.33341783883, 305: 4.75973325378, 306: 1.17680605663, 307: 5.4608443128, 308: 1.23760258712, 309: 1.33379530541, 310: 0.38477822252, 311: -4.73106586365, 312: 1.30349143021, 313: 1.3057284328, 314: -11.135088064, 315: 1.27941209144, 316: -15.3966693656, 317: 1.17232600586, 318: 0.844159617174, 319: -19.8173136206, 320: 1.28295522675, 321: 1.30548222468, 322: 1.39385326286, 323: -7.00194686906, 324: -8.30602920624, 325: -0.836129382991, 326: 15.2966503123, 327: -7.13455766474, 328: 9.69480376787, 329: -11.2270988936, 330: -14.3171547873, 331: -34.165964952, 332: -10.9537177701, 333: 1.01100771493, 334: 0.820121701196, 335: 7.76986393961, 336: 11.0463549409, 337: -2.5300090149, 338: -8.85348877614, 339: 1.31192561598, 340: 9.53954976205, 341: -4.13189420473, 342: -1.19223214298, 343: 3.53255523035, 344: 7.99665680737, 345: 8.88296555971, 346: -4.81956248264, 347: 2.55404046856, 348: -14.7966481042, 349: -9.456536338, 350: 0.406163133626, 351: 82.6049743896, 352: -11.0300910987, 353: 23.2686139868, 354: -6.63935491753, 355: -1.78590667517, 356: -9.65116521366, 357: -15.7286685593, 358: -42.4640552945, 359: -1.22966803484, 360: -7.19219476825, 361: -24.7680094144, 362: 0.50290277203, 363: -6.78066473507, 364: -8.56105279762, 365: 14.4593855405, 366: -11.3989603346, 367: -4.25117940242, 368: -33.4257853637, 369: -32.3912515864, 370: 5.18534152283, 371: 13.3409093997, 372: -5.6064316006, 373: -14.1897385814, 374: -28.5278615235, 375: -17.4937400041, 376: 40.3795244807, 377: -2.81071100633, 378: 0.132261359692, 379: 1.35935635739, 380: -5.40338283213, 381: 14.1536757126, 382: -9.82453245508, 383: -9.93898733863, 384: 9.94725719484, 385: -1.20291581377, 386: 22.2192660451, 387: -6.12676903388, 388: 5.74872845147, 389: -20.810791938, 390: -4.17919079731, 391: 3.82372671548, 392: 4.69697121984, 393: -8.02910095304, 394: 5.62965913116, 395: -9.62120701832, 396: 18.9007198336, 397: 1.03633636644, 398: 15.8885866182, 399: -0.753633321119, 400: 12.4491502667, 401: 24.3518991665, 402: 3.6189165989, 403: -5.93602657426, 404: 9.07479869773, 405: -3.91563989626, 406: 47.0191948722, 407: 11.4262528197, 408: 25.6389639525, 409: -41.4169274679, 410: 18.3024290184, 411: -12.9000971474, 412: -6.8720038982, 413: 34.4777603112, 414: -32.5031430597, 415: -3.07546265598, 416: -2.70135939111, 417: 2.88964082065, 418: 10.9141310174, 419: 3.42709485854, 420: 5.90301796127, 421: 24.1096311463, 422: 3.79624212231, 423: -30.5711381719, 424: -5.50665336958, 425: 7.71485274313, 426: -6.64251359881, 427: -4.76012873365, 428: 48.3877460336, 429: -4.75767915248, 430: -1.00308691896, 431: 0.0885919869879, 432: 2.59393069471, 433: -41.2360893636, 434: 3.26704631049, 435: 9.17187716704, 436: -14.6490260542, 437: 30.9238568571, 438: 3.91754543932, 439: 2.94446109258, 440: -19.245281794, 441: 3.49958071859, 442: 4.31504270063, 443: 3.13826481029, 444: 3.27479569772, 445: 3.00093154648, 446: -9.63832016528, 447: 2.04742866165, 448: 10.1056079671, 449: 1.21151843624, 450: 10.0951774527, 451: -2.80206261746, 452: -9.90166011106, 453: -4.71419804689, 454: -0.497946374812, 455: -36.2333717698, 456: 3.40661341965, 457: 3.82980909475, 458: 2.42893236917, 459: -8.59708587486, 460: 1.24942184325, 461: -16.7502302411, 462: 16.6132450011, 463: 36.6847751804, 464: -50.5779467295, 465: -18.2940868275, 466: 12.0855095329, 467: -13.3672362873, 468: -5.28998769212, 469: -23.5244361656, 470: -20.5059098917, 471: -1.32773826202, 472: -6.07780501405, 473: 0.556571994068, 474: -12.8563193851, 475: 0.520227351274, 476: 1.28495462009, 477: 3.18212861926, 478: 0.555240190082, 479: 0.463106429559, 480: 3.34518131448, 481: -1.33932912076, 482: 12.7263448818, 483: 7.85668027762, 484: -0.117643334279, 485: 7.83114341205, 486: -1.34902962005, 487: 18.1373385405, 488: -11.7577111858, 489: -0.743449779762, 490: -8.43669873018, 491: 39.5279802732, 492: -24.7724167297, 493: 1.78187632008, 494: -1.47554929198, 495: 0.648606448347, 496: 0.690218015526, 497: -3.17671701963, 498: 0.634910005719, 499: 3.42938016249, 500: -3.68099727106, 501: 0.825663420721, 502: 0.721556447745, 503: 0.795896592016, 504: 0.45931770971, 505: 1.8460742011, 506: 0.374240051175, 507: -0.712031744555, 508: -0.292086786746, 509: -2.32563232781, 510: 4.36743061945, 511: 3.87846648646, 512: -27.0065496484, 513: 3.54547998234, 514: -21.5248551739, 515: 48.5534897936, 516: 0.636417034658, 517: 1.91401600724, 518: -5.73949238044, 519: 6.28865571198, 520: -18.018525446, 521: 0.653250078594, 522: -7.70795498076, 523: 0.594576905857, 524: -39.2193649178, 525: -62.8848355817, 526: -7.04095569611, 527: 3.78449189395, 528: 0.604163080473, 529: 0.437795577923, 530: 1.18936480395, 531: 0.97813339891, 532: -3.38481649451, 533: 50.8625441137, 534: 8.49074719905, 535: -20.4650363619, 536: -11.0190967821, 537: -6.59112155349, 538: -2.94430323876, 539: 11.9183399781, 540: -0.979603732649, 541: -17.4219453003, 542: 0.481457378621, 543: 36.1773448523, 544: -0.409529355923, 545: -3.44699853455, 546: -6.71039762857, 547: 9.75989355552, 548: 0.535319874799, 549: 3.52687850276, 550: 0.649041624976, 551: 24.1940701504, 552: 0.5799801504, 553: 0.584409694069, 554: -15.7393042344, 555: 3.79286083841, 556: 40.3962607829, 557: 0.637844320898, 558: 0.683224900969, 559: -1.26414616496, 560: 0.712168019627, 561: 2.59933363919, 562: 5.14176082498, 563: -13.9173930974, 564: -0.462936983022, 565: -0.239405968316, 566: 41.9981449105, 567: 4.03095720743, 568: -0.0362463733404, 569: 4.83170537274, 570: 7.11529800743, 571: -6.62340361773, 572: 1.00679577872, 573: -0.408593714097, 574: 0.321293891503, 575: 0.787763635748, 576: 18.3011391658, 577: 36.3809463465, 578: 1.75401827223, 579: -0.812810922045, 580: -20.6181650005, 581: 0.277011017368, 582: 36.5915884079, 583: -27.2604235975, 584: 4.32462780627, 585: -0.62156574817, 586: 0.596449480677, 587: -7.21370168551, 588: 5.50280892453, 589: 3.2685038111, 590: -1.81342733834, 591: -14.9825378258, 592: -28.0708664893, 593: 6.6616021354, 594: -9.13962110792, 595: -3.41640737166, 596: 1.12737004523, 597: 32.4043919973, 598: -25.771082015, 599: 45.1195082776, 600: -13.8289470334, 601: 0.876900016412, 602: 0.421509724298, 603: 9.22608568738, 604: 6.09487537647, 605: 2.8113876576, 606: -2.4312104282, 607: 3.4620007146, 608: 3.4669479415, 609: -5.13545328439, 610: -23.1413720498, 611: 2.12096461788, 612: 8.57537071444, 613: 3.6780201229, 614: 1.87399389881, 615: 1.92327272733, 616: -30.0158489301, 617: 9.34865088747, 618: 12.152986666, 619: 4.96643866266, 620: 2.99531996858, 621: 11.135795144, 622: 0.57334712664, 623: 61.5864773546, 624: 17.1966345507, 625: -12.9141827489, 626: 4.15410239442, 627: 6.42415830845, 628: 9.95119592054, 629: -33.6855456001, 630: 4.089452204, 631: 25.5931412107, 632: 25.7172718311, 633: 88.5317406024, 634: 36.3176035907, 635: -5.28848239693, 636: 2.72810629843, 637: 11.5713057091, 638: -2.95322626424, 639: 9.14533275199, 640: -31.3645195286, 641: 12.0510941933, 642: -2.1371755545, 643: 5.84275967746, 644: 12.3967555106, 645: -21.2960574338, 646: 2.93863900027, 647: 5.77603954994, 648: -0.721305465205, 649: 10.6499856557, 650: 4.33941994243, 651: -26.3372776248, 652: 2.89924470626, 653: 11.4961118224, 654: -4.22931519149, 655: 3.25173414896, 656: -59.3317414427, 657: 3.71585736585, 658: 3.53404459825, 659: 0.304130626515, 660: 14.1711121119, 661: 3.44766079064, 662: 3.82026984574, 663: 25.435495455, 664: 3.60482929569, 665: -23.0251539949, 666: 3.29956274013, 667: 3.61185267765, 668: -4.64630031628, 669: 2.80498208258, 670: 2.23566914657, 671: -6.31336153031, 672: 20.0792033279, 673: -21.6022203147, 674: 14.6605499575, 675: -85.6172230367, 676: -0.481102743472, 677: 4.66188779668, 678: -24.3854547514, 679: -2.34754985256, 680: -7.23220719234, 681: 12.9128305471, 682: -5.19794204279, 683: 8.69192942108, 684: 1.25310849908, 685: -61.1882792769, 686: -5.12802450052, 687: 0.209162849048, 688: 3.53381113645, 689: 1.98689260073, 690: 8.06079053242, 691: 17.7201022917, 692: -12.5084291139, 693: 3.10632586249, 694: 0.915120839587, 695: -21.6457799152, 696: 17.1234040764, 697: -26.7583896019, 698: -54.0028579574, 699: 0.69380583412, 700: 24.092516536, 701: 17.7780025862, 702: -25.0058721933, 703: 11.6497352222, 704: -32.372396869, 705: 16.5733955319, 706: 9.55808654235, 707: 25.4409541658, 708: -5.31929553458, 709: -20.6920042333, 710: 14.4882108621, 711: -19.6220834344, 712: -27.6398020781, 713: 2.83873759624, 714: -30.9917034676, 715: -22.7688081014, 716: 18.9758293297, 717: 12.0410277261, 718: -7.50107555964, 719: -14.221380978, 720: 34.808828609, 721: 9.81744940687, 722: -6.54313255068, 723: -11.7077215207, 724: 13.1687197133, 725: 6.71308800112, 726: 20.382018199, 727: -15.3542055107, 728: 23.2250475315, 729: 32.7715477694, 730: 23.5940415037, 731: 12.2279370623, 732: 8.32832271912, 733: 4.76410711571, 734: 1.7962383927, 735: 5.60242831957, 736: 40.0602133367, 737: -4.76127481131, 738: 18.6305986161, 739: -0.19218460542, 740: -37.298346165, 741: 48.5239536238, 742: 64.8469121354, 743: -22.3544424541, 744: 13.5275233709, 745: -11.190916778, 746: 22.4864973797, 747: 19.3560980787, 748: 12.8977155299, 749: 12.0957852531, 750: -45.3703623792, 751: 2.8425286904, 752: 8.01119867067, 753: -9.27319466533, 754: 5.457003412, 755: -7.77339486827, 756: 2.20286743355, 757: 1.74759426092, 758: 9.87890109452, 759: 31.8491861951, 760: 17.6866932187, 761: -24.4973106234, 762: -24.8141197182, 763: 12.7970872995, 764: 3.47191086419, 765: 9.44045695779, 766: 3.89607623311, 767: 8.81823523199, 768: 4.37126279664, 769: 0.98957183515, 770: 1.52478910499, 771: 5.07729087302, 772: 33.6184031873, 773: 4.86145785096, 774: 17.1832844572, 775: 4.84830677099, 776: 4.43245378612, 777: -9.98337018066, 778: -0.819040585231, 779: 3.27410819472, 780: -1.89410716039, 781: 4.72668556699, 782: 3.68017188323, 783: -23.9752709056, 784: 2.94299170361, 785: -10.1170488777, 786: -30.720565103, 787: 5.80049486333, 788: -31.2906106282, 789: 2.23926462085, 790: 3.87528916981, 791: -73.8575753568, 792: 3.78205724656, 793: 3.69314350705, 794: 8.60165769197, 795: -8.80901296212, 796: -3.74349458773, 797: 8.90661585568, 798: 6.72630172963, 799: -10.2117126704, 800: -18.0927135683, 801: 19.9992638617, 802: 3.05161138949, 803: 2.60456018595, 804: -40.5773569608, 805: 17.906291899, 806: 1.1824402883, 807: 0.168618381396, 808: -38.327380655, 809: 3.66663404711, 810: 21.0495244366, 811: 62.6044148887, 812: 19.0251526795, 813: -1.278186574, 814: 10.2366615053, 815: 2.73394490568, 816: 23.7717378206, 817: -48.7557795984, 818: 58.5558940862, 819: -40.0465605177, 820: 37.554943315, 821: 4.36164217412, 822: 3.98202555028, 823: 46.1241451219, 824: 0.968639802286, 825: -0.943792971832, 826: -2.35295234555, 827: -25.5447696287, 828: 0.943684388245, 829: 4.37991599229, 830: 8.29534166512, 831: 31.5988356763, 832: -0.954672335033, 833: 0.923090958084, 834: 15.4419492791, 835: 1.94987396523, 836: -32.0066428654, 837: 3.30236784704, 838: 8.01330118029, 839: -23.7306720253, 840: -2.80597571871, 841: 13.7316259043, 842: 3.78627002185, 843: -10.120225984, 844: 1.25205492565, 845: 1.06190678202, 846: 18.4016663475, 847: 1.0694404191, 848: 4.99900089931, 849: -5.06846468683, 850: 1.20105911635, 851: 0.951978072572, 852: 1.10901174224, 853: 0.945381111573, 854: -3.32546240353, 855: -15.5611754289, 856: 2.09867652741, 857: 1.62562141197, 858: 1.91209215801, 859: 5.43499202409, 860: 9.91594402876, 861: -11.1424877993, 862: -13.5266022139, 863: -93.477961211, 864: -24.0908866063, 865: 0.30927083786, 866: -0.0569947516601, 867: 0.92866101175, 868: -0.137777481086, 869: 10.4292936315, 870: 2.08277166516, 871: 9.59428865782, 872: 0.835488149934, 873: -13.9596207601, 874: 53.7295480006, 875: -25.1117328004, 876: -0.237317300823, 877: 0.798836453749, 878: 1.33421630255, 879: 1.22497359382, 880: 0.963376867077, 881: 13.9802144826, 882: -28.5008689297, 883: 4.22645551838, 884: 23.8453506295, 885: -51.6672053559, 886: 4.19853223965, 887: -8.58531756541, 888: -1.85811004949, 889: 4.93676708496, 890: -45.5803698426, 891: 0.404060207362, 892: -43.8892067558, 893: 13.5168020617, 894: 3.85701991734, 895: 2.70419234731, 896: -22.4383810029, 897: 0.988816092837, 898: 5.35814211186, 899: 0.871438077696, 900: -79.4681069556, 901: 0.878345216445, 902: 0.936310453121, 903: 4.67527567549, 904: 5.13701558283, 905: 6.61000358202, 906: 0.882108472184, 907: 0.906921020783, 908: 2.18322695552, 909: 0.937590312818, 910: -10.4552504322, 911: -4.64316470319, 912: -9.89213278381, 913: 2.09543272731, 914: 1.50804202753, 915: 14.3041406608, 916: -0.481766994849, 917: 15.4187501103, 918: 24.8979570161, 919: -14.3743458561, 920: -9.72405219474, 921: -20.9306714334, 922: 4.11836921356, 923: -4.52484832663, 924: 3.16992889225, 925: 0.533985618474, 926: 3.09969675496, 927: -1.30504995907, 928: -1.38109114657, 929: 19.5682317394, 930: -0.306745423421, 931: 12.6488354399, 932: -1.90004592441, 933: 3.83328136554, 934: 1.18757442528, 935: -75.6513674061, 936: -51.088907122, 937: 6.11437449311, 938: 4.42194116132, 939: -9.62920914268, 940: -37.3031284863, 941: 6.30383495023, 942: -31.760052368, 943: 8.04538009369, 944: -37.8773630285, 945: -23.5602758203, 946: -6.52254248133, 947: -6.51625005321, 948: 21.4072112779, 949: 34.381206869, 950: -4.47956147608, 951: -69.1716523445, 952: 7.18231899518, 953: 16.5203641528, 954: 3.57967842557, 955: 3.73527249676, 956: 3.2556817133, 957: 5.08001895746, 958: -6.7709034966, 959: 2.75346927212, 960: 4.35418642857, 961: -6.64337802009, 962: -0.448319100763, 963: -3.54167648434, 964: -13.1788143288, 965: -47.2018253755, 966: -5.46468534035, 967: -13.8565638722, 968: 21.0429647367, 969: -23.5109618601, 970: 33.9889121995, 971: 7.76956120181, 972: 87.4147750956, 973: -11.7174121747, 974: -30.5081697037, 975: -30.4698083055, 976: -33.5461285216, 977: -6.59290997173, 978: 13.0454900667, 979: -32.2690621374, 980: 34.4103992731, 981: -29.2124208348, 982: -81.4007563034, 983: -57.6999423867, 984: 31.3279297837, 985: 4.64968811409, 986: 0.010010776116, 987: 3.99689203625, 988: 7.12412306111, 989: -0.516314207036, 990: -32.1724730712, 991: 13.040625105, 992: -38.3695588372, 993: 56.5609068422, 994: 1.02457257021, 995: -12.5351826965, 996: -12.4459858487, 997: 4.57802760632, 998: -37.6130474974, 999: 5.5462059628, 1000: 12.4090661218, 1001: 13.9448864014, 1002: 4.97633427746, 1003: -14.0890236573, 1004: 3.7534157681, 1005: 2.22227110775, 1006: 5.44560445027, 1007: 5.49353847495, 1008: -0.980857323523, 1009: 58.9607580579, 1010: 5.65696163754, 1011: 4.03895794049, 1012: -30.0291924718, 1013: 5.20901525157, 1014: 42.2120550508, 1015: 4.77440392728, 1016: 5.13508208475, 1017: 54.9260614866, 1018: 4.53049459192, 1019: -14.1750588593, 1020: 37.3413088199, 1021: -37.1523433782, 1022: 0.899525832688, 1023: 53.5586840167, 1024: 50.4102993372, 1025: -9.17536849208, 1026: -51.3301443098, 1027: -4.75482286914, 1028: -14.5628093952, 1029: -11.7822881028, 1030: 1.36029008183, 1031: 4.29536017835, 1032: -0.836264377911, 1033: 6.71007198728, 1034: 21.1136742732, 1035: 5.83234203308, 1036: -1.09259879283, 1037: 5.39196422189, 1038: -7.71816240087, 1039: -16.3958814187, 1040: 0.41686749279, 1041: 19.0869679904, 1042: -21.4042734008, 1043: -1.80858479202, 1044: -5.39433805085, 1045: 8.05377468166, 1046: 7.83443799989, 1047: -17.27772637, 1048: 0.742976316539, 1049: -35.7712544729, 1050: -31.9713956964, 1051: 21.1467037086, 1052: 59.9501685466, 1053: 6.54379239673, 1054: 31.6207544606, 1055: 27.9063799542, 1056: -12.892673251, 1057: -27.2365592823, 1058: 6.95237055175, 1059: -24.2482367003, 1060: -12.1036767763, 1061: -5.11282167271, 1062: -8.72684785634, 1063: -27.9429735842, 1064: -67.5530919749, 1065: -23.9096112484, 1066: -32.3344472917, 1067: -14.7381264942, 1068: -54.2538183849, 1069: -39.959716452, 1070: -0.249524908879, 1071: 20.8073237192, 1072: 7.54168813818, 1073: -14.7935471431, 1074: 34.542798781, 1075: 18.2872328646, 1076: -21.4554000415, 1077: -2.61912033092, 1078: -5.78596100563, 1079: -93.5653604316, 1080: 1.45603958406, 1081: -18.4443056404, 1082: -16.5117990578, 1083: -5.46781671688, 1084: -36.39596057, 1085: -29.4734204947, 1086: -9.18439967631, 1087: 27.2138111499, 1088: -8.79690369245, 1089: 4.97916190448, 1090: 8.55390334751, 1091: -23.160682587, 1092: 3.8996837939, 1093: -16.4000861904, 1094: 11.8121073697, 1095: 4.22152946244, 1096: -55.3779710822, 1097: -19.0943319216, 1098: 42.2436553411, 1099: -8.28330608151, 1100: 4.03616040811, 1101: -4.30562951525, 1102: -28.506071485, 1103: 4.14697190482, 1104: 2.67184316398, 1105: -2.05582221551, 1106: 14.9767991358, 1107: 19.2466604343, 1108: 1.56491500407, 1109: 3.05800827104, 1110: -28.8552229365, 1111: -19.3296878553, 1112: 30.0509896929, 1113: 3.47045896729, 1114: -0.657856430725, 1115: 5.63503930265, 1116: 4.50533012465, 1117: 6.41037809684, 1118: 6.14175059054, 1119: -30.04745617, 1120: 4.97667768909, 1121: 21.8791221674, 1122: -9.03045562812, 1123: 0.678568336958, 1124: -0.016833698515, 1125: -17.423465021, 1126: 1.23118696796, 1127: 50.0237133758, 1128: 3.63152608684, 1129: -0.939561665031, 1130: -2.61328352494, 1131: 6.07616438907, 1132: -31.4525272946, 1133: -27.1383059878, 1134: 5.25723520999, 1135: 19.3938973732, 1136: 1.58667721403, 1137: -20.9986786167, 1138: 4.38293201807, 1139: 4.15147511147, 1140: -42.2731882714, 1141: 5.86238480501, 1142: 6.62527481257, 1143: 4.35143260986, 1144: -11.5794552731, 1145: 6.09864741118, 1146: 12.9586177943, 1147: 5.46470546777, 1148: -4.18872673309, 1149: 5.27678769077, 1150: -20.6916357354, 1151: 2.94432444983, 1152: -0.445113346936, 1153: -47.5854419305, 1154: 5.18061488548, 1155: 4.80429739505, 1156: 1.85181007595, 1157: -14.7961694597, 1158: 4.82973907104, 1159: 38.7857477298, 1160: -9.23507439205, 1161: 21.7149798287, 1162: 29.5472322722, 1163: 14.8618648347, 1164: 63.7288289674, 1165: -15.7257325712, 1166: -43.7000236456, 1167: -88.58729796, 1168: -6.37134670051, 1169: 9.4848735064, 1170: 16.7637425346, 1171: 18.3964951555, 1172: -5.66879532728, 1173: 0.888424844644, 1174: 1.15121461034, 1175: -0.136719542449, 1176: 24.3238710022, 1177: 0.928967239567, 1178: 4.57447265734, 1179: 5.58722334256, 1180: -21.4244381006, 1181: 1.14520697882, 1182: 0.62024983333, 1183: -9.25133781269, 1184: -1.26122714117, 1185: -39.0743241095, 1186: -0.898613443376, 1187: 1.78867452161, 1188: -49.8165678314, 1189: 11.8449479788, 1190: -4.15560073913, 1191: 2.40730702625, 1192: 3.03212905069, 1193: 0.932510378719, 1194: 0.959853547384, 1195: -36.6704182273, 1196: 1.11083604702, 1197: 5.31793549257, 1198: 13.2003973442, 1199: -0.010574317387, 1200: 1.07449616625, 1201: 1.2503841811, 1202: 1.03029542924, 1203: 5.68404004861, 1204: 3.46633532369, 1205: 8.54322369471, 1206: 1.62033313772, 1207: 0.552794821746, 1208: 3.43406199509, 1209: 3.13900409554, 1210: -51.6376988812, 1211: 15.0001631836, 1212: 58.2087208028, 1213: -67.4984159537, 1214: 6.71176238141, 1215: -5.03888749675, 1216: 28.8972445687, 1217: 1.32278284917, 1218: -0.8245476019, 1219: 1.10928526972, 1220: -19.4299123669, 1221: 1.00939994249, 1222: 6.22568456054, 1223: 38.6857687224, 1224: 19.9763112075, 1225: -1.33413762012, 1226: 0.89074986879, 1227: 1.19919321618, 1228: 0.88083656805, 1229: 1.5070670068, 1230: -5.22070126475, 1231: 21.6718881176, 1232: -8.57857159646, 1233: 21.3388885463, 1234: 66.5751769592, 1235: 11.3210700081, 1236: -14.6965086535, 1237: -20.6345628378, 1238: -4.33082046769, 1239: 1.64805606162, 1240: 1.15371353861, 1241: -28.2686525375, 1242: 16.1881850567, 1243: 1.34047096073, 1244: -14.5277042852, 1245: -22.4327941661, 1246: 0.996846286111, 1247: 5.22461286376, 1248: 1.09932566328, 1249: 35.5604647754, 1250: 0.998326544, 1251: 1.07766540167, 1252: -15.6149392504, 1253: 4.7635568403, 1254: -53.5221031513, 1255: 0.849507506657, 1256: 0.965790795538, 1257: 2.35536272252, 1258: 0.908976331989, 1259: -0.183265194947, 1260: -7.08674238281, 1261: 43.8823608073, 1262: 6.04223512915, 1263: -0.976656358418, 1264: -1.38076027624, 1265: 15.6918574371, 1266: 14.290712901, 1267: -22.8929085394, 1268: 8.53404473133, 1269: 20.2153238173, 1270: 19.9072183047, 1271: 4.24682580671, 1272: 11.5226015481, 1273: 1.21136539255, 1274: -23.9914961812, 1275: -1.13723984728, 1276: -6.05886542289, 1277: -9.85574740366, 1278: -30.8850952208, 1279: 1.75876398521, 1280: 7.93485993266, 1281: -21.4696371184, 1282: 4.17137197836, 1283: -15.1009422111, 1284: 1.01756774959, 1285: 21.1733166364, 1286: 3.98518387559, 1287: 4.76987431115, 1288: -0.749362727522, 1289: 8.29392398529, 1290: 11.4230764618, 1291: 11.5431057894, 1292: 1.70587672511, 1293: 32.9932134196, 1294: -7.44662170514, 1295: -27.0630164422, 1296: -79.483368685, 1297: 22.6663475191, 1298: -22.6595607349, 1299: 22.8264233748, 1300: -20.6036821822, 1301: 16.1313426319, 1302: 0.309665882946, 1303: 4.17784818699, 1304: 46.8636305021, 1305: 3.93349419088, 1306: 5.4433139136, 1307: 18.499469947, 1308: 48.8893555511, 1309: 5.30850593295, 1310: -1.42149904365, 1311: 4.01212336895, 1312: 5.7344501384, 1313: 0.790955989579, 1314: -12.7061157901, 1315: 22.6056073412, 1316: 16.7586432445, 1317: 2.31816875303, 1318: 4.37723355615, 1319: -12.7093645978, 1320: -0.554254043901, 1321: -5.9301946772, 1322: 2.04544082126, 1323: 21.6241089297, 1324: 9.53260608962, 1325: 1.00512613957, 1326: 0.0268729185471, 1327: 6.15815576497, 1328: -0.331617598073, 1329: 4.46732908001, 1330: 6.0186677744, 1331: 64.2691708633, 1332: -2.48457890898, 1333: 13.2671151359, 1334: 5.65754230257, 1335: 6.56240435853, 1336: 5.62101202482, 1337: 11.4330368734, 1338: -4.19686073211, 1339: 2.64447790845, 1340: -44.541485403, 1341: -13.8514978554, 1342: -5.07531995998, 1343: 43.628692213, 1344: -10.3417632145, 1345: -3.57794825381, 1346: 10.9950409681, 1347: 11.6672020305, 1348: 3.41657933709, 1349: 18.840767259, 1350: 12.6462226729, 1351: -61.3154802164, 1352: -26.2228675384, 1353: 6.48283164376, 1354: 3.12517184273, 1355: 5.18178942089, 1356: 5.01703386322, 1357: 6.26469736176, 1358: -23.2495148987, 1359: 4.15704300559, 1360: 3.92662495378, 1361: 41.2785140924, 1362: 4.67383910179, 1363: -31.6560614879, 1364: 5.6222012165, 1365: 5.18908388699, 1366: 40.8141380374, 1367: 4.58784764266, 1368: -4.95246096011, 1369: 33.8753612051, 1370: -59.4370551134, 1371: -42.3009024873, 1372: 14.3864369224, 1373: -23.7236253945, 1374: -32.1783775119, 1375: 43.2603783873, 1376: -69.2478973389, 1377: -11.0099328439, 1378: -38.6750878023, 1379: -31.1620943475, 1380: 3.35631205856, 1381: -59.5943777557, 1382: -6.06359524497, 1383: 80.2613589746, 1384: 0.0223667733339, 1385: 29.2053390075, 1386: 5.07558555197, 1387: 20.7684378832, 1388: 14.8178105523, 1389: -14.3170311778, 1390: 7.776562472, 1391: 79.7929767045, 1392: -2.05731560689, 1393: -100.0, 1394: -98.7382750452, 1395: -11.1743998707, 1396: 26.9743153918, 1397: 0.737956672111, 1398: 5.63099094719, 1399: 30.4841566753, 1400: 18.5108202465, 1401: 0.790951986882, 1402: -33.4004091466, 1403: 13.4121202348, 1404: -23.8248844117, 1405: 24.8275889749, 1406: 23.5809197258, 1407: 11.6948996028, 1408: -100.0, 1409: -34.9623173418, 1410: -10.3660862375, 1411: 22.6511339493, 1412: -14.0872396427, 1413: -12.6354049028, 1414: 42.8371504279, 1415: -16.7448772886, 1416: -32.4886504166, 1417: -8.23604586961, 1418: 19.9532723935, 1419: -22.221484532, 1420: -33.3011632522, 1421: -16.4421470581, 1422: 50.6365721616, 1423: 13.5487821689, 1424: 15.0701022486, 1425: -31.389745541, 1426: -13.3717049337, 1427: 16.19469505, 1428: 31.575233845, 1429: 21.9631515974, 1430: 5.93809824012, 1431: 9.50224851849, 1432: 12.6929801674, 1433: -18.842332875, 1434: 4.93592158025, 1435: 4.89530153213, 1436: -24.9675414472, 1437: 19.5118268057, 1438: 4.88237891417, 1439: 0.87508097768, 1440: 55.0890081287, 1441: 1.99193252037, 1442: -9.42413270614, 1443: 8.90584300965, 1444: 5.20352579475, 1445: -16.6042849278, 1446: -0.294897793227, 1447: -16.8790742213, 1448: 3.40857439676, 1449: 5.10039500779, 1450: 15.0674031292, 1451: 20.1386337405, 1452: 1.30996000419, 1453: -17.5318477322, 1454: 10.4806778714, 1455: 12.8980514143, 1456: 11.9991925098, 1457: -1.98169101846, 1458: -29.0651802735, 1459: 18.227597426, 1460: -7.5485630443, 1461: -27.8227915276, 1462: -8.76135511675, 1463: 1.31800115766, 1464: 0.397600586811, 1465: -5.0122139432, 1466: 5.54006371826, 1467: 5.24297479674, 1468: -17.9599367935, 1469: 5.21959310981, 1470: -2.75451003274, 1471: -8.20559196322, 1472: 4.39136371159, 1473: 0.0917389050256, 1474: -6.64184160975, 1475: 21.602108546, 1476: -37.9324897297, 1477: 4.42771828468, 1478: -0.766136841882, 1479: 5.18289602911, 1480: 0.966540553172, 1481: -33.5816895729, 1482: 0.151108446992, 1483: 4.10522443826, 1484: 5.98914978598, 1485: 2.58194523813, 1486: -15.3528012558, 1487: 5.56264526939, 1488: 4.95907685856, 1489: -6.91167803866, 1490: 5.18425427292, 1491: 1.08940270754, 1492: -5.59126737623, 1493: -0.13432830396, 1494: 5.30920177682, 1495: 5.20811401769, 1496: 5.46947029449, 1497: -19.9749937751, 1498: 5.47008974873, 1499: -15.3264801394, 1500: -9.57070690757, 1501: 9.24935517261, 1502: 4.39750044196, 1503: -27.7028728131, 1504: 5.13094088765, 1505: 2.84233199024, 1506: -8.71577261083, 1507: 5.53261429987, 1508: -16.3153768968, 1509: -51.4296757589, 1510: -0.211080193493, 1511: -20.8325786673, 1512: -17.0556996975, 1513: -28.4501674683, 1514: -6.65642915588, 1515: 30.5942151424, 1516: -3.67163710438, 1517: 17.9778283612, 1518: 28.5132678704, 1519: 22.1571296382, 1520: -4.92766284947, 1521: -10.9877767796, 1522: 1.25033358332, 1523: 0.998643384913, 1524: -5.57200311222, 1525: 10.0031767343, 1526: 0.997849328601, 1527: 5.16209518563, 1528: -1.33863896351, 1529: -16.4891627328, 1530: 3.0104924637, 1531: 2.79523614434, 1532: 51.3057780023, 1533: 3.07916884085, 1534: 40.1062664725, 1535: 2.02162112402, 1536: 0.238626645422, 1537: -38.7962240017, 1538: 21.9522782673, 1539: 31.6097667497, 1540: -3.04410235707, 1541: 14.6971410802, 1542: 1.24309262605, 1543: 0.886123083221, 1544: 11.682379815, 1545: 1.27830029721, 1546: 4.96349581861, 1547: -21.3357437694, 1548: 0.0994718288675, 1549: 0.857483936679, 1550: 0.463487412054, 1551: 1.13457844138, 1552: 2.09413745945, 1553: -10.2202987954, 1554: 1.7121961416, 1555: 1.2966244223, 1556: 0.792254815274, 1557: 3.35554796317, 1558: 6.3882334737, 1559: -25.8316845826, 1560: -1.00811372664, 1561: 15.5379880763, 1562: -82.7033086907, 1563: 4.20782541551, 1564: -21.6629167707, 1565: -29.3918749821, 1566: 10.1915063978, 1567: -18.7276686837, 1568: 0.970348927656, 1569: 34.9545658606, 1570: 0.866102984081, 1571: 7.3883465709, 1572: -11.4834917016, 1573: -5.3994110245, 1574: 2.29728533918, 1575: 0.946076662326, 1576: 1.42979278058, 1577: 1.06293555994, 1578: 1.24498106847, 1579: 0.353666975288, 1580: -52.1557532923, 1581: 5.45594103679, 1582: -9.32547028984, 1583: 10.6558056172, 1584: 15.4268050463, 1585: -0.417411186226, 1586: -10.9049387938, 1587: -13.7379235291, 1588: -4.28554870211, 1589: 0.766913600397, 1590: -13.8504035247, 1591: 1.75841744633, 1592: 8.93573545758, 1593: 7.70572148053, 1594: 9.1337352948, 1595: 0.950387285146, 1596: 1.51918709587, 1597: 1.12572822418, 1598: -4.06971249602, 1599: 1.050184457, 1600: 0.90758930019, 1601: 16.4580270139, 1602: 4.97740023958, 1603: -6.93525562944, 1604: 0.900369169779, 1605: 0.901070745128, 1606: 3.40271321323, 1607: 1.10203755851, 1608: 0.877220277796, 1609: 38.5353388851, 1610: -37.8832898753, 1611: 3.00746582, 1612: 2.7207940286, 1613: 8.96827150944, 1614: 6.60769435996, 1615: -34.5872233558, 1616: 59.9096583693, 1617: -1.62164060332, 1618: -14.9781183653, 1619: 0.476743281641, 1620: -0.35084836533, 1621: 12.7700455934, 1622: 1.45693982842, 1623: 23.4905743563, 1624: 13.1470179158, 1625: 5.5479922205, 1626: -3.83560989255, 1627: -25.6382799963, 1628: 1.06933876202, 1629: -61.208648509, 1630: 43.8581858618, 1631: -6.31788656655, 1632: 3.55572200798, 1633: 2.39156269014, 1634: -10.8951116809, 1635: 4.99988008703, 1636: 5.2597967676, 1637: -13.9538609664, 1638: -7.35887104881, 1639: -21.77193361, 1640: 6.68955433121, 1641: 15.8173288925, 1642: -56.7182381296, 1643: -26.014160935, 1644: -8.0626547413, 1645: 27.9239527432, 1646: -13.4930630725, 1647: 27.0613549551, 1648: -28.409808244, 1649: -33.1009251231, 1650: 10.2878996358, 1651: -4.19477221424, 1652: 3.48323519429, 1653: 0.308834691627, 1654: -5.79552931681, 1655: 5.0876236871, 1656: 1.1757532021, 1657: -4.7622943954, 1658: 4.8012711582, 1659: 39.1191160024, 1660: -9.37057674568, 1661: 2.00613407942, 1662: -24.2728067353, 1663: -4.81027194361, 1664: -4.28956062162, 1665: -7.99192101183, 1666: 3.35282546646, 1667: 4.99789033727, 1668: 4.28471365614, 1669: 9.52254801592, 1670: 9.9599413125, 1671: -8.23302997934, 1672: -14.9779173983, 1673: -55.7562120936, 1674: 2.33302888749, 1675: 1.83320430477, 1676: -24.4114772782, 1677: 4.5323605774, 1678: -15.5653994184, 1679: 21.8758229056, 1680: 73.1116287698, 1681: -18.8355343386, 1682: -19.3649917283, 1683: 5.26050865012, 1684: 1.62248553055, 1685: -0.130659918794, 1686: -3.4969803447, 1687: 0.0940709533616, 1688: -10.7312028945, 1689: 6.17770915346, 1690: 54.7761271294, 1691: 43.4194161157, 1692: -10.316059035, 1693: -6.94991822358, 1694: 37.063258776, 1695: -1.70778033596, 1696: -11.507060911, 1697: 15.5931102967, 1698: -16.9344257665, 1699: -2.23596436638, 1700: -23.8370197249, 1701: -20.8253407469, 1702: 1.18143637363, 1703: -2.07029013519, 1704: 1.54907082984, 1705: 1.47155959218, 1706: 2.4875601491, 1707: 26.7129966907, 1708: 5.00404689714, 1709: 4.69757865554, 1710: 0.748028605268, 1711: 5.05212834265, 1712: -5.65129921998, 1713: 5.02573682138, 1714: 5.12209918782, 1715: 24.4933787341, 1716: -2.82677342242, 1717: -1.37112011781, 1718: 34.7365218058, 1719: -41.3155924388, 1720: -86.2741374254, 1721: 27.7844819903, 1722: -12.6376336156, 1723: 10.2929372239, 1724: 36.5224775194, 1725: -27.4371895379, 1726: 4.92750784933, 1727: -0.855490602799, 1728: 50.1052606828, 1729: -1.81809010691, 1730: 7.33712170607, 1731: 11.6943886312, 1732: -40.6122210288, 1733: 71.2173524927, 1734: 81.5597182351, 1735: 5.18029605311, 1736: -27.6731958886, 1737: -35.8296822737, 1738: -1.18525962587, 1739: 38.6085524847, 1740: -6.93547003649, 1741: -40.9096235653, 1742: -28.4603519494, 1743: 17.2432316316, 1744: -50.6351035362, 1745: -19.5798320826, 1746: 0.708560980822, 1747: 41.6288509083, 1748: -34.6278199145, 1749: 14.5002022757, 1750: -40.8575643387, 1751: 25.0712839007, 1752: -56.2121728204, 1753: -7.51887396002, 1754: -40.7850495392, 1755: 17.1625830188, 1756: -19.6653609138, 1757: -21.5911785609, 1758: -19.0450200576, 1759: 20.3583447748, 1760: 0.420061583662, 1761: 27.2674316642, 1762: -58.0578430966, 1763: -7.37191216544, 1764: -61.2803344504, 1765: 5.52616415794, 1766: 3.60546811907, 1767: -18.4233276755, 1768: 63.5213576004, 1769: -18.1812130192, 1770: 36.4581351978, 1771: -44.5450441384, 1772: 64.5665472661, 1773: -6.89765725534, 1774: 20.9464853874, 1775: 26.6508846835, 1776: 21.6064558466, 1777: 23.4803946163, 1778: 9.80604238448, 1779: -6.73454804232, 1780: 14.5463257375, 1781: -6.18886582122, 1782: 19.4886928362, 1783: 51.7574114701, 1784: 6.0373682867, 1785: 0.0717095248458, 1786: -9.26634531886, 1787: -7.53850661202, 1788: -43.7625775465, 1789: -11.5341936581, 1790: -8.9942343732, 1791: 1.30592618431, 1792: -22.1200849916, 1793: 5.46406467376, 1794: 24.4253161752, 1795: -7.6219193606, 1796: -8.55139086796, 1797: -20.4672967062, 1798: 4.04112225949, 1799: 0.482558764208, 1800: -1.91475535897, 1801: 2.6580958797, 1802: 54.208763252, 1803: 16.3116216312, 1804: 54.0001868857, 1805: -27.925536783, 1806: 28.2054942883, 1807: 53.8738612805, 1808: -22.5771778689, 1809: -8.13959311784, 1810: -8.95451728694, 1811: 3.41238893035, 1812: 14.3030410988, 1813: 9.67653965178, 1814: -1.73649731253, 1815: 4.79190418999, 1816: 4.18265307153, 1817: -13.2856967599, 1818: 3.53025118574, 1819: 31.3766355629, 1820: 3.25246161328, 1821: 1.3370866129, 1822: 6.06284814664, 1823: 14.0257752038, 1824: -6.03889277773, 1825: 56.3527051211, 1826: -0.734311159128, 1827: -24.06775868, 1828: -12.733849609, 1829: 4.49473417976, 1830: -12.1194411744, 1831: 3.25156755244, 1832: -6.17946496932, 1833: -30.4771551026, 1834: -11.0270197582, 1835: -1.3911860636, 1836: -22.5345207033, 1837: 4.10786724541, 1838: -29.7852879715, 1839: 4.50745166834, 1840: 5.33056168802, 1841: 5.96005771413, 1842: 2.57438665699, 1843: 5.77110885898, 1844: -0.928729497729, 1845: 1.48764731974, 1846: 1.14138692341, 1847: 21.090586009, 1848: 37.2505105692, 1849: 3.52642395435, 1850: 3.51077255887, 1851: -54.0363688007, 1852: 3.08274614293, 1853: 3.16846673977, 1854: -9.04824798182, 1855: -30.2344221393, 1856: 2.06082349707, 1857: 10.2337618435, 1858: 15.2136949505, 1859: -24.724486909, 1860: 20.2154875653, 1861: 44.7463299694, 1862: -27.7418569208, 1863: 5.11213688034, 1864: 14.8359841191, 1865: 54.9894025453, 1866: -32.1557256625, 1867: -20.876854232, 1868: -28.9933957308, 1869: -27.7640370748, 1870: -34.0913449745, 1871: 0.667568187188, 1872: 1.93829622604, 1873: 1.56221557728, 1874: -56.8294937478, 1875: 0.791587281563, 1876: 3.59553827495, 1877: -5.12748143391, 1878: 35.8860261915, 1879: 1.68228401822, 1880: -3.882097936, 1881: -31.2688477609, 1882: -1.1846580401, 1883: 66.5209800739, 1884: 7.69838779401, 1885: -10.061468299, 1886: 15.6811921486, 1887: 1.99944406195, 1888: 0.0645321323523, 1889: 2.09447391133, 1890: -20.4520662176, 1891: 0.664960410704, 1892: 0.884550936155, 1893: 3.86849050637, 1894: 0.583026239152, 1895: 2.96098882928, 1896: -0.306982127444, 1897: 0.721192176203, 1898: 0.76235579392, 1899: 0.77059827871, 1900: 0.679274424743, 1901: 4.35695089276, 1902: -34.2733204739, 1903: 0.194738433454, 1904: 0.417944546947, 1905: 0.807475186724, 1906: 5.39410695683, 1907: -11.0313931966, 1908: -45.2704566354, 1909: 32.4177406659, 1910: 96.6141704329, 1911: -73.9189695042, 1912: 13.1994850971, 1913: 10.5491057592, 1914: 25.8875427127, 1915: -2.32740916856, 1916: 13.9310804416, 1917: 1.1493479356, 1918: -7.93127465774, 1919: 0.652761502781, 1920: -26.864126532, 1921: -17.9427754207, 1922: -92.300685123, 1923: 0.283626616196, 1924: 0.743416983212, 1925: 0.631001041187, 1926: -0.201380217724, 1927: 1.04190521304, 1928: -0.906183228025, 1929: -43.5099522023, 1930: -13.6739701221, 1931: 15.6405578112, 1932: -38.0468484781, 1933: -1.43474931848, 1934: -4.17036384739, 1935: -1.33944052441, 1936: 25.8092598252, 1937: -14.7069149014, 1938: 0.685351765347, 1939: 41.2840667808, 1940: 5.54949002594, 1941: -0.48959558652, 1942: -2.19243624663, 1943: -4.70545190301, 1944: 0.78610552626, 1945: 3.82761213093, 1946: 1.08985844684, 1947: -9.03763921132, 1948: 0.670923829403, 1949: 0.791604921167, 1950: -51.4700259403, 1951: 2.76852098519, 1952: 50.0904312781, 1953: 0.768822212128, 1954: 0.836661847308, 1955: 1.88801279917, 1956: 0.655073874492, 1957: 8.09209991805, 1958: 33.738563239, 1959: -0.397627330217, 1960: 1.45576149835, 1961: -1.67254832672, 1962: 49.3781877743, 1963: -4.89110319227, 1964: -8.81707643026, 1965: -22.7881043444, 1966: -2.79681821189, 1967: 6.98130773513, 1968: 18.2710341464, 1969: 0.348217446678, 1970: -5.3921636114, 1971: 0.838265711011, 1972: -35.3085457551, 1973: 3.27836162185, 1974: 8.42757561572, 1975: 7.38731710765, 1976: 0.848655056179, 1977: -0.0697419577439, 1978: 1.63704559906, 1979: -39.5565138225, 1980: 2.4613643323, 1981: -2.51166745417, 1982: -27.5688430204, 1983: -49.8620555254, 1984: 0.914920602192, 1985: 3.58621169905, 1986: 38.3499998024, 1987: -29.8131142597, 1988: 6.56794654593, 1989: -13.7538083337, 1990: 43.5022500555, 1991: -16.10600143, 1992: -24.7594323053, 1993: 11.6498784899, 1994: 11.8698603479, 1995: 54.4075964102, 1996: -3.80721462221, 1997: 35.1822981654, 1998: -30.6538058338, 1999: -17.7771639968, 2000: 2.39930663817, 2001: 1.07560748051, 2002: -42.1080363904, 2003: 4.40664749152, 2004: 2.99248328187, 2005: 30.0138463577, 2006: -25.7713510496, 2007: 4.67915332639, 2008: -1.41907252872, 2009: 3.55446347653, 2010: 4.26979790223, 2011: -45.6456267986, 2012: 11.037534893, 2013: 7.54504454277, 2014: 1.14714986333, 2015: -3.47907374452, 2016: 11.7760961267, 2017: 68.7031694598, 2018: -16.764311483, 2019: -67.9071256576, 2020: -1.84615938138, 2021: -65.9878240731, 2022: -8.65665781269, 2023: -25.9975290283, 2024: -6.2307941477, 2025: 10.3709957061, 2026: -38.8297039698, 2027: 18.9789382096, 2028: 12.9890838162, 2029: 39.2205089932, 2030: 0.792818371736, 2031: 48.8041575917, 2032: 4.64874547713, 2033: -0.80725178549, 2034: 13.378222439, 2035: -18.5231435237, 2036: -16.1271308731, 2037: 13.2645960409, 2038: 14.4073769237, 2039: -45.2832614938, 2040: 49.3545727597, 2041: -35.6032344318, 2042: -37.6271843568, 2043: -26.133974277, 2044: 25.1920570484, 2045: -31.6089582818, 2046: 12.548580617, 2047: -12.5082051621, 2048: -4.22071078228, 2049: 25.2975619192, 2050: -5.78094742639, 2051: 5.45078610491, 2052: -4.88888915009, 2053: 3.71892716645, 2054: 3.70479778262, 2055: 1.03953269173, 2056: 42.0765065775, 2057: 3.52625980292, 2058: 3.41507538618, 2059: 69.5273002708, 2060: 2.82861428123, 2061: -11.9906482131, 2062: 4.38920770182, 2063: 4.11466855281, 2064: -28.2726194818, 2065: 3.47347634608, 2066: 11.0358947798, 2067: 37.4296769568, 2068: -6.4610203799, 2069: -41.5477897316, 2070: 27.3763975207, 2071: -16.8192708992, 2072: -10.2445661528, 2073: -29.3894430829, 2074: -38.559180687, 2075: -3.8098266561, 2076: -14.2953188338, 2077: 0.542135369511, 2078: -10.7783533214, 2079: -35.8687858143, 2080: 30.6210671254, 2081: 36.6002171719, 2082: 28.6884990979, 2083: -27.0059624428, 2084: 2.87351282078, 2085: -12.381947425, 2086: -6.24445321809, 2087: 21.6059657456, 2088: 16.7116772226, 2089: 2.76542337257, 2090: -25.1275060041, 2091: -73.0192476795, 2092: -18.1126648752, 2093: -35.8805736388, 2094: -10.6157553831, 2095: -0.160522134942, 2096: 100.0, 2097: 4.88771760525, 2098: -32.4327842763, 2099: 2.43023748665, 2100: -9.5485765543, 2101: -1.06388904309, 2102: -46.5363307745, 2103: -2.30532756234, 2104: -32.6592592941, 2105: 23.9723305576, 2106: -13.9615897328, 2107: 9.04489041595, 2108: -42.3817728465, 2109: -25.0049185643, 2110: -1.85345107772, 2111: 0.954859490904, 2112: -4.2005891959, 2113: -36.6649971648, 2114: -100.0, 2115: -1.31505171738, 2116: 32.6318633618, 2117: 2.13901418421, 2118: 2.93386919676, 2119: -60.5185349472, 2120: -2.68464570924, 2121: -34.2991150127, 2122: 3.23420554183, 2123: -8.45332348504, 2124: 21.7474611788, 2125: 2.39077626316, 2126: -8.42099090891, 2127: -43.8221516855, 2128: 8.98440414031, 2129: 37.9325204844, 2130: 7.65388945055, 2131: 12.4462643572, 2132: -47.4000122581, 2133: -13.2333937685, 2134: 3.05041820002, 2135: -26.8653055636, 2136: -1.53640374886, 2137: 9.44341446551, 2138: 20.2055067813, 2139: -0.813653570433, 2140: -10.6883259099, 2141: 50.538444287, 2142: 0.966447634692, 2143: 7.16412917417, 2144: -5.24773593775, 2145: 6.22445031364, 2146: -7.28488095629, 2147: -0.511296806223, 2148: -6.72418996689, 2149: -4.37030502858, 2150: -21.1996225269, 2151: 1.72318523964, 2152: -7.25226677209, 2153: 33.3003716495, 2154: -24.692158367, 2155: -61.1769220603, 2156: -76.1464823259, 2157: -16.5923285108, 2158: 71.9978578133, 2159: -45.9771494445, 2160: -20.871447917, 2161: -11.4297543423, 2162: 0.755873571871, 2163: 9.73691424003, 2164: 0.655414742368, 2165: 0.925775038287, 2166: 13.9297515043, 2167: -0.760977660445, 2168: -19.5893148039, 2169: 5.200781535, 2170: -6.96745932882, 2171: 4.44396974381, 2172: -34.0209289703, 2173: 79.5686393465, 2174: -77.7774174543, 2175: -24.0645055469, 2176: -13.2305392004, 2177: -0.105062559224, 2178: 4.00812802895, 2179: -4.4159689822, 2180: -43.3128975263, 2181: -10.0349970575, 2182: 18.9560722666, 2183: -0.918775893242, 2184: -3.76520716888, 2185: -3.68401858194, 2186: -0.633252366124, 2187: -16.9895309114, 2188: -0.0741819045627, 2189: 1.33624929377, 2190: -0.242117398023, 2191: 19.7286046173, 2192: 1.38875295197, 2193: -6.62657235518, 2194: 0.645145813235, 2195: 9.21758739403, 2196: -0.104513418776, 2197: -25.4276306702, 2198: -21.3598048116, 2199: -1.34470089617, 2200: -41.7345159444, 2201: -0.483235511849, 2202: -1.58444916339, 2203: -0.722919554147, 2204: -9.27485146962, 2205: -2.03715765295, 2206: 12.2771018168, 2207: -5.92402225133, 2208: 100.0, 2209: 98.7940374999, 2210: -39.3242432618, 2211: 26.3503542571, 2212: 17.2197670207, 2213: 3.24564933323, 2214: -46.9462502755, 2215: -27.2891922375, 2216: 28.3256802191, 2217: 12.1220951665, 2218: 6.2155798042, 2219: 17.0067739336, 2220: -0.175745457854, 2221: 0.0241856107103, 2222: -3.94183899256, 2223: 18.2695281605, 2224: -0.267293450163, 2225: -0.370517840915, 2226: 3.21998475739, 2227: 16.4780568003, 2228: 2.36545032557, 2229: -4.5886874606, 2230: 45.7890485874, 2231: 3.36244149666, 2232: 32.168027208, 2233: 0.569718090734, 2234: 18.0409232561, 2235: 38.5986728505, 2236: 27.9758000241, 2237: -49.1951376788, 2238: 0.615172217388, 2239: -27.9247179003, 2240: -0.238725896792, 2241: -0.136582909289, 2242: 18.5284392391, 2243: -0.120102925224, 2244: -0.672812550599, 2245: 2.84305187147, 2246: 1.67862667316, 2247: -0.232048787487, 2248: 0.621798273269, 2249: -0.0742398107347, 2250: 7.89167004407, 2251: -47.0021115802, 2252: -1.4742158536, 2253: -0.835084699532, 2254: -1.06984606443, 2255: 7.36102316763, 2256: 6.46866084772, 2257: 15.7532125912, 2258: 9.99799795207, 2259: 25.1155906373, 2260: -20.9506459338, 2261: 33.5398661884, 2262: -5.55226083425, 2263: 7.03948772116, 2264: 35.5801905304, 2265: -29.0922987007, 2266: -0.642041351353, 2267: 9.58611820362, 2268: -0.21795623551, 2269: 15.3110667916, 2270: 40.7964646024, 2271: -30.4559841831, 2272: 4.09662955771, 2273: -0.096130944451, 2274: -0.551197131059, 2275: -0.427873960958, 2276: 0.378338860746, 2277: 3.85982170424, 2278: 62.2287776621, 2279: 14.5211429689, 2280: -18.8241760642, 2281: 9.02938247014, 2282: 4.90079418357, 2283: -3.35702488165, 2284: -5.5445598406, 2285: -38.9698830592, 2286: 14.0420588072, 2287: -0.327591245305, 2288: 35.7500595203, 2289: -2.14188556561, 2290: 0.736599101643, 2291: 5.12162610183, 2292: -6.67213540791, 2293: -0.287612776092, 2294: -0.758078462995, 2295: -0.180184697775, 2296: -16.2991834406, 2297: -0.152142312604, 2298: -0.0240686304528, 2299: 28.432047393, 2300: -0.900637595758, 2301: 51.7805963425, 2302: -0.242642410344, 2303: -0.257365647294, 2304: 2.6214006175, 2305: -0.173992690944, 2306: -1.92276597225, 2307: 13.7075636574, 2308: 82.492707454, 2309: 1.5376180455, 2310: -1.77978281233, 2311: 49.2268687307, 2312: 1.57864329887, 2313: -15.7937776075, 2314: 20.8729370091, 2315: -11.3877677988, 2316: 5.93063915108, 2317: 2.42275000539, 2318: -9.83068984694, 2319: 5.11136881183, 2320: -0.682380273391, 2321: -26.4739690706, 2322: -29.9298234533, 2323: -8.68418414206, 2324: 9.01376823367, 2325: -33.4896529954, 2326: 0.563573481274, 2327: 70.6775496287, 2328: 11.4869662063, 2329: -2.47814135139, 2330: -4.91762942713, 2331: 5.51004375433, 2332: -56.1591806449, 2333: -0.672789509193, 2334: -0.349910284835, 2335: -53.2734450627, 2336: -64.1863381149, 2337: -47.6698478578, 2338: -39.3132201368, 2339: -65.6688761292, 2340: -8.5537920223, 2341: -0.659440909227, 2342: 37.6043817948, 2343: -6.56196982604, 2344: 44.6957295177, 2345: -16.0243263652, 2346: -1.94053390308, 2347: -13.996923427, 2348: 3.18928795521, 2349: -5.84284354716, 2350: -0.610075443751, 2351: 13.1908921484, 2352: -0.465107314792, 2353: -0.65075731938, 2354: 19.3719399918, 2355: -65.8408454657, 2356: -0.790321771323, 2357: -18.6814118926, 2358: -3.65229112823, 2359: 8.15331176875, 2360: 11.6667904931, 2361: 25.7155485524, 2362: 17.6074650131, 2363: 25.7215748441, 2364: 7.24605204716, 2365: 6.53688338763, 2366: -10.2977450717, 2367: 7.61839505352, 2368: 55.9837597788, 2369: 21.5416476191, 2370: 22.5885009493, 2371: -17.7861933391, 2372: -16.2740866409, 2373: 70.2283775955, 2374: 19.8779748011, 2375: 13.793653747, 2376: 14.6177598815, 2377: -29.7120998825, 2378: -35.4178599428, 2379: 88.9984009173, 2380: -48.8444450724, 2381: -0.396042784125, 2382: 32.7247124418, 2383: 0.883281609256, 2384: -26.9556038679, 2385: 5.69992012942, 2386: 36.1267584449, 2387: -26.8295687686, 2388: 12.8262774709, 2389: -14.857823052, 2390: -43.5105584973, 2391: 23.6260201432, 2392: 2.91454128562, 2393: 6.29812691246, 2394: -2.40621465882, 2395: 8.79488609528, 2396: -21.536994204, 2397: -19.7588186779, 2398: 24.2003071705, 2399: 16.2769533898, 2400: 1.20535151196, 2401: -55.940575969, 2402: -0.570998640935, 2403: -0.689701728476, 2404: 0.261508374861, 2405: -0.73365912052, 2406: -0.599876337192, 2407: -1.10784752873, 2408: -44.9153190866, 2409: -0.85867705328, 2410: -34.6346671229, 2411: -0.278818906443, 2412: -0.485246541931, 2413: 25.9037138175, 2414: -1.4949385943, 2415: -3.64325592181, 2416: -20.1732744323, 2417: 2.15825521845, 2418: -52.5844207792, 2419: -28.2351642746, 2420: -56.5395726722, 2421: -26.5065371326, 2422: 55.6992533208, 2423: 27.3827083778, 2424: -19.8490559559, 2425: -70.4076297807, 2426: 20.437573372, 2427: -23.1912322153, 2428: 5.74886557237, 2429: 11.6566547621, 2430: -38.861723996, 2431: -19.7401237206, 2432: 35.7909668047, 2433: -0.617630369801, 2434: 5.93784568548, 2435: 11.9840727088, 2436: -4.08709895275, 2437: -44.8849896897, 2438: 48.2541449392, 2439: 18.9198082027, 2440: -81.9746382797, 2441: -0.991949684291, 2442: 21.6263854234, 2443: -2.93950342957, 2444: -0.26650913574, 2445: 24.7856759412, 2446: 49.471225316, 2447: -23.1194528675, 2448: 12.3328130763, 2449: 0.32110517525, 2450: -23.4095277066, 2451: 28.9482316479, 2452: 14.6355223653, 2453: -29.4020493034, 2454: -62.6752897504, 2455: -89.6101648355, 2456: 0.197487491961, 2457: -13.1975410562, 2458: -37.4987982138, 2459: 7.32771638429, 2460: -10.8918471292, 2461: -63.7420932213, 2462: -60.7763395299, 2463: 4.7109731937, 2464: 27.1776348436, 2465: 74.8646325964, 2466: 58.0803518416, 2467: -61.0791240522, 2468: -5.16288042344, 2469: -75.4525191721, 2470: 11.1726210175, 2471: -48.0538882233, 2472: 60.7296384619, 2473: 15.322369272, 2474: -71.232219802, 2475: 3.17654827444, 2476: -24.400116428, 2477: 19.8471412611, 2478: 55.7020387632, 2479: -34.1392362005, 2480: 33.0384628277, 2481: 85.6938418491, 2482: -61.3239779026, 2483: -9.70098001522, 2484: -78.8611446381, 2485: -1.85855418928, 2486: 35.3573500963, 2487: 18.9767815058, 2488: -24.0547795191, 2489: 22.6646659336, 2490: -16.9531328744, 2491: 36.7958642747, 2492: 66.4986324204, 2493: -7.35827399604, 2494: 21.6662176381, 2495: -6.82937030471, 2496: -1.58019358447, 2497: -27.9886060767, 2498: 14.6511713656, 2499: -3.99810188818, 2500: -93.7710890689, 2501: -21.1756408495, 2502: 36.6955481811, 2503: 99.7420349484, 2504: 1.91532746269, 2505: 51.6352537328, 2506: -16.1249531433, 2507: 24.9381064807, 2508: -24.1564288776, 2509: 19.0077877527, 2510: 17.3236345175, 2511: -1.12322217773, 2512: -20.2113201027, 2513: -1.37436449444, 2514: -1.10566892584, 2515: -0.474036916131, 2516: -3.50572483186, 2517: -2.16397957391, 2518: -6.08436648835, 2519: -2.36341639331, 2520: -7.14656023314, 2521: -9.33043236757, 2522: 100.0, 2523: 91.1725118698, 2524: -2.84379972071, 2525: -99.788168865, 2526: -1.41869123587, 2527: 0.18315795376, 2528: -12.8300535844, 2529: -2.58469816013, 2530: 35.9845315513, 2531: -80.4805557703, 2532: -1.62673314036, 2533: -6.75194271427, 2534: 51.6082049703, 2535: -1.67476065417, 2536: -26.7594330934, 2537: -1.33187862679, 2538: -0.911344982313, 2539: -1.5106580107, 2540: 9.04302044828, 2541: -1.07699092943, 2542: -1.30394621137, 2543: -0.835537172218, 2544: -91.1731837818, 2545: -19.3663060226, 2546: -14.7202386607, 2547: -4.64213524897, 2548: -1.80606826987, 2549: -16.1289506031, 2550: -1.51721968129, 2551: -1.83998412169, 2552: -1.50843670106, 2553: -8.20058051385, 2554: -1.37719699246, 2555: 41.5368083558, 2556: -99.6916360811, 2557: 82.3192649715, 2558: -36.8498006635, 2559: 72.1943334694, 2560: -82.5598275102, 2561: -7.72558116256, 2562: 60.4212325987, 2563: 0.658761998594, 2564: -54.8513060055, 2565: 72.7279035761, 2566: -3.59123153822, 2567: -4.1714585864, 2568: -23.5255318632, 2569: 0.116474611084, 2570: -1.85110231135, 2571: -9.09468273726, 2572: -27.5336611008, 2573: 0.0767358718791, 2574: -1.33387167681, 2575: -0.3952764462, 2576: 1.37743106051, 2577: 1.34412021571, 2578: 26.879112283, 2579: -55.1816002743, 2580: 2.77212118067, 2581: 10.5618190299, 2582: 0.341892094399, 2583: -17.1424781693, 2584: 10.3497566928, 2585: 56.8956916149, 2586: -36.0710571812, 2587: 1.6573575436, 2588: -17.9167403074, 2589: -1.03338088816, 2590: -0.162634599482, 2591: 37.0355951438, 2592: -0.242184244798, 2593: -1.51808164271, 2594: 55.4181503445, 2595: 0.550282075744, 2596: -0.24587691599, 2597: 0.103014480541, 2598: -0.344536297206, 2599: -1.40918821683, 2600: 99.6662329783, 2601: 0.355672854838, 2602: 0.957952074773, 2603: -0.655863613249, 2604: -4.31956565938, 2605: -5.91084836441, 2606: -43.277887275, 2607: -4.12746186538, 2608: -28.200162826, 2609: 59.9651014694, 2610: 25.1535635782, 2611: 29.3650088092, 2612: 18.457103797, 2613: 23.3443733726, 2614: -7.13171300227, 2615: -0.231052264341, 2616: -10.1542118854, 2617: -0.263645490358, 2618: -5.75355468218, 2619: 5.32613036393, 2620: -39.8712595174, 2621: 0.645620665457, 2622: -0.0382597600031, 2623: -0.639011388793, 2624: -0.336984347219, 2625: -0.195672472883, 2626: -6.84847551167, 2627: 25.2618586919, 2628: 25.256813714, 2629: -11.7943963028, 2630: -19.9231095983, 2631: 19.42930635, 2632: -19.5883637509, 2633: 17.8991271768, 2634: -23.6222187749, 2635: -17.9066893614, 2636: -0.330902915237, 2637: 71.2621811077, 2638: 0.526417509908, 2639: 0.118283498753, 2640: 1.1038491861, 2641: 10.3455573973, 2642: -0.255419617977, 2643: -1.67439055807, 2644: -0.379411979083, 2645: 54.3810290863, 2646: -0.220794664299, 2647: -0.276968936592, 2648: 13.1728870773, 2649: -1.4990937413, 2650: -22.7666601399, 2651: -0.36371961958, 2652: -0.165242939973, 2653: 0.604321583882, 2654: -0.193259711328, 2655: -11.3580322465, 2656: 46.970022933, 2657: -24.3200461834, 2658: -0.616956335772, 2659: -0.115594233669, 2660: 33.4050036704, 2661: -6.52854200507, 2662: 9.33027068619, 2663: 30.1759626171, 2664: -12.0492372138, 2665: -18.0764020212, 2666: -24.1127186786, 2667: -2.35712834287, 2668: -15.0739333175, 2669: -0.451580491822, 2670: 22.2912573246, 2671: -9.45845240148, 2672: 0.64732526529, 2673: -7.1395075045, 2674: -50.7021528784, 2675: -0.261275673645, 2676: 29.4523097092, 2677: -36.0665464094, 2678: -16.9358610271, 2679: -47.9634872094, 2680: -46.1026452897, 2681: -0.376096342436, 2682: -8.44361684717, 2683: -1.33643233402, 2684: -39.0827390404, 2685: 17.2280444413, 2686: -19.5234391783, 2687: 31.3340861528, 2688: 71.6315669523, 2689: -70.1574715374, 2690: 0.477587362087, 2691: -25.0790123129, 2692: -57.6197010467, 2693: 1.60733007481, 2694: 9.35568568047, 2695: -2.8547601693, 2696: -4.72061265684, 2697: -6.98369356683, 2698: 19.7285783662, 2699: -7.70014387006, 2700: -24.1908423234, 2701: -2.18544558147, 2702: -1.57002819143, 2703: 53.0838598841, 2704: -82.2788238111, 2705: 0.522563961615, 2706: 46.3184754605, 2707: -2.85063818064, 2708: -1.4878220316, 2709: -96.5112975735, 2710: 100.0, 2711: -30.1108316947, 2712: -0.954527095245, 2713: -7.12302784021, 2714: -4.59456618115, 2715: 0.514202850803, 2716: -98.5272665934, 2717: -36.1207176524, 2718: 93.2879868355, 2719: -12.315068316, 2720: 39.1236823295, 2721: 28.61504822, 2722: 10.8782590483, 2723: -34.4487148275, 2724: -5.7350677297, 2725: 99.9951237203, 2726: -6.34680720208, 2727: 11.3875077267, 2728: -33.2567465431, 2729: -31.2267375356, 2730: -7.87300538121, 2731: -0.181477244633, 2732: -3.42016472967, 2733: -6.6697241001, 2734: -2.55032703745, 2735: 33.5107474658, 2736: -30.6712320251, 2737: -17.1618989777, 2738: -26.2866899557, 2739: -11.365636604, 2740: -1.48272129101, 2741: -6.51806609784, 2742: -6.19478685796, 2743: 2.21787740638, 2744: -1.01317697043, 2745: -42.0559359054, 2746: 0.104549318263, 2747: -0.890979501757, 2748: 29.2822954539, 2749: -0.941710118671, 2750: -96.1670227464, 2751: -1.68170416068, 2752: -1.51655657866, 2753: -1.40699358894, 2754: 43.5688214294, 2755: -1.45216768028, 2756: -1.92803710626, 2757: -47.9118266925, 2758: -1.68786622208, 2759: -25.3315671353, 2760: -1.63572355215, 2761: -1.52683344349, 2762: 65.74079325, 2763: -1.56242004057, 2764: -18.6314859796, 2765: 37.9041558255, 2766: -6.76203640731, 2767: 6.2307844608, 2768: -1.14811279064, 2769: 8.82466495534, 2770: -5.03010752892, 2771: 100.0, 2772: -28.2668992399, 2773: -6.33745062215, 2774: -34.0008720466, 2775: -11.138011677, 2776: -3.86312045516, 2777: -6.67017065616, 2778: 5.50169718988, 2779: 45.832661489, 2780: -8.10928184191, 2781: -6.57355430407, 2782: -1.62366412416, 2783: 91.9436143162, 2784: 23.103271658, 2785: -3.5297576546, 2786: -62.6214093411, 2787: -24.3360372601, 2788: -49.8618281183, 2789: 2.56755117916, 2790: 28.7579892801, 2791: -50.9878533453, 2792: -46.1670503258, 2793: -0.478135901044, 2794: -89.55363649, 2795: 23.5949235328, 2796: 9.11559135853, 2797: -63.1055381378, 2798: 29.2067195016, 2799: 29.7780878706, 2800: -15.6102736313, 2801: -2.15968978368, 2802: -15.7005880673, 2803: 27.6568687912, 2804: -0.970037455944, 2805: 4.34475559068, 2806: -6.31320991781, 2807: -0.420613367092, 2808: -0.65747725978, 2809: 47.5768234539, 2810: -25.0616252011, 2811: 8.52223530093, 2812: 10.1912234917, 2813: 0.191245326008, 2814: 10.4065060352, 2815: -40.5908087752, 2816: 10.8328080597, 2817: 14.4590968591, 2818: 28.9490799894, 2819: 6.37827660003, 2820: 12.9372042462, 2821: 5.69522040541, 2822: -29.6631674948, 2823: -36.3557662947, 2824: -0.449173813899, 2825: 17.9935017094, 2826: -7.93166897408, 2827: -5.43474715624, 2828: -26.1526490536, 2829: 19.0602924892, 2830: -36.986539052, 2831: -64.0078293316, 2832: 9.48183344372, 2833: 1.22366861062, 2834: -1.951741022, 2835: -26.2358742291, 2836: 0.906140484596, 2837: -8.07612827521, 2838: 12.1160005933, 2839: 15.6719241676, 2840: 13.4727434554, 2841: 5.30146652805, 2842: 11.1637032441, 2843: 0.161058805734, 2844: -46.804909517, 2845: -2.86848168887, 2846: 8.01148640355, 2847: -4.80634755505, 2848: -3.23741714176, 2849: -10.009103503, 2850: -8.68551330001, 2851: 3.90469004653, 2852: -21.5855012908, 2853: -45.5658027575, 2854: -8.81230621963, 2855: -75.9444554877, 2856: 18.7688812795, 2857: -1.61366549537, 2858: -1.10012892711, 2859: 8.94636721779, 2860: -2.22540267239, 2861: 7.64023019326, 2862: 0.352394692774, 2863: -1.73054080267, 2864: 6.6918995346, 2865: -2.94227296337, 2866: 0.0303904523291, 2867: -15.5177992164, 2868: 6.34113489877, 2869: 6.54609288811, 2870: 7.28504258883, 2871: -65.5545487099, 2872: -32.6901966129, 2873: -0.910820127843, 2874: 14.6908865749, 2875: -1.90919565059, 2876: 27.8892251648, 2877: 30.5104919045, 2878: 21.3752777437, 2879: 2.92506354295, 2880: -21.710746777, 2881: -3.08590967139, 2882: -0.674061552197, 2883: -2.40081088964, 2884: -2.84164227481, 2885: -4.86860616104, 2886: -2.56328751725, 2887: -2.00773417283, 2888: -1.05480177327, 2889: 0.428744871803, 2890: -1.64378688431, 2891: -1.69093647353, 2892: 10.7620813799, 2893: 14.5272446639, 2894: -5.10798193481, 2895: -2.27794532869, 2896: -1.47858912095, 2897: 11.4107953663, 2898: -0.272809154092, 2899: -3.21093162321, 2900: -2.91174621764, 2901: -2.37135037994, 2902: -2.5576451478, 2903: 0.178839509993, 2904: 3.82403758475, 2905: 6.8594573819, 2906: 12.7052800875, 2907: -37.0579991853, 2908: -14.8130175803, 2909: -9.92699894727, 2910: -5.77341383165, 2911: 43.0908423168, 2912: -14.6250059841, 2913: 14.8856935823, 2914: -33.9329760551, 2915: 2.85452461064, 2916: 12.2142751546, 2917: 4.53931004311, 2918: -0.581452620482, 2919: 0.0424622002541, 2920: -35.5345892969, 2921: -8.07970435835, 2922: -0.557729640321, 2923: -1.29148617565, 2924: -0.31958570026, 2925: -17.9931160273, 2926: -2.12494463953, 2927: -0.367252646296, 2928: -15.146828951, 2929: -3.25189715042, 2930: -3.33171287834, 2931: 0.375519524947, 2932: -0.679639673066, 2933: 7.15102789414, 2934: -48.0342402462, 2935: -39.6781733473, 2936: 0.97659282588, 2937: 18.8730133387, 2938: -0.980925193485, 2939: -0.4946031587, 2940: 37.328291539, 2941: -0.439896531495, 2942: -2.81192732264, 2943: -3.20041534254, 2944: -0.367379627766, 2945: -0.408882828802, 2946: -0.603359827927, 2947: -0.59407776605, 2948: -1.89262334188, 2949: -3.7227847278, 2950: -0.409737601351, 2951: 0.24808347692, 2952: -0.587659601602, 2953: -2.57284485238, 2954: -3.14241409877, 2955: -18.4684942016, 2956: -11.329714785, 2957: -53.142940415, 2958: 0.0796776227064, 2959: 14.8664232542, 2960: 7.50683450263, 2961: 9.31991076798, 2962: -2.15109671934, 2963: -5.46611504136, 2964: -0.585595261358, 2965: -18.470011491, 2966: -0.615752573269, 2967: -7.7762137523, 2968: -12.7236156337, 2969: 1.26232464935, 2970: -1.10268426893, 2971: -0.529312830036, 2972: -0.142645902119, 2973: -4.16636176872, 2974: -1.25668437352, 2975: -10.865575717, 2976: 1.88675243455, 2977: 15.0563261786, 2978: 34.4535764316, 2979: -14.4626826952, 2980: 4.18566569001, 2981: 20.5440686188, 2982: -1.83710291188, 2983: -2.81952952833, 2984: -6.09834468335, 2985: -0.392852236392, 2986: -59.6890867165, 2987: -11.0984643348, 2988: -1.12635890804, 2989: -0.449169055735, 2990: -13.7268830889, 2991: -0.405381944146, 2992: -2.88673008032, 2993: -0.636094139365, 2994: -36.3468562511, 2995: -0.483333539826, 2996: -0.495397439619, 2997: 18.4841085114, 2998: -3.12792335123, 2999: 12.4170927732, 3000: -0.466752169479, 3001: -0.512239456809, 3002: 3.61586380905, 3003: -0.227807214974, 3004: -1.63937686688, 3005: -20.1609495356, 3006: 61.1060812512, 3007: 1.13922041367, 3008: -2.2192353158, 3009: -1.31074354731, 3010: 0.17235604526, 3011: -52.2990964318, 3012: -14.9952128249, 3013: -4.13282267239, 3014: 15.965714345, 3015: -0.979397740251, 3016: -2.08360875933, 3017: 11.1596618996, 3018: -0.631980216776, 3019: 21.9018152616, 3020: -8.32144265136, 3021: 1.76850533976, 3022: 3.93208571078, 3023: -8.27069889669, 3024: -2.1136263006, 3025: -11.2172239926, 3026: -1.48814680824, 3027: 3.4144546896, 3028: -12.5144781198, 3029: 31.9327552653, 3030: -7.49168074875, 3031: -3.3472143318, 3032: -1.1393679864, 3033: -15.8274443644, 3034: 5.08365850766, 3035: -1.24515231196, 3036: -12.7483704498, 3037: 46.8309687373, 3038: -8.82074880773, 3039: 5.60032420312, 3040: 44.5922412806, 3041: -0.680280630701, 3042: -17.5044161733, 3043: 38.9766182945, 3044: 12.7608015669, 3045: -11.2671678086, 3046: 24.9468499485, 3047: 12.6839153184, 3048: -2.30207679573, 3049: -13.2462467691, 3050: -1.31299720924, 3051: -2.86195984632, 3052: -3.27499560798, 3053: -1.95195275878, 3054: -1.20983741118, 3055: -2.69977156874, 3056: 7.28502514398, 3057: -1.84519436375, 3058: -3.96648488651, 3059: 5.04688443919, 3060: -1.73079370417, 3061: 5.23603865154, 3062: -3.01175475087, 3063: -2.27614099795, 3064: -30.196141222, 3065: -13.7468666481, 3066: -21.2606415125, 3067: 3.55312348166, 3068: -23.9124610094, 3069: 13.3396313879, 3070: 87.2082626212, 3071: -2.1493484882, 3072: -15.3866067391, 3073: -2.7111440832, 3074: -20.7360251984, 3075: -9.91936183635, 3076: -4.39586447569, 3077: -17.1957569382, 3078: 4.21279945161, 3079: -2.00708868784, 3080: -9.3580834522, 3081: 37.4162860887, 3082: -5.19525049505, 3083: -7.08801576935, 3084: -1.18838734312, 3085: 50.6134769698, 3086: -10.0746164599, 3087: -51.6101469738, 3088: 35.4087481022, 3089: -3.34627846772, 3090: 9.26921716533, 3091: -1.99907346852, 3092: -11.8100345498, 3093: -2.73756081725, 3094: 44.4554254782, 3095: 8.96765822964, 3096: 13.6276616634, 3097: 27.5503371214, 3098: -1.82587676846, 3099: 48.8796313024, 3100: -3.05067935006, 3101: -2.9476865687, 3102: 4.29722919336, 3103: -21.2288384954, 3104: -2.64454137433, 3105: -3.15677471753, 3106: -46.1941440126, 3107: -3.02688150691, 3108: -10.4658968884, 3109: -2.838498659, 3110: -2.81308371667, 3111: 7.33024422603, 3112: -14.8599917573, 3113: -1.15870892378, 3114: -39.4411934692, 3115: -12.4873305662, 3116: 45.010030913, 3117: -3.86333229713, 3118: 45.1366671613, 3119: 10.275325383, 3120: 32.6546010286, 3121: -5.9218876854, 3122: 12.8699929261, 3123: -11.2401809747, 3124: 9.7636951486, 3125: -2.3049045025, 3126: 12.453677962, 3127: -4.0078744795, 3128: 10.9805787474, 3129: -0.371363601267, 3130: -12.0188169432, 3131: -3.08548615757, 3132: 2.44858180166, 3133: 12.7138144631, 3134: 10.1009746606, 3135: -2.13061647196, 3136: -20.5703324222, 3137: 3.0745908473, 3138: 80.3435675733, 3139: -7.47669829804, 3140: -21.8887499778, 3141: 54.0844461707, 3142: 0.629939776868, 3143: 12.8289616807, 3144: 6.38993765339, 3145: 12.0503350157, 3146: 16.1384541333, 3147: -7.07417309131, 3148: -3.52615888117, 3149: -8.70304724317, 3150: 0.563827461282, 3151: 10.6234681755, 3152: 18.590458142, 3153: -3.17487952998, 3154: 11.5913260554, 3155: 29.6785413187, 3156: 1.94166783059, 3157: 8.76811912246, 3158: 5.18460350616, 3159: 2.19819907589, 3160: -19.8379149688, 3161: -1.85405757943, 3162: 6.97234776762, 3163: -25.1248763547, 3164: 32.6430020157, 3165: 3.06304428867, 3166: -1.15966161661, 3167: -7.48958698643, 3168: 9.76657189007, 3169: -2.51870389296, 3170: -43.766906817, 3171: -10.4259635916, 3172: 3.19869147364, 3173: -6.47600020876, 3174: 9.84770602018, 3175: -12.6531386203, 3176: 7.31041469504, 3177: 1.38156734751, 3178: 10.1737930467, 3179: 3.25625823532, 3180: -5.45667057305, 3181: -0.561240271705, 3182: -6.16549773355, 3183: 3.04840232216, 3184: 6.76970023328, 3185: 17.7095546085, 3186: -0.147231271545, 3187: -16.7977735789, 3188: -7.86657822345, 3189: 4.80579053369, 3190: -2.48151733178, 3191: 2.78574651443, 3192: -7.67250748149, 3193: 7.33022132247, 3194: 3.35792369679, 3195: -6.7489301785, 3196: -3.40408576843, 3197: 2.81107135581, 3198: -15.8298987639, 3199: 8.51986706262, 3200: -11.3860919229, 3201: 2.03775083245, 3202: -26.929660954, 3203: 1.63019493492, 3204: 6.11117962168, 3205: 6.28492417929, 3206: 14.5219425033, 3207: 2.45933365232, 3208: 13.4741559706, 3209: 3.21458228698, 3210: 4.7761094354, 3211: 3.17767947692, 3212: 3.30430048134, 3213: 2.48006313853, 3214: 3.22415451205, 3215: 2.95848781646, 3216: 2.33419741368, 3217: 0.603172553322, 3218: 0.111527116605, 3219: 5.39896554773, 3220: 9.28555858289, 3221: 5.43542765351, 3222: 2.9563206899, 3223: -3.76888421004, 3224: 3.30318581087, 3225: 2.41845818303, 3226: -5.56973951838, 3227: -6.81789273998, 3228: 3.28079192858, 3229: 7.114880774, 3230: 3.18901815418, 3231: 1.38955403306, 3232: 3.44806578109, 3233: 3.30352709858, 3234: 2.81718395778, 3235: 3.38761395111, 3236: -0.242645371963, 3237: 3.85982890639, 3238: 12.2950252373, 3239: 3.20965570826, 3240: 15.1618589688, 3241: 6.10424675845, 3242: -18.6090971722, 3243: -5.67384704828, 3244: 0.755345220899, 3245: 2.35063222527, 3246: 2.02453258802, 3247: -1.18532319075, 3248: 3.26107151943, 3249: 3.35516081562, 3250: 3.38886048108, 3251: 0.64790222555, 3252: 3.17423665513, 3253: 9.40030197973, 3254: -31.1837811012, 3255: -35.7008072541, 3256: 9.91667828933, 3257: 2.54687899501, 3258: 19.2612135561, 3259: 17.2965656024, 3260: -20.7050453973, 3261: 5.32733550865, 3262: -29.6916687777, 3263: 3.4932975105, 3264: -0.458799082725, 3265: -8.44814944796, 3266: 13.2172247538, 3267: 0.356032435465, 3268: 0.358883841534, 3269: 3.797895593, 3270: -10.5022671578, 3271: 0.87683948047, 3272: -2.27296696644, 3273: 2.77556371674, 3274: -8.79711306003, 3275: -1.08862247916, 3276: -0.0798916230178, 3277: 5.9888632514, 3278: -0.169758075015, 3279: 0.252014075507, 3280: -1.15818419714, 3281: 1.8699379021, 3282: 6.55165977539, 3283: 2.56909355971, 3284: 1.93697130784, 3285: 0.314445442791, 3286: 0.754142676347, 3287: 0.320085349593, 3288: 0.512233738254, 3289: -27.6727735575, 3290: 0.517244391606, 3291: 3.34649524106, 3292: 0.565765419869, 3293: 0.867171982553, 3294: 0.404325746524, 3295: 0.329621346943, 3296: 0.523442564347, 3297: 3.39399645839, 3298: -0.865931859616, 3299: 0.641949671706, 3300: 0.228903847658, 3301: 0.329865707555, 3302: -12.4526387394, 3303: -15.3727872616, 3304: 15.2570332345, 3305: -1.5680855966, 3306: 13.7146444161, 3307: 27.3989698081, 3308: 1.82515262403, 3309: -0.613822344759, 3310: 0.660303260991, 3311: 2.00882868616, 3312: -12.7865005853, 3313: 0.478575297932, 3314: -15.2088411095, 3315: 0.331085811821, 3316: -5.03651785337, 3317: -21.1245835036, 3318: 11.401020017, 3319: 0.173812922557, 3320: 0.358968064746, 3321: 0.385699230875, 3322: 0.324868298836, 3323: 0.397050871675, 3324: 0.535923085896, 3325: -6.42442569407, 3326: 2.36819205462, 3327: 13.4323175991, 3328: -10.1911135538, 3329: -8.81881703459, 3330: -7.3361798417, 3331: -0.368554319481, 3332: -5.27448864929, 3333: -2.47984856493, 3334: 0.71330405543, 3335: 26.9002361869, 3336: 1.43155319835, 3337: 0.914666095762, 3338: 0.420340579971, 3339: -8.24151753466, 3340: 0.437909021315, 3341: 3.36089329034, 3342: 0.350880903886, 3343: 2.62616938508, 3344: 0.348597301202, 3345: 0.46203001282, 3346: 7.44597051968, 3347: 3.26373120403, 3348: -14.2509928789, 3349: 0.471119476295, 3350: 0.40355974317, 3351: 0.319156054994, 3352: 0.381829298602, 3353: 2.29655016749, 3354: 6.22837803457, 3355: 3.61388251731, 3356: 0.413328415199, 3357: 0.984721936894, 3358: -8.3691296125, 3359: 2.46735579204, 3360: -11.4294754303, 3361: 4.95110331077, 3362: -11.4014943718, 3363: -1.76086629743, 3364: 0.153678980404, 3365: -1.84919741427, 3366: -5.89354979077, 3367: 0.505274500799, 3368: -28.5345323193, 3369: 3.09472087546, 3370: -0.0576288672552, 3371: -0.281160990164, 3372: -8.48253010399, 3373: 0.532665962276, 3374: -16.936511124, 3375: -0.195000201114, 3376: 3.21210917671, 3377: 8.8994362651, 3378: 3.43080369489, 3379: -13.5797481996, 3380: -16.7835832399, 3381: -2.24759379057, 3382: -21.1268298561, 3383: 3.86576210469, 3384: 1.16551616078, 3385: -2.41296163146, 3386: -3.14611690505, 3387: -7.28634815597, 3388: 6.94470511272, 3389: 0.122816560563, 3390: 17.2048869317, 3391: -30.8708148259, 3392: -7.11310620774, 3393: -9.72077768496, 3394: -3.56465880299, 3395: -8.53734764901, 3396: 2.88317845445, 3397: 3.21126558912, 3398: -8.33018281798, 3399: 3.07253708497, 3400: 3.38756545919, 3401: -2.61036292265, 3402: -5.30017966477, 3403: 3.20243137327, 3404: 0.132468756265, 3405: 1.14517425342, 3406: 3.23342466121, 3407: -3.97980004229, 3408: 22.0257995961, 3409: 2.53486462326, 3410: 1.18056377969, 3411: -15.2740711555, 3412: -9.94057306417, 3413: 5.70695811238, 3414: -0.75027550055, 3415: 8.79178468259, 3416: -14.7628193587, 3417: -17.2612617989, 3418: -2.78320662831, 3419: 20.9421941476, 3420: 1.83234432094, 3421: 0.655677703219, 3422: 11.2088626802, 3423: -2.57764670329, 3424: -5.61356552337, 3425: 23.1977764396, 3426: -4.87342723196, 3427: 8.58865475896, 3428: 3.2048618186, 3429: 3.46624481919, 3430: 3.20087548897, 3431: 2.88729944936, 3432: 2.48378234357, 3433: -6.72563414128, 3434: 6.64691017569, 3435: -16.5026626137, 3436: -2.79599016119, 3437: 11.0797721123, 3438: -13.3547036309, 3439: 15.2153764336, 3440: -3.28681422691, 3441: -1.09791369729, 3442: -7.07165179836, 3443: -28.1500180972, 3444: 0.257304282419, 3445: -21.575927658, 3446: 9.87509418264, 3447: -0.298599203374, 3448: -1.27714713826, 3449: 3.46496502466, 3450: 3.377914846, 3451: 3.30678307881, 3452: 0.611032173981, 3453: 3.39518684504, 3454: 3.32627927973, 3455: 13.596203513, 3456: 3.20049097349, 3457: -10.4936263987, 3458: 3.30626558038, 3459: 3.40693612391, 3460: 13.9888291096, 3461: 3.46148548638, 3462: 1.30578175422, 3463: 3.68428045084, 3464: -0.833329653243, 3465: -1.33930767027, 3466: -6.49852771094, 3467: 11.3755569831, 3468: -10.4792343081, 3469: 4.14919652147, 3470: 4.17470633604, 3471: 11.4063514049, 3472: -7.59663398909, 3473: 14.1088047299, 3474: -0.49035059895, 3475: 4.26717485661, 3476: -6.05053438012, 3477: 6.4811764105, 3478: 4.4714335724, 3479: 12.9500454991, 3480: 3.3139337663, 3481: -1.48285575671, 3482: 9.43815065872, 3483: 11.2942377373, 3484: -31.5048133609, 3485: 18.5106126937, 3486: -15.7098560693, 3487: -36.8878180582, 3488: -10.0705275758, 3489: 0.618856144243, 3490: -3.11834520926, 3491: 27.1463816574, 3492: 0.694676324804, 3493: 0.598409084416, 3494: -1.14704542107, 3495: -0.606834076918, 3496: -0.469707423388, 3497: -0.588378469184, 3498: -0.241118626237, 3499: -0.419053680524, 3500: -1.49804307007, 3501: 24.1236455235}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.483960675394, 2: -2.05453402484, 3: 1.92069225744, 4: 11.0465198613, 5: 0.383200854888, 6: -2.99256068371, 7: 5.62280449363, 8: 5.58631752022, 9: 2.48825362398, 10: -6.87873017536, 11: -3.20354687155, 12: -6.22020453918, 13: -1.18032110723, 14: 4.3954045037, 15: -3.62258195659, 16: 10.2294738342, 17: 5.36838983811, 18: -2.49821351497, 19: 5.79797581647, 20: 5.83899853188, 21: 3.80394312452, 22: -13.3302375269, 23: 4.24398202745, 24: 2.30500842816, 25: 4.14753388597, 26: -2.75435858791, 27: -5.6852380983, 28: 4.91889151778, 29: 6.53149209091, 30: 1.22912450103, 31: 0.0291528497867, 32: -3.79846756184, 33: -6.93107470316, 34: 2.26197388311, 35: -0.502531799779, 36: 13.3797809289, 37: -4.68758039168, 38: -4.01131371223, 39: -1.42275078505, 40: 1.21120577373, 41: 19.2194877298, 42: -11.8700783813, 43: 14.5804385549, 44: 15.1877102604, 45: 1.62911255047, 46: 6.31478714333, 47: 4.41780892043, 48: 3.48143958579, 49: 1.8185515597, 50: 0.778093711639, 51: 8.19224791418, 52: 24.061989523, 53: 1.6068457064, 54: 4.71163161883, 55: 7.97576849538, 56: 2.59096540634, 57: 3.15473569771, 58: -4.29817403772, 59: -18.7679213279, 60: -4.78191836221, 61: -5.15016768912, 62: -0.888358224643, 63: 15.0674800091, 64: 6.5901039178, 65: 0.0406236046703, 66: 2.17421395656, 67: -13.8960535102, 68: 0.867820323912, 69: 1.32857257271, 70: 1.7930354141, 71: -0.140416329602, 72: -19.969081044, 73: 1.8378091278, 74: 40.9222116111, 75: 1.12443186517, 76: 4.52010975821, 77: 3.11790310343, 78: 5.82827925392, 79: -12.8445410399, 80: 9.59242044096, 81: -4.03134757458, 82: -23.6084829642, 83: 1.99073662671, 84: 1.93574075135, 85: -0.371777773095, 86: -3.68773030617, 87: 1.23001652574, 88: 0.0555797582487, 89: -3.243808874, 90: 12.4298775416, 91: -0.0765776135629, 92: 2.50712943366, 93: 4.47204561483, 94: 1.97269185355, 95: 0.871128797702, 96: 2.44849959983, 97: -7.14134933318, 98: 1.54344547282, 99: -0.412255098153, 100: 1.2592389474, 101: -23.1914507889, 102: 2.22831002324, 103: -2.42749745818, 104: -0.704764249603, 105: 4.88153335488, 106: -15.3436335493, 107: 2.49286085481, 108: 2.97631746007, 109: -4.12990568311, 110: -8.02953726159, 111: 1.76289915978, 112: 6.30331500931, 113: 14.3480399141, 114: 17.259203674, 115: -15.3563038007, 116: -1.45909705754, 117: 7.96304898075, 118: -3.81549463266, 119: -8.89877553143, 120: 3.16395091321, 121: -6.71556919787, 122: 21.0729430624, 123: -2.68581327933, 124: 20.9481230397, 125: 27.7287923084, 126: 0.494496734847, 127: -0.511474260376, 128: 6.26043754491, 129: -24.1147881611, 130: 0.50551198528, 131: 2.36935720458, 132: -1.17514285161, 133: 13.7330376535, 134: 2.86987009018, 135: 3.88769263936, 136: -11.1098598317, 137: -5.45230466252, 138: -15.701481487, 139: -5.38115584946, 140: -10.7927706847, 141: -0.192331453297, 142: 7.81286028881, 143: 1.59221849616, 144: -2.05929003201, 145: 1.98671635068, 146: 0.411646427688, 147: 0.4244192408, 148: -0.983597764493, 149: 0.399676730241, 150: 2.31731370186, 151: -5.53278209867, 152: 0.765252323686, 153: 0.578803536133, 154: 0.65478667172, 155: 0.477411592836, 156: 3.57580903986, 157: -53.5180906888, 158: -0.509866512623, 159: 0.87433083863, 160: 1.41874923607, 161: -2.37636091765, 162: 5.84938208739, 163: 30.0291394026, 164: -9.72554049115, 165: 11.3899668911, 166: 21.5932438375, 167: -3.13655927279, 168: 11.4832242368, 169: -3.56083883631, 170: 6.97454783569, 171: -8.33658928253, 172: 0.329176457116, 173: -13.02083757, 174: 0.498753862831, 175: -14.0395444297, 176: 11.7410706788, 177: -3.07436659909, 178: -0.784167393181, 179: 0.549626492865, 180: 0.480790927279, 181: 0.533508340387, 182: 0.43433364553, 183: 0.627636562913, 184: -7.45543992431, 185: 8.42759380598, 186: 17.9430086028, 187: 0.920126468924, 188: -13.7683147314, 189: -9.05675573716, 190: -4.4875516253, 191: -5.04666414653, 192: -8.86561575855, 193: 0.469916960734, 194: -22.2210611604, 195: 2.48418372335, 196: 0.861536638824, 197: -3.31292104463, 198: 2.21758028944, 199: 0.53058129606, 200: 2.28856522509, 201: 0.668228478816, 202: -5.2549860689, 203: 0.418360063508, 204: 0.375134088921, 205: -10.5023972322, 206: 1.049952692, 207: -8.72901660137, 208: 0.404084819537, 209: 0.428153105252, 210: 0.300296244345, 211: 0.559935337176, 212: 0.60723148464, 213: -3.49367525808, 214: 4.22976471702, 215: 0.573365270428, 216: 0.92491666939, 217: -31.2731572907, 218: 4.35244823453, 219: -8.29387517069, 220: 19.9198899444, 221: 0.612260264977, 222: 4.79305276057, 223: 1.08426586091, 224: 0.381120026395, 225: -0.259778630524, 226: 0.604617525482, 227: -2.78697355797, 228: 3.58124050541, 229: -1.35314300346, 230: -0.245403038951, 231: 5.49403034284, 232: 0.328740961396, 233: 12.927162458, 234: -19.8988111628, 235: 4.39479895712, 236: 15.2737969791, 237: -15.6597838838, 238: -1.15008807019, 239: 2.04367351875, 240: 2.45838835601, 241: 9.86422122951, 242: -16.7350167316, 243: 7.46289160981, 244: -1.30789081207, 245: -1.95761246486, 246: -2.83786893702, 247: 7.66350463161, 248: -1.17271048681, 249: -23.5475932372, 250: -3.67142138827, 251: -14.8934255264, 252: -3.5790945204, 253: 8.08946116122, 254: 2.3056216986, 255: 6.50285609328, 256: 3.43580202967, 257: -4.64969273585, 258: 2.40108612598, 259: 2.21964471804, 260: -4.2947147784, 261: -3.69543297738, 262: 2.27331933841, 263: -2.66045502337, 264: -1.94401914547, 265: 3.38253884595, 266: 25.635839086, 267: 15.5218632176, 268: -6.32855941199, 269: -2.9029787462, 270: 0.458975602689, 271: 3.02844363443, 272: -16.4424413264, 273: -3.01255050511, 274: -14.0754581537, 275: -8.33750312007, 276: 15.9200367226, 277: 1.58066880215, 278: -2.59632771023, 279: 2.02395413476, 280: 23.4052386639, 281: 5.83640434742, 282: 17.759071909, 283: 27.5693843137, 284: -22.1787945701, 285: 11.6683418662, 286: -15.5730078467, 287: 2.34828396149, 288: 2.88431210145, 289: 1.70315263172, 290: 4.88025065836, 291: 1.17619975223, 292: 3.37947698859, 293: 8.46134885379, 294: -4.91580083563, 295: -6.0974732862, 296: -5.26320933058, 297: -4.96789867449, 298: -17.2091578225, 299: 1.87181532735, 300: -6.0464239807, 301: 0.156517444719, 302: 18.4027950175, 303: -8.02461770824, 304: -2.11811056112, 305: -5.55200284113, 306: 1.00715561142, 307: 12.0054726173, 308: 2.4303361426, 309: 2.32561561934, 310: 0.374683515044, 311: 1.41021701887, 312: 2.24615201038, 313: 2.46254242668, 314: 29.5206025122, 315: 1.02160825536, 316: 1.54252898504, 317: 2.43171057748, 318: 2.42586693478, 319: -8.22990255808, 320: 1.82607228712, 321: -0.671998435494, 322: -6.9361864952, 323: 4.68776485383, 324: 7.86311346464, 325: -3.7394984361, 326: 9.89661215125, 327: 2.94511355278, 328: -37.9196617925, 329: -8.89657263871, 330: 1.46231186362, 331: 12.5523633674, 332: -6.45770291385, 333: -1.41248286336, 334: -0.800499022782, 335: -8.23170500501, 336: -5.45790119744, 337: -8.9836564932, 338: 0.925195290837, 339: 2.43082209833, 340: 4.57529320218, 341: -3.40573760245, 342: -20.072299022, 343: -1.56164179104, 344: 2.8934252737, 345: -3.79716756504, 346: -24.2295619535, 347: -3.97800391289, 348: 7.46315821564, 349: -19.6095807849, 350: 0.510523501628, 351: 10.3577695937, 352: 22.8206383049, 353: 7.09186174879, 354: -1.94380085306, 355: 9.74745381001, 356: -21.8576297658, 357: -15.4058665459, 358: 6.13775240553, 359: 22.2705721113, 360: 13.0591531959, 361: 18.4472540691, 362: -3.15259870058, 363: -7.65687565701, 364: 2.75561141123, 365: -16.2083693871, 366: -30.1412922719, 367: 17.5256575607, 368: -41.7099933303, 369: 10.9787692623, 370: -1.89896335445, 371: 11.8211872891, 372: -7.39862027922, 373: -3.16496611785, 374: 1.57434924687, 375: 12.051868919, 376: -18.3342707816, 377: -9.94571688425, 378: 9.14351294865, 379: -19.9654972322, 380: 12.5684895708, 381: -1.53591387095, 382: -7.27790861666, 383: -4.08044013987, 384: 5.56047545002, 385: -15.1521839773, 386: 12.2774530249, 387: -2.68955665908, 388: -4.86971640047, 389: -7.82067631795, 390: -6.37917986605, 391: 11.5518254298, 392: 16.1319765359, 393: 11.9993681953, 394: -0.0851175195244, 395: -11.4387667487, 396: 15.5758573842, 397: 4.40330685985, 398: -0.201111181242, 399: -1.23193139699, 400: 6.60283793149, 401: -5.68737369026, 402: 1.64711467534, 403: 1.67234374445, 404: -6.17721706881, 405: 2.12092416626, 406: -15.4584045661, 407: -7.45219675089, 408: 15.2901714945, 409: 0.563068900345, 410: 22.1908359961, 411: -46.7563192038, 412: -9.00033902586, 413: -0.828399835475, 414: -2.10772819988, 415: 7.32525518008, 416: 4.34739175658, 417: -2.14180441554, 418: -5.51597591985, 419: 0.0192094852425, 420: 5.75931934316, 421: -10.5276844476, 422: 1.95275279832, 423: 11.1940413431, 424: -1.50586778858, 425: 3.42769100056, 426: -2.11901598798, 427: 2.05193751696, 428: -19.6007769548, 429: -14.3936824142, 430: 6.9879099809, 431: 15.3526380425, 432: 1.7044616077, 433: -1.205063706, 434: 0.114603344532, 435: 6.47857501619, 436: 6.16468687963, 437: -18.0496207386, 438: 2.44727745177, 439: 4.86954997444, 440: 2.08766101416, 441: 2.50481695354, 442: -21.4244378899, 443: 1.84267241115, 444: -1.56704493805, 445: 3.39697600304, 446: 4.80795124573, 447: 1.4051786084, 448: -8.16074801299, 449: 3.52577659178, 450: -5.47775648432, 451: 2.34554758471, 452: 10.6782078732, 453: 2.38158684895, 454: 10.590239966, 455: 29.703441955, 456: 2.12234712996, 457: 0.720049432642, 458: 2.0570667518, 459: 7.87442288899, 460: 4.97391797447, 461: -2.72624436893, 462: -17.1401756206, 463: -5.29922876477, 464: 9.7011863047, 465: -5.67032564294, 466: -3.07075689611, 467: -1.80638863454, 468: -12.064010955, 469: -7.67440435898, 470: 1.18431243791, 471: -6.11192802275, 472: -9.03325485075, 473: -5.13777215415, 474: 4.85802926208, 475: -0.0166397072836, 476: 0.0403517146191, 477: -4.10079331778, 478: -3.646688839, 479: 0.528510141058, 480: 2.36617511504, 481: 0.82384020091, 482: 7.89221188604, 483: 2.91870909496, 484: 8.36066616562, 485: 16.1213546649, 486: 0.653430016801, 487: -18.0105906327, 488: -7.35156843121, 489: -1.62712032563, 490: -17.6200894693, 491: -2.1664897422, 492: 12.2605705529, 493: 0.568934108829, 494: -11.0922196486, 495: 0.743017637209, 496: 0.547426195075, 497: 6.05545595918, 498: 0.472099693685, 499: 1.6809702331, 500: 0.125215933382, 501: 0.987005101303, 502: 0.547343984649, 503: 0.724567013433, 504: 0.803971759828, 505: 1.10913503382, 506: -9.62346521945, 507: 0.339823670773, 508: 1.46420143139, 509: 2.72955397516, 510: -2.82703883766, 511: 1.88276872387, 512: -16.9818180957, 513: -0.815922883743, 514: -15.3530499476, 515: -23.5086704937, 516: -7.16972322423, 517: -0.183830591575, 518: 15.0557267072, 519: -5.56663267816, 520: 10.1719450152, 521: 0.610676688076, 522: 13.4369450426, 523: 0.641903889428, 524: -27.4426186098, 525: 6.15288566959, 526: -18.5098454343, 527: 2.11889874778, 528: 0.512923336769, 529: 0.64008545711, 530: 0.750256482043, 531: 0.407207482274, 532: -3.98963750783, 533: -6.56284876857, 534: -5.78807526098, 535: -5.32808542561, 536: -12.7565910099, 537: 2.12324343992, 538: 6.70617449927, 539: -2.43511717675, 540: -7.60678602099, 541: 7.96412053618, 542: 0.609023158999, 543: -4.08364375464, 544: 3.49377942079, 545: -3.51110646832, 546: -7.82284633728, 547: -12.9856927431, 548: 0.41458941503, 549: 2.25803327984, 550: 0.556523679294, 551: -1.5960561075, 552: 0.60998239734, 553: 0.522644247682, 554: 22.9539783945, 555: 2.42935801211, 556: 18.8572664008, 557: 0.527285545182, 558: 0.50352527668, 559: 1.74809221797, 560: 0.490836588229, 561: 2.5193615281, 562: 9.08740608592, 563: -17.8859245155, 564: 0.060935946024, 565: 0.333642550589, 566: 27.980196839, 567: 7.11510424735, 568: -11.6061308384, 569: -0.792693285017, 570: 3.41828420578, 571: -6.89776418536, 572: -3.56188323487, 573: 13.8684222873, 574: 4.578008154, 575: 1.13722821834, 576: 4.27084725944, 577: 2.20618286311, 578: -1.91691945661, 579: -1.89192448462, 580: 1.04381479157, 581: 0.391295488032, 582: -21.3374371372, 583: 4.92794142287, 584: 17.0301501081, 585: -20.1446964404, 586: -8.20359212621, 587: -4.66514699264, 588: 2.07185889877, 589: 2.39311264767, 590: -6.53249030892, 591: -7.22588131713, 592: -31.1882355132, 593: 12.8054213076, 594: -8.70990643648, 595: -9.62618411735, 596: 16.0751992775, 597: -25.4499718385, 598: 14.3138152626, 599: 19.2807791917, 600: 38.6405617362, 601: -9.58066925487, 602: -10.111792225, 603: 7.30707972214, 604: 0.77912211482, 605: 2.8668211251, 606: 7.4853571219, 607: 2.55561626879, 608: 1.76781672136, 609: 0.858185811989, 610: 2.68660964315, 611: 1.89608253139, 612: 3.75850499144, 613: -2.3957678359, 614: 1.01582302327, 615: 37.6167270253, 616: -20.2086900874, 617: -15.9701128427, 618: 7.14201215666, 619: -5.86795458042, 620: -2.28581313794, 621: -14.957632733, 622: 4.3789599045, 623: 10.5486359619, 624: -0.0163334745693, 625: -19.2381236447, 626: -9.5910084483, 627: 12.1705239067, 628: -16.2554251147, 629: -3.06608814023, 630: 0.323079151383, 631: -16.6410733787, 632: 43.7316927807, 633: -3.24151273075, 634: 19.4374659033, 635: -2.2368320297, 636: 2.49672965054, 637: 3.26474739942, 638: 1.30831996039, 639: 8.19403077213, 640: -1.6276016287, 641: 2.91594931544, 642: 6.6306208424, 643: -7.7200198859, 644: -0.303354482543, 645: -9.35174756123, 646: -25.9800877218, 647: 2.14230089019, 648: 3.13051588309, 649: -10.3522389051, 650: -3.21425356915, 651: 4.2721333322, 652: 4.84373784051, 653: -12.7869057745, 654: -1.63746047889, 655: -1.62145187287, 656: 0.72469691466, 657: 2.20414128058, 658: 2.25315311341, 659: 1.98871082805, 660: -0.815070627343, 661: 2.38531244007, 662: 1.32031032187, 663: -39.8981583675, 664: 2.38091351817, 665: -1.72163353109, 666: 1.96102392022, 667: 2.58314742231, 668: 5.17174479238, 669: 2.00075304534, 670: 2.77732441817, 671: 6.43145826594, 672: 1.79738390521, 673: -6.19543790132, 674: -3.01904219556, 675: -4.58171022971, 676: 5.6087056298, 677: 22.231785253, 678: -11.4302916058, 679: 7.50338849247, 680: 0.215539534301, 681: 8.79560978029, 682: 20.9604695138, 683: 10.5078548975, 684: 2.85164311513, 685: 1.80588066023, 686: 3.02393035552, 687: 1.95814875999, 688: 2.39365417116, 689: 11.3228950358, 690: 1.00674455528, 691: 3.9324589607, 692: -5.38788860569, 693: -3.36347008571, 694: -6.30411986752, 695: 54.4251105185, 696: -10.8675003566, 697: 8.4892517548, 698: -10.6011593192, 699: -0.309176412823, 700: -34.4151000498, 701: 28.6304978256, 702: -2.25204565496, 703: -28.5099021291, 704: 9.04150945637, 705: 6.66117754646, 706: 20.2044403033, 707: 13.5467280044, 708: 8.40455581904, 709: 7.01773199158, 710: -0.609506366293, 711: -6.77829576629, 712: -15.3741977323, 713: 6.29583208522, 714: -6.49308904142, 715: 29.4396432527, 716: -12.1973663657, 717: -5.46397255954, 718: 6.91868627451, 719: 7.42674430721, 720: -4.44535567402, 721: -25.9166531981, 722: 16.0711393486, 723: 18.5288268653, 724: -9.84538133157, 725: -1.65081563652, 726: -12.0296089014, 727: -4.65988163267, 728: -31.9759356155, 729: -44.1242197154, 730: -20.2346952875, 731: -25.809588558, 732: -8.36411598876, 733: 0.570487768759, 734: 0.738046622693, 735: 14.6433485292, 736: -32.2217978877, 737: -46.7843683817, 738: -13.7898173472, 739: -14.2909868076, 740: -1.94260574155, 741: 1.39414417149, 742: -16.0471325797, 743: 2.8727946421, 744: -7.17569216984, 745: 28.9092306283, 746: -0.703661741553, 747: -10.0357346441, 748: 0.903672993865, 749: -12.8081795846, 750: -15.359553149, 751: -1.86997933069, 752: 3.23325936206, 753: -7.52607546427, 754: -1.82883397296, 755: -56.5627163499, 756: -1.36316723942, 757: -21.52051477, 758: 14.8208585185, 759: -6.12246961374, 760: -4.71094082128, 761: -28.4108440214, 762: -10.2422812076, 763: 29.3591524507, 764: 4.48957352998, 765: 35.4104513331, 766: -1.63064539545, 767: 5.645344631, 768: -0.99414019905, 769: -1.74952234853, 770: -37.9229376348, 771: -1.69702482398, 772: -5.89724505839, 773: -11.7145981925, 774: -8.94633996082, 775: -5.25620748807, 776: 9.25927707933, 777: -54.0261924846, 778: -36.9711861359, 779: -1.41580308546, 780: 17.9424965098, 781: -0.480581680729, 782: 1.48596765982, 783: 0.0468052458131, 784: -13.0505291578, 785: 29.7911900888, 786: -14.610752979, 787: -1.97476232221, 788: 16.6902015838, 789: 32.0244156488, 790: -1.68911620853, 791: 7.85870178738, 792: -1.69017023935, 793: -1.44311703037, 794: -2.2966164933, 795: -0.10458575257, 796: -1.47805395135, 797: -11.8882628899, 798: 15.2048200116, 799: 19.052275499, 800: -3.35224859374, 801: -3.51452656269, 802: -1.71674265634, 803: 2.47648593376, 804: 18.2271236033, 805: -1.56462214943, 806: -1.68941906661, 807: -1.37488857259, 808: 17.3336359597, 809: -3.2002958535, 810: 8.69711389913, 811: -1.06542688429, 812: -2.47359057806, 813: 7.42512374452, 814: -5.55103610096, 815: -6.89647250279, 816: -20.7095004799, 817: -3.34044183157, 818: -21.8119279597, 819: 33.7939707992, 820: -30.7203275628, 821: -11.5637843672, 822: -1.38536790425, 823: -6.79627216725, 824: -0.658722425303, 825: -0.359535775195, 826: -18.2436357855, 827: -29.097244792, 828: -0.440011225087, 829: -0.0429734133464, 830: -0.11126794845, 831: -38.219530836, 832: -7.47628037481, 833: 6.9553064493, 834: -4.89452484827, 835: -4.05725769863, 836: -1.21555499996, 837: 2.51907357979, 838: -3.18135854379, 839: -5.26859936828, 840: -9.24857419329, 841: -14.3048503099, 842: 0.566001163326, 843: 3.84795351403, 844: -1.00107987726, 845: -0.421210682127, 846: 17.398045077, 847: -0.371984434295, 848: -1.72572854104, 849: -7.72019640331, 850: -0.203876504443, 851: -0.393610348152, 852: -0.539737927658, 853: -0.434894860103, 854: -3.51814015093, 855: -23.290920852, 856: 0.952065295351, 857: 0.226675084689, 858: -3.52384970148, 859: 3.83300320909, 860: -1.39053540643, 861: -15.8626527157, 862: 14.5369823107, 863: -39.6342776611, 864: -11.6530003402, 865: 33.1810405015, 866: 9.49783036587, 867: 25.9392660946, 868: -0.713323934737, 869: -20.5262092814, 870: 0.0721891091239, 871: -47.4571493947, 872: -0.329397433168, 873: -37.9202963741, 874: 12.709495156, 875: 52.5104281041, 876: 1.24279638242, 877: -0.425456072429, 878: -0.155671297525, 879: -3.23949089224, 880: 0.174416857904, 881: -4.32401362895, 882: -43.1970348904, 883: 8.40119034329, 884: 48.3484099764, 885: -30.2186950193, 886: 12.3020091155, 887: 6.20328292346, 888: -10.8437304251, 889: 2.36593119012, 890: 20.7008835103, 891: -0.238674097995, 892: -12.2629430893, 893: 3.22027402723, 894: -8.74926292069, 895: -2.16252955479, 896: -8.92710193767, 897: -0.241484147413, 898: -1.54569453543, 899: -0.40354176441, 900: -25.6164359509, 901: -0.466546716069, 902: -0.421762271153, 903: 4.84698109872, 904: -2.14359303359, 905: 0.787614380388, 906: -0.371691295548, 907: -0.251459076422, 908: 2.29432484262, 909: -0.454648493325, 910: 2.45958977079, 911: 14.8190904894, 912: -12.2530188106, 913: -1.35726203172, 914: 1.43047911655, 915: -5.12509542148, 916: -0.332448927809, 917: -57.2878065308, 918: -28.3045794589, 919: -1.30715036478, 920: 11.7668389321, 921: 0.984236987972, 922: -1.70541771411, 923: 30.2533298277, 924: -0.449537465137, 925: 28.0675738374, 926: -9.08550770294, 927: -4.41410529987, 928: -1.93786400867, 929: -8.05286418064, 930: -0.938940257431, 931: -12.6237530945, 932: -6.55617019388, 933: 6.75289368305, 934: 3.30509821282, 935: 28.3319653857, 936: -27.0477853626, 937: -1.85229937481, 938: -0.0278268087917, 939: 0.629148386109, 940: 12.3677209116, 941: -14.6653613133, 942: 6.29141527331, 943: 29.1472446423, 944: -5.81558996738, 945: 11.2370682623, 946: 17.5089704795, 947: -29.6124711439, 948: 17.4091749333, 949: 40.1836062802, 950: 15.5788209673, 951: -6.7082069647, 952: 12.4113369225, 953: 4.43596838694, 954: -2.02557427493, 955: 9.60560561902, 956: -2.00278980836, 957: -1.53234011156, 958: -7.89889253548, 959: -7.07008545135, 960: -1.76091849445, 961: -6.77773913603, 962: -5.59505430442, 963: -3.63641946234, 964: -16.8577628636, 965: 2.79940608727, 966: 5.48126955734, 967: -0.467325921057, 968: 3.85333683896, 969: -1.4294696394, 970: 9.63561080114, 971: -4.01629873241, 972: 9.20257959359, 973: 68.1202557148, 974: -6.65021499375, 975: -14.4718009731, 976: 18.0292744959, 977: -3.78425398837, 978: 14.827486261, 979: -16.9504041458, 980: 41.1530121432, 981: 14.779729921, 982: 19.1242501518, 983: -7.89001791669, 984: -48.9611348675, 985: -1.64108013419, 986: -2.49384508182, 987: 26.7423815314, 988: -8.42258329269, 989: -9.1257198145, 990: -1.86762001353, 991: 9.13489699513, 992: -40.2968452794, 993: -23.1996380309, 994: 13.9394466593, 995: 11.2924821939, 996: -1.76473719968, 997: 2.51758995521, 998: 5.19205128, 999: -11.4707857883, 1000: 23.3384357382, 1001: 7.22272979974, 1002: -3.05194946784, 1003: -7.466534982, 1004: -1.21244953164, 1005: 33.7174524821, 1006: -1.62032582811, 1007: -1.59151367588, 1008: 0.469555163793, 1009: -11.9621695541, 1010: -1.87080807679, 1011: -1.831199177, 1012: -19.7300092884, 1013: -2.18687442957, 1014: -16.6022710035, 1015: -1.85254879399, 1016: -1.83861375507, 1017: -18.4064976884, 1018: -1.63456425793, 1019: 4.48200847607, 1020: 3.99557604469, 1021: -18.8018791039, 1022: 49.8518539042, 1023: -28.8345670313, 1024: 10.5750079377, 1025: -4.52515214435, 1026: 42.0156248705, 1027: 7.46776394902, 1028: -14.854866183, 1029: -14.3175567176, 1030: -21.0550265836, 1031: -1.06889828648, 1032: -14.6777986937, 1033: -4.83146673612, 1034: 46.7303368376, 1035: 23.4078530598, 1036: 21.4877688886, 1037: -1.93194723227, 1038: -0.748702631127, 1039: -4.54029431517, 1040: 21.3450874114, 1041: 9.09242585972, 1042: -25.0078947454, 1043: 16.2978965376, 1044: 59.9260238744, 1045: -4.45512328783, 1046: 5.09791398702, 1047: 36.322661976, 1048: -0.233441630845, 1049: -10.8142816527, 1050: 7.16972789896, 1051: 4.02420907662, 1052: 44.5948644758, 1053: -8.92898982044, 1054: -1.66560745605, 1055: 43.8005432792, 1056: -9.6165752476, 1057: -8.60690835916, 1058: -48.4302439926, 1059: -27.1297842267, 1060: 28.7948051716, 1061: 15.8684460143, 1062: -5.43833031335, 1063: -9.16731774823, 1064: 35.0100710761, 1065: -23.8021615239, 1066: 22.0185378429, 1067: -26.507102511, 1068: -14.4836204199, 1069: 11.6243129108, 1070: 26.9129343951, 1071: 24.3207547678, 1072: -14.3947817231, 1073: 17.6867657207, 1074: -8.01079015203, 1075: -17.43271819, 1076: 27.7487243841, 1077: -10.2226237629, 1078: -34.4448027653, 1079: -23.8928678301, 1080: -10.1332921589, 1081: 2.13752795019, 1082: 8.48919153105, 1083: -6.60597028302, 1084: -1.75249052952, 1085: -4.14184134621, 1086: 1.10814327259, 1087: 12.414580097, 1088: 2.65687853309, 1089: -2.48054651172, 1090: 11.3607912997, 1091: -25.0626201715, 1092: 22.4145273939, 1093: 13.9558900034, 1094: -22.7236624561, 1095: -3.11784598181, 1096: 19.684476132, 1097: 0.649914605358, 1098: 11.0649099448, 1099: -6.08897882255, 1100: -2.19077383428, 1101: -2.75128729044, 1102: 7.98957068067, 1103: -2.1461034891, 1104: 11.3793348306, 1105: -3.01744333506, 1106: -9.38142947809, 1107: 84.3024025006, 1108: -10.1117142666, 1109: 26.8185058177, 1110: 8.79126890381, 1111: -15.3306791583, 1112: 19.5196568927, 1113: -4.301913873, 1114: 47.8613465403, 1115: -2.78269749514, 1116: -21.7051536147, 1117: -3.06936589216, 1118: -3.95254403322, 1119: 4.13531634157, 1120: -2.7358274706, 1121: -55.8163849902, 1122: 7.46103968524, 1123: -1.60771659911, 1124: -4.08943084212, 1125: -2.90695267933, 1126: 23.7416678342, 1127: -40.5302728897, 1128: 0.264679899892, 1129: 27.7063440215, 1130: -2.48140893813, 1131: 7.95625103356, 1132: 7.22248572826, 1133: -20.8215938771, 1134: -2.15461062704, 1135: -9.37603563497, 1136: -1.28574790682, 1137: 15.7517738136, 1138: -3.03452367992, 1139: -2.74202147543, 1140: 18.2639216033, 1141: -2.84507600007, 1142: -3.05842153997, 1143: 2.70375332982, 1144: -2.27579650798, 1145: -2.90055519648, 1146: -3.87063606297, 1147: -2.82445599949, 1148: 20.7153924022, 1149: -2.69393233594, 1150: 4.79093985005, 1151: -2.16265470729, 1152: 21.1422049937, 1153: 24.1307090968, 1154: -2.7261471101, 1155: -2.24219266264, 1156: -2.41311637238, 1157: 1.88107615759, 1158: -2.76574586014, 1159: -18.9457089954, 1160: 79.1212752944, 1161: 30.6228252849, 1162: -24.0208282987, 1163: -4.67293283839, 1164: -27.6875701435, 1165: 13.5759489618, 1166: 27.0237096777, 1167: 50.0923555861, 1168: 14.5814953871, 1169: 23.264864168, 1170: -21.9227393681, 1171: 4.78418535077, 1172: 4.93118803501, 1173: -0.304257779097, 1174: -0.58904445445, 1175: 0.713898013135, 1176: -43.464447109, 1177: -0.493256002883, 1178: 0.361303480612, 1179: -1.9280599788, 1180: 32.5045895023, 1181: 0.895252298868, 1182: 3.32204198519, 1183: -45.9488356577, 1184: -5.72504133984, 1185: 28.6938043608, 1186: 3.01648332617, 1187: 10.8709519615, 1188: 12.9796011481, 1189: 5.62676985245, 1190: -11.6165974712, 1191: 0.536987776944, 1192: -13.2269876172, 1193: -0.295870700652, 1194: -0.35725258538, 1195: 8.76425640288, 1196: -0.446691650963, 1197: -1.11783850863, 1198: -6.55695477871, 1199: -1.27277875624, 1200: -0.302113732444, 1201: -0.286453018252, 1202: -0.453926367402, 1203: -3.95141086788, 1204: 53.8777598985, 1205: 7.97466453614, 1206: 1.21335173462, 1207: 0.240560774514, 1208: -12.1276473047, 1209: 8.26866355245, 1210: 14.4341041125, 1211: 23.5836507618, 1212: 10.9685199042, 1213: 23.6809363391, 1214: -36.5404239687, 1215: 9.61626706989, 1216: -21.1518346544, 1217: -2.95626763619, 1218: -4.27709723606, 1219: -0.576976504236, 1220: -21.3172954131, 1221: -0.280773452876, 1222: 60.9421641736, 1223: -14.2018001864, 1224: 0.355925924656, 1225: -1.93734748542, 1226: -0.281404076258, 1227: -0.441018954305, 1228: -0.265598973635, 1229: -0.758056979236, 1230: -4.5742755552, 1231: 1.10037569566, 1232: 4.84859240363, 1233: -34.178570057, 1234: 10.8603091669, 1235: -11.309966113, 1236: -4.26661259698, 1237: -4.38775151269, 1238: 2.24860564751, 1239: -52.3723566379, 1240: -0.469932183061, 1241: -1.25113624286, 1242: -6.03728245946, 1243: -6.14575840689, 1244: -3.82979298902, 1245: -9.05261055602, 1246: -0.382108793096, 1247: -2.56076555166, 1248: -0.36588417443, 1249: -74.876444836, 1250: -0.496842702724, 1251: -0.441820309121, 1252: 12.0211171162, 1253: -2.31682775951, 1254: 45.0218955989, 1255: -0.513073001702, 1256: -0.51865883662, 1257: -2.23627416586, 1258: -0.402764098798, 1259: 1.64213268516, 1260: -10.4435598772, 1261: 7.66507722919, 1262: -1.48585041282, 1263: -0.497762121312, 1264: 6.5168887931, 1265: -7.91842678614, 1266: 1.8998126789, 1267: 15.9288046376, 1268: -8.58249657077, 1269: 5.04692400537, 1270: -9.09279032585, 1271: -2.20736206787, 1272: -14.7485627866, 1273: -0.16422790897, 1274: 33.432912838, 1275: 1.92040166559, 1276: 1.96236682671, 1277: -6.90607637095, 1278: -39.2870648802, 1279: -1.1018421992, 1280: 54.4677483538, 1281: -0.385492487353, 1282: -2.69745124664, 1283: 2.64844125127, 1284: -18.4513173312, 1285: -10.6719501245, 1286: -2.40563652123, 1287: 0.497525473887, 1288: 24.8864909629, 1289: -19.8742975927, 1290: -25.4889384937, 1291: 6.05598482517, 1292: 31.0461416485, 1293: 2.46911829551, 1294: 2.25219591032, 1295: -13.8083982918, 1296: 23.5539498775, 1297: 31.1741148948, 1298: 3.84266544076, 1299: 2.38911401866, 1300: 16.7180825485, 1301: -5.22202184725, 1302: -3.99451255806, 1303: -0.292080960033, 1304: 7.04346395438, 1305: -2.49028421945, 1306: -1.29632939363, 1307: 11.0287731303, 1308: 4.52415408092, 1309: -2.18450751812, 1310: -20.3435261155, 1311: 7.78151380448, 1312: -3.92479872971, 1313: -56.1609473938, 1314: -23.1480023038, 1315: -10.9929139978, 1316: 45.5151608413, 1317: -1.79694059586, 1318: 4.14533898358, 1319: 16.0071456256, 1320: 46.6008968975, 1321: -32.5377684335, 1322: -85.9317006341, 1323: 28.735834219, 1324: 13.5496324367, 1325: 4.47042628253, 1326: -2.16493694322, 1327: 32.6606646117, 1328: 6.2624837248, 1329: -2.32274775362, 1330: -63.3934102172, 1331: -0.53703232771, 1332: -32.0858136811, 1333: 39.2395630078, 1334: -2.5034533734, 1335: -15.3062858475, 1336: -3.45596159078, 1337: -1.26207665067, 1338: -4.33542379848, 1339: 2.89331590381, 1340: 31.5471590497, 1341: -2.52092319788, 1342: -45.5478455865, 1343: -27.4470194638, 1344: 18.9464099601, 1345: -7.29722551623, 1346: -3.97643431301, 1347: 9.02708783215, 1348: 0.670703909593, 1349: -13.4614088303, 1350: -8.19826495036, 1351: -27.5664192939, 1352: 32.154069415, 1353: -3.02473632142, 1354: 28.8350778, 1355: -2.53530341967, 1356: -2.5401734944, 1357: -2.79520715935, 1358: 26.7333538408, 1359: -2.40984111016, 1360: -2.95446370602, 1361: -38.4232754045, 1362: -2.3441747939, 1363: -51.0893873014, 1364: -2.41702622061, 1365: -2.47425502545, 1366: -32.0001127548, 1367: -2.91566835164, 1368: 4.36850217047, 1369: -31.1355745117, 1370: 40.3847986465, 1371: 31.2826297353, 1372: -9.74938141949, 1373: 10.3580535019, 1374: -2.73682402011, 1375: 43.4658445242, 1376: 40.0077275067, 1377: 7.87077080234, 1378: 3.32830647067, 1379: 21.0246379045, 1380: -1.07004016056, 1381: 15.1300360958, 1382: -2.95102645165, 1383: -46.2101606308, 1384: 2.56135926402, 1385: -53.1026235186, 1386: -2.43572353516, 1387: 64.8051736924, 1388: 6.07698904003, 1389: 0.61632764627, 1390: -6.0514845827, 1391: -28.0213400092, 1392: 21.7859546676, 1393: 48.4253124964, 1394: 37.64416771, 1395: 0.276330547654, 1396: -21.1706779895, 1397: -0.985161200304, 1398: -4.46912902288, 1399: 10.6471775738, 1400: 6.80020890677, 1401: 10.7211114271, 1402: 4.87673548789, 1403: -1.87204623494, 1404: -0.549405246035, 1405: 4.89294220238, 1406: 10.3043687284, 1407: 4.67699761505, 1408: -3.01709895929, 1409: 7.32136129882, 1410: 16.0322237768, 1411: -1.42887896459, 1412: -2.13940862136, 1413: 28.6745667621, 1414: 4.47944174604, 1415: -4.75390141889, 1416: -6.70779598157, 1417: 1.91063687508, 1418: 47.0303450189, 1419: 2.47833534302, 1420: 1.12398286004, 1421: -6.14750280364, 1422: 17.5148765818, 1423: 12.0849689423, 1424: -9.66754370063, 1425: 0.965224493163, 1426: -5.67095294371, 1427: -3.61140436867, 1428: -0.992191919526, 1429: -8.82198835849, 1430: -1.30621877028, 1431: 3.31415721875, 1432: 2.77692324151, 1433: 0.0822626845852, 1434: 4.80586879717, 1435: -6.90385043998, 1436: -0.754400157578, 1437: 11.0405709357, 1438: -1.26290343539, 1439: -2.34156317097, 1440: -10.4098790017, 1441: -2.80826875982, 1442: 4.78747890911, 1443: 4.81551020047, 1444: -5.31381653212, 1445: 6.44949985717, 1446: 2.63148090152, 1447: 6.80315038059, 1448: 6.22139206243, 1449: -3.55008586183, 1450: 9.10748335511, 1451: -1.74068712233, 1452: -0.414081909001, 1453: 2.59912721598, 1454: -20.2812558956, 1455: -9.12276481273, 1456: -0.11275099204, 1457: 16.3295586174, 1458: 5.67135422191, 1459: -3.62820714425, 1460: -0.115084408885, 1461: 3.70460286024, 1462: 2.60902114725, 1463: 10.2598739951, 1464: 1.68539485674, 1465: -4.39525286433, 1466: 1.18054452067, 1467: -6.07410354526, 1468: 2.74742099824, 1469: -6.20532125318, 1470: 18.7509063613, 1471: -0.720594277482, 1472: -6.5040190471, 1473: -4.35582278914, 1474: 4.03368171101, 1475: 13.0927480561, 1476: -13.976641132, 1477: -0.630836814606, 1478: 5.85796268311, 1479: 0.100001890117, 1480: 13.1675515719, 1481: -0.00438036710595, 1482: 4.74002138331, 1483: 1.40672797444, 1484: -5.67160748293, 1485: -1.08486399101, 1486: -4.79168237808, 1487: -2.86710977956, 1488: -6.17786722893, 1489: 23.8865511557, 1490: -6.09868443671, 1491: -3.40511322742, 1492: -7.02284011555, 1493: 1.79347766141, 1494: -6.22619080934, 1495: 4.0735017458, 1496: -4.39297371977, 1497: -1.12449631956, 1498: -6.28799107771, 1499: 0.610924312826, 1500: -1.02247302522, 1501: 3.06105926834, 1502: 9.69852576883, 1503: -2.85340754661, 1504: -4.21958840622, 1505: -0.874012227901, 1506: 8.47584307632, 1507: -3.56624293586, 1508: -1.53629594497, 1509: -3.23113677575, 1510: 20.5649550623, 1511: -1.49839624291, 1512: -3.86063161378, 1513: -0.744645341268, 1514: -3.97201845782, 1515: -0.481854693946, 1516: -5.7220763458, 1517: 2.1649048664, 1518: 6.60405333902, 1519: 9.36339584298, 1520: -1.53780320266, 1521: 31.0245000353, 1522: -1.09312477927, 1523: -0.300303267497, 1524: -7.39302308929, 1525: -11.7676695464, 1526: -1.20336753269, 1527: -4.26153244658, 1528: -1.45391423999, 1529: 11.7649419938, 1530: -4.48979110257, 1531: 0.769747167289, 1532: 1.98256841733, 1533: -5.29513302123, 1534: -6.58337336997, 1535: 5.57480061901, 1536: 2.13333486857, 1537: 9.87581201265, 1538: -4.62294627403, 1539: -2.9398420126, 1540: 1.13750112657, 1541: -2.83995952372, 1542: -1.27336045382, 1543: -1.19905345895, 1544: -13.9197499961, 1545: -0.972790434942, 1546: -4.12740121648, 1547: 1.54973735151, 1548: -1.74662547958, 1549: -1.18653548405, 1550: -1.65268220748, 1551: -1.06737032908, 1552: -2.83866269773, 1553: -23.2421697093, 1554: 0.118368476277, 1555: -1.81744861166, 1556: -2.84361686448, 1557: -5.49840619714, 1558: -5.88713635858, 1559: 21.6531374496, 1560: 11.8206009641, 1561: -23.7843343608, 1562: -24.8095066382, 1563: -2.38501968502, 1564: 2.92265764325, 1565: 3.29752551468, 1566: -2.97810090625, 1567: 34.9530259814, 1568: -1.42075167361, 1569: -12.0363254038, 1570: -1.44378958737, 1571: -7.94810013972, 1572: -2.34841366611, 1573: -1.31536242097, 1574: -2.73111775877, 1575: -1.13746504478, 1576: -1.33628702295, 1577: -1.03828289211, 1578: -1.22794115995, 1579: -0.791187960324, 1580: 10.1350898225, 1581: -3.74299919288, 1582: -4.36760092837, 1583: -2.6698829173, 1584: 4.33598605518, 1585: -4.52207209035, 1586: 9.99048203598, 1587: 1.13984201993, 1588: -25.6874505598, 1589: -1.1383636362, 1590: -11.0877098585, 1591: -2.66670180745, 1592: -1.68798306178, 1593: 3.67356966411, 1594: -0.430526740654, 1595: -1.16667810967, 1596: -2.91478610602, 1597: -1.26237731281, 1598: -8.52670613465, 1599: -1.36003048035, 1600: -1.09701556676, 1601: 27.9185969576, 1602: -4.74806709242, 1603: 1.85009030966, 1604: -1.01871506572, 1605: -1.04055388047, 1606: -2.74380813061, 1607: -1.11029501526, 1608: 0.361419641183, 1609: 1.10118668955, 1610: 10.0577597972, 1611: -0.487149826572, 1612: -0.986870717355, 1613: -0.969340998021, 1614: 7.04365679071, 1615: -24.4749574087, 1616: 5.80468654392, 1617: 3.75913151462, 1618: -2.16392872218, 1619: 4.32641730341, 1620: -3.53329883193, 1621: -6.81403921427, 1622: -1.67574943496, 1623: 6.86395343605, 1624: -3.51797625467, 1625: -1.04768696389, 1626: -5.40856927748, 1627: -19.4830116081, 1628: -1.1699176091, 1629: 11.6257382051, 1630: -15.6008820762, 1631: -4.31734509728, 1632: 9.58231845787, 1633: 0.52554488463, 1634: 0.769751809638, 1635: -5.02874401748, 1636: -4.25366411026, 1637: -4.85516504049, 1638: -3.03217743799, 1639: -9.19726739958, 1640: -3.65273814172, 1641: 1.41282846299, 1642: 10.3274196237, 1643: 27.0168822763, 1644: -5.75880825032, 1645: 9.26617128172, 1646: 6.37629938575, 1647: 9.8894680723, 1648: 2.46979194624, 1649: 6.88436042905, 1650: 1.36705735408, 1651: -1.04271306507, 1652: -7.82512366176, 1653: 15.7320853695, 1654: -6.66636742031, 1655: -3.94073924261, 1656: -4.24126559194, 1657: 22.1133833303, 1658: -5.86361169742, 1659: -4.16240909376, 1660: -1.81747872925, 1661: -2.68674577505, 1662: -4.05642279653, 1663: 15.3911709649, 1664: -23.3520586113, 1665: 34.2661360197, 1666: -6.86871411952, 1667: -1.4765604714, 1668: -13.3742381474, 1669: 12.2502660552, 1670: 18.7048029214, 1671: -11.2079878042, 1672: 7.03354954919, 1673: -1.7294973684, 1674: 1.10962635622, 1675: -4.7580780058, 1676: -1.81323851103, 1677: -3.03832861677, 1678: 10.5434596188, 1679: 0.856755793927, 1680: 5.1983656276, 1681: 3.49341417455, 1682: -8.81350503964, 1683: -6.36843644501, 1684: -1.4810508871, 1685: -6.13398506947, 1686: -1.30826917643, 1687: -0.541158032037, 1688: -2.95062943967, 1689: -1.01359977147, 1690: 2.87446391115, 1691: -18.118232621, 1692: -12.1278439539, 1693: -2.29518672552, 1694: 3.7028879537, 1695: 3.3739411545, 1696: -7.42082468815, 1697: -0.612600014116, 1698: 4.24292389727, 1699: -2.4295787122, 1700: -5.47454261015, 1701: 1.72456703869, 1702: -3.44047770963, 1703: 11.0837702366, 1704: -2.84090876709, 1705: -2.77427720297, 1706: -6.46400898655, 1707: 2.6604134177, 1708: 0.0258553898603, 1709: -4.35461765268, 1710: -25.4203704623, 1711: -4.55195656008, 1712: -9.69392815143, 1713: -6.0937590674, 1714: -6.25323609236, 1715: 1.50742803998, 1716: -4.15905870364, 1717: 5.24834597203, 1718: 8.42177843779, 1719: -3.903559001, 1720: -17.1732223179, 1721: 2.84895288727, 1722: -9.99022250551, 1723: -4.7875151813, 1724: 4.86051009291, 1725: 12.6394323576, 1726: 7.1508030303, 1727: 0.518053070533, 1728: 14.6449886169, 1729: 0.578136903208, 1730: 15.9548431049, 1731: 7.51315342258, 1732: -19.2573548676, 1733: -0.321961961686, 1734: 7.03555361654, 1735: -6.41520943517, 1736: 26.8917680609, 1737: 1.2368055092, 1738: -12.5584029917, 1739: 4.72092616652, 1740: 0.451272829324, 1741: 8.5129280147, 1742: 5.49644183064, 1743: 15.288218812, 1744: 0.715763215051, 1745: -3.51889444812, 1746: 0.605009710482, 1747: -9.37331611232, 1748: -16.0708115226, 1749: 25.0628874723, 1750: 10.7758666141, 1751: -20.5816208404, 1752: -4.03517778717, 1753: 0.411584843861, 1754: -18.5626841295, 1755: 9.09459906466, 1756: -0.809968268732, 1757: -30.1236603469, 1758: 15.287308061, 1759: -6.63431773177, 1760: -33.597119406, 1761: -17.7970844428, 1762: -2.84336734926, 1763: -10.8207584144, 1764: -11.6377615757, 1765: 4.81408126743, 1766: -1.48490468029, 1767: -25.9574647383, 1768: 3.35754852731, 1769: 9.04143291327, 1770: -2.34383531122, 1771: -20.6151064172, 1772: 29.7601522308, 1773: -22.5764986748, 1774: -20.6177447347, 1775: 28.2941709125, 1776: -18.0404647882, 1777: 8.49348187884, 1778: 8.72480151547, 1779: -5.86315677708, 1780: 5.28868725411, 1781: -1.87901610578, 1782: 7.55596326699, 1783: 26.8633650633, 1784: -2.67805362739, 1785: -17.7483901231, 1786: 10.0123109053, 1787: -0.837463200167, 1788: -3.74130332905, 1789: -17.5319187406, 1790: -3.43027151571, 1791: 2.17241909182, 1792: -10.1700095552, 1793: 4.85877204136, 1794: -20.7959253901, 1795: -0.311482840227, 1796: 13.5909526777, 1797: 3.78475077804, 1798: -3.63357538877, 1799: -3.13639715399, 1800: 2.42315775588, 1801: 3.30709229237, 1802: 6.46288549297, 1803: 3.09409694495, 1804: -15.5248150804, 1805: 6.15800371914, 1806: 22.1867815945, 1807: 3.26747011923, 1808: 27.0501985734, 1809: 2.31692558431, 1810: 7.4270343554, 1811: 3.24183405191, 1812: 1.44020022865, 1813: 2.07687174685, 1814: -40.3603228981, 1815: 3.17104095856, 1816: 3.08463949203, 1817: 3.10252058397, 1818: 3.45236132947, 1819: -7.77867600072, 1820: 2.03052455178, 1821: -2.42501959851, 1822: -1.99138722602, 1823: 3.3526880616, 1824: 12.3319819915, 1825: -18.8956853124, 1826: 2.69285787946, 1827: -5.28831032267, 1828: 3.42080121551, 1829: 3.43928166197, 1830: 7.41209944957, 1831: -9.47335683078, 1832: -2.3528403825, 1833: -13.9217670144, 1834: -0.633404097824, 1835: 3.1788878594, 1836: 3.10058489234, 1837: 3.46067953832, 1838: -5.66556893463, 1839: 1.19407465309, 1840: 3.19331040326, 1841: -3.03522817323, 1842: 3.49774208998, 1843: 3.27910457193, 1844: 3.11025011446, 1845: 3.56842346441, 1846: -0.992629180099, 1847: 1.40550095947, 1848: 2.40893001307, 1849: 3.36745780193, 1850: 3.27169483344, 1851: -22.9446437473, 1852: 3.33116459794, 1853: 3.28592032589, 1854: -0.847830377505, 1855: 2.80322903621, 1856: 0.979939495021, 1857: -3.64014595331, 1858: -28.7308233978, 1859: -42.6120587718, 1860: -1.5699254541, 1861: 18.9167453032, 1862: -14.032668486, 1863: 6.80935572421, 1864: 39.3965329422, 1865: -4.68654699456, 1866: 12.3040088276, 1867: 2.95354841961, 1868: 11.5447003395, 1869: 7.09809563122, 1870: 14.9646603578, 1871: 0.863119253059, 1872: 0.997814677392, 1873: 3.23847782633, 1874: 3.24280018931, 1875: 0.685677236697, 1876: 3.24030008716, 1877: 0.835294941538, 1878: -1.18003902641, 1879: 0.212057818124, 1880: 0.898976659732, 1881: -16.3995046007, 1882: 1.35528893224, 1883: -10.4247842865, 1884: 0.911063614287, 1885: 1.31277219641, 1886: 5.15856838467, 1887: -3.3409134715, 1888: 23.3974882478, 1889: 0.501622626593, 1890: 13.8843144984, 1891: 0.733281983665, 1892: 0.79511496258, 1893: 2.33768700376, 1894: 0.818661607401, 1895: 3.3796080994, 1896: 3.0716824199, 1897: 0.918556978217, 1898: 0.781384015013, 1899: 0.740404339333, 1900: 0.771876215581, 1901: 3.06410051394, 1902: -0.621946761514, 1903: 0.755997277144, 1904: 1.3126535259, 1905: 1.23964429513, 1906: -0.38426672417, 1907: -1.09959333477, 1908: 1.30749212509, 1909: 4.2294258722, 1910: 2.59055429754, 1911: -19.6520954381, 1912: -0.198150310223, 1913: -1.45971843689, 1914: 0.302331224129, 1915: 3.69659986238, 1916: 2.24517619954, 1917: 0.933766117677, 1918: 19.8628088126, 1919: 0.910161355949, 1920: -14.2608206844, 1921: -20.4816413317, 1922: 28.4403695622, 1923: 1.63732447412, 1924: 0.674878509895, 1925: 0.926418025628, 1926: 0.96726977974, 1927: 1.1605495752, 1928: -2.19339427228, 1929: -14.3698669686, 1930: 0.253200037295, 1931: 4.14375416666, 1932: 5.97329592909, 1933: 2.42547455303, 1934: -9.18886425544, 1935: -3.67364624708, 1936: 3.10617247964, 1937: 4.4392838485, 1938: 0.819700128327, 1939: 5.41485993784, 1940: 0.272023832671, 1941: 2.69668493846, 1942: -6.28637459466, 1943: 17.5437223214, 1944: 0.815892764086, 1945: 3.44706212564, 1946: 0.930829942104, 1947: -5.20733159212, 1948: 0.865868759879, 1949: 0.793609499875, 1950: -11.8763230227, 1951: 3.42109752786, 1952: 4.78636875916, 1953: 0.695139201162, 1954: 0.871265994244, 1955: -1.90746231937, 1956: 0.77276409496, 1957: 1.07709955364, 1958: 13.3459796925, 1959: -34.8384675125, 1960: -1.00802682746, 1961: 1.13415338856, 1962: 10.0967909243, 1963: 2.27909344249, 1964: -4.62103870977, 1965: -1.71081720586, 1966: -2.66471717195, 1967: 4.05360975592, 1968: -1.86094301631, 1969: 3.41240862655, 1970: 1.92357708226, 1971: 1.09403211213, 1972: 31.1498413845, 1973: 2.94077583128, 1974: -5.23605753152, 1975: -0.412256468969, 1976: 2.63626707818, 1977: 1.26751729373, 1978: -5.27356350534, 1979: 5.27931839862, 1980: 3.20941069352, 1981: -1.84437179159, 1982: 4.72934170101, 1983: 3.27289469002, 1984: 4.34500510147, 1985: 3.28372126152, 1986: -8.50132785287, 1987: 0.694130252699, 1988: 2.30150591879, 1989: -0.982058885283, 1990: -6.95688634541, 1991: 20.8507901906, 1992: 15.0234116974, 1993: -0.103715409404, 1994: -6.25703473026, 1995: -7.15924528082, 1996: -17.1095453134, 1997: -0.535065447537, 1998: -5.95500422069, 1999: -19.8768377118, 2000: 4.03596206637, 2001: 3.43581961738, 2002: -8.30091405903, 2003: 3.21690137353, 2004: 3.33626770077, 2005: 2.83549003814, 2006: 3.14953845381, 2007: 3.447368994, 2008: 3.85856511097, 2009: 2.3283276324, 2010: 3.27344946481, 2011: 0.898049648833, 2012: 0.849534970763, 2013: -7.71112600809, 2014: -2.89296794539, 2015: -1.15501410566, 2016: -1.50462608031, 2017: -7.55467227301, 2018: 2.38044320128, 2019: -2.69089138231, 2020: -26.3274994524, 2021: -23.4714278899, 2022: -6.33253595935, 2023: 0.814807952207, 2024: 3.9842735789, 2025: 2.9575276898, 2026: -0.912572430871, 2027: -17.0534305805, 2028: -22.0219381726, 2029: -24.7668947852, 2030: 14.7120943537, 2031: -30.9575946341, 2032: 3.43170973941, 2033: 4.31413739008, 2034: 3.17211578926, 2035: 12.2109575421, 2036: -0.220466463448, 2037: -17.912610826, 2038: 11.4532962251, 2039: -14.1277282821, 2040: 7.40656202913, 2041: -2.21903042665, 2042: -7.28982189484, 2043: 1.21017723559, 2044: 3.36496323652, 2045: 3.92076728189, 2046: 1.65012969864, 2047: -23.4936359054, 2048: -1.41042472596, 2049: 21.160472831, 2050: -14.1905162236, 2051: 3.24086706193, 2052: -15.0983699935, 2053: 3.32322504244, 2054: 3.34584048441, 2055: 3.49667542736, 2056: -13.7650736294, 2057: 3.37093493152, 2058: 3.23294177753, 2059: 9.50102973588, 2060: 3.30024025106, 2061: -5.00789257955, 2062: 3.25606846061, 2063: 3.48019998868, 2064: 8.16257568804, 2065: 3.14424283277, 2066: -0.279096012883, 2067: -2.46223958201, 2068: 29.1148277026, 2069: 6.10736016956, 2070: -14.9081221761, 2071: 6.30934331639, 2072: -2.25295137015, 2073: 11.2751738014, 2074: -2.67081189996, 2075: -10.8222897975, 2076: -4.81894774704, 2077: -1.69064184594, 2078: 3.1542910116, 2079: 2.98153230343, 2080: 9.32409044879, 2081: -0.506078462087, 2082: 7.88955308231, 2083: -6.11612553995, 2084: 3.32515462532, 2085: -2.39101052616, 2086: -12.2206501427, 2087: -29.4499307938, 2088: 15.379701622, 2089: -1.40190458767, 2090: 3.79773051926, 2091: -16.4618277807, 2092: 14.2836740135, 2093: -4.6696080388, 2094: -14.077490117, 2095: -3.94450363906, 2096: -32.494956617, 2097: -7.78085132482, 2098: 62.7194547888, 2099: 6.14172788236, 2100: 21.1147666519, 2101: 17.5330597661, 2102: 20.4528643818, 2103: 29.2332400085, 2104: 100.0, 2105: -13.0719507651, 2106: -46.4265146365, 2107: -1.23564907386, 2108: 41.7150548528, 2109: -32.3549755999, 2110: -53.6858642035, 2111: 70.2217661589, 2112: 51.6939021884, 2113: -82.5715081831, 2114: 40.287199414, 2115: -94.7722966111, 2116: -32.0178070909, 2117: -28.419405706, 2118: 93.854302059, 2119: 35.1565024519, 2120: 80.9768528368, 2121: -100.0, 2122: -37.386751482, 2123: -2.61574451009, 2124: -15.4751618306, 2125: 74.3992916263, 2126: -100.0, 2127: 45.9803835165, 2128: 34.3611635706, 2129: 39.2918344511, 2130: -26.8340580654, 2131: 29.8979275849, 2132: -27.9640178562, 2133: 30.5944452504, 2134: -44.3938692462, 2135: -18.0300352863, 2136: -13.5980989674, 2137: 1.48335798501, 2138: 35.5112560896, 2139: -21.0588212058, 2140: -16.031979007, 2141: 1.8796739421, 2142: -12.0675885064, 2143: 98.228776366, 2144: -4.50912043328, 2145: -10.1451232024, 2146: -11.2628723011, 2147: -12.434736709, 2148: -22.6712265767, 2149: 30.2765091532, 2150: -14.6956141803, 2151: 20.8139241706, 2152: -7.85288781974, 2153: -27.1811480689, 2154: 78.0324587995, 2155: -33.9683546478, 2156: 64.2378180341, 2157: 77.179383584, 2158: 2.90532429936, 2159: -12.4272290719, 2160: -7.00811567491, 2161: 35.285181764, 2162: -13.5103891268, 2163: 66.8958776852, 2164: -13.7808767894, 2165: -13.8606082274, 2166: -34.3827927819, 2167: -13.906798212, 2168: -13.535091505, 2169: 4.25350340187, 2170: 1.45549832884, 2171: 8.33702751363, 2172: -12.5431367956, 2173: -83.6407776143, 2174: -8.63158354266, 2175: -9.02686167967, 2176: 41.1782660997, 2177: -13.4637890744, 2178: -12.6677857597, 2179: 4.20072228183, 2180: -31.462057438, 2181: -10.6981619948, 2182: 48.4301400755, 2183: -21.4458763675, 2184: 13.4034652711, 2185: -7.37686809972, 2186: -13.6535448714, 2187: 18.8024484406, 2188: -14.3318366593, 2189: -9.77714506663, 2190: -9.59616107816, 2191: -12.5873402072, 2192: -14.0627351986, 2193: -22.3261473132, 2194: -8.81401539729, 2195: 26.1769664668, 2196: -5.01670690533, 2197: 17.0777819752, 2198: -7.35805585997, 2199: 1.16962490321, 2200: 4.6206574002, 2201: -13.8556642261, 2202: -14.8891699877, 2203: -21.731080206, 2204: 11.1221624619, 2205: 59.4413981308, 2206: -28.1236084723, 2207: 25.7041336152, 2208: -21.7926188004, 2209: 66.5074521029, 2210: -30.6423578114, 2211: 37.1844221859, 2212: 19.7597183492, 2213: 100.0, 2214: -8.25338303488, 2215: 16.8087932215, 2216: -17.5244781535, 2217: 63.8502036396, 2218: 90.5845784539, 2219: -12.5043612612, 2220: -3.52650824378, 2221: 2.77999998615, 2222: 1.15226906979, 2223: 11.3552258384, 2224: -4.09065128147, 2225: -13.4632461123, 2226: -10.6639104962, 2227: -21.917974146, 2228: -4.52983532218, 2229: 3.34724057611, 2230: 41.1037179053, 2231: -2.28796926154, 2232: -50.9620658783, 2233: 1.58993400707, 2234: -6.6456630318, 2235: 39.4445575168, 2236: -14.3783992295, 2237: -8.41168507079, 2238: -5.97231347253, 2239: -19.5810879886, 2240: -4.08968082991, 2241: -5.54964415845, 2242: 92.007797244, 2243: -4.23537732902, 2244: -13.2797263633, 2245: 0.560103215137, 2246: -7.73009070919, 2247: -3.8500647001, 2248: -3.92279486992, 2249: -4.94393295546, 2250: 6.09485614295, 2251: 28.6177706704, 2252: -5.49486862273, 2253: -10.7474027742, 2254: -4.23971260311, 2255: -13.6662227186, 2256: -13.6003459446, 2257: 2.58763761533, 2258: -25.2472182342, 2259: -61.9213356403, 2260: 20.0940370385, 2261: 16.2293078698, 2262: -4.75755963254, 2263: 1.21261618338, 2264: -7.90469540926, 2265: 21.6917855141, 2266: -4.78306751695, 2267: -16.4638291185, 2268: -4.0817227831, 2269: -19.8823873341, 2270: 22.2386710141, 2271: 39.5354368852, 2272: -7.12218564371, 2273: -4.21140695114, 2274: -5.39514953042, 2275: 1.48428739232, 2276: -8.22290816101, 2277: 35.0747369218, 2278: 19.0341942583, 2279: 14.5630300411, 2280: -91.7623903447, 2281: -20.5486701261, 2282: 2.03713537743, 2283: -1.04051080871, 2284: -27.7558583356, 2285: 35.3607106691, 2286: -13.0097489206, 2287: -4.61256273288, 2288: 12.3157002859, 2289: -7.74674786691, 2290: -4.76882955782, 2291: 2.0282503095, 2292: -55.8809282727, 2293: -3.91599657167, 2294: -13.5540614055, 2295: -1.26366599191, 2296: 16.4872129977, 2297: -3.64485737772, 2298: -3.82299873551, 2299: 8.69799817491, 2300: -13.2486577954, 2301: 21.1242644979, 2302: -3.78593829303, 2303: -3.8066378316, 2304: -2.59925520087, 2305: -3.67287714088, 2306: -11.7888341445, 2307: -1.53068601285, 2308: -0.719432168166, 2309: -5.73583187914, 2310: -5.50808521086, 2311: -15.6574710479, 2312: -3.10243131316, 2313: -12.4795934205, 2314: -15.8337537543, 2315: -5.66005795017, 2316: -12.4286224527, 2317: 8.4665547094, 2318: -13.2807544273, 2319: 33.2169476018, 2320: -4.67143736259, 2321: -19.3569631974, 2322: 7.97454380684, 2323: -5.69019254532, 2324: -7.92772351379, 2325: -53.7539852785, 2326: -6.04821415733, 2327: 80.184984924, 2328: -12.1874923213, 2329: -16.1387183821, 2330: -11.0253747423, 2331: -6.25235265598, 2332: 4.66584014293, 2333: -12.0891459813, 2334: -13.6726251081, 2335: 27.6909261001, 2336: 31.0202643888, 2337: 27.7613958953, 2338: -69.1671541951, 2339: -16.9186973299, 2340: -39.3279741534, 2341: -62.2489862227, 2342: -16.3017677797, 2343: -15.7937644716, 2344: 16.9763746903, 2345: 44.399208367, 2346: -43.1818187258, 2347: 30.5872562395, 2348: 39.5740966936, 2349: 0.16128162191, 2350: 1.65787638514, 2351: 91.7825230061, 2352: -6.04535090261, 2353: -13.2734975107, 2354: 5.54341376127, 2355: 37.7105706813, 2356: -11.3755905303, 2357: -4.62363426545, 2358: 2.78385266135, 2359: 6.57520423667, 2360: -0.914805235863, 2361: 73.0361420292, 2362: 5.28525892583, 2363: 16.9405816521, 2364: -13.5918127496, 2365: -13.4124365908, 2366: 35.2552953748, 2367: -22.1064115794, 2368: 10.7391610187, 2369: -16.1722284898, 2370: -29.6347245018, 2371: 7.80511845552, 2372: -14.2179040962, 2373: -4.99650249049, 2374: -16.7108099155, 2375: -9.86845997241, 2376: 100.0, 2377: -22.6619039781, 2378: 17.3296630964, 2379: 20.9793021237, 2380: 47.3516051695, 2381: -13.9448255143, 2382: 4.69811164379, 2383: -19.2065854647, 2384: 25.7636318336, 2385: 43.3277283704, 2386: 36.7197307375, 2387: 78.0365941013, 2388: 26.7324069249, 2389: 93.0092767488, 2390: 26.9108528551, 2391: 3.97768931582, 2392: -35.3041698099, 2393: 12.9039432361, 2394: 28.222831368, 2395: -9.96829687782, 2396: 72.0364448968, 2397: -14.1236889573, 2398: 4.35908738202, 2399: 100.0, 2400: -9.78590882557, 2401: -7.12651920825, 2402: -13.6398592617, 2403: -13.6861709163, 2404: 3.41334329626, 2405: 33.2803234166, 2406: -14.2956593445, 2407: -13.9218383453, 2408: 8.65229654421, 2409: -13.3202149422, 2410: 44.3781883828, 2411: -14.2331624302, 2412: -13.9873021782, 2413: -72.6878391245, 2414: -1.0462998328, 2415: -9.59354354243, 2416: 3.59415829112, 2417: 6.24956388562, 2418: 10.3100006781, 2419: 46.4789754784, 2420: 0.102426232795, 2421: 13.7124666689, 2422: -36.7301127242, 2423: 31.6489274614, 2424: -41.3254047169, 2425: 93.9522568923, 2426: -46.6068116216, 2427: -20.0098023245, 2428: 24.7244503231, 2429: 9.63351854788, 2430: 40.3664386562, 2431: 21.806416949, 2432: -0.41469890486, 2433: -14.747554401, 2434: 45.3459926176, 2435: 45.5332981854, 2436: -3.61575086034, 2437: 48.1373575588, 2438: -99.6233645177, 2439: -10.0565127095, 2440: -8.50744107842, 2441: 11.7877548646, 2442: -1.75181155117, 2443: -19.6059690231, 2444: -0.102397947425, 2445: -26.435635918, 2446: -58.9170904628, 2447: -49.9721551899, 2448: -27.9621438426, 2449: 53.4695543496, 2450: 14.4820887513, 2451: -23.6929884135, 2452: -3.57441161506, 2453: -40.5033219637, 2454: 26.4441763423, 2455: 12.6403839248, 2456: -8.55703730635, 2457: -7.71784104618, 2458: -14.4722956105, 2459: -4.22434720133, 2460: 23.3491912083, 2461: -20.9621692956, 2462: -29.4041020941, 2463: 33.5558256825, 2464: -26.5073127974, 2465: -14.3681582157, 2466: -10.28312373, 2467: 5.38573629148, 2468: 10.484392619, 2469: 37.154740339, 2470: -1.31073851622, 2471: 15.7785865448, 2472: -45.5151188433, 2473: -66.6554620321, 2474: -18.7551605806, 2475: 75.879748132, 2476: -32.166786048, 2477: -21.5309580246, 2478: -25.1259801618, 2479: -26.7390586857, 2480: -24.587737051, 2481: -17.9102452127, 2482: -38.008838144, 2483: 38.8392965169, 2484: -10.8971349028, 2485: -0.937751357574, 2486: -26.4006179613, 2487: -31.9511064825, 2488: -16.9542772213, 2489: 11.5452065114, 2490: 10.276227586, 2491: 0.555687690712, 2492: -9.93274961509, 2493: -1.4839781696, 2494: -12.0016216086, 2495: -18.9198295997, 2496: -0.993804093395, 2497: -9.61123759785, 2498: -4.88038180169, 2499: -0.557494310612, 2500: -15.4008584086, 2501: -0.573970334667, 2502: -2.70355789374, 2503: 10.3226645548, 2504: -34.1731761821, 2505: 15.902513425, 2506: -29.8417423977, 2507: 11.9111019796, 2508: 0.202562825566, 2509: 0.61952246141, 2510: 8.28782075713, 2511: -0.83059231614, 2512: 7.82190435105, 2513: 9.92166986488, 2514: 0.682646439454, 2515: 25.4785994518, 2516: -0.614793650087, 2517: -31.0824199656, 2518: -7.29128563386, 2519: -18.2115940705, 2520: -23.5299338735, 2521: -23.0341202847, 2522: -19.1434304831, 2523: 25.6520564707, 2524: 1.51117144018, 2525: -2.68630589427, 2526: -0.741924966341, 2527: 60.5583415526, 2528: 7.87879499885, 2529: -9.15702189775, 2530: 33.9210827524, 2531: -6.93777537065, 2532: -0.769949289321, 2533: -4.35449685273, 2534: 30.5217138425, 2535: -0.696007546579, 2536: -17.5144959478, 2537: 3.95679814002, 2538: 0.468253522879, 2539: -1.27450746533, 2540: 37.691600463, 2541: 1.16996043942, 2542: 21.5116054131, 2543: 2.51494775547, 2544: 47.2731519352, 2545: -1.01060415906, 2546: -16.4860550704, 2547: 2.25397788064, 2548: -0.668147342513, 2549: 14.9195797486, 2550: -0.766230434605, 2551: -0.637813879569, 2552: 10.5720458558, 2553: 11.082297851, 2554: 4.3630222496, 2555: 18.2954641402, 2556: -7.43070335904, 2557: -38.2734380828, 2558: -19.0254253416, 2559: -18.7112401556, 2560: 4.73290669776, 2561: -20.1777237768, 2562: -40.0883613772, 2563: -82.2840873005, 2564: 28.6129275474, 2565: 20.2399445274, 2566: 38.7756494157, 2567: 12.753419673, 2568: -51.946216745, 2569: -0.224134968507, 2570: 0.46898736759, 2571: 1.82684731526, 2572: -3.73509213904, 2573: -0.014691944912, 2574: -0.5417254934, 2575: 1.84490958833, 2576: 38.035998888, 2577: -7.14631249678, 2578: -2.71963905461, 2579: 5.76684955044, 2580: 1.27620690948, 2581: -15.2873353095, 2582: 7.85772570986, 2583: -1.25204442301, 2584: -91.614891443, 2585: -4.13670603655, 2586: 9.66814502223, 2587: 1.1831263904, 2588: 11.0386028168, 2589: -1.06024235664, 2590: -0.00228396780891, 2591: -20.7689873461, 2592: -0.0228999455097, 2593: -0.783528145056, 2594: -3.30508885329, 2595: -0.0264053738184, 2596: -0.0198549408006, 2597: -0.154305066797, 2598: -0.0778453449607, 2599: 0.534237062905, 2600: -7.58654059719, 2601: -6.30431338347, 2602: -0.278867362696, 2603: -0.787288289701, 2604: -1.3092167598, 2605: -1.62660928783, 2606: 21.2622995257, 2607: -1.97621974938, 2608: 16.0826505709, 2609: -47.5154807594, 2610: 10.6891679395, 2611: 7.25763520012, 2612: 17.4060539981, 2613: -4.79512290574, 2614: 68.2419995867, 2615: -0.0806857240186, 2616: 63.230340975, 2617: -0.322155497282, 2618: -19.5239992182, 2619: 23.4973266006, 2620: -4.1465167197, 2621: -0.0834830722802, 2622: 0.0162814582659, 2623: 0.251474590206, 2624: -0.578122737075, 2625: 0.40714576181, 2626: -12.3234207611, 2627: -36.6118773046, 2628: -15.1095860799, 2629: 50.2657212068, 2630: 21.2418388508, 2631: -1.30390206608, 2632: 14.5296472831, 2633: -17.522155943, 2634: -2.46656900514, 2635: 37.7739981893, 2636: 0.0605604610376, 2637: -74.8177802564, 2638: -3.53821074438, 2639: -1.15161138413, 2640: -4.37754516149, 2641: -26.3774517376, 2642: -0.0737377253783, 2643: 1.56394162263, 2644: -0.213738726773, 2645: -19.9772972942, 2646: -0.0857792552133, 2647: -0.191914669339, 2648: 18.6883565178, 2649: -0.904876682801, 2650: -23.1461737458, 2651: -0.216097842096, 2652: -0.199616457895, 2653: 2.27998588998, 2654: -0.241845564537, 2655: 17.2397633051, 2656: 11.2559305498, 2657: 26.0677646907, 2658: -0.630056475057, 2659: 7.17258802671, 2660: -0.735643196296, 2661: 15.710617963, 2662: -29.532361724, 2663: -23.4480397263, 2664: -9.46544569532, 2665: 17.9512858059, 2666: 16.0959938614, 2667: -0.279648770824, 2668: 21.7290109821, 2669: 0.030402142309, 2670: 37.0565737024, 2671: -9.22074648131, 2672: -0.705646619952, 2673: -1.56975314853, 2674: -27.1988177862, 2675: 1.58729530686, 2676: -57.1188057632, 2677: 18.5762928731, 2678: 2.0921209399, 2679: 19.9358217884, 2680: 20.6156957632, 2681: 3.68142478307, 2682: -3.85288438086, 2683: -0.61972119828, 2684: 17.1655321352, 2685: -14.6735554879, 2686: 22.0398110489, 2687: 2.20087189384, 2688: 15.0957501773, 2689: -4.11020996739, 2690: 10.3317181183, 2691: -40.0538242301, 2692: -79.9001541952, 2693: 97.4411608949, 2694: 25.1945501198, 2695: 33.8853785542, 2696: -7.65283818219, 2697: -8.41360920013, 2698: 21.2038779604, 2699: -0.865708211672, 2700: 29.3983747356, 2701: -0.643117530018, 2702: -0.801577629477, 2703: -3.44596779333, 2704: 2.9124661132, 2705: -0.815304460934, 2706: -5.2988634509, 2707: -0.490374692228, 2708: 0.710605531654, 2709: 99.8834166305, 2710: -13.7783829084, 2711: 1.30163021381, 2712: 13.8340749632, 2713: -1.51077529032, 2714: -1.21209468447, 2715: -14.0274660481, 2716: -10.234661592, 2717: -37.875528599, 2718: 0.304416060374, 2719: -11.8039956114, 2720: 48.4631667161, 2721: 11.5570765841, 2722: -7.03877222346, 2723: -17.0971430071, 2724: -2.1992899769, 2725: -79.5858315954, 2726: -36.9577979078, 2727: 16.5564569908, 2728: 1.41431616679, 2729: 10.8468485713, 2730: -0.683978702444, 2731: -16.4902002085, 2732: 2.84380046679, 2733: -0.147068069908, 2734: 3.74651892939, 2735: -6.21391217868, 2736: -1.0028778446, 2737: -9.00284001475, 2738: -93.0921346817, 2739: 26.3930284019, 2740: 8.03633420624, 2741: 34.0289556357, 2742: -0.41433270088, 2743: 28.4821757856, 2744: -1.7728423368, 2745: 24.9449149672, 2746: -14.0662682504, 2747: 37.2282415126, 2748: -43.0697403453, 2749: 0.449210464281, 2750: 18.190995369, 2751: 1.49908380498, 2752: 1.50036955508, 2753: -1.76248246327, 2754: 1.31918206109, 2755: -0.362014045085, 2756: -0.803227541397, 2757: 58.2519881348, 2758: -0.86045267163, 2759: 8.65191761386, 2760: -0.835070772229, 2761: -0.703120983991, 2762: 9.76978236589, 2763: 2.68188538621, 2764: 25.2064904362, 2765: -44.863714436, 2766: 26.9243103097, 2767: 30.9648068427, 2768: -56.0305641095, 2769: 4.19751299212, 2770: 14.8132016662, 2771: 28.6836967825, 2772: 68.1593890857, 2773: -12.1850835751, 2774: -12.4611365799, 2775: -6.52257146316, 2776: 0.808529387281, 2777: 4.79577913888, 2778: -20.4877281939, 2779: 37.1676949897, 2780: 30.9288930095, 2781: -4.84473960472, 2782: -0.350750948759, 2783: -28.40513169, 2784: 14.2594140628, 2785: 41.5812761305, 2786: 5.19170855677, 2787: 14.7105790083, 2788: -21.1677515598, 2789: 83.570676256, 2790: -100.0, 2791: -37.4410408715, 2792: 34.8823397434, 2793: 0.488628395744, 2794: -0.354111210463, 2795: -0.166011794212, 2796: -1.45511882819, 2797: -1.95400145162, 2798: -9.28348504415, 2799: 6.99834606434, 2800: -0.322865278318, 2801: -1.61812767419, 2802: -5.1818634487, 2803: 5.01798763936, 2804: -1.93393653844, 2805: -2.8791637358, 2806: -1.39039614543, 2807: 7.83260718658, 2808: 3.41369709255, 2809: 2.70930080395, 2810: -1.04502420811, 2811: -25.5304906367, 2812: -26.7534587904, 2813: 0.951500648223, 2814: -3.53301664569, 2815: 1.31902706759, 2816: -2.70453010194, 2817: 1.88007136692, 2818: 2.47642399929, 2819: 7.55656408913, 2820: 1.58846164711, 2821: -15.2383983239, 2822: 15.8017816669, 2823: -2.48749980889, 2824: 16.2616760096, 2825: 6.38722436289, 2826: 5.09337382743, 2827: -10.4013034997, 2828: 3.27152786489, 2829: -4.77186927614, 2830: 5.15947864928, 2831: 7.06669218265, 2832: 0.231603630511, 2833: 2.89079594567, 2834: -1.19915299843, 2835: 11.5408960173, 2836: -0.37126649468, 2837: -6.5336791936, 2838: 5.9597286935, 2839: 7.45851259422, 2840: 3.0473034163, 2841: 4.9134306447, 2842: -2.8131283873, 2843: -2.514092669, 2844: 25.3545696388, 2845: -8.03418051524, 2846: 0.297553949823, 2847: 2.20086212191, 2848: 1.85827476733, 2849: 43.9653796052, 2850: 4.56590917204, 2851: -16.4132980188, 2852: -18.7095282647, 2853: -2.97249947595, 2854: 2.31199768453, 2855: -7.95649711943, 2856: 5.61850920566, 2857: 4.19386654576, 2858: -0.26982991987, 2859: 1.8277941643, 2860: 2.30031591825, 2861: 3.29355905113, 2862: 2.96924172004, 2863: 0.828486943865, 2864: -9.96015115226, 2865: 3.09252747177, 2866: -22.0163138219, 2867: -2.22184475312, 2868: 5.85351283516, 2869: 5.59856309406, 2870: 7.23660069654, 2871: 2.67599828152, 2872: 12.9667009472, 2873: -4.16467709981, 2874: -9.00761015196, 2875: -6.25756342874, 2876: -8.13278301144, 2877: -5.37733370566, 2878: -8.15569368135, 2879: -4.97007760619, 2880: 2.50902942818, 2881: -1.18677593227, 2882: -2.36361695767, 2883: 2.64009869428, 2884: 2.45693973681, 2885: 8.12068739376, 2886: 2.77146937482, 2887: 0.786358930921, 2888: 4.15470423597, 2889: 14.8600023301, 2890: 2.59869784247, 2891: 1.33651686242, 2892: 1.75644865034, 2893: 3.06035171694, 2894: 2.72157493089, 2895: 0.101949252477, 2896: 0.550903748375, 2897: 3.52332833247, 2898: -16.4544761529, 2899: 2.2201824013, 2900: -3.55163566361, 2901: -1.34871832526, 2902: -5.2879814299, 2903: -0.407635295723, 2904: 0.244502717719, 2905: 24.547983326, 2906: -1.16371464069, 2907: -17.4835444697, 2908: 2.4608228746, 2909: 4.38100472952, 2910: 0.965730523273, 2911: -11.3141811942, 2912: 1.26113571729, 2913: 3.88449014595, 2914: 6.0998879752, 2915: 4.64420352075, 2916: -1.80341052351, 2917: -8.45188449399, 2918: 0.614874844505, 2919: 1.25988835661, 2920: 0.248218206488, 2921: 21.6059370664, 2922: 0.50043676832, 2923: -0.51161736486, 2924: 1.05273171435, 2925: -4.55941653332, 2926: -3.2750614257, 2927: -5.38222712947, 2928: 11.7264346432, 2929: 0.127450112969, 2930: 13.7881907082, 2931: 1.52104267928, 2932: -1.03300702519, 2933: 6.02794112681, 2934: -5.2970161341, 2935: 3.49151045801, 2936: -0.7866948609, 2937: 1.32529187516, 2938: 0.547416564226, 2939: 0.451768407339, 2940: -1.5021283878, 2941: 0.574596605898, 2942: -0.222423621133, 2943: -6.6422127378, 2944: 0.87227484829, 2945: 0.766927427496, 2946: 0.73323975516, 2947: 0.683700980927, 2948: -3.34859414607, 2949: -11.4363130499, 2950: 0.717101904788, 2951: 1.03904779273, 2952: 0.908859788965, 2953: 3.22693179874, 2954: 2.48639221893, 2955: 30.1865139941, 2956: 5.32128822375, 2957: -54.6802761271, 2958: 48.3719144691, 2959: 5.2796786723, 2960: 3.65684616112, 2961: 4.91760053914, 2962: 0.261478218481, 2963: -15.1009869291, 2964: 0.707770365098, 2965: -5.99492148551, 2966: 0.464255297625, 2967: 4.26707635307, 2968: 0.151030068353, 2969: 7.67324561056, 2970: 1.85087182292, 2971: 0.559547043082, 2972: 0.598354491675, 2973: 0.551865972648, 2974: 0.699209001719, 2975: 0.18187588554, 2976: -40.9969112912, 2977: -6.7247509564, 2978: 13.3465081836, 2979: -5.93467646167, 2980: 6.54195178965, 2981: -4.82370140427, 2982: -5.61626214474, 2983: 9.14161946178, 2984: -16.1649842318, 2985: 0.467402264958, 2986: 13.8852244213, 2987: 2.85371715663, 2988: -3.12025670257, 2989: 5.98512742115, 2990: 0.176519655151, 2991: 0.596450364267, 2992: 1.7792477971, 2993: 0.715799249166, 2994: -4.31624753036, 2995: 0.598767909367, 2996: 0.969301341489, 2997: 15.9267543516, 2998: 3.43098412476, 2999: -2.04925530001, 3000: 0.558227479786, 3001: 0.453568318723, 3002: -2.0760408637, 3003: 0.662283399949, 3004: 2.74939677091, 3005: 4.53839886418, 3006: -17.2563361061, 3007: -0.247577933798, 3008: -0.887175648866, 3009: -35.7394420501, 3010: -8.04748227985, 3011: -2.23277023968, 3012: -22.1322322639, 3013: 3.56527299994, 3014: -4.25337109346, 3015: 8.21346481069, 3016: -0.386886026068, 3017: -15.6194957606, 3018: 0.797568743834, 3019: 10.5074083795, 3020: 1.37983693187, 3021: 0.792885908162, 3022: 1.16767793661, 3023: -8.3807249751, 3024: 1.39025121412, 3025: 10.8773917861, 3026: -20.7190232669, 3027: -1.78750973144, 3028: 22.4492056271, 3029: 11.183526034, 3030: -12.2803592916, 3031: 2.97381821319, 3032: -0.683627776427, 3033: 2.87882582511, 3034: 1.98179721732, 3035: 3.60850718079, 3036: -4.2157101437, 3037: -6.3386589482, 3038: 2.99631421608, 3039: -23.7342882425, 3040: 7.86350685708, 3041: -6.42600166868, 3042: -4.54857077916, 3043: -10.2292301244, 3044: 2.9722210329, 3045: 1.97564309848, 3046: -1.49817962351, 3047: 3.25131153125, 3048: 2.82655163126, 3049: -3.43284096573, 3050: 1.46624802967, 3051: -0.169707045125, 3052: -4.60169125524, 3053: -13.4065169496, 3054: -3.09804809132, 3055: 3.03751323908, 3056: 1.46348092558, 3057: -3.33771986809, 3058: -24.8776693473, 3059: -11.1596863148, 3060: -12.3945653456, 3061: 2.16603389484, 3062: 4.41726322795, 3063: 1.99831843128, 3064: 13.0408488301, 3065: 15.4416447145, 3066: 32.1006054649, 3067: -47.7807705082, 3068: -12.0041413918, 3069: 0.75449906259, 3070: -0.754885419512, 3071: 1.05652034688, 3072: -2.66013691948, 3073: -9.98117158037, 3074: 8.73466441011, 3075: -21.3070624946, 3076: -3.33300724484, 3077: -9.97559479361, 3078: 14.9698699557, 3079: 2.94570160551, 3080: 1.54934966371, 3081: 2.56692682521, 3082: -0.506767002915, 3083: -8.27453051876, 3084: -4.09053287612, 3085: -2.29735595634, 3086: -4.5579998359, 3087: 14.4499741759, 3088: 4.29981390244, 3089: -3.86274516606, 3090: -7.8573687219, 3091: 4.52206477636, 3092: -12.5931853843, 3093: 0.398485408075, 3094: -3.44837547942, 3095: 3.49068426654, 3096: -7.42695079002, 3097: -22.1494280059, 3098: 0.82104982447, 3099: -8.14506400443, 3100: 1.63212996913, 3101: 1.83263181569, 3102: 3.43664668217, 3103: -1.09657383109, 3104: 2.3262158211, 3105: -13.8123521795, 3106: 16.1043744549, 3107: 3.41778202589, 3108: -8.80725433625, 3109: 1.14131246286, 3110: 3.31310805989, 3111: -1.190682721, 3112: 0.581101688762, 3113: 2.76071279268, 3114: -7.62426480283, 3115: 19.2220799928, 3116: -13.6458395929, 3117: 4.43874297374, 3118: -6.08701943729, 3119: -4.31957985449, 3120: 0.43140304298, 3121: -24.8793444345, 3122: -7.18335369762, 3123: 0.08426983683, 3124: -6.26576011008, 3125: -3.81325113319, 3126: 2.64508541472, 3127: -6.07050468375, 3128: 10.6012604948, 3129: -5.56354676286, 3130: 18.2572355266, 3131: 2.93418148209, 3132: -13.160494887, 3133: 4.19692724834, 3134: -10.0595439225, 3135: -0.152320979771, 3136: 16.2810973799, 3137: 2.94325171359, 3138: -17.9676233441, 3139: 6.32243609278, 3140: 8.80854808001, 3141: -0.346269814057, 3142: 0.010116537948, 3143: 1.92897595569, 3144: 4.42622616251, 3145: 10.4682312275, 3146: -1.87972498721, 3147: -3.61975061791, 3148: -2.38850899456, 3149: 3.32314853923, 3150: -4.33922013488, 3151: -0.242482231004, 3152: -2.26646838033, 3153: -4.94191700872, 3154: 1.81683726081, 3155: 0.495286242623, 3156: 2.99402294313, 3157: 4.77781419929, 3158: -16.1895206087, 3159: -3.09351983425, 3160: -2.61934628326, 3161: -1.83931153177, 3162: 2.80884378762, 3163: -13.0653043664, 3164: 8.02456139696, 3165: 1.42866478958, 3166: -0.608107216626, 3167: -2.80412957001, 3168: -5.49794923944, 3169: 0.578832527985, 3170: 9.42295098478, 3171: 9.63472645893, 3172: 1.45646207503, 3173: -7.99492463095, 3174: -0.0928230882175, 3175: 4.11714269324, 3176: 0.90302752354, 3177: -1.20402995814, 3178: -1.34991234258, 3179: -2.63640193274, 3180: 8.64698953006, 3181: -4.3621472648, 3182: 6.80764411931, 3183: 0.582683367586, 3184: 1.58447724409, 3185: 4.37760720194, 3186: 4.64039940422, 3187: 1.66671234612, 3188: 5.21845041399, 3189: -2.41046285977, 3190: 2.83019877558, 3191: -0.43677040631, 3192: 7.08315527699, 3193: 4.34834104197, 3194: 0.0280560350238, 3195: 1.90445515229, 3196: 1.33073488616, 3197: 0.0595849579204, 3198: 0.346693934691, 3199: 0.977450259017, 3200: 7.46403805317, 3201: 0.560431780879, 3202: -10.4379705819, 3203: -7.7499516434, 3204: 11.7078926251, 3205: 3.64723502603, 3206: -0.889994958103, 3207: -1.35860116766, 3208: -12.9911028265, 3209: 0.803759199576, 3210: -3.47418676239, 3211: 0.678770943768, 3212: -5.61955090076, 3213: 12.4530387176, 3214: -0.0106337086367, 3215: 0.525501828057, 3216: 2.93212565964, 3217: -0.437814067728, 3218: 2.17833905781, 3219: 0.487206996403, 3220: -4.13887272859, 3221: 15.7140340133, 3222: 0.156582186417, 3223: -9.21563654857, 3224: 0.875502458327, 3225: 6.00405023588, 3226: -3.8128140006, 3227: 14.1901971031, 3228: -1.23463724276, 3229: -13.9627103278, 3230: 1.81116947416, 3231: -3.41913716207, 3232: -0.564477635476, 3233: 0.204716823656, 3234: -6.35798454128, 3235: 0.13942373022, 3236: 1.71821403981, 3237: -0.149413389301, 3238: -3.44542902725, 3239: 0.382477433212, 3240: 1.68735955329, 3241: -0.380587457245, 3242: -24.7939094593, 3243: 0.251452031063, 3244: -1.65492333743, 3245: -1.71298962269, 3246: -0.199186160507, 3247: -5.30464834777, 3248: 0.478097460836, 3249: -0.356659343239, 3250: 2.01707294697, 3251: -12.2454864945, 3252: -0.731981404313, 3253: -0.613916548223, 3254: 17.2494388912, 3255: 1.23580325184, 3256: 3.32697012348, 3257: 4.58660194012, 3258: -0.0252481273701, 3259: -0.976887374965, 3260: 2.37613594808, 3261: 0.0996982155954, 3262: -6.09525756939, 3263: 11.4279419112, 3264: -14.5321050175, 3265: 8.75123366638, 3266: -5.51482137058, 3267: 0.231783371558, 3268: 0.468344259508, 3269: -6.95471372526, 3270: 2.82521258621, 3271: 0.0660043481538, 3272: 0.296653748499, 3273: -0.259431756368, 3274: 10.6615730602, 3275: -2.63942539018, 3276: 1.39227902178, 3277: 1.07678238489, 3278: 0.601011042742, 3279: 3.65027603105, 3280: 3.3201643975, 3281: 2.60739566789, 3282: 4.31453076352, 3283: 7.57310020046, 3284: 4.11177720991, 3285: -1.5695080836, 3286: 6.42893583999, 3287: 0.119526032634, 3288: 0.186299174753, 3289: 4.99527225067, 3290: -0.0151538802851, 3291: 0.212484588767, 3292: -3.5884126476, 3293: 0.397556197847, 3294: -0.0480178095171, 3295: -0.130717559213, 3296: 0.0152148911902, 3297: 3.24980591644, 3298: 7.16130374381, 3299: 0.00509594702155, 3300: 0.157305476878, 3301: -0.306305453003, 3302: 8.90318796976, 3303: 2.82654119734, 3304: 7.80586237479, 3305: 1.40591220728, 3306: 3.058907637, 3307: 13.5932791569, 3308: -0.404687622899, 3309: 4.18005583148, 3310: 3.42599041091, 3311: 3.08203959496, 3312: 1.3635654778, 3313: -0.0312863822391, 3314: 0.953699070506, 3315: 0.104669898849, 3316: 18.2559611123, 3317: -14.8183339469, 3318: -0.246279681505, 3319: 0.407333603505, 3320: 0.0631745037551, 3321: -0.111524643056, 3322: -0.00838249031471, 3323: 0.0328415820847, 3324: 3.53502766701, 3325: -2.14221494028, 3326: -1.29930670712, 3327: -4.54108282754, 3328: 12.3808733315, 3329: -2.98923624428, 3330: 0.682241204979, 3331: -5.5939991843, 3332: -1.47518940614, 3333: -3.3012159602, 3334: -0.182791041706, 3335: 12.2301960521, 3336: 0.496073519062, 3337: -0.126005028867, 3338: -6.61714436846, 3339: 1.4115003088, 3340: 0.0160713516126, 3341: 0.298859519467, 3342: 0.00571636342002, 3343: -6.03158347523, 3344: 0.141530141111, 3345: 0.0778038273442, 3346: 6.10999296096, 3347: -0.0319532967604, 3348: 3.73345023789, 3349: -0.0902830900935, 3350: -0.0335641159672, 3351: -0.619450521359, 3352: 0.0408693866429, 3353: 0.351630748257, 3354: -3.33348448225, 3355: 8.27063448183, 3356: 0.668907543853, 3357: 0.50658162248, 3358: 7.56391426143, 3359: -2.54080618396, 3360: 0.966174600696, 3361: 11.8002719481, 3362: -1.3474079113, 3363: -0.546560763133, 3364: 3.19102166096, 3365: -1.04234415684, 3366: -1.18390389569, 3367: -0.0842650627358, 3368: -9.18262027921, 3369: -5.12348160942, 3370: 0.0110103890616, 3371: 4.32718127205, 3372: 7.70336191434, 3373: 0.591016322693, 3374: 8.95557822857, 3375: -6.69222530233, 3376: -2.75455793118, 3377: 1.81381236023, 3378: -4.91355110178, 3379: -0.471345269517, 3380: -2.36264050713, 3381: 0.231303986684, 3382: 4.86077392465, 3383: -11.7250126625, 3384: -5.00064452928, 3385: 7.21820950499, 3386: -7.40236356392, 3387: -9.58308708466, 3388: -4.05067528823, 3389: 14.2800508748, 3390: 10.0675329038, 3391: -3.6466711458, 3392: -3.75425015465, 3393: -4.0336472274, 3394: -4.39553398111, 3395: -6.65218494162, 3396: -2.21650853544, 3397: -1.68586409233, 3398: -9.90134948846, 3399: -1.65642596124, 3400: 0.0952484015843, 3401: 1.89376364127, 3402: -6.78694402455, 3403: 0.278780837489, 3404: 2.14858735196, 3405: -0.0786842179143, 3406: 3.26066694773, 3407: -5.37991950503, 3408: -9.25022343643, 3409: 2.21026403548, 3410: -2.48534428741, 3411: 10.2623035143, 3412: 3.02752266573, 3413: 8.17830521175, 3414: 7.11631357634, 3415: -12.8235962562, 3416: -7.87673342606, 3417: -5.83379950555, 3418: 1.60755468074, 3419: 4.01261474949, 3420: 8.52047727692, 3421: 6.68879960874, 3422: 1.81086436563, 3423: -1.6832313491, 3424: -6.21927534104, 3425: 2.77816552099, 3426: -2.05866445491, 3427: 2.42713432507, 3428: 0.0159678866848, 3429: 1.67353644369, 3430: 0.682391822261, 3431: -1.54783557422, 3432: 0.579573745901, 3433: 3.93858537993, 3434: 2.16163909636, 3435: -4.68483806402, 3436: -4.94396379427, 3437: 0.290796114566, 3438: -9.35110701857, 3439: -11.1851647315, 3440: 0.321731470267, 3441: -2.51749973383, 3442: -0.88619140318, 3443: -2.5049884924, 3444: 1.86623050495, 3445: 2.57562080222, 3446: -5.0433785755, 3447: 1.60654583086, 3448: 0.221616147834, 3449: 0.382871280024, 3450: 0.412381846696, 3451: -1.42399282261, 3452: 3.46760431224, 3453: -1.4007660426, 3454: 0.378664962181, 3455: 2.08000530449, 3456: 0.149382819049, 3457: 0.0855499575143, 3458: 0.150175791466, 3459: 0.231034619391, 3460: -1.29715251601, 3461: 0.46172568053, 3462: 0.117955931778, 3463: 0.0264966902321, 3464: -5.4095456331, 3465: -0.971050760724, 3466: -3.30183016117, 3467: -10.5519258439, 3468: -0.265390708947, 3469: 7.82295392482, 3470: -24.7303268548, 3471: 1.64062686118, 3472: 2.8113194138, 3473: 1.70356422143, 3474: -1.21625853349, 3475: -14.7991422776, 3476: -5.25511510931, 3477: 1.07081465859, 3478: -5.9580787919, 3479: 0.130429160664, 3480: 0.209573121629, 3481: -3.68291394821, 3482: 1.49702706961, 3483: -9.61419615677, 3484: -5.37045895836, 3485: 7.36402505673, 3486: -2.26043276217, 3487: -28.1702666735, 3488: -6.10130712897, 3489: 4.53210842466, 3490: -5.82634660281, 3491: 0.905480451934, 3492: 0.646029175446, 3493: -1.1559416365, 3494: 0.91565820728, 3495: -1.17807144188, 3496: 26.3143881289, 3497: -0.269615027809, 3498: 0.32871315816, 3499: -1.03388339217, 3500: 0.873528657686, 3501: -1.27446460926}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.378614841719, 2: 52.8119850558, 3: -5.31893507945, 4: -27.4537603368, 5: -36.4456013576, 6: 2.21919264469, 7: -2.3186281128, 8: 29.3727456502, 9: -28.6248326151, 10: 13.1960191275, 11: 22.2262334411, 12: -25.3366614641, 13: -5.68676250864, 14: -26.7492837623, 15: 12.0693655656, 16: -4.70007355036, 17: -36.0815919431, 18: 23.3179131767, 19: -5.28463659798, 20: 20.8259255792, 21: 28.3831859174, 22: 4.2720281119, 23: -7.22684263491, 24: -6.11654300769, 25: 22.3542197356, 26: -17.157959189, 27: 45.2383232693, 28: -13.7642290381, 29: -45.8455668574, 30: 4.97075361059, 31: 7.38012254154, 32: -47.2881970311, 33: 15.2864521143, 34: -3.1008771097, 35: -13.1987963376, 36: -1.4848625861, 37: -0.436753797693, 38: -8.52465038108, 39: 30.4869774195, 40: -29.9552700688, 41: -18.2789981613, 42: 3.04261763073, 43: -14.624973639, 44: -17.2976464564, 45: 3.03599339716, 46: -28.4753725926, 47: 20.6530060242, 48: 3.83445216667, 49: 5.21110782615, 50: 3.53987777419, 51: 0.738954377905, 52: 1.72377524516, 53: 1.91985652167, 54: -19.0442421043, 55: -13.8786111936, 56: 2.25681898728, 57: -27.3450244387, 58: -1.03401873839, 59: -36.5222508206, 60: -28.6771349822, 61: 37.2511528708, 62: -21.3041207815, 63: -36.2449134809, 64: -32.7154661681, 65: -4.1187305456, 66: 5.01739687542, 67: 1.75303206621, 68: 3.28917343929, 69: -39.3025278763, 70: 2.92040419186, 71: 2.74462688346, 72: 0.554581871426, 73: 2.41406320561, 74: -0.673586949236, 75: -0.415911974718, 76: 1.29947260138, 77: 0.259917215791, 78: 4.88263303302, 79: -6.79666111881, 80: -26.5609519916, 81: 1.21637868566, 82: 0.960066426526, 83: 0.975085373494, 84: 2.95274176006, 85: -36.1811255663, 86: 1.94945209509, 87: 1.24467554779, 88: 1.47754144145, 89: 2.54169919082, 90: -9.63162898062, 91: 2.54103632596, 92: 2.56490320166, 93: 21.7341983932, 94: 2.36188806422, 95: 2.87078835028, 96: 2.49806412666, 97: 2.81145017213, 98: 2.87384213179, 99: 2.8740287146, 100: 3.6600286087, 101: 2.08064604594, 102: 2.70400579508, 103: -0.868559581972, 104: 4.49562713093, 105: -10.7879362426, 106: -8.71403427731, 107: 2.80424413966, 108: 2.26456518614, 109: 2.45658115986, 110: 0.235748404071, 111: 1.96066612723, 112: 15.6163597444, 113: -4.0455563102, 114: 17.4470898315, 115: -33.0471887532, 116: -4.90194574394, 117: 33.7190515365, 118: -9.5354534674, 119: -46.191917802, 120: -50.8629384271, 121: -6.55845656041, 122: -4.0302971282, 123: 34.8226525732, 124: 3.76119149014, 125: 3.76925367919, 126: 0.40046389615, 127: 1.05464599758, 128: 3.20768311024, 129: 36.1860146878, 130: 0.582538481075, 131: 2.62148005124, 132: 0.49192129284, 133: -19.3070394587, 134: 0.589566411553, 135: -12.2100421787, 136: 18.2174293847, 137: 4.27024795528, 138: -7.73895182302, 139: -6.0271398371, 140: -11.2112716041, 141: -67.0323088684, 142: 4.23845924514, 143: 40.6227850407, 144: 1.26916787644, 145: 3.57990593841, 146: 0.513614587362, 147: 0.574608644893, 148: 9.39563300525, 149: 0.399635346963, 150: 2.47457459146, 151: -7.25330279248, 152: 0.796045165131, 153: 0.638743604326, 154: 0.783799055515, 155: 0.536698156431, 156: 2.8862867055, 157: 4.00377201442, 158: 2.22828859477, 159: 0.647015022091, 160: -0.408635048039, 161: 1.09136602772, 162: 1.73965690858, 163: -39.0682629413, 164: 3.2521127049, 165: 16.8014358952, 166: -34.8949816295, 167: 4.41284418285, 168: 3.14404256289, 169: 6.36709814007, 170: 3.18549445226, 171: 0.141054942544, 172: 0.449077704592, 173: 46.7477756161, 174: 0.471778473525, 175: 7.09632392225, 176: 7.0874950279, 177: 16.6575870843, 178: -0.470121744971, 179: 0.54894583885, 180: 0.563968322327, 181: 0.326058050692, 182: 0.926926981163, 183: -1.29751363426, 184: 8.68879556995, 185: -6.76824824491, 186: 9.55348355477, 187: -33.9497112446, 188: -2.32762637011, 189: 3.09914074873, 190: 12.3628477216, 191: 13.4646248544, 192: 2.93366186643, 193: 0.518069764973, 194: 4.27127199984, 195: 2.68304840327, 196: 1.38480620611, 197: 4.57735808146, 198: 7.98841453325, 199: 0.559703682747, 200: 2.57748532131, 201: 0.624991985271, 202: 22.2794449679, 203: 0.439442251288, 204: 0.675377732491, 205: 50.1774288949, 206: 2.18230686403, 207: -15.5971594498, 208: 0.602061473312, 209: 0.525044162881, 210: 0.254144523439, 211: 0.572101226254, 212: 2.00242536045, 213: 1.08776105671, 214: -20.3120868, 215: 1.82830717707, 216: 1.33685635546, 217: 1.77671120604, 218: 2.12141142065, 219: 17.5046052337, 220: 39.5343945113, 221: -1.72341550549, 222: 12.589244744, 223: 12.2082539176, 224: 3.19568197702, 225: 8.31708374382, 226: 0.568963557388, 227: -2.40870854099, 228: -1.21686538356, 229: -0.942344848684, 230: 6.89414993277, 231: 11.3028088023, 232: 0.785148766507, 233: -29.298205618, 234: -5.26779570501, 235: 4.72955402684, 236: -12.5324729516, 237: 14.246574835, 238: 8.22663400317, 239: 1.45473406329, 240: 2.69478636056, 241: -4.05282362806, 242: 21.7335140693, 243: -8.51204342328, 244: 14.5780378883, 245: 15.9446295574, 246: 9.44240996869, 247: -1.23403434119, 248: 14.1285166623, 249: 69.7202825634, 250: 35.1325718023, 251: 45.7261844654, 252: -20.4089275835, 253: -22.5891323199, 254: 3.78239414596, 255: 1.13982016757, 256: 2.22283068346, 257: -9.17664642923, 258: 4.91344768233, 259: 2.38048057065, 260: -6.052402199, 261: -4.91317648162, 262: 2.74751393831, 263: -2.83958733988, 264: 0.671385993748, 265: 3.06994694118, 266: -8.0324817359, 267: -15.3044494586, 268: 1.05563301183, 269: 0.703631165394, 270: -0.602782838102, 271: 2.70348482448, 272: -33.9198107799, 273: -9.4917455881, 274: -17.8143711981, 275: -37.0129220705, 276: -30.5727329102, 277: 4.84997794461, 278: 25.3528077529, 279: 2.61554762837, 280: 11.0303371905, 281: 0.332433018238, 282: -15.5881273555, 283: 0.883210884204, 284: 3.89720550265, 285: 4.83752537715, 286: 56.2857601612, 287: 2.99422394798, 288: 1.20305825064, 289: 2.73574546772, 290: -5.15220335053, 291: -3.58392955784, 292: -6.04189132812, 293: -32.5635209218, 294: 11.5663133175, 295: 60.910778832, 296: 23.2980043848, 297: 8.98640906643, 298: 23.3135554804, 299: 7.24328015173, 300: 25.89712179, 301: 1.65578955281, 302: 7.27597183929, 303: -3.66166625325, 304: -66.9749050363, 305: 13.7002761454, 306: 2.86572246586, 307: 1.91248870935, 308: 2.36095350603, 309: 2.37966321309, 310: 3.0220199237, 311: 3.87984216149, 312: 2.21420220553, 313: 2.16072664778, 314: 8.80018261226, 315: 2.22633152925, 316: -2.87116687574, 317: 2.30721897143, 318: 2.4156723993, 319: -22.9017537961, 320: 2.50821990157, 321: 1.25975379697, 322: 11.5187199301, 323: -33.1589680112, 324: -28.1251908483, 325: -31.5985061019, 326: -7.13224763857, 327: 3.1564100743, 328: 23.2447812523, 329: -17.5937121146, 330: 15.8776810834, 331: -13.9070426116, 332: -25.0596984673, 333: 2.40948904747, 334: 8.09144414039, 335: -3.04445129392, 336: 24.4426618078, 337: 0.0915158787556, 338: 62.3285234657, 339: 2.46044774572, 340: 0.905397372796, 341: 2.83134560595, 342: 25.0824058307, 343: 1.5571688673, 344: 35.8245530679, 345: -39.3513068181, 346: -100.0, 347: -73.3391633368, 348: -53.8790323396, 349: -16.6559959556, 350: 0.827908097484, 351: 0.187522108402, 352: -2.25947493267, 353: -13.6857469492, 354: -2.15618063167, 355: -4.28716263595, 356: -6.52706676161, 357: 3.87146799175, 358: 2.18559177971, 359: -18.8408882918, 360: 13.1447961101, 361: -10.609972417, 362: -17.4683941992, 363: 9.68666259979, 364: -2.50889922433, 365: 1.31510835016, 366: 10.7281232965, 367: 11.9762094254, 368: 5.63113663267, 369: -13.9039527618, 370: 8.71972104696, 371: -8.38815447816, 372: -29.337836815, 373: -0.055198405654, 374: 4.69994625641, 375: -36.695324384, 376: 7.94598051192, 377: -20.5584893299, 378: -24.769829173, 379: -1.47833239265, 380: -0.33148559225, 381: -8.39116540835, 382: 3.8037466587, 383: 5.30072260312, 384: -3.00515969034, 385: 2.0436992677, 386: 0.149804313688, 387: -8.6341912087, 388: 1.39433301885, 389: -9.56597899359, 390: -6.83274758577, 391: 3.7738472408, 392: 18.1959000219, 393: -3.31034860166, 394: -11.6415061345, 395: -13.1542967713, 396: 14.5649288439, 397: 4.71719293008, 398: 7.22716147245, 399: -8.54417552623, 400: -11.6437055052, 401: 2.33290121395, 402: 4.05123088122, 403: -2.91844070589, 404: -2.87983264813, 405: 2.10021875317, 406: 29.9249505987, 407: 3.08658658473, 408: -25.9956252054, 409: 6.95532569692, 410: -0.995503477305, 411: 0.364481185942, 412: -1.01747520093, 413: 2.76526021776, 414: 3.59450065693, 415: 1.90718583554, 416: -7.99936332221, 417: 3.20723839438, 418: 3.74486441246, 419: 3.88944658588, 420: 3.88752239966, 421: 3.15136412157, 422: 3.89418258501, 423: 3.14056486947, 424: 1.74319005437, 425: -0.243318160598, 426: -0.158448627995, 427: 3.73247652586, 428: 4.24818027653, 429: 5.55996104639, 430: 4.41610916671, 431: -28.8736076702, 432: 3.44086520362, 433: 3.18953369372, 434: 4.996248796, 435: 2.30525079013, 436: -6.81515626696, 437: -0.743285763747, 438: 3.68921944473, 439: 3.66038937602, 440: 3.76441975855, 441: 3.87079409856, 442: -2.87627441062, 443: 3.81356666744, 444: 3.43570790019, 445: 3.58823078579, 446: 0.80937763369, 447: 3.73816117837, 448: 1.90158716231, 449: 3.37068771194, 450: -22.3883575669, 451: 3.40378451824, 452: -1.07640684771, 453: 2.22680051067, 454: 2.9733781679, 455: -0.35480792889, 456: 3.83503620885, 457: 1.99675469915, 458: 3.6887822017, 459: -3.45071845522, 460: -0.941737465987, 461: -16.9331301793, 462: -31.6496205869, 463: 2.85524803801, 464: 6.81133721731, 465: -9.60301950846, 466: -3.32544860926, 467: -5.95021774268, 468: 9.64048428099, 469: -17.3737482093, 470: -16.4651394853, 471: -9.96024585595, 472: -9.46540243748, 473: -0.0644740177807, 474: -2.59113118267, 475: 0.887784034454, 476: 0.474593477465, 477: 0.669416460695, 478: 3.20808572995, 479: 0.789181327999, 480: 3.76134841354, 481: 0.754017186418, 482: 0.109876132868, 483: 1.63708714823, 484: 1.59625050228, 485: 7.70378137317, 486: 2.17222567667, 487: 4.8761671165, 488: 0.470480807787, 489: -6.98366236293, 490: -22.1659760081, 491: 14.323959601, 492: 10.3399340211, 493: 0.420614001609, 494: 0.92990715624, 495: 0.842438763205, 496: 0.935050700749, 497: 8.64445658164, 498: 0.676425358975, 499: 3.84307921018, 500: 0.43114874309, 501: 0.978307271939, 502: 0.752363160054, 503: 0.9215172273, 504: 0.804470018245, 505: 2.30660889435, 506: -12.5242265516, 507: -0.0685574649333, 508: -0.0981436513318, 509: 1.27030210531, 510: 1.24025386227, 511: 1.15337691047, 512: -3.28088418635, 513: -1.54921177163, 514: 10.3982185226, 515: -9.52011094923, 516: -2.64267711065, 517: -6.54053673996, 518: 1.18656048604, 519: 0.343635015196, 520: -1.21275159506, 521: 0.914331370838, 522: -1.00002211083, 523: 0.71795720661, 524: 10.3096640544, 525: 2.3349239509, 526: 6.23640127638, 527: 1.24503872544, 528: 0.871023706696, 529: 0.603505278348, 530: 0.903615551499, 531: 0.723207582542, 532: -3.09851629736, 533: -22.7156982071, 534: -6.81450972542, 535: 9.61703853201, 536: -7.92044920638, 537: 4.47152256031, 538: 7.20433888893, 539: 1.67039790344, 540: 1.54654102283, 541: -15.041207865, 542: 0.772301971704, 543: 10.2573211899, 544: 1.61814755641, 545: 2.69374077795, 546: 1.29842742928, 547: 4.39808991466, 548: 0.706621375173, 549: 3.97016067855, 550: 0.783336308089, 551: 2.29611174404, 552: 0.636940700234, 553: 0.903274330569, 554: -3.65428978163, 555: 3.86178152071, 556: -3.06091032508, 557: 0.879674538508, 558: 0.870879390196, 559: 0.93588340347, 560: 0.63378186578, 561: 3.45090505465, 562: 0.507007592765, 563: 2.47622157597, 564: 0.681504723491, 565: 1.11178093179, 566: 3.56168820186, 567: -0.43276278921, 568: 5.81260255199, 569: 1.66419285942, 570: -7.1171600248, 571: -7.88971869856, 572: -1.49922688282, 573: 2.80523738873, 574: -4.4203533104, 575: 1.40633868984, 576: 3.48331917969, 577: -0.00155843850626, 578: -3.9832556072, 579: 0.75881742531, 580: -1.18207155335, 581: 0.408810415795, 582: -0.635694503273, 583: -3.25077023114, 584: 0.221526047384, 585: 10.070903257, 586: -8.38542886973, 587: 3.48123274767, 588: -1.47225742159, 589: 3.77672831691, 590: -19.2119760735, 591: 5.64960637449, 592: -14.7035117869, 593: 3.90986939288, 594: -16.694619059, 595: 9.37475547957, 596: -5.66503562944, 597: 2.32160440333, 598: 2.08952383754, 599: 10.1281707462, 600: -20.2789370737, 601: 0.0675030305798, 602: 17.8311828224, 603: -2.29706448271, 604: 3.7325803631, 605: 2.04291236148, 606: -16.541569619, 607: 3.64196122058, 608: 3.90147704065, 609: -0.767530051556, 610: -0.927432874954, 611: 3.78078637914, 612: 0.159511250256, 613: 2.56146682938, 614: 2.41454022346, 615: -20.3632647539, 616: 12.257600743, 617: 1.3020366286, 618: -15.1143668536, 619: 0.71653843855, 620: 1.9435027136, 621: 19.2811382507, 622: 13.6453713594, 623: 17.8101167498, 624: 19.3211387888, 625: 4.01180734423, 626: 6.0160112839, 627: 6.91537683572, 628: -0.239359733089, 629: 8.05080805815, 630: -0.790609655025, 631: -7.82677564135, 632: 11.0485947015, 633: 4.33088584125, 634: -5.94246663212, 635: 6.65563342448, 636: 3.62275703343, 637: 8.1241601807, 638: 3.53857103218, 639: 0.516267558436, 640: 5.46312125305, 641: -26.1445702458, 642: -1.46930804814, 643: 6.95917490895, 644: 7.69817155462, 645: 8.10520451244, 646: 11.534177955, 647: -8.07757746971, 648: 2.33196644985, 649: 23.5489377725, 650: 1.97711916921, 651: 0.443344600374, 652: 2.68982277676, 653: 0.435244432045, 654: -14.410471484, 655: 3.57178057268, 656: -25.513831839, 657: 3.86930383287, 658: 3.94899672654, 659: 2.07159648536, 660: 2.14764859961, 661: 3.88281674047, 662: 3.636545711, 663: -4.93957577997, 664: 3.86618829623, 665: -5.72326819538, 666: 3.71799385513, 667: 3.69884735535, 668: -3.06933670898, 669: 3.902083763, 670: 3.66278460787, 671: -2.39902207964, 672: 4.04687132181, 673: 6.11418909866, 674: -3.28163675362, 675: -14.7482356811, 676: 7.36533818282, 677: 19.5881903935, 678: -4.12239927478, 679: 2.61440721663, 680: -20.3673291945, 681: -5.82861508583, 682: 1.01508298457, 683: -3.26792347243, 684: -5.20400762841, 685: -13.0775901902, 686: -14.3188609303, 687: 7.00212130323, 688: 3.93551001214, 689: 2.33920034901, 690: -14.3748903188, 691: -0.0573866468504, 692: 11.1741380328, 693: -10.0234109258, 694: -6.25608519227, 695: 8.19693750654, 696: -3.24966435289, 697: 1.93703611257, 698: -1.77787938783, 699: -0.828020411609, 700: -50.4114261201, 701: 5.58186204418, 702: 65.7749722992, 703: 4.43083616461, 704: 29.6863561178, 705: -42.5888268407, 706: 28.7489344252, 707: -13.0749378814, 708: 26.7056431262, 709: 23.5179738324, 710: 100.0, 711: 23.8403821808, 712: -27.3654720066, 713: 27.1321205176, 714: -18.877139208, 715: 0.842638461473, 716: 50.803369342, 717: -49.1155031045, 718: -10.4384260884, 719: -57.191987131, 720: 21.7380763455, 721: -60.3961556928, 722: -19.6855481569, 723: 35.7482031905, 724: -49.4333955437, 725: 44.0175064718, 726: -26.9556172275, 727: 6.08227295426, 728: 4.70266463849, 729: -7.44270096726, 730: 17.1368068834, 731: 13.8274035936, 732: -17.3490296209, 733: -16.7103998242, 734: 1.22660564305, 735: 7.93181218702, 736: -26.2949704896, 737: 45.5535543719, 738: -15.1555788605, 739: -31.6581259717, 740: 1.26318120772, 741: 21.0396658136, 742: -22.2074613712, 743: 22.5409096017, 744: -35.7880612987, 745: 7.4091745436, 746: -3.59512403864, 747: -16.1736598941, 748: 2.99769311176, 749: -7.03103504546, 750: -6.53145515154, 751: 1.29137820591, 752: 24.2274290034, 753: 7.93657172271, 754: -7.43680291191, 755: 76.6978906348, 756: -1.36757361944, 757: -29.3545759857, 758: -12.9396367353, 759: 15.1395688764, 760: 24.7378775563, 761: 10.3999717667, 762: 5.96923977355, 763: -27.8373673443, 764: -6.13017513004, 765: -32.7481642959, 766: 5.8656754334, 767: 12.1275478102, 768: -5.40870244786, 769: -5.91088919631, 770: 1.7099229097, 771: -2.53917646171, 772: 6.14466255045, 773: -3.76223213493, 774: 3.28956908923, 775: 1.80420566781, 776: 9.02853289201, 777: -57.5277367658, 778: -17.8041635953, 779: -0.96422043908, 780: 1.06595659074, 781: -5.65663826501, 782: -6.6786318858, 783: 12.1509675744, 784: 17.3223389774, 785: 37.0345356228, 786: 9.16636537075, 787: -8.66995784789, 788: 70.210959673, 789: -2.58159011345, 790: -4.96517880808, 791: 17.5120153733, 792: -4.54036423698, 793: -6.04444251735, 794: -5.65278320184, 795: -1.21299265979, 796: -4.39115580382, 797: -6.17274070115, 798: -7.05524788973, 799: 48.5413186465, 800: -5.91391312076, 801: -8.43273995345, 802: -5.39927072854, 803: -4.09667697243, 804: 99.1783924411, 805: -5.39679520913, 806: -4.36764798205, 807: -8.5991871444, 808: 2.00763981773, 809: -2.58076052192, 810: 8.08638868535, 811: 99.9934864457, 812: 0.921421948791, 813: -34.7749787478, 814: -4.71864814203, 815: 7.23314933027, 816: 7.08109815133, 817: -5.83469451273, 818: 41.2361787819, 819: 10.9243412878, 820: -37.8931523914, 821: -12.2540899357, 822: 1.01227945126, 823: 18.5002919549, 824: -1.02411845122, 825: -1.5613552819, 826: -35.6136386297, 827: -14.9584154032, 828: -0.888276160967, 829: -5.65029837076, 830: 0.716755732083, 831: 33.7213305651, 832: 0.692649195035, 833: -4.91441411056, 834: -28.0429993886, 835: -1.69303966321, 836: 3.65654043779, 837: -4.03324026556, 838: 12.8640678323, 839: 19.9674867744, 840: 35.6570794814, 841: -30.5192494752, 842: -1.9009595956, 843: -27.352465952, 844: -0.950661451344, 845: -0.837717629054, 846: 0.252970064951, 847: -0.851657069631, 848: -5.09248267745, 849: -2.39298311379, 850: -0.938109454177, 851: -1.01816632794, 852: -1.10015618153, 853: -0.881909038744, 854: -10.8435513326, 855: -11.2232426142, 856: -1.97682516283, 857: -1.22667298826, 858: -1.61983759001, 859: 4.13925551008, 860: 1.83786683681, 861: -6.92206314805, 862: 8.45787338019, 863: 15.1395237684, 864: 52.0154588645, 865: 2.74208364541, 866: 8.98894537279, 867: 0.839771325528, 868: -28.8238379849, 869: -21.1565098913, 870: -0.833995825086, 871: 45.2996554534, 872: -0.889632283866, 873: 12.0388358308, 874: 49.9624826527, 875: -7.55080425196, 876: 2.08043444357, 877: -0.882518946865, 878: -0.933154609468, 879: -0.645596244168, 880: -1.44862482783, 881: 3.755213256, 882: -26.644873462, 883: 7.76532035417, 884: 7.68450419622, 885: -10.176777828, 886: 21.8766527101, 887: -11.1766882127, 888: 2.50809827642, 889: 6.98288590975, 890: 41.5230349227, 891: -1.08738307036, 892: -15.4309161916, 893: -0.368216653372, 894: 4.03673338977, 895: -1.85877320096, 896: 71.575838295, 897: -0.83934233866, 898: -5.07468569778, 899: -0.761205714047, 900: 5.10976853923, 901: -0.969649515538, 902: -0.83916546338, 903: -89.5384070243, 904: -4.92107226483, 905: 4.1973186499, 906: -0.77768534476, 907: -0.82564232153, 908: -4.38365279726, 909: -0.924112102388, 910: -5.44255776036, 911: 5.9858808863, 912: 7.67479899053, 913: -0.0305918457853, 914: -1.63366876558, 915: -4.65937617195, 916: -3.97839564028, 917: -0.214653321952, 918: -32.3103700931, 919: 4.84171474156, 920: -7.77915697372, 921: -9.46559483199, 922: -4.87808982353, 923: 3.84756750041, 924: -0.802056569617, 925: 14.639077484, 926: -3.14173425956, 927: -6.8567029681, 928: -10.6564083961, 929: 34.2019713706, 930: -1.38696772734, 931: 9.69360831875, 932: 21.10734332, 933: -4.92324862304, 934: 11.022827289, 935: 16.1170359927, 936: 58.6121531004, 937: -3.23507568145, 938: -5.4466451609, 939: -10.6402820673, 940: -22.1654130774, 941: -41.303900992, 942: 13.9272399756, 943: 4.46602438358, 944: 37.6340801045, 945: 26.1836065125, 946: -39.8455387792, 947: 82.0456722384, 948: -21.4614357628, 949: -15.8297134773, 950: -32.8192923699, 951: 13.1397802858, 952: -1.7836743614, 953: -4.73306721588, 954: -5.09300391272, 955: -28.5740097047, 956: -3.45187707181, 957: -5.08559841082, 958: -1.03277739075, 959: -9.9481914542, 960: -5.01067630485, 961: -10.9831395974, 962: -3.07547238099, 963: -10.7882543317, 964: -25.1208784607, 965: -3.2096366014, 966: -1.57251468712, 967: 12.5893324974, 968: 6.44608212125, 969: -2.24344726521, 970: 43.9841777257, 971: 3.5441332519, 972: 16.4949223788, 973: -27.1716806263, 974: -64.1455212908, 975: 4.95131800518, 976: -9.52625073353, 977: -38.199470108, 978: -7.05802675825, 979: -5.06950605712, 980: -22.6008398528, 981: 25.4793470431, 982: -1.74217814035, 983: 46.1243686622, 984: -13.5022558637, 985: -5.0438684184, 986: 1.15923805953, 987: -6.67347615224, 988: 10.082446876, 989: 1.68580702917, 990: -0.880716512726, 991: -19.7324657337, 992: 55.8340428168, 993: -6.58957318119, 994: 48.3461993526, 995: 22.6101046093, 996: 7.98901139659, 997: 2.57916780486, 998: 8.6001861847, 999: -9.70239786027, 1000: 62.7714870006, 1001: 2.76343126046, 1002: -28.221515479, 1003: 54.8205008667, 1004: -6.06947658468, 1005: -7.87859159818, 1006: -5.21959965223, 1007: -5.03933305568, 1008: -8.93538478065, 1009: 24.2363215026, 1010: -5.14490240424, 1011: -4.7599540853, 1012: -58.0381456711, 1013: -4.95510233792, 1014: -25.27886814, 1015: -5.21571183942, 1016: -5.02043773929, 1017: 66.3467950555, 1018: -3.11160354519, 1019: -5.29419906705, 1020: 9.05070408199, 1021: 11.2117564928, 1022: -51.6470809361, 1023: 21.903624891, 1024: -4.97110583259, 1025: 6.95095539942, 1026: 5.20849418502, 1027: -1.56636342051, 1028: 43.445985356, 1029: -9.27451498401, 1030: 14.8154637856, 1031: -5.2500010975, 1032: 8.45159697982, 1033: -12.3886725002, 1034: 12.6412224906, 1035: -3.42725694932, 1036: -100.0, 1037: -4.45518021011, 1038: -57.1330416358, 1039: -29.9581395785, 1040: 0.446997245765, 1041: 28.5902811011, 1042: -33.3195438813, 1043: 60.0953230658, 1044: 1.91667764204, 1045: -44.4498319445, 1046: -4.94581591744, 1047: 5.29525125792, 1048: 0.287772432586, 1049: 9.76697361939, 1050: -0.922311884165, 1051: -2.20928012975, 1052: 0.654203073176, 1053: 0.0640386452199, 1054: 0.739995993907, 1055: -7.31284163233, 1056: 1.48290635066, 1057: -2.65056167999, 1058: 0.549877963045, 1059: -1.91579722556, 1060: 0.0141299096574, 1061: -0.119420036498, 1062: 2.02165723945, 1063: -2.43197520568, 1064: 3.05369060621, 1065: -3.43190774804, 1066: -3.94352300691, 1067: 1.96892232662, 1068: 3.12730231941, 1069: -3.67960582755, 1070: 3.21043371699, 1071: -0.0557225276547, 1072: -2.61780014377, 1073: 4.58914243594, 1074: -1.45995442136, 1075: 3.33330245814, 1076: -7.73539120118, 1077: -1.25220304056, 1078: 6.57392809715, 1079: 2.30845120147, 1080: 0.0339574320699, 1081: 1.21397967223, 1082: -2.60873301169, 1083: 2.33502692165, 1084: -0.0567197577798, 1085: 0.482769672207, 1086: -1.54041079268, 1087: 4.22203316542, 1088: -3.03724834718, 1089: 0.588774479028, 1090: 3.3155361181, 1091: 1.46539746301, 1092: -1.99665333546, 1093: 1.44852586647, 1094: -0.530328587858, 1095: -0.708048390904, 1096: -0.229644070801, 1097: 0.0701014190193, 1098: -4.17875117911, 1099: -0.791189853427, 1100: 0.920856981993, 1101: -4.30712503114, 1102: -1.9321306083, 1103: 1.3526760194, 1104: 5.41966787166, 1105: 1.05322767324, 1106: 1.99742666411, 1107: 0.769195037902, 1108: -3.43715509268, 1109: -1.81103054363, 1110: 2.30581625278, 1111: -1.85128267952, 1112: -0.41154302769, 1113: 0.376870241467, 1114: -6.21775109215, 1115: 1.46018019444, 1116: 1.46701517548, 1117: 1.20597010006, 1118: 0.0615029130983, 1119: -2.04667013489, 1120: 1.93606503071, 1121: -2.03176255285, 1122: 1.18021718365, 1123: 2.58922031861, 1124: -0.943707860401, 1125: 4.07641552798, 1126: -1.54640352441, 1127: 2.48845662121, 1128: -1.40849491734, 1129: -5.22154372517, 1130: 0.574662802412, 1131: -2.33938877281, 1132: -1.13425608621, 1133: 2.26658346296, 1134: -0.35381177435, 1135: -1.42390961794, 1136: 0.723381890399, 1137: -3.74664271911, 1138: 2.20962332461, 1139: 1.57188299234, 1140: -4.99175222306, 1141: 0.784207612454, 1142: 1.38433630272, 1143: 1.29412494258, 1144: -3.19726430482, 1145: 1.09335080803, 1146: 6.75147036286, 1147: 1.25946799278, 1148: -0.56367701862, 1149: 0.366667194025, 1150: -0.0975001297286, 1151: 0.803313024818, 1152: 0.281529337747, 1153: -8.02543831344, 1154: 1.81275662282, 1155: 0.695843832285, 1156: 0.537653912768, 1157: -0.594536398034, 1158: -2.72843205002, 1159: 2.3916452008, 1160: 6.4133801771, 1161: -0.254052829579, 1162: -4.27970250846, 1163: -1.60850696701, 1164: 5.04893115192, 1165: 0.907006834056, 1166: -5.49488572445, 1167: 0.27059758789, 1168: 5.35837666178, 1169: 5.44815321535, 1170: 8.09110397592, 1171: 0.39716200148, 1172: 2.11984015938, 1173: 0.544342868999, 1174: 1.05652160807, 1175: 2.90731970245, 1176: 8.58305762189, 1177: 0.328775017575, 1178: 1.60542220423, 1179: 0.704748658488, 1180: 3.89274487622, 1181: -1.80138403345, 1182: -7.93736651237, 1183: -2.65820475686, 1184: 0.835918896253, 1185: -5.26093409636, 1186: -1.28735281166, 1187: 4.06609754382, 1188: -0.552150576252, 1189: 1.18541957623, 1190: -1.02662844289, 1191: -1.81314094951, 1192: 1.48071079098, 1193: 0.364703080012, 1194: 0.398671077672, 1195: 1.79753853121, 1196: 0.352669380268, 1197: 1.70614765281, 1198: -0.91556138123, 1199: 0.513462265398, 1200: 0.372480074748, 1201: 0.365727637684, 1202: 0.392163088583, 1203: 1.47668916246, 1204: 4.7424355663, 1205: 1.80490809948, 1206: 0.49414725459, 1207: 0.561762024968, 1208: 1.60067514802, 1209: -1.16046483756, 1210: 2.22602844796, 1211: -0.74044138425, 1212: -10.8258726149, 1213: 1.98674020251, 1214: -0.826912197664, 1215: 1.67594930593, 1216: 1.8582384425, 1217: 1.1856221752, 1218: -3.75487380918, 1219: 0.29746145621, 1220: 0.814796781267, 1221: 0.414196951819, 1222: -0.264149280034, 1223: -0.103831224928, 1224: -4.98509154061, 1225: 0.458869728076, 1226: 0.270047294441, 1227: 0.383113751923, 1228: 0.425529750894, 1229: 0.340785527722, 1230: 1.63875915293, 1231: 2.64738992798, 1232: -1.6078705849, 1233: -0.190554498603, 1234: 5.76370469958, 1235: -1.62021484788, 1236: -1.9570163866, 1237: -1.70394300554, 1238: 1.68963011227, 1239: 5.43344446842, 1240: 0.338423313833, 1241: 3.0330543294, 1242: -0.868390864255, 1243: -1.54404662782, 1244: -0.625144542528, 1245: -3.06560843082, 1246: 0.264229453037, 1247: 1.58578205217, 1248: 0.339367090673, 1249: -3.76395622501, 1250: 0.427665876121, 1251: 0.328507015254, 1252: -5.45262907937, 1253: 1.64975416409, 1254: -4.03964840535, 1255: 0.491856410776, 1256: 0.345572120179, 1257: 0.535566452213, 1258: 0.402645110851, 1259: 1.26303445332, 1260: -3.73410430066, 1261: 2.61431861181, 1262: 0.277158808628, 1263: -0.381927401273, 1264: -2.14023671676, 1265: 0.33033357484, 1266: -1.59785722288, 1267: 2.49675957948, 1268: 0.0697227842917, 1269: -0.678281939885, 1270: 0.259966213133, 1271: 0.823243862263, 1272: -0.690551817815, 1273: 0.895246465389, 1274: -2.00313984215, 1275: 2.92118986578, 1276: -0.588682126716, 1277: -1.18962168506, 1278: 4.57613311136, 1279: 0.106654154222, 1280: 2.23109299523, 1281: -0.181222652453, 1282: -2.64460387195, 1283: -0.338016317665, 1284: -3.21617782137, 1285: 4.0956265366, 1286: 1.23625032608, 1287: 1.70880665721, 1288: 1.82778765018, 1289: -4.98975409817, 1290: -5.3040252803, 1291: -5.60914222091, 1292: 1.19874841767, 1293: -4.17683560483, 1294: -1.02610600527, 1295: 2.76930094265, 1296: 7.49079861566, 1297: -3.48826203105, 1298: -2.92824576335, 1299: 2.49236361455, 1300: -8.1009472331, 1301: 4.2600106221, 1302: 2.60055271621, 1303: 0.880392492304, 1304: 1.97978939624, 1305: 1.95951760515, 1306: 1.64513783162, 1307: -2.21120835406, 1308: -2.89213760551, 1309: 1.26759210168, 1310: -0.962137263746, 1311: 0.00765640479039, 1312: 1.58807302791, 1313: -11.00739006, 1314: 1.50538445529, 1315: -0.252000398028, 1316: -0.229488074073, 1317: -0.448528485474, 1318: 1.21182988388, 1319: 3.04582766264, 1320: 4.5171298152, 1321: 0.991125426729, 1322: -6.27291223823, 1323: 4.27963378279, 1324: -0.183910405794, 1325: 0.109403113754, 1326: 1.26020223096, 1327: -6.20970082717, 1328: 1.1219467254, 1329: -0.295641400401, 1330: 0.381757876793, 1331: 0.334697650592, 1332: 8.19529298428, 1333: -0.204100469679, 1334: 1.15059684459, 1335: -0.515114252255, 1336: 0.647376513972, 1337: 2.02536551847, 1338: 1.84630582275, 1339: 1.18765494495, 1340: 0.203451652242, 1341: -2.00455557556, 1342: -0.688468738613, 1343: -1.13315827694, 1344: -3.09332844406, 1345: 0.460476162029, 1346: 1.79799742463, 1347: -5.06291341662, 1348: -2.76368770592, 1349: -7.12923512276, 1350: -1.09811569544, 1351: -4.2855437874, 1352: -7.36023348473, 1353: 1.29861095431, 1354: -0.937171605777, 1355: 1.60086044672, 1356: 1.63496163569, 1357: 0.986289692149, 1358: 7.76024499714, 1359: -0.123614100214, 1360: 0.954491458206, 1361: 5.83354645797, 1362: 1.79053447043, 1363: 6.34658438477, 1364: 0.339524871957, 1365: 1.88572534082, 1366: -4.54653516722, 1367: 1.61777358016, 1368: 1.31472386604, 1369: 0.434804354033, 1370: -4.15920012119, 1371: -6.65056476157, 1372: 3.68097882758, 1373: 1.57131556477, 1374: -2.51435607992, 1375: -1.46604899693, 1376: -3.62762161265, 1377: 0.260963420836, 1378: 3.52054221274, 1379: -3.38484030379, 1380: -0.0913921750958, 1381: -2.04180067833, 1382: 1.61495026822, 1383: 6.09401135962, 1384: -5.37960957588, 1385: 6.38173676487, 1386: 1.6648339166, 1387: 0.18149826058, 1388: 0.298099263071, 1389: 1.05586605722, 1390: -1.54499609412, 1391: 5.37659673302, 1392: 0.142843898567, 1393: 0.840766656043, 1394: -4.16972375608, 1395: -4.12324516732, 1396: -12.9136566602, 1397: 0.285591341064, 1398: 1.32550872518, 1399: -18.3481678645, 1400: -5.09943732895, 1401: -24.0131080368, 1402: 18.8569734033, 1403: -0.364345314572, 1404: -41.8010820845, 1405: -3.57660751263, 1406: 5.89716897056, 1407: -43.6959948217, 1408: -7.54414255494, 1409: -32.298482714, 1410: -1.53741041893, 1411: -2.5165430644, 1412: -7.02300736902, 1413: -31.9100976603, 1414: -2.80937511878, 1415: -3.07365238593, 1416: 10.0408540926, 1417: 11.7107837028, 1418: -1.2503294114, 1419: 7.90626834916, 1420: -11.3524946734, 1421: -9.30108556882, 1422: 16.6380627729, 1423: -5.12750445186, 1424: 0.571507908758, 1425: 34.9842999235, 1426: -12.7955865888, 1427: 19.1945940969, 1428: 20.4533236173, 1429: -3.18530919114, 1430: -1.65036378309, 1431: 1.82417302966, 1432: 2.69875307188, 1433: 0.0400616274509, 1434: -0.433738745042, 1435: 2.78780507475, 1436: 1.43096217932, 1437: -20.6979606607, 1438: 1.31168271622, 1439: -6.86180628572, 1440: 1.47090355753, 1441: -6.86927562134, 1442: -2.40285158309, 1443: -9.98855487806, 1444: 3.78233867651, 1445: -6.06357219014, 1446: -1.13695138382, 1447: -8.04887001472, 1448: -12.0218396443, 1449: 1.13330994685, 1450: 2.51381433527, 1451: 0.565722940313, 1452: 0.064123362222, 1453: 4.54544247116, 1454: -10.4249753607, 1455: -35.5848765476, 1456: -28.5980851456, 1457: -26.9779435489, 1458: -12.6816068252, 1459: 7.54197064484, 1460: -4.90354891458, 1461: 8.98337056794, 1462: 0.0830544105343, 1463: -16.4762886945, 1464: 0.280743943132, 1465: -0.127011600906, 1466: 2.70717173259, 1467: -1.6377906241, 1468: -2.63972348738, 1469: 2.11047059403, 1470: -13.4443757556, 1471: -5.34146941613, 1472: -0.33750193137, 1473: -3.50810669512, 1474: 4.73445750275, 1475: -15.368408528, 1476: -6.99744854091, 1477: -0.2682311447, 1478: -10.9301074911, 1479: 3.44511298905, 1480: 7.08897952994, 1481: -1.58140911466, 1482: -1.26337282654, 1483: -6.68295620362, 1484: -6.47939000756, 1485: 5.87327376569, 1486: 4.54648502147, 1487: 3.68264173684, 1488: -0.715163477524, 1489: -13.2067021135, 1490: 2.26325749965, 1491: -0.311691450192, 1492: 2.24952848961, 1493: 2.28520468866, 1494: 2.43247969528, 1495: -5.49917692099, 1496: -6.66100000535, 1497: -14.2553771128, 1498: -8.59288147662, 1499: -11.3806582609, 1500: 1.11647217831, 1501: -4.53721249999, 1502: -3.69356775097, 1503: 1.74180098542, 1504: 0.927687574004, 1505: 6.57106423241, 1506: -5.62744253733, 1507: -0.927588040805, 1508: 6.6061433242, 1509: 17.7011147383, 1510: -2.93630273747, 1511: 3.31252780851, 1512: 3.26866336345, 1513: 10.8817286424, 1514: 2.44713689843, 1515: 0.812671768661, 1516: -40.3316035219, 1517: 8.5448947322, 1518: -16.9501897991, 1519: 7.15477827783, 1520: -2.25397146375, 1521: -15.0939903933, 1522: 0.402469470564, 1523: 0.748893323254, 1524: 1.82083361983, 1525: 11.8977951367, 1526: 0.50373719423, 1527: 1.41836613331, 1528: 4.38894078792, 1529: -9.16195112581, 1530: 4.91220190867, 1531: 2.72499446125, 1532: -5.88142667125, 1533: 1.34603125159, 1534: 11.7594814114, 1535: -3.13106003594, 1536: 3.74641774736, 1537: 11.9434040761, 1538: 1.57557413314, 1539: -12.7437244264, 1540: 2.48404109267, 1541: 6.68460041355, 1542: 0.200511410757, 1543: 0.644290805712, 1544: 7.25383113238, 1545: 0.488986859689, 1546: 1.83244594539, 1547: -2.64839338606, 1548: 0.0395772933661, 1549: 0.412723672752, 1550: 0.731308330016, 1551: 0.377066347348, 1552: 0.502816486093, 1553: -10.1990536246, 1554: 0.56235257768, 1555: 1.22326578936, 1556: 0.273605166386, 1557: 31.3446745925, 1558: -18.9240965097, 1559: 17.507528475, 1560: 21.1215712259, 1561: -47.3750989132, 1562: -47.7458705489, 1563: -2.49458989771, 1564: 3.07585527789, 1565: 12.6614392442, 1566: 1.30916014195, 1567: 22.5863205904, 1568: 0.639714315889, 1569: 15.8999257013, 1570: 0.489809375879, 1571: 28.0057150262, 1572: -6.38382233766, 1573: -6.49412639485, 1574: 0.193718756736, 1575: 0.433308789114, 1576: 0.71501533882, 1577: -0.190940734393, 1578: 1.00484987416, 1579: -1.88634447798, 1580: 3.86703081704, 1581: -3.92998435366, 1582: -42.8726087774, 1583: 35.557298602, 1584: 1.53021530302, 1585: 0.571868067252, 1586: 2.4015395946, 1587: 7.59607823892, 1588: 10.2265727957, 1589: 0.601323726737, 1590: 0.437058567681, 1591: 1.82894095536, 1592: 2.68569167885, 1593: -3.67552422331, 1594: 11.4528377288, 1595: 0.346580941037, 1596: 2.97684166813, 1597: 0.323801476168, 1598: -14.4805613317, 1599: 0.441123141429, 1600: 0.645447023967, 1601: -26.2382602307, 1602: 1.615971627, 1603: -26.5549636251, 1604: 0.494665556922, 1605: 0.546850016838, 1606: 1.09550105215, 1607: 0.531438498735, 1608: 1.95168553262, 1609: -8.59398150852, 1610: -4.72915635416, 1611: 1.34052006961, 1612: 0.750481962785, 1613: 32.8319479608, 1614: 1.85557646647, 1615: -21.612342433, 1616: -10.2296907871, 1617: 1.42362724453, 1618: -0.00014508160007, 1619: -3.98993468951, 1620: -1.61428207848, 1621: 1.22680999548, 1622: 0.0643285646592, 1623: -32.7004240021, 1624: -4.74913190385, 1625: 2.66060238001, 1626: 3.30047611011, 1627: 20.6701021347, 1628: 1.19755630102, 1629: 17.5168797182, 1630: 14.6814727885, 1631: -1.30369957349, 1632: -8.18670616642, 1633: -36.1925179533, 1634: 20.2612219553, 1635: 1.75426197546, 1636: 1.2609391638, 1637: -3.91309453873, 1638: -0.588895359338, 1639: 9.43499104748, 1640: 2.15371225701, 1641: -1.39812489648, 1642: -21.7144435277, 1643: -8.88860709677, 1644: 21.6874574596, 1645: -41.8356094275, 1646: 0.0328672248503, 1647: 6.32745484586, 1648: -1.44057063659, 1649: 4.8267301911, 1650: 2.78010322881, 1651: 0.565996117316, 1652: -6.03269591005, 1653: 1.10277694891, 1654: 1.94493846448, 1655: 1.91728129072, 1656: 5.39012785004, 1657: 3.42484924267, 1658: 2.28490467069, 1659: -6.61707060388, 1660: -0.870279783329, 1661: 0.371479212761, 1662: -5.19184273596, 1663: 4.8466555563, 1664: -0.451654441564, 1665: -4.26472034117, 1666: 16.5567686044, 1667: -7.16006332488, 1668: 25.4835560073, 1669: 6.02342149469, 1670: -0.376583626048, 1671: 42.0423799074, 1672: 2.06345096113, 1673: -7.01536265666, 1674: 0.868596813442, 1675: 1.73153606326, 1676: -2.37276020326, 1677: -19.9136454884, 1678: -12.2801069916, 1679: 40.3235666948, 1680: 20.1054715488, 1681: -14.0728260845, 1682: 10.1725571122, 1683: 2.60689541404, 1684: 4.96786840018, 1685: 5.56209318778, 1686: 6.6185523757, 1687: -5.20523695498, 1688: 4.0770203089, 1689: -10.2536677008, 1690: 12.8387023167, 1691: -19.5758712681, 1692: 4.45298982416, 1693: -13.3453133329, 1694: -3.74477197956, 1695: -1.70321911963, 1696: -4.2162928867, 1697: 33.8135854704, 1698: -10.1350468628, 1699: -8.91472360499, 1700: -0.950815235457, 1701: -20.132422999, 1702: -0.133158695232, 1703: 8.52069911533, 1704: 2.90417034762, 1705: 2.79064380785, 1706: 3.13974703499, 1707: -1.56221087929, 1708: 1.7201666487, 1709: -1.55726946418, 1710: 34.7305359614, 1711: 1.70358242128, 1712: -26.7161926368, 1713: 2.56835877364, 1714: 2.44978567693, 1715: -7.79151913988, 1716: -1.30749754965, 1717: 1.13553150755, 1718: 0.0748411626214, 1719: 7.34082483999, 1720: -29.9296594469, 1721: 13.0432973582, 1722: -25.7470713051, 1723: -5.18888786029, 1724: -1.65289562917, 1725: -16.9738927784, 1726: 1.29097018577, 1727: 8.24938003179, 1728: -12.4759305686, 1729: -7.35609468979, 1730: -6.81479306881, 1731: -5.21576528525, 1732: 16.8862594941, 1733: -12.5327151501, 1734: 0.453395954395, 1735: 1.752732064, 1736: 5.89157434581, 1737: -2.82244102186, 1738: 1.34423488307, 1739: -10.9879943142, 1740: 18.0365175893, 1741: 14.5863185825, 1742: -3.75090070582, 1743: 20.238662395, 1744: -25.9310799167, 1745: 6.40444129401, 1746: 0.430812256692, 1747: -1.18498602077, 1748: -5.59309814499, 1749: 11.628000236, 1750: -0.956221031228, 1751: -7.11090067154, 1752: -2.89177450921, 1753: -4.02514252198, 1754: 1.28647136359, 1755: 12.8937608729, 1756: 16.9986134152, 1757: -2.93440191742, 1758: -10.8159753602, 1759: 3.98553149925, 1760: 8.93674517365, 1761: -5.43120650248, 1762: 0.405346453429, 1763: 5.20207251741, 1764: -1.54750885504, 1765: -9.08219382859, 1766: 8.48244732239, 1767: 9.32872962677, 1768: -5.3978856483, 1769: 11.6889614826, 1770: -2.95439462217, 1771: -6.94959691052, 1772: -15.8712351804, 1773: -6.74581956568, 1774: -25.2483171752, 1775: -5.96765085581, 1776: 0.316558465725, 1777: 10.7099735803, 1778: 3.34205238789, 1779: 4.40432816061, 1780: -19.1163958293, 1781: 2.74910796333, 1782: -6.16926063375, 1783: 0.921963228024, 1784: 0.262743911544, 1785: -8.52701598325, 1786: 0.727491329251, 1787: 5.90110578589, 1788: 5.41500343075, 1789: 0.676275656711, 1790: -1.41757819901, 1791: -7.01509633959, 1792: -1.68659301069, 1793: -7.12637273752, 1794: 3.1602226766, 1795: 0.745930723842, 1796: -0.537888879173, 1797: 2.50030238278, 1798: 1.89188582058, 1799: 8.6276925488, 1800: -1.0165784694, 1801: 1.90120430133, 1802: 29.6777447329, 1803: 1.14172307946, 1804: 16.6386155508, 1805: 12.9146483831, 1806: -3.84828197344, 1807: -6.61421287621, 1808: 6.32464487996, 1809: 7.30242598497, 1810: -16.8391779246, 1811: 1.74918243081, 1812: -8.85612032993, 1813: 2.3514954299, 1814: 10.1846051539, 1815: 1.78452320886, 1816: 2.08136545757, 1817: 0.194540546069, 1818: 1.24760209977, 1819: 13.7265664511, 1820: 4.1175486723, 1821: 2.14292406748, 1822: 12.9210431437, 1823: 3.71359380933, 1824: -14.6444315136, 1825: 11.4799331745, 1826: -7.19467852637, 1827: -14.3052029635, 1828: 1.60329248182, 1829: 6.17897322599, 1830: 4.06507475484, 1831: 0.582849401858, 1832: 3.32737927072, 1833: 0.668984942316, 1834: 2.43653947378, 1835: -1.11554692256, 1836: 1.03657045355, 1837: 2.09471744319, 1838: -17.75073456, 1839: 1.78760979634, 1840: 2.62003915243, 1841: 2.14281176503, 1842: -7.85398467442, 1843: 2.29497597105, 1844: 1.83560744136, 1845: 1.77912462934, 1846: 14.4480529757, 1847: 2.08057305999, 1848: -1.16441036697, 1849: 1.91591202458, 1850: 0.547408422503, 1851: 11.4393562492, 1852: 2.17178024063, 1853: 2.71978218923, 1854: 1.6524117511, 1855: 1.38456306666, 1856: 1.32574894745, 1857: -5.01585805338, 1858: 27.7096545993, 1859: -11.9954183577, 1860: 6.30227141056, 1861: 3.19295014808, 1862: -6.76473741017, 1863: -8.29695747025, 1864: -29.3680143505, 1865: 15.9741883274, 1866: 5.68783934, 1867: 13.0411148411, 1868: -8.88264096511, 1869: 0.3672554244, 1870: -6.45135419164, 1871: 0.305793083779, 1872: 0.0432169310617, 1873: -0.206965694128, 1874: -8.74709773922, 1875: 0.59530209713, 1876: 1.55376812905, 1877: -3.14516485096, 1878: 8.82901134864, 1879: 2.70744887435, 1880: 0.393153513152, 1881: -7.58109169451, 1882: 1.78411987037, 1883: -13.6065615547, 1884: -1.35009119043, 1885: -0.184260107851, 1886: -5.62180362839, 1887: 15.1110647725, 1888: 7.17683562898, 1889: 0.763436090648, 1890: 14.6189218951, 1891: 0.240310396505, 1892: 0.424647119868, 1893: -7.43557217124, 1894: 0.495913216659, 1895: 1.80162292228, 1896: -1.2291163771, 1897: 0.0160871451107, 1898: 0.450855026811, 1899: 0.53879097884, 1900: 0.349354977527, 1901: 2.8498044189, 1902: -10.0044040315, 1903: -0.295333979536, 1904: 0.565523743979, 1905: -0.653204545582, 1906: 2.38414258652, 1907: 1.42729214984, 1908: 16.0742200771, 1909: -8.81780294902, 1910: 1.2127375403, 1911: -44.2730105062, 1912: 2.27866826863, 1913: 5.51200061335, 1914: 2.06487972569, 1915: 0.6315687365, 1916: -5.77970301673, 1917: 0.768069402195, 1918: 2.93108215308, 1919: 0.303096651403, 1920: -0.872980343714, 1921: -1.24427268113, 1922: -21.6654039406, 1923: 0.72109285298, 1924: 0.295118670725, 1925: 0.0612792372819, 1926: 0.283604614626, 1927: 0.698674336193, 1928: 7.93926601948, 1929: -5.01029040579, 1930: -0.226180434673, 1931: -12.9637039525, 1932: 18.1436662528, 1933: 4.50443308139, 1934: -4.71224945259, 1935: 0.981410559406, 1936: 2.4560810597, 1937: -7.11102722557, 1938: 0.45169710174, 1939: 15.505341722, 1940: 0.587931031107, 1941: 3.33072844186, 1942: -6.37548766391, 1943: -14.1460135509, 1944: 0.405345733072, 1945: 1.90025439786, 1946: 0.248277190045, 1947: 7.26294625918, 1948: 0.501474046119, 1949: 0.4522911503, 1950: -12.8496421069, 1951: 1.92097128017, 1952: 11.4138030474, 1953: 0.399551133494, 1954: 0.284341564948, 1955: -2.44803120312, 1956: 0.535626763642, 1957: 2.43025554268, 1958: 14.9791661207, 1959: -9.97026222914, 1960: 0.808859992786, 1961: 2.64110006821, 1962: -10.2197421483, 1963: -11.9681580042, 1964: 12.4555587527, 1965: -1.30078908838, 1966: 0.901524852003, 1967: 0.939454717304, 1968: 2.96038055483, 1969: 1.95444786258, 1970: 4.36944097123, 1971: 0.497056969883, 1972: -23.2967760589, 1973: -12.0010576673, 1974: -2.63986702846, 1975: 0.362112244169, 1976: -25.9530198054, 1977: 0.152062954722, 1978: -3.77807727739, 1979: -15.784230873, 1980: -1.27334671801, 1981: 9.92212259058, 1982: -12.1327153819, 1983: 30.5031624051, 1984: -4.60005762484, 1985: 1.65436468602, 1986: 6.71208249839, 1987: -1.05446264836, 1988: 4.16886991947, 1989: -6.41820211829, 1990: -1.09256238959, 1991: -3.99766031376, 1992: -6.65943321652, 1993: -8.97272050434, 1994: 26.2488054476, 1995: 18.8415718519, 1996: -48.2490736989, 1997: -1.39048605065, 1998: 2.15933022674, 1999: 4.98244959196, 2000: 3.02964920744, 2001: 1.54781152528, 2002: 16.8575876577, 2003: 1.88355862421, 2004: 2.03424943001, 2005: 0.117603822686, 2006: 3.16670387714, 2007: 2.03992430247, 2008: 2.27211129553, 2009: 0.283654806219, 2010: 2.81033943814, 2011: -10.6957015552, 2012: 21.7056594797, 2013: -8.05055775245, 2014: -3.54425753029, 2015: 1.78380358153, 2016: 1.68592817901, 2017: -16.129806121, 2018: 9.81561815165, 2019: -5.07998067936, 2020: 37.6082583145, 2021: -1.24131421908, 2022: -14.9095159691, 2023: -11.7387297438, 2024: 0.00484775692533, 2025: -0.934710250563, 2026: -7.45000615606, 2027: 7.98518416618, 2028: -3.46179048931, 2029: 15.85669844, 2030: 2.75396068108, 2031: -30.0437765708, 2032: 1.89790771268, 2033: 5.15932572214, 2034: 2.01286739296, 2035: -3.49059129801, 2036: 6.63846297645, 2037: -26.2119478538, 2038: 10.7636992781, 2039: -8.18877404844, 2040: -11.9308382178, 2041: -11.2499813059, 2042: -6.82985158994, 2043: 11.4671059114, 2044: 1.64035037744, 2045: 1.8905421361, 2046: 1.61955068175, 2047: -11.824305932, 2048: -0.846937304487, 2049: -14.7120901799, 2050: -15.920862274, 2051: 2.74639494557, 2052: -16.9460171494, 2053: 1.93718447769, 2054: 1.81216961129, 2055: 2.90595209704, 2056: -4.86002239754, 2057: 1.31399817471, 2058: 2.10481137903, 2059: 31.1707305217, 2060: 2.05206062709, 2061: -23.3469124339, 2062: 3.12662783094, 2063: 1.9912565697, 2064: 0.37116137616, 2065: 2.0460931262, 2066: 2.42821643914, 2067: 3.07195083518, 2068: 18.1546889533, 2069: -26.7914494241, 2070: 4.86569306317, 2071: -7.67510774363, 2072: 9.5751088172, 2073: 22.1203328043, 2074: -14.4346517019, 2075: -0.538218361848, 2076: 2.67400527637, 2077: 6.5948985923, 2078: 1.93705390285, 2079: 0.0563854392538, 2080: 1.91800008583, 2081: 0.760075730769, 2082: -2.05274724778, 2083: -13.1617751706, 2084: 1.84144386928, 2085: -14.0049632184, 2086: 5.51301122372, 2087: 24.4229240001, 2088: -2.03459044192, 2089: 3.31556003206, 2090: 1.34196759466, 2091: -39.6892635278, 2092: 2.71831105533, 2093: -1.44432488629, 2094: 1.99814773635, 2095: 1.35967745713, 2096: 29.3326136628, 2097: 9.26061001434, 2098: 1.14996683247, 2099: -46.8766172141, 2100: -2.89925982014, 2101: -2.81053389631, 2102: 30.1412060939, 2103: 15.6405119031, 2104: -69.4040729955, 2105: -2.15156313003, 2106: 11.4887960646, 2107: -14.5149414125, 2108: -44.2499462498, 2109: -17.2862458348, 2110: -42.4630888591, 2111: 12.4073017016, 2112: -9.80597278867, 2113: 35.4030209512, 2114: 10.8090930642, 2115: 1.16022829811, 2116: -90.1599278179, 2117: -20.5006878917, 2118: -39.39952302, 2119: 6.69911133786, 2120: -1.59909144804, 2121: -48.7739717107, 2122: 46.0993028585, 2123: -26.1498020187, 2124: 21.3141657578, 2125: 2.62695995886, 2126: -9.70174793583, 2127: -2.84045547156, 2128: -6.83005676532, 2129: -12.8380154201, 2130: 11.0629723488, 2131: 15.5643455387, 2132: 18.051308198, 2133: -28.6684173907, 2134: 46.0511519508, 2135: -24.6457332484, 2136: 7.62730695803, 2137: 1.63059307236, 2138: -12.5483945119, 2139: -5.75045902423, 2140: 19.7173847996, 2141: -0.347198057048, 2142: -8.06184325867, 2143: 9.57005356073, 2144: -2.5109716146, 2145: -22.7402068078, 2146: -11.8387738118, 2147: 9.23584973386, 2148: 8.21072955836, 2149: 6.28666373213, 2150: 6.79472228171, 2151: -24.7280473537, 2152: -14.4283568523, 2153: 9.21524296254, 2154: -22.5061710592, 2155: -21.4735457068, 2156: 14.1983885511, 2157: 0.421347053709, 2158: 5.92289848407, 2159: -9.64391512078, 2160: 7.5660528575, 2161: -19.2326382845, 2162: 2.20016195478, 2163: 20.5283822402, 2164: 9.77662090577, 2165: 9.17559435344, 2166: -48.9823686522, 2167: 7.59293184265, 2168: -16.2222876338, 2169: -9.17141281366, 2170: 15.2437444658, 2171: 1.63583943641, 2172: -8.80638857634, 2173: -49.1748171085, 2174: -36.3947586037, 2175: 0.268231285791, 2176: -37.6906710538, 2177: 8.0628812659, 2178: 11.8428633917, 2179: -5.4000143812, 2180: 0.326751360672, 2181: -11.9855589807, 2182: -14.566987359, 2183: 6.81494879861, 2184: 3.97289486685, 2185: -6.97474536296, 2186: 6.06582294299, 2187: -21.9638384224, 2188: 3.30883338501, 2189: 2.58443504207, 2190: 8.20841338009, 2191: 3.48988496761, 2192: 8.42477093724, 2193: -11.1878108897, 2194: 53.7342055659, 2195: -27.6261380674, 2196: 7.95781104256, 2197: 19.9589162637, 2198: 7.84348847247, 2199: 8.429375479, 2200: -2.36911988792, 2201: 0.0269097672985, 2202: 7.77005725907, 2203: 6.79889746934, 2204: 23.3379886116, 2205: 2.41287427001, 2206: -29.3937421358, 2207: 99.481256097, 2208: 11.6034527276, 2209: -74.1106993791, 2210: -34.6518985276, 2211: -10.3732107703, 2212: 11.9374973275, 2213: 9.26701145639, 2214: -16.7057994432, 2215: -15.3148151938, 2216: 7.64891442123, 2217: 7.41901550453, 2218: 9.19764822804, 2219: 2.26130937404, 2220: 2.16643526238, 2221: 0.0574780769452, 2222: -1.73115402428, 2223: -7.40233387096, 2224: 1.828013614, 2225: 7.92362875748, 2226: 3.32184456293, 2227: -14.7144063106, 2228: 2.69910029108, 2229: 6.77476476348, 2230: -23.2095928961, 2231: 0.667801388338, 2232: 1.45354012861, 2233: 4.86688007671, 2234: 5.43441686245, 2235: 0.870893762551, 2236: -2.00447303507, 2237: 24.7054812555, 2238: 2.70389573563, 2239: -19.4976571771, 2240: 1.7228450802, 2241: 0.969837560368, 2242: -25.1314562084, 2243: 1.48767893453, 2244: 7.91653918137, 2245: 11.9028887587, 2246: 2.42929089359, 2247: 1.57921391504, 2248: 2.0344978616, 2249: 1.80647311447, 2250: 3.15463849288, 2251: -13.2140168676, 2252: 3.95545569645, 2253: 3.43551914675, 2254: 3.66992219126, 2255: 2.81203919185, 2256: -5.27691012065, 2257: -2.60087149037, 2258: 9.7067604506, 2259: 54.1137827946, 2260: -28.5455618286, 2261: -0.493897071334, 2262: -8.57418339, 2263: -5.5623495213, 2264: -24.374047296, 2265: 21.4857221385, 2266: 1.37687534914, 2267: -32.315075145, 2268: 1.04118537024, 2269: -19.3555374871, 2270: 0.967534064759, 2271: 4.79334272156, 2272: 4.55956386565, 2273: 1.51280807441, 2274: 2.88840950411, 2275: 1.33846865707, 2276: 1.94291337277, 2277: 2.74323061173, 2278: 24.1169236776, 2279: -19.3529728893, 2280: -18.5912527899, 2281: 5.43092592313, 2282: 27.6310523428, 2283: -5.06861701382, 2284: -18.7761580869, 2285: -10.2416902781, 2286: -20.6108706832, 2287: 1.04200969843, 2288: -12.6983879276, 2289: 9.13242229879, 2290: 7.32357520244, 2291: 2.36414011247, 2292: 30.4645185198, 2293: 1.59549414324, 2294: 2.7847897914, 2295: 1.5977617605, 2296: -7.45381061568, 2297: 1.37231982328, 2298: 1.66974562787, 2299: 1.32119146746, 2300: 7.72144389337, 2301: -2.7747633966, 2302: 1.38966439797, 2303: 1.60580292829, 2304: 4.73552101808, 2305: 1.82190974099, 2306: 2.05257252846, 2307: 0.476826356891, 2308: -12.440113718, 2309: 2.24364307939, 2310: 2.17736376468, 2311: 0.534587214197, 2312: -13.0565388864, 2313: -2.06962153629, 2314: -14.1718232211, 2315: -5.86701191141, 2316: 10.2506731884, 2317: 13.0205246523, 2318: 7.76909252205, 2319: -5.60459087937, 2320: 1.25717934215, 2321: 7.40463365295, 2322: 3.92302468457, 2323: -1.61133148281, 2324: 2.78712907898, 2325: 20.2896351281, 2326: 1.90438514858, 2327: -27.3338283223, 2328: -0.143737538537, 2329: -15.8611525487, 2330: 2.5792675385, 2331: -31.2395842491, 2332: -12.9816788874, 2333: 4.60172448786, 2334: 8.01164066421, 2335: -17.1328249279, 2336: 15.8639998299, 2337: -12.0044517711, 2338: -44.1932611574, 2339: 15.5015224574, 2340: -10.9789257595, 2341: -11.2327374802, 2342: -97.4642384514, 2343: 62.5262324633, 2344: -9.03045613902, 2345: 22.9512330381, 2346: -4.30540546615, 2347: 4.93192882807, 2348: 16.5959493488, 2349: 3.52578329995, 2350: 13.7565855504, 2351: 13.1000947769, 2352: 8.26494916251, 2353: 7.98565215448, 2354: -12.6509687846, 2355: 17.9788519511, 2356: 7.63383455755, 2357: 5.1023994231, 2358: -2.41468248088, 2359: 3.24206835722, 2360: -10.2556617636, 2361: -14.1879177506, 2362: -10.2424847154, 2363: -39.7968746459, 2364: -0.0259796886574, 2365: -2.84347937763, 2366: 5.84537085705, 2367: 22.8002858663, 2368: -4.58394414539, 2369: 8.05109551398, 2370: 45.3377110113, 2371: 7.68043488382, 2372: -32.5341212439, 2373: -51.6371101927, 2374: 60.0776809962, 2375: -4.61819848657, 2376: 11.6081460824, 2377: 3.41164197595, 2378: 12.8645748851, 2379: -10.1106844905, 2380: 28.573488384, 2381: 7.76719347242, 2382: -6.28515294583, 2383: 7.53838140123, 2384: -2.48826097746, 2385: 1.31532968791, 2386: 5.80103414284, 2387: -1.69900155777, 2388: -1.0811580212, 2389: -27.9785279669, 2390: 11.3560260874, 2391: -8.58730544674, 2392: -4.06055334124, 2393: -7.00656765852, 2394: -6.17994238304, 2395: 0.0825461760543, 2396: 10.0201737199, 2397: 6.26160623495, 2398: 0.379232900172, 2399: -29.1836076204, 2400: 2.50135256692, 2401: -48.5229665189, 2402: 2.67655086172, 2403: 2.73427351662, 2404: 16.3873427898, 2405: 33.8570255832, 2406: 8.58473362949, 2407: 5.95408495396, 2408: 12.089443433, 2409: 7.7403904292, 2410: -12.3366145428, 2411: 8.14652780886, 2412: 7.8433060476, 2413: -19.9978114043, 2414: 5.63883423351, 2415: 2.04248963568, 2416: 0.727244364325, 2417: -14.5534540521, 2418: 0.229105213598, 2419: 9.53468122074, 2420: -1.1463468242, 2421: -5.67524602111, 2422: -38.2731491932, 2423: -31.7302851382, 2424: 12.2058167379, 2425: 15.5844451859, 2426: 18.1842867811, 2427: 7.29428663366, 2428: 2.90752294092, 2429: 0.993903525697, 2430: -27.7210198083, 2431: 4.48007812143, 2432: 20.2099263641, 2433: 7.94905208108, 2434: 35.6849161036, 2435: -42.0287778459, 2436: 9.5555813287, 2437: 30.9383788414, 2438: -26.3370334933, 2439: -30.5114870038, 2440: -59.6407183901, 2441: 18.3457772941, 2442: 45.1087405312, 2443: 11.1735779227, 2444: 0.337259065287, 2445: 1.1830989831, 2446: 2.89508504355, 2447: -4.45471298002, 2448: -4.59491546096, 2449: -6.9968893201, 2450: 1.80009609862, 2451: -8.29995517566, 2452: 0.648937407535, 2453: -8.20619872444, 2454: -9.37205530686, 2455: -8.65399701193, 2456: 8.6048015365, 2457: -2.45754442593, 2458: 3.28943180449, 2459: -5.20942530349, 2460: 8.79886443075, 2461: -3.0607940067, 2462: -21.5972679306, 2463: -11.3420941066, 2464: 0.756464050281, 2465: -5.96583906731, 2466: 0.17822232021, 2467: 6.49595298651, 2468: -14.111586644, 2469: -0.490818910144, 2470: 1.44105446905, 2471: 6.2973161787, 2472: -11.5009723995, 2473: 5.99568010734, 2474: -0.404156684107, 2475: -1.77734063545, 2476: -3.34217416684, 2477: 3.53875587137, 2478: 7.23665142267, 2479: 7.24043106732, 2480: -9.09432895031, 2481: -6.4279925494, 2482: -2.30963367984, 2483: 16.0637199644, 2484: -21.0554162863, 2485: 3.88842217812, 2486: -8.25762308377, 2487: -5.33854982458, 2488: 8.44856385307, 2489: -2.09725380543, 2490: 8.38793284645, 2491: 0.670537050582, 2492: -4.28135417387, 2493: -1.65188445223, 2494: -6.74898685074, 2495: 7.91722652847, 2496: 1.78493785279, 2497: -21.1530852343, 2498: 3.62313871781, 2499: 1.19534403061, 2500: -1.43084948426, 2501: 6.99006171978, 2502: 8.01580864212, 2503: -16.7316826975, 2504: -15.876065481, 2505: 4.03917914105, 2506: -1.24591653338, 2507: 10.100549557, 2508: -9.62034019579, 2509: 1.76025905246, 2510: -20.9519162962, 2511: 4.2330607082, 2512: 3.53652516586, 2513: -2.40669039982, 2514: 2.08628906169, 2515: 12.1048729364, 2516: 1.6950245291, 2517: 15.1925099071, 2518: -3.35861379721, 2519: 0.0239028838573, 2520: -1.33285642523, 2521: 1.7860497228, 2522: -7.80110357695, 2523: 11.9591799924, 2524: -2.37946145644, 2525: -11.5059325586, 2526: 1.3530281878, 2527: 9.04040086166, 2528: -2.44529189173, 2529: -9.46268834869, 2530: -11.6366822665, 2531: -16.80710999, 2532: 3.75688348552, 2533: -4.79312812545, 2534: -12.489041933, 2535: 1.38729762999, 2536: -9.45756130999, 2537: 1.45186485256, 2538: 2.18326543934, 2539: 2.62131993406, 2540: 1.91283835504, 2541: 1.79066868216, 2542: 7.67873056468, 2543: 3.38004009244, 2544: -8.43408875311, 2545: 2.70046493382, 2546: -1.090058075, 2547: 0.524454187753, 2548: 3.95149922693, 2549: -14.312567021, 2550: 1.20688341836, 2551: 1.40918948602, 2552: 3.67563916437, 2553: -13.03136004, 2554: -0.200951046441, 2555: 6.67475711806, 2556: -31.0257504427, 2557: -5.44169752141, 2558: 21.4056516501, 2559: -5.25436173954, 2560: 7.85928661766, 2561: -1.67311170635, 2562: 13.0651665748, 2563: -6.26561818825, 2564: 0.497356462731, 2565: -4.5437943102, 2566: 11.02260211, 2567: -4.06425868977, 2568: -1.30208477497, 2569: 0.381023110589, 2570: -0.823707142835, 2571: 13.7270205604, 2572: -16.4959427645, 2573: 0.368977966472, 2574: 1.42778648116, 2575: -4.12754934919, 2576: 9.87251302251, 2577: 1.1407485693, 2578: 0.304108690911, 2579: -0.917438544529, 2580: 1.61936768517, 2581: 27.3286588961, 2582: 6.14457924463, 2583: -11.22480775, 2584: 20.5205342809, 2585: 6.6731915639, 2586: -5.03678166719, 2587: 1.61452267876, 2588: -9.47095615465, 2589: 0.380126509818, 2590: 0.37582759385, 2591: 6.4837833537, 2592: 0.365248957025, 2593: 1.30158237734, 2594: -9.4266790273, 2595: 0.510627119224, 2596: 0.378832831936, 2597: 0.443662422691, 2598: 0.174790801922, 2599: 1.88077045193, 2600: 38.4203345248, 2601: 1.53340542643, 2602: 0.584719356446, 2603: -0.0534429597218, 2604: -9.43648661659, 2605: 1.22679419323, 2606: -9.10163781404, 2607: 7.57153939166, 2608: 45.6140379305, 2609: -5.83446925817, 2610: 1.28380617202, 2611: 10.9661179458, 2612: -5.73774287722, 2613: -1.73740615566, 2614: -10.8100006583, 2615: 0.155333170198, 2616: 5.40932535566, 2617: 0.498632542937, 2618: -5.81909999507, 2619: -12.1302788153, 2620: -15.4562782404, 2621: -0.914192020631, 2622: 0.33746050996, 2623: 0.273495023094, 2624: 0.0499700945994, 2625: 0.385722708208, 2626: 3.83688044561, 2627: -3.88644488089, 2628: 1.34862017894, 2629: 9.06460491952, 2630: -8.70429701311, 2631: -8.8215767424, 2632: -7.1429613896, 2633: -0.579609142851, 2634: -2.84498101629, 2635: -12.7423372652, 2636: 0.605120165513, 2637: 18.1867649848, 2638: 1.65945899531, 2639: -2.96660060813, 2640: 3.05708320063, 2641: -2.20361793386, 2642: 0.366168536663, 2643: 1.13635374167, 2644: 0.354912854492, 2645: 1.46163677604, 2646: 0.283801888339, 2647: 0.352592459144, 2648: -51.5801314078, 2649: 1.48031931993, 2650: 42.9953932286, 2651: 0.418340860689, 2652: 0.328629006407, 2653: 2.55976473295, 2654: 0.385445168998, 2655: 1.14049304978, 2656: 8.54244275836, 2657: -17.7246052164, 2658: 0.53381931649, 2659: 0.839256639484, 2660: 7.13933697749, 2661: 0.743750048051, 2662: -5.75740025725, 2663: 7.68808891717, 2664: -0.0866743369663, 2665: 3.33895850418, 2666: -0.0413228419231, 2667: -0.544307629896, 2668: -3.59164620246, 2669: -0.958029659977, 2670: -11.6716989701, 2671: -3.60454619958, 2672: -0.333889266702, 2673: 12.4959043156, 2674: 17.014405297, 2675: 1.15339743929, 2676: 11.2060792264, 2677: 5.11580798329, 2678: 0.68540864985, 2679: -11.5954545412, 2680: -4.91966769589, 2681: 7.53072528096, 2682: 1.40026629893, 2683: 1.50783122417, 2684: 8.72608918139, 2685: -15.7530438529, 2686: -17.6810596055, 2687: 23.1644451769, 2688: -4.8508726601, 2689: 4.59268309537, 2690: -9.8102359724, 2691: 16.691651914, 2692: 3.94894883149, 2693: -19.9695366241, 2694: -21.0659758478, 2695: -2.67888240851, 2696: -13.9661084122, 2697: -1.54070280218, 2698: -1.34813301739, 2699: 1.23548274769, 2700: -5.01350153642, 2701: 0.760272401742, 2702: 1.42914265204, 2703: -7.28373088223, 2704: -8.14043359413, 2705: -0.99743970096, 2706: 2.95860245329, 2707: 2.77727057463, 2708: 1.83333292884, 2709: -34.7801268235, 2710: -14.414066323, 2711: 10.01610143, 2712: 14.4782382648, 2713: -12.1173126616, 2714: 5.20992883219, 2715: 1.40576493593, 2716: 1.84609815885, 2717: -36.8646785879, 2718: -8.72728052142, 2719: -9.77915049235, 2720: 12.8795033093, 2721: 3.38567743297, 2722: -3.76596672021, 2723: 1.02244256593, 2724: 2.3116921437, 2725: -17.7164527523, 2726: 9.38043758141, 2727: 10.5853083969, 2728: 40.7170404543, 2729: -9.10343930079, 2730: 1.96021589701, 2731: 15.20123851, 2732: 5.96208664238, 2733: -2.02446135721, 2734: 1.4053413645, 2735: 7.83567779994, 2736: -15.0280651034, 2737: -0.537908812836, 2738: 1.58691762007, 2739: -2.65395738489, 2740: -7.92378707369, 2741: -1.56126008132, 2742: -6.80685812999, 2743: 7.73519710035, 2744: 2.4734817199, 2745: -13.6816939645, 2746: -1.92365655012, 2747: -3.55816221911, 2748: -6.17176737692, 2749: 2.17679611484, 2750: 1.60637484232, 2751: 1.10384460691, 2752: 1.33290021662, 2753: 1.70657890082, 2754: 7.86518435445, 2755: 1.22273167742, 2756: 0.864785751864, 2757: 18.3692228695, 2758: 1.39383468773, 2759: -21.4092356157, 2760: 1.83778700827, 2761: 1.7073352428, 2762: -7.67372192003, 2763: 0.39407387389, 2764: 0.78294108142, 2765: -5.75199675158, 2766: 2.20276795976, 2767: 10.4331037609, 2768: -6.94238439571, 2769: 11.2911867909, 2770: -14.6788851956, 2771: -2.07325045466, 2772: 14.4504611353, 2773: -2.64061378171, 2774: 0.866272889835, 2775: -8.98137110385, 2776: -13.677707973, 2777: -6.00203961657, 2778: 10.1167844552, 2779: 15.1662380321, 2780: -0.995362774505, 2781: 20.3181780888, 2782: 1.42611540063, 2783: -1.15320862086, 2784: 2.99547098806, 2785: -3.07974444387, 2786: -12.75149065, 2787: 39.9693359259, 2788: 3.63090967091, 2789: -37.8538776323, 2790: -12.0431654004, 2791: 14.5631306211, 2792: 3.91661510592, 2793: -0.491131211284, 2794: -29.4749565169, 2795: 3.08306669565, 2796: 24.7306614221, 2797: 5.08201335006, 2798: 13.282343772, 2799: -2.17190141068, 2800: 16.4877487771, 2801: 9.58721767132, 2802: -3.18494622266, 2803: 12.4448377842, 2804: 12.9593428521, 2805: -12.3888566788, 2806: -7.64344176338, 2807: -5.68746240944, 2808: 6.73791072178, 2809: 4.40480949853, 2810: -1.76938478905, 2811: -13.5544522459, 2812: -10.2243458763, 2813: -14.3090941516, 2814: -19.6597342905, 2815: -12.4840583916, 2816: 6.12280935981, 2817: 20.7536589888, 2818: 1.1849276681, 2819: -8.80420132139, 2820: 2.38761999682, 2821: 4.23319919639, 2822: -15.8149119617, 2823: 5.04552938786, 2824: -0.719485321366, 2825: -0.0812968033188, 2826: 0.884555859739, 2827: 4.76838856248, 2828: 0.786994582008, 2829: 1.17132983931, 2830: -13.6285720117, 2831: -12.7696162294, 2832: -4.77393988922, 2833: -7.65067409, 2834: -3.19936326378, 2835: -17.0478669393, 2836: -11.3607250636, 2837: 1.0551090429, 2838: 8.75540091661, 2839: 7.47039741882, 2840: 8.12621341989, 2841: 4.52353861501, 2842: 12.3349237968, 2843: -0.424053419816, 2844: -20.1051379779, 2845: -3.65253510947, 2846: 4.05026676089, 2847: -3.65184503585, 2848: -2.46935964811, 2849: -35.679908332, 2850: -4.76270511351, 2851: -7.73171028616, 2852: 13.6103797133, 2853: -5.36669022465, 2854: 1.21241191578, 2855: -11.213541231, 2856: -0.150215135792, 2857: -1.79055418453, 2858: -0.371420272215, 2859: 8.88342876722, 2860: -2.99126876637, 2861: 17.3129230082, 2862: -3.76201709107, 2863: 2.72541560069, 2864: 4.75523953419, 2865: -3.6307344677, 2866: -1.40460079163, 2867: -0.959246606231, 2868: -10.1432335808, 2869: -5.43204544196, 2870: 4.13455965281, 2871: -22.1018758572, 2872: -16.4844199924, 2873: 2.50344369006, 2874: 15.1531964133, 2875: -2.91950280764, 2876: 20.8794120794, 2877: 14.8959660951, 2878: 7.08477050603, 2879: 0.289533128406, 2880: -10.7696333046, 2881: 2.85401240537, 2882: 8.26912260905, 2883: 12.5874293656, 2884: -3.46727573103, 2885: -13.6603091234, 2886: -3.28738723135, 2887: -2.94472042732, 2888: -0.322837888222, 2889: 6.47586791182, 2890: -2.79237499455, 2891: 4.15617310029, 2892: -18.1124458203, 2893: 2.14591460489, 2894: -1.84564159398, 2895: 2.63710471882, 2896: 3.77878832228, 2897: 2.04102356509, 2898: 9.50720998822, 2899: -3.20191675083, 2900: -3.56910734863, 2901: 3.12456322787, 2902: 10.2256224034, 2903: -1.14735604874, 2904: 6.12782251045, 2905: -9.80941370165, 2906: -30.1173880087, 2907: 3.66816935463, 2908: 2.65891319419, 2909: -14.3124559205, 2910: -7.16413564432, 2911: 18.820996854, 2912: 10.2603840965, 2913: 9.96403087299, 2914: -7.05651645877, 2915: -6.60605314778, 2916: 11.3287492949, 2917: -9.2592283404, 2918: -0.804096900886, 2919: -0.36685802833, 2920: 14.1534774561, 2921: 9.40238662452, 2922: -0.629961174957, 2923: -3.12994335455, 2924: 0.674479126231, 2925: -3.40967262767, 2926: 2.35983951151, 2927: 0.450868756778, 2928: -11.9440675664, 2929: -1.33750187892, 2930: -6.59900051696, 2931: -1.02002141865, 2932: -1.93638561584, 2933: -18.6845603951, 2934: -9.16485337058, 2935: -9.95973264449, 2936: 0.449121153417, 2937: 5.94631538561, 2938: -0.819011182312, 2939: -0.78340143811, 2940: -5.03985223364, 2941: -0.798654469408, 2942: -3.52294854613, 2943: -3.44549179185, 2944: -0.700832716732, 2945: -0.73059890425, 2946: -0.796642377092, 2947: -0.568394678507, 2948: -3.22453651893, 2949: -7.64357054734, 2950: -0.239126173298, 2951: -0.167589380737, 2952: -0.786966583286, 2953: -0.738705475646, 2954: -3.51082304891, 2955: -11.7403885813, 2956: -5.05383094355, 2957: -40.6188385969, 2958: 0.645713495509, 2959: 4.41963134262, 2960: -13.4279186109, 2961: 4.72445247562, 2962: -4.37788271168, 2963: 20.943718655, 2964: -0.652163190327, 2965: -15.6517385829, 2966: -0.77681342997, 2967: -1.63925990872, 2968: 15.9543421708, 2969: -16.3379408391, 2970: 0.846221716982, 2971: -0.730848665703, 2972: -0.668095438011, 2973: -2.18348681533, 2974: -0.278933267562, 2975: -8.6424421837, 2976: 18.1822969524, 2977: 12.5136942964, 2978: 25.786408886, 2979: 34.4388492217, 2980: -0.521932514135, 2981: 20.0143438644, 2982: -9.45463415279, 2983: -4.40110091561, 2984: 6.32196570273, 2985: -0.52197941385, 2986: -46.7002281416, 2987: 1.96248438407, 2988: -3.33521750364, 2989: 2.67506627687, 2990: -0.626448707194, 2991: -0.54358635667, 2992: -3.39630257884, 2993: -0.940049662322, 2994: -13.8893417448, 2995: -0.65363499294, 2996: -0.610209210929, 2997: -8.93901254565, 2998: -3.2252930575, 2999: 21.9570206063, 3000: -0.736346416099, 3001: -0.636432621902, 3002: -0.471930075052, 3003: -0.661624119419, 3004: -1.23124648869, 3005: 2.60660104649, 3006: 9.22050908251, 3007: -0.632881943649, 3008: -0.317145252461, 3009: 11.6898902442, 3010: 2.8425896305, 3011: -10.2921805123, 3012: -21.7614557747, 3013: 1.25057766632, 3014: 6.17574448558, 3015: 4.56622378617, 3016: -0.795782819753, 3017: 9.41387596993, 3018: -0.737733387686, 3019: 10.5597930707, 3020: -11.4563904829, 3021: -2.3234356632, 3022: -1.45072405018, 3023: -18.6116783028, 3024: -0.328535824662, 3025: -4.47074277502, 3026: 8.57450326433, 3027: -0.332782680447, 3028: 0.771923042855, 3029: 3.22375291758, 3030: 8.15672947261, 3031: -3.97295442858, 3032: -3.05411358394, 3033: -8.76597930366, 3034: 4.18255530359, 3035: 13.0118029938, 3036: -5.80227493252, 3037: 11.0215513294, 3038: -6.89146293398, 3039: 8.18153712368, 3040: -2.92376520356, 3041: -13.662190346, 3042: 8.54450922413, 3043: 6.43276540936, 3044: 15.6310062233, 3045: -1.04207371776, 3046: -2.45648646468, 3047: 1.88955584312, 3048: -0.973096661618, 3049: 21.8312605906, 3050: 1.60729352084, 3051: -3.43197011578, 3052: -3.49520132958, 3053: -2.70611929147, 3054: -3.14131745151, 3055: -0.75171448579, 3056: -6.05876130394, 3057: -3.30351108638, 3058: -7.46309710085, 3059: 6.37793260276, 3060: 1.84198475468, 3061: -1.91819053864, 3062: -1.68290868507, 3063: -2.98826707447, 3064: 18.0882376039, 3065: -11.6708898863, 3066: 32.9375692366, 3067: 21.1185428712, 3068: 21.9400474565, 3069: -9.6560064788, 3070: 5.62978532886, 3071: -2.82064354886, 3072: -21.4462266093, 3073: -5.13877424454, 3074: -4.92435974247, 3075: -2.61022186308, 3076: 8.55615623426, 3077: 1.57471985827, 3078: -13.5300443925, 3079: -3.27668414873, 3080: -6.62656097052, 3081: 12.9588801282, 3082: 1.14828262052, 3083: -0.731315788204, 3084: -20.8421680419, 3085: 16.0565668009, 3086: -3.79452869264, 3087: -26.6111993687, 3088: 11.6845599603, 3089: 5.86302849722, 3090: 4.48732930575, 3091: -5.12864490462, 3092: -0.610286272351, 3093: -3.42122457619, 3094: 23.5898131153, 3095: -0.882913776539, 3096: -20.8345589595, 3097: -2.64326766437, 3098: -2.99235596035, 3099: 11.1620809663, 3100: -3.22006843471, 3101: -3.16965032106, 3102: 2.5668984143, 3103: 0.348682368495, 3104: -2.92470527828, 3105: -3.65895448507, 3106: -5.62368353284, 3107: -3.24478439569, 3108: -21.7056174161, 3109: -3.32788723604, 3110: -3.45433095594, 3111: 25.8541340641, 3112: -2.17514107911, 3113: -0.351862297435, 3114: -16.2639659664, 3115: 15.978573141, 3116: 7.40113320035, 3117: -2.87439809568, 3118: -2.10409533226, 3119: -0.0804875565038, 3120: 23.0413425161, 3121: 5.52625668841, 3122: -3.49633062449, 3123: -3.77076683188, 3124: 5.75082854993, 3125: -2.49251592467, 3126: -4.21217310488, 3127: -5.58184822713, 3128: 3.31720669167, 3129: 9.68626868956, 3130: -3.56497230522, 3131: -3.39157393106, 3132: -2.46321969871, 3133: 10.0410930529, 3134: 9.33898127112, 3135: 1.54872053607, 3136: -35.4987307178, 3137: -1.24176526127, 3138: 35.4340831013, 3139: -22.7649679119, 3140: -0.152921332197, 3141: 40.2162024187, 3142: 0.897743067114, 3143: 5.69809831689, 3144: -8.94199832586, 3145: 0.947947088858, 3146: -24.6414285041, 3147: -21.5601182823, 3148: -4.69248438077, 3149: 12.4828774363, 3150: 2.08286300077, 3151: -4.97966723746, 3152: -17.7284210516, 3153: -21.8354825786, 3154: -20.9041264842, 3155: -21.0386437987, 3156: 5.73764543839, 3157: -11.526665164, 3158: 2.69587610545, 3159: 5.73204698948, 3160: 15.0512379938, 3161: -4.75883238996, 3162: -20.8234580014, 3163: -3.82630110761, 3164: -10.7033707569, 3165: 3.99072877091, 3166: -6.6108431157, 3167: -32.8099102688, 3168: -16.7189949948, 3169: 6.10062815361, 3170: 13.9685045794, 3171: -4.57780786813, 3172: -2.87458584798, 3173: -9.11070976374, 3174: 2.68870337666, 3175: 1.99827213779, 3176: -5.21250416241, 3177: 3.58536971602, 3178: 5.626042192, 3179: 3.53211758242, 3180: 17.4915814577, 3181: -18.1552138919, 3182: -1.39485599683, 3183: 3.57095294789, 3184: -5.59766511334, 3185: 0.0349194956675, 3186: -10.6278309128, 3187: 0.625111396891, 3188: 2.64459581029, 3189: 3.23360745324, 3190: -6.83094500388, 3191: 3.11518985491, 3192: -3.6530384147, 3193: -1.42033034274, 3194: 3.54833512295, 3195: 0.669854247202, 3196: 11.4351828028, 3197: 3.42294562789, 3198: 4.69762144011, 3199: 2.18964508878, 3200: 6.18658139401, 3201: -6.44341822895, 3202: -21.2225638291, 3203: -0.456099461982, 3204: -1.3304819065, 3205: -19.2846813965, 3206: -17.523780074, 3207: 3.34286063544, 3208: -6.10850814037, 3209: 4.74387184294, 3210: -1.41218671064, 3211: 3.62261032243, 3212: 3.52896094232, 3213: 3.42236151826, 3214: 3.54432435938, 3215: -7.77227723431, 3216: 0.518063021129, 3217: 3.55195784601, 3218: -3.16541702636, 3219: 1.95126243314, 3220: -8.14343317628, 3221: -12.0836726318, 3222: 3.57209731952, 3223: -9.26747025844, 3224: 3.29215882824, 3225: 3.24358488048, 3226: 4.23040539526, 3227: 4.69473213194, 3228: -2.15440326945, 3229: -10.7487648511, 3230: 3.47862155283, 3231: 1.63409193216, 3232: 3.59988121048, 3233: 3.51333858179, 3234: -11.4063383956, 3235: 3.57597037847, 3236: 1.67344774336, 3237: 3.32918550441, 3238: 3.89786624759, 3239: 3.49879425527, 3240: 2.89271514596, 3241: 3.70869496248, 3242: 6.31298900969, 3243: 4.0213094965, 3244: -2.12305870687, 3245: 1.89502270433, 3246: 0.803805589144, 3247: -2.02431239166, 3248: 3.4951214556, 3249: -4.55543805577, 3250: 3.38482188554, 3251: -15.6249837077, 3252: 3.74116852002, 3253: 11.3469031826, 3254: 20.4217827506, 3255: -5.41454972161, 3256: 11.6519327032, 3257: -8.84824719551, 3258: -7.63436616806, 3259: -1.73627625082, 3260: -28.6587610648, 3261: 12.8560691992, 3262: 10.7951054327, 3263: -12.8925982584, 3264: 3.31292150984, 3265: -2.77219818964, 3266: 3.52776641965, 3267: 0.75384469869, 3268: 1.36346568158, 3269: 0.937391846718, 3270: -4.92511712527, 3271: 0.96809500932, 3272: 3.60954673675, 3273: 1.09124096498, 3274: 3.60312382461, 3275: 1.26263299882, 3276: 1.39701124539, 3277: -10.5343287236, 3278: 0.617086473997, 3279: 19.6826574081, 3280: 2.61611547012, 3281: 1.80928320544, 3282: 1.06884487431, 3283: -3.63240888291, 3284: 1.24131788045, 3285: 0.982419120601, 3286: 8.56322761384, 3287: 0.896500236238, 3288: 0.92605315897, 3289: 5.12642216097, 3290: 0.787485995676, 3291: -3.8361138481, 3292: 8.18718714794, 3293: 0.190469000816, 3294: 0.89417721419, 3295: 1.00286250454, 3296: 1.21252582894, 3297: -1.47144281395, 3298: 3.25640099578, 3299: -0.391896732517, 3300: 0.924095368394, 3301: 1.14098951243, 3302: 3.87593814092, 3303: -0.466296220335, 3304: 3.30532932936, 3305: 4.08181194683, 3306: -35.6628106588, 3307: -29.7183979639, 3308: -4.50786212519, 3309: 3.8971943874, 3310: 14.334281758, 3311: 3.74005656739, 3312: -6.97529910085, 3313: 0.693369928246, 3314: 14.5469729655, 3315: 0.888774314635, 3316: 21.8738416008, 3317: -26.6187595543, 3318: 3.76993506579, 3319: 1.07869867971, 3320: 0.806112927619, 3321: 0.694933134716, 3322: 1.36625733743, 3323: 0.56405440976, 3324: 0.91059729677, 3325: 2.20433639484, 3326: -1.90867656237, 3327: -7.73901468373, 3328: -18.7322404239, 3329: -9.68047042484, 3330: 4.57532263942, 3331: 6.02524127283, 3332: -0.269381265224, 3333: -7.84437360338, 3334: 0.733937467376, 3335: 8.68400133062, 3336: 2.66095805032, 3337: -0.817585111865, 3338: -1.70908519264, 3339: -8.3885479699, 3340: 0.861466651611, 3341: 3.52960626945, 3342: 0.926348507401, 3343: -12.6497245786, 3344: 0.879149736582, 3345: 0.899597574478, 3346: 3.27649872266, 3347: 0.885561791231, 3348: -2.00931032939, 3349: 0.885653462122, 3350: 0.92555479962, 3351: -0.166835815523, 3352: 0.689844278866, 3353: 1.81996462533, 3354: -1.77712165762, 3355: -1.36496990817, 3356: 0.279313264286, 3357: 0.988621111338, 3358: 1.79824994186, 3359: 3.63068186011, 3360: 1.5975938634, 3361: 14.2879824629, 3362: 4.23265815073, 3363: -0.00218647608643, 3364: 0.835846618971, 3365: 3.44299579332, 3366: 1.38502927349, 3367: 1.24318136436, 3368: -36.6736089045, 3369: 2.35457853941, 3370: 0.937340199056, 3371: 1.11599718197, 3372: -16.8456622415, 3373: 0.577480588286, 3374: -7.697637364, 3375: 3.34214485174, 3376: 0.51763695383, 3377: 0.590070632618, 3378: -16.3092319332, 3379: -3.17775718843, 3380: -3.03896775013, 3381: 3.51988134262, 3382: -5.03889696931, 3383: -3.32733874621, 3384: 14.9568137715, 3385: 9.06452908983, 3386: 4.23965246702, 3387: -2.96089262811, 3388: -10.5776279472, 3389: -12.6681789845, 3390: 10.7457398566, 3391: 1.32817441445, 3392: -15.7716034904, 3393: -16.0538499548, 3394: -2.96099439509, 3395: 8.88232864502, 3396: 3.53453063736, 3397: 3.66175864659, 3398: 10.8761144094, 3399: 3.58529417315, 3400: -3.85123443868, 3401: 1.24154472051, 3402: -5.62111740126, 3403: 3.51959143821, 3404: 0.1215731546, 3405: -1.15392588766, 3406: -1.44071003751, 3407: 8.87229359177, 3408: 5.56525075974, 3409: 2.92567589455, 3410: 1.84645292199, 3411: 5.44049781963, 3412: 2.69611428037, 3413: -0.661979030285, 3414: -1.20190382051, 3415: -27.6494535958, 3416: -30.0107550791, 3417: 25.154843216, 3418: -25.9248976692, 3419: 11.9817883435, 3420: 3.7309672294, 3421: -0.505939877579, 3422: -2.4272678023, 3423: -9.56234500642, 3424: 18.694130155, 3425: -10.1651075164, 3426: 7.89149207086, 3427: -2.55326153039, 3428: 3.45831312282, 3429: 2.31106375959, 3430: 3.42221850708, 3431: 0.835075715038, 3432: 0.556492412544, 3433: -0.522778330568, 3434: -7.29242128046, 3435: 11.5539948606, 3436: -17.8360312496, 3437: -10.5287718417, 3438: -11.7247991499, 3439: -0.507104097624, 3440: -7.08017167574, 3441: 1.51360708584, 3442: -0.147770024728, 3443: -4.4153493879, 3444: -2.31484636518, 3445: -30.6705780176, 3446: -14.2802805321, 3447: 1.74996927265, 3448: 1.04053337973, 3449: 3.54606989498, 3450: 3.68956932562, 3451: 3.54683901525, 3452: 3.96582203627, 3453: 3.3254017848, 3454: 3.33891277856, 3455: 1.15649177052, 3456: 0.806444589174, 3457: 10.1233041017, 3458: 3.52090488457, 3459: 3.4765828175, 3460: -9.72718736355, 3461: 3.72078121947, 3462: 0.555715045631, 3463: -5.91882644008, 3464: 0.792672250221, 3465: 4.98335832722, 3466: -10.384309304, 3467: -4.40460577899, 3468: -6.89723812527, 3469: -5.77463062421, 3470: 8.71969755392, 3471: -1.6744269353, 3472: 31.3711711504, 3473: -1.48693840744, 3474: 2.39651007541, 3475: -0.582482510315, 3476: 0.371752468479, 3477: 8.84924531013, 3478: -1.45269277666, 3479: 1.86065088404, 3480: 3.51721331878, 3481: -14.5746062093, 3482: 0.567196079621, 3483: 11.3197443819, 3484: -47.194695337, 3485: -10.8815104439, 3486: -0.0240962503319, 3487: 9.5124309007, 3488: 12.3145351562, 3489: 2.43169833333, 3490: 10.9610869298, 3491: -1.1442387633, 3492: 11.1578603909, 3493: 0.742325650954, 3494: 1.37970160353, 3495: -1.14605699307, 3496: -0.991833924791, 3497: -1.24792748295, 3498: -0.649707376841, 3499: -0.727019622067, 3500: -1.82878268656, 3501: 3.23291319232}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: -0.486058917535, 2: -17.5543071543, 3: 23.5747141661, 4: -26.111323548, 5: -10.672546125, 6: -15.4237190705, 7: 9.62206509701, 8: 19.1673370593, 9: 17.0236393132, 10: 16.0033913673, 11: -4.7947350581, 12: 13.7088510037, 13: -5.47626984634, 14: 14.1235014521, 15: -6.64606042526, 16: -8.85393383309, 17: 36.4226673486, 18: -7.5880889834, 19: 15.1417685208, 20: 12.7387648111, 21: -25.2874845719, 22: 17.5678401905, 23: 10.1423720539, 24: 25.7240911062, 25: -5.50831169248, 26: -12.5101868942, 27: -29.9590507412, 28: -6.93138816446, 29: 21.5874424624, 30: -8.48377475543, 31: 3.75127745322, 32: -62.8514718664, 33: -2.82754726233, 34: 5.72287632541, 35: -0.436079307994, 36: 11.5992371874, 37: 4.10164640017, 38: -12.2423395157, 39: -25.5159709907, 40: 7.24000251833, 41: -16.3612197721, 42: -1.34948578607, 43: -37.1589103945, 44: -30.8498051731, 45: 14.4721359637, 46: 10.3512852039, 47: -6.95284169467, 48: -2.84878526191, 49: 7.96062682519, 50: 1.3394002437, 51: 4.50107493062, 52: -3.36108316802, 53: -4.21946599045, 54: 3.5600351126, 55: 7.57153078933, 56: 0.177048471656, 57: -6.56781066624, 58: 0.528558638991, 59: 4.34067151676, 60: -6.82675140416, 61: -9.27845672156, 62: -17.9228606996, 63: -11.3333101985, 64: 14.2570386661, 65: 7.34314910668, 66: -1.97655998516, 67: 21.862809818, 68: -3.55721732929, 69: -2.04546216508, 70: 1.27514427371, 71: -4.13610012479, 72: -2.68956529578, 73: -4.12529572354, 74: 36.2746015524, 75: -6.0724454859, 76: -4.68094171839, 77: -14.146432156, 78: 7.33055038222, 79: -24.5163418332, 80: 7.83121924321, 81: 7.81098256676, 82: 9.977455789, 83: -4.07761422493, 84: -0.327415131484, 85: 4.08986309253, 86: -3.61911098455, 87: 5.74360943056, 88: 10.9083504149, 89: -2.87127306326, 90: 17.3737761237, 91: 10.9228726609, 92: -4.00743441542, 93: -12.3655917771, 94: -4.19407237211, 95: -4.4564790562, 96: -4.89962632693, 97: 2.13829453649, 98: -3.75976611285, 99: 7.72696432719, 100: 1.33795543237, 101: 13.6757930515, 102: -4.93543715951, 103: -20.5330403682, 104: -3.77986058551, 105: -4.44968414959, 106: 22.7143393598, 107: -1.72493058332, 108: -4.19218543374, 109: -2.83771259786, 110: 36.4031136306, 111: -4.85606497245, 112: 3.18323473442, 113: -8.57252688268, 114: -46.4333801355, 115: 57.6937301015, 116: 4.825444581, 117: -7.03623144473, 118: -7.21100303547, 119: -10.0001079741, 120: -36.2193406704, 121: -3.42807449856, 122: -14.0066996023, 123: 3.95956287042, 124: -2.12302090138, 125: -15.88568575, 126: -0.680500683002, 127: 0.235058213341, 128: -4.9089080481, 129: 17.173681127, 130: -0.721616939808, 131: -3.65691070265, 132: 1.00126183834, 133: -14.859072402, 134: 2.77902002089, 135: 0.0164546557463, 136: -6.67231409149, 137: -0.239561952126, 138: 15.815683765, 139: -0.910019051631, 140: 4.19677711318, 141: 7.97990587312, 142: -12.0925778159, 143: -23.3199814925, 144: -0.372334767446, 145: 13.9878826377, 146: -0.974471698941, 147: -0.769795308012, 148: -9.9644443126, 149: -0.60975263795, 150: -3.87611003789, 151: -3.67358683722, 152: -1.23187422486, 153: -0.833289426192, 154: -0.914745620052, 155: -0.650162599693, 156: -3.38976665672, 157: -0.870137547262, 158: -2.40263078647, 159: -0.689027510848, 160: -3.63741767999, 161: 2.84033935407, 162: -3.35242476914, 163: 7.00872852198, 164: -12.6534459377, 165: -5.9174657091, 166: 2.32377147889, 167: 0.786390985159, 168: 14.281857288, 169: 0.177553828797, 170: -2.92780007988, 171: -7.34159129069, 172: -0.358795821072, 173: -4.41651224458, 174: -0.787986392493, 175: -11.9073920802, 176: -0.244842739204, 177: 14.6318271053, 178: -0.265848167565, 179: -0.811869838943, 180: -0.881019819812, 181: -0.857300837013, 182: -0.700661300026, 183: -10.4187543807, 184: -14.3888929787, 185: 20.4028965604, 186: 32.2461889241, 187: -46.2328083598, 188: 3.30865338156, 189: -4.64888073056, 190: -23.4460447056, 191: 2.51918327049, 192: -16.5822390523, 193: -0.51804952853, 194: -11.3403452965, 195: -3.01470371729, 196: -2.46834324045, 197: -0.785434612441, 198: -5.92667268615, 199: -0.721999773, 200: -4.26877438367, 201: -0.633007896722, 202: -11.891579052, 203: -0.675026406276, 204: -0.756536035836, 205: 6.24975468926, 206: -3.94104524188, 207: -8.22319961562, 208: -0.711999226784, 209: -0.667230345294, 210: -0.855669123244, 211: -0.831812866022, 212: -2.53598049917, 213: 21.5101803849, 214: -3.61914493292, 215: 0.211340581, 216: 1.29794146714, 217: 0.365032434861, 218: 2.03232921994, 219: -4.49372191205, 220: 8.58296863635, 221: -0.558005588577, 222: 10.2213437835, 223: -3.51934745963, 224: -3.76187514173, 225: 12.5312029526, 226: -0.878488381352, 227: -2.75595919851, 228: -3.60809505812, 229: 16.3837078268, 230: 4.26405352919, 231: -42.3978856009, 232: -0.0179889290437, 233: 30.4757089959, 234: -21.4627768598, 235: -1.80451950564, 236: -9.99301682093, 237: -11.2298649784, 238: -14.8433151238, 239: -1.99240487483, 240: -3.64661326146, 241: 3.35099923749, 242: 3.76935376485, 243: 1.74385868179, 244: 15.5151808931, 245: -15.6810414232, 246: -0.774182989844, 247: -13.4250536425, 248: 19.4362635533, 249: 20.3289242666, 250: -22.4035480255, 251: -8.37039227568, 252: 24.1551513684, 253: -11.0959747827, 254: -10.8883003105, 255: -2.13995877981, 256: -3.54991025156, 257: 18.2453867803, 258: -8.14954802528, 259: -3.76709873737, 260: 1.41793304981, 261: 0.83955616352, 262: -3.0903612883, 263: -6.63134517968, 264: 2.79045978318, 265: -3.41388745255, 266: -5.99336463443, 267: 28.3703762728, 268: 12.3762582932, 269: 15.8175453384, 270: 2.61588936201, 271: -3.41593363897, 272: 2.80404621248, 273: 11.5773860857, 274: -15.7912087346, 275: -13.264502128, 276: 2.29428606091, 277: -3.70744453016, 278: -14.295741824, 279: -5.95053580731, 280: -14.3775296896, 281: -15.5924192644, 282: -17.0301610909, 283: 23.3592694174, 284: 19.7757587881, 285: -46.2685248552, 286: 7.11518451244, 287: -2.14982934335, 288: 6.34847994341, 289: -1.38477546406, 290: -1.97802074728, 291: -1.44821032615, 292: -16.0447172386, 293: 19.5188119154, 294: -9.17775158397, 295: -10.439905511, 296: -6.63330036771, 297: -12.8742877908, 298: 0.398886216182, 299: -0.624879931222, 300: 4.36628124877, 301: -8.52773722768, 302: 9.0756584485, 303: -11.1957310754, 304: 36.0612872954, 305: -1.94763038918, 306: -4.54015023905, 307: 32.4063608664, 308: -4.07794774041, 309: -4.17251378176, 310: -1.91395765775, 311: -4.26940368036, 312: -3.81149077427, 313: -4.15726193132, 314: 7.82189822397, 315: -3.871973569, 316: -1.9442290476, 317: -3.98825514428, 318: -3.89702631088, 319: 7.26434674433, 320: -0.090030682577, 321: -2.28302695956, 322: 6.39211983449, 323: 5.43075183348, 324: -4.26562315558, 325: -16.4472859453, 326: 2.30813297383, 327: 10.4574771941, 328: 64.5345252609, 329: 10.4408992722, 330: 13.368257626, 331: -6.62235920902, 332: -2.96160033762, 333: -0.947177289279, 334: -1.31711026793, 335: 5.52244909737, 336: 31.5987487643, 337: -9.01379124264, 338: -5.50687822459, 339: -3.99856149464, 340: 2.55929997876, 341: -6.25150849546, 342: -7.47625901465, 343: -5.72489829413, 344: -11.7234316704, 345: 2.10336416534, 346: 25.6262659734, 347: -0.296071584715, 348: 31.6347784497, 349: 48.0737722606, 350: 0.324139192679, 351: 0.00473853767406, 352: -4.3407774765, 353: -4.00671627016, 354: 0.715579702097, 355: -3.18424842358, 356: -7.65649889259, 357: 17.0947507146, 358: 2.58383839543, 359: -7.4009923094, 360: -12.9178071832, 361: -1.42078443837, 362: -9.51092505005, 363: -1.05626544718, 364: -0.776657811366, 365: -0.706081377845, 366: -6.70444390573, 367: -4.43968294467, 368: -19.2319688112, 369: -2.08690554299, 370: -1.44346845395, 371: 7.08305108764, 372: 7.34719812006, 373: 0.95088130047, 374: 3.85098587758, 375: -15.8446029867, 376: -5.24809477373, 377: 3.65795304591, 378: 8.78441492434, 379: -2.6653741925, 380: 3.90257387464, 381: -8.92805911711, 382: -8.57249723025, 383: 4.34671147946, 384: 3.67640027907, 385: -0.62833744217, 386: -0.258692944138, 387: -9.45623996967, 388: -0.524822204217, 389: 13.6512440211, 390: -6.23677835692, 391: -3.54680863229, 392: 4.89569029398, 393: 18.6706218907, 394: -3.21678875259, 395: 5.11616426689, 396: 3.50460058381, 397: 0.095969534568, 398: 9.39377628012, 399: -6.47086480543, 400: -3.0430143788, 401: 17.3380357336, 402: 1.33319162308, 403: -2.88726340612, 404: 5.19704518104, 405: 0.737964288539, 406: -2.67949073703, 407: 0.288398023862, 408: 9.56264999676, 409: -12.2786728543, 410: 11.8835571392, 411: 1.96150403545, 412: 2.75307397095, 413: 1.3480678884, 414: -3.70335468702, 415: 0.439726509932, 416: 11.4145859293, 417: -2.30448142142, 418: -4.62368497454, 419: -2.04666988033, 420: -1.70597551892, 421: -4.73526162322, 422: 1.02461847281, 423: -9.8915104848, 424: -3.36730191594, 425: 1.85159751129, 426: 0.518031196921, 427: 3.91869350446, 428: 12.5495549848, 429: 6.74785115753, 430: -1.6498992048, 431: 0.75147849761, 432: 0.852237692422, 433: -3.39387838295, 434: -1.10528954729, 435: -16.2651428681, 436: -0.539360649842, 437: 3.92123495315, 438: -1.44434312036, 439: -0.121124015372, 440: 0.923306415223, 441: 0.675887275974, 442: -1.63769853608, 443: 0.500137417011, 444: 0.435723493339, 445: 0.77508310197, 446: -0.0668523313975, 447: 0.473085367218, 448: -0.95626497039, 449: -4.74423398417, 450: -3.83561697439, 451: -1.09348639966, 452: -4.82189286878, 453: 0.569682665046, 454: 1.07341522089, 455: -11.0691602975, 456: 0.603547446233, 457: 1.12764615402, 458: -1.23837828453, 459: -10.2925616647, 460: -4.83073657849, 461: -4.38878856113, 462: -3.14190112556, 463: -1.9719314553, 464: -1.72362008619, 465: -1.85598479385, 466: 7.73000078592, 467: 3.40725731628, 468: -3.84959284451, 469: -7.21724380062, 470: -1.35190955717, 471: 14.4918907509, 472: 6.55464671978, 473: 4.9152228869, 474: 1.6944800442, 475: 0.409935057917, 476: -0.204793949705, 477: 1.63402031761, 478: -15.0561621292, 479: 0.336862454554, 480: 0.172869731546, 481: -0.922675619497, 482: 11.0409513476, 483: 2.19076167376, 484: 3.05987243239, 485: -7.59009017696, 486: -1.19690052629, 487: -10.9237173397, 488: -0.834092755644, 489: 2.67183289061, 490: 3.66656425176, 491: 5.88809365489, 492: 8.70738998652, 493: -0.270148036393, 494: 1.82357564333, 495: 0.224958042068, 496: 0.313294444033, 497: -2.77756988962, 498: 0.194655943681, 499: 0.250100683353, 500: 7.94474752589, 501: 0.421629585659, 502: 0.368023226393, 503: 0.181060715557, 504: 0.365704695203, 505: -1.00615231237, 506: 4.17888148665, 507: -0.317245241355, 508: 0.304819780301, 509: 0.766209101654, 510: -4.63560242766, 511: -3.20691824899, 512: 0.13355068147, 513: -1.51373919589, 514: 2.01550057687, 515: 19.3245355026, 516: -4.16392064849, 517: -1.18197534047, 518: -2.75320262863, 519: -6.01385819581, 520: -3.2761180597, 521: 0.157565143988, 522: 12.1763744691, 523: 0.346216072636, 524: 8.85695632313, 525: 2.99101469315, 526: -0.738894853708, 527: 1.21489036732, 528: 0.205331834029, 529: 0.336311556732, 530: 0.676378482884, 531: 0.0880798221349, 532: -1.12583676779, 533: -5.16189155868, 534: -2.55763270302, 535: -12.9347818151, 536: -26.6015548203, 537: -3.21946747666, 538: -3.13618043587, 539: -8.05407394944, 540: 5.71259071978, 541: 3.71410949565, 542: 0.201653612312, 543: 15.2164735917, 544: -2.67462067094, 545: -1.07192757912, 546: -3.0355612411, 547: -0.639531409062, 548: 0.253864712729, 549: 0.686300689502, 550: 0.362395958791, 551: -0.374412303863, 552: 0.352605630347, 553: 0.222264835242, 554: -3.2097767203, 555: -0.425647845589, 556: -5.83267516253, 557: 0.410265856789, 558: 0.180107911026, 559: 0.445767026112, 560: 0.207773432469, 561: 0.090058959583, 562: -5.71860488007, 563: 1.10415410272, 564: 0.0465866881958, 565: -0.376183001324, 566: -3.20428095486, 567: 4.5817962893, 568: 6.369256357, 569: -6.48514100204, 570: -0.445599853774, 571: 0.316508994069, 572: -0.0453588456048, 573: 0.453200675042, 574: -4.61527050755, 575: 0.893918362168, 576: 6.61934025497, 577: 10.390351648, 578: -1.24757128502, 579: -0.316058823777, 580: 5.97474883857, 581: 0.508686857429, 582: -1.28526363338, 583: -1.36805339542, 584: -0.865596302918, 585: -0.885896078447, 586: -0.951499941454, 587: 18.843057846, 588: 1.24832077916, 589: 0.386037744396, 590: -3.44680905335, 591: -4.28809646521, 592: 3.67984664872, 593: -0.188918809269, 594: -3.24810129741, 595: 3.66278358591, 596: 4.28071419674, 597: 14.9555772687, 598: -0.745547999983, 599: 1.75375267904, 600: -6.96718001528, 601: -13.0648855529, 602: 2.65723276059, 603: 2.96713731012, 604: 2.13175177753, 605: 0.819925724167, 606: -7.97722140785, 607: 0.68645681261, 608: 0.171728841813, 609: 2.54144738371, 610: -1.43753640916, 611: -0.101193885393, 612: 2.17020698622, 613: 0.241708455448, 614: -0.825925221201, 615: 9.26532385849, 616: 4.21953964671, 617: -1.17713816355, 618: -3.4327496098, 619: -2.40372736857, 620: 1.34932443257, 621: 12.1357877825, 622: 1.37385015449, 623: 7.0331201089, 624: 10.3500094882, 625: 10.3256501059, 626: 4.93504905298, 627: -4.85083582699, 628: -9.20859382275, 629: 7.93283857643, 630: 2.6117509403, 631: -5.9721604323, 632: 0.996285884991, 633: -2.35688463734, 634: 0.252061635235, 635: 4.50141149053, 636: 0.3752543863, 637: 2.03492442532, 638: -2.27826695134, 639: 8.02999221566, 640: 2.98220930432, 641: 0.370727098203, 642: 1.08566069777, 643: -0.307084304929, 644: 11.3093355847, 645: 9.11649371548, 646: 6.92934830135, 647: -4.87460316358, 648: 1.05310997807, 649: -15.139365691, 650: 4.44478143559, 651: -17.550485809, 652: -2.50810238507, 653: 4.09634730037, 654: 1.52536428287, 655: 0.566710873854, 656: -11.1292163432, 657: 0.889184859285, 658: 0.784093839666, 659: -1.07038302061, 660: 2.63429803918, 661: 0.153037085411, 662: 1.01692271322, 663: 3.97641064637, 664: -0.470731142503, 665: -0.61164936796, 666: 0.171570549031, 667: 0.955832984967, 668: 1.08864762849, 669: 1.91144844156, 670: -0.573124686855, 671: -1.2986076119, 672: 2.45043042163, 673: 1.15774891626, 674: -3.2076198819, 675: 11.2724239736, 676: -0.321623786973, 677: -2.42461462022, 678: 18.914693551, 679: -8.51231983507, 680: 8.73350545326, 681: -2.6976175322, 682: -2.36288780706, 683: -9.411618859, 684: -0.286771983092, 685: 2.62701615685, 686: 1.745477656, 687: 14.3169094105, 688: 0.614355769072, 689: 3.79579982276, 690: 0.288935311089, 691: -0.700198599667, 692: 0.48755366362, 693: 0.452963407726, 694: 0.128221586362, 695: -3.80714835177, 696: -1.09765674322, 697: 5.35777061984, 698: -17.9846274047, 699: 0.101281627406, 700: -0.804995355268, 701: 9.96383437356, 702: 2.43708171964, 703: -6.41638131941, 704: -2.13743323255, 705: 12.4258051896, 706: -3.46141178668, 707: -2.02158492542, 708: 2.02675002294, 709: -4.02827771635, 710: 8.72099291761, 711: 0.858857682016, 712: 2.58839202829, 713: 1.9604516335, 714: -3.82634693199, 715: 8.24624894268, 716: 1.00529321969, 717: 17.7579605944, 718: 14.8336109991, 719: -3.89252621785, 720: 5.69613231024, 721: -5.05636885778, 722: -0.125416557668, 723: -5.12610837598, 724: 7.55579430088, 725: 2.41620554595, 726: -3.5578851563, 727: -5.05018086522, 728: -3.09256161344, 729: -11.2963674382, 730: -4.57045509498, 731: 2.39500135971, 732: -2.43788764208, 733: -1.69207274573, 734: 0.703465759163, 735: -3.66431685166, 736: 6.11915826408, 737: -1.38737947784, 738: -11.4341697005, 739: 5.03362741247, 740: -0.351671968654, 741: -9.49372635367, 742: -11.6762033692, 743: 5.70766769193, 744: 0.380757601796, 745: 0.210369657773, 746: 0.0919680722454, 747: -6.67776376405, 748: 3.24142958023, 749: -1.08315432403, 750: -10.2302068114, 751: -0.102254692754, 752: 7.05761198459, 753: -5.80876193932, 754: -0.733507929562, 755: 7.69328567258, 756: 7.32305220653, 757: 1.01572233944, 758: -5.22331918908, 759: -10.1829136542, 760: 0.719494971396, 761: 0.922387953304, 762: -1.31557077264, 763: 6.65578544966, 764: -0.561850042896, 765: 15.1011080729, 766: -1.15735744549, 767: 6.19745866324, 768: -0.747276175106, 769: -0.721840156435, 770: -3.18675093671, 771: 0.945770018817, 772: 6.48984365602, 773: -4.363706573, 774: -0.394363426496, 775: -3.31888417941, 776: 0.482939345894, 777: -11.0387329131, 778: -0.953183347852, 779: 2.83270157137, 780: 5.33797607006, 781: -0.459581150129, 782: -5.90506297377, 783: -2.21286036046, 784: 11.5363541904, 785: 2.78815783339, 786: 2.29847680563, 787: 0.223375994229, 788: 2.6052589843, 789: -0.467981575176, 790: -0.54298894423, 791: -5.02213053681, 792: -0.37654335806, 793: 0.463050439866, 794: -0.635057533852, 795: -2.20698941481, 796: -0.736155159518, 797: -2.07226959633, 798: 0.23437401197, 799: 1.73276145366, 800: 0.379841035707, 801: 0.253926617603, 802: -0.575667918137, 803: -0.598159132248, 804: 11.8561779573, 805: -0.378902328881, 806: -0.232188487177, 807: 0.0549931305747, 808: 2.94582859973, 809: 0.371470411239, 810: 0.0930454716012, 811: 0.196052343483, 812: -3.53773754782, 813: -0.837876970771, 814: 2.54185596593, 815: -2.25328550576, 816: 0.888053003077, 817: -2.83344417142, 818: 0.979440276331, 819: 6.08731929795, 820: -7.12961626453, 821: 5.13550408085, 822: -0.321587389406, 823: -5.70879929922, 824: -0.148506131338, 825: 0.183819722627, 826: -0.142770949077, 827: -2.65079254876, 828: 0.0150202750784, 829: -0.73167927887, 830: 1.99396366616, 831: -13.9333768368, 832: 0.755414800943, 833: -0.133365363833, 834: -1.81329862565, 835: -1.28179725725, 836: -7.67497362806, 837: 0.207037936011, 838: -2.63312856767, 839: 4.23464928408, 840: -9.89884491011, 841: 0.152915747871, 842: 0.195598008994, 843: 4.69570403475, 844: -0.0232020364372, 845: 0.00477796505126, 846: 7.40295080904, 847: -0.0737242715738, 848: -0.463309258787, 849: 5.75144008994, 850: 0.362917255373, 851: 0.0347552791079, 852: -0.176573267067, 853: -0.216402237446, 854: -0.615538612372, 855: -9.28119122047, 856: -0.314326407517, 857: -0.518831137405, 858: 0.727046286423, 859: -4.66974727867, 860: -4.12898468486, 861: -21.7128692241, 862: -2.9703123295, 863: 1.83216811881, 864: 8.93061715579, 865: -0.240436658698, 866: 1.81503814848, 867: 9.09872135744, 868: 0.868288378678, 869: -0.212152967348, 870: 0.295256094915, 871: -15.4200767209, 872: -0.0131972511357, 873: -5.36469582284, 874: 2.953253996, 875: 1.88334730762, 876: 0.442359947159, 877: -0.0842720203395, 878: 0.205202021898, 879: 0.173071208271, 880: -0.233183067898, 881: -4.67165100378, 882: -0.218567197766, 883: 2.85744804221, 884: 13.3220554769, 885: 3.64921898425, 886: 5.00370135707, 887: 0.00170201307826, 888: -0.321073775871, 889: 3.08399541161, 890: -4.9436579792, 891: -0.12959656658, 892: -5.38444143749, 893: -1.38083185428, 894: 0.207952145145, 895: -2.11162014032, 896: -0.373699341291, 897: 0.00715504971163, 898: -0.575204353865, 899: -0.160238388192, 900: -0.169173111663, 901: -0.0758775154822, 902: 0.0233050356446, 903: 3.16327096248, 904: -0.445794251734, 905: -3.94028415366, 906: -0.0577737223703, 907: -0.00401539314417, 908: 0.0545504515481, 909: -0.0712240949873, 910: 0.271278695117, 911: -0.587478936357, 912: -0.699159986744, 913: -0.465942367319, 914: -0.291444222659, 915: -13.59940389, 916: -1.59842888299, 917: -3.05302679219, 918: -2.23858116197, 919: -1.17262357797, 920: -0.966209723171, 921: -0.104206334938, 922: -0.0712702954256, 923: -1.75266484112, 924: 1.23085136862, 925: 1.96363644709, 926: 8.93964856682, 927: -0.295855078869, 928: 1.39947974514, 929: -10.3093706535, 930: 0.0663946577417, 931: 3.71716657225, 932: -6.23334308328, 933: 1.94422896897, 934: 1.05129480157, 935: 3.77573652342, 936: 7.17641361332, 937: -0.443719481483, 938: -0.690074066801, 939: 1.49402393589, 940: 6.33896022603, 941: 2.5282982055, 942: 0.468899794488, 943: 6.37236689686, 944: -4.45895582139, 945: 6.96703031344, 946: -4.89155414982, 947: -11.5626283409, 948: -12.703862889, 949: 12.0572367474, 950: 8.05092723653, 951: 4.83164634038, 952: -4.75946867059, 953: 4.76247681035, 954: -0.41541516206, 955: -0.112763932531, 956: -0.240242265304, 957: -0.614726693864, 958: -2.57495615752, 959: -4.52175256853, 960: -0.479620305235, 961: 5.19511609742, 962: -1.3703553685, 963: -0.447875771145, 964: -10.9735367375, 965: 6.30169804472, 966: 4.65866676149, 967: 3.53835862532, 968: -1.48398157205, 969: -3.81628174212, 970: 1.58984106204, 971: 3.37177549374, 972: -15.0131644768, 973: -10.9434631342, 974: 0.443387698507, 975: -5.67845627181, 976: 3.79246017809, 977: 1.68122362028, 978: -7.89388462166, 979: -4.40211211309, 980: 6.43737631847, 981: -3.46430818341, 982: -6.256577058, 983: 9.85784857641, 984: -5.86860764814, 985: -0.48448824075, 986: -4.81768452049, 987: -2.30119412939, 988: 7.32728240055, 989: -2.54061104045, 990: 1.8326590092, 991: 5.53839420339, 992: -4.97282070386, 993: -0.308286041654, 994: -0.532271593478, 995: -5.40726453133, 996: -0.380709177894, 997: 0.707220513643, 998: -0.617708179601, 999: -2.09708498745, 1000: 11.7376485996, 1001: 4.71740366025, 1002: -3.61150239851, 1003: -6.32142165609, 1004: 0.509530583745, 1005: 5.66167412374, 1006: -0.483762429674, 1007: -0.377079113725, 1008: -1.07290306403, 1009: -5.01617383868, 1010: -0.39364548537, 1011: -0.46707377509, 1012: 0.417537045416, 1013: -0.607235793905, 1014: 8.13472834331, 1015: -0.371068140924, 1016: -0.386677930923, 1017: -6.19851420276, 1018: -0.903924345629, 1019: -0.00764009904293, 1020: 3.96086715786, 1021: -10.9994729446, 1022: 6.17027483803, 1023: -0.775884952553, 1024: 1.51377132469, 1025: 0.290450445567, 1026: 9.85259280324, 1027: -6.69056664182, 1028: 2.77958030009, 1029: -2.50923608969, 1030: -3.45492992927, 1031: -0.204844772495, 1032: 1.67448853169, 1033: 3.09261792048, 1034: 0.989626448225, 1035: 0.592337067618, 1036: -7.60271850504, 1037: -0.450713326561, 1038: 4.26214948949, 1039: -1.20684126212, 1040: 1.15844690243, 1041: -0.994586154342, 1042: -6.09964320684, 1043: -3.36705111664, 1044: 14.5860804635, 1045: 3.85976247369, 1046: 4.65253410936, 1047: 5.81219191955, 1048: 0.501367408126, 1049: 24.5711036921, 1050: 0.883422033581, 1051: 7.8822673051, 1052: -0.71893944229, 1053: 5.68405512728, 1054: -17.8480310753, 1055: -18.7197702404, 1056: 9.58568069584, 1057: -5.3849948357, 1058: 13.6300310506, 1059: -6.32811979358, 1060: 10.2864832809, 1061: -8.89685628728, 1062: -7.78534799501, 1063: -4.1253595721, 1064: 7.53174375197, 1065: -50.005255216, 1066: -8.64840751688, 1067: 8.28687955684, 1068: -19.739264637, 1069: -31.7091997385, 1070: -11.8593173817, 1071: -10.9650004938, 1072: 7.5448556685, 1073: 12.988244526, 1074: -17.7156794524, 1075: 11.0073791385, 1076: -25.870805024, 1077: 7.22702490592, 1078: 6.08899793558, 1079: 38.5018669902, 1080: 7.57267095995, 1081: 9.51144742546, 1082: -2.73813682834, 1083: 3.38534524696, 1084: 3.89965587293, 1085: -23.2749019342, 1086: -8.77731105446, 1087: -5.44695677975, 1088: 3.54070753738, 1089: 4.63492514895, 1090: 8.26875090031, 1091: 4.19079456539, 1092: -3.90125741989, 1093: 1.03559026977, 1094: 1.03392059677, 1095: -1.80073120895, 1096: -7.18806549026, 1097: 2.1172505736, 1098: -17.5849218593, 1099: 1.90841708745, 1100: 4.56043596646, 1101: 2.73408449007, 1102: -8.70839332931, 1103: 4.88137171303, 1104: 17.0293917785, 1105: 2.07818942983, 1106: 7.64345262894, 1107: 9.03151167402, 1108: 15.7911507143, 1109: -14.0148499081, 1110: -0.795929337967, 1111: 0.216678624752, 1112: 0.81951651303, 1113: 0.821653448425, 1114: 1.91708658824, 1115: -0.0595773766142, 1116: 2.74261027147, 1117: 4.72692149966, 1118: 4.70098115351, 1119: 3.96464849109, 1120: 4.69091832197, 1121: 4.66324753235, 1122: 2.10772247855, 1123: 4.69212273422, 1124: 3.98437248663, 1125: 4.71208795673, 1126: -2.84244954679, 1127: -26.8490979482, 1128: 3.95736389006, 1129: -3.59686442598, 1130: 4.65129214139, 1131: 13.3522358095, 1132: 6.44927718433, 1133: 4.74470683732, 1134: -5.13276206189, 1135: 10.5051591272, 1136: 4.91848484332, 1137: 3.53651009943, 1138: -4.1311338491, 1139: 4.8581143552, 1140: -4.75808127659, 1141: -4.15170740249, 1142: 4.83893261852, 1143: 0.597010057715, 1144: -3.38855248816, 1145: 4.76184120066, 1146: 5.08813848683, 1147: -8.48219421332, 1148: -3.38000639403, 1149: 4.77660847313, 1150: -4.00907421628, 1151: -2.1252921928, 1152: 3.31611508361, 1153: -17.0239994638, 1154: 4.15063204505, 1155: 4.8888102774, 1156: 4.93752012117, 1157: 0.140700316445, 1158: -4.90046822813, 1159: 13.0558043113, 1160: -30.1641440398, 1161: 16.8826909694, 1162: -25.8083003041, 1163: -1.60348098576, 1164: 0.117118113069, 1165: 3.74818083906, 1166: 4.82423508141, 1167: -16.3129915406, 1168: 28.5999739645, 1169: 9.26840715276, 1170: 0.814898771928, 1171: -0.0729739647537, 1172: 0.248355780474, 1173: 0.579540936012, 1174: 0.592517242137, 1175: 0.00951263109167, 1176: -4.85194053189, 1177: 0.542519612949, 1178: 4.78595440939, 1179: 1.14352423347, 1180: 19.6078755529, 1181: 1.89026242087, 1182: -1.12912698532, 1183: 5.7448511016, 1184: 1.11279107829, 1185: 10.0496779005, 1186: 1.33101452853, 1187: 1.46755308158, 1188: -8.45884921111, 1189: -1.64451787117, 1190: 8.78764444615, 1191: 0.49693567867, 1192: -1.33390095313, 1193: 0.514244895495, 1194: 0.60344917851, 1195: 11.2116216457, 1196: 0.732599744927, 1197: 4.86795780138, 1198: -0.191221131034, 1199: 0.683411110513, 1200: 0.674733779207, 1201: 0.515129292888, 1202: 0.791825530935, 1203: 4.80255260235, 1204: -0.245852134866, 1205: -0.331081429738, 1206: 1.42233845064, 1207: 0.648809830912, 1208: 4.32768789095, 1209: 2.63485552464, 1210: -3.24116504519, 1211: -7.55955379591, 1212: 32.2795488666, 1213: 3.63202230993, 1214: 4.61652863883, 1215: -6.23295526385, 1216: -18.2071493461, 1217: -0.473919555248, 1218: -2.05693175331, 1219: 0.695527893061, 1220: -5.92440656176, 1221: 0.629699223794, 1222: -1.46897767216, 1223: -1.72638090028, 1224: -0.6072501492, 1225: 1.13575384991, 1226: 0.52063027658, 1227: 0.890814640998, 1228: 0.490353495419, 1229: 0.795724614566, 1230: 3.64833059314, 1231: -3.62229412372, 1232: -4.35230885807, 1233: -2.50302896013, 1234: -3.93969599629, 1235: 3.65453785105, 1236: 4.96138989948, 1237: 0.132978526402, 1238: -1.41114021066, 1239: 1.36457851429, 1240: 0.680433397086, 1241: 0.0846121913082, 1242: 1.21971088435, 1243: 1.67794604025, 1244: -1.81699160054, 1245: 4.29668357893, 1246: 0.474983652731, 1247: 4.82010172675, 1248: 0.637199862039, 1249: -0.137363245658, 1250: 0.568014686794, 1251: 0.485321768186, 1252: -37.3532946251, 1253: 4.71829773129, 1254: 22.0334478003, 1255: 0.644722029363, 1256: 0.868506340242, 1257: 1.3151194158, 1258: 0.869002714189, 1259: 2.9121395619, 1260: -4.77659708023, 1261: 1.85027975406, 1262: 0.601828751517, 1263: 0.277605790745, 1264: -4.12333983613, 1265: 2.15348187001, 1266: -1.60528110059, 1267: 2.78367937071, 1268: -6.26615888673, 1269: 3.53747327997, 1270: 4.43436662978, 1271: 2.39626737545, 1272: -3.90079553704, 1273: 0.875347689783, 1274: 0.163922085749, 1275: 4.89884955826, 1276: 1.00980809766, 1277: -1.68996686723, 1278: -6.95601990441, 1279: 1.10819820645, 1280: -18.9067062926, 1281: -2.24543797882, 1282: 3.20973299164, 1283: 8.8008368108, 1284: -14.973081413, 1285: -6.02067395254, 1286: 4.76626962169, 1287: 4.68806653229, 1288: -15.5929209545, 1289: -32.382215068, 1290: -40.3694627149, 1291: -6.50680347365, 1292: 4.83728485927, 1293: -8.05748377655, 1294: -41.0866365486, 1295: 0.686582980926, 1296: 25.5153175375, 1297: -10.3720583515, 1298: 9.21673075569, 1299: -9.42107076154, 1300: 16.7025564028, 1301: -19.8672081603, 1302: 4.64237874721, 1303: 4.6145814645, 1304: -26.3710543474, 1305: 1.16997732263, 1306: 4.78991129721, 1307: -1.68964444523, 1308: -1.43059675057, 1309: 3.33422220043, 1310: 2.4680755239, 1311: 1.89306700347, 1312: 4.65877899382, 1313: -10.9987724072, 1314: -4.27037844556, 1315: -13.7510605634, 1316: -5.59452963568, 1317: 3.96559468279, 1318: 1.27598279403, 1319: -10.766879679, 1320: 2.39217099264, 1321: 5.61115283657, 1322: -13.0983023505, 1323: -18.479070777, 1324: -8.32449548494, 1325: 2.53408445341, 1326: -4.26003471682, 1327: -10.2245366274, 1328: 3.37038859345, 1329: 2.73378928799, 1330: -1.04485314291, 1331: -4.34109755128, 1332: 1.22912980824, 1333: -4.85572524249, 1334: 4.51605337731, 1335: -10.3322673217, 1336: 4.81078300079, 1337: -0.279169745251, 1338: -4.06113430248, 1339: -10.5892845011, 1340: -12.7083432198, 1341: 10.2374946024, 1342: -13.14963661, 1343: 2.75304770942, 1344: 10.1959085538, 1345: -6.64674211325, 1346: 0.742882612076, 1347: -17.5950956286, 1348: 0.288373812475, 1349: 2.08570758058, 1350: 1.04936325032, 1351: -12.4554702259, 1352: -29.5199192451, 1353: 4.712451419, 1354: -25.0293340704, 1355: 4.78671278995, 1356: 4.71501623254, 1357: 4.84365083118, 1358: -4.10906626665, 1359: 5.11069365535, 1360: 5.05990389702, 1361: 7.41835842463, 1362: 4.67167592141, 1363: 9.88807504877, 1364: 4.82593278982, 1365: -11.7008270292, 1366: -6.06026884909, 1367: 0.0239638421935, 1368: 2.81098191097, 1369: -4.76282239657, 1370: 5.34930235696, 1371: 6.69512737607, 1372: 8.03380954225, 1373: -23.4092757832, 1374: 10.088690299, 1375: 0.950811261108, 1376: 20.265350947, 1377: 7.57497039447, 1378: 13.1861190729, 1379: -8.72499111376, 1380: 1.36380813347, 1381: -4.83726453173, 1382: 0.619563178839, 1383: 10.6104315289, 1384: 5.39617745774, 1385: -7.04772518222, 1386: 4.98306329029, 1387: 0.0550781079914, 1388: 6.07439090509, 1389: 12.6189060221, 1390: 0.664063747628, 1391: 7.03354163498, 1392: -13.2191254191, 1393: 4.64147469069, 1394: 0.572737922894, 1395: 8.56174842253, 1396: -5.44953742849, 1397: 0.213236203069, 1398: -25.3009895354, 1399: -26.0245532295, 1400: 41.239654762, 1401: 60.6954256755, 1402: 13.2424319577, 1403: 26.3984679782, 1404: 5.8367579572, 1405: -12.7749740885, 1406: -7.07163335825, 1407: -8.09284439816, 1408: 100.0, 1409: 31.5949758555, 1410: 98.6144381416, 1411: -40.7133506399, 1412: 6.10888371113, 1413: 57.9370029324, 1414: 68.1226625978, 1415: -12.6082154876, 1416: 1.44408916896, 1417: -61.8596511441, 1418: 26.2579085521, 1419: -17.8801084559, 1420: 53.9002672702, 1421: -2.17510230381, 1422: -7.89203046059, 1423: -90.7132475535, 1424: -13.0147434127, 1425: -28.6068129625, 1426: -17.9678208824, 1427: -24.2618443611, 1428: -29.5483341825, 1429: 44.5398727135, 1430: -34.1096518506, 1431: -10.277453921, 1432: 4.16957367723, 1433: -21.0803204467, 1434: -37.3418653189, 1435: 20.7113410562, 1436: -7.43603856775, 1437: 30.7570646816, 1438: -1.17284329457, 1439: 14.5455803896, 1440: -22.920139902, 1441: 7.1605444263, 1442: -32.0700510864, 1443: 24.6365342504, 1444: 6.97212257077, 1445: -33.2095268435, 1446: 0.908293222933, 1447: -21.0715680151, 1448: -51.1240234185, 1449: -0.765695131964, 1450: -5.37928067677, 1451: -48.915048623, 1452: -5.88428531763, 1453: 71.6269882176, 1454: -8.22082289231, 1455: -42.3398327884, 1456: -43.5776882557, 1457: 10.5514619682, 1458: -29.0592444163, 1459: 6.87630042152, 1460: 64.9019467846, 1461: 0.392898987326, 1462: 0.0203740116163, 1463: -28.7511456957, 1464: -2.17496445547, 1465: 7.75675963197, 1466: -1.1107458514, 1467: -0.966014454437, 1468: 7.96212809919, 1469: -0.539008346558, 1470: -7.51398205864, 1471: -1.53274140053, 1472: 7.86154802999, 1473: 4.86999367308, 1474: 7.48913770686, 1475: -69.6363160258, 1476: -80.6589731348, 1477: -0.829295310031, 1478: 36.8057757531, 1479: -0.982982987855, 1480: -22.4313732149, 1481: 6.23938326938, 1482: 21.6136152148, 1483: -6.92417629318, 1484: 9.05246633989, 1485: -5.55055195955, 1486: 41.5234962151, 1487: 8.09391376973, 1488: -0.832624446874, 1489: 4.14680287671, 1490: 3.43761813973, 1491: -0.606045996488, 1492: 3.95751881752, 1493: -0.805753515158, 1494: -0.798884442566, 1495: -0.144132066743, 1496: -4.80467951343, 1497: 15.9051464907, 1498: 0.459289982691, 1499: -81.2553761015, 1500: -0.103676763476, 1501: -2.62586532372, 1502: 85.8383293605, 1503: 9.50124696103, 1504: -0.809193283797, 1505: -5.65223202505, 1506: -27.1618293387, 1507: 3.4344863847, 1508: 5.58913420537, 1509: 100.0, 1510: -23.0342721868, 1511: 6.96027035019, 1512: -72.4318546139, 1513: 31.067531467, 1514: -8.12681651126, 1515: -33.0375639755, 1516: -65.9444752681, 1517: -53.4297678845, 1518: -85.7808319117, 1519: -14.1172295292, 1520: 43.8012478986, 1521: -6.53683515924, 1522: 0.00146681983652, 1523: -0.68472009344, 1524: -8.70381894235, 1525: -26.0854240545, 1526: -0.0995670900176, 1527: -1.03361377187, 1528: 8.58917154713, 1529: -4.37305653299, 1530: -2.37713321114, 1531: -1.98662564873, 1532: -68.8447946166, 1533: -3.06537769114, 1534: 9.53320466694, 1535: 2.02844449467, 1536: -31.5538667656, 1537: -48.7469327108, 1538: 1.03177307749, 1539: -17.2689464749, 1540: 0.115265259917, 1541: -18.7930696138, 1542: 1.61467541821, 1543: 0.108420725348, 1544: -31.6662734697, 1545: -0.00532369010902, 1546: -0.928132263789, 1547: -35.0076103621, 1548: 3.04156230233, 1549: -0.105514286873, 1550: 0.058339354738, 1551: -0.104500734538, 1552: -0.753692239998, 1553: -64.5526618086, 1554: 9.64250688778, 1555: -1.08209085104, 1556: 1.30406903644, 1557: 10.4001422379, 1558: -2.8680050041, 1559: -0.017603400978, 1560: 32.0862121472, 1561: 63.9877807806, 1562: 90.0775840193, 1563: -12.1511662019, 1564: 0.293305443317, 1565: 9.93326181434, 1566: -29.8981303829, 1567: -69.8161285363, 1568: 0.0781533633776, 1569: -9.30690185388, 1570: -0.00204432270558, 1571: 18.8506883602, 1572: 54.3503287493, 1573: 34.7717262248, 1574: 1.26997401424, 1575: 0.0451987988776, 1576: 0.0939191021362, 1577: -0.297388098992, 1578: -0.451985822262, 1579: 9.06246501055, 1580: -49.1990230325, 1581: 14.7087192683, 1582: 12.7443001613, 1583: -32.5110572867, 1584: 12.1673659346, 1585: -18.9628763301, 1586: 3.25120066852, 1587: 6.52494001736, 1588: -24.0720232351, 1589: -0.0654625105829, 1590: -35.8960582072, 1591: 2.14348706084, 1592: -3.48297676374, 1593: 4.78882794909, 1594: 89.6414641286, 1595: 0.211503504406, 1596: -0.871995821845, 1597: 0.223632275672, 1598: -21.710820879, 1599: 0.143104237679, 1600: 0.132803088902, 1601: 6.37059568617, 1602: -0.918616409811, 1603: -30.9754348946, 1604: 0.0482338636865, 1605: 0.131405934064, 1606: -0.183710049618, 1607: -0.0689210005003, 1608: -1.04106345976, 1609: 9.1201048846, 1610: -6.08175932408, 1611: 0.15062458424, 1612: 0.287042524607, 1613: -43.2056447337, 1614: -3.98646973848, 1615: 4.24300632531, 1616: -28.7826416637, 1617: 31.8606233663, 1618: -12.6291458036, 1619: -5.05461884193, 1620: 1.93648910749, 1621: -1.14262458809, 1622: 0.222807004153, 1623: -64.7012929461, 1624: -1.6817735337, 1625: 3.22047799473, 1626: 0.512753374262, 1627: 16.4561956962, 1628: 0.0446117159064, 1629: -21.9608916125, 1630: 1.61548968401, 1631: -2.12364611063, 1632: 14.2114089154, 1633: -15.6529072073, 1634: 38.681428383, 1635: 8.66145380917, 1636: -0.879052422259, 1637: 15.2534626103, 1638: 27.6857529615, 1639: 60.2489853248, 1640: 10.6051705506, 1641: -29.8714116427, 1642: 33.7077663901, 1643: 2.54933292381, 1644: -78.3969864952, 1645: -61.5779096772, 1646: 20.1573999764, 1647: -83.4231959463, 1648: -51.702577509, 1649: 99.8823458506, 1650: -5.65880732922, 1651: -47.0435667033, 1652: -0.13611048739, 1653: 32.5130413522, 1654: 5.20975893625, 1655: -0.893207043577, 1656: -37.5187574008, 1657: 41.7687604811, 1658: 4.86091684031, 1659: 6.88489070471, 1660: 12.4927837829, 1661: -0.688005993506, 1662: 45.6374031131, 1663: 53.689810306, 1664: 31.7467270681, 1665: 26.2436602642, 1666: 8.34012197816, 1667: -2.04715516947, 1668: 25.2043101985, 1669: -30.8980147271, 1670: 29.3240071663, 1671: -100.0, 1672: 24.2072091015, 1673: 25.9744095208, 1674: 14.0464413307, 1675: -42.8354827589, 1676: -4.56994379528, 1677: -0.240366620841, 1678: -17.7513590553, 1679: -19.535910362, 1680: -100.0, 1681: 14.1002613536, 1682: 22.9883466886, 1683: -2.02131582472, 1684: 10.3874041657, 1685: 20.601829986, 1686: 6.35845199528, 1687: 17.1121584494, 1688: -26.1580976218, 1689: 17.3501717687, 1690: 40.4123039222, 1691: 64.5829821044, 1692: 12.3180749602, 1693: 32.7335737875, 1694: 12.5218590781, 1695: 7.92778259668, 1696: 45.4214587904, 1697: 1.81220779195, 1698: 35.4336806833, 1699: 5.17357838417, 1700: 18.8452208649, 1701: 26.6332963627, 1702: -0.375819400453, 1703: 13.4395315615, 1704: -0.833523741441, 1705: -0.761346437548, 1706: 9.37366717366, 1707: 17.1561510264, 1708: -0.803201360735, 1709: -0.58739387685, 1710: -3.53842843082, 1711: -0.890768935265, 1712: 20.4292228379, 1713: -0.97103863146, 1714: -0.965932324763, 1715: -10.2221257774, 1716: 0.260917723734, 1717: -0.987360845184, 1718: 0.18868964223, 1719: -8.76792098721, 1720: -9.00580389829, 1721: -18.1259489599, 1722: 75.7707311897, 1723: 6.34610585934, 1724: 3.76097001306, 1725: 31.9627672834, 1726: 25.3823627963, 1727: -12.6080953643, 1728: -34.1192722567, 1729: 1.62548194244, 1730: 21.2502739126, 1731: -21.2459864399, 1732: 58.8503481889, 1733: -45.8651207827, 1734: -68.9619176167, 1735: -0.902598611099, 1736: -51.6331315724, 1737: 3.34797009607, 1738: 6.55119285811, 1739: -2.60261561814, 1740: 20.6197357883, 1741: 20.2962418846, 1742: 79.7007916485, 1743: 23.7071887184, 1744: 32.2358975777, 1745: 21.0012031743, 1746: -0.355937346161, 1747: -100.0, 1748: -9.27782004499, 1749: 4.80504474347, 1750: 27.1167419899, 1751: 29.3290814141, 1752: 37.0824869015, 1753: -2.02502969321, 1754: 44.0443073637, 1755: -5.51941476091, 1756: 3.57860137413, 1757: 23.4369433934, 1758: -11.5792341744, 1759: -16.4127123053, 1760: -0.446555201744, 1761: -15.5498922691, 1762: -6.61411648369, 1763: 1.15394636112, 1764: -70.5711250248, 1765: -1.29636948203, 1766: 0.59857424517, 1767: -3.20854065525, 1768: -11.1660711857, 1769: 17.149762606, 1770: 21.0018220226, 1771: 15.998080529, 1772: -44.0776257553, 1773: -1.08645028226, 1774: 11.0520587124, 1775: -61.8795321721, 1776: -42.5043495843, 1777: 21.0031325472, 1778: -12.5136485192, 1779: -28.7767718476, 1780: -7.11114458223, 1781: -19.1904399428, 1782: 12.7101027524, 1783: -8.65336589056, 1784: -42.3841046024, 1785: 30.0270074566, 1786: 6.7545834837, 1787: -18.7699468842, 1788: -17.7881390983, 1789: -7.63195861506, 1790: 0.687102244145, 1791: -7.53584229389, 1792: -24.4349599976, 1793: 11.9775294584, 1794: -4.69982706941, 1795: -13.1437587121, 1796: 3.51195262359, 1797: -32.2533766159, 1798: -2.80395780489, 1799: -2.3462975314, 1800: -6.02477478358, 1801: -2.80105580665, 1802: 6.8746397885, 1803: -2.55289114855, 1804: -13.3921044973, 1805: -9.71638095834, 1806: -7.35706074305, 1807: -16.3844351041, 1808: -53.503612179, 1809: 7.51077031853, 1810: 23.5026061807, 1811: 7.23917065884, 1812: 13.5652465332, 1813: -2.55945904471, 1814: -10.1211357009, 1815: -2.66734516665, 1816: -2.49779375063, 1817: -8.79147548752, 1818: -2.84855094314, 1819: -41.4066072192, 1820: -5.41935145169, 1821: -18.826685992, 1822: -16.3955852546, 1823: -8.64902966636, 1824: -68.9974990221, 1825: -31.4163255734, 1826: -1.46388715062, 1827: 17.3340555618, 1828: -3.00990996578, 1829: 18.1762742801, 1830: 16.0311794473, 1831: -6.52353088975, 1832: 8.85190584463, 1833: -30.4794834083, 1834: -2.47542047176, 1835: -7.56772952865, 1836: -1.39855688999, 1837: -2.67996742216, 1838: -38.8433227036, 1839: -0.53641934591, 1840: -2.27168673947, 1841: 7.96632481758, 1842: 3.76513173668, 1843: -2.20673584145, 1844: 13.5068610982, 1845: 42.7245132164, 1846: 21.8854044136, 1847: -2.43221863829, 1848: 4.37622315241, 1849: -1.401763502, 1850: 1.78767050838, 1851: -5.4953909122, 1852: -2.84846420224, 1853: -3.44851687658, 1854: -2.6867873966, 1855: 1.63874360364, 1856: -4.28448101604, 1857: -8.04637830163, 1858: -0.404975404462, 1859: -30.7621632824, 1860: 24.2542651855, 1861: -26.3688733633, 1862: -3.0217544898, 1863: -5.25080523788, 1864: 13.3885870648, 1865: -32.3113197547, 1866: 2.2810626818, 1867: -37.0613052301, 1868: 7.28893961995, 1869: -16.3850706391, 1870: 28.9599850385, 1871: -0.521645642171, 1872: -0.732825698761, 1873: -10.0894624168, 1874: 24.2444269012, 1875: -0.513143559493, 1876: -2.67137839054, 1877: -0.653029401312, 1878: -35.841146763, 1879: -1.91496001664, 1880: -2.647020811, 1881: 16.1113022262, 1882: 1.51295677135, 1883: -3.71797277496, 1884: 1.7440665663, 1885: -1.13260429138, 1886: -32.1157115908, 1887: -16.8236658685, 1888: -19.9518770001, 1889: 0.604064385825, 1890: 15.6601470742, 1891: -2.09845789993, 1892: -0.629272580221, 1893: 5.04902114342, 1894: -0.50286709287, 1895: -2.75332042527, 1896: -2.65041299108, 1897: -0.523601951594, 1898: -0.528950371928, 1899: -0.58884028122, 1900: -0.420370816731, 1901: -2.92588335941, 1902: -1.48862331598, 1903: -0.590504653705, 1904: -0.413067144695, 1905: -1.05068430246, 1906: -2.66756771568, 1907: -2.86981589957, 1908: -13.9834419043, 1909: 4.22410216698, 1910: -11.6791749361, 1911: 18.4811070693, 1912: 7.42744801151, 1913: 3.5817346618, 1914: 7.66964618439, 1915: 5.14502758757, 1916: 6.36895469242, 1917: -0.358258723026, 1918: -45.7541371129, 1919: -0.618423750184, 1920: 26.8051028302, 1921: -15.609798766, 1922: 41.8736061647, 1923: 1.10578599788, 1924: -0.496117791258, 1925: -0.535035507156, 1926: -3.29156099689, 1927: -0.0680708848094, 1928: -21.2498026968, 1929: -3.84494164689, 1930: 23.1850947381, 1931: 34.0831228891, 1932: -37.9908371819, 1933: 6.87842575042, 1934: -3.42108218571, 1935: 0.960585075214, 1936: 1.56215639508, 1937: 13.9112203705, 1938: -0.565282690132, 1939: -53.4194512624, 1940: -11.6438420628, 1941: -0.282560732322, 1942: -8.52889042, 1943: -11.510759033, 1944: -0.357616868245, 1945: -3.53014100131, 1946: -0.46741889819, 1947: -33.4678970234, 1948: -0.588067501346, 1949: -0.497304303578, 1950: 38.5402831726, 1951: -2.5997949117, 1952: 16.3062007759, 1953: -0.363100521634, 1954: -0.427946851277, 1955: -1.4891369206, 1956: -0.420718592616, 1957: 0.629007988622, 1958: -15.4598056785, 1959: 13.6808064482, 1960: 1.19161466408, 1961: -0.311475491355, 1962: 38.6276804565, 1963: 8.93745567019, 1964: -10.2583196079, 1965: -25.5580763994, 1966: 2.42545748032, 1967: 4.83986915043, 1968: 5.08961842883, 1969: -2.83821606803, 1970: 10.4634959191, 1971: -0.730212236946, 1972: 17.0885821084, 1973: -17.3229948508, 1974: -7.04344280712, 1975: 0.279280650325, 1976: 13.2867640723, 1977: 0.71751176814, 1978: -11.6090630539, 1979: 7.06678990527, 1980: -3.19695866179, 1981: 1.57724153409, 1982: 40.8152164402, 1983: 13.6457182042, 1984: -3.14887491413, 1985: -2.63971307694, 1986: 1.66597291277, 1987: 36.1837628806, 1988: -9.50236900967, 1989: 17.0087827288, 1990: 24.9190073793, 1991: 12.9052961031, 1992: 8.14735385114, 1993: -16.5812120509, 1994: 18.9445760756, 1995: -18.4338335879, 1996: 14.0531289347, 1997: 39.4314872675, 1998: 18.1578308314, 1999: 18.9764201843, 2000: -5.39848971957, 2001: -2.42561157657, 2002: 3.6968311655, 2003: -1.8175818679, 2004: -2.55761077215, 2005: -2.51950071214, 2006: 2.64862708436, 2007: -2.91306223672, 2008: -4.12539212427, 2009: -7.29554907734, 2010: -2.9895722435, 2011: -2.17971962239, 2012: 6.83966475588, 2013: -3.07032248394, 2014: 4.23619887695, 2015: -2.74764325988, 2016: -2.52275301469, 2017: -6.23881540501, 2018: -3.82684163766, 2019: 38.6758606655, 2020: 38.4659413551, 2021: 4.22880330535, 2022: 18.0962289484, 2023: 14.2792851305, 2024: 2.58159085156, 2025: -2.32619773727, 2026: -4.71315303855, 2027: 24.6897496133, 2028: -4.55783607646, 2029: -8.19915077739, 2030: 2.92223558145, 2031: -12.5230874423, 2032: -2.47389508511, 2033: -8.56181880117, 2034: 25.0339512987, 2035: 4.49427408686, 2036: 10.0398235708, 2037: 2.24551044195, 2038: 11.3787435398, 2039: -30.2419766706, 2040: -5.18196625562, 2041: 44.6192564352, 2042: 13.7832132288, 2043: 13.0566370654, 2044: 0.624725760419, 2045: -23.7028572787, 2046: -4.4311454184, 2047: 36.9855186225, 2048: -4.69989557231, 2049: 16.1865281159, 2050: 31.4828987797, 2051: -2.43230041845, 2052: 35.5787322401, 2053: -3.46239080367, 2054: -3.50061093844, 2055: -1.46096234001, 2056: -19.5222531845, 2057: -2.42615559948, 2058: -3.03267913843, 2059: -34.9105985487, 2060: -2.7086630355, 2061: -22.8803188722, 2062: -2.69533998552, 2063: -2.76270541277, 2064: -21.4966192391, 2065: -4.78924861837, 2066: 1.88884449736, 2067: -14.4884262253, 2068: -13.0112384814, 2069: 35.3014915499, 2070: -14.7292281135, 2071: 51.3936578931, 2072: -2.64633369981, 2073: -2.337614135, 2074: -29.5347577863, 2075: -9.78402366822, 2076: 16.6855380601, 2077: -15.8157418153, 2078: -3.13651752153, 2079: -33.6640503468, 2080: -1.65083248703, 2081: 100.0, 2082: 2.85538197513, 2083: 15.8468313867, 2084: -2.85564221341, 2085: 14.7460005975, 2086: 12.7531232739, 2087: 0.582298816743, 2088: -4.81686298224, 2089: -51.2911069298, 2090: -9.65433974857, 2091: 97.0674231781, 2092: 3.13912774851, 2093: -12.0313353272, 2094: 37.9763208171, 2095: -0.27788174798, 2096: -2.70414129611, 2097: -3.14668332973, 2098: -5.38761996737, 2099: -2.54686253904, 2100: 2.73237877459, 2101: 8.69119154678, 2102: 15.7695232759, 2103: -16.2522845254, 2104: -13.4256549864, 2105: -17.1938387668, 2106: 21.0395833171, 2107: -11.856511837, 2108: 25.5796473196, 2109: 6.06104137319, 2110: -14.1631685967, 2111: -19.8944842609, 2112: 12.489900543, 2113: 27.9880548508, 2114: 13.2667246848, 2115: -11.8985143312, 2116: 39.557793557, 2117: 17.2720285269, 2118: -8.91597487947, 2119: -4.24454379162, 2120: -8.38224304089, 2121: 7.10905134256, 2122: -2.13027863487, 2123: 70.364771754, 2124: -13.4953780925, 2125: -10.7217728157, 2126: -41.8966029256, 2127: -6.7734340617, 2128: 4.09763654599, 2129: 11.3763730127, 2130: -6.27589034906, 2131: -6.5702319333, 2132: -3.32211267719, 2133: 5.38644245582, 2134: 7.12826259385, 2135: 3.54095702522, 2136: -2.23866832386, 2137: -4.27841461165, 2138: -16.9740569792, 2139: -3.90440009495, 2140: 3.91796771598, 2141: 9.33571987589, 2142: 2.34122217112, 2143: 14.3121547836, 2144: -3.09296361746, 2145: 6.00063028705, 2146: -25.9039742264, 2147: -4.73279980438, 2148: 5.77731826644, 2149: 7.03494122274, 2150: -2.23386673564, 2151: -3.90917639407, 2152: -2.13981919405, 2153: 12.7140378516, 2154: -10.3086834733, 2155: 20.2209876304, 2156: -1.55950306951, 2157: -6.78418177878, 2158: -1.83951387753, 2159: -6.5606491928, 2160: -2.63239024266, 2161: 10.4575428665, 2162: -2.38316140223, 2163: -4.76299663422, 2164: -2.31805288166, 2165: -2.52540063315, 2166: 8.70083873684, 2167: -2.86676875696, 2168: -15.5720280068, 2169: -1.5620830929, 2170: -2.54609564626, 2171: 0.574747830883, 2172: 3.67055189481, 2173: 2.88560715932, 2174: 11.6000099416, 2175: 2.42615480942, 2176: 23.9521863179, 2177: -2.90463814768, 2178: -1.65255008509, 2179: -0.367611942315, 2180: -2.62969787954, 2181: 3.60240871977, 2182: 6.70589631125, 2183: 0.680025536207, 2184: -2.9116803226, 2185: -0.265075796497, 2186: -2.29036349728, 2187: 11.5208311338, 2188: 3.80327442418, 2189: -2.68516632786, 2190: -1.61977516854, 2191: 7.18675165987, 2192: -2.4127887157, 2193: -14.6458399632, 2194: 0.594933087236, 2195: -0.551773058339, 2196: -2.41236510467, 2197: -3.192139534, 2198: -2.5069636235, 2199: -1.53177585525, 2200: 16.1164574327, 2201: -2.34117511459, 2202: -2.57586900878, 2203: 0.578290415103, 2204: 2.89053369572, 2205: 3.07129259312, 2206: -3.46946881828, 2207: 39.3006892034, 2208: -14.5030396885, 2209: 8.07238428926, 2210: 14.4897830463, 2211: -3.93883444379, 2212: -12.8512575497, 2213: -10.5559081626, 2214: 5.96447596351, 2215: -24.3115716112, 2216: -11.045111202, 2217: -11.8715842908, 2218: 3.14458801759, 2219: 3.23505422553, 2220: -0.44887049292, 2221: -0.493524807305, 2222: 4.37291291981, 2223: 18.3757562405, 2224: -0.320838198216, 2225: -2.37245986062, 2226: 3.76197586901, 2227: -6.85877621597, 2228: -0.063640112433, 2229: -1.01572781925, 2230: -6.5721742529, 2231: -0.031042273804, 2232: 20.8322050766, 2233: 0.101601749108, 2234: -5.61585064615, 2235: 10.240147357, 2236: -2.63330416156, 2237: 4.44150154544, 2238: 0.527768618013, 2239: 1.32638395619, 2240: -0.462420916947, 2241: -0.346806069558, 2242: -20.4183396708, 2243: -0.413712875561, 2244: -1.99783678954, 2245: 2.43853014625, 2246: -1.13004414178, 2247: -0.558561536258, 2248: -0.379339193657, 2249: -0.571389767595, 2250: -2.80661453153, 2251: 13.4012217196, 2252: -0.206651722799, 2253: -0.501708399293, 2254: -0.826736688102, 2255: 1.237676359, 2256: -6.85138079512, 2257: 21.1352115723, 2258: -0.376400929838, 2259: 18.6734115424, 2260: -27.3623581859, 2261: -2.65334177861, 2262: -1.28993493525, 2263: -3.26668412061, 2264: 2.45838367257, 2265: 18.037163592, 2266: -0.711750767084, 2267: -4.19032133429, 2268: -0.345377011642, 2269: 18.9303703125, 2270: -22.0506092732, 2271: 11.9342422784, 2272: -0.190679619018, 2273: -0.3040193698, 2274: -0.757095475512, 2275: -0.541608613231, 2276: -0.422940752578, 2277: -0.663120833461, 2278: 6.15038455885, 2279: 3.32569748977, 2280: -2.51107228573, 2281: -7.46579141246, 2282: 1.7236908015, 2283: 3.82548241209, 2284: 3.3897758252, 2285: -2.45979819963, 2286: -10.7384241709, 2287: -0.947097969651, 2288: 14.8434654875, 2289: -0.475427379268, 2290: -0.252517864673, 2291: 0.141376903338, 2292: 3.97981617924, 2293: -0.497476509722, 2294: -2.45253653604, 2295: -0.522491002028, 2296: -11.863441954, 2297: -0.628186731292, 2298: -0.49142768611, 2299: 2.81455578742, 2300: -2.48194458483, 2301: -24.7885569171, 2302: -0.451586482446, 2303: -0.442001302242, 2304: -0.442385350802, 2305: -0.727729605628, 2306: -2.09395251716, 2307: 0.562391314499, 2308: -2.4596324926, 2309: -0.529332805001, 2310: -0.712618897814, 2311: -9.8966485985, 2312: -2.80015184795, 2313: -4.06202126132, 2314: -6.39375525355, 2315: -1.89225470075, 2316: 1.45417015931, 2317: 0.294490961124, 2318: -0.776355286965, 2319: 0.712112422819, 2320: -1.49925938826, 2321: -6.14379621961, 2322: -2.98155631342, 2323: -0.737797463811, 2324: -0.633722200055, 2325: -3.84307490365, 2326: -0.341704739845, 2327: -3.25976950732, 2328: -0.0503947748874, 2329: -2.17361031592, 2330: -8.14644424977, 2331: -0.708654741539, 2332: 4.80055669444, 2333: -2.79092251272, 2334: -2.43723420556, 2335: -3.68897005688, 2336: 0.620799512952, 2337: 1.0685947675, 2338: 1.43581152992, 2339: 11.016847984, 2340: -19.9678300671, 2341: 26.5852443284, 2342: -5.19755606508, 2343: -10.5732746058, 2344: 13.5900133404, 2345: 20.4658287749, 2346: -1.7169066452, 2347: -13.979300216, 2348: -5.19823882126, 2349: -2.75528183033, 2350: -2.72440325575, 2351: 25.0529237327, 2352: -2.53473895803, 2353: -1.98651101491, 2354: 3.78448542365, 2355: 1.28522057621, 2356: 1.31042178326, 2357: 0.332752376519, 2358: 0.356449797681, 2359: -2.80049312574, 2360: -3.41677446366, 2361: 8.74565216982, 2362: -1.50410212262, 2363: 6.26550860881, 2364: -4.42572852309, 2365: -2.5750186909, 2366: 3.11992385683, 2367: -10.7508644612, 2368: -5.379783425, 2369: 39.2812730493, 2370: 19.738334399, 2371: -10.705270141, 2372: -9.32804511049, 2373: 7.51007263846, 2374: 20.193893679, 2375: 3.37392904715, 2376: -7.98611702197, 2377: 11.4051075184, 2378: 8.66458325129, 2379: 7.24925503899, 2380: -5.47643247682, 2381: -2.7987539049, 2382: 1.31612405374, 2383: -2.31342775267, 2384: -8.04133410393, 2385: -3.64413448739, 2386: 10.7237585441, 2387: 1.41755454016, 2388: -3.64448058625, 2389: -16.2986987095, 2390: 0.255169735461, 2391: -7.90447091757, 2392: 7.64187068232, 2393: -2.7544330822, 2394: -4.59671836073, 2395: 3.22750639961, 2396: -7.72795269351, 2397: -1.18499641849, 2398: 4.91400598216, 2399: 19.2806575758, 2400: -2.72943340759, 2401: 20.169253042, 2402: -2.52797138861, 2403: -2.40291999459, 2404: -2.72225892577, 2405: 5.59389515146, 2406: 3.37277528177, 2407: -2.25959693468, 2408: 4.13794304603, 2409: -2.37786218404, 2410: -15.6479895407, 2411: -2.34953681556, 2412: -2.45644792961, 2413: 31.5420518164, 2414: -2.68980494762, 2415: -1.90985497719, 2416: -0.367042049038, 2417: 0.192077188311, 2418: 4.28486737992, 2419: 2.7997286242, 2420: 15.1918349856, 2421: -6.59197874173, 2422: -7.28172612861, 2423: 22.4325204448, 2424: 9.95925406856, 2425: -14.896223198, 2426: 22.7445127361, 2427: 0.744035695425, 2428: 2.01676342098, 2429: 17.0451373492, 2430: -13.9375960571, 2431: 2.34942323016, 2432: -17.7781021051, 2433: -2.33254955414, 2434: 6.21755394178, 2435: -12.2598528944, 2436: 5.4285855114, 2437: -5.88723777833, 2438: -17.7281694263, 2439: -1.49356636372, 2440: -5.98228365197, 2441: -1.53449434857, 2442: 9.00171456606, 2443: 3.01329543335, 2444: 0.651037991125, 2445: -0.348562452028, 2446: 17.242997817, 2447: 42.7038507994, 2448: -55.7035998741, 2449: 9.07442706196, 2450: 6.25115118091, 2451: 34.1536621559, 2452: -3.54136303242, 2453: -53.9311010656, 2454: -100.0, 2455: 15.9685480821, 2456: 4.64584805411, 2457: -12.2189696004, 2458: 23.4819635272, 2459: -6.16261407483, 2460: -9.53328807607, 2461: -15.6503621251, 2462: -37.1691890288, 2463: 54.0043077596, 2464: -7.77733711757, 2465: -13.9375139394, 2466: -3.15147423548, 2467: -43.5795389367, 2468: -4.24086882431, 2469: -16.6796915741, 2470: 20.2573516867, 2471: 5.24461848561, 2472: -35.8878250677, 2473: 5.17792838836, 2474: -53.1716809229, 2475: 27.0609723538, 2476: -13.1719056701, 2477: 29.6049311612, 2478: 44.0641937776, 2479: 6.07221585258, 2480: -30.3070910533, 2481: -21.7359380623, 2482: -1.27133811719, 2483: -6.24219070373, 2484: 1.98728836055, 2485: -22.8878253397, 2486: -30.5943765578, 2487: -12.7558233016, 2488: -13.5005080758, 2489: 15.2786443569, 2490: -0.505320954241, 2491: 5.42506281938, 2492: -39.9449268674, 2493: -11.79617339, 2494: -6.40747851342, 2495: 14.0877499751, 2496: 4.3872996341, 2497: -22.8057551353, 2498: -11.4677401979, 2499: 7.6818524872, 2500: -16.5865353071, 2501: 11.7109559365, 2502: 27.4442846289, 2503: -37.0653570402, 2504: 28.4343758844, 2505: -2.49628425165, 2506: 54.3820254256, 2507: -8.70442635739, 2508: -40.2925726828, 2509: -2.51522051097, 2510: 30.9986707449, 2511: 4.54153835911, 2512: 23.5852070062, 2513: 4.65656518257, 2514: 4.78326680036, 2515: -20.6023539516, 2516: 4.2695601726, 2517: -36.3488451045, 2518: 2.09357592912, 2519: 2.15284281554, 2520: 13.8707151893, 2521: 4.80396926484, 2522: -27.9700727124, 2523: 1.30334852494, 2524: -0.0323093223327, 2525: -18.9146239404, 2526: 1.3277168226, 2527: 3.46692274855, 2528: -14.902957091, 2529: 3.77830490943, 2530: 3.82896638348, 2531: 2.40083039257, 2532: 5.00498324593, 2533: 5.20358568637, 2534: 6.29116169502, 2535: 4.12640057597, 2536: -62.0905359106, 2537: 4.62489833143, 2538: 3.09388123458, 2539: 0.52961424143, 2540: 1.99859388164, 2541: 4.47739623168, 2542: 4.36662509873, 2543: 5.82560710264, 2544: -37.5048209622, 2545: -4.01905945817, 2546: -14.4936242217, 2547: 9.73982371411, 2548: -2.58416222672, 2549: -17.2590342359, 2550: 15.5728050326, 2551: 2.99929682527, 2552: 5.24325402947, 2553: -39.4676131392, 2554: 0.854351936633, 2555: -29.2264823452, 2556: 20.5541723497, 2557: 26.1397325517, 2558: 1.75537134279, 2559: 4.19795740993, 2560: 15.5657327566, 2561: 1.51815422971, 2562: -7.81233881227, 2563: -18.1125802157, 2564: -29.46883296, 2565: 12.9401571267, 2566: 34.4825988087, 2567: -22.2088160877, 2568: 59.4023748565, 2569: 1.22767115904, 2570: 2.20509922025, 2571: 2.07083215959, 2572: 35.3499802334, 2573: 0.929725549622, 2574: 1.93769808261, 2575: -4.11164860155, 2576: 10.6614229565, 2577: 0.792051470155, 2578: 4.89748616562, 2579: -90.9135247098, 2580: 3.45852446393, 2581: -39.7388471385, 2582: 1.36879040605, 2583: -0.660632243251, 2584: 66.6652416019, 2585: 26.7734366503, 2586: 46.864140455, 2587: 4.02062858478, 2588: -10.1456841776, 2589: 0.86773431033, 2590: 0.925626416977, 2591: -4.05692104541, 2592: 0.735206285912, 2593: 4.94443222479, 2594: 1.24786581651, 2595: 1.18375656283, 2596: 0.807518799727, 2597: 0.950575777898, 2598: 0.847057647174, 2599: 7.66379246243, 2600: -0.186548181099, 2601: 0.530635959043, 2602: 1.02449893952, 2603: 1.3152169432, 2604: -12.1589920518, 2605: -12.3032683588, 2606: 0.0744772959554, 2607: 29.8096952354, 2608: -100.0, 2609: -37.019974132, 2610: 0.372467384805, 2611: 0.734752732661, 2612: -11.3651880449, 2613: 5.53585361665, 2614: -1.29969239716, 2615: 0.711968870834, 2616: -14.8547898658, 2617: 0.739473412927, 2618: -15.9170648391, 2619: -12.584625218, 2620: 21.74457665, 2621: 3.28807764596, 2622: 0.920077945242, 2623: 0.272800275471, 2624: 0.804392677975, 2625: 1.1911771784, 2626: 10.0423881512, 2627: 57.8322028302, 2628: -29.8345818677, 2629: 25.2866567169, 2630: 35.5508210918, 2631: -0.0214491034716, 2632: -29.2137154671, 2633: -16.6921662646, 2634: -17.5734409605, 2635: -30.0138362236, 2636: 1.71674565167, 2637: 50.4084728345, 2638: 1.01925863395, 2639: -1.44211522427, 2640: -2.56642213607, 2641: 1.13353315415, 2642: 0.846646743335, 2643: 2.68245793657, 2644: 0.941504235701, 2645: 19.8651012306, 2646: 0.896377694825, 2647: 0.973402155297, 2648: -30.4683915454, 2649: 4.6117858942, 2650: 91.0004754435, 2651: 0.821003513162, 2652: 0.866338376655, 2653: 9.93457304467, 2654: 0.802560719357, 2655: 5.20868236578, 2656: -7.765321013, 2657: 10.3804570318, 2658: 2.31814225378, 2659: 2.9622115429, 2660: -18.1320078927, 2661: 2.67121838661, 2662: 41.947661989, 2663: -0.763853538883, 2664: -3.95869275385, 2665: 15.2481844778, 2666: 38.550421389, 2667: 0.3356937829, 2668: -8.25120882824, 2669: -0.939861242762, 2670: -2.88108578895, 2671: -6.62111525229, 2672: 11.7939861615, 2673: 3.04657097919, 2674: -12.8262620816, 2675: 1.16571844401, 2676: 31.3906326757, 2677: -50.0799476305, 2678: -12.1665278441, 2679: 2.26880485195, 2680: 7.28996940641, 2681: 38.5562509204, 2682: 4.12886118658, 2683: 1.94350009125, 2684: -46.5931133982, 2685: -27.1585337953, 2686: -36.4815195524, 2687: 11.7436522627, 2688: 41.5576931384, 2689: -37.4844377463, 2690: -17.3038574155, 2691: 10.7562559752, 2692: 1.56320943743, 2693: -36.1183744536, 2694: -17.4939383334, 2695: -8.33945466801, 2696: 15.96217686, 2697: -24.2747490281, 2698: 6.07193573221, 2699: 2.89463022427, 2700: -57.8286635268, 2701: 6.13572113123, 2702: 5.04899123798, 2703: 1.06476289664, 2704: 8.69248661455, 2705: 3.51083500762, 2706: 11.8876411308, 2707: 5.77066846695, 2708: 7.89001505644, 2709: 15.8955836521, 2710: 53.5358545358, 2711: 13.3668579322, 2712: -33.3287377529, 2713: -22.0830945061, 2714: -8.93297814607, 2715: 48.53212629, 2716: 61.511468513, 2717: 81.0573174579, 2718: -41.9134498051, 2719: 33.1327527953, 2720: 12.4219178809, 2721: -0.947306465529, 2722: 7.78501119617, 2723: -18.7755795216, 2724: 7.82883652908, 2725: 1.94921207744, 2726: 13.1793660794, 2727: 29.7653068185, 2728: 55.2643643744, 2729: -21.9827144003, 2730: 4.59819568345, 2731: 32.4130617696, 2732: 4.71261376092, 2733: -13.3431131339, 2734: 9.56740547146, 2735: -24.1712471975, 2736: -12.2218232179, 2737: -38.1880708355, 2738: -2.62056867669, 2739: 60.1520103308, 2740: -1.05118393598, 2741: -50.8586837702, 2742: -8.36505216332, 2743: 17.0845185076, 2744: -17.2038625067, 2745: 3.70092577622, 2746: -0.095757089172, 2747: -29.0162363122, 2748: 66.2618712406, 2749: 3.29238095714, 2750: -77.9079264886, 2751: 2.73281414216, 2752: 2.6632020295, 2753: 1.57448950043, 2754: 18.2251503548, 2755: 4.46308162599, 2756: 3.4437642262, 2757: 19.1150782374, 2758: 4.76059104118, 2759: -99.9989366463, 2760: 4.3999523635, 2761: 4.32642205349, 2762: -19.1975090632, 2763: 2.9227096557, 2764: 5.31288616969, 2765: -30.0093379985, 2766: 4.33469565366, 2767: -38.7704924259, 2768: -13.9660249168, 2769: -34.8621182366, 2770: -7.9538381018, 2771: 4.68783720175, 2772: -52.2470664808, 2773: 2.42274591636, 2774: 23.3093098516, 2775: -50.2078815537, 2776: -2.56319245455, 2777: -38.0545420032, 2778: 1.52998209308, 2779: 20.2769291304, 2780: -13.0987204005, 2781: -37.1064686353, 2782: 4.11677448313, 2783: 8.06280129881, 2784: -10.5679106829, 2785: 25.8893311144, 2786: 1.58584673981, 2787: 29.2250109096, 2788: -14.836874293, 2789: -74.963935131, 2790: 49.2730074294, 2791: 41.0938480133, 2792: -12.1092623833, 2793: 0.23897504755, 2794: 3.02592212564, 2795: 0.625009415352, 2796: -2.75867047628, 2797: -1.41757501604, 2798: 3.8360471483, 2799: -5.79784055638, 2800: -1.42861881914, 2801: 0.171220588426, 2802: -3.90765212685, 2803: 3.19345855116, 2804: -7.55694473916, 2805: 2.91400001703, 2806: -3.015122684, 2807: 0.752918394119, 2808: -2.01949168007, 2809: -0.755701836567, 2810: -4.63470345276, 2811: -1.79475777781, 2812: 2.57133166998, 2813: -0.178022560658, 2814: -7.06645988583, 2815: -0.389361589286, 2816: -3.89929693521, 2817: 2.47428014497, 2818: 0.554225351956, 2819: -1.58957575673, 2820: 6.59516313633, 2821: 1.99263274609, 2822: 1.84099124104, 2823: 6.7321222817, 2824: 3.93481296231, 2825: 0.267122088216, 2826: 4.1579759789, 2827: -0.261912753511, 2828: 0.107047150391, 2829: 3.44587680708, 2830: -2.67238514289, 2831: -1.69993032868, 2832: 3.80870673543, 2833: -6.4622427721, 2834: -7.8543217943, 2835: 7.9490461716, 2836: -1.00062357401, 2837: -3.46629918407, 2838: 0.590955143449, 2839: -4.22359665013, 2840: -0.426414966235, 2841: 1.76655202398, 2842: -2.18828433097, 2843: 2.61511430541, 2844: 1.30885031396, 2845: 1.26338975523, 2846: -4.53364808129, 2847: 0.506401889917, 2848: 0.360082717235, 2849: 0.888800177708, 2850: -0.537949093053, 2851: -3.25674998154, 2852: 3.95716774374, 2853: 11.0284626836, 2854: -1.65312359302, 2855: 0.636680242453, 2856: -1.62719150798, 2857: -5.0132474435, 2858: 1.19433756438, 2859: -4.40594008217, 2860: 0.566211137708, 2861: -2.51970932033, 2862: 0.876091104307, 2863: -5.67529605329, 2864: -2.96927264188, 2865: 1.29946201637, 2866: 0.0514973604169, 2867: 1.67435042554, 2868: 0.401125896618, 2869: 7.25578084953, 2870: -2.27006325347, 2871: 4.14309174935, 2872: 4.96107137304, 2873: -0.784976413963, 2874: -8.84186914907, 2875: 0.173351043829, 2876: -0.391832205772, 2877: 4.02096216711, 2878: 7.74465586058, 2879: -0.988875804189, 2880: 2.54035007348, 2881: -1.79350475062, 2882: -1.92309647053, 2883: -0.125610578919, 2884: -0.0236072424779, 2885: -0.9385061082, 2886: 0.371148539565, 2887: 0.824851522063, 2888: 1.08484646743, 2889: 0.289508061053, 2890: 0.847780151095, 2891: 0.689706344631, 2892: -0.846164617987, 2893: 2.93407801537, 2894: -3.98789233958, 2895: -4.75865183804, 2896: 3.73182440551, 2897: -1.91535969866, 2898: -4.46750264637, 2899: 0.130658064379, 2900: 1.48829559703, 2901: -1.79135028856, 2902: -0.359713416208, 2903: -0.0231979827896, 2904: 1.95064263759, 2905: 11.6018784531, 2906: 2.25662496323, 2907: -0.619730828836, 2908: 3.59572840121, 2909: -1.62182629039, 2910: 1.2649094832, 2911: -3.86317346836, 2912: 4.33221129832, 2913: -2.16506208227, 2914: 6.55373120409, 2915: 4.72603412606, 2916: -2.13499960172, 2917: -0.93864166352, 2918: 0.263663964192, 2919: -0.0128836186257, 2920: 2.12427136162, 2921: 0.875118282552, 2922: 0.100577067935, 2923: 1.01524374488, 2924: -0.390577992816, 2925: -3.99253185651, 2926: -0.724785120808, 2927: 0.618189266671, 2928: 2.28582686111, 2929: 1.21316319214, 2930: 4.44249021658, 2931: -0.84905190191, 2932: -3.59587827469, 2933: -3.57517888778, 2934: 5.20003600725, 2935: 2.89378780883, 2936: -0.402586780208, 2937: 1.8833239469, 2938: 0.293216650304, 2939: 0.195689820821, 2940: -2.31630543949, 2941: 0.163001890355, 2942: 0.731792848796, 2943: 1.97199675245, 2944: 0.330585166297, 2945: 0.278061188448, 2946: 0.122683432454, 2947: 0.105906455668, 2948: -0.359148524775, 2949: -3.06623461981, 2950: 1.70078453102, 2951: 0.543622519096, 2952: 0.488515699808, 2953: -2.10230649076, 2954: -0.19914051611, 2955: 17.0682793027, 2956: -1.42932929272, 2957: -16.7301884501, 2958: 3.6068557654, 2959: -2.9154054255, 2960: 2.23243762892, 2961: -5.68273924232, 2962: -2.6964511478, 2963: 4.20985561397, 2964: 0.179388320714, 2965: 6.40035662564, 2966: 0.25712839694, 2967: 10.6407156669, 2968: -7.67042297968, 2969: -1.33586701573, 2970: 0.984280041415, 2971: 0.218369561273, 2972: 0.263076349927, 2973: 0.304502136986, 2974: 0.283724564967, 2975: -0.746534769311, 2976: 16.8694714837, 2977: 2.36022519526, 2978: -2.02778249273, 2979: 13.9718032235, 2980: -5.02941277691, 2981: 1.26126699264, 2982: -1.36799612697, 2983: 3.69458551964, 2984: -5.1629514525, 2985: 0.00832495764037, 2986: -0.304434647962, 2987: -0.382430266061, 2988: -3.10869743149, 2989: 2.62790612664, 2990: -5.89018207093, 2991: 0.203529180764, 2992: 1.01734235692, 2993: 0.250365335404, 2994: -12.8748778883, 2995: 0.114723179389, 2996: 0.118475352498, 2997: -16.6655085784, 2998: 0.958646811725, 2999: -12.5619205391, 3000: 0.199592133066, 3001: 0.0603340707115, 3002: -0.584544839716, 3003: 0.253603207884, 3004: -1.75771514602, 3005: -3.58518620867, 3006: 4.9066400269, 3007: 0.0228181694484, 3008: -0.734602395441, 3009: 0.848761807396, 3010: -1.60271676111, 3011: 8.65781106916, 3012: 2.48646985653, 3013: -0.838336618867, 3014: 2.3421237315, 3015: 0.385164007158, 3016: 0.24390248013, 3017: -2.2929116271, 3018: 0.240318592327, 3019: -2.55377247181, 3020: 1.97260110058, 3021: -1.07531571543, 3022: 0.882582432413, 3023: 0.0757633663861, 3024: 0.271664008387, 3025: 4.81068687304, 3026: -2.20317701246, 3027: 0.212282381004, 3028: -9.0532362926, 3029: -4.86350598576, 3030: -0.659967591769, 3031: -1.57498531214, 3032: 0.888601058846, 3033: 1.54074148548, 3034: -1.85709936927, 3035: -3.37629502739, 3036: 0.882981435891, 3037: 1.08347907887, 3038: -3.03233635616, 3039: -6.00553184999, 3040: 2.61038611794, 3041: -2.01659168033, 3042: 2.47335438412, 3043: -2.39839971687, 3044: -2.05332011541, 3045: -3.51219381065, 3046: -1.79739645596, 3047: 0.250814293755, 3048: -1.43972989938, 3049: -2.12785584536, 3050: 4.38018828943, 3051: 0.794616347924, 3052: 4.12051465775, 3053: -3.92475951363, 3054: -1.54328434419, 3055: -1.43388610923, 3056: -2.72171908911, 3057: -0.541414763878, 3058: 7.75437446661, 3059: -7.66435748425, 3060: 0.734281591691, 3061: -2.71508680932, 3062: -1.45258831991, 3063: -0.00934795888643, 3064: -6.84662531023, 3065: 3.99164199939, 3066: 9.65726720539, 3067: -0.939884908751, 3068: -6.39161030412, 3069: 2.3604332892, 3070: 5.43949818871, 3071: -2.0468807767, 3072: 3.17866700783, 3073: -0.6346671869, 3074: -7.4641700599, 3075: 0.536746884787, 3076: 8.89508775594, 3077: -12.6953513883, 3078: 8.68132512627, 3079: 0.325481830252, 3080: 2.65387479746, 3081: -1.26235708358, 3082: 1.91520864972, 3083: -3.34687114901, 3084: -13.7167258365, 3085: -1.05294178319, 3086: -4.03052434168, 3087: -17.9964495237, 3088: -0.476172371188, 3089: 0.760562164988, 3090: -1.36576652014, 3091: -0.704237419979, 3092: 7.01573964554, 3093: 2.25130577904, 3094: -0.299880253673, 3095: -2.88665008791, 3096: -4.69904592864, 3097: 3.38465584691, 3098: 0.98003352334, 3099: -0.231862637124, 3100: 0.904484655208, 3101: 0.843738810048, 3102: 0.759188977425, 3103: 10.1284414323, 3104: 0.665429825695, 3105: -0.073388307097, 3106: 10.9408634207, 3107: 0.83797850166, 3108: 5.41546325138, 3109: 0.0981667226987, 3110: 0.911690612028, 3111: 8.67193673367, 3112: -3.16452267627, 3113: -3.08565984895, 3114: 2.6224198925, 3115: -3.00756410378, 3116: -4.13837830767, 3117: 0.868217889184, 3118: 2.04824756887, 3119: -1.70043936142, 3120: -1.57301822618, 3121: -0.677983290916, 3122: 6.19414492648, 3123: 0.761760671779, 3124: 3.54148279242, 3125: -2.30188958789, 3126: 1.74629465283, 3127: 0.340049945457, 3128: 0.885243645773, 3129: -0.463486493199, 3130: -5.87148624535, 3131: 1.07693335337, 3132: 6.52545562202, 3133: -2.16010844635, 3134: -0.925884391791, 3135: 2.34466986493, 3136: 1.33999668908, 3137: 4.23729485101, 3138: -15.5390076814, 3139: -3.70917353362, 3140: 2.97546593221, 3141: 2.48598162157, 3142: 1.0944472068, 3143: -11.7256598877, 3144: -33.7428467005, 3145: 38.9030443995, 3146: -10.875908965, 3147: 0.18852706311, 3148: 30.1413500223, 3149: 20.2931147935, 3150: -33.555107347, 3151: -22.5107063256, 3152: 37.7258480859, 3153: -30.5031190186, 3154: 7.82534584964, 3155: -24.6584551289, 3156: 24.5339787734, 3157: -3.25909092948, 3158: 14.0111869083, 3159: 10.7240253637, 3160: 1.48010960495, 3161: 3.71768827887, 3162: -41.4392040547, 3163: -19.5025171433, 3164: -30.0153435148, 3165: 8.18009597719, 3166: -1.57653431169, 3167: 3.50824843489, 3168: 54.6094528599, 3169: 19.1426247692, 3170: -58.7994710031, 3171: -5.29302275008, 3172: -51.9740259646, 3173: -99.7787965814, 3174: -17.3689323372, 3175: -26.1495312215, 3176: -15.8264241965, 3177: -29.6022737637, 3178: -62.0084842383, 3179: -16.1844562494, 3180: -21.2496029235, 3181: -30.2034702158, 3182: -17.4024691305, 3183: 2.87013172868, 3184: -21.2431079155, 3185: -37.5947411132, 3186: -0.58812857135, 3187: -24.9503188694, 3188: -2.44224975278, 3189: 3.81737892967, 3190: -43.6038148724, 3191: 0.384061062871, 3192: -3.34252489527, 3193: -21.4384284592, 3194: 1.85522947427, 3195: -14.2124743562, 3196: -23.3728306249, 3197: 2.44692404686, 3198: -7.93863475042, 3199: 0.733961580011, 3200: -13.3367949495, 3201: 31.2963198876, 3202: -6.94609786763, 3203: -8.72103582901, 3204: -10.7101976139, 3205: -18.493055998, 3206: 23.8370272643, 3207: 2.64829151183, 3208: -1.89974394838, 3209: 3.18970071649, 3210: 24.2502810566, 3211: 3.75135906191, 3212: 3.3342280232, 3213: -5.388718909, 3214: 2.93300641921, 3215: 20.5486305103, 3216: -8.68002306721, 3217: 1.38335589973, 3218: 2.3634949722, 3219: 6.20640727016, 3220: -8.79835775656, 3221: -25.1008832855, 3222: 2.250387348, 3223: -1.1233107285, 3224: -0.761664849738, 3225: 3.17694806066, 3226: -30.9565284428, 3227: -5.03811859319, 3228: 2.63291021034, 3229: 21.7743455474, 3230: 3.66788181396, 3231: 0.72158636817, 3232: 2.73142501535, 3233: 2.65855783647, 3234: -24.6147970946, 3235: 3.05255987385, 3236: 0.380709188914, 3237: 3.18635439696, 3238: -4.9988244466, 3239: 3.30674491015, 3240: 8.6810786415, 3241: 4.8668355165, 3242: -5.54186312696, 3243: 1.95785935399, 3244: -2.90074250437, 3245: 2.54188103999, 3246: -17.5886557822, 3247: -65.0355894882, 3248: 2.99171996361, 3249: 2.66210131089, 3250: 3.76300479348, 3251: -24.7854897909, 3252: 2.71164022719, 3253: 18.6913493737, 3254: -65.4946502035, 3255: 8.88914162267, 3256: -18.962112384, 3257: 17.1833013289, 3258: 39.6276916317, 3259: -7.91764036502, 3260: -30.762511103, 3261: -73.7311807729, 3262: 10.8155522649, 3263: 16.7774462541, 3264: 5.4833319255, 3265: 43.7673067655, 3266: 5.17688220837, 3267: 1.0180482296, 3268: 2.11288096547, 3269: 2.82241605233, 3270: -5.33391274463, 3271: 1.25686630129, 3272: 2.71906290002, 3273: 0.719313484178, 3274: -33.5210848778, 3275: -1.1397441933, 3276: 2.85598447027, 3277: 4.17983487915, 3278: 2.36390241218, 3279: -4.39268233381, 3280: 2.86773238203, 3281: 4.35751079809, 3282: -52.3235789352, 3283: 7.26280940234, 3284: 3.40092700213, 3285: 3.49475839528, 3286: 8.64376969702, 3287: 0.993483108462, 3288: 1.28616542247, 3289: -6.57673290216, 3290: 1.1713065615, 3291: 1.7329226765, 3292: -6.44987793295, 3293: 0.761829378852, 3294: 1.11491421244, 3295: 1.28579179322, 3296: 1.19434128565, 3297: 0.934672424903, 3298: -4.730830094, 3299: -0.372593248092, 3300: 0.993697748752, 3301: -0.40705667062, 3302: 2.7787680564, 3303: 2.06215806219, 3304: 20.0313508169, 3305: -0.16646412049, 3306: 35.5569509973, 3307: -1.78484286941, 3308: 14.5670983268, 3309: -0.322955578562, 3310: 23.7870096461, 3311: 2.96249852524, 3312: 7.93859209241, 3313: 1.24494379751, 3314: -1.40579292542, 3315: 0.970950095765, 3316: -0.399103917152, 3317: 42.5123663478, 3318: 17.3912492352, 3319: -0.760745420792, 3320: 0.973137199378, 3321: 1.20155395283, 3322: 0.844508461363, 3323: 1.72217130403, 3324: -0.0106912243989, 3325: 22.0740434643, 3326: 3.82283647081, 3327: 19.7217672507, 3328: 37.5167712571, 3329: 13.1752599865, 3330: 12.4094635115, 3331: -3.56067349007, 3332: 13.383855432, 3333: 21.3412575976, 3334: 1.17915407251, 3335: -41.5848701261, 3336: 5.05207067935, 3337: 1.12670764348, 3338: 12.8497930777, 3339: -21.9382332079, 3340: 1.01008182909, 3341: 3.04589228255, 3342: 1.07478272507, 3343: 53.3306062738, 3344: 1.19290763639, 3345: 1.23553554061, 3346: -13.4392420404, 3347: 2.98668224394, 3348: 3.33493667435, 3349: 1.13945194563, 3350: 1.27336765616, 3351: 1.06765239548, 3352: 1.1901575424, 3353: 0.894300317889, 3354: -9.30788219017, 3355: 42.9722392717, 3356: 4.62269526164, 3357: 0.100185971923, 3358: -16.7528772007, 3359: 11.9156179731, 3360: 9.13926064263, 3361: -6.5029574663, 3362: 8.89293537141, 3363: 5.05019434151, 3364: 3.34312142692, 3365: 2.07028990343, 3366: 4.99759399377, 3367: 1.14626194056, 3368: -34.5344664807, 3369: -3.63717002573, 3370: -6.3125385768, 3371: -7.58249967688, 3372: -51.379264973, 3373: 1.06451442694, 3374: -22.9055884506, 3375: -15.8105145464, 3376: 2.68054096704, 3377: -16.7846630174, 3378: 6.22557692459, 3379: -5.76559758216, 3380: 1.628390159, 3381: 2.62558936182, 3382: 5.5836541218, 3383: 23.1120007391, 3384: 43.8865174085, 3385: -3.72698730969, 3386: 37.245850161, 3387: 47.1384454293, 3388: 11.4917597163, 3389: -31.0295536912, 3390: -100.0, 3391: 65.6194394179, 3392: 29.4175172558, 3393: 15.9917672644, 3394: -60.0826501726, 3395: 9.92613495601, 3396: 4.33847603471, 3397: 1.92004577454, 3398: 34.6459625567, 3399: 2.7149920028, 3400: 1.86594316196, 3401: -6.83059694794, 3402: 36.1087473516, 3403: 3.40302997687, 3404: -0.879524235298, 3405: 7.44338324358, 3406: 0.877711673696, 3407: -1.054954567, 3408: -27.4316160832, 3409: 11.0078436618, 3410: 18.8662141361, 3411: 2.15110165026, 3412: 2.54081711486, 3413: -41.0104556954, 3414: -27.4421854183, 3415: 44.006610508, 3416: 72.1174443796, 3417: -35.6532266746, 3418: 2.96309203627, 3419: 15.7144186027, 3420: 2.78206883677, 3421: -1.66923677843, 3422: -0.362698902307, 3423: 0.0203913409847, 3424: 1.43216397434, 3425: 41.5419396087, 3426: -1.30229912151, 3427: -5.72512290663, 3428: 4.36207426546, 3429: 7.28766235091, 3430: 3.50713165205, 3431: -0.622224046427, 3432: -4.6933790073, 3433: 1.97749943291, 3434: -26.5733793483, 3435: 8.88121540105, 3436: -17.1157718949, 3437: 10.1435382906, 3438: -15.4823081014, 3439: -1.27732069995, 3440: 11.9862296178, 3441: -6.70767738152, 3442: 2.90189613299, 3443: 55.7386063797, 3444: -1.37519469061, 3445: -30.4299273783, 3446: -12.4470657754, 3447: 0.312198167859, 3448: 20.6908814243, 3449: 2.90104570328, 3450: 3.12673367191, 3451: 3.55071379175, 3452: -51.7581361207, 3453: 2.82939234842, 3454: 2.42390148178, 3455: 43.9186135824, 3456: 3.07769591871, 3457: -64.5110872993, 3458: 2.64524268911, 3459: 2.95084185272, 3460: 53.4479461017, 3461: 3.37294364577, 3462: -0.793393836366, 3463: 25.2645703603, 3464: -33.2958712262, 3465: -26.1281958627, 3466: 2.03960811777, 3467: 0.413630688415, 3468: -0.332130955359, 3469: 37.8170194105, 3470: -13.7939394805, 3471: -3.13552267369, 3472: -24.7355607658, 3473: 9.05155785467, 3474: -0.819983739248, 3475: -17.5957242983, 3476: -3.58951916437, 3477: -11.2582922401, 3478: -8.28072401634, 3479: 17.4188111955, 3480: 2.85954629712, 3481: -0.349941504894, 3482: 11.6025823065, 3483: -0.489865786574, 3484: -9.47624887838, 3485: 81.5232078764, 3486: -4.22237349438, 3487: -100.0, 3488: -94.9275746602, 3489: -59.2669504096, 3490: -7.26824355585, 3491: -0.65665963301, 3492: 0.964685896253, 3493: 1.06738422855, 3494: -2.29215509629, 3495: 0.60774154583, 3496: -1.34472906153, 3497: -1.83292787265, 3498: 0.417601502232, 3499: -0.543395591563, 3500: 1.24932450438, 3501: -1.34429989177}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: -0.0434663318255, 2: -95.8362272437, 3: 40.3257644838, 4: -16.521359586, 5: 17.1351743266, 6: 10.7617161019, 7: 52.5994922822, 8: -9.42927980539, 9: 46.4672679519, 10: -18.7581190623, 11: -12.4839020809, 12: -22.6835090379, 13: -9.22901743614, 14: -6.66344831793, 15: -6.41903071256, 16: -7.078628666, 17: 4.14326319679, 18: -3.15707987258, 19: -52.4877357734, 20: 29.2029094661, 21: -24.8064737443, 22: -13.123119215, 23: -17.3987911615, 24: 29.1704033003, 25: 9.58877872221, 26: 27.5668304967, 27: -17.7237413662, 28: 21.750272386, 29: 31.0524359988, 30: -26.4325900646, 31: -57.3298283442, 32: -8.3405696689, 33: -14.5775618151, 34: -14.2779085876, 35: -35.4459498755, 36: -25.8275440499, 37: -20.1432681022, 38: -9.08722928406, 39: -50.5665461288, 40: 9.40347136379, 41: 17.9018882012, 42: -1.27585841632, 43: -20.8476282277, 44: -2.58595157214, 45: -1.79695549273, 46: 6.95529053642, 47: -7.30679660332, 48: -4.25789617419, 49: -14.3356134916, 50: 1.7186552285, 51: 1.01571364249, 52: -24.9317439527, 53: -1.10341739159, 54: 0.522429137591, 55: -7.84499039656, 56: -1.25514024847, 57: 5.20266636662, 58: -0.00811688219344, 59: -12.4115182118, 60: 47.2987043037, 61: -44.23540677, 62: 18.9036994161, 63: -48.5051206449, 64: -12.3622382971, 65: 37.3393036738, 66: 10.3437315695, 67: 9.37330089013, 68: -0.433258568959, 69: -2.3371939351, 70: 0.434132754739, 71: 0.239805535965, 72: 16.5428388467, 73: -0.909354523265, 74: -2.50648611274, 75: -10.637341048, 76: -15.4473884027, 77: -13.5824863871, 78: 0.17805013309, 79: -55.664692092, 80: -31.146221419, 81: 2.21888955354, 82: 29.5665911349, 83: -0.664658179659, 84: -14.5995808611, 85: 0.382124185535, 86: 1.67015767957, 87: 4.28313014201, 88: -23.3834489736, 89: -1.10949419992, 90: 6.51125616212, 91: 4.38173435254, 92: -1.22597722056, 93: 29.5989926088, 94: 1.08314778013, 95: -0.421047123942, 96: -0.491006308325, 97: 24.9326995947, 98: -0.0253815231015, 99: -10.5250024311, 100: 17.6246546493, 101: 0.313443747771, 102: -1.49893387, 103: -5.09126028825, 104: -0.68862555367, 105: 18.1024194195, 106: 18.8628658126, 107: -0.838362487592, 108: -1.029460633, 109: -0.860030567558, 110: 16.7336905931, 111: 7.50401041662, 112: -1.87036943413, 113: -78.5963277349, 114: -1.83967134054, 115: 34.6604587415, 116: -19.0647082017, 117: 7.22089880541, 118: -2.21868826303, 119: 4.12238961228, 120: 0.831746320495, 121: 14.0588952219, 122: -20.5065932422, 123: -8.54315856825, 124: 4.57999594397, 125: -15.649439222, 126: -0.254911901646, 127: -0.249719010523, 128: 10.5692446917, 129: -0.319167128893, 130: -0.132418723594, 131: -0.36173209544, 132: 1.97238375451, 133: -21.7634856427, 134: -4.33081628421, 135: -2.34577217527, 136: -14.3938440351, 137: 0.531257221174, 138: -7.72750139931, 139: -0.573375867472, 140: 1.18446813689, 141: -29.899180108, 142: -25.9520968553, 143: -29.2176889961, 144: 0.384179717038, 145: 4.26110594959, 146: -0.080084670958, 147: -0.200385235355, 148: 21.6301471796, 149: -0.0893593626084, 150: -0.824803242625, 151: -3.04087889742, 152: -0.0119808749425, 153: -0.25999826015, 154: -0.0532654188346, 155: -0.100917905215, 156: -1.01888148398, 157: 0.646473174873, 158: 0.357535806277, 159: 0.163463588022, 160: -0.21970109876, 161: -0.197054533954, 162: -0.908200910752, 163: -6.74950091962, 164: -4.19427580522, 165: -24.655009066, 166: -5.54533039615, 167: 13.0662208575, 168: 2.98872613959, 169: 18.1390263971, 170: 0.552400510954, 171: -11.7085029102, 172: -0.0543833487274, 173: -31.503602705, 174: -0.270010630208, 175: 3.50049341297, 176: 24.7199771456, 177: -27.6304513241, 178: 0.361520711443, 179: -0.0782504509184, 180: -0.0909340765318, 181: -2.90137716122, 182: 0.572809350552, 183: -9.65771709904, 184: 4.25838602699, 185: 3.52528556347, 186: 34.5675326476, 187: -15.8112661676, 188: 12.4097890258, 189: 21.119574309, 190: -10.8719462372, 191: 2.09260789035, 192: 22.8789663103, 193: -0.0633664349955, 194: -18.9572801931, 195: -10.4664289927, 196: -1.78569530432, 197: -4.27211578779, 198: -29.5882851383, 199: -0.0936304840862, 200: -0.811221208218, 201: -0.231546409737, 202: -46.3403825907, 203: -0.148005096818, 204: -0.213367894421, 205: -6.27125477394, 206: -0.824117540382, 207: -6.36878883616, 208: -0.198167487336, 209: -0.0575630613657, 210: -1.43497599418, 211: -0.201443042099, 212: 0.304754971322, 213: 15.0680632621, 214: 25.0727360948, 215: -0.192956843758, 216: 1.91920636185, 217: -0.591504749189, 218: 5.48500641618, 219: -18.7413815731, 220: -16.4484511651, 221: 7.99024341118, 222: 8.38670641415, 223: -8.9987876284, 224: -0.909237174365, 225: 5.3114728474, 226: -0.181689930484, 227: -2.08829624489, 228: -3.1218884023, 229: 2.63349761912, 230: -3.57786616623, 231: -39.2452844036, 232: -0.223983190147, 233: -13.0165418693, 234: -7.05390178464, 235: 2.93586472637, 236: -31.336812688, 237: 30.8314245024, 238: 1.44773735446, 239: -1.06313252006, 240: -0.389094155461, 241: -16.555911627, 242: 27.0607084525, 243: 30.5306326815, 244: 21.4033142738, 245: 38.606536382, 246: 2.39859038607, 247: 15.4277321611, 248: 23.5915802477, 249: -31.2192291952, 250: 7.75132622283, 251: 36.7074572349, 252: 41.1667763339, 253: -5.3966682655, 254: 1.36648903835, 255: -7.81976089461, 256: -0.0345588355239, 257: 18.1907676805, 258: -0.987138540945, 259: -0.834877052865, 260: -3.36305148771, 261: 4.36372018521, 262: -0.72635290034, 263: -9.33643780898, 264: -2.33272705785, 265: -1.23333518556, 266: -3.0638167479, 267: -22.9101427208, 268: -10.312863562, 269: -3.85086118327, 270: -0.0329142735183, 271: -0.417570662132, 272: -5.92655192787, 273: -19.9973725863, 274: -13.7559341794, 275: -16.6099254131, 276: -47.3617244318, 277: -5.288011493, 278: -8.07765880681, 279: -6.42097001085, 280: -29.8849566967, 281: -0.825484694287, 282: 37.1969929653, 283: 4.93246459474, 284: -21.9200197488, 285: -6.46804466507, 286: 9.32090211243, 287: -0.270597811634, 288: -21.5020069544, 289: 28.2012336918, 290: -18.8503569848, 291: -4.5743331562, 292: 5.82704509273, 293: -2.65410995585, 294: -12.3853517293, 295: -40.1252970568, 296: 43.1031990397, 297: 27.5733290818, 298: 17.3826218711, 299: 3.02049954849, 300: -19.6223282487, 301: -1.33519905128, 302: 6.26571561699, 303: -9.42092413584, 304: 14.1211239124, 305: 17.5546099539, 306: -0.554485288782, 307: 47.8184815454, 308: -0.806654251963, 309: -0.948668374161, 310: 2.52595449858, 311: 21.3366297416, 312: -0.566268053139, 313: -1.58153292812, 314: -12.0024895308, 315: -0.812003196775, 316: -42.955957214, 317: -0.874120634416, 318: -0.934621107421, 319: -13.1874364794, 320: -1.81603592259, 321: -0.423840867219, 322: -10.2263580137, 323: -11.42738089, 324: 51.2341172777, 325: -36.3491607531, 326: 44.9048921041, 327: 13.0117916921, 328: 52.2770975581, 329: 2.53218887636, 330: 12.0979210749, 331: 23.3548639199, 332: -24.2372419875, 333: -1.63735317857, 334: -12.3817549297, 335: -1.92707555624, 336: 81.7729107954, 337: 25.5580077679, 338: 12.2967006929, 339: -0.749513749856, 340: 17.1059977281, 341: 13.3119947282, 342: 8.77089147215, 343: 12.4959428163, 344: 4.07475983628, 345: -34.1371490341, 346: 79.187221316, 347: -8.36047897985, 348: -18.8729639995, 349: 24.7364355467, 350: -0.628892257681, 351: 11.4004940586, 352: 6.83520641146, 353: 17.6889517705, 354: -6.66048583726, 355: -9.67785274854, 356: 10.8164682065, 357: -45.2460524629, 358: -8.3553471824, 359: -26.1159485318, 360: 11.7094097434, 361: 35.4895245488, 362: -6.31876383905, 363: 70.4482527189, 364: -17.6084399329, 365: -16.1163733206, 366: 39.0179678798, 367: -6.03942298886, 368: 28.3351374235, 369: 2.48388044686, 370: 11.478237827, 371: -14.1195516278, 372: -14.563698847, 373: -31.8441079023, 374: -17.7277466104, 375: 28.2333101008, 376: -8.04386365671, 377: 6.32013709178, 378: -12.8177023465, 379: 1.96387392231, 380: -18.8302634245, 381: 76.6798282, 382: -0.800844046496, 383: 3.39597171932, 384: -11.739648409, 385: 12.4400264112, 386: 17.9673500098, 387: 20.2675381801, 388: 6.16100994212, 389: 11.3663658507, 390: -7.15252650116, 391: -3.65510853369, 392: 2.65953571597, 393: -26.2577392737, 394: -23.9616253505, 395: 0.300059692932, 396: -11.2922174464, 397: -4.2857190679, 398: -4.91870606963, 399: 6.49511502122, 400: 7.34810238007, 401: 36.5878731442, 402: -2.92013726307, 403: 10.0234304809, 404: 15.1246380374, 405: -2.4766574311, 406: -3.85667895974, 407: -6.18002369946, 408: -15.1648820032, 409: 49.0800440038, 410: -13.7675992124, 411: -0.458932290059, 412: -3.79252956018, 413: 21.9469645106, 414: 19.8625242535, 415: -1.83545125945, 416: 2.27807937619, 417: -3.97876632503, 418: -1.15011858165, 419: -4.88962987333, 420: -4.64499185356, 421: 30.2665789042, 422: -3.4757203463, 423: 11.9450025638, 424: 2.01849316077, 425: -2.75578573296, 426: -1.91758215406, 427: -4.81005160488, 428: -6.91567275783, 429: 34.453718103, 430: -4.17773753049, 431: 7.71638170039, 432: -3.84113277001, 433: -3.55303430524, 434: -11.6283328483, 435: 0.860044555278, 436: -2.52693094197, 437: -24.0025416764, 438: -5.64414948649, 439: -2.08165170739, 440: -3.6320413876, 441: -3.28092407221, 442: -2.44984860144, 443: -3.84745489972, 444: -4.2206425559, 445: -5.04815442331, 446: 5.22887321375, 447: -4.44979518881, 448: -0.974573744041, 449: -4.04441078984, 450: 27.8318077223, 451: -2.64960506239, 452: 2.40075626877, 453: -2.05902915791, 454: 2.59796024788, 455: 42.2805431869, 456: -3.73769493552, 457: -3.15893111963, 458: -5.70536123404, 459: 7.43538235571, 460: -3.04811629019, 461: -21.5289983913, 462: 23.3868286846, 463: 2.28959166206, 464: -19.0336862361, 465: -27.7367123258, 466: -18.2425015568, 467: 6.56723785073, 468: 36.1281829375, 469: 44.3168278277, 470: -2.90082884469, 471: -29.2209124918, 472: -8.28596413062, 473: 7.08465198692, 474: 9.12779107565, 475: -0.659203106407, 476: -1.079182146, 477: 24.9909574982, 478: 14.4731791479, 479: -0.781261314514, 480: -3.99625007204, 481: -0.810909594994, 482: -14.4037019003, 483: 0.0871750350711, 484: -7.8629648872, 485: -14.737081704, 486: 4.66366635477, 487: 19.5013441945, 488: -3.63033448074, 489: -1.11615629946, 490: 31.3695394622, 491: -13.8902071543, 492: -20.1421754174, 493: -2.45867411887, 494: 20.259716792, 495: -0.706245400826, 496: -0.791185917948, 497: -7.46022118473, 498: -0.613250377601, 499: -2.92256776329, 500: 21.6077908318, 501: -0.77431390856, 502: -0.716537251379, 503: -0.941940892168, 504: -0.770884082291, 505: -8.59504131192, 506: -37.1023607317, 507: -3.30391648906, 508: -0.827186214997, 509: -0.0640029507478, 510: -1.71024281192, 511: -2.74309141763, 512: 17.5240325367, 513: -10.9547097493, 514: -6.20403244913, 515: 100.0, 516: -14.5867082565, 517: 1.78483172571, 518: -8.54030935289, 519: -6.65978624629, 520: -8.85160237739, 521: -0.797162832816, 522: -43.4931021732, 523: -0.535663589532, 524: 6.75061074736, 525: -4.03340873627, 526: -37.8837307926, 527: 2.21898646555, 528: -0.681342821827, 529: -0.642548200466, 530: -0.521189661868, 531: -0.492244703214, 532: -2.36855865127, 533: 13.5740301033, 534: 10.7367366745, 535: -5.15317760843, 536: -14.1898941258, 537: 3.29083483063, 538: 0.699297182496, 539: 3.69887933838, 540: 0.976845246812, 541: -4.00240141466, 542: -0.634382416384, 543: 46.631764371, 544: -2.82733248462, 545: 10.058336432, 546: 2.53273858076, 547: 8.71623021274, 548: -0.725874509803, 549: -3.70056201403, 550: -0.518649559476, 551: 3.21551351109, 552: -0.775149957838, 553: -0.699686913856, 554: 2.44013055031, 555: -3.5780411874, 556: -20.6580596252, 557: -0.760716341934, 558: -0.7267637827, 559: -0.773495401736, 560: -0.627993524226, 561: -0.411036546959, 562: 3.77663251083, 563: -13.8763588134, 564: -2.3547940291, 565: -0.311967111366, 566: 4.00957024351, 567: -4.85336819001, 568: -3.99000814008, 569: -48.7550807734, 570: -13.0625145069, 571: -11.8854600123, 572: -6.1220194751, 573: 5.21118254352, 574: -6.55663040238, 575: -0.802506031844, 576: 15.7449248004, 577: -0.330374206036, 578: -9.54767656992, 579: 5.92266511387, 580: -7.05635315146, 581: -0.80474768924, 582: -5.38453201678, 583: -18.3669146612, 584: -3.36269631452, 585: 6.0629497477, 586: -5.46978136452, 587: -45.0278854563, 588: -3.31433711785, 589: -4.01930129342, 590: 4.65296879831, 591: 18.138201616, 592: 6.88258977194, 593: -2.70053267757, 594: 2.56918494747, 595: 18.1344785423, 596: 31.2269643662, 597: 15.0739165956, 598: -44.4048827977, 599: -15.2840014591, 600: -4.42391768471, 601: 13.5445374254, 602: 26.2011484541, 603: 3.78699518465, 604: -4.20095494159, 605: -4.33380606106, 606: 11.568905153, 607: -2.98322428896, 608: -2.8771621198, 609: 15.8989653873, 610: -32.0540092067, 611: -4.08234255002, 612: -0.0183048093722, 613: -3.60591204456, 614: -8.74940546254, 615: 51.2186993983, 616: 8.7991779927, 617: 10.2123840153, 618: -27.3929942003, 619: -0.58240305467, 620: -3.2254062607, 621: -3.35045024627, 622: -24.5456351433, 623: -35.7556455192, 624: -27.0263340903, 625: 3.072764288, 626: 21.3783067138, 627: -10.1896966449, 628: -8.21434706314, 629: -13.1046736076, 630: -2.91329260383, 631: 15.6329126293, 632: -36.3324566902, 633: 26.962282565, 634: 55.7660357833, 635: -32.8474474785, 636: -4.18298918684, 637: -6.14304931602, 638: -4.08809813911, 639: -10.8491327425, 640: -7.79906936518, 641: 3.51830506241, 642: -1.79885108177, 643: -6.21494050186, 644: -16.9236084511, 645: -12.2617674364, 646: 1.04152361292, 647: -7.53816277775, 648: 11.7855126551, 649: -4.9859472262, 650: -2.32978436006, 651: -30.6810506396, 652: 8.10818018136, 653: -13.993349372, 654: 19.6576219251, 655: -4.24504250354, 656: -11.2355732857, 657: -3.73615878452, 658: -3.70373581585, 659: -4.65927009311, 660: 18.3755183053, 661: -3.52789836829, 662: -3.00730638106, 663: -3.23483208979, 664: -3.6439202621, 665: -4.22650591051, 666: -3.7792939301, 667: -3.54295406212, 668: -5.33752396843, 669: -2.69238801387, 670: 2.47678459965, 671: -25.2450027402, 672: 17.2321607176, 673: 27.6100918819, 674: -9.80066923051, 675: 45.8908943455, 676: 31.5423210753, 677: 51.3718994473, 678: 68.904148856, 679: -20.0414227912, 680: 57.3710522038, 681: 16.8945781973, 682: 33.8082092359, 683: -3.03507159205, 684: 2.87052298638, 685: -69.3422325129, 686: 32.6760649246, 687: -44.0244229942, 688: -3.70559198978, 689: -12.7958324628, 690: -10.6202207341, 691: 2.46575306171, 692: -12.466457153, 693: -34.1463877359, 694: 29.5397751838, 695: 100.0, 696: 38.236507674, 697: 48.4758062132, 698: 1.98203181907, 699: 0.289181107276, 700: 14.1848216094, 701: 17.8079298093, 702: 30.4119936925, 703: -74.2481639036, 704: -19.4873357439, 705: -27.6491414668, 706: 17.0808810275, 707: 7.50680713318, 708: 17.2107220457, 709: 51.5757359222, 710: -23.8208924616, 711: 24.4621253135, 712: 22.9569396003, 713: 4.93621444215, 714: 9.70923452018, 715: -1.28673652211, 716: -70.3583987551, 717: -27.6080415521, 718: -21.3556687192, 719: 37.3072738026, 720: 46.0502240833, 721: 16.1439897589, 722: 22.664831055, 723: 16.6820709177, 724: -21.2454067613, 725: -8.04547612744, 726: -48.8645915788, 727: -27.4530031538, 728: -1.91601404205, 729: -66.264184921, 730: 17.0384751767, 731: 15.5500353692, 732: -6.06602046678, 733: -31.2297775646, 734: 4.80002075206, 735: 14.7603919639, 736: -27.1219566108, 737: -19.3094797218, 738: -9.38987490262, 739: -7.32631764045, 740: 2.97562998443, 741: -38.4242617331, 742: -0.0769452642662, 743: -14.024829341, 744: 0.929472058567, 745: -34.2674047882, 746: 4.4145547796, 747: -2.88277886279, 748: 8.47655175564, 749: 2.87189889149, 750: -15.9918427724, 751: -0.237647169246, 752: -11.7150603116, 753: -2.75805444575, 754: 0.900530149551, 755: -17.2802159567, 756: -5.85456978087, 757: 25.4812819075, 758: -13.0525482667, 759: 25.2048333145, 760: 26.3126260085, 761: 10.0005639606, 762: -12.7918259716, 763: -8.90839117937, 764: 3.61202407351, 765: -3.89956256263, 766: 2.767710676, 767: -26.0065935163, 768: 6.11094472324, 769: 6.23490906613, 770: -3.44405974954, 771: 0.228340654852, 772: 1.80239831626, 773: -3.41934629507, 774: 0.151853986373, 775: -4.93296901578, 776: 2.88077367522, 777: 43.7031396179, 778: -10.8541597072, 779: -3.21082763808, 780: -14.3724552883, 781: 3.58632889742, 782: 1.99709586456, 783: -25.9198833816, 784: 3.18069255272, 785: -8.59067693893, 786: 32.3534803734, 787: 0.0596912509337, 788: -5.35029119414, 789: -19.7271656661, 790: 2.71141648848, 791: 46.8632641213, 792: 2.77482236192, 793: 0.858586700549, 794: 3.53494939565, 795: 6.54270770677, 796: 3.55421608445, 797: 2.27494182963, 798: 7.5832031698, 799: 21.2756728943, 800: 2.62257373089, 801: -0.454554655188, 802: 0.55628826583, 803: 3.95280001656, 804: -68.2750708196, 805: 2.9050562179, 806: 2.18877097085, 807: -1.45113027569, 808: -11.5225261859, 809: 0.321397660378, 810: 21.784854494, 811: 16.0333844564, 812: 5.84660213692, 813: 9.48567479769, 814: -2.80404742098, 815: 20.1095456311, 816: -5.85804510091, 817: -14.2596595058, 818: 28.3464130575, 819: 12.6046158202, 820: -19.6632047746, 821: -3.29678975843, 822: -28.3407043755, 823: -2.01414770906, 824: 0.763806931192, 825: 0.141003423276, 826: -0.00950948892087, 827: -6.92726220861, 828: 0.768418421368, 829: 3.0249516785, 830: 1.90913555985, 831: 91.534895453, 832: 0.9448577834, 833: 24.3303450788, 834: 9.61283720627, 835: 2.02233298643, 836: 21.5540848145, 837: 1.82062147115, 838: 24.0121682353, 839: -7.65750996554, 840: -14.6649431249, 841: -16.2042117258, 842: 1.57626132119, 843: 7.61797319262, 844: 0.323439631103, 845: 0.606871391893, 846: 8.84832140106, 847: 0.536352006289, 848: 2.91331606106, 849: 43.0488455736, 850: 3.27177937964, 851: 0.134862151963, 852: 0.314990844999, 853: 0.337298432314, 854: 2.66276893414, 855: -27.7814772867, 856: 0.066058125863, 857: 1.33998070722, 858: 0.944025313073, 859: -3.7338654191, 860: 3.20329710119, 861: 0.507842420399, 862: -13.9747957286, 863: -46.2097701174, 864: -57.4206039165, 865: 8.81554208328, 866: 1.40329010758, 867: 3.8156583265, 868: -16.5420889623, 869: 12.1015116963, 870: 0.311877615895, 871: 7.50925978175, 872: 0.108186162763, 873: -23.0483760887, 874: -44.4619214872, 875: 25.9081489252, 876: -1.33948789225, 877: 0.384705608085, 878: 0.409367442721, 879: -0.477934175175, 880: 0.556402580382, 881: -3.64416669943, 882: -6.60570110221, 883: -3.95357645842, 884: -0.290341690068, 885: -14.0315356476, 886: -2.96005259306, 887: -8.93143337054, 888: -10.6887951508, 889: -0.0961901720947, 890: 17.0331640765, 891: 0.680245467541, 892: 63.7048117133, 893: 1.4723082453, 894: 6.64279945818, 895: 0.228057837551, 896: 18.9412906208, 897: 0.354341636398, 898: 2.87749817548, 899: 0.592264691292, 900: 24.9453079978, 901: 0.350783159138, 902: 0.398186810322, 903: 33.9392644168, 904: 3.43277788668, 905: 14.2628828729, 906: 0.970417842051, 907: 0.39090397403, 908: 1.22681017728, 909: 0.545873799308, 910: 1.13255942978, 911: -14.5511733686, 912: -2.00007102026, 913: 0.731505668628, 914: 0.781447295426, 915: 8.34314965532, 916: 10.1952517105, 917: -54.9798103924, 918: -33.5660115013, 919: -17.748347197, 920: 7.244880319, 921: 8.54486192263, 922: 0.230126591206, 923: 1.10081677092, 924: 0.6250533974, 925: 60.386016798, 926: 1.23217649051, 927: 1.32466824465, 928: 2.62034117242, 929: -41.5662648416, 930: 0.692929468855, 931: -22.8249249723, 932: -5.43310682191, 933: -12.356467319, 934: -36.5858453307, 935: -23.2349549394, 936: -9.44699077417, 937: -2.29033832083, 938: 2.96299194891, 939: -8.9698798355, 940: -62.3201685984, 941: -54.7190144346, 942: 24.4144662864, 943: -5.19289873425, 944: -28.3854149118, 945: -21.1712604742, 946: -39.2973050988, 947: 100.0, 948: -8.73582183903, 949: 26.6784116981, 950: 23.0946289337, 951: -2.48538392947, 952: 5.97232001909, 953: -3.66571992203, 954: 0.231305663439, 955: 29.2344318521, 956: 2.00129927591, 957: 3.00193794517, 958: 6.34421677576, 959: -60.3408285208, 960: 3.7183664313, 961: 27.7050959033, 962: 9.8210938292, 963: 2.6940214715, 964: 37.193907996, 965: -5.5331392466, 966: -16.6459567131, 967: -4.05765847842, 968: -6.50899451092, 969: 5.27629399638, 970: 30.9620357163, 971: 21.767911084, 972: -56.1470310733, 973: 23.4554133943, 974: -25.7347483279, 975: -5.84163435169, 976: -5.63280325184, 977: -22.3595916854, 978: -10.953282516, 979: 3.60204044616, 980: 4.6488807335, 981: 12.1647772501, 982: 53.5327209489, 983: -19.5005471047, 984: 12.536729127, 985: 4.54331634355, 986: -2.78691786338, 987: 4.75259095203, 988: -2.14118927255, 989: -3.69159573834, 990: -6.33411880108, 991: -13.3364408016, 992: -4.65293778035, 993: 3.1989922737, 994: 13.7787750547, 995: 6.8209893501, 996: -28.1348786483, 997: 3.59439641811, 998: -15.1392516563, 999: -4.88438617125, 1000: -20.41063364, 1001: 1.33400114547, 1002: -51.9828250508, 1003: -56.9091039205, 1004: 0.810241225838, 1005: -16.5055094507, 1006: 2.99501548133, 1007: 2.98663034034, 1008: 3.30957873481, 1009: -6.81324679659, 1010: 2.32805874582, 1011: 2.21363038132, 1012: 3.3698636792, 1013: 3.39220004086, 1014: -18.2595326824, 1015: -17.1684183926, 1016: 2.82562490347, 1017: 22.9366641586, 1018: -2.21687640584, 1019: -0.885530212843, 1020: -9.53936471167, 1021: 23.1885691881, 1022: 24.24078972, 1023: -21.0317210535, 1024: 9.94744099671, 1025: -5.42844577546, 1026: 35.934273132, 1027: 41.6799668739, 1028: 17.750504619, 1029: -25.3409913173, 1030: -34.8274886769, 1031: -4.1470127257, 1032: -34.5196580552, 1033: -9.08789337796, 1034: -16.9078680152, 1035: 4.70645582542, 1036: -4.94399496489, 1037: 2.76329917903, 1038: -43.3563547554, 1039: 13.4026978931, 1040: 3.32988243009, 1041: 0.060975846831, 1042: 15.5622505195, 1043: 20.2748579365, 1044: -48.6704491446, 1045: 16.8042298453, 1046: -30.0375683805, 1047: -10.729897098, 1048: -0.165801922076, 1049: 23.792417405, 1050: -10.0303923456, 1051: -23.8056958967, 1052: -6.34074304666, 1053: -8.6117926087, 1054: -0.205226867406, 1055: -1.40360460345, 1056: -5.68547659594, 1057: -12.5570677298, 1058: 10.5702578902, 1059: -0.867776522976, 1060: -12.3507967903, 1061: -17.4302474074, 1062: 19.8634505014, 1063: 16.6983331493, 1064: -28.6750307902, 1065: -10.0230007915, 1066: 7.57627611978, 1067: -12.8907248507, 1068: 8.26475853131, 1069: 13.5406090566, 1070: 50.148110071, 1071: -12.6611992431, 1072: -19.0066763569, 1073: -18.5800804564, 1074: 13.151543218, 1075: 10.9702334654, 1076: -16.1628545091, 1077: -6.15085119846, 1078: 4.53637668898, 1079: 15.946709542, 1080: -8.00419163191, 1081: 8.31920229582, 1082: 12.526674179, 1083: 1.21853080015, 1084: 13.8449473905, 1085: 8.29783998307, 1086: 8.85636678787, 1087: 1.94090119663, 1088: -18.078218478, 1089: 0.432717119294, 1090: 12.0463538597, 1091: 10.4391953943, 1092: 1.0946481098, 1093: -8.50106586336, 1094: -33.9777454922, 1095: -1.20432656598, 1096: 22.7464094301, 1097: -6.43341289664, 1098: 4.59501837092, 1099: -10.0233347523, 1100: -1.43116129256, 1101: -7.39054531348, 1102: 9.17094293525, 1103: -2.18893879477, 1104: -39.697411533, 1105: -1.04648017437, 1106: 9.09240323412, 1107: -22.9580541642, 1108: 4.54129161883, 1109: 2.1358750947, 1110: 13.6725205573, 1111: 8.29316676531, 1112: -28.2962361942, 1113: -1.98708370521, 1114: -5.89275059195, 1115: -0.70716026622, 1116: 2.99852632976, 1117: -1.23869851615, 1118: -1.29869340266, 1119: 3.30322432207, 1120: -1.46080527024, 1121: 22.763876213, 1122: 2.15049499466, 1123: -1.78137745658, 1124: 6.69158279528, 1125: -1.03713560974, 1126: 10.5920209763, 1127: -20.8316282139, 1128: -2.22785055675, 1129: -13.5139857988, 1130: -1.40379377431, 1131: 0.233968983617, 1132: 1.9158465943, 1133: -3.81570051386, 1134: -4.48734473652, 1135: -13.2734965286, 1136: -1.56042995058, 1137: 4.14457995835, 1138: 0.437802142135, 1139: -1.60283752409, 1140: -22.8109293045, 1141: -1.09103505716, 1142: -1.53549774047, 1143: -0.709588792721, 1144: -2.53302447831, 1145: -1.29207571903, 1146: -1.03747593504, 1147: -1.36278358724, 1148: 12.3541909322, 1149: -0.919759986936, 1150: 2.89437043622, 1151: -1.83585688206, 1152: 1.54035420601, 1153: -6.23428051078, 1154: -1.38356466195, 1155: -1.58375284091, 1156: -1.62255038998, 1157: -3.55928836611, 1158: -1.0220694271, 1159: -11.856724333, 1160: 28.7454286797, 1161: 43.4799954017, 1162: 21.0820770743, 1163: -17.2488346049, 1164: 9.22342640273, 1165: 8.16628168658, 1166: 9.00369139623, 1167: 2.79878979569, 1168: 0.455659837521, 1169: 26.2139316175, 1170: 2.98917039159, 1171: 0.350948164501, 1172: 0.867653773052, 1173: -0.336988010058, 1174: 0.261503832352, 1175: 8.30026666004, 1176: 3.84355293811, 1177: -0.179571813893, 1178: -1.2842160108, 1179: -2.88412847966, 1180: -10.2344886579, 1181: 0.942181022364, 1182: -11.5104696053, 1183: 5.74113037608, 1184: -0.6539659366, 1185: 3.9301224434, 1186: 0.465058609951, 1187: 1.43928469314, 1188: 17.2452192122, 1189: -7.67275444162, 1190: 10.2631474564, 1191: 0.447224379063, 1192: 6.85635675635, 1193: -0.585191799253, 1194: -0.23917138099, 1195: -10.415528114, 1196: -0.31854997157, 1197: -1.60704985309, 1198: 22.7501880356, 1199: -0.889135315092, 1200: -0.177486162386, 1201: -0.226324481134, 1202: -0.310066824345, 1203: -1.32029616436, 1204: 5.48045102153, 1205: 3.2621115177, 1206: 0.340386784238, 1207: 0.0600994125356, 1208: 21.8901309509, 1209: -11.0059401076, 1210: 14.8548845485, 1211: 6.80275973943, 1212: 22.65315263, 1213: 37.4317801354, 1214: 1.0975161375, 1215: -4.68915505761, 1216: -2.71814890717, 1217: -0.408208944083, 1218: 5.17565709235, 1219: -0.485818743351, 1220: 13.1095170546, 1221: -0.325362200566, 1222: 8.41748189532, 1223: 8.16460405743, 1224: 15.2956131785, 1225: 0.4505279844, 1226: -0.242084322147, 1227: -0.737055639142, 1228: -0.416048620618, 1229: -0.074536281698, 1230: 3.83052040633, 1231: -15.7734219177, 1232: -1.34280607401, 1233: -5.32578515109, 1234: 10.4097665335, 1235: -3.73299496627, 1236: -2.74356345949, 1237: -2.34252584277, 1238: -20.9462217077, 1239: -6.20618702519, 1240: -0.40405775212, 1241: 4.99891349417, 1242: 0.490277979206, 1243: 0.769849087138, 1244: 6.55060223263, 1245: 16.9679614578, 1246: -0.344820979943, 1247: -1.63232958587, 1248: -0.198483648377, 1249: -2.6334282635, 1250: -0.190017566338, 1251: -0.350819157238, 1252: 9.38950057686, 1253: -1.5446128603, 1254: -2.27234632228, 1255: -0.12076105366, 1256: -0.361552787915, 1257: 1.38264692118, 1258: -0.219665624448, 1259: -0.930702288455, 1260: -3.43416146708, 1261: 14.3234428943, 1262: 0.738497606313, 1263: -0.738342121101, 1264: -12.883664344, 1265: 10.9033217705, 1266: 30.5950980241, 1267: 35.6060244167, 1268: 4.73398410003, 1269: -4.13384437825, 1270: 4.4529333265, 1271: -1.73154778755, 1272: 9.40744211635, 1273: -1.29252809361, 1274: 11.5949456016, 1275: 2.03281204875, 1276: 10.0952996482, 1277: 1.29359382457, 1278: -32.0746297602, 1279: 1.15352960962, 1280: 2.6083327045, 1281: 1.28974500431, 1282: -1.9310825012, 1283: 3.0341333787, 1284: 21.2154581351, 1285: -10.0320125418, 1286: -1.57784162037, 1287: -1.40894843588, 1288: -10.2117863344, 1289: 0.329449461333, 1290: -12.778669512, 1291: -19.9808795397, 1292: -3.85711001069, 1293: -23.1931890876, 1294: 2.79804204704, 1295: -5.26068498067, 1296: 0.551423116453, 1297: 5.72508719714, 1298: 28.1105310832, 1299: -31.1872832923, 1300: 8.13527719768, 1301: 9.51176563205, 1302: 3.09360396275, 1303: -0.615436134143, 1304: -4.03864240356, 1305: -1.48944271448, 1306: -1.59438646726, 1307: 21.9170354262, 1308: 22.3426133288, 1309: -2.41819549297, 1310: -4.949116108, 1311: 1.06772315386, 1312: -1.31165816128, 1313: 15.0271577559, 1314: -10.8971722191, 1315: 12.1701700671, 1316: -4.19407138124, 1317: 4.18235401039, 1318: -5.30515321224, 1319: 18.8966757505, 1320: -26.6537517699, 1321: -20.8022382231, 1322: -60.8085915745, 1323: -12.7333766794, 1324: 9.04782002697, 1325: 14.1795914335, 1326: 0.739798105761, 1327: 9.70811967423, 1328: 4.38586810209, 1329: -5.82525202789, 1330: -21.6996549936, 1331: 31.3938412267, 1332: -6.8385780024, 1333: -19.3573238921, 1334: -1.14158914152, 1335: 6.7300432283, 1336: -1.10855328297, 1337: -6.89542402167, 1338: -0.282833567117, 1339: 12.4194535478, 1340: -5.54505838501, 1341: -5.62734035125, 1342: -6.43219547734, 1343: -11.0989466231, 1344: 4.36229418732, 1345: -12.4574495602, 1346: -9.64307872929, 1347: 19.1147258763, 1348: -6.03270043685, 1349: -2.9185461908, 1350: -1.43325145931, 1351: -4.65324907387, 1352: 3.8711840279, 1353: -1.53537909148, 1354: -26.3216634336, 1355: -1.53490095961, 1356: -1.42907698575, 1357: -7.58692950454, 1358: 6.52499281483, 1359: -1.52140728719, 1360: -1.75139244368, 1361: -0.243416988027, 1362: -1.52535442735, 1363: -5.30453769905, 1364: -1.44514310808, 1365: -1.64770496981, 1366: 4.11933132126, 1367: -1.51650844037, 1368: -1.03310926201, 1369: -5.88618483206, 1370: 3.23282143424, 1371: -8.74580995471, 1372: 20.4230660891, 1373: -9.71437522825, 1374: -8.35586759993, 1375: -28.4083683219, 1376: -18.6641788483, 1377: -5.84855294634, 1378: 11.8017125933, 1379: -22.2660929293, 1380: -3.0288794713, 1381: -9.26782421205, 1382: -4.03945461933, 1383: 23.3488620588, 1384: 0.59093034942, 1385: 38.8231836786, 1386: -1.46888580587, 1387: 20.2106850662, 1388: -5.73590742695, 1389: -19.7606440588, 1390: -14.3064861969, 1391: -17.6928018502, 1392: -0.273749775632, 1393: -37.4185447397, 1394: -9.4742952966, 1395: 0.85693515034, 1396: -11.8776152932, 1397: -0.283670330736, 1398: 13.2748370617, 1399: 0.353759445623, 1400: 6.02381785974, 1401: 4.41817212009, 1402: 0.701343934125, 1403: 7.38993672459, 1404: -0.115155170736, 1405: 7.31239762391, 1406: -5.47526378652, 1407: 5.00918614241, 1408: 4.48050159949, 1409: -0.596001420481, 1410: 8.23955734106, 1411: -0.629476729388, 1412: -3.09228631984, 1413: -13.9186131849, 1414: -0.435844179666, 1415: -6.65870547972, 1416: -5.08704394432, 1417: -1.17267264421, 1418: -5.20353258684, 1419: -11.8493791266, 1420: 2.87189925531, 1421: 0.258360870993, 1422: 8.753110121, 1423: -12.6901537123, 1424: 2.05553451658, 1425: 0.0856260504326, 1426: 2.00048420217, 1427: 1.8085459594, 1428: 5.63197617017, 1429: -1.06589020479, 1430: 2.86528223293, 1431: 7.03449566319, 1432: -1.94660798766, 1433: 0.696515193144, 1434: 30.2889875798, 1435: -6.1588697738, 1436: -1.35617844343, 1437: 2.21955811658, 1438: -1.39335015688, 1439: 0.536363034904, 1440: 1.36399491501, 1441: 3.70262493082, 1442: -0.621041104614, 1443: -3.25398185545, 1444: -1.59145626369, 1445: -0.451538959508, 1446: -1.10597020957, 1447: -1.71090406266, 1448: -5.90783463345, 1449: -1.24629741723, 1450: -2.07669021417, 1451: 1.89469964001, 1452: -1.08559917021, 1453: 0.84600346186, 1454: -2.2111983035, 1455: 1.65709147005, 1456: -1.17971084067, 1457: -1.55961430395, 1458: 3.62021234805, 1459: -14.5598769752, 1460: -3.64478017054, 1461: 2.96473757279, 1462: -1.50622739618, 1463: 0.174598316521, 1464: -1.19443867108, 1465: -1.7352026523, 1466: -1.32045618772, 1467: -0.704476036931, 1468: 0.0253299060319, 1469: -1.43139078642, 1470: -11.5130612331, 1471: -1.77014855903, 1472: -1.25925960595, 1473: -1.9872073987, 1474: -2.02886447393, 1475: -1.07254703959, 1476: 2.74812419024, 1477: -2.14362493505, 1478: 9.91528430227, 1479: -1.54259487429, 1480: -1.12631080663, 1481: -4.29058549566, 1482: -0.633695014707, 1483: 2.93641934475, 1484: 0.016235331799, 1485: -1.34879515577, 1486: -1.58783112739, 1487: 0.190706456975, 1488: -1.30144814175, 1489: -0.966462342989, 1490: -1.0219005532, 1491: -1.08826728379, 1492: 1.66599873638, 1493: -1.11953504627, 1494: -1.48003037211, 1495: -1.66967368401, 1496: -1.34865180163, 1497: -1.86902062567, 1498: -1.98091540385, 1499: -3.95401316806, 1500: -1.63075199442, 1501: 0.649914899308, 1502: 26.6345454315, 1503: -1.23838378183, 1504: -1.63584085937, 1505: -1.47619577392, 1506: 1.53797287717, 1507: -1.79616121466, 1508: 5.03987074223, 1509: -14.1237977831, 1510: -4.00011504472, 1511: -17.9896665837, 1512: -7.31953295962, 1513: -1.32093667555, 1514: 2.05469133362, 1515: 4.43397439224, 1516: -0.438863074832, 1517: -4.50827164576, 1518: -1.05305157015, 1519: -0.413769168548, 1520: 2.84992944742, 1521: 10.7537687411, 1522: -0.248764050222, 1523: -0.246794453716, 1524: -0.410506205994, 1525: 0.0160188762383, 1526: -0.337176865471, 1527: -1.1994225118, 1528: -0.324804859274, 1529: -1.91260838383, 1530: -0.488914637586, 1531: -0.215595153561, 1532: 5.10001014604, 1533: -0.738523021887, 1534: -2.19395176965, 1535: -0.411683843895, 1536: -0.723972875175, 1537: -11.2382290954, 1538: 2.4115034329, 1539: -3.71919006569, 1540: 0.0619456294754, 1541: -4.58958012963, 1542: -0.161033188888, 1543: -0.263589601329, 1544: 3.50298728306, 1545: -0.254712629422, 1546: -1.35095556657, 1547: 1.91025185554, 1548: -0.56658867066, 1549: -0.241408972295, 1550: -0.318355569888, 1551: -0.271203097507, 1552: -0.82862727038, 1553: 6.03874215833, 1554: -0.397462653777, 1555: -0.378407479698, 1556: -0.444079154488, 1557: 0.992161130811, 1558: -0.77780774674, 1559: -7.37279274475, 1560: -1.92125674168, 1561: 12.5451772095, 1562: -2.16191275932, 1563: 0.399151636235, 1564: 0.640268317137, 1565: -1.72509777077, 1566: 1.73351187306, 1567: -4.73523391552, 1568: -0.338660918464, 1569: 0.680934072402, 1570: -0.211831951839, 1571: 0.221187477171, 1572: -4.72563309851, 1573: 1.22153867172, 1574: 0.120959527208, 1575: -0.371008462493, 1576: -0.208725533227, 1577: -0.225011038844, 1578: -0.320921051837, 1579: -1.65345584327, 1580: -4.62473665654, 1581: -1.05620032474, 1582: -1.79643742956, 1583: 10.4389926373, 1584: 2.84490354533, 1585: 0.879457776353, 1586: 1.00392410558, 1587: -0.194245057145, 1588: -4.54091764042, 1589: -0.335938989577, 1590: -3.88236131572, 1591: -0.628784980869, 1592: -0.175934236388, 1593: -0.261297530945, 1594: -5.67260126876, 1595: -0.28170226224, 1596: -1.4291095135, 1597: -0.234791188717, 1598: 1.36854774314, 1599: -0.217070390167, 1600: -0.176461520579, 1601: 0.00577588027963, 1602: -1.53989843116, 1603: 2.82541441885, 1604: -0.292355327031, 1605: -0.312261150253, 1606: -0.182399885254, 1607: -0.291209537186, 1608: -0.867445643617, 1609: 2.17215652971, 1610: -0.938811077207, 1611: -0.228146407072, 1612: -0.428183732535, 1613: 8.13080699195, 1614: -0.929681340476, 1615: 4.45341777271, 1616: -0.334017303389, 1617: -0.512951040455, 1618: 1.5878608781, 1619: -1.21898769391, 1620: -1.63557340607, 1621: 1.02771474835, 1622: -0.3618044304, 1623: 19.8797807072, 1624: -1.47791033012, 1625: -0.838588105194, 1626: 0.324460078285, 1627: -3.73024020347, 1628: -0.249381364968, 1629: 3.80105622352, 1630: -10.2446935642, 1631: -1.76413681966, 1632: -2.64112702632, 1633: -0.548043165147, 1634: 4.85379419932, 1635: -1.44023397324, 1636: -1.1879529331, 1637: 7.17925506511, 1638: -0.103889665154, 1639: 2.69344578773, 1640: 6.02550062199, 1641: 4.13410082455, 1642: 5.34978144757, 1643: 2.23537543244, 1644: -3.49754240133, 1645: 1.01119074346, 1646: 14.5601404406, 1647: -3.44469130846, 1648: 9.65801222517, 1649: -4.65826848816, 1650: 6.45155144099, 1651: -1.49380632949, 1652: 1.69321384226, 1653: -0.73974652618, 1654: -1.22507797241, 1655: -1.45437180452, 1656: 2.03944475172, 1657: 1.00662886955, 1658: -1.70079619163, 1659: -0.970082407063, 1660: -1.96698849363, 1661: -0.905130231933, 1662: 5.16608399893, 1663: 4.26774049156, 1664: 4.70949027389, 1665: 1.12042491811, 1666: 1.510193504, 1667: -1.24354727495, 1668: -3.94144787255, 1669: 4.0698335714, 1670: 20.5930978582, 1671: 7.63394171378, 1672: -5.60742483828, 1673: 4.94076357807, 1674: 4.97987871118, 1675: 3.22378236263, 1676: -9.26617334366, 1677: -1.03185893553, 1678: -4.10381093723, 1679: -0.54675554157, 1680: 1.74277650538, 1681: 1.50710349222, 1682: -3.58504184539, 1683: -1.46181846569, 1684: -0.561584494682, 1685: -1.65400645531, 1686: -1.80941532218, 1687: -1.71930696041, 1688: 3.26655596025, 1689: -5.16895922106, 1690: 4.36040890921, 1691: 11.5732001275, 1692: 2.48445134436, 1693: 7.63839035548, 1694: 4.62001482384, 1695: -1.0824491265, 1696: 2.14691476746, 1697: 1.63595367003, 1698: 6.92461649962, 1699: -0.11033376095, 1700: 5.49805862894, 1701: 5.00680082213, 1702: -1.12853367918, 1703: 0.199726056483, 1704: -1.36081619443, 1705: -1.31428303763, 1706: -1.71544363614, 1707: -0.616080767287, 1708: -0.138027316617, 1709: -1.51802046023, 1710: 0.137279036341, 1711: -1.40645078259, 1712: 0.753845360803, 1713: -1.40512542907, 1714: -0.151037525992, 1715: -3.21575936919, 1716: -0.990928643571, 1717: -0.822542000162, 1718: 0.3046913026, 1719: 1.76075318023, 1720: 2.56984105939, 1721: -4.23461446839, 1722: 15.752301787, 1723: -5.41259288312, 1724: 5.43080332306, 1725: -3.17767711064, 1726: 0.3692546766, 1727: 1.3526254583, 1728: 3.40630592776, 1729: -1.14239188307, 1730: 0.534273205515, 1731: 0.690710013575, 1732: -2.53812003248, 1733: -2.24200588088, 1734: -1.20929915177, 1735: -1.32060335618, 1736: 1.20053315991, 1737: 1.50982046461, 1738: 1.31665822534, 1739: 6.19649840329, 1740: -1.4842277994, 1741: -1.90129624255, 1742: -1.63900771872, 1743: -4.56650778654, 1744: -0.996045340221, 1745: 3.12952486996, 1746: 0.992431336315, 1747: 7.79382390323, 1748: 1.02194928718, 1749: -0.347859069253, 1750: 8.91912733042, 1751: -1.23770417069, 1752: -9.93851158084, 1753: -0.406243710674, 1754: -1.73604684263, 1755: -2.59726677535, 1756: -1.88274326188, 1757: 7.73503767891, 1758: -9.55298184381, 1759: -14.4666737257, 1760: -6.58310543574, 1761: -15.2640621167, 1762: 18.4621057219, 1763: 16.3038497069, 1764: 19.7646444493, 1765: -1.70643570603, 1766: -9.61415322791, 1767: -17.2173187794, 1768: 2.23522041578, 1769: -18.9774638399, 1770: 4.32830882214, 1771: 5.32011495647, 1772: 16.2659729633, 1773: 2.72178647466, 1774: -7.42071555811, 1775: 6.9890727738, 1776: 12.5267387284, 1777: 4.54345015585, 1778: -10.8576513531, 1779: 4.74078244664, 1780: 3.0430247243, 1781: -1.79811045926, 1782: 10.947177367, 1783: -6.76840832771, 1784: 8.33752903649, 1785: 4.93799039985, 1786: 2.56395891591, 1787: -0.468906325828, 1788: -30.3856930119, 1789: -2.82648508083, 1790: 0.612184241215, 1791: 10.2971032886, 1792: -23.0295317313, 1793: 0.13359661705, 1794: 5.04662905962, 1795: -6.90139077896, 1796: 4.59814808198, 1797: 13.1095035111, 1798: 4.45220408581, 1799: -11.4594219717, 1800: 13.4758654406, 1801: 4.88935885737, 1802: 4.64477736668, 1803: -9.2865177913, 1804: 5.02524360858, 1805: -1.70462679759, 1806: 1.72722918647, 1807: -1.40256340481, 1808: -14.0462108207, 1809: 3.77222312268, 1810: 5.26929586645, 1811: 5.02064225291, 1812: 16.6076057163, 1813: 4.81537716419, 1814: 6.35237622319, 1815: 5.79450645584, 1816: 4.2443548959, 1817: -6.84594539443, 1818: 5.11773857252, 1819: -0.243599909443, 1820: -1.12141306599, 1821: -1.49136936011, 1822: -4.91924485572, 1823: -4.4587884115, 1824: -12.4543596631, 1825: 2.36361620643, 1826: 1.95021277119, 1827: -3.85424512273, 1828: 5.32208056141, 1829: 0.0899020505129, 1830: -2.44515846328, 1831: -3.76149881831, 1832: 0.700466958301, 1833: 14.3539266219, 1834: 4.81704962196, 1835: -17.1836210103, 1836: 3.44195907557, 1837: 4.52916649546, 1838: -19.3901210288, 1839: 1.36979832128, 1840: 3.33871926727, 1841: 7.30058425652, 1842: 4.68549944227, 1843: 5.37234404832, 1844: 4.87495964307, 1845: 5.3398065335, 1846: -32.5146725439, 1847: -1.9666776192, 1848: -0.239527812593, 1849: 5.28253269164, 1850: 5.35111500519, 1851: -9.34965510526, 1852: 4.46567871282, 1853: 4.74245408268, 1854: 5.56264862806, 1855: 2.22211895892, 1856: 0.934725271637, 1857: -2.14962444221, 1858: 1.74561154921, 1859: -8.51886333243, 1860: 17.463787546, 1861: 1.00267729301, 1862: -6.33297773462, 1863: -13.5164539585, 1864: 0.934290074526, 1865: 7.89728822657, 1866: 1.47152603014, 1867: -9.71649496134, 1868: 11.6904129309, 1869: -5.75278466108, 1870: -19.6430997483, 1871: 1.12117958412, 1872: 1.85357097334, 1873: 5.60437303368, 1874: -10.0573274113, 1875: 1.15415753752, 1876: 6.20904750669, 1877: -0.600370893187, 1878: 1.26348386395, 1879: 0.0733538352217, 1880: -1.26397559674, 1881: 0.263740514198, 1882: 3.59005603484, 1883: -1.86741559394, 1884: 2.3851673184, 1885: 0.105568563657, 1886: 1.28177611086, 1887: -8.63663255192, 1888: -2.17640220355, 1889: 1.30672079427, 1890: 11.4640222928, 1891: 0.817906673423, 1892: 0.916341093129, 1893: -10.3620054923, 1894: 0.99131908317, 1895: 5.35825736276, 1896: -1.0080393684, 1897: 1.05206213864, 1898: 1.02728598436, 1899: 1.25998493616, 1900: 1.07636882317, 1901: -1.55003019805, 1902: 18.6735363195, 1903: 0.218664692351, 1904: 1.56479066818, 1905: 1.45933246569, 1906: 6.73345237666, 1907: -1.10437597709, 1908: -3.74779738676, 1909: -5.94750169197, 1910: -13.3009128807, 1911: -0.187914075143, 1912: 3.2609788247, 1913: -4.16764404773, 1914: -2.04981612837, 1915: -7.65766714768, 1916: 8.22264954839, 1917: 0.891982098582, 1918: -19.1225372788, 1919: 1.0176617303, 1920: -13.7406490664, 1921: 13.9091862099, 1922: -0.962346063298, 1923: -1.74109826491, 1924: 0.944061507644, 1925: 1.18045753765, 1926: 0.835204080763, 1927: 1.18101209124, 1928: 0.56877722665, 1929: 14.6929143393, 1930: 0.174615804042, 1931: 10.6019292195, 1932: 12.9517722289, 1933: 6.25861218491, 1934: -12.5078659864, 1935: -2.90750443038, 1936: -17.7699666454, 1937: -4.93154712476, 1938: 1.15428049159, 1939: -4.65987827342, 1940: 0.0752155346014, 1941: -0.592301019646, 1942: 0.971758091893, 1943: 1.48097299027, 1944: 1.01167784533, 1945: 2.54833327635, 1946: 1.1042388147, 1947: -5.21161843378, 1948: 0.883082719461, 1949: 0.967288494705, 1950: 7.86859999863, 1951: 5.4395783824, 1952: -9.63625287361, 1953: 0.973261905433, 1954: 1.05201662472, 1955: 5.57578019921, 1956: 1.1013614777, 1957: 5.22064653933, 1958: -2.41312040509, 1959: -8.16290407298, 1960: 1.60123169134, 1961: 1.28876485681, 1962: 1.74384464509, 1963: 1.72333095872, 1964: 27.5052182469, 1965: -3.41646273623, 1966: 4.65081156021, 1967: -4.0093985367, 1968: -2.63789354288, 1969: 4.26721723066, 1970: 4.25356107168, 1971: 0.716641252394, 1972: 0.442199688051, 1973: 1.07875261288, 1974: -2.47097851408, 1975: -0.179396769598, 1976: 21.3845861599, 1977: 1.37698579071, 1978: 3.79899327441, 1979: 11.6575143446, 1980: -0.20220557005, 1981: 0.936873069067, 1982: 5.21652560181, 1983: -7.89080469594, 1984: 3.8664515329, 1985: 6.30129531083, 1986: 0.78894195843, 1987: -0.523119850169, 1988: 7.02244310362, 1989: -3.67970584476, 1990: 4.60083085315, 1991: -16.8217345461, 1992: 16.5416903355, 1993: -14.5795197641, 1994: -1.59580468816, 1995: 7.21852490506, 1996: -2.65830703957, 1997: 9.4281969038, 1998: -21.3877675902, 1999: -23.37880468, 2000: 8.55746413299, 2001: 4.99511487684, 2002: -0.283781414147, 2003: 4.98883264239, 2004: 5.34830619847, 2005: -14.4228347565, 2006: 1.76436297298, 2007: 2.96843347574, 2008: -4.09346833039, 2009: 2.40791064588, 2010: -1.57774252165, 2011: -14.5101170524, 2012: -2.02635820349, 2013: -5.93902400256, 2014: -3.57532288594, 2015: -3.71353707775, 2016: 1.38086453871, 2017: -2.12273291052, 2018: -1.85948304106, 2019: -15.1205255476, 2020: 4.68202377529, 2021: -5.81443220867, 2022: 0.522395097875, 2023: -6.74618076322, 2024: -16.1943421324, 2025: 1.91722106754, 2026: 4.40496128005, 2027: 15.5266773748, 2028: -1.3026087362, 2029: -9.97362577115, 2030: -14.9852847264, 2031: 6.27420333274, 2032: 3.8490359112, 2033: -1.19653950527, 2034: 6.12601973119, 2035: -0.26420423873, 2036: 2.80154645725, 2037: 2.48938860044, 2038: -5.63371453448, 2039: -3.82075072349, 2040: -12.5147648744, 2041: -1.65104565023, 2042: -16.4125891639, 2043: -6.08293642969, 2044: -4.50560983313, 2045: -27.4581167291, 2046: -3.74903566239, 2047: 6.01338745856, 2048: 1.00056484961, 2049: 7.66832250433, 2050: 0.0822174047365, 2051: 3.48065262594, 2052: -4.33447406507, 2053: 2.33605962794, 2054: 2.3785416669, 2055: 5.53818013706, 2056: 8.81987642097, 2057: 4.83634410045, 2058: 4.30771722215, 2059: 4.09976923788, 2060: 5.44347085254, 2061: 11.2922321326, 2062: 4.87596567262, 2063: 5.02925712356, 2064: 6.07337145827, 2065: 2.74317419494, 2066: 4.73032236618, 2067: -7.90413035732, 2068: -19.9962693448, 2069: -8.34619930649, 2070: 4.14378013606, 2071: -25.1017210446, 2072: -1.37007585531, 2073: -29.9211280019, 2074: 2.17672903239, 2075: -1.96575330875, 2076: -10.4093968352, 2077: -2.16223257546, 2078: 2.44805714543, 2079: 2.81232779954, 2080: 4.08538139801, 2081: -3.28774819957, 2082: -4.48320588698, 2083: -3.55405747898, 2084: 5.15027899282, 2085: 1.08862142868, 2086: -20.3044821683, 2087: -5.44722842302, 2088: 22.3727584577, 2089: 6.6544479623, 2090: -6.49666243692, 2091: -6.74282881462, 2092: -9.65040529741, 2093: 5.43286876397, 2094: -17.6147567803, 2095: -1.28550794589, 2096: -0.548341332381, 2097: 13.2382475211, 2098: -50.267133432, 2099: -38.668909453, 2100: 0.0350334217636, 2101: 6.89160936178, 2102: 10.0303816943, 2103: -18.0020135072, 2104: 15.0300189687, 2105: -45.9987735494, 2106: 1.36235158366, 2107: 8.57220116976, 2108: 23.3695494251, 2109: 2.07506202266, 2110: 26.820175942, 2111: 40.1576335667, 2112: 77.0279660522, 2113: -24.8957292738, 2114: 21.9305804471, 2115: 17.7259476294, 2116: 4.49286506035, 2117: 22.1222289415, 2118: -23.3851148837, 2119: -2.09294454902, 2120: -11.8818468492, 2121: 13.7504916786, 2122: -29.8729105164, 2123: 57.5468840766, 2124: -28.6607389428, 2125: -2.3149249529, 2126: -12.660056495, 2127: 27.1598705293, 2128: 19.4614441323, 2129: -2.93346576367, 2130: 3.46395301243, 2131: 42.9747178179, 2132: -9.55668267964, 2133: -10.1100591596, 2134: 19.1533486761, 2135: 30.4095819348, 2136: -7.15312006941, 2137: -4.62720227908, 2138: 9.09089716639, 2139: -9.38560361687, 2140: -11.6869361916, 2141: -23.9857996203, 2142: 0.207767139086, 2143: 12.9710493252, 2144: -3.17037221559, 2145: -19.6897228756, 2146: 1.60857886661, 2147: -6.18439378751, 2148: 15.9522050586, 2149: 19.6558042577, 2150: -6.32566549524, 2151: -10.9950624943, 2152: 8.06201831392, 2153: 30.8246365048, 2154: -10.7433042651, 2155: -32.7871229339, 2156: -16.4501784646, 2157: 18.3537721503, 2158: -22.9744629392, 2159: 18.6785628355, 2160: -5.5616208138, 2161: 2.71857181577, 2162: -7.71198433183, 2163: -12.6201577927, 2164: -7.99418391135, 2165: -7.75318737042, 2166: -5.78567190888, 2167: -7.08781385298, 2168: -11.5383682096, 2169: 18.0023106244, 2170: -11.1556870189, 2171: 17.1876000719, 2172: -9.81772505081, 2173: -13.6916326955, 2174: 11.4045491968, 2175: -9.37698076276, 2176: 18.4686718307, 2177: -6.13904748779, 2178: 1.73179350484, 2179: 40.5105918163, 2180: -7.14356391033, 2181: -15.9874405791, 2182: -34.6265469779, 2183: -6.95724033432, 2184: -6.69098969545, 2185: -5.54987773226, 2186: -6.9219000756, 2187: 96.2268857661, 2188: -7.41140345793, 2189: -1.34928674324, 2190: -8.95459946515, 2191: 13.3165875817, 2192: -7.72175522859, 2193: -4.94914569285, 2194: -14.5216080531, 2195: -2.97750164664, 2196: -6.26346729859, 2197: -0.658428139004, 2198: -5.4080245987, 2199: 3.85559182863, 2200: 20.4574141582, 2201: -7.09762847341, 2202: -6.98471398123, 2203: -7.23211389078, 2204: 11.3990873839, 2205: -3.61439331779, 2206: -32.9646983669, 2207: -97.313579698, 2208: -17.9466517878, 2209: 32.1360382323, 2210: 5.80646408169, 2211: -28.8760412425, 2212: 6.08445408456, 2213: 40.8876067873, 2214: 46.8826333606, 2215: -35.5479215156, 2216: 23.522525515, 2217: -4.01548299106, 2218: -15.9869320498, 2219: 8.44236515562, 2220: -1.46297025898, 2221: -1.61733366031, 2222: -4.07316313053, 2223: -22.3704083618, 2224: -1.45804274388, 2225: 0.229134558276, 2226: -2.14790768606, 2227: -26.0967927951, 2228: 0.913004770346, 2229: -0.641960163256, 2230: 43.8781529411, 2231: -2.50777917055, 2232: -3.93956222224, 2233: -6.90243425775, 2234: -8.62165453039, 2235: 17.5958976377, 2236: 7.16319275719, 2237: 9.52447303119, 2238: -2.61381130523, 2239: -9.76636701995, 2240: -1.43116169444, 2241: -1.4066338033, 2242: -13.7948770313, 2243: -1.50846728114, 2244: -7.44353671932, 2245: 7.55478517589, 2246: -1.73466489084, 2247: -1.39384275119, 2248: -1.62418170542, 2249: -1.28172988282, 2250: -6.71832690146, 2251: -34.4649713062, 2252: -1.8850774586, 2253: -1.80930752731, 2254: -2.53225463871, 2255: 4.54355177634, 2256: 7.80674468902, 2257: 36.0736704184, 2258: 6.41047204082, 2259: 15.7636002593, 2260: 26.9941942686, 2261: -19.4668972127, 2262: 9.28727524656, 2263: -18.935452916, 2264: -4.76663388037, 2265: -13.7986386749, 2266: -1.00187271673, 2267: -18.6218588534, 2268: -1.3374237778, 2269: 4.44107731966, 2270: 10.7688490788, 2271: 6.56564669764, 2272: -1.62048678767, 2273: -1.36498257566, 2274: -1.46801311454, 2275: -1.27058274906, 2276: -1.80844888836, 2277: -8.9340117743, 2278: -9.74764665019, 2279: 0.863050604316, 2280: 23.7349341106, 2281: -25.1372357771, 2282: -0.988307777021, 2283: -4.84187884747, 2284: -0.405727650004, 2285: -2.09736375373, 2286: 25.4456696475, 2287: -1.75071729566, 2288: -22.0972738203, 2289: -4.43127084506, 2290: -1.69455771835, 2291: 1.70419039348, 2292: 47.1442395943, 2293: -1.45354197602, 2294: -3.78051792772, 2295: -1.46970025977, 2296: -46.4978108613, 2297: -1.39893400486, 2298: -1.39820532969, 2299: 25.8993652739, 2300: -7.12843132355, 2301: 39.8916387536, 2302: -1.41484077042, 2303: -1.36994423611, 2304: -4.74087533922, 2305: -1.70198902881, 2306: -2.24606520617, 2307: 48.9007105423, 2308: 22.8444662846, 2309: -3.94580648267, 2310: -5.09018989675, 2311: 11.5288501442, 2312: -1.07354142674, 2313: 1.69104493942, 2314: 7.82454388759, 2315: 22.7206744794, 2316: 12.8937283268, 2317: 12.6712421243, 2318: -6.49178390313, 2319: 6.65761973381, 2320: -1.77159215906, 2321: 22.2461821132, 2322: 1.57521773715, 2323: -4.02435187818, 2324: -2.6160907628, 2325: 4.5352949179, 2326: -0.0894478913134, 2327: -18.4714205432, 2328: -34.9290777673, 2329: -4.92467705517, 2330: 9.63913243953, 2331: 39.4882444523, 2332: -14.1046995376, 2333: -2.04712435895, 2334: 0.243992738746, 2335: 28.3064309675, 2336: 19.3215113225, 2337: 9.42578127555, 2338: -4.91394047345, 2339: -35.3499080633, 2340: -29.2576327148, 2341: 7.33288722951, 2342: 24.7630458992, 2343: -20.3090394043, 2344: 17.4819057374, 2345: 4.78444451383, 2346: -14.3881835892, 2347: 30.6068636277, 2348: -6.13610048692, 2349: 9.74861532971, 2350: -5.71714998532, 2351: -4.80158975193, 2352: -6.21452676028, 2353: -7.32211604342, 2354: 7.80573356021, 2355: -15.626175781, 2356: -7.32094628245, 2357: -3.97279177423, 2358: 0.298331363765, 2359: -6.76593578452, 2360: 21.0691878691, 2361: 16.9520433606, 2362: -2.32659524712, 2363: -10.1695295428, 2364: 8.06283525747, 2365: 2.40318669482, 2366: -0.751587984697, 2367: 17.1478301197, 2368: -35.2732764796, 2369: 8.6481984173, 2370: -24.8277777555, 2371: 16.6782435337, 2372: -1.73702549584, 2373: -4.24188566575, 2374: 19.2237671849, 2375: -35.1447916589, 2376: -5.9664686365, 2377: 2.608387782, 2378: 24.4998970505, 2379: 11.6545058698, 2380: 11.6425602586, 2381: -8.77804803071, 2382: -49.2912738553, 2383: -8.35366678541, 2384: 10.9973497451, 2385: -2.82198120145, 2386: 19.187327873, 2387: 26.2727356008, 2388: 34.1849407672, 2389: 25.0001786255, 2390: -0.249222358052, 2391: -1.2527313519, 2392: -10.4460712642, 2393: -6.90450586548, 2394: -53.7001609607, 2395: 3.66143188201, 2396: -6.15197982722, 2397: -7.59061916682, 2398: -34.7573216645, 2399: 9.72612348539, 2400: -1.39593140391, 2401: -15.6710361054, 2402: -3.88039443759, 2403: -3.74913398147, 2404: -4.28377074306, 2405: 36.4989635901, 2406: -5.11869795373, 2407: -6.92639982123, 2408: -7.64275323568, 2409: -7.11710801939, 2410: 26.8714443834, 2411: -5.98040420373, 2412: -6.94785025602, 2413: -29.0725030058, 2414: 4.40321505676, 2415: 1.70798023336, 2416: -41.1983099277, 2417: -16.2106050669, 2418: 33.984683536, 2419: -19.9459682638, 2420: 39.2910621913, 2421: 2.93804923675, 2422: 21.2425017037, 2423: 36.2661656536, 2424: 20.387970068, 2425: 47.3778130166, 2426: -6.96863857872, 2427: -5.35143743534, 2428: 5.48146436291, 2429: 59.6622184815, 2430: -12.3623535441, 2431: -9.98821962216, 2432: 25.0648859664, 2433: -7.15833132024, 2434: 50.4908497427, 2435: -8.12401897802, 2436: -31.258964489, 2437: 2.07733669291, 2438: -29.4331515989, 2439: -7.76399722169, 2440: 81.3875471872, 2441: 35.4338323555, 2442: 1.77438350539, 2443: 3.05387560331, 2444: -0.102672288589, 2445: 28.5889266361, 2446: -4.59672014109, 2447: 13.1057562443, 2448: -25.5831226295, 2449: 8.26711811114, 2450: -17.2583099724, 2451: 0.630175786795, 2452: -4.32113321638, 2453: 4.71455708092, 2454: -12.3833127253, 2455: -7.95689977154, 2456: 7.88649109764, 2457: 2.91718965169, 2458: -0.376683032392, 2459: -4.19716221906, 2460: 0.301040704649, 2461: 13.4683015184, 2462: -10.7957683888, 2463: 9.15919297389, 2464: 10.5542226653, 2465: -17.3621400063, 2466: 4.35441335464, 2467: -1.9670014679, 2468: -4.26370455384, 2469: 3.00333299486, 2470: 3.14841776948, 2471: 3.66367303352, 2472: 3.33401315715, 2473: 2.91557429936, 2474: 5.82930628214, 2475: 6.44233382163, 2476: 0.0527419098817, 2477: 8.92961632926, 2478: 20.4529143497, 2479: -8.16683632019, 2480: 8.59025224585, 2481: 1.37162746549, 2482: 11.1586923174, 2483: -6.38081274936, 2484: -15.4894178214, 2485: -26.7727125852, 2486: 12.4180449993, 2487: 19.8198823317, 2488: -2.66365533424, 2489: -2.65285584664, 2490: -18.2769780205, 2491: -7.15717975493, 2492: -7.96944851633, 2493: -7.44146455879, 2494: -4.82003916392, 2495: 24.9282589458, 2496: -1.3371730741, 2497: -0.583880935016, 2498: 4.44530561065, 2499: -0.647555453693, 2500: 13.3228400497, 2501: -3.13467159228, 2502: -17.036821886, 2503: -10.7793784162, 2504: -0.549391409673, 2505: 4.44777908204, 2506: 8.46964862936, 2507: 16.7483825554, 2508: -9.9272812581, 2509: -1.62493401899, 2510: -3.10760245471, 2511: 5.59831590411, 2512: 2.62577695656, 2513: -0.648087706445, 2514: -1.24655010255, 2515: -2.58790877099, 2516: -1.26790234014, 2517: -8.39112353634, 2518: 2.82133771709, 2519: 4.00098746948, 2520: 11.0258881162, 2521: -1.8879921232, 2522: 4.97677655351, 2523: 14.3018439833, 2524: -7.35882719254, 2525: -20.1275588483, 2526: -1.25489800468, 2527: -6.59293266246, 2528: 15.4870014637, 2529: 0.519520631676, 2530: -3.41552874954, 2531: 0.984603783363, 2532: -1.60066945643, 2533: -5.90849399834, 2534: -4.09787863909, 2535: -1.6366501987, 2536: 3.89290419916, 2537: -1.08624322123, 2538: -2.24330992639, 2539: -0.93228851584, 2540: 2.85065874205, 2541: -0.960758794953, 2542: 0.724229271898, 2543: -0.931730940513, 2544: -8.63788314794, 2545: -1.8213297827, 2546: 17.6601767443, 2547: -2.07560873434, 2548: 0.228848174824, 2549: -28.0000423322, 2550: -3.84059628487, 2551: -1.97032051517, 2552: -1.13724370207, 2553: -7.4131526842, 2554: -4.22340056278, 2555: -19.0720513639, 2556: -37.1785398066, 2557: -18.4171085044, 2558: 23.8207236774, 2559: 0.231690736178, 2560: -7.70609967634, 2561: -10.5527606504, 2562: 19.2558100417, 2563: -15.6018352601, 2564: -16.2927872786, 2565: -18.3849269932, 2566: -28.7799585335, 2567: -21.6156982511, 2568: 40.3937371303, 2569: -0.0211860002577, 2570: -0.971095894414, 2571: 3.65399488735, 2572: -5.59496430741, 2573: -0.0306214051611, 2574: -0.759497251383, 2575: -5.01029760295, 2576: 10.6919218689, 2577: -2.34931969945, 2578: -0.082931727823, 2579: -14.0664009811, 2580: 7.47023128153, 2581: 16.0417577131, 2582: -4.01540954087, 2583: -1.83437241861, 2584: 28.4562625786, 2585: 32.93672189, 2586: -1.42402894328, 2587: -0.608844803511, 2588: -20.1863364118, 2589: -0.206248732348, 2590: -0.237883256138, 2591: 4.14792979061, 2592: -0.133327153205, 2593: -1.55705695649, 2594: 3.33646280659, 2595: -0.342760810632, 2596: -0.296369459191, 2597: -0.146334142676, 2598: -0.25067448031, 2599: -2.10923582219, 2600: 32.3815747573, 2601: -0.889520401489, 2602: -0.74281865098, 2603: -0.39512638699, 2604: 11.3908019477, 2605: -12.6396973763, 2606: 29.9941516091, 2607: -1.72210210603, 2608: 14.4744181807, 2609: -31.5903055747, 2610: -7.36807186302, 2611: 15.8358303971, 2612: -1.93007348287, 2613: 4.25144847341, 2614: 1.56778806515, 2615: -0.702106595293, 2616: 11.0862142624, 2617: -0.333705562652, 2618: -4.84722868552, 2619: -13.3044216804, 2620: -38.5649325155, 2621: -2.75550888482, 2622: -0.214996337351, 2623: -0.420426464203, 2624: -0.932052070004, 2625: -0.123512019507, 2626: 11.1097691957, 2627: -30.521670033, 2628: 2.45726904124, 2629: 10.1832440411, 2630: -23.847363526, 2631: 3.15461656642, 2632: 3.74411575147, 2633: 0.879454543553, 2634: -2.31206881716, 2635: 31.544489399, 2636: -0.115770030715, 2637: 3.7048685052, 2638: -0.618385973265, 2639: 3.74457464128, 2640: -6.81332600902, 2641: -7.19478952716, 2642: -0.282485706405, 2643: -1.2221741916, 2644: -0.183647155118, 2645: 3.83821289336, 2646: -0.147969420447, 2647: -0.199261062895, 2648: 13.116888033, 2649: -1.2126741411, 2650: -6.52865738035, 2651: -0.100391193899, 2652: -0.176160245209, 2653: 2.15651258641, 2654: -0.226024176231, 2655: -0.929131858572, 2656: -24.1895682155, 2657: 15.5665243744, 2658: 0.588497973802, 2659: -0.803718838198, 2660: -4.93383089381, 2661: 3.17527353783, 2662: 5.39620847476, 2663: -2.42196977345, 2664: 3.34983233345, 2665: 23.598650642, 2666: 3.59342106138, 2667: -1.53438084091, 2668: 1.8652339805, 2669: -1.3655116831, 2670: 10.6865251926, 2671: 3.17266258174, 2672: -2.2701463864, 2673: 4.81533934178, 2674: 3.92674521382, 2675: -0.811627797164, 2676: -18.5300626103, 2677: -23.1939944407, 2678: -5.55110306392, 2679: 3.56182760338, 2680: -1.14928563325, 2681: 13.4161250963, 2682: -3.04280664015, 2683: -0.685719140885, 2684: -1.36253109845, 2685: 5.78312629103, 2686: -0.761724583183, 2687: 19.4829899353, 2688: -4.21822098043, 2689: 9.11455639591, 2690: -8.77278616175, 2691: -10.1427946613, 2692: 31.2193411588, 2693: -5.87311365873, 2694: -20.6522038739, 2695: 3.41522166504, 2696: -40.9150567326, 2697: 3.91755789652, 2698: -4.25311554689, 2699: -4.54177230054, 2700: -16.1359462, 2701: -3.32501654096, 2702: -1.59683228169, 2703: 4.25453628636, 2704: 5.22054668977, 2705: -0.786932602237, 2706: -9.91514163925, 2707: -1.86795875094, 2708: -1.88583460392, 2709: 13.9296854105, 2710: 19.1374770434, 2711: 12.1202484861, 2712: 18.673522503, 2713: -3.75454407577, 2714: 2.58417638822, 2715: -24.0235965514, 2716: 17.2286208763, 2717: -0.483379004798, 2718: 20.3232516902, 2719: -25.7028607548, 2720: 17.9678642079, 2721: 6.28563857379, 2722: 7.3022305686, 2723: 7.23196743871, 2724: 5.55637151519, 2725: -10.1920071237, 2726: -16.8333363875, 2727: 46.7524528721, 2728: 43.373839031, 2729: -7.73824123048, 2730: 2.23936510088, 2731: -3.85222364898, 2732: 5.85850432679, 2733: -8.19377011851, 2734: -6.28253829918, 2735: 0.0686743734818, 2736: -4.27353008029, 2737: 5.0117774867, 2738: 24.0399003775, 2739: -2.73491773036, 2740: -8.35848534293, 2741: -1.55727285088, 2742: -1.85635003209, 2743: -9.16362274902, 2744: 8.89478857108, 2745: -3.85408646822, 2746: 2.68937018153, 2747: -6.56609006991, 2748: 4.971528602, 2749: -2.35888152946, 2750: -20.21855919, 2751: -1.31972093877, 2752: -1.31902128243, 2753: -1.14110398042, 2754: 11.6556406181, 2755: -1.1838591135, 2756: -2.15149075088, 2757: -11.3569806514, 2758: -1.0971455945, 2759: -8.79842795092, 2760: -1.12806851018, 2761: -3.67963286888, 2762: 11.3401932113, 2763: -2.54437648045, 2764: -1.05249856613, 2765: -20.4636371012, 2766: 10.0488525066, 2767: -9.70827023037, 2768: 29.2467735297, 2769: -0.596536303564, 2770: -13.1651168599, 2771: -7.94264617945, 2772: -18.3808378176, 2773: 3.14320147204, 2774: 4.2789497502, 2775: -8.60639368283, 2776: -0.465481819274, 2777: -2.51815796862, 2778: 8.50366991567, 2779: 6.30918696444, 2780: -11.7199082695, 2781: -11.4614047239, 2782: -1.38056188783, 2783: 1.18833754385, 2784: 9.19915947092, 2785: -10.8502047205, 2786: 1.33311861022, 2787: 14.9642016877, 2788: -6.14444031832, 2789: -39.3366090191, 2790: -1.74242224214, 2791: -9.16522778959, 2792: 5.50963909808, 2793: 0.281099956482, 2794: -9.39450306714, 2795: -1.10802221409, 2796: 0.867145003492, 2797: -3.60870399512, 2798: -0.101472766851, 2799: 9.58936750363, 2800: -8.85392756842, 2801: 5.41636595802, 2802: -3.30645389949, 2803: 1.60989147369, 2804: 0.570898834257, 2805: 2.35722896545, 2806: 9.37805825783, 2807: -7.69353991117, 2808: 5.70956465252, 2809: 4.15797905004, 2810: -0.797923424546, 2811: -6.18055777818, 2812: 2.13924988085, 2813: -4.76769075706, 2814: 15.6363877502, 2815: -11.1671779359, 2816: 6.41253542453, 2817: -1.74719485566, 2818: 1.30432272495, 2819: -12.3365707176, 2820: -5.68862477744, 2821: -3.71533515286, 2822: 0.0145581855949, 2823: -7.61992035442, 2824: -5.54739884115, 2825: -0.919655645693, 2826: -1.84276138879, 2827: 1.32929562992, 2828: 2.58648518999, 2829: -6.21621821592, 2830: 10.7191362087, 2831: -8.72571408913, 2832: -0.854208859021, 2833: 3.14293668535, 2834: -0.494093779324, 2835: -7.19880803709, 2836: -15.4652609212, 2837: 5.6757750024, 2838: -4.87821393387, 2839: 2.69615174608, 2840: 0.449102950242, 2841: -1.94529824992, 2842: 1.13819509125, 2843: -0.948675402612, 2844: -12.4227727188, 2845: 0.534119641164, 2846: 3.15305664104, 2847: 14.9298895846, 2848: 0.140819738785, 2849: 11.1645282551, 2850: 1.13336139789, 2851: -0.889285934815, 2852: 9.30844738381, 2853: -13.3443362726, 2854: -9.39740847257, 2855: -1.23882091202, 2856: -6.22256843873, 2857: 5.68746343331, 2858: 1.21234486624, 2859: -6.14943367953, 2860: 0.888256010321, 2861: -3.81879227668, 2862: 1.71428067706, 2863: 1.75446395651, 2864: -14.314292536, 2865: 0.79829127816, 2866: 9.67335414407, 2867: 0.151331327739, 2868: -3.25518113433, 2869: -5.89121471622, 2870: -11.0345808623, 2871: -7.1634344467, 2872: -0.377325494493, 2873: 3.07292698474, 2874: 7.43824733991, 2875: 1.11353221013, 2876: 1.33845660063, 2877: -8.64213633032, 2878: 2.75880027669, 2879: 4.53731957789, 2880: -14.5258276959, 2881: 1.14683424839, 2882: 2.3996734874, 2883: 1.71067232005, 2884: 0.600894482233, 2885: 1.62017415915, 2886: 1.35259308815, 2887: 0.658218519996, 2888: 2.2031534424, 2889: 0.536998194019, 2890: 1.67046765358, 2891: -0.830439939722, 2892: 2.31031717249, 2893: -2.86264633011, 2894: 1.7068012638, 2895: -0.6579334364, 2896: 0.733310050912, 2897: 0.524282011817, 2898: 15.9912166893, 2899: 0.210138032364, 2900: 0.810773669775, 2901: 1.49362985586, 2902: 4.18007248932, 2903: 2.53303438371, 2904: 0.773819257479, 2905: -25.8869463461, 2906: -10.8051273847, 2907: 1.25489287945, 2908: 1.00477895721, 2909: -3.47551342216, 2910: -4.93020178081, 2911: 6.16719718781, 2912: -1.65409561242, 2913: -8.72632234153, 2914: 0.571876088986, 2915: 3.9343798104, 2916: 10.5234139283, 2917: -1.39845867148, 2918: 0.150893188714, 2919: 0.307410706871, 2920: 3.8142132382, 2921: -0.512839690935, 2922: 0.0611184281503, 2923: 1.96660689276, 2924: -4.74873215417, 2925: 3.57719890279, 2926: 1.24225607998, 2927: -3.6972716343, 2928: 9.18065700829, 2929: 2.24575201218, 2930: -2.98052147472, 2931: -0.923537921202, 2932: -4.02714237025, 2933: -10.9473162987, 2934: 8.93150286413, 2935: 3.30817765295, 2936: -0.00941171851262, 2937: 0.768886353967, 2938: 0.112628534612, 2939: 0.230975540888, 2940: -7.3492212402, 2941: 0.11179106876, 2942: 0.695804409161, 2943: -13.6155730276, 2944: 0.930402640286, 2945: 0.0450332968029, 2946: 0.215077746805, 2947: 0.237083849026, 2948: -0.940152459158, 2949: 3.22975330641, 2950: 0.638904109163, 2951: 0.167629376519, 2952: 0.153512009538, 2953: -2.79066291842, 2954: -0.225925929157, 2955: -2.54714734422, 2956: -0.142659943514, 2957: 12.3474246227, 2958: 17.1165919674, 2959: -0.522242398901, 2960: -2.18357647944, 2961: -5.91408984729, 2962: 3.46062854859, 2963: -23.8771858572, 2964: 0.137880038685, 2965: 8.84396609666, 2966: 0.149600236427, 2967: 9.3241270229, 2968: -10.2981956786, 2969: 8.55793437806, 2970: -1.79149111534, 2971: 0.170552166381, 2972: 0.301409350017, 2973: -0.132092546602, 2974: 0.0920761573089, 2975: -2.95362734219, 2976: -2.86610395414, 2977: -0.620973055713, 2978: 7.80290183723, 2979: -5.94491901899, 2980: -0.00764391021201, 2981: 3.82242849805, 2982: 1.01741790167, 2983: 2.08143910013, 2984: -3.9124735062, 2985: 0.28879446169, 2986: -11.6151265414, 2987: 0.0452105444615, 2988: 1.59949964451, 2989: -1.86881660892, 2990: 3.97554647009, 2991: 0.277803888942, 2992: 0.764127926215, 2993: 0.181272430603, 2994: 1.95749901827, 2995: 0.224026373274, 2996: 0.170665747641, 2997: -9.8596525554, 2998: 1.06853907343, 2999: -1.60213817579, 3000: 0.258365325278, 3001: 0.276435115861, 3002: 0.766346206243, 3003: 0.035494675053, 3004: 2.07566204099, 3005: -2.19902880166, 3006: 4.23364967972, 3007: 0.668577022307, 3008: -0.651367264961, 3009: 6.69619582851, 3010: -0.867197161139, 3011: -7.50813967013, 3012: 5.13364975617, 3013: -2.20435225095, 3014: -4.05272965684, 3015: 3.98791271666, 3016: 1.28309173602, 3017: 3.05479038773, 3018: -0.0325730265051, 3019: -0.0287965715684, 3020: 0.409328097713, 3021: -0.496572686982, 3022: -0.599559143457, 3023: 0.914926262343, 3024: 0.628871915884, 3025: 0.101671719459, 3026: 1.60656997947, 3027: 1.76721371347, 3028: 7.43264127458, 3029: -3.1828704442, 3030: -11.7003409307, 3031: 0.946783063786, 3032: 1.78295990911, 3033: 4.30780190147, 3034: -4.13490772946, 3035: 0.90991352585, 3036: 2.25582488979, 3037: -0.0508362047181, 3038: 12.9326041785, 3039: -1.61624163978, 3040: -5.44733147227, 3041: -28.2624405166, 3042: 6.67612677954, 3043: -12.9398150896, 3044: -7.53657472443, 3045: 3.49042911199, 3046: 0.998532799983, 3047: -3.42469201455, 3048: 3.29141709564, 3049: 3.13502738234, 3050: 0.735341548953, 3051: 0.635182382808, 3052: -10.3945140012, 3053: 4.56632903988, 3054: 1.37309976051, 3055: -0.180169778487, 3056: 0.838137195299, 3057: -0.984907407095, 3058: -8.40060259417, 3059: 0.430563803498, 3060: 2.26311333094, 3061: -0.145095004995, 3062: -6.0549463908, 3063: 1.31024092147, 3064: -3.3873395147, 3065: -5.59188176095, 3066: 16.8396207269, 3067: 6.49218999666, 3068: 12.5608724435, 3069: 4.38044642354, 3070: -7.78221355947, 3071: 13.0632926461, 3072: 13.0014388074, 3073: 5.28272067383, 3074: -8.79505080251, 3075: -7.81528422221, 3076: 5.11041731822, 3077: -10.295886358, 3078: -2.65809813229, 3079: 2.69886550755, 3080: -3.86552853229, 3081: 4.86438534203, 3082: -1.70278029886, 3083: 3.40784790132, 3084: 13.3173328079, 3085: 3.72105668302, 3086: 3.14740590012, 3087: 0.857480821095, 3088: -1.38462165321, 3089: 8.72988397396, 3090: 5.0795607813, 3091: 0.752294619859, 3092: -9.5728908848, 3093: -5.03305863089, 3094: 9.10888436365, 3095: 1.65805719544, 3096: 4.63245546027, 3097: 16.9582697526, 3098: 0.856644069497, 3099: 9.9640558107, 3100: 0.713844432531, 3101: 0.713003687358, 3102: 1.84970421163, 3103: -7.44843478831, 3104: 1.22928000048, 3105: 0.572687243597, 3106: 1.93263463107, 3107: 1.00351638206, 3108: -3.09381630983, 3109: 1.52750885436, 3110: 0.619221713023, 3111: 3.57196679463, 3112: 0.688843421163, 3113: 2.17320027218, 3114: 8.68898369327, 3115: -3.16340227936, 3116: 6.35870618775, 3117: -15.1673176527, 3118: 2.5138083476, 3119: -0.625365967299, 3120: 8.36354831053, 3121: 0.197817334674, 3122: 0.240320421515, 3123: 4.21539840572, 3124: 14.7477146658, 3125: 0.656534527085, 3126: 3.28298759322, 3127: -6.09171450934, 3128: -19.5879233387, 3129: 5.07320170208, 3130: -0.302756308704, 3131: 0.873114913854, 3132: 2.98715436794, 3133: -1.27902514414, 3134: 11.4319845747, 3135: -6.13685064036, 3136: -12.6022233395, 3137: 1.91270143975, 3138: 26.3987115843, 3139: 3.11630180153, 3140: -2.99615004464, 3141: 6.89695371044, 3142: 0.00185743476266, 3143: -6.77828615541, 3144: -0.890267454014, 3145: 8.01468367333, 3146: 7.84327614033, 3147: 4.24365726411, 3148: 6.45369061782, 3149: 5.60226506714, 3150: -4.89453783436, 3151: 15.8084187518, 3152: 0.619665040362, 3153: 10.9760785856, 3154: 2.39955063237, 3155: 6.84353608303, 3156: -6.54549885635, 3157: -11.4231408206, 3158: -2.05546387395, 3159: -3.85494928939, 3160: -19.4688496604, 3161: -18.7726415965, 3162: -7.95360827934, 3163: -17.9500793322, 3164: -14.2743512993, 3165: -0.187811012693, 3166: 13.3396523674, 3167: 8.92171414678, 3168: -4.19343593452, 3169: -10.2902777628, 3170: 19.4107909287, 3171: 0.305100642214, 3172: -3.51984594451, 3173: -2.74629980229, 3174: 10.6090348087, 3175: -5.15960443095, 3176: -12.8948587595, 3177: 2.19133786737, 3178: -2.79394548021, 3179: -25.8784662638, 3180: -6.91716726741, 3181: 0.483342855118, 3182: 3.24951250399, 3183: -4.51342564003, 3184: -14.4831786068, 3185: -15.8316065106, 3186: 3.67053965803, 3187: 6.49136050455, 3188: 11.6792467526, 3189: -0.450404569065, 3190: -2.22276343133, 3191: -0.561087091167, 3192: -4.04487208898, 3193: 12.1011670339, 3194: -1.257171136, 3195: 0.0627114883596, 3196: -7.75628333392, 3197: 1.33973924939, 3198: 1.5549068543, 3199: -4.29962980612, 3200: 10.482167826, 3201: 9.67489709907, 3202: 7.88545941409, 3203: -5.51112133176, 3204: -15.2399440146, 3205: 0.0528307371318, 3206: 10.6698848559, 3207: 1.80827658095, 3208: -0.292845546068, 3209: -3.88963803729, 3210: -14.9793572125, 3211: 0.411324720192, 3212: 5.48774678296, 3213: 7.32055008732, 3214: -1.04603615627, 3215: -20.4141866636, 3216: 4.12275463108, 3217: -14.2305762388, 3218: -10.0365310256, 3219: 4.86617131385, 3220: -14.0657885243, 3221: 3.79898336246, 3222: 1.24814042896, 3223: 45.6270576679, 3224: -0.432541316385, 3225: -7.19178586049, 3226: 6.06210056467, 3227: -9.58523803464, 3228: 11.5183259689, 3229: -8.29548500797, 3230: 0.62394137427, 3231: 0.168305852321, 3232: 3.42905473539, 3233: -0.854687611006, 3234: 10.8876042061, 3235: 2.95038629888, 3236: -1.0489232724, 3237: 4.85157786334, 3238: 5.35434494474, 3239: -0.300624518752, 3240: -1.16793029296, 3241: 0.399835703922, 3242: 0.0256927040011, 3243: 1.94614635754, 3244: -14.2303151234, 3245: 1.3862474402, 3246: -0.228855894504, 3247: 17.2339533904, 3248: -0.889672962384, 3249: -1.31930856069, 3250: 0.889121021123, 3251: 7.1936959738, 3252: 2.95096682848, 3253: 6.9852006495, 3254: 26.1111260322, 3255: -7.55391311446, 3256: -29.8852658007, 3257: 3.21123162992, 3258: -2.3219451449, 3259: 4.75452455763, 3260: -14.9762564909, 3261: -28.3680604399, 3262: -2.31182268514, 3263: -12.6122917945, 3264: -16.0785047428, 3265: 4.47798097233, 3266: -18.5261468541, 3267: 0.024685405844, 3268: 0.0144287403048, 3269: 8.43974654563, 3270: -20.1231317886, 3271: 0.0394816510962, 3272: -0.533098398377, 3273: 3.39273489593, 3274: -3.73301173439, 3275: 0.860725953615, 3276: 4.01917416961, 3277: 5.83909084759, 3278: 0.7547397271, 3279: 2.3041716849, 3280: 2.54904277721, 3281: -4.65623936178, 3282: -10.2778215375, 3283: -7.69187470008, 3284: -9.31373391852, 3285: 0.755596241137, 3286: 15.4342895485, 3287: 0.0289955475979, 3288: -0.177894577635, 3289: 12.2452109033, 3290: 0.0202695900012, 3291: -0.766790475878, 3292: 0.0708278251733, 3293: -0.295174882412, 3294: -0.0968558492604, 3295: -0.163781041752, 3296: -0.137692993306, 3297: 0.363777015096, 3298: 14.5085852291, 3299: -1.26219529783, 3300: 0.42194633072, 3301: -0.112667342824, 3302: -2.73343390765, 3303: -8.14485504166, 3304: -1.08761797647, 3305: 6.27548366339, 3306: 8.9934198518, 3307: -55.3599657084, 3308: -3.07667017492, 3309: 1.04506524804, 3310: 0.79695795118, 3311: -4.06348692922, 3312: 4.88091743905, 3313: -0.349208648172, 3314: 1.92452168217, 3315: -0.106273993633, 3316: 1.52502018891, 3317: -15.1347666819, 3318: 22.6140092186, 3319: 0.867481012985, 3320: 0.00586398182025, 3321: -0.360094712894, 3322: -0.242841844026, 3323: -0.527042787074, 3324: -7.21645222284, 3325: 9.33596635448, 3326: -1.83675065447, 3327: -4.33895254248, 3328: 2.02322811433, 3329: 9.60547896517, 3330: 6.90427860038, 3331: -1.6573302798, 3332: 3.08665395861, 3333: 9.71388834739, 3334: -0.492334038524, 3335: 8.45048634787, 3336: -8.82783123793, 3337: 2.51087842996, 3338: -2.69854394136, 3339: -5.82817080458, 3340: 0.0365168792996, 3341: -0.841688564667, 3342: 0.000380488224265, 3343: 0.978847656243, 3344: -0.0927273289269, 3345: 0.0257343938222, 3346: -12.8604791969, 3347: -0.515860967156, 3348: 19.2866872796, 3349: 0.0870445705741, 3350: 0.0723401926206, 3351: 1.56137341486, 3352: -0.092901252132, 3353: -0.0894312154475, 3354: 5.95970486547, 3355: 6.67809238163, 3356: -0.0979068595959, 3357: 0.640806472274, 3358: -1.47498435279, 3359: -9.63571262102, 3360: -16.4643250199, 3361: -14.4646565241, 3362: 0.945764494275, 3363: 5.74567033, 3364: -4.59857555129, 3365: 0.535753591456, 3366: 5.17328849281, 3367: -1.19236431329, 3368: -32.4058956246, 3369: 0.385629310186, 3370: -6.38415616849, 3371: 3.73579330011, 3372: 14.3362125328, 3373: -0.857694932738, 3374: 17.6565896021, 3375: -0.644862545773, 3376: -0.855240908396, 3377: -2.3913096291, 3378: -18.77117415, 3379: -30.4288070406, 3380: -1.25610125812, 3381: -0.615587837675, 3382: -1.06586114097, 3383: 2.54304762021, 3384: 7.13119377631, 3385: 4.17579094659, 3386: -14.7305903096, 3387: -15.9478868517, 3388: 1.23074308562, 3389: 8.27015833752, 3390: 29.395552175, 3391: 9.19805253653, 3392: -10.6008886557, 3393: 22.8686349742, 3394: 31.481968918, 3395: -15.0039735863, 3396: -5.13804300152, 3397: 1.25117874544, 3398: 10.1768439947, 3399: -1.32879583044, 3400: -0.656161658741, 3401: -9.91788195391, 3402: 18.4554295216, 3403: 0.0798080549583, 3404: -2.57268411405, 3405: -0.926326809344, 3406: 0.274347273517, 3407: -26.3302664183, 3408: 12.0158534509, 3409: -0.65508129461, 3410: -1.65586875548, 3411: -3.91998071937, 3412: 3.1400500912, 3413: 7.16332523986, 3414: 8.00684205158, 3415: -31.6785059435, 3416: 14.8878830776, 3417: 10.0795689843, 3418: -6.13671843927, 3419: -4.97631168531, 3420: -7.44304686219, 3421: -15.3114094699, 3422: 3.20103856781, 3423: -12.0711228488, 3424: -5.22434700659, 3425: -9.48444892109, 3426: -21.1120589503, 3427: -2.97407186823, 3428: -0.28412316389, 3429: 1.34889070573, 3430: 0.695887062872, 3431: -3.79999658424, 3432: -4.39921897124, 3433: 0.452643372058, 3434: 2.16581270109, 3435: -2.1631795616, 3436: -10.4853300058, 3437: 21.3893137935, 3438: -1.38035344055, 3439: 8.10709740501, 3440: -6.00866704577, 3441: -7.85810116655, 3442: 10.4611743665, 3443: -4.09866567711, 3444: 2.23247559788, 3445: 18.2715456934, 3446: -5.23218115089, 3447: -1.19643204809, 3448: 17.1813119576, 3449: -0.608694770975, 3450: -0.687920337301, 3451: 2.2398940133, 3452: -21.463410871, 3453: -0.84407777747, 3454: -1.11702884465, 3455: 10.2418422274, 3456: -0.653395705997, 3457: -4.95822200756, 3458: -0.976435686675, 3459: -0.958352397395, 3460: -2.41913806079, 3461: -1.04761210464, 3462: -0.352190820117, 3463: 7.65619870739, 3464: 4.65311170351, 3465: 7.76084807565, 3466: -38.3370574987, 3467: 20.1064223863, 3468: 17.3330860445, 3469: 26.0782784214, 3470: -12.3318994544, 3471: 11.442597519, 3472: -0.291634948184, 3473: 10.8394276013, 3474: 2.2397134744, 3475: -9.72277073775, 3476: 1.48811258426, 3477: -1.54899687703, 3478: 16.7400663058, 3479: -9.51236445689, 3480: -0.790236099957, 3481: -13.8983096925, 3482: -0.851057825671, 3483: 8.60232602738, 3484: 8.15315725642, 3485: 6.05643066354, 3486: 0.416986310227, 3487: 27.2130794512, 3488: -9.61898088223, 3489: 0.288063535388, 3490: 15.0902490691, 3491: -1.69710606003, 3492: 1.06073778289, 3493: -0.897447586653, 3494: -0.851602156214, 3495: 0.901132767772, 3496: -0.770016169854, 3497: 0.722921990373, 3498: -0.916502567659, 3499: -0.809473339976, 3500: -0.891578988137, 3501: -0.904442154454}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.0639943168929, 2: 7.23352258258, 3: -0.0122012907123, 4: -3.07688991828, 5: 0.838993891704, 6: -1.79714900332, 7: -1.26569649064, 8: 1.24045532241, 9: -0.702298691675, 10: -0.187867161815, 11: -3.76080693675, 12: 3.11791125451, 13: -0.0484285556186, 14: -1.41518040099, 15: -6.92143029191, 16: 2.45590322105, 17: -7.18359194329, 18: -0.0781199228023, 19: 1.80984361535, 20: 4.1906607614, 21: -3.05806386979, 22: -4.55280140688, 23: 0.112914610812, 24: -2.96237697351, 25: -0.337244527946, 26: -0.11769538299, 27: 3.34620419468, 28: 1.26016766703, 29: 3.45619288474, 30: 4.8072608901, 31: 3.42965642776, 32: -2.38013704788, 33: -0.205740328122, 34: -0.794529920241, 35: -0.285575124099, 36: -1.6083166997, 37: 2.12648058058, 38: 3.28530081713, 39: -0.16146975767, 40: 1.71052909936, 41: -5.26246763497, 42: 4.01174107215, 43: -0.907474671265, 44: 0.5711019565, 45: 11.8342286153, 46: 2.05907854563, 47: -2.10085956956, 48: 0.465756088592, 49: 4.07680369855, 50: 2.3428577889, 51: -0.051128666183, 52: -0.841385652081, 53: 1.50177987392, 54: -0.680548958553, 55: -1.40292946416, 56: 0.704113049287, 57: 6.66287115571, 58: 1.17722965902, 59: 0.980481056845, 60: -0.195312505165, 61: 1.24024768507, 62: -1.11543633079, 63: 0.290060269287, 64: -1.62898864639, 65: -0.280919763273, 66: 1.03193397412, 67: -2.05863601173, 68: 2.50147837987, 69: -3.20966839199, 70: 0.340365345591, 71: -0.0921666619865, 72: -4.84576685789, 73: 0.257258161114, 74: -15.3749875376, 75: -1.47346706071, 76: 0.303458178045, 77: -1.04881298562, 78: -4.01757352694, 79: -18.3144836422, 80: 5.40670233459, 81: -0.512962358724, 82: -10.2377398606, 83: 1.0778887178, 84: 2.20898182861, 85: 0.265054399006, 86: -2.74726025374, 87: 0.289799313719, 88: -1.64282177938, 89: -0.297359198408, 90: -1.14932304533, 91: -0.0313475537079, 92: 1.23576255246, 93: -7.36881892698, 94: -0.428431297398, 95: 0.994455133556, 96: 0.108396265827, 97: 1.55973584246, 98: 1.14083856946, 99: 1.20903317025, 100: 0.495659666729, 101: -17.9006047324, 102: -0.144712025868, 103: 0.107588731499, 104: 1.01597074386, 105: 0.894100921624, 106: 0.64194616096, 107: 2.52614273885, 108: 0.808090089459, 109: 0.323252824429, 110: -1.71704789644, 111: -0.731393805036, 112: -1.67939951139, 113: -3.7021348535, 114: -1.13501960079, 115: -5.51348758089, 116: -0.578749129983, 117: -2.69904861573, 118: 2.20333510983, 119: 1.24024107922, 120: 1.4519397269, 121: 1.45129165793, 122: -0.903200760916, 123: 5.23446080429, 124: 0.256490580253, 125: -4.64744888314, 126: 0.230806152621, 127: 0.174253467066, 128: -2.07361656109, 129: -0.911475530114, 130: 0.0109259562225, 131: 0.904833514037, 132: 0.421039060722, 133: -5.59481302869, 134: -0.153938164842, 135: -1.29140853995, 136: 3.35369242159, 137: 0.36782660577, 138: -2.43263693953, 139: 2.40148783991, 140: -2.01357846316, 141: 2.44482179091, 142: -3.25477200657, 143: -0.367458190166, 144: 0.44149764466, 145: 4.75938270743, 146: 0.176830390683, 147: 0.0891319487164, 148: -1.4762556508, 149: 0.267404676109, 150: 0.993653710479, 151: -2.98505504133, 152: 0.178044263539, 153: 0.129467904594, 154: 0.07222840432, 155: 0.104096585569, 156: 0.300846170889, 157: -10.4334663496, 158: 0.365515538099, 159: 0.291156670099, 160: 0.199738170447, 161: 2.95437712765, 162: 0.523984578648, 163: 6.2890538441, 164: 0.347508899834, 165: 7.49102982141, 166: -2.20144073796, 167: 0.575065148898, 168: -3.76318607269, 169: -2.48846846038, 170: 0.446779754855, 171: 0.469270707261, 172: 0.0611711068541, 173: 7.2655184942, 174: 0.179630139171, 175: -0.915273151038, 176: 4.68282435398, 177: -0.459378985443, 178: -0.136162842678, 179: 0.176033302643, 180: 0.323180228945, 181: 0.0469599160953, 182: 0.169114402127, 183: -1.9073541133, 184: -6.66931920773, 185: -3.79131405373, 186: -1.27456518523, 187: 1.37563555415, 188: 2.49808088396, 189: 1.3347723533, 190: 1.56571773384, 191: -0.283573602002, 192: -3.3290930598, 193: 0.21862479987, 194: -3.98649170738, 195: 1.48764746732, 196: 0.275723106742, 197: 0.995127776573, 198: -0.734526870433, 199: 0.17242336375, 200: 0.904126908803, 201: 0.299347634405, 202: -0.232438618298, 203: 0.186768655951, 204: 0.230296290558, 205: -7.66673015784, 206: 1.10200223807, 207: 6.59017972947, 208: 0.185728790138, 209: 0.255768499957, 210: 0.0971952754435, 211: 0.149875938244, 212: -1.10032890756, 213: 0.797061012834, 214: -0.973670330983, 215: 0.36885388412, 216: 0.0651260473285, 217: 4.05523184546, 218: 1.53968696965, 219: -5.80909923929, 220: 0.911053130999, 221: 2.75058419264, 222: 2.72997173075, 223: 1.27090274573, 224: 0.455316930872, 225: -0.689909461442, 226: -0.0584469006803, 227: 5.41564999536, 228: 1.84271469267, 229: 1.6681729122, 230: -0.505705929938, 231: 2.74307885255, 232: 0.59786337471, 233: 1.4515851462, 234: 0.262771091305, 235: 0.454831995241, 236: 1.47365055698, 237: 1.30927095047, 238: -8.93587296544, 239: 2.05630777496, 240: 0.849357754908, 241: 2.63610180433, 242: 5.02111031912, 243: 0.486047575497, 244: -4.0972336149, 245: 2.09474869393, 246: -2.49478631103, 247: 3.05677534654, 248: -1.54919911054, 249: -5.44490141333, 250: -2.75065806195, 251: 3.05935065023, 252: 0.415028524004, 253: -1.99638655798, 254: -6.44674419227, 255: 0.0446518312545, 256: 0.740981176493, 257: -0.236103946072, 258: 0.534670777049, 259: 0.98358231443, 260: -3.39450045256, 261: 1.85479059353, 262: 0.172771639637, 263: 1.3443508224, 264: 0.183377832067, 265: 0.308495030353, 266: -8.87854742683, 267: 4.32009247005, 268: -2.38238158582, 269: -3.09732122034, 270: 0.212896415042, 271: -4.88247967022, 272: 1.2126156458, 273: -1.33998968286, 274: -2.27968498493, 275: 5.02610158059, 276: -4.40008232228, 277: -1.47397995318, 278: -3.30346262075, 279: 3.71427887904, 280: -1.10562865956, 281: 0.346557892095, 282: -2.88802227568, 283: -4.89592800392, 284: -1.84994335783, 285: -0.0675020461866, 286: 0.475073014865, 287: 1.04557666445, 288: 0.631063988136, 289: 1.57645284121, 290: -1.1548363338, 291: 0.154967662088, 292: -0.834324706352, 293: -4.90442599897, 294: 0.473318836196, 295: 0.997142102776, 296: 3.57534059649, 297: -0.0424825434556, 298: 1.65069458257, 299: 0.300091019835, 300: -1.34863755974, 301: 1.89216302665, 302: -0.157493020949, 303: 1.54896651191, 304: 2.41657949591, 305: -5.74380557499, 306: 0.90125526694, 307: -7.9158291383, 308: 0.958133594584, 309: 0.726720343366, 310: 0.126612589438, 311: -2.17337476806, 312: 1.66554383015, 313: 1.73281059747, 314: 15.5137775555, 315: 1.07425747845, 316: -1.82221863429, 317: 1.05767971946, 318: 0.95696443449, 319: -4.11000640874, 320: 1.26942216114, 321: -1.10182387543, 322: -4.53418552946, 323: -1.64541270693, 324: -2.66104000538, 325: 6.26215918866, 326: 0.400009773552, 327: -2.20111722757, 328: 4.98792006326, 329: 3.54053306468, 330: -1.37452765778, 331: 1.97732631796, 332: -0.77837157223, 333: 0.71536947799, 334: -1.96469634802, 335: 1.90141741381, 336: -5.56791067462, 337: 0.820041737692, 338: 8.91273867109, 339: 1.03386212443, 340: 0.642810702512, 341: -0.552728740435, 342: -3.23992050932, 343: 5.4855515578, 344: 5.14263822737, 345: -8.25757961415, 346: -5.01918894475, 347: -0.58431950451, 348: 3.62577345315, 349: -2.84826087941, 350: 0.29348303494, 351: 5.68295027897, 352: -0.195774632691, 353: -9.71105420567, 354: -0.401335580231, 355: -2.16751175703, 356: -1.23094674028, 357: 3.24595703639, 358: -2.05213018188, 359: -1.73285789136, 360: -6.54587843349, 361: 0.758834481429, 362: 2.45760171302, 363: -0.0522045639363, 364: 0.438510119776, 365: -1.02408335198, 366: -5.98295596258, 367: 2.52138340875, 368: -3.33493841401, 369: -0.719846922224, 370: -7.10408381986, 371: -1.30763577579, 372: -5.84670630689, 373: -0.23933094898, 374: 0.385821062882, 375: -5.95819882497, 376: 3.25292039227, 377: 0.676083508758, 378: -3.90999511253, 379: 2.39006773808, 380: 1.20904095392, 381: 0.252164591109, 382: 2.63899933265, 383: -0.77256989306, 384: -0.293278401141, 385: -1.02522795173, 386: 1.40999136783, 387: 1.29318934126, 388: -2.15900640825, 389: 0.600436502497, 390: -1.76915173485, 391: -1.75490964066, 392: -5.42594324857, 393: -6.21760647379, 394: 0.746997630665, 395: 2.34848512208, 396: -0.339535625968, 397: 1.44002358507, 398: 1.43435793005, 399: 0.171612820128, 400: -4.64719630611, 401: -2.04813083894, 402: 0.917973146887, 403: 1.86115134577, 404: -2.58072211577, 405: 0.756644222003, 406: -0.226467246149, 407: 0.239002138504, 408: 3.40445504167, 409: 0.593878804689, 410: 1.84610407933, 411: 2.13237829823, 412: -2.49439673019, 413: -3.58757804445, 414: 1.24217411073, 415: 1.10659487929, 416: -2.94901650839, 417: 0.645664975572, 418: -0.04318025008, 419: 1.06279081316, 420: 1.06581811228, 421: 1.07651690966, 422: 1.49006388646, 423: -1.72647380473, 424: -2.63187218967, 425: 5.10635123984, 426: -0.294665267147, 427: 1.3650607051, 428: 1.84434677267, 429: -3.31937240407, 430: 1.19892515206, 431: 2.4660703519, 432: 1.53738484066, 433: -0.466744560872, 434: 2.87431417531, 435: 1.57404299794, 436: -0.856600957336, 437: 4.32540370074, 438: 0.687624971944, 439: 0.751874566414, 440: 0.918959448783, 441: 1.28962064204, 442: -0.730628143623, 443: 1.16361036029, 444: 1.16090618159, 445: 1.64585391509, 446: 1.93126891404, 447: 0.805400140407, 448: 0.792744980456, 449: 0.515393253567, 450: 2.16004692429, 451: 1.50425993208, 452: 3.89022176152, 453: 0.965300435535, 454: 1.55139447634, 455: -1.86895888392, 456: -0.220627117486, 457: 1.53439971751, 458: 0.450495256513, 459: -2.25262228675, 460: 1.01173452007, 461: 0.870876565414, 462: -4.75683982854, 463: -9.80655543056, 464: -0.839583719329, 465: -2.84929890668, 466: -0.0651970380575, 467: -0.371432815558, 468: -4.62020768537, 469: 7.74135445575, 470: 5.13393499468, 471: -10.6925446238, 472: 0.722048177059, 473: -5.36149012586, 474: 0.221510266427, 475: 0.353777355506, 476: 0.274627451179, 477: 0.840846223054, 478: 2.36101258721, 479: 0.223406661063, 480: 0.915419100037, 481: 0.31340055852, 482: -1.21442866434, 483: 0.795869270898, 484: 0.0999789669109, 485: -3.68776786472, 486: 0.375973790572, 487: 4.39254552496, 488: -0.112228148777, 489: 0.0726001547216, 490: -4.04578588587, 491: -1.65407338078, 492: 1.85767771301, 493: 0.33433584862, 494: 4.71880237914, 495: 0.203159350382, 496: 0.199975978418, 497: -0.190305116971, 498: 0.11710444862, 499: 0.661975316807, 500: 0.783157531702, 501: 0.548858632179, 502: 0.306020952922, 503: 0.213478337219, 504: 0.190169002923, 505: 0.907215982489, 506: 0.410833938018, 507: -0.0365039627482, 508: 0.488162473103, 509: 0.583369786675, 510: 0.962486164239, 511: 0.327507132955, 512: 1.49151270204, 513: -0.938906879686, 514: 15.0219135915, 515: 5.11083365404, 516: 0.989064221803, 517: -0.586852300122, 518: -1.66264303701, 519: 1.55447914052, 520: -6.70430927346, 521: 0.24332558675, 522: 0.888402162105, 523: 0.121276025568, 524: -3.99200327735, 525: -2.86894083034, 526: -6.32321819046, 527: 0.698601091042, 528: 0.158239803551, 529: 0.376017530176, 530: 0.32271581042, 531: -0.0646602358504, 532: -0.282597626066, 533: -3.03194258233, 534: -1.39517844748, 535: -2.80604397789, 536: 3.12929583823, 537: -1.38489303109, 538: -2.41144240683, 539: 1.41703543652, 540: -1.83338883988, 541: 0.970826024719, 542: 0.211472432798, 543: 2.05304542284, 544: 1.10284778119, 545: 0.0433038357139, 546: 0.226188859515, 547: -3.95327363424, 548: 0.146611189219, 549: 1.23145169988, 550: 0.167231234391, 551: 0.308523279079, 552: 0.344337348982, 553: 0.361669222566, 554: 3.39758110277, 555: 1.29542513451, 556: -8.438601288, 557: 0.272713446776, 558: 0.367301136247, 559: 0.212468845245, 560: 0.262550240863, 561: 1.11438570662, 562: -1.35411069347, 563: -5.7737248478, 564: -0.174663180902, 565: 0.495134723522, 566: -1.51282400543, 567: 0.119213855808, 568: -4.07839932611, 569: -2.5110327955, 570: -0.388599258759, 571: 0.865249631156, 572: 0.280536548575, 573: 0.998045469858, 574: 0.37011075202, 575: 0.554682160254, 576: 1.61609052146, 577: 0.00415800235188, 578: 0.410368854561, 579: 1.92506137616, 580: -4.99057866585, 581: 0.142152660301, 582: 5.59758943255, 583: -2.32562206649, 584: 1.2544959354, 585: 1.89941144188, 586: 2.09985794354, 587: 4.17750834009, 588: 2.37264617282, 589: 1.10760940139, 590: -1.96327046815, 591: -0.023068644869, 592: 0.585707073854, 593: -0.999961087419, 594: 1.66394944038, 595: 1.6667949936, 596: -4.6236009458, 597: -4.13955167088, 598: 1.5666509249, 599: 3.61657216875, 600: -4.13996227063, 601: -4.66069523529, 602: 0.991965538584, 603: -4.16303458434, 604: 1.5945253256, 605: 0.0561620696596, 606: 3.17000374764, 607: 1.63945074383, 608: 0.829987243478, 609: 0.657595372059, 610: -3.40350921738, 611: 1.04985013032, 612: 2.15577718381, 613: 1.65166288848, 614: 0.963952712786, 615: 3.88668636711, 616: -1.61849486309, 617: -2.08955957718, 618: -5.77658908713, 619: 1.3057906694, 620: 1.57631828376, 621: -1.11734113334, 622: 3.36872640564, 623: 0.0675153301258, 624: -1.30384780333, 625: 0.0801477094554, 626: 0.714511312117, 627: 1.58971672562, 628: 1.55207399658, 629: -7.04144510048, 630: -0.988941385043, 631: 1.16879867182, 632: -6.38090359559, 633: 0.109933275164, 634: 2.4511215208, 635: 4.63071970643, 636: 0.700382081717, 637: -0.892499278035, 638: 0.501658848962, 639: 1.35576055383, 640: -1.95838827911, 641: -3.78804089419, 642: -0.0414003546022, 643: -0.817999544301, 644: 5.20980689739, 645: 2.2460604121, 646: 1.56338114529, 647: 2.94755929749, 648: -0.306904304675, 649: 0.585754792616, 650: 0.578057826471, 651: -6.28303561192, 652: 2.1266502772, 653: -5.413545807, 654: -4.5075267533, 655: 1.08507132255, 656: -4.92725847403, 657: 1.35451613722, 658: 1.33912642204, 659: 1.57043272472, 660: -2.23932287674, 661: 0.937393422326, 662: -0.186162230243, 663: 5.22717241473, 664: 1.2129593722, 665: -11.5264514268, 666: 1.53259364866, 667: 1.32786375572, 668: -2.90525467645, 669: 1.37713752258, 670: 0.89598268921, 671: -2.91322605556, 672: 4.24437539682, 673: 0.853568280972, 674: 0.709912532748, 675: -2.38807790842, 676: -0.8348783904, 677: -8.26744354297, 678: 6.09444008656, 679: -2.81024473945, 680: -4.74536121321, 681: -7.72681033415, 682: 0.609785331707, 683: 0.719371510356, 684: -1.8696196233, 685: 2.67423641897, 686: 5.44173926079, 687: 4.54269675531, 688: 1.25238887987, 689: -5.87089338688, 690: 1.32168260826, 691: -0.710965176357, 692: -2.96751718134, 693: 2.59601881175, 694: -0.280804363881, 695: -8.19351404698, 696: 0.452411783986, 697: 1.32429470287, 698: -0.737029288816, 699: 0.224190696697, 700: 4.72339704013, 701: -0.180542593264, 702: -1.09083271934, 703: -1.91678129535, 704: 0.931745389769, 705: -0.0973579626142, 706: -0.732270231809, 707: -1.20113782727, 708: 2.50842180014, 709: 2.41626410621, 710: 1.51807325503, 711: 0.20706946538, 712: -0.41007177852, 713: 4.23874630345, 714: -0.826848205681, 715: -15.6292539777, 716: 1.3769851151, 717: -3.69883136395, 718: -3.25740782033, 719: -0.752371708533, 720: -0.808893488574, 721: 1.60653017944, 722: -0.12585755669, 723: 2.30418125508, 724: 0.139570299905, 725: 1.28440001408, 726: 0.635575411171, 727: -6.41691569258, 728: 1.47250715343, 729: 2.135814734, 730: 4.38551990278, 731: -0.470772733584, 732: 0.324789096183, 733: -1.84465122756, 734: 0.311111755955, 735: 2.45914825874, 736: 0.284342100009, 737: 0.44298694856, 738: 1.2447207105, 739: -2.71416652341, 740: 0.960250763963, 741: -5.65530609189, 742: -8.54395668708, 743: 0.136471259039, 744: 2.7382446802, 745: 1.18234515024, 746: 1.52242929187, 747: 1.83159350142, 748: 1.81036787412, 749: 1.15153099865, 750: 4.65573953153, 751: -0.962515536835, 752: 2.33188096127, 753: -2.37342638624, 754: 0.0141288992134, 755: 3.92941681054, 756: -1.69451664963, 757: 7.63010957805, 758: 2.40275430009, 759: -5.32632248115, 760: 1.6440852054, 761: -6.61732849832, 762: -2.60438462505, 763: 1.91683490045, 764: -0.533982650236, 765: -2.714926441, 766: 0.558901967545, 767: 3.16491729365, 768: 0.644306798795, 769: 0.714030905712, 770: -2.35487942443, 771: 0.260519412291, 772: 3.12349034355, 773: -1.63180171923, 774: -0.145421895218, 775: 0.714425060083, 776: 1.65327868472, 777: -5.86896690665, 778: -3.36752150776, 779: -1.49855622989, 780: -4.23559027485, 781: 0.695333032291, 782: -0.0573668262223, 783: 1.28558699657, 784: -1.61798328413, 785: 1.28084072331, 786: 3.33387472018, 787: 0.346212902086, 788: 1.33038594604, 789: 0.426311888172, 790: 1.29100915084, 791: -9.84103211189, 792: 0.294073243918, 793: 0.146014918148, 794: 0.421970697618, 795: 0.443738569102, 796: 0.907046961858, 797: 2.73101413311, 798: 1.0195766372, 799: -2.06448022627, 800: 0.21729610241, 801: -0.946005769849, 802: 0.255163411939, 803: 1.27754865008, 804: -1.1936239107, 805: 0.39653853974, 806: -0.301537536941, 807: 1.60062891375, 808: -1.72196662251, 809: -0.474567757302, 810: -1.42150373147, 811: 4.39092317907, 812: -1.43762776113, 813: 2.6058126819, 814: -0.9190610278, 815: -0.0680522391595, 816: -2.86754445246, 817: -14.3076943546, 818: -4.29099456148, 819: 2.29925372561, 820: -3.4827327318, 821: -2.25828947088, 822: -1.04789884585, 823: 1.1412595429, 824: 0.213192613114, 825: 0.282427467246, 826: -0.00903101301928, 827: -1.0515390829, 828: 0.108585762437, 829: -0.368818540235, 830: 0.159067875599, 831: -1.95253744967, 832: 0.588574545606, 833: 0.710985378547, 834: -4.06869827973, 835: 0.323627159388, 836: 3.87676061966, 837: 0.245475299816, 838: -0.146923325688, 839: -1.41236337528, 840: 3.90328831252, 841: 2.73287177281, 842: 0.57092613499, 843: 5.17167079563, 844: 0.14615752192, 845: 0.161159119527, 846: -3.95232879152, 847: 0.157664085839, 848: 0.431385512384, 849: -2.68440934208, 850: 0.198197490134, 851: 0.0937915850807, 852: 0.0693079536295, 853: 0.0856696554944, 854: -1.49507444749, 855: -2.9588340798, 856: 1.71946517783, 857: -0.0323477689563, 858: -0.140583870788, 859: 0.531274797012, 860: 0.875536542488, 861: -6.64642478092, 862: 1.25068905208, 863: 3.92180535625, 864: -6.6334925551, 865: -0.527901569694, 866: 3.70438690765, 867: 2.01713655535, 868: -0.820014714824, 869: 0.692694488125, 870: 0.130639604562, 871: -3.84606200165, 872: 0.130991881911, 873: 0.644376191639, 874: 5.14926113935, 875: 0.268492818671, 876: -0.345112616824, 877: 0.124893447479, 878: -0.0157472396085, 879: -0.118713039711, 880: 0.165847138216, 881: 0.00487310503289, 882: -5.25351902386, 883: 0.482112220105, 884: 1.62174457419, 885: -4.1825173078, 886: 3.34831574028, 887: -2.93134926789, 888: -2.09403715546, 889: -2.11263680546, 890: 0.201169313087, 891: 0.0825296406865, 892: -3.47068365359, 893: -0.646405943179, 894: 0.375618514735, 895: 1.62205076966, 896: -0.735736730483, 897: 0.0491520140804, 898: 0.408120443813, 899: 0.0262024338636, 900: 4.05813680043, 901: 0.172437675676, 902: 0.105356250898, 903: -0.771482995046, 904: 0.409252833531, 905: -3.49718397926, 906: 0.172875111247, 907: 0.221481849646, 908: 0.433756570835, 909: 0.0248957741742, 910: 0.179435863574, 911: -4.31192964669, 912: 2.33311342981, 913: 0.347587379451, 914: 0.135757638661, 915: 0.602366581611, 916: -1.71288409269, 917: 9.71778931992, 918: -2.44523935797, 919: 2.14391192908, 920: 0.365488990489, 921: 0.333349556435, 922: 0.0258455811323, 923: 0.774428196836, 924: 0.0900515491057, 925: -0.0982276348762, 926: 1.55750497449, 927: 2.28118850933, 928: -1.19862021027, 929: 0.236033136273, 930: 0.277884163243, 931: -3.01011100001, 932: -4.54026717335, 933: -0.307497074707, 934: 2.78881154798, 935: -0.0907448207437, 936: -1.24544472952, 937: 0.183110930068, 938: -0.287656233454, 939: -5.09797014583, 940: 2.40869722259, 941: 3.6075036676, 942: 2.49618040135, 943: 1.99525921663, 944: 4.26423364191, 945: -5.80511881611, 946: -5.76188642918, 947: -1.29316724843, 948: 7.75408321634, 949: -6.98435748104, 950: -2.3363909042, 951: -5.33250096743, 952: -1.93860672205, 953: -0.382565019174, 954: 1.76286453985, 955: 4.41260002505, 956: 0.305742706033, 957: 0.392078008658, 958: -2.49451043767, 959: 6.94101393274, 960: 1.55045663439, 961: -0.977316357266, 962: 0.599071703871, 963: -1.35859420942, 964: -0.355355966959, 965: 1.22442442189, 966: -0.701539577274, 967: -1.5876219574, 968: 0.171487757589, 969: 1.08631236606, 970: 3.44620678064, 971: 3.47555620889, 972: 0.415689701138, 973: -2.06134643601, 974: -3.0081342094, 975: 3.33366360741, 976: 1.01831266905, 977: -1.18842772989, 978: -1.10712095964, 979: -0.0575263033119, 980: 4.29773095845, 981: -3.13727738561, 982: -3.12055571329, 983: -7.43903247964, 984: 6.48857590662, 985: 1.0475064722, 986: 0.378397287153, 987: 2.87999802963, 988: -1.61116055638, 989: 2.53173427103, 990: 0.955615952775, 991: 1.32405428118, 992: -1.27552980198, 993: 6.14974298834, 994: 2.32979629757, 995: -2.46462695402, 996: -1.94930479629, 997: -0.894217906374, 998: -6.18580511642, 999: 0.0476247809, 1000: 2.87292921825, 1001: -1.34398491724, 1002: -2.59731771659, 1003: -1.06898901595, 1004: 0.0470345103097, 1005: -5.54695963141, 1006: 0.450639760929, 1007: 0.37483839957, 1008: -0.603883466493, 1009: -1.6190791576, 1010: 0.518452546374, 1011: 1.45158866025, 1012: 1.14139559539, 1013: 0.397802076634, 1014: -9.9416598697, 1015: 0.771603474464, 1016: 0.191759790691, 1017: 4.09399156571, 1018: 0.599347225775, 1019: 0.283186067334, 1020: -2.59179129335, 1021: 3.4671853029, 1022: -3.95334012542, 1023: 2.18435431607, 1024: -6.84607871724, 1025: 0.404369654974, 1026: -3.24035759206, 1027: -9.26452601607, 1028: -4.83822555622, 1029: -1.52602867929, 1030: -4.41772999875, 1031: -4.08487108231, 1032: -1.10877488179, 1033: -0.884358574431, 1034: 6.13283847485, 1035: -5.14954491779, 1036: 2.23128206828, 1037: 0.585913500882, 1038: 0.0524676631826, 1039: 0.973906782005, 1040: 0.954463977762, 1041: 4.64492450748, 1042: 5.45524168602, 1043: -0.758044593003, 1044: -7.25339695046, 1045: -7.85577459548, 1046: -1.01767833906, 1047: -2.01761972149, 1048: -2.131412289, 1049: -82.3892065987, 1050: -47.5293625844, 1051: 1.07130230336, 1052: 54.7268864685, 1053: 66.7339687171, 1054: 95.0083818456, 1055: 29.0180429139, 1056: 75.678766229, 1057: -74.012192908, 1058: -15.2465662302, 1059: 34.9843925429, 1060: 6.29568009553, 1061: -32.0700034903, 1062: 40.8995338398, 1063: 6.09995646705, 1064: 87.3925665646, 1065: 58.943176119, 1066: 10.7507361371, 1067: 4.51930852666, 1068: -18.3174786142, 1069: -51.4039154798, 1070: 27.6268653266, 1071: -52.5047578856, 1072: -55.123367439, 1073: -13.9663540621, 1074: -32.5008215922, 1075: 5.44997711967, 1076: 50.1183636741, 1077: 21.6938173778, 1078: 84.5330504721, 1079: -33.5431037233, 1080: -12.5146759167, 1081: -3.55370852394, 1082: 55.8517091543, 1083: -11.219084069, 1084: -34.1433014697, 1085: 85.0829256752, 1086: 49.6518867552, 1087: 69.419872581, 1088: -79.6603360325, 1089: 5.00963387795, 1090: 13.730825553, 1091: 60.9816617065, 1092: 66.1237352768, 1093: -41.4323578931, 1094: 12.5163659517, 1095: 35.7446480347, 1096: 86.5402062797, 1097: 7.36271506293, 1098: 23.8649836155, 1099: 18.1730816118, 1100: -9.97496424449, 1101: 39.551739538, 1102: 2.59946360194, 1103: -8.9518221955, 1104: 17.7191343583, 1105: -11.3914124336, 1106: -73.4503288782, 1107: -61.7241281668, 1108: -20.1347365636, 1109: -29.7450515431, 1110: 65.4835254474, 1111: -1.03566059774, 1112: -14.5010612785, 1113: -7.26954066923, 1114: 0.970970996624, 1115: -5.28160292921, 1116: 42.6063414444, 1117: -12.3626514902, 1118: -6.17711064244, 1119: -17.7641974998, 1120: -6.83743859717, 1121: 35.3474007649, 1122: -29.4531641502, 1123: -5.59140611736, 1124: 1.84014582796, 1125: -3.35885901325, 1126: -82.7863254722, 1127: 7.04284015442, 1128: 11.5505122925, 1129: 2.34434745539, 1130: -7.68021588966, 1131: -3.54965358797, 1132: 100.0, 1133: -1.66915657769, 1134: -4.6021885293, 1135: -24.6768461006, 1136: -9.68880817725, 1137: -5.89729809182, 1138: -8.02216159579, 1139: -17.7131734755, 1140: 40.8418087791, 1141: -9.08034566082, 1142: -7.92811017033, 1143: -8.79527551084, 1144: -7.10098374662, 1145: -7.82304431219, 1146: -16.1896530095, 1147: -18.6206279419, 1148: -32.4684141486, 1149: -8.90023260304, 1150: -9.26439795209, 1151: -9.01579813406, 1152: 4.93024684854, 1153: -10.699095922, 1154: -8.06898518342, 1155: -13.1214508972, 1156: -11.6055553407, 1157: 31.1772834351, 1158: 4.50534937619, 1159: -63.6799850996, 1160: 3.33008949969, 1161: -13.9831592725, 1162: 99.9820388945, 1163: 63.2301364683, 1164: 34.1229022278, 1165: -0.50865912177, 1166: -49.3141615513, 1167: -31.5081729996, 1168: -12.7113702623, 1169: 2.68687470302, 1170: 81.4306336876, 1171: -10.4614422651, 1172: 38.3224795832, 1173: -2.88543726146, 1174: -1.5945812699, 1175: -0.639885358433, 1176: -82.4851573337, 1177: -2.62053871856, 1178: -9.74879213086, 1179: -13.1941592966, 1180: 26.8488974047, 1181: -16.2970127345, 1182: 15.0164822507, 1183: 75.3427845501, 1184: -4.42758087416, 1185: -7.80954810202, 1186: -5.53422156463, 1187: -6.21030888951, 1188: -55.7363971608, 1189: -19.4291901684, 1190: 2.17020325378, 1191: -5.86707652068, 1192: 54.876764527, 1193: 0.0799464375506, 1194: -3.61067152518, 1195: 3.77512921725, 1196: -2.50656264765, 1197: -8.4854389907, 1198: 5.7057207976, 1199: -2.27057305974, 1200: -1.85981183526, 1201: -2.88960047572, 1202: -2.65838569203, 1203: -6.97889966927, 1204: -53.4960404854, 1205: 3.8900547964, 1206: -4.33700851606, 1207: -5.11348474592, 1208: -5.70830467844, 1209: -2.42135344575, 1210: -50.1103142693, 1211: 12.7909021499, 1212: -46.3833049027, 1213: -48.5609856159, 1214: -2.82583426414, 1215: -2.78410148583, 1216: 53.9263251324, 1217: 25.4362129681, 1218: 81.8112878252, 1219: -1.95769852466, 1220: -33.5428742985, 1221: -2.23850475683, 1222: 46.6489495563, 1223: 79.576720253, 1224: -65.3272539044, 1225: -5.64454346181, 1226: -2.72529893011, 1227: -1.27687209171, 1228: 0.181562329952, 1229: -2.7346965052, 1230: -20.2843707907, 1231: 40.5916746319, 1232: 0.221132204638, 1233: 48.3594624342, 1234: 56.7654340346, 1235: 14.3962886334, 1236: 19.9272937338, 1237: -2.09911327442, 1238: -4.59086680454, 1239: 12.5356435996, 1240: -2.97932339411, 1241: -82.9922687884, 1242: 20.7226605865, 1243: -0.956576244812, 1244: -23.3514478757, 1245: -71.6633422259, 1246: -2.76298968186, 1247: 2.27661947476, 1248: -2.28871133442, 1249: 14.3673602377, 1250: -2.69594853699, 1251: -1.44285964649, 1252: -24.597145793, 1253: -7.67704269028, 1254: -51.7384794575, 1255: -2.37462263339, 1256: -1.30594387744, 1257: -2.77453049373, 1258: -2.3166941478, 1259: -9.44607109253, 1260: 15.4045452933, 1261: 14.4048878054, 1262: -4.20943848915, 1263: -7.15380679737, 1264: -12.3418025745, 1265: -50.3293126473, 1266: -19.0812828775, 1267: -54.7515888736, 1268: 39.9980067787, 1269: -7.73526165192, 1270: 5.1200266692, 1271: -11.101225883, 1272: -17.0919874403, 1273: -1.41923377103, 1274: -66.7434873943, 1275: -10.6835003189, 1276: -22.3073335406, 1277: -2.5613886417, 1278: -12.4482308723, 1279: -4.96958694531, 1280: -11.8634069028, 1281: 47.4777936407, 1282: 8.04297993932, 1283: -99.6413013519, 1284: 99.9296827433, 1285: 46.2289544634, 1286: -0.365250559923, 1287: -9.7560778868, 1288: -23.4312991704, 1289: 18.2461576354, 1290: 100.0, 1291: 40.0166848788, 1292: -55.1426266998, 1293: -18.7607836328, 1294: -35.4038150817, 1295: 38.8549959401, 1296: -25.748389884, 1297: 10.2213261429, 1298: -20.5994357911, 1299: 41.5986550098, 1300: 15.3072861542, 1301: 28.7944707082, 1302: -6.94887454028, 1303: -12.3171904882, 1304: -40.4735169031, 1305: -8.31060846222, 1306: -8.52081116772, 1307: -4.65035169296, 1308: 58.2551013641, 1309: -11.4399335925, 1310: 18.9212148145, 1311: -0.201272842835, 1312: -7.02391791968, 1313: -98.7721317348, 1314: -100.0, 1315: 46.789733703, 1316: -33.2807004916, 1317: 1.28574657605, 1318: -7.46019385388, 1319: 23.8920707943, 1320: -2.36212629216, 1321: -75.7818503158, 1322: 67.3555047697, 1323: -8.79254417485, 1324: -68.3579587811, 1325: -57.9530134079, 1326: 25.2444794999, 1327: 97.9199759427, 1328: -4.07877077053, 1329: 8.45352010239, 1330: 79.9258769667, 1331: 57.0897417133, 1332: 66.5626413213, 1333: -41.1574232432, 1334: -5.70145893757, 1335: 3.20546105862, 1336: -15.5450941321, 1337: -29.9322212796, 1338: -17.5904488907, 1339: 63.2235784767, 1340: 33.0206953665, 1341: -17.3347491266, 1342: -8.20679905822, 1343: -58.1741601145, 1344: -45.9591351903, 1345: -45.6146793067, 1346: -5.5123177106, 1347: 17.8133634706, 1348: -1.03551197437, 1349: 23.7029389802, 1350: 43.8688459476, 1351: -47.5144898279, 1352: 99.999442535, 1353: -7.9361402995, 1354: -20.7467405764, 1355: 2.32941317244, 1356: 2.16391036043, 1357: -8.86305843048, 1358: 14.5225677102, 1359: -7.61823173931, 1360: -17.3431755722, 1361: 26.159423756, 1362: -7.69322893733, 1363: -6.19545794843, 1364: -19.4222817063, 1365: -9.34765646168, 1366: 2.41973173423, 1367: -8.55791852867, 1368: -9.86882272891, 1369: 59.1014549399, 1370: 14.4444003226, 1371: 0.270709399916, 1372: 58.6325088332, 1373: 38.477683513, 1374: -5.69448685264, 1375: -60.3902212378, 1376: -64.2554540371, 1377: -19.7583597207, 1378: -60.7900480349, 1379: 30.5900159047, 1380: -11.6844664974, 1381: -2.37228454716, 1382: 12.2944304751, 1383: 95.0715822562, 1384: -3.9577349068, 1385: 85.7971840054, 1386: -12.7222955029, 1387: 1.31496344896, 1388: -25.1358405816, 1389: -4.97542108654, 1390: -68.9258133909, 1391: -5.41560950202, 1392: 20.186769989, 1393: 50.4317126161, 1394: 16.7147736281, 1395: 43.5237383867, 1396: 71.2448231035, 1397: -0.32590319594, 1398: 9.39651473088, 1399: -1.70024355384, 1400: 36.8419442799, 1401: 33.5309015845, 1402: -4.47786407937, 1403: 40.6672525552, 1404: -7.91270440072, 1405: -5.93363090584, 1406: -23.4665745324, 1407: -26.4290458171, 1408: 15.9177475892, 1409: 16.4324333126, 1410: 0.957605824459, 1411: 16.7375470752, 1412: -5.1499692934, 1413: 34.8719998564, 1414: 0.340428014648, 1415: 10.4840203566, 1416: 3.06419097282, 1417: -5.62234308659, 1418: 2.26801627615, 1419: -6.02305964727, 1420: -5.65425505808, 1421: -15.9479569243, 1422: 19.5874564201, 1423: 12.1164207392, 1424: 4.95669078055, 1425: 17.4993245978, 1426: 4.61365785233, 1427: -19.0564999294, 1428: 17.1964293282, 1429: -4.44723458817, 1430: 5.23033839606, 1431: 13.9712203908, 1432: 3.0648674387, 1433: -5.20979800632, 1434: -12.6564124548, 1435: 55.993980924, 1436: 16.8504135731, 1437: 1.46084474188, 1438: 2.14050893553, 1439: -2.75915665619, 1440: -26.0332114227, 1441: -3.86512391288, 1442: -0.262679306058, 1443: -33.9292274499, 1444: -2.01483783443, 1445: 23.5522077817, 1446: -12.5084627269, 1447: -17.4931181302, 1448: 11.0172065473, 1449: -7.27360411402, 1450: 0.11696022347, 1451: -27.8051794202, 1452: -2.05092012992, 1453: -32.5100957498, 1454: -3.42595212673, 1455: -12.2566656338, 1456: 40.8915183613, 1457: 3.00631323608, 1458: -5.62916088345, 1459: -44.8729919896, 1460: -6.83278533246, 1461: 27.1701687502, 1462: -2.01722643652, 1463: 4.95688423332, 1464: -2.61230337749, 1465: -18.4623430624, 1466: -2.34457659726, 1467: 0.59990387233, 1468: -7.75500901362, 1469: -0.81276085358, 1470: 33.7673851041, 1471: -7.48066025265, 1472: -1.40036430992, 1473: -4.56059082319, 1474: -1.06338417625, 1475: -31.585957138, 1476: 37.7790892515, 1477: -4.14122772114, 1478: -11.5229841335, 1479: -0.281013863898, 1480: -0.319756134499, 1481: -2.8773933184, 1482: -6.04779864294, 1483: 20.4401990363, 1484: 6.99005383731, 1485: -1.83439668981, 1486: 1.04725793268, 1487: -3.51048298167, 1488: -7.00497824709, 1489: -2.63525089647, 1490: -1.7608476235, 1491: -0.884605553651, 1492: -2.68271941718, 1493: 13.0735796842, 1494: -2.63761967169, 1495: -2.65886637982, 1496: 0.630002610695, 1497: 5.52294858901, 1498: -2.72050806961, 1499: -18.2873563056, 1500: -2.13398749822, 1501: -0.610867925019, 1502: 60.0651578151, 1503: -2.29374807972, 1504: -2.15396817278, 1505: -3.65188883706, 1506: 9.66606386272, 1507: -2.8908270573, 1508: -20.16546248, 1509: -49.3667000057, 1510: -13.6402565834, 1511: 50.1216604548, 1512: -11.7495420373, 1513: -5.81248041136, 1514: 17.0288941921, 1515: -4.68168909484, 1516: -3.78040375605, 1517: -5.99062931057, 1518: -5.14731032728, 1519: -2.60998886388, 1520: 4.29787026977, 1521: 15.873510221, 1522: -0.386754241434, 1523: -0.229784956883, 1524: -7.03794506015, 1525: -10.4376263395, 1526: -0.0995203516261, 1527: -1.92276954178, 1528: 5.01317729202, 1529: -13.6272855017, 1530: -2.68909395813, 1531: 5.68618888651, 1532: -24.8366052505, 1533: -3.40076694246, 1534: -20.3163578074, 1535: -1.78229681187, 1536: -3.85216853516, 1537: -1.78204035884, 1538: -8.8289885348, 1539: 10.9371866306, 1540: -0.960171757455, 1541: 7.03201160755, 1542: -0.267170673935, 1543: -0.5243577144, 1544: -32.5553052839, 1545: -0.334106405903, 1546: -1.77938096053, 1547: 16.053581365, 1548: 0.760987622701, 1549: -0.468176836568, 1550: -0.518375974842, 1551: -0.448863055648, 1552: 1.20520366016, 1553: 3.01095609686, 1554: -1.84183778704, 1555: -0.456929442228, 1556: 0.934615143294, 1557: -21.9263809314, 1558: 5.24774205865, 1559: -6.23511203917, 1560: -5.58240601373, 1561: -53.5499234755, 1562: 6.18697234393, 1563: -1.41195463983, 1564: 9.48255302213, 1565: 0.0529809787738, 1566: 5.78106154955, 1567: 27.015094874, 1568: -0.390518012803, 1569: -1.93932646635, 1570: -0.267338411711, 1571: -3.21546918039, 1572: -22.8095137603, 1573: -14.7822055344, 1574: -0.175691844777, 1575: -0.491211261427, 1576: -0.450940333735, 1577: -0.219668761804, 1578: -0.5329430615, 1579: 0.183164054489, 1580: 32.6958094788, 1581: 10.8303795312, 1582: 47.2578978199, 1583: -12.0705551564, 1584: -2.38479651809, 1585: -0.875519051076, 1586: 0.0484558929338, 1587: 8.89232760521, 1588: 0.0428865999084, 1589: -0.51267555673, 1590: -29.1021926241, 1591: -1.17569312186, 1592: -18.2373352269, 1593: 3.77009557744, 1594: -15.2873451714, 1595: -0.381802234289, 1596: -2.08668568919, 1597: -0.321606269063, 1598: -37.2052882521, 1599: -0.401772650226, 1600: -0.117280900502, 1601: -4.79399597233, 1602: -2.1037813256, 1603: 24.2283485323, 1604: -0.419353216764, 1605: -0.404153760416, 1606: 0.845289552696, 1607: -0.504543882567, 1608: 1.10808904713, 1609: 12.2242744119, 1610: 17.3050587506, 1611: -1.42650877161, 1612: 0.191059145188, 1613: 13.4123697511, 1614: 1.19467336384, 1615: 13.9812846644, 1616: -19.4133080006, 1617: -6.35603468527, 1618: 0.617269048981, 1619: 2.05542990379, 1620: -1.73248251424, 1621: -3.08083690486, 1622: 0.0318826313308, 1623: 17.8149640693, 1624: -11.2251357962, 1625: -11.0976988191, 1626: -1.11656204431, 1627: 20.8571680888, 1628: -1.01814239068, 1629: -28.8931499923, 1630: 21.5093917222, 1631: -2.09761796018, 1632: -2.11537129246, 1633: 15.0471962963, 1634: -14.0726538071, 1635: -12.0412953534, 1636: -1.90745849556, 1637: -7.80316311614, 1638: 0.175509045239, 1639: 0.276596748311, 1640: 0.278468420574, 1641: 30.8792736139, 1642: -8.16258568145, 1643: 26.0302590985, 1644: -1.04765536661, 1645: -24.5406406285, 1646: 33.0739893109, 1647: -21.0704585537, 1648: 10.496916199, 1649: 12.0222177332, 1650: -38.9907107519, 1651: -2.6654047685, 1652: -2.52713348547, 1653: 1.32601301717, 1654: -1.76520468762, 1655: -1.69764906422, 1656: -9.36570283695, 1657: -10.9393804219, 1658: -2.39665081938, 1659: 3.51703156952, 1660: -0.855541705739, 1661: 1.16486466566, 1662: -2.42863728261, 1663: 43.7195163065, 1664: 4.10324601329, 1665: -17.2546818333, 1666: -9.45350198079, 1667: 4.25397280194, 1668: 1.98224424388, 1669: 30.5784225461, 1670: 55.7104649809, 1671: 1.43706613203, 1672: -6.93720928959, 1673: 34.5443989836, 1674: -16.1957797802, 1675: 4.94324920071, 1676: -5.43930585269, 1677: 6.07987641095, 1678: 33.6924588809, 1679: 7.47911933228, 1680: 26.1769651684, 1681: -23.3551283175, 1682: -1.80727525191, 1683: -1.96677556093, 1684: -0.274776952636, 1685: -4.65853946283, 1686: 2.39485475266, 1687: 7.90256389066, 1688: -7.50712450941, 1689: -9.02021505995, 1690: 1.71250126408, 1691: -30.438989373, 1692: -24.0142155857, 1693: 32.0959123789, 1694: 5.20107241035, 1695: 2.01742886179, 1696: -19.6088958013, 1697: -2.65938816121, 1698: 25.3615385235, 1699: 1.4142717347, 1700: -13.0630097073, 1701: 24.5788905671, 1702: -1.06652417772, 1703: 30.8691647805, 1704: -2.32648710318, 1705: -2.2274154046, 1706: 2.72378288492, 1707: 30.7523871619, 1708: -2.09490749423, 1709: -12.2635761454, 1710: -8.44798363594, 1711: -2.20320276135, 1712: 5.62240199402, 1713: -1.23796403832, 1714: 7.33193372282, 1715: -28.8631259029, 1716: -2.82774764538, 1717: 2.28407544857, 1718: 9.1826803399, 1719: -24.9511973251, 1720: 14.1094980953, 1721: -30.2237795742, 1722: 12.436807203, 1723: 45.8698373272, 1724: 2.81940940616, 1725: 29.4575759387, 1726: 0.768025838377, 1727: 48.6438348109, 1728: -4.3994598685, 1729: -1.01049589449, 1730: 4.83801590194, 1731: 1.01722515875, 1732: -27.690175293, 1733: -13.4009847156, 1734: -3.19729744541, 1735: 0.247887348706, 1736: -4.67453815273, 1737: -20.3694735609, 1738: 12.66113343, 1739: 10.2420375767, 1740: 18.750899498, 1741: -11.9087679566, 1742: 36.950400598, 1743: 24.0057877382, 1744: -22.6486359544, 1745: 0.545831151778, 1746: 0.309110743366, 1747: 2.63135155774, 1748: -7.50567848034, 1749: -15.2406545762, 1750: 4.35243980737, 1751: -1.40145510326, 1752: -3.64027180542, 1753: 3.84400725351, 1754: 1.54537696466, 1755: -1.0288725213, 1756: -6.75894714151, 1757: 0.0514440214226, 1758: 4.64232863169, 1759: 0.38142414068, 1760: 1.15510415595, 1761: -2.57556492979, 1762: -8.66464881354, 1763: 1.99926713432, 1764: -24.727329605, 1765: 0.215593169252, 1766: -2.29679605508, 1767: 12.8875875287, 1768: 1.27111995659, 1769: 4.1975968045, 1770: 7.02051900314, 1771: -13.2644970993, 1772: 2.33261330603, 1773: -3.48476454145, 1774: -1.32340485446, 1775: 0.676362833262, 1776: -5.11588347555, 1777: -3.48956220068, 1778: 2.2383472155, 1779: -0.635529081459, 1780: -1.65824798788, 1781: -7.65986674749, 1782: -6.47569650537, 1783: -2.57988303271, 1784: -6.67868826446, 1785: -2.41732453216, 1786: 2.32565815455, 1787: 2.19617421558, 1788: 4.72666995542, 1789: -7.49184297404, 1790: -1.24754036728, 1791: 6.75806411038, 1792: -2.05285393999, 1793: 0.378870271614, 1794: 8.68516510491, 1795: 8.03830920098, 1796: -2.39561664884, 1797: -3.4924206126, 1798: 1.67072523934, 1799: -3.43184726685, 1800: -3.54467353381, 1801: 1.19994799315, 1802: -2.6246420703, 1803: 0.289438081283, 1804: 2.80072205664, 1805: 0.735180765149, 1806: 3.28976319052, 1807: 9.12818752763, 1808: -5.44870871426, 1809: -10.0148291822, 1810: 1.43429238053, 1811: 1.06156609963, 1812: -6.78018533632, 1813: 2.21036315517, 1814: -8.51689891422, 1815: 2.36752931751, 1816: 2.2524342118, 1817: 1.53535213885, 1818: 2.08348666126, 1819: 2.01237354069, 1820: -3.07591382383, 1821: 0.0805935325176, 1822: 1.00195496138, 1823: 1.70677876365, 1824: 6.49418803848, 1825: 2.3920816568, 1826: -0.630441118793, 1827: -4.25799481258, 1828: 2.21774768463, 1829: 4.92682319324, 1830: 0.61012178276, 1831: 2.17580165417, 1832: 3.72663630315, 1833: -2.81557952453, 1834: 2.06342557965, 1835: -15.1199544106, 1836: 2.44390581849, 1837: 2.41442797896, 1838: 8.86853329968, 1839: 2.00111062799, 1840: 0.636147471661, 1841: -2.22531168552, 1842: 3.80148852961, 1843: 2.60069915322, 1844: 3.73880308316, 1845: 2.41201868844, 1846: -1.46675038681, 1847: -0.843061904751, 1848: -1.07485165911, 1849: 0.342086708646, 1850: 2.38572556827, 1851: 3.53866603631, 1852: 4.46116574365, 1853: 1.69846533413, 1854: 2.42250325389, 1855: -2.67140138345, 1856: 2.55495717799, 1857: 7.32986776204, 1858: 5.8292412249, 1859: -4.51914775999, 1860: -9.79142114608, 1861: 0.530532154228, 1862: -1.91532686277, 1863: -3.45313579376, 1864: -0.0314586059171, 1865: 14.2219979162, 1866: 2.36940828158, 1867: 7.36468291826, 1868: 1.41673224436, 1869: -0.978288463993, 1870: -2.5343080048, 1871: 0.335518369578, 1872: 0.629206541595, 1873: 1.86491268165, 1874: 10.9319236832, 1875: 0.333114579129, 1876: 1.87607057101, 1877: -0.29028869966, 1878: 3.05455738375, 1879: 0.33951064731, 1880: 0.464042376068, 1881: -1.61383720787, 1882: 1.6153352262, 1883: 3.67991841963, 1884: 0.304159374085, 1885: 0.0435347047589, 1886: -10.9471949659, 1887: -9.32968616965, 1888: -7.1257362809, 1889: 0.979567959135, 1890: -5.6016619763, 1891: 0.0716470243019, 1892: 0.380463029096, 1893: -3.30143639101, 1894: 0.516038143605, 1895: 2.04184278961, 1896: 0.818402101961, 1897: 0.532041539, 1898: 0.443254924852, 1899: 0.516431440345, 1900: 0.467120501356, 1901: 2.38164492934, 1902: -0.0995367892175, 1903: 2.69463589169, 1904: 0.441916782219, 1905: 0.0763303362401, 1906: 0.211772621718, 1907: 2.96704555634, 1908: -4.17714272339, 1909: -3.54814832476, 1910: 4.49189732237, 1911: 2.55695772834, 1912: 0.320901168566, 1913: 1.15917494489, 1914: 3.87965824555, 1915: 3.27410802592, 1916: -0.501822189935, 1917: 0.262222629358, 1918: 14.192309096, 1919: 0.257272270014, 1920: -6.75470683491, 1921: -5.00362681253, 1922: 1.59653275941, 1923: 0.176105851003, 1924: 0.38831469488, 1925: 0.511005737783, 1926: 0.251201723081, 1927: 0.64321467713, 1928: -0.50298248481, 1929: 4.93029057848, 1930: -4.54238596194, 1931: -0.468620908586, 1932: 16.8813068746, 1933: 0.0654140314793, 1934: -3.02261790826, 1935: 0.212271369357, 1936: 1.20015637835, 1937: 0.047826621406, 1938: 0.402760409047, 1939: 8.86890111223, 1940: 5.79800853642, 1941: 2.64219862536, 1942: 2.32525854803, 1943: -10.6265568201, 1944: 0.46177719175, 1945: 2.04160327816, 1946: 0.382529240463, 1947: -3.29162250516, 1948: 0.468882697775, 1949: 0.253545350491, 1950: 4.1445705668, 1951: 1.92609408631, 1952: -6.37160686951, 1953: 0.526881551769, 1954: 0.279132407795, 1955: 0.350636485233, 1956: 0.3344059817, 1957: 1.46104904761, 1958: 11.5953013891, 1959: 2.42408867204, 1960: 1.0870220514, 1961: 0.152898301112, 1962: 5.42525449846, 1963: 0.0117564011312, 1964: 0.149178845407, 1965: -3.89789368759, 1966: -5.56160438025, 1967: 2.59781820084, 1968: 1.85728562352, 1969: 1.87339279125, 1970: 1.08561871752, 1971: 0.319172845454, 1972: 0.439724012799, 1973: -0.293167588698, 1974: -1.06187514419, 1975: -1.98655468877, 1976: -7.53726628561, 1977: 0.90598282566, 1978: 3.79756050465, 1979: -4.70008216434, 1980: 1.49527127935, 1981: -10.9453867285, 1982: 1.62985667304, 1983: 2.45599324402, 1984: -0.369715675021, 1985: 1.88200712956, 1986: 3.61816419837, 1987: -4.78503726647, 1988: -7.35920874641, 1989: -1.61428154147, 1990: 0.456982763675, 1991: -1.78739705813, 1992: -16.0042452624, 1993: -7.22849438223, 1994: 23.3588673423, 1995: 12.2947142293, 1996: 15.7576288831, 1997: 5.37633637797, 1998: -9.65217162978, 1999: -2.37620010206, 2000: 6.66991081587, 2001: 0.612729806117, 2002: 4.41600121531, 2003: 0.971159086861, 2004: 2.11383297626, 2005: -0.827342356949, 2006: 1.61442798605, 2007: 3.10231352273, 2008: -1.23378993894, 2009: -2.12428770366, 2010: 2.31616222713, 2011: 7.27182163298, 2012: 2.64497253162, 2013: -9.47938360837, 2014: 2.05236390905, 2015: -0.938999684193, 2016: 2.5441238428, 2017: 2.31802058149, 2018: 3.73385222967, 2019: 2.99781251109, 2020: -9.39634869798, 2021: -5.86399154786, 2022: 0.911570717848, 2023: 11.0067374107, 2024: 5.8412922193, 2025: 8.64495341624, 2026: 1.53490710587, 2027: -6.26710928802, 2028: -7.20113191552, 2029: 13.9479037282, 2030: -15.8468535592, 2031: -7.27573179325, 2032: 1.31884664272, 2033: -0.540704925116, 2034: -0.350142574392, 2035: 0.79684384606, 2036: -3.46042627909, 2037: -5.49994496097, 2038: 2.82569359302, 2039: -10.4207668571, 2040: -13.594313251, 2041: 8.02041430165, 2042: -2.7387154712, 2043: 2.16600785328, 2044: 2.89258864498, 2045: 7.61317844878, 2046: 2.68251025589, 2047: -10.4618352811, 2048: 6.48551156334, 2049: -3.7980343724, 2050: 6.92884472494, 2051: 0.681740360178, 2052: 5.4294972673, 2053: 2.17667391942, 2054: 2.13112148641, 2055: 2.43667571943, 2056: 0.0424409325994, 2057: 1.89899398941, 2058: 2.20234232715, 2059: -9.28207306487, 2060: 2.10308526602, 2061: 7.48993034691, 2062: 1.11546043751, 2063: 1.9762697307, 2064: -1.61607687394, 2065: 1.03828959163, 2066: 0.970646954987, 2067: 2.81861026101, 2068: -0.20101475755, 2069: -9.46707110413, 2070: -4.93882742213, 2071: -6.94660305256, 2072: -0.363590543037, 2073: 0.0213673569158, 2074: 3.52857680125, 2075: 4.76889841861, 2076: -12.3940981319, 2077: -4.96794837549, 2078: 0.809133434155, 2079: -4.55015779928, 2080: -1.18586580263, 2081: -1.4209433316, 2082: 3.70638931802, 2083: -1.97308833659, 2084: 2.08030477181, 2085: -7.05393589235, 2086: 4.34849565081, 2087: -6.64062429084, 2088: -12.7684164049, 2089: 12.1752516339, 2090: -10.3544166819, 2091: -28.2446767173, 2092: -10.3234619718, 2093: -9.29775809204, 2094: -7.51198940866, 2095: -0.36721452684, 2096: -7.94186146289, 2097: -10.6207671469, 2098: -7.7469246666, 2099: -44.7241521475, 2100: 9.32498155932, 2101: 4.22634082433, 2102: -2.89405559412, 2103: -18.8902500544, 2104: -1.30522765298, 2105: -12.7393470513, 2106: 47.6573844537, 2107: 2.19470766006, 2108: -4.73147673374, 2109: 0.557169517015, 2110: -9.07473094095, 2111: -2.48863220341, 2112: 4.7555575376, 2113: 21.4471136949, 2114: 12.4999434513, 2115: -2.69686847818, 2116: -13.0160243885, 2117: -6.8622688998, 2118: -21.8289750346, 2119: 25.9978788153, 2120: 12.8526758137, 2121: -6.71266970363, 2122: 1.76737384965, 2123: 45.3642242282, 2124: -15.8444870031, 2125: -18.1414193925, 2126: 10.5986668204, 2127: 6.13323182471, 2128: 2.25641938068, 2129: -5.29051227502, 2130: 3.32645826311, 2131: -6.6864338899, 2132: -12.1115749819, 2133: 7.17725602706, 2134: -7.16961854164, 2135: 0.983553264381, 2136: 17.8419363981, 2137: 20.1585667172, 2138: 16.903933893, 2139: 7.03068454021, 2140: -6.76606219249, 2141: 17.4079537621, 2142: -1.07241871957, 2143: -6.36018758758, 2144: -16.1632256781, 2145: 2.14240482555, 2146: 28.8735876806, 2147: -1.98327869872, 2148: 1.53857988501, 2149: -4.98691524908, 2150: -1.59283478301, 2151: 16.7150921618, 2152: -0.360548060367, 2153: 3.77383855681, 2154: -6.16395639428, 2155: 13.5872518607, 2156: 24.6141997531, 2157: -1.6209864396, 2158: -14.4177506161, 2159: 19.2572714988, 2160: -1.34369835984, 2161: -11.8488177037, 2162: 0.110211703734, 2163: 0.650021657258, 2164: -1.64850503399, 2165: -1.50764910408, 2166: 2.64870131317, 2167: -1.96261806172, 2168: 20.5260678056, 2169: -0.471781675964, 2170: -1.48372676402, 2171: -0.358412200746, 2172: 6.41018106607, 2173: 5.97345124874, 2174: -7.36932752439, 2175: 6.04957633309, 2176: 6.86446739358, 2177: -1.68348232752, 2178: -4.94974309662, 2179: 8.02375850344, 2180: -0.886740983064, 2181: 4.66479632195, 2182: 3.94245715684, 2183: -1.16848898111, 2184: 0.841579690159, 2185: -1.16376223198, 2186: -1.54484492965, 2187: 42.6871214122, 2188: -1.90125150556, 2189: -1.5026057308, 2190: -1.88349475614, 2191: 5.6270260318, 2192: -1.52090326353, 2193: 0.179534934153, 2194: -1.47087558377, 2195: 9.86821076606, 2196: -1.48742787686, 2197: -1.75415426976, 2198: -1.41201785663, 2199: 0.443840475647, 2200: 2.65411916525, 2201: -1.66564333897, 2202: -1.5405591637, 2203: -0.888037793401, 2204: 26.6178028066, 2205: -1.03301007764, 2206: 17.6822002331, 2207: 13.6840446739, 2208: -21.3700746615, 2209: 6.31778491161, 2210: 19.909717503, 2211: 14.5095907024, 2212: -0.342958840791, 2213: -15.0124293096, 2214: -16.5240597483, 2215: -17.2789334394, 2216: -4.65388411603, 2217: -4.66416518636, 2218: -6.68612734794, 2219: -10.8683869596, 2220: -0.488186730215, 2221: -1.40865325284, 2222: -1.36434910104, 2223: -42.4741112991, 2224: -0.636575218858, 2225: -1.72583564417, 2226: 2.1328217568, 2227: 3.10132939134, 2228: 0.227602167232, 2229: -3.78047410436, 2230: 3.4891414725, 2231: -0.309979251782, 2232: -20.0030771954, 2233: 3.62861415281, 2234: -1.05958117651, 2235: 11.2602160966, 2236: -7.43670668294, 2237: 0.14120131953, 2238: 1.59684616922, 2239: 8.57302539518, 2240: -0.482366770096, 2241: -0.470141425134, 2242: -40.7489754559, 2243: -0.383390609607, 2244: -1.13298453321, 2245: -14.6694635016, 2246: -1.22201645333, 2247: -0.458413828184, 2248: -0.450743971178, 2249: -0.464607068884, 2250: -1.86785842369, 2251: 14.2445069092, 2252: -1.2234845079, 2253: -0.255429021962, 2254: -0.919845703213, 2255: -0.189707792141, 2256: 1.75469096241, 2257: -26.9076045606, 2258: 12.7447605366, 2259: -9.23461422108, 2260: 33.1650654162, 2261: 4.92264823349, 2262: -10.040793517, 2263: -4.80564474351, 2264: 4.1855242273, 2265: -14.2930033004, 2266: -0.548134149165, 2267: -9.34488059345, 2268: -0.416159542271, 2269: -14.9908425561, 2270: -10.5053969556, 2271: 4.03111151053, 2272: -1.12256771142, 2273: -0.455974450569, 2274: -0.413220264569, 2275: -0.407106254412, 2276: -0.486055930747, 2277: -1.43602429127, 2278: -6.19967477422, 2279: -7.66281076665, 2280: 17.851334558, 2281: -15.1371148917, 2282: 7.55386551321, 2283: -0.0831787848931, 2284: -3.63277493374, 2285: 12.0048223365, 2286: 7.81236243213, 2287: -0.477204958318, 2288: 5.28557640858, 2289: -1.06039665688, 2290: -0.674969353843, 2291: -4.50167998516, 2292: 3.36581149641, 2293: -0.385959755168, 2294: -1.8261028658, 2295: -0.431107317709, 2296: 6.00281285268, 2297: -0.448217737583, 2298: -0.484898808297, 2299: 15.9707860393, 2300: -1.52020337026, 2301: 13.6570662955, 2302: -0.752534724968, 2303: -0.436634104759, 2304: -0.515939244605, 2305: -0.417897692654, 2306: -1.67077947431, 2307: 3.64522989003, 2308: -1.47585352488, 2309: -0.530838706403, 2310: -0.834102925158, 2311: -16.3385073931, 2312: 0.533678677071, 2313: -24.505046619, 2314: 10.694389769, 2315: -6.50018894521, 2316: 1.72961381669, 2317: 5.28225980116, 2318: 0.411069476968, 2319: 3.56302172204, 2320: -0.286766097126, 2321: -6.02330835733, 2322: 0.359674487881, 2323: 1.32984359451, 2324: 0.856392112636, 2325: 5.00654917395, 2326: 0.0253578691905, 2327: -1.71538566223, 2328: -7.70023509514, 2329: -1.14203046362, 2330: 12.7611436857, 2331: -13.4904814006, 2332: 4.43263690641, 2333: 6.66707592853, 2334: -1.64471391829, 2335: 1.93247547175, 2336: -3.78331722239, 2337: -4.45695662567, 2338: 14.0237439466, 2339: -7.73150369305, 2340: 33.8985338536, 2341: 19.1083135792, 2342: -4.57813175394, 2343: -12.5955350682, 2344: -2.69800435104, 2345: 19.7724574321, 2346: -9.522280662, 2347: -1.47828464572, 2348: 4.57382648544, 2349: -1.76228540623, 2350: 2.00090375482, 2351: 7.59673161407, 2352: -1.60459284202, 2353: -1.17597530667, 2354: -10.7099918065, 2355: 27.3290068191, 2356: -1.83497541607, 2357: -8.42237127941, 2358: -1.56706788272, 2359: -1.72833874765, 2360: -3.6951697669, 2361: -14.6847003667, 2362: -10.5749290046, 2363: 5.83991579725, 2364: -0.427642464152, 2365: 0.490671276006, 2366: 22.9777191569, 2367: -9.49008272176, 2368: 27.9926517359, 2369: 19.3779324045, 2370: 23.8119882264, 2371: -17.3893486051, 2372: -7.39803666102, 2373: 13.491470333, 2374: -12.5273689006, 2375: 1.16420459414, 2376: 12.0564737608, 2377: 5.51492306233, 2378: 22.9255943904, 2379: 6.1704757594, 2380: -7.06067745109, 2381: -1.46688684275, 2382: -1.28479722115, 2383: -1.10489754643, 2384: -3.17612874229, 2385: -3.28388252876, 2386: -13.1599884015, 2387: -21.7506224424, 2388: -6.06146458664, 2389: 5.84813844361, 2390: 1.8124131536, 2391: 12.3254778067, 2392: 0.862018684759, 2393: 5.51301225679, 2394: 4.21729869486, 2395: 0.313769044247, 2396: -14.2549481042, 2397: -1.15002030733, 2398: 13.7245138067, 2399: 10.747427585, 2400: -1.55577178102, 2401: 13.8105363189, 2402: -1.66139906226, 2403: -1.77553548071, 2404: 2.20038226013, 2405: 10.4738977256, 2406: -0.764219037872, 2407: -1.1941090669, 2408: -25.5484590551, 2409: -1.56068344569, 2410: 12.6765356879, 2411: 7.72582627895, 2412: -1.43450069507, 2413: -39.4363884006, 2414: -0.735602570541, 2415: -1.44883685596, 2416: 4.13470816699, 2417: 0.270260159773, 2418: 7.75723462081, 2419: 4.26563399213, 2420: -40.1850242787, 2421: 3.37968882681, 2422: 10.6587885119, 2423: -26.7189460065, 2424: 12.5177236113, 2425: -11.306650811, 2426: 20.6162607365, 2427: 3.36234620121, 2428: 5.03475680167, 2429: 0.149815350084, 2430: 0.867444061767, 2431: 10.1037866249, 2432: -9.8642144387, 2433: -1.5401026258, 2434: 18.281165981, 2435: -8.44785193001, 2436: 3.10635933776, 2437: 5.64925897505, 2438: -0.644515358649, 2439: 1.23882454479, 2440: 4.3841325473, 2441: 22.0770574486, 2442: -23.8808553703, 2443: 12.4811741335, 2444: -0.0533032991949, 2445: -2.07452348726, 2446: 2.00893130022, 2447: 2.54606278409, 2448: 3.30592742235, 2449: 2.33144942239, 2450: 1.49522415267, 2451: -2.12287518824, 2452: -0.555988223372, 2453: -4.80416929442, 2454: -2.83246942282, 2455: -2.99033184129, 2456: 2.55911891679, 2457: 1.62684666098, 2458: -0.169026344966, 2459: 0.69483983357, 2460: 13.2778094208, 2461: 0.906085256679, 2462: -3.52690384671, 2463: -4.34202797867, 2464: 1.03358254062, 2465: 0.00289447296498, 2466: -0.0463931472528, 2467: -0.130240505891, 2468: -3.55775416876, 2469: 0.243179356412, 2470: -0.446979652216, 2471: -1.66106030092, 2472: 2.29517255403, 2473: -1.52921576442, 2474: -1.51964234412, 2475: -0.521604593512, 2476: -0.262473462901, 2477: 1.04551347779, 2478: 7.73191491756, 2479: 0.265442960114, 2480: -0.880985560262, 2481: 0.995436673335, 2482: 2.2830955243, 2483: 2.10771574629, 2484: -0.532740146495, 2485: 1.6826984299, 2486: -1.20094712448, 2487: 4.81731284282, 2488: -9.33915515828, 2489: -3.07081133493, 2490: -1.07747763376, 2491: -3.01725735652, 2492: -0.648160329575, 2493: -1.99728637094, 2494: -1.63071017413, 2495: 1.6284213486, 2496: -0.0975014056669, 2497: -1.17136783628, 2498: 3.4286830213, 2499: -0.870820080043, 2500: -6.05951514104, 2501: -3.83814492802, 2502: -6.0477730963, 2503: -2.99976062139, 2504: 0.876797032572, 2505: -1.95759588436, 2506: -0.69570948779, 2507: 3.17568699579, 2508: -1.53090307504, 2509: 1.22526106769, 2510: 5.13750790458, 2511: 0.133510516519, 2512: 1.98876561459, 2513: 0.0566969499964, 2514: -0.48232043918, 2515: 1.25405724609, 2516: -5.7296047076, 2517: 11.1688650641, 2518: -0.986356328837, 2519: -4.12826293601, 2520: -2.54454348126, 2521: -3.08583769849, 2522: 1.03034321817, 2523: -1.93649480893, 2524: 1.6610319126, 2525: 7.40285434143, 2526: -0.101119807938, 2527: 1.62188348168, 2528: 2.72451184906, 2529: -2.46238574463, 2530: 0.320839810083, 2531: -3.99368169209, 2532: -1.12436642476, 2533: 4.69624269258, 2534: 2.10667195379, 2535: 0.767195375632, 2536: -0.618996188466, 2537: -0.540938787616, 2538: -0.261161732368, 2539: -0.513653178916, 2540: 3.06596616168, 2541: 0.296379812475, 2542: 0.561576079262, 2543: -0.296963950153, 2544: -0.770161876514, 2545: -0.535856699693, 2546: 0.584938475019, 2547: -0.346293725652, 2548: 0.589775425782, 2549: 2.03966982712, 2550: -0.0594708573226, 2551: -0.579655520288, 2552: 0.968087382215, 2553: 4.75674657015, 2554: 0.281942461659, 2555: -1.84726220531, 2556: -11.4266409347, 2557: -0.838342312647, 2558: 6.22497159747, 2559: -0.164022468162, 2560: -1.83953146914, 2561: 0.533449418176, 2562: 9.92788762599, 2563: -1.97042900877, 2564: -3.09591578067, 2565: -3.88072387336, 2566: -2.65135580692, 2567: -0.80121488094, 2568: -0.588932558785, 2569: -0.0355405671854, 2570: 0.0357037697052, 2571: -0.745764930504, 2572: -1.0506899316, 2573: 0.0510219101697, 2574: -0.247749402135, 2575: 0.230872822795, 2576: 4.47900487407, 2577: -0.687165143917, 2578: -0.53230930528, 2579: -1.58076436871, 2580: 0.876011328856, 2581: 7.39099184369, 2582: -0.223495851761, 2583: -0.142977291647, 2584: -0.108792301854, 2585: 2.40449730893, 2586: -0.990521083505, 2587: 0.405533584783, 2588: 0.524276586935, 2589: -0.0538611866506, 2590: 0.135301964368, 2591: 3.08294452215, 2592: -0.0256689116275, 2593: -0.224815749739, 2594: 1.17370711394, 2595: 0.0591712184141, 2596: -0.0480460406356, 2597: -0.129357927413, 2598: -0.0449146692993, 2599: -0.582993841893, 2600: 10.2449520654, 2601: 0.756121818309, 2602: -0.244670112735, 2603: -0.197244871667, 2604: 0.409806463938, 2605: -0.564617150573, 2606: -3.25230654314, 2607: -0.034271668326, 2608: 10.2448238037, 2609: -3.9435413153, 2610: 0.0206624763178, 2611: 0.516812337841, 2612: -1.71926575378, 2613: -1.28983663698, 2614: 0.278261088032, 2615: -0.135020962197, 2616: -6.53285678611, 2617: -0.0412956119307, 2618: 5.92325438963, 2619: -14.4794936909, 2620: -1.33115913001, 2621: -0.0582565546655, 2622: -0.133101120579, 2623: -0.173483947664, 2624: -0.277997851126, 2625: 0.0944569163407, 2626: -3.15936292144, 2627: 5.47726876691, 2628: -0.0776331258912, 2629: -1.12694886918, 2630: -4.01499608364, 2631: -2.33639974985, 2632: 1.64081151164, 2633: 0.287603707767, 2634: -3.40897814263, 2635: -0.499746253743, 2636: -0.0179737106938, 2637: 0.671133406093, 2638: 0.646225831838, 2639: 0.230536124255, 2640: -0.568108304415, 2641: 0.761456581317, 2642: -0.0796984580266, 2643: -0.649707318231, 2644: -0.090953437902, 2645: 0.296530106085, 2646: -0.184488480563, 2647: -0.0512263256143, 2648: 4.22260433262, 2649: -0.492399256871, 2650: 4.28465558845, 2651: 0.140633306387, 2652: 0.0459773392339, 2653: -0.565159340887, 2654: -0.0292076518085, 2655: 0.654360897611, 2656: 2.52204506353, 2657: -4.38986959174, 2658: 0.0833521800476, 2659: 0.0945178104657, 2660: -2.83592298374, 2661: -0.0697267897795, 2662: -3.0128617237, 2663: -0.360538590834, 2664: -0.851403716065, 2665: -0.46814120419, 2666: 1.98951101812, 2667: 0.708252494894, 2668: 0.887851960489, 2669: -0.153194036746, 2670: -4.62222198558, 2671: -0.0704890446613, 2672: -0.296640550023, 2673: 0.106159973522, 2674: -3.1201393115, 2675: 0.306516751941, 2676: 2.93235581255, 2677: 4.02188386865, 2678: 0.245228765534, 2679: -1.60743372187, 2680: -0.534126925878, 2681: 1.89186353585, 2682: 0.406114881189, 2683: -0.0615873255536, 2684: -1.29556300453, 2685: -3.1021127606, 2686: -0.713271377064, 2687: -1.31692269089, 2688: -1.08628767754, 2689: -2.4067108572, 2690: -0.782887564004, 2691: 4.04132624186, 2692: -2.96258789232, 2693: 2.71074852237, 2694: -4.39276123071, 2695: 1.32310917178, 2696: 1.22006050768, 2697: -1.50880243683, 2698: -3.00383664027, 2699: -4.06732446466, 2700: -3.25386140648, 2701: -0.847391001666, 2702: -0.0765892746695, 2703: 1.22102438304, 2704: -3.83719579464, 2705: 0.651418519628, 2706: -1.75866748214, 2707: 1.75239409061, 2708: -0.392379413342, 2709: -8.38101202179, 2710: 6.36971564863, 2711: 1.06094929811, 2712: 0.539061346742, 2713: -1.43142880592, 2714: -1.54912963871, 2715: -5.38562587828, 2716: -0.924256312026, 2717: -5.56363660166, 2718: 11.6167206852, 2719: 3.41560727293, 2720: 1.23604173496, 2721: -2.18075464146, 2722: -1.41289778517, 2723: -2.33205503935, 2724: -0.886450625077, 2725: 1.05265418704, 2726: 0.410480837968, 2727: 13.7742925969, 2728: 2.91493830959, 2729: -2.69475912096, 2730: -0.0797871400164, 2731: -0.605345648599, 2732: 1.32871043148, 2733: -1.54668413049, 2734: -3.39117348077, 2735: 3.46528197241, 2736: -1.39924526403, 2737: 1.76315330152, 2738: -2.71800978764, 2739: -3.47393667593, 2740: 2.71083850467, 2741: 0.651143799333, 2742: 0.0671455004902, 2743: 0.769741569837, 2744: -0.572517099948, 2745: 1.80199725346, 2746: -0.509709551817, 2747: -2.1775325732, 2748: 5.86694376966, 2749: -0.171686677872, 2750: 4.34223613631, 2751: -0.631135149159, 2752: -0.602807944898, 2753: 0.54437284815, 2754: -0.909294858811, 2755: 0.231940811526, 2756: 0.722922344841, 2757: -8.53796304023, 2758: -0.652041189962, 2759: -5.67605840514, 2760: -3.90130739909, 2761: -0.799865141952, 2762: 6.77298426299, 2763: 0.0604317269399, 2764: 1.29508743282, 2765: 4.16840464976, 2766: 3.10647416789, 2767: 5.27755418865, 2768: -7.13877127806, 2769: -1.50981184658, 2770: 2.89917942975, 2771: 5.28097630283, 2772: 4.36105808408, 2773: 3.99474293117, 2774: -0.369954358359, 2775: 3.599372287, 2776: 0.479447199263, 2777: 0.833644747979, 2778: -0.140873707347, 2779: -4.64642176788, 2780: 5.58554486552, 2781: 1.08270237238, 2782: -0.220176466224, 2783: -1.55268127697, 2784: -0.0377552643, 2785: -3.70894109622, 2786: -7.51561916136, 2787: -8.33571334174, 2788: 3.43376167792, 2789: 13.9801585438, 2790: 2.97843804239, 2791: -5.08888939209, 2792: 2.76932259679, 2793: -0.785781884065, 2794: -82.6974353358, 2795: 14.9868790583, 2796: -0.346102640934, 2797: -24.1658072364, 2798: 9.9433313405, 2799: 50.7003655265, 2800: 2.98458132602, 2801: 15.2804750378, 2802: -8.83253815715, 2803: -9.51952647477, 2804: -11.4471948229, 2805: -3.0077091807, 2806: -14.0969015121, 2807: -11.1426037399, 2808: -2.38409567305, 2809: 22.4196031678, 2810: -4.39384672891, 2811: -33.9535272061, 2812: -4.22230790854, 2813: -9.71364750792, 2814: -19.8019330965, 2815: -39.5788726706, 2816: 29.4642265423, 2817: 4.0530320469, 2818: 50.112093296, 2819: 8.81331862807, 2820: 0.39881567849, 2821: 23.3747820281, 2822: -20.5219541828, 2823: -40.7628794528, 2824: 10.6082289378, 2825: 1.19063688505, 2826: -1.21720003585, 2827: -2.53109478226, 2828: -19.9131988641, 2829: -11.4064539499, 2830: -33.1681381452, 2831: -27.6637735557, 2832: 6.42538946352, 2833: -5.81097275884, 2834: -14.8566430769, 2835: -38.0984344032, 2836: 14.0306573607, 2837: -8.18288893459, 2838: 14.5003523111, 2839: -2.95022061485, 2840: 4.62388617967, 2841: -9.50723962622, 2842: 0.299317553444, 2843: -16.1080006035, 2844: -41.2179170335, 2845: -4.94247201608, 2846: -6.62457306323, 2847: -6.15450117829, 2848: -5.88124771469, 2849: -20.8908490623, 2850: -18.0775376184, 2851: 7.03316604756, 2852: 33.0804758256, 2853: -50.6757701287, 2854: 18.4974235873, 2855: -23.1857433865, 2856: -5.35749622769, 2857: 34.6242986096, 2858: 5.15419498153, 2859: 7.18250298446, 2860: -3.80493109403, 2861: 6.96821913343, 2862: -3.29310517063, 2863: -2.68470723305, 2864: -5.96447862293, 2865: -4.28647592693, 2866: -13.8429007527, 2867: -10.1980093964, 2868: -7.54859980316, 2869: -5.63807922447, 2870: -12.0761753437, 2871: -40.5448677354, 2872: -16.3452764481, 2873: -2.15918444961, 2874: 0.708839850292, 2875: -3.91024879476, 2876: 6.70839348632, 2877: 17.1625976901, 2878: -4.89334033928, 2879: 12.5359721595, 2880: -13.0827693876, 2881: -4.86982122249, 2882: 21.2171331375, 2883: 1.67072232847, 2884: -4.70148906003, 2885: 9.05004780594, 2886: -4.05895785007, 2887: -3.62162461517, 2888: -4.14756503206, 2889: 23.6538812844, 2890: -2.78798107382, 2891: 1.92247200785, 2892: 12.9733358202, 2893: 25.2122801268, 2894: 5.78518622669, 2895: 25.7224737046, 2896: -0.496031432031, 2897: -4.09069331781, 2898: 17.329602479, 2899: -3.75569158883, 2900: -5.06554311681, 2901: -3.01359103503, 2902: 14.1524431884, 2903: -1.26419155231, 2904: -5.90722709426, 2905: -54.9329125789, 2906: 0.620641834651, 2907: 27.634153791, 2908: 1.59707550655, 2909: -3.97212356562, 2910: -6.06583128884, 2911: 29.4423774475, 2912: 38.7187391813, 2913: 15.6646715175, 2914: -20.9811038973, 2915: -1.58809277948, 2916: 25.3847481908, 2917: -15.4708835717, 2918: -0.858857256645, 2919: 0.127689522117, 2920: 37.2403275895, 2921: -10.8260586079, 2922: -0.761370215887, 2923: -2.28151342043, 2924: -0.857552714844, 2925: -7.00690257233, 2926: -1.91995247867, 2927: -1.14491209367, 2928: -13.7700244304, 2929: -1.45319647751, 2930: -50.011710379, 2931: -0.574967952582, 2932: -1.36870411265, 2933: -25.6177981178, 2934: 1.95279485833, 2935: -21.3057902058, 2936: 1.81163177234, 2937: -1.41915125934, 2938: -0.875882470775, 2939: -1.007368567, 2940: 50.3280751085, 2941: -0.784114675119, 2942: -3.97683464479, 2943: -5.37928364136, 2944: -0.717124871346, 2945: -0.818387816349, 2946: -0.852163351416, 2947: -0.731295797668, 2948: -4.64209635285, 2949: -10.8334869719, 2950: -0.243400091766, 2951: -1.90415735485, 2952: -0.126509124059, 2953: -3.91341329154, 2954: -4.07214479732, 2955: -24.2837133447, 2956: -5.68452485882, 2957: -16.9751894801, 2958: -18.5851594055, 2959: 16.9531205826, 2960: -13.0072103797, 2961: 16.3612637351, 2962: -1.13160245694, 2963: 15.0100298224, 2964: -0.704958514753, 2965: -63.4339952279, 2966: -0.800693006571, 2967: -11.5230391046, 2968: 1.80822724046, 2969: 2.08547689952, 2970: -0.446135575182, 2971: -0.907730894588, 2972: -0.992541742599, 2973: -1.62003808509, 2974: 0.0288229850846, 2975: 1.66780330185, 2976: -19.2127814503, 2977: 16.3288007284, 2978: 55.7679138363, 2979: 20.6960510943, 2980: -0.0486824015191, 2981: 25.2195941794, 2982: -4.85243016884, 2983: 2.31057784937, 2984: -5.36125716109, 2985: -0.716824696007, 2986: -35.3026344454, 2987: -2.74302244394, 2988: -1.4153916259, 2989: 1.9285846233, 2990: -28.649176282, 2991: -0.72561129601, 2992: -4.01554611822, 2993: -0.842032273467, 2994: -23.7968213267, 2995: -0.893592754211, 2996: -0.710443396204, 2997: 14.8269481358, 2998: -4.10419602501, 2999: 23.2271242975, 3000: -0.8900309955, 3001: -0.825178512009, 3002: -1.52874024374, 3003: -0.86821813598, 3004: -0.625048661085, 3005: 21.4333776552, 3006: 13.3200431516, 3007: -1.65314675366, 3008: 1.22987021397, 3009: 31.664840289, 3010: 10.7632822026, 3011: -22.6070200758, 3012: -18.7971483968, 3013: -12.376321936, 3014: 12.2282518755, 3015: 11.8377499982, 3016: -3.85594748816, 3017: 20.4605604023, 3018: -1.05055639486, 3019: 6.25362939235, 3020: -12.1941988006, 3021: -4.68238460478, 3022: -2.69162448011, 3023: -37.9883451047, 3024: -0.852449891303, 3025: 4.84718515938, 3026: 26.3659303413, 3027: -4.55240719448, 3028: -7.37936692784, 3029: 29.4851492356, 3030: -10.6829270647, 3031: -5.1078887995, 3032: -2.17757605884, 3033: -2.6722983355, 3034: 10.9717586941, 3035: -9.40977117307, 3036: 17.9282446291, 3037: 28.2903462183, 3038: 11.2148681704, 3039: 45.4303376242, 3040: 32.4849394042, 3041: -15.3967138103, 3042: 3.16235689957, 3043: 11.5563253993, 3044: 38.5956122568, 3045: -18.9600118665, 3046: -2.96851876742, 3047: -8.25150117531, 3048: 3.57749902508, 3049: -7.8729044236, 3050: -4.37010263224, 3051: -3.93601767036, 3052: -5.66431853266, 3053: 2.04410752404, 3054: -4.17242275632, 3055: -10.6744563876, 3056: -9.98263810839, 3057: -4.55008924081, 3058: -9.56111812087, 3059: 11.3871523533, 3060: 24.7638563638, 3061: 6.393587653, 3062: -4.29263797532, 3063: -3.47881868444, 3064: -12.3378292223, 3065: 15.9007095533, 3066: -11.0283074822, 3067: -44.6393802999, 3068: -43.3971447564, 3069: -27.5033733696, 3070: 0.211836154736, 3071: -5.92995579061, 3072: 14.3620305177, 3073: -4.09442420675, 3074: 42.2464330049, 3075: -18.9989294786, 3076: -1.80779954168, 3077: 11.3269591052, 3078: -20.6862051195, 3079: -3.46940857711, 3080: -11.710862201, 3081: 4.51244906228, 3082: -24.181933666, 3083: -2.04140177805, 3084: 10.921168586, 3085: 16.6243296681, 3086: -6.78663568506, 3087: -49.467323239, 3088: 47.068005972, 3089: 23.671957896, 3090: 14.6508684272, 3091: 1.25554767156, 3092: 11.2549812733, 3093: -4.46801776771, 3094: 26.0749957728, 3095: 1.21050127511, 3096: 4.13488263379, 3097: -1.93600597063, 3098: -3.63980643133, 3099: 57.7015934735, 3100: -4.11285034292, 3101: -4.05440389508, 3102: -1.483475509, 3103: -7.27908204347, 3104: -4.45890296924, 3105: -5.11512191221, 3106: -17.8360000192, 3107: -4.19073567859, 3108: -48.2469287578, 3109: -4.04368033569, 3110: -4.63370049275, 3111: 6.9319553733, 3112: -5.05575662877, 3113: 0.916283585142, 3114: -22.6256783926, 3115: -2.24526529142, 3116: 35.5046188234, 3117: -32.3299820336, 3118: 44.0972861774, 3119: 3.43604467471, 3120: 2.70836837328, 3121: -2.08293447283, 3122: -8.25977284593, 3123: -16.1546801822, 3124: -12.6872474826, 3125: -3.68111331032, 3126: 16.4556068941, 3127: -4.99954431489, 3128: 54.8126674602, 3129: 45.2111536466, 3130: -29.892598331, 3131: -4.12296997404, 3132: 6.98675291486, 3133: 27.9629535979, 3134: 2.42162177864, 3135: 2.81027386884, 3136: -27.7134197401, 3137: -34.6958173748, 3138: 57.9383739825, 3139: -54.1462574015, 3140: -7.49865661115, 3141: 35.6981555646, 3142: -0.00122361010024, 3143: 6.37731537313, 3144: -2.62513351672, 3145: -7.64484575143, 3146: -10.3996886479, 3147: 7.16665121699, 3148: -4.61205187755, 3149: -12.0610458138, 3150: -1.19937369236, 3151: 6.68462631393, 3152: 0.808939413213, 3153: -3.3657292189, 3154: 5.49294397185, 3155: 6.36145141802, 3156: -0.303715690056, 3157: 3.82125176806, 3158: 15.9439707937, 3159: -0.801966604329, 3160: -17.2517031559, 3161: -5.91139692539, 3162: 0.676350007555, 3163: 0.300043223954, 3164: -2.42382638317, 3165: -3.80168222465, 3166: 7.74291229716, 3167: 9.53997506931, 3168: 2.12125111865, 3169: -10.9611366784, 3170: -7.89479741605, 3171: 12.0612215606, 3172: -0.799806582756, 3173: 6.35744563567, 3174: -0.00920982024632, 3175: 10.5378595054, 3176: 8.51548092066, 3177: -5.23518258544, 3178: 8.50157461923, 3179: 10.6288421594, 3180: -3.56636937508, 3181: -0.707466714913, 3182: 10.2418635676, 3183: -1.28975263896, 3184: 2.49910127287, 3185: 8.09076322047, 3186: -6.59392009367, 3187: 4.2253130112, 3188: -9.4804240889, 3189: -1.86094454815, 3190: 15.7692768902, 3191: 3.81415440074, 3192: 8.06593096257, 3193: -0.0616171800721, 3194: -1.16504407368, 3195: -2.83342530718, 3196: 2.29662999024, 3197: -1.28906682239, 3198: 12.7968248032, 3199: 4.07516246118, 3200: 2.43218804694, 3201: 11.4210309538, 3202: 0.277763278943, 3203: 3.50969450374, 3204: -6.64853572814, 3205: -7.02092738088, 3206: 3.7830375414, 3207: -1.96043286653, 3208: -9.27862582347, 3209: -1.30335679546, 3210: -2.28884362329, 3211: -1.80021617263, 3212: -1.57478025344, 3213: -1.4575612791, 3214: -1.38997848606, 3215: 22.7901748038, 3216: -1.14404603303, 3217: -1.62050598823, 3218: -7.88903637531, 3219: -5.61937704899, 3220: 3.80822460456, 3221: 4.42406020809, 3222: -0.207519511034, 3223: 5.74851325396, 3224: -1.02656937034, 3225: -2.26148489749, 3226: 5.4881884096, 3227: -1.03661974753, 3228: 1.48244164872, 3229: 8.12449989421, 3230: 0.164577370386, 3231: -0.890210437534, 3232: -1.7996649577, 3233: -1.40472364005, 3234: -0.19794397221, 3235: -1.54242443526, 3236: -0.440016793307, 3237: -0.572848337309, 3238: -2.44717586638, 3239: -1.95457680575, 3240: -3.46923633208, 3241: -1.01623225501, 3242: 0.598901074063, 3243: -1.35797106784, 3244: 2.28223170936, 3245: -2.43608464665, 3246: 0.370007949884, 3247: 8.68839065571, 3248: -2.98872015745, 3249: -1.02388031225, 3250: -0.148012105808, 3251: 2.36702198431, 3252: -0.743619325899, 3253: -0.331626182263, 3254: -11.5603739923, 3255: 4.29785886638, 3256: -19.8241150752, 3257: -0.180052117155, 3258: -13.7772307284, 3259: 2.34297869354, 3260: 6.81699189809, 3261: 13.2459605736, 3262: -3.50640034342, 3263: 4.13783156267, 3264: -8.09416113821, 3265: 0.258072323351, 3266: 7.77572499167, 3267: -0.402291574099, 3268: -0.331878667596, 3269: -1.74903855683, 3270: 0.742312235288, 3271: -0.0373824031905, 3272: -1.19384802863, 3273: -0.939631806886, 3274: 6.62017744016, 3275: 0.44856815275, 3276: 1.70593115615, 3277: -5.59411564424, 3278: 0.770767518245, 3279: 8.09738245722, 3280: -0.288151767672, 3281: -1.84983265737, 3282: 3.99905687035, 3283: 10.3588457998, 3284: -9.23474125919, 3285: -0.714948682987, 3286: 5.48011447366, 3287: 0.0106366681667, 3288: -0.0515823384947, 3289: -3.6733060146, 3290: -0.163168233309, 3291: -1.6177482056, 3292: 7.47191446617, 3293: -0.0188529890705, 3294: -0.250116225139, 3295: -0.268750473921, 3296: -0.275708142344, 3297: -2.08491978471, 3298: -5.87222493102, 3299: -2.44135886423, 3300: -0.434866124159, 3301: 0.689343379879, 3302: -3.24965153984, 3303: -0.250657573223, 3304: 12.6935734953, 3305: -0.529474296655, 3306: -6.11104946623, 3307: 3.14858558226, 3308: -7.1338250374, 3309: 3.86029889178, 3310: -2.69574870611, 3311: -2.36357553566, 3312: 3.77065548654, 3313: -0.240500890134, 3314: -8.90455633131, 3315: -0.0809173257354, 3316: 4.90367734439, 3317: -9.30146269584, 3318: -9.27703648442, 3319: 0.719271508495, 3320: -0.0639168719782, 3321: -0.345104740807, 3322: 0.00312718053388, 3323: -0.452093536904, 3324: -2.07992026696, 3325: -10.2840996914, 3326: -1.75645672534, 3327: -2.90561021298, 3328: -7.9355050716, 3329: -0.133050064767, 3330: 5.03303041011, 3331: -1.23438335194, 3332: 0.77686111459, 3333: -7.86057504168, 3334: -0.332003118653, 3335: 9.08414097167, 3336: 0.147449065808, 3337: -1.86856624845, 3338: 0.690346630633, 3339: -0.490659438754, 3340: -0.179938273666, 3341: -1.13773876219, 3342: -0.204998004453, 3343: -3.54521867498, 3344: -0.225387455241, 3345: -0.238810728674, 3346: 16.3128148524, 3347: -1.23801390264, 3348: 25.8352517544, 3349: -0.0876229331096, 3350: -0.108738816796, 3351: -1.73056207479, 3352: -0.250954595423, 3353: -2.6343000642, 3354: -10.3641183202, 3355: 1.64390689599, 3356: -0.660748992382, 3357: -0.295747801488, 3358: 2.56319510784, 3359: -5.39594068849, 3360: 8.89730174584, 3361: -8.00414009055, 3362: -1.06069163518, 3363: 6.39644705917, 3364: 1.01496825337, 3365: -1.66359197339, 3366: 2.14346342401, 3367: -0.0599037268964, 3368: 0.231333062866, 3369: -4.89874973448, 3370: -1.5968126186, 3371: -1.63438817492, 3372: -4.7668592916, 3373: -0.341092226135, 3374: 5.77349944852, 3375: -7.58483639473, 3376: 6.46697540971, 3377: -0.283577066527, 3378: -3.20677857052, 3379: -12.0364764429, 3380: -1.15252837277, 3381: -1.12609881564, 3382: 3.89547608488, 3383: 3.54190670789, 3384: 4.48090975453, 3385: -4.14757864404, 3386: -2.03400108434, 3387: 6.96261714835, 3388: -15.5573201864, 3389: -4.21135683485, 3390: 12.1296269307, 3391: 11.6653603046, 3392: -6.17193364274, 3393: 4.38632345711, 3394: -16.4269374607, 3395: -4.10465784809, 3396: -0.0656561807321, 3397: -9.18273662086, 3398: 3.5104435883, 3399: -1.54646191552, 3400: -1.67875925431, 3401: 2.52106509433, 3402: -4.2850906952, 3403: -1.2669949594, 3404: 2.21595811659, 3405: -1.00747766313, 3406: -2.24886368339, 3407: -17.2238270982, 3408: 11.2007144555, 3409: 0.716404025665, 3410: 7.22333666759, 3411: -1.67071127163, 3412: 0.0129530090619, 3413: -2.61625410806, 3414: 2.48779434946, 3415: 4.83904993349, 3416: 0.0191888108353, 3417: -5.67981710065, 3418: 7.88472248391, 3419: 4.42840018631, 3420: -3.71510257192, 3421: -6.87810967222, 3422: 2.75250244921, 3423: -1.07749214622, 3424: 6.04610820937, 3425: -3.40573528628, 3426: 17.4925830061, 3427: -11.2923289416, 3428: -1.52952064168, 3429: 2.31339069757, 3430: -1.29899227017, 3431: 6.50120154344, 3432: 5.71034456838, 3433: -3.84548245733, 3434: 6.64843681765, 3435: -2.33087489152, 3436: -7.66421002732, 3437: -0.912360757364, 3438: 1.98214635122, 3439: -4.9990760795, 3440: -1.38505743214, 3441: 8.42302770385, 3442: 1.3480845424, 3443: -3.21658538603, 3444: 4.23791846283, 3445: -6.89258251185, 3446: -0.295930021219, 3447: -0.437374578557, 3448: -3.52300440214, 3449: -1.15120980762, 3450: -1.00814746034, 3451: -1.91165644983, 3452: 1.84556641073, 3453: -1.13248100667, 3454: -1.30883364457, 3455: -33.0302017634, 3456: -1.16011685194, 3457: 0.793189530464, 3458: -1.39202372629, 3459: -1.40551231678, 3460: 24.4254128506, 3461: 0.400579662876, 3462: -3.78686272105, 3463: 11.2337924005, 3464: 2.69217507302, 3465: 0.128852896441, 3466: 6.12258462522, 3467: 6.55073009289, 3468: 6.66429747589, 3469: 8.71238459491, 3470: 20.0997915653, 3471: -2.36748055837, 3472: -3.26859368643, 3473: 2.0180790604, 3474: -0.946604214078, 3475: 6.53337496306, 3476: 1.45413612232, 3477: -14.043305567, 3478: -3.95287117348, 3479: 12.5651060661, 3480: -1.46818140591, 3481: 6.63489923431, 3482: 8.14869857923, 3483: -10.954818992, 3484: -0.428342433078, 3485: -5.14592274943, 3486: -0.568679133691, 3487: 25.9139589549, 3488: 4.23593407127, 3489: 0.470991563775, 3490: -2.93307017673, 3491: -1.76289609894, 3492: 11.4980091688, 3493: -4.74327031583, 3494: 0.367370672381, 3495: 0.663618664094, 3496: -1.68034312803, 3497: -1.75293588854, 3498: 1.05093329052, 3499: -1.73447788702, 3500: -1.66543621637, 3501: 1.36557976123}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.279699921028, 2: 2.45706109108, 3: 1.09872806894, 4: -0.825138572941, 5: -0.138333933935, 6: 0.337743378013, 7: -3.35928491643, 8: 1.50456790969, 9: -0.647658133342, 10: -0.42841984809, 11: 2.0428819359, 12: -1.11184670326, 13: -2.54324741676, 14: -0.860859784852, 15: 2.38951024034, 16: -0.583312786654, 17: 2.84839287692, 18: -0.858188832358, 19: -14.376705224, 20: 2.52122974375, 21: -2.68648685707, 22: -0.853207350454, 23: -2.89184620315, 24: -2.65549895661, 25: -0.241507549276, 26: 1.20611206891, 27: 5.74400059359, 28: 2.07624729762, 29: -5.80551443729, 30: 3.30148691459, 31: 4.10187316954, 32: -1.79369117081, 33: -2.23430663647, 34: -0.330238222759, 35: -3.90070426362, 36: -0.198261713297, 37: 0.109401913329, 38: -0.427984138772, 39: -0.728660752726, 40: 1.16747885701, 41: -0.30466395395, 42: 0.784734276075, 43: 2.63808729932, 44: -1.70225582488, 45: 3.12543564275, 46: 0.0159354929389, 47: -1.44922087458, 48: 1.39737421935, 49: 0.482886324193, 50: -1.22244810384, 51: -2.93285315913, 52: -2.0566312162, 53: -2.56878357996, 54: 1.55496138096, 55: -3.26077355765, 56: 0.434326112108, 57: 7.12521501019, 58: 1.71995572087, 59: -5.73488467228, 60: -0.605852272373, 61: -5.07728995111, 62: 0.311656878711, 63: 1.50539458546, 64: -0.742538987093, 65: -1.69790976026, 66: 0.208006177071, 67: -4.88372870824, 68: -0.608705289399, 69: 2.35700334218, 70: 1.0475492226, 71: 0.307046122406, 72: -6.06656273802, 73: 1.19590401261, 74: -9.06695859109, 75: 0.410004079445, 76: 1.93892763689, 77: 0.397093652744, 78: -0.168792821712, 79: 1.50746041018, 80: 1.8313590337, 81: -2.61955722464, 82: -5.80095050023, 83: -0.714517047735, 84: -1.2696688562, 85: -0.302907607875, 86: -0.506776114041, 87: -1.56085955189, 88: 0.174274402044, 89: -0.578212042063, 90: -6.08935077532, 91: 2.41441679899, 92: 0.439152387512, 93: -7.58927001238, 94: 0.615221607651, 95: 0.35894761061, 96: 0.853562338861, 97: -3.35406795526, 98: -1.32152153562, 99: -0.396432522972, 100: 0.62455036568, 101: -0.194785789556, 102: -1.31273372056, 103: -0.307763408126, 104: 0.302022549374, 105: -0.409344087031, 106: -2.13883706124, 107: 0.622293072874, 108: 0.490832427612, 109: -0.699759510617, 110: -4.23680477682, 111: 0.484631860402, 112: -0.370131975298, 113: -3.99417379829, 114: -9.71761791883, 115: -6.42921617223, 116: -2.0580173186, 117: -0.264764215176, 118: -0.742057848631, 119: 1.89550719361, 120: 0.840155224108, 121: 2.0100304981, 122: 2.86751458353, 123: 5.35805019935, 124: 2.16290751912, 125: -3.03784834578, 126: 0.129725819278, 127: 0.107424730604, 128: 0.318432796723, 129: 10.597990353, 130: 0.237878154675, 131: 0.833977943647, 132: -2.10038082079, 133: 1.15637972184, 134: 0.233979571641, 135: 0.117036964778, 136: 2.61211486357, 137: 0.841984017639, 138: -4.93524276667, 139: 1.78033351784, 140: 1.09779816852, 141: -2.79026278047, 142: -0.247293839112, 143: 2.818146111, 144: 0.163262696644, 145: 5.79441845924, 146: 0.111498704229, 147: 0.187122291093, 148: -0.427788354743, 149: 0.260026431224, 150: 0.877352739073, 151: -1.28178731341, 152: 0.268089908699, 153: 0.342947459836, 154: 0.229597552621, 155: 0.234846159018, 156: -0.491405658809, 157: -11.1387833162, 158: 0.87297107306, 159: 0.0765111448632, 160: 0.415741353152, 161: 0.590399284856, 162: -0.495982633054, 163: -2.62325383458, 164: -3.98960662397, 165: -4.05405294421, 166: 8.88330037494, 167: 0.587460254832, 168: 1.94581582868, 169: 0.329938102498, 170: -0.052231504753, 171: 2.01800811961, 172: 0.204266018773, 173: -2.05820200032, 174: 0.224996300374, 175: 1.09643523789, 176: 8.40453970549, 177: -8.45732193546, 178: -0.0786482901499, 179: 0.234300023046, 180: 0.3577684221, 181: 0.291085043238, 182: 0.0945929605808, 183: 1.04510515724, 184: 6.08755706574, 185: -1.90684640423, 186: -3.25458837014, 187: -3.39156567785, 188: -0.742429835341, 189: 3.29711006842, 190: -0.0126278189356, 191: 0.055362952497, 192: -8.55896751233, 193: 0.156180151844, 194: 3.4273708593, 195: -0.497801457526, 196: -1.89425036829, 197: 0.737558678631, 198: -1.587168511, 199: 0.190005323286, 200: 0.864400622355, 201: 0.266347543524, 202: -1.50741606138, 203: 0.121831810837, 204: 0.382168107795, 205: 2.74254017902, 206: 0.153936241764, 207: -8.82939253688, 208: 0.186869783775, 209: 0.251652640345, 210: 1.07474627389, 211: 0.290724662834, 212: -0.0330351608624, 213: -1.80434411184, 214: -4.50552132706, 215: 0.0515230962969, 216: 0.389251918185, 217: -6.68048299992, 218: -0.534411252799, 219: 9.46023121766, 220: -2.460539839, 221: -1.2041501239, 222: 0.580633497485, 223: -0.869763180159, 224: 0.560751783412, 225: -3.09212703694, 226: 0.891386990157, 227: 1.30333602488, 228: 0.202419325725, 229: 0.322491018514, 230: 0.508159890219, 231: 0.857312190954, 232: -0.1567052981, 233: 0.533861384337, 234: -5.20508647594, 235: 1.56677315648, 236: -1.22385839515, 237: 2.86946008619, 238: 6.5119005741, 239: -0.201654733845, 240: 0.89686666712, 241: -0.101570257291, 242: -0.0103275433889, 243: 2.68392277659, 244: -0.995709816595, 245: -1.57453395398, 246: -2.25193091352, 247: 0.475806605177, 248: 3.37013386917, 249: 5.97398276156, 250: -0.609119536032, 251: 1.95035688016, 252: -3.4205679148, 253: -0.803464918553, 254: -1.23537656199, 255: 3.38422184639, 256: -0.074295271609, 257: 5.59217754062, 258: 1.14705711363, 259: 0.85008978374, 260: -1.49398450014, 261: -0.753549385849, 262: 0.249473182481, 263: -1.56452824436, 264: 0.304110368482, 265: -0.506673935546, 266: 7.48658240068, 267: -1.22964942997, 268: 0.690070828727, 269: -3.85964770442, 270: 3.29765320195, 271: -0.729573072096, 272: -1.2026218916, 273: 1.03941088995, 274: 3.32349045128, 275: 2.17182607524, 276: 1.19890057503, 277: 0.173931652337, 278: 3.35379571194, 279: 0.27733133853, 280: 7.68009423664, 281: 0.615626901698, 282: -0.404175953645, 283: 4.59449810221, 284: -6.47843569131, 285: 6.82430060034, 286: 1.63364904647, 287: -0.746696309627, 288: 1.07443584659, 289: 0.0822850744585, 290: 3.62676101353, 291: -0.574310306963, 292: -1.58563783603, 293: -2.33585969874, 294: 1.18757263288, 295: 5.66177526946, 296: 4.22513223468, 297: -2.39491422727, 298: -0.505630115423, 299: 0.181849127284, 300: 5.04022934542, 301: -1.9106571538, 302: -4.51707210341, 303: 1.97906548199, 304: -1.4625741114, 305: -2.17874728405, 306: 0.371149775475, 307: -2.98103521742, 308: 0.939416216622, 309: 0.763307411068, 310: 0.850543153379, 311: -4.07428296023, 312: -0.0825822725547, 313: -1.97423524425, 314: 2.96860649856, 315: 0.10411889059, 316: -8.15820277121, 317: -1.17588058638, 318: 1.27802159156, 319: 10.9170215696, 320: 1.12208174085, 321: -0.179070350756, 322: 3.1927415189, 323: -3.70784522369, 324: -4.77595894588, 325: 1.34175935429, 326: 6.72854306531, 327: -0.379381332753, 328: -10.2755031206, 329: 1.10844948382, 330: -3.55562922709, 331: -1.44501512585, 332: -2.31561963867, 333: 0.38934039634, 334: -2.14572470184, 335: -0.839606479308, 336: -2.3975966037, 337: -1.62868176691, 338: 10.2733838493, 339: 0.757854005227, 340: 2.53016933418, 341: 1.95867258814, 342: 6.82971949731, 343: -2.93343399901, 344: 1.26498115955, 345: 1.22882244964, 346: -0.914572451681, 347: -5.2451583001, 348: 0.0687531871299, 349: -5.73113700893, 350: 0.532636501824, 351: 3.1321729705, 352: -7.70358955774, 353: -28.6106053917, 354: -5.89615505859, 355: -2.48812339747, 356: 6.08676429319, 357: -18.7119687587, 358: -9.49756795606, 359: -1.55643043965, 360: 4.26315924699, 361: -15.9398129921, 362: 1.02031597767, 363: 6.74737950758, 364: -9.02358901896, 365: -0.476684517906, 366: -2.19268578743, 367: -35.9517116218, 368: 6.49700803288, 369: 7.36515913749, 370: -9.13874570481, 371: -5.24835000543, 372: -3.53367711831, 373: -2.50065386958, 374: -5.41364045752, 375: 0.384045986797, 376: -3.6362570036, 377: -4.80012431348, 378: -16.0188554216, 379: 1.1337303831, 380: -1.44002862209, 381: -17.1102268971, 382: -11.5467985711, 383: -4.10870310231, 384: 0.327517337395, 385: -4.92032786209, 386: -12.4894805493, 387: -8.78151770791, 388: -3.31511234238, 389: -4.74126619249, 390: 6.29612263161, 391: 2.71928889299, 392: 1.26230788212, 393: 20.6977995357, 394: -6.67689000326, 395: 1.47642465782, 396: 10.7241595769, 397: 0.613355381512, 398: 5.76374143125, 399: 5.0167547334, 400: -3.51290587801, 401: -12.712785782, 402: 2.60216217806, 403: 0.793831229444, 404: 1.5373220974, 405: 3.70909879383, 406: -27.4470315329, 407: 1.45688977201, 408: -28.4747430574, 409: -15.3705953414, 410: -3.85115049904, 411: -20.6530175216, 412: 5.90018308146, 413: -5.45645366023, 414: 10.0670916144, 415: -0.74169606865, 416: -0.459046221763, 417: 2.97404852459, 418: 4.59919932612, 419: 1.02802107064, 420: 3.30331888645, 421: -5.54589097617, 422: 2.55222988872, 423: 4.26614733363, 424: 2.62317571336, 425: -6.87788530244, 426: 2.52291106374, 427: 3.2369216492, 428: 12.3871449648, 429: -0.34440794409, 430: 1.16654730149, 431: -8.61156785534, 432: 5.61498458858, 433: 2.41245548988, 434: 2.03075774605, 435: -0.292015216128, 436: 0.00717644062419, 437: 7.52732916691, 438: 5.03061048435, 439: -3.14524643884, 440: -0.253005577011, 441: 3.49746145681, 442: -7.51824079997, 443: 3.64300754487, 444: 3.99420815238, 445: 0.489667068091, 446: 3.09203412685, 447: 3.43316312118, 448: 5.62044916066, 449: -1.37172291223, 450: -10.5550820215, 451: 3.03049560976, 452: -10.5022532422, 453: 1.0781722196, 454: 1.78307164686, 455: -4.38897702945, 456: 2.34799044678, 457: 2.28197376638, 458: 4.06200694882, 459: -6.60436569945, 460: 5.24600118192, 461: 2.87599621858, 462: 4.6786143282, 463: 15.0627618772, 464: -20.8971724087, 465: 4.31128227804, 466: -3.0228174106, 467: -8.7477170002, 468: 10.0862599904, 469: 9.04410976793, 470: -3.50269530783, 471: 16.4252011832, 472: -15.3223315026, 473: 13.1295856954, 474: -4.36899690503, 475: 0.613240821258, 476: 0.854153001316, 477: 3.78963958918, 478: 0.963605753093, 479: 0.502089140097, 480: 1.91453152567, 481: -0.495093442242, 482: 8.01117672633, 483: 3.22560083369, 484: 1.17503100802, 485: 9.79590376363, 486: 0.0206832279185, 487: 3.32681442562, 488: -0.831520044073, 489: 2.29208638932, 490: -18.1307368571, 491: 0.296824867335, 492: -17.7365428381, 493: 1.37493286907, 494: -3.02556302652, 495: 0.336685359095, 496: 0.594457005042, 497: -1.88906363777, 498: 0.638481688984, 499: 0.643068575436, 500: 1.40105157102, 501: 0.812353738171, 502: 0.633211276501, 503: 0.640594412802, 504: 0.402839854874, 505: 4.16420331163, 506: -13.4799752409, 507: 1.68558883354, 508: 0.507167344165, 509: 1.13698326418, 510: 1.351285723, 511: -0.447582009842, 512: 10.0930915969, 513: -15.5765554387, 514: 16.6968249879, 515: 23.685960018, 516: 1.63206824937, 517: 3.27595853115, 518: 0.480922457096, 519: 1.99539793802, 520: 16.1175491977, 521: 0.631587983736, 522: -12.5081611339, 523: 0.550779433147, 524: 15.7350650461, 525: 4.88348393935, 526: 5.1705858757, 527: 1.60228664126, 528: 0.546614177392, 529: 0.638630495279, 530: 0.635638376519, 531: 0.448141482141, 532: 2.75038851738, 533: -16.8009382575, 534: -0.544355081752, 535: -1.82465529291, 536: 6.80939912115, 537: 4.05627429158, 538: 4.90135274038, 539: 3.17354752599, 540: 5.72244586735, 541: 1.4149048792, 542: 0.697164418794, 543: 2.55274028194, 544: 5.62994256219, 545: 2.41865703438, 546: -3.12464811272, 547: -3.78883537307, 548: 0.658430081936, 549: 2.89005027744, 550: 0.588397587216, 551: -11.1297984085, 552: 0.509449941033, 553: 0.470942882155, 554: -11.9598399541, 555: 1.53981954576, 556: -6.49808743842, 557: 0.546948014525, 558: 0.472457567941, 559: -1.84037329736, 560: 0.447098099329, 561: 5.13999567043, 562: 1.61037496924, 563: -3.13157835858, 564: 1.01333574238, 565: 1.06727223137, 566: 11.6555440026, 567: -1.11923300286, 568: 9.15483869506, 569: -3.63474933627, 570: -0.670864956088, 571: 4.39405640934, 572: -0.864610067069, 573: 1.96557820639, 574: 4.20110974213, 575: 0.522315756189, 576: -9.16153292692, 577: 1.34321312562, 578: 0.14092752746, 579: 2.49130221607, 580: -3.3060280891, 581: 0.635510410483, 582: 4.16443703137, 583: 5.37031846799, 584: 3.61895209731, 585: 3.29107578622, 586: -2.51926159729, 587: 3.09938446341, 588: 1.7125142717, 589: 1.93135311497, 590: -6.93166988173, 591: -2.67620620151, 592: 12.6786119908, 593: -1.63924353427, 594: -2.48891514408, 595: -2.8926630815, 596: -1.83465514057, 597: 2.52192292089, 598: -6.6058461866, 599: 8.890290236, 600: 9.55177091118, 601: 16.5397402811, 602: -19.6280122063, 603: -1.47249918888, 604: 6.45994601621, 605: 2.55383770839, 606: 7.39484895503, 607: 0.276380442305, 608: 0.772547235101, 609: 2.41472828199, 610: -6.83042103617, 611: 2.75686108376, 612: -2.3589920535, 613: -0.413764991088, 614: 3.94956667607, 615: -10.3578527821, 616: 11.5163059478, 617: 5.67532904197, 618: -25.3909957989, 619: 0.355497610921, 620: 0.817532597252, 621: -13.0426696243, 622: -2.22939496322, 623: -27.8896406083, 624: -12.0051385259, 625: 10.471107676, 626: -4.10055398605, 627: -0.640907727496, 628: 1.98633424262, 629: 17.2548340863, 630: -5.85601953366, 631: 5.87471512843, 632: -2.07494614592, 633: -4.99089342278, 634: -18.9718295437, 635: 3.38072942221, 636: 3.87903941914, 637: 0.378724689005, 638: 1.47650088528, 639: 1.76472943491, 640: -1.18587564736, 641: 9.33087839937, 642: 1.85838915859, 643: -4.21525643706, 644: -4.08868614793, 645: -1.3641756614, 646: -3.5074913404, 647: -1.11589692818, 648: 4.21180016221, 649: -3.89800718316, 650: 1.06599625601, 651: -14.3886598793, 652: 7.23007831038, 653: -14.3897541084, 654: -0.718316050254, 655: 3.92198408287, 656: -5.82291722597, 657: 3.00292584505, 658: 2.94098085523, 659: 1.83179317314, 660: 7.92872108656, 661: 0.365640781251, 662: 3.55879793037, 663: 11.6225673087, 664: 1.63923217987, 665: -16.3151866256, 666: 7.07375976604, 667: 2.27493489815, 668: 18.9186917909, 669: 2.74208654768, 670: 2.05267210763, 671: 6.78982372161, 672: -6.95210428326, 673: -9.54898807434, 674: 23.1583334737, 675: -6.74598959886, 676: -3.73073713786, 677: -0.97342046719, 678: -5.20786855276, 679: -1.27606672434, 680: -18.665646736, 681: 11.9175085719, 682: 2.40428023274, 683: -8.12751684957, 684: 0.815985764022, 685: -28.5296760224, 686: 3.13958698531, 687: 20.6916800286, 688: 2.75985311151, 689: 12.0587142942, 690: 3.17982429842, 691: 18.6061902059, 692: -15.7143968195, 693: 2.43742330106, 694: -15.9812140059, 695: -1.71189535486, 696: -7.48848974913, 697: -0.827463469281, 698: -2.20711757945, 699: -0.114674926822, 700: -8.42177804573, 701: -0.175880011099, 702: -6.1764669042, 703: -7.67863682226, 704: -0.169590737365, 705: 1.40194186782, 706: 3.46127656158, 707: -1.01889277735, 708: 1.7710739153, 709: -4.95602308935, 710: -1.27463162086, 711: 0.169498528873, 712: 4.45273081625, 713: 1.19175845963, 714: 0.202794333331, 715: -0.649867945433, 716: -0.285625008592, 717: 14.1451601702, 718: 4.77976207117, 719: 1.19597020826, 720: 0.882391276181, 721: 5.07522996643, 722: 6.44630846206, 723: -1.12403264655, 724: -7.17469300859, 725: -6.65008904779, 726: -6.26403029876, 727: 3.74963555438, 728: -3.48274447008, 729: -1.89466816216, 730: 0.652203730649, 731: 2.46843768056, 732: -0.30140514368, 733: 1.72010170367, 734: 0.743230762995, 735: -1.26664645297, 736: -4.4567196048, 737: 0.330178764754, 738: -1.11739820774, 739: 0.313258515406, 740: -1.50034085869, 741: 5.28267093844, 742: 2.57009026533, 743: -1.73695576575, 744: -0.827964427032, 745: 12.664063009, 746: -0.244483845872, 747: 0.643467897782, 748: -2.71333306578, 749: 2.60499955254, 750: -4.57804401786, 751: 4.929294082, 752: -1.93706596132, 753: 6.0514161575, 754: -1.52277796261, 755: 3.15851144499, 756: -6.36670200552, 757: -3.57877212929, 758: -2.9502413233, 759: -12.079224352, 760: -1.23871409699, 761: -0.0144614648223, 762: -0.230566777279, 763: 2.26916404758, 764: -7.07036531984, 765: -1.02943624545, 766: -4.2764531715, 767: -0.00344142867059, 768: -1.54862131166, 769: -3.7080396211, 770: -1.73217579932, 771: -0.668474299236, 772: 5.01790720413, 773: 1.12224064313, 774: 0.498418936164, 775: -1.43002521071, 776: 0.532812193431, 777: 2.58634778122, 778: -5.76948491189, 779: 4.75980690361, 780: 12.2626933683, 781: -1.34990950295, 782: -6.27463639965, 783: 0.797437690202, 784: 4.93843727316, 785: 0.364565304881, 786: 8.9278464801, 787: -1.33533122587, 788: 5.12754704095, 789: 2.50810729485, 790: -0.559539633509, 791: -0.467780528707, 792: -1.46714029025, 793: -2.10508408486, 794: -0.725798118359, 795: -0.300670969361, 796: -1.40528914897, 797: -2.87588524497, 798: -1.27278996814, 799: 17.5163580477, 800: -0.0129328414528, 801: -3.4901795258, 802: -0.393617489516, 803: -0.327118810025, 804: 0.412343554542, 805: 1.1565323507, 806: -1.2890958758, 807: -0.710016191017, 808: 6.8913668043, 809: 0.270570073455, 810: -0.775052176244, 811: 1.66647112493, 812: 8.01394857216, 813: -5.28959379381, 814: -0.436067528135, 815: -0.803581810004, 816: 1.83819675228, 817: 1.29863612845, 818: 4.01213934412, 819: -9.79556672262, 820: 5.4052815621, 821: 6.05926470588, 822: 6.08955877278, 823: 7.88153728792, 824: -0.326158992572, 825: -0.290799254404, 826: 0.389182132361, 827: -10.1086716473, 828: -0.109400437367, 829: -0.0105530050194, 830: -1.12798394875, 831: 3.55015056252, 832: -0.251565444523, 833: -1.81894899187, 834: -5.99531680008, 835: -0.820256469375, 836: 1.84626862066, 837: -0.805736105524, 838: 1.80477058803, 839: -0.180524919252, 840: 4.47904386005, 841: -5.35099601014, 842: -0.321902491213, 843: -6.81088068167, 844: 0.0518983975302, 845: -0.09088804425, 846: 3.31708376228, 847: -0.188530610016, 848: -1.29367110789, 849: -1.44225921389, 850: -0.184683264635, 851: -0.214913478927, 852: -0.223958565276, 853: -0.166928568279, 854: -1.06569655869, 855: -15.7371447854, 856: -0.22569071558, 857: -0.46863331329, 858: 0.0743644100672, 859: -6.23622913505, 860: 7.02901761329, 861: -12.1620780929, 862: -1.37130259683, 863: 9.06456850723, 864: 1.88979090792, 865: -2.39672162905, 866: 0.508414466356, 867: 3.15665394365, 868: -0.567952675773, 869: -9.39819595243, 870: -0.128575348909, 871: -1.90245919048, 872: -0.125732850266, 873: -5.2403806376, 874: 2.98028493092, 875: 9.08672940292, 876: -1.33424692791, 877: -0.266269291292, 878: -0.2210545791, 879: -0.138033617439, 880: -0.129638526239, 881: -1.03542289948, 882: -6.78065998915, 883: 1.20232402708, 884: 4.77935603003, 885: 4.91717783724, 886: -2.33448492753, 887: 3.05096900069, 888: 1.46520904432, 889: 0.352524549011, 890: 12.2975334077, 891: -0.147613549753, 892: -1.44661379093, 893: -1.25245163774, 894: -2.67695010729, 895: -0.432681887588, 896: -2.49459383106, 897: -0.204723802359, 898: -0.789661273642, 899: -0.123914621194, 900: 7.49474231229, 901: -0.220961804508, 902: -0.231501729874, 903: 5.42403609044, 904: -1.147506453, 905: 2.68841278923, 906: -0.23560929735, 907: -0.113657829436, 908: 0.329050443864, 909: -0.132490155682, 910: -3.51009532258, 911: -6.66171370229, 912: 5.22276034822, 913: -0.373412289657, 914: 0.0279434933257, 915: 0.688716536915, 916: -6.73094892719, 917: -1.50247396794, 918: -0.0727367936764, 919: 1.86002408721, 920: -1.77240589565, 921: 2.13392492241, 922: 0.320478572785, 923: 4.6015609935, 924: 1.26904340788, 925: 4.22440720244, 926: -1.90035496375, 927: -0.469634865759, 928: -1.48468904129, 929: 7.66637490323, 930: -0.0226070210208, 931: 4.27227452321, 932: 8.53076331708, 933: 5.29547295807, 934: 0.142504321467, 935: 0.23779600161, 936: -0.439762590587, 937: 0.852689096422, 938: 0.0754082371525, 939: 0.843130430018, 940: -4.10354526869, 941: 3.56970436847, 942: -0.764505652943, 943: -0.210918117882, 944: 3.16777437187, 945: 4.85599772403, 946: 8.77775546278, 947: -2.15258484655, 948: -0.179441485227, 949: -3.94391663711, 950: 1.26918104371, 951: 18.609101905, 952: 1.26268578213, 953: -3.15553270138, 954: -1.36031308948, 955: -1.25258665086, 956: -0.0269709069231, 957: -1.1567078606, 958: -12.290097776, 959: 6.08244418151, 960: 0.25202820448, 961: -0.858354267631, 962: 2.22919986836, 963: -0.927649820305, 964: 10.6500329478, 965: 16.3097938687, 966: 8.29108793951, 967: -2.92834507034, 968: -1.36517756357, 969: 3.80064359789, 970: 7.75611581528, 971: -0.525586544147, 972: -23.3188907187, 973: -0.364652878703, 974: -1.41216514341, 975: -1.637542294, 976: 5.08586008216, 977: -3.59211794431, 978: -7.21325217776, 979: 3.26176407482, 980: 1.14265628681, 981: -0.334012376004, 982: -4.96703036178, 983: -8.13546671157, 984: -1.94071907918, 985: -0.760115224086, 986: 0.55703333666, 987: 0.509385193601, 988: -0.315296098698, 989: -2.229468098, 990: 4.24296667925, 991: 2.4977512113, 992: -2.67024217419, 993: -8.73618710119, 994: -1.38650625964, 995: 4.58659748027, 996: 4.66294934094, 997: 2.77731391119, 998: -2.97559618242, 999: -5.24512693316, 1000: 0.710317308879, 1001: -1.32857840876, 1002: -7.99240700372, 1003: 15.1067207893, 1004: -1.87821081398, 1005: 8.82928986933, 1006: -0.802608626085, 1007: -0.897091005805, 1008: -0.123906935329, 1009: 13.0869906814, 1010: -1.31334797865, 1011: -1.09248024164, 1012: -15.8151625831, 1013: -1.22204815265, 1014: 11.3451143069, 1015: -1.16537259106, 1016: -0.936467120754, 1017: -1.53239483141, 1018: -0.288401176899, 1019: -5.68383775097, 1020: -11.3417353831, 1021: 3.00316271177, 1022: -1.14706011162, 1023: 10.1706677886, 1024: 8.05173303987, 1025: -1.48923562472, 1026: -4.5504876527, 1027: 1.44380059927, 1028: 2.60192583061, 1029: 0.387601335937, 1030: -3.09029536738, 1031: 1.62501113892, 1032: 3.88374124615, 1033: -4.28005435023, 1034: 5.91575249346, 1035: 4.82727002894, 1036: -8.58445648012, 1037: -0.83631018577, 1038: 3.57175526025, 1039: -1.95998710062, 1040: 2.16686940154, 1041: -8.75469758553, 1042: -2.39266391372, 1043: 1.95858377377, 1044: 4.49020178672, 1045: 3.16745884577, 1046: 4.15187903824, 1047: 1.88423662258, 1048: -1.25356019215, 1049: -26.4647981143, 1050: 18.4865770469, 1051: -3.84154580729, 1052: -25.7214715599, 1053: 8.81853337034, 1054: -11.6880187269, 1055: 16.2751802433, 1056: -6.94040180258, 1057: 5.61549137284, 1058: -14.3662228604, 1059: 12.4740463049, 1060: 28.7244044101, 1061: 33.7027496298, 1062: 0.334404647538, 1063: 6.21079995821, 1064: 45.839465383, 1065: 31.4549493047, 1066: 7.06914749888, 1067: 21.1927472583, 1068: 16.3552957547, 1069: 10.9984361611, 1070: -26.4275602186, 1071: -11.2319370307, 1072: -7.5934135333, 1073: 43.0370599231, 1074: -3.25395161288, 1075: -13.7170060081, 1076: -15.2704507362, 1077: -7.60629842479, 1078: 2.41739756073, 1079: -21.4745471659, 1080: -21.596061915, 1081: -6.00492097635, 1082: 6.18873077601, 1083: 1.79107812489, 1084: 1.31139549017, 1085: 0.623703493857, 1086: -0.925630238381, 1087: 1.77054889586, 1088: -10.4011046494, 1089: -8.75292390469, 1090: -13.3739008044, 1091: 14.03262571, 1092: 12.2028956496, 1093: -0.744895756851, 1094: 2.7285907021, 1095: -10.66674462, 1096: 11.1695334271, 1097: -5.89942103398, 1098: -1.30099653199, 1099: -17.3437092848, 1100: -8.07239963249, 1101: 6.06050403111, 1102: 16.8792086032, 1103: -9.10659733495, 1104: 4.9999159828, 1105: -0.94140366731, 1106: 4.32262308208, 1107: 14.4882282085, 1108: -38.618942962, 1109: -1.05499120875, 1110: -2.03422298461, 1111: -6.7920254206, 1112: 14.5156399135, 1113: -0.0683371523584, 1114: 16.2121342754, 1115: -5.12232804202, 1116: 10.0183893063, 1117: -7.03573796807, 1118: -4.05026108155, 1119: -4.9598000882, 1120: -8.89504190131, 1121: 9.12999376046, 1122: -14.2181128436, 1123: 0.263098174434, 1124: 4.51196315568, 1125: -4.84999527839, 1126: -18.9152057834, 1127: 44.3476449503, 1128: -10.0694583667, 1129: 23.952561726, 1130: -6.60956265846, 1131: -0.572565129508, 1132: -7.90636755592, 1133: -15.8397036571, 1134: -8.82307617254, 1135: 7.76571487876, 1136: -8.79195602916, 1137: -5.03353206223, 1138: -7.99063304924, 1139: -6.50482163063, 1140: 5.16069857918, 1141: -9.14416194381, 1142: -4.93404019625, 1143: -2.5850149152, 1144: -8.8961511988, 1145: -8.19528310896, 1146: -11.2789450663, 1147: -9.46438272399, 1148: 30.436027104, 1149: -1.28882765102, 1150: 10.8198801942, 1151: 1.04323500296, 1152: -5.71409384683, 1153: 29.6065742564, 1154: -9.3505117849, 1155: 8.86874427631, 1156: -5.10965696048, 1157: 7.47610829375, 1158: -0.0767751155918, 1159: 1.70689997135, 1160: -41.0188271997, 1161: 43.7261913003, 1162: -9.72353275669, 1163: -7.6347835822, 1164: 2.30735745301, 1165: 9.81393161258, 1166: 37.9240220901, 1167: -12.51109555, 1168: -19.3685230138, 1169: 3.27987405636, 1170: 11.910621804, 1171: 0.0286015866434, 1172: 24.8492281149, 1173: -1.62596988505, 1174: -1.58514935162, 1175: 11.992200698, 1176: -18.4234828232, 1177: -1.44612370139, 1178: -9.25765237417, 1179: -2.95864132675, 1180: -3.23636063595, 1181: 5.02720007174, 1182: -3.63431428012, 1183: 8.1398283485, 1184: 0.981185838353, 1185: 0.14681038539, 1186: -3.2887672254, 1187: -9.4735071946, 1188: 10.4445707642, 1189: -1.24899775722, 1190: -10.6614708381, 1191: -1.34826179923, 1192: 0.669339469028, 1193: -2.0455968486, 1194: -1.60332623381, 1195: 1.85158011036, 1196: -1.87069473826, 1197: -5.89943015985, 1198: -10.3415274675, 1199: -1.91963702853, 1200: -1.53892314568, 1201: -1.59757125188, 1202: -1.55280076276, 1203: -1.60814359174, 1204: 13.3565165476, 1205: 0.0579033445068, 1206: -1.2362117179, 1207: -2.12124351315, 1208: -4.79210162872, 1209: -3.74024206309, 1210: 15.690913156, 1211: -6.1339169606, 1212: -14.7824510788, 1213: -31.2391853446, 1214: 7.01447849746, 1215: -14.3148805958, 1216: 18.3840035469, 1217: 7.90631145828, 1218: -2.75140049018, 1219: -1.93675540611, 1220: -28.9391822061, 1221: -1.7196166177, 1222: 23.4634866918, 1223: 31.2934238713, 1224: 20.5923247853, 1225: 0.189234285452, 1226: -1.6804742091, 1227: -1.66271795688, 1228: -2.22973316686, 1229: -3.34451041921, 1230: -0.478272522776, 1231: -0.866745931172, 1232: 21.8339915269, 1233: 15.2056998989, 1234: -10.4263137763, 1235: -7.76146037616, 1236: 24.5793205283, 1237: -1.21699076634, 1238: -14.6026358995, 1239: 5.13845503076, 1240: -1.55730197402, 1241: 13.0781095355, 1242: -2.63790764874, 1243: -3.17746195326, 1244: -2.74297956846, 1245: 17.4444623456, 1246: -1.67957589193, 1247: -7.89532506604, 1248: -1.56266895632, 1249: 0.639873650817, 1250: -1.7361325784, 1251: -1.65631821727, 1252: 4.83970740304, 1253: -8.9605145087, 1254: 14.7211175798, 1255: -1.31110717697, 1256: -1.65112545326, 1257: -2.499232906, 1258: -1.4113857395, 1259: -3.16195945095, 1260: -1.5587989426, 1261: -16.9561405207, 1262: -1.86623460948, 1263: -2.46485443745, 1264: 19.5974518326, 1265: -6.56074950435, 1266: -62.42938209, 1267: 3.15936048557, 1268: 10.7730388013, 1269: 16.3836626518, 1270: 7.48489840582, 1271: 7.70914756177, 1272: 5.85636276383, 1273: -2.27323792857, 1274: 39.479267362, 1275: -8.9408489628, 1276: 0.68723642936, 1277: -0.0769647883251, 1278: -6.37157790463, 1279: -1.82932442611, 1280: -10.6365699127, 1281: -28.028429812, 1282: -9.03474048064, 1283: 2.93983414471, 1284: -9.14837439186, 1285: -36.0667341837, 1286: -9.08399491085, 1287: -9.30549658696, 1288: 30.1666770827, 1289: -4.64430227801, 1290: 16.9539289398, 1291: -7.79760357856, 1292: -4.70488835342, 1293: -13.5581532913, 1294: 29.4401571093, 1295: 100.0, 1296: -22.0588491128, 1297: 4.3450236113, 1298: 7.04430623862, 1299: -6.01669559997, 1300: 20.019100416, 1301: 7.84749717262, 1302: -21.6920349113, 1303: -4.74434122029, 1304: 11.1957256741, 1305: 1.39753119752, 1306: -5.76264155562, 1307: -3.26806939102, 1308: -40.0368820975, 1309: -8.31243712444, 1310: -27.4280738221, 1311: -4.14878030837, 1312: -1.41335136203, 1313: 11.4604776443, 1314: 0.882747783368, 1315: -17.5958094675, 1316: 8.10819902436, 1317: -5.25694888514, 1318: -1.9188035679, 1319: -26.8761216566, 1320: 11.2043914079, 1321: 36.2901747445, 1322: 2.43414844166, 1323: 21.7790990799, 1324: -15.6522644977, 1325: 1.06763666196, 1326: 17.4671274096, 1327: -34.2929951531, 1328: -1.03184365576, 1329: 1.77051557181, 1330: 7.06254358159, 1331: -29.0786796298, 1332: 6.25469656779, 1333: -19.3482509075, 1334: -8.63045213766, 1335: -9.9395984604, 1336: -8.8584936144, 1337: -4.22586768625, 1338: -6.08975077148, 1339: 8.55412298198, 1340: 15.9024900529, 1341: -4.31860520045, 1342: -4.61087188635, 1343: -9.87027505286, 1344: 15.6708349831, 1345: 13.3672917061, 1346: -3.77762227137, 1347: -5.12000251231, 1348: -4.75270767691, 1349: -3.16307626562, 1350: 5.3839354186, 1351: -7.8094656339, 1352: 65.8458241007, 1353: -4.74144142046, 1354: -10.6145444722, 1355: -7.972896923, 1356: -7.93421338181, 1357: -9.24717330691, 1358: -41.6656185332, 1359: -0.566110211583, 1360: -6.70309950499, 1361: -0.853574950038, 1362: -9.07198716203, 1363: 6.80306352944, 1364: -11.9315447945, 1365: -8.89692767481, 1366: 5.85980988944, 1367: -9.70381630766, 1368: -0.941135465764, 1369: -24.0290649546, 1370: -9.77023495897, 1371: 16.5879785161, 1372: 5.18584018868, 1373: 7.98472093708, 1374: -1.30172642723, 1375: 4.70463348248, 1376: -1.25003290295, 1377: 2.63332692593, 1378: 14.1030998557, 1379: -9.26382167728, 1380: 43.8295391222, 1381: -9.28767971446, 1382: 1.64705130372, 1383: -12.4760213475, 1384: 5.94522342929, 1385: -0.304386424964, 1386: -8.24348322409, 1387: 25.4154603097, 1388: 30.2326095015, 1389: -8.78224943757, 1390: -31.6763501681, 1391: 21.6828127286, 1392: 13.1901473021, 1393: -10.4278348222, 1394: 1.68494424989, 1395: 9.83570360546, 1396: 25.7573044765, 1397: 0.46186017278, 1398: 5.31706680022, 1399: -0.648736880899, 1400: -3.00769270104, 1401: -12.9128064129, 1402: 1.29267775322, 1403: -2.10894961419, 1404: 7.27537646204, 1405: -4.61697264059, 1406: 3.41636385065, 1407: 1.25871171319, 1408: -0.252660858472, 1409: -2.06629192359, 1410: -1.29482823573, 1411: 5.67380460455, 1412: -0.974660995841, 1413: -17.9004127244, 1414: 4.54893503641, 1415: -12.6542730123, 1416: 7.35472581968, 1417: 2.0063009821, 1418: 5.84691868956, 1419: 7.96475827157, 1420: 6.30036765803, 1421: 6.77638802474, 1422: -9.57683339788, 1423: 10.2161242884, 1424: -0.671005890983, 1425: 4.81754509016, 1426: -2.80850388705, 1427: -0.93255939844, 1428: 1.27141443119, 1429: 5.69841566308, 1430: -3.86050977208, 1431: -8.03836419493, 1432: -1.70213521573, 1433: -3.30732300679, 1434: -10.3770994443, 1435: 3.4764053464, 1436: -10.5108031293, 1437: 1.18430424234, 1438: 1.40301043421, 1439: 3.76590587769, 1440: 6.14165826252, 1441: 7.09572605495, 1442: -6.04203982958, 1443: -9.24807592065, 1444: 1.13202556499, 1445: -3.58731263112, 1446: -0.337880945525, 1447: 3.86850223143, 1448: -14.6458681857, 1449: 0.851084575655, 1450: 5.97852122992, 1451: 2.06366339862, 1452: 1.30743672051, 1453: -15.3483640557, 1454: -4.87790918099, 1455: 10.4814058886, 1456: 4.79464734711, 1457: -14.3896419692, 1458: 9.00469059312, 1459: 0.760046049238, 1460: 0.253176909396, 1461: 3.0395220504, 1462: 0.254977527118, 1463: -1.90156438481, 1464: 1.47803150088, 1465: -2.07605946921, 1466: 1.83840572331, 1467: 1.51376728735, 1468: 1.3958709918, 1469: 1.32791036042, 1470: 5.05691133463, 1471: -1.31179985471, 1472: -0.191008428946, 1473: -1.95419231517, 1474: 3.57555346427, 1475: 1.3423831171, 1476: -23.7923596796, 1477: 0.155907427964, 1478: -7.57374703267, 1479: 7.51103036837, 1480: 3.82381368288, 1481: -4.61831445583, 1482: 2.2376895583, 1483: 0.26752190147, 1484: 7.64687085272, 1485: 0.939207731486, 1486: -2.36052747228, 1487: 0.914774710374, 1488: 1.27764040307, 1489: 5.5092582284, 1490: 1.55995794577, 1491: 2.02206888594, 1492: 3.42255592842, 1493: -2.38087447256, 1494: 1.72569141117, 1495: 10.7402923385, 1496: 1.64771894161, 1497: 2.30591754488, 1498: 1.40155028166, 1499: 2.19216212, 1500: 1.7386288581, 1501: 1.22562735744, 1502: 3.35114501461, 1503: 1.17934193894, 1504: 0.768051117347, 1505: 1.37701535065, 1506: -1.34536143673, 1507: 2.22950599215, 1508: -2.7618735394, 1509: 15.692449462, 1510: 11.5410190013, 1511: -6.57221435949, 1512: -5.35949872372, 1513: 4.3523665052, 1514: -12.5364183113, 1515: -15.7422595301, 1516: -4.66843281845, 1517: 8.45733520164, 1518: 1.93745942753, 1519: 9.92031149844, 1520: 3.22460529716, 1521: -14.1816407296, 1522: 0.467900389303, 1523: 0.789445992101, 1524: 0.262298840986, 1525: 20.1511074924, 1526: 0.35851452005, 1527: 0.352395317538, 1528: 0.443389060435, 1529: 5.10706201784, 1530: 1.66566816577, 1531: -2.37127611569, 1532: -2.67812669968, 1533: 0.212279135558, 1534: 6.19897901592, 1535: 0.356095100839, 1536: 7.47615269106, 1537: -17.5029462986, 1538: -2.38237913851, 1539: 2.80804601729, 1540: 1.03746166945, 1541: -0.542149694189, 1542: 0.249112656672, 1543: 0.362323840264, 1544: 4.61962123855, 1545: 0.266474783908, 1546: 1.92263747566, 1547: -9.29406201599, 1548: 0.407180044895, 1549: 0.35655221106, 1550: 0.530884939233, 1551: 0.276465669494, 1552: 1.46359229189, 1553: -22.1256330673, 1554: 0.271621569729, 1555: 0.837991891659, 1556: -0.885369460194, 1557: 0.311635542105, 1558: 1.85589426121, 1559: -44.2757677209, 1560: 6.40354328439, 1561: -4.94847598403, 1562: -46.2696821392, 1563: 1.70127779269, 1564: 3.73485470164, 1565: -5.56070332652, 1566: 0.358033704331, 1567: 0.915950373162, 1568: 0.585821522149, 1569: -2.26276700826, 1570: 0.471157646422, 1571: 1.95874864676, 1572: 8.71620077452, 1573: 6.44030418334, 1574: -0.471600423953, 1575: 0.469022918177, 1576: 0.357166850343, 1577: 0.313485273491, 1578: 0.251978894137, 1579: -1.58185401972, 1580: -7.27739504902, 1581: -5.97541982901, 1582: 3.52914394645, 1583: 4.50092487241, 1584: -0.895888878415, 1585: -2.49748391164, 1586: -0.643416871669, 1587: 2.8655143683, 1588: 13.177488305, 1589: 0.356891945444, 1590: 4.23194530832, 1591: 4.18012463146, 1592: 0.942964026266, 1593: 4.11353515446, 1594: -4.57033565627, 1595: 0.35956640361, 1596: 1.50089601145, 1597: 0.432780946392, 1598: -2.64664679824, 1599: 0.388581282036, 1600: 0.47119196187, 1601: -1.08119390691, 1602: 1.25452764458, 1603: -19.8275546852, 1604: 0.440027991169, 1605: 0.463691615759, 1606: -1.51902366005, 1607: 0.288655930305, 1608: 0.146949843126, 1609: 1.16903162886, 1610: 3.41964372296, 1611: 1.06798910795, 1612: 0.190805191803, 1613: 1.98413354055, 1614: 0.398658146156, 1615: 37.0688116782, 1616: 3.84963909482, 1617: -4.72843523969, 1618: -4.47923189704, 1619: 4.60015903214, 1620: 0.437962847583, 1621: 9.92081777891, 1622: 0.426699992482, 1623: 7.42914846724, 1624: 1.93894016231, 1625: 2.66567824643, 1626: 5.85905839134, 1627: -17.1224940845, 1628: 0.851641942225, 1629: -14.7825585782, 1630: 17.6618354722, 1631: 0.136370605323, 1632: 1.78015665104, 1633: 3.04514744714, 1634: -1.92815092069, 1635: 0.401204925758, 1636: 0.248044908435, 1637: -2.63895848922, 1638: -2.01955821508, 1639: -6.14997615316, 1640: 3.63802493456, 1641: -0.510229114, 1642: 9.29935469855, 1643: -1.92500773322, 1644: -14.6113807624, 1645: 47.7388913469, 1646: 17.5299552769, 1647: 18.7166896953, 1648: -1.72844635241, 1649: -6.08541935887, 1650: -2.57618302273, 1651: 1.55261406842, 1652: -0.673550749199, 1653: 4.5077439357, 1654: 1.44979570854, 1655: 1.95140656948, 1656: -2.46775977784, 1657: 9.34666393302, 1658: 1.80898460505, 1659: -0.519273612759, 1660: 2.27032751328, 1661: 1.4113895448, 1662: 21.4343589564, 1663: 1.55459025723, 1664: 2.42964284382, 1665: -8.11663430908, 1666: -1.0323303789, 1667: 1.64461501055, 1668: 3.70632511999, 1669: 0.431721442067, 1670: 21.2489999961, 1671: -22.5357838858, 1672: 0.858158263782, 1673: -5.35837498354, 1674: -9.70232567807, 1675: -0.801067563168, 1676: -0.526649565538, 1677: -4.37102974648, 1678: 5.24302169401, 1679: 2.50675622936, 1680: -16.1393825468, 1681: -6.47128577342, 1682: 13.1215426978, 1683: 1.48418968705, 1684: 5.18939636138, 1685: 0.0677010500292, 1686: 3.84000047075, 1687: 0.798976295596, 1688: 1.12792969915, 1689: -2.97283556913, 1690: -9.01466767224, 1691: -5.68855191911, 1692: 3.46262073523, 1693: -8.57164773523, 1694: 2.02133925336, 1695: 2.80670896603, 1696: 5.28501538546, 1697: 0.0794521405921, 1698: -0.54019952685, 1699: -3.68427265871, 1700: -14.4891754201, 1701: -1.19148030382, 1702: 2.01581414503, 1703: 4.89857410655, 1704: 1.52580398259, 1705: 1.5277404927, 1706: 1.85341406039, 1707: 9.83742078271, 1708: 0.83058587214, 1709: 1.33319291594, 1710: 5.09856070377, 1711: 1.38756810956, 1712: 22.9201664458, 1713: 1.29125997211, 1714: 1.1776423173, 1715: -20.014891449, 1716: 0.584297346267, 1717: -0.658688659795, 1718: 10.1913490872, 1719: -12.3363526829, 1720: -12.3542509932, 1721: -1.71625435412, 1722: -13.0476761721, 1723: -5.8720325822, 1724: -23.5193122876, 1725: -5.15707545327, 1726: -4.46628099131, 1727: -16.0661593601, 1728: -13.1830727677, 1729: 3.28839266306, 1730: -10.9491404525, 1731: -3.66538218054, 1732: 13.1330493478, 1733: 1.26797290047, 1734: 5.14223205321, 1735: 1.47798591333, 1736: -13.3899502438, 1737: 3.87341882942, 1738: -2.82060941576, 1739: 1.86438265253, 1740: -0.0692984821002, 1741: -9.67382118443, 1742: -21.6071233602, 1743: -27.6498365276, 1744: -2.32159666276, 1745: -11.5762173571, 1746: 1.21709369389, 1747: 1.64221107301, 1748: 16.7464389623, 1749: 12.5299094615, 1750: -29.570066453, 1751: -39.7087023573, 1752: -24.2882993475, 1753: 28.8408393709, 1754: -12.5991219976, 1755: -29.4350048657, 1756: -28.3158482125, 1757: -42.3865817908, 1758: 8.56891866816, 1759: -2.00780701224, 1760: -2.84459738976, 1761: 7.11134724201, 1762: -20.0542801802, 1763: -21.7811133356, 1764: -48.8243039761, 1765: 30.3421976435, 1766: -10.2863128835, 1767: -50.0330428923, 1768: 2.0117348166, 1769: 7.07564244788, 1770: -50.6374932156, 1771: -54.2282724627, 1772: 12.8465867307, 1773: 12.7408248639, 1774: 22.9972551799, 1775: 35.9382852118, 1776: -0.799943177687, 1777: 5.1036413714, 1778: 22.8930985937, 1779: 11.1462625024, 1780: 31.570613677, 1781: -1.51658352885, 1782: 3.47861048549, 1783: -40.1405888249, 1784: 10.3032515153, 1785: 12.1477276688, 1786: -23.0117409012, 1787: -31.423169287, 1788: -14.7308601334, 1789: -5.96571904276, 1790: 33.6340939769, 1791: 3.78731743228, 1792: -15.8114544047, 1793: -21.782925582, 1794: -10.9886971494, 1795: -11.4597501527, 1796: -25.306103963, 1797: 48.3051844529, 1798: 5.80112675232, 1799: -12.8168016232, 1800: 7.33436757423, 1801: 4.74343487361, 1802: 44.2429856112, 1803: -3.9585666178, 1804: -25.9190660759, 1805: 2.04732005876, 1806: -5.809516634, 1807: 8.80293160075, 1808: 16.4404195052, 1809: 31.0378091484, 1810: -28.2457675196, 1811: 4.35657003283, 1812: -12.550650203, 1813: 8.90045052098, 1814: 23.7720786831, 1815: 3.73388705264, 1816: 9.3935352837, 1817: 28.6131577576, 1818: 6.37798193953, 1819: -79.0683344511, 1820: 15.5121392445, 1821: 2.54421060485, 1822: 11.8441292394, 1823: 8.09793527723, 1824: 97.513040782, 1825: -21.5341958532, 1826: -18.9708792855, 1827: -23.6043287545, 1828: 8.26125094587, 1829: 6.13514200887, 1830: -9.99012189007, 1831: 1.06675521122, 1832: -7.93037999051, 1833: -6.64985507892, 1834: 8.61579905239, 1835: -3.78743510484, 1836: -4.31181259488, 1837: 6.22419926546, 1838: -47.2454292665, 1839: 6.84965763054, 1840: 7.73222504315, 1841: 7.86877704753, 1842: -6.05097560492, 1843: 8.15285811941, 1844: -12.2822503473, 1845: -5.52452923696, 1846: -6.76463659667, 1847: -9.00292958474, 1848: 29.5021231583, 1849: 4.60738524483, 1850: -4.78923433936, 1851: -22.1254531336, 1852: 7.34349269846, 1853: 7.44664663271, 1854: 9.5966938532, 1855: -18.0098716668, 1856: -18.7182774539, 1857: -18.563686523, 1858: 39.9526745312, 1859: -34.7368968162, 1860: 12.624460665, 1861: -33.9262490608, 1862: -20.0328644641, 1863: -22.8989935011, 1864: 63.0747418353, 1865: 30.8158558301, 1866: -5.47269501963, 1867: 16.4853346201, 1868: 18.9490251897, 1869: 27.9086393145, 1870: 28.3377090251, 1871: 2.07001222166, 1872: 2.8262741975, 1873: 35.7946519939, 1874: 21.7409114724, 1875: 1.23713504166, 1876: 7.9948190823, 1877: 4.10785666614, 1878: 63.3285723917, 1879: -0.493911108749, 1880: 11.9819171078, 1881: -13.6592630096, 1882: -0.723509750402, 1883: 41.7918746664, 1884: -8.43449668022, 1885: 2.40398638337, 1886: 60.3443128865, 1887: 29.6235976414, 1888: 37.2644964901, 1889: 4.51388669464, 1890: 11.4881906821, 1891: 0.968563480701, 1892: 1.24182688648, 1893: -18.7206521591, 1894: 1.37325121101, 1895: 7.50548659021, 1896: 7.61555414031, 1897: 1.17572057866, 1898: 1.3793558085, 1899: 1.59291355429, 1900: 1.53237568006, 1901: 11.1973942732, 1902: 24.6784802795, 1903: 1.73971065303, 1904: 0.918006372506, 1905: 1.67648719232, 1906: 8.78957124968, 1907: 21.0431063746, 1908: 17.5180892601, 1909: 8.85824357687, 1910: 73.3510082233, 1911: -100.0, 1912: 0.830634289544, 1913: 9.41706863664, 1914: -11.8505089345, 1915: 4.09383181754, 1916: 14.1724357701, 1917: 0.984113780575, 1918: -2.45225327179, 1919: 1.4865527304, 1920: -100.0, 1921: -6.48694171061, 1922: -23.329174046, 1923: 7.02980557168, 1924: 1.31816728348, 1925: 1.17044172287, 1926: 0.532210650977, 1927: 1.00671327969, 1928: 6.90795363931, 1929: -25.2587241157, 1930: -12.9572655472, 1931: -4.04573342424, 1932: -65.1527147666, 1933: -33.3502809674, 1934: 8.67335572987, 1935: -0.789253368135, 1936: 2.43064517938, 1937: 9.77508028098, 1938: 1.78097184907, 1939: 90.3315714815, 1940: 7.47576216312, 1941: 1.18284444515, 1942: -0.933647216538, 1943: 29.193994805, 1944: 1.21290959698, 1945: 6.65250528396, 1946: 1.18771187583, 1947: -35.6781723107, 1948: 1.28017084025, 1949: 1.09897054368, 1950: -17.2306349739, 1951: 6.64675728511, 1952: 4.19996624916, 1953: 1.12427077449, 1954: 1.24076559494, 1955: -3.21457199044, 1956: 1.25110615567, 1957: 2.90634057478, 1958: 5.11764747157, 1959: -48.7211947707, 1960: 1.46582005008, 1961: -5.63533504793, 1962: 13.8246655401, 1963: 8.41819775547, 1964: -61.6834520589, 1965: 8.28608614898, 1966: -8.63286046508, 1967: -21.6991833831, 1968: -3.0623985636, 1969: 5.4103056567, 1970: -2.19205520662, 1971: -2.29343527229, 1972: -35.0292660289, 1973: 18.3265051869, 1974: 12.8563529077, 1975: -14.0508303792, 1976: 56.8164665814, 1977: 2.03261466009, 1978: 92.3814932101, 1979: -68.0723231949, 1980: -20.6518894037, 1981: -28.1401648201, 1982: 10.6064075658, 1983: 10.5721080169, 1984: 4.62212296502, 1985: 8.11682980537, 1986: 24.2489124395, 1987: -49.9251773938, 1988: -34.62530968, 1989: 42.7348693502, 1990: -0.908823736855, 1991: -5.00409265802, 1992: 13.5724520793, 1993: -18.9310779208, 1994: -1.08036726331, 1995: 8.60283030339, 1996: -28.01025058, 1997: -46.795508755, 1998: 51.5500769385, 1999: -15.3912823235, 2000: 2.0260574805, 2001: 7.18642470627, 2002: -2.65813668953, 2003: 6.00539216339, 2004: 7.55181188935, 2005: 7.53912883945, 2006: 7.47708494174, 2007: 5.2583066776, 2008: 1.91878741532, 2009: -9.91202057255, 2010: 11.2414269571, 2011: 21.6283479329, 2012: -3.9068699653, 2013: 46.5639034043, 2014: -0.503406755148, 2015: 2.38221953645, 2016: 11.178076858, 2017: 26.5205485606, 2018: -1.01589455741, 2019: -72.7660569467, 2020: 67.8361264194, 2021: -51.1506862031, 2022: 17.9556011963, 2023: -5.66276270706, 2024: 2.77468755057, 2025: 16.537158505, 2026: 7.91421897139, 2027: 3.04004367607, 2028: 27.818823214, 2029: 46.7736455496, 2030: -7.8463725104, 2031: 5.79110214826, 2032: 5.17243300242, 2033: -1.92838362816, 2034: 13.538012069, 2035: -2.56966278357, 2036: -8.74101320383, 2037: 9.32672185019, 2038: -35.4141204311, 2039: 13.1008483999, 2040: 47.0688243082, 2041: 4.23795446004, 2042: 14.7322886098, 2043: -11.4309525566, 2044: 3.09599394368, 2045: 2.05437101439, 2046: -3.91748711699, 2047: -52.2306966645, 2048: -17.9011798224, 2049: 3.27781989102, 2050: -0.208846885103, 2051: 7.73403083074, 2052: -40.7621332097, 2053: 6.65546099312, 2054: 6.53902745448, 2055: 7.26727049531, 2056: 7.33114805547, 2057: 10.5027883592, 2058: 9.6076108143, 2059: 26.2882751091, 2060: 6.78879883572, 2061: -98.983813438, 2062: 9.84392847941, 2063: 8.01569984048, 2064: 32.5785906761, 2065: 8.80618906197, 2066: 3.00160921458, 2067: 8.67131473296, 2068: 21.2573785337, 2069: 0.357004047807, 2070: -80.2912341439, 2071: -100.0, 2072: 14.5101223725, 2073: -31.468373333, 2074: 6.50283229161, 2075: -34.4049798967, 2076: 22.4966459339, 2077: -12.2326600768, 2078: -5.58372577149, 2079: -14.9985644901, 2080: -35.6942679837, 2081: -12.5262649191, 2082: -10.4052625855, 2083: 13.4238315635, 2084: 6.04405312086, 2085: 44.6611449888, 2086: -0.91499227045, 2087: 31.3644895405, 2088: -45.8360589403, 2089: -10.5461532891, 2090: 0.452047606341, 2091: -13.9307476152, 2092: -36.470962985, 2093: -9.70264491413, 2094: -18.4455087094, 2095: -0.755027538272, 2096: -54.6406983348, 2097: -4.07390435114, 2098: -1.78949602125, 2099: -27.6867013785, 2100: 20.0559813273, 2101: 27.7960762013, 2102: 88.8043458258, 2103: 8.13404170692, 2104: -5.49428163615, 2105: 14.2176855538, 2106: -10.7306756531, 2107: -23.1449251416, 2108: 1.76746852164, 2109: 19.1089328287, 2110: -10.1150934298, 2111: -13.7955276456, 2112: 34.4784350145, 2113: -44.8862931545, 2114: -33.9500539634, 2115: 23.158509349, 2116: -6.38483483041, 2117: 25.4703742496, 2118: 35.379556104, 2119: 60.6774895877, 2120: 2.78432025231, 2121: -1.229280775, 2122: 1.59760854163, 2123: 50.9278015963, 2124: -90.1783371371, 2125: -24.8243583234, 2126: -31.1601870562, 2127: 8.17344468053, 2128: -13.5060258501, 2129: -61.7259425463, 2130: -19.8595149728, 2131: -7.04510617847, 2132: 13.8217145705, 2133: -61.5203219635, 2134: 32.4233295246, 2135: 20.7176631106, 2136: -5.36797776917, 2137: -3.08974768245, 2138: -41.4630751589, 2139: 16.4417027466, 2140: -12.6592705788, 2141: 3.6318387736, 2142: -4.51331972001, 2143: -23.3155251758, 2144: -13.9728943371, 2145: 48.5726008517, 2146: 17.6669932201, 2147: -6.04386993588, 2148: 19.3518061583, 2149: -13.3171980728, 2150: -5.27672576401, 2151: 34.712884539, 2152: -6.64109829722, 2153: 10.9695800326, 2154: 41.4594797921, 2155: -32.1452344183, 2156: -28.3614045147, 2157: -9.76837014523, 2158: 22.0078786385, 2159: 43.0722582512, 2160: 5.9114275615, 2161: -8.77772114843, 2162: -5.11002381921, 2163: -16.3437532037, 2164: -5.58590584632, 2165: -1.99517429406, 2166: -9.98879975272, 2167: -5.66715451703, 2168: 7.20966309935, 2169: 12.788969124, 2170: -10.5309247665, 2171: 11.6668060507, 2172: -3.03749433239, 2173: -87.410782213, 2174: 6.27913915073, 2175: -7.83356651258, 2176: 30.4093640741, 2177: -4.68429001005, 2178: 11.8355894706, 2179: 39.2219818225, 2180: -14.1566965239, 2181: 37.0171719429, 2182: 11.3218420353, 2183: -5.44694048603, 2184: 31.3651513547, 2185: 5.72198924544, 2186: -5.63013200314, 2187: 23.2333172101, 2188: 5.82773570885, 2189: -5.10446799194, 2190: -11.6559987662, 2191: -43.9644138449, 2192: -4.70005852401, 2193: 18.4716503971, 2194: -9.27924573211, 2195: 72.1640690313, 2196: 24.9180741708, 2197: -18.1259364543, 2198: -7.55744434661, 2199: -4.67433611065, 2200: 3.64419852589, 2201: -5.23923674986, 2202: -6.12311407267, 2203: -4.90318968075, 2204: 46.4658262035, 2205: 6.57903253498, 2206: 37.5454237043, 2207: -100.0, 2208: -13.5324994161, 2209: -33.4676996637, 2210: 34.7066932705, 2211: 11.8123520229, 2212: 15.1828269604, 2213: -49.7789432909, 2214: -17.5702547855, 2215: 39.2344607975, 2216: -44.7161412875, 2217: 28.486299907, 2218: -67.9406199836, 2219: -26.0276925503, 2220: -0.836422462125, 2221: -0.289384399484, 2222: 0.187203754048, 2223: 33.0921533533, 2224: -0.812285783038, 2225: -4.1538861348, 2226: -0.705058428891, 2227: -42.846769962, 2228: -3.60404779351, 2229: -0.170265765893, 2230: -13.9898339489, 2231: -5.95861667541, 2232: -5.18692349386, 2233: -2.08995318658, 2234: 7.37367517908, 2235: -9.89954716441, 2236: 27.363240182, 2237: -63.5634236008, 2238: -0.709714487393, 2239: 8.2900002109, 2240: -1.24095244476, 2241: -0.993811382207, 2242: -7.19064178392, 2243: -1.06483081435, 2244: -5.27027460685, 2245: -1.99120057652, 2246: -1.19423669382, 2247: -1.12086470966, 2248: -0.956632719212, 2249: -1.09518858488, 2250: -6.45252407186, 2251: 9.72453452136, 2252: -15.7680822751, 2253: 5.71588996158, 2254: 1.18899696886, 2255: -3.42934341899, 2256: -5.53515685724, 2257: -47.5173815687, 2258: -7.33828809502, 2259: 7.38610624718, 2260: 7.4197870709, 2261: 14.1348612461, 2262: -1.05984313868, 2263: 19.8608073946, 2264: -5.44924348143, 2265: -15.985368886, 2266: -0.97540281298, 2267: -84.7676551018, 2268: -1.03967867141, 2269: -31.7891932029, 2270: -2.54250993377, 2271: 31.7234978178, 2272: -0.187167157305, 2273: -0.991074413381, 2274: -0.991979352909, 2275: -1.21107590698, 2276: -0.983362421287, 2277: 8.55185055785, 2278: -36.3952635619, 2279: 0.578239621263, 2280: 33.795962884, 2281: -4.8239821698, 2282: 16.5691980181, 2283: 9.5094566031, 2284: -14.4473766502, 2285: -0.68786346995, 2286: 34.0344139176, 2287: -0.85689743654, 2288: -64.153285402, 2289: -3.94186653057, 2290: -0.477141777412, 2291: -2.14407695909, 2292: -48.325911822, 2293: -1.0974916291, 2294: -4.96669843629, 2295: -0.924766667526, 2296: -21.4154153093, 2297: -0.902373501887, 2298: -0.989692470886, 2299: 21.8819276029, 2300: -5.40064243096, 2301: -73.4313974386, 2302: -0.934715333231, 2303: -0.959679324444, 2304: -0.482390940585, 2305: -1.20445279981, 2306: -1.42561602514, 2307: -2.23646342227, 2308: 8.18988249775, 2309: -1.70943599372, 2310: -0.752572650123, 2311: -52.7834171485, 2312: 5.16889218409, 2313: 88.2133464641, 2314: -45.2898700047, 2315: -8.40215564863, 2316: 5.39314495059, 2317: -27.7403122987, 2318: -6.64799550951, 2319: 34.9186563107, 2320: -1.75598076283, 2321: 14.3474969535, 2322: 1.68700345258, 2323: 4.8752671556, 2324: 1.97117704574, 2325: -14.4753893738, 2326: 0.503795148403, 2327: 14.8279817117, 2328: -9.80464712271, 2329: -7.32927767769, 2330: -6.06310461333, 2331: 1.47191342253, 2332: 24.3778989616, 2333: -7.70152741256, 2334: -4.04606798469, 2335: 7.72579003983, 2336: 23.4925559507, 2337: -4.05138553577, 2338: 2.21802736494, 2339: 82.2242114149, 2340: -6.25674094541, 2341: -44.0527580544, 2342: 28.994097272, 2343: -25.2477038506, 2344: -35.0659684816, 2345: -51.8222379734, 2346: 81.4266150096, 2347: -39.5390146324, 2348: 9.02061629287, 2349: -4.47719073305, 2350: -4.25743888412, 2351: 31.4495757879, 2352: -4.99717186382, 2353: -5.14701716926, 2354: -2.70230230499, 2355: -0.80358773425, 2356: -2.77240173682, 2357: -5.79386163161, 2358: -5.96528848394, 2359: -6.43570535559, 2360: -22.6082167491, 2361: 4.66140741173, 2362: 31.8781883135, 2363: -5.90629725339, 2364: -4.21712106918, 2365: -4.72611410242, 2366: 37.6407659529, 2367: 4.06800716994, 2368: -9.31028262085, 2369: 96.0950871685, 2370: 78.9182713677, 2371: -28.986818567, 2372: 31.052043709, 2373: -16.2301019834, 2374: -39.4648195572, 2375: -5.92043377529, 2376: 51.1517824923, 2377: -18.2218460801, 2378: -42.7702625286, 2379: 80.1964640474, 2380: -13.8675805519, 2381: -5.7218043761, 2382: -41.1412119632, 2383: -2.41784416911, 2384: -7.46296236333, 2385: -9.70614733193, 2386: -4.77418117915, 2387: -15.7854420555, 2388: -15.8848570212, 2389: -55.1301020756, 2390: 48.4676934597, 2391: -14.1140468897, 2392: 52.0552761, 2393: -1.19310915923, 2394: -41.1365975263, 2395: -6.10769542947, 2396: 15.7905136507, 2397: -14.6946309673, 2398: -13.5655078116, 2399: -3.66343617176, 2400: -5.02960712523, 2401: 49.7080444097, 2402: -5.04965369443, 2403: -5.01556139595, 2404: 1.50600979748, 2405: 22.3340903421, 2406: -5.02553447282, 2407: -5.41025175398, 2408: -0.288713774799, 2409: -5.30471008557, 2410: 21.8434802731, 2411: -6.00793104165, 2412: -6.1217078394, 2413: -38.9698076365, 2414: -5.89712159379, 2415: 0.229994910932, 2416: -23.1170465202, 2417: -30.6039200964, 2418: 30.0885526378, 2419: -7.81624299535, 2420: 74.4198022883, 2421: -9.10731414454, 2422: 22.451457219, 2423: 18.4788529843, 2424: -0.215409561893, 2425: -27.4575182306, 2426: 24.3953683485, 2427: -8.36571462976, 2428: 13.2801031169, 2429: -13.7213541511, 2430: 65.8600102419, 2431: 24.6632734317, 2432: -10.7104812371, 2433: -5.47731531588, 2434: -9.08021649395, 2435: 31.844538923, 2436: 65.041495848, 2437: 25.161094057, 2438: -100.0, 2439: -25.9433847228, 2440: 45.4364661228, 2441: 45.0818695711, 2442: -56.0254651291, 2443: 90.3315769956, 2444: 0.609127704091, 2445: 6.09485588474, 2446: -3.81079413939, 2447: -7.4754203227, 2448: 5.64157451825, 2449: -5.01502599204, 2450: 1.21010354291, 2451: -7.52807786636, 2452: -12.5006061022, 2453: -2.5079450625, 2454: -1.54909945382, 2455: -6.12564193613, 2456: -0.114805333755, 2457: 0.440757547516, 2458: 0.468397042249, 2459: 1.19807941482, 2460: -19.2715046397, 2461: -0.136987974882, 2462: -1.34405617085, 2463: 3.24107473086, 2464: -1.86855379042, 2465: -2.90421485216, 2466: 5.32356624936, 2467: -2.4148735184, 2468: -1.85603057856, 2469: -9.31073286502, 2470: 2.48297680093, 2471: -1.90164383915, 2472: 5.4236984575, 2473: -0.519849427746, 2474: -2.80998264286, 2475: -3.07993960645, 2476: -2.81848805873, 2477: -5.28089626912, 2478: -5.49359361462, 2479: -1.84312055992, 2480: -3.37806453999, 2481: -1.76818835071, 2482: 1.63704557665, 2483: -16.1376354819, 2484: 4.53484067951, 2485: 0.808899348127, 2486: -6.62539766508, 2487: -2.47284069589, 2488: -2.30009314635, 2489: -0.65884299279, 2490: -0.88173235682, 2491: 2.99481892362, 2492: -3.44987235892, 2493: 4.92960426856, 2494: -2.9650763236, 2495: -4.87117407903, 2496: 3.8251704223, 2497: 3.37208801233, 2498: -6.47075715898, 2499: 1.39952984321, 2500: -8.14472287969, 2501: 1.69158821095, 2502: 9.25593975195, 2503: -13.4333546712, 2504: 1.18216817411, 2505: -7.73638812885, 2506: 8.71518014653, 2507: -1.00320524979, 2508: 6.89555009123, 2509: 0.96139827879, 2510: 3.08316370441, 2511: 2.10907293604, 2512: -1.36961526639, 2513: 2.97625755693, 2514: 3.02430180263, 2515: 3.00753136521, 2516: 3.39434446403, 2517: -14.710730753, 2518: 0.168990363533, 2519: 12.8064892075, 2520: 3.02987231087, 2521: 0.816911605738, 2522: 2.15735466065, 2523: 1.0723995544, 2524: -2.4574558818, 2525: -6.62241898823, 2526: -3.94375419249, 2527: 2.64336279571, 2528: 0.987576628477, 2529: 2.78227453678, 2530: -0.899944107458, 2531: 4.40229787701, 2532: 1.52290548572, 2533: 0.435136234026, 2534: 2.44603161606, 2535: 3.52185071215, 2536: -13.0763625443, 2537: 3.05819642239, 2538: 1.30641532216, 2539: 2.94931818561, 2540: 0.298827254953, 2541: 2.85578045294, 2542: 1.30797843915, 2543: 3.32340225453, 2544: 5.02858567449, 2545: 2.17067033013, 2546: -9.50418370928, 2547: 0.410416483059, 2548: 0.901736172925, 2549: -0.064275599589, 2550: 3.99821496696, 2551: 3.26976142085, 2552: 1.2081252397, 2553: -3.50366210168, 2554: 2.60995829462, 2555: -3.51177021873, 2556: 8.16724246657, 2557: 11.7888879683, 2558: -18.5007880985, 2559: -3.47930311817, 2560: -2.85289788076, 2561: -2.53026810534, 2562: -1.84684012165, 2563: 1.9097612723, 2564: 10.2547811492, 2565: 4.14391433011, 2566: -9.32603534056, 2567: 5.87052357232, 2568: -8.19312592088, 2569: 0.808964268668, 2570: 1.00494881861, 2571: 0.82381852432, 2572: 6.50552995759, 2573: 0.590174455291, 2574: 3.40836725657, 2575: 2.96587473281, 2576: 6.13337537647, 2577: 0.298918203074, 2578: 0.977670123768, 2579: 4.20894903029, 2580: 0.255759299399, 2581: 8.20386306354, 2582: -1.41239696711, 2583: 0.118267151375, 2584: -8.50904088847, 2585: 1.82004334541, 2586: -6.22673939384, 2587: 1.55070465854, 2588: 3.96376488088, 2589: 0.864768352599, 2590: 0.646065386994, 2591: 4.45732593228, 2592: 0.783668965246, 2593: 3.06093185357, 2594: 6.38687740767, 2595: 0.610091383032, 2596: 0.62025829623, 2597: 0.786253959124, 2598: 0.719829100518, 2599: 1.9032867739, 2600: -7.93909958509, 2601: 1.39170847941, 2602: 0.925589092205, 2603: 2.27269595528, 2604: -0.380935414779, 2605: -0.652002524645, 2606: -41.1441357953, 2607: -5.05287685547, 2608: 4.56708369445, 2609: -8.35602809606, 2610: -1.22994979279, 2611: 1.41842561388, 2612: -6.28643617529, 2613: 0.483185195445, 2614: -4.61186919763, 2615: 0.903064731556, 2616: -4.79851465205, 2617: 0.600616958075, 2618: 0.307250686746, 2619: -8.78702962678, 2620: -5.64239237674, 2621: 0.781011401701, 2622: 0.526505188516, 2623: 0.686855903412, 2624: 0.838444475157, 2625: 0.93175912907, 2626: 0.827277397442, 2627: -17.2942813038, 2628: 0.491080587355, 2629: -18.1500559998, 2630: 1.14171030614, 2631: -0.821241712983, 2632: -0.594785682638, 2633: 10.5399794402, 2634: 11.7999929506, 2635: 6.45671500006, 2636: 0.719209476404, 2637: 7.48741829894, 2638: 2.8110284124, 2639: 1.27926699186, 2640: -2.62104046085, 2641: -0.820771001666, 2642: 0.633375402275, 2643: 3.15499670228, 2644: 0.764580658531, 2645: -1.85786609322, 2646: 0.551751045323, 2647: 0.703570193696, 2648: -3.08772237004, 2649: 2.77043589437, 2650: -15.3608377923, 2651: 0.921589163918, 2652: 0.533530636045, 2653: 2.01811331986, 2654: 0.685000870425, 2655: 2.93623134131, 2656: 0.477900686856, 2657: -7.31228352819, 2658: 0.257934049504, 2659: 0.934962306875, 2660: -0.411203088358, 2661: 0.605283593961, 2662: 7.07474615921, 2663: -5.50423456663, 2664: -2.32985145425, 2665: 0.251566832905, 2666: 2.25068954866, 2667: 1.75285870649, 2668: -4.1182013074, 2669: 0.94053218777, 2670: 9.41855925783, 2671: -0.51531535146, 2672: 1.64367438072, 2673: -2.17373387605, 2674: -21.7755220775, 2675: 0.869456278462, 2676: 0.89827904445, 2677: 8.83561158123, 2678: -2.83755313418, 2679: 6.06142570523, 2680: -4.0418680566, 2681: 4.85661948017, 2682: 1.08792094358, 2683: 3.4769252399, 2684: 2.27221065803, 2685: -0.829450259493, 2686: -1.8417567374, 2687: -0.155227974786, 2688: 2.28416045462, 2689: -8.63500401191, 2690: -3.81474056805, 2691: 3.53814967367, 2692: 11.0484614168, 2693: 5.74340309594, 2694: 8.90548597657, 2695: 5.05216419824, 2696: 1.12356452368, 2697: -10.6813293526, 2698: 0.681573237704, 2699: -0.319252582873, 2700: 5.18355676678, 2701: 2.79395852011, 2702: 2.87720582365, 2703: 1.75209021757, 2704: 2.20282409103, 2705: 3.34516119543, 2706: -2.69435518767, 2707: -0.591683425981, 2708: 1.98032760459, 2709: -8.76346235878, 2710: 0.345410553758, 2711: -0.28408678534, 2712: -4.92273149776, 2713: -0.462264578125, 2714: -0.903282584018, 2715: 9.74192067722, 2716: 1.019062564, 2717: 2.71033992794, 2718: -14.2391076326, 2719: 8.71008492471, 2720: -6.10684772945, 2721: -2.84029003905, 2722: 1.26368338116, 2723: 3.41949149472, 2724: -1.72848222357, 2725: 0.851736594265, 2726: -0.28045248625, 2727: -18.1453248459, 2728: -1.15335742857, 2729: 5.84742995773, 2730: 3.73597234969, 2731: 0.379296153404, 2732: 1.76781057439, 2733: -6.33582687164, 2734: -0.0677271607393, 2735: 9.31812952473, 2736: 3.43753112199, 2737: 2.17496676455, 2738: 2.99167342022, 2739: -5.20875108299, 2740: -12.7815249321, 2741: 1.23817523018, 2742: 2.43769041343, 2743: -6.23301443655, 2744: 0.348969151999, 2745: -13.4957954054, 2746: 1.85803871551, 2747: -1.83493864142, 2748: -16.1716049884, 2749: 1.34298099943, 2750: -8.47704909163, 2751: 3.27608963974, 2752: 3.3257957852, 2753: -1.00602758906, 2754: -5.85159928838, 2755: 3.15302680682, 2756: 3.11727579572, 2757: 9.94627898489, 2758: 2.77495631363, 2759: 11.0279788274, 2760: -2.48938080857, 2761: 3.58915501827, 2762: -3.12002381717, 2763: 3.35936011032, 2764: 7.10028710361, 2765: -2.22459952702, 2766: 2.17949942771, 2767: 7.91464346658, 2768: -1.84668383864, 2769: 2.06915031981, 2770: -8.43225476858, 2771: -6.14730867235, 2772: 5.25680519743, 2773: -7.17055920083, 2774: -10.4228379237, 2775: 7.0143339072, 2776: 0.00405601592606, 2777: 1.58351880557, 2778: -1.72911533407, 2779: -23.7680458742, 2780: 2.39480524136, 2781: 13.1281972758, 2782: 3.21450801869, 2783: -14.6885391961, 2784: 3.92944126862, 2785: 2.52800783942, 2786: -9.30778387626, 2787: -7.1645613826, 2788: -15.1249757079, 2789: 4.28143444895, 2790: 2.66554143367, 2791: -0.803842796478, 2792: 2.73952180153, 2793: -0.88950342327, 2794: -61.9014782558, 2795: -6.170289507, 2796: -5.26808021569, 2797: -4.61838941174, 2798: 44.0536749713, 2799: 13.1889901972, 2800: -11.0281182228, 2801: -14.1562876193, 2802: -23.7225808637, 2803: 48.6534582762, 2804: 26.6136902062, 2805: -11.4679468053, 2806: -14.0091942013, 2807: 21.6622193646, 2808: 0.541104914737, 2809: 31.4002923472, 2810: -15.2938386929, 2811: -8.69025686384, 2812: -12.0772177775, 2813: -22.0354022124, 2814: -50.8573410209, 2815: -10.9071395434, 2816: -1.48954392471, 2817: 1.48108380236, 2818: -1.53977673684, 2819: 3.11569849666, 2820: 10.8171761013, 2821: -24.3842226866, 2822: 9.51854877253, 2823: 15.5465495543, 2824: 25.8465999632, 2825: -26.1251002833, 2826: 0.73569071187, 2827: 3.21968333019, 2828: 8.62690568007, 2829: 19.16068786, 2830: -4.80451560724, 2831: -46.5610099825, 2832: -0.393286602203, 2833: -30.5561836318, 2834: -10.1710314466, 2835: -16.3812289076, 2836: -1.14800264318, 2837: 1.85529299747, 2838: -2.63910732152, 2839: 24.765294756, 2840: -2.92109703983, 2841: -9.75177045614, 2842: -9.92513239515, 2843: 2.79895812007, 2844: -5.20377941241, 2845: -4.71968938421, 2846: -0.969308508215, 2847: 23.9322589906, 2848: -3.43244291336, 2849: -15.3967268779, 2850: -11.1829268715, 2851: -31.3528566602, 2852: 6.51911168239, 2853: -11.2919785189, 2854: 14.3651391472, 2855: -23.0875857979, 2856: -24.4979447348, 2857: -7.58391021268, 2858: -0.810802001309, 2859: -16.3467212719, 2860: -4.45785615825, 2861: 25.3795146697, 2862: -2.57366236063, 2863: -2.87409913846, 2864: 26.2628851937, 2865: -5.34625751472, 2866: 30.4992858676, 2867: -11.6815728678, 2868: 1.21116801474, 2869: 0.683787971751, 2870: -6.94959425926, 2871: -46.4212036327, 2872: -12.0285178176, 2873: -3.53655157764, 2874: 13.9569814515, 2875: -3.98415192164, 2876: 5.69076932802, 2877: 12.3170884897, 2878: -5.05637927912, 2879: -3.92542719029, 2880: -34.0255071879, 2881: -4.75788921512, 2882: -3.1001512483, 2883: 0.276731511534, 2884: -4.57590755593, 2885: -6.56702620604, 2886: -3.54783099149, 2887: -4.0626334793, 2888: -7.20489607896, 2889: 5.66829023611, 2890: -3.40172084406, 2891: -11.783091762, 2892: -8.13894231523, 2893: 31.9121669806, 2894: -5.97999879881, 2895: -4.63864157273, 2896: -0.149174914724, 2897: 6.3046008637, 2898: -8.05597379012, 2899: -5.05341274486, 2900: -4.35344181577, 2901: 5.34961957389, 2902: 19.6326619422, 2903: -5.82542380679, 2904: -2.53269896481, 2905: -30.3616425201, 2906: -30.4640376144, 2907: 20.3429584815, 2908: -7.10780819021, 2909: -4.42105092258, 2910: -5.13034805258, 2911: 5.82284042278, 2912: 3.73612620395, 2913: 9.17031207305, 2914: -29.9725028681, 2915: -4.56516940242, 2916: 7.37638651513, 2917: -25.6449936606, 2918: -0.6351829188, 2919: -0.391600722028, 2920: 14.8507081919, 2921: 0.165304589068, 2922: -0.949021438419, 2923: -3.93060363947, 2924: -11.2102564632, 2925: -20.5120844306, 2926: -1.17940496354, 2927: -0.985843654951, 2928: -18.4459812531, 2929: 3.14606562299, 2930: 8.82102001432, 2931: -3.27848158963, 2932: 6.51741236664, 2933: -10.858580053, 2934: -40.5459640399, 2935: -14.5601218954, 2936: 0.27578342069, 2937: 18.7818144317, 2938: -1.03973607035, 2939: -0.733282703555, 2940: -51.2147873079, 2941: -0.859992200726, 2942: -3.82525997494, 2943: -10.6773648929, 2944: -1.79539742258, 2945: -0.809368593045, 2946: -0.898524602275, 2947: -0.972443665249, 2948: -4.79659980616, 2949: 2.75799911485, 2950: -1.96019318339, 2951: 1.14532247496, 2952: -0.394095766033, 2953: 1.28178511832, 2954: -4.73154817807, 2955: -19.5057675511, 2956: 15.1785724953, 2957: -40.0934907487, 2958: 9.73216273971, 2959: 0.394935503967, 2960: -25.5030053072, 2961: 8.81921609852, 2962: -3.76296413187, 2963: 16.2630224782, 2964: -0.884440808628, 2965: -34.0895799459, 2966: -0.848842368149, 2967: -1.93388776307, 2968: -5.26990863246, 2969: -68.7801370126, 2970: 1.20503989079, 2971: -0.724464463425, 2972: -0.77194033976, 2973: -1.13512158548, 2974: -1.02186979982, 2975: -4.4962552143, 2976: -31.4839451282, 2977: 9.49278973124, 2978: 10.6802504449, 2979: 42.6161253048, 2980: -2.61264726135, 2981: 16.1445181105, 2982: 29.0566744263, 2983: 10.6569807183, 2984: -20.4315654918, 2985: -0.754288215556, 2986: -52.0264694371, 2987: 20.4986248726, 2988: 3.14407777038, 2989: -11.4601470905, 2990: 0.861070496358, 2991: -1.00097083439, 2992: -2.89593962215, 2993: -0.86795958982, 2994: -10.2591002295, 2995: -0.805860344408, 2996: -0.826158565578, 2997: -15.901479278, 2998: -4.22026663021, 2999: 50.2297628657, 3000: -0.746367546176, 3001: -0.739099267679, 3002: 1.96089981797, 3003: -0.873753652418, 3004: 5.41265648195, 3005: -1.61905603708, 3006: 26.8590760724, 3007: -0.715367413948, 3008: 1.91943518186, 3009: 49.4220396553, 3010: 9.85982964049, 3011: -46.0674009034, 3012: 3.69393916845, 3013: -4.7400184686, 3014: 2.60272921154, 3015: 15.7177275261, 3016: -2.38770403406, 3017: -2.29549005962, 3018: -0.898713881803, 3019: 17.340312577, 3020: -16.1749431012, 3021: -10.4341706569, 3022: 0.38355850603, 3023: -13.223944233, 3024: -3.41438755321, 3025: -42.6280064234, 3026: -25.8513989447, 3027: -13.4153197532, 3028: -29.6517328656, 3029: -1.38252418546, 3030: -21.5810527479, 3031: -4.20589789012, 3032: -4.07048823125, 3033: 27.2699618522, 3034: 4.55125690255, 3035: 20.5845148968, 3036: 0.966259462575, 3037: -6.81486271931, 3038: -23.1735896627, 3039: 9.04371323288, 3040: 30.2923312638, 3041: 0.236888852419, 3042: -24.7455982679, 3043: 41.1075886814, 3044: 9.78977408156, 3045: -28.3718079751, 3046: -5.95924589615, 3047: -3.52062948744, 3048: -3.39810887916, 3049: 35.692725114, 3050: -1.43783431128, 3051: -3.96585052596, 3052: 1.89165299409, 3053: -11.7427021526, 3054: -5.61589968019, 3055: -5.24545117207, 3056: -4.99425199147, 3057: -4.7291784987, 3058: 20.4205403245, 3059: 20.1543149734, 3060: 5.29006967065, 3061: 8.84237492852, 3062: -0.0144411799185, 3063: -3.71360086822, 3064: 2.88902586211, 3065: 3.73831821958, 3066: 2.32494487554, 3067: 19.7421558855, 3068: 45.3553385468, 3069: -0.0979660147856, 3070: 40.0497872872, 3071: -4.62662811769, 3072: 2.08067924358, 3073: -6.51431672444, 3074: -25.0800083497, 3075: 7.62257431621, 3076: -27.3628774879, 3077: 22.5485841661, 3078: -16.9171071658, 3079: -4.95765492225, 3080: 2.45931655848, 3081: -2.15151816122, 3082: 5.29671021779, 3083: -9.38567916162, 3084: 16.0440684878, 3085: 13.3371048538, 3086: 7.47556312002, 3087: -48.4801678612, 3088: 25.4800252687, 3089: 16.5913910814, 3090: 0.405400399558, 3091: -2.21278269769, 3092: -9.3079513282, 3093: -2.64101515038, 3094: 52.0364694743, 3095: 5.890299782, 3096: 52.8317001806, 3097: -6.82067036581, 3098: -4.00789579067, 3099: 19.4262032942, 3100: -2.71716501652, 3101: -2.68349732015, 3102: -3.42295860695, 3103: 29.3737504583, 3104: -3.90156214925, 3105: -4.08330059765, 3106: -5.27440493661, 3107: -4.15890614004, 3108: -36.2233435242, 3109: -4.24235161011, 3110: -5.46297307681, 3111: 15.2585825248, 3112: -3.82589862565, 3113: 9.13828628526, 3114: -49.6765771183, 3115: 49.0041058147, 3116: 12.5659935079, 3117: -11.6528694339, 3118: -13.3887977827, 3119: 3.79165261594, 3120: 74.8938988017, 3121: 6.87202711588, 3122: 7.3563263228, 3123: -11.4529414228, 3124: -2.5731195164, 3125: -3.60256810289, 3126: -13.728069151, 3127: -4.28129463478, 3128: 29.7234279382, 3129: -18.5164778234, 3130: -23.2105669236, 3131: -3.93995770214, 3132: -6.77886186743, 3133: -2.22217711234, 3134: -7.59753104423, 3135: -13.5610861556, 3136: -12.4527560436, 3137: 39.9486859562, 3138: 36.8511787952, 3139: -12.3672302488, 3140: 30.2149774816, 3141: 100.0, 3142: -0.32201089031, 3143: 0.676884532022, 3144: 9.35632821075, 3145: -22.8387963677, 3146: -10.7540115122, 3147: 25.7856981606, 3148: -7.26014099065, 3149: 0.86704555025, 3150: -32.891264404, 3151: 13.4374592853, 3152: -35.3176685784, 3153: -12.5954440245, 3154: -12.6469601709, 3155: 19.8790666701, 3156: -9.08408773983, 3157: -14.9750812857, 3158: 2.14013717319, 3159: -6.28953864676, 3160: 45.6137299981, 3161: 24.0173554037, 3162: -12.7718502917, 3163: 81.1976507122, 3164: -27.3938272402, 3165: -23.7849556427, 3166: 11.9977630667, 3167: 29.4658649396, 3168: 32.066282636, 3169: -4.05420642784, 3170: 29.0094366267, 3171: -1.84663241307, 3172: 6.06151351859, 3173: 3.33497776195, 3174: 0.116450243677, 3175: -4.36009354105, 3176: 11.7574577664, 3177: 24.1892374612, 3178: -9.35248830479, 3179: -18.8705351046, 3180: -8.67199824408, 3181: 0.761567765651, 3182: 19.878847913, 3183: -3.22786120089, 3184: -10.7401054705, 3185: -6.30398509939, 3186: 14.6367041801, 3187: -3.74609360175, 3188: 3.54219090176, 3189: -3.78134992324, 3190: 49.1604078633, 3191: -3.74975763097, 3192: 12.7531843788, 3193: -1.60310061095, 3194: -4.66888370002, 3195: -8.31828095321, 3196: 1.71526531291, 3197: -3.30772568139, 3198: 2.56842891899, 3199: -3.48015386849, 3200: 24.6329716662, 3201: -28.6543168606, 3202: -23.341267025, 3203: -19.5048812431, 3204: -6.36021237096, 3205: -21.9995452995, 3206: -39.5611709506, 3207: -3.25345920625, 3208: -6.03943009311, 3209: -1.70780080544, 3210: 16.8123361624, 3211: -2.10043008768, 3212: -2.71466810783, 3213: -2.58393082685, 3214: -3.19596873224, 3215: 11.0236595577, 3216: -3.72079134727, 3217: -2.20494272474, 3218: -2.03467013651, 3219: 0.380863991318, 3220: -17.2384272132, 3221: 23.5091134814, 3222: -5.01917873835, 3223: 11.4572788813, 3224: -2.80823134631, 3225: -2.64267500918, 3226: -5.97179410131, 3227: -18.5680906776, 3228: 28.1159968613, 3229: 9.27174901511, 3230: -4.40776323712, 3231: -2.00265960674, 3232: -6.0773350355, 3233: -3.98732088982, 3234: -23.3596702435, 3235: 4.5205566626, 3236: -2.5044887169, 3237: -0.775352569223, 3238: 17.7381215788, 3239: -2.23650818472, 3240: -3.12372982883, 3241: -2.319146263, 3242: 55.0764716822, 3243: -3.15033221751, 3244: -6.18016251186, 3245: -3.19970206818, 3246: -4.20712281665, 3247: 77.0724875113, 3248: -3.01747630051, 3249: -3.62453891939, 3250: -1.99421164985, 3251: 36.4726580035, 3252: 4.70849630413, 3253: -10.0600470299, 3254: -87.2676725951, 3255: 2.71520485823, 3256: -63.564528432, 3257: 6.79047200311, 3258: -9.80079338075, 3259: 12.5184129659, 3260: 36.6636484554, 3261: 10.9765009335, 3262: -33.5544259666, 3263: 16.336924904, 3264: -25.9497564891, 3265: 32.589892648, 3266: -0.431944734758, 3267: -0.459910543239, 3268: 0.660691556261, 3269: -17.7541417848, 3270: 9.64818788271, 3271: -0.561438068222, 3272: -2.74717819322, 3273: 0.230944280162, 3274: -7.46260322885, 3275: 2.0920346552, 3276: -0.592718776376, 3277: -4.27209799159, 3278: 3.2778957782, 3279: -12.6010986959, 3280: -0.478609212841, 3281: 2.58546304901, 3282: -24.0113662354, 3283: 10.1092349182, 3284: 8.47924208493, 3285: -0.570791832157, 3286: -7.09620233096, 3287: -0.498431183256, 3288: -0.410172264575, 3289: 78.8866538445, 3290: -0.533792276629, 3291: -3.21498907498, 3292: 1.35022186283, 3293: -1.02538801271, 3294: -0.588411170738, 3295: -0.526965360097, 3296: -0.471290642024, 3297: -2.60649466843, 3298: 7.39187520685, 3299: -1.13430716865, 3300: -0.867627064529, 3301: -0.898535522685, 3302: -7.77354256856, 3303: -9.42499321714, 3304: -30.7932272571, 3305: 6.70995024904, 3306: 11.4191524825, 3307: 78.2097528329, 3308: 15.8441130917, 3309: -0.890326479218, 3310: -0.733531049473, 3311: -4.96948980772, 3312: 14.3126376317, 3313: -0.541300165894, 3314: 22.4874133808, 3315: -0.458615151137, 3316: -7.79038979914, 3317: 23.3768895702, 3318: 29.6915358535, 3319: -1.24457168357, 3320: -0.43716214407, 3321: -0.686440091132, 3322: -0.630812749762, 3323: -0.756902896206, 3324: -1.39648512825, 3325: -26.0881102753, 3326: -11.6711802594, 3327: 29.1143199321, 3328: -13.1311434939, 3329: 7.26491139882, 3330: -7.8737675455, 3331: -5.17519640113, 3332: 10.9556788919, 3333: -5.57065555647, 3334: -0.337820530949, 3335: 7.69613858898, 3336: 1.56304759066, 3337: 1.92896341468, 3338: -6.39505508497, 3339: -34.1728710497, 3340: -0.610752958176, 3341: -3.10020090475, 3342: -0.567075425708, 3343: 5.97736237848, 3344: -0.4822647456, 3345: -0.520108581894, 3346: 29.8833029525, 3347: -3.57830822981, 3348: 45.9347338389, 3349: -0.607479622042, 3350: -0.445203056626, 3351: -4.94049272055, 3352: -0.655803388582, 3353: -3.59115709079, 3354: -13.1889176142, 3355: 10.0441708197, 3356: 0.935108954593, 3357: -2.37091040875, 3358: -1.27165679675, 3359: -1.33451373208, 3360: 4.13514436115, 3361: -31.7029487524, 3362: -1.98304861734, 3363: -8.60787782575, 3364: -2.01138116211, 3365: -2.97613289203, 3366: -2.42877238643, 3367: -0.681412201943, 3368: -26.1807726248, 3369: -1.67724775052, 3370: -3.24699606549, 3371: 2.18001610385, 3372: 8.31059811464, 3373: 0.00503948819356, 3374: 27.2284749568, 3375: 0.295301934548, 3376: -4.07250842876, 3377: 2.34317035424, 3378: 9.80056874294, 3379: 8.19885686095, 3380: -2.34865481504, 3381: -2.86429904622, 3382: 4.12369688437, 3383: 9.12122408352, 3384: 53.9060780511, 3385: -1.19184909423, 3386: 6.81729173771, 3387: -21.762062925, 3388: 4.75441643855, 3389: 28.462828292, 3390: -6.61595620763, 3391: 47.2465886023, 3392: 31.6721853419, 3393: 1.74346924289, 3394: -41.8059655353, 3395: -21.1503296593, 3396: -8.19658333768, 3397: -3.53794676698, 3398: 18.2823556198, 3399: -3.19909640042, 3400: -3.31577362972, 3401: -15.3113071673, 3402: 38.9004124702, 3403: -3.27235602996, 3404: -14.8427831514, 3405: -5.73159005392, 3406: -2.43702409208, 3407: 11.4894187427, 3408: 13.0595449128, 3409: 5.12550684119, 3410: 1.63577930998, 3411: -13.3100737436, 3412: -0.693774421287, 3413: 15.6241044624, 3414: 10.2913355983, 3415: -18.0715115125, 3416: -47.785357589, 3417: 12.0760363927, 3418: 13.2071308454, 3419: 0.0868698099131, 3420: -9.09486469776, 3421: 21.2567505723, 3422: -1.22226535531, 3423: -9.76391005616, 3424: -6.62903741247, 3425: 36.2831591701, 3426: -71.0570118447, 3427: 26.0810913804, 3428: -4.1950437637, 3429: -3.05212756152, 3430: -4.7561323523, 3431: -2.75392593217, 3432: 2.27523944557, 3433: 15.274109623, 3434: -7.18434701586, 3435: 6.25468964529, 3436: 4.59556203474, 3437: 23.528963965, 3438: -29.8764214644, 3439: 17.0239383773, 3440: 9.203116039, 3441: -1.65752344796, 3442: -6.64048011142, 3443: -50.9065159315, 3444: 0.647015223596, 3445: -28.7657907808, 3446: 15.7811418525, 3447: -2.44894686543, 3448: 45.8553906899, 3449: -3.24509903487, 3450: -3.12943134386, 3451: -2.42449087073, 3452: -37.4237909273, 3453: -2.43769055564, 3454: -4.09792938258, 3455: -22.1863253991, 3456: -3.47636598473, 3457: -42.2138727587, 3458: -3.37973891779, 3459: -3.29785809408, 3460: -5.7474341223, 3461: -3.58385056268, 3462: -4.34998966433, 3463: -7.42890829536, 3464: -17.35784189, 3465: -11.5953629348, 3466: 24.283442644, 3467: 6.77144353245, 3468: 10.0600660365, 3469: 30.7615035714, 3470: 26.5168763576, 3471: 11.8028205041, 3472: 11.7132486547, 3473: 32.205160895, 3474: -3.32083254555, 3475: -3.95636086373, 3476: 2.80423772187, 3477: -43.5399004427, 3478: 22.4028808716, 3479: 45.7144932737, 3480: -3.2869377596, 3481: -11.020504574, 3482: -10.0580674507, 3483: -3.33535797501, 3484: 2.72522335939, 3485: -7.7618265795, 3486: -6.89263464983, 3487: 36.7535865197, 3488: 16.9867811723, 3489: -49.8911252782, 3490: -8.39795701466, 3491: 1.8917292422, 3492: -2.01182387663, 3493: 1.81046452099, 3494: -1.65072056457, 3495: -1.767983627, 3496: -0.532978043193, 3497: -0.758474714499, 3498: 1.24726976265, 3499: 3.47195502614, 3500: -1.12381972918, 3501: -1.27059858449}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.414960903209, 2: 28.4995390328, 3: -73.4636024591, 4: -16.4839079655, 5: 33.7508929163, 6: 35.8910902138, 7: -45.2013787704, 8: -43.2574926918, 9: 17.2082477435, 10: 0.721987125237, 11: -13.5481456228, 12: 46.8865882549, 13: -31.0562704688, 14: -45.9497800962, 15: -17.8342405164, 16: 89.8000289136, 17: -45.288653255, 18: -11.4560889193, 19: -31.9440311613, 20: -30.0875002344, 21: -55.6555041471, 22: 37.2655217812, 23: 45.4282126499, 24: 25.2138917374, 25: -48.7513995868, 26: -23.0491413981, 27: -52.2750012142, 28: -5.27216772462, 29: 50.6587796276, 30: -6.65677461029, 31: 18.8334236559, 32: 39.7433722661, 33: -16.3567918791, 34: -13.1723822488, 35: -3.82558472868, 36: -13.246393273, 37: 22.5777021396, 38: -23.2080203245, 39: 45.1395756781, 40: 0.108451445687, 41: 11.2750583117, 42: 2.38324992018, 43: -29.277430435, 44: 19.264657264, 45: -0.292730404796, 46: 20.6896345996, 47: 16.732856206, 48: -9.62445361366, 49: 35.855648577, 50: -28.2242259117, 51: -5.30350595291, 52: -42.406936305, 53: 2.9496624275, 54: -20.7301541884, 55: 26.3202462841, 56: 0.603167943707, 57: -30.3435616236, 58: -13.2569334635, 59: -23.1674103113, 60: 78.201558329, 61: 63.2858266522, 62: -29.4816808047, 63: -39.5694412672, 64: -11.1902270487, 65: 63.730635781, 66: 4.39969473311, 67: 6.55076509603, 68: 3.08310758642, 69: 32.4788986873, 70: 4.652984966, 71: 3.76292900275, 72: 46.0295170354, 73: 2.91359365362, 74: 42.0650449377, 75: 11.0099832978, 76: -3.14419820227, 77: 4.09391691317, 78: 12.9574409369, 79: 16.0098227735, 80: -41.2573191209, 81: -1.21054794615, 82: -13.8072455722, 83: 3.32154933592, 84: 27.2310649601, 85: -30.7308825348, 86: 0.985229473327, 87: 3.49651503334, 88: -73.8524916679, 89: 2.48249747769, 90: 21.706982318, 91: 3.33123752788, 92: 2.10228820711, 93: -78.2131795872, 94: 3.05941274625, 95: 3.29798546643, 96: 3.71214082711, 97: 10.5800234906, 98: 3.22971752898, 99: -8.25098128814, 100: 3.6900881892, 101: -52.2998120136, 102: -3.84896408178, 103: -17.6868054627, 104: 1.45848836918, 105: 6.61727651929, 106: -0.603677448378, 107: 2.84616206446, 108: 1.96581549119, 109: 2.53191271161, 110: -6.45664464404, 111: 8.29679244818, 112: 45.2070232472, 113: 46.8953574032, 114: 96.6518660308, 115: -10.1997663187, 116: 2.40903916674, 117: 47.1634060625, 118: -43.0145583577, 119: 20.7879091692, 120: -0.413781830776, 121: 2.35104551787, 122: -18.3913852605, 123: 5.40226088123, 124: -16.5716642542, 125: 2.97939668094, 126: 0.66470393598, 127: 1.31645058439, 128: 5.69726443599, 129: -73.598644087, 130: 0.844026256727, 131: -0.0235645832361, 132: 0.557810888518, 133: -12.6158556103, 134: -1.24031374612, 135: -0.0538902717121, 136: -19.8915460893, 137: 3.02604570917, 138: 34.5005445893, 139: 1.59378160471, 140: 10.5906977283, 141: 10.1565577145, 142: 0.972782367449, 143: 21.3719128336, 144: 1.51653744815, 145: 4.70985305452, 146: 0.204318233961, 147: 0.616640484131, 148: 37.0098937808, 149: 0.665466277646, 150: 2.63385637479, 151: 31.6643750702, 152: -0.65182886401, 153: 0.551533973167, 154: 0.801088435251, 155: 0.649750982083, 156: -4.61702458537, 157: 89.6800188559, 158: 7.53342214612, 159: 3.41652421216, 160: 0.3604403923, 161: -0.583219840323, 162: 14.8860643736, 163: 54.8427222194, 164: 8.03416476768, 165: -10.4807622198, 166: 16.5099259727, 167: 76.9175454585, 168: -5.53123604577, 169: 25.1257648567, 170: 7.11955106255, 171: 14.1240371328, 172: 0.330225639579, 173: -17.547446158, 174: 0.470444173377, 175: -93.1756407086, 176: 71.5902512545, 177: 7.57646267214, 178: -1.86093842826, 179: 0.589204527321, 180: 0.251343041993, 181: 0.468651970661, 182: 0.720458876407, 183: -8.21127572247, 184: 8.73318660143, 185: 0.223013293471, 186: -46.7598927606, 187: -33.9733683475, 188: 2.00284223226, 189: 3.23332833651, 190: -15.6225432376, 191: -10.2565159853, 192: 11.1367203707, 193: 0.317018730447, 194: 84.3259169072, 195: 0.647024920525, 196: 13.8788147596, 197: -1.75335606295, 198: 12.2999730362, 199: 0.438342867366, 200: 2.56675759161, 201: 0.646759076133, 202: -26.3044767417, 203: 0.606742044061, 204: 0.582915348059, 205: -6.25519690809, 206: 2.58413469628, 207: 40.2517112452, 208: 0.479151520125, 209: 0.621618392913, 210: 0.874411703007, 211: 0.568195805693, 212: -7.96349615064, 213: 9.94424961018, 214: -80.4208114423, 215: -0.812344152877, 216: -3.4948161115, 217: 6.44981631339, 218: 2.77484072755, 219: 11.2101044974, 220: 19.4422843287, 221: -6.674347707, 222: 18.1432473885, 223: -1.90173561015, 224: -7.92076297916, 225: -9.13626073244, 226: 0.60778105883, 227: 30.2318003164, 228: 1.60971338503, 229: -5.3783045123, 230: 2.93303997177, 231: 20.5007917017, 232: 3.95504154336, 233: -13.4054546651, 234: -25.4064583946, 235: 0.0469168526048, 236: -8.27830174869, 237: -13.8699578503, 238: -56.2833935222, 239: -5.04813228279, 240: 0.0665452092964, 241: 16.9699010015, 242: -50.8981997541, 243: 7.29605489909, 244: 14.1130204667, 245: 2.93995058276, 246: -9.61353323165, 247: -97.4238841389, 248: -33.6643516942, 249: 70.3099940243, 250: -6.72670774993, 251: 10.3378674667, 252: -52.5471224844, 253: -12.3881309312, 254: -42.3596872207, 255: 14.7400366402, 256: -2.17416660709, 257: -26.3537175678, 258: 2.61004858014, 259: 2.47024735576, 260: 28.0913053038, 261: 33.7511982826, 262: 2.14833896068, 263: -3.256032252, 264: 1.63167853197, 265: -4.54218401062, 266: -35.8837061413, 267: -5.55049631574, 268: -19.9247812977, 269: 31.3588684516, 270: -6.6885382529, 271: 10.8276435945, 272: -21.2047530999, 273: 25.5724303338, 274: -15.9190931966, 275: -13.0929295595, 276: 30.7792424833, 277: 25.7714481125, 278: -7.91775149719, 279: 7.93231752776, 280: -12.4842335226, 281: 14.0854518269, 282: 15.0351420031, 283: -98.2640030624, 284: -19.968537961, 285: 2.41406987112, 286: -27.1484223562, 287: 3.39918477026, 288: 1.15281192426, 289: 3.10302830063, 290: -5.37480085779, 291: 2.83953490408, 292: -14.0721873258, 293: -78.8828696244, 294: -31.7094339275, 295: 11.4287777277, 296: 27.1626553414, 297: 6.66693019767, 298: -11.6618926402, 299: -8.15326345745, 300: 6.27968729383, 301: 7.87793018701, 302: -8.18758591068, 303: 4.63637580932, 304: 18.0824493335, 305: 34.1374277079, 306: 3.16305484109, 307: -2.61922538359, 308: 2.61743248833, 309: 2.76611782833, 310: 4.62018747882, 311: 10.2497739796, 312: 2.32901411663, 313: 0.946875087402, 314: -9.20294058638, 315: 2.64669584757, 316: -20.9969863494, 317: 2.94515885928, 318: 2.85970948503, 319: 0.0624545601959, 320: 0.806592546603, 321: -13.8279459092, 322: 25.8178112696, 323: -82.461522972, 324: 19.6772056358, 325: 66.9957799268, 326: -15.280783777, 327: -83.1933905701, 328: 43.3862742539, 329: -6.59421319616, 330: 4.85212692683, 331: 5.11295928279, 332: 1.72863823708, 333: -8.94116449018, 334: -15.0050877078, 335: -42.7202143436, 336: -18.7095006203, 337: 5.5639789167, 338: -0.805850110818, 339: 2.82648738163, 340: -65.2679133245, 341: -16.0155113568, 342: -28.0047119843, 343: -22.9117265959, 344: 14.7627072015, 345: -29.5996777218, 346: -100.0, 347: -100.0, 348: -35.0848008687, 349: -56.2433934126, 350: -0.793446560267, 351: -29.3832832368, 352: 7.81725246994, 353: 24.4458892041, 354: -5.08892409659, 355: 3.92341806192, 356: 35.5299358591, 357: 10.990737802, 358: -0.292713744413, 359: 2.44779037972, 360: -38.1122115797, 361: -13.6456842264, 362: 13.241214555, 363: -12.9867048985, 364: 66.9582776599, 365: 3.17964599982, 366: 15.4357900827, 367: -14.2240834425, 368: -8.39151371942, 369: 8.07473924701, 370: 33.4802637459, 371: 24.9860623252, 372: 26.6559951468, 373: -17.0882521679, 374: -7.05408401365, 375: 19.294961781, 376: 21.5295917432, 377: -22.0709086579, 378: 48.0528240796, 379: 31.1564775982, 380: 31.7629148423, 381: -4.70002803707, 382: -1.5351104446, 383: -9.61002086211, 384: 16.3882061787, 385: -1.63837210501, 386: -24.9571388844, 387: -9.59339093139, 388: 7.21253477846, 389: -12.524251058, 390: -0.326822087858, 391: -3.82742213104, 392: 5.13982899237, 393: 2.9249277576, 394: 4.39541176665, 395: 0.801531593624, 396: 11.002480519, 397: -1.52944009039, 398: 16.3962609627, 399: 0.479802565548, 400: 0.78696187176, 401: -0.0822746443792, 402: -4.70269027338, 403: 0.679886990803, 404: 4.3602364497, 405: -4.77944719594, 406: -52.1802323045, 407: -2.18172858162, 408: 33.8676600187, 409: -25.3703301883, 410: -10.449789394, 411: 9.61743442666, 412: -57.3647231097, 413: -4.93703439231, 414: 28.9611891432, 415: -4.84789016842, 416: -7.28356126644, 417: 0.866984313721, 418: -18.3175516821, 419: -4.97849876451, 420: -4.71557825715, 421: -12.3614081616, 422: -4.92584941486, 423: -1.94155387526, 424: -4.16315243761, 425: -5.01934264144, 426: -4.13427059201, 427: -5.43900738341, 428: -13.8403073766, 429: -1.77792518273, 430: -4.70173095014, 431: 1.52629081857, 432: -4.76097810581, 433: -4.73519244385, 434: -4.92738949853, 435: -11.7313045505, 436: 0.309716314185, 437: -11.5743655272, 438: -3.3458974058, 439: -0.827554534037, 440: -4.26747344694, 441: -4.78238944702, 442: 10.1045546956, 443: -4.71790857339, 444: -2.92364740671, 445: -2.45962306254, 446: 12.840796076, 447: -4.82640510133, 448: -3.85634623957, 449: -3.88021467027, 450: -2.84537748535, 451: 13.9688595574, 452: 4.65982627405, 453: -4.84805264791, 454: 5.02492323022, 455: -10.3092413658, 456: -4.76968836255, 457: -4.72893981374, 458: -3.28026349872, 459: 0.890491582572, 460: -4.56328089534, 461: 1.7651939558, 462: -25.1538716947, 463: -10.404537448, 464: 5.24978833567, 465: 12.1014495032, 466: -17.1128405913, 467: -0.251558980307, 468: 12.6517743575, 469: 0.51183472001, 470: 62.9597671643, 471: 1.5540134245, 472: -20.3470083414, 473: -34.871167367, 474: 13.6858193347, 475: -0.928818855467, 476: -1.53484892446, 477: -0.522029655368, 478: -30.6394406506, 479: -1.08654937601, 480: -4.80243514252, 481: -1.69815968905, 482: 30.4147637616, 483: -2.21074545953, 484: -2.21043032926, 485: 47.1328078541, 486: 0.126790674872, 487: -6.09082143949, 488: -3.80624328104, 489: -0.170283282619, 490: 6.96348416834, 491: 2.49784359091, 492: -19.6819631846, 493: 0.72561928808, 494: -7.67680026932, 495: -0.650784769239, 496: -1.02575646859, 497: 3.13889590672, 498: -0.952216647587, 499: -4.9414048067, 500: 1.87432810961, 501: -1.53901639657, 502: -0.939692570795, 503: -1.10464442091, 504: -0.970153659577, 505: -4.13825372119, 506: 27.7001931882, 507: 0.300743557016, 508: -1.15788698381, 509: -1.51989639012, 510: -1.29978713304, 511: 0.419114286932, 512: 36.1386908241, 513: 12.2793086143, 514: 31.2191602172, 515: -23.1224316126, 516: 4.65934432787, 517: -2.61762002165, 518: -5.9605577838, 519: -2.26300106238, 520: 0.196955084661, 521: -0.904458489171, 522: -17.4950188281, 523: -0.972994566679, 524: -7.75240005251, 525: 40.3034377774, 526: -4.20658639369, 527: -3.59229245524, 528: -1.03316174116, 529: -0.906303501865, 530: -1.20076643364, 531: -1.13284403685, 532: -6.77345634868, 533: 27.1430420849, 534: 2.60292267819, 535: 30.2117184225, 536: -44.2828293188, 537: -8.52161067689, 538: -4.55416237429, 539: -12.3638769717, 540: -5.40450700199, 541: -8.71098001197, 542: -1.25112500978, 543: -61.2385097861, 544: -3.6747916374, 545: 5.84243369659, 546: 5.27422357342, 547: -33.8294722234, 548: -0.967286894666, 549: -4.89920181269, 550: -0.949443624666, 551: 4.90828267878, 552: -1.14855142595, 553: -1.0265787778, 554: 6.16043793173, 555: -2.47339272229, 556: 26.2972912603, 557: -1.00114440111, 558: -1.06120081491, 559: 1.25790470049, 560: -1.10206331712, 561: -1.32754329781, 562: 15.4216764741, 563: 19.044867983, 564: 0.911080216878, 565: -1.08453408131, 566: -7.76967319941, 567: 0.637125819346, 568: -5.26750737118, 569: 37.1460171008, 570: 15.3613733372, 571: 6.0834640114, 572: 5.68293783654, 573: -4.76833719972, 574: 3.04654723024, 575: -1.03226016853, 576: -13.0039382926, 577: -5.52417295306, 578: -1.12619590127, 579: 1.20466907856, 580: -25.670985384, 581: -1.50370489057, 582: -31.097637848, 583: -62.8532788695, 584: -4.8480823531, 585: -10.7715337517, 586: 8.40989356279, 587: -22.3260990253, 588: -4.92889546583, 589: -4.80383142932, 590: 26.2211983858, 591: 15.1790839053, 592: -7.45758383933, 593: 48.5492187094, 594: 8.40070038919, 595: -87.7652395143, 596: -4.63303475095, 597: -2.87334987693, 598: 45.3387385924, 599: -8.45050778559, 600: -0.836394621348, 601: 27.2575336541, 602: 9.53778842116, 603: 8.9885807151, 604: -9.97418935889, 605: -4.6118084957, 606: -26.0704011748, 607: -4.83189669771, 608: -4.85296897872, 609: 1.85607278374, 610: 1.80490113655, 611: -4.61622975607, 612: -1.50671515327, 613: -4.11056575449, 614: -4.2525719647, 615: 25.2867023103, 616: -5.02468125718, 617: 3.65788746702, 618: 8.71926387193, 619: -0.726802244219, 620: 0.768920310156, 621: 45.7221713001, 622: 5.37984292546, 623: 6.17954289812, 624: -33.3023208428, 625: 15.4162367252, 626: -27.4393129742, 627: 25.3948537537, 628: -1.3318323201, 629: 8.26788546674, 630: -1.03375455069, 631: 10.3503436389, 632: -7.68566363184, 633: -13.7224092703, 634: -15.4928949748, 635: 9.05943770319, 636: -4.76332702015, 637: 0.0137233823451, 638: -4.62454732185, 639: -4.06447509491, 640: -4.04837495453, 641: 31.5425549582, 642: -9.23884360538, 643: 2.35348365662, 644: -2.84223574402, 645: -6.59461190074, 646: 23.941177258, 647: 15.9665787925, 648: -5.40858936914, 649: -0.113079647271, 650: 0.680151301124, 651: 27.7946976691, 652: -4.13396941743, 653: 1.0323530419, 654: -35.3616263209, 655: -2.76659901046, 656: 44.3819095207, 657: -4.82183190408, 658: -4.78385619917, 659: -4.89422785945, 660: -13.5553552498, 661: -0.233587726894, 662: -4.7440894995, 663: -48.6011734479, 664: -2.27762769832, 665: 44.4584539841, 666: -4.70399004148, 667: -4.86772397401, 668: -30.0338422591, 669: -4.85065180174, 670: 1.21112958002, 671: 26.4118306508, 672: -27.0706356492, 673: -13.2433229356, 674: 14.1916894004, 675: 39.2107423383, 676: -10.002534584, 677: -13.2357856689, 678: 44.343329305, 679: 50.4137588196, 680: 2.31553919771, 681: 23.2060304285, 682: -4.62368584117, 683: 2.15060886009, 684: -0.884691442812, 685: 37.2878627792, 686: -27.9789378505, 687: -12.3076793783, 688: -4.91623589274, 689: -37.2043723479, 690: -6.60723570179, 691: 4.49275961497, 692: -21.602045914, 693: 30.1502871889, 694: 40.3867018671, 695: 5.12634695209, 696: -3.60048505749, 697: 38.0954868663, 698: -3.02548419449, 699: 0.532749333108, 700: 11.7291516617, 701: -1.99668267981, 702: -45.5556330791, 703: -35.632356328, 704: -15.1787007183, 705: 41.3557779741, 706: -17.3991941398, 707: -12.9134503946, 708: -44.2754224848, 709: -39.8788958948, 710: -47.4270364902, 711: 8.82704649, 712: 9.27860743176, 713: -78.4997911196, 714: 3.18326922264, 715: -8.1101668633, 716: -100.0, 717: 10.8638465709, 718: 6.23325485256, 719: 53.9618364653, 720: 23.6823655848, 721: 55.2235632593, 722: -18.3042901884, 723: -20.0418638269, 724: 8.79707416099, 725: 50.4050995584, 726: 32.3431926873, 727: -18.6667435496, 728: -26.4834059939, 729: -1.33108880322, 730: 32.9248058557, 731: -20.5973535093, 732: -2.35517673006, 733: -13.3817400633, 734: -10.6079992586, 735: 12.0945647645, 736: 17.9360215287, 737: -65.2524984952, 738: 57.8056741563, 739: -4.3691604619, 740: 3.89220120195, 741: -3.22021187995, 742: 66.3629635279, 743: -28.6905885383, 744: 30.0515027976, 745: -17.3700721244, 746: 0.186121768598, 747: 38.2826876224, 748: 4.17954398932, 749: -7.63472718913, 750: 2.68852800638, 751: 1.61534837656, 752: 16.4636216926, 753: 27.1480346078, 754: 2.71035388391, 755: 47.1202472109, 756: 1.8801612279, 757: -54.1147143431, 758: -10.0757385868, 759: -2.36437609227, 760: -55.5484645246, 761: 75.3205207707, 762: -16.465098158, 763: 53.0646551928, 764: 10.0527693711, 765: 9.96261163046, 766: 3.81600252985, 767: -23.2408427008, 768: 2.88936506677, 769: 4.00613452636, 770: -53.3563511564, 771: -0.00289752669298, 772: -7.65958911005, 773: -3.05495151202, 774: -11.8555183603, 775: -15.4952953552, 776: -2.5379826528, 777: 69.7072151781, 778: 11.3071293915, 779: -4.32570881358, 780: -9.06006250303, 781: 4.38396734044, 782: 15.427902363, 783: -7.86196358497, 784: -0.538381085543, 785: -1.48998238888, 786: -1.52013126055, 787: 3.44308530683, 788: -10.2910908578, 789: -16.7210242429, 790: 2.23218497038, 791: -14.0167729267, 792: 3.06033040786, 793: 4.16882127378, 794: 3.98773079506, 795: 4.20994919521, 796: 4.39386437864, 797: 0.439854822375, 798: 9.10309642811, 799: -78.723842658, 800: -1.50042976333, 801: 1.84628642548, 802: -0.192335857039, 803: 1.4257178857, 804: -91.2385275308, 805: 3.42112512094, 806: 0.829739359974, 807: 4.3200687413, 808: -8.13941404509, 809: 0.212885728737, 810: -6.92339468019, 811: -60.2236490724, 812: -22.7495817888, 813: 53.254513922, 814: 11.4194269803, 815: -8.25905895694, 816: 2.73174887131, 817: 40.4248727109, 818: -50.5050573849, 819: -78.9177297455, 820: 36.5848362403, 821: 2.47126251803, 822: -30.783603045, 823: 2.77380006138, 824: 0.681146189563, 825: 0.829142040509, 826: 10.4544570289, 827: -17.589382639, 828: 0.812663930928, 829: 4.2693873, 830: 1.00682759255, 831: -6.99118005366, 832: 7.04652534669, 833: 4.46697410894, 834: 4.91317988166, 835: 2.46959605649, 836: 28.0930365191, 837: 2.54814951462, 838: 7.50240638511, 839: 54.4413127691, 840: -14.0499281486, 841: -21.4023907009, 842: 0.816585830604, 843: 37.9792920043, 844: 0.77136546703, 845: 0.538783603947, 846: -35.9571385612, 847: 0.692699934446, 848: 3.74838195693, 849: -19.1500348344, 850: 0.238831668305, 851: 0.462438266483, 852: 0.957678737672, 853: 0.721570865446, 854: 2.45393869196, 855: -23.5743845994, 856: 1.50453534118, 857: 0.9548994436, 858: 0.889103374811, 859: 3.4352317817, 860: 3.65382639625, 861: -0.871569545834, 862: -51.9600453774, 863: 46.6994556943, 864: -2.89347192123, 865: -8.70976212679, 866: -16.3110590476, 867: 27.2994186567, 868: 21.7026137439, 869: -6.15444412715, 870: 0.470791514985, 871: -5.85753806861, 872: 0.521974512349, 873: -38.8144925031, 874: 40.711058774, 875: 1.51551128445, 876: -0.0255582806688, 877: 0.571124948774, 878: 1.13488104239, 879: 0.466181632693, 880: 1.21148902137, 881: -26.3016781928, 882: 17.4053607435, 883: -19.1750515302, 884: -22.4528090263, 885: -28.2137849891, 886: 7.73154129423, 887: 2.08787601152, 888: -7.76053639529, 889: -26.4196879832, 890: 6.57416903479, 891: 0.640099463139, 892: 2.79359740016, 893: -3.46610045002, 894: -1.09177219285, 895: -1.38704404832, 896: -92.5924513006, 897: 0.570877824878, 898: 3.47191329728, 899: 0.518841630446, 900: -9.85096275367, 901: 0.593947270074, 902: 0.548860523028, 903: -7.85689374449, 904: 3.56201719474, 905: 16.9924642782, 906: 0.474026212635, 907: 0.697089969243, 908: 1.05522572916, 909: 0.618139220289, 910: 4.53338402697, 911: 1.64477846523, 912: 2.22094427741, 913: 0.681034321937, 914: 0.608358415309, 915: 33.5610251476, 916: 13.1685484592, 917: -39.9928054732, 918: -13.9120387652, 919: 0.529149237332, 920: 24.4642475253, 921: 8.52072316189, 922: 2.08848915691, 923: -1.34450143672, 924: 0.42936875878, 925: 24.1927443938, 926: 2.17556862751, 927: -3.46873098278, 928: -2.34675511446, 929: 25.1225980427, 930: 0.573556170157, 931: -30.3821205592, 932: 1.92772459351, 933: 3.74801901219, 934: 0.838413745098, 935: -8.63354830682, 936: -18.5010319855, 937: -3.05195155254, 938: 4.25543242347, 939: -38.2619658615, 940: -6.8268548564, 941: -28.5680051584, 942: -17.1574129166, 943: 48.7323852338, 944: -13.9428030082, 945: -52.2480439347, 946: -47.0140711302, 947: -67.4980762993, 948: 35.0213149988, 949: -32.1445120934, 950: 42.9710766715, 951: 10.4730681878, 952: -41.4241892233, 953: -8.90794859669, 954: 1.78134478447, 955: 15.624990097, 956: 0.243046086517, 957: 3.67357561189, 958: -28.4525885695, 959: 19.4727578599, 960: 3.20650304694, 961: -11.1144778384, 962: 0.295211406865, 963: 2.52705396265, 964: -24.1294756727, 965: -18.5276732132, 966: 5.16061815455, 967: -0.957349167929, 968: 2.54110176182, 969: 4.48667707284, 970: -20.9040111105, 971: -36.4495130725, 972: 36.5038408243, 973: 27.5230584702, 974: -27.4235716766, 975: 22.366919837, 976: -27.4670390635, 977: 25.530797807, 978: 18.5355136129, 979: 5.13503182537, 980: 13.5966630981, 981: -40.368733929, 982: 60.1964518735, 983: -13.9542933617, 984: -7.04447751132, 985: 5.21206054969, 986: -19.0544469788, 987: 4.29033454018, 988: -5.55566058363, 989: -49.2212069039, 990: -27.8445413862, 991: -8.20895842345, 992: -5.44488538557, 993: -15.1987302763, 994: 9.45454241469, 995: 46.3823450685, 996: -20.4216062385, 997: -15.7272413758, 998: -22.5626849823, 999: 3.5467587692, 1000: -59.9811524933, 1001: -8.07145923396, 1002: 24.4109238973, 1003: -8.78035006163, 1004: 4.054581911, 1005: -20.2906672344, 1006: 3.47500912351, 1007: 3.49809918541, 1008: 0.497557007815, 1009: -17.513155395, 1010: 4.38617621138, 1011: 1.26026833548, 1012: 58.8517655758, 1013: 3.42587843545, 1014: 15.279315333, 1015: 3.83308518746, 1016: 1.95589026946, 1017: 44.8841931159, 1018: 2.08842464993, 1019: 4.47270997351, 1020: 1.87703431538, 1021: 1.56047394938, 1022: 3.02362996315, 1023: 2.63814556278, 1024: 61.0379059325, 1025: 22.5002982069, 1026: -32.0530105495, 1027: 19.693343591, 1028: -49.060329873, 1029: 2.07568390655, 1030: 17.4948819997, 1031: 0.583166678731, 1032: -6.39062021223, 1033: 4.70323548266, 1034: 13.2842280091, 1035: 2.80204039259, 1036: 19.6487693964, 1037: 3.49791735935, 1038: 46.8945129782, 1039: 16.8226627756, 1040: 7.48013244367, 1041: -22.6132953522, 1042: -33.0247483781, 1043: -45.3471778275, 1044: -11.783173439, 1045: 75.7710618623, 1046: 32.3512183433, 1047: 30.6745583028, 1048: 0.800564111737, 1049: -0.736961955169, 1050: -0.574047903357, 1051: -6.9745804228, 1052: -3.76997455455, 1053: -8.27009482185, 1054: 2.70932582638, 1055: -6.31392314938, 1056: 6.91182259674, 1057: -5.64078858519, 1058: -9.31157534469, 1059: 0.81482043982, 1060: 1.96446971004, 1061: 3.35070171633, 1062: 4.39463292085, 1063: -2.31706692932, 1064: -10.7187083104, 1065: -9.98662718197, 1066: -6.45971135395, 1067: 2.30448116762, 1068: 8.00246063038, 1069: 15.0835985525, 1070: -1.17876978272, 1071: 1.19746927879, 1072: -16.0123115826, 1073: 2.34137397193, 1074: -1.19041997517, 1075: 4.5038284174, 1076: 0.67232228468, 1077: -0.230437642375, 1078: 6.77514751332, 1079: 3.47784572758, 1080: -3.49173808301, 1081: 0.191067420195, 1082: 4.96117015638, 1083: 5.09387809137, 1084: 2.8721970053, 1085: -10.6840759783, 1086: 6.40492451338, 1087: 7.90723139568, 1088: 13.3838065115, 1089: 5.19532602544, 1090: 1.7884891694, 1091: 12.8931717229, 1092: 0.508950339045, 1093: 2.2890219703, 1094: 11.3219713496, 1095: 1.95684129219, 1096: 11.7124140465, 1097: -12.5888794515, 1098: -1.59369837611, 1099: 3.06981081306, 1100: 5.3242531949, 1101: 6.33199914779, 1102: -3.68775536645, 1103: 3.04149737362, 1104: -0.956495500891, 1105: -0.274036814972, 1106: 15.1155157556, 1107: 3.29745642838, 1108: -0.358883623074, 1109: -10.8018512908, 1110: -1.32649034853, 1111: -1.22397438621, 1112: 2.82479359914, 1113: -0.119416733258, 1114: -9.81277194297, 1115: 3.69697492822, 1116: 5.90036080305, 1117: 3.43497821214, 1118: 3.75642744732, 1119: -15.7293184795, 1120: 4.71472012524, 1121: -6.87572500878, 1122: 4.77638949967, 1123: 7.22744036327, 1124: 0.49575321097, 1125: 7.80108213521, 1126: 1.17511637034, 1127: 13.6673047509, 1128: 1.74793013357, 1129: 0.481698152433, 1130: -1.19318646082, 1131: -9.14749397918, 1132: -6.16710624157, 1133: -0.0120414320498, 1134: 1.52949461296, 1135: 0.738941370971, 1136: 5.25444372432, 1137: -7.2153388084, 1138: 3.18737445974, 1139: 5.13267211236, 1140: -16.3758795102, 1141: 4.45365851254, 1142: 4.07271840484, 1143: 2.78067378443, 1144: -6.7858323418, 1145: 3.87769588422, 1146: -0.379311645319, 1147: 1.36041423157, 1148: -8.58058774491, 1149: -13.5352244628, 1150: -9.00478099246, 1151: 2.15351855671, 1152: -8.41201232477, 1153: -11.732396339, 1154: 4.86963742459, 1155: -2.75156321505, 1156: 4.56695003655, 1157: -14.4168276972, 1158: -3.38025619528, 1159: 0.926552061069, 1160: -0.337778466251, 1161: 21.3870457754, 1162: -7.90521236339, 1163: -4.9935786528, 1164: 3.3021775582, 1165: -0.129902166168, 1166: -2.96307847129, 1167: -1.61459106707, 1168: -9.50387075562, 1169: 11.28254311, 1170: 2.24818446405, 1171: 0.671707136461, 1172: 43.4371273128, 1173: 0.878480223902, 1174: 2.13953854845, 1175: 13.589665305, 1176: -17.4730987691, 1177: 0.976653286464, 1178: 4.13824816575, 1179: 5.39755429646, 1180: -11.6689857075, 1181: 1.1031475358, 1182: 3.10208331877, 1183: -7.62805029279, 1184: -0.16074495595, 1185: -3.32136892711, 1186: -0.435568405745, 1187: -9.05491093467, 1188: 1.5258581271, 1189: -9.39027307209, 1190: -2.52007751827, 1191: -1.54289344355, 1192: -1.52672074539, 1193: 0.611151581188, 1194: 0.903426385084, 1195: 12.7597263843, 1196: 0.782527299284, 1197: 4.59546773712, 1198: 2.96893693275, 1199: 1.66283641943, 1200: 1.08388836497, 1201: 1.09735302938, 1202: 0.958016298643, 1203: 3.19280854812, 1204: -14.7552532583, 1205: -1.38114564897, 1206: 0.354081719533, 1207: 0.143028891212, 1208: -0.327147381603, 1209: 10.1200536845, 1210: -11.8025957057, 1211: -1.73563257863, 1212: -9.07680513823, 1213: 34.302818121, 1214: -19.1346515367, 1215: 1.98137305645, 1216: 3.34701747558, 1217: -2.77645129691, 1218: 3.49177217221, 1219: 1.00798708849, 1220: -7.46117186944, 1221: 0.810776105582, 1222: -7.56244787916, 1223: 10.8775683354, 1224: -9.55676142564, 1225: 1.99420795053, 1226: 0.883353168806, 1227: 1.15075242649, 1228: 1.03347170385, 1229: 0.84483441893, 1230: 3.24796519163, 1231: -4.43418901182, 1232: -5.43769548223, 1233: -1.70334191081, 1234: -13.3825896573, 1235: -0.674676545719, 1236: 4.12718130254, 1237: -11.4837574383, 1238: 8.63608003515, 1239: 2.09126966274, 1240: 0.868337105085, 1241: 2.82052544249, 1242: -0.971176444962, 1243: 4.51133446854, 1244: -2.1240316629, 1245: -10.21532517, 1246: 1.02600929406, 1247: 2.25420341607, 1248: 0.857567323225, 1249: -8.14159426859, 1250: 0.865848575264, 1251: 0.847639632275, 1252: 0.355789752539, 1253: 0.861802845842, 1254: -3.0652957449, 1255: 1.16980372504, 1256: 0.907183131085, 1257: -0.468952710502, 1258: 0.906717234596, 1259: 2.07080929592, 1260: 3.21161923051, 1261: -6.1775589174, 1262: 1.08690364362, 1263: 0.713376927543, 1264: -13.077363916, 1265: -2.25532221302, 1266: -4.04321702838, 1267: 13.9289637432, 1268: -0.386135730787, 1269: -5.49722153649, 1270: -4.29005271675, 1271: 1.08491830475, 1272: 2.31924946713, 1273: 2.82780533244, 1274: 11.3293130202, 1275: 5.08129099281, 1276: 2.89341039445, 1277: -9.75145210449, 1278: 5.06577418919, 1279: 3.99394411171, 1280: -4.58412239547, 1281: -28.0199828395, 1282: 3.43193158462, 1283: -1.36487789004, 1284: -8.24100834995, 1285: 7.80968656249, 1286: 2.22304207976, 1287: 4.24735830299, 1288: 6.08219059531, 1289: 4.15214690483, 1290: 0.668491736866, 1291: 3.49547377858, 1292: 10.2450684003, 1293: -0.971776716817, 1294: -5.32974506941, 1295: 5.57967041112, 1296: -24.6149844095, 1297: -3.78650326357, 1298: 5.66539595288, 1299: 1.45244930175, 1300: -3.51597296587, 1301: -3.42470289606, 1302: -2.51344403223, 1303: -4.7034340995, 1304: -2.24973681727, 1305: 4.93300338371, 1306: 4.65116889908, 1307: 5.20080056306, 1308: 2.56898263398, 1309: 1.6624402618, 1310: 2.3133054039, 1311: 1.94897824444, 1312: 3.20593965149, 1313: 27.3022409617, 1314: -9.94201691278, 1315: -1.58321539501, 1316: 4.46130610032, 1317: 7.10605589921, 1318: -10.6662009003, 1319: -1.94509556948, 1320: 4.77475849548, 1321: -3.53038838671, 1322: -23.7641783234, 1323: -5.39782785451, 1324: -1.94864646074, 1325: -0.0886845242339, 1326: -3.32237725544, 1327: -3.93569601397, 1328: 2.32156757386, 1329: 7.60338802254, 1330: -20.7633655437, 1331: 3.28394782126, 1332: 7.31497488636, 1333: 8.36197089523, 1334: 3.85451163858, 1335: 0.188330849361, 1336: 2.63474554129, 1337: 9.75725042683, 1338: 1.00832579868, 1339: 0.904306559284, 1340: -5.0122781357, 1341: -1.13927325461, 1342: 1.30468716342, 1343: -0.866652417164, 1344: 6.80523366492, 1345: -13.3722856218, 1346: 4.8745392805, 1347: 3.22456972581, 1348: -1.46039172779, 1349: -1.01225587855, 1350: 0.466906327544, 1351: -4.92070805617, 1352: -2.43702124305, 1353: 4.02914234115, 1354: -7.08768272756, 1355: 2.25932399822, 1356: 2.23602012567, 1357: 1.59164944162, 1358: -0.811100993822, 1359: 3.80961138891, 1360: 2.2459391461, 1361: 10.1994794248, 1362: 0.827396655595, 1363: -5.0195457213, 1364: -1.73980716703, 1365: 5.07567010069, 1366: 17.3920485676, 1367: 2.68730137505, 1368: 1.63371882929, 1369: -9.86314711235, 1370: 2.63021759607, 1371: -3.79881390432, 1372: 11.7074567589, 1373: 5.375176615, 1374: -11.2244658197, 1375: -0.839790612856, 1376: -1.65171604638, 1377: -8.08516099658, 1378: -2.97943499553, 1379: -1.67467981658, 1380: -2.29413199317, 1381: -11.5119943792, 1382: 1.95730610518, 1383: 0.960359915072, 1384: -13.0709586838, 1385: 21.7002900111, 1386: 4.57585938566, 1387: 2.47666664825, 1388: -1.23214739978, 1389: 2.29562302878, 1390: -5.98947301027, 1391: 3.56328589598, 1392: -3.4762198354, 1393: -9.09178580174, 1394: 13.8407289878, 1395: 7.59548840878, 1396: -4.8316148946, 1397: -0.215229006206, 1398: -17.0935966551, 1399: 7.14026949678, 1400: 3.92097342387, 1401: -6.10169975492, 1402: -16.2087769009, 1403: -1.34294303306, 1404: 22.9399333643, 1405: -2.0508048297, 1406: -2.09521037903, 1407: -35.6587448151, 1408: 16.7935381532, 1409: 7.58614910474, 1410: 8.07543190351, 1411: -13.7863907376, 1412: -6.90050675266, 1413: 48.4956393943, 1414: -7.05842855425, 1415: 18.806987133, 1416: -9.41894092123, 1417: -25.5380085592, 1418: 12.5057457195, 1419: 16.528077908, 1420: 15.80996232, 1421: -14.4117335502, 1422: 16.2681027713, 1423: -14.007419128, 1424: 5.92415121969, 1425: 1.90050763154, 1426: 3.63808402566, 1427: -18.9957061888, 1428: 35.9676204477, 1429: 3.07602429397, 1430: 13.9893163169, 1431: 24.9140138194, 1432: -6.64866255137, 1433: 2.07661723925, 1434: -1.52711713459, 1435: -0.871827257557, 1436: 0.455511826199, 1437: 15.4817714891, 1438: 0.162046286726, 1439: 27.9261941862, 1440: -25.5619362781, 1441: -4.3897235604, 1442: 8.35978195359, 1443: -8.33867809653, 1444: -1.9256210087, 1445: 51.1542735332, 1446: -1.79714586052, 1447: 2.46955845671, 1448: -6.22247348361, 1449: -1.65714767928, 1450: 6.07673634849, 1451: 16.7517973773, 1452: -0.732214559421, 1453: -6.18140151075, 1454: 5.4047632732, 1455: 19.7909876219, 1456: 55.5788819365, 1457: -8.8491159962, 1458: -0.00711572085994, 1459: -10.3237914145, 1460: 0.647130151932, 1461: -0.604928342197, 1462: 3.34535277095, 1463: 32.967557388, 1464: -1.68208492567, 1465: 0.651965400241, 1466: -2.62462955979, 1467: -3.61134562387, 1468: -5.31886093102, 1469: -1.70858707066, 1470: 1.90779321661, 1471: -0.63447202199, 1472: 0.450029306504, 1473: -5.87105511637, 1474: -21.2667788192, 1475: 11.6175052907, 1476: -56.9100846703, 1477: -2.01359363649, 1478: 21.0004771633, 1479: -3.63052107177, 1480: -5.47074559223, 1481: 30.3336662126, 1482: -10.6594173689, 1483: -1.98167775506, 1484: -11.0200873415, 1485: 5.93153701978, 1486: 1.24476193535, 1487: -0.953224602278, 1488: -0.948819287597, 1489: -75.8012948772, 1490: -1.56579445924, 1491: -0.554451391223, 1492: -2.15687999586, 1493: 7.66103387265, 1494: -1.87925699458, 1495: -0.190099380874, 1496: -1.79214049678, 1497: 26.315933367, 1498: -1.60360391953, 1499: 9.21466817104, 1500: 1.80192330849, 1501: -4.4528026605, 1502: 33.4326780779, 1503: -1.59267123324, 1504: -0.18462290287, 1505: 5.29982639878, 1506: -2.77776857769, 1507: -3.85079030557, 1508: -9.41370335927, 1509: -32.0866443842, 1510: -3.0469112024, 1511: -26.8220048348, 1512: -9.66402307618, 1513: -58.5385008513, 1514: 8.12373998268, 1515: 59.4089032573, 1516: 29.0094711392, 1517: 7.00524336469, 1518: -11.5167566143, 1519: -27.4737225799, 1520: -1.41381176739, 1521: -9.14013537117, 1522: -0.373848212222, 1523: -0.98252335305, 1524: -0.782838449978, 1525: 21.521186228, 1526: -0.363259151421, 1527: -1.06150320358, 1528: -0.149513343389, 1529: -2.74815969986, 1530: 0.511963866134, 1531: 7.27440989678, 1532: -34.5826593558, 1533: -0.727746402917, 1534: 3.51822189859, 1535: 1.20696318716, 1536: 5.99505792483, 1537: 44.1703478264, 1538: -8.40686238602, 1539: 26.1614272856, 1540: -0.795034187602, 1541: 0.472735911431, 1542: -0.15611830002, 1543: -0.135169604974, 1544: -19.2227923945, 1545: -0.323933022901, 1546: -1.52638121316, 1547: -11.7070041022, 1548: 0.181966920954, 1549: -0.357108891968, 1550: -0.533021271213, 1551: -0.149937461057, 1552: -1.83804952336, 1553: -14.2900553371, 1554: 0.550941971384, 1555: -1.6306516703, 1556: 4.82876765894, 1557: -27.496268853, 1558: 5.33861323851, 1559: 11.0509584849, 1560: -9.38794010319, 1561: -13.0422669657, 1562: 15.1466294257, 1563: -4.0773681135, 1564: -8.41047640998, 1565: -18.1306353609, 1566: -13.0871970792, 1567: -5.44593234981, 1568: -0.200568818803, 1569: 0.278549173938, 1570: -0.270683047044, 1571: -20.0082995091, 1572: 24.4329427158, 1573: 15.3500745039, 1574: -1.14156749043, 1575: -0.297213936428, 1576: -0.103575292839, 1577: -0.231868170507, 1578: -0.314297866737, 1579: -6.37731057119, 1580: 36.8680181697, 1581: 3.41549663406, 1582: -2.46909684689, 1583: 17.5549090902, 1584: 7.20017535302, 1585: -4.31493620158, 1586: 14.6382232288, 1587: 6.88787079783, 1588: 0.560863734002, 1589: -0.167131060233, 1590: 36.4691451784, 1591: -8.20633825016, 1592: -2.71417463684, 1593: 6.75799583448, 1594: -20.1925459285, 1595: -0.335497743942, 1596: -1.72158877385, 1597: -0.600478160128, 1598: -1.86670208702, 1599: -0.289415773337, 1600: -0.375577484512, 1601: 10.2574197832, 1602: -1.56005436901, 1603: -0.421493950988, 1604: -0.188052243143, 1605: -0.20257443897, 1606: 0.832587136924, 1607: -0.391870592334, 1608: -1.42737420499, 1609: -0.0539245447435, 1610: 4.6494119722, 1611: -1.40269538489, 1612: -1.12120928736, 1613: -13.3552513149, 1614: 3.35159645433, 1615: -29.1461082255, 1616: -45.0474364009, 1617: -5.16198596122, 1618: -15.4583655396, 1619: -8.13931530929, 1620: 0.363581886866, 1621: -0.714366699267, 1622: 0.153283848916, 1623: 4.56533287444, 1624: 0.548042814023, 1625: -0.617264746372, 1626: 0.64114629226, 1627: -17.2915361205, 1628: -0.195485992317, 1629: -4.73344278529, 1630: 1.89744216626, 1631: 2.01301602665, 1632: 6.34028594515, 1633: -9.16557971275, 1634: -57.1350139298, 1635: -1.39930263637, 1636: -1.18420398801, 1637: 0.556905034852, 1638: -11.2236012209, 1639: 29.6792117931, 1640: 3.13706954862, 1641: 14.7728566597, 1642: 25.8186424506, 1643: -23.2581913146, 1644: 1.60613762433, 1645: 35.9887892611, 1646: -2.67240195101, 1647: 9.83851709374, 1648: -19.1719943467, 1649: 25.0995777217, 1650: 23.9022501868, 1651: -7.16528956305, 1652: -2.69889981798, 1653: 16.7272529089, 1654: 3.9354417948, 1655: -1.56144913729, 1656: -8.45597835813, 1657: -14.5612896737, 1658: -0.966097139884, 1659: 2.42187563098, 1660: 6.14915877754, 1661: -1.84531963973, 1662: 22.3202713858, 1663: -12.6931276388, 1664: 29.3285431146, 1665: 3.70378768475, 1666: -11.565283769, 1667: 3.07449289622, 1668: -9.20673347881, 1669: 15.1919658135, 1670: 19.7941083956, 1671: -3.70937293586, 1672: -2.84456986748, 1673: -11.8464914087, 1674: -0.496950623917, 1675: -18.6242951673, 1676: 3.07641923467, 1677: 2.39513703542, 1678: -17.6428916379, 1679: -20.3780164823, 1680: -51.0948352271, 1681: 2.49059549642, 1682: -9.04789767495, 1683: -1.77568172069, 1684: 2.57274656527, 1685: -2.13598280978, 1686: 5.26740095929, 1687: -3.11778300734, 1688: 1.76367882313, 1689: 20.6396217739, 1690: -12.640247701, 1691: -26.08042361, 1692: -10.1382332139, 1693: -4.81944372842, 1694: 30.4001962613, 1695: -10.010948582, 1696: -42.9297996125, 1697: -1.76067417467, 1698: -40.061666961, 1699: -5.84795166103, 1700: 24.4804907231, 1701: 81.0024329901, 1702: -0.722258725699, 1703: -1.04638889193, 1704: -1.83505999018, 1705: -1.68148904646, 1706: -2.50946698822, 1707: -21.7249232805, 1708: -1.54878932053, 1709: 0.0200470305095, 1710: -53.2076996942, 1711: -1.59525154331, 1712: -25.6717213732, 1713: -2.26250840215, 1714: -1.61069210226, 1715: -3.89287799543, 1716: 0.869164700445, 1717: -1.36571639919, 1718: -14.5356432518, 1719: 17.3504931496, 1720: 28.9008071509, 1721: 37.1627926779, 1722: 1.1355503179, 1723: 18.354011253, 1724: 23.3211781269, 1725: 56.7868506642, 1726: -9.74124244698, 1727: 16.7762459338, 1728: 31.6740820828, 1729: -0.0961030440898, 1730: -1.00951522926, 1731: -9.40559083731, 1732: -36.489169231, 1733: -8.61132056146, 1734: -39.8173304279, 1735: -1.77209497719, 1736: 27.2423879684, 1737: 7.63074859144, 1738: 3.56895885479, 1739: 4.84358619589, 1740: -28.3955582921, 1741: 1.08679264286, 1742: 64.6030423659, 1743: 100.0, 1744: -19.5425014177, 1745: -4.93686572482, 1746: 0.324958772178, 1747: 6.87701874029, 1748: 24.4783163702, 1749: -4.02890050866, 1750: -35.6124056421, 1751: -12.8497529367, 1752: 6.8553540151, 1753: 21.4070538419, 1754: 4.02336563537, 1755: -9.1574428871, 1756: -31.1001139933, 1757: -10.3974030224, 1758: 12.6352511466, 1759: 10.8061283809, 1760: 17.7267408575, 1761: -8.89394489153, 1762: -52.2475389326, 1763: -17.2264567863, 1764: -25.4038331259, 1765: -1.64940640042, 1766: 28.0227053335, 1767: 34.4773503826, 1768: 15.5290096948, 1769: 18.1375229961, 1770: -36.5590022165, 1771: -8.71005631695, 1772: 22.4610155366, 1773: 8.57857098522, 1774: 9.18818630751, 1775: 20.32839646, 1776: 11.603803693, 1777: 6.87541054265, 1778: -13.0890859217, 1779: -7.13526629508, 1780: 4.06042783369, 1781: 23.9044454807, 1782: 12.1900769019, 1783: -46.1542279591, 1784: 26.9209671914, 1785: 15.5284346377, 1786: -5.44690373253, 1787: 7.13797890952, 1788: 1.62326372765, 1789: 13.8077292578, 1790: 15.6366333124, 1791: -0.827712778357, 1792: 16.7473085456, 1793: 8.39562731883, 1794: 12.356926441, 1795: -15.0359234633, 1796: 18.9650314042, 1797: 14.9906594692, 1798: 0.780220088778, 1799: 0.124551244845, 1800: -25.2579400307, 1801: -1.02105746909, 1802: 38.9394392432, 1803: -7.52845825835, 1804: 27.8160842613, 1805: -12.3284925664, 1806: -50.9611601082, 1807: -25.5407068858, 1808: 17.4568256064, 1809: 12.8714545911, 1810: -46.4880733338, 1811: 1.23343337132, 1812: -17.6414982031, 1813: 1.42456293264, 1814: 9.39298046174, 1815: 0.339521234642, 1816: 4.01548006658, 1817: -12.2216514994, 1818: 1.08602661948, 1819: -4.19863532576, 1820: -10.46700606, 1821: 6.24736509929, 1822: 12.3183907105, 1823: 0.318525241134, 1824: -49.9924625865, 1825: 36.8913803209, 1826: -0.408764166383, 1827: -17.7803773835, 1828: 1.66357543368, 1829: -0.288310241033, 1830: -20.8438123512, 1831: 0.471508754188, 1832: -9.89417286911, 1833: 12.3632594859, 1834: 3.06328068553, 1835: -3.26908568261, 1836: 7.50571101916, 1837: 0.244618442479, 1838: -62.2108053065, 1839: 1.59568806101, 1840: 1.86866277283, 1841: 2.81036222983, 1842: 19.9592785532, 1843: 2.69709779026, 1844: 21.235356619, 1845: 4.39254347408, 1846: -49.056828422, 1847: 6.51255230701, 1848: -4.19738744773, 1849: 2.40365727649, 1850: -5.0313137843, 1851: -21.0415185414, 1852: 0.0869130704302, 1853: -3.76144132863, 1854: 3.03577092249, 1855: -28.082751131, 1856: -0.15695805409, 1857: -26.2991588585, 1858: -10.9926929655, 1859: 1.04459678787, 1860: 11.5369961227, 1861: -11.8633865114, 1862: -1.84857635153, 1863: -45.9583693916, 1864: 15.8665428486, 1865: -48.2085682314, 1866: -25.687616816, 1867: -16.7703330619, 1868: 12.0743014084, 1869: -5.0155301514, 1870: -16.1712462123, 1871: 0.219134166237, 1872: 1.24800670418, 1873: 4.82334958862, 1874: -35.68412746, 1875: 0.386384369727, 1876: 1.85055729824, 1877: -5.08334081343, 1878: 20.0454250852, 1879: 1.83587224867, 1880: 8.75361855912, 1881: 2.50287193696, 1882: 3.78481557994, 1883: 22.7054942518, 1884: -8.41589863696, 1885: -3.53721790958, 1886: 40.7740333288, 1887: -4.61152879159, 1888: -20.9551406192, 1889: 4.19293362149, 1890: -22.8699799491, 1891: 0.367180972214, 1892: 0.214816277618, 1893: -16.488873229, 1894: 0.116293486718, 1895: 1.68493210488, 1896: 18.2938381382, 1897: 1.21318779973, 1898: 0.211807715883, 1899: 0.535333994202, 1900: 0.0564371560922, 1901: 3.95907168218, 1902: 13.3389497989, 1903: -0.509243753315, 1904: 1.60488268731, 1905: 0.766765412939, 1906: 6.87460108048, 1907: 0.0297373980752, 1908: 0.930934151997, 1909: 30.0678044101, 1910: -64.6358921855, 1911: -45.2201935169, 1912: 4.4485282179, 1913: -0.860953226461, 1914: -25.2925197097, 1915: -2.72373462686, 1916: 0.244794035614, 1917: -0.751488129181, 1918: -5.41020612677, 1919: 0.219818052963, 1920: 31.8528941752, 1921: 17.5916200726, 1922: 32.4109818049, 1923: 1.39843666034, 1924: 0.302079515809, 1925: 0.359001666856, 1926: -0.288789162055, 1927: 0.114050491348, 1928: 11.388210095, 1929: 9.95480275321, 1930: -0.234255368892, 1931: -12.5377723374, 1932: 63.5798858957, 1933: 5.65122275031, 1934: -0.0349579936222, 1935: -9.33087105701, 1936: -17.1343741591, 1937: -19.7482311302, 1938: 0.266200843025, 1939: 50.7608482075, 1940: 13.0057902073, 1941: -3.91823819095, 1942: 0.163858488204, 1943: -16.6212635485, 1944: 0.29775234899, 1945: 1.47433373342, 1946: 0.264151934822, 1947: 5.07140453606, 1948: 0.327317650083, 1949: 0.224760852517, 1950: 0.0501399948348, 1951: -0.458358950962, 1952: 38.1168487224, 1953: 0.231559352986, 1954: 0.334547204074, 1955: 0.792327468305, 1956: 0.624015070501, 1957: 2.18810599615, 1958: -15.0366919155, 1959: 17.9446049712, 1960: 0.292394924625, 1961: 2.58894110263, 1962: -31.8823996008, 1963: 19.0741440876, 1964: -0.230779329155, 1965: -2.34609947325, 1966: 7.66715699124, 1967: -3.90679642892, 1968: 2.18750764672, 1969: 1.22975530178, 1970: -5.51446279455, 1971: -3.84432258899, 1972: -10.9951793844, 1973: -1.91944418082, 1974: 6.95478462149, 1975: 5.45835306618, 1976: 8.95743314087, 1977: 0.23927543245, 1978: 24.9060984032, 1979: 1.41226850599, 1980: 1.41293792955, 1981: 21.3960108129, 1982: -24.1559044929, 1983: -11.3988970967, 1984: 0.366947260671, 1985: 1.90694284304, 1986: 42.4732626845, 1987: -35.1419967288, 1988: -12.3019336006, 1989: -3.2419198745, 1990: -20.1096265268, 1991: -17.0813502445, 1992: 41.0692790127, 1993: 30.5458834458, 1994: 46.088619748, 1995: -8.31400339804, 1996: -13.8057035319, 1997: 15.7447857136, 1998: -3.0883590568, 1999: 16.8749779623, 2000: -9.41742989279, 2001: 8.1044214207, 2002: 18.6222787843, 2003: 2.6726579996, 2004: 1.60819505136, 2005: 14.2064541714, 2006: -22.1213022325, 2007: 1.24952480007, 2008: -2.67583121012, 2009: 1.83289812854, 2010: 3.9655000798, 2011: 11.2188202646, 2012: 30.517330827, 2013: -6.57477748849, 2014: -8.90976233788, 2015: 7.53372813037, 2016: 7.31131803709, 2017: 5.15438757774, 2018: 13.5938599014, 2019: 16.1482242148, 2020: 22.0639115668, 2021: 20.0482480723, 2022: -2.36439925298, 2023: 0.358030154598, 2024: -6.3546725441, 2025: 16.0357024348, 2026: 27.4172324609, 2027: -25.4544859998, 2028: -29.8922545486, 2029: 18.4801079325, 2030: 24.1780597993, 2031: -13.9622140892, 2032: 2.61324403834, 2033: -1.01314379534, 2034: 8.31046334535, 2035: -2.39913903074, 2036: 0.683058979288, 2037: -29.9226758716, 2038: -9.15369483344, 2039: 11.3065010004, 2040: -62.3209097743, 2041: 44.4863153737, 2042: -9.10314904308, 2043: -45.5133681884, 2044: -7.95014984201, 2045: 22.3827837841, 2046: 27.4179564909, 2047: 2.59269491871, 2048: 2.04842661728, 2049: -13.7550076525, 2050: 18.7453925319, 2051: 1.73995710825, 2052: -10.1374590479, 2053: 1.53870407774, 2054: 1.46073305702, 2055: -2.4633712048, 2056: 7.74863563288, 2057: 2.14261817307, 2058: -0.62570672094, 2059: 28.9445132095, 2060: -0.385922878576, 2061: -89.0710669925, 2062: 3.43221459601, 2063: -2.52539906716, 2064: 34.0962820761, 2065: -12.2297858695, 2066: 1.40433142779, 2067: -38.1011685659, 2068: -6.28210125637, 2069: -3.8563731728, 2070: -13.2940648952, 2071: 5.22149227246, 2072: -8.68275845165, 2073: -5.92581486475, 2074: -13.866095083, 2075: 6.22382255651, 2076: -10.1837741574, 2077: -7.84471662677, 2078: -1.40494335259, 2079: -18.3947018793, 2080: 24.9618412455, 2081: -44.2333748771, 2082: -3.5608764956, 2083: 100.0, 2084: 1.15069697857, 2085: 14.4879919771, 2086: 24.4545149055, 2087: -30.9787611764, 2088: -37.504455949, 2089: 22.2110234712, 2090: -22.0007862644, 2091: -49.1483065303, 2092: 5.42555168236, 2093: -36.6900907329, 2094: -4.81818853271, 2095: -0.616940546766, 2096: -34.0767730794, 2097: 1.4651692983, 2098: -17.9950442549, 2099: -10.4047781874, 2100: 8.9896043086, 2101: 26.4747710142, 2102: 5.63228662288, 2103: 1.63274190342, 2104: -2.89410134164, 2105: 24.1413089562, 2106: 7.50537892696, 2107: -7.26991086914, 2108: -11.1876808442, 2109: -6.39802867847, 2110: -6.97663989091, 2111: 15.5786895951, 2112: -12.7211210478, 2113: -8.99648571263, 2114: -10.5938821159, 2115: -23.5420146318, 2116: -17.6322136998, 2117: -17.3926136447, 2118: 19.0506640581, 2119: 9.44838843098, 2120: 17.0770089707, 2121: 7.95619703729, 2122: 3.80778215092, 2123: -12.9115053141, 2124: -10.2614557242, 2125: -32.7897217199, 2126: -6.72624952148, 2127: -2.04035497919, 2128: -13.3484336419, 2129: 10.4912143059, 2130: -3.136962799, 2131: -9.4656668301, 2132: -13.1005812993, 2133: -10.1585488736, 2134: 6.7501691809, 2135: 2.03777703454, 2136: -4.14631150153, 2137: -21.1971850732, 2138: 0.625575755709, 2139: 4.57193445003, 2140: 8.24079489635, 2141: 19.8376738271, 2142: 3.71525722001, 2143: 1.58607809709, 2144: 6.52454517537, 2145: -3.25946378006, 2146: -1.5395368935, 2147: -4.35195253391, 2148: 2.4628129307, 2149: -5.27392922535, 2150: -3.82707437115, 2151: -8.40396483, 2152: 2.88645853298, 2153: -26.1248562051, 2154: 7.52752749505, 2155: -21.851869391, 2156: 7.58729904812, 2157: -17.5499891763, 2158: -5.81929312203, 2159: 11.3003552598, 2160: -0.308201026238, 2161: 7.568755541, 2162: -3.74679444802, 2163: -0.0733803326629, 2164: -3.73695080609, 2165: -1.55133488737, 2166: 1.53079431347, 2167: -4.20452707607, 2168: 8.90621082453, 2169: -1.48971810188, 2170: -20.4115828975, 2171: -7.74904466647, 2172: -5.67255313605, 2173: -29.1413077644, 2174: -10.0283786649, 2175: 4.7801876126, 2176: -9.90261731931, 2177: 0.452463371866, 2178: 7.20515187202, 2179: 4.17528037391, 2180: -1.08941197248, 2181: 14.0588050884, 2182: -12.2759140292, 2183: -4.67320522083, 2184: 13.5189596466, 2185: 24.9951626358, 2186: -3.9573552434, 2187: -7.92135811445, 2188: -3.72599612834, 2189: -3.71487287444, 2190: -2.66658876118, 2191: 1.44641700022, 2192: -2.64021062644, 2193: -2.42959281002, 2194: 4.80775353101, 2195: 12.0955293999, 2196: -2.53082335381, 2197: -4.83593446162, 2198: -1.27769894433, 2199: 4.3524488937, 2200: 7.45594051488, 2201: -3.89099751058, 2202: -5.34707620817, 2203: -4.09622033517, 2204: 19.7598295017, 2205: -2.46511472032, 2206: 5.70178559787, 2207: -32.278981977, 2208: -20.7722535211, 2209: 6.14301713306, 2210: 2.46715490984, 2211: -5.39376046698, 2212: -9.58943688355, 2213: 9.27403473979, 2214: 15.7681782605, 2215: 12.4866955078, 2216: -3.75138143954, 2217: 4.85061656917, 2218: 4.90675768287, 2219: -1.08883644388, 2220: -0.902753159702, 2221: -0.297949394815, 2222: 28.3271655571, 2223: -3.21868725411, 2224: -0.63459490643, 2225: -3.65249170582, 2226: -0.284435611159, 2227: -10.5873257064, 2228: 0.767727018663, 2229: -1.21252261986, 2230: -5.69050497003, 2231: 2.7070399559, 2232: 3.38346158481, 2233: -1.05658447656, 2234: -1.83355931811, 2235: -23.9006047517, 2236: -0.430389518507, 2237: -3.83473797897, 2238: -0.964284372888, 2239: -2.556222888, 2240: -0.0271721710799, 2241: -0.793182336277, 2242: 9.60984400741, 2243: -0.735320844794, 2244: -4.00092206683, 2245: -4.47245938809, 2246: -0.649098801852, 2247: -0.839890450548, 2248: -0.906572576572, 2249: -0.84909378528, 2250: 1.3403658492, 2251: 2.42783294873, 2252: -0.450488800406, 2253: 2.30186004268, 2254: -0.293175926882, 2255: -4.08007245702, 2256: -4.21763851621, 2257: 5.33860081804, 2258: 1.25299348103, 2259: 29.1242377346, 2260: 0.0603821157591, 2261: 20.9204194263, 2262: -1.44280668034, 2263: 16.7999555662, 2264: -1.26446726932, 2265: 14.5999900277, 2266: -0.745208262578, 2267: -7.32809795731, 2268: -0.652766113904, 2269: -17.2749329015, 2270: -14.3892209324, 2271: -17.7319998365, 2272: 0.288552842105, 2273: -0.788622159229, 2274: -0.849014222411, 2275: -2.27973117799, 2276: -0.155283684893, 2277: -7.44601990864, 2278: -27.969113965, 2279: 3.20846514332, 2280: 38.2438773158, 2281: -14.9544370097, 2282: -2.22579689819, 2283: 11.5899355844, 2284: -0.331896216747, 2285: 0.263657536715, 2286: -3.79644358456, 2287: -0.734018541728, 2288: -38.5505451815, 2289: -0.527635723762, 2290: -1.26145058164, 2291: -9.38531323363, 2292: 1.06931614333, 2293: -0.830959206704, 2294: -3.6878564617, 2295: -0.688215150435, 2296: -12.9621306484, 2297: -0.780202237983, 2298: -0.73197145998, 2299: -9.33319927843, 2300: -3.91188592289, 2301: -11.3333553695, 2302: -0.707267871003, 2303: -0.825864654436, 2304: 1.9671122062, 2305: -0.811533539354, 2306: -0.620393367965, 2307: -4.23874062214, 2308: 47.2000488512, 2309: -0.930718266878, 2310: -1.92151425185, 2311: 16.5883383849, 2312: 5.96582459281, 2313: -29.0583861935, 2314: -4.83001851854, 2315: 4.5056125953, 2316: 4.10226895129, 2317: -3.9116726989, 2318: -1.49723939839, 2319: -1.98927554346, 2320: -0.85015424751, 2321: -5.64815286276, 2322: -15.0457973455, 2323: -0.96785820175, 2324: 4.83879844703, 2325: -0.989074956133, 2326: -2.69668448827, 2327: -21.2229433115, 2328: 6.66224764421, 2329: 3.63844939207, 2330: 1.48333089937, 2331: -8.04094924963, 2332: -3.41151471131, 2333: -4.70441702621, 2334: -3.52749323788, 2335: -15.7252498498, 2336: 5.94719629211, 2337: 11.1441017682, 2338: 14.8118424951, 2339: 14.7113028645, 2340: 11.7609045501, 2341: 0.119341964064, 2342: -7.35048623359, 2343: -38.3681740734, 2344: 17.482092116, 2345: 1.123159169, 2346: 18.0728584202, 2347: -10.3622586395, 2348: 6.45264134954, 2349: -13.9507396521, 2350: -2.75468086447, 2351: 1.58189847598, 2352: -2.98459263688, 2353: -3.94992376231, 2354: -4.64828402605, 2355: -0.79962749408, 2356: -0.122276928987, 2357: -3.23013830779, 2358: -9.02928544517, 2359: 1.61980326237, 2360: 2.09269437599, 2361: 1.21305002673, 2362: 19.0384205361, 2363: 8.53928281914, 2364: -4.58554932773, 2365: -3.45095046178, 2366: -0.482973223073, 2367: -8.22743896637, 2368: 21.7260100217, 2369: 11.1173510698, 2370: -14.7145636515, 2371: 8.69400185794, 2372: 3.37024328922, 2373: -2.2148428815, 2374: -19.2073567177, 2375: -4.18085861195, 2376: 5.87510128717, 2377: -20.0878242545, 2378: 10.2939620932, 2379: 16.264441375, 2380: -17.1151241913, 2381: -3.34695358022, 2382: -2.37302854864, 2383: 11.7958070507, 2384: -8.54221763489, 2385: -1.41434038127, 2386: 19.9991063807, 2387: 10.7478297329, 2388: -14.8846570488, 2389: -9.84727713308, 2390: -1.84660133297, 2391: 9.72569518413, 2392: 20.0803925225, 2393: 2.3513634446, 2394: -11.5961113136, 2395: -4.15356422333, 2396: 20.9815642117, 2397: -0.207022278185, 2398: -2.18492487089, 2399: -10.8935791625, 2400: -3.81035428793, 2401: 23.2232327602, 2402: -3.62440456101, 2403: -3.6689463932, 2404: 0.269304444806, 2405: -10.3444443323, 2406: -3.4615501695, 2407: -4.92968043578, 2408: 9.28377490734, 2409: -3.81362996896, 2410: 4.54720676929, 2411: -3.60028384172, 2412: -4.5472474435, 2413: -6.17931090242, 2414: -4.08234379101, 2415: 0.843525807601, 2416: -16.8479883426, 2417: -13.0538929527, 2418: 13.9006177286, 2419: -4.04658688478, 2420: -8.49142896039, 2421: 4.30422684326, 2422: 18.8286517917, 2423: 3.81313087986, 2424: -3.69084261272, 2425: -5.93102268854, 2426: -5.07808204832, 2427: -0.827645021665, 2428: 10.8281287337, 2429: -6.99795753415, 2430: 49.5443318873, 2431: 9.73516051068, 2432: -4.51680196283, 2433: -3.94337134803, 2434: 13.287545703, 2435: 7.33035815718, 2436: 9.92970017738, 2437: 20.2654138636, 2438: -0.937190962882, 2439: 0.29543709582, 2440: 27.8065219434, 2441: -19.2878544212, 2442: -16.9706145632, 2443: 26.2393707312, 2444: 0.567269386076, 2445: 44.2179748457, 2446: 23.4920462888, 2447: -55.0841822049, 2448: -68.7566593222, 2449: -21.9927778888, 2450: -16.6421093009, 2451: 19.7838324037, 2452: -16.2940944619, 2453: -14.5262289171, 2454: 23.078227512, 2455: 6.52921218845, 2456: 36.2765338056, 2457: -0.0725929301526, 2458: 3.95057262096, 2459: -33.981673412, 2460: 3.15811420652, 2461: 9.46551280713, 2462: 0.214643690322, 2463: -11.5661683736, 2464: 27.0862575549, 2465: 34.8136082343, 2466: 8.09138161199, 2467: 7.05104513951, 2468: -11.3694428255, 2469: -39.8125748136, 2470: 16.6443873041, 2471: 0.0733217447622, 2472: -66.8149112025, 2473: 6.43631741941, 2474: -31.7987477927, 2475: -0.726745076818, 2476: 0.832299897856, 2477: -13.7929523882, 2478: 2.24051480642, 2479: -2.39291924784, 2480: 6.11786537084, 2481: -23.6085956189, 2482: -3.60335070528, 2483: 5.50846587605, 2484: -5.71449927521, 2485: 18.6045512924, 2486: -25.1421624173, 2487: -52.0661483936, 2488: 3.94807384218, 2489: 10.8164065702, 2490: -3.37961026079, 2491: 1.16854406163, 2492: 6.24373340983, 2493: -27.950300865, 2494: 3.43698009354, 2495: 7.9070830783, 2496: 3.71759900705, 2497: -0.269680886291, 2498: -10.6728186144, 2499: 3.70631896099, 2500: 23.06000656, 2501: 4.36161460424, 2502: -1.66838280354, 2503: 18.0708850096, 2504: 55.6703151441, 2505: 15.7207030822, 2506: 34.1807380693, 2507: -6.56534660014, 2508: 7.25639493152, 2509: 3.8251189576, 2510: 3.27600813338, 2511: 3.33631691837, 2512: 1.8580385628, 2513: 3.74119702579, 2514: 3.7078224568, 2515: 1.58352552083, 2516: 3.29118011706, 2517: -41.1898401527, 2518: 0.360743419242, 2519: 3.50594038577, 2520: -3.99997443473, 2521: 3.83717626719, 2522: -18.4382085684, 2523: 10.5068811096, 2524: 3.97776369343, 2525: 1.19057139899, 2526: 3.65287976102, 2527: 3.82782336054, 2528: -22.8130908794, 2529: -9.46076010894, 2530: 3.74043080749, 2531: 15.8914668218, 2532: 3.46535626875, 2533: -17.3991537286, 2534: -1.12668899079, 2535: 3.64266392159, 2536: 7.25194894004, 2537: 3.66785359383, 2538: 3.71470050607, 2539: 3.6953360426, 2540: 2.55341781027, 2541: 3.63236640921, 2542: 3.43303414982, 2543: 3.84927643975, 2544: 12.0924521467, 2545: 3.52295615339, 2546: 1.18495071894, 2547: 4.06849393257, 2548: 4.50713104249, 2549: 34.563715746, 2550: 3.57372331854, 2551: 3.80755919336, 2552: -6.01028507013, 2553: -13.4834082543, 2554: 3.77416633805, 2555: 9.16766634906, 2556: -38.3404306404, 2557: 21.0806816208, 2558: -41.3373165075, 2559: -36.0311444576, 2560: -28.5933973664, 2561: -12.8194416331, 2562: 10.4909078315, 2563: 18.0661298579, 2564: -42.3109713489, 2565: 11.5364805141, 2566: -2.64058857399, 2567: 15.1687428222, 2568: -45.0924685134, 2569: 0.503221188622, 2570: 1.12174183331, 2571: -0.40775365883, 2572: -2.26823353368, 2573: 0.835914361479, 2574: 3.54041763273, 2575: 3.59545782824, 2576: -44.8306114248, 2577: 2.59964103602, 2578: -0.809734295348, 2579: -8.29527428546, 2580: -1.92226217732, 2581: 26.235564295, 2582: 4.23788673869, 2583: 5.35892839527, 2584: 6.06106839745, 2585: 6.57309174236, 2586: 39.8564446581, 2587: 0.874761624155, 2588: -4.7464802978, 2589: 1.05382431265, 2590: 1.15439193725, 2591: 22.5149557432, 2592: 0.531413522393, 2593: 3.60124306208, 2594: -5.55009981454, 2595: 1.83014217379, 2596: 0.620965740947, 2597: 0.614251737449, 2598: 0.817370820644, 2599: 3.77169814448, 2600: -38.3837868293, 2601: 1.44970461688, 2602: 1.5333718554, 2603: -0.0559974961459, 2604: 2.75884779422, 2605: 2.93249075072, 2606: -26.4376545414, 2607: 4.33104947142, 2608: 36.6673220364, 2609: 26.4025617058, 2610: 2.66082793296, 2611: 20.72555239, 2612: -20.7621720261, 2613: -2.65001068522, 2614: -9.71519106517, 2615: 1.21890879627, 2616: -11.4879379608, 2617: 0.769007608412, 2618: -63.8585803965, 2619: -0.317161627385, 2620: 3.63569809021, 2621: 2.04810055589, 2622: 0.663943221826, 2623: 0.967498652472, 2624: 0.44052253664, 2625: 0.79559382713, 2626: -2.09181763022, 2627: -12.3815198386, 2628: -10.5624306967, 2629: -12.4135846082, 2630: -5.04851914345, 2631: 23.2833180778, 2632: 15.9094915077, 2633: -7.14272892695, 2634: 3.4817035257, 2635: -29.637802549, 2636: 1.13276479188, 2637: 43.8435570419, 2638: 2.07678737055, 2639: 0.821945495204, 2640: 1.525578543, 2641: 18.2893644168, 2642: 0.69385671308, 2643: 3.75954549728, 2644: 0.671403159067, 2645: 15.0258772231, 2646: 0.626672857883, 2647: 0.759404239677, 2648: -38.2139927365, 2649: 3.78703845171, 2650: -17.3974255236, 2651: 0.60273457101, 2652: 0.702166713447, 2653: 1.96934327169, 2654: 0.637141311852, 2655: 3.62887184632, 2656: -14.1834825578, 2657: 11.7864050348, 2658: 0.340027536144, 2659: 0.0347041189998, 2660: -23.4720496255, 2661: -0.442149630772, 2662: 16.45787932, 2663: -14.3209334811, 2664: -0.954562231083, 2665: -7.11424088527, 2666: 14.0812163207, 2667: 3.15831091963, 2668: 11.8512114853, 2669: 1.36104875307, 2670: 11.431450254, 2671: 3.56904739055, 2672: 0.383121401987, 2673: 1.13277808279, 2674: 2.67291849245, 2675: 0.159651873049, 2676: -5.16646534129, 2677: 62.1106927357, 2678: 5.17188217041, 2679: 3.62173579481, 2680: -21.7262046432, 2681: -5.18532655508, 2682: 2.74174041428, 2683: 3.50964929969, 2684: -36.1348948098, 2685: 33.6711809724, 2686: -14.1068556604, 2687: -0.0799018146955, 2688: 18.3981653459, 2689: 59.2657361518, 2690: -27.4509118282, 2691: -10.2646257186, 2692: 14.4132889494, 2693: -6.12193995377, 2694: 17.7398806076, 2695: -93.8797646428, 2696: -5.37277388875, 2697: -31.3682745548, 2698: -4.81639207382, 2699: -17.5313393757, 2700: 25.5832340028, 2701: 3.76690051843, 2702: 3.58985101418, 2703: -6.14168194409, 2704: -10.7328425725, 2705: 3.7338761511, 2706: 5.25518722626, 2707: -2.13443181237, 2708: 3.83575300591, 2709: 30.6862583955, 2710: -9.08291118621, 2711: -4.91823937189, 2712: 22.8537982235, 2713: 2.56634744326, 2714: 3.19003967982, 2715: -19.2169781489, 2716: 16.9449709094, 2717: -5.9466919583, 2718: -9.57760930615, 2719: -32.3518098675, 2720: 7.41165416113, 2721: 22.2683161864, 2722: -6.28727248868, 2723: -16.0155204871, 2724: 4.59222136704, 2725: -5.03026517625, 2726: 27.418704138, 2727: 27.8949793552, 2728: 1.91323826517, 2729: -8.02541785161, 2730: 3.76920369979, 2731: -1.62842045107, 2732: 3.89924442127, 2733: 3.76766424677, 2734: 1.76522619223, 2735: -19.9702752897, 2736: -23.8919750548, 2737: -38.0902372524, 2738: 4.42327218925, 2739: -3.67376214056, 2740: 14.7405532443, 2741: 8.16608294667, 2742: 3.33477005323, 2743: -17.5022564593, 2744: 2.97417221145, 2745: -32.4540923165, 2746: 2.65745269465, 2747: -6.80361508513, 2748: 56.5594901405, 2749: 3.70836199167, 2750: -19.4500736123, 2751: 3.72518794644, 2752: 3.63149842425, 2753: 3.67850765888, 2754: 15.6433474346, 2755: 3.88408386715, 2756: 3.60593868947, 2757: -38.5220958404, 2758: 3.59913877908, 2759: 0.187753303275, 2760: -3.80599568851, 2761: 3.63291438668, 2762: -20.0179408206, 2763: 3.51738874167, 2764: 3.54859702007, 2765: -24.0665760775, 2766: 24.5890068149, 2767: 43.526753007, 2768: -44.8975434185, 2769: -16.0029619275, 2770: -13.2734057303, 2771: -31.1275090191, 2772: 47.3536176288, 2773: -53.1433597872, 2774: -6.02631217661, 2775: -15.518993548, 2776: 2.06692643571, 2777: -2.54707526314, 2778: -14.7607706008, 2779: -19.0936317662, 2780: 20.2524226594, 2781: 6.6374618759, 2782: 3.51864616159, 2783: -20.7609436338, 2784: -21.2792960365, 2785: -8.65710912171, 2786: 4.99735541944, 2787: -28.8976279446, 2788: -46.2646068353, 2789: 6.95705235966, 2790: 18.680003285, 2791: 30.2401516837, 2792: -25.9884069005, 2793: 0.0549911722424, 2794: 8.35319841126, 2795: 5.59398717868, 2796: -4.90911039563, 2797: -0.16607433759, 2798: -3.40732736681, 2799: 0.795143296709, 2800: 3.8997663916, 2801: 5.19915158468, 2802: -0.659065408506, 2803: -18.706736281, 2804: 0.434330762683, 2805: 0.59399756386, 2806: 4.68824145242, 2807: 3.73885851252, 2808: -3.33593102773, 2809: 0.0733780087392, 2810: -1.05916764812, 2811: 19.8783368747, 2812: 7.34607546794, 2813: 1.19512986968, 2814: -3.2319514499, 2815: 18.620811557, 2816: 7.29487189795, 2817: 2.02305134857, 2818: -31.3218093769, 2819: -0.0678570171124, 2820: 6.91778935396, 2821: -27.7234215326, 2822: 8.15183161688, 2823: 5.6204270451, 2824: 2.75670029489, 2825: -6.22910307597, 2826: -3.86988322556, 2827: -8.16730585826, 2828: 0.269095081909, 2829: -0.460434739726, 2830: 2.96662187936, 2831: -5.58737549927, 2832: 2.38279018446, 2833: -1.02876334352, 2834: -0.411106575964, 2835: 8.33370641887, 2836: 9.33865905767, 2837: -3.47905224496, 2838: -2.90237620664, 2839: 2.05208739165, 2840: -0.0733338597502, 2841: -9.09243299528, 2842: 2.16926096556, 2843: -10.7306987864, 2844: -9.7868649736, 2845: 2.38533722718, 2846: -7.40783414878, 2847: 0.344365499729, 2848: -1.20258289989, 2849: 7.90399247217, 2850: 3.50293148175, 2851: -0.394668552685, 2852: 0.243899828798, 2853: -13.3820475723, 2854: 4.11519884302, 2855: 11.2179533106, 2856: -10.129683865, 2857: 0.73022724756, 2858: -0.279362204797, 2859: 11.1131708324, 2860: 0.954964565679, 2861: 0.2828389047, 2862: 2.36079547258, 2863: 2.27286391514, 2864: 2.18406668769, 2865: 2.41522915903, 2866: -0.934385340513, 2867: -2.40571337072, 2868: 2.4912510706, 2869: -2.00364893725, 2870: 0.811179603967, 2871: 0.991122953098, 2872: 8.4556058255, 2873: 1.92088394746, 2874: -5.57288492961, 2875: 1.76397656398, 2876: 1.35754935585, 2877: -12.2998379477, 2878: -0.571962518084, 2879: -10.910078419, 2880: -7.91593419876, 2881: 0.121485982922, 2882: 2.36392023993, 2883: 0.778586554921, 2884: 2.48989984767, 2885: -9.75305246419, 2886: 2.39159094777, 2887: 1.4945824007, 2888: 1.46101673279, 2889: -6.9915687008, 2890: 2.50247801797, 2891: 2.32184496924, 2892: 0.346190673573, 2893: -2.68458327303, 2894: -10.5072998742, 2895: 4.59712216758, 2896: -0.297776980716, 2897: -10.2126710759, 2898: -60.2629447149, 2899: 2.46690198872, 2900: -0.0632596428594, 2901: 0.140447442741, 2902: -14.6939199062, 2903: -0.9437193455, 2904: 4.64541650996, 2905: 12.3566106482, 2906: -2.28113588372, 2907: 10.6834883425, 2908: -8.91323235668, 2909: 7.39085517503, 2910: 2.73814088966, 2911: -4.92802585493, 2912: -5.24422749092, 2913: 16.8411315684, 2914: 7.61099927504, 2915: -2.73519845669, 2916: -4.92516598989, 2917: 2.57084667901, 2918: 0.310487778397, 2919: 0.39739606638, 2920: 0.15429236927, 2921: -2.77826414302, 2922: 0.33245313689, 2923: 2.64370651248, 2924: 0.66803838354, 2925: 20.2149990762, 2926: -1.40705183903, 2927: 0.314163507933, 2928: -2.73309427585, 2929: 1.42357030797, 2930: -3.1286946396, 2931: 1.88217497033, 2932: -2.15959594247, 2933: 11.4753965469, 2934: -8.36007988833, 2935: -6.02806262895, 2936: 0.562931495821, 2937: -0.46609119673, 2938: 0.266239595492, 2939: 0.269234067726, 2940: -7.92082362876, 2941: 0.240483443892, 2942: 2.14687516838, 2943: -2.86251798381, 2944: 0.29346890994, 2945: 0.240950647039, 2946: 0.362793510404, 2947: 0.344850314996, 2948: 1.00031337793, 2949: -2.59988680957, 2950: 0.442697325443, 2951: 0.0252323820304, 2952: 0.405698729544, 2953: 1.06113198155, 2954: 6.5495525829, 2955: 2.89085460526, 2956: -2.33011643771, 2957: -10.4426961201, 2958: -6.95024232621, 2959: 3.70847438342, 2960: 1.64562041703, 2961: 4.66675203072, 2962: -2.99293475295, 2963: -13.9023441018, 2964: 0.195989786561, 2965: -8.52768539402, 2966: 0.336231256742, 2967: -4.32725675772, 2968: 3.7631789945, 2969: -13.1485320991, 2970: 2.23918981266, 2971: 0.0909692364659, 2972: 0.316020654314, 2973: 0.19705418178, 2974: 0.410942187657, 2975: 1.85311197893, 2976: -5.57181237187, 2977: 10.8482663421, 2978: -4.54060709543, 2979: -7.98146176841, 2980: -6.63375771856, 2981: -7.54256391462, 2982: 6.31279984179, 2983: -2.1545879581, 2984: -0.0598940111592, 2985: 0.188389423908, 2986: 0.0217781268451, 2987: 0.668040079211, 2988: -1.66814288576, 2989: -2.29842422283, 2990: 1.05330825638, 2991: 0.319343168672, 2992: 2.28342702694, 2993: 0.343081815557, 2994: 0.444502307364, 2995: 0.656931160469, 2996: 0.317178686577, 2997: -11.1020031438, 2998: 2.34728209262, 2999: 6.8097810061, 3000: 0.205222675075, 3001: 0.271405064912, 3002: -1.07560321289, 3003: 0.171793544856, 3004: 2.29260478461, 3005: 2.79043128486, 3006: 1.2137607782, 3007: 0.375824272475, 3008: 0.614188421173, 3009: 0.509404577682, 3010: 4.2342891466, 3011: -18.3054877013, 3012: -15.0047241807, 3013: -0.456903413604, 3014: -0.100528787816, 3015: -0.258433840632, 3016: 0.816791306501, 3017: 10.2428528765, 3018: 0.976820664967, 3019: 14.2770377507, 3020: 0.580473422004, 3021: -0.935633647975, 3022: 0.82060361623, 3023: -1.37494149296, 3024: 0.505269978479, 3025: -12.7378052944, 3026: 8.46971264421, 3027: 2.32003126389, 3028: -1.09285415487, 3029: -20.0726808747, 3030: 4.64871891136, 3031: 2.43277192313, 3032: 2.44427219871, 3033: 0.509443790793, 3034: -18.636851722, 3035: -0.438393639251, 3036: 3.63415654876, 3037: 0.658354587565, 3038: -13.5519459198, 3039: 0.700133365909, 3040: 5.4553092806, 3041: 8.40206854849, 3042: -19.6770253187, 3043: -1.6281337813, 3044: 2.38075313429, 3045: -0.178153829707, 3046: 11.3793054418, 3047: 2.11997885403, 3048: 2.41109199844, 3049: 0.447860011357, 3050: 3.35121661137, 3051: 2.13595774521, 3052: 1.94116870398, 3053: -0.723955857237, 3054: 2.45933209868, 3055: -1.28749040069, 3056: 2.20044708095, 3057: 0.914272241567, 3058: -9.71201977243, 3059: 6.93603184815, 3060: -1.68017141148, 3061: 2.18564627376, 3062: 1.01332027557, 3063: 1.42677501414, 3064: 7.47959629558, 3065: -2.05476703519, 3066: 4.3441891898, 3067: -7.68622889357, 3068: -3.01854365507, 3069: 10.7741374514, 3070: -1.21971723302, 3071: 3.55555300749, 3072: 2.05439907555, 3073: 6.36250295358, 3074: 13.3244714315, 3075: 9.83527786721, 3076: -0.986489255784, 3077: 14.9615949864, 3078: 1.1151557637, 3079: 2.33872472183, 3080: 0.374123105997, 3081: 2.46884743263, 3082: -11.125042704, 3083: 10.9980093253, 3084: -15.1673497717, 3085: -0.533775912017, 3086: 3.59549174929, 3087: 7.76168032332, 3088: 5.12052217756, 3089: 3.05680361119, 3090: -4.19010146224, 3091: -1.40620654095, 3092: -2.51830049975, 3093: -2.82934427801, 3094: 2.2039513803, 3095: 0.378796011372, 3096: 2.79838539234, 3097: -3.3075822825, 3098: 1.6382061541, 3099: -6.85419377221, 3100: 2.3357237578, 3101: 2.50157141577, 3102: 2.79637231504, 3103: -1.01057105374, 3104: -9.2251856096, 3105: 2.40405582022, 3106: 7.04476631936, 3107: 2.31996190371, 3108: -6.21659380761, 3109: -0.265688781085, 3110: 2.50546772081, 3111: -7.40591378488, 3112: 2.09973292959, 3113: 2.32004616844, 3114: 6.16460764413, 3115: -0.142867600305, 3116: -2.42648999962, 3117: -2.91791064096, 3118: 11.3404533052, 3119: 5.40463610775, 3120: 1.59751947765, 3121: 7.67585623348, 3122: -15.2914449874, 3123: 9.8497806917, 3124: -10.8074195176, 3125: -1.77741546127, 3126: -7.54538172068, 3127: -8.66629198463, 3128: 22.4011020481, 3129: -9.92426599559, 3130: -6.78306830905, 3131: 2.40210750152, 3132: -7.53359925338, 3133: -5.94607807946, 3134: -5.65349366119, 3135: 3.1320706556, 3136: -6.41165261665, 3137: -1.57787125643, 3138: 3.97983129818, 3139: -13.6007413446, 3140: 13.4098152008, 3141: -8.21684494267, 3142: 0.907062561155, 3143: 10.2751767831, 3144: 35.0241516164, 3145: 6.61991591851, 3146: -2.15513275369, 3147: -25.8936878779, 3148: -27.0724513461, 3149: -0.429392638639, 3150: 1.28751226126, 3151: 61.4596056402, 3152: 10.7076334038, 3153: 1.69518615143, 3154: -34.4081899596, 3155: 17.4175936392, 3156: -23.9329272312, 3157: -25.5531251863, 3158: 38.6327769961, 3159: 28.3864759876, 3160: 40.0774772765, 3161: -9.52865750481, 3162: 12.6450926056, 3163: -2.9622041881, 3164: 19.8690858984, 3165: 9.25935389033, 3166: 14.8952687583, 3167: -30.9310530245, 3168: -48.9508499783, 3169: -19.7440554335, 3170: -45.9786486087, 3171: 32.48567101, 3172: 15.6885156329, 3173: -49.8484384849, 3174: 6.09356502295, 3175: 25.6484379082, 3176: 24.4742661169, 3177: -41.0514698944, 3178: -10.3455954986, 3179: 5.37792112982, 3180: -54.2286670485, 3181: -29.1103957373, 3182: 15.0556199828, 3183: 5.99892909256, 3184: -1.15979409329, 3185: -64.8503008876, 3186: -2.98518287908, 3187: -16.5771939077, 3188: -1.7618560607, 3189: -11.629512881, 3190: -19.4152369798, 3191: 5.18279135287, 3192: 14.4647940422, 3193: 0.81208511427, 3194: 3.69451225739, 3195: 28.784241409, 3196: -16.2366677883, 3197: 5.28688641153, 3198: -22.9620587243, 3199: 0.540721604665, 3200: 11.3634759393, 3201: 25.2843216478, 3202: -2.87247617141, 3203: 0.957472771596, 3204: -5.27504472218, 3205: -13.6342753006, 3206: -38.0982183662, 3207: -3.81840563052, 3208: -12.4532977788, 3209: 5.02645024009, 3210: 39.3549421063, 3211: 5.15741763472, 3212: 5.22175347428, 3213: 1.99520773132, 3214: 5.31205918986, 3215: 7.689222236, 3216: 5.68526224926, 3217: 3.63837482049, 3218: 4.79003900867, 3219: -2.24701447807, 3220: -9.41045288829, 3221: 54.394353795, 3222: 3.25091635216, 3223: -7.57867914577, 3224: 5.26833542545, 3225: 4.28866831799, 3226: 25.1029651266, 3227: 4.4874674564, 3228: 5.10266272931, 3229: -82.2834378626, 3230: 4.57133666911, 3231: -22.5081993894, 3232: 5.45667607664, 3233: 5.28187603227, 3234: -60.455863152, 3235: 4.05078411554, 3236: 6.59939929316, 3237: 5.47086815263, 3238: 0.759411970123, 3239: 5.30112527101, 3240: 5.33498718874, 3241: 5.29448561636, 3242: -56.6167402769, 3243: 5.28960740495, 3244: 4.98171589539, 3245: -3.82671899648, 3246: 7.80422704059, 3247: -49.9713207874, 3248: 5.11320388353, 3249: 5.19472459771, 3250: 4.70377522387, 3251: -6.91508045786, 3252: 3.84869809895, 3253: 13.8589146655, 3254: -0.20383953426, 3255: -10.8968047504, 3256: 6.56742758113, 3257: 13.0109768422, 3258: -11.9731157823, 3259: -5.31902881924, 3260: -22.6880617162, 3261: 19.1578534074, 3262: -16.0065284034, 3263: 50.8869428492, 3264: 16.5221150811, 3265: 0.951560847551, 3266: -5.37672926709, 3267: 1.30363964718, 3268: -0.0946762853699, 3269: -6.01169315932, 3270: 14.6328029609, 3271: 1.35215298089, 3272: -1.03828642693, 3273: 2.6106642212, 3274: -6.87751203168, 3275: 11.339134838, 3276: -2.85004029458, 3277: -13.4795871231, 3278: 2.70838122043, 3279: 26.5177252188, 3280: 8.55309126615, 3281: 3.83145177815, 3282: 5.41783828716, 3283: -11.1635118337, 3284: -67.1068622646, 3285: 0.974879206412, 3286: 13.8061619882, 3287: 0.670261844988, 3288: 1.00787679976, 3289: 34.1676398847, 3290: 1.16269042027, 3291: 5.38174182461, 3292: -11.3334754315, 3293: 0.922743743531, 3294: 0.796687421288, 3295: 0.957161644485, 3296: 0.912257320202, 3297: 3.04610851349, 3298: 35.2508711234, 3299: 2.60994627462, 3300: 2.09073176061, 3301: 1.24166953298, 3302: -0.732978455946, 3303: 2.08329312008, 3304: 44.6858221057, 3305: -28.8189135612, 3306: -42.9980160159, 3307: 18.9039979727, 3308: 24.0776734212, 3309: 4.24365789779, 3310: -5.27871939343, 3311: 18.0394381427, 3312: 0.873592752993, 3313: 0.874723035421, 3314: -6.80632534464, 3315: 0.944344953599, 3316: 23.7580726937, 3317: -0.404081550472, 3318: -28.3647659781, 3319: -3.40029841409, 3320: 1.0311109191, 3321: 0.899911409103, 3322: -0.000398886728605, 3323: 1.00315412558, 3324: 6.59574382175, 3325: 29.004476392, 3326: -0.341682325305, 3327: -71.8554592708, 3328: 28.3357497211, 3329: -11.6548815496, 3330: -2.28881994216, 3331: -0.0934672926652, 3332: 6.90549915493, 3333: 16.8383213212, 3334: 1.01727432277, 3335: -11.9798454036, 3336: 8.31489977188, 3337: 2.49536326458, 3338: -3.83621882274, 3339: -13.5659841121, 3340: 1.11401280885, 3341: 5.13861254451, 3342: 1.55052708353, 3343: 6.18288809703, 3344: 1.0211696419, 3345: 0.932444838807, 3346: -35.6873957827, 3347: 5.37201647697, 3348: -16.8378162648, 3349: 0.550538849606, 3350: 1.0054007358, 3351: 1.83698119522, 3352: 1.40014691908, 3353: 5.39774078089, 3354: -21.6870352004, 3355: 31.9255774682, 3356: 1.79733790445, 3357: 0.317710912096, 3358: -3.55977317618, 3359: 11.7222386777, 3360: 81.2341271366, 3361: -1.5948127662, 3362: 1.72210694759, 3363: -6.00342266479, 3364: 8.39926200188, 3365: 2.14925886982, 3366: 20.4747312492, 3367: 0.853046566843, 3368: 1.18750133601, 3369: 1.12004200325, 3370: 5.85457284641, 3371: -1.2787765076, 3372: -78.2730707932, 3373: 1.19570910037, 3374: -52.5963286473, 3375: 10.0904596385, 3376: -10.4634501387, 3377: -20.9287009567, 3378: 18.7069318135, 3379: 36.0622513183, 3380: 1.74062231917, 3381: -1.07338222547, 3382: -9.23705598487, 3383: 29.055079343, 3384: 13.2278366303, 3385: -1.19638196499, 3386: 40.2511331837, 3387: -42.5563648665, 3388: -100.0, 3389: -10.3118338642, 3390: -14.2785468144, 3391: 4.67790236197, 3392: -88.820846743, 3393: -3.41287384946, 3394: -11.0239646955, 3395: 36.9266814357, 3396: -2.06970034659, 3397: 6.86388696642, 3398: -16.2332104189, 3399: -4.01913965592, 3400: 5.29060494029, 3401: -8.99733071785, 3402: 0.694092722287, 3403: 5.28926953788, 3404: -3.35533443745, 3405: 6.64585592472, 3406: 3.13516637177, 3407: -25.6695670929, 3408: -1.68206016748, 3409: -12.9008545804, 3410: -1.38688126851, 3411: -1.79062353696, 3412: 3.31480667554, 3413: -7.21830127936, 3414: -24.7617423646, 3415: 56.8677862447, 3416: 16.8286720672, 3417: -8.06170395746, 3418: -15.8918809878, 3419: -13.8015986573, 3420: 26.6398067745, 3421: 37.8685975786, 3422: 11.6486196125, 3423: 43.4205596333, 3424: -27.2415951267, 3425: 24.8501858927, 3426: 4.5220615077, 3427: 12.0045937796, 3428: 5.16907528006, 3429: 11.4408965301, 3430: 6.15079622854, 3431: 15.2175866062, 3432: 16.3370631636, 3433: -5.92466519821, 3434: 11.9219721536, 3435: -8.09419901902, 3436: -10.0486561935, 3437: -55.6864479596, 3438: 17.5915492249, 3439: 37.4397295165, 3440: 6.87248994382, 3441: -53.5855354568, 3442: 9.11738433766, 3443: -30.6328683855, 3444: 10.558537275, 3445: -94.879226441, 3446: 11.6247634103, 3447: 6.71151075186, 3448: -52.9057006221, 3449: 5.11796486097, 3450: 5.23227154035, 3451: 2.25304248846, 3452: -18.9861584481, 3453: 5.38204359405, 3454: 5.39106005186, 3455: 40.3782159728, 3456: 5.39662867671, 3457: 7.5883381955, 3458: 6.11548772246, 3459: 5.09029813817, 3460: 26.4316079973, 3461: 5.29464178831, 3462: 5.53737111265, 3463: -8.35606328099, 3464: 33.670984438, 3465: -25.4440655665, 3466: 6.40479808037, 3467: -19.4819530092, 3468: 0.887523340311, 3469: 13.244854827, 3470: -4.81982085352, 3471: -35.840837835, 3472: -100.0, 3473: 28.2352797684, 3474: 3.28741520737, 3475: 4.09871080383, 3476: -30.6408102256, 3477: 49.3591221357, 3478: -13.3462510909, 3479: 71.9209979589, 3480: 5.33880249242, 3481: 13.0613872108, 3482: -7.6570382934, 3483: -11.2167980783, 3484: 26.5577537793, 3485: -8.42126853207, 3486: 5.79850287358, 3487: -21.3980476181, 3488: 2.91148503083, 3489: -10.0185934918, 3490: -29.0004529239, 3491: -0.477497287142, 3492: -1.80613803871, 3493: -0.747040310563, 3494: 0.728112441235, 3495: 0.886657526084, 3496: -0.653518362043, 3497: -1.86762901456, 3498: 1.0537051768, 3499: -1.03011830253, 3500: 1.99336909354, 3501: -0.686241480058}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.285346866637, 2: 3.38740893042, 3: -13.6938036143, 4: 16.6515641483, 5: -14.3088786027, 6: -7.07898399066, 7: 3.79228649329, 8: -4.09941934268, 9: 9.35112194757, 10: 5.00497562179, 11: -6.72671311005, 12: -10.3002589451, 13: 0.0232175434721, 14: -1.55139366366, 15: -2.15987022672, 16: 0.624594870132, 17: -7.55837250893, 18: 7.41134037492, 19: 4.9975099924, 20: -2.26006152213, 21: 7.52469260932, 22: -21.784552939, 23: 14.1302423572, 24: 3.58073875879, 25: -3.55842587412, 26: 9.84555647751, 27: -11.8008975877, 28: -3.61436037441, 29: -4.30770743714, 30: 3.14498556987, 31: -10.4909132271, 32: 6.90415207745, 33: 4.52517458436, 34: -0.327784168182, 35: 7.31678916855, 36: -3.94196734856, 37: -2.32299232547, 38: 3.64938835413, 39: 4.86973650238, 40: 3.20469259479, 41: 2.0240662288, 42: 0.810118829296, 43: 7.66529341064, 44: 3.44697213299, 45: -1.93627460614, 46: -5.75652580277, 47: 13.8763317143, 48: 0.81633355252, 49: 1.77707043058, 50: -3.7158567949, 51: -9.48226214248, 52: 9.74384437246, 53: 1.78082243777, 54: -1.11454338129, 55: -10.1288473277, 56: 0.821198180778, 57: -7.02724547417, 58: 0.869679852771, 59: 15.6033285022, 60: 1.10945370331, 61: -20.5804084645, 62: -3.30042614061, 63: 9.51764327137, 64: 7.84296329779, 65: -7.78723230843, 66: -3.99537137531, 67: -0.307708787118, 68: 2.48205633269, 69: 2.20874431038, 70: 4.68699145724, 71: 1.57958379616, 72: -1.00766765815, 73: 1.83205743717, 74: 1.11264683089, 75: 5.63593025089, 76: -1.67541236592, 77: 2.52111728043, 78: 1.87450661793, 79: 2.94581581435, 80: 12.098322294, 81: -1.07913956598, 82: -9.24577814111, 83: 2.3473185961, 84: 12.7738756433, 85: -3.64354495187, 86: -1.1000739358, 87: -12.2940384645, 88: 3.42052582187, 89: 1.90642246615, 90: -1.45243143842, 91: 5.2994886765, 92: 2.4281875322, 93: -12.5260014123, 94: 2.92483682051, 95: 3.09011127717, 96: 0.920051281489, 97: 14.7115344399, 98: 2.60516526373, 99: 2.50992464725, 100: 5.26280506146, 101: -11.4505680375, 102: 2.14919868182, 103: 1.86742521878, 104: -0.0775803491506, 105: -0.00299377001943, 106: -13.9021606742, 107: 2.22579117168, 108: 1.06432839578, 109: 2.66764900752, 110: -5.49440285963, 111: -1.25222615479, 112: -5.7990831518, 113: -2.08280277748, 114: 5.7602169415, 115: 20.7134932159, 116: 0.7127628515, 117: -0.46613635606, 118: -9.17951276284, 119: 26.830572629, 120: -3.12384069889, 121: 10.5051918931, 122: -9.18268926696, 123: 0.678592775681, 124: -4.94492423469, 125: 3.79131270883, 126: 0.475315438582, 127: 0.330923932183, 128: -3.77855126172, 129: -9.63517476481, 130: 0.445796921008, 131: 2.55106277824, 132: -1.34616081625, 133: -0.565841307327, 134: -2.42294320526, 135: -2.29465494744, 136: 0.983602917195, 137: 1.24029251717, 138: 7.098507307, 139: 1.49861471534, 140: -0.571314931458, 141: 10.1573829239, 142: -5.44668979352, 143: -7.6919143953, 144: -0.5075936414, 145: -8.85757031166, 146: 0.349208955932, 147: 0.370857245722, 148: -15.4458723032, 149: 0.509170098723, 150: 2.28319145739, 151: 11.1008076034, 152: 0.806682560019, 153: 0.373806517796, 154: 0.475559511385, 155: 0.405754473557, 156: 2.60277570534, 157: 27.9334571249, 158: 0.685972496857, 159: 0.428811153419, 160: 1.02521710114, 161: -1.11021298721, 162: 7.64695523926, 163: -13.0195949701, 164: 7.95415385797, 165: -3.13161526967, 166: -3.81917322578, 167: -2.56430844713, 168: -4.23991937211, 169: -2.64531326575, 170: 10.1365546137, 171: 4.27818246875, 172: -0.00758813705337, 173: 10.1948746074, 174: 0.243856571875, 175: 9.87380027622, 176: -10.9747751209, 177: -4.08845128329, 178: -1.0090935441, 179: 0.5127795009, 180: 0.360631789684, 181: -0.257117700747, 182: 0.9740307923, 183: 1.21375416388, 184: 0.309862613798, 185: -4.80814446093, 186: -16.4207657209, 187: -15.6832184857, 188: -4.89371566765, 189: 1.2257091585, 190: -6.31765444462, 191: -3.69441689297, 192: -10.8192733983, 193: 0.30453864558, 194: -7.95410129169, 195: 6.80219383865, 196: 6.43305452071, 197: 1.02138055088, 198: -1.60548208248, 199: 0.25759751228, 200: 2.64987834074, 201: 0.4965287751, 202: 2.27715335751, 203: 0.404981012316, 204: 0.49114326609, 205: 1.85375210486, 206: 2.20268539563, 207: 23.1727116259, 208: 0.321684000281, 209: 0.475145721632, 210: 0.472901729836, 211: 0.440543727408, 212: 2.10233214597, 213: -6.4676631661, 214: 2.42401272017, 215: 0.504789183821, 216: -0.986898635824, 217: -14.1329869232, 218: 6.3987470299, 219: 2.93592090252, 220: 1.39394385065, 221: -3.7774697704, 222: -5.14602347923, 223: -0.781921643679, 224: 1.85776854853, 225: -8.07702850696, 226: -0.010014642568, 227: -8.05014268761, 228: -3.09946767361, 229: -1.76919622356, 230: -2.76618873902, 231: 1.64437918966, 232: 0.634800983916, 233: -1.03118854661, 234: -11.2143466297, 235: -3.83747598254, 236: 16.9343532446, 237: -12.7350744155, 238: 14.8028160122, 239: 1.02003941651, 240: 2.34983116726, 241: -1.54567901282, 242: -1.0112252437, 243: 6.80477074267, 244: -1.57276464071, 245: -2.95904200756, 246: -2.4557011509, 247: -10.6640140911, 248: -1.3360951609, 249: 14.587039779, 250: 14.5638879142, 251: 9.0028733484, 252: 0.392021710498, 253: -4.80096256744, 254: 2.03033130039, 255: -0.922815813798, 256: 1.79375804086, 257: 11.221488767, 258: 1.15655341148, 259: 2.41490365659, 260: 8.97763831945, 261: -10.9159934333, 262: 2.29611624852, 263: 0.806106063959, 264: -2.26269204023, 265: 2.6723162457, 266: -6.83125370176, 267: 10.2467219075, 268: -7.97234910697, 269: -16.12379906, 270: -6.16685771414, 271: 8.66629957344, 272: 13.3141831587, 273: 22.79099864, 274: -17.8308882182, 275: -21.0890709495, 276: 10.1584288373, 277: 5.77885408218, 278: -12.675878123, 279: 18.3387582929, 280: 9.36252154965, 281: 17.3561295717, 282: -4.61427128901, 283: 2.04955462394, 284: 17.9695905209, 285: 15.6393047492, 286: 5.31993357252, 287: 1.69853131937, 288: -5.14486416379, 289: 6.90853150726, 290: 5.48146636951, 291: 5.52436436054, 292: -0.243684094154, 293: 2.01809598703, 294: 4.01264671957, 295: 23.5006339571, 296: -8.55414675762, 297: -7.83616451685, 298: -4.33004538152, 299: -2.90199928853, 300: -2.87915809903, 301: -1.73491461492, 302: 2.53429100116, 303: 3.16966943355, 304: -2.32973821157, 305: 4.10774975294, 306: 3.07751448729, 307: -4.36041582025, 308: 2.65599047333, 309: 2.69840639089, 310: -0.463275431715, 311: -2.94890904102, 312: 1.73726139371, 313: 2.02212595554, 314: -8.64319092783, 315: 2.05555332381, 316: -7.4108002855, 317: 2.12793654312, 318: 2.30961462285, 319: -10.6799991706, 320: 3.23870763245, 321: 2.540853173, 322: -5.84919399747, 323: 11.2915021897, 324: 6.69904629576, 325: -9.40218922422, 326: 0.648891461241, 327: -4.03048686768, 328: -10.3867100762, 329: -10.7078151792, 330: -3.17613071205, 331: -0.908539083073, 332: -9.8354494748, 333: 0.0655971532343, 334: -8.33776191459, 335: -3.40547902031, 336: -5.09724723279, 337: 4.52893024931, 338: 16.2937338868, 339: 2.13647507842, 340: -2.99781238177, 341: 4.04440025006, 342: -2.5818004993, 343: -1.28160507448, 344: 2.93651732466, 345: 1.75033839797, 346: -15.7002031395, 347: -15.7389368846, 348: 0.065267657039, 349: 0.392955817055, 350: 0.851352723678, 351: 0.824481789428, 352: 13.028438426, 353: 6.24680760375, 354: 13.8221640198, 355: -10.3969092344, 356: -17.5551200556, 357: 5.85231459347, 358: -26.7271665587, 359: -7.01913702101, 360: 5.55056169296, 361: 13.8594077987, 362: 13.23608156, 363: -5.60130265839, 364: 7.28896743757, 365: 2.04000129524, 366: -13.9650455404, 367: -2.62989818717, 368: -4.60129444061, 369: 7.717209649, 370: 1.98937476852, 371: 17.3438347944, 372: -74.326945593, 373: 14.0192498701, 374: 9.37594108247, 375: -7.39869316188, 376: 12.8785886336, 377: 7.02002928358, 378: -1.24787086816, 379: -3.58661372672, 380: 22.4566625569, 381: -30.3366476537, 382: 6.50400513153, 383: -9.46275847264, 384: 7.0976541097, 385: -22.1123209091, 386: 1.40875374827, 387: 4.69884154851, 388: 10.0072945158, 389: 2.07082350549, 390: 15.3666390687, 391: 4.28135992783, 392: -40.3506567394, 393: 18.3513578067, 394: -24.7797569039, 395: -2.50506800916, 396: 1.64608136508, 397: 5.96466530649, 398: 2.96344850167, 399: 1.12957738539, 400: -3.73882150004, 401: -18.0754252826, 402: 5.6242548108, 403: 3.52006364231, 404: -2.64206707733, 405: 1.65791132808, 406: -8.99445250865, 407: 0.632069493361, 408: 5.30795591744, 409: -9.85986685241, 410: 16.4497669304, 411: -14.6929996715, 412: -5.71445021328, 413: 9.90869169535, 414: -8.8248638704, 415: 3.14194780059, 416: 11.3529870636, 417: 1.28753745588, 418: 7.6528155954, 419: 3.00683210478, 420: 3.03930604595, 421: -2.2229559315, 422: 3.93581408203, 423: -2.6678149704, 424: -1.00385377376, 425: 3.71445457933, 426: -4.54187906529, 427: 1.91347267662, 428: 13.4876290265, 429: 2.96937108826, 430: -7.51117704705, 431: -22.6090505829, 432: 2.8125512965, 433: 1.07936312663, 434: 13.6559225178, 435: -4.23233232534, 436: -3.62418290194, 437: -10.4342651345, 438: 3.78858037912, 439: -20.9941585517, 440: -1.5547603118, 441: 2.16488589723, 442: 10.4754159895, 443: 1.38496328282, 444: 2.93612174999, 445: 3.6800556419, 446: -5.69934104726, 447: 2.68771646881, 448: 3.69797268871, 449: 1.68435819524, 450: -22.2467382208, 451: 2.2367407978, 452: -9.67614038874, 453: -0.666046618759, 454: 5.96531790677, 455: 0.189985888113, 456: 3.7339813566, 457: 7.41688394532, 458: -2.72833186515, 459: -1.5468773672, 460: -1.95279954033, 461: -15.8888210524, 462: -21.4278561793, 463: 15.0255835832, 464: -13.4361180603, 465: -3.46759069109, 466: -13.1435687432, 467: 5.94768956738, 468: 10.926291258, 469: 0.705659948324, 470: 0.505438735571, 471: 2.66465350847, 472: -7.22482015839, 473: 7.31796212573, 474: 8.35151413967, 475: 0.720327481978, 476: 2.1137985772, 477: 0.0807625820723, 478: -16.4444817322, 479: 0.954291312593, 480: 3.35059773011, 481: 0.724408817164, 482: 0.317122355701, 483: 1.56639680745, 484: 5.60442713671, 485: 41.4442746827, 486: 1.64673001572, 487: 23.1933233389, 488: 0.984594414771, 489: 4.63824148661, 490: 1.56719572251, 491: 4.83139162284, 492: 17.1863636994, 493: 1.2778051701, 494: -1.96668416794, 495: 0.908781241148, 496: 0.626176315976, 497: 20.0811383167, 498: 0.672701484986, 499: 3.06575719245, 500: 22.2183486113, 501: 1.74188531857, 502: 0.832396683478, 503: 0.790860971629, 504: 0.969010495285, 505: 2.85757316682, 506: 9.93310957078, 507: 0.198136971009, 508: 1.23600222467, 509: 1.16856800883, 510: 4.37809246777, 511: 2.64243487695, 512: -9.75834250428, 513: -0.736831567179, 514: -20.0991845023, 515: 8.37001950504, 516: -11.0013673909, 517: -8.19412574094, 518: -10.7388949917, 519: 1.93250302285, 520: -1.78314276966, 521: 1.1758042014, 522: 16.8573179813, 523: 0.880519275217, 524: 9.63851645508, 525: 14.6295838898, 526: 11.764314454, 527: 0.760605417504, 528: 0.682837114323, 529: 0.904958229833, 530: 1.40185952457, 531: 1.16693725585, 532: -2.76729008962, 533: 6.57392898651, 534: 7.28995238668, 535: -21.5990895812, 536: -17.4910162824, 537: 2.37988029224, 538: 9.77870174468, 539: 12.3622189454, 540: 0.525347352738, 541: -12.5963460879, 542: 0.856279377973, 543: 15.5912806201, 544: -1.36477727034, 545: 2.30811710863, 546: -0.9088585848, 547: 2.48766733239, 548: 0.869540424857, 549: 3.38294075526, 550: 0.845140040548, 551: -3.25537929403, 552: 0.710606496123, 553: 0.810288335537, 554: 11.2580919847, 555: 3.76395045028, 556: 12.3164478713, 557: 0.933728175914, 558: 0.945351665916, 559: 0.753754063106, 560: 0.75354177625, 561: 1.42331258475, 562: -4.11968894606, 563: -10.0919542831, 564: 0.275876938636, 565: 1.68635061001, 566: 10.9233201251, 567: -1.09063600984, 568: 19.1994970356, 569: -14.0709466797, 570: -9.0839399732, 571: -2.36749231297, 572: 4.53368307937, 573: -2.24292250033, 574: 0.340279099491, 575: 1.51971069148, 576: 5.68087754254, 577: -1.11827150852, 578: 1.7802114299, 579: -5.34427840667, 580: -27.4426726481, 581: 0.755993103603, 582: -14.7158063295, 583: 9.20586326805, 584: 4.59271329047, 585: 8.18925215528, 586: -17.3904815831, 587: -12.6837516532, 588: -3.20054335366, 589: 3.33745850489, 590: -7.35379408257, 591: 8.24123841529, 592: -2.30095636136, 593: -6.50311860665, 594: -12.1803904921, 595: -26.8133181955, 596: 38.1242805276, 597: 8.21358918402, 598: 33.3322225917, 599: 2.03530592476, 600: 8.66461456396, 601: -6.62843129329, 602: 12.3608587356, 603: -12.3263746801, 604: 0.322086547368, 605: 2.83654907967, 606: -13.7396963833, 607: 3.26421751967, 608: 3.05330436891, 609: 6.27503528222, 610: -19.7030373025, 611: 0.847457051689, 612: 7.45789075704, 613: 0.784681493168, 614: 2.86483929704, 615: -2.16883145392, 616: 6.57863110709, 617: -9.27991048062, 618: 2.8578449356, 619: 5.45224424094, 620: 1.9677763377, 621: 15.8406364396, 622: -18.773782444, 623: -21.6099516284, 624: -42.6935018155, 625: -0.965628974811, 626: -8.55735734264, 627: -8.53419328435, 628: 0.466200488884, 629: 8.16419954288, 630: -3.98328132349, 631: -5.06323204517, 632: 8.05601980643, 633: 6.10763398946, 634: 19.8379751562, 635: -10.902537035, 636: 4.20227499377, 637: -0.0689205520926, 638: -0.230206049983, 639: -3.48568736958, 640: -0.828772271239, 641: -21.6353400176, 642: 2.61801420675, 643: -6.89827539095, 644: 12.9315714496, 645: -33.1357467889, 646: -43.0097134379, 647: -29.5561660678, 648: 1.55033813496, 649: -31.890509591, 650: -4.97650766257, 651: -2.28062629276, 652: 6.38073603334, 653: -31.2192333169, 654: 8.42797065553, 655: 2.97391612037, 656: 15.6863666536, 657: 3.36810475852, 658: 3.47530360918, 659: 6.96567359482, 660: -0.704994090675, 661: 3.80852371381, 662: 4.01906097133, 663: -17.4732084237, 664: 3.78114342185, 665: 3.44202153345, 666: 3.10024129999, 667: 2.83587765976, 668: -20.878071839, 669: 3.07732072516, 670: 1.62763252769, 671: -12.8464968153, 672: 9.90725172077, 673: 3.55337948141, 674: 17.9770757733, 675: -6.23838792043, 676: -28.1710681623, 677: -3.64441597442, 678: -23.6335873089, 679: 2.09376137463, 680: -1.62028128727, 681: 5.36239034514, 682: -1.29210646277, 683: -1.93573195739, 684: -11.3836792585, 685: -35.6686075072, 686: -0.87188657823, 687: 19.6540021621, 688: 4.76303168871, 689: 2.76430311577, 690: 10.9034014835, 691: 4.9144977913, 692: -4.16581256166, 693: -29.7506241385, 694: -33.9007449001, 695: 10.4292612606, 696: -3.69316379079, 697: -26.1237596708, 698: 15.9985687912, 699: 0.427105716598, 700: -2.0197507465, 701: 4.79306049296, 702: -1.35779036742, 703: 3.73152306972, 704: -3.09625057445, 705: 3.88399325956, 706: 3.36015114662, 707: -8.37219677298, 708: -0.268996359365, 709: -2.27938655012, 710: -1.99141712402, 711: -2.61270748848, 712: 3.30908897643, 713: -0.478795402409, 714: -1.36313123518, 715: -16.4185835364, 716: -24.0403259104, 717: -6.86528215852, 718: -4.77085689277, 719: 3.69872833244, 720: -12.8234686668, 721: 3.91530632772, 722: -0.826205969435, 723: -1.38450256933, 724: -0.0626294797238, 725: 11.3849168131, 726: 2.62529853086, 727: -15.3798348714, 728: 10.5123133104, 729: 5.3121552681, 730: 9.28418256891, 731: 3.35662343813, 732: -1.49909732971, 733: -7.58434384642, 734: 2.99404467452, 735: -0.303322283699, 736: -0.0530194595395, 737: 5.89464219498, 738: 1.46125806153, 739: -1.9584617833, 740: -0.0073831654171, 741: -8.81804728221, 742: -6.30850317114, 743: 1.25115229398, 744: 1.79681223843, 745: -7.3009219811, 746: -0.570291923626, 747: -15.0620853323, 748: 5.72872628023, 749: -5.38799514693, 750: -1.24313611102, 751: 1.54053566846, 752: -9.65916448291, 753: -0.421925032381, 754: 1.5726065009, 755: 2.47753982676, 756: 1.38017090772, 757: 2.64110661218, 758: 0.982147407003, 759: -7.25541337157, 760: 4.20144107312, 761: -6.15219970732, 762: 2.03201889314, 763: -6.39353923188, 764: 2.34263841719, 765: -0.893569160034, 766: -0.409693121482, 767: -1.18199225429, 768: 1.16293532904, 769: 1.31754600996, 770: -3.49004796041, 771: 1.48790158851, 772: -11.3030045914, 773: -1.68470851485, 774: 1.42621069108, 775: -2.75838828844, 776: -2.83652650462, 777: -3.93948198626, 778: 6.66016529317, 779: 1.24114948231, 780: -8.34473373203, 781: 1.36281663845, 782: -0.391390805424, 783: -1.80689292154, 784: 4.10768132355, 785: -2.80075554322, 786: -7.1235776159, 787: -3.09819913224, 788: 1.20848627532, 789: 1.53813419912, 790: 1.55865410255, 791: -8.65432005224, 792: 1.15120394935, 793: 1.32112434822, 794: 1.5740220712, 795: 0.476998040421, 796: 1.52672859129, 797: 0.91972414051, 798: 0.250667379007, 799: -5.69922718248, 800: -5.70003912546, 801: 0.66480012606, 802: 2.07768704977, 803: -4.91604215665, 804: -10.0781499505, 805: 1.49905873224, 806: 1.57977114395, 807: -3.10305515934, 808: -2.36883171298, 809: 1.19374140428, 810: 2.43762730277, 811: -10.4177166989, 812: -5.71658733825, 813: -12.6380323403, 814: -7.9003462354, 815: 3.79533357874, 816: 2.26889752548, 817: 0.981642975516, 818: 2.6017588162, 819: 2.57777186721, 820: 10.6691012904, 821: 3.42234897439, 822: 9.49781917834, 823: 1.1310355696, 824: 0.282463570808, 825: 1.1265982661, 826: 0.841292629318, 827: -2.54985808309, 828: 0.447085036514, 829: 2.10083737318, 830: 0.548501120709, 831: -2.34022728794, 832: 1.09825507388, 833: 1.42477078548, 834: -6.92790756224, 835: -0.22010545767, 836: 5.99641811566, 837: 0.586703269086, 838: -1.95869029306, 839: 3.88509881483, 840: 0.750426932324, 841: 4.66411548266, 842: 0.400774599782, 843: 0.0763854117002, 844: 0.377878703476, 845: 0.397798112058, 846: 4.53242529941, 847: 0.352354244155, 848: 1.49904988818, 849: -1.83031386014, 850: 0.669601232865, 851: 0.352488006079, 852: 0.336599294232, 853: 0.436764043524, 854: -0.528934078228, 855: -16.487924783, 856: 0.0450033362188, 857: 0.707079976265, 858: 0.542321151153, 859: -3.39764037363, 860: 1.73948441219, 861: 7.81299141012, 862: -5.89979292389, 863: -2.6810186678, 864: 17.0196037087, 865: 0.966199771501, 866: 8.72945251492, 867: 1.48522786357, 868: 2.62247644867, 869: 9.1942421982, 870: 0.353390899227, 871: 9.52900505283, 872: 0.428379902959, 873: 9.30989493792, 874: -7.8912316753, 875: -0.24241499971, 876: 0.931823540654, 877: 0.415291840701, 878: 0.464350886012, 879: 0.358172715835, 880: 0.319358932914, 881: -0.399454813078, 882: 11.7294222396, 883: 0.171540835399, 884: -6.73729563363, 885: 5.87726590962, 886: -7.81301644187, 887: 0.721035647048, 888: -3.39096056134, 889: -0.903124594526, 890: 0.167407142366, 891: 0.389262315712, 892: -0.287196933362, 893: -1.30106938099, 894: 0.10523563188, 895: -3.7765449783, 896: -6.77264196573, 897: 0.398801055594, 898: 1.36152343554, 899: 0.571796081215, 900: 6.11779865408, 901: 0.304512810068, 902: 0.441594126826, 903: -7.98158948336, 904: 1.40764758328, 905: 6.13926474289, 906: 0.275015208978, 907: 0.366194365605, 908: 0.822164218802, 909: 0.379783522656, 910: 1.56151496149, 911: 0.541131298131, 912: -2.17386882656, 913: 0.463905425304, 914: 0.656403105907, 915: -0.264744640212, 916: -1.05985215099, 917: 11.0543869117, 918: -18.1093152746, 919: -1.27928671932, 920: 4.49961555993, 921: -1.01605870838, 922: 1.80708162618, 923: -3.23969605687, 924: 0.633971235274, 925: -2.13484653775, 926: 2.02495928907, 927: -0.42561782489, 928: -0.691933509847, 929: 14.9541199323, 930: 0.254467550798, 931: 7.56295329111, 932: -4.59301881418, 933: 0.393105897345, 934: 1.85210271623, 935: -9.38818836365, 936: 9.65033343035, 937: 1.53124696835, 938: 2.25366400127, 939: 3.44268170845, 940: 0.894057229215, 941: -4.77014612082, 942: 9.70697771646, 943: 1.01795204693, 944: -14.1520984949, 945: -19.830739503, 946: 6.35093881206, 947: 7.30002904493, 948: -16.3561418847, 949: -4.40214024464, 950: -2.49530023325, 951: 8.75149980557, 952: 3.92086905071, 953: -0.142689716944, 954: 1.36481234549, 955: -4.65613123332, 956: 2.22449629006, 957: 1.49315152268, 958: -0.0720174250561, 959: -8.69866201904, 960: 1.35171642871, 961: 6.41110971167, 962: 1.73416974776, 963: -0.506114723959, 964: -3.64811107144, 965: 11.1401680122, 966: -5.40632147308, 967: 2.65870928503, 968: -2.28424405333, 969: 1.53202974038, 970: 2.83887697541, 971: 4.30145642676, 972: 1.18824706327, 973: -21.6201551352, 974: -3.74085746355, 975: 6.9647938353, 976: 6.28115583732, 977: 2.94375050946, 978: -11.7287420182, 979: 2.96159501922, 980: -1.11537864289, 981: -9.36543868034, 982: 17.2454119755, 983: -5.4040927111, 984: 3.11598633755, 985: 1.36243233783, 986: -1.06294755717, 987: 1.43740976259, 988: 3.47069270182, 989: 5.45875272011, 990: -5.03866521256, 991: 3.30996385955, 992: -1.70965645387, 993: 7.52850051449, 994: 6.36442348527, 995: 3.91898747047, 996: -3.88089821714, 997: 1.337049622, 998: -9.95974066443, 999: -2.2206406339, 1000: 0.786200999697, 1001: -2.60719495506, 1002: -8.4716457524, 1003: -5.25990534309, 1004: 1.39542615384, 1005: -3.52429730145, 1006: 1.52507225343, 1007: 1.39688194984, 1008: 1.4158183608, 1009: -3.20581521764, 1010: 1.49888256516, 1011: 1.5071559066, 1012: 18.4282140501, 1013: 1.41363391033, 1014: -6.74177990642, 1015: 1.43906907444, 1016: 1.50402636097, 1017: -2.8585005267, 1018: 1.39285777041, 1019: 1.41487753249, 1020: 2.55590218213, 1021: -4.33433028508, 1022: -2.97966531703, 1023: -12.0524624881, 1024: 1.4313306657, 1025: 3.98568901602, 1026: 5.33934635177, 1027: 3.22132884024, 1028: -6.47162590625, 1029: -4.68234480094, 1030: -22.9989010162, 1031: 1.7466198977, 1032: -2.64575947575, 1033: -5.19164386301, 1034: 21.7138591338, 1035: -6.06837263403, 1036: 11.2148633641, 1037: 1.36957616697, 1038: -8.71804372103, 1039: -4.1780225167, 1040: -2.5999945404, 1041: 8.78681757634, 1042: 10.1533124095, 1043: -4.2497593711, 1044: -17.4492577659, 1045: -9.53325308971, 1046: 9.86338656975, 1047: 4.92939070285, 1048: 0.451992966846, 1049: 22.7829680406, 1050: 22.666490375, 1051: 18.6620005492, 1052: 17.1241002028, 1053: -21.7034412568, 1054: -2.39847917387, 1055: 15.376226532, 1056: -17.1004645324, 1057: -28.641261375, 1058: -58.6969396875, 1059: -6.12305451691, 1060: 16.4765602452, 1061: -5.1452845892, 1062: -15.0333034728, 1063: -19.2977606934, 1064: -53.4223340255, 1065: -16.8252209626, 1066: -27.6925641411, 1067: -65.7016142536, 1068: 0.26743662905, 1069: 65.9372891202, 1070: -11.8500186611, 1071: -15.1331670896, 1072: -61.2068375014, 1073: 2.60663301826, 1074: 1.92772088527, 1075: 19.1600885149, 1076: -7.69004369254, 1077: 7.8344417304, 1078: 6.70958106305, 1079: -22.898019184, 1080: 5.1410707001, 1081: 21.0649721312, 1082: 41.9531346522, 1083: 8.52284106575, 1084: 20.8910612324, 1085: -10.3435922127, 1086: -6.324478135, 1087: 12.4055568062, 1088: 21.2803976871, 1089: 2.50293471301, 1090: 14.6272442056, 1091: 1.40798619796, 1092: -38.4918708688, 1093: -13.1844699425, 1094: -48.7217504716, 1095: -2.50206264425, 1096: -8.66146318259, 1097: 0.97387843514, 1098: -5.16251441304, 1099: -13.648159204, 1100: 3.4063971458, 1101: -6.2232880972, 1102: 7.79695643994, 1103: 3.22648474106, 1104: -1.24872692356, 1105: 6.79873861536, 1106: 56.8193918674, 1107: 11.0602574575, 1108: 30.1444151121, 1109: -3.03344052757, 1110: 20.0539409562, 1111: -12.0939288414, 1112: 13.849558252, 1113: 0.423984174982, 1114: 3.21652896089, 1115: -6.21859586125, 1116: 26.3778068126, 1117: 4.07436703355, 1118: 3.60176080401, 1119: 0.193315996675, 1120: 3.20766537819, 1121: 8.28233341508, 1122: -6.53299700427, 1123: 3.73855312656, 1124: 2.8330889783, 1125: 0.165342979609, 1126: 63.1498952065, 1127: -3.51722963558, 1128: 5.43832556112, 1129: -11.202428289, 1130: 3.35116857866, 1131: -0.710412717106, 1132: -13.2744384474, 1133: -5.70458925047, 1134: -11.8318859051, 1135: -13.741912814, 1136: -2.18441782096, 1137: 6.18961866704, 1138: 3.11243377103, 1139: 3.33750607356, 1140: -42.5728396067, 1141: 3.44778359176, 1142: 3.59462844439, 1143: 3.2497841637, 1144: -10.2464860427, 1145: 3.38226354664, 1146: -23.2581883421, 1147: 3.13090610745, 1148: 25.9218007154, 1149: 3.72992489602, 1150: -15.2932974266, 1151: 4.03742577308, 1152: 1.96765960934, 1153: 0.947890055922, 1154: 3.13959908136, 1155: 3.22108556688, 1156: -2.24156889817, 1157: -13.1764531922, 1158: 4.49506052596, 1159: 18.294924777, 1160: 55.5818322883, 1161: 84.9764969825, 1162: 29.2415939916, 1163: -30.0892211317, 1164: -17.9708975679, 1165: 3.94041613248, 1166: 34.4438516602, 1167: -21.3492636885, 1168: -28.8671797498, 1169: 33.1831195392, 1170: -30.0973379548, 1171: 9.85712914754, 1172: 23.1732300124, 1173: 0.609619943314, 1174: 0.368399154803, 1175: 16.4995468795, 1176: -0.411318313874, 1177: 0.743160122176, 1178: 3.096076332, 1179: 0.724334801732, 1180: -21.1909609285, 1181: 0.636078052028, 1182: 0.148136819638, 1183: -12.4084965846, 1184: 3.54813224836, 1185: 50.8390806488, 1186: -2.21244410617, 1187: 8.92004142376, 1188: 41.0698083005, 1189: 27.5556216609, 1190: -25.9123721528, 1191: 2.12474611265, 1192: 8.45438919831, 1193: 1.46036980826, 1194: 0.70489305597, 1195: 27.7883000461, 1196: 0.694685127022, 1197: 3.41358467581, 1198: 8.00407340515, 1199: 0.343883410395, 1200: 0.6455190686, 1201: 0.776258226511, 1202: 0.547305139471, 1203: 3.2209017592, 1204: -50.905571397, 1205: 1.24267431821, 1206: 1.37959282684, 1207: 0.26763841714, 1208: 11.712358559, 1209: -11.883374843, 1210: 24.0478853619, 1211: -5.59761684588, 1212: 31.3472290344, 1213: -24.0223984067, 1214: 13.4993565497, 1215: -16.7505704646, 1216: 9.45690688049, 1217: -7.1672563768, 1218: 4.33050112451, 1219: 0.581163354801, 1220: -67.5583686357, 1221: 0.500643596819, 1222: 2.28489494439, 1223: 54.5375508207, 1224: 65.2536827449, 1225: 0.721229997059, 1226: 0.580818317919, 1227: 0.363957023836, 1228: 0.656570339034, 1229: 0.754205083539, 1230: -5.01149710162, 1231: 7.7883100019, 1232: -3.28804454294, 1233: 10.0737023049, 1234: -8.56445589858, 1235: 8.42007880475, 1236: -11.3976285083, 1237: 6.4274805401, 1238: 1.15962232561, 1239: -27.9888057746, 1240: 0.769371666894, 1241: 60.013894505, 1242: 9.92055762041, 1243: -0.709852765118, 1244: -5.68031973061, 1245: -3.17586430417, 1246: 0.546856211979, 1247: 3.39275303018, 1248: 0.597348184843, 1249: 18.5310026464, 1250: 0.66209139995, 1251: 0.739221356337, 1252: 15.8989640466, 1253: 3.43606426865, 1254: 31.4531063374, 1255: 0.513159135838, 1256: 0.67163181802, 1257: -2.25730164733, 1258: 0.544958485411, 1259: 4.72665513072, 1260: 9.72726226213, 1261: 4.59113569809, 1262: 0.799271575949, 1263: 0.592694413172, 1264: 23.7678990178, 1265: -22.9419946823, 1266: -28.1798616935, 1267: 43.1965241388, 1268: 7.61629868555, 1269: -13.2186640166, 1270: -0.681838335765, 1271: 4.01730954558, 1272: -6.22434421284, 1273: -0.029557904005, 1274: -25.6256079695, 1275: 0.623047001401, 1276: 16.7485222112, 1277: -7.29801534784, 1278: 29.0652976049, 1279: -0.218168675925, 1280: 0.404827776317, 1281: 15.0039325073, 1282: 3.40515650628, 1283: 8.38117757334, 1284: -29.0247153696, 1285: -11.991429829, 1286: 3.39854766541, 1287: 3.0532881831, 1288: -9.82724754739, 1289: -18.1203138113, 1290: -35.5608309715, 1291: -15.8991255383, 1292: 11.8539806955, 1293: -12.7564071002, 1294: -43.2563729918, 1295: -27.9278415388, 1296: 6.90890781374, 1297: -6.630118834, 1298: -18.6912352701, 1299: -1.20086979786, 1300: 35.9246648108, 1301: 12.5337973912, 1302: -21.254155961, 1303: 4.46451106918, 1304: -29.7550391979, 1305: 3.57878488006, 1306: 3.34247754363, 1307: 8.68258902178, 1308: 4.33770170598, 1309: 3.19285516086, 1310: -9.00538669228, 1311: 2.60456991552, 1312: 3.1946681215, 1313: -4.85546775835, 1314: 31.3698216521, 1315: -41.2193932652, 1316: -5.39279505332, 1317: -10.5429269949, 1318: -9.19936538335, 1319: -19.8399898907, 1320: -40.6067613043, 1321: -21.8430288752, 1322: -65.0911121627, 1323: 10.8702913745, 1324: 28.4196655854, 1325: 31.5763159147, 1326: -9.51892096225, 1327: -5.53484864182, 1328: -2.75852629151, 1329: 21.3284516973, 1330: -67.5429152822, 1331: 76.3148522356, 1332: -51.3456628785, 1333: -54.9874001475, 1334: 3.37750339861, 1335: 18.551343694, 1336: 3.87899657992, 1337: -3.62308652101, 1338: -4.45592528119, 1339: -5.39867740595, 1340: -27.7246575443, 1341: -29.7008612517, 1342: 3.51802654695, 1343: -13.639609601, 1344: 9.50003809157, 1345: -20.5478211559, 1346: 1.10477113307, 1347: 3.6061783443, 1348: 3.24414323084, 1349: -8.83052954842, 1350: -6.17491454909, 1351: -17.1946700584, 1352: 12.0584729145, 1353: 3.56850389415, 1354: -18.9327977423, 1355: 3.39564231327, 1356: 3.35066383587, 1357: 13.1084758394, 1358: -0.00662722578712, 1359: 3.3122425074, 1360: -3.61944863908, 1361: 39.3489765755, 1362: 3.4720318336, 1363: -26.9425783545, 1364: 3.31532050812, 1365: 3.36093054993, 1366: -8.91305159624, 1367: 1.65545145813, 1368: 5.67491513061, 1369: -3.01290937422, 1370: 8.19188284934, 1371: 9.45048407262, 1372: -14.8532536926, 1373: 18.5260838647, 1374: -8.05607002618, 1375: 36.8339763573, 1376: -15.336199649, 1377: -12.3580460854, 1378: -20.8251259428, 1379: 17.2586734881, 1380: 3.46314410661, 1381: -12.4817505046, 1382: 20.4149799883, 1383: -54.9843935402, 1384: -23.6160921517, 1385: 24.8563990309, 1386: 3.30988520999, 1387: -3.99675449059, 1388: -11.1058410112, 1389: -6.4939389737, 1390: -36.041259491, 1391: -6.66088138879, 1392: -1.0285451259, 1393: -53.331661319, 1394: 24.0966386013, 1395: 8.1690991952, 1396: -12.7759799015, 1397: 0.506265566629, 1398: 21.6909910931, 1399: -0.122967529291, 1400: -2.34122148975, 1401: 12.4804789218, 1402: -8.88532734353, 1403: -11.8331739413, 1404: -5.40143556935, 1405: 2.21015745025, 1406: 9.59865125599, 1407: -1.64188824028, 1408: -4.06784167931, 1409: 1.72316357854, 1410: 6.16293250419, 1411: -10.6174556853, 1412: -1.70915365871, 1413: -12.9834252222, 1414: 7.73796504947, 1415: 10.230956554, 1416: -4.08974042441, 1417: 10.7702080996, 1418: 4.61169652634, 1419: 17.1754943092, 1420: -8.39648858009, 1421: -0.752511841705, 1422: -7.02811970019, 1423: -0.435879076062, 1424: -6.24318435215, 1425: -8.41224855868, 1426: 8.94799474013, 1427: 4.63105353587, 1428: 6.51103224565, 1429: 2.68217105916, 1430: 2.16572720033, 1431: 1.24080749099, 1432: 0.243029424237, 1433: 4.56173015877, 1434: 0.208188901795, 1435: 19.6814121964, 1436: -9.71091947291, 1437: -1.69917386515, 1438: 2.0178929551, 1439: 13.0100430607, 1440: 1.77603471307, 1441: -10.3740609553, 1442: -1.47469465529, 1443: -8.9048840579, 1444: 1.48615718407, 1445: 1.16292293653, 1446: -4.25058740596, 1447: -3.19605969079, 1448: 15.2578479591, 1449: 1.90987347963, 1450: -3.8138031991, 1451: 4.50807735655, 1452: 1.96212224037, 1453: 14.3698534896, 1454: 7.14024241546, 1455: 7.39604113421, 1456: -5.88404158176, 1457: 14.6889052934, 1458: 2.13529658516, 1459: 3.00353502805, 1460: -1.24573980586, 1461: -9.15938803462, 1462: 0.568541422245, 1463: -14.5245303901, 1464: 1.17544285735, 1465: -4.21594909878, 1466: -3.52127906951, 1467: 0.962526270664, 1468: -10.4606563874, 1469: 1.81225015494, 1470: 0.42496415875, 1471: 6.11557437549, 1472: 9.29170104807, 1473: 2.83488077242, 1474: 3.42627264665, 1475: 19.4858729151, 1476: 7.61561236358, 1477: -1.44004180607, 1478: -2.82682435527, 1479: 0.0158015699674, 1480: 3.8412651054, 1481: -4.93725773004, 1482: -0.762836190183, 1483: -1.35239087815, 1484: 11.9058349, 1485: 1.72886403091, 1486: -9.5883325872, 1487: -11.3105702133, 1488: 1.73995270783, 1489: -6.29986167075, 1490: 1.31349103373, 1491: 1.17685320341, 1492: 1.14423735459, 1493: -4.79078765795, 1494: 1.04497883081, 1495: -4.33831426927, 1496: -4.16729231102, 1497: 3.14303077781, 1498: 1.59895095364, 1499: -2.75443101832, 1500: 1.21737107286, 1501: 1.61277639913, 1502: 1.58974174788, 1503: 1.69283256895, 1504: 1.89226736653, 1505: 1.35957330449, 1506: -5.67487918748, 1507: 1.9635906596, 1508: -2.67655778242, 1509: 16.5459218299, 1510: 6.13552348845, 1511: -19.2749373938, 1512: -7.43835634303, 1513: 2.47065122879, 1514: 6.51433501455, 1515: 7.59199135234, 1516: 9.56465113406, 1517: -10.6204332199, 1518: 6.43554417552, 1519: 2.34195002682, 1520: -3.3761146699, 1521: 6.73054478617, 1522: 0.454078023574, 1523: -0.114077827654, 1524: -8.83297523461, 1525: 3.55316354552, 1526: 0.326682606615, 1527: 1.39967008274, 1528: 0.939471910846, 1529: 5.54498240459, 1530: 3.05685585065, 1531: 0.312336420845, 1532: -6.37037627933, 1533: -2.70298408056, 1534: 8.57265133032, 1535: -0.674431769681, 1536: -0.237372599981, 1537: 12.0090851139, 1538: 1.10803490146, 1539: -5.32132135464, 1540: -0.426610835006, 1541: -5.82015655611, 1542: 0.565090812204, 1543: 0.492075217831, 1544: -3.15898100879, 1545: 0.348061974475, 1546: 1.97327075505, 1547: 2.44427384496, 1548: 0.402935614706, 1549: 0.394060773283, 1550: 0.378004143372, 1551: 0.495183755187, 1552: 0.909690349697, 1553: 1.40700761702, 1554: 0.0728538719767, 1555: 1.52123638439, 1556: 1.35405716385, 1557: -2.92448298984, 1558: 2.03160300962, 1559: 8.29429418521, 1560: 4.90868916237, 1561: 2.85280461702, 1562: -4.9243595851, 1563: -3.32967963624, 1564: 1.3815650398, 1565: -8.36914703497, 1566: 1.54488497088, 1567: 5.6254485363, 1568: 0.399184590242, 1569: 18.6864682653, 1570: 0.518656050263, 1571: 0.788476784062, 1572: -8.39413152315, 1573: 6.22015611129, 1574: 0.35065248225, 1575: 0.421142223747, 1576: 0.688754594615, 1577: 1.51954660346, 1578: -0.263571697701, 1579: 13.4162703094, 1580: 6.74855817662, 1581: 1.08583758518, 1582: -31.302247131, 1583: -10.3606047691, 1584: -4.38689320066, 1585: -5.32139555246, 1586: -5.60416371102, 1587: 2.05136877239, 1588: -3.26952507846, 1589: 0.481371238938, 1590: 15.0933134669, 1591: -0.456140841716, 1592: 0.998318396198, 1593: 2.16751875535, 1594: 3.27246724209, 1595: 0.328817855946, 1596: 0.674438905782, 1597: 0.587363016255, 1598: 3.28163216684, 1599: 0.351229739834, 1600: 0.457323852327, 1601: 16.314873603, 1602: 1.26510071882, 1603: -3.14862488422, 1604: 0.333967650678, 1605: 0.303291574484, 1606: 2.10351568883, 1607: 0.503600265261, 1608: -1.44989293388, 1609: 3.20307895139, 1610: -10.3746694648, 1611: 0.481827761646, 1612: 1.1788521376, 1613: -12.8658015591, 1614: -2.80412973927, 1615: -12.3950236345, 1616: -1.03029502816, 1617: -3.52917209937, 1618: 5.57291950029, 1619: -6.26541406383, 1620: 1.66984507191, 1621: 0.140455383183, 1622: 0.716640367113, 1623: -14.0467660904, 1624: 14.845702726, 1625: -0.912452361002, 1626: 1.32870872086, 1627: 2.62479642405, 1628: 0.680137226446, 1629: 8.3091349044, 1630: 2.54182226079, 1631: 2.43885891372, 1632: 14.8617893526, 1633: -0.672735195654, 1634: 3.09366232789, 1635: 1.97758399419, 1636: 1.41184396907, 1637: -3.04911797215, 1638: -4.02188936653, 1639: -11.2636033475, 1640: -7.99500685154, 1641: -0.0864202897416, 1642: -6.23510771214, 1643: -7.88712075758, 1644: -0.00771656963798, 1645: 17.5910427875, 1646: -10.8834955049, 1647: 3.48160797543, 1648: -7.96482440659, 1649: 16.1966795736, 1650: 3.69197937777, 1651: 2.06129967936, 1652: 1.65601808629, 1653: -4.11038096929, 1654: 1.53564848927, 1655: 1.96841838633, 1656: 2.51571871241, 1657: 0.843647843376, 1658: 1.81637068068, 1659: 3.63714728906, 1660: 1.22672367035, 1661: 0.917907585093, 1662: 1.73489801336, 1663: 3.69288798542, 1664: -3.66278890507, 1665: -9.37764629985, 1666: 0.977801193434, 1667: 1.56958631074, 1668: -8.97195771552, 1669: 8.69213113643, 1670: 12.4890244342, 1671: -14.1453751376, 1672: 11.9522615165, 1673: 5.94666758178, 1674: -6.89600361098, 1675: 1.87461071669, 1676: -5.66298392617, 1677: -1.07907102502, 1678: -6.3787847927, 1679: 0.37511951707, 1680: 9.45846334832, 1681: 6.48549186097, 1682: 7.99713568308, 1683: 1.62643843494, 1684: 1.09430231624, 1685: -8.54020207821, 1686: -2.67678412165, 1687: -1.60053308254, 1688: -4.70471475939, 1689: 1.09577026435, 1690: 6.8835526692, 1691: 36.0060585851, 1692: -12.1976398575, 1693: 1.19514595383, 1694: 4.27156186245, 1695: 0.116602303908, 1696: -6.20439284861, 1697: 1.95831983378, 1698: -20.4773778111, 1699: -3.75554526465, 1700: -10.6412021555, 1701: -0.394192507329, 1702: 1.21847288765, 1703: -12.4361086647, 1704: 0.895555931717, 1705: 0.809581129591, 1706: -4.620991346, 1707: 4.96205578361, 1708: 1.93897044039, 1709: 1.76380265279, 1710: -4.78689258491, 1711: 1.25367402262, 1712: 13.3409868003, 1713: 1.75443623336, 1714: 1.82830265356, 1715: 8.00408237893, 1716: -1.14039174186, 1717: -3.62420725258, 1718: 1.81509303781, 1719: -1.17916966556, 1720: -4.24664343642, 1721: 2.2042049327, 1722: 14.8857880002, 1723: -4.33577644616, 1724: 2.79656312241, 1725: -3.16637511659, 1726: 1.29128653172, 1727: 16.2498946598, 1728: 5.16792077793, 1729: 1.91052158844, 1730: -6.36510176091, 1731: -0.909900496135, 1732: -22.8850999402, 1733: -5.86829099713, 1734: -14.3601292297, 1735: 1.71586666933, 1736: -11.3059269599, 1737: -1.58379941614, 1738: -8.74643553722, 1739: -8.80051393777, 1740: 18.3896789348, 1741: 3.41650754942, 1742: -26.6827818471, 1743: 22.972593332, 1744: 1.01002830903, 1745: -30.5946402133, 1746: 0.804404531866, 1747: 2.82936469189, 1748: 3.63844793801, 1749: -14.4350272979, 1750: 10.4891603795, 1751: 12.7192120762, 1752: 5.03782105402, 1753: -17.4958982868, 1754: -18.2640559996, 1755: 8.32956998911, 1756: 13.2254273751, 1757: -2.04403925164, 1758: -8.72752482534, 1759: 8.95776756806, 1760: -18.243041549, 1761: -4.02315556849, 1762: -12.7057575911, 1763: -3.02516944339, 1764: 14.0190258536, 1765: 9.21712728687, 1766: -13.1386058324, 1767: 18.4479474856, 1768: 12.2522132056, 1769: 13.482897736, 1770: 0.661256021963, 1771: 4.68252198049, 1772: 16.0543934535, 1773: 9.95096350354, 1774: 4.15036297217, 1775: 2.06499609729, 1776: 9.7714025161, 1777: -20.6593832609, 1778: 10.5670562157, 1779: 1.61843336904, 1780: 8.41481394397, 1781: -0.977038730487, 1782: 3.07515301626, 1783: 4.78337936568, 1784: 14.0328085994, 1785: -14.2474399027, 1786: 8.06490163783, 1787: 0.667131336359, 1788: 7.67177519554, 1789: -6.52444296101, 1790: 2.70602884643, 1791: -17.6001549676, 1792: 0.160051792257, 1793: 0.549039187791, 1794: -13.647550964, 1795: 0.665503114295, 1796: 9.85009662688, 1797: 0.00613668624334, 1798: 0.572615604641, 1799: -3.04860641829, 1800: -5.61011871462, 1801: 0.743533303196, 1802: 10.9955838097, 1803: -3.66400505188, 1804: -10.2779346351, 1805: -5.95279985879, 1806: -2.37756824298, 1807: -2.50101011432, 1808: 4.81355853644, 1809: 3.30644372282, 1810: -1.39287516673, 1811: 0.686991187102, 1812: -0.384584454874, 1813: 0.702735416811, 1814: -7.7533610542, 1815: 1.69511170311, 1816: 0.957668737041, 1817: 2.46785394614, 1818: 0.79706864772, 1819: 15.9835920438, 1820: -2.95360387953, 1821: 0.300831496843, 1822: -2.18160433314, 1823: 5.34491761489, 1824: 1.87738723173, 1825: 13.1761749346, 1826: -0.273482927124, 1827: -2.60513850926, 1828: 1.26291215942, 1829: 1.06515361211, 1830: -0.779566193132, 1831: 0.478177727223, 1832: 13.1805199617, 1833: 23.4064106005, 1834: 0.268790198387, 1835: 0.582769191638, 1836: -1.65391530394, 1837: 0.844420554314, 1838: -4.10720335829, 1839: -2.80415460894, 1840: -0.18909970291, 1841: 1.69954839984, 1842: -0.0794830463976, 1843: 1.00617117318, 1844: 1.19812216141, 1845: -1.20656547474, 1846: 28.7605127029, 1847: 0.0220048780902, 1848: 1.40305306976, 1849: 1.16845351391, 1850: -2.93928051829, 1851: -28.610558185, 1852: 0.912369612762, 1853: 0.589446471789, 1854: 0.705967577327, 1855: -5.08674019744, 1856: -3.89603653565, 1857: -7.22273272472, 1858: 17.6256215028, 1859: -13.1641989197, 1860: 16.3059186313, 1861: 5.27937834721, 1862: 0.246430938416, 1863: -2.18876848405, 1864: 0.502248883463, 1865: -18.0342923865, 1866: -5.47595618014, 1867: 2.55723738055, 1868: 7.09201220484, 1869: -2.72738991879, 1870: -2.40742398983, 1871: 0.502590266868, 1872: 0.800944886065, 1873: -0.69455090472, 1874: 3.0311389357, 1875: 1.09488888667, 1876: 0.718740476228, 1877: -0.684786026045, 1878: -4.06682293372, 1879: 0.108834806119, 1880: 0.838022804095, 1881: 27.9605488951, 1882: 1.25540275893, 1883: 2.95564032043, 1884: 0.485494333294, 1885: -2.344271506, 1886: -4.51214080545, 1887: -15.405902176, 1888: 10.427793735, 1889: -0.403163929847, 1890: -9.21224367217, 1891: 0.680200714014, 1892: 0.654764801146, 1893: -9.98076578837, 1894: 0.473700590615, 1895: 0.833137435834, 1896: -3.1164364044, 1897: 0.137365010619, 1898: 0.321913796475, 1899: 0.232467701471, 1900: 0.660208485344, 1901: 0.625268935832, 1902: 6.80291427892, 1903: -0.813182642806, 1904: 1.37553075682, 1905: 0.617477535205, 1906: 1.45789174326, 1907: -1.93969045543, 1908: 2.60015448376, 1909: 2.51397375654, 1910: 39.690600465, 1911: -10.4671489296, 1912: 1.1201582575, 1913: 4.19164483992, 1914: 5.30483535158, 1915: -0.236737336245, 1916: -19.2574060134, 1917: 0.565412015652, 1918: 9.76725521607, 1919: 0.254135854764, 1920: 1.01126033921, 1921: 19.2724473271, 1922: 0.0588527846382, 1923: -0.604657843115, 1924: 0.652054790836, 1925: 0.737221050498, 1926: -0.763469603839, 1927: 1.0693224749, 1928: -1.63028424101, 1929: -9.52260104392, 1930: 8.31423355571, 1931: 5.37441432842, 1932: 18.5731179805, 1933: -3.07980610616, 1934: 2.78985377337, 1935: 0.25016814429, 1936: -1.69539710049, 1937: -9.19545208447, 1938: 0.498198622566, 1939: -12.7320276888, 1940: 4.29189150455, 1941: 1.57017288582, 1942: -5.67201261119, 1943: -13.8696662585, 1944: 0.333730945874, 1945: 0.718561845173, 1946: 0.55137282452, 1947: -2.93494173559, 1948: 0.612243939064, 1949: 0.462610439633, 1950: 34.8577527265, 1951: 0.918504309746, 1952: 20.0270898277, 1953: 0.636281712448, 1954: 0.356013073852, 1955: -0.869579929557, 1956: 1.32127341407, 1957: -0.139229802563, 1958: 20.2331161991, 1959: -14.5982747385, 1960: 2.5402020021, 1961: 0.365205666232, 1962: -2.69187242002, 1963: -4.77104592769, 1964: 7.03914214561, 1965: -3.85373201077, 1966: 5.2432290445, 1967: 1.67791263144, 1968: 0.552308255309, 1969: 0.543361801557, 1970: 7.77795260027, 1971: 0.639022116813, 1972: -8.56153970792, 1973: -0.915062555552, 1974: -0.314237301776, 1975: 1.89914356217, 1976: -8.1210801323, 1977: 0.601657481989, 1978: -31.042882172, 1979: -4.56818448342, 1980: -2.59675485805, 1981: -1.93333729059, 1982: -13.0633046744, 1983: 0.79587929151, 1984: -0.464308945272, 1985: 0.638196329833, 1986: -3.32877915709, 1987: -0.949218559209, 1988: -3.750204536, 1989: 10.8037099903, 1990: -20.1026396892, 1991: -11.4716542438, 1992: 4.98882339276, 1993: -8.33892131796, 1994: 16.3394810637, 1995: 11.581028501, 1996: 9.6317148424, 1997: 4.67157800554, 1998: -3.79677013645, 1999: -5.42035433606, 2000: 0.14324778078, 2001: 0.701683154045, 2002: 18.2223458009, 2003: 0.803780412983, 2004: 0.840746514156, 2005: -2.41420225241, 2006: -1.67474062083, 2007: 0.934478960021, 2008: 0.499093938811, 2009: -0.594832129027, 2010: 0.735914407154, 2011: 11.445211116, 2012: -4.18058364037, 2013: -13.7170346042, 2014: -1.98282214157, 2015: 0.0486776415844, 2016: -1.04412714458, 2017: 10.9147143429, 2018: 11.2451243068, 2019: -11.6938891021, 2020: -25.514392567, 2021: -11.0342930955, 2022: -9.1477029103, 2023: 2.33878181604, 2024: -0.912956579936, 2025: -31.5710068221, 2026: -1.05247800858, 2027: 2.28244438929, 2028: 2.00981941397, 2029: 12.3922500238, 2030: -5.14406846262, 2031: -3.71798239017, 2032: 0.649259072918, 2033: -1.73477055724, 2034: 2.2347636243, 2035: -1.85708242552, 2036: -2.25691249907, 2037: -15.9620388217, 2038: 12.2990571731, 2039: -20.9577023656, 2040: -1.08943526738, 2041: 10.0296595733, 2042: -15.5881484038, 2043: 4.24372726044, 2044: 0.516400478675, 2045: -17.5795770518, 2046: -0.43155187319, 2047: -14.6351763221, 2048: -5.43769047422, 2049: -20.0391294347, 2050: -5.0141502173, 2051: -0.0146355007863, 2052: 5.43304425756, 2053: 0.711064020235, 2054: 0.795982091513, 2055: -0.899067618856, 2056: 10.1571438939, 2057: -0.797368442091, 2058: 0.9132635107, 2059: -30.6297283975, 2060: 0.875454161844, 2061: 2.15612233116, 2062: -1.55803787953, 2063: 0.76664550767, 2064: -1.0958034238, 2065: 0.0545783505854, 2066: -0.430950559582, 2067: -7.1190100685, 2068: -22.4156635061, 2069: -22.1471950203, 2070: 15.6390216477, 2071: -3.70793047371, 2072: 3.56915009689, 2073: 20.9126384707, 2074: -9.41588348892, 2075: -5.92526416002, 2076: 4.12670560886, 2077: 8.04548827208, 2078: 0.221347073192, 2079: -6.31844111274, 2080: -1.66020517617, 2081: 14.4627075547, 2082: -6.4094102139, 2083: 8.77777885094, 2084: 0.940560402476, 2085: 0.0149888833938, 2086: 15.5051496132, 2087: 0.996580091885, 2088: -0.932468687232, 2089: -12.3492631798, 2090: -5.84506219611, 2091: -38.3874376683, 2092: -1.77532179458, 2093: -49.4291545045, 2094: -14.3641542552, 2095: 0.12162459028, 2096: 9.69248349314, 2097: -1.01726334479, 2098: -0.574564530278, 2099: 13.9205109644, 2100: 4.58837033543, 2101: -4.2352480025, 2102: -2.61468318742, 2103: -6.8487535364, 2104: 1.99973504808, 2105: 17.4900393629, 2106: 11.4706720409, 2107: 6.27499083588, 2108: -1.32053646958, 2109: -6.88565754302, 2110: -2.52971907215, 2111: -18.0647558654, 2112: 1.93748706016, 2113: -9.14795949553, 2114: -3.17980565977, 2115: 2.97890613831, 2116: -6.24535726375, 2117: 8.35782955087, 2118: -4.68870281218, 2119: 16.261908355, 2120: -8.09914727642, 2121: 14.9852919505, 2122: -10.2365150108, 2123: -10.7803286595, 2124: 7.10988735804, 2125: 11.3170021755, 2126: -3.49000335364, 2127: 1.40489044142, 2128: -1.98368909652, 2129: -3.99319045118, 2130: -4.28206924087, 2131: -0.275202610013, 2132: 3.13690436199, 2133: 3.2138986892, 2134: -0.32037738154, 2135: -9.17219593681, 2136: 0.011917301915, 2137: 1.84010104205, 2138: 0.726399285475, 2139: -2.95339132651, 2140: 5.11311321348, 2141: -7.98578965341, 2142: -0.302605508623, 2143: 5.77594356407, 2144: 4.62358700615, 2145: 4.05429516447, 2146: 1.78787175546, 2147: 0.383967220553, 2148: -1.75539449639, 2149: 5.96839712072, 2150: -0.611423792436, 2151: 7.47135433855, 2152: 9.62524756619, 2153: -1.25386803698, 2154: -4.40443745578, 2155: -10.136683542, 2156: 5.46010426997, 2157: -11.1916178214, 2158: -11.0486802834, 2159: -10.5882517527, 2160: 0.039181609101, 2161: -0.288252408571, 2162: -0.730763665035, 2163: -0.214868436683, 2164: 0.625830369494, 2165: 0.723653939574, 2166: -1.54384951115, 2167: 0.309412289849, 2168: 9.84119491808, 2169: -0.176280713518, 2170: -0.919322524266, 2171: 0.807976264166, 2172: -4.63945539743, 2173: 9.79813738177, 2174: 11.7501144764, 2175: -1.16042786195, 2176: -4.98929150463, 2177: 0.539117120994, 2178: 8.41515783267, 2179: -2.82014404821, 2180: -11.3445995907, 2181: -1.3590378939, 2182: -20.4009309625, 2183: 2.36419512601, 2184: -3.23348871866, 2185: -1.7770051114, 2186: 0.0224991046968, 2187: 6.25146262794, 2188: 0.445352913651, 2189: 0.875446106717, 2190: 0.21517186832, 2191: 0.548979371619, 2192: 0.590963881861, 2193: 1.85195496703, 2194: 0.698407031225, 2195: -9.35347098715, 2196: 1.85403403774, 2197: -4.38383750171, 2198: -0.110189508295, 2199: -0.979989305981, 2200: 1.13220668837, 2201: 0.208038217564, 2202: 0.208336780442, 2203: 2.38438071742, 2204: -4.85000358852, 2205: 0.84493872225, 2206: 1.07940108617, 2207: -30.0601320945, 2208: 8.32593490643, 2209: -17.0463383556, 2210: -2.32086777928, 2211: 1.83818852937, 2212: -13.7353266898, 2213: 2.16023286432, 2214: 9.11028320587, 2215: -9.02277722585, 2216: 13.2127082491, 2217: 17.6985322306, 2218: 7.55001285991, 2219: 8.76967702864, 2220: 0.072741598998, 2221: -0.18858338357, 2222: -4.86912854183, 2223: 5.52680490629, 2224: 0.128049976491, 2225: -0.502279688815, 2226: 0.480058062098, 2227: 5.48436465012, 2228: -0.920139029928, 2229: -0.133262387952, 2230: 8.83084057774, 2231: 0.597798603046, 2232: -1.11788504145, 2233: -0.141032214341, 2234: 1.46518330763, 2235: -6.8336114501, 2236: -6.74729044067, 2237: 5.57755229302, 2238: -0.0396391592733, 2239: -2.49484472988, 2240: -0.485813963267, 2241: 0.0994463378068, 2242: -1.52321489274, 2243: -0.0556497176812, 2244: 0.295638454228, 2245: -14.5176793115, 2246: -1.0295789663, 2247: 0.0597570221386, 2248: 0.192298365923, 2249: 0.0758333488471, 2250: 0.909145213892, 2251: -16.6403731709, 2252: -0.277956370335, 2253: 0.129718973104, 2254: 0.0780968578758, 2255: 8.19590858314, 2256: -1.92737968311, 2257: 2.49240670868, 2258: -7.62328511523, 2259: 6.73946837831, 2260: -8.99373035394, 2261: 8.00186778992, 2262: -0.718519905533, 2263: 4.07843343583, 2264: 3.48760067785, 2265: -3.7256619482, 2266: 0.0858015063549, 2267: 16.1521674819, 2268: 0.0738057784378, 2269: -9.49109210067, 2270: -2.96964940324, 2271: -5.25480191321, 2272: -0.593071177198, 2273: 0.0294733304618, 2274: 0.0545235103966, 2275: -0.112465572443, 2276: 0.20552101007, 2277: 0.703859890328, 2278: -3.45327680738, 2279: 1.61156958748, 2280: -12.6085945709, 2281: 3.58513650706, 2282: 1.66349870986, 2283: -6.90822928039, 2284: 4.07867838804, 2285: 3.39610462323, 2286: -4.24325369987, 2287: 0.134449669999, 2288: 0.705778525545, 2289: 2.1963715223, 2290: -0.816055341564, 2291: -0.886827309341, 2292: 0.0167944487338, 2293: 0.132982861374, 2294: 0.398116480827, 2295: -0.14528186926, 2296: 2.38655735868, 2297: 0.148588158557, 2298: 0.0382825698684, 2299: -13.4579588992, 2300: 0.318178808876, 2301: 0.306271052709, 2302: 0.0899390184183, 2303: -0.0649249795197, 2304: 0.688180876729, 2305: 0.0289641921465, 2306: 1.49721877062, 2307: -7.06948134643, 2308: 3.11294875222, 2309: 0.443703271775, 2310: -0.359510942941, 2311: -0.953298822436, 2312: 0.0335959748315, 2313: -0.0259536933054, 2314: 8.12421442895, 2315: -0.0555471701676, 2316: 0.290913508921, 2317: 8.08417821493, 2318: -0.0293241075599, 2319: -5.97111766991, 2320: 0.0367663128664, 2321: 3.71703229721, 2322: -1.71554626459, 2323: 0.32918082854, 2324: -1.07781625965, 2325: 5.23328499588, 2326: 0.368402208547, 2327: 0.717462223479, 2328: -0.994886268675, 2329: 0.0603513425454, 2330: -1.74447038651, 2331: 9.73855515342, 2332: 1.68242310069, 2333: -1.71687164886, 2334: -0.308243200879, 2335: -5.25266879722, 2336: 2.23386110127, 2337: -1.54614109721, 2338: -0.836941056109, 2339: 3.50199643468, 2340: -4.11426534371, 2341: 1.48380762876, 2342: -4.42577196255, 2343: 11.1775048189, 2344: 10.9843725239, 2345: 16.0361450475, 2346: -0.943472200424, 2347: 4.28986750265, 2348: 5.98150890854, 2349: 10.3446352527, 2350: -0.171072327745, 2351: -4.80276963832, 2352: 0.609965085965, 2353: 0.114996832087, 2354: -0.843912894754, 2355: 5.03107810394, 2356: 0.33306188054, 2357: -7.86504070984, 2358: -4.14433810752, 2359: 0.946149702991, 2360: 1.51778162153, 2361: 3.89694590433, 2362: -10.7371584336, 2363: -4.85042528805, 2364: 1.75269053, 2365: -0.6529168602, 2366: 1.30449722063, 2367: -1.51758159482, 2368: 8.39780563665, 2369: -1.33654090234, 2370: -10.5739794141, 2371: 0.614480082707, 2372: -3.15583165087, 2373: 4.92990540982, 2374: -0.536248098965, 2375: 0.831516789692, 2376: -5.26931687174, 2377: -2.32391959819, 2378: -4.24443638125, 2379: -2.55219273612, 2380: 7.01028584616, 2381: 0.423570447224, 2382: -1.82882466915, 2383: 0.512839813126, 2384: -5.03009083869, 2385: -0.580102737926, 2386: 6.23956750132, 2387: 1.03136322642, 2388: 12.5291664575, 2389: 5.28517664231, 2390: 9.70968028822, 2391: 1.79100393034, 2392: 4.29564858575, 2393: -4.7077970903, 2394: 13.6024524277, 2395: 0.943956201635, 2396: -0.13614598754, 2397: 0.634677941475, 2398: 1.85424014974, 2399: 5.50811847449, 2400: 0.926113113223, 2401: 5.41259398854, 2402: 0.250205683435, 2403: 0.452296855194, 2404: 0.895251833474, 2405: -0.420564307739, 2406: 0.421170635452, 2407: -0.0941057840338, 2408: 30.5720613871, 2409: 0.263458353252, 2410: -3.1077753273, 2411: 0.419917526442, 2412: 0.372126716951, 2413: -5.78789209815, 2414: -1.47013848948, 2415: 2.25985701334, 2416: 2.42192426855, 2417: -1.6550302383, 2418: -0.542925099453, 2419: -2.42183211021, 2420: -3.58081219162, 2421: -5.8977630748, 2422: -9.66211252272, 2423: -20.3464034346, 2424: 11.3803472084, 2425: 0.763650609835, 2426: 6.78305651868, 2427: -0.756787617953, 2428: -10.7214748051, 2429: -2.18153159092, 2430: -18.9678810111, 2431: 0.0174495200832, 2432: 7.77263021863, 2433: 0.373076349874, 2434: 1.76316159743, 2435: -3.11208956295, 2436: -8.96218547448, 2437: 6.90033321788, 2438: 9.72751522122, 2439: -14.2140767931, 2440: -25.3288389214, 2441: -12.1966318122, 2442: -27.9514517478, 2443: -3.18348294825, 2444: -2.70761779245, 2445: 13.5873373561, 2446: -9.87595096824, 2447: 78.9001509548, 2448: 23.8617215974, 2449: 2.411124173, 2450: 27.061053772, 2451: 45.2591220591, 2452: -45.6337458442, 2453: -6.92418288884, 2454: -99.9551309627, 2455: 80.408802453, 2456: -53.077646382, 2457: 48.7469469451, 2458: -32.0426305257, 2459: -5.11880546864, 2460: 84.4891057039, 2461: -30.7913060764, 2462: -61.0362795402, 2463: 0.861831042177, 2464: 21.7414645866, 2465: 35.2684563353, 2466: 20.9278614612, 2467: -54.2057560746, 2468: 58.3853710412, 2469: -95.8694699412, 2470: 19.1599009063, 2471: 6.9916785536, 2472: 92.7052721122, 2473: 88.3158044435, 2474: -49.5473796873, 2475: 4.79878533038, 2476: 9.31660315152, 2477: -11.4609296286, 2478: 11.6932052569, 2479: 15.0830829385, 2480: -5.48483499665, 2481: 16.8073270635, 2482: 41.7380872879, 2483: -6.46757712431, 2484: 100.0, 2485: -12.6045166295, 2486: -35.363353862, 2487: -39.7562768264, 2488: -7.16884537873, 2489: 59.3843326106, 2490: 27.3164825304, 2491: -10.5719444593, 2492: -17.2982881907, 2493: 19.6330574049, 2494: 7.5696241985, 2495: -24.4985873488, 2496: -6.52259598489, 2497: -9.43115418386, 2498: 33.8561309354, 2499: -10.0437294739, 2500: 28.7311062633, 2501: 16.9762541323, 2502: 50.6162889934, 2503: 35.6374175647, 2504: -25.9751370023, 2505: 61.1781491842, 2506: -7.27402172274, 2507: -34.4138287159, 2508: 26.2726889376, 2509: -9.95324168806, 2510: -0.828065434037, 2511: -2.32537626616, 2512: -36.0832446, 2513: -9.62596117161, 2514: -10.2844639296, 2515: -8.14963488635, 2516: -10.5248606235, 2517: -6.87712825363, 2518: -5.32157889879, 2519: -10.8666169469, 2520: -12.5821141071, 2521: -5.2648894072, 2522: -0.691610361804, 2523: -56.2616780457, 2524: 35.2367549494, 2525: 20.9338178986, 2526: -9.8906107336, 2527: -9.60588042933, 2528: -51.1135478046, 2529: -10.9426135564, 2530: 38.1889431655, 2531: 16.1063258972, 2532: -9.96661943694, 2533: -7.37212799815, 2534: -22.7770339853, 2535: -10.2877264426, 2536: 42.1296975516, 2537: -21.8633126696, 2538: -10.5744730529, 2539: -9.82218990342, 2540: 25.9702405182, 2541: -10.7329584992, 2542: 36.0291687477, 2543: -10.5302931064, 2544: 20.2386284763, 2545: -2.75007769293, 2546: 1.4381166441, 2547: -10.4475476079, 2548: -2.94245616219, 2549: 35.9828940223, 2550: -10.3677548892, 2551: -10.4975821231, 2552: -9.67757803544, 2553: 25.1480017133, 2554: 9.92408521554, 2555: 33.1329980064, 2556: -63.0618614284, 2557: -70.8532929297, 2558: -1.58589905438, 2559: -11.892856375, 2560: -4.75930208237, 2561: 11.0071327102, 2562: 67.7115765593, 2563: -29.8762616541, 2564: -39.1410178551, 2565: -6.67251643227, 2566: 49.872092637, 2567: 25.7447500061, 2568: 45.9095142881, 2569: -2.67167663806, 2570: -2.02979396972, 2571: 7.79452852513, 2572: 22.5202195591, 2573: -2.30589425644, 2574: -10.2645030612, 2575: -3.0113606729, 2576: -34.9187756238, 2577: -1.45781613411, 2578: -18.9687926044, 2579: 7.99839199231, 2580: -3.34750884214, 2581: 21.8541332361, 2582: 18.8945573121, 2583: -26.3663236978, 2584: -14.7076999726, 2585: 79.1252803039, 2586: 31.7751596263, 2587: 1.99329497253, 2588: 31.7802917072, 2589: -1.66993292051, 2590: -2.56695716512, 2591: -4.26253653005, 2592: -2.21038466802, 2593: -10.2986560643, 2594: 7.90557725441, 2595: -5.34798554362, 2596: -2.44769160783, 2597: -2.31803959731, 2598: -2.65017931487, 2599: -10.6423506223, 2600: -32.3049429045, 2601: -14.7845834747, 2602: -4.33565274519, 2603: -2.91426263644, 2604: -5.76646244479, 2605: -11.6850445521, 2606: 4.03465182853, 2607: -13.4074029936, 2608: -29.5016455971, 2609: 14.2551870248, 2610: -11.4567596318, 2611: -20.2757785277, 2612: 28.995230419, 2613: 3.07904083497, 2614: -5.00368882035, 2615: -3.07567797486, 2616: 21.234330042, 2617: -2.51795077984, 2618: 78.0164415803, 2619: 42.1609224407, 2620: -36.1978576247, 2621: -3.60928498177, 2622: -2.54553380998, 2623: -3.18464962573, 2624: -2.98795456033, 2625: -2.41755198564, 2626: -5.85884441639, 2627: 24.2030776101, 2628: 24.1697274961, 2629: 100.0, 2630: -36.210342562, 2631: 12.0645780679, 2632: -28.5870616954, 2633: 14.4730098035, 2634: -12.75264049, 2635: 39.2505845213, 2636: -3.05743585669, 2637: -51.5544861704, 2638: -8.14375660477, 2639: -1.61957584703, 2640: -7.93532406568, 2641: -21.208201325, 2642: -2.29309504919, 2643: -10.2267489935, 2644: -2.56013492764, 2645: -36.1029052576, 2646: -2.10092081051, 2647: -2.50663931608, 2648: 3.42747187106, 2649: -10.2024341494, 2650: -8.37254135589, 2651: -2.20965576352, 2652: -2.39352469045, 2653: -8.78810551484, 2654: -2.22740220889, 2655: -10.705569932, 2656: 13.0537159043, 2657: -37.7990704307, 2658: -3.4538950774, 2659: -5.35624331887, 2660: -18.1767539903, 2661: -13.4535130168, 2662: 6.4661143123, 2663: -67.2846009636, 2664: 12.1480868294, 2665: 24.0925519264, 2666: 46.4525186974, 2667: -8.2388051339, 2668: -10.7943424962, 2669: -2.59468621549, 2670: 27.8047571479, 2671: -8.21437439537, 2672: 12.7056600781, 2673: -10.9558508899, 2674: 1.46251530743, 2675: -1.2560689588, 2676: 92.8801150373, 2677: -12.1537066369, 2678: -9.69291190087, 2679: -7.16988409791, 2680: 9.8392473314, 2681: -28.6790289674, 2682: -14.9536007149, 2683: -10.2564835865, 2684: -10.2605651348, 2685: 17.0424282124, 2686: -23.7425138315, 2687: 21.1067044513, 2688: 55.3375485557, 2689: 40.4213371438, 2690: 34.9170200021, 2691: 21.872650554, 2692: -37.431199572, 2693: 2.4507252267, 2694: -86.228815764, 2695: 37.7042432326, 2696: -59.1674717404, 2697: -3.33648970375, 2698: -27.73136216, 2699: 4.0647053772, 2700: -26.5906419575, 2701: -10.6811144585, 2702: -10.3947631342, 2703: 11.2533986706, 2704: 5.29778479333, 2705: -10.5890028335, 2706: -11.0356555529, 2707: -10.0740874891, 2708: -10.5581672666, 2709: 41.9313208236, 2710: -11.2206951782, 2711: 31.9751622335, 2712: -8.44434862699, 2713: -4.61355205988, 2714: -7.43307517583, 2715: 5.27739877617, 2716: -69.5119976394, 2717: -28.1827479031, 2718: -22.4320158934, 2719: -23.9938531254, 2720: -4.15999374594, 2721: -89.2393654321, 2722: 10.9317278251, 2723: -2.45181166852, 2724: -4.1394218476, 2725: 56.5945850347, 2726: 67.615391659, 2727: -30.3953386322, 2728: -72.6291057766, 2729: -12.5418355728, 2730: -2.0683420458, 2731: -7.46274185822, 2732: -10.2314944022, 2733: -9.34905046854, 2734: 5.78665600562, 2735: -17.7083591764, 2736: -52.9127944019, 2737: -61.737869874, 2738: 33.1625628885, 2739: 36.8780183126, 2740: 41.4323979659, 2741: 24.6062445909, 2742: -11.7388408529, 2743: 15.5378574648, 2744: 6.76511388175, 2745: -39.4982149651, 2746: -10.7720211737, 2747: 100.0, 2748: -78.1057532892, 2749: -10.5158808966, 2750: 41.4101532383, 2751: -10.2908392309, 2752: -10.3491837467, 2753: -10.8020229517, 2754: 28.6247517704, 2755: -12.5693552269, 2756: -10.3953987568, 2757: -25.3520603475, 2758: -10.3820184827, 2759: 29.2211542498, 2760: -13.2266007522, 2761: -10.6699238983, 2762: 100.0, 2763: 4.28116700928, 2764: -10.6993180878, 2765: 8.56621984259, 2766: -26.1077226376, 2767: -9.13004779966, 2768: -28.8166657485, 2769: -25.5026958955, 2770: 0.380603740616, 2771: -9.81847048742, 2772: 15.2386276873, 2773: 31.1225347877, 2774: -12.7939501647, 2775: 5.7070390071, 2776: -5.57525341462, 2777: -8.02254439114, 2778: 5.06502461797, 2779: -63.9877070581, 2780: 3.39714061911, 2781: 71.2848609996, 2782: -10.3407409095, 2783: 61.1213900882, 2784: -33.5501081817, 2785: 43.1763250272, 2786: -54.9836003861, 2787: -93.2131880425, 2788: -21.6840266569, 2789: 100.0, 2790: 42.4455790302, 2791: 14.7507927826, 2792: 65.0245186863, 2793: -0.470032066626, 2794: -26.7626920309, 2795: 1.22482084228, 2796: -42.4161423469, 2797: 22.1574616709, 2798: 11.9502976316, 2799: 4.36241054765, 2800: -31.9736920447, 2801: -20.1570713525, 2802: -5.06231861936, 2803: -20.8348376038, 2804: -28.3594663241, 2805: 10.2307549306, 2806: 12.4417439977, 2807: 0.289287546659, 2808: 35.4198929334, 2809: 81.6092827004, 2810: 4.84116289633, 2811: -23.7792477069, 2812: -5.68081656325, 2813: -1.50587714709, 2814: 26.0027727793, 2815: -4.93168328027, 2816: -9.85738714193, 2817: -21.713983504, 2818: 21.4021111415, 2819: 25.0050450336, 2820: -5.53309463706, 2821: -21.0980580092, 2822: -1.34714284677, 2823: 35.476129097, 2824: 23.4156958488, 2825: -8.21754701218, 2826: -3.15001783763, 2827: 25.5257822634, 2828: 11.6386586321, 2829: -4.28066624535, 2830: -2.04898832777, 2831: -12.7218801894, 2832: -6.68331903274, 2833: 1.24708387193, 2834: 2.01445603475, 2835: -48.9566156454, 2836: -42.422354134, 2837: 1.23709118632, 2838: -7.16906244096, 2839: 28.1947971842, 2840: 0.112555366352, 2841: -20.1630151628, 2842: -0.316223147541, 2843: 0.0927923341804, 2844: 23.0430126849, 2845: -0.0379652496867, 2846: 17.5073349732, 2847: -11.168678691, 2848: 0.142031090618, 2849: -68.6895769169, 2850: 9.69870558756, 2851: 5.19788482849, 2852: 16.5029233934, 2853: -19.2951694269, 2854: -8.46141052471, 2855: -33.7059707635, 2856: -22.4159859513, 2857: 12.0400287614, 2858: 0.12200437924, 2859: -17.612719628, 2860: 0.160466724733, 2861: 12.5430660449, 2862: 0.214780981409, 2863: 0.227256705843, 2864: 32.2950026937, 2865: 0.128159745775, 2866: -31.787174653, 2867: -0.895522025342, 2868: 0.131331443269, 2869: 0.14761669283, 2870: -6.73321828747, 2871: -13.9638536744, 2872: 19.6282042777, 2873: 0.325353385, 2874: -0.730152700852, 2875: 0.0394719066568, 2876: -0.230558532558, 2877: 0.550095006934, 2878: 4.6220111781, 2879: 9.42626550236, 2880: 10.5345659798, 2881: 0.0783392576798, 2882: 9.82941008495, 2883: 65.4395154255, 2884: -0.0084211386363, 2885: -9.1507702307, 2886: 15.106553182, 2887: -0.0622440623904, 2888: 1.01893784821, 2889: -17.8809769197, 2890: -0.0449111974843, 2891: -12.163787812, 2892: 0.692476806638, 2893: 8.84874487846, 2894: 0.0574331598795, 2895: -0.0418412409626, 2896: 3.60935418743, 2897: 0.0260103879699, 2898: 32.574864092, 2899: -0.0243828727526, 2900: 0.0593564436918, 2901: 0.0585781231514, 2902: 14.1256482615, 2903: 14.7774057203, 2904: 5.76642511524, 2905: -32.3260560996, 2906: -7.441274862, 2907: -20.1093572109, 2908: -10.306322224, 2909: -16.8951329049, 2910: -2.42005600424, 2911: 4.00210543449, 2912: -1.26129268371, 2913: -33.793426095, 2914: 3.92384999829, 2915: 40.0834601681, 2916: 8.01901941831, 2917: -86.9866841425, 2918: -0.276387265137, 2919: 1.52381356167, 2920: 0.0234976531182, 2921: 22.1809838417, 2922: -0.275741321107, 2923: 0.0133504759734, 2924: -1.07590322078, 2925: 6.87065463333, 2926: 4.5079982585, 2927: -8.14724450259, 2928: -36.6556498815, 2929: -0.561969100827, 2930: 9.97505626278, 2931: -9.00866435648, 2932: -0.545115458566, 2933: -4.12745125396, 2934: -6.27371011185, 2935: 24.521787293, 2936: 1.07957385743, 2937: 12.1909675622, 2938: -0.183168599579, 2939: -0.0491070070245, 2940: 15.1443668209, 2941: -0.168793822948, 2942: 0.0172587134286, 2943: 7.88331928656, 2944: -2.58230644802, 2945: -0.2655140174, 2946: 0.0953536498363, 2947: -0.245792040487, 2948: 0.06103929002, 2949: -6.27116558691, 2950: -4.50039232895, 2951: 0.154408890983, 2952: -0.0917460742764, 2953: -2.80143052376, 2954: 0.696236457536, 2955: -15.7092809339, 2956: -2.060426799, 2957: 48.741800157, 2958: 39.0216085721, 2959: 0.852594281173, 2960: -12.4260296344, 2961: -24.023386588, 2962: -6.03248260255, 2963: 26.4184865629, 2964: -0.417117046885, 2965: 27.9106472725, 2966: -0.174760987162, 2967: -10.2538143526, 2968: -1.78377700059, 2969: 2.99624175811, 2970: -0.911831244293, 2971: -0.125595927617, 2972: -0.279022699102, 2973: -0.840695330133, 2974: -0.43026878265, 2975: -0.31551050168, 2976: 27.9132110203, 2977: -11.3609396606, 2978: 15.0790667493, 2979: 25.442769577, 2980: -31.39670216, 2981: 3.64745631467, 2982: 15.1375736655, 2983: -20.5669461283, 2984: 14.6087736833, 2985: -0.176280953153, 2986: -5.14629925954, 2987: 2.62300521614, 2988: 11.2994807161, 2989: -1.678362507, 2990: 27.0906344933, 2991: -0.233646070458, 2992: 0.0231972390952, 2993: -0.110721636092, 2994: 34.9195278359, 2995: -0.248261279832, 2996: -0.161287482735, 2997: -39.0352377943, 2998: 0.10313549138, 2999: -43.0586681724, 3000: -0.243387171835, 3001: -0.151353958221, 3002: 0.314119929336, 3003: -1.09412820659, 3004: 5.79102167659, 3005: 4.05161732992, 3006: 11.368555733, 3007: -0.137226982618, 3008: -0.729981298171, 3009: 32.7909911454, 3010: -2.62952292832, 3011: -4.05505022716, 3012: 17.3630629722, 3013: -4.99252854204, 3014: 6.34412409972, 3015: 12.168417119, 3016: 0.199571177802, 3017: 3.20844313682, 3018: 0.212866412375, 3019: 35.5443488556, 3020: 9.80892770789, 3021: 3.35257822335, 3022: -3.90221485619, 3023: 11.2537291156, 3024: 0.142614330225, 3025: -32.6622386118, 3026: 8.87625052633, 3027: 0.280738923779, 3028: 24.0919587607, 3029: 17.1345917195, 3030: 22.2777691525, 3031: 0.0330388055859, 3032: 0.149673511332, 3033: 37.1294677283, 3034: 1.27255247585, 3035: 17.1009598554, 3036: 16.8233713151, 3037: 20.8093316135, 3038: 28.2155871328, 3039: -27.0731883336, 3040: 9.30074611319, 3041: 6.60538678371, 3042: -54.6618042162, 3043: 3.72497092325, 3044: -38.5161351645, 3045: 28.3329887694, 3046: -17.7633728469, 3047: 1.15230719346, 3048: 0.0133738710888, 3049: -27.29179668, 3050: 0.0672670612933, 3051: 0.0420363101672, 3052: 7.04753756536, 3053: 10.2229673417, 3054: -0.0458517946216, 3055: -7.43204702234, 3056: -0.0160624209237, 3057: 0.0782008066191, 3058: 69.3038043009, 3059: 0.780685255497, 3060: 15.4459728657, 3061: -3.41192116117, 3062: -5.10965711219, 3063: 0.635123189245, 3064: -7.00255851813, 3065: -15.4123738532, 3066: 33.9722825799, 3067: -12.8755339459, 3068: 17.2683198276, 3069: -20.0413300949, 3070: 34.6190856194, 3071: -13.1810262271, 3072: 51.7128085568, 3073: 1.3193259504, 3074: -31.7298622494, 3075: 25.4917098586, 3076: -12.390412035, 3077: 31.3495350439, 3078: -23.1192282616, 3079: -0.0422963795652, 3080: -10.22855252, 3081: 9.92551555202, 3082: 9.31601331206, 3083: 1.36540092535, 3084: -12.2748345788, 3085: -18.6143238041, 3086: -4.67183922815, 3087: -22.1175146397, 3088: -9.83815155635, 3089: 41.2807820088, 3090: 4.87335106106, 3091: -22.545794596, 3092: -53.0411781519, 3093: -1.85167066408, 3094: 26.7572970884, 3095: 4.78490690587, 3096: 6.53590119553, 3097: 26.2976494179, 3098: -0.119090452751, 3099: -15.3113000126, 3100: 0.0137761787083, 3101: 0.044105266047, 3102: -0.135387048693, 3103: -26.5697553281, 3104: -0.0179286662065, 3105: -0.0616653459379, 3106: 58.4970213141, 3107: -0.0538712155653, 3108: -43.7099785222, 3109: 0.03229350006, 3110: -0.0876320778082, 3111: 22.3573841937, 3112: 19.6492597913, 3113: 7.8472844449, 3114: 12.0270710748, 3115: 9.73200826546, 3116: -10.1613244562, 3117: -36.588063586, 3118: -41.6205220433, 3119: 6.57538857444, 3120: 38.6692651152, 3121: -30.0993700902, 3122: 4.59539758846, 3123: -5.43969956304, 3124: 27.0387118686, 3125: 1.34110105981, 3126: 0.354547506447, 3127: -4.07151311667, 3128: -44.8632748713, 3129: -27.0408912075, 3130: -14.4301482515, 3131: 0.0739808597375, 3132: 8.74028319294, 3133: 4.93692597979, 3134: 0.200431036159, 3135: -4.00612708474, 3136: 13.0480563808, 3137: -2.63524353607, 3138: 27.232836832, 3139: 31.2017300149, 3140: -19.2644840719, 3141: 22.3043168311, 3142: -0.0983436656685, 3143: -8.3893159007, 3144: 2.52372033933, 3145: -5.47692085334, 3146: -8.39901172582, 3147: 5.48471614469, 3148: 2.38951930604, 3149: 16.6537458923, 3150: -0.217343028587, 3151: 9.23207111654, 3152: 15.6171461005, 3153: 2.74210543937, 3154: -0.711983202408, 3155: -0.767518202271, 3156: 7.85616078688, 3157: 3.92304316741, 3158: -2.9838037893, 3159: 2.86083109183, 3160: -17.6947456018, 3161: 8.65997377798, 3162: -2.31218773934, 3163: -24.7918654907, 3164: -10.3559999297, 3165: 11.1417139994, 3166: 15.1417663268, 3167: -10.1262772029, 3168: 3.65599896787, 3169: -5.31007039653, 3170: 11.488083022, 3171: -1.98223967879, 3172: -5.5171765199, 3173: -1.38280333054, 3174: -0.132111051796, 3175: -1.99708820656, 3176: -14.5859495279, 3177: 0.522463458087, 3178: -6.24428085024, 3179: -9.59458696217, 3180: -11.2894754826, 3181: -13.8961754368, 3182: -18.1482751013, 3183: -5.23357570597, 3184: -19.7434050125, 3185: 3.46910574425, 3186: 6.02478542082, 3187: 6.85896980221, 3188: -7.22023139626, 3189: -1.41655806309, 3190: 0.306567860793, 3191: -4.30757829474, 3192: 16.0912716325, 3193: -21.0683736815, 3194: -2.10553543865, 3195: 0.420956921172, 3196: 3.98308543457, 3197: -2.58903242636, 3198: -2.93644451547, 3199: -3.50655969017, 3200: 9.55014286821, 3201: 5.86428144063, 3202: 6.97033477217, 3203: 7.58221482448, 3204: -19.5142838375, 3205: 15.9872716317, 3206: -2.5735335462, 3207: -1.62283338652, 3208: 1.66475643359, 3209: -2.00688129787, 3210: -0.266273195692, 3211: -1.02573626144, 3212: 1.28869144759, 3213: 3.8046866109, 3214: -2.41042140369, 3215: -1.37723389166, 3216: -2.74620295665, 3217: 1.64905859971, 3218: -4.65061205144, 3219: -0.254848275564, 3220: -23.1194685744, 3221: -7.37317002623, 3222: 0.925916916993, 3223: 1.98554923398, 3224: -1.84152177636, 3225: 0.201426795463, 3226: 9.81977365983, 3227: -15.6514730103, 3228: 7.33477306672, 3229: -4.38686154343, 3230: -2.38194477188, 3231: 2.90264237188, 3232: 17.7809611237, 3233: -2.08024668393, 3234: 2.50487114191, 3235: 4.95300268432, 3236: -1.51794410418, 3237: -1.82541902241, 3238: 2.56607691547, 3239: -1.59883759241, 3240: -8.65694017585, 3241: 7.75966377656, 3242: -10.2876951315, 3243: 4.55023135124, 3244: 10.9018460676, 3245: -1.68627661555, 3246: -2.68835180317, 3247: 5.21515292438, 3248: -1.96949512364, 3249: -2.46996961138, 3250: -1.90768061322, 3251: 13.2403143347, 3252: 5.32703776985, 3253: 3.6042576927, 3254: -15.3794887743, 3255: -32.2312469247, 3256: -11.8196624658, 3257: 5.3304031173, 3258: 7.14078999532, 3259: 2.33505115577, 3260: -26.0484045216, 3261: -7.98470629998, 3262: 11.9223399036, 3263: -8.12949679906, 3264: -12.0112552734, 3265: -0.205635974939, 3266: -11.9077974017, 3267: -0.331182344569, 3268: 0.262605979111, 3269: 0.31699978962, 3270: -18.1252949172, 3271: -0.326830071896, 3272: -1.70529148605, 3273: -2.47730048516, 3274: 6.79941070331, 3275: -3.15225054511, 3276: 0.451908869488, 3277: 1.08654653204, 3278: -0.81875059215, 3279: 1.65874401016, 3280: 2.38090395361, 3281: 0.0289463588606, 3282: -19.9401542708, 3283: 2.58819871468, 3284: -3.50745815781, 3285: 0.697749586991, 3286: 5.09264494267, 3287: -0.744763331667, 3288: -0.391283403731, 3289: -4.5855000827, 3290: -0.404711549535, 3291: -2.01428554365, 3292: -2.13308594285, 3293: -0.717273171146, 3294: -0.420610236188, 3295: -0.378370717293, 3296: -0.281390562573, 3297: -1.50262087215, 3298: 16.1110226706, 3299: 0.38088911056, 3300: -1.12709006253, 3301: -0.759475807264, 3302: -2.73039192967, 3303: -3.27123098686, 3304: -12.066592193, 3305: 6.08995828087, 3306: 9.03980235936, 3307: -14.8520512019, 3308: 10.095740675, 3309: 7.44088038146, 3310: 1.76148371916, 3311: 2.4131277498, 3312: 0.164957508094, 3313: -0.162850496442, 3314: 1.42056458969, 3315: -0.339818702775, 3316: 9.20046235126, 3317: -17.0862225905, 3318: -8.37909171223, 3319: -1.39818192946, 3320: -0.389226175247, 3321: -0.370897323454, 3322: -0.449075583934, 3323: -0.423997048944, 3324: 0.158377720251, 3325: -4.39159279643, 3326: -3.31065551217, 3327: -11.1129574316, 3328: 17.9647968192, 3329: 2.75464068034, 3330: 6.79326608928, 3331: -5.07011351358, 3332: 3.58465583925, 3333: 7.93080665539, 3334: -0.322125167182, 3335: 11.2414689972, 3336: 2.25064325392, 3337: 0.119990471081, 3338: -2.5156422757, 3339: -3.28028343141, 3340: -0.40796792362, 3341: -1.97061652617, 3342: -0.337233225701, 3343: -11.0102344584, 3344: -0.242584770248, 3345: -0.416226878739, 3346: 14.0974435205, 3347: -2.16688457878, 3348: -9.26400121611, 3349: -0.168099730882, 3350: -0.378857755236, 3351: -1.10703417326, 3352: -0.381266472896, 3353: -0.748581070384, 3354: 6.30029089736, 3355: -2.5404528095, 3356: -0.539876833091, 3357: -0.992678280162, 3358: 7.35237351478, 3359: -1.08145291779, 3360: 1.40641418589, 3361: -1.65918892443, 3362: -1.5634503825, 3363: 6.28131173246, 3364: 2.16646645964, 3365: -1.65069402584, 3366: 1.74981240381, 3367: -0.688017671738, 3368: -3.76397737432, 3369: -2.18682577583, 3370: 1.71298793583, 3371: 2.67633261887, 3372: -6.71692275181, 3373: -0.109754550317, 3374: 15.7998388048, 3375: -5.40708821258, 3376: -3.17637867011, 3377: -14.2867814022, 3378: 5.98069492793, 3379: 4.2681404967, 3380: -2.21036269953, 3381: -1.88456652114, 3382: 12.5957933956, 3383: 3.6176066432, 3384: -2.75335799126, 3385: 4.99005107033, 3386: -10.8299727072, 3387: -24.7276936876, 3388: 3.07656492558, 3389: 16.5928568761, 3390: 16.1727899621, 3391: 2.62635627072, 3392: -3.72053832958, 3393: 7.18164483568, 3394: -8.56553297807, 3395: -6.87188547701, 3396: 6.13951654985, 3397: -1.63299760587, 3398: 9.14202883714, 3399: -1.82630383391, 3400: -2.1257630514, 3401: 1.21689170525, 3402: 6.34492613922, 3403: -1.88796309711, 3404: -2.32547408641, 3405: -4.02858775039, 3406: -1.64572997196, 3407: -3.24004911069, 3408: -7.0476249102, 3409: 3.88597667638, 3410: -2.57397863127, 3411: -3.05602515269, 3412: -2.69387813939, 3413: 14.6570090256, 3414: -4.71428547096, 3415: -4.90743877242, 3416: 7.70915313263, 3417: 1.65033375013, 3418: -9.44152648293, 3419: 1.8266295067, 3420: 4.93043049976, 3421: 11.7997546581, 3422: -3.86694107468, 3423: -3.78005129539, 3424: -4.61430107793, 3425: -17.9145586747, 3426: 2.85119021196, 3427: 5.43708805089, 3428: -1.75662768624, 3429: -4.8646854066, 3430: -0.0259781115633, 3431: -3.93814954737, 3432: -5.56717989401, 3433: -7.10735096592, 3434: 3.3673161909, 3435: 18.0339508842, 3436: -12.6076235035, 3437: 15.5710732926, 3438: -5.326236872, 3439: 11.3623574484, 3440: 2.02778884883, 3441: -0.0620168141738, 3442: -2.94458218467, 3443: 0.360136625362, 3444: -5.08027816472, 3445: 16.9944003672, 3446: -4.19792340096, 3447: -1.34337637182, 3448: 10.6228827554, 3449: -1.843888654, 3450: -1.92545582624, 3451: -3.03762823679, 3452: 2.30486616048, 3453: -1.49996397112, 3454: -2.19136666392, 3455: -25.6938251732, 3456: -2.15689043092, 3457: 0.541147679426, 3458: -2.01101498562, 3459: -2.12484961798, 3460: -1.47482144781, 3461: -3.51155175814, 3462: -0.113338638656, 3463: -3.33132385016, 3464: -6.79064567153, 3465: -4.37178440307, 3466: 8.84151508911, 3467: 18.8345191132, 3468: 2.97437629145, 3469: 14.5719038038, 3470: 18.298948614, 3471: 13.1852854218, 3472: 9.852006669, 3473: -0.304398064519, 3474: -2.36975060923, 3475: 8.43674265337, 3476: -2.80757418973, 3477: 26.7916061591, 3478: 6.9431364489, 3479: -7.06549267132, 3480: -2.05038373012, 3481: -8.9390854056, 3482: 0.804553190633, 3483: 0.713906942183, 3484: 14.6829238717, 3485: 7.50904002227, 3486: 1.82657615722, 3487: 5.15545629403, 3488: -12.5618142323, 3489: -8.48474594968, 3490: 12.8228303512, 3491: -0.801369161784, 3492: 1.81657341669, 3493: 7.98315922301, 3494: -0.655744520515, 3495: 1.72230049449, 3496: -64.3749358931, 3497: -1.34308386367, 3498: 0.341045716703, 3499: -0.489382736285, 3500: -0.703909711935, 3501: -0.885983185354}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.582256350219, 2: 5.75239330221, 3: -15.8269885383, 4: 4.42838927836, 5: 6.45472095099, 6: -6.08065226547, 7: 2.14412192107, 8: -4.85663316142, 9: 0.43365568297, 10: -1.94704952115, 11: -3.40610652216, 12: -1.01165832508, 13: 0.553239872003, 14: 0.27745890693, 15: 2.188846929, 16: -12.1508873496, 17: -3.49303092741, 18: -0.908817929905, 19: 1.62601939875, 20: -3.40753835203, 21: -35.8262540183, 22: -10.8100772215, 23: 0.31268469046, 24: -2.93783543868, 25: 4.12422795302, 26: -6.39912167227, 27: 0.230714336082, 28: 1.23516583305, 29: 9.68242880272, 30: 0.345503191546, 31: -2.07805825659, 32: -5.49583855958, 33: 4.86911369303, 34: 2.01248447931, 35: -22.7154933304, 36: -6.27963952797, 37: -0.688585097907, 38: -0.942127836255, 39: -9.01125615428, 40: -6.39098605345, 41: -3.71292673002, 42: -0.420771478336, 43: 12.0310033471, 44: 12.7765538773, 45: -2.73268542504, 46: 1.27444858402, 47: 0.547732034314, 48: 1.4863469124, 49: -0.687658321829, 50: 2.14392694947, 51: 0.917245465525, 52: -13.2161625561, 53: 2.22758203502, 54: 2.6649669655, 55: -0.109825413477, 56: 1.72299692539, 57: -14.8617137445, 58: -2.74117596896, 59: 1.00186481181, 60: 5.16775892499, 61: 2.19866808451, 62: 4.4524139201, 63: 7.05065321606, 64: -2.24940056269, 65: 6.16265867241, 66: 1.15659682704, 67: 12.5759156094, 68: 0.080355313085, 69: -7.35455429674, 70: 1.49348432798, 71: 1.62447870892, 72: 0.889872076098, 73: 1.91318307466, 74: 2.43300915056, 75: -4.1124017669, 76: 3.42795342718, 77: -3.24995127794, 78: -2.26473166641, 79: 6.49685682937, 80: -5.57081596105, 81: 1.2344639461, 82: -2.10178683742, 83: 1.66499498254, 84: 1.69989398686, 85: 1.77872823476, 86: 9.187480524, 87: 0.415632280591, 88: -1.64499604442, 89: 1.89501722521, 90: 1.80844462884, 91: 1.36064951983, 92: 1.73377517811, 93: 1.33741015121, 94: 0.869649318563, 95: 1.13679428352, 96: 3.63389546941, 97: 0.0131623844008, 98: 1.51100678762, 99: -2.03117951145, 100: 1.57241119668, 101: -3.56509670835, 102: 1.65322888958, 103: 12.1433434582, 104: 1.72134398761, 105: -2.00077327751, 106: -15.5269775024, 107: 0.292111625526, 108: 1.87883083351, 109: -0.731858887616, 110: -2.87928234067, 111: 0.36522273163, 112: -3.94159188889, 113: 21.3848025611, 114: -7.56486216655, 115: 16.6632923414, 116: 6.15275842936, 117: 4.56196497657, 118: 1.1813786258, 119: -21.4614455838, 120: 5.30192215876, 121: 8.51332970927, 122: -6.61824835512, 123: 1.17643818747, 124: -3.80334157196, 125: 6.19601859004, 126: 0.567834501165, 127: 0.21418343393, 128: 0.194618594506, 129: -8.6700897995, 130: 0.452025730392, 131: 1.80795341446, 132: 0.299123270895, 133: 4.69181507044, 134: 1.30362163015, 135: 0.981895644727, 136: -5.70126634288, 137: 0.591315143456, 138: 3.37445291704, 139: 0.93739896958, 140: -0.791639180344, 141: 16.1418701894, 142: 2.00317752283, 143: -0.303324904051, 144: 0.767706692192, 145: 1.16790262092, 146: 0.580655872473, 147: 0.621519690725, 148: 13.9283189453, 149: 0.606779392831, 150: -0.65550859261, 151: 1.81810972852, 152: 0.767484939956, 153: 0.573682222333, 154: 0.589013666052, 155: 0.437309853806, 156: -0.753049843502, 157: 3.48167957681, 158: 0.783214036398, 159: 0.994070662048, 160: 0.674685262787, 161: 1.29269510733, 162: 0.0699603077185, 163: -6.23944768748, 164: -3.16227192451, 165: -20.4184549299, 166: -10.863462847, 167: -1.57719578491, 168: -1.44296031681, 169: 0.847672087939, 170: 1.59953526715, 171: -4.60400646072, 172: 0.619712235452, 173: 4.04828311707, 174: 0.449697707075, 175: 4.77327108763, 176: -10.9163362062, 177: 5.37117269034, 178: -0.361200077043, 179: 0.563942533438, 180: 0.633849475326, 181: 0.749777401323, 182: 0.386571180482, 183: -0.658238391493, 184: -7.68778776026, 185: -3.1968054733, 186: -5.98381693821, 187: -7.19576263847, 188: 2.88679999458, 189: -3.45675510777, 190: 0.807315648732, 191: 3.10816576311, 192: 1.44504630872, 193: 0.502751062522, 194: 3.87270303921, 195: 0.630938846441, 196: -2.39858764826, 197: -0.114715094856, 198: -6.20575007481, 199: 0.426276034287, 200: 1.78381507629, 201: 0.603472792427, 202: 0.0607472558076, 203: 0.395299559334, 204: 0.392373260984, 205: -14.3073515235, 206: 1.92161304487, 207: 3.09177322106, 208: 0.431410283335, 209: 0.43443389586, 210: 0.106754866512, 211: 0.532891541516, 212: 1.03806146353, 213: 0.498259028912, 214: 0.0489683454476, 215: 0.280445643676, 216: 0.308578340395, 217: 0.95127791607, 218: 0.458206208162, 219: 0.562007187711, 220: 9.86312074122, 221: -0.113822414188, 222: -1.20397241767, 223: 1.25461801449, 224: 1.49695916748, 225: 1.8249482444, 226: 0.63913651124, 227: 3.8105331257, 228: -0.939903313575, 229: 1.34466087831, 230: -0.846760869004, 231: -11.7798879422, 232: 0.368877175709, 233: -5.58092827886, 234: -1.99915965831, 235: 3.26609020939, 236: 1.73386069179, 237: -5.12971305489, 238: -6.44817724895, 239: 1.52453567187, 240: 1.86978606105, 241: 4.56872777998, 242: 2.22756537633, 243: 9.92623523099, 244: -6.96254514572, 245: 5.41702275981, 246: -11.2777734452, 247: -11.4627009953, 248: -5.70309134558, 249: -0.601085107104, 250: -22.32822622, 251: 12.2194593561, 252: 2.81326145955, 253: 6.87429542377, 254: -6.66279941272, 255: 1.01853083588, 256: -0.0533181730655, 257: -9.94032127813, 258: 1.35304691416, 259: -0.79676935768, 260: 2.03972876823, 261: -2.3341941355, 262: 1.629058548, 263: -2.55072718744, 264: 1.32962632677, 265: -0.902981371146, 266: 18.4705099317, 267: -1.05616772479, 268: -3.17938870525, 269: -0.601832723751, 270: 3.32004384549, 271: 0.197761559592, 272: 1.50953735718, 273: 1.45329139805, 274: -3.21227132823, 275: 3.22474471212, 276: 8.21186022545, 277: -7.71663077091, 278: 4.00740501514, 279: 0.702658834149, 280: -5.14570796965, 281: -2.89849998291, 282: -1.82708684677, 283: 2.69026665971, 284: -2.60323007657, 285: -7.85421890936, 286: 10.2137310689, 287: 2.01061007128, 288: -0.45835562717, 289: -5.75009323435, 290: -2.48545539185, 291: -0.495316445047, 292: -7.04586409011, 293: -3.30568749021, 294: -8.23930794335, 295: -0.903678574092, 296: -4.68191097535, 297: -10.2591473853, 298: -9.3176989207, 299: 1.03472435485, 300: -1.22717161926, 301: 2.4063991256, 302: 8.90430917491, 303: 2.61863148667, 304: -10.0052157041, 305: -10.1137220683, 306: 1.09133935794, 307: -3.14520393445, 308: 1.72737436923, 309: 1.91924454774, 310: 2.11828353883, 311: -0.903845153329, 312: 2.19207423394, 313: 1.83445857552, 314: 3.25190745549, 315: 1.95915233077, 316: -0.455575998002, 317: 0.272984389435, 318: 1.75167198552, 319: -4.05645430243, 320: 1.79996769498, 321: 0.770909949771, 322: 6.36291942928, 323: -1.51467666582, 324: 0.971245094252, 325: -6.4057476159, 326: -3.12215476737, 327: -0.0948164276707, 328: -9.15313655573, 329: 2.27814216052, 330: -6.06132550309, 331: -0.814678887799, 332: -4.7089518997, 333: 1.33286646598, 334: -0.251621626762, 335: 1.11017401685, 336: -6.37215056655, 337: 11.7447279707, 338: -0.430979267606, 339: 2.42573428369, 340: -2.2028059696, 341: 1.82633893814, 342: 2.9010245072, 343: -2.41816443513, 344: -5.60220656191, 345: -6.10992525168, 346: -4.25785953773, 347: 6.24610092725, 348: 9.12877496956, 349: 2.28335806644, 350: 0.17261368529, 351: 50.1443623984, 352: 34.0701880204, 353: -47.7211039662, 354: 45.3150939778, 355: 26.5286600289, 356: -51.2818723683, 357: -5.53699318929, 358: -46.856222401, 359: 7.54345913724, 360: 9.60655491112, 361: -18.0730007484, 362: 19.4516401017, 363: -53.260500315, 364: 16.9067335282, 365: -32.6932814848, 366: -41.2956353981, 367: 8.46589745433, 368: -72.6450046868, 369: -55.8366392439, 370: -70.6700235931, 371: 24.9080096609, 372: 7.21591183318, 373: -7.9495979593, 374: 23.8822369118, 375: 0.435729752007, 376: -57.0159616079, 377: -20.2293551097, 378: 5.58459116925, 379: -4.28915286613, 380: -16.8463084813, 381: 55.8683161101, 382: 1.99586053724, 383: -45.5021622359, 384: -3.01949985035, 385: -0.715857245377, 386: 23.9638492934, 387: 20.8748774591, 388: -36.4597193846, 389: -2.92973630483, 390: 16.6483766289, 391: 1.18286380503, 392: 14.7156402358, 393: 30.1683662562, 394: 11.9116247054, 395: -17.0214662455, 396: 29.2381108345, 397: -10.5053839427, 398: 11.6186034984, 399: -2.14960345169, 400: 24.5429611677, 401: 24.391880339, 402: -0.655594908566, 403: 42.4665128387, 404: 1.1805649331, 405: -0.0640491310483, 406: 73.1674604258, 407: -9.56293654524, 408: 23.9823281553, 409: -36.4432857192, 410: 2.39633988557, 411: -19.1787739918, 412: 21.0524243517, 413: 23.1991996436, 414: 27.0384196267, 415: -8.01293836406, 416: 15.4488130801, 417: 1.25854199624, 418: 3.30319204506, 419: 2.78779948069, 420: 4.79491254177, 421: 27.9151593511, 422: 0.660925534863, 423: -72.5425601863, 424: 4.14473783431, 425: 1.15861918264, 426: 2.99944266152, 427: 1.17883448775, 428: -8.7966406884, 429: -8.10138007838, 430: 23.226893782, 431: 63.9309636115, 432: 1.10348057652, 433: 13.5547430174, 434: -14.1725051733, 435: 2.89766189174, 436: 9.09951091386, 437: -35.6396562575, 438: -5.15980168114, 439: 38.0023292162, 440: -1.84593694239, 441: 1.1744780076, 442: -8.19058285416, 443: 1.15484547827, 444: 1.40601989704, 445: 1.1467091548, 446: 1.95408037888, 447: 1.33905384739, 448: 3.0481887183, 449: 5.77513692175, 450: 33.3979435773, 451: -0.349490163548, 452: -45.5870406422, 453: -9.98874531682, 454: 5.10762572601, 455: 30.3338839708, 456: -13.5125183375, 457: -2.06366039741, 458: -4.95093153501, 459: 19.9272233085, 460: 8.35545402769, 461: -8.42399051565, 462: -71.4594959106, 463: -4.90271492884, 464: -50.1701172608, 465: -22.8246114421, 466: -49.5723841331, 467: 35.6083044797, 468: 25.9927318184, 469: 54.3457068264, 470: 100.0, 471: -0.460497325139, 472: -37.8245976445, 473: -45.303145462, 474: 4.05661131265, 475: 1.48670947475, 476: -1.98871734572, 477: 17.2281481594, 478: -53.8975491471, 479: 0.229798103284, 480: -4.83785415087, 481: 2.2357513075, 482: -32.8777817222, 483: 6.02920896365, 484: 15.0460708719, 485: 3.66463981234, 486: -1.61705208641, 487: -6.70907638778, 488: -3.02271897152, 489: 18.861998344, 490: -26.1374607249, 491: -34.1716048053, 492: -20.6967898243, 493: 0.0109630357202, 494: -10.6520797273, 495: 0.916539820312, 496: 0.399685128135, 497: -95.1873115695, 498: 0.569601117509, 499: -2.24188397272, 500: 40.2866361795, 501: -0.61733997742, 502: 0.271306853401, 503: 0.539656198889, 504: 0.338279514117, 505: 4.2928112072, 506: 23.4157777598, 507: -0.200197716575, 508: 0.318480545562, 509: 0.267052366736, 510: 7.66903963224, 511: 0.740150061341, 512: 8.58196420192, 513: 12.3837258315, 514: 54.7067599327, 515: -75.4699896551, 516: 38.2285748005, 517: -17.3230098885, 518: -1.53765125846, 519: 17.4648626221, 520: 1.3445615132, 521: -0.115788390455, 522: -29.9932698379, 523: 0.285717338722, 524: -5.42402669908, 525: -99.8028507936, 526: 100.0, 527: 5.11901796011, 528: 0.155049997687, 529: -0.416225409827, 530: 0.392218133755, 531: 0.428624773552, 532: 3.6446467358, 533: -11.7405508776, 534: -22.4518747357, 535: -53.6880020641, 536: 16.0491695936, 537: 18.7379078543, 538: 39.5991209077, 539: 27.9077663379, 540: 12.1083867641, 541: 57.6950059149, 542: 0.23237346994, 543: 29.8038722581, 544: 4.8198098165, 545: 8.37977292879, 546: 1.92741452285, 547: -12.1045005109, 548: 0.244341840204, 549: -0.40752767496, 550: 0.517000437565, 551: -12.6933365194, 552: 0.343317584413, 553: 0.314270904947, 554: -95.6123767483, 555: 4.42916099595, 556: 66.8847551321, 557: 0.437099011381, 558: 0.399638578554, 559: 2.95751449725, 560: 0.276397561237, 561: 1.4527106143, 562: 36.8199516621, 563: -29.8210820426, 564: -0.0933703011555, 565: -3.31969050781, 566: -27.7836788516, 567: 8.01154814813, 568: 57.2375592777, 569: -12.9631362427, 570: -5.00090499941, 571: -6.79720206318, 572: -14.4035032768, 573: -4.19820712905, 574: -14.3007490342, 575: 0.222424058091, 576: 23.4559156497, 577: 1.6611484325, 578: -3.94075826893, 579: 15.5473277255, 580: -7.97584233084, 581: 0.187212370826, 582: 43.3739527644, 583: 5.15120294672, 584: -39.8889170086, 585: -35.0624324307, 586: 18.4128311707, 587: -52.8713152475, 588: 3.84939242464, 589: -4.71497899822, 590: 10.22068565, 591: 29.6726949314, 592: -20.8323077847, 593: -5.36818173883, 594: -29.6871798477, 595: -31.6496355697, 596: 69.7517426513, 597: 2.24440431383, 598: -67.0624227987, 599: 28.7426627529, 600: 8.40108134208, 601: 29.5638637601, 602: -26.578257422, 603: -10.7964873116, 604: -41.5321203322, 605: -0.964036382169, 606: 22.5711045016, 607: -9.13840052942, 608: -2.23907499856, 609: -2.77812222856, 610: 50.4153330688, 611: 2.62414573663, 612: -11.9448362638, 613: 2.03368429875, 614: 4.13100761437, 615: 2.74846260718, 616: -43.8534992532, 617: 32.4768023203, 618: -14.291296442, 619: 9.09400356495, 620: 1.7501434545, 621: 97.7015414258, 622: -30.0403370953, 623: -15.5417876109, 624: 100.0, 625: -31.5049728875, 626: 3.61496280747, 627: -25.226099619, 628: -40.3067560105, 629: -3.66533232381, 630: 19.7021615538, 631: 38.4555323416, 632: -81.2172985134, 633: 34.404840792, 634: -57.1603897897, 635: -60.0883945417, 636: 1.4167651974, 637: 9.64494314373, 638: 1.20194552288, 639: 5.10299908033, 640: 6.20066769424, 641: -10.3278826703, 642: -34.3087449981, 643: 66.7477769192, 644: -10.698977585, 645: -9.9728200165, 646: 12.0310141356, 647: -70.1898537936, 648: 0.813725442667, 649: 26.7626464017, 650: 2.35749272502, 651: -75.7159969317, 652: 4.82984500126, 653: -59.1395776023, 654: -79.6939123513, 655: 1.41952669058, 656: 46.7861283254, 657: -0.341712548958, 658: -0.323630863631, 659: -5.78258373533, 660: -11.9540174237, 661: -0.946186287539, 662: 6.64369874967, 663: -6.97570445253, 664: 4.46452711168, 665: 52.7216775667, 666: 1.08837461497, 667: -0.35788606071, 668: -55.9705522898, 669: 3.14162003105, 670: 1.53280907541, 671: -11.3956991736, 672: -32.7085262147, 673: -12.4086447504, 674: -4.86469986699, 675: -20.1499107093, 676: -15.7913366572, 677: 38.4647368574, 678: 19.0742533589, 679: 32.9243784375, 680: -24.164597188, 681: 26.435330189, 682: -23.1757942582, 683: -4.28025487092, 684: 1.85578616872, 685: -14.1346337026, 686: -7.99159743958, 687: 12.0075677104, 688: -3.68422510372, 689: -31.5506570981, 690: -57.3819669258, 691: -2.77534625689, 692: 43.5909302835, 693: -92.1714601537, 694: 46.0885418102, 695: 100.0, 696: 40.6484982711, 697: 3.27280966815, 698: 35.2005838422, 699: 0.251913960266, 700: 4.14725322771, 701: 7.27557282159, 702: 4.42226746616, 703: 7.54395623911, 704: 8.32827209976, 705: 3.55707540371, 706: -9.35826510609, 707: -2.10175295904, 708: 13.6934514259, 709: 14.3982542424, 710: 11.9642880252, 711: 2.32765378378, 712: -10.07818492, 713: 10.9136921524, 714: -6.6783384675, 715: 16.2694770228, 716: -0.268596303084, 717: -10.6957364825, 718: -1.8616530418, 719: -4.72982055263, 720: -4.42048223002, 721: -11.0216762136, 722: 4.6138024812, 723: 8.45180566054, 724: 0.655679257161, 725: -11.9574390701, 726: -5.83460114329, 727: 8.17155211127, 728: -19.6167843139, 729: -4.73804009604, 730: 7.55986829811, 731: 11.2298713852, 732: 1.50224269862, 733: -6.88908084747, 734: 0.0657253357759, 735: -3.46081539479, 736: -5.89792088485, 737: -13.2752818798, 738: -22.861003293, 739: -7.46158023268, 740: -5.58636333419, 741: -22.9090819267, 742: 0.790873906529, 743: 3.37051266467, 744: 4.64069295941, 745: -4.86176751921, 746: 2.27751515468, 747: -2.69722289943, 748: 6.2993703258, 749: 6.68117008446, 750: -26.8061348403, 751: 1.93743628287, 752: 3.07879827111, 753: -9.17382800487, 754: 1.98083965722, 755: -3.82779353641, 756: 1.05180728098, 757: 1.41349636913, 758: 10.3843034567, 759: -27.3157567599, 760: -1.89267348841, 761: -8.0600409513, 762: -3.3329836923, 763: -7.68404649836, 764: 2.09049830744, 765: -8.16315180712, 766: 2.18045324669, 767: -9.99990151882, 768: -0.709876836074, 769: 4.46185873584, 770: -3.66130822316, 771: 0.631194165764, 772: 11.3380880873, 773: 6.51645791851, 774: -8.18688266349, 775: -9.66680484054, 776: 6.0476760257, 777: -14.2734825554, 778: -6.15126889859, 779: 4.51638275588, 780: 19.1692871505, 781: 3.33334287262, 782: -2.79770892556, 783: -1.40652667458, 784: 4.74866523942, 785: 11.2818129161, 786: -7.32129680096, 787: 2.04369082389, 788: 8.99757535944, 789: 8.00335536425, 790: 2.3622596913, 791: 4.86342374645, 792: 3.24944758063, 793: 2.95418630734, 794: 2.56557960654, 795: 6.98204637704, 796: 2.43813022094, 797: 11.6338435695, 798: 9.06902632792, 799: 10.0890240039, 800: 6.99611249958, 801: 1.05981060147, 802: 1.74523406049, 803: -0.50812612623, 804: 4.33072343562, 805: 0.186787871255, 806: -3.67055175604, 807: 4.73106247771, 808: 14.914946674, 809: 5.30193817083, 810: -2.05176800708, 811: -14.2421693575, 812: -10.4276818444, 813: -19.2869552569, 814: 8.66836975729, 815: -8.93041709009, 816: -6.35430609488, 817: -8.72204119455, 818: -23.4764585149, 819: 12.6826685919, 820: -17.5828136534, 821: -12.8315344535, 822: -7.21810532751, 823: 21.2198538031, 824: 0.876952768129, 825: 0.70402506045, 826: 2.88966319711, 827: 4.09120041107, 828: 0.424393316165, 829: 2.07791299914, 830: -2.19625625336, 831: 4.30521179459, 832: 2.13385420578, 833: -2.81125784415, 834: 5.59423252908, 835: 1.89681614693, 836: 9.69462699795, 837: 2.02130630934, 838: -6.61845668813, 839: -1.9350764093, 840: 1.13075801431, 841: 7.87344922714, 842: 0.906939532981, 843: -2.39697802171, 844: 0.382187963481, 845: 0.493258937476, 846: -3.42736084692, 847: 0.311065326689, 848: 3.61419010462, 849: -4.38316731711, 850: 0.669713900851, 851: 0.425829156137, 852: 0.540321009537, 853: 0.291906434231, 854: 2.5264762064, 855: -2.96787678636, 856: 0.139152019487, 857: 0.767067236925, 858: -0.801104755294, 859: 0.94754016387, 860: -0.237192157468, 861: -3.82429393607, 862: 0.52326521914, 863: -17.2628870656, 864: -35.5046437386, 865: 3.37541987194, 866: 3.08478572876, 867: -10.1086345467, 868: 1.42566745693, 869: 1.07750508734, 870: 0.37746119683, 871: -1.4678074214, 872: 0.552460362831, 873: -4.87941808176, 874: -16.2834134242, 875: -6.24116960516, 876: 1.88591291385, 877: 0.41848404475, 878: 0.213697404137, 879: 0.172271645575, 880: 1.4296323382, 881: 13.0054232134, 882: -47.04077332, 883: -2.7898305939, 884: 37.3284298341, 885: 14.0724327149, 886: 5.43590086805, 887: 0.971564698591, 888: -3.96248651298, 889: 3.40310848178, 890: 1.54436092236, 891: 0.374227850625, 892: -14.3890773999, 893: 6.11261543747, 894: 2.82170219491, 895: 0.591377818077, 896: 7.87607890987, 897: 0.480131475131, 898: 1.85683597744, 899: 0.501751522428, 900: -1.77565946209, 901: 0.375681278206, 902: 0.364785321869, 903: -9.86946881747, 904: 0.240768576226, 905: -4.94265198622, 906: 0.424519765011, 907: 0.436097768414, 908: 0.492593141005, 909: 0.41569178278, 910: 1.73195889402, 911: 0.962151226829, 912: 1.10603234411, 913: 0.135716138098, 914: 1.74527555468, 915: 16.4803416701, 916: -6.42550949876, 917: -2.2143410508, 918: -9.04288744809, 919: -3.30879309031, 920: -5.88139941163, 921: -1.68730822403, 922: 1.17592596241, 923: 5.3510271812, 924: 0.653669340783, 925: -12.5922302333, 926: -2.20915581934, 927: 7.68757258735, 928: -4.62261166975, 929: -10.3305971514, 930: 0.84727845485, 931: -8.78755481911, 932: -2.0847474491, 933: -14.4393685006, 934: 2.5386530605, 935: 9.29528982205, 936: -14.7351192929, 937: 1.08782363451, 938: 2.10963947121, 939: -2.76228251208, 940: 4.01567557991, 941: 17.859956166, 942: 5.10155959966, 943: 0.941080428503, 944: -3.11323406487, 945: -8.48197112129, 946: 0.691275714186, 947: -8.02784892973, 948: 1.78462749275, 949: 2.7928210958, 950: -0.678397297942, 951: -12.8805619925, 952: -0.84058068123, 953: 4.37393690164, 954: 2.42246844623, 955: 2.62543098731, 956: 2.61579641525, 957: 3.51507392728, 958: -4.91576476656, 959: 7.65467075631, 960: 2.65827948409, 961: -0.533129926778, 962: 1.39168354252, 963: 2.58992844519, 964: -2.75382028422, 965: -1.80703669908, 966: 7.12724278043, 967: -0.123411793783, 968: 3.08419394253, 969: 0.924621479323, 970: 1.37580605619, 971: 3.849328687, 972: 2.98134197778, 973: -28.0209174993, 974: 5.99531950236, 975: -3.5926212277, 976: 6.44788839523, 977: 0.290483479983, 978: 16.0927572427, 979: 3.092369422, 980: -1.75884586399, 981: -5.59855484585, 982: 15.8520237521, 983: 11.2699722029, 984: -6.61555820528, 985: 1.41985801147, 986: -1.61602081072, 987: 4.09937327239, 988: -6.73177376798, 989: 8.73314184071, 990: 20.8508420253, 991: 9.60306561928, 992: -15.9704025424, 993: -15.2019637705, 994: 6.42015226294, 995: -13.8517670966, 996: 3.19749120379, 997: 4.30794398715, 998: 3.33362352113, 999: -3.34342950867, 1000: 16.0900450296, 1001: -6.11885078326, 1002: 13.8453363716, 1003: 1.51066449329, 1004: 3.06216719462, 1005: -0.962596586885, 1006: 1.8604885491, 1007: 1.77641875407, 1008: -0.155320313715, 1009: -3.34513078941, 1010: 2.00115146556, 1011: 2.3572631445, 1012: -13.9857830287, 1013: 0.202404883934, 1014: -2.55514134622, 1015: 2.51271776718, 1016: 2.05243798204, 1017: -1.464748129, 1018: 0.802684001551, 1019: 2.31877075005, 1020: -2.02932794964, 1021: 3.14854732879, 1022: 17.4299668858, 1023: -2.20588162197, 1024: -2.84587050851, 1025: -6.08193143738, 1026: -8.49925469601, 1027: 39.5592248858, 1028: 2.84308905086, 1029: -7.09414734643, 1030: 7.59328805566, 1031: -0.828761085101, 1032: 8.97777652288, 1033: -21.0503324214, 1034: 8.88196954207, 1035: -31.5732730905, 1036: 4.05332594035, 1037: 1.0854005701, 1038: -2.04054635642, 1039: 4.13949276002, 1040: 3.9211665312, 1041: 2.78706837181, 1042: -3.28817127331, 1043: 7.88310812448, 1044: 10.0370066422, 1045: -13.3849389975, 1046: -6.56846181327, 1047: 21.5370527712, 1048: 0.598448949569, 1049: 16.1539413839, 1050: 56.3876582661, 1051: 28.6160940396, 1052: -4.53522169877, 1053: -17.51045848, 1054: 21.7555714819, 1055: -31.0110032282, 1056: 6.08768784696, 1057: -99.9084427044, 1058: -15.6025660635, 1059: -36.4639386378, 1060: 1.64151833252, 1061: 45.9205268022, 1062: -17.3747994663, 1063: -21.6009286181, 1064: 53.6653252541, 1065: -50.3456034816, 1066: 56.055406556, 1067: 7.9841935659, 1068: -25.3627116564, 1069: 24.3188358851, 1070: 36.9989349607, 1071: 20.9875523999, 1072: -2.4560233015, 1073: 25.5107441945, 1074: -100.0, 1075: -36.9825535816, 1076: -90.2635805278, 1077: -6.38814582364, 1078: 20.5869524929, 1079: 41.8942804, 1080: -22.9956288075, 1081: -26.3774333922, 1082: -8.34755342203, 1083: -62.2517905442, 1084: 7.71342304606, 1085: -5.40999491785, 1086: 41.5324136133, 1087: 51.6767630254, 1088: 60.2263293658, 1089: 0.582544837494, 1090: 1.64602643125, 1091: -12.5478175862, 1092: 20.41468573, 1093: -21.1953239552, 1094: -25.1166728336, 1095: 2.60907612466, 1096: 39.3972897522, 1097: -13.4009863477, 1098: -4.14422586586, 1099: -3.677893264, 1100: 3.24120336904, 1101: -8.27173081672, 1102: -14.0431866932, 1103: 2.86263598139, 1104: 67.2750382935, 1105: 9.11094920585, 1106: -7.48254534103, 1107: -53.8476344072, 1108: 29.5142683526, 1109: -75.7728799789, 1110: 33.1578798344, 1111: -4.85256842787, 1112: -25.9480692738, 1113: 1.60953419847, 1114: -21.9091765791, 1115: -1.47926238183, 1116: 3.21744524336, 1117: 2.69379843677, 1118: -5.44790487909, 1119: -13.6294543463, 1120: 2.61765789291, 1121: -6.69773399837, 1122: 7.9143169229, 1123: -1.4746783393, 1124: -3.00133284439, 1125: 2.40676788264, 1126: 4.13495521203, 1127: 2.58810505134, 1128: -13.3376639431, 1129: 34.2339428287, 1130: -3.02636918803, 1131: 8.06159756994, 1132: -26.1463708149, 1133: 0.134635386645, 1134: 0.307936091281, 1135: 16.7095182662, 1136: 3.17067860184, 1137: -25.0425615811, 1138: -5.16786569226, 1139: 2.85883453848, 1140: -9.8244831834, 1141: 3.03950945403, 1142: -0.0130036881522, 1143: 2.48375311358, 1144: -21.8783547777, 1145: 2.96207490085, 1146: -0.0648040376503, 1147: 2.85360520962, 1148: -45.6977185679, 1149: -1.62208069748, 1150: -51.6968095726, 1151: 4.31584500613, 1152: -1.56774715759, 1153: -24.8958995616, 1154: 2.62277065053, 1155: 2.91230109792, 1156: 1.98004537677, 1157: -61.3391085268, 1158: -16.3559566144, 1159: -33.6932419594, 1160: 81.1357374665, 1161: -24.3664644273, 1162: 23.2976652454, 1163: 29.7011902773, 1164: 29.4733903552, 1165: 14.4257105419, 1166: 15.2479691078, 1167: -36.8346425688, 1168: -31.2349204867, 1169: 51.5974778659, 1170: 44.9265273111, 1171: 19.3643047848, 1172: -25.8938678052, 1173: 0.286836081519, 1174: 0.555813344364, 1175: 2.63538429058, 1176: -18.0918943903, 1177: 0.467070796189, 1178: 2.45454946847, 1179: 1.95964418335, 1180: 21.0491209682, 1181: 4.43861258344, 1182: -1.39007938348, 1183: -28.2935666406, 1184: 3.13660279361, 1185: -55.0579843734, 1186: 6.90874739048, 1187: -0.748936245172, 1188: 5.21397811175, 1189: 15.4717413299, 1190: 64.2081838063, 1191: 3.42951379798, 1192: -24.5460773451, 1193: 0.669157488053, 1194: 0.597779838482, 1195: 21.2226608391, 1196: 0.63840343687, 1197: 2.17172500701, 1198: -3.88689420177, 1199: 1.10598660082, 1200: 0.555461384815, 1201: 0.552524257819, 1202: 0.318388688239, 1203: -4.50842348431, 1204: 41.0667542027, 1205: -4.72804944107, 1206: 0.180834404357, 1207: 1.40277369856, 1208: 0.322192931292, 1209: 2.98987103689, 1210: 47.5588817033, 1211: -14.8124527299, 1212: 5.52081502775, 1213: 8.2822164948, 1214: 1.23350926599, 1215: -8.31647349758, 1216: -4.90190628182, 1217: -4.53886736399, 1218: -3.61338338593, 1219: 0.513243236564, 1220: 7.57763039425, 1221: 0.559900804051, 1222: 12.4029512319, 1223: -26.3819422296, 1224: 41.711064377, 1225: -0.251168375542, 1226: 0.424507048447, 1227: 0.412556694677, 1228: 0.759392142299, 1229: 0.37166563018, 1230: -3.81471253754, 1231: -14.5297022003, 1232: 23.1945691843, 1233: -5.42018826093, 1234: 7.69989036192, 1235: 24.6594505668, 1236: 31.3892203202, 1237: 9.26790933994, 1238: 0.469420808422, 1239: -30.9596617045, 1240: 0.821846199517, 1241: 56.598836189, 1242: 2.87147324862, 1243: -2.63449485783, 1244: 2.24398723128, 1245: 17.5107164813, 1246: 0.439568943739, 1247: 2.42483082658, 1248: 0.410868819548, 1249: -22.0531058783, 1250: 0.623851548491, 1251: 0.450331656498, 1252: -6.11859408754, 1253: 2.30465704413, 1254: -16.3516027861, 1255: 0.62658017728, 1256: 0.456143382463, 1257: 2.46618428305, 1258: 0.417835134695, 1259: -7.91090424065, 1260: -3.55387435713, 1261: -24.9192379236, 1262: 0.173967785589, 1263: -0.12425129018, 1264: 19.3017961217, 1265: -1.56284174946, 1266: -12.2258734196, 1267: 16.8650136592, 1268: -10.4911881263, 1269: -5.91021570006, 1270: -3.53662518681, 1271: -2.58056876366, 1272: 1.71971040594, 1273: -0.326911452122, 1274: 4.84192567567, 1275: 1.17930846855, 1276: -13.3939625472, 1277: 0.751120517533, 1278: 29.9940698031, 1279: 0.482288282369, 1280: 9.40330853881, 1281: 25.360042809, 1282: 5.6487688637, 1283: -12.1387285284, 1284: -17.3229274797, 1285: -8.27036203792, 1286: 2.72810281631, 1287: 2.534879018, 1288: -14.9830045148, 1289: -24.3279341618, 1290: -52.5522008949, 1291: -29.0434437658, 1292: 12.2480491266, 1293: -31.6164649646, 1294: -99.9108698222, 1295: -13.8066859148, 1296: 6.11742354289, 1297: 3.47221137588, 1298: -1.42366372657, 1299: -64.9200784683, 1300: -12.1307068979, 1301: -37.3435336467, 1302: -5.68299482934, 1303: 4.5232646208, 1304: -6.70453967254, 1305: 2.23668848988, 1306: 2.00569901509, 1307: -7.39980596475, 1308: -0.990320772298, 1309: 1.97628708501, 1310: 0.762590022374, 1311: 3.81555559047, 1312: -4.88271966866, 1313: -27.0383994674, 1314: -48.9560064793, 1315: -12.8825561613, 1316: -11.1490318848, 1317: 0.0194076837071, 1318: 4.64178925794, 1319: 7.62281489856, 1320: 43.0480979425, 1321: 50.297236789, 1322: 18.6195326715, 1323: -55.3376565481, 1324: 16.9557777692, 1325: -24.219511595, 1326: -7.78031333352, 1327: 6.95180224932, 1328: 8.11915620896, 1329: 0.0463164279127, 1330: -26.2189968409, 1331: -44.1692327777, 1332: 31.8953528328, 1333: 18.3664173687, 1334: 2.20301913387, 1335: -6.29468417517, 1336: 0.0732409148646, 1337: 1.84811481833, 1338: -4.53759023439, 1339: -41.0580926299, 1340: 40.0306755395, 1341: 0.0612975253496, 1342: 65.8862813495, 1343: -56.2179113105, 1344: -21.0173840343, 1345: -22.2685014317, 1346: -2.25079981301, 1347: -42.1824438993, 1348: -7.12998498156, 1349: 33.4015741103, 1350: 1.77183564854, 1351: 40.8014385383, 1352: 40.0157013854, 1353: -0.231880984573, 1354: -29.6475867208, 1355: 2.55839229329, 1356: 2.37684414082, 1357: 3.83071289018, 1358: -14.6757816388, 1359: 2.61874949476, 1360: 3.35787191453, 1361: 4.36249326562, 1362: 2.31949819084, 1363: 1.87595123496, 1364: 2.3287832838, 1365: 2.48042970799, 1366: -61.8857069861, 1367: 3.728437925, 1368: -16.7586636602, 1369: -0.465715038435, 1370: -28.365463967, 1371: 6.5453692006, 1372: 15.3971854767, 1373: 22.5604100592, 1374: 39.938825037, 1375: 51.1776742038, 1376: -36.1973868761, 1377: -88.3251325795, 1378: 18.8456112986, 1379: -5.42374176177, 1380: 1.81361362683, 1381: 2.77742006685, 1382: 12.2251032203, 1383: 0.885993786396, 1384: -1.09894091081, 1385: 19.5099950078, 1386: 2.52325835605, 1387: 5.61616124963, 1388: -1.40837987309, 1389: -9.18835325684, 1390: -1.52036681453, 1391: -65.5620156461, 1392: -51.8063279642, 1393: 2.38576253047, 1394: 23.0078911443, 1395: 4.78879714529, 1396: 26.5761585247, 1397: -0.209841051394, 1398: 11.2697683203, 1399: 3.74352562519, 1400: -7.38280076352, 1401: -10.8848629672, 1402: 0.997928181272, 1403: 2.41583899608, 1404: 6.61467317043, 1405: -2.82818636437, 1406: 18.1820963528, 1407: 6.70237588568, 1408: -7.96323771667, 1409: 1.24105844756, 1410: -0.804806678294, 1411: -5.7086617118, 1412: 1.4012876355, 1413: -9.78348416786, 1414: -14.2083475921, 1415: 11.2481774791, 1416: -11.5961334178, 1417: -14.5241712814, 1418: 3.05920672456, 1419: -23.3155359447, 1420: 18.0353468352, 1421: -2.89244296835, 1422: 15.0627539492, 1423: -10.3128971584, 1424: -15.0478865928, 1425: -14.9253185641, 1426: 7.62115136624, 1427: -2.30946268022, 1428: 15.2919746359, 1429: -8.59586306371, 1430: -1.0024351296, 1431: 7.52348199827, 1432: 10.0640209378, 1433: -3.04170488413, 1434: -1.82824370017, 1435: -5.35651422047, 1436: -3.18917613475, 1437: -12.1687060591, 1438: -0.883873916646, 1439: -3.22262311202, 1440: 6.60828989511, 1441: -2.19271051304, 1442: -1.89559423158, 1443: 11.8285264361, 1444: -1.71421706323, 1445: -8.10010288179, 1446: -9.58618080257, 1447: 3.17323343024, 1448: -2.09744336678, 1449: -1.1629541764, 1450: -4.40338121469, 1451: -10.8973779203, 1452: -0.794512841601, 1453: 28.7265890074, 1454: -0.189194110021, 1455: 7.4209829185, 1456: -21.6118074193, 1457: -1.20691180863, 1458: 8.58577815958, 1459: -16.0692698816, 1460: 3.16833651768, 1461: 10.834924771, 1462: -1.32370908194, 1463: -17.2789286429, 1464: -1.49501237111, 1465: 2.1059126651, 1466: -1.44730218896, 1467: -1.35081303411, 1468: 21.8064520811, 1469: -1.09399878297, 1470: -16.5733972285, 1471: -1.53713884735, 1472: -0.0959437839312, 1473: -4.69987276387, 1474: -4.90421661424, 1475: -0.162850814335, 1476: -20.9156523297, 1477: -1.06209847958, 1478: 70.7722517201, 1479: -2.57085064677, 1480: -5.6687631931, 1481: 0.0571173554738, 1482: -11.6641893947, 1483: 13.1858366834, 1484: 5.28049000377, 1485: -1.02624921724, 1486: -6.86292114368, 1487: -1.22633520091, 1488: -1.07612323527, 1489: 16.1595584061, 1490: -1.34835469151, 1491: -2.38293515797, 1492: 1.77548982932, 1493: 7.43932405196, 1494: -1.40144097513, 1495: -0.805142573307, 1496: -1.62214017236, 1497: 9.88584236329, 1498: -1.1846501576, 1499: 1.8072968807, 1500: -0.768637408354, 1501: -6.47577902971, 1502: 0.0203993621714, 1503: -1.16922619337, 1504: -1.15420281466, 1505: -1.15126575524, 1506: 2.09275076598, 1507: -1.51018737217, 1508: 18.0187622529, 1509: 10.1208961687, 1510: -17.6643699523, 1511: -16.2721051667, 1512: 9.98605324966, 1513: 8.59167064994, 1514: 2.2516382216, 1515: -9.3196860736, 1516: -8.38732408414, 1517: -2.95934229307, 1518: -12.1150261567, 1519: -14.4063208719, 1520: 1.22599946471, 1521: -8.30104936148, 1522: -0.126562279966, 1523: -0.207334387769, 1524: 9.13212094684, 1525: 1.01618786306, 1526: -0.177859892767, 1527: 8.69173640032, 1528: -2.08433659575, 1529: -9.40241307461, 1530: 0.508268578624, 1531: 5.72106863059, 1532: -7.2578020314, 1533: -0.165977821907, 1534: 6.63700667515, 1535: -3.47466523008, 1536: 4.99280428481, 1537: -10.1509285523, 1538: 11.3257123588, 1539: -2.2768220266, 1540: -0.0900660212092, 1541: -7.21400720438, 1542: -0.0219230898113, 1543: -0.393064159588, 1544: -12.7830231015, 1545: -0.195363274227, 1546: -1.40305414218, 1547: 4.38347163126, 1548: -0.207737610216, 1549: -0.262552529001, 1550: -0.342450135129, 1551: -0.0568297079104, 1552: -3.29091240375, 1553: 12.8874231323, 1554: 0.719350931118, 1555: -0.450000114477, 1556: 0.907582565153, 1557: -12.7022770102, 1558: 3.410463106, 1559: 16.6223447221, 1560: -0.474009855816, 1561: 3.48774604375, 1562: 4.98652831503, 1563: 1.74111230888, 1564: 2.72662996691, 1565: 6.58930002189, 1566: 5.78775245792, 1567: -1.19293629434, 1568: -0.330369771087, 1569: -23.8855601853, 1570: -0.192531415686, 1571: -5.86171238262, 1572: -34.9791945819, 1573: 7.362086153, 1574: 2.37070484064, 1575: -0.29580517588, 1576: -0.21432140475, 1577: -0.241492361462, 1578: -0.284125773239, 1579: -2.24271561243, 1580: -16.688218521, 1581: -4.25090242202, 1582: 8.94529783293, 1583: 5.1841102478, 1584: -0.846278658919, 1585: 12.2724632451, 1586: -3.889717477, 1587: -9.32606025126, 1588: 0.828347116168, 1589: -0.327444260986, 1590: 2.66111124811, 1591: -0.0791619682278, 1592: -1.69690482001, 1593: -7.10754000368, 1594: -7.55221022355, 1595: -0.280177777806, 1596: -1.10626272929, 1597: -0.309137746303, 1598: -2.79955029319, 1599: -0.305339189862, 1600: -0.291192573085, 1601: -11.2059799747, 1602: -1.1104046036, 1603: -0.856168418644, 1604: -0.191747193463, 1605: -0.278760725587, 1606: 0.0520410291372, 1607: -0.264198097237, 1608: 7.18825135023, 1609: 2.59400850936, 1610: -0.549772977431, 1611: -0.512845845097, 1612: -0.490621450607, 1613: -34.2721129386, 1614: 6.73771432162, 1615: -2.55257930148, 1616: -15.6401911674, 1617: -5.60182976331, 1618: 6.41814874836, 1619: -7.62254344059, 1620: -0.939939634585, 1621: -8.55760516235, 1622: 0.0286555987886, 1623: 25.6274164803, 1624: -0.0180283512028, 1625: 1.16357879519, 1626: 1.64234595467, 1627: 16.6979782589, 1628: -0.385093735894, 1629: 25.4700056791, 1630: -13.1029720682, 1631: -3.69100935241, 1632: 13.2424911504, 1633: -14.141733873, 1634: -37.5081450298, 1635: -0.855117239046, 1636: 8.54688748909, 1637: 17.34413617, 1638: -7.09539758718, 1639: -5.98567127923, 1640: -14.2475012613, 1641: 3.1129901153, 1642: -27.6246229617, 1643: 23.0325945314, 1644: 23.9877836518, 1645: -20.7306401495, 1646: 15.9892923823, 1647: -16.6579475828, 1648: 2.17903536356, 1649: 10.4081713647, 1650: 3.39861813714, 1651: -5.81503884975, 1652: 1.18753925012, 1653: 7.04301823174, 1654: -1.25048719334, 1655: -1.35984482253, 1656: 2.93760486247, 1657: 5.72688465032, 1658: -1.03441931028, 1659: 4.36087795834, 1660: -3.7922161721, 1661: -3.16824923363, 1662: -21.0023624936, 1663: -0.512248331723, 1664: 10.979358642, 1665: -3.28542506184, 1666: -7.78851610596, 1667: 2.64739434058, 1668: 10.5896497577, 1669: 25.1062892658, 1670: -29.5080150142, 1671: 5.09525041688, 1672: 11.1156878028, 1673: -13.3807616855, 1674: 12.5856371088, 1675: 9.40232829297, 1676: -1.10767713385, 1677: 2.18318483661, 1678: 9.24334375896, 1679: -3.8433409644, 1680: 12.2180919961, 1681: 17.8150290138, 1682: 4.23556785606, 1683: -1.24576746489, 1684: -9.4092415779, 1685: -1.4684478268, 1686: -0.457333158689, 1687: -1.58065290314, 1688: 4.72159223789, 1689: -12.2267453827, 1690: 17.1675132572, 1691: -9.11802862691, 1692: 0.397078241461, 1693: 2.03015272589, 1694: -15.9187285477, 1695: -4.58008309406, 1696: -12.8635801378, 1697: 2.55923886805, 1698: -4.97788761258, 1699: 0.540129394088, 1700: 12.1609135587, 1701: -1.55008519622, 1702: -2.56176120881, 1703: 31.9185296264, 1704: -1.10902230596, 1705: -1.15124597775, 1706: 2.67921562721, 1707: -23.1247666332, 1708: -1.16357714691, 1709: -0.89600755212, 1710: 12.0836866015, 1711: -1.17590505312, 1712: -5.98227407088, 1713: -1.24050648195, 1714: -1.03848921613, 1715: -7.63801044575, 1716: -0.898074361028, 1717: 11.3296841062, 1718: -12.583017819, 1719: -15.1048968176, 1720: 15.4872549339, 1721: -16.823255411, 1722: 2.44898506378, 1723: -1.79636387526, 1724: -40.1773828758, 1725: 34.8520554696, 1726: 2.20362837468, 1727: -3.07099071208, 1728: -4.87775007093, 1729: -0.826402776558, 1730: -2.20742989717, 1731: 2.28128141433, 1732: 16.4005742699, 1733: 11.7492908585, 1734: -15.7257989347, 1735: -1.23794490422, 1736: 1.4133157758, 1737: -0.336119876912, 1738: 8.33688949028, 1739: -7.76729957252, 1740: 5.2589262829, 1741: 15.3502133655, 1742: 8.93625719675, 1743: 6.92673794987, 1744: 24.2934250944, 1745: 9.57187167561, 1746: -0.24040812072, 1747: 11.1905163943, 1748: -11.1092101366, 1749: -6.50747182597, 1750: 38.9761830584, 1751: -4.2955800836, 1752: -3.2487713534, 1753: 11.1382726761, 1754: -10.250037166, 1755: 25.6507659687, 1756: 20.8140319466, 1757: 15.1287110782, 1758: -23.5023028998, 1759: 15.9004023852, 1760: 24.8294656123, 1761: -1.90429224097, 1762: 30.6760492543, 1763: 5.89236776349, 1764: 100.0, 1765: -8.12641704832, 1766: -26.9178823256, 1767: 84.0183723231, 1768: -8.14923346214, 1769: 15.8547425865, 1770: 27.0327484737, 1771: -17.3856836603, 1772: -3.38079065518, 1773: -18.114287799, 1774: 6.92545814863, 1775: -15.9956996742, 1776: -7.00502703675, 1777: -41.6961671522, 1778: 14.7983171998, 1779: -4.18923026379, 1780: 11.7876337042, 1781: -0.99577281547, 1782: 1.87674498047, 1783: 5.81725378109, 1784: -6.37903988513, 1785: -27.5519505964, 1786: 2.20076320523, 1787: 1.79931469499, 1788: 1.91761013441, 1789: -37.224949651, 1790: 17.1143501026, 1791: -0.186183229755, 1792: -12.8189318035, 1793: -2.13682218359, 1794: 27.7034995986, 1795: -2.90712463138, 1796: 14.1416928856, 1797: 0.933350801915, 1798: -2.60120057351, 1799: -2.21879255166, 1800: 13.5202668236, 1801: -2.69877622987, 1802: -2.0248483223, 1803: -21.2359756958, 1804: 46.2870815897, 1805: -21.9068068851, 1806: 0.55750716158, 1807: -2.33690094566, 1808: 0.653092296026, 1809: 10.7598758015, 1810: 5.49743972145, 1811: -2.57350150581, 1812: 11.428379221, 1813: -2.71103707848, 1814: -14.5956062063, 1815: -2.61292435191, 1816: -2.5521489963, 1817: 3.8722162544, 1818: -2.70514260557, 1819: -10.0067254116, 1820: 4.20356294347, 1821: -2.57291298052, 1822: 1.24497534284, 1823: 3.04066049483, 1824: -1.5452576575, 1825: 13.715288755, 1826: 3.44368208761, 1827: 16.4503668295, 1828: -2.65739635756, 1829: 1.71753585726, 1830: -1.63313240339, 1831: 6.78961899042, 1832: 2.92535912153, 1833: 49.3456562861, 1834: -4.89885079803, 1835: 37.0233746201, 1836: -1.16257567077, 1837: -2.46900610953, 1838: -14.5292965768, 1839: 25.101896109, 1840: -2.52087878296, 1841: -1.29823950625, 1842: -18.9424657815, 1843: -2.71556803259, 1844: 12.7309524081, 1845: -2.3673662794, 1846: -11.4770684196, 1847: -2.55254587573, 1848: 4.55603742926, 1849: -2.50828709833, 1850: 0.626001089415, 1851: 53.0347232094, 1852: 4.5393107184, 1853: -2.52421594602, 1854: -4.73709414203, 1855: 3.79327705038, 1856: 24.9073352435, 1857: 7.24970384879, 1858: -84.2349237649, 1859: -7.43230889629, 1860: 52.7180022962, 1861: 26.2220066726, 1862: 16.1314161793, 1863: -6.48634693209, 1864: -23.7445000106, 1865: 15.7757591288, 1866: -34.7966153042, 1867: -35.732792636, 1868: -8.60252564915, 1869: -14.7695378762, 1870: 10.1593034739, 1871: -0.362547881432, 1872: -0.548466766638, 1873: -0.476501906442, 1874: -27.6162544856, 1875: -0.378426251554, 1876: -2.82237404312, 1877: -6.25012665953, 1878: 10.8707189624, 1879: 0.546981755044, 1880: -25.0939569191, 1881: 18.7681070111, 1882: -1.02183906443, 1883: -6.96915621949, 1884: 0.373533082434, 1885: 5.19052595642, 1886: 15.8817241139, 1887: -10.4310462608, 1888: 30.0092425342, 1889: -0.986466820523, 1890: -5.59393267944, 1891: -0.454026302657, 1892: -0.440312308136, 1893: 0.793818897684, 1894: -0.370555333086, 1895: -0.378590208705, 1896: 15.8725834838, 1897: -1.17950128515, 1898: -0.446006631479, 1899: -0.518955989467, 1900: -0.493368916901, 1901: 1.38186203317, 1902: -26.2928247685, 1903: -1.76986209029, 1904: -0.0951879628209, 1905: -0.728096137415, 1906: -7.3687690801, 1907: 2.879842425, 1908: -42.9158543032, 1909: -1.10479468168, 1910: -12.4996306503, 1911: -44.5233058055, 1912: 2.94956424961, 1913: -5.73305328041, 1914: 13.599124138, 1915: 4.063098424, 1916: -11.0203302354, 1917: -0.737586489137, 1918: -9.87588489175, 1919: -0.435581441518, 1920: -24.6602462617, 1921: 14.4412951064, 1922: 12.2544932664, 1923: -2.11933455203, 1924: -0.435137716424, 1925: -0.68046906535, 1926: -0.353703299297, 1927: -0.725791020596, 1928: 2.11280778392, 1929: 72.0343723851, 1930: -5.67710249363, 1931: -6.11214658394, 1932: 1.75339807417, 1933: 30.0116553653, 1934: -19.2592281909, 1935: 4.06934827941, 1936: -1.28084785639, 1937: 9.92007707815, 1938: -0.628883862378, 1939: 7.85145536877, 1940: -2.21347884223, 1941: 4.23789219766, 1942: -20.7809190756, 1943: 0.668783103094, 1944: -0.437014547298, 1945: -2.64198706077, 1946: -0.393435290381, 1947: 1.61164882968, 1948: -0.497787601194, 1949: -0.550205247208, 1950: -6.46291343323, 1951: -2.65573004461, 1952: -23.448259557, 1953: -0.324950935573, 1954: -0.487583985211, 1955: -5.84656495151, 1956: -1.40159794895, 1957: -3.03879826177, 1958: -0.0634390400538, 1959: -6.16521882731, 1960: -0.632267937901, 1961: -0.417809907281, 1962: 35.6382843427, 1963: 1.06813667721, 1964: 17.0583399983, 1965: 17.1244905697, 1966: 10.759941896, 1967: 4.42694773413, 1968: -7.68527421209, 1969: -2.5573775838, 1970: -7.30623903728, 1971: -1.26949815185, 1972: -11.6746376311, 1973: -5.49625753477, 1974: -2.28559646286, 1975: 11.7746475598, 1976: -34.8763007842, 1977: -0.543407902399, 1978: 20.3907003622, 1979: 11.7332855577, 1980: -3.41990496616, 1981: 22.0952052635, 1982: 15.3969242096, 1983: 19.6241825848, 1984: -2.81840223343, 1985: -2.86402494116, 1986: -5.43088854095, 1987: -7.95737220008, 1988: -2.80870431153, 1989: -3.38131233141, 1990: -11.6334767864, 1991: -24.5392155587, 1992: 10.2219827168, 1993: 4.34984333679, 1994: 1.78771567048, 1995: -2.79921608158, 1996: -7.16397351698, 1997: -17.354484604, 1998: -36.8141469376, 1999: -28.1270588297, 2000: -2.71611357261, 2001: -2.57956980364, 2002: -8.53937241385, 2003: -2.56415144172, 2004: -0.548585229104, 2005: 16.2070813405, 2006: 18.2118056404, 2007: -2.60310435867, 2008: 7.56116065411, 2009: -1.84000270172, 2010: 1.55591084464, 2011: 23.4572642091, 2012: 37.6330720098, 2013: 0.20857311813, 2014: 2.70336592241, 2015: -10.7240798309, 2016: 1.3207056877, 2017: 32.4097631076, 2018: -21.4738870453, 2019: -20.5582092494, 2020: -26.8487265537, 2021: 17.3679897219, 2022: 37.9527535703, 2023: -6.03592465749, 2024: 6.51397215237, 2025: -11.8445663821, 2026: 12.0104027859, 2027: 15.1941103914, 2028: -4.08264090286, 2029: 34.8648951817, 2030: 27.4381657434, 2031: 15.9753681414, 2032: -3.34751448321, 2033: 0.571733510433, 2034: -2.47770317373, 2035: 2.50395512243, 2036: 5.90931532257, 2037: -9.36685270521, 2038: 15.8835248793, 2039: 28.8181778398, 2040: -20.1210602991, 2041: -8.63549518996, 2042: -33.9656651569, 2043: 7.29310404611, 2044: -1.24601085545, 2045: -0.633901465867, 2046: -3.6497563751, 2047: -2.90573362362, 2048: -2.6132149295, 2049: -45.889659088, 2050: -10.6034375159, 2051: -2.48925734414, 2052: 18.6630887658, 2053: -2.47435175451, 2054: -2.66298578631, 2055: -4.36228120418, 2056: -4.9875840614, 2057: -2.81224940019, 2058: -2.35054749571, 2059: 11.7470536834, 2060: -2.56024211214, 2061: 5.65982860584, 2062: -2.48135005417, 2063: -2.62265727262, 2064: 17.4376618283, 2065: 4.09066736824, 2066: -3.38538988036, 2067: -0.394843296854, 2068: 1.09354479619, 2069: 14.5322878338, 2070: -6.13039891018, 2071: -42.0685911535, 2072: 11.0424767868, 2073: 27.0462907326, 2074: -23.7984549745, 2075: 10.6874099928, 2076: -27.8118541213, 2077: 2.60867999966, 2078: -2.39703258715, 2079: 7.3894661693, 2080: -7.44682121798, 2081: -15.4921671244, 2082: -27.2878875985, 2083: 28.4614771586, 2084: -2.80774856892, 2085: -15.7953510593, 2086: 11.3982662766, 2087: 3.45863465933, 2088: -4.07013764462, 2089: -8.14181484855, 2090: -9.91858500006, 2091: 8.90563017387, 2092: -1.24731359238, 2093: -1.39024033877, 2094: -11.7907045185, 2095: 0.51916724172, 2096: 0.879825424573, 2097: -4.67110022325, 2098: -3.96163362252, 2099: -5.23077805825, 2100: -4.60953426147, 2101: -1.88590000545, 2102: 6.1460779807, 2103: -1.17942112829, 2104: -13.0459879823, 2105: -12.1225580046, 2106: 1.66055509505, 2107: 5.92931302539, 2108: 7.8290089385, 2109: -1.27323928982, 2110: 3.56861554239, 2111: -15.131020988, 2112: 3.90108491545, 2113: -0.107028370316, 2114: 0.006256950183, 2115: 8.76030050167, 2116: -1.23778917465, 2117: -18.3487774137, 2118: 1.38815044677, 2119: -4.28744543158, 2120: -15.1195786766, 2121: -2.10954304698, 2122: -2.17106464276, 2123: 1.54908104463, 2124: -3.19912948352, 2125: -2.87657815059, 2126: -5.35039620711, 2127: 3.99822926213, 2128: -1.37657036496, 2129: -2.17477512382, 2130: -0.643259200387, 2131: 0.653194479544, 2132: -3.1833885737, 2133: -3.0549229761, 2134: 2.64040900451, 2135: 5.94338802503, 2136: 3.61181533331, 2137: -9.57664401204, 2138: -6.78318772791, 2139: 4.324837312, 2140: -0.482662095636, 2141: -11.9002211234, 2142: 2.61701058685, 2143: -0.0605966847799, 2144: -0.825016151361, 2145: -8.71916667906, 2146: -3.99400534581, 2147: 4.28144432916, 2148: -3.59131379505, 2149: -6.46704361443, 2150: 1.35871439037, 2151: 13.4174375892, 2152: 1.71056330062, 2153: 18.4943102564, 2154: -7.9558914628, 2155: 5.87848568904, 2156: -11.6586099126, 2157: 3.97345379838, 2158: -0.0503915385062, 2159: 3.6127862586, 2160: 0.409753202209, 2161: -0.931787973387, 2162: 3.75533953422, 2163: 0.442302250025, 2164: 4.13340325597, 2165: 3.89679297933, 2166: -1.64278036605, 2167: 4.289336725, 2168: -10.8240647817, 2169: -3.59835926867, 2170: 6.74109845484, 2171: -0.272286500562, 2172: 8.73123804638, 2173: 7.42382020589, 2174: 5.89600372542, 2175: 1.27618611502, 2176: -5.91764375038, 2177: -2.50979900215, 2178: 1.62673886612, 2179: 2.66738974463, 2180: 4.29106158438, 2181: 2.6320006724, 2182: 4.2676641089, 2183: 4.26278527955, 2184: 2.65985726159, 2185: 4.68178659408, 2186: 4.75897657571, 2187: -4.38657187762, 2188: 3.91156954588, 2189: 2.25069801905, 2190: 3.77264321542, 2191: 1.75094920013, 2192: 3.92664004949, 2193: 4.33634909113, 2194: 0.0409903869203, 2195: 1.44878533606, 2196: 4.81457623291, 2197: 1.1286725087, 2198: 0.893185238044, 2199: 1.26212051884, 2200: -6.18055123945, 2201: 3.96626682197, 2202: 4.16573731873, 2203: 4.18301810895, 2204: -3.89179893603, 2205: 2.3771345036, 2206: -6.74089023422, 2207: 5.10214752347, 2208: 0.721412183818, 2209: -2.78256149566, 2210: -5.09778970823, 2211: -5.13288415713, 2212: -1.68641792449, 2213: 10.9020933505, 2214: 10.9341482911, 2215: 5.98595706504, 2216: -9.72052040881, 2217: -0.108098062325, 2218: -4.539593597, 2219: -7.73604926136, 2220: 0.680534495395, 2221: 1.1235053207, 2222: 0.942454680905, 2223: -0.920116432302, 2224: 0.672051740804, 2225: 3.70782661665, 2226: -0.137512941662, 2227: -8.67286735644, 2228: 1.06626175236, 2229: -1.19508532414, 2230: -7.18212589309, 2231: 1.09631115878, 2232: 8.62071566051, 2233: 2.13584877002, 2234: 1.3303395593, 2235: -1.11213271737, 2236: 4.28141372094, 2237: 6.82008924548, 2238: 0.901610961419, 2239: 5.42429993882, 2240: 0.827111177415, 2241: 0.815423261072, 2242: -1.08014348322, 2243: 0.683786299859, 2244: 3.24456858823, 2245: 3.6074180745, 2246: 0.945322752217, 2247: 0.845795907144, 2248: 0.907524331441, 2249: 0.831853946838, 2250: 2.07400695594, 2251: -14.5111422404, 2252: -0.28940436926, 2253: 1.19831717556, 2254: 1.85017580258, 2255: 0.496576916415, 2256: 1.94851577245, 2257: -27.5574890666, 2258: -2.48680422394, 2259: -17.4856062151, 2260: -21.9518332102, 2261: -0.884838565727, 2262: -3.0852294053, 2263: -2.43239965422, 2264: 1.74024579923, 2265: -2.93889429288, 2266: 0.884497486363, 2267: 2.47036903065, 2268: 0.600111511554, 2269: 6.85744098412, 2270: 1.53059033094, 2271: -15.2600849582, 2272: 0.275269559417, 2273: 0.743335491876, 2274: 1.32178473423, 2275: 0.847760532856, 2276: 0.703156740218, 2277: -4.15676403786, 2278: -5.13536703915, 2279: 0.00399364681947, 2280: 4.73334111271, 2281: 0.240815249339, 2282: -0.318839966667, 2283: -1.14145584067, 2284: 0.500850789695, 2285: 7.88635040113, 2286: -6.95775707669, 2287: 0.654170564435, 2288: 6.01536608202, 2289: -0.223082738088, 2290: 0.161572678532, 2291: 1.82806478443, 2292: 4.93605386102, 2293: 0.679174386182, 2294: 3.68908350048, 2295: 0.594391071428, 2296: -9.61282015221, 2297: 0.675634558324, 2298: 0.869420314807, 2299: -3.01194290424, 2300: 4.22743874523, 2301: -18.9761986835, 2302: 0.636655348234, 2303: 0.566143330739, 2304: 0.0181248981272, 2305: 0.7405339309, 2306: 3.04885561406, 2307: 1.12281897209, 2308: -8.630493513, 2309: -0.0062309913366, 2310: 0.776747169093, 2311: 16.5287663088, 2312: -3.15875375602, 2313: 12.4268842266, 2314: -1.53648716356, 2315: -2.32345222623, 2316: -5.98087033111, 2317: -1.87212643229, 2318: 1.14313638345, 2319: -2.37099552659, 2320: 1.24147766322, 2321: -7.37781627358, 2322: 3.01500329817, 2323: 1.63458879784, 2324: 0.510688666246, 2325: -5.96987153378, 2326: 0.899449479726, 2327: -5.75463568783, 2328: 3.72609245869, 2329: 1.52349737301, 2330: 12.1653769083, 2331: -12.3380371552, 2332: 15.7286434312, 2333: 3.41642121158, 2334: 3.90922405021, 2335: -8.17768259515, 2336: 4.40125531454, 2337: -0.41616944528, 2338: 1.55657563896, 2339: -0.425252955138, 2340: 18.9287316325, 2341: 3.0079348234, 2342: 1.43399356464, 2343: 7.09048617061, 2344: -0.25954993788, 2345: -6.36718164719, 2346: -4.97077863483, 2347: 4.72031155525, 2348: -2.84993225542, 2349: -0.615804665615, 2350: -0.0499403723921, 2351: -3.37631794831, 2352: 3.646698915, 2353: 3.11525508321, 2354: 1.59484628351, 2355: -2.56739658848, 2356: 3.03779014159, 2357: 1.41046606088, 2358: 5.59641510702, 2359: 2.18114830239, 2360: 14.8584890377, 2361: -2.11989258418, 2362: 1.95663942693, 2363: -15.6616958582, 2364: 0.497695377754, 2365: 2.73467266165, 2366: 0.905883365619, 2367: -5.84076376616, 2368: 12.5784901404, 2369: 13.4725478793, 2370: 0.737337878803, 2371: 8.80447838169, 2372: -0.204872405756, 2373: 1.86374512815, 2374: 3.44158126578, 2375: -1.72762844632, 2376: -3.90549962969, 2377: 1.19955207597, 2378: -3.42562770483, 2379: -17.9030352241, 2380: 22.1420225019, 2381: 3.64997120268, 2382: -0.519095495645, 2383: 3.2418877193, 2384: 1.13041636245, 2385: 9.65642145507, 2386: -4.887101307, 2387: 0.667491744717, 2388: -9.83531379948, 2389: 0.332481867064, 2390: 4.69387690498, 2391: 0.947741010701, 2392: 3.89249330732, 2393: 4.02975919057, 2394: 14.4390177105, 2395: 0.861122064142, 2396: -8.32665789124, 2397: 1.15773630278, 2398: -13.9829603805, 2399: -7.95855758698, 2400: 2.12247393219, 2401: -24.1630279465, 2402: 3.65763273497, 2403: 3.76867184172, 2404: 3.5854430412, 2405: -18.4545888221, 2406: 4.80639354802, 2407: 1.38348658669, 2408: 0.863524059352, 2409: 4.35982227601, 2410: -28.0010680421, 2411: 3.7793684688, 2412: 4.51048673181, 2413: 23.4262908778, 2414: 3.88087912033, 2415: 3.04771152425, 2416: -5.55617958062, 2417: 5.92006209836, 2418: 3.71238588575, 2419: -1.07138326781, 2420: -28.997853093, 2421: -5.37171851737, 2422: 6.14396958922, 2423: 0.991529560983, 2424: -10.4690355073, 2425: -19.7413413802, 2426: -7.21076020592, 2427: -3.50577208344, 2428: -4.7615294012, 2429: -0.233095392041, 2430: 2.88369986143, 2431: -14.5587611261, 2432: 8.49923022188, 2433: 4.19979875046, 2434: -8.8446202751, 2435: 4.624172274, 2436: -0.214306451157, 2437: -2.5368672155, 2438: -7.25256235633, 2439: -13.9252871363, 2440: 12.853296543, 2441: -3.20547317277, 2442: 4.11693875256, 2443: 0.624349392441, 2444: 0.0168796234989, 2445: -4.82063047503, 2446: 4.86941001404, 2447: -3.09809384275, 2448: 7.56107806237, 2449: 1.06972921296, 2450: -1.72679206987, 2451: -1.76596036757, 2452: 0.503505383031, 2453: 5.50809266459, 2454: 6.1108424383, 2455: 5.50605038462, 2456: 0.996237410964, 2457: -3.47944369741, 2458: 4.86125879132, 2459: -2.95549700998, 2460: 8.02591843824, 2461: 0.689003525701, 2462: -2.4063734671, 2463: 0.210668650927, 2464: -2.0249519444, 2465: 4.00443883493, 2466: -2.72077201076, 2467: 0.449056175924, 2468: 4.05210039679, 2469: -3.02783168916, 2470: -5.58788710187, 2471: -1.94595651268, 2472: 2.11270206365, 2473: -6.55077364047, 2474: -0.175808561385, 2475: -0.752668680644, 2476: 1.92815585659, 2477: -3.27807899125, 2478: -3.53242347284, 2479: -1.17075686637, 2480: -0.225440239815, 2481: 1.14983003386, 2482: -3.66307831205, 2483: -4.2438292748, 2484: 0.253435376055, 2485: 0.886034773374, 2486: 3.61076925989, 2487: -3.31105119793, 2488: -0.3523785812, 2489: 1.22075446992, 2490: 4.64038790643, 2491: 1.31556344355, 2492: 0.909306809046, 2493: 1.36673480043, 2494: 3.73466660586, 2495: 2.99096859279, 2496: 0.432029819936, 2497: 1.61742876247, 2498: -0.530128618717, 2499: -1.15145938158, 2500: -12.9759756842, 2501: 0.523721425076, 2502: -4.90529141606, 2503: -0.494495366813, 2504: -5.94380270824, 2505: -1.86261658517, 2506: 0.0981771897729, 2507: -2.30335356297, 2508: -0.146378684249, 2509: -0.338320697706, 2510: 0.774955380245, 2511: -3.24127445956, 2512: -3.42524202952, 2513: 0.260866277326, 2514: -1.64550381935, 2515: 1.38654281875, 2516: 0.45329430915, 2517: -0.334907070581, 2518: 3.87217472609, 2519: 0.315210650887, 2520: 4.57104762078, 2521: -1.28569494232, 2522: 2.36472280306, 2523: 0.442469340519, 2524: 3.08411329482, 2525: 10.5197654812, 2526: -1.36117553553, 2527: 0.543755673737, 2528: 0.486374807313, 2529: -6.40129202046, 2530: -2.63700098837, 2531: -2.56152608258, 2532: 0.227157366732, 2533: 2.59664641523, 2534: 7.09685470143, 2535: -0.805060132897, 2536: -0.114413049302, 2537: -4.82669878845, 2538: -1.39239103247, 2539: -0.75624384324, 2540: -0.427310609538, 2541: -0.603545114371, 2542: 0.163334553957, 2543: 0.426840040808, 2544: 5.18281478305, 2545: 1.94928707012, 2546: -0.272277839421, 2547: 0.176813549067, 2548: -1.18498303294, 2549: 2.71216266531, 2550: 0.085123287005, 2551: 0.409975784215, 2552: -0.675728193774, 2553: 7.57832009268, 2554: 1.4052916672, 2555: -0.67744161333, 2556: -10.640094714, 2557: 0.500795858147, 2558: -7.730926722, 2559: -0.739430220399, 2560: 0.00193795612325, 2561: 2.13168331809, 2562: -8.62630050028, 2563: 2.16257902493, 2564: 2.68218392844, 2565: 0.908635812483, 2566: 2.52273211592, 2567: -1.52016055213, 2568: 4.89082201927, 2569: 0.10514708938, 2570: -0.123457287595, 2571: -1.02636136303, 2572: 7.03123701015, 2573: -0.0988309509107, 2574: -0.488322675868, 2575: -0.134347441936, 2576: 2.46156187968, 2577: 0.0163155342266, 2578: -0.481288274907, 2579: 3.89352476006, 2580: 0.0235620866179, 2581: -4.04990839044, 2582: 0.888523909562, 2583: 1.13921462487, 2584: -1.60977826051, 2585: -3.67343055178, 2586: 2.83322793912, 2587: -0.216631929935, 2588: -0.559070083329, 2589: 0.0820186156334, 2590: -0.0353992904233, 2591: 0.866717328342, 2592: -0.0403339870578, 2593: -2.67089344621, 2594: 0.522968720706, 2595: 0.0697426185444, 2596: 0.00131029067814, 2597: 0.0434566047424, 2598: -0.0868194732145, 2599: -1.3055953094, 2600: -2.69026014213, 2601: 0.283528589714, 2602: -0.0262507614554, 2603: 0.3965192408, 2604: -2.59847290006, 2605: 1.27898622033, 2606: -0.925665348015, 2607: -1.24969210115, 2608: 1.66925511191, 2609: 11.7952705753, 2610: 1.32607214335, 2611: -2.57029874669, 2612: 2.24134173594, 2613: -1.78718049621, 2614: 6.8895881353, 2615: 0.0200418092673, 2616: 2.84708348258, 2617: -0.0510201604203, 2618: -6.44986708767, 2619: 5.68833337049, 2620: -5.76372276415, 2621: -0.113626664952, 2622: -0.00866039225083, 2623: 0.0691376816808, 2624: 0.0858076890119, 2625: -0.215740209583, 2626: -4.14914755147, 2627: 5.41609606001, 2628: 0.610313229925, 2629: 8.22986722393, 2630: 1.25524809268, 2631: 0.737894680026, 2632: 2.92207620897, 2633: -0.69443426411, 2634: 1.80525718916, 2635: 5.6210536829, 2636: -0.143329519, 2637: -7.56896120157, 2638: 0.667005198859, 2639: -0.840764919672, 2640: -0.188911449023, 2641: 0.280486916671, 2642: -0.100280184611, 2643: -0.199275760166, 2644: -0.0783075649006, 2645: -0.0374333507223, 2646: -0.0908145960966, 2647: -0.127814545862, 2648: 0.966271528707, 2649: -0.228771681103, 2650: -3.17471623504, 2651: 0.00659172679253, 2652: 0.0766526589197, 2653: -0.0303312215621, 2654: 0.096973725484, 2655: -0.509747280712, 2656: 0.506223967427, 2657: 1.52866213623, 2658: 0.240300463474, 2659: 0.668244604659, 2660: -4.29604041559, 2661: -1.0430085499, 2662: -0.900407632712, 2663: -3.0876335759, 2664: 0.2598099208, 2665: -3.03749251841, 2666: -2.39683439535, 2667: -5.42228283556, 2668: 0.955542337413, 2669: 0.474025798482, 2670: -0.218725868944, 2671: -1.23251962039, 2672: -0.160301282701, 2673: -0.219298191549, 2674: -3.36918599699, 2675: -1.13105822124, 2676: -3.97913926252, 2677: -0.040192079319, 2678: -6.19697391288, 2679: -3.42800806711, 2680: 5.7529804672, 2681: -0.286900469642, 2682: -0.383818139711, 2683: -0.671929063926, 2684: -2.2038633537, 2685: 1.72627580828, 2686: 0.523110671464, 2687: 5.77842395041, 2688: -2.16535809688, 2689: 5.66286809213, 2690: 1.61971609638, 2691: 1.78084813388, 2692: -4.19961752096, 2693: -0.396483306644, 2694: 3.71529382857, 2695: 1.62452922512, 2696: -0.73105420587, 2697: 0.0723288511887, 2698: 0.366415209598, 2699: -0.424353121008, 2700: 1.35376522971, 2701: -0.624573900977, 2702: -2.74056461458, 2703: -2.6577475983, 2704: 3.84950991715, 2705: 0.841673557381, 2706: -1.0912013653, 2707: -1.36346764244, 2708: -1.50466972639, 2709: 3.01764715349, 2710: -2.14612305474, 2711: 1.01512627381, 2712: 4.73350504973, 2713: -0.410836479856, 2714: 2.66213294503, 2715: -0.536347838338, 2716: -1.95099869871, 2717: -7.99036582818, 2718: -9.49345034113, 2719: -5.94836433261, 2720: -1.2836182273, 2721: 1.52530350437, 2722: -5.24408204304, 2723: 0.64258975906, 2724: -1.10408750963, 2725: 1.95386070078, 2726: 4.51311370307, 2727: -5.45563261259, 2728: 9.46650837322, 2729: -4.2122224638, 2730: 0.278473285905, 2731: 1.1609050564, 2732: -0.858905336436, 2733: 3.39931319437, 2734: 4.10723123957, 2735: -1.53572410508, 2736: 7.74606670688, 2737: 0.502153949337, 2738: -8.66304768795, 2739: 0.502464097258, 2740: -1.53656428265, 2741: -0.979034874075, 2742: 3.2194972402, 2743: 5.6763293447, 2744: -2.06055518307, 2745: 4.12790761228, 2746: 2.56839876334, 2747: -2.90061911466, 2748: -0.299406273731, 2749: -1.26902355693, 2750: 0.410417308343, 2751: -0.164514282969, 2752: -0.306617643046, 2753: -0.734505338409, 2754: 2.24932970772, 2755: -0.0241885482987, 2756: -0.509716213363, 2757: -6.8106258621, 2758: -0.238820281869, 2759: 3.72692113709, 2760: -1.24561818535, 2761: -0.327477923306, 2762: -1.89972013879, 2763: 0.180623936937, 2764: -0.453946126387, 2765: -0.176697986281, 2766: -3.04529848296, 2767: -2.10293710948, 2768: 2.86261779663, 2769: -0.0472383968271, 2770: -1.430831186, 2771: -3.69189169217, 2772: 0.992820828521, 2773: 1.63223450685, 2774: -0.562023979865, 2775: 0.158352056169, 2776: -2.74724202125, 2777: 0.211692070643, 2778: -3.72348716477, 2779: 5.39045156746, 2780: -13.9209095776, 2781: 0.475964396139, 2782: 0.528109594826, 2783: -4.29255396135, 2784: -0.856276379531, 2785: 9.06944802318, 2786: 1.3534141053, 2787: 2.61068410389, 2788: 3.99511403626, 2789: 0.120587196094, 2790: 3.92819035887, 2791: 5.18090259169, 2792: 8.77427842441, 2793: -0.625306910144, 2794: -99.6472916588, 2795: -23.3877467331, 2796: -60.6199284475, 2797: -3.20830839893, 2798: 9.0564466247, 2799: 41.3595428883, 2800: 41.5362211776, 2801: -12.3474506054, 2802: -41.2379792808, 2803: 34.0880351243, 2804: -16.6314926356, 2805: -10.3143498502, 2806: -2.8612633495, 2807: 15.9073219713, 2808: -6.59848150624, 2809: 21.3833019933, 2810: 9.54511985123, 2811: 13.8475946084, 2812: -3.2428611405, 2813: -19.8623920567, 2814: -82.6609027161, 2815: -17.5457128869, 2816: 36.0707496525, 2817: 16.333437329, 2818: -4.50148609858, 2819: 4.23967423654, 2820: 10.6063881819, 2821: -20.9623399441, 2822: 6.72158932223, 2823: -23.1937007548, 2824: -30.0156727484, 2825: -34.7201811709, 2826: -26.4458908897, 2827: -13.5601908889, 2828: -21.3340459651, 2829: 3.43799711211, 2830: -43.2668659466, 2831: -7.61676029939, 2832: 30.9215396478, 2833: -18.7723413966, 2834: -5.06667462164, 2835: 3.43377555074, 2836: -17.1656360235, 2837: -15.8460242592, 2838: 7.51754800142, 2839: -1.9631276529, 2840: -24.3093070839, 2841: 5.68274556469, 2842: -5.26122478574, 2843: -8.87751553842, 2844: -29.7395939087, 2845: -4.94122372582, 2846: -6.82068951645, 2847: -6.60793029135, 2848: -5.0155438761, 2849: -0.621559202585, 2850: -6.23393103357, 2851: -16.8799785871, 2852: 33.7988413637, 2853: 27.7838617086, 2854: -49.8495871173, 2855: -60.7056854256, 2856: -32.006868574, 2857: 22.9422881163, 2858: -3.10379502399, 2859: 58.9270175006, 2860: -4.8692185821, 2861: 28.288985244, 2862: -5.84177877832, 2863: 9.1998147225, 2864: 3.54244595007, 2865: -5.26853335115, 2866: 2.94036294371, 2867: 0.53440080297, 2868: -21.2260799431, 2869: -7.50671998908, 2870: -15.5546456013, 2871: -52.5577757257, 2872: -15.6072209808, 2873: -1.90957983692, 2874: 2.53215718548, 2875: -2.87531770058, 2876: -3.8853687905, 2877: 5.34293751108, 2878: 6.71679080124, 2879: -2.37461758455, 2880: 0.274247103718, 2881: -5.04438225761, 2882: -6.16094889113, 2883: 5.24721465492, 2884: -4.96266524586, 2885: 19.073726864, 2886: -4.40565291573, 2887: -4.49168349036, 2888: -4.7483289102, 2889: 2.72384735403, 2890: -4.39308547786, 2891: -22.0854848874, 2892: -3.55464183696, 2893: 5.18026315433, 2894: -5.58991858456, 2895: -1.61035071829, 2896: -4.62757734657, 2897: -5.07145043985, 2898: 38.4003109009, 2899: -4.87578193232, 2900: -5.08526417529, 2901: 16.5252907619, 2902: 32.1976638571, 2903: -6.92379943466, 2904: -12.1619169073, 2905: -45.3078573011, 2906: 20.6925675802, 2907: 2.0306883525, 2908: 12.4631656178, 2909: -15.6446587877, 2910: -7.64653658513, 2911: 43.5554736196, 2912: -28.1643654214, 2913: 9.08128291072, 2914: -7.44636454404, 2915: -9.4416789428, 2916: 38.4438065595, 2917: -13.350783302, 2918: -1.02360318244, 2919: -0.206357190301, 2920: 15.4558429342, 2921: -14.5322018284, 2922: -0.874894363105, 2923: -2.29905411311, 2924: -0.721068403252, 2925: 1.30104404743, 2926: -0.278014221991, 2927: -0.59966692715, 2928: 2.5413575391, 2929: -0.372476708657, 2930: 4.17581879097, 2931: -0.931555446296, 2932: -1.92260161369, 2933: -14.0691502246, 2934: -3.80439993802, 2935: 15.2927586153, 2936: -0.224688931601, 2937: 39.7084978274, 2938: -0.901059400131, 2939: -0.771900337062, 2940: -16.0531424198, 2941: -0.893536561303, 2942: -4.77886865802, 2943: -5.07744927694, 2944: -0.735403217261, 2945: -0.745222423627, 2946: -0.856775625719, 2947: -0.785077214195, 2948: -5.13669783124, 2949: -19.0414895618, 2950: -2.8911848088, 2951: -1.78883096786, 2952: -1.48920496697, 2953: -4.39780406958, 2954: -4.8383023659, 2955: 13.0096141316, 2956: 1.60427815363, 2957: 9.61014078565, 2958: -24.1237938471, 2959: 1.82015718883, 2960: 0.899571013265, 2961: 3.70704630014, 2962: -3.02719511252, 2963: 23.7590396919, 2964: -0.786027140446, 2965: -19.0414729745, 2966: -0.704809503485, 2967: -1.05588798859, 2968: 18.4870401751, 2969: -25.2854002932, 2970: -1.88980001713, 2971: -0.922038213396, 2972: -1.24646738393, 2973: -1.35762638303, 2974: 0.760363052666, 2975: -28.1775714618, 2976: 2.64563968521, 2977: 4.77532884407, 2978: 52.2194129231, 2979: 26.6486340702, 2980: 0.989928722589, 2981: 15.0414312204, 2982: 4.23526627466, 2983: -3.63570027249, 2984: -34.1927342565, 2985: -0.913304332181, 2986: -34.4372378682, 2987: 18.1257330517, 2988: -0.459188513776, 2989: -1.13022374299, 2990: -12.9858874727, 2991: -0.688287243239, 2992: -7.8372506445, 2993: -0.776194991311, 2994: -17.419888509, 2995: -0.743579299411, 2996: -0.839675550983, 2997: 19.0183213759, 2998: -4.84476454767, 2999: 10.8457439179, 3000: -0.731993270901, 3001: -0.812647322619, 3002: -2.30229174176, 3003: -0.837611793857, 3004: -2.86431139227, 3005: 3.88730885484, 3006: 32.9652670899, 3007: -0.448260467093, 3008: -1.08147087555, 3009: -7.45045906183, 3010: -4.1445274787, 3011: -4.576375468, 3012: 2.30329521407, 3013: 7.73488680697, 3014: -8.91882854913, 3015: 1.54349594363, 3016: -4.48019094248, 3017: -2.19253229104, 3018: -0.963609041223, 3019: 34.4940418974, 3020: -17.0875977916, 3021: -1.58519249918, 3022: 3.26070726095, 3023: -1.19586243038, 3024: -1.1907977056, 3025: -24.8557154716, 3026: -12.9000136492, 3027: -0.222989594216, 3028: 35.9664158269, 3029: -4.71005062748, 3030: -14.0171442136, 3031: -4.41629547371, 3032: -2.35236561461, 3033: 4.16315880139, 3034: -6.08792808696, 3035: 19.8106018132, 3036: 17.1467006635, 3037: 20.714305206, 3038: 4.2072340217, 3039: -6.3163514087, 3040: -12.2052589421, 3041: 0.612051364219, 3042: 0.141355550389, 3043: 23.2885936824, 3044: 22.1651396468, 3045: -5.52477375056, 3046: -36.9743128406, 3047: -2.68544059265, 3048: -4.39798501117, 3049: -7.95446499716, 3050: -3.92928307974, 3051: -4.86208695351, 3052: -5.18085920212, 3053: -3.59452422474, 3054: -5.07679583547, 3055: -4.15876993056, 3056: -17.8923786422, 3057: -5.10122011274, 3058: 10.7736947357, 3059: 32.2754585672, 3060: 31.3478115721, 3061: 26.2809604643, 3062: -4.36533718361, 3063: -4.51016577663, 3064: 17.5951888401, 3065: -1.49414597262, 3066: 85.6648889167, 3067: -36.7109793735, 3068: 21.7698570874, 3069: 5.73616619533, 3070: 7.0329432748, 3071: -2.77809009481, 3072: -19.3467143385, 3073: -4.82636421964, 3074: -36.69248541, 3075: -10.5079889108, 3076: 14.2272670423, 3077: -14.9305677487, 3078: 20.1679370743, 3079: -5.03915972489, 3080: 21.8271396629, 3081: 1.19651418741, 3082: -23.7837526484, 3083: -7.39658904632, 3084: 4.19838288096, 3085: 20.3348491392, 3086: -28.6339822603, 3087: -28.4615020873, 3088: 68.1111044792, 3089: 32.2624354659, 3090: 43.6182446419, 3091: -2.57272885223, 3092: 11.7134241383, 3093: -4.6246242801, 3094: 14.175162347, 3095: -5.95113181003, 3096: -20.6455665175, 3097: 2.00345803034, 3098: -4.65765566359, 3099: 32.880839316, 3100: -7.86731289787, 3101: -7.73923196279, 3102: 1.19181829706, 3103: -22.3511378756, 3104: -4.78912317069, 3105: -4.96580931892, 3106: 2.45993746197, 3107: -4.8298311832, 3108: -40.4026715215, 3109: -4.79316307757, 3110: -4.91839659556, 3111: -4.91367668045, 3112: -5.56478329057, 3113: -1.02169664428, 3114: 6.80034458451, 3115: 17.5799671358, 3116: 26.2140470818, 3117: -15.644308465, 3118: 0.896603686411, 3119: 28.0216768429, 3120: 22.3298365949, 3121: 25.3017037474, 3122: 4.56758838043, 3123: -19.4390125013, 3124: -18.2271580031, 3125: -5.14710036343, 3126: 7.77856646118, 3127: -6.94966254388, 3128: 23.8387149196, 3129: 38.0717553884, 3130: 4.82436439248, 3131: -6.90264476297, 3132: 28.545385319, 3133: 26.4790796259, 3134: 14.3708630689, 3135: -9.2804364134, 3136: -26.2826893325, 3137: -34.6781715993, 3138: 98.6390257442, 3139: -100.0, 3140: -5.32583916524, 3141: 75.5603298878, 3142: 0.0370937507959, 3143: 14.2550576744, 3144: -12.5033480267, 3145: -4.30678659271, 3146: 0.943488292909, 3147: -15.1797491486, 3148: 6.38876787167, 3149: 5.56089890405, 3150: 17.1709977102, 3151: 13.6824664599, 3152: -2.03179056106, 3153: -13.59055082, 3154: 7.63581576141, 3155: 12.5779095924, 3156: -3.37213639135, 3157: -1.1581027663, 3158: -19.5241052438, 3159: -2.58926503318, 3160: -12.052462924, 3161: -17.474860436, 3162: 15.6660948028, 3163: 15.039390861, 3164: 15.5646578015, 3165: 25.2685207251, 3166: 1.77563960672, 3167: -28.138809856, 3168: -19.1978827808, 3169: -17.1163654267, 3170: 24.5173922097, 3171: 6.01941444515, 3172: 1.74322644096, 3173: -2.8796211661, 3174: 6.73382118033, 3175: -10.1637763246, 3176: -16.4672695523, 3177: -7.9370153979, 3178: -13.6153222004, 3179: -5.39496462262, 3180: -25.4145772477, 3181: -1.7179741673, 3182: 9.61789775769, 3183: 0.0439261359531, 3184: 6.03776563736, 3185: 7.89838156168, 3186: 10.3567457689, 3187: -37.2894926667, 3188: 13.8512544573, 3189: 0.651499137182, 3190: -29.4541129996, 3191: 0.794218573605, 3192: -7.65224590895, 3193: -11.4947215268, 3194: -0.173806856712, 3195: 0.87112204337, 3196: -18.6802634939, 3197: 0.617791599916, 3198: 18.9083685589, 3199: -4.33103479293, 3200: -7.8168419268, 3201: 2.19450716029, 3202: -1.52283884896, 3203: 8.0427650562, 3204: -8.983857975, 3205: 11.2058811999, 3206: 5.98527777813, 3207: 0.875658681505, 3208: -17.0877181998, 3209: 1.23327773989, 3210: 2.90111583103, 3211: 1.94972809485, 3212: 5.68540984176, 3213: 33.8858100327, 3214: 0.474205704056, 3215: -7.47759869257, 3216: -0.757725713027, 3217: -1.32682496642, 3218: 6.64856812902, 3219: 10.1424490333, 3220: 58.4814949556, 3221: -24.4314116809, 3222: -0.325422410156, 3223: -8.8520445454, 3224: 1.9260939972, 3225: 0.233343522807, 3226: 5.72761578142, 3227: -13.565812616, 3228: 0.63611599955, 3229: 13.6672861857, 3230: 0.0270637576261, 3231: 3.32898659492, 3232: -4.47908951219, 3233: 0.478414013973, 3234: 5.52082275787, 3235: 0.98888271411, 3236: 0.596097267526, 3237: 0.158267413932, 3238: -11.3865334878, 3239: 1.67986986939, 3240: 0.997008635391, 3241: 6.69463476853, 3242: 5.4563710435, 3243: 0.696047318799, 3244: -1.07533175971, 3245: 0.811473703834, 3246: 0.18788086439, 3247: -54.5895991678, 3248: 0.920325721817, 3249: 0.678204301959, 3250: 0.173620864389, 3251: -3.98430264596, 3252: -0.219798927983, 3253: 12.9158538828, 3254: 36.9210248817, 3255: -33.4294446121, 3256: -0.485758512714, 3257: 5.00583859758, 3258: 18.9980549339, 3259: -5.69569881615, 3260: -24.1020474795, 3261: 12.7709456732, 3262: 8.90857186435, 3263: 22.3610576896, 3264: 8.46694813365, 3265: 4.96564533898, 3266: 1.5024828317, 3267: 0.230509204628, 3268: 0.492257014506, 3269: -0.514036716383, 3270: 25.0388343795, 3271: 0.183610306159, 3272: 1.08625489532, 3273: -1.49929859697, 3274: 17.1309516502, 3275: 0.300879363336, 3276: 1.45409452667, 3277: -8.59757248075, 3278: 0.171677710657, 3279: -3.14105888941, 3280: -0.0951250713413, 3281: 1.81683898779, 3282: -22.6093501282, 3283: 21.6271159909, 3284: -2.20805133059, 3285: 1.35321233946, 3286: 6.8797248409, 3287: -0.455303372828, 3288: 0.211341482928, 3289: 11.3267691981, 3290: 0.20916702434, 3291: 0.979633826471, 3292: -11.3847683431, 3293: 0.177566354232, 3294: 0.0560701282244, 3295: 0.294700521674, 3296: 0.0626470088119, 3297: 0.717274402829, 3298: 9.61999819424, 3299: -0.57677964625, 3300: 0.0142719646378, 3301: -3.34607738174, 3302: 1.4849072484, 3303: 0.825437771946, 3304: -13.9078715755, 3305: 1.25510641805, 3306: -31.2244513712, 3307: 62.7966884257, 3308: 5.30267459189, 3309: -4.57270091784, 3310: 26.3238316194, 3311: 0.44745383715, 3312: -22.0753851749, 3313: 0.0375374532898, 3314: -17.2662720695, 3315: 0.180137832177, 3316: 14.9886100687, 3317: -29.1869172115, 3318: 11.5257801024, 3319: 0.160693611575, 3320: 0.142275520784, 3321: -0.171397388296, 3322: -0.179105590048, 3323: 0.234179959918, 3324: 4.77133369844, 3325: 2.84404688012, 3326: -5.43343054727, 3327: 6.96960453866, 3328: -26.734956107, 3329: -4.20234219038, 3330: 1.33469027131, 3331: -7.61023247818, 3332: 9.92151857204, 3333: -3.87764684305, 3334: 0.254166753891, 3335: 22.7858779942, 3336: 1.30173540262, 3337: 4.34654485526, 3338: -0.550423976879, 3339: -17.906598807, 3340: 0.101791149622, 3341: 0.797634911546, 3342: 0.184663196972, 3343: -6.83060451375, 3344: 0.274218051282, 3345: 0.299319108782, 3346: 4.35695256906, 3347: 0.290268397382, 3348: -25.6306638745, 3349: 0.233398465222, 3350: 0.207612761806, 3351: 4.39605070043, 3352: -0.0124114186376, 3353: 1.73291552858, 3354: 1.10173860188, 3355: -17.1124821241, 3356: 1.17023540887, 3357: 2.19501639084, 3358: -21.4837163566, 3359: 0.336679606612, 3360: 3.80347664894, 3361: 6.664529462, 3362: 1.89796668499, 3363: -1.05905204845, 3364: 4.20909880846, 3365: 0.78335633334, 3366: -1.85068460462, 3367: -0.0271495702951, 3368: -4.88238797807, 3369: -1.09276758117, 3370: -0.559274067332, 3371: 6.01007361627, 3372: 9.32991582411, 3373: 0.0427810541308, 3374: 5.72179190894, 3375: -10.1673724716, 3376: 0.172177423917, 3377: -14.3722847667, 3378: -21.0305080859, 3379: 15.7965088724, 3380: -0.885591773875, 3381: 1.05921529528, 3382: 5.59044178371, 3383: -6.11313318436, 3384: -2.87431796078, 3385: 24.3484116967, 3386: 1.4867267765, 3387: 8.94393156609, 3388: -12.5219215356, 3389: 6.2404122765, 3390: -35.4660743561, 3391: 20.5612810984, 3392: -21.5939001992, 3393: 2.2768949721, 3394: -13.8472236766, 3395: -3.240091397, 3396: 15.678419476, 3397: 0.0644596929505, 3398: -0.407609261213, 3399: 0.909893073182, 3400: 0.984938677506, 3401: -12.4857058138, 3402: 14.93593956, 3403: 1.22564319593, 3404: 0.473859869838, 3405: -1.23198236146, 3406: 0.672316328635, 3407: 13.0680062209, 3408: -6.73061687922, 3409: -1.9945671914, 3410: -16.0808813396, 3411: 1.24334379768, 3412: 1.40628016241, 3413: -4.72783157666, 3414: 10.3155996945, 3415: 9.67459820532, 3416: 1.00350631525, 3417: -13.9260517126, 3418: 3.40081103694, 3419: 15.8708867543, 3420: 0.239401041192, 3421: 25.5357832503, 3422: 0.738734874001, 3423: 20.8165221899, 3424: 2.95895639004, 3425: 37.0530715673, 3426: -13.9395908598, 3427: -2.5931552957, 3428: 1.02345298507, 3429: 2.82554863244, 3430: 1.18684821814, 3431: -2.25025862129, 3432: -1.61650663046, 3433: 1.96909220012, 3434: -10.824606288, 3435: -11.2884913147, 3436: 5.22409741286, 3437: -1.2643725772, 3438: -19.2860716968, 3439: -7.32988864073, 3440: 9.7076352198, 3441: 28.3261388984, 3442: 1.56544487407, 3443: -9.51398294743, 3444: -3.02681287811, 3445: -9.56665303307, 3446: 7.20708560475, 3447: 0.494416872982, 3448: 20.6205115187, 3449: 0.769890537333, 3450: 0.880990292679, 3451: -0.16739830234, 3452: 4.25528095456, 3453: -0.0550465501079, 3454: 0.384249189931, 3455: 2.391288352, 3456: 0.283201832132, 3457: -7.55339427819, 3458: -0.0134962999358, 3459: 0.661052358335, 3460: -13.5419608563, 3461: 0.707389463078, 3462: 1.97284797896, 3463: 7.07306030607, 3464: -0.443010822908, 3465: -17.7176853868, 3466: 8.27562726568, 3467: -11.6459652299, 3468: -17.1918531378, 3469: -14.5380478333, 3470: 2.39716290784, 3471: 2.43714318573, 3472: -10.9159654273, 3473: -18.2625637356, 3474: 0.842940119706, 3475: 7.1152299966, 3476: -5.07627421279, 3477: 37.2806929848, 3478: 0.339970688157, 3479: 32.5059346928, 3480: 0.767098698542, 3481: 4.57657341824, 3482: 4.12924297042, 3483: -8.51301398144, 3484: -22.7959722742, 3485: 8.34786251778, 3486: -22.3362391647, 3487: -41.0725688706, 3488: -53.9153226975, 3489: -16.9072913321, 3490: -28.1279548244, 3491: 26.4232557601, 3492: -0.43660205463, 3493: -0.748231685736, 3494: 0.855346468895, 3495: 0.74043994591, 3496: -0.835486474436, 3497: 1.78098514423, 3498: -0.325100382809, 3499: 1.19577861007, 3500: -1.3287065893, 3501: -1.23826837378}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.358339828798, 2: -13.6997538041, 3: -4.95888582056, 4: 0.898660865032, 5: -25.7331273506, 6: -3.75426117081, 7: 5.84899530549, 8: 19.5667165878, 9: 39.5392275722, 10: 2.79493059492, 11: 7.71608292024, 12: -28.3032445575, 13: -4.66737976263, 14: 0.338046538961, 15: 11.1382655949, 16: -8.30642763678, 17: -56.8242853528, 18: -11.2591904588, 19: 1.4629901954, 20: 5.86099718862, 21: -6.78023333649, 22: -6.92681746155, 23: -0.813036779765, 24: 28.4532238684, 25: 16.1820249107, 26: 0.918396247294, 27: -14.6686079078, 28: 20.8777174255, 29: -3.42718940395, 30: 17.5846024624, 31: 15.4811737437, 32: -21.7056802573, 33: -17.7042393757, 34: -10.6736298554, 35: -6.55008610465, 36: -12.1259454308, 37: -24.1949330951, 38: -1.73309993391, 39: -9.20073157209, 40: -17.1004760702, 41: 1.69487045989, 42: 1.62786425201, 43: -25.4022167548, 44: -3.12270437863, 45: 3.84425650509, 46: -5.82503590325, 47: -21.473477801, 48: 1.36189987477, 49: 15.1076975315, 50: 0.196362502975, 51: -1.13589413228, 52: -18.4052048115, 53: 1.30610986021, 54: 7.19362241965, 55: 2.99488808411, 56: 0.380611703405, 57: -1.96991071437, 58: 0.220206093699, 59: 34.8652990461, 60: -9.48422037303, 61: -45.5382149709, 62: 12.4387557229, 63: -19.7460797375, 64: -1.47190857237, 65: -14.8056793063, 66: -0.897919525655, 67: -16.7337454787, 68: 1.75321633455, 69: 6.38047650039, 70: 3.93400902377, 71: 2.77970439665, 72: -0.180464108832, 73: 1.67995663668, 74: 11.1084780135, 75: -6.42710885005, 76: -0.743319090999, 77: 1.34936712599, 78: 4.45185826763, 79: -7.81480083802, 80: 45.4867151164, 81: 0.322655478501, 82: -13.5252912195, 83: -1.28117779998, 84: -1.63474352313, 85: -13.400377847, 86: 1.14748361318, 87: 2.36037178383, 88: 28.5746009802, 89: 2.29085023451, 90: -0.996350194779, 91: 3.48834837236, 92: 1.432699809, 93: 11.0260412571, 94: 2.23413097806, 95: 1.88172410904, 96: 1.47905483763, 97: 7.38290837592, 98: 2.84498315369, 99: 2.89129273759, 100: 2.04948985677, 101: 10.4975117753, 102: 2.22261031401, 103: 7.48421119755, 104: 0.701912652043, 105: -9.55939639697, 106: -25.7470051977, 107: 2.331100873, 108: 1.26738083521, 109: 2.58624391318, 110: -19.2423045226, 111: 4.53402128152, 112: 14.5215771369, 113: 0.936428014994, 114: 22.0707633718, 115: -29.2711787738, 116: 10.5360544602, 117: 35.1199920674, 118: 9.57717117856, 119: -38.1703465545, 120: -34.1245579924, 121: 7.18389467903, 122: 25.720893579, 123: 42.9255573138, 124: 9.77505656985, 125: -0.073349180226, 126: 0.263071398728, 127: 1.05445136961, 128: 0.731710715244, 129: -0.961074724725, 130: 0.268992405151, 131: 2.5363228443, 132: 0.774209336559, 133: -8.04617678194, 134: -0.759661562333, 135: 0.918960355984, 136: -8.5754422375, 137: 0.510793217005, 138: -7.86281429333, 139: 0.198410157668, 140: 1.89444028407, 141: -13.0618631479, 142: 0.962223865414, 143: 4.55220586138, 144: 1.01066265234, 145: 19.5757683529, 146: 0.241864701854, 147: 0.23319923211, 148: -0.396414977422, 149: 0.315396023215, 150: 1.45448452881, 151: 14.0466153227, 152: -0.571925998076, 153: 0.308534464313, 154: 0.366249415986, 155: 0.0581654163865, 156: -1.12117856261, 157: 3.58021136466, 158: 2.47533749101, 159: 0.298476911371, 160: -0.11400735309, 161: -0.0843632978402, 162: 2.47664514555, 163: 24.5243616654, 164: 1.86382951244, 165: 19.8175887653, 166: -1.36404953495, 167: 5.22173045069, 168: 3.03734884927, 169: 9.30665954578, 170: 1.22566022747, 171: 3.2538369652, 172: 0.435614473548, 173: 27.6042361383, 174: 0.339350260683, 175: -3.32641098651, 176: 14.6824308216, 177: 6.26694399891, 178: -0.324790102667, 179: 0.441363596532, 180: 0.40919208383, 181: 0.190132544574, 182: 0.9997094472, 183: -2.88673666053, 184: -10.6783199596, 185: -6.17359230474, 186: 30.6635991299, 187: 24.0937306053, 188: 2.53508420478, 189: 1.3762653393, 190: -5.17883882579, 191: 3.49994368176, 192: 21.9250712012, 193: 0.131461323329, 194: -0.690883202164, 195: 10.4524603795, 196: 1.39901957719, 197: -2.49064956917, 198: -17.1157856222, 199: 0.204952069775, 200: 2.48037317524, 201: 0.389073683393, 202: 21.7956181065, 203: 0.301791318409, 204: 0.259648167274, 205: -12.4934885778, 206: 1.85332100872, 207: -56.6050371685, 208: 0.403531100998, 209: 0.369324747925, 210: -0.0656322985841, 211: 0.391015399031, 212: -0.640807819125, 213: 2.56381063561, 214: -14.819228018, 215: 2.0377372692, 216: 1.75235989938, 217: 19.0233979821, 218: -5.90328608034, 219: 36.0315147553, 220: -1.20239227117, 221: -5.52907636244, 222: 5.11522420653, 223: -0.00377519841274, 224: 0.332759788308, 225: 1.34730173776, 226: 1.192753106, 227: -5.14703993397, 228: -1.60404472713, 229: -0.220567363914, 230: 1.81005775194, 231: -34.3757836848, 232: 0.498434204101, 233: 8.75474927101, 234: -1.93744168215, 235: 1.46985694319, 236: -8.33632811223, 237: -9.80842521444, 238: -3.52972251376, 239: 1.32897768236, 240: 2.51934874931, 241: -17.8230673191, 242: 3.92145132227, 243: -2.6235845816, 244: -26.6105288074, 245: 19.1928617304, 246: 19.3978134801, 247: 7.72488808889, 248: 13.7131489699, 249: -65.3050912024, 250: 7.58038312464, 251: 11.6591993303, 252: -1.66561482211, 253: -25.7386701092, 254: -12.9025472103, 255: 3.67577228755, 256: 3.52398835537, 257: 10.6923108098, 258: 2.18684747909, 259: 1.47689424136, 260: 17.3990761714, 261: -3.6686192879, 262: 2.03796093528, 263: 0.56518886554, 264: 2.59425601311, 265: -0.999976065731, 266: 3.41529858889, 267: -7.88764417183, 268: 10.8249253575, 269: -20.0414422592, 270: -4.23046717526, 271: -2.93655505561, 272: -12.8262115775, 273: -14.0097060748, 274: -4.46866508125, 275: 1.38422983876, 276: 6.31291988401, 277: -25.7531083601, 278: 4.72788639675, 279: 1.86818230747, 280: 2.42453956285, 281: 0.286434958092, 282: 13.0487226516, 283: -3.62243586706, 284: 3.93185322063, 285: -2.45938417861, 286: 3.56396728729, 287: 2.32391231056, 288: 21.363990338, 289: 1.22807858861, 290: 1.88553178947, 291: 1.916190197, 292: -0.362528989414, 293: -12.9752969613, 294: -30.2784681009, 295: 1.53986986127, 296: 24.0964652233, 297: 4.34430153834, 298: 6.91169078889, 299: 13.5140556845, 300: -5.36748201556, 301: 5.09431171315, 302: 0.837122966101, 303: -0.458055773905, 304: -25.4722189459, 305: -22.7521097226, 306: 1.94374725787, 307: 9.59210632215, 308: 2.3056409943, 309: 2.29386632527, 310: 2.27397552884, 311: -22.6480856882, 312: 0.502239648881, 313: 0.997099651965, 314: 23.5054491508, 315: 1.79565240002, 316: 24.7686429409, 317: 0.915957733994, 318: 1.67005952611, 319: 23.1738534332, 320: -2.80759140747, 321: -2.1554922869, 322: 27.5061671863, 323: -35.3118562481, 324: -50.8631900071, 325: -2.597514192, 326: 12.4279070414, 327: 9.44576449206, 328: -63.0489953266, 329: -3.79282730285, 330: 6.42672267122, 331: -27.5531527261, 332: 5.97946940626, 333: -2.35886752799, 334: -5.55869625732, 335: -0.816003946448, 336: 33.0799233378, 337: 9.27067656014, 338: 14.4152399253, 339: 1.76283036291, 340: 21.4886231988, 341: 14.0946682046, 342: -28.1258756443, 343: 13.3199838254, 344: 20.7240253739, 345: 8.81339415714, 346: -97.586559072, 347: -50.1855294137, 348: -61.9352102374, 349: 10.2429871937, 350: -0.183465467863, 351: -3.96796984799, 352: -8.61454756673, 353: -13.964120365, 354: -43.094050312, 355: 9.53658631946, 356: 18.3523133466, 357: 15.0205681329, 358: -12.6162718078, 359: -5.79808856439, 360: -10.3967870292, 361: 9.84850349976, 362: 10.7779875937, 363: -6.26109214839, 364: 0.102457524959, 365: -35.768331238, 366: 14.0182412879, 367: 34.068988402, 368: 15.0414184667, 369: -10.5240115673, 370: -20.8853692634, 371: 2.68516199757, 372: -15.8804792349, 373: -6.42667292523, 374: -3.13518688179, 375: 20.8677603737, 376: 9.04244291337, 377: 1.62059297779, 378: 3.50274289555, 379: 1.27359539005, 380: -7.44985224109, 381: 36.4712265101, 382: 18.0433060749, 383: 2.36980311606, 384: 6.20838759638, 385: -2.0620685721, 386: 3.34680643159, 387: -2.79793011248, 388: 55.8214084001, 389: 18.7994837142, 390: 25.6272872834, 391: -1.38076333717, 392: -4.71682009463, 393: 9.45363408856, 394: 4.80539946064, 395: -13.2750315684, 396: -6.50551950781, 397: 0.56727661966, 398: -44.8774656825, 399: -3.69352465165, 400: 2.68377587323, 401: -8.56905816932, 402: -3.77023221293, 403: 3.50007852228, 404: -12.6788787321, 405: 0.697221688509, 406: 43.9887612442, 407: -1.93707353054, 408: 19.6928811829, 409: -6.43275631312, 410: 36.9474701871, 411: -15.8116218217, 412: 0.509831394841, 413: -1.21632626711, 414: -13.8743860762, 415: -8.3342744399, 416: -27.5524941386, 417: -0.00582183997324, 418: 7.59361824379, 419: -3.66836417018, 420: -0.082904847755, 421: -1.47301632574, 422: 1.70710616641, 423: 9.04625427998, 424: -7.05629555203, 425: -0.784444054992, 426: -10.6727194763, 427: -7.51035627948, 428: -23.1538578872, 429: 19.641571831, 430: 0.310427789873, 431: 13.415629397, 432: -1.20779180099, 433: -2.72291533703, 434: 2.09871185341, 435: 12.8622334914, 436: 4.0126441935, 437: 1.03794505929, 438: -1.70069973781, 439: 7.21459735819, 440: 1.64217778617, 441: -1.46045343632, 442: -24.6091594734, 443: -2.31139963799, 444: -1.896515077, 445: -2.18382988569, 446: 9.43309599432, 447: -2.09815393338, 448: -8.08159239156, 449: -7.0343076965, 450: -12.5219750052, 451: -4.5488518084, 452: 4.61262983861, 453: -6.94929170039, 454: -0.421669803772, 455: 20.3137124868, 456: -1.28248462606, 457: -0.978558614858, 458: -2.58624241994, 459: 12.7811109029, 460: 4.03816145073, 461: -16.7011869497, 462: -13.3395666638, 463: 12.9835727719, 464: 0.499092789764, 465: 1.45134870575, 466: 2.36152853479, 467: 19.0278715724, 468: -37.6287137403, 469: 2.39818392407, 470: -1.79505780393, 471: -14.8864324032, 472: 2.96098345283, 473: -6.30047151721, 474: -0.999094958884, 475: -0.568902615017, 476: -2.32470655136, 477: -18.0355999465, 478: 22.7802805648, 479: -0.64771554244, 480: 1.20883030497, 481: -2.79612068406, 482: 22.5999489434, 483: 2.2057800906, 484: 2.18636972997, 485: -16.3342705277, 486: -2.66533291339, 487: -12.7069520884, 488: -2.67610425916, 489: 5.03152781148, 490: 27.6740332549, 491: 12.0610699022, 492: 1.09936347092, 493: -0.0155911658342, 494: -8.51581676307, 495: -0.320044974496, 496: -0.423598131909, 497: -20.0491958886, 498: -0.02358136655, 499: -1.99892882339, 500: 7.15412397009, 501: -0.517247892137, 502: -0.204580503598, 503: -0.472715165453, 504: -0.145787896759, 505: 11.5182436916, 506: 25.0323175509, 507: 0.195251956019, 508: -0.494796650991, 509: 1.85705135192, 510: -2.55117712138, 511: -5.73512471393, 512: -6.14345970952, 513: 3.70963140559, 514: -12.4480684844, 515: 66.5924979789, 516: -5.72716575699, 517: -1.7996101698, 518: -0.0614398717462, 519: -0.220317569054, 520: -11.2813735621, 521: 0.251223226117, 522: 3.11595529873, 523: -0.146777062607, 524: -19.9049294924, 525: 21.2168775467, 526: -38.0804008963, 527: -1.40004622003, 528: -0.461188435101, 529: -0.181505509797, 530: 1.17426045258, 531: -0.461889246359, 532: 9.28670754143, 533: 3.13877689141, 534: 6.23484907707, 535: 7.13760399312, 536: -25.1502868902, 537: -4.48232631807, 538: -3.11224404145, 539: -2.57941211648, 540: 5.16212861484, 541: 36.9921257608, 542: -0.370449312927, 543: -27.851788692, 544: -1.74807896542, 545: 2.85260535024, 546: -14.9073703927, 547: -8.90979501698, 548: -0.403082327076, 549: -1.90946175232, 550: -0.363076634278, 551: -4.11627399919, 552: -0.490299887237, 553: -0.345939290702, 554: 14.7886935071, 555: -2.15815443058, 556: -7.64073740032, 557: -0.362267557814, 558: -0.314442188374, 559: 1.13421524816, 560: -0.336860664273, 561: -3.55159313107, 562: 11.0657431397, 563: -2.59347246041, 564: 0.0503610709995, 565: 0.723311282869, 566: 41.8153252327, 567: -2.14231134445, 568: -8.78162651389, 569: -13.4844283798, 570: 18.9886939077, 571: 4.91325799824, 572: 0.112703402053, 573: -3.72978327355, 574: 4.21611657122, 575: 0.504005499563, 576: 12.9785647858, 577: -8.21812733713, 578: 4.79962789346, 579: -1.52366311005, 580: 1.77015690434, 581: -1.44363586091, 582: -14.2043466213, 583: 0.0274212787462, 584: 5.69170591769, 585: -24.5299125776, 586: 9.23984733597, 587: -3.80419320737, 588: 6.89828904631, 589: 1.013198672, 590: 44.2952234492, 591: -14.0365557815, 592: 8.01061309144, 593: -36.5864045617, 594: 13.6941776306, 595: -20.0231626543, 596: 12.2137578937, 597: -18.4192904831, 598: 21.6733746123, 599: -12.3024877824, 600: 26.1569197086, 601: 24.7573623679, 602: -29.8934621212, 603: 21.6698275358, 604: -0.976598401916, 605: -1.19364269589, 606: 18.1206347254, 607: -9.07748538877, 608: -2.1062469398, 609: 14.8131946701, 610: 18.4834533345, 611: -2.43828051783, 612: 3.86273788955, 613: -6.89089384903, 614: 11.3994553866, 615: 20.7341685596, 616: -29.8211800471, 617: 1.68389780645, 618: -21.583875727, 619: 7.95416084233, 620: -9.45085952154, 621: 13.4023874938, 622: 2.61844080074, 623: 54.8507263159, 624: -38.5031606962, 625: 27.7887255884, 626: 10.5027159711, 627: -12.7408837918, 628: 7.04405167384, 629: -28.0460018424, 630: -18.5619853879, 631: 2.99830199109, 632: -22.2704122827, 633: -22.3903801399, 634: 54.464801422, 635: -17.970215048, 636: 1.94636780595, 637: 2.16081270865, 638: -15.5813989757, 639: 3.08771876473, 640: 10.1670276107, 641: -6.28191688092, 642: 6.19075564896, 643: -0.299393642244, 644: -0.415371572181, 645: -3.04455113027, 646: -36.0420119058, 647: -8.2850073425, 648: 13.2125920948, 649: 13.9189269286, 650: -5.15270070304, 651: 38.0781969584, 652: -1.49985810955, 653: 3.60277004506, 654: 19.9981912499, 655: -1.95854379762, 656: 18.007561162, 657: -2.04340702234, 658: -2.03466890617, 659: -0.377755204638, 660: -16.6162397902, 661: 1.64621769872, 662: -1.00680110965, 663: -42.9828166702, 664: -2.15061330949, 665: -4.35058279355, 666: -2.13870588032, 667: -0.951064690841, 668: 11.9354113467, 669: -1.46919809988, 670: -4.19729232279, 671: 15.8791785436, 672: -13.1153331082, 673: -30.2701882079, 674: 13.100179948, 675: 39.2794957881, 676: 0.611762075431, 677: -32.2335640096, 678: 1.21758063558, 679: 9.90032561926, 680: 8.12989023335, 681: -0.0346307823681, 682: -0.887056219311, 683: 29.1508073559, 684: -1.22811885276, 685: 22.3098859332, 686: -10.1763569505, 687: 0.630153442637, 688: -1.42420060821, 689: 9.88330350664, 690: -8.85575401972, 691: 0.315808010799, 692: 9.61106373699, 693: -42.9022135919, 694: -11.4431786062, 695: 11.2793433364, 696: -32.9628332643, 697: 15.3160440887, 698: 2.20569510108, 699: -0.363610824882, 700: -9.39334861686, 701: -7.56437047295, 702: -36.9609473818, 703: -49.736242913, 704: 51.7049500019, 705: -2.7801044596, 706: -5.4316278378, 707: -23.0563572713, 708: 51.8473612852, 709: 5.10762555024, 710: 14.530729529, 711: -8.09699140296, 712: 26.2377466842, 713: 1.38725112371, 714: -0.949898027921, 715: -12.8927558113, 716: 35.4968401619, 717: 38.077584412, 718: 7.87810827123, 719: 6.42142362652, 720: 34.780812438, 721: 85.0167590019, 722: -20.5385376199, 723: 5.31292269645, 724: -7.86737472213, 725: 10.7337874194, 726: -19.3197969853, 727: 3.09531698271, 728: -20.8562156598, 729: -0.0406346090185, 730: 2.34417302816, 731: 5.59382760334, 732: -17.634616443, 733: -5.0573336553, 734: -3.65178800346, 735: -2.19938807068, 736: 15.7303330712, 737: -16.2732587127, 738: -23.6162150693, 739: -31.6176370216, 740: -7.05180058338, 741: -48.909922146, 742: -41.2490515303, 743: -8.31685845759, 744: -25.623346867, 745: 13.6227537838, 746: -1.07621360749, 747: -32.3010824701, 748: -3.54224950113, 749: 23.6409740141, 750: -25.4121328929, 751: -2.96573159498, 752: -19.2585534162, 753: 4.20131356413, 754: -2.77148309228, 755: 46.6115464519, 756: 11.3505002603, 757: -28.2823940625, 758: -1.30380763108, 759: 37.2497813088, 760: 13.4464393188, 761: -55.9744597204, 762: 7.07690815452, 763: 12.2357829856, 764: -2.0247593105, 765: 5.764485048, 766: -3.24082131353, 767: -59.7650140532, 768: 0.207337711435, 769: -2.96966345231, 770: 6.40629321665, 771: -3.31690570658, 772: 38.7602692882, 773: 1.43962955543, 774: -3.47031344829, 775: 1.71182196245, 776: -1.92728110205, 777: -24.1258212706, 778: -14.296964425, 779: 1.12368346926, 780: 6.09920178273, 781: -2.14846741075, 782: -0.741760732471, 783: -13.1380650203, 784: 31.5014945393, 785: -3.08147938816, 786: -17.1991326931, 787: -1.23043192706, 788: -0.256389533662, 789: -1.15491941562, 790: -2.57801628281, 791: -31.9320577096, 792: -2.33623813385, 793: -2.445156541, 794: 1.93952796769, 795: 0.160985514084, 796: -1.960278627, 797: -4.70061762784, 798: -0.236343030516, 799: -1.62841561843, 800: -1.16271597519, 801: 3.39825748959, 802: -2.25390561796, 803: -4.5465630986, 804: -6.2452001355, 805: -2.40744927083, 806: -0.901901027139, 807: -0.925893712835, 808: 0.713028928552, 809: 7.69689863364, 810: 33.815477159, 811: 28.7935124617, 812: -18.7999634124, 813: 15.0107497668, 814: -36.9002966004, 815: -25.6749238126, 816: -4.39000594058, 817: -23.9653177885, 818: 27.557435869, 819: -13.1657577165, 820: -11.8219710638, 821: -6.45980369118, 822: -10.8202184578, 823: 3.13092295804, 824: -0.482794968394, 825: 1.09259199012, 826: 6.274691686, 827: 5.11710329771, 828: -0.460200318171, 829: -2.17569902571, 830: -2.52424396814, 831: 14.9246786082, 832: -2.60369298639, 833: 0.9461500683, 834: 26.0377560616, 835: -0.498646623287, 836: 20.0174910237, 837: -4.75749389997, 838: -1.94009040576, 839: -21.8390990506, 840: 16.7700472923, 841: -1.09795579403, 842: -0.144675516551, 843: -0.200277299209, 844: -0.500649246534, 845: -0.593622636245, 846: 42.523149893, 847: -0.403564776275, 848: -2.54844962695, 849: -8.86839288791, 850: -1.68499928757, 851: -0.501261443185, 852: -0.522674428039, 853: -0.571085731503, 854: -2.50908692985, 855: -57.3875178539, 856: -2.16165129318, 857: -0.993126872439, 858: -0.967630199315, 859: -6.45106518835, 860: 0.570675296771, 861: -49.9119723463, 862: -7.29448279252, 863: 78.2244479925, 864: -1.66159373527, 865: 5.68611538729, 866: -3.95039561322, 867: 1.12348564782, 868: 0.0779224206532, 869: 15.3849458184, 870: -0.659647664288, 871: 0.0461191827667, 872: -0.454763069492, 873: 13.9884354918, 874: -39.3588707323, 875: -16.1581141149, 876: -1.45938315629, 877: -0.353429789892, 878: -0.719461084736, 879: -0.482219970338, 880: -0.561745549339, 881: 0.493727885974, 882: 19.8500124159, 883: 1.22729878395, 884: 8.56659179462, 885: -44.3051647649, 886: 13.5864973058, 887: -1.32470182381, 888: -0.0523496970965, 889: 27.0616266242, 890: 12.5587639174, 891: -0.282228327302, 892: -9.48289625501, 893: -1.1797311425, 894: 0.950446559031, 895: 0.970416709619, 896: 6.71550464023, 897: -0.507513785665, 898: -2.2777066434, 899: -0.501584869011, 900: -18.4240041694, 901: -0.396877344435, 902: -0.533028221762, 903: 1.30565564499, 904: -2.33054136289, 905: -15.5048388423, 906: -0.421340287306, 907: -0.507592647059, 908: -0.759608198683, 909: -0.447132224726, 910: -2.36928567176, 911: -2.40232960636, 912: -22.6663952484, 913: 0.374902854242, 914: -0.96068932601, 915: -2.08417955054, 916: 3.50412304555, 917: -38.7994495115, 918: -4.22840282396, 919: -4.77688668368, 920: -13.7841794329, 921: 10.9852193495, 922: -2.23450880679, 923: -1.81563798274, 924: -0.716306562393, 925: 30.245684165, 926: -0.650191704827, 927: 5.27347287074, 928: -5.79541761479, 929: 8.78898410313, 930: -0.68520315728, 931: -13.5424819919, 932: -14.0562725998, 933: -0.687197653339, 934: -7.18384433628, 935: 10.5312527744, 936: 14.4895902173, 937: -3.27534302184, 938: -2.22751039906, 939: -5.10698298686, 940: -27.690568943, 941: -51.4888393374, 942: -43.5600013655, 943: 17.5181452145, 944: -12.6697402469, 945: 0.827278466394, 946: -46.912427548, 947: 51.428649268, 948: 40.1397548959, 949: -22.7458096949, 950: 35.4099743631, 951: 47.2826899157, 952: -23.4501091635, 953: -1.85193486905, 954: -1.40019340833, 955: 49.1890708636, 956: -2.38703354898, 957: -2.52331980088, 958: 0.865796360083, 959: 8.09620546129, 960: -2.40540831521, 961: 12.380749863, 962: -2.75992667428, 963: -2.50994187787, 964: -1.09573257734, 965: 43.1798301225, 966: 7.33301314832, 967: 2.33961657389, 968: -8.68835086159, 969: -0.268328259531, 970: 14.3176584379, 971: -39.7127197833, 972: 41.0400599794, 973: -38.4026712384, 974: -7.48904477912, 975: 27.8771778425, 976: -3.68811885775, 977: -1.57070735775, 978: 6.69423023898, 979: 5.97357820269, 980: -24.5544529458, 981: 12.1424053796, 982: -32.1460048025, 983: -13.0569016351, 984: -32.4820151668, 985: -2.72359121913, 986: 6.20303647261, 987: -1.06646206099, 988: -2.53430414951, 989: -0.577107398849, 990: 17.3598529542, 991: 19.0585781826, 992: -33.8778111805, 993: 30.0513899909, 994: 46.3063966063, 995: -0.22880462851, 996: 50.7723180196, 997: 7.35635936508, 998: -8.56578441539, 999: -6.34482983912, 1000: 35.291871443, 1001: -3.29970892078, 1002: 9.77599534109, 1003: 29.7612178383, 1004: -2.54535459598, 1005: -29.9592685125, 1006: -2.46693798711, 1007: -2.47466370862, 1008: -1.17167650145, 1009: -9.8964529188, 1010: -2.47479868178, 1011: -2.85206884164, 1012: -47.7917201773, 1013: -2.17380386782, 1014: 22.37978918, 1015: -2.51378579222, 1016: -2.71055629857, 1017: 42.218372479, 1018: -2.44151328159, 1019: -2.44311761777, 1020: 20.1000135115, 1021: -11.1214427012, 1022: -69.3913686653, 1023: 59.3111845963, 1024: -27.7733963775, 1025: -1.40790716027, 1026: 52.6476260886, 1027: 2.32766334373, 1028: 63.6350532692, 1029: 50.5307946427, 1030: 1.77049075644, 1031: -2.54478646097, 1032: -2.05455683258, 1033: -5.33518752523, 1034: 29.1727538202, 1035: 9.83823458682, 1036: -6.01301274313, 1037: -2.53844125003, 1038: 3.73727318119, 1039: -4.14608442204, 1040: 33.7367339733, 1041: 8.1485644434, 1042: -26.3015116149, 1043: 33.7547462513, 1044: 71.6996699872, 1045: -51.683578403, 1046: -68.0148722378, 1047: 40.9943450333, 1048: -0.0860739963094, 1049: 15.4060646668, 1050: -4.42539634534, 1051: 6.95680072099, 1052: -1.47745714519, 1053: -5.73784846347, 1054: -1.49436571517, 1055: 9.0000262659, 1056: 8.22041283397, 1057: 4.12024091484, 1058: -10.523750244, 1059: 2.44024598933, 1060: 6.75185455744, 1061: 6.40327501036, 1062: 8.01610323188, 1063: -3.77669976771, 1064: 2.47687130879, 1065: 12.426423283, 1066: -9.45401501568, 1067: -0.397516976942, 1068: -1.01229285588, 1069: -8.03297705194, 1070: 11.6905452939, 1071: 0.474154049753, 1072: 0.612617900572, 1073: -5.16450414697, 1074: -3.25357604992, 1075: 2.93069646259, 1076: 12.5864873912, 1077: -2.79687811736, 1078: -2.22724266427, 1079: 0.545112329938, 1080: 17.1981301639, 1081: 8.64720958462, 1082: 12.0433710007, 1083: -1.01153817672, 1084: 5.76990463189, 1085: -1.06489751583, 1086: 4.47143940039, 1087: 9.71023650098, 1088: -4.97451879561, 1089: 0.673548251734, 1090: -1.22443825317, 1091: 6.26364137528, 1092: -8.48254977526, 1093: 4.46116827983, 1094: 5.54825394699, 1095: 0.771344467006, 1096: -3.87192058923, 1097: -0.891811821646, 1098: 3.26848402267, 1099: 4.06733906856, 1100: -0.900208550584, 1101: 0.975414978366, 1102: -2.07646961911, 1103: -1.13277603156, 1104: -6.10895871293, 1105: -0.554554144772, 1106: -8.3113489849, 1107: -0.742811952699, 1108: 12.5765722823, 1109: -3.5127210374, 1110: 20.9829371214, 1111: -0.888757070155, 1112: -8.90671622101, 1113: -1.76344963279, 1114: -11.3714032328, 1115: -5.3068277269, 1116: -4.8896250338, 1117: -0.317344745907, 1118: -0.38954821509, 1119: -2.90517687928, 1120: -0.791075745912, 1121: -5.45861671626, 1122: 0.6672506639, 1123: -1.6080847433, 1124: 4.21488059406, 1125: -4.05928993789, 1126: 1.65671973096, 1127: 13.2456505915, 1128: -1.05379320899, 1129: -7.68102525857, 1130: -4.75939847441, 1131: 15.2328004793, 1132: 1.09792277745, 1133: -6.36488066462, 1134: -0.975750722842, 1135: -8.94333084328, 1136: -1.09987754687, 1137: -5.375439962, 1138: 0.0108313920878, 1139: -0.932144006278, 1140: 3.86252264095, 1141: -0.696735064947, 1142: -1.58690634977, 1143: 3.42722564681, 1144: -8.96279758669, 1145: 0.425984012634, 1146: -1.29949868604, 1147: -0.560079894546, 1148: 6.32040001137, 1149: -5.63192036381, 1150: 6.2947274745, 1151: -0.69178161343, 1152: -1.08054604089, 1153: -4.13646549706, 1154: -0.384523739302, 1155: -2.1310007282, 1156: -1.07307856496, 1157: -6.6156527827, 1158: 0.125576082394, 1159: -10.5359959637, 1160: -3.29875815962, 1161: -7.13509854057, 1162: 21.5066732504, 1163: 7.74497208796, 1164: -5.21391964076, 1165: 1.23799463968, 1166: 2.42972896571, 1167: -0.79688300377, 1168: -4.84195746984, 1169: 0.184172601357, 1170: -6.14348196688, 1171: -1.11931912144, 1172: 2.5682086465, 1173: -0.240811245252, 1174: -0.569774653466, 1175: -8.42194481576, 1176: -14.2081685752, 1177: -0.253570342436, 1178: 0.0109888728053, 1179: -0.214443875286, 1180: 9.12284232292, 1181: -0.802400574014, 1182: 2.90242890842, 1183: -14.2831760517, 1184: 1.47101541307, 1185: 18.8058941261, 1186: 3.10492602765, 1187: -2.60069111159, 1188: -3.24114424107, 1189: 9.79447843243, 1190: -1.32340386282, 1191: 0.24307869318, 1192: -6.81720788329, 1193: -0.242530220342, 1194: -0.166511315174, 1195: -6.67498647889, 1196: -0.247374696423, 1197: -0.719904370321, 1198: -8.76410184065, 1199: -0.701363288744, 1200: -0.124360899588, 1201: -0.0366551005032, 1202: -0.144737857234, 1203: -1.33480143835, 1204: 18.7118166573, 1205: 1.34640152148, 1206: -0.547385146663, 1207: -0.617085817853, 1208: -2.35186797559, 1209: -3.03548878689, 1210: 22.0249788834, 1211: 2.50685375972, 1212: -13.2917184965, 1213: -18.6585599906, 1214: -1.08903538851, 1215: 0.491409817558, 1216: -2.2114499374, 1217: 0.586917463398, 1218: 7.19197451525, 1219: -0.493334838622, 1220: 24.777684439, 1221: -0.126225761868, 1222: -17.1627949399, 1223: 10.5871260688, 1224: 10.7241112252, 1225: 0.255558073389, 1226: -0.197141592315, 1227: -0.365675236465, 1228: -0.447370684993, 1229: -0.0875796364838, 1230: -1.43050564986, 1231: 0.418706571337, 1232: -2.63589781371, 1233: -3.84075324997, 1234: -10.6566006521, 1235: -7.51620390178, 1236: -3.59162385253, 1237: 1.80488095997, 1238: -1.87938953454, 1239: -10.361001645, 1240: 0.146212977818, 1241: 3.00447814301, 1242: -0.0304996744201, 1243: -0.719020291682, 1244: -2.19665094101, 1245: -0.140280627128, 1246: -0.266562742354, 1247: -0.676252168357, 1248: -0.211432785788, 1249: -1.56320810318, 1250: -0.303277010573, 1251: -0.136961149875, 1252: -10.2863054649, 1253: -0.85551859703, 1254: 14.179320366, 1255: -0.236624242142, 1256: -0.124264964547, 1257: 0.537010849436, 1258: -0.0825302241068, 1259: 0.932816457258, 1260: -8.2236662872, 1261: 2.25873821605, 1262: 0.239742668649, 1263: 0.655169453504, 1264: 8.61150819005, 1265: 1.53803712656, 1266: 6.84428997168, 1267: 18.6634210571, 1268: -4.22149560989, 1269: -0.773548488356, 1270: 3.89471955822, 1271: -1.17888755107, 1272: 1.43609251515, 1273: -0.590790030895, 1274: -13.8648200125, 1275: -4.30422224927, 1276: 4.49922643555, 1277: 3.46993922268, 1278: 2.73666959397, 1279: 0.0609240699121, 1280: -0.835893173185, 1281: 4.74194847742, 1282: 0.391778720458, 1283: 5.25527277168, 1284: 3.99319785453, 1285: 11.7655730605, 1286: -0.864768434504, 1287: -0.226167849705, 1288: 2.75513621016, 1289: -16.1647829146, 1290: -21.5225436211, 1291: -0.7581273002, 1292: 2.73131817115, 1293: -15.1827743008, 1294: -8.10022490315, 1295: -8.96398232958, 1296: 12.72951436, 1297: -2.79401712943, 1298: -9.58126423813, 1299: 2.67921153516, 1300: 1.49679337285, 1301: 16.9444768979, 1302: -7.60616422268, 1303: -0.87886121225, 1304: -15.1075375598, 1305: -0.702955600784, 1306: -0.684848207338, 1307: -4.9192296078, 1308: 0.206538492429, 1309: -0.884182511794, 1310: 0.706180240575, 1311: -0.101627411341, 1312: -1.44823152385, 1313: -18.8742563243, 1314: 6.95620158468, 1315: 2.72728238, 1316: 4.12502744069, 1317: -5.76289985608, 1318: 0.63144825865, 1319: 13.2499405728, 1320: 9.61639424449, 1321: 28.1412316469, 1322: 27.6002663322, 1323: 20.3447383098, 1324: 5.46599654091, 1325: -9.01212264037, 1326: 1.26478398036, 1327: 0.463361891122, 1328: 10.1464482528, 1329: -10.8819404705, 1330: -4.18106546434, 1331: 1.70191596035, 1332: 9.77611858429, 1333: -17.072618443, 1334: -0.525125780838, 1335: 4.13794159042, 1336: 2.26769681569, 1337: -3.29359863979, 1338: 3.63575110011, 1339: 12.1032549782, 1340: -7.17960785687, 1341: 1.01277102257, 1342: 10.8980831064, 1343: -3.90959877997, 1344: -29.7742785144, 1345: -5.09849267677, 1346: -1.11244325989, 1347: 2.24844442068, 1348: -3.81638702716, 1349: -2.92926316417, 1350: -2.7934316607, 1351: 5.77072823945, 1352: 9.41608276885, 1353: -1.56877518636, 1354: -6.67808382442, 1355: -0.73850722001, 1356: -0.562731540202, 1357: 0.813704714434, 1358: 15.4845050022, 1359: -0.479654636682, 1360: -1.46575773088, 1361: 1.14929158591, 1362: -0.908699238726, 1363: -11.623391282, 1364: -0.774747926238, 1365: -0.763277031352, 1366: -17.9880067733, 1367: -0.693903818575, 1368: 1.39259698417, 1369: 5.27463100965, 1370: 13.2920502844, 1371: 7.56430677789, 1372: -10.6990502186, 1373: 1.46035055224, 1374: -8.73394809209, 1375: 10.5508372729, 1376: -10.0787989141, 1377: -1.41109586089, 1378: -1.71690803534, 1379: 1.89453907395, 1380: -2.08142126077, 1381: -0.627643653128, 1382: 7.19476576882, 1383: -6.10735692213, 1384: 1.1893742018, 1385: -8.77313821478, 1386: -0.962771272778, 1387: -5.72439004791, 1388: -0.0937280736464, 1389: 11.0302240192, 1390: 5.36913024556, 1391: 7.16934839738, 1392: -4.49756683746, 1393: -30.3953966027, 1394: 8.66181591015, 1395: -1.56280911571, 1396: -6.37755144425, 1397: -0.49977934052, 1398: -66.9460264758, 1399: 30.292425876, 1400: 9.76035055044, 1401: -4.01078834611, 1402: 17.0413035583, 1403: 42.868745035, 1404: -3.25748371502, 1405: 16.3995561831, 1406: -3.30110499674, 1407: 4.82488762813, 1408: 14.7559115526, 1409: -7.35404387102, 1410: 2.00054206565, 1411: -12.3647522126, 1412: 7.61702213784, 1413: 58.9900156703, 1414: -22.2017946456, 1415: -13.0555630404, 1416: 31.9237680605, 1417: -4.94729278551, 1418: -4.46753100526, 1419: 24.9688222196, 1420: 28.5957049543, 1421: -2.05522858309, 1422: 10.6249328782, 1423: 9.55300370818, 1424: 0.58274666616, 1425: 14.8281213572, 1426: -19.0995098803, 1427: -36.3906029547, 1428: 34.8948626808, 1429: -14.4081786212, 1430: -18.9886771789, 1431: 20.036124883, 1432: 3.30066677683, 1433: -0.697667059208, 1434: 5.74132200665, 1435: -0.0479411466893, 1436: 14.3237323579, 1437: 4.41952886331, 1438: -8.25519814813, 1439: 3.86747984617, 1440: 7.61057842284, 1441: 1.86448382338, 1442: 4.28107263708, 1443: 14.0200605163, 1444: 9.05158254101, 1445: -11.7456778281, 1446: -1.25371405106, 1447: 7.35422368724, 1448: -4.46502225857, 1449: -3.72941082982, 1450: -9.73935642464, 1451: -4.43687249682, 1452: -3.1462513362, 1453: -16.5357954691, 1454: -4.54908295299, 1455: -3.12797706759, 1456: 5.81902696871, 1457: -14.3412928235, 1458: -2.97315094691, 1459: -40.6248706972, 1460: -4.9217753675, 1461: 33.3012784703, 1462: 4.40278102648, 1463: 10.9832593348, 1464: -9.08256944569, 1465: -8.21443406065, 1466: -2.6415879145, 1467: -3.16986412932, 1468: 17.6838309027, 1469: -3.65012690474, 1470: 4.60481565896, 1471: 12.0654215804, 1472: -24.1471487562, 1473: -1.69170189583, 1474: 6.38504292735, 1475: -64.9188062459, 1476: -18.0335848587, 1477: -1.15040638532, 1478: -3.37741617036, 1479: -4.41422277003, 1480: 9.38317910181, 1481: 12.2695834696, 1482: -1.54058876948, 1483: -4.46808811704, 1484: -19.78118529, 1485: -3.42956196789, 1486: 9.24164016611, 1487: 2.29306403363, 1488: -3.71132954508, 1489: -19.9094562429, 1490: -3.10005311207, 1491: -3.26734200804, 1492: -3.17379913173, 1493: 14.1156986637, 1494: -2.99932835464, 1495: 6.03966185549, 1496: 33.2518569643, 1497: -19.4617950432, 1498: -3.06560407387, 1499: 4.12193372765, 1500: -1.85040799132, 1501: -15.278385942, 1502: 31.52625815, 1503: -3.56216685518, 1504: -3.81361238011, 1505: -0.436804929713, 1506: 1.86289068004, 1507: -2.30192279993, 1508: -10.9228194386, 1509: -2.50850248332, 1510: -35.1239329117, 1511: 66.3554597178, 1512: -16.5611179566, 1513: 3.23982467921, 1514: 1.29979455998, 1515: 6.85373485855, 1516: 18.3445664649, 1517: 13.263065613, 1518: -28.6933638917, 1519: -15.399528441, 1520: -7.69799453695, 1521: -8.60074119699, 1522: -0.702844558229, 1523: -1.05878552746, 1524: 0.0104885561737, 1525: -6.94629297838, 1526: -0.42418430276, 1527: -3.37402690437, 1528: -1.28441326673, 1529: -18.29037829, 1530: -0.660261529614, 1531: -0.372285082365, 1532: -34.2296445727, 1533: 0.905937735836, 1534: -10.3828147493, 1535: -0.354051815561, 1536: -0.248239870652, 1537: -17.1987535267, 1538: -25.4422325648, 1539: -6.06388900734, 1540: -0.778941101686, 1541: 13.5496398812, 1542: -0.431092201257, 1543: -0.5757302453, 1544: -11.5248607133, 1545: -0.474163140763, 1546: -3.55120981977, 1547: -3.70689801718, 1548: -0.486071079093, 1549: -0.529123412332, 1550: -0.536186292038, 1551: -0.414721621096, 1552: -3.38004472321, 1553: -38.2129539186, 1554: 0.159831545476, 1555: -0.590459342681, 1556: -1.71708049816, 1557: -9.47424599673, 1558: 4.55525039745, 1559: -5.15221619867, 1560: -4.03025920681, 1561: 5.92823044764, 1562: -24.6575036107, 1563: 6.03672546159, 1564: -1.83019938694, 1565: 20.7919105189, 1566: 3.40042624846, 1567: -4.23598034098, 1568: -1.2195279517, 1569: -29.7138733979, 1570: -0.603979058388, 1571: -7.52410564345, 1572: -16.4061007514, 1573: -8.91925318008, 1574: 0.0152909999628, 1575: -0.65295484276, 1576: -0.694048775001, 1577: -2.64329395797, 1578: 0.145383069822, 1579: -17.3877721707, 1580: -21.6522458268, 1581: 5.67392836499, 1582: 22.4610685428, 1583: -23.0176372024, 1584: -1.10404734936, 1585: 20.1961127781, 1586: 3.33284338834, 1587: -0.441001585132, 1588: -18.7337245068, 1589: -0.195380526476, 1590: -29.7789558677, 1591: 7.44416486266, 1592: 0.133247749795, 1593: -7.30538417816, 1594: -20.6516222473, 1595: -0.471525220819, 1596: -3.14878623874, 1597: -0.429855844779, 1598: -32.102255729, 1599: -0.590656482979, 1600: -0.403457691677, 1601: 22.0361780396, 1602: -3.25818503429, 1603: 3.60202688908, 1604: -0.545131836796, 1605: -0.512529143388, 1606: -2.01951967653, 1607: -0.681321069902, 1608: -0.911089745441, 1609: -3.13886763981, 1610: 43.3639546909, 1611: -1.49938032858, 1612: -0.496657129438, 1613: -7.23518701825, 1614: 6.58274807269, 1615: -27.3671119973, 1616: -5.35364141641, 1617: 4.81678575144, 1618: 0.86066236524, 1619: -0.0351370495132, 1620: -2.61345580145, 1621: 22.8180932288, 1622: -4.04669825176, 1623: 4.06661866141, 1624: -19.8723978238, 1625: -1.97309717389, 1626: 1.58950680115, 1627: -34.3957890239, 1628: -1.09845084138, 1629: -14.9511400049, 1630: -4.04766208944, 1631: 5.94594020754, 1632: -10.9436164704, 1633: 14.8196887172, 1634: 3.14949403101, 1635: -3.96336002475, 1636: -3.35302646823, 1637: 15.3071301987, 1638: 5.39559972289, 1639: 5.62600662474, 1640: -12.0349437607, 1641: 26.3340661968, 1642: 7.28434848504, 1643: -6.26300709136, 1644: -20.6688320266, 1645: 16.1407371466, 1646: -33.4999491337, 1647: 22.3324220136, 1648: 25.3473893306, 1649: -5.37454752404, 1650: 2.15565055499, 1651: -14.2246184193, 1652: -1.4045645467, 1653: 33.8133947636, 1654: -2.93173612315, 1655: -3.47505057254, 1656: -3.65869028808, 1657: 0.388948854103, 1658: -1.66750694411, 1659: -2.85848086271, 1660: -9.44602977197, 1661: -3.31242246851, 1662: -35.7166802408, 1663: -3.59160051052, 1664: -4.49324637782, 1665: 24.6138255204, 1666: -7.65807094899, 1667: 9.2881755225, 1668: -21.1650707966, 1669: 9.52583732323, 1670: 15.3044182453, 1671: 16.5008258158, 1672: -5.87534594624, 1673: 1.97946586023, 1674: 8.07515713225, 1675: 1.71561010072, 1676: -12.0060534741, 1677: 32.9149042105, 1678: 8.42364656631, 1679: -6.50292863763, 1680: -10.0524680353, 1681: -6.00055388155, 1682: -9.91306463866, 1683: -3.58563267415, 1684: -14.2714857133, 1685: 19.5327202809, 1686: -6.0496344849, 1687: -5.59307576728, 1688: -23.2826733507, 1689: 13.9670212338, 1690: -22.5930198997, 1691: 3.99570715578, 1692: 18.58150475, 1693: 35.2993345661, 1694: 16.8483247651, 1695: -1.5080340421, 1696: -3.82188865421, 1697: -23.9525981354, 1698: 25.8097788176, 1699: 5.68724128573, 1700: -1.56447251955, 1701: 17.7312995131, 1702: -3.36337815767, 1703: 38.7592987211, 1704: -3.19476493179, 1705: -3.17707264821, 1706: -3.21585797662, 1707: -11.2426962847, 1708: -3.85030127662, 1709: -3.79185583673, 1710: -36.6895476939, 1711: -3.42827913721, 1712: -39.715042112, 1713: -3.68099039613, 1714: -3.69785069223, 1715: -10.2874048896, 1716: -0.210821533309, 1717: 0.86481043733, 1718: -2.27386591972, 1719: -18.0800756959, 1720: 43.1351900266, 1721: -5.92442571824, 1722: 0.115804605625, 1723: 11.9310030082, 1724: 19.473293999, 1725: 5.82985320812, 1726: 4.05006888863, 1727: -4.28448120554, 1728: -26.7687377003, 1729: -2.63097141827, 1730: -6.30026444857, 1731: 11.3403766979, 1732: 80.5975888019, 1733: 13.9565929293, 1734: 22.6055991975, 1735: -3.58397167196, 1736: 13.1135755467, 1737: 15.9091829867, 1738: 25.5871435614, 1739: -2.15165799597, 1740: -22.2435726887, 1741: -20.3312838695, 1742: 84.0674913368, 1743: -39.6465567235, 1744: -0.442449226323, 1745: 25.3486546921, 1746: -0.090540202053, 1747: -18.9036386338, 1748: 14.3371086026, 1749: 2.70625774964, 1750: -16.4241057104, 1751: -8.88055356211, 1752: -6.97221960302, 1753: 21.3303093492, 1754: 9.08725154741, 1755: 9.2745966406, 1756: -15.5071462148, 1757: -9.33865722155, 1758: -3.84158423173, 1759: 15.5947758212, 1760: 7.41051429946, 1761: -10.9027355506, 1762: -13.8358315173, 1763: -24.0649634687, 1764: 7.28036646907, 1765: -21.0065149803, 1766: -2.3139759144, 1767: 3.65746366776, 1768: -5.91923941218, 1769: -15.005403309, 1770: 9.76865529433, 1771: 5.10319197873, 1772: -29.5618634767, 1773: 10.4747953191, 1774: -8.0556630975, 1775: -6.64124465375, 1776: 15.2237125403, 1777: -7.78945647363, 1778: 4.46150335235, 1779: -6.71884951498, 1780: 4.18195090263, 1781: -0.299967148431, 1782: 7.96162474742, 1783: 5.68724135329, 1784: 7.23422140954, 1785: -4.4171386254, 1786: 1.76901316547, 1787: 0.82319027381, 1788: -21.4031470576, 1789: 23.5965164127, 1790: -0.515151861249, 1791: -16.889775956, 1792: -7.51346393454, 1793: 0.665034707303, 1794: -26.4994919305, 1795: 0.61879963686, 1796: 4.00593484626, 1797: -1.1519177868, 1798: 0.584643071072, 1799: 2.43602323332, 1800: 1.18312050588, 1801: 0.626915872338, 1802: -8.60869698043, 1803: -0.442605283101, 1804: -14.20775797, 1805: 8.87878241888, 1806: -2.62087143151, 1807: -38.4653037259, 1808: 1.70775803222, 1809: 19.1877174426, 1810: 5.17267864485, 1811: 0.753202779111, 1812: 2.15300895923, 1813: 0.728243375568, 1814: 6.2243219166, 1815: 0.734449827894, 1816: 0.787550714114, 1817: -0.73700924444, 1818: 0.697343390362, 1819: 8.05975463213, 1820: -0.69829233346, 1821: 0.63193666135, 1822: -0.763652613856, 1823: 0.682318862002, 1824: 3.01288495021, 1825: 2.65728558277, 1826: 0.54991389289, 1827: 1.32350248577, 1828: 0.70845024947, 1829: 0.536196237849, 1830: 0.0457971451435, 1831: 0.7886675372, 1832: 0.705465474371, 1833: -1.54963391982, 1834: 0.715003451035, 1835: 0.583998955628, 1836: 0.639035192815, 1837: 0.6374760876, 1838: -19.838228356, 1839: 0.52827196525, 1840: 0.668005882897, 1841: 0.563591049853, 1842: 0.602374191755, 1843: 0.662051811983, 1844: -19.002480963, 1845: 0.566407327343, 1846: -12.2109411577, 1847: 0.735245400777, 1848: 0.987732027588, 1849: 0.731733615611, 1850: 0.81217695894, 1851: -22.9941392886, 1852: 0.689639107059, 1853: 0.702509791716, 1854: 0.672612480319, 1855: -0.324738042686, 1856: 0.868170600193, 1857: -13.138231164, 1858: 12.5503687844, 1859: -7.11772845058, 1860: -0.272067867338, 1861: 6.07305635089, 1862: 8.03534578072, 1863: 1.65274552829, 1864: 13.2992939524, 1865: 1.23981725526, 1866: 14.4823186892, 1867: 10.1858691913, 1868: 13.1650582583, 1869: 7.3855228519, 1870: -2.46429902945, 1871: 0.0704774497122, 1872: 0.550975918253, 1873: 1.22102306538, 1874: -3.83860855249, 1875: 0.00692468721089, 1876: 0.778354288075, 1877: 0.166980930275, 1878: -5.39270989826, 1879: -0.784062208473, 1880: 3.15729576422, 1881: 20.9973350975, 1882: 0.558531917589, 1883: 6.06612498443, 1884: 0.79496636302, 1885: 7.84683269554, 1886: -0.842555844911, 1887: 6.39982130519, 1888: -4.34852246222, 1889: -0.0554450727689, 1890: -4.51395176605, 1891: 0.0737776235171, 1892: -0.0246302531444, 1893: 17.9821813059, 1894: -0.0236713651505, 1895: 0.641891032164, 1896: -1.88824267008, 1897: -0.0888815506108, 1898: -0.0829839154197, 1899: 0.0942606677312, 1900: 0.00808780804705, 1901: 0.782074087367, 1902: -10.5932565386, 1903: 2.31795188431, 1904: 0.189997952806, 1905: 0.350569131057, 1906: -0.238555973563, 1907: -0.0645557108194, 1908: -1.37864644016, 1909: -1.03959707793, 1910: 7.68098194811, 1911: 10.574056015, 1912: -2.65173171121, 1913: -1.9303691002, 1914: -5.39654072841, 1915: -1.34544172093, 1916: -3.03222314585, 1917: 0.124105126833, 1918: -7.0702200565, 1919: -0.0574332511025, 1920: 0.0819313627843, 1921: -15.0882375411, 1922: 0.069392508714, 1923: -0.0771325982489, 1924: 0.0899308065746, 1925: 0.13639048835, 1926: -0.00333048042141, 1927: 0.0892815683326, 1928: 0.0656129451901, 1929: -9.60209676648, 1930: 0.256571418054, 1931: 4.2295542045, 1932: -3.85357653693, 1933: -3.11441596158, 1934: -1.97139511879, 1935: 1.5450414972, 1936: -0.15421588702, 1937: 6.28986699105, 1938: -0.0624658668476, 1939: -6.22128316894, 1940: 0.148506288049, 1941: -1.30017121775, 1942: -1.41468674973, 1943: -1.61149256063, 1944: 0.0660218743179, 1945: 0.722242560134, 1946: 0.0297033613779, 1947: -7.61379353106, 1948: 0.0776332045013, 1949: -0.118087179479, 1950: -2.89359851893, 1951: 0.654454812541, 1952: -0.37178164766, 1953: -0.0520940213372, 1954: -0.147332787783, 1955: 0.854526900997, 1956: -0.075144667891, 1957: 0.600334591388, 1958: -1.37575073476, 1959: -4.27306891061, 1960: 0.0303610656685, 1961: 0.234454479105, 1962: 12.4182985805, 1963: 1.67726944651, 1964: -6.43943186859, 1965: 2.1475227081, 1966: 5.67297371282, 1967: -2.69055732517, 1968: -1.28923415513, 1969: 0.797515900654, 1970: -2.13019002301, 1971: 0.00809729740425, 1972: -7.69825716716, 1973: 0.863416654499, 1974: -0.351044473288, 1975: -0.148262469225, 1976: -24.9244865839, 1977: -0.0295777837446, 1978: 4.77782207198, 1979: 4.70247018619, 1980: 0.648118529792, 1981: 0.879112489622, 1982: 4.45893956232, 1983: -3.22274793142, 1984: 0.633001385112, 1985: 0.787198934173, 1986: 8.74009195054, 1987: -5.95547928743, 1988: 3.11795511176, 1989: -12.7932307481, 1990: -8.5964385555, 1991: -5.53053415538, 1992: 15.8719948699, 1993: -19.757289704, 1994: 10.2215742627, 1995: 12.1836127473, 1996: -14.7301399833, 1997: 5.12971086628, 1998: 10.0119558844, 1999: 9.04660396859, 2000: 0.799249376739, 2001: 0.564476235993, 2002: -6.49489402718, 2003: 0.71418411161, 2004: 0.738732361281, 2005: 1.28248446605, 2006: 0.395301248288, 2007: 0.612332501025, 2008: 1.19909896404, 2009: 0.755159251108, 2010: 0.571119728462, 2011: 33.3830603522, 2012: -1.30467451151, 2013: -2.66788478266, 2014: -8.09386969181, 2015: 1.0036063547, 2016: 0.61721664355, 2017: -7.60192065385, 2018: 5.92780120064, 2019: 8.46759226287, 2020: 3.09895898791, 2021: -7.47646068628, 2022: 1.25453375971, 2023: 6.97084488513, 2024: 1.62610835343, 2025: -4.63686921007, 2026: -0.99712991366, 2027: 7.2027263736, 2028: 0.786478164414, 2029: -12.3349380582, 2030: -5.10114989181, 2031: 8.458774564, 2032: 0.728703156587, 2033: -0.141861672969, 2034: 0.73534090937, 2035: 2.37006583515, 2036: 2.0260360764, 2037: -8.03674205763, 2038: -0.250548034002, 2039: -2.95036307353, 2040: 20.4605248666, 2041: 10.3077671564, 2042: -12.5203330668, 2043: -5.24865104689, 2044: 0.268643579825, 2045: -1.32322452726, 2046: -0.139635018585, 2047: 3.87212973342, 2048: -0.0396907963252, 2049: 1.07659980086, 2050: 7.21089980611, 2051: 0.699704018313, 2052: -3.76460240344, 2053: 0.527853377611, 2054: 0.712292014917, 2055: 0.760598835002, 2056: 0.506010099805, 2057: 0.582167860402, 2058: 0.706030859629, 2059: -12.4050010378, 2060: 0.811449768655, 2061: -6.32345957142, 2062: 0.723954284496, 2063: 0.665467149854, 2064: 13.538786449, 2065: 0.827776629201, 2066: 0.828734448856, 2067: -4.25511916456, 2068: -6.60084278435, 2069: -9.92934214529, 2070: -5.00779786104, 2071: 16.4023863064, 2072: -2.60743519901, 2073: -16.0218495887, 2074: -11.8867889377, 2075: 12.3180942429, 2076: -26.9401686326, 2077: 3.50631861196, 2078: 0.689518997127, 2079: -3.55302362229, 2080: -0.504190092678, 2081: -24.287404579, 2082: 2.16516531793, 2083: 15.4815874181, 2084: 0.67778157254, 2085: -0.660245303365, 2086: -3.36226061988, 2087: 0.748527829557, 2088: -10.4334149163, 2089: -5.60151002785, 2090: 8.59948467127, 2091: -15.6828562267, 2092: -3.32745054927, 2093: -4.385306322, 2094: 11.3750055916, 2095: 0.243344299059, 2096: 5.75304711079, 2097: -8.36884828015, 2098: -17.6984727354, 2099: 1.66038743309, 2100: 26.062133203, 2101: -15.7174860331, 2102: 2.67089654075, 2103: -42.4618878348, 2104: 57.5246864432, 2105: 19.8030232413, 2106: 18.3635245245, 2107: -19.9173949201, 2108: 18.5576834248, 2109: 41.6296607794, 2110: 72.2101451569, 2111: -81.1782552835, 2112: 100.0, 2113: -56.5119566678, 2114: -26.0194529252, 2115: -6.86807550023, 2116: 44.906222968, 2117: -27.2635439231, 2118: 70.27703397, 2119: -3.04855125495, 2120: -27.0341089389, 2121: 61.1461126258, 2122: -39.3518628908, 2123: -26.7704769924, 2124: -13.6652003767, 2125: -30.6575012983, 2126: -9.28314050711, 2127: 9.20442713999, 2128: 0.160647773091, 2129: -12.9405984519, 2130: 6.77696282973, 2131: -12.6037190717, 2132: -43.7667145051, 2133: 42.9413606841, 2134: -30.055705794, 2135: 68.0366142818, 2136: 0.691848082214, 2137: 26.4788780795, 2138: -48.2083029702, 2139: -20.9913700421, 2140: -12.7506795628, 2141: 25.1320026684, 2142: 7.0123408886, 2143: 5.40895907949, 2144: 5.17894503633, 2145: 14.0628454493, 2146: -36.6998590435, 2147: 0.00469415301573, 2148: 0.83262251504, 2149: -16.4659600527, 2150: 1.68851140459, 2151: -36.8182587522, 2152: 5.73317769211, 2153: -22.0333735767, 2154: 24.3350763892, 2155: -25.7018751116, 2156: 0.285673825802, 2157: -37.2174428919, 2158: -39.7999412107, 2159: -16.5712629437, 2160: -0.291862549036, 2161: 22.1619414192, 2162: -0.133999110095, 2163: -4.48792875024, 2164: -0.374195172279, 2165: -0.277276758009, 2166: 6.44276509412, 2167: 0.0174086489748, 2168: 55.2585187782, 2169: 4.06606875599, 2170: 0.359576263262, 2171: 2.7960729078, 2172: -3.54557396716, 2173: -42.7744698199, 2174: -36.5267582373, 2175: 0.766507803612, 2176: 15.360985153, 2177: -0.154536834107, 2178: -11.9957514591, 2179: 53.9680754331, 2180: -0.016783963962, 2181: -0.280108467626, 2182: -14.3314357272, 2183: -0.209392920052, 2184: 24.3428815711, 2185: 20.2198647638, 2186: 0.0636243974349, 2187: -19.9313199001, 2188: -1.71916171802, 2189: -0.358017877841, 2190: 3.82441608428, 2191: 0.00502402786027, 2192: -0.497415757, 2193: -1.14221098664, 2194: -17.4795300077, 2195: 31.7094526359, 2196: 9.58548151884, 2197: -66.8386625909, 2198: 1.14225450707, 2199: 0.101006437541, 2200: 81.7287775047, 2201: -0.135634121038, 2202: -0.0372818611183, 2203: -0.83006442193, 2204: 1.77559977318, 2205: -2.30552925153, 2206: 13.9480830149, 2207: 72.0084419126, 2208: 11.8948992473, 2209: -47.4845389338, 2210: -45.4785241273, 2211: -20.6260441213, 2212: -5.03393785399, 2213: -6.48106542267, 2214: 6.79857098292, 2215: 16.3316880916, 2216: 43.3592461582, 2217: -0.414449100459, 2218: 33.013480642, 2219: -14.9440157534, 2220: 0.175682669703, 2221: -0.938273738624, 2222: 6.53181008188, 2223: 76.9219122082, 2224: 0.0261057323246, 2225: -0.273091000368, 2226: -1.23328949913, 2227: -40.0475161781, 2228: -0.436063915397, 2229: -5.75603033346, 2230: -11.4181781763, 2231: -2.34134782264, 2232: -22.7549434216, 2233: -2.07611644003, 2234: -3.40367968888, 2235: -51.8878787159, 2236: 11.0364428029, 2237: -50.4202014556, 2238: 0.954451917756, 2239: 0.17871769826, 2240: 0.24568403559, 2241: -0.00306530578053, 2242: -10.6951645894, 2243: 0.209605478852, 2244: -0.134258467499, 2245: -2.69318263164, 2246: 0.209152724305, 2247: 0.104384870585, 2248: -0.390284008006, 2249: 0.0166794385622, 2250: -0.22584388706, 2251: 74.6121107434, 2252: 2.09567674403, 2253: -0.735621966146, 2254: -0.208631821053, 2255: -36.5905284753, 2256: 3.83824278349, 2257: -15.8227327839, 2258: 4.78266094503, 2259: -25.9895106564, 2260: 16.901298864, 2261: -15.9158417194, 2262: 0.606533066315, 2263: 2.12206531425, 2264: -42.2772268502, 2265: -21.4476978219, 2266: 0.369957435458, 2267: -62.9301337838, 2268: 0.198089427044, 2269: 14.6346892454, 2270: 10.9104240117, 2271: 61.6523582618, 2272: 0.344415990578, 2273: 0.245555371965, 2274: 0.0663598830264, 2275: 0.258681507312, 2276: -0.0602403895849, 2277: 4.04983130831, 2278: 47.3596067999, 2279: 3.83592561006, 2280: 1.60555182271, 2281: 67.229276739, 2282: 17.1196799434, 2283: -7.56611244952, 2284: 10.4103685384, 2285: 13.9904036017, 2286: 17.8339334088, 2287: 0.318571268132, 2288: 11.6421208705, 2289: 0.562835509348, 2290: 2.04229082936, 2291: 1.49806787517, 2292: -10.6317357965, 2293: 0.136523130782, 2294: -0.141450633378, 2295: -0.0666910331844, 2296: 1.28426080478, 2297: 0.261443879981, 2298: 0.0944568896444, 2299: -23.2098107816, 2300: -0.17398139204, 2301: -100.0, 2302: 0.0315233779372, 2303: 0.168590777442, 2304: -0.276284874672, 2305: 0.126732754221, 2306: -2.55760938606, 2307: 2.78381592989, 2308: -1.20305742884, 2309: 0.0251490164646, 2310: 0.170386737321, 2311: -26.6510297586, 2312: 21.6607684876, 2313: -51.302018674, 2314: -14.0528488183, 2315: 4.39357515581, 2316: -7.53495320312, 2317: -24.5317957216, 2318: 0.681426293733, 2319: 7.02454998703, 2320: 1.11201918431, 2321: 34.0824177626, 2322: 1.77287478701, 2323: -0.0816415479464, 2324: -1.46168725614, 2325: -72.6533703303, 2326: 0.0156406546336, 2327: 22.8686155039, 2328: 10.9931826254, 2329: -1.28616184314, 2330: 10.076285034, 2331: 50.7593018259, 2332: 85.3559805804, 2333: 0.0523686668909, 2334: -0.367411917265, 2335: 51.8367624142, 2336: -8.1899542044, 2337: 36.8590425451, 2338: 32.6723753716, 2339: -38.1019140688, 2340: 0.861285707331, 2341: -5.7515719867, 2342: 72.6265877457, 2343: 56.9610301022, 2344: 35.5380078433, 2345: -25.9662005164, 2346: 3.99486721997, 2347: -27.6016002603, 2348: -7.86278892454, 2349: 0.468775594683, 2350: 0.274979465305, 2351: 11.2305647973, 2352: -0.486503198523, 2353: -0.210316349522, 2354: -12.3153905157, 2355: 9.36070739058, 2356: -0.0518757926931, 2357: -62.4870522558, 2358: 0.232625076837, 2359: -0.0537271841577, 2360: 87.0845154575, 2361: -9.78658878595, 2362: 20.0620020612, 2363: 6.77356439652, 2364: -17.8165803742, 2365: 2.40215344895, 2366: -41.6308577117, 2367: 83.6984584629, 2368: -4.22962228548, 2369: -10.3152787023, 2370: 24.3191933191, 2371: 2.93912872876, 2372: 65.9268237047, 2373: -56.8172668104, 2374: -26.933660958, 2375: -3.73032658963, 2376: -26.4789853226, 2377: 15.4309343045, 2378: -40.7973151722, 2379: 58.5276836817, 2380: 26.8365668782, 2381: -0.423723219097, 2382: 10.7887929403, 2383: -0.34143419363, 2384: 2.41074134182, 2385: 6.32870550859, 2386: -38.2502128583, 2387: -19.0444673482, 2388: 14.8338112175, 2389: 18.6273401665, 2390: -48.9096359757, 2391: -11.0647889208, 2392: 43.2285863063, 2393: 9.09367059274, 2394: 22.5229737467, 2395: -5.12946695514, 2396: -2.80503050596, 2397: 1.06057057503, 2398: -61.968903909, 2399: 20.0372834665, 2400: -0.449252802275, 2401: 36.2431359612, 2402: -0.0761829523843, 2403: -0.0906643348337, 2404: 21.5229866568, 2405: 1.54383947541, 2406: -0.296263283516, 2407: 0.0306599419697, 2408: 45.835561276, 2409: -0.178400193736, 2410: -51.7321303416, 2411: -0.10361180728, 2412: -0.159103956002, 2413: 47.5846820775, 2414: 2.78629852804, 2415: -3.21757675336, 2416: 2.56217668637, 2417: -1.03596372331, 2418: 1.38112339417, 2419: -1.61978144195, 2420: -7.58340244034, 2421: -38.9580362765, 2422: -95.0085684397, 2423: 2.55459895387, 2424: 54.7221847117, 2425: 25.0347484687, 2426: -87.3798732986, 2427: 1.78196714063, 2428: 4.31665339309, 2429: -22.1849317, 2430: 39.2972865962, 2431: 3.28100114305, 2432: 46.9222935117, 2433: -0.217310221669, 2434: -25.9056724129, 2435: 9.89346911417, 2436: -0.296080897375, 2437: 26.0231649727, 2438: 14.3223496841, 2439: 100.0, 2440: 5.45125776176, 2441: -48.1797701251, 2442: 8.69157339714, 2443: -24.7793906609, 2444: -0.82105063413, 2445: -35.9360470616, 2446: 58.8748943187, 2447: 80.534919524, 2448: -37.5339285422, 2449: -25.953185621, 2450: 34.1741818562, 2451: 56.3957882651, 2452: 34.352546415, 2453: 21.0538628912, 2454: -26.5320162398, 2455: 34.0636289863, 2456: -24.5353762777, 2457: 55.1616497222, 2458: -1.93128340248, 2459: -26.3320214013, 2460: 53.6398667137, 2461: 10.3510752173, 2462: -21.5775573156, 2463: 2.10004421057, 2464: -42.3849786194, 2465: -97.8604286245, 2466: 0.965684696187, 2467: 4.75292456441, 2468: 88.3968879454, 2469: 6.39002787055, 2470: -25.7706713735, 2471: 39.310699716, 2472: 83.8210862987, 2473: 6.89564512036, 2474: -34.8639950229, 2475: -65.362324988, 2476: 16.6389354868, 2477: -30.1912266431, 2478: 44.0065320648, 2479: -29.579800295, 2480: -17.5378405448, 2481: -5.40157740762, 2482: -73.3244172074, 2483: -61.9817920726, 2484: -55.9719707328, 2485: -6.32316901425, 2486: -54.5939960478, 2487: 28.1923018108, 2488: -5.64014266863, 2489: -8.83473279437, 2490: 27.8892010742, 2491: -6.90374526288, 2492: -76.1633856384, 2493: -7.61016147394, 2494: 0.910412000073, 2495: 42.1595372722, 2496: -6.14222889175, 2497: 1.72339057032, 2498: -56.8603248416, 2499: -5.97374851168, 2500: -55.7499949793, 2501: 6.48719818914, 2502: -53.7541362141, 2503: -2.94096364818, 2504: -38.9189716807, 2505: -29.301591984, 2506: 8.83713867885, 2507: 81.8666385598, 2508: -15.9565272714, 2509: -1.30490882301, 2510: -4.32294919405, 2511: -6.20210465901, 2512: -30.3214226062, 2513: -6.63256092174, 2514: -6.30130550212, 2515: 6.85907527892, 2516: -6.19101480449, 2517: 37.883699159, 2518: -4.49248548563, 2519: -5.73028605738, 2520: 9.31599817786, 2521: -3.76121077892, 2522: -52.0160355137, 2523: -28.3105360584, 2524: -6.26328091076, 2525: 13.5206597403, 2526: -6.8655423959, 2527: -6.72464266355, 2528: 47.1613304294, 2529: 14.3178989635, 2530: 4.52098809453, 2531: 14.2580839833, 2532: -6.187403393, 2533: -2.36394521492, 2534: -6.6175060669, 2535: -6.23847587755, 2536: 39.5088185437, 2537: -6.38986968447, 2538: -6.5748248099, 2539: -6.34825721206, 2540: 8.24898310419, 2541: -6.36243942494, 2542: -6.88528413658, 2543: -6.5946196402, 2544: 86.5954380782, 2545: -6.31154780616, 2546: 10.4246809454, 2547: -0.579075121298, 2548: -4.48746019582, 2549: -22.1541375994, 2550: -6.44712151491, 2551: -6.286708848, 2552: -6.14725153062, 2553: -6.10002653061, 2554: -6.44763872846, 2555: -89.0218353811, 2556: -100.0, 2557: -10.8913445006, 2558: 9.12980167076, 2559: -23.0159134695, 2560: -38.4755034753, 2561: 13.9823616502, 2562: 22.8990025942, 2563: -38.2352727497, 2564: -0.256287391233, 2565: 10.3847165534, 2566: -14.0641331832, 2567: 21.8646876747, 2568: 19.5806012799, 2569: -1.07722847046, 2570: 0.00481059771276, 2571: 1.77344592019, 2572: -51.80898327, 2573: -1.41048255525, 2574: -6.72458472872, 2575: 1.11733060763, 2576: 22.918922472, 2577: -6.11921593529, 2578: 29.80381977, 2579: -24.5641156892, 2580: -7.58940664344, 2581: 34.3626653453, 2582: -5.31192612056, 2583: 18.0087858924, 2584: 8.05894236571, 2585: 50.8413407242, 2586: 100.0, 2587: -4.14349996719, 2588: -23.6176387437, 2589: -0.33617192322, 2590: -1.32091913896, 2591: 59.2873361853, 2592: -1.1997624445, 2593: -6.18258080782, 2594: -3.74681414714, 2595: 1.21750352035, 2596: -1.00617424175, 2597: -1.38847722027, 2598: -1.22750894037, 2599: 3.73154318338, 2600: -8.29835965545, 2601: 6.97928100083, 2602: -0.581806537859, 2603: 2.1614681904, 2604: 0.0803048130121, 2605: -9.63818577527, 2606: 10.6375660366, 2607: -1.75017998617, 2608: 48.1936910662, 2609: 42.2300094546, 2610: -3.33375506247, 2611: -3.99193912077, 2612: -36.2240784151, 2613: -2.36718897444, 2614: 5.36324069017, 2615: -1.01282870001, 2616: 17.0833650016, 2617: -0.978275885512, 2618: 21.2878127908, 2619: 49.0614169694, 2620: 5.58180817306, 2621: -2.22753217868, 2622: -1.06934981975, 2623: -0.837208055291, 2624: -0.736862340154, 2625: -1.71454701374, 2626: 4.67785542337, 2627: 4.10676379689, 2628: 21.1282250488, 2629: 2.1229969897, 2630: -6.77931001868, 2631: -23.0787193172, 2632: 31.6134874755, 2633: -59.6011376068, 2634: -3.36765378804, 2635: 23.2645204926, 2636: -1.66434718154, 2637: 10.2439738871, 2638: -9.80271338543, 2639: -8.07634066273, 2640: -0.717990947765, 2641: -96.9504520044, 2642: -1.11084359796, 2643: -6.30553874973, 2644: -1.18639522036, 2645: 93.1962737257, 2646: -1.12320645074, 2647: -1.11760698065, 2648: -7.26565896521, 2649: -6.3526630593, 2650: -91.7251214075, 2651: -1.00010328855, 2652: -1.26251328035, 2653: -2.06029521986, 2654: -1.18562801706, 2655: -5.07816651059, 2656: 5.80641903226, 2657: -38.6089839439, 2658: -4.09527564883, 2659: 0.52086160147, 2660: 0.974317807483, 2661: -0.804258592525, 2662: 6.52237022341, 2663: -10.082927984, 2664: -37.9748063281, 2665: -13.5300791728, 2666: 25.8904137819, 2667: -4.2016756251, 2668: -5.77285538382, 2669: -1.04793612778, 2670: 4.79414158127, 2671: 7.76287246621, 2672: 0.782045788387, 2673: -21.7198459633, 2674: 16.9207567199, 2675: -2.39275533389, 2676: 22.3894759549, 2677: 33.3103187596, 2678: -6.12914234804, 2679: -26.8611336733, 2680: 38.8277789914, 2681: -10.2118050289, 2682: -1.02439925078, 2683: -6.54876542072, 2684: 5.31330243172, 2685: 23.1537584734, 2686: -58.8922602719, 2687: -69.1345569481, 2688: 31.8611052256, 2689: 85.1882182197, 2690: 49.5790724013, 2691: 78.8493410584, 2692: -7.60450548643, 2693: -43.3213832532, 2694: -100.0, 2695: -12.5435768157, 2696: 37.7172630175, 2697: 43.9482561211, 2698: -24.1086603018, 2699: -6.59493247612, 2700: -6.57996609372, 2701: -0.891011461293, 2702: -6.2821671046, 2703: -11.1460010404, 2704: -30.2734328105, 2705: -6.23203113847, 2706: 0.00322519102987, 2707: -6.14545338764, 2708: 3.63334443953, 2709: 42.75934317, 2710: -68.1198274931, 2711: 32.7451250194, 2712: -20.7624827496, 2713: 2.33566080065, 2714: -8.12702858304, 2715: -19.9438381035, 2716: -30.4803379854, 2717: -14.0173230558, 2718: -100.0, 2719: 54.2105222642, 2720: 13.6352797536, 2721: 13.3074842429, 2722: -5.89894206169, 2723: 17.8281881614, 2724: -8.82230998633, 2725: -17.5278089715, 2726: -24.391433429, 2727: -100.0, 2728: 83.5585074971, 2729: 26.8695798429, 2730: -6.64331066648, 2731: 7.69098159991, 2732: -6.43822494522, 2733: -2.79386716817, 2734: -1.36507661001, 2735: 6.90322455769, 2736: 28.2776171361, 2737: -35.5604291525, 2738: -16.9650099546, 2739: 54.0896740331, 2740: 42.3676485863, 2741: 3.48120839653, 2742: -6.91979485518, 2743: -4.97238187046, 2744: -3.16128573796, 2745: 4.86832981564, 2746: 15.0162003562, 2747: -30.5129871702, 2748: 17.7744547567, 2749: -6.68604973791, 2750: -13.4169928382, 2751: -6.32311774262, 2752: -6.3812483078, 2753: -6.54249007647, 2754: -44.8080631307, 2755: -6.19342337629, 2756: -6.15565492828, 2757: 6.94914464856, 2758: -6.23803435863, 2759: 7.05751033085, 2760: -6.20863062849, 2761: -6.13257202124, 2762: 48.4796850476, 2763: -6.30300676774, 2764: -4.26893136745, 2765: -29.1836366423, 2766: 0.647324105874, 2767: 47.2654759261, 2768: -0.410447875053, 2769: -12.1927318425, 2770: 59.2340760972, 2771: -61.6538323619, 2772: -39.9964819693, 2773: 15.9945007709, 2774: 29.4097711506, 2775: 75.0073433492, 2776: -4.78399657083, 2777: -33.9899896572, 2778: 16.0424077394, 2779: -42.5122754918, 2780: 68.7171996187, 2781: -21.4125714661, 2782: -6.15755279904, 2783: -34.0945915448, 2784: 45.0129223741, 2785: -0.269168860921, 2786: 69.6730235871, 2787: -13.2636167339, 2788: 48.0148231273, 2789: 100.0, 2790: 73.3698594265, 2791: -24.4081081189, 2792: 93.0057379406, 2793: -0.415025034107, 2794: -15.7144897695, 2795: -1.47680901723, 2796: 1.97554171036, 2797: -5.75333779394, 2798: 4.56353700139, 2799: -0.0924123225902, 2800: 6.35751678214, 2801: -6.38273546993, 2802: -7.7401610726, 2803: 0.539425787898, 2804: -2.1779516883, 2805: 3.22398284648, 2806: 0.491622284596, 2807: -4.72236953743, 2808: -9.39809882571, 2809: 2.21198526734, 2810: -4.84028404545, 2811: 12.4606588124, 2812: -20.1647767434, 2813: -1.71340117138, 2814: 9.03238430413, 2815: -21.1992930354, 2816: -0.73067310274, 2817: 1.8730747139, 2818: 7.91898313641, 2819: -4.29809556178, 2820: 0.95006137532, 2821: -1.23650191292, 2822: -3.3457946017, 2823: -1.6985231009, 2824: -0.400734012251, 2825: 1.98750812243, 2826: -0.762111309585, 2827: 14.0992550145, 2828: -5.06621384319, 2829: -0.146604112343, 2830: -3.83521004015, 2831: 19.1775718247, 2832: 3.96037355185, 2833: 20.1991794987, 2834: 0.572416015024, 2835: -9.72866753514, 2836: -1.90299864771, 2837: 16.0435219059, 2838: -5.80348373165, 2839: -5.22263605431, 2840: 1.40260763178, 2841: -22.0836475849, 2842: -0.281491296564, 2843: -7.44751877242, 2844: 5.43850615744, 2845: 1.2824319636, 2846: 1.62723211428, 2847: -6.94751057774, 2848: -2.2882258422, 2849: -11.1158993086, 2850: -1.26166660912, 2851: 4.0387328862, 2852: -0.958318798445, 2853: -1.41877497946, 2854: -10.7484325229, 2855: -1.22685430721, 2856: 9.74490875817, 2857: -6.01287479455, 2858: 0.142382188692, 2859: 4.96197883293, 2860: -1.88526849386, 2861: 8.99430006177, 2862: -1.32919589323, 2863: -0.587179641597, 2864: 12.6774192487, 2865: -2.92883821349, 2866: 6.82569499019, 2867: 4.03440211928, 2868: -10.2381336823, 2869: 2.41806630515, 2870: 5.52451467164, 2871: 6.14714029371, 2872: 13.7392390622, 2873: -0.206362457114, 2874: 2.93574391661, 2875: -2.33928388651, 2876: -2.90446598376, 2877: 1.60691883734, 2878: 2.47511472722, 2879: -0.284718312635, 2880: 1.39921486773, 2881: -1.09750826081, 2882: 13.6750133181, 2883: -2.26704228712, 2884: -0.271067589476, 2885: 11.0655178819, 2886: 1.5538335847, 2887: -1.36293258498, 2888: 1.0055878837, 2889: 15.1638533152, 2890: -0.660237863672, 2891: 4.40992908496, 2892: -3.56989509977, 2893: -6.76238114661, 2894: -1.61863662314, 2895: 5.46762490297, 2896: -1.85088306618, 2897: -4.02137262327, 2898: 27.9384234728, 2899: 7.38702641854, 2900: -1.39867263643, 2901: -0.730516230615, 2902: 24.4884677576, 2903: -3.68489899717, 2904: -14.018946597, 2905: -24.6663048074, 2906: 19.7517444844, 2907: -3.87957399315, 2908: 1.54729793999, 2909: -8.80449201053, 2910: 1.96080858779, 2911: 18.0262526043, 2912: -10.1747160215, 2913: 11.4826673435, 2914: -6.70337849828, 2915: -8.05322476614, 2916: 0.129343771385, 2917: 22.7951560051, 2918: -0.4117433761, 2919: 0.0379424654468, 2920: -3.17612441229, 2921: 23.7289656062, 2922: -0.31297303199, 2923: 2.97778708435, 2924: 0.0232405615332, 2925: -0.878966191384, 2926: -4.39714863695, 2927: -1.74924606901, 2928: 0.339202439895, 2929: 2.99810199259, 2930: -0.658837133944, 2931: -3.43193552742, 2932: -3.50111942857, 2933: 10.1923817426, 2934: -4.01053475069, 2935: 9.04562365213, 2936: 1.10314592491, 2937: -1.25408703833, 2938: -0.430099157134, 2939: -0.409490964278, 2940: -23.4968185274, 2941: -0.378061641054, 2942: -2.30679279207, 2943: 15.087355473, 2944: -0.550818271579, 2945: -0.365663065283, 2946: -0.233893398777, 2947: -0.280788639426, 2948: -2.44450647355, 2949: 31.7400845996, 2950: 1.22910849467, 2951: -0.714933208265, 2952: -1.06988719687, 2953: -0.428489133353, 2954: 1.23713883086, 2955: -0.273704278963, 2956: -1.54892543592, 2957: 13.6646197768, 2958: -0.594063973925, 2959: -0.493842059936, 2960: 0.449614456237, 2961: -1.88326191287, 2962: -11.7503621306, 2963: -6.17852444662, 2964: -0.561045949065, 2965: -8.85426075883, 2966: -0.350232364221, 2967: -0.927329895084, 2968: -11.116523202, 2969: 0.488283872985, 2970: -3.09080990234, 2971: -0.553463570611, 2972: -0.568473179462, 2973: -0.674426483985, 2974: -0.0938334269138, 2975: -2.42636598043, 2976: 23.8489579839, 2977: 6.00881243788, 2978: 5.81959273362, 2979: -8.7108495355, 2980: -2.20214521236, 2981: 1.0060224313, 2982: 0.890431105084, 2983: 9.40013523897, 2984: 17.7096799786, 2985: -0.342554172363, 2986: 2.28083807249, 2987: -2.23885960813, 2988: 1.98991521785, 2989: 0.0125060146411, 2990: 7.85090744879, 2991: -0.37886652749, 2992: -2.01485382781, 2993: -0.492957926144, 2994: -0.759906304278, 2995: -0.453763568059, 2996: -0.390172327764, 2997: 15.4175155122, 2998: -1.84873549522, 2999: 0.589717174444, 3000: -0.355496741258, 3001: -0.508981314154, 3002: -1.43570634421, 3003: -0.691831788123, 3004: 1.69542206377, 3005: 2.55771392626, 3006: -4.51801738747, 3007: -0.337947732287, 3008: 2.53484087223, 3009: 12.8042422351, 3010: 0.21594793222, 3011: -4.42239976956, 3012: -7.91839208648, 3013: 6.90340854615, 3014: 4.01113202976, 3015: -5.76881214168, 3016: -1.62907463868, 3017: 2.43382286719, 3018: -0.702917869432, 3019: -7.49233600685, 3020: -4.36724744327, 3021: -2.88506209725, 3022: 0.915056528905, 3023: -8.91835734829, 3024: -0.569888405285, 3025: 3.10864921131, 3026: -19.902505325, 3027: -1.50689575724, 3028: 17.299020287, 3029: 3.48369987993, 3030: -11.8156715915, 3031: -2.72014560757, 3032: 2.74915882247, 3033: 3.20018946212, 3034: 16.379982878, 3035: -4.71180826422, 3036: -16.0840938028, 3037: -2.02084076647, 3038: -3.7377972125, 3039: 12.8928825849, 3040: 0.924346055847, 3041: -5.5490619363, 3042: 14.1454439176, 3043: 12.1269133474, 3044: -11.0685888126, 3045: -26.9197942669, 3046: -5.10252610072, 3047: -8.01262301691, 3048: -0.983279255552, 3049: 18.6501162369, 3050: -2.12462228759, 3051: -2.29443679431, 3052: 14.9328700423, 3053: 5.98969727129, 3054: -0.868613342927, 3055: 11.8614678581, 3056: -3.86888262419, 3057: -2.47574308189, 3058: -7.98899224564, 3059: -6.32665603755, 3060: 2.656044731, 3061: 13.2267315216, 3062: -3.26484031544, 3063: 2.43756127025, 3064: -1.50401124581, 3065: -5.74668116155, 3066: -11.9393408799, 3067: -8.71304776187, 3068: 4.80956882632, 3069: 8.54196802065, 3070: 7.97314034274, 3071: -23.3948337883, 3072: -18.874345029, 3073: 6.28044466404, 3074: 6.29590651468, 3075: 2.95002398577, 3076: 7.76129114255, 3077: 10.7317510742, 3078: -2.23855679247, 3079: 3.14322418203, 3080: 5.46861304754, 3081: 1.15989192111, 3082: -3.27795664422, 3083: -0.51223905656, 3084: -11.6077371398, 3085: -2.98356518651, 3086: 7.6229126548, 3087: 4.0490569797, 3088: -3.22955852288, 3089: -11.2642381826, 3090: -1.7455856831, 3091: 5.27927894425, 3092: 15.1651715282, 3093: -8.08515079846, 3094: 0.562758217138, 3095: -0.875513001673, 3096: 2.00218706872, 3097: 5.94625443371, 3098: -1.23717856298, 3099: 12.472023207, 3100: -1.93172680116, 3101: -1.82286995811, 3102: -1.07836994864, 3103: -7.88589725152, 3104: -0.160792510732, 3105: -0.314689262767, 3106: -9.00386486719, 3107: -1.80409936322, 3108: -13.3288998851, 3109: -2.63822576775, 3110: -2.31379361341, 3111: -0.672137132621, 3112: 6.21631350667, 3113: 2.12564491361, 3114: -8.35258644284, 3115: 8.43336724875, 3116: -5.39757404039, 3117: -3.51399395584, 3118: 7.84698174089, 3119: -1.97550137605, 3120: 8.56741573444, 3121: 6.3517519622, 3122: 2.18968766797, 3123: -7.72985478222, 3124: 6.51371924012, 3125: -2.08606936001, 3126: 1.08093350004, 3127: -3.73415641239, 3128: -8.93169805066, 3129: -16.5625664761, 3130: -3.27664364018, 3131: -1.89954002491, 3132: -5.33983542604, 3133: -1.71481973623, 3134: 14.2387043465, 3135: -1.69943967993, 3136: -18.3233116937, 3137: 0.103422095585, 3138: -0.185662346389, 3139: 8.52883039123, 3140: -13.6138571489, 3141: 24.256183445, 3142: -0.0195664083558, 3143: -29.3029335392, 3144: 1.22114972684, 3145: -10.4884899695, 3146: 0.338688578831, 3147: 14.8114152468, 3148: 7.84420330117, 3149: 11.5081419354, 3150: -23.1688421968, 3151: 3.21051010942, 3152: 4.80655427461, 3153: 5.8783691926, 3154: -13.0148204833, 3155: -17.9391563432, 3156: -0.447602477371, 3157: 3.59545751927, 3158: -14.6378642842, 3159: -15.2994542128, 3160: -7.64733584431, 3161: -5.08999329363, 3162: 0.119499507082, 3163: -10.4514981512, 3164: -25.899897021, 3165: 5.67387450662, 3166: 11.9177410722, 3167: 2.41639824831, 3168: 19.9907803278, 3169: -10.3244019237, 3170: -4.44158815835, 3171: -0.277188527759, 3172: -9.33670481263, 3173: -22.8812037837, 3174: -21.0252769013, 3175: -9.22082743206, 3176: -6.12670080646, 3177: 2.22879944888, 3178: -12.5539341793, 3179: -9.67675177333, 3180: -14.7120147333, 3181: -23.1221440901, 3182: 5.05724216177, 3183: 0.572079409652, 3184: -9.0957834736, 3185: -27.3866219457, 3186: 8.20173009977, 3187: -0.557703204149, 3188: 1.89982969512, 3189: 1.63132742805, 3190: 3.7983243091, 3191: -1.22698297361, 3192: -7.91142510656, 3193: -6.92710807095, 3194: -0.863991933222, 3195: 4.11722980809, 3196: -15.5115571832, 3197: -1.19314803911, 3198: -3.37941646462, 3199: -1.39951487432, 3200: -18.7388731797, 3201: 0.158075144254, 3202: -14.9803641728, 3203: -14.3333791921, 3204: -33.2825409333, 3205: 15.2021086507, 3206: -2.53248305557, 3207: -0.872484772282, 3208: 10.4956094902, 3209: -0.58201172145, 3210: 17.5318135156, 3211: 0.141774249174, 3212: -0.0273923490752, 3213: 3.50377637394, 3214: -0.955263096983, 3215: 43.691706888, 3216: -3.98683906886, 3217: -3.11509287219, 3218: -8.01244264426, 3219: 1.00346269045, 3220: -13.0882374857, 3221: -19.3321728654, 3222: 1.61284791176, 3223: 8.06435371335, 3224: -0.517175591111, 3225: 5.24569576475, 3226: 6.34444741769, 3227: 3.00787105769, 3228: 15.1841009301, 3229: -11.9905397083, 3230: -0.627214258805, 3231: 10.820652586, 3232: 1.12492007454, 3233: -0.954920904235, 3234: -16.2781685178, 3235: 1.15287749122, 3236: 0.492738558614, 3237: 7.48386430813, 3238: -9.64957077221, 3239: 2.3747438583, 3240: -0.294322153089, 3241: 0.105523970525, 3242: 12.2992501612, 3243: -0.526233164157, 3244: -8.79284095757, 3245: 11.633622394, 3246: -1.16959742357, 3247: 2.54632485233, 3248: -0.589597085234, 3249: -1.09965461181, 3250: 9.63632136655, 3251: 9.30781439708, 3252: -2.47344521195, 3253: 12.4680841644, 3254: 8.7581802561, 3255: 1.75798760302, 3256: -3.61868720934, 3257: -8.39308116705, 3258: 5.14354566751, 3259: -9.41732729133, 3260: -4.86208347789, 3261: -15.1728070508, 3262: 3.78022826638, 3263: -10.6408291766, 3264: -1.91340409963, 3265: 0.466467471577, 3266: 3.38768396309, 3267: -0.112845761897, 3268: 0.430871460049, 3269: -1.79683414728, 3270: -11.096292566, 3271: -0.161232219545, 3272: 0.368057527566, 3273: 0.116791747778, 3274: -13.3474079231, 3275: 1.08631516417, 3276: 0.175120686405, 3277: 12.374598539, 3278: 0.0282451207024, 3279: -16.9057844566, 3280: 0.0556912372255, 3281: -1.89943295007, 3282: 2.13009610641, 3283: -2.85225123004, 3284: -1.85881795927, 3285: -0.106836703592, 3286: 10.2008858377, 3287: -0.0644558962222, 3288: -0.134206500066, 3289: 12.782702258, 3290: -0.167062267184, 3291: -0.718664316201, 3292: -5.21260348234, 3293: -0.512001680211, 3294: -0.172952063601, 3295: 0.00178615994938, 3296: -0.178726684002, 3297: -0.847481205903, 3298: 3.53074916556, 3299: 1.65027633148, 3300: -0.177384633228, 3301: -2.29773389661, 3302: 2.61055512265, 3303: 0.466845833064, 3304: -28.7008624801, 3305: 5.46592132085, 3306: -19.0780517701, 3307: -30.0936489783, 3308: 2.20648066347, 3309: -2.1531448759, 3310: -0.280399996975, 3311: 1.00635650692, 3312: 3.68135777348, 3313: -0.199094406225, 3314: -39.7417090956, 3315: -0.0661123241221, 3316: -1.21019237458, 3317: 11.5650207287, 3318: -44.7886428478, 3319: 0.397756533149, 3320: -0.124195790945, 3321: 0.168309937186, 3322: -0.75077110612, 3323: -0.310234904284, 3324: -2.8720673276, 3325: -7.92023692235, 3326: 4.94648056673, 3327: 10.6371794085, 3328: -7.91303817173, 3329: 7.046108519, 3330: 8.60415336785, 3331: -4.87423905462, 3332: 7.27036195084, 3333: 2.20638957382, 3334: 0.0271544194366, 3335: -30.561456572, 3336: -5.43336507334, 3337: 0.181788228037, 3338: -1.16325793424, 3339: -1.63751125076, 3340: -0.00160824333214, 3341: -0.26727618012, 3342: -0.010676792671, 3343: -3.44925526004, 3344: -0.176390415806, 3345: 0.0794935700568, 3346: 10.6016557335, 3347: -0.479764901531, 3348: -8.60801130013, 3349: -0.0196243358449, 3350: -0.150063718706, 3351: 1.13870113818, 3352: -0.12458571042, 3353: 0.0816557058222, 3354: -1.70388243363, 3355: 6.79322838497, 3356: 0.833797518162, 3357: 2.18701149087, 3358: -1.65983034703, 3359: 0.348689849898, 3360: -20.0486776632, 3361: -24.6903825448, 3362: 1.90347529263, 3363: 2.08478479842, 3364: 18.6659388931, 3365: 0.0619761505849, 3366: 12.4835489408, 3367: 1.48413989688, 3368: -2.19776420214, 3369: -7.38684938261, 3370: -2.21867345512, 3371: 7.43469467836, 3372: 6.33944826953, 3373: -0.500326534949, 3374: 3.22258767926, 3375: -3.0339230202, 3376: -0.353007089273, 3377: -24.5278091724, 3378: 20.9510202416, 3379: -10.8631205241, 3380: -0.837073958562, 3381: 0.376651766834, 3382: 1.6769690484, 3383: 21.6024510312, 3384: 31.8410527384, 3385: 8.54149004557, 3386: -1.1674999238, 3387: -4.49914587021, 3388: -3.37751995833, 3389: 20.6168451929, 3390: -4.27076005577, 3391: 15.3871233585, 3392: 1.69487996213, 3393: 8.42689289212, 3394: -14.6050903139, 3395: -17.12847517, 3396: -0.525415812967, 3397: 0.504778748966, 3398: 10.725168678, 3399: -0.542872020115, 3400: -0.705398735594, 3401: -4.92946194345, 3402: 20.4842722385, 3403: -0.675764535549, 3404: 1.88121550217, 3405: -0.541178585768, 3406: -0.879949726294, 3407: 13.1280153179, 3408: -6.78104628215, 3409: -1.10582171484, 3410: 7.07369409415, 3411: 2.74385439447, 3412: 0.6849578657, 3413: -27.5914405238, 3414: -3.32842539136, 3415: 6.59232026234, 3416: -19.0914305022, 3417: -25.2735974447, 3418: -0.517711777014, 3419: 9.60273606306, 3420: 2.01543104382, 3421: -8.2408561252, 3422: 9.07992656619, 3423: 3.45227163621, 3424: 1.75681817088, 3425: 9.61585722693, 3426: 26.6285636168, 3427: 9.34064991222, 3428: -0.331508478021, 3429: -4.1320285644, 3430: 7.17054327707, 3431: -2.655276119, 3432: -1.80518827626, 3433: 10.1024962247, 3434: 8.05997328564, 3435: -0.86573799295, 3436: 3.6775939765, 3437: 17.6525289706, 3438: 42.1638487132, 3439: 15.3753982007, 3440: 9.93298440533, 3441: 22.183134457, 3442: -7.73431943152, 3443: 16.5107021357, 3444: -8.8602586466, 3445: -2.22191885472, 3446: -16.0054291069, 3447: 0.335416696205, 3448: 15.7564818079, 3449: -0.260598629901, 3450: -0.27699155826, 3451: 4.89607281966, 3452: -16.4750215597, 3453: -0.0963521351088, 3454: -1.22685597265, 3455: 5.2084405283, 3456: -0.685435166905, 3457: 1.02933860389, 3458: -0.619914928174, 3459: -0.906427320266, 3460: 24.3433349353, 3461: 0.399205820897, 3462: 0.343429307643, 3463: 0.653606310646, 3464: -18.9650532825, 3465: -11.1103921015, 3466: 13.6934508531, 3467: 2.33861327144, 3468: 17.1873785901, 3469: -19.5120112976, 3470: -14.0495873015, 3471: -0.650353516744, 3472: -6.43724150143, 3473: 7.56383265765, 3474: 3.41068599365, 3475: 21.1827300304, 3476: -27.6499100206, 3477: 14.3952317669, 3478: -4.51173635791, 3479: 24.2921193871, 3480: -0.596357516379, 3481: 3.57819238063, 3482: 10.4076601319, 3483: 3.41114642142, 3484: -3.44127006124, 3485: 4.99985952502, 3486: 2.51915606428, 3487: 17.9462125224, 3488: -16.172402331, 3489: -7.15880942747, 3490: 13.1600768875, 3491: -1.11887412366, 3492: 0.708469626492, 3493: -0.485653062288, 3494: -1.43851013645, 3495: -1.46660988987, 3496: 15.9809097301, 3497: 0.63874966325, 3498: 0.383497976372, 3499: 0.444939641652, 3500: -0.811332401569, 3501: -1.140820246}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.504759797515, 2: 3.43735032587, 3: 7.09595553255, 4: -6.54268383939, 5: -10.2291739542, 6: -13.2084882486, 7: -5.96351730455, 8: 5.963585152, 9: -7.10203083276, 10: -2.57407095049, 11: 4.19170084361, 12: 18.3336172447, 13: -1.39440743969, 14: -9.90516717222, 15: -7.25240318256, 16: 3.68021084651, 17: -6.79760561911, 18: -4.05248917741, 19: 30.5704518963, 20: -12.3263256137, 21: -12.2176544172, 22: -4.06384116573, 23: -2.00138149455, 24: 0.476788559757, 25: 8.1023449636, 26: -15.1401847394, 27: 22.4534910598, 28: -1.0435970616, 29: -18.477102045, 30: 1.00668598778, 31: -7.06681582434, 32: 3.82386510155, 33: -7.46607891826, 34: -1.26163876158, 35: 0.918754189819, 36: -4.74519906356, 37: 4.25281730868, 38: -1.54556432307, 39: -8.54775334792, 40: -0.043498477788, 41: -6.48288161511, 42: 2.8927069024, 43: 22.6759064874, 44: -7.39116975622, 45: -2.31551909117, 46: 6.19171979595, 47: 5.16341429507, 48: -5.11795582647, 49: 11.4575337293, 50: -0.467283342718, 51: -2.98740557248, 52: 5.53940430263, 53: 3.29560733213, 54: -0.219814735665, 55: 13.8412786792, 56: 2.5576912644, 57: -6.14451211492, 58: -1.67847976729, 59: 3.00119750084, 60: 4.64079066727, 61: 15.9852117827, 62: 10.6092190343, 63: 0.960413871794, 64: -1.07545525325, 65: 7.855415886, 66: 3.3236334441, 67: 21.2723859285, 68: 3.97606116616, 69: -2.41471153709, 70: 3.54457575988, 71: 4.21258499455, 72: 5.3518416165, 73: -0.424684402538, 74: 7.64722417284, 75: 2.58099852028, 76: -1.22596014598, 77: 1.43634043342, 78: -14.8544334337, 79: -12.2817034538, 80: 8.23902003919, 81: -2.25721231247, 82: -2.44423176369, 83: 3.14137554588, 84: 3.49370996042, 85: -8.09004633017, 86: -3.22185630315, 87: 0.946456926431, 88: -4.9467223802, 89: 2.76459009926, 90: 2.31116345272, 91: 0.792613783099, 92: 2.8664562185, 93: -16.1450537536, 94: 1.07012950397, 95: 3.06615201312, 96: 3.83918245967, 97: 4.38268492404, 98: 3.88888220861, 99: 11.6170528041, 100: -0.148888844401, 101: -26.4023086899, 102: 5.58122111248, 103: -6.17530442128, 104: 3.82903016095, 105: 3.50500652395, 106: -4.36662236329, 107: 3.06811633617, 108: 2.71445461488, 109: 1.78213272162, 110: 0.915499420832, 111: -0.441807168183, 112: 3.73664170944, 113: 22.3572836824, 114: 22.6762462467, 115: -42.2202807073, 116: -2.14088967158, 117: -17.6309887909, 118: 11.6519642523, 119: 18.5974923569, 120: -8.97254440513, 121: 14.1960674916, 122: 19.8351205344, 123: -23.8667271512, 124: 6.75452313186, 125: 0.201773101886, 126: 0.435046447002, 127: -0.0336301102001, 128: -0.290496236999, 129: -12.4261573354, 130: 0.616340761812, 131: 4.12284815865, 132: 2.33429728763, 133: -10.2652360699, 134: 0.658241053648, 135: 2.56543475959, 136: 11.6987373051, 137: 0.921196301617, 138: 16.2913225138, 139: -1.04148476519, 140: 1.31350696963, 141: 7.41945121079, 142: -7.50552047931, 143: 5.33961005053, 144: 1.15666377464, 145: 7.50510637531, 146: 0.359791829399, 147: 0.613669566193, 148: 15.4702055062, 149: 0.56077732283, 150: 3.05149419302, 151: 3.31240751632, 152: 1.30437833748, 153: 0.583658714056, 154: 0.748680393467, 155: 0.625127427607, 156: 3.11414029552, 157: -4.36490654133, 158: -1.72205033106, 159: 0.990674780402, 160: 1.03366563997, 161: 5.21674067792, 162: -3.47451331357, 163: -13.530084526, 164: -6.39569103212, 165: -5.97181602701, 166: 10.7714857342, 167: 3.7310717731, 168: -9.65792803999, 169: -14.4551634851, 170: 1.64470755879, 171: -10.1235705335, 172: 0.701669277325, 173: -8.10826723279, 174: 0.415414722217, 175: -7.54610628871, 176: 2.99781152977, 177: 15.3234950279, 178: -0.834093871667, 179: 0.407414996113, 180: 0.502253982041, 181: 0.334454741628, 182: 0.564047261406, 183: 1.11405559383, 184: 10.3480831261, 185: -4.22375109316, 186: -14.7065231679, 187: -2.17644682731, 188: 1.76178777505, 189: 5.2971758111, 190: 1.27693212496, 191: -3.97086934109, 192: 1.1753314978, 193: 0.647300894901, 194: 20.3470395517, 195: 2.89250420346, 196: 2.95987930972, 197: 2.62042780119, 198: 7.88428155681, 199: 0.390730264826, 200: 2.60819740056, 201: 0.630177352295, 202: -5.58929310679, 203: 0.542960519037, 204: 0.595298196494, 205: -28.0288432903, 206: 3.01125727052, 207: 4.95689743861, 208: 0.661858922656, 209: 0.579697978563, 210: 1.23762273528, 211: 0.644860432121, 212: 7.62995579956, 213: 7.9219889699, 214: -4.26740877198, 215: 0.934342039524, 216: -0.717724119153, 217: -25.4669865189, 218: -6.30545608106, 219: -28.5105841859, 220: -17.4559737792, 221: -11.2411938636, 222: 22.5518887766, 223: 7.9812881397, 224: 1.33945954287, 225: 5.12515471388, 226: 0.528541968237, 227: 35.9651172132, 228: 2.33341528751, 229: -8.75418222125, 230: 5.0826183846, 231: 3.20725898446, 232: 1.83734338362, 233: -16.6527895918, 234: 0.160401431751, 235: 0.141192122488, 236: -18.8038479376, 237: -19.0028031051, 238: -19.0582078036, 239: -5.67619561353, 240: 4.05070785332, 241: -15.9669344531, 242: 13.0900543452, 243: -29.2614544389, 244: 1.13999127468, 245: 2.81210088193, 246: 4.22545416584, 247: -12.9056693326, 248: 4.4777799397, 249: 20.2359745589, 250: -0.267381505384, 251: 23.4715617071, 252: -3.08312430806, 253: 5.65441613289, 254: -8.31716101004, 255: 3.87500237654, 256: 1.19008641254, 257: -8.40743409908, 258: 3.4135926227, 259: 3.20730916502, 260: -1.15901795189, 261: -24.2622557928, 262: -0.216097073946, 263: 0.114758480313, 264: -1.29538674135, 265: 3.05597308022, 266: 14.6331807169, 267: -16.0876221471, 268: -17.3156906617, 269: -2.66159916665, 270: -1.19814330928, 271: 1.81260009947, 272: -15.3678810165, 273: -15.3685503734, 274: -23.429086011, 275: -10.3424156878, 276: -16.2572594004, 277: -12.7157388935, 278: -1.66928366977, 279: 0.181565124937, 280: 5.88392180082, 281: -9.46152606776, 282: 7.12602446855, 283: -12.936783485, 284: 9.32855133932, 285: -24.7414962631, 286: -7.05468923709, 287: 3.16349569199, 288: 13.064715127, 289: 3.0478758893, 290: -0.00119551233459, 291: 2.63039238456, 292: -8.69163460109, 293: -10.2256942221, 294: 6.57213582506, 295: -6.98304637011, 296: -2.2354868263, 297: 1.18661614328, 298: -11.7121266802, 299: -1.07733334232, 300: 3.55049303318, 301: -9.66433065961, 302: -0.251293422533, 303: 4.87343874887, 304: 22.2402565944, 305: -15.670726871, 306: 3.03801401796, 307: -16.8581514568, 308: 2.65243856735, 309: 2.70881231331, 310: 0.580546910822, 311: -5.8250404042, 312: 2.83941122539, 313: 2.70394321544, 314: 25.7023810066, 315: 3.24817002455, 316: 9.84400775018, 317: 0.206811282611, 318: 2.8832367591, 319: -12.7476653346, 320: -2.05146320839, 321: 12.7482119906, 322: 16.4898801836, 323: -5.25825911918, 324: -0.832164192786, 325: -0.255942367761, 326: -23.3858483791, 327: 4.57501853151, 328: 1.90987762637, 329: 31.6742646064, 330: -15.7114376702, 331: -2.7754102696, 332: 8.0081518771, 333: -0.107470054309, 334: -1.11667963429, 335: 8.27107543272, 336: -30.2912337776, 337: 1.55001130631, 338: -1.824693243, 339: 2.86374630698, 340: -5.26227690023, 341: -11.4403944957, 342: 17.7018848064, 343: 34.3358070321, 344: -4.27729909041, 345: -18.3444053904, 346: -7.39656453433, 347: 14.4975197254, 348: -1.53643559205, 349: -6.62327120171, 350: 0.660906675898, 351: 8.93242023649, 352: -5.71929210933, 353: -2.31670731774, 354: 0.162070958496, 355: 2.61667282774, 356: -19.5692275252, 357: -5.89094116577, 358: 4.88736796875, 359: -3.58954415578, 360: 0.207163090068, 361: -11.0328348611, 362: -3.02670135478, 363: -26.6632596768, 364: -4.18503509353, 365: 7.84076712517, 366: -16.5037564994, 367: 4.13187316735, 368: 12.6003737058, 369: 4.62628861357, 370: 8.60988215955, 371: -2.71749759792, 372: 13.9102683062, 373: -17.4661943792, 374: 6.03096457463, 375: -24.5231372704, 376: -21.1682568307, 377: 2.72117700611, 378: 4.63919109412, 379: -8.00603085346, 380: 6.84206054679, 381: -13.4777040726, 382: -6.08626988869, 383: 6.87338864802, 384: -6.80535497169, 385: -6.83779990833, 386: 3.52824103977, 387: -2.56544199153, 388: -5.67502433119, 389: -26.4592843127, 390: -5.80824007272, 391: 3.13935044484, 392: -6.95739115969, 393: -3.16024575333, 394: -1.590138759, 395: -1.80100090383, 396: -3.55904358502, 397: 2.40095395985, 398: 8.53656378107, 399: 2.07839656913, 400: -9.68207472985, 401: -1.48560882128, 402: 3.12565068655, 403: 6.44377706529, 404: 4.17024738486, 405: 1.93679537857, 406: 11.7029325547, 407: 3.37181762098, 408: -2.3844027187, 409: -11.632750716, 410: -6.71920061304, 411: -2.97963947086, 412: 3.02919589893, 413: -16.6750524493, 414: 4.85439162544, 415: 0.426471354676, 416: -4.629876058, 417: 1.31751096357, 418: -0.0232302831403, 419: 2.60664076077, 420: 2.8983345171, 421: 1.109324549, 422: 3.0846296149, 423: 3.15863615387, 424: 1.00234402658, 425: 2.40488725106, 426: 1.8355151789, 427: 4.22604391454, 428: -4.40241755148, 429: -15.2786903076, 430: 0.198370252498, 431: -1.3973631567, 432: 3.27269686997, 433: 3.54898394354, 434: 9.63640380009, 435: 0.471139594384, 436: -1.5278635366, 437: 1.13950157801, 438: 3.19268271671, 439: 2.60718411376, 440: 1.13468957977, 441: 3.30113548692, 442: 13.0038833593, 443: 2.8567826049, 444: 2.97188892445, 445: 2.66533823444, 446: -0.0664731704505, 447: 2.73812185107, 448: 2.97724253868, 449: 1.22307172359, 450: -13.9602276475, 451: 2.58375407184, 452: -2.43143613463, 453: 0.591245413275, 454: -1.27794512479, 455: 9.23814358381, 456: 2.6728150403, 457: 3.60625530503, 458: 2.76678199716, 459: -7.4225680769, 460: -2.02894455262, 461: 5.88481209995, 462: 8.08008897926, 463: 2.13293009409, 464: -16.9780491257, 465: -2.7736796061, 466: -11.731829944, 467: -1.56236798963, 468: 6.1864163938, 469: 1.86224511883, 470: 2.53234690641, 471: -7.30186082902, 472: -0.679431691931, 473: 1.9068345285, 474: 10.4465003816, 475: 0.88210920439, 476: 1.32728910247, 477: 0.216157952093, 478: -3.704021943, 479: 1.11165801233, 480: 2.87639675257, 481: 1.34719052693, 482: -3.5952756066, 483: 0.82449580139, 484: 1.84518395405, 485: -0.341968627766, 486: 0.746096911882, 487: 10.0084224476, 488: 1.08147339676, 489: 1.68409321674, 490: -7.17012990927, 491: 6.52318679592, 492: 3.79473121072, 493: 0.608177558327, 494: 3.53362907062, 495: 0.683502546065, 496: 0.807038459783, 497: 8.3700480185, 498: 0.793194287361, 499: 3.18657130956, 500: 7.19493467208, 501: 0.748591669139, 502: 0.878249738981, 503: 0.910805245196, 504: 0.914461450886, 505: 2.03238900183, 506: -2.10380591006, 507: 2.00447593215, 508: 0.895702250737, 509: 1.12945814118, 510: 0.272739003261, 511: 0.336939421044, 512: -13.478687001, 513: 0.218729270107, 514: -13.7543200592, 515: -6.26035445345, 516: -0.424195817646, 517: 1.71875149642, 518: -3.22179378714, 519: 1.06248565135, 520: -10.152008142, 521: 0.966207361606, 522: -1.16732785934, 523: 0.851382252875, 524: -2.45550206295, 525: -6.00439206954, 526: -8.74919802624, 527: 1.87692924161, 528: 0.919419330588, 529: 0.662116400505, 530: 0.957455485803, 531: 0.973509592831, 532: 1.45942078364, 533: 10.5663405873, 534: -2.67267391686, 535: -4.08635308607, 536: -17.6346095501, 537: 3.12937400832, 538: 4.87964509538, 539: 5.05531382765, 540: 4.07991071788, 541: -7.29581538174, 542: 0.892776460939, 543: 12.0392483244, 544: 1.13127434763, 545: 0.502356503592, 546: -2.88200615716, 547: -10.1146307117, 548: 0.848132461821, 549: 2.94795818298, 550: 0.722525254898, 551: 0.792688599091, 552: 0.719914532373, 553: 0.808640156272, 554: -4.59211659771, 555: 3.12609323879, 556: 6.09095913784, 557: 0.83374403185, 558: 0.650701719082, 559: -0.432625117385, 560: 0.70660104538, 561: 4.01334547203, 562: -2.48664043881, 563: -17.5656379118, 564: -0.899136656685, 565: 0.175341082905, 566: -0.781399296526, 567: 0.596020322976, 568: 6.6339395948, 569: 4.38884896214, 570: -2.43730714465, 571: 0.195463613905, 572: -0.95150902, 573: 1.6423671981, 574: 2.55180847911, 575: 0.972534442691, 576: -2.80612334382, 577: 0.638929519146, 578: 2.16833358781, 579: 1.66622724655, 580: -16.5438551661, 581: 0.610599656232, 582: 9.00526284931, 583: 1.06467593629, 584: -5.11632836922, 585: 0.322995587761, 586: -9.91810047032, 587: -2.44010041075, 588: -11.453591452, 589: 2.65149732765, 590: -3.47508766132, 591: 6.24891102052, 592: 8.04097702661, 593: -0.280958840643, 594: -2.2711374463, 595: -1.29847278404, 596: 6.64860552067, 597: -8.28883612284, 598: -11.8601087069, 599: 6.37234520473, 600: -3.46662799816, 601: 2.39678173483, 602: 3.50281821082, 603: -4.81742117603, 604: 3.00289553793, 605: 1.44041000062, 606: -18.9232761697, 607: 1.77551246534, 608: 3.10054684597, 609: 6.97337167879, 610: -6.34205547537, 611: 2.92898086872, 612: 4.41778841495, 613: -2.34403094108, 614: 1.96297037553, 615: -2.65961267427, 616: 0.109299625077, 617: 9.01383008041, 618: -1.25687019658, 619: -0.549929584494, 620: 0.874638221284, 621: -4.56740564612, 622: -11.590828196, 623: -2.9395786376, 624: -3.98543065526, 625: 26.765914725, 626: 9.38402239114, 627: -9.34943240066, 628: 1.0402583656, 629: -1.29321466597, 630: 0.134934248617, 631: -9.85433722971, 632: -8.60687598021, 633: -2.50618126802, 634: 2.5216278671, 635: -4.37875862943, 636: 2.3060445438, 637: 4.02323446967, 638: 2.08602192377, 639: -0.763942427893, 640: -1.11753275957, 641: -3.06622239167, 642: -8.1343052078, 643: -8.99015206109, 644: -23.1556722658, 645: -1.74729192008, 646: 8.03133419754, 647: -14.6133609985, 648: 2.26260983244, 649: -5.35999895485, 650: -0.240666857391, 651: -1.10554776109, 652: 2.39602614608, 653: -2.4660176968, 654: -10.995271896, 655: 2.9671122801, 656: -9.90972709265, 657: 3.10977623207, 658: 2.94787094246, 659: 2.83268264209, 660: -2.30366919333, 661: 3.0146631319, 662: 3.51641108988, 663: -16.8855824417, 664: 2.96806198871, 665: -3.55673515037, 666: 2.69411362389, 667: 3.12678015011, 668: 6.51284576781, 669: 3.06271414762, 670: 1.74036799793, 671: -9.2827371734, 672: 9.28105387665, 673: 5.27861437284, 674: -0.857449825923, 675: -5.90429640225, 676: -4.8676431736, 677: -1.86821434691, 678: -4.24061747791, 679: -16.5920220421, 680: 0.369172517523, 681: -12.3696568371, 682: 0.898479286801, 683: 2.17491571372, 684: -0.711455784358, 685: 4.59940250244, 686: 6.11114895239, 687: 10.6612741089, 688: 3.20702291022, 689: -11.7344223417, 690: -5.68437649882, 691: -1.21603387675, 692: -36.3543399658, 693: -15.4103175353, 694: -0.417188178433, 695: 13.8369298639, 696: 9.57201956485, 697: 13.2862991933, 698: 6.58084685283, 699: 0.252071215544, 700: 4.49519376537, 701: -5.27048144759, 702: -5.01344703076, 703: -1.90929627396, 704: 3.3469939514, 705: 1.27295197552, 706: -4.66599316573, 707: -4.90127443459, 708: 0.0480173241426, 709: -7.1794529468, 710: -1.25019913997, 711: -5.30193162833, 712: -5.87518333019, 713: 0.329795944444, 714: -2.92069867116, 715: 0.316953199417, 716: 7.36022086769, 717: -13.3082190933, 718: -12.157994966, 719: 5.05281204109, 720: -4.08351054026, 721: -1.23113349619, 722: -13.523342127, 723: 3.73092804898, 724: 8.24559651592, 725: 2.64562587645, 726: 2.72337178834, 727: -1.82116728033, 728: -1.34589642411, 729: 4.15679226338, 730: -1.92909707182, 731: -2.31068944208, 732: -2.84882921827, 733: -0.675725393752, 734: -2.72254117661, 735: 1.4306109014, 736: 10.4461519703, 737: -0.0126562810483, 738: -2.81783656395, 739: 0.637086070841, 740: -5.41322888728, 741: 0.749306669152, 742: -0.880732346567, 743: -3.28498589962, 744: 1.58142014995, 745: -6.41651739858, 746: 0.493877976771, 747: 7.41972777967, 748: -1.66121906886, 749: -6.34951899264, 750: -4.57304057133, 751: -0.208549434153, 752: -0.905839634971, 753: -4.08599001803, 754: -0.161240766965, 755: 0.56942329393, 756: 3.76096067582, 757: 10.358825579, 758: 1.43706012394, 759: 1.63843086346, 760: -14.5846719602, 761: -3.99271894212, 762: 3.18876870725, 763: 2.65533133801, 764: -0.852478527162, 765: -11.3944596783, 766: 0.745578232819, 767: -1.99355889844, 768: 0.572191135664, 769: 1.96843129668, 770: -3.87829608775, 771: 1.57965077566, 772: 4.91018578462, 773: -3.58866863864, 774: 0.821551061382, 775: 0.241571390186, 776: -0.782129330604, 777: -10.2250966688, 778: 2.06448815392, 779: -3.31777446671, 780: -4.40913015349, 781: 1.67795929812, 782: 2.20887345389, 783: 1.18049754909, 784: -0.200264661526, 785: 0.804300287456, 786: 3.9216574167, 787: -0.097234845698, 788: 0.321038814444, 789: 0.878900587094, 790: 1.50151951501, 791: -8.3801092177, 792: -1.3720266885, 793: 1.24887234244, 794: 2.76225355594, 795: 6.76633776294, 796: 2.22077448147, 797: 1.27243314762, 798: 2.06788967143, 799: -1.35517495484, 800: 1.12731156497, 801: 7.07168887781, 802: 0.19872671838, 803: 3.06813016906, 804: -7.18687204977, 805: 1.67261417026, 806: 0.619057790885, 807: 0.331480513507, 808: -2.49301112749, 809: -0.698606840031, 810: -8.38032092953, 811: 3.67024768369, 812: -5.57312714922, 813: 2.00319398493, 814: -0.17950593754, 815: 3.70498375626, 816: 2.20999556339, 817: -6.583784078, 818: -2.60371552226, 819: -4.75450584749, 820: -3.72501236908, 821: -3.99175136708, 822: -2.42940590396, 823: 3.93098145694, 824: 0.238020490868, 825: 0.382863253799, 826: 1.70644974764, 827: -3.53561955652, 828: 0.346621016321, 829: 1.67676818279, 830: 0.0956276998559, 831: -6.8714859938, 832: -1.96540297496, 833: -3.04950565343, 834: 9.21571197592, 835: 1.04586330418, 836: 7.90257249408, 837: 2.33400635753, 838: -10.0021488641, 839: -5.17941416612, 840: 0.62280935253, 841: 5.51409477977, 842: 1.64714734046, 843: 5.55174378141, 844: 0.193990596419, 845: 0.170303584239, 846: 1.00247472767, 847: 0.237334872358, 848: 0.590164416706, 849: -4.60622977991, 850: 0.251068703789, 851: 0.264529294925, 852: 0.312877787566, 853: 0.0950799550189, 854: 0.0112680797285, 855: 0.418619526927, 856: -0.2921291911, 857: 0.324397025716, 858: 0.448701223944, 859: 4.70588397222, 860: -3.41532399337, 861: -11.5134943943, 862: -0.027466787888, 863: -6.94075607714, 864: 10.6165908664, 865: 0.814950904836, 866: 3.66259535469, 867: 0.91758935662, 868: 1.58432095474, 869: 6.15016681873, 870: 0.182180621589, 871: 4.42332618475, 872: 0.351808818798, 873: -5.60908070185, 874: 0.192354978402, 875: -2.34276251143, 876: -0.642674293743, 877: 0.14749620616, 878: 0.451863544612, 879: 0.180800992989, 880: 0.290247127444, 881: -0.406347498416, 882: 1.24898013701, 883: -3.1174325739, 884: -1.92943330139, 885: -11.8991315025, 886: 7.13388345657, 887: -4.46416026927, 888: 3.77199610533, 889: -1.58682389492, 890: -0.252526287066, 891: 0.239135288758, 892: -6.62738334883, 893: 1.23680618306, 894: -0.609802095977, 895: -1.40390165587, 896: 9.32055346961, 897: 0.315340884193, 898: 1.49138917505, 899: 0.229443053602, 900: 0.967371289579, 901: 0.210540969605, 902: 0.326982442096, 903: -6.79624867598, 904: 1.4575576618, 905: 2.5088961444, 906: 0.232503670977, 907: 0.209703878911, 908: -0.137935712674, 909: 0.49714353479, 910: 0.68127871823, 911: 1.05221162296, 912: -3.67694172433, 913: 0.741179553614, 914: 0.503903922392, 915: -0.760447240205, 916: -0.359924322944, 917: 4.65256814997, 918: -3.23444599067, 919: 5.13555761406, 920: 0.915017851182, 921: 1.89639663772, 922: 1.06385413763, 923: 0.688554455103, 924: -0.766551538391, 925: 0.690767234357, 926: 1.74031994007, 927: -0.384334584889, 928: 1.11416855621, 929: 8.86644753872, 930: 0.965163843421, 931: -2.72133260409, 932: -2.21102091134, 933: 1.18691225787, 934: -0.236940563703, 935: -3.59722615311, 936: -4.7272664689, 937: 1.19513564573, 938: 1.54070338605, 939: -0.921258209696, 940: 4.60703762365, 941: 2.53597551654, 942: -13.4705847321, 943: 0.362388944742, 944: -3.45196017276, 945: -2.20035390565, 946: 5.47803586954, 947: -29.651779856, 948: 14.4046729, 949: 0.20761324269, 950: -2.96123071366, 951: -5.28370242692, 952: 1.92542213405, 953: 1.91933674224, 954: 2.26328742364, 955: -4.38153341965, 956: 1.66086009747, 957: 0.658356224937, 958: -2.82417976358, 959: 7.50194271732, 960: 1.42104665164, 961: -2.37542376824, 962: 1.90011910343, 963: 0.184797045867, 964: -9.02724212519, 965: 8.06657911917, 966: -5.40195898007, 967: -8.13133135226, 968: -1.82104383571, 969: -2.13865223433, 970: 11.1751486801, 971: -5.32084928229, 972: -0.984822405803, 973: 6.89095570104, 974: 0.462336152314, 975: 2.88601987291, 976: -7.79088128935, 977: 1.8462669659, 978: 0.32278896313, 979: 0.134983568709, 980: -3.08760767412, 981: -11.0422157644, 982: 2.0683577282, 983: 1.68001675089, 984: -1.03936294012, 985: 2.00892932031, 986: -0.540612799855, 987: 2.91183191124, 988: -4.80112828567, 989: 2.09894056243, 990: -3.54303439948, 991: -2.96979354145, 992: -1.89269146915, 993: 2.79425748303, 994: 5.89324023353, 995: -1.24466171703, 996: -2.0247308437, 997: 1.89371299396, 998: -13.1286997654, 999: 1.88451155916, 1000: 3.20663993193, 1001: 1.96864812816, 1002: 1.11743173507, 1003: 0.617031131519, 1004: 1.21931416535, 1005: -8.00072246863, 1006: 1.64322606048, 1007: 1.49418938886, 1008: 0.80672388353, 1009: -0.970965037443, 1010: 1.75786099985, 1011: 1.31954841859, 1012: 10.8557510873, 1013: 1.47063978342, 1014: -2.52317405421, 1015: 1.60590796934, 1016: 1.61852612285, 1017: -2.07353166255, 1018: -1.43880090078, 1019: 2.24570855797, 1020: -2.35641862488, 1021: 5.55592660872, 1022: -9.40272574398, 1023: 7.25659835354, 1024: 14.7335430465, 1025: 1.36945442835, 1026: -11.0733236539, 1027: 0.648946341869, 1028: -0.978526479882, 1029: 3.49010242537, 1030: 1.42150802188, 1031: -0.249522648274, 1032: -0.660501555286, 1033: 3.81035208588, 1034: -8.48225925172, 1035: 7.21184364344, 1036: 4.7302337747, 1037: 1.65462718937, 1038: 0.0170288605815, 1039: -2.43389363661, 1040: 1.65650990433, 1041: 5.86425946926, 1042: 14.508047135, 1043: 1.73101055123, 1044: -14.3915575495, 1045: -5.77759008283, 1046: 6.32701560044, 1047: 0.0965245401709, 1048: -0.148726414424, 1049: -3.19533124397, 1050: 9.36688349968, 1051: 4.83806952418, 1052: -11.092764406, 1053: -0.70787908087, 1054: 4.98316173765, 1055: 10.5437706904, 1056: 4.98794103323, 1057: -4.09573258368, 1058: 2.03338965434, 1059: 1.53799789148, 1060: -0.856469246289, 1061: 1.72063548897, 1062: -2.22076884575, 1063: -1.88409882125, 1064: 7.4012064813, 1065: -1.42720378185, 1066: -8.47131449291, 1067: 2.97072134524, 1068: 0.0690869353408, 1069: -8.39563621975, 1070: 7.95323523897, 1071: 7.97222410942, 1072: 9.82784574722, 1073: 8.42300395976, 1074: -6.88504652198, 1075: 6.06493736961, 1076: -4.02797847327, 1077: -3.4608363429, 1078: -5.35612855486, 1079: 0.608991771537, 1080: 1.20572235623, 1081: 4.57176607171, 1082: -4.06499449554, 1083: 3.70570489238, 1084: 1.44273364193, 1085: -2.36465121508, 1086: -5.30343284533, 1087: 2.18917343143, 1088: -6.92272731273, 1089: -1.40285732826, 1090: -21.2072975668, 1091: -6.31154705112, 1092: -2.77369703154, 1093: 0.145131265226, 1094: -5.9076272313, 1095: 1.71939001436, 1096: -1.67503918858, 1097: 3.31912313323, 1098: 0.229195092313, 1099: -14.5045911715, 1100: -0.364084101777, 1101: 0.646570994424, 1102: -3.15982609271, 1103: -0.379783959306, 1104: -18.4133354107, 1105: -5.07611296017, 1106: 0.983649157517, 1107: -1.5755079005, 1108: -33.1736310107, 1109: -21.2829648215, 1110: -0.884872251856, 1111: 0.600070573111, 1112: 3.57196585938, 1113: 0.201977136502, 1114: -4.01789875562, 1115: 0.11144747167, 1116: -3.02547615598, 1117: 0.446954923119, 1118: 1.53851491503, 1119: 7.27478115896, 1120: -0.212817469871, 1121: 7.89699664343, 1122: -6.16253657711, 1123: -2.78366500126, 1124: -2.99553843857, 1125: -9.62792287629, 1126: -11.798387569, 1127: -7.18599688011, 1128: 1.79574760744, 1129: 9.97123595974, 1130: 0.120598994119, 1131: 1.66626461909, 1132: 3.85253213461, 1133: -2.67429159587, 1134: 5.52931033028, 1135: -6.70801072857, 1136: -0.538727398319, 1137: 14.5198294254, 1138: 4.26110426531, 1139: -0.306654741207, 1140: -0.79768531915, 1141: -0.311465885772, 1142: 0.0635323804083, 1143: 0.534689162284, 1144: -1.62550818236, 1145: 0.576709342979, 1146: -0.26027791528, 1147: -0.21300466591, 1148: 5.12855020523, 1149: 2.18492924923, 1150: 8.98976677712, 1151: 0.125024094635, 1152: 2.18841638081, 1153: 8.76027348354, 1154: -0.141972587044, 1155: -0.695168417919, 1156: -0.00518851558712, 1157: 3.9198542397, 1158: 1.41231018194, 1159: 3.52303622995, 1160: 14.2644154259, 1161: -3.81720626293, 1162: 0.621788325861, 1163: -2.41043923502, 1164: -11.4288994304, 1165: 3.58126213347, 1166: 5.80896648127, 1167: -1.92070738107, 1168: -2.03590417054, 1169: -8.42260820376, 1170: -21.9789538079, 1171: 2.3126697535, 1172: -10.7806010449, 1173: -0.263219785084, 1174: 0.250616144304, 1175: -1.7105975513, 1176: -9.70156062189, 1177: -0.201166138245, 1178: 0.985660037081, 1179: -1.69059784078, 1180: -6.01550683585, 1181: -2.62157893106, 1182: -8.39232745705, 1183: -3.4805926468, 1184: 4.23459711113, 1185: 17.4514605124, 1186: 1.35470053455, 1187: -4.63387502303, 1188: 5.18854674845, 1189: 20.1888429691, 1190: -1.30108050889, 1191: 0.222857815256, 1192: -6.78304444352, 1193: -0.537874955303, 1194: -0.138926003946, 1195: -15.0256884031, 1196: -0.369707104756, 1197: -0.158589308925, 1198: -7.64464514203, 1199: -0.713354465813, 1200: -0.337820693007, 1201: -0.134378837819, 1202: -0.34768227559, 1203: -0.553421454476, 1204: -4.27503483168, 1205: 0.272766570188, 1206: 0.42809498205, 1207: 0.314375793451, 1208: 0.797478458042, 1209: -0.343994697231, 1210: -1.75168087668, 1211: 7.28753609407, 1212: 27.1301900222, 1213: -10.0243424286, 1214: 1.39848377069, 1215: 3.55531507491, 1216: -0.0822243615614, 1217: 1.97196118064, 1218: 8.36733541708, 1219: -0.17885355065, 1220: -12.2753280239, 1221: -0.15876449588, 1222: -5.63611847374, 1223: -7.46076327287, 1224: -3.07459042417, 1225: -0.520885136125, 1226: -0.17809805103, 1227: -0.185212956486, 1228: -0.414495045409, 1229: -0.174403657901, 1230: -3.07429521742, 1231: 4.43130406242, 1232: 3.32879170227, 1233: 6.1668595847, 1234: 0.584037889382, 1235: -2.33310801382, 1236: 7.2411218299, 1237: -1.87446259226, 1238: 7.77331978159, 1239: 3.27840324781, 1240: -0.261153832047, 1241: -5.97591698612, 1242: 3.20774889511, 1243: -2.7293767101, 1244: -1.15219558326, 1245: -9.45080640828, 1246: -0.261024403132, 1247: -0.202049541535, 1248: -0.245707944379, 1249: 2.98764875657, 1250: -0.249460605896, 1251: -0.221846262405, 1252: 7.20377180721, 1253: -0.266450128623, 1254: 0.772170717384, 1255: -0.280810401008, 1256: -0.278256810764, 1257: 2.87852608528, 1258: -0.211714831028, 1259: 2.99470921791, 1260: 3.82206904721, 1261: 7.31148036997, 1262: -0.199700170532, 1263: 1.15345960169, 1264: 13.3325301547, 1265: 1.47701945118, 1266: -18.5943579061, 1267: -0.244597563998, 1268: -4.98143027334, 1269: 1.026196881, 1270: -1.63005037119, 1271: 0.127292470129, 1272: 0.273092836484, 1273: -0.65784298545, 1274: -0.430743646955, 1275: -0.122760871796, 1276: 6.69795278067, 1277: 2.52269339576, 1278: -13.945080573, 1279: -0.484649463536, 1280: 14.489448527, 1281: 7.38626951635, 1282: 1.19316261025, 1283: 15.7047574011, 1284: -5.40916624818, 1285: -19.8035961643, 1286: -1.37322419199, 1287: 0.905057571005, 1288: -9.01300428885, 1289: -1.18015352162, 1290: 3.36452660136, 1291: -7.2899364443, 1292: 0.277786624993, 1293: -10.2932394344, 1294: -6.175584565, 1295: 1.90392646722, 1296: 12.0126911252, 1297: 0.637631440986, 1298: -20.1814608785, 1299: -2.5953370133, 1300: -25.0484985479, 1301: 0.220257699408, 1302: 3.19144452597, 1303: 2.03420758573, 1304: 11.0621999068, 1305: -0.00956329443887, 1306: -0.149073582656, 1307: -9.85146572091, 1308: 20.2039252829, 1309: -0.130510282991, 1310: -1.92983992547, 1311: -0.292992669605, 1312: -0.424785671401, 1313: -7.97442684816, 1314: 6.63029773833, 1315: 4.80400874719, 1316: -2.35424946659, 1317: -0.803943449753, 1318: -0.0695047657569, 1319: 3.85836822881, 1320: 0.597632051374, 1321: -35.2200422719, 1322: -19.401437226, 1323: 6.64185761908, 1324: 7.25196087341, 1325: 3.41721203929, 1326: 1.31402889451, 1327: 3.36619004596, 1328: 2.10504847723, 1329: 4.14334519092, 1330: 3.30546696934, 1331: 2.00210668897, 1332: -7.81443340708, 1333: 5.4475651083, 1334: 0.185976738585, 1335: 0.709000795026, 1336: 1.91647280865, 1337: -2.792343219, 1338: -2.27514618314, 1339: 1.28019785761, 1340: 9.37640668053, 1341: -6.53172863539, 1342: 3.70160954581, 1343: -3.04923869225, 1344: 2.51091120393, 1345: 5.67159844238, 1346: 2.23150866913, 1347: 0.548074219188, 1348: 1.19654241386, 1349: 6.08534869788, 1350: 3.75513506569, 1351: 6.69287543393, 1352: -2.75362722615, 1353: -0.0163629740954, 1354: 14.5807936259, 1355: -0.184302202208, 1356: -0.224502556276, 1357: 2.43612217647, 1358: -4.34896131408, 1359: -0.264377551402, 1360: -0.656972760345, 1361: -6.93378982625, 1362: -0.236739637924, 1363: 2.57688541922, 1364: 0.340979511936, 1365: -0.369791057434, 1366: 0.925267745783, 1367: -3.2221799267, 1368: 4.32218774507, 1369: -8.99125091045, 1370: -0.980649955281, 1371: 11.3866546023, 1372: -9.64690803526, 1373: 14.6361427007, 1374: 0.405297177117, 1375: 16.9721363112, 1376: 0.302044906988, 1377: 1.22574304169, 1378: -5.50926808755, 1379: -0.663096154505, 1380: 0.414502731673, 1381: 9.4649700024, 1382: -10.4404994102, 1383: 6.78690617963, 1384: 5.17354179273, 1385: -4.6782524048, 1386: -0.369117042611, 1387: 1.54780193715, 1388: 4.0947063269, 1389: 13.7029081846, 1390: 6.1534264025, 1391: -8.31585391926, 1392: -5.53170054449, 1393: 8.85033084828, 1394: 2.11086639359, 1395: -4.10470587122, 1396: 13.345907573, 1397: -0.229568631418, 1398: -5.66835411125, 1399: 2.79831012871, 1400: 5.60201221716, 1401: 18.3273120848, 1402: -0.220718206595, 1403: -1.9788354459, 1404: 1.68445819335, 1405: 0.66802549156, 1406: -2.89683650443, 1407: -3.90949453651, 1408: -6.7930299214, 1409: 6.40144444197, 1410: 6.00827932678, 1411: -4.52147964263, 1412: 6.1954052684, 1413: 15.2159436565, 1414: 4.65887777352, 1415: 6.82168235318, 1416: -5.00022065202, 1417: -0.677219448681, 1418: 3.01746113335, 1419: -3.44952348157, 1420: 5.09117110025, 1421: -7.27168946805, 1422: -4.18038453285, 1423: -3.69049817253, 1424: -2.42637049298, 1425: 5.07548705677, 1426: -0.19089965204, 1427: -6.32140396687, 1428: -4.24368847841, 1429: 2.24031721913, 1430: 1.82798213526, 1431: 4.79833537594, 1432: 2.63076765785, 1433: -1.69067609273, 1434: -4.33250249033, 1435: 2.05650867969, 1436: 0.522303150613, 1437: 3.03926856409, 1438: -4.66912033331, 1439: -10.548975749, 1440: 2.50077147758, 1441: 3.85695452483, 1442: -3.70277706751, 1443: 0.197055549348, 1444: -1.43162148923, 1445: -8.67403885023, 1446: 1.56812658956, 1447: 3.13784802458, 1448: -2.95869896238, 1449: 1.71236797312, 1450: 5.8865656665, 1451: -1.77711167501, 1452: -0.991924928501, 1453: 2.73760367338, 1454: -0.409246605462, 1455: -9.98969056012, 1456: -2.87414337745, 1457: -1.76239568685, 1458: 0.463851167838, 1459: -2.00499872143, 1460: 7.12646993616, 1461: -7.96082898588, 1462: 0.0994245599728, 1463: -4.79501858449, 1464: -0.701479537812, 1465: 2.7512233686, 1466: -0.981419630768, 1467: 0.428042836951, 1468: 2.40670134635, 1469: -0.930691924427, 1470: 0.19729510035, 1471: -0.165060941488, 1472: -2.92929610411, 1473: 0.825409322375, 1474: 0.41698295135, 1475: 4.39653120983, 1476: 9.07008922671, 1477: 0.492744863311, 1478: 11.8099135247, 1479: -0.572196181518, 1480: 0.573338933965, 1481: 2.15249541312, 1482: -7.43737493889, 1483: 0.547010865383, 1484: 5.23083607828, 1485: -0.78554573711, 1486: 2.05051336589, 1487: -0.0141040869922, 1488: -0.544276940762, 1489: 10.1113879589, 1490: 0.0522269296341, 1491: -0.298560532632, 1492: 0.427774316054, 1493: 2.11428781119, 1494: 0.211576577705, 1495: 0.239489853525, 1496: 0.593885885499, 1497: 10.5249835175, 1498: 0.513696695849, 1499: 0.783152929223, 1500: 1.56015835919, 1501: -0.226789422062, 1502: 1.27916799677, 1503: 1.07686050011, 1504: -0.665537239444, 1505: 0.363159091794, 1506: 8.12521320322, 1507: 0.790405478691, 1508: -0.519882263971, 1509: -29.2733717112, 1510: 0.540735208636, 1511: 12.2251808534, 1512: -4.37127173716, 1513: 0.00691617443554, 1514: 9.86581138619, 1515: 8.60759824102, 1516: 4.50692203557, 1517: 0.328161373097, 1518: -7.47055469667, 1519: -1.67542688261, 1520: -1.49277870247, 1521: 3.40901922737, 1522: -0.150980091327, 1523: -0.154104687202, 1524: 3.65819662167, 1525: 3.32644358569, 1526: -0.217166691264, 1527: 0.495763460113, 1528: -0.545144465115, 1529: 3.54766319651, 1530: -2.38765552002, 1531: -2.33266563706, 1532: -6.31052001276, 1533: -0.680196365788, 1534: 0.253448400966, 1535: 0.607210968612, 1536: -1.93312085376, 1537: 4.27417183927, 1538: -4.19955044103, 1539: -0.897254739781, 1540: 0.437338670246, 1541: -2.52516455077, 1542: -0.0675580998759, 1543: -0.124428387888, 1544: 4.77702865179, 1545: -0.113806124361, 1546: 0.139433587073, 1547: 0.714915723439, 1548: -0.146441380848, 1549: -0.122061619541, 1550: -0.236959947038, 1551: -0.0673549510613, 1552: -0.205499497505, 1553: 4.37359855513, 1554: -0.520677313312, 1555: -0.0260201868588, 1556: -0.39022066024, 1557: -3.19074268943, 1558: -0.61816679484, 1559: -0.95364425515, 1560: 3.65625235679, 1561: 9.46732428816, 1562: 0.334864767993, 1563: -1.91204447732, 1564: 1.10849454408, 1565: -0.24115323792, 1566: 1.31305013733, 1567: -2.03349324392, 1568: -0.110554235949, 1569: -5.6570636068, 1570: -0.257087863947, 1571: 5.98466811709, 1572: -3.46312952501, 1573: -1.22971598895, 1574: -0.145488727447, 1575: -0.292195408982, 1576: -0.264505126251, 1577: -0.344119692152, 1578: -0.227419980933, 1579: 0.432300194183, 1580: 13.1851321079, 1581: -2.2135545808, 1582: 1.00053279602, 1583: 10.865999009, 1584: -2.84874879902, 1585: 1.19756664317, 1586: -0.664656916172, 1587: -3.90387547083, 1588: 3.65705615604, 1589: -0.0927626248179, 1590: 6.75329240013, 1591: 0.514251987175, 1592: 1.83300877052, 1593: 0.307477146545, 1594: -2.08535830473, 1595: -0.173727986414, 1596: -0.658741113685, 1597: -0.238047477472, 1598: 3.77837923142, 1599: -0.252726181213, 1600: -0.117470194764, 1601: 4.916364455, 1602: -0.642723379485, 1603: 0.780177380346, 1604: -0.253039252065, 1605: -0.282052877731, 1606: -0.958950719471, 1607: -0.190592177134, 1608: 4.95665916945, 1609: 1.935269993, 1610: 0.10634792515, 1611: 0.105423962735, 1612: 0.316511303295, 1613: -17.3149179539, 1614: 0.783320443433, 1615: 10.7066291272, 1616: -7.73773078482, 1617: 0.158824015683, 1618: -1.59580175273, 1619: 0.374584882454, 1620: -0.219065618097, 1621: 2.53878647007, 1622: -0.705383240409, 1623: -2.14480111553, 1624: -0.69011444103, 1625: 2.07464415677, 1626: 0.886209808438, 1627: -5.65504526764, 1628: 0.219633192155, 1629: 0.538915136279, 1630: -1.62123151189, 1631: -0.70862953596, 1632: -12.6157100856, 1633: 1.00969792216, 1634: -26.580615468, 1635: 0.515844158755, 1636: 0.606670057272, 1637: 5.94491585594, 1638: -1.04041076759, 1639: -3.57775583224, 1640: 1.62325195499, 1641: -1.9454493465, 1642: -4.36435311665, 1643: 6.70480868157, 1644: 10.1705758176, 1645: 7.72069540728, 1646: -4.52822240637, 1647: -5.54091019414, 1648: -0.706300415657, 1649: 12.8570504709, 1650: -2.64177865841, 1651: -7.48308237933, 1652: -1.02946314023, 1653: 1.13396882705, 1654: 0.368266573615, 1655: 0.198752537052, 1656: -0.833216795835, 1657: 3.62707118537, 1658: -0.407338552579, 1659: -0.638463977678, 1660: -0.357733013103, 1661: -0.436868624419, 1662: 3.66614856741, 1663: 8.99207338488, 1664: 0.548540328377, 1665: 3.69295701457, 1666: -5.76191429344, 1667: 0.746055466918, 1668: -6.4198301148, 1669: 2.44275465308, 1670: 10.5160493733, 1671: 0.48556999384, 1672: -2.74486145058, 1673: 5.3364369524, 1674: 1.90043748888, 1675: 2.37001493811, 1676: -0.964984936658, 1677: 1.74172456222, 1678: -0.477494284261, 1679: -0.399922786863, 1680: 9.51828757845, 1681: 5.58066912346, 1682: -6.85659316252, 1683: -0.0425741230085, 1684: 2.27361018973, 1685: 2.56906857677, 1686: -2.40590404545, 1687: -2.04808069258, 1688: 13.6909365385, 1689: -3.95570735204, 1690: 3.68752635776, 1691: -7.18705581171, 1692: -2.79954370611, 1693: -1.80050201169, 1694: 3.92870322572, 1695: -2.69212136164, 1696: 3.53444316777, 1697: -3.07127941452, 1698: -4.04908481837, 1699: 0.0804196412177, 1700: -7.68103040545, 1701: -5.2815829239, 1702: -0.415037355226, 1703: 3.01153918873, 1704: -0.780637568186, 1705: -0.867802429155, 1706: -0.602568089592, 1707: -6.97642392899, 1708: -0.814150146049, 1709: -0.517908313353, 1710: 1.09646911264, 1711: -0.583196510839, 1712: -5.73706307005, 1713: -0.477047393381, 1714: -0.706315400226, 1715: 4.76817387094, 1716: -1.03339518916, 1717: 6.1481250708, 1718: -5.57085369552, 1719: 7.40920370575, 1720: 2.14951652357, 1721: -11.9428302425, 1722: 3.26095476852, 1723: -0.361429555307, 1724: -2.43250733219, 1725: -8.06686101562, 1726: 1.90438062641, 1727: 3.17662281681, 1728: 0.920767245799, 1729: 0.365171159892, 1730: -0.221095761196, 1731: -8.07161163016, 1732: 4.29214322705, 1733: -10.4438586032, 1734: 1.84284663785, 1735: -0.710989575028, 1736: 6.03946908466, 1737: 3.41207977223, 1738: -7.26431234518, 1739: -14.3931017882, 1740: -5.1656468649, 1741: 2.12731095418, 1742: 13.7015494107, 1743: 6.96216239723, 1744: -1.4671313156, 1745: 3.39915473691, 1746: -0.482748287259, 1747: -18.9477364471, 1748: 10.0824916745, 1749: -5.7513297657, 1750: -2.98069297058, 1751: -12.7189838932, 1752: 22.2468906336, 1753: 39.8818162223, 1754: -12.8881676974, 1755: -0.833369485779, 1756: -34.0926303786, 1757: -10.7886165839, 1758: 17.0590370441, 1759: 1.03760439339, 1760: -8.27406366535, 1761: -1.77457047911, 1762: -1.1922041461, 1763: 5.99019055264, 1764: -28.1007437187, 1765: 7.46984409972, 1766: -2.04861895026, 1767: -28.8370325973, 1768: -21.1266940529, 1769: -5.91706517928, 1770: 16.1280182377, 1771: -8.90793097904, 1772: 22.0930727184, 1773: -15.532818783, 1774: -6.61333419227, 1775: -10.0738858335, 1776: -12.4609170592, 1777: -0.225841385993, 1778: -14.5365235826, 1779: -4.02903623426, 1780: 0.972861791429, 1781: -0.286599327507, 1782: -4.65746461925, 1783: -8.47378099572, 1784: 0.707936532401, 1785: 22.2329102635, 1786: 60.3260974602, 1787: -2.68434012728, 1788: -0.293372812261, 1789: 5.98497244939, 1790: -21.8678032211, 1791: 4.09035034506, 1792: -5.74753862203, 1793: -3.38508963092, 1794: 3.01799640373, 1795: 3.2996488039, 1796: -8.49911838739, 1797: -21.6619561089, 1798: -2.54166934695, 1799: -9.41698741668, 1800: -22.2735758988, 1801: -1.72824035759, 1802: -0.135503896666, 1803: 6.2037749026, 1804: 5.28612739288, 1805: 10.1432054596, 1806: -34.9952560372, 1807: -4.35768955027, 1808: -26.1359113608, 1809: 6.78788312108, 1810: 1.87778779143, 1811: 0.960317916206, 1812: 2.1945609973, 1813: -4.83377025553, 1814: -18.8027847982, 1815: -3.32408415875, 1816: -3.42187978938, 1817: -1.97460307653, 1818: -2.64250069932, 1819: 2.78786496795, 1820: -3.06128898508, 1821: 1.79547497165, 1822: 0.0166974923181, 1823: -1.35686957156, 1824: 16.5849973313, 1825: 11.8343778856, 1826: -3.20427460461, 1827: 27.9952830477, 1828: -1.34119565522, 1829: -2.39417352676, 1830: 13.8186488362, 1831: -3.83283626979, 1832: 6.66620647951, 1833: -10.8130289742, 1834: -2.37294147779, 1835: -2.34375906031, 1836: -1.86368653439, 1837: -2.41286216011, 1838: -21.8844839357, 1839: -2.7509151717, 1840: -3.20808480889, 1841: -3.02192532192, 1842: -6.94231155457, 1843: -3.51238970096, 1844: 3.64524768108, 1845: -2.37461931862, 1846: -1.09423018594, 1847: -2.98985985666, 1848: 3.84327355231, 1849: -2.08652392706, 1850: -1.32555077313, 1851: 14.2645629999, 1852: -2.86121269016, 1853: -1.98148913365, 1854: -3.01690029724, 1855: 15.577988209, 1856: -3.10042036257, 1857: 17.4209191458, 1858: -4.91139502932, 1859: 10.4337727819, 1860: -1.78631917224, 1861: -11.3893675953, 1862: -8.84660080171, 1863: 20.5787695267, 1864: 0.856879592592, 1865: 13.4294266555, 1866: 4.1895153195, 1867: 12.0147945685, 1868: 1.73370304874, 1869: 5.42666240103, 1870: 6.15762485207, 1871: -0.656950385497, 1872: -0.561809531152, 1873: -0.779048790573, 1874: -8.26375857935, 1875: -0.505582218059, 1876: -3.30991863034, 1877: 2.28823638139, 1878: -15.9136592672, 1879: -1.51916906329, 1880: 3.28928406696, 1881: -2.25474223153, 1882: -4.52458096823, 1883: 20.2853690316, 1884: -2.26751051563, 1885: 3.7620854913, 1886: 11.1291697483, 1887: 5.68634027232, 1888: 7.84391765032, 1889: -0.971588024787, 1890: 3.03091613214, 1891: -0.291359701973, 1892: -0.522210765099, 1893: -7.78637086351, 1894: -0.636345410403, 1895: -2.65027109189, 1896: -7.03527093998, 1897: 0.115888242794, 1898: -0.538739986784, 1899: -0.73947906559, 1900: -0.5742642306, 1901: -2.13291413059, 1902: -14.220531266, 1903: -0.545813656563, 1904: -1.16934359986, 1905: -2.60332278537, 1906: -10.0616062844, 1907: -2.30927927475, 1908: -10.0721220666, 1909: -22.333112838, 1910: 24.2475135221, 1911: 47.0601654668, 1912: 5.49144603976, 1913: 0.0586689547711, 1914: 13.2712913153, 1915: -0.0951821367058, 1916: -20.035494383, 1917: -0.561708048433, 1918: -3.96685533599, 1919: -0.644346787628, 1920: 9.93834198499, 1921: 3.08880700449, 1922: -8.54947404486, 1923: 2.73627597313, 1924: -0.516627814713, 1925: -0.750548143746, 1926: -0.396816112357, 1927: -0.850463309834, 1928: 0.893518405113, 1929: 31.5148869313, 1930: 5.04301376761, 1931: 32.5940634559, 1932: -3.85262424052, 1933: -1.25045445179, 1934: -16.4935948287, 1935: -6.46578063506, 1936: -4.57827863356, 1937: 5.14462757463, 1938: -0.568429071152, 1939: -17.9660488885, 1940: -5.30121277678, 1941: -3.82335665621, 1942: 16.3608187034, 1943: -17.7197456584, 1944: -0.648344076488, 1945: -2.72121640961, 1946: -0.493004631714, 1947: 1.70720325025, 1948: -0.528428345202, 1949: -0.495436900445, 1950: 27.8884189818, 1951: -2.69013437632, 1952: 6.17617695216, 1953: -0.645644589922, 1954: -0.685325926888, 1955: -3.11274184649, 1956: -0.593393684131, 1957: -5.43263870173, 1958: 5.31232890044, 1959: 21.0428489939, 1960: -0.936391676376, 1961: -1.51775124026, 1962: 7.32445719411, 1963: -12.2185679219, 1964: -5.46533797976, 1965: -20.0381758472, 1966: 9.55193629452, 1967: -1.15278874202, 1968: -16.2334982317, 1969: -1.18786718903, 1970: 6.12175824289, 1971: -0.603044834504, 1972: 0.284974716143, 1973: -5.17071627295, 1974: -1.31892078537, 1975: 1.30353967823, 1976: 30.6053131959, 1977: -1.19442660899, 1978: 26.4591092323, 1979: 14.3260851829, 1980: 4.32074737727, 1981: -1.16263825892, 1982: -25.4675054691, 1983: -3.20549343088, 1984: -3.10863064981, 1985: -3.45262167721, 1986: -1.62831121664, 1987: -1.16887298786, 1988: 82.214011643, 1989: -8.02395267589, 1990: 13.0037886977, 1991: 7.92256505831, 1992: 11.0765903081, 1993: 23.0931717569, 1994: 23.1877825129, 1995: 4.56241002189, 1996: -11.7496252433, 1997: -6.55379622782, 1998: -3.74973570163, 1999: -5.29991982842, 2000: -5.94128985591, 2001: -2.4734949925, 2002: 22.5313314929, 2003: -2.71647336744, 2004: -2.58055209112, 2005: -9.47309789593, 2006: 15.3507232847, 2007: -2.49195316375, 2008: 0.00861917267276, 2009: -2.85219569533, 2010: -2.23760508697, 2011: -18.1423373011, 2012: 30.2547619246, 2013: 11.2817499084, 2014: -0.701490883941, 2015: -9.80536939459, 2016: -1.7727008829, 2017: 0.197208518796, 2018: -4.10093774919, 2019: -56.0573906484, 2020: -1.42589453598, 2021: 4.31299971809, 2022: -10.199292227, 2023: 5.4572370377, 2024: -0.034848367891, 2025: -21.6203666684, 2026: -1.30520744962, 2027: 8.73334929691, 2028: 8.32200445552, 2029: 19.2635637426, 2030: 44.7482017937, 2031: -34.2511090239, 2032: -2.95540835364, 2033: 9.93117355456, 2034: -4.24502567725, 2035: -10.8765593581, 2036: 5.06298641225, 2037: -26.8717821896, 2038: -11.9214290226, 2039: -14.6016955366, 2040: -20.0523111573, 2041: -5.96679398086, 2042: -5.04301419757, 2043: 0.985203029595, 2044: -3.90549374472, 2045: 2.62851485964, 2046: -10.6016013787, 2047: 22.1604153352, 2048: 6.66523209497, 2049: 17.8201163555, 2050: 5.33284245006, 2051: -3.19790607464, 2052: 26.2451318353, 2053: -2.83671783823, 2054: -2.66615348995, 2055: -3.64140514691, 2056: -7.75172813725, 2057: -2.1396119411, 2058: -2.02688138111, 2059: -15.641766893, 2060: -2.50983857252, 2061: 30.9814074924, 2062: -2.8091558621, 2063: -2.58038135686, 2064: -34.7617450768, 2065: -2.57031056943, 2066: -7.02852762066, 2067: 3.11308668718, 2068: -1.68183737105, 2069: -2.84042800368, 2070: -9.18593321412, 2071: 21.9863654681, 2072: 13.2009208764, 2073: -32.1482690281, 2074: -0.0543778969996, 2075: -5.87651193126, 2076: 55.8365720597, 2077: 5.78493753332, 2078: 0.789389316262, 2079: -18.7604401569, 2080: 5.60894849605, 2081: 6.49155862525, 2082: 14.3390477672, 2083: 14.6295147663, 2084: -2.49982480684, 2085: 11.7649969118, 2086: -10.5792554229, 2087: -9.08656575688, 2088: 0.3640885471, 2089: 14.1502031106, 2090: -0.960973510954, 2091: 22.4319294734, 2092: 6.21890218255, 2093: 35.5248188475, 2094: 10.6135165196, 2095: -0.369908592348, 2096: 0.54056534364, 2097: 10.7320898826, 2098: -1.9565613182, 2099: 7.5772688625, 2100: 0.52282265087, 2101: -0.720305665743, 2102: -0.763022172198, 2103: -1.20951632779, 2104: 0.320439643601, 2105: -4.56563106858, 2106: 16.292159709, 2107: 2.13136286808, 2108: -6.18542537772, 2109: -1.06357818248, 2110: 7.18951646976, 2111: 11.283574538, 2112: 0.749998083614, 2113: -4.29054981654, 2114: -2.92311691528, 2115: -4.59396600955, 2116: 18.3681901427, 2117: -2.65479815582, 2118: -3.21316236374, 2119: -1.67214393785, 2120: 2.71360499996, 2121: -1.26197431036, 2122: 2.48244568795, 2123: -0.515506837146, 2124: -6.30882308526, 2125: 4.9788692748, 2126: -9.30055447238, 2127: 1.51237389697, 2128: 3.67123668955, 2129: 3.06358245835, 2130: 3.74526810858, 2131: -0.620905085798, 2132: 21.0471316056, 2133: -0.225175879091, 2134: -0.516633749118, 2135: 2.86208197664, 2136: 6.12555945449, 2137: -0.393800165627, 2138: 7.46894270142, 2139: 4.08333891957, 2140: -4.04972140993, 2141: 9.85582787124, 2142: 0.00685708722177, 2143: -7.440137197, 2144: -0.122391447054, 2145: 4.49788201802, 2146: 17.1371641106, 2147: 2.4912844393, 2148: 7.77848999262, 2149: -2.67263549222, 2150: 1.09467009145, 2151: -6.85464166242, 2152: -0.0103798531963, 2153: -7.18176747067, 2154: 2.23856715902, 2155: 2.00892102873, 2156: 2.49054197099, 2157: -3.75881201046, 2158: 3.25852650461, 2159: -1.60429412295, 2160: 1.2007580281, 2161: 1.25847101728, 2162: -1.82367588294, 2163: 4.47172963098, 2164: -1.04305615056, 2165: -2.5515527122, 2166: -2.24589206044, 2167: -2.35619261287, 2168: 8.09069013642, 2169: -0.517580655012, 2170: -2.20671323405, 2171: -2.32014441254, 2172: 0.216454981093, 2173: 9.10106118533, 2174: -7.66081631248, 2175: 2.56466633341, 2176: 15.0278534851, 2177: -2.25449499427, 2178: -1.10334979493, 2179: 5.87440949198, 2180: -0.706016759479, 2181: 0.097098969494, 2182: -3.05987142064, 2183: -2.10957464795, 2184: 2.31861190937, 2185: -1.52954825576, 2186: -1.36227731178, 2187: 10.163471645, 2188: -2.97640366487, 2189: -1.40592731579, 2190: -1.57247445409, 2191: 3.99391063749, 2192: -2.26709880463, 2193: 0.254139533448, 2194: -0.822712543301, 2195: 6.11868256812, 2196: 2.36625200099, 2197: 4.46220347692, 2198: 0.458884355872, 2199: 1.76459518375, 2200: -5.27954298255, 2201: -1.17999652259, 2202: 4.76001164022, 2203: -2.69700188069, 2204: 3.60193164735, 2205: 3.1054411689, 2206: -0.0395254100182, 2207: 2.28911550846, 2208: -2.0401798585, 2209: -0.872814717319, 2210: -4.33859755943, 2211: 2.74304645192, 2212: -0.749930003883, 2213: -1.47425017257, 2214: 3.35364864876, 2215: 8.38219208557, 2216: 0.799858186715, 2217: 5.6977225961, 2218: -2.09503334317, 2219: -16.2576203844, 2220: -0.397983177169, 2221: 0.0107591627232, 2222: -3.32330115121, 2223: -11.0730667614, 2224: -0.286077791692, 2225: -2.72307804533, 2226: -0.571385353923, 2227: -5.96170728465, 2228: -1.71927581915, 2229: -2.9504578443, 2230: 6.44035816097, 2231: 0.848430996601, 2232: 20.1266323905, 2233: 3.17744619863, 2234: -8.11743305655, 2235: -1.35531240984, 2236: 0.741842388167, 2237: -11.9685659198, 2238: -1.41008726226, 2239: -12.8553518808, 2240: -0.413180001169, 2241: -0.480052920582, 2242: 2.1504128727, 2243: -0.149655096892, 2244: -1.30815767401, 2245: -1.21310068726, 2246: -0.354842053829, 2247: -0.410824688438, 2248: -0.431805056953, 2249: -0.265944120113, 2250: -2.43495591919, 2251: 6.82672971396, 2252: -0.387109561696, 2253: -0.403055739567, 2254: -0.3492030005, 2255: 1.87724763572, 2256: -3.15068442989, 2257: -11.0518621535, 2258: 5.77214954812, 2259: 6.53675179662, 2260: 2.28069177562, 2261: 1.22752520183, 2262: 1.50590035718, 2263: 1.64950919446, 2264: -2.10954262341, 2265: 1.88882245774, 2266: -0.270696347729, 2267: -11.210154622, 2268: -0.2000516475, 2269: -7.31850718663, 2270: 13.0616525408, 2271: -15.5065100515, 2272: 0.199796926148, 2273: -0.356695538896, 2274: -0.501850921715, 2275: -0.398621281169, 2276: -0.400977516021, 2277: -1.5388335964, 2278: -10.7840196323, 2279: -0.721316005796, 2280: -1.41798874957, 2281: 23.0518652008, 2282: 0.287575459761, 2283: 1.43632364588, 2284: 6.87390381543, 2285: 1.56133171817, 2286: -1.9692106834, 2287: -0.210522489434, 2288: 2.69542234399, 2289: -1.18501041438, 2290: 1.47263375937, 2291: -0.711618862855, 2292: 4.03384237042, 2293: -0.397305029212, 2294: -0.132034472185, 2295: -0.164182309632, 2296: 1.34300222707, 2297: -0.23808768135, 2298: -0.26749015127, 2299: -1.0735314414, 2300: 0.285300263559, 2301: 0.507639657275, 2302: -0.642491815935, 2303: -0.234163580129, 2304: 0.223376605292, 2305: -0.351778263917, 2306: -0.610059336824, 2307: -4.64811069239, 2308: -8.36792572306, 2309: -0.524520851729, 2310: -1.07640652247, 2311: 0.218761069602, 2312: -4.23891962108, 2313: -6.78335660717, 2314: 2.97948112023, 2315: 0.306457082888, 2316: 2.31268141202, 2317: -0.256889876381, 2318: -0.309204659943, 2319: 4.01684967688, 2320: -0.423169813307, 2321: 12.7689665254, 2322: 0.433307604906, 2323: -2.96426457907, 2324: -1.10455401481, 2325: -8.54214612737, 2326: 0.647940638095, 2327: -0.147081734505, 2328: 13.1320763155, 2329: 5.25032212385, 2330: -10.7187066254, 2331: 3.71322759449, 2332: 3.08890416385, 2333: 2.12323535348, 2334: -2.83895941251, 2335: 0.132100369367, 2336: 0.189179592509, 2337: -2.66902127071, 2338: -5.80913397605, 2339: -3.45919778983, 2340: -3.87057159404, 2341: -6.24632063824, 2342: -9.23909028332, 2343: 9.37046894503, 2344: -3.84959623619, 2345: -10.1215723111, 2346: 4.5783530836, 2347: -9.97495724101, 2348: 7.70649289367, 2349: 1.86875004336, 2350: -3.50668947114, 2351: -3.31330043337, 2352: -2.43296913288, 2353: -1.45766766823, 2354: -1.94995977519, 2355: 3.51096680429, 2356: -0.726065762189, 2357: 0.862664522939, 2358: 3.38133502192, 2359: -2.57319674705, 2360: 4.44531155728, 2361: 3.57598046558, 2362: -0.917546396471, 2363: 0.174384625138, 2364: -0.294679920958, 2365: -1.35941086737, 2366: -4.34742262894, 2367: -8.31390549637, 2368: 30.5355875302, 2369: -5.06362577338, 2370: 5.04202298281, 2371: 6.17718824146, 2372: 5.24758677446, 2373: -6.12743661778, 2374: -1.0026763611, 2375: 1.36246313121, 2376: 3.72579278708, 2377: 6.8406896872, 2378: -13.5924225619, 2379: 0.929320536554, 2380: 4.04981893618, 2381: -1.92927312668, 2382: 1.26306749327, 2383: -0.0937205213248, 2384: -0.302675279487, 2385: 3.52355490183, 2386: 4.81334699793, 2387: 10.0922556734, 2388: -1.8823326823, 2389: -5.55438334881, 2390: -9.46420644261, 2391: 1.1744745617, 2392: 6.82197597763, 2393: -0.252545371023, 2394: -0.0415969906155, 2395: -1.48759111865, 2396: -2.48289663881, 2397: 2.51097845861, 2398: 4.94841797041, 2399: -2.96075773098, 2400: -1.54379153088, 2401: -1.32905618601, 2402: -0.0789627527211, 2403: -0.185178238405, 2404: -2.47058325087, 2405: 3.59032575634, 2406: -0.089411007272, 2407: -1.54642505134, 2408: 4.08161035943, 2409: 0.52017184025, 2410: -0.800380246357, 2411: 8.90903485452, 2412: -1.85012759172, 2413: -12.7371916176, 2414: -1.41839303147, 2415: -1.83485546842, 2416: -4.39186339198, 2417: -11.6840764851, 2418: 14.6279992543, 2419: 1.01012219184, 2420: 15.0986533618, 2421: -0.291267532948, 2422: -8.26516594136, 2423: -6.93642161972, 2424: -0.524531164962, 2425: -3.66775664965, 2426: 3.21047261812, 2427: 2.96565411799, 2428: 2.11650201186, 2429: -5.02591171734, 2430: 1.19686473783, 2431: 3.22927469671, 2432: 15.2284962709, 2433: -1.50454239205, 2434: -3.39020404994, 2435: -2.61603133999, 2436: -4.46625973945, 2437: 1.90681945397, 2438: -3.91383926967, 2439: -1.42725635159, 2440: 14.1308624439, 2441: -1.08331609888, 2442: 1.78785930968, 2443: -6.75973841485, 2444: -0.768784825266, 2445: -52.2548457767, 2446: 15.7209571944, 2447: 17.9466535103, 2448: 6.30754194063, 2449: -30.2895978123, 2450: 28.2632030034, 2451: -10.9190756707, 2452: 26.8991708234, 2453: 5.17166885522, 2454: -7.21088631983, 2455: -17.8772825403, 2456: 27.0338777059, 2457: 91.48150589, 2458: -19.4970881545, 2459: 10.5534507716, 2460: 100.0, 2461: 27.5062022943, 2462: -34.0748081804, 2463: 36.600593716, 2464: 20.098758421, 2465: -41.662693517, 2466: -62.437101789, 2467: -9.09924357774, 2468: -16.7196367699, 2469: 93.7384834594, 2470: -45.5370137138, 2471: 45.6586761922, 2472: -0.0834563706882, 2473: 23.6801358075, 2474: -60.8848439207, 2475: -20.6793899045, 2476: 9.08122328332, 2477: -6.94763188387, 2478: 53.1139295967, 2479: 0.14720293368, 2480: 25.610724593, 2481: 66.7512522287, 2482: -71.3683802113, 2483: 16.364667421, 2484: -56.5346953204, 2485: -14.8491967339, 2486: -5.15491092888, 2487: -0.211399940349, 2488: -1.81156237113, 2489: 69.0675480748, 2490: 3.59355872576, 2491: -4.12116686523, 2492: 98.9507961763, 2493: 3.12285396359, 2494: 1.26284717051, 2495: -14.8146519983, 2496: -4.05408658022, 2497: 17.1228921628, 2498: -60.0084464221, 2499: -3.75704716654, 2500: 23.0098854624, 2501: -1.62514007923, 2502: -72.3850753923, 2503: 74.8375953856, 2504: -17.0178211909, 2505: -36.4831394473, 2506: 13.1441663992, 2507: -70.8462786166, 2508: 19.8563672911, 2509: -4.0214805826, 2510: -19.2028839074, 2511: -4.21099808904, 2512: 79.8668318318, 2513: 2.140105046, 2514: -4.29646836569, 2515: 20.2700080119, 2516: -3.85948422525, 2517: 37.2495644441, 2518: -6.29325772261, 2519: -3.25956273532, 2520: -6.97752950166, 2521: -6.81080261763, 2522: -1.6429771492, 2523: 4.70885279603, 2524: -3.44177723643, 2525: -4.25152631045, 2526: 2.04897284324, 2527: -4.72388649089, 2528: 12.7928846942, 2529: 33.8379290508, 2530: -2.81399226322, 2531: -45.0673209197, 2532: -1.32565093198, 2533: 28.8577821735, 2534: 21.0494063805, 2535: -3.93656856403, 2536: 19.8759421257, 2537: -4.1019296378, 2538: -4.54816006353, 2539: -4.0422506046, 2540: 64.8488729119, 2541: -4.13239506712, 2542: -3.76377872034, 2543: -4.09306554504, 2544: -33.0043102842, 2545: -3.91955460425, 2546: 40.5092843947, 2547: -3.79518103627, 2548: 5.24592030931, 2549: -17.9480441584, 2550: -4.05415020057, 2551: -3.75994324397, 2552: -1.9890085824, 2553: 0.363186307407, 2554: -4.39789163292, 2555: -73.7986651177, 2556: -83.6901706433, 2557: -49.2341816679, 2558: 22.8615921953, 2559: 11.0221650408, 2560: -62.223706597, 2561: 21.9687886241, 2562: 57.5068764847, 2563: -12.7708680632, 2564: -20.596068829, 2565: 36.0391583971, 2566: -57.4230772261, 2567: 34.0378730385, 2568: 3.25876621243, 2569: -0.98809931321, 2570: -0.779316422987, 2571: -4.24894238756, 2572: -38.1936160041, 2573: -1.09235925501, 2574: -4.38704493294, 2575: 2.00526514775, 2576: -15.7513092074, 2577: 8.38449669745, 2578: 2.01711939169, 2579: 43.1244596289, 2580: -7.33057595205, 2581: 28.0661319234, 2582: 11.0395445286, 2583: 24.4415878717, 2584: -22.4064861832, 2585: 53.0694700744, 2586: -24.4950785232, 2587: -3.24409582675, 2588: 7.86887373724, 2589: -0.405718909535, 2590: -1.46968339543, 2591: -39.4675618184, 2592: -1.08017816633, 2593: -3.93298342962, 2594: 44.5831838277, 2595: 0.194928404646, 2596: -1.07343292212, 2597: -1.41754840521, 2598: -1.01741807631, 2599: -4.36321934215, 2600: -1.45392292266, 2601: -1.32449683385, 2602: -1.3525746973, 2603: -1.93053781511, 2604: -3.8093742164, 2605: -5.32234727186, 2606: -31.7539720582, 2607: -19.5263121194, 2608: -65.4434836643, 2609: 100.0, 2610: -20.2129008014, 2611: -4.73259166629, 2612: -22.5401733702, 2613: -1.17300822312, 2614: 12.74813474, 2615: -1.05751756819, 2616: -95.8227111758, 2617: -1.1207305921, 2618: -56.7131194861, 2619: 15.3413827007, 2620: -21.9429420469, 2621: 1.09618778297, 2622: -0.908806201764, 2623: -0.753885214619, 2624: -0.874075016466, 2625: -1.16724231, 2626: -6.9810426064, 2627: 32.7173824787, 2628: 1.86882782994, 2629: -39.5787201444, 2630: 5.94278665924, 2631: -1.35237123031, 2632: 25.4712160729, 2633: 9.78909937499, 2634: 9.73697695774, 2635: 43.6239477004, 2636: -1.25943198338, 2637: 9.32142587217, 2638: -4.95014429891, 2639: -11.6410072786, 2640: -2.20325034763, 2641: -54.5343548629, 2642: -0.952137889237, 2643: -3.99683974408, 2644: -1.00800472086, 2645: -19.0339729471, 2646: -1.11752762055, 2647: -0.925079172605, 2648: -43.9479320924, 2649: -4.04991654944, 2650: -4.8119036973, 2651: -1.06904908091, 2652: -1.14533958917, 2653: 1.51172707089, 2654: -0.838242433746, 2655: -3.46125782792, 2656: -5.93711036601, 2657: -2.15986349583, 2658: -0.809841844735, 2659: 0.167408018493, 2660: 6.11549360177, 2661: -20.6810178257, 2662: 100.0, 2663: -32.7418523206, 2664: -27.463484495, 2665: -16.9765337079, 2666: 11.507447688, 2667: -3.75702913563, 2668: -2.20837121826, 2669: -0.815570947423, 2670: 52.6793571063, 2671: -3.46023426592, 2672: -7.61802115631, 2673: -12.3847073126, 2674: -15.5055994252, 2675: -9.68879221836, 2676: -2.11281840887, 2677: -32.6798421664, 2678: -3.48204684573, 2679: -6.20868996291, 2680: 11.8908294073, 2681: -64.1770255721, 2682: -3.91024379096, 2683: -4.44858440776, 2684: -9.24421212165, 2685: 34.3605771323, 2686: 8.33263815268, 2687: -67.1577603819, 2688: -10.5512926763, 2689: 85.1842827272, 2690: -16.2915262939, 2691: 94.7606567723, 2692: -100.0, 2693: 9.49596290097, 2694: -6.10492798947, 2695: 30.6866107653, 2696: 7.5194635956, 2697: -16.7629046861, 2698: -11.9068302712, 2699: 8.75657488917, 2700: 7.44601314724, 2701: -4.0176100715, 2702: -3.97346522535, 2703: 8.54951755649, 2704: 7.89347052011, 2705: -3.79741956961, 2706: 20.6239139355, 2707: 1.13639525203, 2708: -4.41702081593, 2709: -60.1693372241, 2710: 20.6392091088, 2711: -4.52144687866, 2712: -27.8823247466, 2713: -2.43465922503, 2714: -5.19652843005, 2715: -1.5217119519, 2716: 62.1530901756, 2717: -28.7462998451, 2718: -88.5241245266, 2719: 15.130371625, 2720: 24.7947277278, 2721: 16.4406646788, 2722: 6.89986156724, 2723: 41.5028314374, 2724: -5.13097860202, 2725: -9.50625472871, 2726: 100.0, 2727: -24.277293351, 2728: 79.6962917894, 2729: -49.12060917, 2730: -4.21970207432, 2731: -12.0578561894, 2732: 2.76412562845, 2733: -2.06961025466, 2734: -7.85312145414, 2735: 56.1990144435, 2736: 96.0939906057, 2737: 51.6530638597, 2738: -9.68208097045, 2739: -19.6051798176, 2740: -50.8402752296, 2741: -8.81563366669, 2742: 12.0776475038, 2743: -34.0719627084, 2744: -2.72539414185, 2745: -3.95640596177, 2746: 3.37911663788, 2747: -93.3028856808, 2748: -9.24328015127, 2749: -4.51055945056, 2750: -1.18434045585, 2751: -3.97119356272, 2752: -4.08506979273, 2753: -4.4458428702, 2754: -24.8052193106, 2755: -3.90035430374, 2756: -3.69840269403, 2757: 0.233588075101, 2758: -3.85180481778, 2759: -18.249388984, 2760: -4.01341421749, 2761: -4.00333069783, 2762: 35.1137661169, 2763: -3.84998388123, 2764: -2.74978727706, 2765: -13.3398937029, 2766: 20.3547207572, 2767: -22.288463057, 2768: -91.0770285217, 2769: -8.16970791303, 2770: 16.3136325097, 2771: 58.7863624248, 2772: 11.6468441614, 2773: 6.61969681592, 2774: 35.4743584975, 2775: 14.7697158224, 2776: -3.62093040617, 2777: 3.59502704343, 2778: 23.5178543781, 2779: -34.3589778461, 2780: -6.43288337924, 2781: 23.0984094937, 2782: -3.92657094653, 2783: 93.7940700025, 2784: 14.5301989063, 2785: -13.5727966401, 2786: 21.0724891292, 2787: 42.1117738308, 2788: 9.07226666932, 2789: 100.0, 2790: 37.4098549972, 2791: 20.3280379977, 2792: 47.3307317272, 2793: 0.050059928404, 2794: 25.980672422, 2795: 14.8617474207, 2796: 35.1131052152, 2797: -0.333326392299, 2798: -19.9923900013, 2799: -29.6647980833, 2800: -43.4530197839, 2801: 12.8322473792, 2802: 4.11214757193, 2803: 18.9056254843, 2804: -5.10986543051, 2805: 15.4603865859, 2806: 19.4953614223, 2807: -23.1880336881, 2808: -7.8517171104, 2809: 5.4568317645, 2810: 6.01308712751, 2811: 12.2623729782, 2812: -4.95649615009, 2813: 24.1284180025, 2814: 21.9251780272, 2815: 37.6833463152, 2816: -23.5170992462, 2817: -12.4058047455, 2818: 5.36474278596, 2819: -7.93810918135, 2820: -6.28504256503, 2821: -21.0812063946, 2822: 21.3382740375, 2823: 27.5300479359, 2824: 33.407189354, 2825: 12.9870634798, 2826: 14.0392423732, 2827: -1.3216680988, 2828: 3.3080578557, 2829: 18.6808443398, 2830: 41.8305767158, 2831: 17.8501539704, 2832: -20.735278948, 2833: 5.06533317264, 2834: 2.23693516932, 2835: 18.4354611829, 2836: -17.0487019736, 2837: 2.03671245421, 2838: 7.84266974156, 2839: -19.6185783581, 2840: -3.36070735696, 2841: 15.2543144725, 2842: -4.92606377691, 2843: 10.2503869453, 2844: 2.59900161454, 2845: 1.8067633164, 2846: -0.737326365058, 2847: 4.26406173478, 2848: 2.18607056296, 2849: 40.1286906548, 2850: 1.77644313439, 2851: 25.3584911715, 2852: -17.5079375308, 2853: 53.6094878713, 2854: -32.0735027601, 2855: 23.3420993434, 2856: 5.91482002901, 2857: -15.1474850165, 2858: 2.47232942716, 2859: -5.69348379764, 2860: 1.72239049896, 2861: -12.3534623518, 2862: 0.791606421363, 2863: -0.791009612966, 2864: -9.04225086724, 2865: 1.89992434283, 2866: -44.1074880167, 2867: -4.58069955504, 2868: 13.8258861446, 2869: 6.57104725586, 2870: 2.98606430706, 2871: 43.0510460511, 2872: 20.9567080168, 2873: -0.00279121028383, 2874: -7.27027829692, 2875: 1.48094363988, 2876: 0.649622057047, 2877: -8.22213227992, 2878: -1.88222764008, 2879: -4.54555799649, 2880: 24.1463678526, 2881: 1.99350744765, 2882: -20.6220504783, 2883: -16.5978133281, 2884: 1.94046520702, 2885: -18.9102419285, 2886: 5.95958277856, 2887: 2.95201447534, 2888: -5.89924906691, 2889: 0.280305051184, 2890: 0.471246531679, 2891: 3.01869616247, 2892: -2.74983078382, 2893: -12.0865475372, 2894: 0.319244325963, 2895: 9.53987349216, 2896: 0.202016817838, 2897: 2.15882936206, 2898: 8.26620135346, 2899: 1.60473823858, 2900: 2.24693312255, 2901: 2.07322544404, 2902: -10.2989717268, 2903: 9.13286836038, 2904: -28.0617146986, 2905: -1.22443533082, 2906: 35.0259921171, 2907: -16.5795682453, 2908: -5.77917360158, 2909: -11.6527598832, 2910: 7.73617546023, 2911: 22.9554146036, 2912: -9.3156910702, 2913: -15.2157567108, 2914: 23.3182547063, 2915: -18.1567046339, 2916: -9.77892308092, 2917: -0.848191012165, 2918: 0.135386929311, 2919: 1.19509239993, 2920: -13.7519619999, 2921: -1.95536642714, 2922: 0.357866485989, 2923: 0.998455530306, 2924: -1.66149838019, 2925: 3.08710245451, 2926: 2.18978940197, 2927: 0.888156599344, 2928: -23.155539972, 2929: 3.30860369548, 2930: 5.49215641112, 2931: 1.14508541582, 2932: 0.59358877516, 2933: 27.8977703695, 2934: -21.9750831059, 2935: -2.55637638049, 2936: -1.44471531154, 2937: -9.43855487336, 2938: 0.717903440578, 2939: 0.195057701612, 2940: 17.6375807875, 2941: 0.214654631765, 2942: 1.90506637989, 2943: 2.25496963539, 2944: 0.144717680851, 2945: 0.171159449218, 2946: 0.145702328251, 2947: 0.200504534394, 2948: 1.51742988631, 2949: 2.65593786584, 2950: -0.365433798875, 2951: 0.490509807258, 2952: 0.819155886425, 2953: 1.91268601339, 2954: 2.43988890782, 2955: 27.2676182551, 2956: 7.26525063416, 2957: 31.6851881652, 2958: 10.5676481456, 2959: -12.8447277189, 2960: 5.6439390347, 2961: -8.24874262046, 2962: 1.85244790105, 2963: 4.77172916856, 2964: 0.216631420088, 2965: 20.5830815366, 2966: 0.182869057363, 2967: 0.316856271663, 2968: 13.1708129504, 2969: 27.4044445227, 2970: -2.64523284889, 2971: 0.301504330733, 2972: 0.394125546761, 2973: 1.77568254932, 2974: 0.108067617137, 2975: 9.43390590987, 2976: 46.1532559085, 2977: 0.752580348013, 2978: -0.204889247782, 2979: 11.8189486452, 2980: -17.6535059289, 2981: -13.6336337089, 2982: 7.12134115573, 2983: 1.6723782268, 2984: -3.68541669, 2985: 0.122757393428, 2986: 35.8540065767, 2987: 2.95844824391, 2988: -5.3646010199, 2989: 1.25618793904, 2990: 18.1287617255, 2991: 0.307925200879, 2992: 1.66384543601, 2993: 0.362443509166, 2994: -3.75461761777, 2995: 0.35881170554, 2996: 0.311290797793, 2997: 16.4631499562, 2998: 1.70298309475, 2999: -12.545946551, 3000: 0.351380792985, 3001: 0.190836971731, 3002: -1.46620037693, 3003: 0.223957619592, 3004: 0.463787772306, 3005: 4.36127529382, 3006: -19.6947344538, 3007: -1.86019201213, 3008: -2.28343610868, 3009: -20.3957117254, 3010: -0.698608769716, 3011: 20.2152499292, 3012: -4.01626277954, 3013: -0.85858209039, 3014: -3.29006121877, 3015: 9.34586982169, 3016: 1.6682822253, 3017: -19.0250516182, 3018: 0.402997920457, 3019: -16.0003613148, 3020: 1.18601598596, 3021: -1.42217860167, 3022: 5.30644313425, 3023: 28.8054770351, 3024: -0.577249095589, 3025: 20.6077070745, 3026: 19.0046551734, 3027: 1.40733748346, 3028: 23.7852283384, 3029: 0.920183252387, 3030: 0.272772155698, 3031: 1.71455663765, 3032: 0.96996349842, 3033: -0.796116221196, 3034: -2.6868573056, 3035: -19.2068685357, 3036: 5.18632614988, 3037: 14.4665006734, 3038: 3.55626882544, 3039: -12.6027380887, 3040: -5.71758157999, 3041: -3.16443372641, 3042: -43.7196223663, 3043: -13.5731943006, 3044: -18.04587167, 3045: 13.1556471515, 3046: 3.08781614947, 3047: -9.14442116167, 3048: -0.672973121255, 3049: -34.1391240763, 3050: 2.02074155243, 3051: 1.87696240289, 3052: 2.31831598157, 3053: 1.52413369034, 3054: 1.69195332037, 3055: 2.08500065134, 3056: 1.41905556065, 3057: 1.46155966254, 3058: 2.68411073385, 3059: 10.3604745735, 3060: -5.80244051954, 3061: -17.373530006, 3062: 2.45132606073, 3063: 1.99897655797, 3064: -11.7944342487, 3065: 28.9727138, 3066: -41.0410130571, 3067: -2.2292422822, 3068: -22.9291918387, 3069: 13.3544900716, 3070: -4.89822343292, 3071: 1.74881337583, 3072: 24.9538248133, 3073: 3.05873873812, 3074: 4.1440327588, 3075: 0.356558123909, 3076: 17.4389904209, 3077: -34.0892302696, 3078: 27.6443386683, 3079: 1.23328410721, 3080: 8.62491645745, 3081: -11.3418916865, 3082: 0.202267492442, 3083: 4.44854308056, 3084: -5.29776354227, 3085: 11.0769102433, 3086: -14.8293345112, 3087: 65.9482694585, 3088: -65.717789171, 3089: 5.48959384845, 3090: 18.9260810783, 3091: 1.20700316646, 3092: -78.0774402637, 3093: 2.77461833217, 3094: -50.8802582121, 3095: 2.10990894946, 3096: 7.35641109951, 3097: 6.90310019837, 3098: 2.90637162945, 3099: -43.3280867162, 3100: 1.69135330902, 3101: 1.61494727699, 3102: -2.90965955999, 3103: -4.33909291264, 3104: 1.55135089576, 3105: 2.1865632688, 3106: -10.8873757126, 3107: 1.91028046037, 3108: -2.23470849015, 3109: 1.92679592107, 3110: 1.92907598759, 3111: 30.0674163775, 3112: 1.55129660419, 3113: 0.218750808898, 3114: 1.0021455062, 3115: 16.1457559425, 3116: 16.31042748, 3117: 8.24195798662, 3118: -13.4727911173, 3119: -5.18441010941, 3120: -22.4308105946, 3121: -0.790455055511, 3122: 3.415694926, 3123: 27.2359199415, 3124: 9.33048461275, 3125: 1.0891266244, 3126: -5.20193536686, 3127: 5.10742758049, 3128: -40.5998718305, 3129: 3.24002882439, 3130: -49.7337645649, 3131: 2.02494983348, 3132: 8.60129352061, 3133: -0.892656895016, 3134: -27.9090866059, 3135: -8.0424689622, 3136: 39.8520245561, 3137: -0.373776443028, 3138: -46.846871057, 3139: 81.1162733721, 3140: 10.6181289779, 3141: -74.2357440906, 3142: 0.499759438084, 3143: 4.27780623825, 3144: -2.71051488607, 3145: -0.315793476994, 3146: 2.49242319287, 3147: -0.20478254019, 3148: -1.77937950653, 3149: -3.8005436586, 3150: -0.797504484073, 3151: 3.71681213051, 3152: 3.90634239632, 3153: -12.9645033653, 3154: 2.06029502935, 3155: -0.558256680134, 3156: -0.0940459053196, 3157: -0.12205320489, 3158: -0.378867519541, 3159: 0.397002955796, 3160: -19.9310103147, 3161: 1.61619756926, 3162: -3.78193848941, 3163: 0.254031666536, 3164: 1.05526812451, 3165: -1.89742420565, 3166: 0.328123340935, 3167: -6.01208891152, 3168: 2.96383044846, 3169: -3.82651902653, 3170: -4.06830680642, 3171: 0.12865905599, 3172: 3.74572852436, 3173: 2.78630623789, 3174: -2.82551773082, 3175: 1.66250955839, 3176: 0.32793660956, 3177: -1.50585559576, 3178: -1.18109541479, 3179: -1.27797977163, 3180: -1.54495661593, 3181: 1.1190228675, 3182: 1.38462198675, 3183: 0.310250728138, 3184: -8.92432659254, 3185: -4.99704456697, 3186: 0.81954526406, 3187: 1.86216217034, 3188: 1.30567731296, 3189: -1.84177291735, 3190: 2.98373307699, 3191: 1.97415566141, 3192: -0.828255591739, 3193: 1.86581671837, 3194: 0.920574773069, 3195: 0.645488021777, 3196: -2.56878242308, 3197: -0.0678224984324, 3198: 3.14745315379, 3199: -0.92438805288, 3200: 4.14453431115, 3201: 4.85140968471, 3202: 2.22390173667, 3203: -5.87068114696, 3204: -10.6679580143, 3205: -11.2704298271, 3206: 1.15301073751, 3207: -0.219221958667, 3208: 1.36959947339, 3209: 0.385903628173, 3210: 0.833848640708, 3211: 0.0427363217469, 3212: 0.377239385509, 3213: 1.59751360827, 3214: 0.3318062256, 3215: -5.03306302332, 3216: -1.55784217989, 3217: -0.146421332732, 3218: 0.170499468126, 3219: 4.98314269998, 3220: 6.0851925484, 3221: 1.24123877636, 3222: -0.0746974530281, 3223: -5.43703936552, 3224: 0.176794512645, 3225: -0.908837565352, 3226: 1.46980429309, 3227: 2.04302062122, 3228: -0.188182318769, 3229: -1.79927010712, 3230: 0.0794754133289, 3231: 0.768281796476, 3232: 0.226653221611, 3233: 0.0961803487084, 3234: -10.5168589272, 3235: 0.348495633598, 3236: -0.0293054966788, 3237: 0.374523093879, 3238: -2.86268002658, 3239: 0.597786164814, 3240: -0.0775021052697, 3241: -0.270410818123, 3242: -6.03988626371, 3243: -4.7638053586, 3244: -11.5041287097, 3245: 0.0994283030099, 3246: 0.240293744199, 3247: 1.89065081875, 3248: -0.049271896459, 3249: -0.0770220982624, 3250: 0.407422946549, 3251: -2.98704606133, 3252: -1.17975461359, 3253: -2.7793657374, 3254: 4.78918049606, 3255: -1.5857634911, 3256: -2.41042848727, 3257: 1.27489018777, 3258: 0.344592821299, 3259: 0.469777741884, 3260: -1.99459490253, 3261: -1.83912843408, 3262: -2.3823799506, 3263: 4.59592231038, 3264: -1.45352263973, 3265: -2.20895795054, 3266: -10.7235657864, 3267: 0.261722380528, 3268: 0.515595531893, 3269: -0.458010769596, 3270: 1.09369802388, 3271: 0.40638444647, 3272: 0.507058772433, 3273: -1.0418979463, 3274: 8.99747912641, 3275: 0.509009049072, 3276: -0.374123594753, 3277: 0.268611831883, 3278: -2.73450506438, 3279: -1.4519722483, 3280: -0.667814709434, 3281: 1.97098436862, 3282: -4.87619691594, 3283: -2.93226921264, 3284: -2.99058682765, 3285: 0.803661797101, 3286: -2.35826435018, 3287: 0.24403265039, 3288: 0.37496266369, 3289: 2.59086542774, 3290: 0.360310296873, 3291: 0.196371361878, 3292: -0.844230840686, 3293: 0.522546779187, 3294: 0.419030935652, 3295: 0.334966570049, 3296: 0.34134265998, 3297: -0.700605002827, 3298: 0.0922873601497, 3299: 0.679873483723, 3300: 0.949515745558, 3301: 0.200203839684, 3302: -0.329423587497, 3303: -0.581959535476, 3304: -8.03217330311, 3305: -3.96904129726, 3306: 2.34659293499, 3307: 4.85563154205, 3308: 3.14092408237, 3309: 2.09834019939, 3310: 2.31237421496, 3311: -0.736764618044, 3312: -3.59143532206, 3313: 0.465264644032, 3314: -0.220008933915, 3315: 0.473831897807, 3316: -6.81085941394, 3317: -12.8747758867, 3318: 6.29982753855, 3319: 0.69967519349, 3320: 0.457239226064, 3321: 0.387540838669, 3322: 0.454733390175, 3323: 0.350307184007, 3324: 0.407421243799, 3325: 2.27664822046, 3326: -0.21654666181, 3327: 0.834106485188, 3328: 6.2430913179, 3329: 2.33476824832, 3330: -0.651862908657, 3331: 1.40988547599, 3332: -1.72641189014, 3333: 4.7364012795, 3334: 0.336761290277, 3335: 2.05287519503, 3336: 0.561678124878, 3337: 1.06904094953, 3338: 1.08627422279, 3339: -2.71307208806, 3340: 0.441908825516, 3341: 0.25333245099, 3342: 0.320674458789, 3343: -0.0727680429886, 3344: 0.286387693689, 3345: 0.36769135155, 3346: -8.51079855566, 3347: 0.230138062124, 3348: 9.30242967299, 3349: 0.47989404455, 3350: 0.507857867541, 3351: 0.843074397519, 3352: 0.307526146211, 3353: 0.697415948678, 3354: -3.40166779245, 3355: -4.55990710135, 3356: 0.682606978231, 3357: 0.809991098336, 3358: 1.31002569521, 3359: 2.25886243456, 3360: 4.35482861809, 3361: 7.64919133118, 3362: -0.684195476963, 3363: -1.96980900213, 3364: 2.38034960515, 3365: 0.15211708831, 3366: -3.09636970735, 3367: 0.26929378587, 3368: -7.91492721856, 3369: 0.470668942621, 3370: -0.70754339537, 3371: 0.405661042759, 3372: 7.52877553359, 3373: 0.0458820898764, 3374: 2.65573158417, 3375: 7.96207936975, 3376: 0.691819009429, 3377: 1.29110953558, 3378: 3.64820108241, 3379: 2.13172269187, 3380: 0.477320563422, 3381: 0.55481747732, 3382: -0.719518410763, 3383: -5.82935500774, 3384: 2.40827066286, 3385: 2.45294500195, 3386: 2.36652344585, 3387: 0.400215954463, 3388: -3.96943938272, 3389: -3.7126667725, 3390: -7.89098053829, 3391: 7.3233832744, 3392: 2.92089901001, 3393: 2.9276827663, 3394: -1.42685057471, 3395: -3.94702557477, 3396: 0.293360033037, 3397: -0.648383883518, 3398: -3.2781624073, 3399: 0.2429226387, 3400: 0.118164685689, 3401: -1.87288740161, 3402: 0.284936321921, 3403: 0.20785337509, 3404: -0.173719401424, 3405: -1.99747494304, 3406: -0.703431752316, 3407: -1.32910823857, 3408: 3.65943278912, 3409: 1.44061269696, 3410: -1.79535974232, 3411: -1.15292734999, 3412: 0.379089327639, 3413: 0.2571814948, 3414: 1.6053424218, 3415: -7.09110834426, 3416: -1.04596190423, 3417: -2.10361951759, 3418: 2.33236964806, 3419: 2.86887801826, 3420: -1.10929852892, 3421: -9.3044814327, 3422: -0.113774499328, 3423: 0.914175362554, 3424: 8.00006117695, 3425: 1.65488006012, 3426: -6.65117156638, 3427: 3.26174991214, 3428: -0.230867199586, 3429: 1.50738306884, 3430: -2.41622763592, 3431: -2.6983798432, 3432: -1.5607973599, 3433: 1.62989679699, 3434: -0.14400336711, 3435: -4.85749770087, 3436: -4.24520536643, 3437: -2.67176468826, 3438: 0.432696781103, 3439: -1.33885018778, 3440: -0.906813065904, 3441: -13.5590251178, 3442: -1.32935345747, 3443: -2.54687431974, 3444: -0.149087772778, 3445: -1.70288987656, 3446: 9.04357169031, 3447: 0.0933443841219, 3448: -3.42963873725, 3449: 0.20186745107, 3450: 0.194721654211, 3451: 0.482395338637, 3452: 0.130623625865, 3453: 0.183789624915, 3454: 0.026372879959, 3455: -4.73471551982, 3456: 0.333881982389, 3457: -5.21204536201, 3458: 0.17790254772, 3459: 0.28845857754, 3460: -1.01968237033, 3461: -0.994910654624, 3462: 1.57695005143, 3463: 3.45517738706, 3464: -0.0297045474129, 3465: 1.64010225267, 3466: 4.47916756436, 3467: -9.93089753771, 3468: 2.38882807127, 3469: -6.1790118906, 3470: -5.15355652727, 3471: -1.21272935993, 3472: 4.9150273873, 3473: -2.9341379041, 3474: 0.0162794607275, 3475: -3.95768856769, 3476: -0.271834398016, 3477: 6.78462697166, 3478: 8.06615951407, 3479: 7.04221453221, 3480: 0.217971697639, 3481: -1.73926662426, 3482: -1.8202664154, 3483: 3.53802076468, 3484: -0.963922332052, 3485: 10.5238957835, 3486: -0.472406881002, 3487: -4.65457543816, 3488: -1.97216054407, 3489: -2.73585897097, 3490: -3.83000169628, 3491: -0.852288429149, 3492: 23.2950037917, 3493: -1.83125896885, 3494: -0.764686881264, 3495: -1.80812661967, 3496: 0.63648197703, 3497: 1.13937996758, 3498: -1.39742520735, 3499: 0.476452694571, 3500: 1.33990322656, 3501: -2.49800463107}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 1.09911463261, 2: 14.8918180534, 3: -36.6537073715, 4: -24.4720717396, 5: -26.4395455895, 6: 42.7020550325, 7: -41.9403870549, 8: -12.6110906136, 9: -38.1270713434, 10: -46.7289728442, 11: -3.81201523011, 12: 6.66063794498, 13: -50.0984033066, 14: -21.9366786601, 15: -17.1888211651, 16: 26.6304400516, 17: 6.16113908032, 18: 36.9586350188, 19: -4.48884915634, 20: 48.2357928766, 21: 33.091822916, 22: 19.6553185098, 23: 5.52969892503, 24: -100.0, 25: 16.6316784309, 26: 34.0147105593, 27: 100.0, 28: 50.099199294, 29: -80.1107643951, 30: 7.8870745739, 31: -3.25278040036, 32: 24.5022871922, 33: 3.52350644029, 34: 3.50849782232, 35: -4.88537188954, 36: -14.4687309064, 37: 14.9320573857, 38: -17.1840255674, 39: 34.288367985, 40: 12.3831962526, 41: 38.1346004314, 42: 7.72310669596, 43: -7.03302499494, 44: -22.3024625097, 45: 5.45746961697, 46: 40.387197541, 47: -14.5521812449, 48: 11.7660450248, 49: 20.0336081082, 50: 21.3789864185, 51: 8.34551285804, 52: 26.3262687768, 53: 7.63802754592, 54: -20.1041327675, 55: 48.5996710546, 56: 6.18851323801, 57: -58.5665478092, 58: -9.52556946238, 59: -20.3634199508, 60: 13.3686838952, 61: 12.8087410462, 62: 39.043504509, 63: 12.8844826123, 64: -11.9150359746, 65: -75.5145062743, 66: 9.00517433799, 67: 13.2824726946, 68: 8.56542652063, 69: 14.1863760966, 70: 8.35044435435, 71: 8.04841923777, 72: 9.37148097364, 73: 7.77598476153, 74: 62.8887654116, 75: -0.560833684248, 76: 5.99601562724, 77: -4.78110967721, 78: -7.7862198232, 79: -19.7982349327, 80: 42.1053275345, 81: 7.18067111619, 82: -24.3409977208, 83: 7.98643250966, 84: 17.1524094003, 85: -23.7656221494, 86: 6.84316167004, 87: -8.5949574417, 88: -31.092850345, 89: 7.29698332906, 90: 1.19869160626, 91: -13.4837512687, 92: 7.6108140788, 93: -41.971722637, 94: 7.97885827977, 95: 8.45447465763, 96: 7.50529669143, 97: 12.6263877564, 98: 7.95448975545, 99: 9.05773891755, 100: 8.05950884472, 101: -69.4412213028, 102: 8.69740073596, 103: -15.2568924444, 104: 6.05138237792, 105: 25.0519896462, 106: 4.64952247088, 107: -3.41663515215, 108: 7.62168252555, 109: 7.74575148522, 110: 12.8098223754, 111: 8.78988006695, 112: 11.5102542669, 113: -17.3914379088, 114: -35.0847689402, 115: 54.0859560547, 116: 20.5821453383, 117: 5.83681239087, 118: -39.2930339622, 119: -47.4110483828, 120: -81.376723482, 121: -37.3029561286, 122: 1.31177391621, 123: -9.31879477843, 124: -11.7302914655, 125: -6.3956844879, 126: 1.21144632499, 127: 1.21072539078, 128: 9.31310483151, 129: 28.1512730578, 130: 1.17766121926, 131: 6.77374337916, 132: -1.42250516561, 133: -62.5783512616, 134: -0.675426660441, 135: -38.287140395, 136: -26.4895072676, 137: 6.8207178439, 138: -40.6444537358, 139: -5.3248378839, 140: -22.4061974659, 141: 23.5436547942, 142: -14.740688993, 143: 31.3185611593, 144: 0.400562783053, 145: -32.179275922, 146: -0.125152490572, 147: 1.23450633753, 148: -48.4835446997, 149: 1.3525613266, 150: 8.212871156, 151: 33.9585514624, 152: -0.55649359523, 153: 1.17639218378, 154: 1.44269227457, 155: 1.24672647966, 156: 7.81398283613, 157: 10.4981548648, 158: -7.90182983677, 159: 2.84817006444, 160: -0.139050654692, 161: 8.51488964053, 162: 11.6516027765, 163: -12.6635685223, 164: 0.935912370635, 165: -4.91496415026, 166: -37.9966872736, 167: 4.58984354163, 168: -28.414852536, 169: 19.6598148847, 170: 8.71582243478, 171: -2.26617137161, 172: 1.20951118187, 173: -12.2565796766, 174: 1.30101603853, 175: 7.60805678976, 176: 42.1930453627, 177: -25.8011709553, 178: -0.225384827358, 179: 1.23734522069, 180: 1.2432998594, 181: 1.29502080283, 182: 2.67079472678, 183: -7.05667219478, 184: 34.5841249182, 185: 10.7504642272, 186: 7.69395287812, 187: 48.1202257492, 188: -0.453097842603, 189: -14.317365796, 190: 33.3215493849, 191: -4.31397887239, 192: 12.4685965287, 193: 1.38174874446, 194: 67.3949226058, 195: 13.0435593366, 196: -3.59574347344, 197: 8.09955441281, 198: 8.70250909299, 199: 1.17880538885, 200: 8.17130226448, 201: 1.25724818073, 202: 14.7780934845, 203: 1.30272636102, 204: 1.37346597668, 205: 46.4748359119, 206: 7.71517962677, 207: 2.72139020218, 208: 1.12551122832, 209: 1.22429991917, 210: 1.75267028359, 211: 1.42788399244, 212: 7.79660779114, 213: 2.60254279487, 214: -35.9204419197, 215: 3.07392461346, 216: 0.998558494523, 217: 10.974916757, 218: 17.0245984987, 219: -33.704028271, 220: 28.7676365727, 221: 17.4370609351, 222: 7.87705882918, 223: 9.48211682078, 224: 6.27541480178, 225: 5.81112009953, 226: 1.19023925542, 227: 9.0940802058, 228: 7.21127151153, 229: 5.66109823162, 230: -1.86591519459, 231: 26.091855396, 232: 1.82152212858, 233: -91.0099132983, 234: -24.9766469808, 235: 6.58365098025, 236: -9.40220451466, 237: 32.7276501786, 238: 1.78772866371, 239: 7.0605859159, 240: 6.66080976057, 241: -39.4446427996, 242: 15.5823794652, 243: -21.9820239715, 244: -69.7388448024, 245: 43.0631970466, 246: -85.4317436998, 247: 12.1326787053, 248: -10.6102317706, 249: -25.8986982598, 250: 29.0821566032, 251: 65.4366213638, 252: 24.5785363617, 253: 4.07495833754, 254: 47.1422166067, 255: 39.9982477935, 256: 6.79253817403, 257: 15.3953460121, 258: 7.11102781264, 259: 8.1991074956, 260: 30.2756388218, 261: 14.0212903308, 262: 6.8934937053, 263: 6.72686066204, 264: 7.1487870922, 265: 7.70887864745, 266: 99.9336562903, 267: -6.82707397516, 268: -29.49797057, 269: 27.8136585197, 270: 4.49308208697, 271: 11.3240272883, 272: 76.8330787739, 273: -44.6959240408, 274: 27.1222380237, 275: 31.6436097469, 276: 15.6030827914, 277: 7.1250125433, 278: -62.4209914512, 279: 8.07333388782, 280: -45.2628726418, 281: 14.045179003, 282: 28.7311481634, 283: 7.64907680429, 284: 24.21412381, 285: -20.8055839309, 286: 21.7052088612, 287: 8.06351141457, 288: 15.6088544664, 289: 8.09510027239, 290: -30.8560739421, 291: -18.8078073731, 292: -0.347935147978, 293: -37.294092687, 294: -43.4006788174, 295: -67.3176163105, 296: -36.7293926211, 297: -19.2832780438, 298: 29.3397446642, 299: 0.871338107737, 300: 9.14097207558, 301: 8.49421671775, 302: -8.40874319855, 303: 10.600156702, 304: 43.7964628932, 305: 2.7028629529, 306: 8.62959318278, 307: -16.9031199526, 308: 7.95184163592, 309: 8.03611036536, 310: 9.31058092251, 311: 15.7814528886, 312: 7.56986049048, 313: 7.30683401241, 314: -65.0556455544, 315: 7.60646977738, 316: 78.624523638, 317: 7.85580149582, 318: 7.618031892, 319: 24.4137701621, 320: 2.73571043334, 321: 7.31830182399, 322: 36.5393326145, 323: -50.7592188962, 324: -36.4328554005, 325: 39.7246835993, 326: -26.6051576575, 327: -35.3094643745, 328: -20.1465375516, 329: 42.5891979661, 330: 22.6363408242, 331: -6.69916978897, 332: -8.73945835358, 333: 4.6464345055, 334: 9.34428240492, 335: -39.6799520315, 336: 7.71874597629, 337: 37.2616813157, 338: -4.7485295996, 339: 7.85592103574, 340: -3.57095313476, 341: -95.1095713355, 342: 0.150550350766, 343: 31.7339083414, 344: -25.3874440636, 345: -64.3944677357, 346: -100.0, 347: -48.6042477652, 348: -40.4432411987, 349: -28.1247124199, 350: -1.47247324542, 351: -3.16705810918, 352: 2.74052047236, 353: -20.1324459352, 354: 6.17192043383, 355: -4.91942768156, 356: 51.6336490214, 357: -0.492397325422, 358: -2.73762821553, 359: 12.2806837205, 360: -26.5966237433, 361: 19.8797638024, 362: 0.411988365694, 363: 25.8668650447, 364: 1.26275435924, 365: -3.342468321, 366: 6.21725001105, 367: 12.9259866838, 368: 55.6519287416, 369: 5.74547833692, 370: 42.5049840712, 371: -11.7389313352, 372: -13.8615852044, 373: -6.72286896854, 374: -14.4609718054, 375: 15.7347395406, 376: 18.0702776144, 377: -10.6983877688, 378: 3.99913184093, 379: 16.0349473342, 380: -20.3003103757, 381: -11.5904455165, 382: -18.2608220932, 383: -6.83217648696, 384: 12.470024915, 385: 4.58775701083, 386: -6.43171696037, 387: 9.16732075155, 388: -5.00383102364, 389: 20.8607700957, 390: 15.5857911541, 391: -1.07407906686, 392: -26.2253605537, 393: 1.0924854844, 394: -3.1072879793, 395: 6.6372767934, 396: -8.56922445612, 397: -2.03895411681, 398: -4.93312303777, 399: -1.78361788708, 400: 5.04949674859, 401: -14.152074662, 402: -2.42356539832, 403: -10.2713705414, 404: -0.802002128755, 405: -2.10336184058, 406: 15.0760524246, 407: 1.07506153666, 408: -22.0008318249, 409: 0.519210547964, 410: 14.9709503899, 411: 10.1951753841, 412: -8.83413058591, 413: -6.69640865449, 414: -5.73201660377, 415: -2.0390994326, 416: -22.6494646455, 417: -3.60965760518, 418: 19.2220391845, 419: -2.17599475505, 420: -1.97141994776, 421: -1.99434146501, 422: -2.1388155728, 423: -25.8106876588, 424: -1.17665268157, 425: -2.12921932165, 426: -0.340592842527, 427: -1.02139506783, 428: 1.45387492996, 429: -15.6802108576, 430: 12.832136784, 431: 82.9779023232, 432: -2.05205901219, 433: -2.32712901109, 434: -26.3492549493, 435: -2.18960216408, 436: 24.0919686342, 437: 7.14546659984, 438: -2.20799581353, 439: 11.5787778092, 440: -2.35552795342, 441: -2.00195297419, 442: 25.8736919063, 443: 9.67668158325, 444: -2.24617988125, 445: -2.19900601269, 446: -6.90005076065, 447: -2.14972584668, 448: 2.76679886436, 449: -2.1573465691, 450: 21.8171298155, 451: -1.88550097109, 452: 13.9385411715, 453: -2.16082105685, 454: -2.51628544874, 455: -23.6480693106, 456: -2.15728733203, 457: -2.24601771462, 458: -2.17550889073, 459: 29.354478645, 460: 15.6411876105, 461: 15.7211916908, 462: -35.315242687, 463: 24.0145552717, 464: -40.613448206, 465: -0.659413860585, 466: -7.88370451748, 467: 24.0833425703, 468: 9.9089552103, 469: 13.6481962767, 470: -5.47564772141, 471: 12.598189516, 472: -12.1778648079, 473: 4.78499306308, 474: 23.5427314481, 475: -2.08614275576, 476: 0.351427961706, 477: -0.759355287379, 478: 3.30892874868, 479: -1.09382739733, 480: -2.06224700169, 481: -1.62682549411, 482: 10.0403796432, 483: -2.04715915036, 484: -4.96505605945, 485: -14.1317761693, 486: -0.05983714617, 487: -29.7945401128, 488: -4.60711817922, 489: -2.27935858547, 490: -2.60900614016, 491: 9.17706431938, 492: 0.51691040099, 493: 0.767929475994, 494: 4.6766339425, 495: -1.14009756015, 496: -1.19614646928, 497: -6.49842025497, 498: -1.17281980711, 499: -2.2046251014, 500: -0.709923253752, 501: -1.40867983229, 502: -1.17301509613, 503: -0.983637123972, 504: -1.24337131861, 505: -1.0823908166, 506: 16.056951348, 507: 1.81649742805, 508: -0.300408873455, 509: -1.14598984525, 510: -16.7434180401, 511: 4.45302597319, 512: 13.3039252409, 513: 6.21829764566, 514: 38.4894404753, 515: -48.9188855648, 516: -4.37390884022, 517: -5.05094227438, 518: -17.2888361974, 519: -1.06468470133, 520: 26.4409498433, 521: -1.40047626123, 522: -40.628367581, 523: -1.18497382859, 524: 2.8476838647, 525: 28.9257889172, 526: 10.9099615633, 527: -1.02334783962, 528: -1.22965887808, 529: -1.21346514406, 530: -1.13809577644, 531: -1.0682295966, 532: -0.591377885385, 533: -0.924526068413, 534: -1.48830530811, 535: -22.1093728282, 536: -62.5018529348, 537: -10.629002104, 538: -3.84987496533, 539: -0.338571977933, 540: -0.083391146844, 541: 4.54025040634, 542: -1.05041281815, 543: 7.59902208021, 544: -1.45455918261, 545: -4.74085012463, 546: -0.89100998203, 547: 5.27408172664, 548: -1.16686117418, 549: -2.23333651058, 550: -1.22124966867, 551: -9.0305607241, 552: -1.05136805286, 553: -1.27577072147, 554: 28.7641867981, 555: -2.1605667403, 556: 6.60735619156, 557: -1.03890094205, 558: -1.07805153896, 559: -2.73616294627, 560: -1.58352200974, 561: 1.33784165279, 562: -7.74578138431, 563: -5.69196226769, 564: -0.714087225725, 565: -1.17905143005, 566: -21.7564126788, 567: -1.39535764274, 568: -3.61786280951, 569: -8.18874198729, 570: 5.05686588157, 571: 8.84475704782, 572: 13.5213873554, 573: -2.19721049587, 574: -2.05948687273, 575: -1.75944857409, 576: 11.8952431292, 577: 1.63725524527, 578: 5.450649648, 579: 4.61802491181, 580: 37.4455726277, 581: -1.20495939952, 582: 3.97897375821, 583: -27.9088921034, 584: 30.8108897762, 585: -2.09503181213, 586: 5.37655509882, 587: 2.03064852791, 588: -2.30154292812, 589: -2.02467542788, 590: 3.65211145215, 591: 7.08009597386, 592: 9.27726824067, 593: 27.4562413337, 594: 3.90515390212, 595: -7.65311794904, 596: 53.1264350798, 597: 18.001546223, 598: 39.1285304716, 599: -10.3216098731, 600: -14.6214114934, 601: -27.8757757639, 602: -26.5439951441, 603: 27.67220756, 604: -2.68045667657, 605: -2.91691826727, 606: 10.7112915302, 607: -2.16464960665, 608: -2.21763602604, 609: 2.86355481526, 610: 3.06384349064, 611: -2.29263281887, 612: -2.34289078357, 613: -1.7150596906, 614: -1.20001938199, 615: -36.6796615902, 616: 0.633888409217, 617: -11.1376787915, 618: -1.1970506414, 619: -1.88435419075, 620: -0.0723316849311, 621: 4.12426835659, 622: -52.8859632409, 623: 7.2583983352, 624: -22.2256362567, 625: 26.6732632487, 626: -6.13948841129, 627: -10.930523279, 628: -0.107762945213, 629: 42.2757327184, 630: 11.9659951728, 631: 22.6743731865, 632: 3.08746226713, 633: 21.1264672526, 634: 39.0803979804, 635: -43.7579369008, 636: -2.18345132099, 637: -2.32976978873, 638: -2.27303774755, 639: -2.68745374375, 640: -1.65702180855, 641: 7.14480999262, 642: 8.50493920318, 643: -5.89234280091, 644: 56.8785837231, 645: -1.21816123483, 646: 19.1068073516, 647: 8.79921307357, 648: 1.50463901648, 649: 3.56325285909, 650: 3.8316557555, 651: 1.02857572946, 652: 13.2056766269, 653: 17.0974139428, 654: 0.0619594562249, 655: -2.40130167046, 656: 5.69980234002, 657: -2.10112630256, 658: -2.20409080484, 659: 0.129832778287, 660: 2.89186722922, 661: -2.15597023186, 662: -2.12177693927, 663: -36.9773757296, 664: -2.10649636858, 665: -4.61597872377, 666: -2.04873038331, 667: -2.20753490668, 668: 16.5823393332, 669: -2.32248627067, 670: 2.34829070177, 671: -4.4936743809, 672: -3.41137064708, 673: -0.459311225827, 674: 19.1256441633, 675: 41.4543807717, 676: 14.8593538589, 677: 2.15808821727, 678: -23.8489380385, 679: -0.381679507471, 680: 9.67408455088, 681: 2.2135950727, 682: -2.27476135481, 683: -1.64662251076, 684: -0.1480212719, 685: 48.765185623, 686: 19.1877620959, 687: -46.6904303987, 688: -2.17690121024, 689: -21.1274676128, 690: -0.70046367616, 691: 2.49680426176, 692: -5.93104072673, 693: -22.1101157116, 694: -15.8820455619, 695: 12.25933867, 696: -4.04968891591, 697: -4.68329994339, 698: 4.70674193508, 699: -0.134938947266, 700: -28.8251482454, 701: -0.360614676465, 702: -1.04500738241, 703: -18.3351574819, 704: -1.19517935693, 705: 2.51981829479, 706: 32.4448027926, 707: 5.79116696579, 708: 3.47991228676, 709: -22.053641907, 710: -14.1864188538, 711: 2.95671245885, 712: -8.78607732761, 713: -10.6157307371, 714: 5.88329107657, 715: -25.9402241194, 716: -9.41363166456, 717: 22.5668480322, 718: 9.5409249466, 719: -25.3696923982, 720: -3.14671458594, 721: -9.8361435926, 722: 31.1025428813, 723: 5.23236365666, 724: -27.9606803583, 725: 3.51582390926, 726: -12.1888779613, 727: 34.0070124179, 728: 23.8247320983, 729: 15.9963013058, 730: -53.6005466582, 731: -9.28427719635, 732: -5.55059535775, 733: -10.3431720946, 734: -6.24920845573, 735: -10.2171424945, 736: -2.96134600131, 737: -3.32348373163, 738: -0.446054754503, 739: 8.58458136953, 740: -0.520590391311, 741: -12.1057166159, 742: -10.5256422859, 743: 20.5218211532, 744: -17.2887419775, 745: 19.1671970444, 746: 0.2353386523, 747: -6.6116267862, 748: -2.92152217758, 749: -5.58731654384, 750: 7.99170216958, 751: -3.38338700658, 752: -8.53774054619, 753: -9.64965642621, 754: -2.88174699379, 755: -7.85267725055, 756: -10.4587238898, 757: 3.32965531667, 758: -9.08884726787, 759: -15.5243060276, 760: 3.08536538816, 761: -31.6517705845, 762: -15.8762818363, 763: -2.29547522517, 764: -3.45278768208, 765: -15.0808102109, 766: -0.279373596214, 767: 1.60406174803, 768: 3.76457409903, 769: 6.32786590034, 770: 38.8182753597, 771: -0.793053624182, 772: -13.4490926134, 773: -5.85126079869, 774: -2.71638543631, 775: -3.10306983181, 776: 5.65387950728, 777: 6.45141517749, 778: 34.9673838158, 779: -1.00719307376, 780: 8.42059724945, 781: -3.91383981441, 782: -4.21489052531, 783: -28.5286732166, 784: -18.5256306794, 785: 0.666815700454, 786: 18.0757481646, 787: -4.54616289011, 788: -3.66689441653, 789: 2.72130873571, 790: -1.7583817347, 791: -2.06270475492, 792: -0.746349529878, 793: -1.18765598035, 794: -1.40336085248, 795: -28.1419868515, 796: 0.260513342279, 797: 1.28745482949, 798: 1.03951212512, 799: 8.79742669333, 800: 0.0487996216911, 801: 10.2117188991, 802: -3.47384805596, 803: -11.624522693, 804: -5.54441242021, 805: -0.488750536403, 806: -2.13466476391, 807: -1.28801629531, 808: -2.42506257148, 809: 0.654926419144, 810: 12.3172193838, 811: 9.91488901112, 812: 7.86023745512, 813: 14.2500229053, 814: 17.5073158759, 815: 29.582963912, 816: -5.95833839458, 817: -24.7378690011, 818: -27.0162534629, 819: 5.33444500542, 820: 19.1087226892, 821: 16.6156977424, 822: 14.6628578456, 823: 8.04557390972, 824: -0.187220094891, 825: 0.336832138572, 826: -1.2725150658, 827: 34.0711631647, 828: -0.106951719946, 829: -2.37783787058, 830: -0.452761171755, 831: 11.0477915834, 832: -0.337129906423, 833: 0.73245053698, 834: 4.1964600657, 835: -2.9713492493, 836: -7.24571204845, 837: 3.04186473971, 838: -0.118507562605, 839: -46.1129450811, 840: 14.9658501032, 841: 0.455486368138, 842: -0.171998714806, 843: -8.36491210504, 844: -1.22795038366, 845: -0.136192252549, 846: -16.3298061266, 847: -0.251988970002, 848: 0.339239387284, 849: -3.81692551203, 850: -0.26062084956, 851: -0.16081295028, 852: -0.0520620750269, 853: -0.304793330077, 854: -0.47255584271, 855: -7.08276136692, 856: 1.10884271403, 857: 0.178418993931, 858: -1.41855013163, 859: 0.183846207199, 860: -0.91587220177, 861: -29.6790218436, 862: 5.07157360644, 863: 32.1294307908, 864: 6.19978933456, 865: 11.0822458292, 866: -4.15933321161, 867: 7.80108438424, 868: -3.20173694813, 869: 0.899401969954, 870: -0.172817416371, 871: 2.08156189892, 872: -0.267250696195, 873: -17.8527980623, 874: -33.6028284087, 875: 25.2123711687, 876: 0.37606385949, 877: -0.0962941478308, 878: -0.0266582715416, 879: -0.309687793019, 880: 0.176333789732, 881: -5.05300777435, 882: 19.5048993484, 883: 0.126053542008, 884: 31.6967983105, 885: -3.22790082122, 886: 2.10381295183, 887: 1.53920930719, 888: -21.2073866605, 889: -4.09359357503, 890: 25.4812873288, 891: -0.202334342043, 892: -10.5643036453, 893: 16.0923719334, 894: 0.884415060906, 895: -2.02777064842, 896: -7.85030809743, 897: -0.206429165772, 898: 0.0922315275729, 899: -0.240007959413, 900: -48.1378044095, 901: -0.210923004005, 902: -0.247800823825, 903: -17.3525367053, 904: -1.34574917218, 905: -26.8066684341, 906: -0.204770530727, 907: -0.230612356729, 908: 0.84794053087, 909: -0.258720658367, 910: 0.123932361826, 911: 6.42693268002, 912: 7.96056366276, 913: 1.38986662258, 914: -0.770629920843, 915: 4.02354656496, 916: 10.1488399541, 917: 21.68700981, 918: -4.99460660824, 919: 14.980077944, 920: 2.17891532022, 921: 2.73481840004, 922: -2.05027507007, 923: 10.0918777946, 924: -0.164755540271, 925: 7.46247307751, 926: -7.40637606471, 927: 2.60266189976, 928: 2.66091419687, 929: -3.25988611526, 930: -0.302326157952, 931: 8.2076835275, 932: -26.2314563432, 933: -5.8991565209, 934: -18.314269354, 935: -7.85607823323, 936: 8.48355003663, 937: -2.22687229514, 938: -2.46267749891, 939: 4.36778127976, 940: -2.32488957071, 941: -1.90833025262, 942: 10.0300724313, 943: 7.10949293173, 944: 11.9360258223, 945: -0.913952975646, 946: -11.1299054608, 947: -26.0705329414, 948: 28.1999249822, 949: 19.3022817489, 950: -2.40823415642, 951: -6.29542295017, 952: 5.6066984718, 953: 10.4248103137, 954: -1.6756400033, 955: 21.2234582885, 956: -1.6659329138, 957: 0.247884221547, 958: -4.28942428224, 959: 5.72482969588, 960: -0.104444707318, 961: -1.77458381467, 962: -4.51729825414, 963: -0.544669555085, 964: 14.9330155564, 965: -5.88470492685, 966: -0.942562520438, 967: 7.91462372564, 968: -0.631272930463, 969: -0.778067259867, 970: -33.2486825127, 971: -2.07155417049, 972: 16.2716952245, 973: 18.7210809603, 974: -23.8013478686, 975: -38.0248855624, 976: -8.21314647166, 977: -3.67057082739, 978: 17.7274062717, 979: -1.21182369811, 980: 11.3887374383, 981: 45.061967276, 982: 90.855878877, 983: -17.6022450482, 984: -9.50000686603, 985: -0.982591466234, 986: -1.54515028971, 987: -5.70083867912, 988: 9.84225358005, 989: -1.85610853628, 990: -10.2856073406, 991: -13.1601620202, 992: -5.10502071958, 993: -3.42623718301, 994: 16.0370346257, 995: 1.48817475917, 996: 1.53402554629, 997: -2.3082037593, 998: 2.9027348362, 999: -0.587004299984, 1000: 16.7973344834, 1001: -4.77933740183, 1002: -24.0739344296, 1003: 2.11167368324, 1004: -1.17880605536, 1005: 18.386904927, 1006: 0.125712632643, 1007: 0.0592575498616, 1008: 0.351571747909, 1009: 23.1796416833, 1010: -3.13684556562, 1011: -2.00646611481, 1012: 21.5063938082, 1013: -1.36007723554, 1014: -13.7087447028, 1015: -2.45810878847, 1016: -0.88001529184, 1017: -5.26122127024, 1018: -1.33851500525, 1019: -0.162983003256, 1020: 28.581099204, 1021: -10.0341935723, 1022: -32.9961410574, 1023: 10.6279975738, 1024: -34.6935371353, 1025: -1.6892086426, 1026: 3.21095853058, 1027: -26.598262095, 1028: -9.75104149083, 1029: -42.7240732456, 1030: -13.344457069, 1031: -4.10472045395, 1032: -7.54829986608, 1033: -6.29257065559, 1034: 38.6426941504, 1035: -8.40609568733, 1036: 36.1106588104, 1037: -0.650746529365, 1038: 14.5082730283, 1039: 7.32527650824, 1040: 16.427063177, 1041: -1.7785683208, 1042: 53.625961205, 1043: 0.265028065239, 1044: -89.286593397, 1045: -33.6933767412, 1046: -16.4331822404, 1047: 16.1591189538, 1048: -0.669364616377, 1049: 1.48858531503, 1050: 0.543075721345, 1051: 3.34649927668, 1052: 21.5973692123, 1053: 0.187846748132, 1054: 0.476957341396, 1055: 5.21099003693, 1056: 5.46729201554, 1057: 13.0871481286, 1058: 6.85238054502, 1059: -2.10624714915, 1060: -2.09665669943, 1061: -4.35229862521, 1062: -0.99804207722, 1063: -1.20013073322, 1064: -1.28403142527, 1065: -6.29191816866, 1066: -16.3057062722, 1067: -1.71997917574, 1068: 5.02851321837, 1069: -5.99541548587, 1070: 1.74550246894, 1071: 6.7403174813, 1072: 6.42536544746, 1073: 4.27734551566, 1074: 7.01460863081, 1075: -0.333887019917, 1076: 0.494396545117, 1077: 8.78719864954, 1078: 3.16692312044, 1079: -3.56043626317, 1080: 2.76318578907, 1081: -0.722180822397, 1082: 4.0854490074, 1083: -0.18771244384, 1084: -3.04224526235, 1085: -2.04794331489, 1086: -5.8827340504, 1087: 3.3518424294, 1088: -3.17109465202, 1089: 2.04793117674, 1090: -1.43734764071, 1091: 5.6153674381, 1092: -11.0030608394, 1093: -1.82786678411, 1094: 4.76152069436, 1095: -0.874625593016, 1096: 1.94549404055, 1097: 1.10327661427, 1098: 1.33149625996, 1099: 1.83320077736, 1100: -0.648857051305, 1101: -1.84009080428, 1102: 34.8256936238, 1103: 0.0475021405792, 1104: -8.03862762246, 1105: -1.61981096439, 1106: 9.41005526393, 1107: 19.9391410223, 1108: 4.41137796632, 1109: 10.0425821194, 1110: -1.84074157849, 1111: -4.49602541566, 1112: 2.69171299904, 1113: 0.748026368004, 1114: 1.7614353729, 1115: -0.796150512788, 1116: -1.41594306252, 1117: -0.135994109515, 1118: -0.704065025097, 1119: 8.28016103789, 1120: -0.517922115995, 1121: -0.797210256721, 1122: 0.104049157638, 1123: -1.21035177466, 1124: -1.34026998219, 1125: -0.167648752482, 1126: 4.96311541513, 1127: -2.91054528373, 1128: 2.4880261985, 1129: 13.8859903069, 1130: 2.69577136531, 1131: -1.23143235217, 1132: 4.13676105199, 1133: 0.304650520211, 1134: -0.651257897728, 1135: 1.52371831465, 1136: -1.4177936064, 1137: 1.84244729948, 1138: 1.90349854262, 1139: -0.539362964135, 1140: 7.79236608608, 1141: 7.4340556837, 1142: -0.393000073828, 1143: -0.615855156151, 1144: -2.20206786018, 1145: 1.69965084873, 1146: 0.922921370105, 1147: 1.95831902713, 1148: 4.15970111654, 1149: -0.231869171305, 1150: 7.5180892532, 1151: 0.0924529192732, 1152: -0.956443235086, 1153: 10.2337541557, 1154: -0.541490897777, 1155: -0.597128008869, 1156: -1.65681979139, 1157: 6.11522344267, 1158: 6.26457592773, 1159: -4.56324436114, 1160: -10.292535002, 1161: -6.01642392645, 1162: 1.36557963435, 1163: 0.292940826358, 1164: -6.07484395167, 1165: -1.69792479846, 1166: -1.49125005688, 1167: -5.88971460607, 1168: 2.15898575288, 1169: -0.542585234308, 1170: -1.53476826114, 1171: -0.962253195208, 1172: 20.3852593902, 1173: -0.546888769347, 1174: 0.601117318174, 1175: -2.56324405571, 1176: -5.38490415569, 1177: -0.453181311934, 1178: -0.348115825535, 1179: -0.687781203579, 1180: 10.2889819066, 1181: 1.51327868997, 1182: -0.446690990641, 1183: -6.17724202904, 1184: -0.147663620709, 1185: -0.706461561704, 1186: -1.24199714968, 1187: 4.69275548309, 1188: -14.1633839624, 1189: -3.57131495995, 1190: -9.19509836779, 1191: 0.231687649852, 1192: 1.5690182869, 1193: -0.530532358094, 1194: -0.4114996602, 1195: -9.32695657816, 1196: -0.560168384813, 1197: -0.527850127009, 1198: -18.2939706789, 1199: -0.630638004147, 1200: -0.443431582069, 1201: -0.442281336235, 1202: -0.343644549714, 1203: 0.51833614737, 1204: 39.9611674945, 1205: -1.66903569754, 1206: -0.17185016361, 1207: -0.396127784347, 1208: 0.403124451266, 1209: -4.16005833673, 1210: -9.27823264663, 1211: -1.10904412159, 1212: 10.847075344, 1213: -15.2585392859, 1214: -0.549236204928, 1215: -1.90347765811, 1216: -1.74764382548, 1217: 3.06668784313, 1218: 7.38019051604, 1219: -0.445215583051, 1220: -3.62732734985, 1221: -0.41271340807, 1222: -7.73673228546, 1223: 7.12335710268, 1224: 5.39030080292, 1225: -0.578833576148, 1226: -0.376058265152, 1227: -0.325035880636, 1228: -0.702905553552, 1229: -0.542226440043, 1230: -1.18848494043, 1231: -10.1406703978, 1232: 0.957531457914, 1233: -14.2422486808, 1234: 6.49853275048, 1235: 3.44400248466, 1236: 0.523438340662, 1237: -6.2726383888, 1238: -6.49039861615, 1239: 3.08227801007, 1240: -0.459704345598, 1241: -10.503136008, 1242: -0.675651363431, 1243: 0.26934805439, 1244: -1.39268993385, 1245: -4.68911869832, 1246: -0.386854852895, 1247: -0.427282795097, 1248: -0.324399565882, 1249: 4.87508519789, 1250: -0.527286566711, 1251: -0.555695037257, 1252: 19.197717152, 1253: -0.757024328734, 1254: -7.705889387, 1255: -0.358956581457, 1256: -0.44434638463, 1257: -0.374888699757, 1258: -0.698328641146, 1259: 3.41070349269, 1260: 3.63927588498, 1261: 12.0617270562, 1262: -0.263599312266, 1263: -0.569967690844, 1264: -14.4714074112, 1265: -0.805129703857, 1266: -13.1489124379, 1267: -4.04646090089, 1268: -8.70448500797, 1269: -0.26282343355, 1270: 1.81424560179, 1271: -0.380295307845, 1272: 1.51006379123, 1273: -0.632222132954, 1274: 0.496634564384, 1275: 1.54014690507, 1276: 3.58841897029, 1277: -0.41969199008, 1278: -2.75362223293, 1279: 0.127805722319, 1280: 3.63426301256, 1281: -10.9137275738, 1282: -1.2220871905, 1283: -9.85561244839, 1284: -1.43577369281, 1285: -16.1076118916, 1286: -1.03492485022, 1287: -0.286094232301, 1288: 2.94156142309, 1289: -1.09152730327, 1290: -3.03024612171, 1291: 3.1711634044, 1292: 7.03361743638, 1293: -1.75592793038, 1294: 0.182910325557, 1295: -4.92132395078, 1296: -0.371871226914, 1297: 10.7291695367, 1298: -8.15426422952, 1299: 8.04379898834, 1300: -0.363763494166, 1301: 23.5077430526, 1302: 4.40215017597, 1303: -0.88174063015, 1304: 2.19549841635, 1305: -0.401549711042, 1306: -0.438174726269, 1307: -14.343463653, 1308: -8.06147934705, 1309: -1.0961148484, 1310: -0.174646147057, 1311: -0.0145804929021, 1312: 0.600645893055, 1313: -37.2524494871, 1314: -12.9051993691, 1315: 2.58256381223, 1316: -0.357356869553, 1317: -0.89172427587, 1318: -2.48023067466, 1319: 3.24992924689, 1320: 4.25341914208, 1321: 5.38095324823, 1322: 7.31897909028, 1323: 2.7526111762, 1324: 8.10248984534, 1325: -5.39042387123, 1326: 2.26586778, 1327: 10.8891387616, 1328: 1.14936204569, 1329: 11.2456634806, 1330: -0.826278399985, 1331: -11.0451802799, 1332: -21.9629450094, 1333: 11.0068157935, 1334: -0.550936878575, 1335: -2.41915099428, 1336: 2.5859276417, 1337: -0.801957958519, 1338: 1.39182325875, 1339: -1.29625327084, 1340: 2.43053883865, 1341: 13.54390043, 1342: -1.50043212168, 1343: 4.91800884123, 1344: 9.87625880395, 1345: -3.94641208582, 1346: 2.94726187544, 1347: -7.97904464125, 1348: 3.43005587566, 1349: -0.119282830391, 1350: -1.06690823694, 1351: 14.045687227, 1352: -6.83116325236, 1353: -0.520394635353, 1354: 13.9168243271, 1355: -0.632613418712, 1356: -0.652451978229, 1357: -0.988763219643, 1358: 1.42081779671, 1359: -0.721416506441, 1360: -0.575602541188, 1361: -8.44795373551, 1362: -0.783517066512, 1363: 0.436278881474, 1364: -0.650698046806, 1365: -0.897516754405, 1366: -3.04879393842, 1367: -0.789718250238, 1368: 6.44623207451, 1369: 8.05846663161, 1370: 17.1716937723, 1371: 4.57382222534, 1372: -14.1939930257, 1373: 0.638548554746, 1374: 12.2556639894, 1375: 23.6278437748, 1376: 44.2352419806, 1377: 3.06116415676, 1378: -2.63447613749, 1379: 1.68206885715, 1380: -0.240700538505, 1381: 12.0808031892, 1382: -3.41091987, 1383: -8.58776731882, 1384: -2.90245033689, 1385: 2.47415993279, 1386: -0.402582104216, 1387: -2.5341092614, 1388: 0.249120209635, 1389: -14.2120099809, 1390: 10.7735300043, 1391: 0.850745603762, 1392: -4.42417388814, 1393: 4.9254056685, 1394: -1.98828669955, 1395: -10.7809779629, 1396: -5.03148793836, 1397: 0.207992955988, 1398: 19.4667543122, 1399: -1.26540902683, 1400: 4.23213075235, 1401: 8.46205349499, 1402: -6.99386307528, 1403: -13.8512465889, 1404: 4.99870165171, 1405: -7.85473549523, 1406: 4.12322508767, 1407: -6.6806878147, 1408: 0.00481537788098, 1409: 4.0706915166, 1410: 2.06246478795, 1411: 2.63188965999, 1412: -1.54620135304, 1413: -5.58827918199, 1414: 2.61120338619, 1415: 12.2235980367, 1416: 3.75139298015, 1417: 6.56736891848, 1418: 5.9954109984, 1419: 3.94553793648, 1420: -8.23228845885, 1421: -6.97574495694, 1422: -9.51870397807, 1423: 4.40732970868, 1424: -1.11027762901, 1425: -2.86352668269, 1426: 3.55137295952, 1427: 13.7543891306, 1428: -0.23388500753, 1429: 2.04263995827, 1430: -0.818367809221, 1431: 1.93804387765, 1432: 5.80744977779, 1433: 6.68752905928, 1434: 5.93384862209, 1435: 22.0324268012, 1436: -3.07886702077, 1437: 2.39352808638, 1438: -0.259165414717, 1439: 13.185522634, 1440: 0.557397457459, 1441: -5.68752872611, 1442: -2.21918211585, 1443: 9.04695968815, 1444: -0.332494823568, 1445: 1.84297027564, 1446: 1.8018725509, 1447: -5.25449128567, 1448: 23.8211284031, 1449: 1.36115541361, 1450: -1.28574047764, 1451: 2.83745254749, 1452: 0.398838009692, 1453: 9.00474829038, 1454: 1.34168107128, 1455: 2.79562738909, 1456: -9.82336525923, 1457: 21.0239504895, 1458: -6.19137786184, 1459: 5.47925553705, 1460: 2.62902071284, 1461: -4.67847022477, 1462: -2.39680870815, 1463: 0.866677466895, 1464: 0.787262410539, 1465: -0.489284684811, 1466: 0.305526916891, 1467: -0.994796715243, 1468: -9.58161259099, 1469: 1.28327887213, 1470: -13.8795580981, 1471: 1.41545168024, 1472: 5.57051638589, 1473: 3.85040349788, 1474: 3.14961925948, 1475: 13.8638797905, 1476: 10.3298484935, 1477: 0.01648016958, 1478: -4.84154363803, 1479: -1.15775791114, 1480: -9.59369177178, 1481: -2.20811014921, 1482: 3.97015440825, 1483: -6.03651695219, 1484: 5.08661727579, 1485: 1.28541966831, 1486: -4.22979721794, 1487: -2.82221032605, 1488: 1.19400778709, 1489: -2.29082741602, 1490: 0.97857182136, 1491: 0.902369442865, 1492: 0.88155455095, 1493: -2.70673461756, 1494: 0.541127784467, 1495: 3.21874311517, 1496: -4.73156963031, 1497: -6.02701813068, 1498: -1.80105377014, 1499: -5.46603217695, 1500: 0.304138072001, 1501: -1.75775377453, 1502: -1.83952824432, 1503: 0.955276739611, 1504: 1.61628025487, 1505: 0.537977350144, 1506: -7.29784837224, 1507: -7.39579191801, 1508: -5.20964690741, 1509: -1.4555197207, 1510: 7.32813415079, 1511: -10.6836425403, 1512: -0.155783271767, 1513: 0.4871400347, 1514: -3.27227772642, 1515: 3.74200491514, 1516: 7.08708885178, 1517: -1.68313336503, 1518: 8.28930214868, 1519: 8.36533321607, 1520: 0.237066807013, 1521: -1.89598655436, 1522: 0.165080562953, 1523: 0.0403950811981, 1524: -0.335425361512, 1525: 2.58545072991, 1526: 0.089203104295, 1527: 0.476130359921, 1528: 2.08041375531, 1529: -0.41329639095, 1530: -0.342586710183, 1531: -0.139533564548, 1532: 0.376216818054, 1533: 1.5323034958, 1534: 5.65862910712, 1535: 0.0591120368048, 1536: -0.94322343566, 1537: -5.84426953615, 1538: 4.89022661995, 1539: 4.95475019474, 1540: -0.928887993326, 1541: -7.02417312235, 1542: 0.733122688858, 1543: 0.138226832026, 1544: 2.66479767884, 1545: 0.266743475238, 1546: 0.792915923572, 1547: 4.29488290138, 1548: 0.168186581682, 1549: 0.269111666078, 1550: 0.254758020422, 1551: 0.124846165224, 1552: 0.389558451577, 1553: 3.06439638665, 1554: 0.118435824638, 1555: -0.483823842569, 1556: 0.0835040494269, 1557: 0.230386774198, 1558: 0.257489694491, 1559: 11.7346996252, 1560: 4.84757753194, 1561: -7.71169564974, 1562: 15.9287802297, 1563: -4.93289235876, 1564: 1.98446032121, 1565: -3.83825494623, 1566: 0.186846818379, 1567: -2.2287849199, 1568: 0.0992851685408, 1569: 8.85764586635, 1570: 0.243867480692, 1571: 10.8267304942, 1572: 15.3934872899, 1573: 0.314071796943, 1574: 0.191053006732, 1575: 0.207638271666, 1576: 0.298407619926, 1577: 0.70984022431, 1578: 0.0518536220024, 1579: -1.79477035596, 1580: 10.4629685122, 1581: -0.55552130024, 1582: -4.56234352437, 1583: -10.0457924249, 1584: -1.81630101923, 1585: -4.727585201, 1586: 3.86127627822, 1587: 0.131115879116, 1588: -4.06855698136, 1589: 0.0985368063086, 1590: 8.66614471742, 1591: -3.40382787251, 1592: 0.929117553287, 1593: 0.783616689645, 1594: 3.39465497094, 1595: 0.0857699628599, 1596: 0.670711975566, 1597: 0.0951365265029, 1598: 15.1240459285, 1599: 0.0832781292857, 1600: 0.178822389675, 1601: -3.15620822456, 1602: 1.15107178228, 1603: -5.76109549325, 1604: 0.16535736289, 1605: 0.213440274378, 1606: -0.402817097331, 1607: 0.100741635373, 1608: -3.84024868587, 1609: -13.0481843723, 1610: -2.82076337165, 1611: 0.263788711492, 1612: -1.59374120043, 1613: -2.37473980636, 1614: 0.262229201293, 1615: 6.49940654282, 1616: -1.90384533705, 1617: 3.77891541246, 1618: 0.319042608379, 1619: -0.733181013988, 1620: 0.917390884396, 1621: -6.1686428194, 1622: 0.379472257845, 1623: -0.0314445231408, 1624: 1.17407012082, 1625: 0.101171608186, 1626: -2.18548002998, 1627: 6.83821701229, 1628: 0.0932040404755, 1629: 3.02207055454, 1630: -3.1178357873, 1631: -1.36608428054, 1632: 6.46504463701, 1633: -8.84403066578, 1634: 3.64321134684, 1635: 1.6493142595, 1636: 0.628672780972, 1637: -4.61942612608, 1638: 2.72625008621, 1639: -3.58466321382, 1640: -3.53927175685, 1641: -3.34979930808, 1642: 1.76499249783, 1643: 2.35976668643, 1644: -11.8282367219, 1645: 2.77999179817, 1646: 6.32291091381, 1647: 0.293403738169, 1648: -8.90236657117, 1649: 9.61760220957, 1650: 0.099752554209, 1651: -8.92068535709, 1652: -0.714426481275, 1653: -5.13048700942, 1654: 0.98240507116, 1655: 0.789142028554, 1656: 4.04224465117, 1657: 0.092471143108, 1658: 0.314496786664, 1659: 0.206515811529, 1660: 1.5192605869, 1661: 0.323912100737, 1662: 3.06852411667, 1663: -4.68497664076, 1664: -2.41304390777, 1665: 5.43483738869, 1666: 0.87857337567, 1667: 0.333297148189, 1668: 2.7294936533, 1669: -3.04649444606, 1670: 23.6881016281, 1671: -12.774621282, 1672: 2.43284109164, 1673: 0.325064661809, 1674: 2.28007356448, 1675: 2.30370246859, 1676: 0.420690769839, 1677: 1.80500361282, 1678: -5.11115142902, 1679: 3.94552132755, 1680: 2.5098819934, 1681: 14.4583798682, 1682: -11.52799206, 1683: 0.488743050301, 1684: 7.10654723339, 1685: -6.15587819014, 1686: 2.05366652428, 1687: 4.87448384359, 1688: -0.723231144772, 1689: -3.87138347667, 1690: 5.96536547877, 1691: 16.6728664385, 1692: -10.7363350079, 1693: -0.308674639933, 1694: -6.19624902428, 1695: -2.18761240097, 1696: 1.38021827129, 1697: 2.37187979064, 1698: -11.7885789731, 1699: 0.461315291351, 1700: -11.7930382385, 1701: 3.75665724131, 1702: 0.900240663575, 1703: -8.18772278143, 1704: 0.493175738779, 1705: 0.573386273015, 1706: -0.402642182256, 1707: -10.4695682277, 1708: 0.757835914016, 1709: 1.46079201343, 1710: -2.10839296982, 1711: 1.00839289863, 1712: 9.67475286401, 1713: 1.26141659323, 1714: 1.37005154872, 1715: -9.04750343751, 1716: 1.25699657885, 1717: -5.95303027291, 1718: 9.92379486034, 1719: 2.61587819513, 1720: -16.3403176521, 1721: 5.99898220541, 1722: -37.1809845821, 1723: 0.317439049705, 1724: -8.75167698649, 1725: 0.475658852551, 1726: 0.399967198496, 1727: -3.10762005419, 1728: 0.04838258921, 1729: 1.12490384711, 1730: -6.61209015755, 1731: -3.45414273593, 1732: -15.8903999097, 1733: -8.5136843979, 1734: -2.87416115376, 1735: 0.940068849316, 1736: 2.01288496248, 1737: -8.00826791064, 1738: -6.65944676582, 1739: -4.88968431617, 1740: 2.46443048319, 1741: -0.476260575389, 1742: -5.88995489424, 1743: 2.82759134107, 1744: 5.31486961593, 1745: -17.9791765312, 1746: 0.442054342279, 1747: -13.2178899391, 1748: -9.91635315894, 1749: -13.1469187741, 1750: 3.5499828077, 1751: -2.17962022306, 1752: 9.3779674398, 1753: -15.9565688491, 1754: 17.24102558, 1755: 8.1997561866, 1756: -13.8301128578, 1757: -3.45321676353, 1758: -4.18583571419, 1759: -0.336923042901, 1760: -1.66554271643, 1761: -1.2310139464, 1762: -17.6191055166, 1763: -8.11206915538, 1764: -14.2972320009, 1765: -14.4801183182, 1766: -6.49117400215, 1767: -4.29218889093, 1768: -0.0715697637217, 1769: 9.91583515195, 1770: 2.0425631276, 1771: 9.87258030916, 1772: -6.92520947699, 1773: -2.7175058111, 1774: -5.37292437765, 1775: 9.67232391092, 1776: 7.18839446641, 1777: -2.18328097447, 1778: 3.70681467152, 1779: -1.11855985529, 1780: -27.672413251, 1781: -5.3666513934, 1782: -2.7964660047, 1783: 2.5854934728, 1784: -11.9158635047, 1785: 0.0176325223867, 1786: -4.93857944281, 1787: 2.71828190027, 1788: 3.13655940385, 1789: 1.34722563035, 1790: -7.46428097661, 1791: 4.33407798493, 1792: 1.9565183185, 1793: -0.721089947186, 1794: 6.43267290814, 1795: 8.62278139344, 1796: 1.20051318177, 1797: 9.50776494258, 1798: 0.402213720297, 1799: -9.66721967624, 1800: 1.23489297103, 1801: 1.35481098258, 1802: -3.62730209719, 1803: 3.22910572602, 1804: -10.4559544303, 1805: 3.45293976639, 1806: -14.828784286, 1807: 21.1767808047, 1808: -7.03235341885, 1809: -12.6947544005, 1810: -6.294771219, 1811: 0.948209827096, 1812: 2.90599282878, 1813: 2.10508385273, 1814: 2.50888227302, 1815: 0.154867607083, 1816: 2.44140801924, 1817: -15.7807355008, 1818: 3.01518989647, 1819: -29.5502974179, 1820: -0.472184038954, 1821: 7.38382934532, 1822: 1.41310588848, 1823: 0.412791550107, 1824: -7.30737938913, 1825: -13.5627609694, 1826: 0.0375920569654, 1827: -2.21492786574, 1828: 2.38814521854, 1829: 3.94484573494, 1830: -9.11840409232, 1831: 4.24497179944, 1832: 1.53727893697, 1833: -0.709064134836, 1834: 3.13781777055, 1835: -7.47178944104, 1836: 1.42201789128, 1837: 0.630159583651, 1838: -9.79428157614, 1839: 2.62776419851, 1840: 2.70613239411, 1841: 2.22892808607, 1842: -2.57864857623, 1843: 2.27097353225, 1844: 1.00267285292, 1845: 2.51194662107, 1846: -17.7384363868, 1847: 0.316907374506, 1848: 5.92155142552, 1849: 1.48618313001, 1850: 3.27285421522, 1851: -7.00572520574, 1852: 3.21596369982, 1853: -0.341088802384, 1854: 0.730245280634, 1855: -7.46226746884, 1856: -1.14375097633, 1857: 5.43061710017, 1858: -20.9975726575, 1859: -22.6499019263, 1860: 16.0444011167, 1861: 7.20168796264, 1862: 4.99260049782, 1863: -2.49429234765, 1864: -4.80795718388, 1865: -2.99117725502, 1866: -10.1493713255, 1867: -4.5752623199, 1868: -6.20313623254, 1869: 3.11419432541, 1870: -2.53800755225, 1871: 0.613775196588, 1872: 0.808407779945, 1873: 1.81092072842, 1874: 2.02245222251, 1875: 0.647000433893, 1876: 2.96286018594, 1877: 0.733632770235, 1878: 13.2988927813, 1879: 1.07351556404, 1880: 0.864598249541, 1881: -6.01099961848, 1882: -0.822827417259, 1883: -9.91534370257, 1884: 0.691174462289, 1885: 1.73165983271, 1886: 13.1699708164, 1887: 2.98483937165, 1888: -5.62484338678, 1889: 0.170329272073, 1890: 5.72510445908, 1891: 1.1252479019, 1892: 0.538966191608, 1893: -4.63368553981, 1894: 0.47534077663, 1895: 1.16020983197, 1896: 2.76546117098, 1897: 0.697463111047, 1898: 0.542760447347, 1899: 0.672007453124, 1900: 0.491255527674, 1901: 0.194138677739, 1902: 2.08052982696, 1903: 0.762076415671, 1904: 0.778476406727, 1905: 0.954529330851, 1906: 5.58941951588, 1907: 0.51013534862, 1908: 1.80343422377, 1909: 2.13462012751, 1910: -13.0495445819, 1911: 6.61387267322, 1912: 0.731859612227, 1913: 0.099129833888, 1914: 2.5101649818, 1915: 1.71662123482, 1916: 1.10445921093, 1917: 0.673991630159, 1918: -8.09583553785, 1919: 0.673845912325, 1920: 5.33775040526, 1921: 4.08740565858, 1922: 6.15254796511, 1923: 0.85181960379, 1924: 0.46435641737, 1925: 0.673792073775, 1926: 0.74987281124, 1927: -0.0749940718287, 1928: 4.10360222684, 1929: -3.19490308328, 1930: 3.51758462171, 1931: 16.9631242284, 1932: 0.315983874393, 1933: -0.0778550382852, 1934: 1.78768111242, 1935: -1.22236727806, 1936: 0.155060132934, 1937: -6.73689058163, 1938: 0.405522646127, 1939: -13.5045842294, 1940: 2.399322471, 1941: -0.834373155723, 1942: -1.04069889678, 1943: 6.41581117723, 1944: 0.576312718152, 1945: 2.50831775942, 1946: 0.545038268939, 1947: 13.4941300805, 1948: 0.508830796724, 1949: 0.62619316066, 1950: -0.157596811967, 1951: 2.52934242229, 1952: -1.26913176778, 1953: 0.531926234211, 1954: 0.639179631666, 1955: -0.387653541629, 1956: 0.393984630915, 1957: 2.7017619204, 1958: -3.96273230998, 1959: 7.5167430156, 1960: 0.58014245354, 1961: 0.228609064974, 1962: 10.2705815223, 1963: 3.04739152689, 1964: -2.54078545463, 1965: -0.454161775168, 1966: -3.42877356979, 1967: -7.35817728577, 1968: 0.710487741969, 1969: 1.69257193105, 1970: -2.38115713766, 1971: 1.16250671651, 1972: 5.70234423248, 1973: 5.1919892604, 1974: -2.42053299095, 1975: 0.136576460365, 1976: -3.52754997656, 1977: 0.622899076069, 1978: 2.68995405224, 1979: -3.61123386872, 1980: 2.19253991753, 1981: 3.88852234742, 1982: -2.16181198698, 1983: 2.12536256424, 1984: 3.42786424523, 1985: 3.03196162105, 1986: -23.2454350716, 1987: -7.38771361313, 1988: 10.5170246812, 1989: -9.48442095267, 1990: 2.13623132091, 1991: 9.98312300262, 1992: -7.50951397804, 1993: 10.9994443172, 1994: -11.0421425839, 1995: -15.6018447407, 1996: 2.44906861881, 1997: 9.63737776917, 1998: -11.0003250329, 1999: 0.716851922542, 2000: 19.7909750443, 2001: 2.73575505447, 2002: -1.25357810823, 2003: 3.07517653214, 2004: 0.993929028256, 2005: 2.56206669467, 2006: 2.56465833239, 2007: 2.5349313275, 2008: -0.319673658254, 2009: 0.105459181292, 2010: 0.0148463240357, 2011: 2.40484441175, 2012: -3.40305709881, 2013: -0.344691106484, 2014: -3.97980258209, 2015: 7.3957095838, 2016: 0.15646179155, 2017: 6.0887288568, 2018: 4.3078329739, 2019: -8.3513321404, 2020: -5.46348418792, 2021: -2.81642183707, 2022: -6.18237908839, 2023: -18.1987064895, 2024: 2.29362486777, 2025: 2.00463253642, 2026: -0.967624193204, 2027: 12.4093668978, 2028: 2.43280216979, 2029: 8.64025515154, 2030: -3.33013038658, 2031: 0.661236133489, 2032: 1.42162388854, 2033: 2.93703240388, 2034: -0.609691930822, 2035: 2.19526945549, 2036: 2.33218855406, 2037: -18.6819245491, 2038: 3.88064213153, 2039: -14.4753069847, 2040: -12.3410918115, 2041: 5.10925305687, 2042: -2.09849285731, 2043: 4.15656960896, 2044: -0.0547713794959, 2045: -5.45028356418, 2046: 3.91751849849, 2047: 16.8178602354, 2048: 0.533223467791, 2049: -5.04363269205, 2050: 9.2263756727, 2051: 2.62119066352, 2052: 0.334575693345, 2053: 2.46996502301, 2054: 2.34712163738, 2055: 1.26213930102, 2056: -5.7867256116, 2057: 1.2965390198, 2058: 0.934505366538, 2059: 5.30599380311, 2060: 2.63579969182, 2061: -3.48192747648, 2062: 2.86613519616, 2063: 2.96760200503, 2064: -4.94062896336, 2065: 6.20188472739, 2066: 2.79062788202, 2067: 9.72356802744, 2068: 4.76591359423, 2069: -11.2211267692, 2070: -2.01294145511, 2071: -11.183907944, 2072: -1.53826162979, 2073: -16.2278396556, 2074: -0.555676442799, 2075: -8.57578736703, 2076: 1.38425433378, 2077: -2.97078224063, 2078: -0.358949671648, 2079: -0.23462362964, 2080: 4.56108375395, 2081: 12.9629432793, 2082: 3.8148627834, 2083: -1.79018386969, 2084: 0.953362375982, 2085: 2.83749397766, 2086: 5.3262061722, 2087: -3.89191050498, 2088: 8.31901922297, 2089: 4.53871731122, 2090: -0.695245577611, 2091: -17.2071839567, 2092: -6.13643961606, 2093: 2.2681697744, 2094: -4.63041232703, 2095: -0.819966426285, 2096: -13.6975785759, 2097: -4.28407760036, 2098: 4.49726602198, 2099: -4.43559873827, 2100: 32.121369956, 2101: -9.30256606263, 2102: -17.7391366812, 2103: -12.4961206966, 2104: 6.23426372508, 2105: 19.3653656413, 2106: 16.3126411241, 2107: 18.6349744286, 2108: 50.5723799914, 2109: 15.0042215824, 2110: 7.80796412899, 2111: 9.61507093978, 2112: 39.7112729818, 2113: 15.9817419908, 2114: -7.75210245371, 2115: 8.51345177473, 2116: 4.72409233609, 2117: 14.1766298968, 2118: -13.8688516021, 2119: 18.4410323079, 2120: 3.65732412463, 2121: -7.18134170989, 2122: -10.8425742479, 2123: 34.9796269525, 2124: -15.2011297177, 2125: 4.20742044086, 2126: 34.6895573927, 2127: 17.792859781, 2128: -0.050494370917, 2129: 1.38263194944, 2130: 5.24978321626, 2131: 7.93276416506, 2132: -6.15533759082, 2133: 10.0699676602, 2134: 0.320370518615, 2135: 6.8854581326, 2136: -6.31112328606, 2137: 0.436399790198, 2138: -0.424295963052, 2139: 2.5036729821, 2140: -5.86703275963, 2141: 2.77052948785, 2142: -8.60442735068, 2143: -7.62074066706, 2144: -0.451088681189, 2145: -4.21690778983, 2146: -17.0229953438, 2147: -7.96800787697, 2148: -6.14289900647, 2149: 14.28024572, 2150: -5.50204065345, 2151: 20.3903147865, 2152: -6.3547891145, 2153: 26.497015413, 2154: 4.71406715177, 2155: 7.57930360015, 2156: 21.8203474831, 2157: 2.71724111919, 2158: -8.92072823766, 2159: 3.58921471921, 2160: -3.82568643806, 2161: -14.3728679698, 2162: -16.2520738567, 2163: 5.95307339711, 2164: -6.16129429979, 2165: -15.1062048292, 2166: -7.35699801565, 2167: -4.79617844738, 2168: 49.4281461947, 2169: 18.7620382784, 2170: -4.69556263318, 2171: 12.8311987236, 2172: 0.525957673232, 2173: 13.2459606014, 2174: -19.4324116393, 2175: 6.65823625121, 2176: 7.99234338463, 2177: -6.58849795954, 2178: -6.4673894467, 2179: 6.31412863109, 2180: 12.3084065744, 2181: 0.484975533267, 2182: 0.906471951496, 2183: -6.05935772558, 2184: 17.7939867236, 2185: 15.7849228089, 2186: -5.91628340632, 2187: 13.469702884, 2188: 3.96267206865, 2189: -6.13913252707, 2190: -7.15533857563, 2191: 15.3079128184, 2192: -7.49130565564, 2193: -8.3254811153, 2194: 4.11166199943, 2195: 30.6557586683, 2196: -1.48161666141, 2197: 4.13318323093, 2198: -4.9062421033, 2199: -4.63644081757, 2200: 26.1687605367, 2201: 6.94710225449, 2202: -2.59823437575, 2203: -8.13062602387, 2204: 18.8932651224, 2205: 4.55409167476, 2206: 1.27508782909, 2207: 40.7828145673, 2208: -6.71178149064, 2209: 11.8025001739, 2210: -17.0666304428, 2211: 15.7711981719, 2212: -0.259720529024, 2213: -30.6418601826, 2214: -24.8225470019, 2215: -10.0354236913, 2216: -10.4108423644, 2217: 14.1349441296, 2218: 5.22214138111, 2219: -5.79926876622, 2220: -0.908083536417, 2221: 1.21127144005, 2222: -11.1405256168, 2223: 16.1522730342, 2224: -1.07079337851, 2225: -6.84193409514, 2226: -5.47713875666, 2227: -16.2205841631, 2228: -1.23263671106, 2229: -11.7431282371, 2230: -17.7665024651, 2231: -4.39215534286, 2232: 1.89246225408, 2233: -5.32958750267, 2234: 28.9553646854, 2235: 3.79675570488, 2236: -6.78861984116, 2237: 2.97812443292, 2238: -1.52760232769, 2239: -16.2775319769, 2240: -1.04221510906, 2241: -1.13157828623, 2242: -5.24650668229, 2243: -1.18555382196, 2244: -6.49309051676, 2245: 1.32469881537, 2246: -1.48722194433, 2247: -1.14663992843, 2248: -1.24064740335, 2249: -1.08189053067, 2250: -6.06741250628, 2251: 7.04138972333, 2252: -2.27256024672, 2253: -1.50555743502, 2254: -1.14519726647, 2255: -8.77586099757, 2256: 5.52729876345, 2257: -8.12111792427, 2258: 8.27292083558, 2259: 26.1830276128, 2260: 35.2645584683, 2261: -4.80559683022, 2262: -7.21352112415, 2263: 5.68411408412, 2264: -4.49949790198, 2265: -16.5651371762, 2266: -1.36745783062, 2267: -26.4320126715, 2268: -0.9525941651, 2269: 1.71776721936, 2270: 19.7438038952, 2271: 31.9302622292, 2272: -5.31412615411, 2273: -0.971931595117, 2274: -1.71046335659, 2275: -0.912367799597, 2276: -1.37672676225, 2277: 8.15693853875, 2278: -18.7752617325, 2279: 19.1903892478, 2280: 46.2681348098, 2281: -24.9446065941, 2282: 7.05980764504, 2283: -16.3086727644, 2284: -1.07339077619, 2285: -0.945136890148, 2286: 14.4221548009, 2287: -1.12504186679, 2288: 2.63087090747, 2289: -7.44655267879, 2290: -4.69574765253, 2291: -8.12481849383, 2292: 21.622001531, 2293: -0.944465599488, 2294: -7.14376231082, 2295: -1.25554843474, 2296: -18.0025843211, 2297: -1.03765885758, 2298: -1.22668500673, 2299: 55.6964748232, 2300: -6.15428932889, 2301: 9.79564476148, 2302: -1.09398082096, 2303: -1.03324808294, 2304: -4.29718534214, 2305: -1.4894259018, 2306: -6.17963906409, 2307: 2.60826068833, 2308: 3.82153374945, 2309: -1.44758911583, 2310: -1.04439788892, 2311: 16.4837522857, 2312: 2.89457641006, 2313: 0.00492130511253, 2314: -40.2868902003, 2315: 0.594720087572, 2316: -0.698216662962, 2317: -1.20129824608, 2318: -5.16409891389, 2319: -0.430248201532, 2320: -0.690979805979, 2321: -4.24535837696, 2322: -6.69335971643, 2323: 10.2535879987, 2324: -1.73209987356, 2325: -4.32185684751, 2326: -1.83327192748, 2327: -22.6658208905, 2328: -3.77940202007, 2329: -9.79580128999, 2330: -26.507617152, 2331: -1.29874989641, 2332: -15.2305963331, 2333: -3.70377814777, 2334: -6.97696283686, 2335: 7.75737687758, 2336: 16.3698217315, 2337: 1.81644204801, 2338: -15.2977833661, 2339: 3.95817623086, 2340: -3.25822322576, 2341: -7.22120302135, 2342: -29.4611991189, 2343: 43.5156351294, 2344: 11.1690878084, 2345: 36.2097271198, 2346: 4.07167946453, 2347: -15.3479720043, 2348: 13.2553492388, 2349: -7.18482051102, 2350: -3.04781999753, 2351: 29.7832648698, 2352: -5.05495319628, 2353: -6.64502323401, 2354: -0.174614601131, 2355: 10.3999850944, 2356: -5.41550578877, 2357: -4.76918114547, 2358: -5.16019460915, 2359: -6.26751345468, 2360: 6.82189469834, 2361: 27.9439378738, 2362: 4.90424276161, 2363: -1.02913739255, 2364: 1.72948006423, 2365: 13.0723644235, 2366: 3.29911254956, 2367: -16.7676522685, 2368: 6.57478033797, 2369: -19.3004772925, 2370: 6.56273672216, 2371: -9.02290364969, 2372: -2.07223655779, 2373: -4.46074575553, 2374: 1.6648592716, 2375: 9.66644860859, 2376: 15.2611366289, 2377: -5.36604165176, 2378: -47.3311325151, 2379: -22.7793431821, 2380: 3.41632050679, 2381: -6.42194738029, 2382: -7.13321728888, 2383: -8.59955054409, 2384: 5.02262488161, 2385: 3.67717528528, 2386: -8.58017138836, 2387: 22.8682852968, 2388: -14.5267935184, 2389: 10.1676160612, 2390: -9.13039575761, 2391: -15.4906593499, 2392: 9.1168350906, 2393: 10.1725521159, 2394: -0.89040417578, 2395: -7.88675307659, 2396: 16.0293368357, 2397: -6.94810287967, 2398: 16.56584897, 2399: -6.80453994671, 2400: -6.2382286316, 2401: 2.58649572773, 2402: -7.09923280552, 2403: -7.03380809563, 2404: -0.0673301024754, 2405: 15.9366950575, 2406: -3.59913643801, 2407: -5.32723273278, 2408: -52.9913232471, 2409: -6.26751693812, 2410: -19.9172737477, 2411: -6.45463704989, 2412: -5.93422646656, 2413: -0.514622766575, 2414: 4.63337608648, 2415: -6.45466400449, 2416: 3.70572095195, 2417: 2.7885091895, 2418: 11.6838262305, 2419: 10.4104529163, 2420: 29.5424490349, 2421: -4.85421979391, 2422: 9.18316433452, 2423: 67.8435721834, 2424: -8.38568825074, 2425: -7.67757606536, 2426: -6.90570666884, 2427: -3.6766811114, 2428: 2.93051161749, 2429: -24.1776669699, 2430: 25.6663974748, 2431: -37.5889184865, 2432: -3.79484739955, 2433: -5.52370014469, 2434: 6.71499807272, 2435: 0.0390743671776, 2436: -22.8520962142, 2437: 5.04682411986, 2438: -31.3282054107, 2439: -1.46791497032, 2440: -0.275432362539, 2441: 18.5804518709, 2442: -10.7470651564, 2443: 11.5303306094, 2444: -1.10894810234, 2445: -19.4476476083, 2446: 60.7879510262, 2447: -19.3829185272, 2448: -83.6452299088, 2449: -55.0436028441, 2450: 55.570071603, 2451: 62.0220759104, 2452: 62.4281460168, 2453: 15.2088933112, 2454: -62.9858985308, 2455: -59.1052491115, 2456: 47.7730518154, 2457: 52.5911240225, 2458: -60.1788722914, 2459: -5.22466779669, 2460: 100.0, 2461: 2.16141687503, 2462: -44.9545408875, 2463: -1.43955797382, 2464: -39.2310689948, 2465: -82.374764388, 2466: -76.8522169564, 2467: -11.5705149429, 2468: 20.7243668247, 2469: 9.18972894199, 2470: 52.3944487507, 2471: -64.0003147909, 2472: 57.288117556, 2473: 34.4189545844, 2474: -69.9670783146, 2475: 7.39175118843, 2476: -68.3301563733, 2477: 1.34351810569, 2478: 20.1836959361, 2479: -34.6115721325, 2480: -9.53791817678, 2481: 3.07995719659, 2482: 18.832429872, 2483: 85.4735962264, 2484: 50.9914951833, 2485: 17.4892635756, 2486: -54.4110234026, 2487: -10.9332375725, 2488: -30.4954356768, 2489: 16.5498090638, 2490: 25.0435391029, 2491: -6.80663463915, 2492: 7.75472504327, 2493: -5.95305752552, 2494: -4.13794367074, 2495: 26.7771652198, 2496: -7.34277871713, 2497: 22.5720910956, 2498: -30.4828069131, 2499: -2.30636713255, 2500: -12.8931895991, 2501: -3.23580047576, 2502: 0.935900068941, 2503: -37.5262448772, 2504: 22.4836375024, 2505: 79.2042148149, 2506: -80.4208898108, 2507: -8.62652699993, 2508: -21.7910158983, 2509: -2.00892240648, 2510: -3.86713204795, 2511: -23.127228164, 2512: 48.6509581807, 2513: -8.65853995333, 2514: -10.8346422613, 2515: -12.6375923713, 2516: -7.84136378316, 2517: -14.4967188466, 2518: -0.773854609383, 2519: -4.18096871934, 2520: 26.7222207764, 2521: -10.7433262905, 2522: -8.70656196835, 2523: -4.76426396852, 2524: 8.29593347561, 2525: 99.986182123, 2526: -7.19742250687, 2527: 9.59456857475, 2528: 93.8831773182, 2529: -6.59501807179, 2530: -7.84422448383, 2531: -27.7823500439, 2532: -8.0034255513, 2533: 5.20257409159, 2534: -8.29616828923, 2535: -7.92880436207, 2536: 19.3069952822, 2537: -7.61885519136, 2538: -7.47293683637, 2539: -8.20426464023, 2540: -17.0453052052, 2541: -7.68805024009, 2542: 18.6525635061, 2543: -2.61771564291, 2544: -16.7240463312, 2545: -8.26738792167, 2546: -20.5733530265, 2547: -2.02257884421, 2548: 8.81096389597, 2549: 48.5676964482, 2550: -7.87087788516, 2551: 12.0871033576, 2552: -7.97813799251, 2553: 37.7880039225, 2554: 1.21490972097, 2555: 36.9250887982, 2556: 40.4800726848, 2557: 17.3510129832, 2558: 26.795939865, 2559: -1.32291026284, 2560: -59.8189271902, 2561: -2.7338146643, 2562: 15.7454910549, 2563: -2.50938977607, 2564: -7.25390706364, 2565: 100.0, 2566: -49.6432433664, 2567: 33.5948886697, 2568: 27.8733989083, 2569: -2.10214550179, 2570: -0.0134000017755, 2571: -9.1074582874, 2572: -1.16539205903, 2573: -2.72532521374, 2574: -4.88097604793, 2575: -0.0728001060851, 2576: -18.9706957457, 2577: 0.584853902719, 2578: 9.47946856854, 2579: -28.5084019042, 2580: -6.19920472648, 2581: -21.7099075798, 2582: 20.7926450964, 2583: 11.0831459083, 2584: 1.58486131146, 2585: -5.35518968611, 2586: 46.9325878864, 2587: -3.2764298393, 2588: -2.2060597178, 2589: -1.4815810362, 2590: -2.19169811188, 2591: 8.80444620609, 2592: -1.78294247184, 2593: -7.84561375312, 2594: -0.576830795628, 2595: -0.870174074094, 2596: -1.83033921473, 2597: -2.0400345757, 2598: 0.835131145146, 2599: -3.53057310079, 2600: 9.71020984339, 2601: -8.32651712555, 2602: -3.89580890425, 2603: 4.65675801221, 2604: 6.54982511583, 2605: 2.93515856352, 2606: 21.4438288232, 2607: -6.10264437202, 2608: -67.2153631967, 2609: 8.64343587809, 2610: -0.859451983889, 2611: -16.3190229005, 2612: -33.5749093558, 2613: 9.06151458782, 2614: 34.7204937044, 2615: -0.661137837374, 2616: -84.9862746197, 2617: -1.61693638524, 2618: -2.34574027726, 2619: 43.9474109892, 2620: 74.100411054, 2621: -7.7849115111, 2622: -1.94029747078, 2623: 0.0634232623246, 2624: -0.550183977725, 2625: -1.70840190411, 2626: -2.00683982509, 2627: -26.6504380702, 2628: 15.082298384, 2629: 56.6208537789, 2630: 75.1344609929, 2631: 24.7012950106, 2632: -17.041009714, 2633: 0.751060892096, 2634: 8.24533114749, 2635: -12.9059938348, 2636: -0.467205628723, 2637: -11.3325851043, 2638: -2.66008562801, 2639: -16.5183464903, 2640: -4.3114503366, 2641: -16.5075826145, 2642: -1.87873703969, 2643: -7.79118109847, 2644: -1.77972321724, 2645: -5.49921483383, 2646: -1.83758244156, 2647: -1.64274928141, 2648: 91.3207802011, 2649: -8.02511186331, 2650: -59.924561569, 2651: -1.79572679196, 2652: -1.87554209117, 2653: -9.9928750476, 2654: -2.15986915495, 2655: -8.03258706441, 2656: 30.7343408629, 2657: 19.55774019, 2658: -5.5451805904, 2659: -1.95273761116, 2660: -73.8701355176, 2661: 14.1346512383, 2662: 22.3298631728, 2663: 10.5209477074, 2664: 8.3638170323, 2665: 26.0900849661, 2666: -23.2265705771, 2667: -5.33850903632, 2668: 9.99557339765, 2669: -3.28138543602, 2670: 2.68512336583, 2671: 4.96938614694, 2672: -11.3783947861, 2673: -16.0937672447, 2674: 12.1043110588, 2675: -5.07162899977, 2676: -31.2568335688, 2677: 13.9573456798, 2678: 3.90411456571, 2679: -84.2509303117, 2680: -10.1720246108, 2681: -1.99829436781, 2682: -0.734477661351, 2683: -5.04692727092, 2684: -0.729503274754, 2685: 98.4122040671, 2686: -78.6131087684, 2687: 15.1248164803, 2688: 53.5493438321, 2689: 12.9844312619, 2690: -31.999888818, 2691: -22.1858846253, 2692: -22.4723278806, 2693: 100.0, 2694: -7.77153216343, 2695: -40.8809150937, 2696: -17.1774806311, 2697: 18.5843222487, 2698: 8.78500407577, 2699: -6.65780283589, 2700: -67.6972873581, 2701: -8.04778589257, 2702: -7.87219687607, 2703: -35.025312127, 2704: 1.16542137268, 2705: -8.11208031944, 2706: 2.48219484368, 2707: -57.8821876528, 2708: -3.6929359609, 2709: 6.68988655511, 2710: -28.383310362, 2711: 13.9619484765, 2712: -42.997912427, 2713: 7.51203825498, 2714: 2.22704755856, 2715: -22.1958243051, 2716: 73.6308217214, 2717: 100.0, 2718: 62.7974415361, 2719: 26.8778231215, 2720: 68.5083222452, 2721: -43.4407059639, 2722: 13.3026591438, 2723: 57.3477458596, 2724: 21.2524007827, 2725: 50.4444897321, 2726: -79.6841419217, 2727: -86.0614450509, 2728: 99.2973215746, 2729: 100.0, 2730: -7.83103355499, 2731: -56.3389047918, 2732: -10.1727992503, 2733: 10.0827230372, 2734: 0.514130532343, 2735: 1.49613103284, 2736: -41.6710901742, 2737: -42.4702781203, 2738: -99.3847883174, 2739: -63.2280928617, 2740: 97.7885235115, 2741: -6.86321026903, 2742: 0.455037039166, 2743: -20.8292250099, 2744: -1.08147113025, 2745: 16.5752357032, 2746: 19.5569001039, 2747: -43.6269065334, 2748: 1.0697983816, 2749: -7.61484525056, 2750: 54.7661301757, 2751: -7.99313943772, 2752: -7.92259959819, 2753: -8.10113823431, 2754: -7.86128496876, 2755: -7.75829072559, 2756: -7.96065394814, 2757: -50.2211798328, 2758: -8.2321130874, 2759: -14.2172871234, 2760: -7.96962993535, 2761: -7.97446488023, 2762: 5.62014888265, 2763: -7.38597886485, 2764: -8.43368056875, 2765: 35.2949557154, 2766: -25.3512054666, 2767: 77.6760027038, 2768: -16.6128314395, 2769: 35.9360863718, 2770: 36.4096102961, 2771: -44.4798182233, 2772: 13.9228482335, 2773: 34.099700506, 2774: 13.0685862624, 2775: 83.1911230143, 2776: -1.62234282607, 2777: 13.1193376572, 2778: 40.2954840689, 2779: -98.9534602397, 2780: 39.8170037338, 2781: 29.3084611125, 2782: -9.00948299105, 2783: 100.0, 2784: -34.3338380866, 2785: 38.4502647945, 2786: 41.0579913885, 2787: -15.0398259785, 2788: -78.7610113367, 2789: 74.2719564395, 2790: 29.673380042, 2791: -1.32048359481, 2792: 9.57132650772, 2793: 0.41009955562, 2794: 19.4797876474, 2795: 1.64631493548, 2796: -2.74972144148, 2797: -29.2981051773, 2798: -15.729425037, 2799: -9.76936427714, 2800: 17.0025163241, 2801: -9.27083587736, 2802: 29.3148281773, 2803: -13.3116292006, 2804: -10.657682924, 2805: 27.7923097387, 2806: 7.15164453555, 2807: 4.75482974234, 2808: -17.1152636951, 2809: -44.9589877022, 2810: 15.7232200207, 2811: -19.9002654582, 2812: -31.0389583064, 2813: 17.0603163114, 2814: -14.8092225932, 2815: 30.768493613, 2816: 5.40876432155, 2817: -13.7024280796, 2818: -28.4178889852, 2819: -11.0615516556, 2820: -4.92959932067, 2821: -8.82041165261, 2822: -5.55220211832, 2823: 34.2149383659, 2824: 13.0759018336, 2825: 17.9773428, 2826: -5.37254856806, 2827: 9.28010101907, 2828: 17.7364258436, 2829: 13.5839534058, 2830: -1.46777776608, 2831: 33.2037236147, 2832: -15.7780443802, 2833: -21.6081680091, 2834: 1.36439406001, 2835: -9.42405215275, 2836: -2.00741113556, 2837: 2.69025896078, 2838: -8.99430862969, 2839: 36.8598903199, 2840: -2.10046616364, 2841: -7.0489926117, 2842: -0.485575395474, 2843: -11.8807720823, 2844: 48.5211266867, 2845: 1.21953999167, 2846: 1.42531046645, 2847: 8.27304480654, 2848: 3.06680336607, 2849: -24.1474641778, 2850: -15.1174640267, 2851: 19.8360474162, 2852: 13.1225987234, 2853: 40.4217660111, 2854: 1.90455117781, 2855: 34.1096249702, 2856: 23.3182887397, 2857: -26.7827733727, 2858: -0.290641557444, 2859: -10.2563838456, 2860: 3.93437335821, 2861: 25.2030391776, 2862: 3.49450597385, 2863: 5.01904383832, 2864: -2.41832850747, 2865: 1.97455721955, 2866: 31.9675501692, 2867: 10.609615215, 2868: -1.40267231712, 2869: 7.06367469938, 2870: -3.02402815691, 2871: 11.8431613912, 2872: 33.8864861828, 2873: -5.09469364995, 2874: -56.0388336187, 2875: 0.24938398499, 2876: 35.0748205499, 2877: 15.6398597759, 2878: 2.5688551131, 2879: -3.67703507441, 2880: -29.7882192887, 2881: 1.99516134586, 2882: 3.39454034029, 2883: 3.40279633398, 2884: 1.81768621957, 2885: -38.5991069537, 2886: 2.11456051733, 2887: 3.05979475418, 2888: 2.58009457415, 2889: 30.8372607305, 2890: 3.36311083006, 2891: 6.03721993441, 2892: 10.3680417357, 2893: 4.27208810174, 2894: 5.07282678597, 2895: -8.5772939925, 2896: 0.717680861863, 2897: -5.06250689652, 2898: 7.49606695549, 2899: 3.05972549279, 2900: 0.474531643085, 2901: 3.01945142914, 2902: -53.2153620684, 2903: 1.78996844586, 2904: -4.49262141147, 2905: -52.2652888192, 2906: -6.30761013395, 2907: 37.3033892333, 2908: -13.9427433609, 2909: -14.662925408, 2910: 7.74517963862, 2911: 52.2479043643, 2912: -11.1041959031, 2913: 2.82088140547, 2914: -1.99871487173, 2915: 10.7285387453, 2916: 4.63112608804, 2917: -30.1127846067, 2918: 0.841182376681, 2919: 0.935600551565, 2920: -19.3051473218, 2921: 37.6392777925, 2922: 0.448850290855, 2923: 3.42504546494, 2924: 3.66950318052, 2925: -1.64677562603, 2926: 4.61617572321, 2927: -2.40496900542, 2928: -23.8211192729, 2929: 5.40582735876, 2930: 48.1851064635, 2931: 3.40948763089, 2932: -9.00984228923, 2933: 24.5958819057, 2934: 22.4236094421, 2935: 3.72883434385, 2936: 1.68468522822, 2937: -19.1008336289, 2938: 0.489968118334, 2939: 0.438047339071, 2940: -4.94917239432, 2941: 0.500621210663, 2942: 3.21908543477, 2943: -10.1469686465, 2944: 0.892333255978, 2945: 0.403895877746, 2946: 0.612456500497, 2947: 0.602511067118, 2948: 5.22496373923, 2949: 79.6017538167, 2950: 2.95113845447, 2951: 1.25885855686, 2952: -0.826539457465, 2953: 11.4156476765, 2954: 8.45690461659, 2955: 13.7548233345, 2956: 10.64606106, 2957: 14.5520148158, 2958: -45.5528311139, 2959: 2.62587009065, 2960: 10.5299696849, 2961: -14.5280499389, 2962: 13.083418377, 2963: -2.64627583389, 2964: 0.226435365815, 2965: -14.7901722967, 2966: 0.47888107963, 2967: -29.0468490107, 2968: 30.9323195073, 2969: 14.3883905461, 2970: -0.0393479076411, 2971: 0.529043205016, 2972: 0.136511006923, 2973: -2.36591673035, 2974: 0.655170784699, 2975: 10.0521460782, 2976: -8.23797446708, 2977: -6.23329076088, 2978: 2.51267922559, 2979: -7.98471029591, 2980: -2.63835362907, 2981: -6.49760045607, 2982: -11.7074812598, 2983: -21.9267778289, 2984: -26.5562868516, 2985: 0.379573049133, 2986: 4.99468340876, 2987: 2.90797184359, 2988: 2.20932054757, 2989: 0.222885282563, 2990: 18.8293054634, 2991: 0.542486191855, 2992: 1.97222280386, 2993: 0.549147052531, 2994: 6.87904244969, 2995: 0.579938062537, 2996: 0.526974940226, 2997: 3.28757494136, 2998: 2.47367102031, 2999: -12.1887562646, 3000: 0.576148322342, 3001: 0.403560413347, 3002: -3.52906744568, 3003: 0.361781036704, 3004: 4.61708468337, 3005: -13.736049144, 3006: -3.93797062265, 3007: -0.135744901701, 3008: 0.319687233731, 3009: 24.7394823322, 3010: 7.35228114031, 3011: 37.5452909604, 3012: -0.954829909964, 3013: 9.45658772773, 3014: -2.75210152689, 3015: 10.2519598516, 3016: -14.8079541279, 3017: 12.0869763315, 3018: -0.0557572083009, 3019: -20.5046032035, 3020: 3.85097428179, 3021: -6.95549468135, 3022: 1.57331771822, 3023: -12.5215144476, 3024: 0.108654397656, 3025: 9.6632600168, 3026: -5.76550649393, 3027: -6.02615386696, 3028: 22.1637774805, 3029: -65.2668798023, 3030: -19.5411180125, 3031: 1.92539281184, 3032: 3.59921152867, 3033: -20.6325272891, 3034: 21.8029852615, 3035: 18.2045002467, 3036: 16.5416765404, 3037: -29.3596266344, 3038: 15.1094697824, 3039: -32.1262783721, 3040: -18.5855985492, 3041: 36.9408950532, 3042: 43.8797332889, 3043: 0.997317476007, 3044: -15.1268086094, 3045: 21.3675783991, 3046: -26.6878764897, 3047: -1.08199980207, 3048: 2.21220965236, 3049: -1.09968649055, 3050: -0.381309566411, 3051: 3.31729618784, 3052: -8.01639018367, 3053: 0.722736747838, 3054: 2.81405419714, 3055: -12.8962715672, 3056: -10.5013273117, 3057: 5.39247773348, 3058: -3.13137329256, 3059: -23.5508738692, 3060: -2.79423793341, 3061: -7.41365340768, 3062: 10.2992671574, 3063: 9.1486893201, 3064: -4.94825784305, 3065: -54.3556686274, 3066: 2.04567431466, 3067: -12.3787013956, 3068: 37.9776907974, 3069: 12.2718660227, 3070: -3.00240405597, 3071: 17.317513173, 3072: 7.1172006658, 3073: 5.75819317963, 3074: -8.57786819801, 3075: -24.9804325644, 3076: -22.0187596208, 3077: 36.8976748528, 3078: -22.255697111, 3079: 3.28336114928, 3080: 13.7300366856, 3081: 33.9786956944, 3082: -1.85444257654, 3083: -1.38667625585, 3084: 10.3656599565, 3085: -5.58754232721, 3086: 30.0503753006, 3087: 32.0696873913, 3088: -4.55946925478, 3089: 0.895242617234, 3090: 1.32545433377, 3091: -10.5873802337, 3092: 17.1863839071, 3093: -3.12855282567, 3094: -17.0718368458, 3095: -7.38489416858, 3096: -35.1404416315, 3097: 32.3660237929, 3098: 2.84151613266, 3099: -29.2217035466, 3100: 2.01364842833, 3101: 2.02081663273, 3102: 5.460821829, 3103: -0.663500953158, 3104: 2.64326719979, 3105: 1.26682625069, 3106: 9.53482524585, 3107: 2.25874738908, 3108: -63.9167071028, 3109: 2.61113753143, 3110: 1.57057113247, 3111: -26.9154257941, 3112: 1.91770291332, 3113: 4.61519715941, 3114: 2.05609574119, 3115: -5.99540247042, 3116: 22.8899849892, 3117: -61.6291679681, 3118: -97.4078090965, 3119: -1.90950141506, 3120: -42.1059793091, 3121: 16.1844138281, 3122: 3.02348992057, 3123: -14.6010941762, 3124: 5.6782368203, 3125: -24.2830246889, 3126: 14.1028575624, 3127: 15.0180125501, 3128: -32.2136813395, 3129: -27.5693538836, 3130: -0.867708750877, 3131: 2.24446121839, 3132: 17.6829933399, 3133: -6.47456650445, 3134: 0.657925030765, 3135: -0.193981473935, 3136: -33.3880692371, 3137: 7.73763079, 3138: -31.3821256928, 3139: -8.78493664007, 3140: -25.8793861635, 3141: -7.04127616699, 3142: 0.33511906874, 3143: 1.95085891122, 3144: 2.36902166159, 3145: -5.83916799418, 3146: 0.980400409618, 3147: -5.13106204238, 3148: 3.44517902659, 3149: 5.72486518362, 3150: -0.356471449908, 3151: 3.29720341654, 3152: -6.79900370996, 3153: 4.85624925189, 3154: -6.50369103071, 3155: -9.55109081725, 3156: 1.47800150361, 3157: -2.3880531485, 3158: -4.14701170462, 3159: -8.21304032598, 3160: -27.7579351566, 3161: 6.55499774398, 3162: 8.31983568216, 3163: -8.53234445063, 3164: -2.47784120606, 3165: 1.53090351754, 3166: -10.0126922024, 3167: 2.87408415455, 3168: 7.7151933531, 3169: 4.86008541364, 3170: -12.2557794095, 3171: 6.79223987751, 3172: 3.15387005673, 3173: -3.77056746696, 3174: -1.62416287321, 3175: -2.64612529395, 3176: -4.02797313547, 3177: 2.68138756788, 3178: -1.93095411058, 3179: -1.12093882552, 3180: 1.52835368273, 3181: -0.389027721016, 3182: -7.58284448366, 3183: 1.22422888679, 3184: -3.96202690219, 3185: 11.1437574438, 3186: -0.783545328597, 3187: 1.70260791399, 3188: -1.32043325467, 3189: 3.15631877257, 3190: -4.30645431852, 3191: 0.251243534048, 3192: -5.49020176697, 3193: -0.9989077514, 3194: 2.25244320761, 3195: 0.448346046108, 3196: 2.18546810836, 3197: -0.917634209125, 3198: -0.60644758984, 3199: 2.56018448175, 3200: 5.46230550122, 3201: 4.92739760398, 3202: -12.5841782875, 3203: -1.08664715562, 3204: 7.35368022521, 3205: 3.13806186428, 3206: 3.02759856203, 3207: -1.05820290249, 3208: -3.13188411193, 3209: 2.60210583064, 3210: -1.28539014463, 3211: 2.40249008395, 3212: 1.16724022156, 3213: 8.35908122268, 3214: 2.40723663355, 3215: 5.34667757945, 3216: -0.729135013853, 3217: 2.67500683816, 3218: -0.946762997139, 3219: 1.98265008585, 3220: 4.08784226475, 3221: 5.55585993244, 3222: 1.34427086848, 3223: 0.698446358753, 3224: -0.0628941479824, 3225: 2.19972448366, 3226: -5.92311901868, 3227: -2.69916597526, 3228: -4.17240747788, 3229: 0.952207188584, 3230: -1.22062934087, 3231: 2.08441724729, 3232: 0.00880296865603, 3233: 2.12476250894, 3234: -3.75842884553, 3235: 1.94795120612, 3236: 2.35641424869, 3237: 2.45729595987, 3238: 0.835303986261, 3239: 2.46746253136, 3240: 2.56461213695, 3241: 2.07900586522, 3242: 10.3465741678, 3243: 2.43419713777, 3244: 2.46941576454, 3245: -1.29039902335, 3246: 0.47058084302, 3247: -7.95290024815, 3248: 2.60031634645, 3249: 2.47981314525, 3250: -1.11046789266, 3251: -3.79422570529, 3252: -0.257681980412, 3253: 1.4891474669, 3254: -0.0741266457539, 3255: -1.63157604905, 3256: -11.4324726814, 3257: -0.566484566654, 3258: -1.96636624638, 3259: -1.57191256318, 3260: 2.98878315063, 3261: 7.19989542835, 3262: 10.4366985003, 3263: 9.05727802593, 3264: 10.10973622, 3265: 1.3302437661, 3266: -7.95883223347, 3267: 0.373733563489, 3268: 0.621015694492, 3269: -1.65098581191, 3270: 0.628078129319, 3271: 0.431676561368, 3272: -3.49140227068, 3273: -0.648276529454, 3274: -2.27334523516, 3275: 1.19383236368, 3276: 0.419085689783, 3277: -5.29597192854, 3278: 1.05529978182, 3279: -5.93719451224, 3280: 2.01129804215, 3281: -6.23839850269, 3282: -3.54631828752, 3283: -1.93673283386, 3284: 7.63021645491, 3285: -0.0670627567886, 3286: 2.14476599468, 3287: 0.388797507477, 3288: 0.365327556191, 3289: 1.87559731479, 3290: 0.441709754375, 3291: 1.55139095583, 3292: 1.11470823743, 3293: 0.815384160272, 3294: 0.245540031356, 3295: 0.704258997539, 3296: 0.448612306811, 3297: 2.769282811, 3298: -5.20474528663, 3299: 0.693857872223, 3300: 0.727150193957, 3301: 0.737784305713, 3302: -1.85551831504, 3303: 0.0552667729286, 3304: -4.50191731345, 3305: -0.389867671033, 3306: -1.2859818661, 3307: 16.799835509, 3308: 0.98926932267, 3309: -1.73820734238, 3310: 2.68775761717, 3311: 2.19392648843, 3312: -1.64910520495, 3313: 0.432344389066, 3314: 5.98508083394, 3315: 0.430312788604, 3316: 3.17068542457, 3317: -1.56246414872, 3318: -16.836326976, 3319: -0.0846155967427, 3320: 0.295141628885, 3321: 0.530589548035, 3322: 0.323402111881, 3323: 0.450834259407, 3324: -0.263784647307, 3325: -7.73340843387, 3326: -2.22549498127, 3327: -14.977661444, 3328: 2.46465770674, 3329: -7.23493857143, 3330: 3.09489106801, 3331: -3.13654273387, 3332: -0.302364498783, 3333: 1.08869855824, 3334: 0.356577993136, 3335: 1.40331457604, 3336: -0.0179534517306, 3337: -0.49786468389, 3338: 2.018602217, 3339: -6.36164695071, 3340: 0.480648829358, 3341: 2.29217879669, 3342: 0.398627828584, 3343: 7.79105089265, 3344: 0.459738630303, 3345: 0.380455004998, 3346: -7.65861450162, 3347: 2.41880498403, 3348: 7.76651067812, 3349: 0.345147497553, 3350: 0.26697578576, 3351: -0.330747649368, 3352: 0.318481809529, 3353: 2.27094844663, 3354: -3.165676151, 3355: -2.66939192841, 3356: 0.446709301965, 3357: 1.09274381034, 3358: -8.13294774575, 3359: -0.154903544936, 3360: 2.28428104937, 3361: -3.29588291542, 3362: -1.39018797206, 3363: -5.39993562719, 3364: 1.98383378384, 3365: 1.05037873184, 3366: -2.46851639141, 3367: 0.793053621386, 3368: -3.61223640296, 3369: 2.41842300326, 3370: 0.432377443795, 3371: -2.19778254541, 3372: 11.3553808431, 3373: -0.242476867387, 3374: -3.53717662121, 3375: -2.74563166597, 3376: 2.34647779355, 3377: -0.171536360451, 3378: 8.74545390668, 3379: -6.86436040043, 3380: 2.62447700599, 3381: -3.53132046136, 3382: 5.6111101787, 3383: -2.56802164029, 3384: 5.41013206089, 3385: 6.27827995155, 3386: 0.970764472657, 3387: -0.598728160289, 3388: 5.52878366548, 3389: 7.86303761942, 3390: -2.35386930084, 3391: -5.72493121702, 3392: -7.08738721709, 3393: -5.69924827967, 3394: -3.03086092026, 3395: 5.20053703319, 3396: 2.83980287788, 3397: 2.04045101082, 3398: -3.51263300473, 3399: 2.00553290006, 3400: 1.42297296019, 3401: 2.31412947441, 3402: -7.119536916, 3403: 2.31173070911, 3404: 2.11650699228, 3405: 2.69688413016, 3406: 2.84268060857, 3407: -4.56226322111, 3408: -0.339473915139, 3409: -1.28660767491, 3410: -0.05490270035, 3411: -2.17149579117, 3412: 1.99227592785, 3413: 1.9262350179, 3414: -0.515174637834, 3415: -0.883478496157, 3416: -22.8693583014, 3417: -2.19299654089, 3418: -3.35389429507, 3419: 0.890147398385, 3420: 4.38917110701, 3421: 0.211197767641, 3422: -2.22395802522, 3423: 1.59778587384, 3424: 2.8210476916, 3425: 4.19165338347, 3426: 12.688535285, 3427: 1.62217666162, 3428: 2.47151990165, 3429: -1.61276641436, 3430: 2.19787226051, 3431: 8.20706950586, 3432: 1.0821819652, 3433: -4.84617348242, 3434: 1.14470268602, 3435: 8.46026354231, 3436: 3.38976215, 3437: 7.64357754747, 3438: 2.37433623251, 3439: -0.361688739091, 3440: 3.23623767571, 3441: 4.01147669882, 3442: 0.796194082121, 3443: -8.94958354463, 3444: -0.494664864945, 3445: -4.66354529614, 3446: -2.28343651708, 3447: 2.21204605762, 3448: -3.04424306594, 3449: 2.35855718075, 3450: 2.41188687761, 3451: 2.38986141653, 3452: -2.02766050132, 3453: -1.9487407449, 3454: 0.401650883588, 3455: 13.563981789, 3456: 2.37030989833, 3457: -14.1628249814, 3458: 2.35404771028, 3459: 2.49471954528, 3460: 0.661164855323, 3461: 2.6102648509, 3462: 2.31681916677, 3463: -2.15185242636, 3464: -2.47408759159, 3465: 0.881076430553, 3466: -3.40447716657, 3467: -16.6368350732, 3468: -3.3602257122, 3469: 2.0308917204, 3470: -0.143152499623, 3471: -3.15145624096, 3472: -2.79453103572, 3473: -13.2288786515, 3474: -0.933151122056, 3475: -5.0534118098, 3476: 3.35166900311, 3477: 20.5713348548, 3478: 1.1131554575, 3479: -1.69041560807, 3480: 2.50544336929, 3481: -8.86743762896, 3482: -2.79937510144, 3483: -1.78531359079, 3484: 1.15914948384, 3485: 5.99717569062, 3486: 2.71768514514, 3487: -6.91542009091, 3488: -9.46807464856, 3489: 3.1029822047, 3490: -3.72082623351, 3491: -0.485331564021, 3492: -1.45976421672, 3493: -1.09610296433, 3494: -0.834625887489, 3495: 1.95075799199, 3496: 3.46628443057, 3497: 0.982496050633, 3498: -1.06318944946, 3499: 0.620281727139, 3500: -0.50449960458, 3501: 2.78303792009}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.36097854043, 2: 6.22359982976, 3: 13.7720232119, 4: 16.6266722013, 5: 18.3617832745, 6: -29.6102751023, 7: -2.11722744826, 8: 26.8599451378, 9: -17.5233620389, 10: 21.3553361053, 11: 28.1527041753, 12: 5.9693718709, 13: -19.1333710805, 14: 6.24305863806, 15: 6.81660602269, 16: -1.31151214416, 17: -25.726533357, 18: -9.61773339101, 19: -6.6107585072, 20: -26.854369555, 21: 34.734682166, 22: 17.0430631949, 23: -26.8069527825, 24: 3.78396097886, 25: 22.4046654669, 26: -30.0423560733, 27: 6.70214997929, 28: -20.2106457623, 29: -30.5818784048, 30: -15.470789675, 31: -18.1996654819, 32: -7.37924304147, 33: 26.4665981355, 34: 11.1387343465, 35: -7.70520787589, 36: 7.24789904604, 37: 9.70809773356, 38: 0.369689793815, 39: -36.0858569037, 40: -26.7033213799, 41: 2.46380492458, 42: 3.66498324761, 43: 24.7437522363, 44: -10.3795979944, 45: -18.8906427898, 46: 18.2027736971, 47: -0.991936239741, 48: 4.52554769302, 49: -4.09854332775, 50: 4.1203473219, 51: 6.29279235011, 52: -16.2485106627, 53: -4.27857579528, 54: -4.02312547667, 55: 18.6215225274, 56: 3.23395572805, 57: 0.188916869264, 58: -6.21386604289, 59: 19.3253611105, 60: 18.0696854458, 61: -1.315472925, 62: 0.223376304871, 63: -4.77986571198, 64: -11.137506826, 65: -9.26625489138, 66: 2.09285839761, 67: 25.9200899782, 68: 0.0613690934911, 69: 22.3384738766, 70: 3.90584176922, 71: 3.29164793558, 72: -14.7145249069, 73: 1.98957308114, 74: 26.9517980706, 75: -0.327026508362, 76: 2.36236405228, 77: 6.8879856432, 78: -2.66066706977, 79: 38.5611093906, 80: 3.07234651821, 81: -11.2602680975, 82: -39.3351483955, 83: 2.85662683754, 84: -0.6536626444, 85: 2.03705488418, 86: -21.6737395893, 87: -31.4597314129, 88: -19.7522766862, 89: 3.42365713118, 90: -12.1097658708, 91: -0.719189117352, 92: 3.09000256108, 93: -34.9409359602, 94: 1.6300612172, 95: 2.4570836465, 96: 3.7409544858, 97: -9.98575101521, 98: 3.84525444197, 99: 1.95727407705, 100: -11.8456039642, 101: -24.7433780534, 102: 2.85192097641, 103: -17.3483128752, 104: 2.18315255782, 105: 0.625688222902, 106: -35.944042474, 107: 3.25352972724, 108: 0.247761959163, 109: 5.25015183688, 110: -35.6841890486, 111: -1.84180800841, 112: 24.1426164533, 113: 27.9926193462, 114: -9.08129955027, 115: -34.102478756, 116: -1.65388594629, 117: 11.9844966501, 118: -3.78819228461, 119: -7.01712070581, 120: -10.5375884052, 121: 5.48100559284, 122: 37.0048346317, 123: 35.1012554605, 124: 0.81121230646, 125: 9.27046370717, 126: 0.500011785009, 127: -0.879370086412, 128: 5.69427786791, 129: -7.62987966107, 130: 0.41190299365, 131: 3.36818637007, 132: -0.672685450252, 133: 15.0956440279, 134: -0.0215390761423, 135: 7.14453176545, 136: 3.8534511256, 137: -4.50663236869, 138: -8.30247525575, 139: 2.00676910639, 140: -4.8047380073, 141: 4.58207701188, 142: -25.5851378258, 143: -1.41680920955, 144: 1.14658076044, 145: -3.41105821588, 146: -0.851173919707, 147: 0.690356881133, 148: 9.16207202512, 149: 0.619295022318, 150: 3.26484010338, 151: 4.41373378556, 152: 0.300619964135, 153: 0.520119818114, 154: 0.577884416202, 155: 0.497103136208, 156: -2.25184164254, 157: 22.7398162213, 158: 0.757165094656, 159: 4.21017303252, 160: -1.0695051708, 161: 1.16405757885, 162: 0.852181499139, 163: -4.35798925375, 164: 3.95704492306, 165: -2.50291261775, 166: 23.9050192034, 167: -0.541795381233, 168: -3.75496174042, 169: -2.0511585, 170: -6.30286542427, 171: 3.09109987923, 172: 0.512515849319, 173: -89.3446349922, 174: 0.408609082653, 175: 8.01031419986, 176: 33.1623920096, 177: 53.149300923, 178: -1.99861020791, 179: 0.666014910193, 180: 0.386433444723, 181: -0.151372480839, 182: 0.51885767846, 183: 1.88055942505, 184: -14.3470902735, 185: -16.8208226088, 186: -3.22502106543, 187: -3.03202949639, 188: -14.2411408193, 189: 5.8001086138, 190: 9.3192072882, 191: 0.833488710096, 192: -3.18280457648, 193: 0.514058346213, 194: 30.2446799562, 195: 1.86856345175, 196: 9.43241426341, 197: 4.40694959623, 198: -18.4475364903, 199: 0.611744549671, 200: 3.21968047667, 201: 0.48651338561, 202: -3.34755605032, 203: 0.458211496568, 204: 0.460060678686, 205: -2.91190438378, 206: 1.85316129484, 207: 39.0619807471, 208: 0.696595394362, 209: 0.484310990583, 210: 0.960309631868, 211: 0.532800439275, 212: 1.35994165475, 213: 0.886345056309, 214: 1.52453804376, 215: 1.11992961259, 216: 1.15953346938, 217: -32.5415317479, 218: 5.4780626264, 219: -2.98062191299, 220: 3.95344187072, 221: -4.46363230219, 222: -7.85188533179, 223: 0.178167103525, 224: 1.52969078596, 225: -3.9932561405, 226: 0.424157223781, 227: -25.362307154, 228: 7.55858044769, 229: 4.55116534436, 230: 5.88443617229, 231: -3.94190954607, 232: 1.14573394083, 233: 32.902312786, 234: -3.11577137659, 235: -0.106814007825, 236: -12.5093399779, 237: -23.4713268585, 238: 21.3471363648, 239: -0.0251685353146, 240: 3.17882519231, 241: 9.06491016439, 242: -42.5156561406, 243: -8.05117010373, 244: -5.47062798756, 245: -0.577744334005, 246: 20.397805594, 247: 14.7008661709, 248: -20.2930722037, 249: -5.35648153081, 250: -19.0482897032, 251: -10.8197593224, 252: 9.73722025314, 253: 22.6730487269, 254: 7.42692370382, 255: 14.0422127618, 256: -0.544267491267, 257: 19.6726117096, 258: 1.99767791937, 259: 3.17097248619, 260: 17.332335978, 261: -28.680501075, 262: 1.52909760635, 263: -5.1717123687, 264: 0.251019788327, 265: -2.06849516034, 266: -8.0253377629, 267: -1.42294600884, 268: 9.22691017939, 269: 19.1227813242, 270: -1.4903881218, 271: 1.57865484648, 272: -16.0984633791, 273: -16.7384996533, 274: -4.81703451179, 275: -22.2627711028, 276: 38.9027693152, 277: 5.80903118082, 278: 21.6683687427, 279: -18.4905792594, 280: -28.3922990458, 281: 2.12559353004, 282: 34.402888226, 283: -3.771575999, 284: -0.663339301443, 285: -66.5590494997, 286: 8.82759635434, 287: -0.0492060143229, 288: 0.357809974967, 289: 8.83985218503, 290: -2.51642299071, 291: -7.89494338397, 292: -5.85857272796, 293: 20.9302028047, 294: -11.0754749331, 295: 10.1367851511, 296: -13.2406664661, 297: 27.5340821167, 298: -28.4089952789, 299: 1.41769566591, 300: -23.5344293641, 301: -0.254995935579, 302: -14.421031866, 303: -3.6501043855, 304: 13.9516173964, 305: -13.1012963889, 306: 2.3925935796, 307: -0.430107005101, 308: 3.12852510603, 309: 3.12952785686, 310: -0.391541778376, 311: -4.71909857382, 312: 3.8129084109, 313: 3.07023866891, 314: 20.7537164802, 315: 1.80250511441, 316: -11.8012112315, 317: -3.49633068384, 318: 4.33935614267, 319: -5.12235224508, 320: 3.63753674701, 321: 0.826355258347, 322: 12.0775226871, 323: -10.0542855646, 324: 12.5374920488, 325: -32.4705689438, 326: -20.6996284254, 327: 12.1630337291, 328: 29.1698811762, 329: 28.1234423959, 330: -5.60054603701, 331: 6.0860057661, 332: -20.7518471652, 333: 0.705212563377, 334: -5.02479641014, 335: 9.37759930715, 336: -1.5458363273, 337: 12.6465312596, 338: -8.18785418909, 339: 2.97395404196, 340: 4.24000900226, 341: 4.89324119847, 342: -4.87211099545, 343: -12.6933930308, 344: -23.7451787414, 345: -1.27614928985, 346: 11.9936010486, 347: 10.5181354385, 348: -4.86901611059, 349: -16.0203916922, 350: 0.642836618662, 351: -4.11708620093, 352: -1.27434236443, 353: 1.88425290498, 354: 4.95255720345, 355: -2.01604201524, 356: -3.97305116355, 357: 0.39767515893, 358: 3.72193258327, 359: 5.45323857975, 360: 2.0096768796, 361: 1.47548971149, 362: -3.09719410569, 363: 1.17463335144, 364: 3.77513934974, 365: -4.42611052383, 366: -0.302387479392, 367: 1.10134359357, 368: -5.70809280465, 369: -12.907110661, 370: 3.89924311978, 371: 5.29926692276, 372: 3.50244602538, 373: 3.23351094256, 374: 5.52195142074, 375: -5.78505007873, 376: 0.869752590251, 377: -2.1971336812, 378: 6.87109019362, 379: 0.745922168755, 380: 4.21710474345, 381: 0.481392675019, 382: 2.76245627671, 383: -1.74774086588, 384: 1.23250979072, 385: 1.34707018087, 386: -0.00408347630503, 387: -0.770970742549, 388: 0.960073284139, 389: -6.15723224667, 390: 0.912788922686, 391: 3.30499851805, 392: -3.82706723398, 393: -6.92348656377, 394: -0.625560876555, 395: 0.973220463531, 396: -1.58970859314, 397: 3.13201827452, 398: -11.7990179915, 399: 2.89220635734, 400: 0.524014970436, 401: -0.392474432872, 402: 3.35305100383, 403: -4.84451333712, 404: -2.43884919684, 405: 3.59427762062, 406: -9.45647042188, 407: -0.493034203015, 408: -13.4211050003, 409: -10.4494289977, 410: 5.45994848838, 411: 3.39105025511, 412: 7.65491024458, 413: -3.34388091659, 414: -1.94668986017, 415: 0.866233309053, 416: -4.84266517032, 417: 2.23013249443, 418: -0.750992340992, 419: 3.07787393698, 420: 2.9895039614, 421: 3.43612179682, 422: 3.02621688996, 423: -18.6950236756, 424: 0.334625144467, 425: 2.44031275773, 426: -1.21481406633, 427: 3.51669183218, 428: -7.21594708868, 429: -0.593420207506, 430: 0.0456356436432, 431: -7.39512318478, 432: 0.808165945158, 433: 0.586022992659, 434: -7.74673333142, 435: -3.12668804484, 436: -0.368407595545, 437: -4.57784604828, 438: 2.83550963376, 439: -1.87181820467, 440: 3.01709947833, 441: 2.15863915148, 442: -21.9925643133, 443: 3.15064373366, 444: 1.01234245543, 445: 3.35071095035, 446: -6.52183559217, 447: 3.18259731122, 448: 2.79243549342, 449: 3.14815851225, 450: -7.88526117322, 451: 2.80413977888, 452: -8.83279459756, 453: 2.94647420708, 454: -0.200438740057, 455: 4.27154510526, 456: 2.98545818779, 457: 2.98017104691, 458: 2.92496600421, 459: -8.39897372382, 460: -0.918769967414, 461: 1.75370838549, 462: 12.2802237825, 463: 2.84355970539, 464: 0.721569530837, 465: -0.0124283247287, 466: 3.07706675986, 467: 0.933901334382, 468: -5.99012817072, 469: 3.20284147519, 470: 5.24936572704, 471: 2.34391221027, 472: 3.35414084168, 473: -0.0772904437247, 474: 4.20686106096, 475: 0.577175298508, 476: 1.13282108421, 477: -1.17444465333, 478: -3.80726797723, 479: 0.61271251848, 480: 2.76438969692, 481: 0.103023450783, 482: -4.45994036358, 483: 0.47606115772, 484: 1.70219517636, 485: -3.52703208219, 486: 0.969011656225, 487: 4.60114252525, 488: -5.13756843762, 489: -1.5776232566, 490: -8.57514342378, 491: -1.60842395965, 492: 3.25588701952, 493: 0.863456891907, 494: -0.689043977499, 495: 0.576928666864, 496: 0.858970469926, 497: 3.76577734374, 498: 0.768056115764, 499: 2.84096164913, 500: -4.15355718451, 501: 0.79246880769, 502: 0.689203261141, 503: 0.721705115078, 504: 0.860455004256, 505: 2.74509210222, 506: 7.51808053967, 507: -0.262221299309, 508: 0.825612840624, 509: -0.298167954058, 510: -1.59788914111, 511: -4.30935887332, 512: -1.18874105284, 513: 3.8195458135, 514: -6.82099306081, 515: -12.6936987755, 516: 2.28935484336, 517: 1.48509503101, 518: -1.70914261383, 519: -0.513557379691, 520: -9.51459906521, 521: 0.638737898, 522: 0.80239241287, 523: 0.606970537276, 524: -5.72297397732, 525: 3.48493796101, 526: 13.7975799157, 527: 0.967296483229, 528: 0.623579063327, 529: 0.584645008447, 530: 0.41245503659, 531: 0.963428724864, 532: -0.30931053844, 533: 3.10844724236, 534: -2.31071791086, 535: -26.0477088546, 536: 6.11014020317, 537: -0.475411674456, 538: -1.87183366585, 539: -2.73396713324, 540: -2.0832325028, 541: -5.75534538896, 542: 0.902166791855, 543: 8.40959992044, 544: 2.84467875454, 545: 3.35662511576, 546: -3.0677543666, 547: -6.23673396029, 548: 0.82369512371, 549: 1.14150602188, 550: 0.63863814046, 551: -4.63477455705, 552: 0.688017264009, 553: 0.771316838383, 554: -2.9437793034, 555: 2.30562550426, 556: -16.220206012, 557: 0.606195200798, 558: 0.744296355692, 559: 1.85721273584, 560: 0.710452518308, 561: 0.00899109038183, 562: -6.94796546793, 563: 2.17187621523, 564: 1.82136788, 565: -1.42512007592, 566: 6.69101190756, 567: -1.47272464391, 568: -2.1310922746, 569: -14.3474591539, 570: -2.26043440955, 571: -1.3504221356, 572: 5.12314517037, 573: 2.87817383495, 574: 3.14847722738, 575: 0.669174294369, 576: 6.64871137324, 577: 1.35979902575, 578: 1.62573939356, 579: 2.60735702815, 580: -10.84947831, 581: 3.26854590382, 582: -15.7159276765, 583: 5.28012695404, 584: 2.58009746271, 585: -2.4306928982, 586: 1.37777069188, 587: 12.5533824587, 588: 1.94033225637, 589: 2.92303344852, 590: -4.78400917441, 591: 3.56982808629, 592: -17.2949348177, 593: 4.64386407185, 594: 0.480698963361, 595: 3.88053454457, 596: -7.12072350964, 597: -4.09057063831, 598: 8.10546749583, 599: 5.59184779284, 600: 4.18339310206, 601: -4.54968918362, 602: -2.6804320497, 603: 1.27787025499, 604: -4.62509932941, 605: 1.353018907, 606: -2.27320983147, 607: 2.24349274906, 608: 2.80125625616, 609: 1.03735966208, 610: 4.90511920216, 611: 4.97262534179, 612: 1.76091377835, 613: -2.0367707784, 614: 2.60357277254, 615: 5.36911680862, 616: -2.67078663521, 617: -6.65219817927, 618: -0.306323353141, 619: -4.10231344972, 620: -1.68096915812, 621: -1.4491808621, 622: -5.28262585217, 623: 7.7140076126, 624: -0.420408148591, 625: 0.773297461568, 626: 1.59604652044, 627: -1.29894138486, 628: -1.76562783001, 629: 1.61634951259, 630: 6.23186614599, 631: 0.688291454369, 632: -8.264789739, 633: -10.6798085167, 634: 1.80153179171, 635: -16.2210193424, 636: 3.11279100385, 637: 4.14969766742, 638: 1.35741210026, 639: 0.171957979427, 640: -0.0984314051307, 641: -1.33225142484, 642: 0.0298454942965, 643: -3.4286265615, 644: -6.61269594262, 645: 0.878298283362, 646: 0.143846502958, 647: 0.795340204238, 648: 1.21242183475, 649: 2.3042521402, 650: -2.1943009111, 651: -1.6032979665, 652: -0.0213741828248, 653: -3.2559670013, 654: -5.20438959776, 655: 0.874926828564, 656: 1.77691675243, 657: 1.0683043832, 658: 0.995325327206, 659: 1.61618985469, 660: 3.19692674398, 661: 3.13922921226, 662: 2.0156618182, 663: 6.63592792258, 664: 2.48218725339, 665: 9.51146045944, 666: 3.78481599595, 667: 3.08566233924, 668: -4.61973752605, 669: 1.96366277927, 670: -1.37902171699, 671: 7.09562077653, 672: -2.52444708993, 673: -1.09917433402, 674: -1.96825986024, 675: -1.5220565208, 676: -0.520110622807, 677: 8.10768593435, 678: 3.47884783873, 679: -4.22979514504, 680: 3.34506874536, 681: 0.47008204594, 682: 2.82484440354, 683: -3.40554949128, 684: -8.23276730811, 685: 0.82151615475, 686: -0.971434931575, 687: -4.3757693886, 688: 2.93648484313, 689: 3.01696699725, 690: 2.91631702118, 691: 4.88941093389, 692: -2.22266546594, 693: -7.20010709688, 694: 2.25262095483, 695: -6.01130750427, 696: -3.14570356254, 697: -9.81979140057, 698: -3.42663932669, 699: 0.307409505242, 700: 99.2928736937, 701: -12.0697805285, 702: -8.97258239138, 703: -9.65431505902, 704: -12.4767717843, 705: -17.5012870469, 706: -23.7020169619, 707: -18.3940459311, 708: 10.0724192165, 709: 11.1009521491, 710: -21.4301111793, 711: 16.0836375497, 712: 8.80758730017, 713: 8.17827125305, 714: 0.950502059174, 715: -27.9801903835, 716: 18.7575035216, 717: 34.780538779, 718: 4.44754036253, 719: 12.8351716128, 720: 14.7452944902, 721: 8.4683613832, 722: -28.3627089869, 723: -26.4306876367, 724: -7.23763930378, 725: 5.62570232616, 726: -0.826594025616, 727: -14.1147130048, 728: 17.0421688174, 729: 20.4706574262, 730: -19.4391259799, 731: 15.5086781658, 732: 17.6498378003, 733: -20.6479711799, 734: 32.1461286443, 735: 13.5834898443, 736: 54.3453882098, 737: 17.2584452861, 738: 14.6441356342, 739: -11.2003643121, 740: 1.5735603551, 741: 25.9063871844, 742: 17.1203317917, 743: 24.0819986776, 744: -16.7522795808, 745: -14.6717219781, 746: -2.0835608958, 747: -8.17143232576, 748: -7.71367497167, 749: -5.79250407343, 750: -1.58885092358, 751: 2.08194535154, 752: -9.24509743639, 753: 3.45893672357, 754: 1.96215980816, 755: 70.1167700996, 756: 4.84681871811, 757: -4.28154416144, 758: -24.2310275971, 759: 55.7938151612, 760: -14.9528970763, 761: 37.4422717698, 762: 5.65039658637, 763: -15.7810441298, 764: -0.409083670118, 765: -17.0557171415, 766: 1.98076615543, 767: 2.71304905288, 768: 1.39989054149, 769: 1.2085053207, 770: -25.8763690726, 771: 1.93123458921, 772: 43.7126438962, 773: 18.4342697597, 774: 11.6210327889, 775: 9.93692317311, 776: 9.02317120392, 777: 73.1677866341, 778: 35.3481621335, 779: -5.69349391544, 780: -5.82437713449, 781: 1.42205028146, 782: -27.187222592, 783: 4.00021639307, 784: -6.02650912829, 785: -5.54315892152, 786: 18.2887994592, 787: 2.04498201046, 788: -8.32592017392, 789: -0.0843057771247, 790: 1.86146075126, 791: -4.95686171464, 792: 1.81559817721, 793: 1.92285780052, 794: 1.11153713102, 795: 10.9779594358, 796: 1.73373639928, 797: -15.1088999289, 798: -0.527774395919, 799: -11.9561839588, 800: 1.72550826014, 801: 2.72892045475, 802: 2.1145835083, 803: -3.11338708511, 804: -9.57962581331, 805: 1.95749209925, 806: 1.88782736144, 807: -3.54727437792, 808: -12.827046045, 809: -0.778172687501, 810: -5.89793866801, 811: -11.5664281241, 812: 19.4548100467, 813: -43.0927031772, 814: -5.68380808387, 815: 0.0440908139347, 816: 2.25514929861, 817: -4.70961159907, 818: -6.33683964529, 819: -14.5145496728, 820: 5.75067514542, 821: 25.1534180979, 822: -24.2425887, 823: 12.6943770943, 824: 0.371868719353, 825: 0.244029959933, 826: 3.42552266166, 827: 18.2983645548, 828: 0.389806488308, 829: 1.75206837655, 830: 0.51180307062, 831: 10.3281698532, 832: 2.23450276388, 833: 0.37884471631, 834: 6.07295670209, 835: 1.25746962683, 836: 4.41197220456, 837: -4.57544932972, 838: -0.0355602842995, 839: 26.1376296034, 840: 41.7223013093, 841: 35.725634621, 842: -0.355395939384, 843: -9.37789542855, 844: 3.16966827519, 845: 0.531808697873, 846: 16.667222531, 847: 0.401063650042, 848: 1.96114788819, 849: 2.55937329317, 850: 0.430922418506, 851: 0.303102773348, 852: 0.527492494944, 853: 0.29812756123, 854: 1.67415518767, 855: 5.87190717077, 856: -0.855630238942, 857: 0.999860642057, 858: 0.330549568235, 859: -3.41113534375, 860: 2.9046813492, 861: -5.39923368065, 862: 6.698452231, 863: 21.8378512241, 864: 8.67966014815, 865: -12.92006135, 866: 11.3890655461, 867: -11.9981762886, 868: 2.44163973711, 869: -16.7536228634, 870: 0.362865109827, 871: 13.138976438, 872: 0.361167581707, 873: 10.5442608207, 874: 15.1701361804, 875: -5.19708430712, 876: -1.39455458116, 877: 0.359881361717, 878: 0.123633883633, 879: 2.00357198895, 880: 0.097016388668, 881: 3.11872846575, 882: 40.2044642179, 883: 10.2519794867, 884: -38.179924207, 885: -26.1478585673, 886: -5.62408579586, 887: -16.9267830012, 888: 2.00864816886, 889: 0.79383453607, 890: -8.40749386452, 891: 0.432450290048, 892: 63.5881015255, 893: 9.40781316282, 894: -3.82149132944, 895: -15.1190978253, 896: 12.4029186178, 897: 0.406455770415, 898: 2.00279560068, 899: 0.553717858491, 900: -19.2378877504, 901: 0.50874611978, 902: 0.292228239533, 903: -3.78161014356, 904: 1.71962396896, 905: -31.6315370646, 906: 0.271935141351, 907: 0.303489402823, 908: -2.4487663701, 909: 0.4585285484, 910: 4.2843072816, 911: 16.6874898802, 912: -66.9507198457, 913: -1.15803955913, 914: 1.61377993222, 915: -6.03008088046, 916: -10.9503398607, 917: 30.4372289577, 918: 28.6145867775, 919: -4.89051671073, 920: -13.3873851665, 921: 3.35359180688, 922: 1.72421279021, 923: -7.09434059135, 924: 0.465205410635, 925: -1.31038394628, 926: 8.55363811726, 927: -0.766573297167, 928: -1.9448287335, 929: 25.8380799158, 930: 0.212758755406, 931: 15.1562661414, 932: 8.7224144869, 933: 1.96251971842, 934: -22.8056247137, 935: 3.96230933351, 936: 18.2854349159, 937: 2.21713319947, 938: 1.76795713661, 939: -10.8920137741, 940: -13.0666070003, 941: -15.0267900824, 942: -10.3165372202, 943: -47.7785350171, 944: -4.93595253575, 945: 8.80097253344, 946: 38.8757937774, 947: 13.9193353549, 948: -4.51872961545, 949: -26.9182523166, 950: -36.1462508217, 951: 15.4352102376, 952: 9.6035423954, 953: 5.47660295165, 954: -0.415021886786, 955: -19.2273358424, 956: 2.03026210437, 957: 2.05007196584, 958: 2.50485539823, 959: 1.08521044052, 960: 1.6670002719, 961: 6.24491217234, 962: 6.85486502947, 963: 1.72026050211, 964: 5.77874074997, 965: 1.17656221777, 966: -25.6811207874, 967: 10.7941861046, 968: -0.802089770418, 969: 2.1588173332, 970: 14.2626986917, 971: 15.2320358068, 972: -35.7267519577, 973: 11.2479231129, 974: 38.1532416237, 975: 14.0363765737, 976: -29.9425287516, 977: 4.14694387741, 978: -7.38893844085, 979: 2.52035119053, 980: -26.5159343445, 981: 19.0554634499, 982: 11.3673824634, 983: -19.7410080992, 984: -6.81130485133, 985: 1.64993387726, 986: 26.9599809707, 987: -13.4479387768, 988: 27.5582382377, 989: -6.58630318422, 990: -48.1180846857, 991: -24.9629630447, 992: -3.90259545512, 993: 82.182367773, 994: -17.5586670894, 995: -27.4310138401, 996: -7.35677388451, 997: 0.685876071626, 998: 10.0206720748, 999: 0.996896115552, 1000: -43.063212708, 1001: -7.48828358222, 1002: 18.4713941563, 1003: -2.74707834556, 1004: 1.75667085646, 1005: -48.6941311448, 1006: 1.89016585782, 1007: 1.9302272599, 1008: -10.4260807322, 1009: 37.3991842712, 1010: 1.78539221452, 1011: 1.9739568171, 1012: 19.49180603, 1013: 1.78208925594, 1014: 57.5668516614, 1015: 1.80144672356, 1016: 1.94439186151, 1017: -37.3343425301, 1018: 2.29760947669, 1019: 1.38052886043, 1020: -0.533622384317, 1021: 24.3921063921, 1022: -41.880651342, 1023: -17.6382799545, 1024: 15.4820963253, 1025: -3.02135740606, 1026: -22.2060609639, 1027: -6.79935411729, 1028: 5.66087845399, 1029: 0.893142462463, 1030: 9.32405392647, 1031: 1.9205720388, 1032: 19.0147468592, 1033: 0.929766537962, 1034: -52.3785875435, 1035: -38.3196935576, 1036: 19.2992721262, 1037: 1.9785752349, 1038: -13.2092409707, 1039: -19.8049452316, 1040: -4.34600755269, 1041: 6.62052260256, 1042: 15.3344629022, 1043: 26.1044374014, 1044: -64.3111994905, 1045: 40.4234955428, 1046: -8.72366695927, 1047: -55.5294050952, 1048: -0.207221587925, 1049: -2.05661502264, 1050: -16.5430527979, 1051: 13.202543419, 1052: 25.0793741667, 1053: 0.802724689606, 1054: 11.2651584765, 1055: -18.8470360208, 1056: 1.97359004818, 1057: -9.02833575344, 1058: -29.4349921843, 1059: -5.60872284274, 1060: -7.66229886137, 1061: -8.30983709557, 1062: 14.0481221425, 1063: 21.4214803064, 1064: 6.75559896636, 1065: 8.25281228894, 1066: 10.4160269392, 1067: -3.62633482027, 1068: 5.41296618444, 1069: 27.1257433668, 1070: 17.6747543612, 1071: 2.17082286817, 1072: -14.9206493936, 1073: -4.34524420422, 1074: 9.6408033034, 1075: 15.2690364547, 1076: 10.3791348032, 1077: -13.249607788, 1078: -14.8931278213, 1079: 5.74567875196, 1080: 8.91650455096, 1081: -12.1213333808, 1082: 21.1603874742, 1083: -3.68265086203, 1084: 2.52517579706, 1085: -2.28911135579, 1086: 5.56518296549, 1087: -8.57420207465, 1088: 10.8858011866, 1089: 7.25442552866, 1090: 8.80572821957, 1091: 24.5801268443, 1092: -13.6387239609, 1093: -1.26762585031, 1094: 11.1116656235, 1095: -5.8915710855, 1096: 21.9195082426, 1097: 13.9684896072, 1098: 16.894093372, 1099: 39.1254638161, 1100: -1.23473084797, 1101: -19.3715929055, 1102: 2.31963558603, 1103: -8.01352080978, 1104: 43.9924862781, 1105: -0.711671817819, 1106: 13.6896845568, 1107: -4.87354116846, 1108: 20.489916449, 1109: -5.89139515028, 1110: 24.1175661325, 1111: 22.8073227982, 1112: 8.87731262379, 1113: -4.21637578555, 1114: 14.4756867131, 1115: -1.52610439391, 1116: -8.55534284693, 1117: -1.57782127091, 1118: -11.4204258451, 1119: 3.94935420642, 1120: 0.254159644455, 1121: 0.420080605165, 1122: 1.15537309401, 1123: 3.41466843635, 1124: 6.90543602748, 1125: 8.75191286802, 1126: 49.3315746041, 1127: 21.3535311964, 1128: 15.0675040994, 1129: -21.6762830666, 1130: 2.16784943028, 1131: -12.398676221, 1132: 20.1538740539, 1133: -15.2449541665, 1134: -3.36094535073, 1135: -21.7032599162, 1136: 5.8843869749, 1137: -33.5915988228, 1138: -0.476364755116, 1139: 3.3983473716, 1140: 1.39941047163, 1141: -1.56263921214, 1142: -0.649655098863, 1143: -2.24023846868, 1144: 5.32104091069, 1145: -1.29806667113, 1146: 22.2398823245, 1147: -2.52312330203, 1148: -56.1903706967, 1149: 27.6866387937, 1150: -3.04078385682, 1151: -1.75984619321, 1152: -0.957687770367, 1153: -22.2460889841, 1154: 0.667614158081, 1155: -8.48182087375, 1156: 4.55984568324, 1157: -10.2311474877, 1158: -19.4635191285, 1159: -2.95165002519, 1160: -3.80924509719, 1161: 22.7079874653, 1162: 4.40282701509, 1163: -9.13625994534, 1164: 8.92402919059, 1165: 1.05336142909, 1166: 19.3538053286, 1167: -1.86731951207, 1168: -2.18012009488, 1169: 8.09061102211, 1170: 3.26421321948, 1171: -12.8208625618, 1172: -12.2221473298, 1173: -0.120931195213, 1174: -1.9506405849, 1175: 2.4485203491, 1176: 12.4058480142, 1177: -0.200365390967, 1178: -0.689534743241, 1179: -1.99626492861, 1180: 4.68493033197, 1181: 0.175029619718, 1182: -3.26056777, 1183: -27.9260039639, 1184: 2.26622308443, 1185: -16.6617236603, 1186: 11.5502764891, 1187: 1.68905641193, 1188: -22.7762728331, 1189: -3.41254503386, 1190: -6.42121405672, 1191: -2.96130629818, 1192: -6.65580880659, 1193: -0.20700891363, 1194: -1.08875605622, 1195: 22.154017974, 1196: -0.308550512052, 1197: 0.567479901947, 1198: 18.4198232611, 1199: 0.803880199644, 1200: -0.54327285556, 1201: -0.164653511339, 1202: -0.199467859168, 1203: -4.22837417963, 1204: -0.242804374174, 1205: -1.31845893855, 1206: -0.920891932221, 1207: 0.0143443201522, 1208: 0.213181127437, 1209: 0.288576784522, 1210: -23.750287169, 1211: -20.8008000283, 1212: -36.0557119245, 1213: 33.4801915474, 1214: -14.3908374453, 1215: 2.02007778245, 1216: -2.88565469046, 1217: 10.4148082709, 1218: 12.9258217001, 1219: -0.250601798649, 1220: 9.72536800029, 1221: -0.349348068786, 1222: 23.1441899586, 1223: 33.7018818856, 1224: 22.7247643121, 1225: -0.458942797564, 1226: -0.279342347532, 1227: -0.40861245869, 1228: 0.042493302112, 1229: -0.417616863625, 1230: 2.36328775122, 1231: 13.1754116518, 1232: 1.69109466843, 1233: -18.9241875023, 1234: -9.97037784423, 1235: -9.45970765393, 1236: -10.8123985927, 1237: 19.8984878767, 1238: 6.42552834709, 1239: -30.1176776691, 1240: -0.324292617731, 1241: 13.8265385372, 1242: -3.17549107427, 1243: 0.887607178618, 1244: -5.63082863247, 1245: -10.1969105922, 1246: -0.348991948015, 1247: -0.0587462434293, 1248: -0.443948896168, 1249: -41.6065027705, 1250: -0.206795677887, 1251: -0.503334738475, 1252: -25.0096156527, 1253: -1.52872212314, 1254: -12.7495320224, 1255: -0.285388282537, 1256: -0.502088857557, 1257: -1.25587346583, 1258: -0.343271554052, 1259: 12.5236248687, 1260: -3.46684521021, 1261: -3.79364621588, 1262: -1.00945164814, 1263: -0.267454724418, 1264: 3.46598853963, 1265: -5.17815435983, 1266: 17.2368849945, 1267: 19.0959994815, 1268: 11.165566197, 1269: -2.10851324194, 1270: 8.98560796208, 1271: -11.5000885946, 1272: 13.7488420391, 1273: -0.311510894499, 1274: 4.04126348756, 1275: 5.33447789437, 1276: -2.61636779878, 1277: -8.13866553812, 1278: 12.4100814783, 1279: -2.77298527644, 1280: -42.9501794617, 1281: -3.52346085853, 1282: -0.207777237488, 1283: 30.5012036274, 1284: 27.9304598295, 1285: 6.14742186686, 1286: 1.16442607498, 1287: -0.542734886845, 1288: 2.47911407812, 1289: -19.125171588, 1290: -6.74658631698, 1291: 13.6177937808, 1292: 34.0067511672, 1293: 16.8721218087, 1294: 7.91398689743, 1295: -19.5244728408, 1296: -24.5799649943, 1297: 12.4453476399, 1298: -21.3419642705, 1299: 17.5183480492, 1300: 8.36444186992, 1301: 1.09841804724, 1302: 3.32986203043, 1303: 6.43280767139, 1304: -15.2407803512, 1305: 1.57740499136, 1306: 0.681449142168, 1307: -10.0291269904, 1308: -8.83620980195, 1309: 11.6859568513, 1310: -5.56958633401, 1311: 6.07383748626, 1312: -4.31843412096, 1313: -11.707181817, 1314: -16.0915404281, 1315: -18.385053395, 1316: 6.49699812059, 1317: 1.041884482, 1318: -2.44520781048, 1319: 15.8823367837, 1320: 4.22358126122, 1321: 55.1842401281, 1322: 9.4907690614, 1323: -5.46367186524, 1324: 34.0628012765, 1325: -4.77733151435, 1326: 12.6621859701, 1327: 4.94328269836, 1328: -5.14216208818, 1329: 15.6631423224, 1330: -28.6949713317, 1331: 62.7821383952, 1332: 22.9784629828, 1333: 0.517720536767, 1334: -1.32980524014, 1335: 8.8389303088, 1336: -4.57524200611, 1337: 11.0257159527, 1338: -11.0932375668, 1339: 8.97562630007, 1340: 13.9920744504, 1341: 19.7819774332, 1342: 4.80445388651, 1343: -13.3296233041, 1344: 1.13871672108, 1345: 15.6636735223, 1346: 8.10196947969, 1347: -13.9271613493, 1348: 1.23839169823, 1349: -29.8841836918, 1350: -5.41014293489, 1351: -5.87274395884, 1352: 8.96411257658, 1353: -0.628256856571, 1354: -18.2932835192, 1355: 0.152179841296, 1356: -0.0680504188287, 1357: 0.869769880389, 1358: 25.8034714214, 1359: -3.38024110194, 1360: 5.26749026249, 1361: 26.3816034042, 1362: -1.54744357614, 1363: 35.5950499052, 1364: -3.22735971844, 1365: -1.41716417594, 1366: -22.276243552, 1367: 8.30927745454, 1368: 11.6099955638, 1369: -17.6500431329, 1370: -6.30838416904, 1371: 18.0553766867, 1372: -8.29587322654, 1373: -72.8256188696, 1374: -15.7460326335, 1375: -12.1136371846, 1376: -11.7471158796, 1377: 1.9884260189, 1378: 15.8892843805, 1379: -6.3402730549, 1380: 8.98038170555, 1381: -4.70564083686, 1382: 13.2888316672, 1383: -23.8046710839, 1384: -10.9069371627, 1385: 15.5856263956, 1386: -0.777633725908, 1387: -5.56224159152, 1388: 4.51141168721, 1389: 11.5159387528, 1390: 3.32842248905, 1391: -27.5337473293, 1392: -0.301411459868, 1393: -28.1949711298, 1394: 25.3641182585, 1395: -27.2085320637, 1396: -23.8088875961, 1397: -0.164924744424, 1398: 1.20596800379, 1399: 5.73550687089, 1400: -3.22488641736, 1401: 2.19799822143, 1402: 7.34133941775, 1403: 2.47369352933, 1404: -10.9918726508, 1405: 0.866734298663, 1406: -7.92399192305, 1407: -5.56402650134, 1408: -8.48591128709, 1409: 16.4969318457, 1410: 7.63012531009, 1411: -1.8238615024, 1412: 3.34801997305, 1413: 4.960919325, 1414: 5.92278917882, 1415: -5.92380926807, 1416: 6.04576613833, 1417: -5.3812797048, 1418: -5.74371909405, 1419: -3.74155767559, 1420: -1.47510464446, 1421: -13.6683291211, 1422: 3.33793889977, 1423: -2.04292187822, 1424: -0.278257729324, 1425: 7.87182347216, 1426: -3.53858831867, 1427: -4.88625601695, 1428: -5.59747221493, 1429: -5.69084519552, 1430: -3.49839710116, 1431: 8.2027296823, 1432: -0.938116654356, 1433: -3.24192784651, 1434: 1.32310388393, 1435: 3.39143945967, 1436: 6.85382962568, 1437: 2.24362067853, 1438: -2.98568820807, 1439: -13.0200870962, 1440: -4.61014703258, 1441: -0.0189945026604, 1442: -7.12901654056, 1443: -0.954862172748, 1444: -3.3951201938, 1445: -1.76206281752, 1446: 2.30804934171, 1447: -1.13659233187, 1448: -4.07753196654, 1449: 2.96636579428, 1450: -1.92275037629, 1451: 0.788702713319, 1452: -1.11476045075, 1453: 9.79375358624, 1454: -0.97858970397, 1455: -3.37077500758, 1456: -8.85697511804, 1457: -5.29517692343, 1458: -1.56625560759, 1459: 5.79965102392, 1460: 10.7857252499, 1461: 1.04290293372, 1462: -1.01841250948, 1463: -3.30076042349, 1464: -0.660025151698, 1465: -4.66241871519, 1466: -1.27440169799, 1467: 1.09456481708, 1468: 19.8107948873, 1469: -1.9468651706, 1470: -6.01251369881, 1471: 3.39497164584, 1472: -7.3729201624, 1473: -2.14366075279, 1474: 6.23049938512, 1475: -0.299863075806, 1476: -8.10568377837, 1477: 3.18370686365, 1478: 9.67941873352, 1479: -1.06273921622, 1480: -3.14030207872, 1481: 2.76251595667, 1482: -1.28388150117, 1483: 6.80231586341, 1484: 5.10558472272, 1485: 2.62726908275, 1486: 6.26912960352, 1487: -5.29984435812, 1488: -1.25983421082, 1489: 10.0948300265, 1490: 0.00948284267178, 1491: -2.25544440477, 1492: 0.739367903312, 1493: 5.5741665138, 1494: -0.604241917258, 1495: 0.451349069972, 1496: 0.632966046071, 1497: 5.5980059392, 1498: -3.24893515551, 1499: 17.7635493863, 1500: -1.00007972557, 1501: -0.913012328489, 1502: 5.81714820813, 1503: -1.09204480857, 1504: -1.88074430125, 1505: 3.89785912901, 1506: 10.7135524715, 1507: 2.38995011632, 1508: -6.64973450536, 1509: -27.5359966215, 1510: -4.78785962533, 1511: 3.03290551681, 1512: -2.22839214861, 1513: -7.54640360839, 1514: -0.0407318830789, 1515: 19.621360192, 1516: -13.9072636744, 1517: -5.58461927637, 1518: -15.6878703385, 1519: -19.9984534108, 1520: -5.85226323478, 1521: 10.4283881474, 1522: -0.273779537141, 1523: -0.503836301782, 1524: 2.23611199285, 1525: 7.65159993411, 1526: -0.397874117334, 1527: -0.928820412773, 1528: -1.22649190593, 1529: -4.06434609548, 1530: -2.83624584612, 1531: -9.31629083129, 1532: -9.30804969203, 1533: -1.6685114, 1534: 7.90946352266, 1535: -4.1682794938, 1536: -3.80586669616, 1537: -3.94512057889, 1538: 11.5069200729, 1539: -4.94836553293, 1540: 1.47175561413, 1541: 2.02841045431, 1542: -0.438476460542, 1543: 0.110570090701, 1544: 7.79000528472, 1545: -0.231843163384, 1546: -1.01157326253, 1547: -9.56436742475, 1548: -0.0265067351608, 1549: -0.253275210584, 1550: -0.200159242811, 1551: -0.411535005519, 1552: -0.664197152032, 1553: 0.890694054789, 1554: -0.314282958856, 1555: -0.584391050361, 1556: -0.793289578529, 1557: -7.14893190661, 1558: 0.514056821595, 1559: -5.1005117399, 1560: 2.94994449316, 1561: 17.0819052466, 1562: -0.851767844326, 1563: -1.43947853025, 1564: -0.77286271269, 1565: 0.589848461953, 1566: 3.54696668515, 1567: -1.8307412793, 1568: -0.494675502754, 1569: 4.64444935451, 1570: -0.162256888645, 1571: 4.54998431115, 1572: -23.4149286076, 1573: 4.68996521864, 1574: -1.91184716689, 1575: -0.25322640847, 1576: -0.506361248891, 1577: -0.560149879463, 1578: -0.487375546061, 1579: -4.22129274506, 1580: 11.4497753426, 1581: 1.9825358526, 1582: -4.57304334687, 1583: -17.8201118978, 1584: -3.62104733335, 1585: 7.84869873893, 1586: 4.80612348771, 1587: -7.41326759848, 1588: 32.6434819963, 1589: 0.0231479024549, 1590: 2.81380473758, 1591: -4.65961766844, 1592: 4.68008181817, 1593: 5.15054988944, 1594: 8.90574140217, 1595: -0.245478188648, 1596: -1.44767064375, 1597: -0.302110506985, 1598: -19.4283959122, 1599: -0.318034984627, 1600: -0.303228725121, 1601: 7.95233925461, 1602: -0.87653343085, 1603: 3.69324090694, 1604: -0.18685364596, 1605: -0.109832263546, 1606: -0.660730677864, 1607: -0.22102464425, 1608: 4.66697996254, 1609: 4.26288190142, 1610: -7.49513341223, 1611: 0.146583642471, 1612: 0.668063309266, 1613: -27.8020751388, 1614: -4.3802904205, 1615: -3.23661100653, 1616: -12.7331424518, 1617: -0.489407058685, 1618: 3.19362040969, 1619: -0.445251833888, 1620: -0.618873340747, 1621: 4.53425184543, 1622: -0.329785092712, 1623: -4.75627348908, 1624: 4.94053214386, 1625: 0.266031491157, 1626: 2.5042729536, 1627: -0.920023348873, 1628: 1.14257188372, 1629: -5.65832899407, 1630: -2.67969227507, 1631: -0.689064784332, 1632: -0.840729276559, 1633: 6.40313949463, 1634: 10.6498750049, 1635: 3.65204086526, 1636: -0.894630437949, 1637: 3.95443961075, 1638: 9.01071272319, 1639: 0.666508132902, 1640: -7.69862225611, 1641: 3.7793603482, 1642: -8.72409885701, 1643: -7.60364713421, 1644: -6.75466661106, 1645: 7.76050896723, 1646: 3.73527408892, 1647: -11.9468080501, 1648: 4.06422335524, 1649: -8.34292084625, 1650: -0.325808139252, 1651: 3.85045859192, 1652: -3.20198308687, 1653: -5.46709582928, 1654: -1.22575885044, 1655: -1.05901100178, 1656: -4.83405208662, 1657: -1.14510569947, 1658: -0.984149599509, 1659: -1.1591550805, 1660: 0.146810427081, 1661: -0.625206501842, 1662: -1.94991405263, 1663: -0.437049789907, 1664: 4.73058903106, 1665: 9.7625933119, 1666: -9.12049305138, 1667: 6.36275006211, 1668: 0.540277480349, 1669: 11.5927593673, 1670: 7.85231526112, 1671: 4.60088969383, 1672: -7.49976814784, 1673: 7.0005921263, 1674: -0.399167446888, 1675: 6.34074018756, 1676: -1.19702455541, 1677: 8.62435754763, 1678: -12.7609282063, 1679: -0.446491443884, 1680: 25.6011713823, 1681: 7.81378582916, 1682: -10.2386424976, 1683: -2.01495152882, 1684: 6.89059240123, 1685: -1.22113308897, 1686: 0.0436403849876, 1687: 5.42736635783, 1688: 6.38714213437, 1689: -9.82396560324, 1690: 18.3183373187, 1691: 15.3372130659, 1692: -3.08646409002, 1693: -10.6233383734, 1694: 10.9873250459, 1695: 0.28621484936, 1696: 11.6941408056, 1697: -2.97340522406, 1698: -2.05592591702, 1699: -5.96430957965, 1700: 2.1934405767, 1701: 8.23547450742, 1702: -2.23754108505, 1703: -1.91515302782, 1704: -1.46037089839, 1705: -1.40136285128, 1706: -0.6379912278, 1707: 7.50184544308, 1708: -0.640272540089, 1709: -1.16462715833, 1710: -15.4441412421, 1711: -0.992059255018, 1712: -9.33228243003, 1713: -0.496177774971, 1714: -1.1425038928, 1715: 14.1244982803, 1716: -0.675729548431, 1717: 6.55043813352, 1718: -3.97570178902, 1719: -1.56887941673, 1720: 7.57557282233, 1721: 0.158855235723, 1722: 8.32848336165, 1723: -4.60998061772, 1724: -7.27209321798, 1725: -11.5000463373, 1726: 5.70703274965, 1727: -3.29377041293, 1728: 9.52013400665, 1729: -0.53145711257, 1730: -0.0112626855361, 1731: 0.22340202282, 1732: 0.0576725016643, 1733: -14.749978202, 1734: 7.41704231818, 1735: -0.454866039311, 1736: 3.85096013742, 1737: -0.643668482166, 1738: 2.56401025802, 1739: -9.10646891614, 1740: 7.88585792707, 1741: -0.133658689492, 1742: 2.29623954267, 1743: 8.04926468931, 1744: 8.61186641236, 1745: 17.7605406765, 1746: 0.48359479275, 1747: -3.3648859274, 1748: 19.3209404856, 1749: -10.9621859985, 1750: -30.665942549, 1751: -20.1568641016, 1752: -12.0535134489, 1753: 24.5898980332, 1754: 7.30381561512, 1755: -1.95195059616, 1756: -0.310952559336, 1757: 23.8523095971, 1758: -19.3190443415, 1759: -9.15793812887, 1760: -17.2435353748, 1761: 13.3439990062, 1762: -10.1003521166, 1763: 12.4408413801, 1764: 12.2939346266, 1765: 9.14762717776, 1766: 6.62750531974, 1767: 9.02544404548, 1768: 22.3620511195, 1769: 6.55904134456, 1770: 13.498197767, 1771: -8.30595590372, 1772: -8.55526692048, 1773: -6.45493458286, 1774: -51.0424714828, 1775: 14.3926202722, 1776: 8.8326628434, 1777: -9.62977985399, 1778: -15.2030316768, 1779: 1.86355842887, 1780: 8.16300082788, 1781: -2.19014285317, 1782: 10.8736276312, 1783: 17.4562876131, 1784: -12.0670945787, 1785: 6.98221476748, 1786: -7.50128861441, 1787: 58.3348810651, 1788: 2.74424251149, 1789: -6.4187984088, 1790: -1.89599052708, 1791: 6.41800565541, 1792: -5.18696418254, 1793: 1.70505732911, 1794: 18.4899143421, 1795: 1.92495251516, 1796: -4.31168517508, 1797: 17.2175919509, 1798: 2.14228166525, 1799: -10.4201273898, 1800: 5.93915146928, 1801: 5.27416182057, 1802: 9.02912959428, 1803: 14.1009770963, 1804: -59.1304388991, 1805: -8.20723537616, 1806: 44.6076718286, 1807: 2.06343839493, 1808: -2.95631499876, 1809: -1.20038956443, 1810: 11.2516638529, 1811: 0.921588762237, 1812: -13.9871690383, 1813: -5.95320228378, 1814: -6.61481978085, 1815: 3.22852843342, 1816: 3.27026034508, 1817: -11.6091658346, 1818: 2.54747781563, 1819: -9.64410297575, 1820: 1.34058005206, 1821: 0.111217243031, 1822: 3.52688599035, 1823: -10.4658176657, 1824: -6.74637895742, 1825: 21.3414664564, 1826: 4.67315436713, 1827: -14.1440472383, 1828: 3.29346291736, 1829: 18.7522415029, 1830: -1.36891780294, 1831: 0.147921525913, 1832: 6.80798240013, 1833: -8.78457464098, 1834: 2.13631412481, 1835: -4.32599968829, 1836: -27.8724523836, 1837: 0.614627652667, 1838: 15.7962695226, 1839: 2.18985837922, 1840: 1.78538058111, 1841: 3.54371131533, 1842: 2.70723857178, 1843: 2.71053683066, 1844: 2.60880031796, 1845: 2.50685953768, 1846: -7.33762481267, 1847: 4.63415606713, 1848: -16.1517985088, 1849: -5.80792078069, 1850: 3.01257384143, 1851: -12.1554002025, 1852: 3.67901245806, 1853: 2.66927370877, 1854: 4.71548490996, 1855: 1.61729367199, 1856: 1.05023618221, 1857: 8.43624364242, 1858: -19.359138493, 1859: 13.1553818319, 1860: -17.1546775117, 1861: -3.14886104787, 1862: -30.0294719182, 1863: 4.51589107467, 1864: 12.7201692079, 1865: 23.1248499496, 1866: 2.59452370157, 1867: 10.9460481306, 1868: -0.877791955298, 1869: -10.0781042952, 1870: 4.65649059647, 1871: 0.435036577904, 1872: 0.316443709895, 1873: 1.47033768979, 1874: 3.54141238034, 1875: 0.634118967555, 1876: 0.162763835679, 1877: 2.79034830833, 1878: -33.8276447718, 1879: 1.05208610158, 1880: -11.5805613316, 1881: 16.7752245955, 1882: -2.24137369123, 1883: -3.58688917385, 1884: 2.66934572073, 1885: -0.00773171937457, 1886: 15.8304000532, 1887: -13.3919035298, 1888: 31.300100581, 1889: 1.43238956547, 1890: -19.0934700408, 1891: 0.447650902394, 1892: 0.535945177211, 1893: 20.474865422, 1894: 0.630137578035, 1895: 2.47384013245, 1896: -6.07489001501, 1897: 0.584788927584, 1898: 0.549019268197, 1899: 0.609134925475, 1900: 0.476410321721, 1901: 3.25004258825, 1902: -23.9063615647, 1903: 0.27593931547, 1904: 0.607988857226, 1905: 0.0574908323019, 1906: -2.45127184989, 1907: 2.12088307699, 1908: 33.7889053105, 1909: -1.10094556619, 1910: 26.1597974171, 1911: 10.1554892356, 1912: 7.69804995671, 1913: -17.9087490104, 1914: -17.1517750241, 1915: -5.57060500993, 1916: -3.05513820352, 1917: 0.212499931298, 1918: -2.60099687842, 1919: 0.454826823759, 1920: -10.0211495912, 1921: 33.3693840007, 1922: -14.6609532748, 1923: -2.07885724419, 1924: 0.624343542452, 1925: 0.419088218119, 1926: 0.0428036839928, 1927: 0.554251075613, 1928: 2.30624604912, 1929: 36.6417767685, 1930: -1.64097275724, 1931: -22.8460730709, 1932: -15.2507195862, 1933: 11.9091846442, 1934: 0.860119225225, 1935: -3.13320950668, 1936: -20.5093516474, 1937: -23.200223685, 1938: 0.69316776932, 1939: 9.36148936461, 1940: -2.56173995532, 1941: 0.687300657904, 1942: 3.64698809489, 1943: 16.2046687057, 1944: 0.540105804038, 1945: 2.03338945558, 1946: 0.492596638693, 1947: -5.4691558849, 1948: 0.400134007715, 1949: 0.684376076836, 1950: -11.6386973878, 1951: 3.14451424183, 1952: 11.2267634792, 1953: 0.392044885832, 1954: 0.503300989119, 1955: 1.46243985073, 1956: 0.537863515887, 1957: 3.7469289488, 1958: 1.15825468228, 1959: 1.28001157668, 1960: 0.756912137887, 1961: 0.688096801927, 1962: -14.9750159198, 1963: -0.431107287638, 1964: -8.53432225442, 1965: 12.5341239074, 1966: 3.00546146388, 1967: 5.73322703366, 1968: 7.55603550458, 1969: 1.70790186961, 1970: -3.75769800403, 1971: 0.312193371496, 1972: 4.98866636127, 1973: 6.0350223667, 1974: -4.39225002285, 1975: 5.48172114735, 1976: 27.4130576172, 1977: 1.09653493172, 1978: -9.76057976143, 1979: 38.4395896234, 1980: 2.30781145629, 1981: 7.22348697244, 1982: -23.989006184, 1983: -3.15284913784, 1984: 2.61554921491, 1985: 0.162418650151, 1986: -55.7579948482, 1987: 25.9630138309, 1988: -14.3784586198, 1989: -8.4220266111, 1990: 13.7583934112, 1991: -9.71673291048, 1992: 2.2553467956, 1993: -11.749904639, 1994: -16.8343387188, 1995: 9.86070699955, 1996: 21.7150050114, 1997: -10.9614862465, 1998: 22.8396161905, 1999: 13.9951145154, 2000: -7.60855857042, 2001: 1.80590215808, 2002: -16.3341198837, 2003: -1.89565961253, 2004: 2.25397671876, 2005: 18.0764042833, 2006: -13.5533789558, 2007: 1.75733967872, 2008: 3.42810791441, 2009: 4.3076499558, 2010: 3.26759175141, 2011: -37.0303011176, 2012: 13.3285736746, 2013: -2.44447469929, 2014: 4.02633831502, 2015: -4.13778425476, 2016: 2.04427043232, 2017: -14.7225273644, 2018: -8.98649966815, 2019: -6.53486036274, 2020: 20.8919340941, 2021: -22.1928986419, 2022: -16.7498099861, 2023: -26.1099793414, 2024: -15.5633795289, 2025: 3.90873066298, 2026: 12.4854275457, 2027: 9.88518927466, 2028: 1.34405118752, 2029: -12.5906646707, 2030: -26.9199791259, 2031: 2.6828447892, 2032: 3.00473777047, 2033: -7.49774135142, 2034: 7.63953053021, 2035: -3.69538300392, 2036: 1.39214631261, 2037: -14.1273692997, 2038: -23.4511031644, 2039: -17.4252711099, 2040: -11.0668110255, 2041: 1.86635185279, 2042: 21.7829055271, 2043: 6.83600242493, 2044: 0.0429353521825, 2045: 7.73933599369, 2046: 1.75386310111, 2047: -12.9718271211, 2048: -1.77262639905, 2049: -10.281981277, 2050: 2.6848667006, 2051: 1.87561170987, 2052: -8.47819696285, 2053: 2.07419978411, 2054: 2.03943184888, 2055: 0.81557647891, 2056: -6.38167029709, 2057: 4.63118368065, 2058: -0.495815995897, 2059: 35.9973476795, 2060: 3.23192275384, 2061: 2.51492236801, 2062: 3.30802545408, 2063: 3.06866745483, 2064: -36.5633286266, 2065: -2.21360427441, 2066: 4.20328839852, 2067: 2.16328045239, 2068: 2.74675042902, 2069: 0.0977493307216, 2070: -14.6201218952, 2071: -15.6897109654, 2072: -1.06379563955, 2073: 17.7107072358, 2074: 28.2713985273, 2075: -2.89915485673, 2076: -9.1453173773, 2077: 6.3561879365, 2078: 0.908820321602, 2079: -6.72237858479, 2080: 0.932218230614, 2081: -1.42622268408, 2082: 1.27499060119, 2083: -30.1815608237, 2084: 3.25484151427, 2085: -7.28372979316, 2086: -6.98502935221, 2087: 1.42089638855, 2088: 16.3629841535, 2089: -6.89510455307, 2090: -11.3171716367, 2091: -17.2218879403, 2092: 36.8820978459, 2093: -19.406580908, 2094: -6.66467571431, 2095: 0.318045868461, 2096: -1.98992124631, 2097: -15.1149572206, 2098: 0.405148354597, 2099: 2.88938059185, 2100: 4.13893276309, 2101: 7.51710259836, 2102: -26.7024336377, 2103: -0.190882223337, 2104: 3.16814468314, 2105: 23.3579156636, 2106: -20.1952273319, 2107: -2.15908115596, 2108: 5.23722151313, 2109: 1.74947115352, 2110: -12.1683412836, 2111: -25.8826646939, 2112: -0.0690991088461, 2113: -3.17182242516, 2114: 7.47548705323, 2115: 5.94375627198, 2116: -6.021033841, 2117: 0.728019428408, 2118: 4.34343418709, 2119: 1.5477722747, 2120: -1.97546754124, 2121: 0.984954006254, 2122: -8.71576044481, 2123: -19.1205936465, 2124: 2.2767211393, 2125: 1.96807190846, 2126: -3.75905227755, 2127: -3.96748099087, 2128: -4.74437946358, 2129: -4.14029535931, 2130: 4.64234402154, 2131: -5.29061868305, 2132: 6.84427407759, 2133: 2.38165535309, 2134: 7.8122201859, 2135: -10.9159680708, 2136: 2.59040243875, 2137: -2.19518747002, 2138: 25.166877528, 2139: 10.1469054285, 2140: -3.05686131811, 2141: -1.56506032494, 2142: 2.19837878975, 2143: 3.37059129999, 2144: -2.75050527058, 2145: -6.03271016007, 2146: 0.690968254484, 2147: 2.51864906511, 2148: 0.382220224886, 2149: -1.26210664906, 2150: 1.68361504183, 2151: 7.77203540886, 2152: -5.34183756149, 2153: -23.4601044393, 2154: 1.77378431575, 2155: 14.2696695519, 2156: -5.08540449525, 2157: 2.72287738675, 2158: 1.0668822658, 2159: -6.23920787147, 2160: -1.09257461209, 2161: -15.6833169664, 2162: 2.58102763547, 2163: -4.27673013587, 2164: 2.69414289982, 2165: 2.56157783485, 2166: -4.3186190576, 2167: 2.50074155433, 2168: 3.45233431523, 2169: -1.01239788094, 2170: 1.81053176657, 2171: 2.17015837575, 2172: 3.27883780838, 2173: 5.59160021776, 2174: 14.1817011455, 2175: 4.12833887099, 2176: -18.3469407577, 2177: 2.62147286974, 2178: 7.45859934821, 2179: 2.3239921174, 2180: -4.08027388153, 2181: -5.50474675028, 2182: -13.3108080895, 2183: -4.80234032767, 2184: -8.52338217849, 2185: 3.70032524302, 2186: 2.29035132142, 2187: 3.69689807469, 2188: 2.63740662812, 2189: -0.776180494482, 2190: 2.94002004956, 2191: 3.59494977065, 2192: 2.59275962959, 2193: 1.23926348337, 2194: 2.52251489513, 2195: -9.6779336057, 2196: -0.43247021885, 2197: -6.55215473733, 2198: -1.56732707079, 2199: 0.987347443811, 2200: -5.11986405104, 2201: -3.55897150958, 2202: 1.10810615393, 2203: -4.63520087682, 2204: 5.80089890969, 2205: 3.08547466675, 2206: 8.82935997639, 2207: 10.6814984391, 2208: 5.07162600359, 2209: -0.538724712797, 2210: 8.53938483821, 2211: 6.09402272613, 2212: -3.7694344222, 2213: -7.92231195302, 2214: -9.49753826953, 2215: -12.4825637524, 2216: 10.1750165347, 2217: -1.50751646319, 2218: -5.54350639574, 2219: 0.973827201369, 2220: 0.375987856023, 2221: 0.676439589972, 2222: 1.75922660328, 2223: 0.610123888866, 2224: 0.535031269259, 2225: 1.84606309284, 2226: 0.512264716748, 2227: -2.60391163928, 2228: 1.65050115322, 2229: -2.93285981418, 2230: 15.5868950764, 2231: 0.309386362831, 2232: -7.55959354133, 2233: -2.3741070253, 2234: 7.18442610684, 2235: -10.5555697784, 2236: -1.77259267378, 2237: 3.8511937083, 2238: 0.295336234388, 2239: -15.5440758502, 2240: 0.326368973908, 2241: 0.463016595208, 2242: -6.95770887888, 2243: 0.492457666028, 2244: 1.86049317042, 2245: -10.2759885524, 2246: -0.152193899753, 2247: 0.433948252789, 2248: 0.714700730047, 2249: 0.119613575079, 2250: 1.7616092421, 2251: -4.31495062327, 2252: -0.132213041921, 2253: 0.948229518322, 2254: -1.85059982728, 2255: 2.44461078017, 2256: 0.811230474369, 2257: -10.2670127771, 2258: -2.40653698185, 2259: -12.7799434378, 2260: 9.4272269277, 2261: 3.81328627977, 2262: 8.84763696023, 2263: -2.90381062401, 2264: 0.758694427456, 2265: 30.1863517462, 2266: 0.524868303807, 2267: 27.097404589, 2268: 0.318103459577, 2269: 4.94928136567, 2270: 4.95316278634, 2271: -3.18900544957, 2272: 0.197952613113, 2273: 0.415261403961, 2274: 0.278681810715, 2275: 0.505949846503, 2276: 0.647340256659, 2277: -1.34823584971, 2278: -11.1876498384, 2279: -4.48235018964, 2280: 5.55264687275, 2281: 7.75832373325, 2282: -2.24030845835, 2283: -2.19920808982, 2284: -2.42714431278, 2285: -13.6774511009, 2286: -0.197885647563, 2287: 0.382255602965, 2288: -11.8119037543, 2289: 1.75124235691, 2290: 0.768022867086, 2291: 2.67980083816, 2292: 5.29208047064, 2293: 0.418748322975, 2294: 2.50350182197, 2295: 0.744635763855, 2296: 13.5902668438, 2297: 0.282196798942, 2298: 0.33248838184, 2299: -14.8928435552, 2300: 2.70716370374, 2301: 24.6924956206, 2302: 0.274147193317, 2303: 0.443480665805, 2304: 0.297726090814, 2305: 0.400558762535, 2306: 2.13568675673, 2307: -4.90601961169, 2308: 4.17794396199, 2309: 0.772965574734, 2310: 0.195846256749, 2311: 23.2169355706, 2312: -1.63915650545, 2313: -8.95606555274, 2314: 27.1887838806, 2315: -8.07096286708, 2316: 4.21734487526, 2317: -3.62606767275, 2318: 1.11579994522, 2319: -3.05527764838, 2320: 0.575189726964, 2321: -7.64848543594, 2322: -4.59773614382, 2323: -0.183559692147, 2324: -1.680578737, 2325: 26.6166574687, 2326: 0.501654420067, 2327: 0.393161046388, 2328: 6.0897502486, 2329: 2.4756053236, 2330: -7.19393230684, 2331: -0.492122771759, 2332: 24.140862967, 2333: 2.21747062616, 2334: 1.84282837605, 2335: -23.4258179664, 2336: -1.30128650892, 2337: 14.6983760126, 2338: 4.01927352526, 2339: -14.2522930132, 2340: -16.4369304831, 2341: 1.87822982013, 2342: -3.7015102484, 2343: -22.8704492289, 2344: 7.2313739718, 2345: -3.18803030542, 2346: 8.43409142633, 2347: 5.68953464728, 2348: -1.25123831245, 2349: 2.50612336619, 2350: -2.22486930255, 2351: 1.71737551748, 2352: -0.800181660319, 2353: 1.93415534971, 2354: 2.1040619773, 2355: 26.2888533798, 2356: 2.47407895801, 2357: -1.29984202502, 2358: 3.58024646831, 2359: 1.75238760334, 2360: 4.26663377711, 2361: 2.58328343737, 2362: -11.2650444382, 2363: 7.33414507519, 2364: 0.683515350969, 2365: 1.61953147777, 2366: 8.94996986525, 2367: 5.30306860288, 2368: 12.784962741, 2369: -19.2444587664, 2370: 9.09850084987, 2371: 4.21793796802, 2372: -11.7887952803, 2373: 0.236510615468, 2374: -9.4398563015, 2375: 1.38614622126, 2376: -18.5871141493, 2377: 4.9760149807, 2378: 4.79826369707, 2379: -8.72721194531, 2380: -0.690421517313, 2381: 2.0247111938, 2382: -5.50313572759, 2383: 2.41059953775, 2384: -10.6742511662, 2385: 2.02300591229, 2386: -18.5489353305, 2387: -10.142375014, 2388: -2.60829086739, 2389: -1.10911581091, 2390: 10.0481646333, 2391: -8.89378954292, 2392: -5.2574190423, 2393: -2.3367574661, 2394: -10.1546275723, 2395: 1.43282035199, 2396: -0.340750375779, 2397: -6.12296647588, 2398: -17.666567344, 2399: -18.0758643967, 2400: -0.944080132113, 2401: 9.14059142385, 2402: 2.5547489282, 2403: 2.72214890342, 2404: 2.88659934711, 2405: 2.04347008009, 2406: 2.39325818266, 2407: 2.39774988945, 2408: 30.1467879424, 2409: 2.64513576242, 2410: 8.93136030201, 2411: 2.66590905494, 2412: 2.52751035378, 2413: -5.9123015463, 2414: 2.19879781316, 2415: 1.71297020949, 2416: -11.6047986889, 2417: -5.30851413317, 2418: -6.49558403373, 2419: -4.74268172586, 2420: 7.38012875131, 2421: -2.88222692169, 2422: 1.51166858245, 2423: -26.1228462706, 2424: 1.17808744278, 2425: -7.1197355555, 2426: -9.32914661938, 2427: 0.689133332991, 2428: 0.257680037625, 2429: 0.0152973096759, 2430: 19.3993904175, 2431: 10.4888746892, 2432: 22.6750978313, 2433: 2.69188656951, 2434: -0.223417212738, 2435: -10.0910315829, 2436: 10.9696033539, 2437: 3.02314038888, 2438: 14.9868786324, 2439: -6.43091734014, 2440: -31.5088222278, 2441: 0.209803918458, 2442: -17.860116552, 2443: 8.80119669249, 2444: 0.110236822028, 2445: -0.643980786798, 2446: 1.25977077777, 2447: -11.5617575917, 2448: -14.9079438609, 2449: -11.2100665265, 2450: 0.427898365247, 2451: 28.2759538793, 2452: 4.0107868953, 2453: -4.13125804338, 2454: -8.55935905163, 2455: -2.79688386868, 2456: -6.59743651342, 2457: -5.70639110333, 2458: 6.04557430932, 2459: -1.63583961235, 2460: 4.94834388555, 2461: -1.38905078451, 2462: -36.3800285105, 2463: -15.1034136644, 2464: -16.9314379522, 2465: -8.16588802488, 2466: 0.414662672508, 2467: 13.8473730684, 2468: 5.72010624855, 2469: -9.48086227355, 2470: 0.877306236628, 2471: 11.2322999037, 2472: 9.22500720237, 2473: -10.4665904224, 2474: -17.7337809308, 2475: -21.2811526155, 2476: 2.82557838631, 2477: -2.53819015448, 2478: -15.6548347824, 2479: 2.359311928, 2480: 1.62179283945, 2481: -7.75950665688, 2482: -7.65194841903, 2483: -2.55222315207, 2484: -21.0481907398, 2485: -4.06914754711, 2486: -5.05554164807, 2487: 8.81301853503, 2488: -2.58114829526, 2489: -11.8094878986, 2490: -1.01943172142, 2491: 0.042137754742, 2492: -23.2990383604, 2493: 4.98013637778, 2494: 7.24946939225, 2495: -5.68187083895, 2496: 0.792905590418, 2497: -6.85435681478, 2498: 2.92598517123, 2499: 0.92523440075, 2500: -19.5252324851, 2501: 5.26539525558, 2502: 6.27347030564, 2503: 22.0244444719, 2504: -16.8482717557, 2505: -9.39612889522, 2506: -15.714832084, 2507: -3.58501033691, 2508: -11.9208936256, 2509: 0.972248750301, 2510: 5.50668650385, 2511: 0.836032175854, 2512: 15.8528019058, 2513: 1.29975016836, 2514: 1.26381423087, 2515: 8.50209119516, 2516: 0.804705455855, 2517: 6.51264668229, 2518: 0.254143470755, 2519: -0.684889178631, 2520: -1.09379160705, 2521: 4.23760852454, 2522: 13.5255264867, 2523: -20.877212636, 2524: 6.91737168988, 2525: -0.96635853069, 2526: 0.891992693762, 2527: -4.16233813993, 2528: -13.8847459785, 2529: -23.1015387398, 2530: -3.49956055332, 2531: 8.69717058393, 2532: 0.855913570887, 2533: 15.2587695216, 2534: 1.26742583397, 2535: 0.963117274096, 2536: -23.4511908348, 2537: 1.0309040575, 2538: -1.51485963715, 2539: 2.90655979481, 2540: -8.11809407186, 2541: 1.15539920112, 2542: 1.27084400133, 2543: 0.866164196794, 2544: 0.826375779054, 2545: 0.882470312865, 2546: 0.92517819718, 2547: 1.39507571307, 2548: -0.274822112953, 2549: -63.3993194354, 2550: 1.01272835409, 2551: 0.897873301444, 2552: 1.06541930174, 2553: -1.2594300494, 2554: 1.40917237702, 2555: -4.71924253291, 2556: -17.2399305481, 2557: 4.61345517626, 2558: 8.5651702606, 2559: -5.02332820697, 2560: 5.99682928152, 2561: -9.53702391013, 2562: -26.0871851557, 2563: -12.6108666892, 2564: 4.71848019192, 2565: 7.45678375469, 2566: 15.1340965382, 2567: 7.48885634322, 2568: 7.14229274223, 2569: 0.0497783495781, 2570: 0.459529402219, 2571: 1.64374458759, 2572: -25.9252218819, 2573: 0.0715741195773, 2574: 0.953157251647, 2575: -0.972473414743, 2576: -10.2939497237, 2577: -0.485372048138, 2578: -0.569201349101, 2579: 8.33571053044, 2580: -2.81253750461, 2581: -3.77779569323, 2582: -1.92397247403, 2583: 6.65193130367, 2584: 18.8920917987, 2585: -4.66508774293, 2586: 3.71436718302, 2587: 0.289216816371, 2588: 9.94882214083, 2589: -0.634414916824, 2590: 0.311236702147, 2591: -0.537037807519, 2592: 0.181193632355, 2593: 0.858117152779, 2594: -3.57769846846, 2595: 0.317521315166, 2596: 0.222799975843, 2597: 0.125535703297, 2598: 0.150472212965, 2599: -1.34448950471, 2600: -14.3755318946, 2601: 2.04032486511, 2602: 0.328801058499, 2603: -0.251738780165, 2604: 4.38107054398, 2605: -2.40559511207, 2606: 5.9204007606, 2607: -0.0307133488316, 2608: 16.3576938887, 2609: 47.016184187, 2610: 0.528449171337, 2611: -4.81882324941, 2612: -1.06232643298, 2613: -2.49284891221, 2614: -13.6086770411, 2615: 0.117714835147, 2616: -25.3354265864, 2617: 0.0888695790745, 2618: 16.1614432773, 2619: -1.37817780767, 2620: 3.47011005905, 2621: -2.38016471918, 2622: 0.293027178169, 2623: 0.196532815547, 2624: -0.0554909191668, 2625: 0.287250478552, 2626: -0.715974343699, 2627: -4.52562803731, 2628: 5.70922707565, 2629: 30.7909405734, 2630: -7.26893203305, 2631: 8.05097114558, 2632: -0.960849440767, 2633: -1.13863066147, 2634: 6.18120725914, 2635: 23.1948120365, 2636: 0.219596023154, 2637: 1.01735286981, 2638: -1.43752608179, 2639: 7.61015543251, 2640: -6.05608566235, 2641: -5.50113371508, 2642: 0.201737605785, 2643: 0.879495653313, 2644: 0.140467145595, 2645: 10.2053828891, 2646: 0.270143512222, 2647: 0.193761992449, 2648: -9.77770449026, 2649: 1.08111603037, 2650: -35.3540025253, 2651: 0.296817230104, 2652: 0.112067476983, 2653: 5.83459925004, 2654: 0.274999076847, 2655: 0.797709530945, 2656: -2.38257430943, 2657: -5.08190252771, 2658: 1.2411272123, 2659: -0.788694551653, 2660: -12.748258145, 2661: -3.1082638419, 2662: 6.11386395133, 2663: -26.8463245585, 2664: 12.9783999929, 2665: -3.8362484698, 2666: -0.913401935808, 2667: 0.948463737366, 2668: 3.84227698675, 2669: -0.014483058611, 2670: 23.8358631794, 2671: 6.3495048519, 2672: 11.9721758219, 2673: -1.7173922663, 2674: -5.2688194498, 2675: 3.12085236695, 2676: -2.34559265121, 2677: -4.3546533347, 2678: 2.36631639889, 2679: 12.0201644383, 2680: -8.54422331344, 2681: 24.866658949, 2682: 0.699011807158, 2683: 1.12057120847, 2684: 3.30584945169, 2685: -2.38176053144, 2686: -7.28694407133, 2687: -0.355569210092, 2688: 9.28277088589, 2689: -1.01158671285, 2690: 7.88722215385, 2691: -2.89347351543, 2692: 23.7246394152, 2693: -29.4241492886, 2694: 9.24628359612, 2695: -9.76582052655, 2696: 0.74038606053, 2697: -0.359858029407, 2698: 8.55286996609, 2699: -0.670088617517, 2700: 3.50831870706, 2701: 0.871148734221, 2702: 0.874649503534, 2703: -1.51407517644, 2704: -2.93496689555, 2705: 1.12842866981, 2706: 1.07335739582, 2707: -4.80295922451, 2708: -1.45844090528, 2709: -4.85495546365, 2710: -20.491685569, 2711: -13.4572491853, 2712: -14.539786926, 2713: -1.00777090923, 2714: -0.0427078550469, 2715: -13.7318822996, 2716: -1.83619247525, 2717: 7.51105330035, 2718: -8.64660228978, 2719: -28.7844717288, 2720: -6.76926468456, 2721: -8.15144414879, 2722: -4.68828089431, 2723: 7.60070815246, 2724: -0.287555074242, 2725: 33.7740421178, 2726: -0.102847153202, 2727: 21.9217556071, 2728: 11.03669612, 2729: 2.85928053199, 2730: 0.933075150331, 2731: 19.7595414856, 2732: 1.65212150679, 2733: -5.99009258662, 2734: -0.952683681908, 2735: 3.33704607776, 2736: 0.24533780332, 2737: -12.0647202463, 2738: 24.4742431556, 2739: 14.5390691499, 2740: 13.1964518274, 2741: 6.12604829757, 2742: 2.68834303304, 2743: 1.4855686464, 2744: 0.375993175364, 2745: 9.96693532152, 2746: -4.90723207915, 2747: -7.43030756161, 2748: 9.85853663922, 2749: -1.48919935716, 2750: -8.45300073468, 2751: 0.953898528616, 2752: 0.880597093596, 2753: 0.831019158374, 2754: -3.8609324821, 2755: -0.0670701640537, 2756: 0.849164431799, 2757: 2.23708242167, 2758: 1.11393756827, 2759: 3.37734644296, 2760: 1.02727110008, 2761: 0.884424028706, 2762: 6.68473436534, 2763: 0.902074437333, 2764: 0.835409100159, 2765: 13.0532283131, 2766: 16.0618197109, 2767: -34.922147339, 2768: -6.53094196487, 2769: 4.73052311453, 2770: -3.49365143679, 2771: 18.5048360772, 2772: -26.8583221667, 2773: 6.02639424707, 2774: -10.7374121625, 2775: -5.0541573802, 2776: 0.698559016099, 2777: -5.50088386637, 2778: 2.58876109425, 2779: 13.6547415736, 2780: 1.37292759213, 2781: 30.4359784575, 2782: 1.26558934284, 2783: 3.01651077111, 2784: 3.68676928761, 2785: 17.5399922445, 2786: 17.9106484791, 2787: 22.8762155305, 2788: -4.0431568855, 2789: -50.6052132609, 2790: -40.8504382737, 2791: 10.5203909702, 2792: 5.30551272135, 2793: 0.973216961022, 2794: -17.4351682319, 2795: -9.51267281445, 2796: 28.7604319871, 2797: 31.8635520904, 2798: 28.2286185858, 2799: -13.7018542666, 2800: 12.5215412263, 2801: -67.097618688, 2802: 12.4397194586, 2803: -70.2627852231, 2804: -10.721701057, 2805: 1.1148100663, 2806: -27.5786221523, 2807: 3.98858099824, 2808: -10.9393692844, 2809: -59.0309114724, 2810: -8.05420829377, 2811: 16.1974046281, 2812: -17.1735065363, 2813: 11.720018288, 2814: -15.3916569162, 2815: -2.86357844803, 2816: -33.5740734584, 2817: 13.2133601019, 2818: 4.79413414244, 2819: 2.05695990023, 2820: 51.9649786226, 2821: 20.7513012055, 2822: -21.8121435003, 2823: -23.3957448595, 2824: 14.1926089396, 2825: -3.06023526172, 2826: 17.6054091442, 2827: -16.2525025423, 2828: -27.7263472789, 2829: -21.9510036384, 2830: -64.9313330884, 2831: -10.9093168145, 2832: -54.4146458132, 2833: -11.0368038779, 2834: 16.7775346668, 2835: 38.5526078832, 2836: 17.3212447074, 2837: 1.90088016533, 2838: 24.6444407453, 2839: -24.6884607949, 2840: -19.3155226019, 2841: 31.3520680444, 2842: 11.5425425296, 2843: 34.9038134483, 2844: -28.8125608982, 2845: 4.99440902852, 2846: 0.832658930686, 2847: 7.85728302453, 2848: 15.4417468854, 2849: 67.3860772071, 2850: 6.24499127216, 2851: -4.99999159585, 2852: 25.8299056994, 2853: -37.2639810567, 2854: -24.93992872, 2855: -12.0122662251, 2856: 47.485352524, 2857: -35.9434672302, 2858: 12.2856684513, 2859: 27.9705498383, 2860: 6.39884607655, 2861: -40.3106014565, 2862: 6.45055787547, 2863: 20.5190831979, 2864: -30.7541133406, 2865: -6.80924586599, 2866: 43.5183218103, 2867: 2.9652364904, 2868: 5.51092452332, 2869: 5.22163247063, 2870: -4.40069001106, 2871: -8.85897455993, 2872: 1.80563198617, 2873: 22.5996256533, 2874: -9.07561338627, 2875: 5.06678071457, 2876: -1.72310817153, 2877: 17.3235270552, 2878: -16.2768197653, 2879: -28.5003462063, 2880: -63.9507227925, 2881: 9.00833929686, 2882: -37.3196526895, 2883: -13.2159769933, 2884: 9.04738405398, 2885: 4.03505851155, 2886: 6.22226261558, 2887: 4.00868480023, 2888: 6.10072839084, 2889: -1.05982801249, 2890: 5.43891019254, 2891: -1.04045907558, 2892: 5.99408800362, 2893: -99.9999915658, 2894: 5.76689710999, 2895: -21.7823029908, 2896: 8.99021318204, 2897: -0.797829474245, 2898: -32.7424114916, 2899: 3.75648205382, 2900: -7.99150451303, 2901: 10.0587559562, 2902: 0.0106407191746, 2903: 4.9172964862, 2904: 10.6199164882, 2905: -1.22062730883, 2906: -95.1823412965, 2907: 11.0402016848, 2908: -11.4785283098, 2909: 38.9760270905, 2910: -6.074796851, 2911: 51.5738974471, 2912: -37.1518607026, 2913: -26.6639045911, 2914: 12.5823703089, 2915: 14.3344356559, 2916: 2.7255476718, 2917: -30.697786099, 2918: 1.0684999522, 2919: -0.0258065385547, 2920: 2.89217525024, 2921: -2.62141218697, 2922: 1.20490705649, 2923: 5.36924091696, 2924: 4.45011016189, 2925: 42.8070284371, 2926: 4.52283255898, 2927: -2.23481368932, 2928: -88.9422523587, 2929: 3.70273090777, 2930: -8.89182847267, 2931: 1.00628341417, 2932: -17.1234358245, 2933: -2.77135394522, 2934: -8.55365945921, 2935: -55.7205954167, 2936: 3.8546494376, 2937: 5.5543561301, 2938: 0.903242366603, 2939: 0.992536116667, 2940: -24.0192861574, 2941: 1.37577959983, 2942: -2.35994697336, 2943: -29.0524669332, 2944: -0.667975287289, 2945: 1.09722657237, 2946: 1.06175977234, 2947: 1.07166108789, 2948: 3.19741929351, 2949: -5.71859885075, 2950: -4.38276155517, 2951: 1.48155035903, 2952: 0.573117317656, 2953: 14.721550065, 2954: 13.4263185675, 2955: 19.5137065929, 2956: -12.5315324148, 2957: 23.5257550428, 2958: 6.22789106181, 2959: 27.5981029378, 2960: 7.53481852956, 2961: -0.672269361322, 2962: 2.9278575883, 2963: -6.16085315642, 2964: 0.485716051742, 2965: 44.4656222075, 2966: 0.886093767405, 2967: -15.3204797045, 2968: 73.3666732735, 2969: 84.0456533195, 2970: -2.0315821718, 2971: 1.1231744343, 2972: 0.910619182699, 2973: 0.153152704818, 2974: 1.78624723702, 2975: 3.4481369086, 2976: -15.2380250681, 2977: 14.2214002225, 2978: -54.0282713294, 2979: -15.6472022014, 2980: 8.37052262066, 2981: -17.9773864743, 2982: 13.6664747214, 2983: 4.64781471931, 2984: -74.0737349516, 2985: 1.0294866265, 2986: -29.1134255986, 2987: 11.8846338347, 2988: 11.035384936, 2989: 5.91094579864, 2990: 45.4968774185, 2991: 1.08039370795, 2992: 4.09279883441, 2993: 0.857424025277, 2994: -21.198646647, 2995: 1.27071452736, 2996: 0.873510545421, 2997: 16.3727792671, 2998: 5.98084740097, 2999: 50.671580848, 3000: 0.98869127498, 3001: 0.950307893651, 3002: 3.64017882352, 3003: 1.02538362977, 3004: 19.1134306068, 3005: -28.2800652705, 3006: 8.51427596036, 3007: 2.56172123826, 3008: 0.714009095991, 3009: -10.671416342, 3010: -21.1272979287, 3011: 18.552573243, 3012: 7.01865451551, 3013: 28.0033806231, 3014: 8.056431973, 3015: 2.45999731296, 3016: 5.54819324192, 3017: -9.36726748466, 3018: 0.591590808297, 3019: -52.322431066, 3020: 5.39958656072, 3021: -14.5348704356, 3022: 7.11014832479, 3023: -50.8209983653, 3024: 3.36854756651, 3025: 1.61665284927, 3026: 21.1009564589, 3027: 6.21350753571, 3028: -5.50800999761, 3029: 15.8451909775, 3030: 2.32319999644, 3031: 5.77732082095, 3032: 5.20818993081, 3033: 11.3183685939, 3034: -71.6662408585, 3035: 28.1695874421, 3036: -43.3898093021, 3037: 66.5104374864, 3038: 12.6261720725, 3039: 19.4036650188, 3040: 26.4802777857, 3041: 14.4642525418, 3042: 49.0404414283, 3043: -28.1368520583, 3044: 24.5863017, 3045: -10.0744790232, 3046: -11.2024503091, 3047: 3.48060879644, 3048: 8.74420084987, 3049: 41.8372355144, 3050: -1.7003221029, 3051: -2.48023379319, 3052: 16.262344675, 3053: 19.2606823742, 3054: 5.79398729867, 3055: 2.24293218259, 3056: 2.27195931729, 3057: 3.15966069313, 3058: -100.0, 3059: 84.1563124007, 3060: -5.68367126084, 3061: 8.33181922821, 3062: 13.3872171109, 3063: 12.7241881126, 3064: 4.02644666414, 3065: -7.42969325836, 3066: -100.0, 3067: 100.0, 3068: 20.4607660188, 3069: -28.2206677167, 3070: 30.9215372311, 3071: -6.60384634343, 3072: -5.01368543159, 3073: 25.9887580685, 3074: -1.63630591403, 3075: -12.6635145306, 3076: 100.0, 3077: -62.0473926524, 3078: 72.0496317158, 3079: 6.78442318035, 3080: -3.92885449014, 3081: 13.6353819171, 3082: -13.6083427447, 3083: -16.7691774368, 3084: -1.75339689069, 3085: 8.40613109074, 3086: 29.1338394297, 3087: -2.67410448553, 3088: -38.3068867368, 3089: -85.2825860337, 3090: -15.1703159415, 3091: 3.95525290705, 3092: -17.6644464061, 3093: 11.8378799319, 3094: 17.563853393, 3095: -1.62035283485, 3096: 40.53533747, 3097: 10.1136777592, 3098: 4.08545131747, 3099: -12.2227383396, 3100: 4.06522340553, 3101: 4.11546939427, 3102: 5.94302733365, 3103: -24.6391277036, 3104: 2.05090264205, 3105: 8.24133906516, 3106: 29.133275342, 3107: 6.022461065, 3108: -87.1606778535, 3109: 5.83429985592, 3110: 5.9281843076, 3111: 8.16876605263, 3112: 3.50607481358, 3113: 24.5086936707, 3114: -29.7232742737, 3115: 34.5406554164, 3116: 3.8103241356, 3117: 6.12327331009, 3118: -6.57199107545, 3119: 12.5271332749, 3120: -25.1620599766, 3121: 0.434097698434, 3122: 3.41740248273, 3123: 3.63635111496, 3124: -9.88442508516, 3125: 7.31816206884, 3126: -0.858353465639, 3127: -37.8641985064, 3128: -8.48163089023, 3129: 53.7831030517, 3130: -14.0701844789, 3131: 5.96237806333, 3132: -59.9703802548, 3133: 3.88129039951, 3134: -22.0624346614, 3135: -8.89338672494, 3136: 15.8636577665, 3137: 18.3504682903, 3138: -70.3996160867, 3139: -47.8745004803, 3140: 35.6872698483, 3141: -56.6154951932, 3142: 0.988368981718, 3143: 57.8751482045, 3144: -21.9223289171, 3145: -5.83478650771, 3146: -28.041994591, 3147: 5.1920328493, 3148: 3.27796473897, 3149: -48.9322531174, 3150: 29.073834456, 3151: -5.4293433983, 3152: -5.99956368, 3153: 17.7058149684, 3154: -15.1605157059, 3155: -14.6615292527, 3156: -4.8286921521, 3157: 39.2029134823, 3158: 31.8490017016, 3159: 2.55134547776, 3160: -27.9445545009, 3161: -39.3226596704, 3162: -33.3353326463, 3163: 32.937860282, 3164: 3.92806762368, 3165: 9.02873781021, 3166: -36.2475025376, 3167: 17.8770825106, 3168: 11.374894434, 3169: 9.70195734649, 3170: -43.1762756896, 3171: 11.2072164401, 3172: 14.6477443471, 3173: 67.9620084746, 3174: 5.43227448595, 3175: 1.77944569564, 3176: 47.1752323569, 3177: -12.0629734007, 3178: 20.3178243539, 3179: 21.2624427602, 3180: 18.1903936711, 3181: 15.5589994825, 3182: -13.6159883361, 3183: -17.6345456572, 3184: -15.6309525179, 3185: -6.87100935932, 3186: 7.04465464946, 3187: -9.54061539569, 3188: 8.54401170542, 3189: 7.70405076914, 3190: 28.6229154288, 3191: 17.1507745024, 3192: 14.1778738099, 3193: -8.37369822605, 3194: 6.41483987842, 3195: -34.0684866157, 3196: 30.9266819785, 3197: 6.09942725526, 3198: 81.2607277434, 3199: 7.18205836667, 3200: 12.8986987177, 3201: -10.2433034446, 3202: -38.8050200044, 3203: 20.1450589204, 3204: -28.1366138194, 3205: 27.2315326695, 3206: 26.2410466527, 3207: 5.27744202626, 3208: 54.6888446796, 3209: 6.22676840442, 3210: 22.6865169897, 3211: 5.58803483713, 3212: 7.21378786099, 3213: 7.74735052884, 3214: 5.94990735611, 3215: 15.6249311452, 3216: 5.21055361493, 3217: -5.69416366224, 3218: 5.00055834836, 3219: 5.10160826393, 3220: 11.5154392703, 3221: 73.9556675199, 3222: 22.0380585453, 3223: -100.0, 3224: 6.60663509093, 3225: 13.3686510293, 3226: 30.823036385, 3227: 4.34773688527, 3228: 20.3708294629, 3229: -70.93987409, 3230: 6.56923588605, 3231: -20.5380147681, 3232: -16.2077972864, 3233: 6.25660403141, 3234: -55.564640645, 3235: 3.90506250273, 3236: 5.8088478541, 3237: 6.88849298458, 3238: 2.45012971336, 3239: 6.77020929754, 3240: -11.2098803609, 3241: 16.3358712539, 3242: -93.7589034752, 3243: 1.97401244462, 3244: -68.5251783042, 3245: 6.00274212025, 3246: 7.14384224443, 3247: 3.51505485409, 3248: 6.52504776041, 3249: 6.33811156162, 3250: 6.96313326942, 3251: -33.9140665067, 3252: 2.84765227877, 3253: -49.147046884, 3254: -18.533138543, 3255: 45.8624013258, 3256: 63.1015829084, 3257: 26.9757354934, 3258: 0.91998052749, 3259: -23.5507855228, 3260: -5.20640417534, 3261: 57.0166848824, 3262: -52.2621860383, 3263: -100.0, 3264: -19.3246475747, 3265: -20.8039635471, 3266: -41.8094986888, 3267: 1.3133682445, 3268: 2.04180133733, 3269: 5.48193773023, 3270: -24.746864836, 3271: 1.31763754788, 3272: 6.7879586446, 3273: 1.98069431049, 3274: 41.4936859775, 3275: -1.24061771646, 3276: -0.394247838298, 3277: 47.9830919343, 3278: -5.43809326503, 3279: 9.99473590297, 3280: 2.1931638003, 3281: -10.0336167754, 3282: -14.4112691091, 3283: 21.315378797, 3284: 0.323282827676, 3285: 0.704819412164, 3286: -21.5897291564, 3287: 0.842283816193, 3288: 0.716772106857, 3289: 7.83408755645, 3290: 1.1305800758, 3291: 6.25692056578, 3292: 37.6409083457, 3293: 0.424236774056, 3294: 0.769177562927, 3295: 1.5304028139, 3296: 1.45446145427, 3297: 6.32290351508, 3298: 39.3196257185, 3299: -0.858873927725, 3300: 2.04233376813, 3301: 0.499522524212, 3302: 3.40841108181, 3303: -11.4531816391, 3304: 66.6405354619, 3305: -6.95415535152, 3306: -7.24789255282, 3307: -54.1790685178, 3308: -8.33893743376, 3309: 2.94770620404, 3310: -11.0214933045, 3311: -11.8921671248, 3312: -5.06756875376, 3313: 1.38837534641, 3314: 28.7223072452, 3315: 1.35774493569, 3316: -32.2090473788, 3317: 45.341649043, 3318: -34.9300669123, 3319: -0.96723696999, 3320: 1.33741381403, 3321: 0.911371281938, 3322: 0.761568960895, 3323: 1.57341146171, 3324: -5.92754038481, 3325: 18.3022219043, 3326: 10.2588574519, 3327: -41.7883326722, 3328: 35.2675512003, 3329: 27.046682961, 3330: -13.4642115672, 3331: 7.0229459412, 3332: -14.6119673152, 3333: -58.3833620496, 3334: 1.67796135868, 3335: 10.6585721204, 3336: -4.32709156684, 3337: 1.95360296742, 3338: 3.30900409658, 3339: -28.3000031976, 3340: 1.18699924365, 3341: 4.57718440378, 3342: 1.38698602795, 3343: 24.040523859, 3344: 1.15965197095, 3345: 1.34423508628, 3346: 1.03023898391, 3347: 6.24323922709, 3348: 100.0, 3349: 1.22589446531, 3350: 1.25819519129, 3351: 3.78174468663, 3352: 1.44727042414, 3353: 8.287622417, 3354: -8.93927252159, 3355: -13.5018994291, 3356: 1.87607992947, 3357: 1.21566688998, 3358: 36.9405740258, 3359: 1.5732238225, 3360: 5.75991380122, 3361: 58.4495863779, 3362: 4.8044624858, 3363: -9.75419984841, 3364: -1.94968300597, 3365: 5.03701723866, 3366: 8.84034002058, 3367: 1.10781946281, 3368: -33.6826359291, 3369: 1.38699211465, 3370: 5.43119984088, 3371: 12.8746534035, 3372: 10.5382338811, 3373: 2.15521949252, 3374: 0.578911106994, 3375: -28.7799350317, 3376: 1.39774643346, 3377: 15.4366301047, 3378: 22.4386539323, 3379: -38.1885190801, 3380: 2.58214903406, 3381: 6.80241909259, 3382: 12.1576994006, 3383: -23.6363906415, 3384: -19.8931435549, 3385: 15.9432216167, 3386: -38.9774542256, 3387: -32.5984166723, 3388: 6.02785592642, 3389: -18.4345539045, 3390: -46.0088016956, 3391: 26.506602615, 3392: 2.80921556748, 3393: 42.7017296512, 3394: -44.9893464793, 3395: -34.9203429119, 3396: 12.6744299864, 3397: 28.6171380763, 3398: -50.2596366045, 3399: 5.83128479975, 3400: 6.25057289033, 3401: 39.5331857335, 3402: 6.83821535447, 3403: 21.3709683038, 3404: 0.0843207957406, 3405: 0.523614858576, 3406: 6.5025907298, 3407: -1.10158559876, 3408: -33.9030071465, 3409: -39.1221680347, 3410: 34.3426639735, 3411: 2.19988026352, 3412: -4.12918785107, 3413: -3.79375681837, 3414: 5.87851691314, 3415: 23.8176725279, 3416: -28.098228188, 3417: -0.478047969901, 3418: 33.5152484009, 3419: -45.4303714752, 3420: -25.8218556978, 3421: -12.4096972014, 3422: -5.26036214344, 3423: -13.7126719801, 3424: -18.3770761868, 3425: -47.3803329161, 3426: -35.0479760057, 3427: -5.24718099275, 3428: 4.54942913089, 3429: -1.81838129986, 3430: 8.7041224652, 3431: -12.272363256, 3432: -5.87359982112, 3433: 10.6689013868, 3434: 48.023791824, 3435: -49.9576599082, 3436: -38.6967410727, 3437: 52.0996195337, 3438: 4.1236302235, 3439: 24.2253182999, 3440: -5.4751182911, 3441: 3.47935953812, 3442: -1.23817788211, 3443: -14.2180958587, 3444: 3.00650225529, 3445: 34.8351657341, 3446: 52.6478942289, 3447: 5.7042197556, 3448: -25.2846562968, 3449: 4.64201710332, 3450: 4.74479029201, 3451: -6.77477186164, 3452: 21.2906451771, 3453: 4.89073143397, 3454: 5.54449915648, 3455: -50.0861344059, 3456: 6.29436400935, 3457: -10.8350155984, 3458: 6.36280886713, 3459: 6.59560373732, 3460: -12.0354178184, 3461: 5.49468847763, 3462: 9.21136752225, 3463: -15.9620137582, 3464: -28.0536959407, 3465: 12.37266922, 3466: 9.14309202754, 3467: -44.8342528912, 3468: -33.7653715241, 3469: -80.4643188682, 3470: 17.1477212752, 3471: 13.4139468939, 3472: 16.2220909221, 3473: -33.6781171949, 3474: -9.92758730069, 3475: 10.8409971852, 3476: -3.50724534137, 3477: -22.1317708957, 3478: -21.0101911024, 3479: -37.1519917629, 3480: 5.83125245505, 3481: 7.24989359981, 3482: -4.48239469524, 3483: 10.4173572165, 3484: 18.6291388504, 3485: -40.6667350298, 3486: -20.2986451925, 3487: -46.7923309417, 3488: -14.8978669447, 3489: -43.7146432892, 3490: -7.59441267301, 3491: -0.767885067611, 3492: -5.00101800147, 3493: 1.59413308329, 3494: 0.566084228093, 3495: -1.12375416196, 3496: -0.582976226678, 3497: -1.2216285369, 3498: -0.590393563257, 3499: -1.47261473852, 3500: -0.459739757166, 3501: -0.471150500169}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.17997714671, 2: 8.26170291191, 3: -5.57898067395, 4: -3.33099038996, 5: -10.8244181303, 6: -8.69386948556, 7: 7.37529373527, 8: -2.76230091308, 9: 0.857719669448, 10: -6.43535606682, 11: -4.89639743632, 12: 10.7983297756, 13: -9.22103008095, 14: -8.73850206942, 15: -4.86881881579, 16: 5.15625487304, 17: -12.7110900876, 18: 3.56089268025, 19: -13.1447660958, 20: -13.0788058064, 21: 7.36879553652, 22: -6.89051231548, 23: 9.04908307673, 24: 4.64244755863, 25: -7.93900447214, 26: 5.73324305367, 27: -11.4451215324, 28: 4.00488985543, 29: 1.09103270943, 30: 0.0624949624532, 31: -1.12640706995, 32: 8.15806252651, 33: -1.27924137631, 34: 2.4392469517, 35: 5.00986426093, 36: 4.66512975585, 37: 2.34316157433, 38: -2.07179082294, 39: 8.59211300094, 40: 5.02439233449, 41: -12.6959544981, 42: -8.52139233742, 43: 2.71661407493, 44: -7.79042137165, 45: -0.666942934301, 46: 4.88400808729, 47: -1.11116550244, 48: 2.24906755382, 49: 2.58299193597, 50: -0.522123877974, 51: -15.4654131215, 52: 3.83875871829, 53: 1.71787099941, 54: -0.261504893051, 55: 2.08586831141, 56: 1.19642226568, 57: 18.3436066825, 58: -3.70206857679, 59: 6.20551868769, 60: -14.6227113449, 61: 14.5973750463, 62: 16.7690614814, 63: 4.97700655169, 64: 4.29990466779, 65: -0.733528943905, 66: -0.121750428587, 67: -6.17247482115, 68: 1.95014392596, 69: -5.01496996667, 70: 2.09745051156, 71: 1.82715340226, 72: 4.77860409701, 73: 1.40748630782, 74: 3.62428517484, 75: 3.73563723022, 76: 4.04380511368, 77: 7.7563172895, 78: 2.73856995308, 79: 14.4170509003, 80: -0.250137427411, 81: -1.45073942124, 82: 2.85747616001, 83: 1.39281483675, 84: 7.77386295391, 85: -5.26210400667, 86: -1.21871307708, 87: 1.29005664156, 88: 0.103990794937, 89: 1.46999338549, 90: 4.43233214804, 91: 1.39795609944, 92: 1.4922772762, 93: -19.0446503109, 94: 2.86857469628, 95: 1.93385860413, 96: 1.62365979217, 97: 1.19923338112, 98: 1.81919271615, 99: 0.547639325667, 100: 1.99898582351, 101: -7.93287362862, 102: -0.827062133069, 103: -8.60432421065, 104: 1.43659242011, 105: 0.719715464139, 106: -9.51496327141, 107: 1.66876069832, 108: 1.38733482535, 109: 1.6427343325, 110: -7.57377575863, 111: -4.75856975752, 112: -6.30962200863, 113: 14.704431715, 114: 23.72238916, 115: -7.61545629013, 116: -10.087133644, 117: -3.68673654135, 118: 2.13764184543, 119: 19.0291812988, 120: -5.03043303192, 121: -2.10707671909, 122: 7.5174368619, 123: -19.0372024559, 124: -3.13855185537, 125: 19.814258661, 126: 0.465879784701, 127: 0.108241679275, 128: -6.68534263491, 129: -4.59070402811, 130: 0.284691598589, 131: 0.94562488133, 132: 0.263827736123, 133: 4.34582575444, 134: 2.95109363233, 135: -4.22069309804, 136: 0.685564316965, 137: 2.19505029073, 138: 12.3283630225, 139: 4.42436257969, 140: -3.87185440992, 141: 2.64926518182, 142: 13.9240521665, 143: 2.64094863151, 144: 1.29683604574, 145: -3.88433607412, 146: 0.324061525644, 147: 0.209112322084, 148: -15.4858318638, 149: 0.251373414083, 150: 1.51827629882, 151: 7.55823890516, 152: 0.123601687965, 153: 0.206219942536, 154: 0.315639556923, 155: 0.244527495351, 156: 0.701080847639, 157: 9.5876070068, 158: -1.33798145111, 159: 0.38345136591, 160: -0.251625352703, 161: 14.4851881687, 162: -1.40853524147, 163: 10.4058472799, 164: 1.70837466139, 165: 7.51477334668, 166: 34.6858145085, 167: -3.68786755268, 168: -3.19796459163, 169: -1.49977586584, 170: -4.43077875056, 171: 1.459040783, 172: 0.252062759485, 173: 11.8272956634, 174: 0.111003165008, 175: -5.52859073823, 176: 2.99432494462, 177: 10.095582362, 178: 0.0237754833501, 179: 0.219011625813, 180: 0.398850289332, 181: 0.141459026667, 182: 0.0939949726433, 183: 8.56261435621, 184: -2.47011865683, 185: -2.29748266041, 186: 5.65664781403, 187: 7.65632276345, 188: -5.57871470815, 189: 6.55141532014, 190: -2.33591311107, 191: -9.1024673281, 192: 7.30884947668, 193: 0.244797230421, 194: 13.8957778789, 195: 2.94611779256, 196: 3.5125233837, 197: -0.923867479707, 198: 10.2483312165, 199: 0.332960497188, 200: 1.86031228905, 201: 0.224026333426, 202: -0.0833126394781, 203: 0.201275001055, 204: 0.365792133326, 205: 3.15385997987, 206: 1.46409479083, 207: 9.28634973897, 208: 0.344787593182, 209: 0.353003459607, 210: -1.27600149579, 211: 0.17829336735, 212: 2.18667196151, 213: 2.28311179103, 214: 1.83984603033, 215: 0.59247759964, 216: -1.29831824961, 217: -6.67350161391, 218: -0.956885353754, 219: -15.9334626895, 220: -1.31499204531, 221: -1.73241665634, 222: 14.0574775428, 223: -0.137254438413, 224: 0.75498469335, 225: -1.19533357916, 226: -0.277415262214, 227: -4.20115539102, 228: -1.75414604037, 229: -0.61839221559, 230: 4.30713341615, 231: 2.28053975585, 232: 0.066623127228, 233: 12.2048595359, 234: -4.96462237518, 235: -1.83964680914, 236: 2.64984578171, 237: -14.9614855215, 238: -6.69987625675, 239: 1.62447980974, 240: 0.887986553854, 241: 4.47250287219, 242: -0.208669326137, 243: 1.70035566571, 244: -10.6717244814, 245: -3.05452824828, 246: -5.4790977219, 247: -2.71597463699, 248: 5.51279209365, 249: -1.71409505514, 250: 20.9481399484, 251: -3.00656380723, 252: -2.30323937218, 253: -2.16764325349, 254: -5.60756525271, 255: -1.63608503364, 256: 1.31216693784, 257: 6.64694590962, 258: 1.87416035844, 259: 1.48685338083, 260: 8.45321103817, 261: 16.3783369593, 262: 1.47811258092, 263: 3.42742303045, 264: 1.98536520641, 265: 0.746066032418, 266: -16.0689249291, 267: 1.82136650222, 268: -1.61373773335, 269: 14.244268111, 270: -2.52872371075, 271: 2.76938611919, 272: -0.363509177211, 273: 1.0179782497, 274: -22.6929794414, 275: -55.2344362713, 276: -5.06720688816, 277: 5.06299441508, 278: -14.0030174709, 279: -6.29818933881, 280: -9.3493011714, 281: 2.54705871393, 282: -3.9904366316, 283: 9.88636094998, 284: -1.32664762404, 285: -29.5230888859, 286: 12.1493622119, 287: 1.74739978823, 288: -0.0691520685186, 289: 3.30179516876, 290: -3.24615662646, 291: -2.43037011376, 292: -1.2129716674, 293: -14.7709776299, 294: -0.982229868691, 295: 1.83348036277, 296: 1.80188361907, 297: 8.01657216566, 298: -5.11427189277, 299: 1.58751760935, 300: -33.2467139872, 301: 14.0152435617, 302: -9.00938848199, 303: -1.0914841009, 304: 12.9697032694, 305: 3.58551076299, 306: 1.96928992958, 307: -6.01975904861, 308: 1.95700567511, 309: 1.90725223163, 310: 2.21160503725, 311: 3.09104476849, 312: 1.64926535074, 313: 1.36020622009, 314: -4.15223467675, 315: 1.45663672711, 316: -0.871326239844, 317: 1.65910562445, 318: 1.59863082307, 319: -5.20048428483, 320: -1.37739333792, 321: 2.34099164314, 322: 5.88040139449, 323: 16.0703928518, 324: -11.9532537321, 325: -0.727776066826, 326: 0.824172749952, 327: -1.39085195407, 328: -13.0303217477, 329: -9.9860106549, 330: -5.16607682932, 331: -0.365690140538, 332: -1.50207645745, 333: -0.656282244831, 334: -7.26160987634, 335: -12.8352787451, 336: 1.47310987939, 337: -4.33327383885, 338: -8.27636518277, 339: 1.82251910368, 340: 2.37746459415, 341: 5.22923043148, 342: -5.07100981022, 343: -5.66528987442, 344: -6.12052875803, 345: -6.96009377627, 346: -19.3408573061, 347: 0.104423711492, 348: -10.3840142263, 349: -5.50398519817, 350: 0.498627453559, 351: -21.8711466657, 352: -7.68812775155, 353: -14.6158419208, 354: 14.3914391727, 355: -1.15119040755, 356: -6.15955420306, 357: 0.251063569867, 358: 4.55119354065, 359: -7.06945283542, 360: -36.7268027224, 361: 4.89310556768, 362: -9.97197341606, 363: 2.78593341992, 364: -5.51000797889, 365: 12.1132575528, 366: -23.3883034581, 367: 11.3055810265, 368: -18.4147948566, 369: 0.59854407017, 370: 6.43760266147, 371: -28.4547371154, 372: -21.4155658692, 373: -3.4084593938, 374: 10.5994999563, 375: 14.8154741113, 376: 2.45074170207, 377: -0.258037755334, 378: -65.6617606462, 379: 10.7529780255, 380: 6.69626459285, 381: 0.0250051511266, 382: 7.4902842567, 383: 1.47257720407, 384: 5.17508331504, 385: 6.50464559431, 386: 4.18763486668, 387: 13.5509124965, 388: 4.20755575324, 389: 17.6571253899, 390: 11.771957481, 391: -7.66996081794, 392: 9.09837175483, 393: -6.4702207424, 394: -5.84049978061, 395: 7.7658751364, 396: 12.4183697401, 397: -10.2789080239, 398: -85.2829999601, 399: 5.29901175908, 400: -4.80732561158, 401: 28.2076261593, 402: 2.1089707965, 403: -7.79130932616, 404: -2.27539636552, 405: 4.14535797028, 406: 9.94894914342, 407: 3.78748227203, 408: -20.7595283602, 409: 8.8957375626, 410: -19.8015636773, 411: 18.7218175636, 412: 15.2443946447, 413: -20.7346135689, 414: 14.126590484, 415: 3.60421016552, 416: 5.39341739117, 417: 2.73670169145, 418: 6.68184341146, 419: 2.69722386615, 420: 2.44515713118, 421: -2.13271175429, 422: 3.81133646733, 423: -38.8540541074, 424: -4.37724931781, 425: 0.914995793689, 426: -3.18054574803, 427: 5.13764579636, 428: 14.41206266, 429: 8.79397018845, 430: 2.69007219669, 431: 0.490938867021, 432: 2.9066478941, 433: 6.36155493152, 434: -6.70190125656, 435: -4.70984571522, 436: 4.08697086691, 437: 6.05462323198, 438: -1.43802047708, 439: 6.1388973132, 440: 7.01229404432, 441: 3.98014731436, 442: -11.1719945858, 443: 2.8508186144, 444: 2.81628660668, 445: 2.1999844734, 446: 3.02309961823, 447: 2.57337194303, 448: 11.7003284543, 449: -1.75985627166, 450: 11.4576230332, 451: -5.62633233494, 452: -0.100187748457, 453: 4.04502814051, 454: 0.374572627301, 455: -50.7237070356, 456: 2.84700280873, 457: 4.08041213227, 458: -1.79344374079, 459: -9.67585941953, 460: 3.54470047072, 461: 9.80182598291, 462: 14.8057471304, 463: -14.3397223051, 464: -21.6697644749, 465: -16.9693052254, 466: -11.4431487707, 467: -1.88524230174, 468: 15.7335980052, 469: 31.6052745188, 470: 17.3956354643, 471: 18.569464706, 472: 10.2153922867, 473: 2.72091623556, 474: -20.5711955121, 475: 0.578152482963, 476: 1.37406729416, 477: 0.0910545291906, 478: 14.6511931824, 479: 0.697568661881, 480: -5.60853877933, 481: -0.171174793429, 482: 3.88039671414, 483: -1.1536812017, 484: -4.20181870564, 485: -1.06864466169, 486: -2.56623703913, 487: 5.13030944283, 488: 3.10558216538, 489: -16.320358861, 490: -1.58032510933, 491: -15.9426696074, 492: -7.43227889301, 493: 0.395863867579, 494: 9.57844335057, 495: 0.70359640455, 496: 0.692482696309, 497: 11.2067699022, 498: 0.701580253219, 499: -0.168597199751, 500: -1.50766406899, 501: 0.767737731349, 502: 0.77861750298, 503: 0.547436134636, 504: 0.677760879611, 505: 3.14410005936, 506: -13.2260957983, 507: 0.882684051483, 508: 0.924349321078, 509: 1.09789037558, 510: 2.01995407333, 511: 0.0302807940037, 512: 26.8718037243, 513: -1.32020340437, 514: 9.50176922694, 515: 9.87384802117, 516: -1.77615198322, 517: -2.07891903968, 518: 1.40193457418, 519: 0.237261227077, 520: -8.51278237074, 521: 0.845817279182, 522: -27.1937252699, 523: 0.77630996406, 524: -8.82327506795, 525: -20.4862835227, 526: -8.89340334545, 527: 2.33960736015, 528: 0.685609342747, 529: 0.92483090201, 530: 0.923730390468, 531: 0.865858690436, 532: 1.51786560297, 533: 3.50608438845, 534: 1.82839825059, 535: -14.2746386164, 536: 41.8573691555, 537: -4.7416511414, 538: 2.61751901491, 539: 1.41108357662, 540: 3.18706959475, 541: -11.041762775, 542: 0.639303302671, 543: 11.3459460163, 544: -5.9430840849, 545: -1.98609840311, 546: -3.16783598886, 547: 9.4355798574, 548: 0.556702687186, 549: 3.40102263081, 550: 0.911852206946, 551: 16.6511744953, 552: 0.722646078247, 553: 0.644009533177, 554: -31.1925236385, 555: 3.28760187817, 556: -7.50152798019, 557: 0.807322512042, 558: 0.624701663341, 559: 4.00578050806, 560: 0.713771472127, 561: 2.91274722186, 562: 7.39555743992, 563: -14.5881359965, 564: 0.505533842865, 565: 1.46087189519, 566: 23.1041014016, 567: 1.14893054872, 568: 2.20929825327, 569: -5.73170442822, 570: -7.2875652627, 571: 7.90163148984, 572: 13.4510010767, 573: 1.3733428674, 574: -2.42642638636, 575: 1.24079518301, 576: 13.9634234414, 577: 6.51656240825, 578: 2.0261602447, 579: -10.8041321877, 580: 20.6126582045, 581: 0.627312949909, 582: -24.6040399202, 583: -16.5960115556, 584: 4.95550683569, 585: 13.648964206, 586: 2.11384762688, 587: 14.0254077301, 588: 3.57376636224, 589: -5.62444246723, 590: -37.2403738583, 591: -14.5470264576, 592: 8.85059883801, 593: -1.51426837408, 594: 4.69435908613, 595: 14.0487115463, 596: 0.804579051472, 597: 26.697159165, 598: -8.78640525467, 599: -30.7726732151, 600: 12.8640713482, 601: 5.91433937907, 602: -14.3149685059, 603: 6.83723042589, 604: 4.82897448184, 605: 4.22066702831, 606: -0.839519422531, 607: 2.99582252344, 608: -0.186400141797, 609: -1.52590754445, 610: 1.75536296225, 611: 2.96809667401, 612: -4.3093406589, 613: 4.15633755647, 614: 3.22114970173, 615: -11.4881175617, 616: 30.6188718253, 617: 11.6484273099, 618: -3.57278457624, 619: 3.14332183485, 620: 0.417508899445, 621: 3.74419404467, 622: 0.994278397011, 623: -37.645525944, 624: -6.68140516059, 625: 5.26243080891, 626: -6.06764382913, 627: 3.66485433398, 628: -0.828912723808, 629: -18.709286658, 630: -3.23341993767, 631: 20.8229453457, 632: 0.536041528036, 633: 3.13142577871, 634: -20.6508624275, 635: 15.4581523116, 636: 2.21799769488, 637: -1.63055594693, 638: 1.24348620311, 639: 12.4227014182, 640: 9.65525470127, 641: -28.2523211536, 642: 0.186543573647, 643: 21.5465404798, 644: -7.87970013554, 645: 2.28094167303, 646: 13.1856055636, 647: 5.79780148797, 648: 1.96626008673, 649: -3.37665304603, 650: 1.03258899438, 651: -16.7052869583, 652: -6.60655172229, 653: -12.2618583522, 654: -31.6923631563, 655: 2.69155500789, 656: -12.9856423496, 657: 3.43297694206, 658: 3.31918321693, 659: 2.4657987731, 660: -8.014073128, 661: 4.39273332532, 662: -0.700411664655, 663: 48.3029697855, 664: 3.25853726502, 665: 6.60859096832, 666: -0.344922652954, 667: 3.59276968427, 668: 11.6940294719, 669: 5.14736878585, 670: 2.96746745615, 671: 17.9271636939, 672: -16.5609011936, 673: 0.690123875604, 674: -20.821773331, 675: -25.5720965102, 676: -1.00498564849, 677: -4.87101432215, 678: 13.3637456547, 679: -15.7043890386, 680: -9.05693766421, 681: -1.63214559593, 682: 2.69549816014, 683: -5.43676305857, 684: -3.70028882627, 685: 9.09553329574, 686: -3.5993859236, 687: -21.8036002889, 688: 3.24690413069, 689: -21.5395524434, 690: 1.11178547828, 691: 0.627347304753, 692: 15.0088569118, 693: -3.19815565287, 694: -8.95790235342, 695: -35.6283706684, 696: 4.08468950902, 697: 25.1908250339, 698: 10.6707002822, 699: 0.844013006977, 700: 19.1253397916, 701: -6.67236632736, 702: -2.85063681329, 703: 1.0845420396, 704: -20.3790635934, 705: 2.9814041055, 706: 5.28780902803, 707: 20.5477175256, 708: -30.829779214, 709: 1.08843370599, 710: -26.2674908892, 711: -10.6031868339, 712: -4.97653612354, 713: -7.37371670116, 714: 1.31430399566, 715: -2.97939817662, 716: 2.39016710661, 717: 3.03253114541, 718: -2.81986720907, 719: 0.0284339110474, 720: -7.33485710944, 721: -0.224958138262, 722: 13.4672325306, 723: -5.02515797293, 724: 10.2324342857, 725: -15.7790300407, 726: 23.5679758822, 727: -39.8606112315, 728: 3.00230633569, 729: -27.6651768533, 730: -30.7918101277, 731: -7.98953357422, 732: 6.81106510895, 733: -9.67465835661, 734: -7.93053235004, 735: -7.10058023088, 736: -3.17364917227, 737: -4.87432276671, 738: 11.8939591041, 739: 15.6990775636, 740: -0.480172751489, 741: -12.4687486185, 742: 24.3434937393, 743: -4.91876336204, 744: -2.97937661078, 745: 5.87671435617, 746: -6.1248956467, 747: 7.02554942023, 748: 2.65636122312, 749: 2.52339668487, 750: -2.86166511261, 751: 3.3502596461, 752: -12.0388429747, 753: -0.416063060054, 754: 1.70276139048, 755: -18.009013582, 756: 1.48832816731, 757: -23.1615831246, 758: 18.3654832806, 759: 23.7410558958, 760: -26.7997044104, 761: -28.8028977928, 762: 4.5534046567, 763: -18.3550645979, 764: 2.07988837194, 765: 17.8022657264, 766: 7.98250150084, 767: 0.974523176936, 768: 3.24630831731, 769: 4.99582181933, 770: 14.2585308612, 771: 2.14047518269, 772: 18.5789099415, 773: -9.15367141569, 774: 5.53690146106, 775: -8.05544693362, 776: -2.80724956299, 777: 23.7533389863, 778: -21.3037491388, 779: 1.85936410547, 780: -22.9710970975, 781: 5.30246521884, 782: 3.65112834239, 783: -7.64677794926, 784: -10.4437733633, 785: -3.05144530902, 786: 18.6618520992, 787: -5.05865630395, 788: 2.22556830956, 789: 0.851733292842, 790: 5.13920053614, 791: -16.0427106722, 792: 4.0748301759, 793: 5.94031748034, 794: 4.72779772479, 795: -1.19771503671, 796: 4.82668396187, 797: -5.00784664068, 798: 5.10337376319, 799: 23.4373852157, 800: -5.00212775951, 801: 1.37133485483, 802: 1.78532524472, 803: 0.912091296847, 804: -75.3474864547, 805: 1.37827168342, 806: 4.76648624445, 807: -4.73964526796, 808: -17.1910420396, 809: 1.89943267417, 810: 8.40313753182, 811: -18.8811379566, 812: -0.677439935224, 813: 22.0793551757, 814: 2.59665569155, 815: 0.497527330288, 816: -11.7605102738, 817: -22.8707568551, 818: -9.34551345799, 819: 9.28933686181, 820: -1.34935998007, 821: -5.59480897198, 822: -2.83153342903, 823: 3.30962135246, 824: 1.18091023727, 825: 1.41298788733, 826: 10.5998987956, 827: 18.0344842935, 828: 1.02876462792, 829: 5.3241030987, 830: -2.29930225562, 831: 4.2763001306, 832: 3.30723451227, 833: 22.6801158366, 834: 15.4444493207, 835: 2.17760693408, 836: 12.5772301712, 837: 2.26040079245, 838: -2.5489353688, 839: -5.64007578, 840: 2.64169529965, 841: -4.56929248144, 842: 1.59508229024, 843: -7.74207473279, 844: 0.895749835625, 845: 1.11089493974, 846: 10.223458815, 847: 1.06428016439, 848: 5.09949019175, 849: 7.66236970497, 850: 1.71842438908, 851: 1.03544410503, 852: 1.19625620977, 853: 0.875551405861, 854: 5.20222040275, 855: -14.8059991611, 856: 1.24737089415, 857: 1.5034029519, 858: 0.649438914256, 859: 1.53774447411, 860: 5.6299357566, 861: 2.11085673752, 862: -3.15081783728, 863: -26.7198005336, 864: 4.61842380811, 865: 2.39187333322, 866: 2.52112815826, 867: -5.5525552984, 868: 4.98059930454, 869: 5.54120846389, 870: 0.931999367844, 871: -9.25234459911, 872: 1.04231315235, 873: 11.5813921485, 874: -31.4966131659, 875: 33.2654488175, 876: 3.01858513729, 877: 0.964738787868, 878: 1.68068024521, 879: 0.713422262391, 880: 0.973186332227, 881: -3.3661075229, 882: -1.46282219049, 883: -8.5923898891, 884: 24.5995850706, 885: 2.20988229705, 886: -7.92522686024, 887: -4.61191486562, 888: 4.47146769272, 889: -24.2530638062, 890: 0.015255892023, 891: 0.853349343117, 892: 20.2272262799, 893: 6.57212101996, 894: 2.2156014848, 895: -9.27979171674, 896: -14.6325687421, 897: 1.02540725676, 898: 5.10561544799, 899: 1.34381252292, 900: -2.24337466571, 901: 1.08605104443, 902: 0.984009139781, 903: 4.11778393739, 904: 3.3520195637, 905: -5.55534892511, 906: 1.06031396431, 907: 0.845402175318, 908: 1.71834118234, 909: 1.06392700009, 910: 1.63003292222, 911: -7.99680747125, 912: 21.8120958292, 913: 1.36179059585, 914: -0.478206566246, 915: -6.57933697395, 916: -2.173521564, 917: 9.31395517877, 918: 29.2223494501, 919: -8.27185478357, 920: 4.31047294084, 921: 0.942668260198, 922: 3.91823911798, 923: -4.73631541245, 924: 1.35322127863, 925: 5.0827346257, 926: 0.792004861828, 927: 4.06962054832, 928: 7.08865878688, 929: 22.2334149858, 930: 0.649445471581, 931: 2.73805049205, 932: 12.3633541044, 933: 0.211872999641, 934: 8.23622211751, 935: 0.651600717213, 936: -1.99686316474, 937: 2.53240315264, 938: 5.24763469473, 939: -15.7186276084, 940: -8.90719017545, 941: -40.8567115962, 942: 20.336478031, 943: -10.7716413631, 944: -22.9268906371, 945: -1.31472453036, 946: -0.117687622101, 947: -36.7437434319, 948: -10.0854328205, 949: -20.1125027049, 950: 14.5674099565, 951: 0.12198592378, 952: -5.14404165133, 953: 10.4187020013, 954: 3.1895878531, 955: -23.5885194325, 956: 5.12110802824, 957: 5.06284270867, 958: 7.25736759525, 959: -16.5681077929, 960: 5.28201591224, 961: 5.66960104386, 962: 8.09602259973, 963: 5.31127180144, 964: -13.6754960513, 965: -1.16589999704, 966: -12.3239191758, 967: -6.35320335897, 968: -0.258986203858, 969: 3.13172638875, 970: 11.2416017629, 971: 1.42738273036, 972: 36.2596148714, 973: 44.6727613183, 974: -4.857278762, 975: 18.6392900345, 976: -9.8856404717, 977: 5.75844057001, 978: -6.74440131183, 979: 1.8029387631, 980: 21.0287982037, 981: -12.4484915315, 982: 45.6055598414, 983: -10.8167859811, 984: -25.8040787816, 985: 4.559135329, 986: 0.377504386351, 987: 5.71012664584, 988: -1.61347057994, 989: -3.09211655973, 990: -29.2970071996, 991: -35.9862530792, 992: -0.963361972299, 993: 9.7609687141, 994: -3.3151955242, 995: -4.96655452058, 996: 1.53705249203, 997: -4.04304038872, 998: -6.85413210408, 999: 4.45526978362, 1000: -13.1370721403, 1001: 2.14663681236, 1002: -12.5056070244, 1003: -26.9358361607, 1004: 5.90690096642, 1005: -12.2086499838, 1006: 4.98828457302, 1007: 5.05605426516, 1008: 4.18250577166, 1009: 8.15232779246, 1010: -0.706671170422, 1011: 5.39039189101, 1012: 24.6161380065, 1013: 3.35830287391, 1014: 2.98155944316, 1015: 5.30053154033, 1016: 4.80774593883, 1017: -4.45706651307, 1018: 0.618798820618, 1019: -1.28254086643, 1020: -20.8278182312, 1021: -2.84949545782, 1022: 15.3440690112, 1023: 5.81478349908, 1024: 29.0968231444, 1025: -10.105251711, 1026: 10.9181160009, 1027: -47.8612163661, 1028: -0.0889286344797, 1029: -1.0264488527, 1030: -21.6664179488, 1031: 1.44075578116, 1032: 1.72013087277, 1033: 1.553644526, 1034: 3.53305416611, 1035: 10.2566731022, 1036: 9.19964974414, 1037: 3.31117802553, 1038: -18.7248394715, 1039: 3.68245269242, 1040: -26.7317412821, 1041: 12.1444499616, 1042: 27.2471471989, 1043: -11.2406579637, 1044: -0.96664538066, 1045: -13.9945641154, 1046: 0.767813288465, 1047: -30.563010028, 1048: 0.736374279129, 1049: 14.9387339872, 1050: -4.18706597057, 1051: 3.79542677188, 1052: 1.54611143755, 1053: -3.94230697091, 1054: -2.40772448634, 1055: -2.24230893189, 1056: 4.46393816512, 1057: 3.12941797493, 1058: 1.55636322468, 1059: 2.90732292958, 1060: 1.2954345316, 1061: 4.64923888123, 1062: 1.36242115901, 1063: -4.76530924931, 1064: -12.7765445153, 1065: -0.198717471902, 1066: -20.8512850102, 1067: 4.63819537034, 1068: -0.24027996404, 1069: 6.42927078648, 1070: 4.2587772159, 1071: 8.98375387481, 1072: 6.8943199429, 1073: -4.94302604221, 1074: -3.58732119717, 1075: 0.549816705724, 1076: -3.61426275988, 1077: -0.803645902033, 1078: 3.10419430628, 1079: 7.66156582596, 1080: 4.72865447877, 1081: -0.166948970006, 1082: -0.700249288177, 1083: 1.60871657953, 1084: 0.112871586514, 1085: -4.01811257222, 1086: 1.29929363466, 1087: -12.5086404644, 1088: -13.2810096287, 1089: 1.30726215369, 1090: -1.92657490534, 1091: -7.39270314915, 1092: 2.1423154503, 1093: -0.52799592289, 1094: -3.98901845777, 1095: 3.14640016214, 1096: 8.49796754084, 1097: -1.4295861279, 1098: -2.17076959558, 1099: -0.58169236455, 1100: 1.97108080132, 1101: 0.922027742482, 1102: -9.97550422474, 1103: 1.91181247027, 1104: 5.52386697411, 1105: -0.930453377264, 1106: -2.13482688934, 1107: -0.223455916956, 1108: 5.93177416207, 1109: 10.0689290856, 1110: 3.08684809955, 1111: -4.28258677388, 1112: 0.493501550552, 1113: 2.00871333425, 1114: -3.90715232415, 1115: 2.5822423809, 1116: 0.922309666149, 1117: 2.66151975305, 1118: 2.71733203094, 1119: 2.19661960643, 1120: 2.42326356579, 1121: 3.27931440737, 1122: -0.00228271574243, 1123: 1.71531133561, 1124: -7.00087590542, 1125: 1.49724527533, 1126: 6.4124026414, 1127: -7.68889535803, 1128: 1.41123158021, 1129: -0.258979919285, 1130: 2.32068585639, 1131: 2.0423366678, 1132: -2.36692327339, 1133: 2.73632532475, 1134: 2.83710775922, 1135: 5.82328435545, 1136: 1.99217658577, 1137: 2.37024968664, 1138: 2.05608522935, 1139: 2.65440866947, 1140: -15.9906699672, 1141: 2.53233390441, 1142: 2.70773259147, 1143: 2.37060114034, 1144: 2.12028806101, 1145: 2.59307101798, 1146: 2.31419419274, 1147: 2.44449402865, 1148: 11.1601511622, 1149: 3.22006515184, 1150: 4.49590263881, 1151: 1.77862304314, 1152: 1.7876688659, 1153: -8.98301957892, 1154: 2.53597511254, 1155: 1.88324982322, 1156: 2.26275437926, 1157: -1.83333694946, 1158: 2.59721905141, 1159: -7.55514480682, 1160: 19.0302389293, 1161: 7.47980974997, 1162: 3.0674450129, 1163: 0.226253199443, 1164: 6.85070923014, 1165: -8.07428142811, 1166: -20.2237599916, 1167: 8.74503039547, 1168: 2.97566536268, 1169: 5.87133697048, 1170: 16.0192321377, 1171: -0.254192733888, 1172: 3.5458742702, 1173: 0.75393380269, 1174: 0.641679783336, 1175: -0.887801390921, 1176: 5.95179415604, 1177: 0.766369093413, 1178: 2.10981441583, 1179: 0.255702336335, 1180: 0.911211004699, 1181: 2.03905373141, 1182: 0.686069477708, 1183: -0.737637044605, 1184: 2.26163862185, 1185: 2.22289073961, 1186: -1.36431302419, 1187: 2.35702290925, 1188: -17.2145498174, 1189: 3.10757504931, 1190: 1.85068202109, 1191: 1.05537387251, 1192: 10.6746403258, 1193: 0.56006012588, 1194: 0.675515703984, 1195: 7.58017691448, 1196: 0.615924998928, 1197: 2.20659010672, 1198: -2.01936297571, 1199: 0.765972270342, 1200: 0.578350132236, 1201: 0.669616069184, 1202: 0.746288368926, 1203: 3.19625554613, 1204: 8.05539019695, 1205: 0.000184048632605, 1206: 0.957571399274, 1207: -0.216849511162, 1208: 2.48571471463, 1209: -0.0574894640565, 1210: -8.06440422843, 1211: -2.16868262623, 1212: -1.13494380208, 1213: 1.55899468163, 1214: 4.04117567237, 1215: 2.71407389955, 1216: -1.04874823432, 1217: -0.589778278003, 1218: -7.06269231615, 1219: 0.809732380615, 1220: -1.99917404794, 1221: 0.691253994737, 1222: 0.813648737137, 1223: -4.692592453, 1224: 1.01774100797, 1225: 1.56645337929, 1226: 0.5954900874, 1227: 0.929476654833, 1228: 0.606014298572, 1229: 0.760659864428, 1230: -3.65367215746, 1231: 6.02915309662, 1232: 7.1334667077, 1233: 0.0311675108011, 1234: 14.4705186315, 1235: -1.89400356007, 1236: -1.21847438279, 1237: -4.15697454503, 1238: 1.49586765772, 1239: 0.0388558804163, 1240: 0.72385078521, 1241: -9.40114318274, 1242: 2.64000502557, 1243: 1.11906871085, 1244: -2.15119863179, 1245: -8.27941200506, 1246: 0.554252643885, 1247: 2.52144018918, 1248: 0.734866629467, 1249: -7.2189779625, 1250: 0.641525145622, 1251: 0.583895781352, 1252: 8.24175520844, 1253: 2.22699267306, 1254: -13.3817975183, 1255: 0.699718049622, 1256: 0.604449978112, 1257: 1.32005567672, 1258: 0.665672213655, 1259: 2.03548706994, 1260: -3.50321983137, 1261: 0.328031834791, 1262: 1.50302124644, 1263: 0.572751520556, 1264: 4.54182158864, 1265: 1.40134728893, 1266: -1.25782707882, 1267: 1.87200950932, 1268: -1.5048464015, 1269: -3.34766434768, 1270: -2.68585765548, 1271: 2.18185301771, 1272: -1.80023922136, 1273: 0.437954464446, 1274: 8.28528383398, 1275: 2.4721514408, 1276: 3.80158721516, 1277: -0.710544925123, 1278: -17.6423362752, 1279: 1.66295956544, 1280: 0.701141133694, 1281: -9.06296780271, 1282: 1.12631071967, 1283: -0.965064688011, 1284: 7.79976215808, 1285: -2.46736962597, 1286: 2.45861909407, 1287: 2.24237556066, 1288: -9.85520988414, 1289: 6.56532445254, 1290: -5.26641651867, 1291: 3.9338600131, 1292: 0.674266169466, 1293: 0.413640264099, 1294: 6.11169409662, 1295: -1.01642251218, 1296: -12.6518124004, 1297: 26.9999934489, 1298: 1.40484298415, 1299: 3.15201908091, 1300: -6.92898050517, 1301: -9.96540632737, 1302: 4.42188933158, 1303: 1.9725724953, 1304: -6.67416728571, 1305: 2.23097171679, 1306: 2.30162482854, 1307: -1.1905822844, 1308: 8.16133263046, 1309: 2.47792217188, 1310: 1.74023589613, 1311: 3.10228325505, 1312: 3.27846262339, 1313: -3.31482612721, 1314: -9.85646964557, 1315: -4.13670126059, 1316: 1.24223452934, 1317: 0.77542727205, 1318: 1.22344844348, 1319: -8.80374976038, 1320: 2.28666110441, 1321: -2.70883443044, 1322: -10.1885835541, 1323: -17.3473672871, 1324: -0.198936508956, 1325: -10.0576810951, 1326: -2.41839530448, 1327: 1.15957407376, 1328: 1.63440358033, 1329: 9.41351642693, 1330: -4.94049519756, 1331: -0.612225961782, 1332: -5.27285760815, 1333: -36.247418244, 1334: 2.53048519816, 1335: 6.05601388492, 1336: 3.0389696647, 1337: 3.85680092017, 1338: -0.21391684268, 1339: 2.73409890218, 1340: -1.84214428876, 1341: -15.9205826356, 1342: -20.1175761121, 1343: -8.83992085064, 1344: 2.1162659386, 1345: -1.06479010762, 1346: 4.33848464436, 1347: -9.20675226099, 1348: -0.400860577919, 1349: 5.15787882857, 1350: -1.73173711094, 1351: -18.7303055727, 1352: 3.46836515468, 1353: 2.65495410435, 1354: 0.299003703076, 1355: 2.29187327593, 1356: 2.28768777285, 1357: 2.58329324028, 1358: 3.67056461078, 1359: 1.8563817958, 1360: 2.60968377634, 1361: -12.0607596375, 1362: 2.31033742603, 1363: -12.731257484, 1364: 2.72701440149, 1365: 2.24988379537, 1366: -10.8015767989, 1367: 2.48297422461, 1368: 1.43011848437, 1369: -9.62607945261, 1370: -8.40583753162, 1371: 5.44757465908, 1372: -13.7212407673, 1373: -6.14423374003, 1374: -6.73739204087, 1375: -1.98651334429, 1376: -15.1926870786, 1377: -5.86136665635, 1378: -8.00637187688, 1379: -13.6452990546, 1380: 1.6712018909, 1381: 0.8542110762, 1382: -3.42674583226, 1383: 15.4579112459, 1384: 5.36764923535, 1385: 25.0201056846, 1386: 2.5375018831, 1387: 3.12057053703, 1388: 5.8150590209, 1389: 4.87399379756, 1390: 1.342616949, 1391: 0.858669805061, 1392: -8.25839324652, 1393: -8.00144305927, 1394: -30.8194645499, 1395: -6.44004968172, 1396: -23.1046216425, 1397: -0.0658447618279, 1398: 8.43439259772, 1399: 3.18650735752, 1400: 3.58183614657, 1401: -10.3250467062, 1402: 2.1408812432, 1403: -7.88229261585, 1404: 15.8614874172, 1405: -0.369968075024, 1406: 6.615795925, 1407: -11.5305145406, 1408: -11.3579248471, 1409: 3.84481261565, 1410: 10.8699939734, 1411: 5.55461154427, 1412: -7.99019725503, 1413: 7.09298917252, 1414: 0.0630699211243, 1415: -4.45869707919, 1416: 10.2538964896, 1417: -9.97977883876, 1418: 0.653637838268, 1419: -2.08187167332, 1420: -2.241151585, 1421: 10.3502504598, 1422: -11.7512130449, 1423: 12.8550939622, 1424: -4.4016071942, 1425: 3.98246564052, 1426: 3.49395790817, 1427: -0.256601394387, 1428: -9.19428936904, 1429: 11.136869729, 1430: 1.07223291292, 1431: -1.29760660122, 1432: -6.60873899862, 1433: 1.70265643881, 1434: -2.57951074555, 1435: -11.6777917024, 1436: -6.92770138335, 1437: 10.5322816218, 1438: -0.293453463605, 1439: -5.27041162486, 1440: 11.253601389, 1441: -6.54216838455, 1442: -4.15676907826, 1443: -7.5792554543, 1444: -0.89522006784, 1445: -11.1613448485, 1446: -3.35639997929, 1447: 8.8926694122, 1448: -10.6674786697, 1449: -1.52415076317, 1450: 0.260963903114, 1451: -3.11077193967, 1452: -1.72273194128, 1453: -4.74224383194, 1454: -4.58623811809, 1455: -5.20317592782, 1456: 15.8039898933, 1457: -23.0109284592, 1458: 0.577577053721, 1459: -7.31911333093, 1460: -1.61369675041, 1461: 0.135329320803, 1462: -0.745020712359, 1463: -10.3276880942, 1464: -0.969368142762, 1465: 11.691491764, 1466: -1.14784706965, 1467: -1.11272062171, 1468: 2.23464236406, 1469: -1.42114490881, 1470: -11.5631543164, 1471: -4.94688155972, 1472: -2.03681531048, 1473: -5.77839196082, 1474: -15.0283714434, 1475: -19.1226065443, 1476: -3.92010243553, 1477: -0.0221278398173, 1478: 12.4842983906, 1479: -1.40643529458, 1480: 8.39145445069, 1481: -0.167924113774, 1482: -1.85173153915, 1483: 6.96202595621, 1484: -0.880688363515, 1485: -0.448167523984, 1486: 6.50184485925, 1487: -1.63860199631, 1488: -1.56150277246, 1489: 3.03588006066, 1490: -0.116794317695, 1491: -1.03172821166, 1492: -1.57727729177, 1493: 3.64140408441, 1494: -1.13257449333, 1495: -5.162390722, 1496: -0.959820451849, 1497: 10.0813383538, 1498: -0.243575383417, 1499: 0.0514101723794, 1500: -1.42972779133, 1501: -0.0569048001846, 1502: 18.7284614019, 1503: -1.31705560574, 1504: -1.56350246249, 1505: -0.047063619517, 1506: 4.84661025459, 1507: 1.01964251246, 1508: 9.82770485046, 1509: -13.071158636, 1510: -28.2156846586, 1511: 11.5104210845, 1512: 11.0692690881, 1513: 2.65722274767, 1514: -0.266024784378, 1515: -18.9764255195, 1516: 7.73528689136, 1517: -13.9806000797, 1518: -12.610300013, 1519: -3.62931013287, 1520: -3.16663374847, 1521: -7.84437342643, 1522: -0.167002343352, 1523: -0.0945778397475, 1524: -0.531686826044, 1525: 0.404984827517, 1526: -0.29462996165, 1527: -1.1934056724, 1528: -0.0394548522272, 1529: 8.53721408266, 1530: 0.376242625817, 1531: -0.948226461688, 1532: -1.51514605949, 1533: 1.05198317739, 1534: 3.61468704483, 1535: 3.06577245904, 1536: -1.64539170335, 1537: 4.06053970842, 1538: 6.42390349916, 1539: -3.28165506637, 1540: -0.303768676032, 1541: -5.17261215959, 1542: -0.428246792374, 1543: -0.299397976652, 1544: 11.5288916272, 1545: -0.21121336371, 1546: -1.51922856475, 1547: -9.30606960231, 1548: -0.299999646715, 1549: -0.25738082811, 1550: -0.373143732516, 1551: -0.346593207403, 1552: -1.20602974992, 1553: 4.55762993109, 1554: -1.6888312274, 1555: -0.38898768813, 1556: -0.485142768695, 1557: 4.84693117512, 1558: -17.797898393, 1559: 23.0434035523, 1560: 6.26084702581, 1561: 8.1409665806, 1562: -13.1259439412, 1563: -3.26720928648, 1564: 13.4691188187, 1565: 2.69246369315, 1566: 6.61004929604, 1567: 23.142420065, 1568: -0.281782888582, 1569: -19.7148950251, 1570: -0.189844952936, 1571: -19.064656504, 1572: -12.8841443779, 1573: 13.3635927403, 1574: 1.7315663106, 1575: -0.175013083541, 1576: 0.0364636254401, 1577: -0.306995478669, 1578: -0.316626319628, 1579: -3.07366859988, 1580: 1.78357885799, 1581: 3.56312102257, 1582: -10.9072415105, 1583: 1.48548813526, 1584: 3.39492953017, 1585: -5.91697754536, 1586: 6.26839678168, 1587: -4.39725869179, 1588: 4.7091917789, 1589: -0.185344260318, 1590: 0.662148764425, 1591: 0.488690196831, 1592: 2.01568356282, 1593: 1.40049027956, 1594: -5.8170783466, 1595: -0.232860468819, 1596: -1.36758772812, 1597: -0.277848625445, 1598: 21.2859593394, 1599: -0.290535093138, 1600: -0.113204678922, 1601: -9.77301368671, 1602: -1.66686341675, 1603: -2.76198645362, 1604: -0.221225774346, 1605: -0.320213004474, 1606: -3.56189170278, 1607: -0.180903422293, 1608: 0.555964646256, 1609: 5.29712667204, 1610: -4.5808284802, 1611: -0.679284171204, 1612: 1.8279666988, 1613: 9.79108117353, 1614: 1.39684323441, 1615: -21.7633413384, 1616: -6.60053201438, 1617: -2.01149782632, 1618: -5.09225485978, 1619: -1.42555476261, 1620: -1.45383946801, 1621: 2.42569445058, 1622: -0.791841357532, 1623: 0.363471131006, 1624: -4.90991003647, 1625: 2.84521770476, 1626: 3.16058704137, 1627: -9.65705904314, 1628: -1.33786558988, 1629: 7.68513030639, 1630: -11.7502013847, 1631: 1.22583304951, 1632: -2.38337612719, 1633: 1.20733335654, 1634: -0.497885620134, 1635: -1.95263481358, 1636: -1.17622421722, 1637: -5.55532353837, 1638: -8.79869165072, 1639: 1.98536916484, 1640: -4.5507477234, 1641: 1.85507938641, 1642: -2.45136743503, 1643: -9.7934838113, 1644: 5.84480509204, 1645: 18.1002226828, 1646: -21.8232557451, 1647: -17.1918473151, 1648: 4.36308956522, 1649: 20.2467402423, 1650: 3.49153631006, 1651: 1.6576779221, 1652: -1.77485262696, 1653: -14.1089374605, 1654: -1.41361055785, 1655: -1.455275114, 1656: -2.6860474413, 1657: 7.69056121052, 1658: -1.49757435552, 1659: -0.0103628864623, 1660: -0.267526585506, 1661: -1.15898915762, 1662: 5.77570066749, 1663: -8.79040562618, 1664: 6.24719692512, 1665: -10.2577799509, 1666: -13.8752860615, 1667: -8.21170299046, 1668: 1.11417268529, 1669: 0.612213541483, 1670: -2.29341227126, 1671: 3.30152441311, 1672: 6.31859942555, 1673: 10.9587815843, 1674: 10.3190576628, 1675: 10.6011073978, 1676: -5.09596842558, 1677: -3.49545571797, 1678: 8.37164033657, 1679: 4.83892701389, 1680: 8.84780060751, 1681: -2.314156829, 1682: -10.8232614961, 1683: -1.15436693776, 1684: -9.58866757621, 1685: -0.461671816529, 1686: -0.676841196541, 1687: -3.31085627332, 1688: -3.76497662795, 1689: 6.46617287828, 1690: 4.66521108278, 1691: 15.771953141, 1692: -0.990600452463, 1693: -8.11519415743, 1694: 16.5651050151, 1695: -1.0614434289, 1696: -12.0971152869, 1697: 21.903555005, 1698: -5.72320665944, 1699: 2.01173143007, 1700: -0.121588460727, 1701: 3.41331796089, 1702: -0.899784402137, 1703: 8.49935274734, 1704: -1.4605575926, 1705: -1.50622915106, 1706: 1.64179850631, 1707: -13.5016345789, 1708: -1.54427420387, 1709: -1.43805149773, 1710: -1.09008082589, 1711: -1.79945599949, 1712: 7.77642378672, 1713: -1.31130886373, 1714: -1.47302376379, 1715: 14.9807602458, 1716: -1.42555316955, 1717: 1.00316515438, 1718: 0.973081188236, 1719: -12.8451645395, 1720: 13.9880816784, 1721: -1.91199187147, 1722: 1.60363431849, 1723: 3.9106227623, 1724: 28.211834182, 1725: -4.88883606622, 1726: 11.0883833854, 1727: 3.73448222496, 1728: 6.5212389935, 1729: -1.79775530164, 1730: 2.3982675178, 1731: -3.46648457842, 1732: -3.36237624501, 1733: 11.6953000536, 1734: -3.98240895398, 1735: -1.34609750556, 1736: -1.16988180578, 1737: -6.6548664971, 1738: 5.771779566, 1739: 8.4621650973, 1740: -9.49703321793, 1741: 5.7436868105, 1742: 22.3023705086, 1743: -6.45325596117, 1744: -9.40705963582, 1745: 3.82311369917, 1746: 0.0759363771016, 1747: 5.00908156617, 1748: -2.82974638364, 1749: 0.618719007873, 1750: 10.0348540168, 1751: 1.79842654063, 1752: -2.94511748746, 1753: -12.4955231159, 1754: -0.661119917636, 1755: 0.585420519447, 1756: 6.46856415211, 1757: -4.02922637213, 1758: -28.8075842792, 1759: -1.64051535785, 1760: -0.189533799972, 1761: 4.73295644736, 1762: 2.31315663527, 1763: 6.27877116645, 1764: 1.38444385914, 1765: -9.58896379898, 1766: 9.16954537688, 1767: 6.20138702538, 1768: -11.0672615691, 1769: -12.6535613573, 1770: 0.192332287776, 1771: 1.23827300429, 1772: 10.1060011667, 1773: 2.32956700158, 1774: 2.03793585333, 1775: 0.461059105473, 1776: 5.04430225893, 1777: 1.11125195526, 1778: -1.34028767099, 1779: -2.99072704206, 1780: 1.40823001867, 1781: 1.76498305686, 1782: 2.38070985791, 1783: 9.14695365779, 1784: -0.203587920716, 1785: 5.24090668763, 1786: -1.28679229817, 1787: -0.0653251124436, 1788: 13.2960097228, 1789: 4.18754119441, 1790: -1.46908766674, 1791: 2.33828015423, 1792: -6.20892266411, 1793: 0.730247454859, 1794: 3.0408044161, 1795: -4.55320701781, 1796: -1.06047316611, 1797: 23.4120756266, 1798: 0.16488286211, 1799: 1.62528092639, 1800: 9.31738576379, 1801: -0.533177948364, 1802: -17.9991275181, 1803: 3.05020309066, 1804: -8.84767539353, 1805: 2.9290562202, 1806: 10.5466400223, 1807: 1.60524801265, 1808: 4.76493751503, 1809: -2.11499563492, 1810: 7.3623895972, 1811: 0.38779128037, 1812: -1.42563280123, 1813: 3.10954478524, 1814: -3.67198668766, 1815: 0.211978329933, 1816: 0.113503139943, 1817: 2.25693945665, 1818: 0.155307579286, 1819: -1.218943259, 1820: 0.374286065022, 1821: -0.47239355572, 1822: 1.93892380757, 1823: -1.4546856468, 1824: -3.72479804174, 1825: 6.19447316817, 1826: -0.314949581087, 1827: -8.04020634298, 1828: 0.380685637522, 1829: 8.01532363722, 1830: -4.14363664927, 1831: -8.63565728918, 1832: -0.870210129986, 1833: 1.67693385048, 1834: 0.0909445466215, 1835: -9.53081498131, 1836: -0.0430937878352, 1837: 0.171085605859, 1838: -11.1609543195, 1839: 0.292550185475, 1840: -0.146446324456, 1841: -2.36304471733, 1842: 1.68534540692, 1843: 0.327536380977, 1844: -0.529065749253, 1845: 0.290590886301, 1846: -6.65023742935, 1847: 0.540645653294, 1848: -4.55295148305, 1849: -1.37291090652, 1850: -1.42813857542, 1851: 1.43251260882, 1852: 0.0109675198813, 1853: 0.262784779035, 1854: 0.244932272769, 1855: 1.29390875849, 1856: 0.437554705972, 1857: 2.91937545442, 1858: 13.6833306086, 1859: 7.20399755922, 1860: -20.6303495108, 1861: -1.64154194606, 1862: -6.8383631196, 1863: -2.67180192168, 1864: 6.36878374306, 1865: -1.31895256366, 1866: 13.390737444, 1867: 14.3200492963, 1868: 4.16495316921, 1869: -5.31004334599, 1870: 1.27303927282, 1871: -0.0137177694186, 1872: -0.535851390167, 1873: 4.36680987674, 1874: -1.40628939558, 1875: 0.0660392419807, 1876: -0.016726816802, 1877: -0.196066830327, 1878: -9.2399746088, 1879: -0.661437169677, 1880: -2.19634845829, 1881: 6.20094021612, 1882: 1.82290856901, 1883: -1.71100713348, 1884: 1.62199163027, 1885: 4.14211977926, 1886: 3.81248338506, 1887: -2.72697886113, 1888: 4.98502199179, 1889: -0.608244840163, 1890: -0.0228309919996, 1891: 0.12921789288, 1892: 0.0554706531378, 1893: 0.994238079137, 1894: -0.0742308195313, 1895: 0.334542111124, 1896: -15.2969722284, 1897: -1.35797144919, 1898: 0.020784489406, 1899: 0.0494292116636, 1900: 0.108963776814, 1901: 0.137909199348, 1902: 6.60621244791, 1903: -2.74336558052, 1904: 0.414452033265, 1905: -1.40231947252, 1906: 12.8601228419, 1907: -7.73279012403, 1908: -3.19089800134, 1909: -1.40708912998, 1910: -26.536715692, 1911: -1.24336909613, 1912: 10.8912813931, 1913: -0.341060911505, 1914: 2.26912635182, 1915: -2.64990148918, 1916: 15.7738263626, 1917: -0.104199883954, 1918: 10.5472697079, 1919: -0.0575442231327, 1920: 16.6644199537, 1921: -0.27678909788, 1922: 3.08276889824, 1923: 0.370821980573, 1924: 0.0537537927294, 1925: -0.224465112257, 1926: -0.146842664147, 1927: 0.0764840990768, 1928: 0.331684127514, 1929: -3.56114224728, 1930: -8.0377714175, 1931: -5.19173082948, 1932: -2.85026505241, 1933: 6.81694573313, 1934: -1.13879509138, 1935: -1.34273777605, 1936: -8.78084729975, 1937: -4.5658731212, 1938: 0.0289848682103, 1939: 7.28832396658, 1940: 0.368059166778, 1941: -1.21564195042, 1942: 4.51835459395, 1943: 7.94079480804, 1944: -0.0307365528916, 1945: 0.268728262498, 1946: 0.0597633899591, 1947: -2.01880242362, 1948: 0.042213103908, 1949: -0.0133714035007, 1950: 6.47123179065, 1951: 0.0807310349736, 1952: 0.660632583472, 1953: 0.0970073848772, 1954: 0.123335460678, 1955: 0.661906455725, 1956: 0.0432037532072, 1957: -4.73245651787, 1958: -0.142366381685, 1959: 0.147841304655, 1960: 0.963396005492, 1961: -0.00771850488672, 1962: 4.72984727378, 1963: 1.41386931178, 1964: -3.76223255407, 1965: 2.48462715074, 1966: -0.355470149561, 1967: -2.65541225, 1968: 7.06637876666, 1969: -0.427430826806, 1970: -0.366588756691, 1971: -1.08091576745, 1972: 3.37112158709, 1973: 0.497215603099, 1974: 0.898630108698, 1975: 0.265952108788, 1976: 8.57535422563, 1977: 0.303194034319, 1978: -18.9633568829, 1979: -3.00005797939, 1980: -0.0819379198974, 1981: -10.1910105011, 1982: -5.08291045838, 1983: -18.5315008907, 1984: -1.20653456881, 1985: -0.220494100287, 1986: -15.6641669378, 1987: -4.35805205015, 1988: 5.37597464012, 1989: -3.87065121689, 1990: 3.61313425499, 1991: 11.205368486, 1992: 0.392494835277, 1993: -2.52949573321, 1994: 17.7178802474, 1995: -5.0396401028, 1996: 3.92229547918, 1997: -2.71601239736, 1998: -3.52971280184, 1999: 8.57736654, 2000: -3.98912640168, 2001: 0.889967208553, 2002: 1.19044259337, 2003: -0.591749800177, 2004: 0.357454066851, 2005: 5.90913198684, 2006: -6.00708687326, 2007: -0.134846067152, 2008: -13.2932476946, 2009: 1.48431337701, 2010: 0.144609168463, 2011: 11.3072961406, 2012: -7.8084767507, 2013: -3.85854175737, 2014: -0.795564915303, 2015: 3.34580581387, 2016: -1.42987464266, 2017: -0.388833411336, 2018: 0.863919464887, 2019: 12.441842982, 2020: -3.8060425068, 2021: 4.9716551565, 2022: 1.07507003896, 2023: -3.03464118334, 2024: 12.8045987587, 2025: -10.6085391377, 2026: 2.47679979254, 2027: -6.79226592275, 2028: -1.36295291897, 2029: 9.55202529235, 2030: -5.89585529966, 2031: -0.356723822224, 2032: 0.674858745669, 2033: 0.987813095105, 2034: -0.237722801484, 2035: -3.58342857653, 2036: -4.49225379128, 2037: -15.8122388491, 2038: -10.6785091101, 2039: 0.468268181255, 2040: 0.141824516798, 2041: 5.85947742523, 2042: 2.03439195347, 2043: -1.31787493284, 2044: 4.0643400144, 2045: 9.00734274344, 2046: 4.75084966324, 2047: -9.74903053991, 2048: 0.771439088115, 2049: -4.08287342364, 2050: -27.2478736892, 2051: -0.0436681305676, 2052: -15.707450234, 2053: 0.258894936207, 2054: 0.168660222224, 2055: 0.532820320137, 2056: 5.42681452472, 2057: 0.240993211298, 2058: 0.115999993954, 2059: 3.1424805476, 2060: 0.148605281843, 2061: 6.08126309707, 2062: 0.349574135349, 2063: 0.242112273775, 2064: 11.8238053381, 2065: -0.471392617057, 2066: -6.86331100817, 2067: 10.9363459419, 2068: -2.67958234176, 2069: -22.0369736664, 2070: 8.72258050163, 2071: -25.3609389012, 2072: -4.19699854245, 2073: -6.05912798203, 2074: 6.74190212775, 2075: 1.92747476307, 2076: -0.0752606362995, 2077: 6.81811934647, 2078: -0.812609060861, 2079: -0.524452146853, 2080: 9.2473829452, 2081: -18.1398969866, 2082: 4.21752137587, 2083: 10.9742672461, 2084: 0.130261006728, 2085: -9.87677270915, 2086: -0.598405553641, 2087: -4.06295905994, 2088: 6.54574382949, 2089: -1.39365380317, 2090: -10.1622653412, 2091: -11.7610463492, 2092: 0.880032244065, 2093: -8.02932474528, 2094: -1.71168279844, 2095: -0.547122113848, 2096: -51.0075653259, 2097: -5.73983767914, 2098: -9.75994812063, 2099: -9.83756070069, 2100: 15.2399755089, 2101: 25.1429841879, 2102: 16.4432559187, 2103: -12.8476998853, 2104: -7.89226364586, 2105: 34.5450020207, 2106: -9.47430345324, 2107: -9.23664664076, 2108: -9.25755941035, 2109: 22.5985494244, 2110: 4.35921691842, 2111: 1.19026263045, 2112: -4.49295933439, 2113: 7.9902221294, 2114: 5.62173492277, 2115: -12.1664479288, 2116: -23.2768176708, 2117: -15.7458077228, 2118: 16.4853440518, 2119: 2.81923055214, 2120: -6.09912542965, 2121: 9.45505286108, 2122: -2.51373439244, 2123: 5.406973762, 2124: -8.98854638178, 2125: -20.8265409678, 2126: -31.3217910314, 2127: -2.67646992678, 2128: -0.669340806216, 2129: 0.316293811579, 2130: -12.290515017, 2131: -19.9003698179, 2132: -33.8256777524, 2133: -17.037296929, 2134: 17.0476472332, 2135: -11.0049186156, 2136: -4.04607085498, 2137: -28.5654347045, 2138: -12.7071481254, 2139: -4.90793069077, 2140: 1.04556898219, 2141: 1.06473310731, 2142: -5.45676916748, 2143: -6.73369854564, 2144: 5.29328718729, 2145: 5.41791312069, 2146: -25.155251159, 2147: -4.05841253586, 2148: 5.18920710146, 2149: -1.67649683761, 2150: -4.11561952923, 2151: -71.1368819398, 2152: -13.9140772562, 2153: 20.2485674729, 2154: 15.2463141578, 2155: -35.7639606359, 2156: -19.8057858624, 2157: -7.65385693025, 2158: -16.7667166539, 2159: 2.16082373625, 2160: -4.4433257433, 2161: 19.2206847521, 2162: -3.87004969125, 2163: 19.5399452175, 2164: -1.41081576332, 2165: 2.38691944032, 2166: 0.552566732309, 2167: -3.93410230069, 2168: 10.4891755073, 2169: -4.97188446771, 2170: -11.2935756091, 2171: -13.3384241023, 2172: -4.5527248657, 2173: -21.7626799758, 2174: -13.4221454558, 2175: -4.11084700601, 2176: -19.5017515294, 2177: -3.87802826749, 2178: 14.9121302019, 2179: 7.85682588208, 2180: -6.51195705684, 2181: 7.32648920715, 2182: -16.2437891187, 2183: -3.79464703909, 2184: 7.5673360459, 2185: -2.50560482962, 2186: -3.77875971656, 2187: 11.9343282071, 2188: 3.74780784255, 2189: -3.87507775378, 2190: -2.58976123879, 2191: 8.76169938184, 2192: -3.4691540129, 2193: -4.27914690802, 2194: -0.764692167272, 2195: 15.4359135154, 2196: -4.74652472466, 2197: -1.98951331243, 2198: -3.67504252695, 2199: -6.35920386702, 2200: 10.4291585708, 2201: -3.85027552079, 2202: -3.81797428784, 2203: -2.68417682308, 2204: 15.4504342679, 2205: 5.10091928338, 2206: 17.4081331572, 2207: -37.4170667241, 2208: 9.85587963497, 2209: 5.63808722763, 2210: 3.70618777694, 2211: 3.27353444519, 2212: -11.5706159205, 2213: 6.96930106116, 2214: -0.0775283937462, 2215: 5.11846471013, 2216: -20.4549491859, 2217: -15.5121789372, 2218: 12.2192344909, 2219: -32.1863125642, 2220: -0.737887803332, 2221: -0.4889306583, 2222: 4.6276697824, 2223: 29.6993417611, 2224: -0.84589745403, 2225: -3.15852500773, 2226: -4.64469628197, 2227: 8.0547999146, 2228: -3.58244452788, 2229: -0.239818178185, 2230: 0.838357355991, 2231: -1.95855676791, 2232: -24.3603775409, 2233: -0.846983744312, 2234: -0.353658938297, 2235: 0.908503637312, 2236: -1.69151221924, 2237: -27.4351063566, 2238: 1.02902887174, 2239: 2.33870983938, 2240: -1.13208823916, 2241: -0.704356058669, 2242: -7.1816480193, 2243: -0.692722908701, 2244: -3.95653436169, 2245: -3.88719082642, 2246: -1.62168124246, 2247: -0.759982544204, 2248: -0.837536410772, 2249: -0.510359687789, 2250: -3.35518099053, 2251: -2.44883623441, 2252: -0.643523816644, 2253: -0.72022676402, 2254: -1.78535321989, 2255: -2.62908275488, 2256: -5.4177332375, 2257: 4.68545283962, 2258: 5.30372165694, 2259: 15.1627852443, 2260: -32.0216620701, 2261: 10.0038804266, 2262: -2.11329520057, 2263: 1.32812144915, 2264: -4.57096088154, 2265: 18.4032686181, 2266: -0.665318547526, 2267: -30.3482533646, 2268: -0.801280245748, 2269: 12.6419855448, 2270: -0.421895474224, 2271: -12.0087501712, 2272: 0.884748515411, 2273: -0.825570383142, 2274: -0.456643365719, 2275: -2.76350656108, 2276: -2.17395783005, 2277: 0.782054201137, 2278: -20.0207591926, 2279: 13.9307389106, 2280: 58.0226102252, 2281: 3.30697226558, 2282: 2.40796496216, 2283: 7.82294959017, 2284: -5.22694739931, 2285: 16.911796003, 2286: -4.0531294551, 2287: -0.72086309451, 2288: -40.9008620852, 2289: -3.07216664558, 2290: -1.71055731092, 2291: 16.7803480319, 2292: -10.1390605413, 2293: -0.667371851529, 2294: -3.47759949886, 2295: -0.766946511592, 2296: -28.8820845248, 2297: -0.728744705788, 2298: -0.635460123202, 2299: -16.0972811154, 2300: -3.99849978272, 2301: 10.6657570941, 2302: -0.855302604184, 2303: -0.840694084844, 2304: -1.24628821403, 2305: -0.879293309219, 2306: -3.1605130374, 2307: 7.92798246049, 2308: 16.7244805637, 2309: -0.446732867246, 2310: 0.558538803307, 2311: -1.99823101384, 2312: 9.18743906761, 2313: -6.13421244218, 2314: 5.73194138236, 2315: 3.6866567343, 2316: 10.6340358353, 2317: 4.96503943395, 2318: -4.02982307532, 2319: 19.6403777705, 2320: -0.874519875614, 2321: 10.1689756373, 2322: -30.8768589018, 2323: -4.98059162876, 2324: -0.138876149951, 2325: -30.7110011181, 2326: -0.684897056574, 2327: -15.8141515455, 2328: 23.6609504896, 2329: -3.5543265074, 2330: -12.8447853465, 2331: -2.41705031989, 2332: 18.5714346305, 2333: -3.95705137267, 2334: -3.2637061764, 2335: -10.2606471728, 2336: -8.99775839653, 2337: 19.3492246508, 2338: -14.5803437207, 2339: 9.66179621162, 2340: -6.56501786958, 2341: -0.483379598794, 2342: -10.9125897089, 2343: -11.3397063578, 2344: -2.28995496532, 2345: -4.73827882607, 2346: 36.5866215168, 2347: -23.5822505276, 2348: -5.44208308916, 2349: 3.83760799776, 2350: -3.51166043787, 2351: 23.21042023, 2352: -3.68114828936, 2353: -3.90329536158, 2354: -3.77829565731, 2355: 27.5965303885, 2356: -3.84817352503, 2357: 0.294080986864, 2358: -8.62619834172, 2359: -3.12347862366, 2360: -2.6666373363, 2361: 11.5280410079, 2362: 11.5017082118, 2363: -15.5576757547, 2364: -3.6064801111, 2365: -4.56290560257, 2366: -33.3575648665, 2367: 5.21489401016, 2368: -16.9855727124, 2369: -47.6333652825, 2370: 16.5064545239, 2371: 13.4506676496, 2372: -12.4745728619, 2373: -4.77031743755, 2374: 52.4639737285, 2375: -7.12977156622, 2376: 7.38792919846, 2377: 5.69168911959, 2378: -36.9926012825, 2379: 7.19312359314, 2380: 4.99814741788, 2381: -3.4779973839, 2382: 7.45643791148, 2383: 18.9854183928, 2384: 1.12228470654, 2385: -15.5925873235, 2386: 32.5001588444, 2387: 21.0285056109, 2388: -12.0891350591, 2389: -22.8511285672, 2390: 41.006094189, 2391: 11.0688633621, 2392: 5.48985584696, 2393: 12.1159604202, 2394: 24.4106510638, 2395: -5.47397710117, 2396: 31.1544884417, 2397: 4.18230249274, 2398: -1.70389112994, 2399: 18.2645951278, 2400: -3.76364736243, 2401: 27.4381034191, 2402: -3.57995514148, 2403: -3.52524679911, 2404: -1.17560209624, 2405: 10.5149677472, 2406: -3.97038730128, 2407: -4.0191201746, 2408: 24.7878092754, 2409: -3.9595365254, 2410: -29.5710332689, 2411: -3.66293300625, 2412: -4.05434634784, 2413: 13.7487862813, 2414: -3.72632397408, 2415: -3.26772547703, 2416: -1.6080254652, 2417: -8.30501554684, 2418: 18.6742827235, 2419: 3.52043710847, 2420: 28.6442895908, 2421: 16.1891352371, 2422: 21.478166295, 2423: 35.4568698109, 2424: 6.59763631425, 2425: 0.692917162628, 2426: 7.64938885836, 2427: -4.60487166094, 2428: -15.7419940587, 2429: -6.71506902649, 2430: 27.4023875201, 2431: 15.5993265756, 2432: 12.0839882655, 2433: -3.69646643281, 2434: -10.8463232524, 2435: -0.314156785131, 2436: -0.0502373717868, 2437: 2.31910264633, 2438: -34.8964882446, 2439: -0.350563059106, 2440: 32.879480379, 2441: -42.3817981019, 2442: -12.6692861207, 2443: 61.5963776174, 2444: -0.204589536072, 2445: -27.6073475583, 2446: 32.1565156615, 2447: 97.1672445749, 2448: -86.7934785532, 2449: -38.3958503516, 2450: 57.0102341856, 2451: 16.048493434, 2452: 18.7858429918, 2453: -17.054076359, 2454: -46.6531055061, 2455: -17.0328052024, 2456: -49.90805402, 2457: -30.1413957107, 2458: 61.1404836244, 2459: 72.5293773969, 2460: 53.17044875, 2461: 73.5745516303, 2462: 54.7115510352, 2463: -9.40788024704, 2464: 41.7849037293, 2465: -2.15021524703, 2466: 11.3483130625, 2467: -35.9179731321, 2468: -61.1244280527, 2469: -2.54242210995, 2470: 51.115875487, 2471: 81.0619568695, 2472: 4.09195652862, 2473: -36.2144559301, 2474: -31.9253470972, 2475: 43.556955512, 2476: -1.12349441016, 2477: -24.1847027764, 2478: 9.00429865687, 2479: -33.5323747132, 2480: 9.64316152636, 2481: -9.18730817103, 2482: 13.525600838, 2483: -4.44841314708, 2484: 31.1620887715, 2485: -0.196657118872, 2486: 6.380106355, 2487: -26.4122414229, 2488: 24.8935829447, 2489: 8.9596715945, 2490: 4.9398350774, 2491: 1.44575792896, 2492: 40.7112918152, 2493: 0.685012356377, 2494: 2.30317672495, 2495: 2.16004767036, 2496: -0.433108144997, 2497: -0.0403918973684, 2498: -6.04905369606, 2499: -0.29616734609, 2500: 60.2703065476, 2501: -0.778612251483, 2502: 3.39225259005, 2503: 14.2675380743, 2504: -76.2370176398, 2505: -69.3307810553, 2506: -4.31462922019, 2507: -5.36936092254, 2508: 58.9686351779, 2509: -0.466771335489, 2510: 1.28150855763, 2511: 0.119101376178, 2512: 16.3999859879, 2513: -0.954440637598, 2514: -0.469384098031, 2515: -4.0222482003, 2516: -0.216392288717, 2517: -71.3183945614, 2518: -5.48143115297, 2519: 0.5676686768, 2520: -2.8266687952, 2521: 0.734324088254, 2522: -34.3510994172, 2523: -43.2254470617, 2524: 2.25290434705, 2525: -0.376051031409, 2526: -0.366946518435, 2527: -1.18610552283, 2528: 5.68773912767, 2529: -0.107360178038, 2530: 1.79357674501, 2531: 9.65753215193, 2532: -0.201113618263, 2533: 0.869433598164, 2534: 0.00986987529699, 2535: -0.412090485268, 2536: 2.00423567826, 2537: -0.221261790623, 2538: -0.390120729299, 2539: -0.0657155233819, 2540: 41.3428874812, 2541: -0.231763647015, 2542: -0.833922388307, 2543: 21.8608829202, 2544: 30.5339734605, 2545: -0.383417487258, 2546: 4.72049697701, 2547: -0.204891891674, 2548: -0.0671878619535, 2549: 97.8205557769, 2550: 0.746794204713, 2551: -1.06803905976, 2552: -0.360024799141, 2553: -2.15563801521, 2554: -0.0137404127168, 2555: 65.8048557728, 2556: 21.3655012, 2557: -32.3250367346, 2558: 69.8085176276, 2559: 4.88961700032, 2560: -14.4956604975, 2561: 0.183376206827, 2562: -56.2936522836, 2563: 1.86744735846, 2564: -6.43094070648, 2565: -3.79570615181, 2566: -44.2789434432, 2567: 19.908090702, 2568: 2.09598262819, 2569: -0.491408358802, 2570: -0.792061669775, 2571: 0.207553445982, 2572: 4.09450777092, 2573: -0.209751013675, 2574: 0.850822388821, 2575: -2.10987275526, 2576: 28.0525077735, 2577: -1.20342509882, 2578: 1.58628302424, 2579: -1.22773848562, 2580: -1.26261835789, 2581: 3.0039213305, 2582: -8.00684303797, 2583: 13.4803828506, 2584: -21.4530632512, 2585: -0.0192522165608, 2586: 51.7875523681, 2587: -1.03705246541, 2588: 2.93744529127, 2589: -0.249355486391, 2590: -0.187013769829, 2591: -68.0153013247, 2592: -0.189623508336, 2593: -0.559399303716, 2594: -28.3411117162, 2595: 1.12811418376, 2596: -0.216812617376, 2597: -0.430509165538, 2598: -0.0771341903601, 2599: -0.303871335271, 2600: 10.2808889012, 2601: 1.89321314778, 2602: -0.763044333081, 2603: -0.219679421887, 2604: -0.391469465817, 2605: -0.0512718765498, 2606: -6.86655505396, 2607: 10.6653491065, 2608: -65.9150987607, 2609: 40.6655738914, 2610: 9.32356234398, 2611: 7.59708800872, 2612: 35.8263750204, 2613: 22.6110493607, 2614: 2.37832959353, 2615: -0.324700896138, 2616: -28.1944216159, 2617: 0.198586889741, 2618: 2.49588868216, 2619: 12.9224424664, 2620: 22.8886485275, 2621: 0.245169108818, 2622: -0.249045044859, 2623: 0.0494843358377, 2624: -0.00386687041169, 2625: -0.304416120257, 2626: 3.33620036252, 2627: 9.46117220256, 2628: 8.77279779684, 2629: 24.3094550112, 2630: -6.2648389723, 2631: -17.5307311905, 2632: -7.24511255959, 2633: -4.08901006815, 2634: -14.768859881, 2635: 19.2127959076, 2636: -0.519149283377, 2637: 8.31018633649, 2638: 0.185697852355, 2639: -1.30755667326, 2640: -0.301873868629, 2641: -5.90787651495, 2642: -0.20478104178, 2643: -0.381313904487, 2644: -0.0342155938969, 2645: 0.914268244908, 2646: -0.0377770915441, 2647: -0.0604832096141, 2648: -5.22756817543, 2649: -0.263721780017, 2650: 2.50478170311, 2651: -0.248755013998, 2652: -0.116285439033, 2653: -0.058529765015, 2654: -0.198100949791, 2655: -0.0390385567904, 2656: 16.4294594197, 2657: -34.6520018284, 2658: -0.141203700392, 2659: -1.2215801289, 2660: -21.6518798983, 2661: 1.58374529436, 2662: 6.3232213822, 2663: -64.8108798187, 2664: -5.81291115278, 2665: -2.23512834143, 2666: -1.80015398583, 2667: -0.318928337545, 2668: -0.422147705744, 2669: 0.277882132587, 2670: -10.4386463283, 2671: -0.716970272099, 2672: -0.0210957455481, 2673: -1.12198373158, 2674: 7.45220453479, 2675: 2.24420322096, 2676: 8.53976547508, 2677: 3.44621700293, 2678: 0.455783651162, 2679: -1.15560432991, 2680: 56.5678398946, 2681: 4.89199717136, 2682: -0.111841303811, 2683: 0.777977330581, 2684: 3.1984283747, 2685: 18.5453622222, 2686: 11.8580026053, 2687: -15.7673764144, 2688: -24.1178883713, 2689: -28.2401263072, 2690: -28.6149289017, 2691: 31.9891066574, 2692: -4.6875654191, 2693: 37.1380981342, 2694: -66.9292843597, 2695: -30.4256165946, 2696: -0.108772350591, 2697: 27.883216289, 2698: -1.01038755028, 2699: 1.38636166918, 2700: 51.3200066845, 2701: -0.412438272413, 2702: -0.644430403304, 2703: -42.4396983957, 2704: -8.64313652746, 2705: -0.237628556934, 2706: 13.3897664163, 2707: -0.341950637279, 2708: -0.0660490877729, 2709: 33.4701427523, 2710: -87.20190999, 2711: 16.7297342215, 2712: -0.0746838129726, 2713: -0.516740087795, 2714: 0.231065535492, 2715: -11.1204086398, 2716: -17.294584894, 2717: -3.20491568175, 2718: 5.93973271326, 2719: -8.40415518045, 2720: 16.4444055416, 2721: -37.1019886245, 2722: 40.955780917, 2723: 3.48065030593, 2724: -0.221218538386, 2725: -0.368825676594, 2726: 1.86622978263, 2727: 14.4339466987, 2728: 15.7268592202, 2729: -24.1647110394, 2730: -0.0768485538889, 2731: 5.16605625615, 2732: -0.389983447883, 2733: 1.0859802829, 2734: 15.382604227, 2735: 35.7441817136, 2736: -85.6119782965, 2737: 65.5758729457, 2738: 22.5501822068, 2739: -28.446487519, 2740: -36.3483633793, 2741: 35.2006005145, 2742: -9.563495723, 2743: 8.62456828185, 2744: -0.0664899202236, 2745: -1.21146552919, 2746: -0.0464136512044, 2747: -25.3144802247, 2748: 0.458091713241, 2749: -0.507463839449, 2750: -4.94655562556, 2751: -0.33699213885, 2752: -0.316451901475, 2753: -0.86367283451, 2754: 2.76850088009, 2755: -0.435611270984, 2756: -0.544652447411, 2757: -4.08109855769, 2758: -0.268864252176, 2759: 4.01269209758, 2760: -0.104062440865, 2761: -0.197129422435, 2762: 2.35832198082, 2763: -0.359391536428, 2764: 0.114051587689, 2765: 29.8619020414, 2766: -35.8261071011, 2767: -31.1419022164, 2768: 36.1460823053, 2769: -21.8435535659, 2770: -15.3865761669, 2771: 1.77077397257, 2772: -1.43032119511, 2773: 38.883699998, 2774: 67.9078724991, 2775: -30.7201865905, 2776: -0.312577307701, 2777: -0.00128281326984, 2778: 0.610339344196, 2779: -18.320981485, 2780: -0.798889218757, 2781: 2.89732549952, 2782: -0.621037940995, 2783: 4.48933071747, 2784: -56.7782106579, 2785: -8.44201531697, 2786: 54.147821302, 2787: -2.34169039043, 2788: 38.8976798604, 2789: -10.6484338297, 2790: 17.8845240644, 2791: -15.6227559109, 2792: -13.988673066, 2793: 0.701638452525, 2794: 33.6639312388, 2795: -15.0673315151, 2796: -1.86537814947, 2797: 17.6177747555, 2798: -9.13022154844, 2799: 1.93075286673, 2800: -32.4962845986, 2801: 12.8674444867, 2802: 5.04489886399, 2803: 2.52325481415, 2804: -1.56711032549, 2805: 0.747387093587, 2806: 4.40249062692, 2807: -2.34207228254, 2808: -4.1842096116, 2809: 7.58288223946, 2810: 5.75568646609, 2811: 0.794005024943, 2812: 5.02199945249, 2813: 10.132593489, 2814: -23.49349232, 2815: -16.9619457994, 2816: -0.728092689623, 2817: -10.1750299684, 2818: -12.2103474024, 2819: -19.8384491827, 2820: -16.3155512566, 2821: 18.24370511, 2822: -4.91159931625, 2823: -6.3049940581, 2824: -2.09965706545, 2825: -3.67606360043, 2826: -14.6737981414, 2827: -6.31380674844, 2828: -2.37597915143, 2829: 3.74347474852, 2830: -1.06533132729, 2831: -13.4637609769, 2832: -35.0285747296, 2833: 8.03191614097, 2834: 4.70341629288, 2835: -1.98225209015, 2836: 9.21456826323, 2837: 0.917459484686, 2838: -4.04136986179, 2839: -12.0846038205, 2840: 3.30879118818, 2841: -5.12921658729, 2842: 0.966405513908, 2843: -5.27316917596, 2844: -2.69045701494, 2845: 5.00226680383, 2846: 3.35027099657, 2847: -17.3550885499, 2848: 0.725255627963, 2849: -10.7244707478, 2850: 4.64625622664, 2851: 7.57849667794, 2852: -35.1880105859, 2853: 31.5527119678, 2854: -3.76675277155, 2855: 10.1546666088, 2856: -13.4959418946, 2857: 10.3217961358, 2858: -7.74971375881, 2859: 16.9885961681, 2860: 3.1204010034, 2861: 5.48092241199, 2862: 4.4106709617, 2863: 3.17889590854, 2864: 4.23127629456, 2865: 4.71906837838, 2866: 7.97150847231, 2867: -1.89228852306, 2868: 5.17299446311, 2869: 5.73798980936, 2870: 4.03634866419, 2871: 23.6430677489, 2872: 8.19652696866, 2873: 2.0392025617, 2874: -4.27250982324, 2875: 4.58731041933, 2876: 3.7083399765, 2877: 8.96744739554, 2878: 4.98177810557, 2879: -4.52870995469, 2880: 14.2589024527, 2881: -0.212550462202, 2882: -5.49886760113, 2883: 5.80316823652, 2884: 4.25826595083, 2885: -19.085950632, 2886: 4.8014133733, 2887: 4.79537123858, 2888: 1.07861416113, 2889: -3.33313039077, 2890: 4.29792842301, 2891: 4.34063550976, 2892: 3.8630666848, 2893: 21.1163599423, 2894: 3.96440820602, 2895: 1.05599844972, 2896: -7.66234374094, 2897: 4.70613446415, 2898: -7.71963615165, 2899: 4.12004683268, 2900: 4.70026775906, 2901: -0.015870748486, 2902: -29.8804743717, 2903: 4.01093939872, 2904: -9.53583437935, 2905: 23.002210125, 2906: 0.413867791047, 2907: -27.9140078419, 2908: 5.0329652895, 2909: -9.23300256759, 2910: 25.2216692226, 2911: -16.1973171935, 2912: 23.1025855929, 2913: 22.1837364833, 2914: -17.6407873803, 2915: -3.79774194123, 2916: -21.4787525615, 2917: 6.13761898099, 2918: 0.760602408902, 2919: 0.922854927453, 2920: 1.96521034319, 2921: 17.5211881574, 2922: 0.977004858872, 2923: 4.14523667006, 2924: 0.846335611933, 2925: -4.74854696244, 2926: 1.74062465544, 2927: 0.986740199962, 2928: 1.25941140713, 2929: 1.33839352009, 2930: 11.6064357292, 2931: 1.27914564979, 2932: 6.29617910382, 2933: -6.40705640811, 2934: -6.08428841929, 2935: -3.79136481782, 2936: 1.06131512691, 2937: 1.21329349603, 2938: 0.902931899259, 2939: 1.00497160568, 2940: -3.15474149172, 2941: 0.970379671919, 2942: 4.69689591337, 2943: 23.3467657344, 2944: 2.04804698029, 2945: 0.719413818862, 2946: 1.07418946298, 2947: 0.718948308277, 2948: 4.23556121268, 2949: -19.5315535696, 2950: 0.738683177075, 2951: 3.61994468034, 2952: 3.3643709728, 2953: -2.25101636409, 2954: -1.42846620391, 2955: -7.7521429615, 2956: 6.63480234524, 2957: -6.21656608838, 2958: 17.4925882482, 2959: 1.22372362828, 2960: -0.615503256829, 2961: -23.4491162962, 2962: 8.74708408293, 2963: -22.140849375, 2964: 1.13684199463, 2965: 2.04852818271, 2966: 0.723140452526, 2967: 8.5397811283, 2968: -22.6221018052, 2969: -5.39850889791, 2970: 0.500127088045, 2971: 0.903215327613, 2972: 1.02145084315, 2973: 1.03132090056, 2974: 1.17104349229, 2975: -11.7883599811, 2976: -23.2342233137, 2977: 2.67712720758, 2978: -22.5603634053, 2979: -15.3753490552, 2980: -3.95644984422, 2981: -3.58773090295, 2982: 10.6674170189, 2983: -9.30026982432, 2984: -3.02558537386, 2985: 0.793244796906, 2986: 21.7976815837, 2987: 0.935702011738, 2988: 1.02241410714, 2989: -0.249688148717, 2990: 12.6364145258, 2991: 0.934676701325, 2992: 4.61294653914, 2993: 0.979091019614, 2994: 15.6245549469, 2995: 0.887508946949, 2996: 0.882897012483, 2997: 4.23688212202, 2998: 5.743478541, 2999: 6.66900871072, 3000: 0.979845079447, 3001: 0.80206961467, 3002: -0.513884299506, 3003: 0.734230121603, 3004: 4.44660633271, 3005: -5.32923410113, 3006: -25.9973711215, 3007: -0.356638335661, 3008: 1.31328567064, 3009: 3.74910222127, 3010: -8.08911534952, 3011: -1.07070232212, 3012: -23.6932635803, 3013: -7.81422809919, 3014: 5.98752424587, 3015: 0.379664238758, 3016: 0.008505214549, 3017: -8.81398525638, 3018: 2.28576070865, 3019: 25.8951754049, 3020: -8.82977063402, 3021: 4.19960493653, 3022: -1.99852841734, 3023: -19.3688184563, 3024: 0.660498657446, 3025: -15.342995345, 3026: 4.3286052313, 3027: 4.8543191731, 3028: 4.64160194774, 3029: -2.49089107722, 3030: 14.1971456195, 3031: 1.51184448664, 3032: 4.11974151902, 3033: -2.12750678116, 3034: -6.7070637742, 3035: 1.84199425728, 3036: -3.39449206604, 3037: 2.16348998378, 3038: -9.5396785492, 3039: 9.38469536867, 3040: -12.6813816169, 3041: 23.7712284715, 3042: 9.40214578245, 3043: -16.5054665016, 3044: 20.5765829268, 3045: 12.601521744, 3046: -7.65310175677, 3047: 3.92191193432, 3048: 1.47999634804, 3049: 4.56692068186, 3050: -3.36362138499, 3051: 4.87430693548, 3052: 7.8269958507, 3053: 24.4798390532, 3054: 4.65298057341, 3055: 4.81131225484, 3056: 10.8274249161, 3057: 4.09532226711, 3058: 38.1068965835, 3059: -3.44429820233, 3060: 3.2816505095, 3061: -15.7288971535, 3062: -4.10414824496, 3063: -5.61639787387, 3064: -26.1969348765, 3065: 9.42414711441, 3066: 0.179359573924, 3067: -12.8037121312, 3068: 8.18094400963, 3069: 6.92254424251, 3070: -42.3164458424, 3071: 19.1607562475, 3072: 26.9482240991, 3073: -0.239456829082, 3074: 2.67257102734, 3075: 1.08392521142, 3076: -18.4628875993, 3077: -1.39145498359, 3078: -7.88508964918, 3079: 4.97181219343, 3080: -0.471724926069, 3081: 3.06765590202, 3082: 0.51797836679, 3083: 0.526716474453, 3084: 2.00790292436, 3085: -7.4696169847, 3086: -17.3192629959, 3087: 14.1768621475, 3088: 0.810928325486, 3089: 7.02319717833, 3090: 6.89555546832, 3091: -11.6535366563, 3092: 22.847070198, 3093: 8.59072799594, 3094: -29.2079616693, 3095: 4.98527756388, 3096: 2.98817785419, 3097: -19.2920993691, 3098: 4.85782686877, 3099: -50.3648983783, 3100: 4.67420105867, 3101: 4.69530036952, 3102: 3.71597535346, 3103: 2.33440780779, 3104: 6.10852684626, 3105: 4.19077840839, 3106: 14.4351079938, 3107: 5.77974536135, 3108: -9.02017668888, 3109: 0.275939856439, 3110: 4.79560247469, 3111: -2.92274903897, 3112: 3.88054023376, 3113: 4.94370148821, 3114: -12.6211669946, 3115: -10.9485958628, 3116: 25.2240956437, 3117: 12.2618573216, 3118: -9.68581031645, 3119: -3.44401415321, 3120: 8.15048934169, 3121: 14.7001989694, 3122: -3.41316190727, 3123: -12.0302346329, 3124: 1.65441746654, 3125: -2.15282874322, 3126: -14.3078363627, 3127: -3.45874155667, 3128: -23.1551830044, 3129: 6.43244627037, 3130: -2.89412950587, 3131: 4.59709154626, 3132: -35.2090225038, 3133: 3.16476591, 3134: 7.58739270638, 3135: 4.22315897405, 3136: -28.2121714222, 3137: -16.8345617826, 3138: 1.6952724685, 3139: -12.8148425735, 3140: 10.8782280865, 3141: -0.978965363546, 3142: 0.776314140717, 3143: 14.3766441169, 3144: -4.73381602363, 3145: 2.93969904287, 3146: -8.50954552524, 3147: -9.70689913522, 3148: -1.07255169528, 3149: 4.83699320634, 3150: 8.75482838903, 3151: -1.05369574396, 3152: 3.32236897036, 3153: -1.79189408478, 3154: -6.86365929697, 3155: -2.52761064121, 3156: 1.30279240703, 3157: -8.41782667581, 3158: -3.98909084443, 3159: 0.614681240579, 3160: -12.7734532442, 3161: -0.0383528580626, 3162: -1.44339024742, 3163: 2.43193409178, 3164: 4.71519035094, 3165: 4.66764873838, 3166: 3.21794911769, 3167: -1.89887268729, 3168: -5.2136253242, 3169: 3.73550247149, 3170: -7.83651735493, 3171: 2.06955116382, 3172: 0.661576683359, 3173: -0.0939513002256, 3174: 3.33498429887, 3175: 1.55554401534, 3176: -3.55273438621, 3177: 1.2130841816, 3178: 2.36384046261, 3179: 0.507927763229, 3180: -2.65969790257, 3181: -0.14867849686, 3182: 0.484080146946, 3183: 0.436674897565, 3184: 2.62994258658, 3185: -1.67022079752, 3186: -1.39764653371, 3187: -0.243900549675, 3188: -3.16429267632, 3189: 0.440332521016, 3190: -1.24326959778, 3191: 0.119562114218, 3192: -0.525038795771, 3193: -11.3567321796, 3194: 3.32224963741, 3195: 0.108291756005, 3196: -3.33483685014, 3197: 1.63496720041, 3198: -5.37031106644, 3199: 2.78572896318, 3200: 7.07104914818, 3201: 0.594059150613, 3202: 14.7486029474, 3203: 0.23495341781, 3204: -2.5569952723, 3205: -11.0097733852, 3206: -1.6614222814, 3207: 1.22690010752, 3208: -2.51936726862, 3209: 1.38212904882, 3210: -4.12105594484, 3211: -1.16561191312, 3212: -0.306141727219, 3213: -1.75364209274, 3214: 2.11767157022, 3215: -10.424704441, 3216: -1.30096186563, 3217: -0.225696546959, 3218: 0.00482022243677, 3219: -2.0040639446, 3220: -6.38228272681, 3221: 1.06674293293, 3222: -1.67410907442, 3223: -2.73628514947, 3224: 1.01413262562, 3225: 2.18219680828, 3226: -5.63718889921, 3227: 0.533593141983, 3228: 1.73778186219, 3229: 3.29703640546, 3230: 1.46668406492, 3231: -1.13584449104, 3232: 1.60441519417, 3233: 1.41538513099, 3234: -20.4207874325, 3235: -2.77783131282, 3236: 0.667337437703, 3237: 1.39404756342, 3238: -1.04315081305, 3239: 0.792419569012, 3240: -0.257931509029, 3241: -1.72551199435, 3242: -4.34580837078, 3243: -3.94892456703, 3244: -2.47690291631, 3245: 1.14619093362, 3246: 0.826499681212, 3247: -6.24998213655, 3248: 1.16710885307, 3249: -1.11396688699, 3250: 0.0737307929189, 3251: -2.43204701048, 3252: -2.62707867648, 3253: -0.304582410302, 3254: 12.6861966336, 3255: 4.65531597168, 3256: -2.2215624123, 3257: 0.551827909524, 3258: 2.82217261929, 3259: 2.55121251649, 3260: -13.9279958845, 3261: -0.400945046169, 3262: -1.24605129294, 3263: 5.61143443303, 3264: 1.9589838457, 3265: -2.57080994932, 3266: 9.62405538641, 3267: 0.747233660839, 3268: 0.483658110382, 3269: 0.0555641888606, 3270: -5.1518040458, 3271: 0.834116216868, 3272: 1.30136000161, 3273: 1.26010993982, 3274: 1.71873209906, 3275: 1.53826573825, 3276: 1.47837568536, 3277: 4.27701373595, 3278: 0.799233646577, 3279: -0.0663774299241, 3280: 0.442066615237, 3281: 0.470182675606, 3282: -0.810204028318, 3283: 15.6033250018, 3284: 6.59056301341, 3285: -0.214666594867, 3286: -7.10188688773, 3287: 0.83868995468, 3288: 0.853979510152, 3289: -1.05375407491, 3290: 0.71968695597, 3291: 1.42172401222, 3292: -2.82171280025, 3293: 1.09994868069, 3294: 0.903136943208, 3295: 0.63637020607, 3296: 0.675242212392, 3297: -0.0537006406005, 3298: 4.42946340407, 3299: 1.13032891349, 3300: 0.709957389791, 3301: 1.79473067841, 3302: -1.26512340087, 3303: 1.26153404519, 3304: -1.77203127039, 3305: -2.63417524567, 3306: -22.2588826455, 3307: -6.06771847578, 3308: 1.16738980478, 3309: 0.232004471772, 3310: 2.24761402024, 3311: 1.04512562408, 3312: 8.66610021101, 3313: 0.980477665152, 3314: 8.20437802342, 3315: 0.919313424553, 3316: -3.93231529554, 3317: -3.45592441938, 3318: 0.970270801661, 3319: 0.805457779933, 3320: 0.754882066147, 3321: 0.896505501297, 3322: 1.09880407942, 3323: 0.761160645313, 3324: 0.230852299896, 3325: 6.33728719481, 3326: 1.60807244842, 3327: 6.85066186504, 3328: 5.36623532961, 3329: 2.57090751334, 3330: -2.2575655374, 3331: -0.637795342471, 3332: 1.48000096454, 3333: -14.0028132743, 3334: 0.753246726306, 3335: -4.71349712441, 3336: -2.86003838748, 3337: -2.14887902304, 3338: 2.04826196441, 3339: -4.44265076346, 3340: 0.786389695237, 3341: 0.52647488223, 3342: 0.831342874098, 3343: -3.74422344205, 3344: 0.660147669266, 3345: 0.623308501449, 3346: -15.0863598813, 3347: 1.55061123262, 3348: -18.2262170006, 3349: 0.878922167479, 3350: 0.840991464082, 3351: 0.781740381512, 3352: 0.922809329569, 3353: 1.26111079017, 3354: -0.14143724947, 3355: 1.22449382775, 3356: 0.475468500846, 3357: 0.560713293318, 3358: 5.74233619585, 3359: 3.44314094859, 3360: 1.18945424088, 3361: -6.36488178988, 3362: -2.104415439, 3363: -3.38454209834, 3364: 0.186013312708, 3365: 1.28272441205, 3366: -1.27749452147, 3367: 1.22132624218, 3368: 6.62052058269, 3369: 0.700880768479, 3370: 3.28897536079, 3371: 1.22334634858, 3372: -4.32361733804, 3373: 0.723808337994, 3374: 5.18010065603, 3375: 1.48674836428, 3376: 0.0919890612323, 3377: 0.090343461493, 3378: -3.19768274054, 3379: -3.41935928722, 3380: 1.78779651949, 3381: 1.11865901154, 3382: -4.59081313119, 3383: 2.15430376899, 3384: 1.52867924543, 3385: -24.3983948035, 3386: -2.55388527039, 3387: -6.08636945829, 3388: -2.39968931404, 3389: -7.46701544059, 3390: 14.3688590872, 3391: 6.88021936181, 3392: -8.82840604414, 3393: 0.211415728654, 3394: -0.8883530318, 3395: 2.83492803178, 3396: 2.28444585927, 3397: 0.322991517079, 3398: -7.98892815534, 3399: 1.29785138902, 3400: 1.31376933586, 3401: -3.09914065548, 3402: -3.41658805127, 3403: 1.19741163006, 3404: 0.940184246768, 3405: 1.56650948793, 3406: -0.116346273154, 3407: 12.6824165542, 3408: -15.9041368866, 3409: -0.175791064759, 3410: -3.64059555292, 3411: 1.61654593805, 3412: -2.84492958637, 3413: 0.658981756939, 3414: 3.11609050868, 3415: 12.91869434, 3416: -24.9525253305, 3417: -5.40003584987, 3418: 3.46686194622, 3419: -6.12375790838, 3420: 0.840642985463, 3421: 10.7904144631, 3422: -3.66590826594, 3423: -0.887007228531, 3424: -0.755710363769, 3425: 7.97884956606, 3426: -8.19639942064, 3427: 3.00759915384, 3428: 1.60385879926, 3429: -1.53317263111, 3430: -1.28349393626, 3431: 0.899237626176, 3432: 3.87408444412, 3433: -2.52210130627, 3434: -12.887922798, 3435: -4.30920613486, 3436: 1.68725632513, 3437: -3.68897111063, 3438: 2.61862828174, 3439: -2.06639003508, 3440: 0.583427990395, 3441: -9.10578013777, 3442: 4.64909082723, 3443: 1.3939237468, 3444: 0.22386483265, 3445: 0.462545441362, 3446: -12.8743046299, 3447: 0.593011343143, 3448: -3.54480566533, 3449: 0.623021300912, 3450: 0.637963669001, 3451: 0.36139965219, 3452: 0.0711489346226, 3453: 2.56073606487, 3454: 1.47621839822, 3455: 3.22922063305, 3456: 1.560130823, 3457: 36.2582273029, 3458: 1.45607581566, 3459: 1.45821634265, 3460: 3.58086421399, 3461: -3.06127981118, 3462: 1.58360974122, 3463: 1.39658579093, 3464: -1.49426445124, 3465: 0.466944385304, 3466: 0.59856078209, 3467: 4.21575711681, 3468: -3.31883093274, 3469: -4.38610366442, 3470: -16.8134349504, 3471: -2.71796221555, 3472: -2.60548274853, 3473: -6.11794237866, 3474: 2.03196389367, 3475: -1.78422835392, 3476: -7.11345374233, 3477: 7.52020704204, 3478: -2.92573813689, 3479: 2.61448338164, 3480: 0.419640875104, 3481: 1.30178679853, 3482: 1.18193071209, 3483: -8.57799467037, 3484: 4.94260478814, 3485: 2.4674514003, 3486: -3.98253985595, 3487: 3.55565827724, 3488: -7.62548171045, 3489: -6.35831677255, 3490: -14.485294995, 3491: -0.953652060954, 3492: 1.82012037913, 3493: -1.4333283505, 3494: -2.17368138785, 3495: -0.875088027022, 3496: -1.67099745191, 3497: -1.61837794281, 3498: -0.103708191863, 3499: 0.860947860357, 3500: 1.7722317859, 3501: 2.26728020847}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.434375973982, 2: 8.43224137086, 3: -4.27366976074, 4: -5.42480025186, 5: -2.85828158146, 6: -2.46597991585, 7: 3.71813801637, 8: 3.96013653321, 9: -1.8670562931, 10: -8.76409855277, 11: -14.7219890792, 12: 4.99639280476, 13: 0.679932949996, 14: -5.06855703354, 15: -11.5103466394, 16: 2.57904960589, 17: -1.77207599748, 18: 5.57675446254, 19: 4.20831213812, 20: 2.73177805587, 21: 5.2963830673, 22: 10.0152629774, 23: 5.79898022209, 24: -7.06169650243, 25: -10.2925476059, 26: 7.6475245015, 27: -0.734431687174, 28: 2.19806900246, 29: -20.6601674682, 30: 1.84123417269, 31: 1.92871620136, 32: 0.0628080159215, 33: -0.491652784206, 34: 3.86447765529, 35: 3.97932142409, 36: 0.799513916794, 37: 4.83266648536, 38: 2.74069223206, 39: 5.50549718725, 40: 2.28859954092, 41: 3.89716385245, 42: 16.1758623915, 43: 2.07706445897, 44: 6.07507783423, 45: 7.06219199535, 46: -2.92331917909, 47: 3.66635678714, 48: 1.61066108804, 49: 2.89494833277, 50: -0.2545865586, 51: -1.21786475167, 52: 13.005773167, 53: 1.71502407267, 54: -5.42978656368, 55: -4.55112516582, 56: 1.73205192335, 57: -14.8414679059, 58: -2.05216187953, 59: -9.10770596686, 60: -10.798634248, 61: -15.5278616318, 62: 0.769322542699, 63: 1.86837117967, 64: 1.85696475756, 65: 0.736181643999, 66: 1.70323148914, 67: 4.48208137177, 68: 2.00903711038, 69: -1.14733741516, 70: 1.72409753081, 71: 1.53894208367, 72: 14.1637496, 73: 1.71748264457, 74: 1.89457775461, 75: -1.25906112749, 76: 1.26945618981, 77: -0.228286884559, 78: 2.32677147919, 79: 3.61755553237, 80: 6.01616811042, 81: 1.68213964373, 82: -8.18404286764, 83: 1.69664065415, 84: 1.35183041743, 85: -8.26798998653, 86: 2.48148525942, 87: -2.77480160983, 88: -4.52104732391, 89: 1.74166092359, 90: -1.17633261187, 91: 0.852260368246, 92: 1.60172574453, 93: -10.496440494, 94: 1.78206285999, 95: 1.72479399377, 96: 1.75506472872, 97: 2.44245204962, 98: 1.32049211438, 99: 1.57273002202, 100: 1.41602194179, 101: -20.5855838917, 102: 1.79758866188, 103: 1.62307416353, 104: 1.73679769073, 105: 2.96397182179, 106: -11.2580761806, 107: 1.56508762319, 108: 0.0391141347856, 109: 1.75962457769, 110: 2.16056434273, 111: 2.090657578, 112: -22.6191116472, 113: -1.8234748881, 114: -0.437248236728, 115: -14.7270876313, 116: 0.641774144008, 117: -2.49593965639, 118: -5.28237673681, 119: -2.00080954364, 120: -0.228493191928, 121: 2.45179836615, 122: -20.1651830529, 123: -1.4148871251, 124: -5.7712698803, 125: -15.4772222838, 126: 0.353279287395, 127: 0.233113051302, 128: 0.376246458249, 129: -9.52115709288, 130: 0.367493974976, 131: 0.0174502910395, 132: -1.25908377122, 133: 5.73748892461, 134: 0.856300647423, 135: -1.51061389096, 136: -1.83335864493, 137: 0.881063225463, 138: -1.13508223308, 139: 1.25217867377, 140: -6.59212744194, 141: -13.8772184465, 142: 1.22227804176, 143: -5.08096800897, 144: -0.624539046885, 145: -2.30132594462, 146: 0.45824997133, 147: 0.4171429965, 148: 9.60099528204, 149: 0.334825623318, 150: 1.79843053773, 151: 0.278664509828, 152: 0.432758894453, 153: 0.498300945491, 154: 0.283095008065, 155: 0.278626845785, 156: 1.93307640372, 157: 12.417020256, 158: 0.793609101748, 159: 0.696493123927, 160: 0.872062552719, 161: -0.747101255418, 162: 1.97340659022, 163: 6.17467809186, 164: -2.12996422822, 165: -4.48786879041, 166: -13.3522717108, 167: 1.32835644373, 168: -0.48400630442, 169: -1.97676046163, 170: -4.96821278954, 171: 12.074194428, 172: 0.579168909615, 173: -20.6439305388, 174: 0.426578544703, 175: 2.13442306728, 176: 16.3270113403, 177: -10.0706077752, 178: -0.349807369902, 179: 0.461683348009, 180: 0.383851837897, 181: 0.435372819828, 182: 0.326461064374, 183: -0.368715264015, 184: 10.4294015587, 185: -0.0391526823192, 186: 10.7937204659, 187: 5.55662093729, 188: 6.31309734362, 189: -0.373802275312, 190: 2.50879185074, 191: -2.27395228647, 192: -3.17565512871, 193: 0.141145784631, 194: 3.57915137353, 195: 1.1697444979, 196: 1.24627991086, 197: -2.07700081513, 198: -1.65372942833, 199: 0.429020260466, 200: 1.1245843721, 201: 0.27005336329, 202: -6.41146486809, 203: 0.257919046532, 204: 0.337275889201, 205: -11.4318980185, 206: 1.77633864954, 207: -12.253908529, 208: 0.488730834424, 209: 0.496112641601, 210: 0.970712303418, 211: 0.486257237932, 212: 1.19062421116, 213: -4.75450394798, 214: 4.46832399872, 215: 0.143878330284, 216: 1.06748234605, 217: 3.64847196572, 218: 0.317401058539, 219: -1.52720737082, 220: 1.68596714022, 221: -0.341904162086, 222: 1.3369596883, 223: -1.9372039903, 224: 1.66352298316, 225: -1.3182114463, 226: 1.11051690894, 227: -5.22294128177, 228: 1.1520309213, 229: 2.35558015097, 230: -0.263449061869, 231: 12.7874049868, 232: -0.252369544633, 233: -0.197319028292, 234: 3.92324887074, 235: -0.611139284889, 236: 2.32784467056, 237: 1.60317051215, 238: -5.37825875325, 239: 1.70737796323, 240: -0.0466987235987, 241: 8.57301815891, 242: -0.768234183643, 243: 8.07407604687, 244: -0.99607512208, 245: 1.83736059367, 246: -11.2749898468, 247: -2.49525875776, 248: 6.05734936642, 249: -3.16513798158, 250: 12.0559766849, 251: -9.22063431926, 252: 2.55819727834, 253: -2.70389902993, 254: -22.9239447734, 255: -0.138889519773, 256: 1.63273542087, 257: -10.1003833854, 258: 1.57393193505, 259: 1.64816883081, 260: 5.7506176037, 261: 0.816129270028, 262: 1.6574979918, 263: 2.32910010388, 264: 1.56286795162, 265: 2.04566471565, 266: 5.12934038701, 267: -14.7888605346, 268: -4.20197042246, 269: -7.41064093246, 270: 1.28761570252, 271: -1.05564592867, 272: -9.21074961207, 273: 1.59257328485, 274: -14.6984433572, 275: 16.323022148, 276: 0.0760938687848, 277: 0.125357275979, 278: 3.5116197941, 279: -3.48044977607, 280: -2.02358897908, 281: -3.86705205704, 282: 12.199890585, 283: -1.62756800081, 284: -10.268896625, 285: 17.739346967, 286: -0.975353317894, 287: 1.74772825021, 288: 0.467487699268, 289: 1.73239477327, 290: 0.416696098059, 291: -0.416279901296, 292: -8.73136366531, 293: -15.7822230863, 294: -2.49250642566, 295: -4.69006548337, 296: -0.0776545297424, 297: -0.584381366209, 298: -4.23560409636, 299: 0.395674014974, 300: -0.612087301211, 301: 7.43244657618, 302: 1.4198368784, 303: 1.03088723493, 304: -13.6933946993, 305: 1.26186452066, 306: 1.69962373738, 307: 4.34119621943, 308: 1.07065307236, 309: 1.03804653367, 310: 0.802550943335, 311: 11.6948632728, 312: 1.77274235254, 313: 1.65636948422, 314: 8.69234120609, 315: 1.78665182834, 316: -6.26725038586, 317: 1.50408915791, 318: 1.58937958533, 319: -3.9200937388, 320: -2.21096977639, 321: 1.38293695155, 322: -12.2581079033, 323: -0.196860836112, 324: 13.342604071, 325: -16.6184745922, 326: -19.4365829764, 327: -1.36096994404, 328: 5.66155936281, 329: 1.10287368023, 330: 1.2945745948, 331: -1.44725170244, 332: -4.73105953491, 333: 1.62935717494, 334: -2.85960368229, 335: 0.656034844632, 336: 1.36142441309, 337: 2.20547402525, 338: 7.17520472173, 339: 0.983544296122, 340: 7.535817849, 341: -4.50935389338, 342: 3.47185507262, 343: -12.2871616354, 344: 2.13992361142, 345: -0.320871439109, 346: -8.7511508462, 347: 6.1324350882, 348: 18.1132042506, 349: 3.07193122493, 350: 1.13449991967, 351: 1.94433136354, 352: 5.5581816383, 353: 5.49427823771, 354: 15.4291208527, 355: -13.2274160971, 356: -19.2628736863, 357: -3.68313442152, 358: 0.304349661981, 359: -3.59044010028, 360: 3.28657180969, 361: -38.0615075985, 362: 6.86284687467, 363: 9.96455640553, 364: 15.3757887969, 365: 10.0139413688, 366: -12.3422161312, 367: -1.28197265563, 368: 3.38985465326, 369: -10.1927448691, 370: 9.7635214486, 371: -33.0270035888, 372: 14.0285518348, 373: 2.86392293831, 374: -3.12648478242, 375: -14.8910125831, 376: -0.0461991325745, 377: 8.83660879351, 378: -18.1735930144, 379: -9.85488900511, 380: 14.6130646005, 381: -6.71676059949, 382: 0.973667379846, 383: -8.32605107114, 384: -0.0615998265134, 385: -51.9218188545, 386: -19.3530068014, 387: -2.44836934958, 388: 5.8579422411, 389: -6.9655134179, 390: -28.3270556825, 391: 6.13571559449, 392: -5.25642104116, 393: -21.9500320416, 394: -4.33454894701, 395: 11.6332557152, 396: -1.45161932094, 397: -3.37275197269, 398: 19.7153191436, 399: 0.943381942988, 400: -5.39612301658, 401: 12.9999914174, 402: 3.2491079831, 403: 3.33041376629, 404: 7.06644651004, 405: -3.44532503909, 406: 5.75944521459, 407: -3.42732323454, 408: 15.5709805141, 409: 1.21345545754, 410: -6.4521858086, 411: 13.7504438055, 412: -39.3157027358, 413: 2.50097572977, 414: -2.65502922153, 415: -2.66438560606, 416: 0.905676029236, 417: -2.19321128217, 418: -34.4381817962, 419: 6.33870993645, 420: -3.35478315774, 421: 12.6033842497, 422: 6.6274335982, 423: 21.2400531603, 424: 4.73136135951, 425: -2.47467582381, 426: -3.21665079938, 427: 1.72382461132, 428: -4.00948247574, 429: -0.385044441283, 430: -10.3871495099, 431: 3.09516193687, 432: 3.58723172576, 433: 8.46464914881, 434: -5.3900027501, 435: -2.73564887444, 436: 7.01154090238, 437: -8.94363978753, 438: 5.67248864051, 439: 5.53957614367, 440: -7.63999181008, 441: 5.60634947216, 442: -21.0940488105, 443: 1.89898743924, 444: 2.40809204313, 445: 6.07445295161, 446: -7.82336169415, 447: 9.01322342582, 448: 6.1635930754, 449: 23.1148967608, 450: -6.35380690599, 451: -4.1006369342, 452: -4.70132318302, 453: 1.24516784254, 454: 1.90918881977, 455: -40.8155972161, 456: 5.66445148072, 457: 9.74807498111, 458: 6.14616632343, 459: -15.3410009176, 460: -10.3461006557, 461: 6.88442491864, 462: -4.44461042386, 463: -24.054953154, 464: -41.70949339, 465: -21.4138534434, 466: -31.7852572937, 467: 7.73473599294, 468: 31.2424393126, 469: 19.3451649013, 470: 2.69956652634, 471: -0.358548634671, 472: 8.85274868093, 473: -12.7247845782, 474: 9.34740878773, 475: 1.50399910707, 476: 1.77557998405, 477: 2.64410563478, 478: 2.28432647041, 479: 1.22141689619, 480: 6.53992011841, 481: 1.56030780635, 482: 3.17529250577, 483: 2.36321374093, 484: 2.48096531808, 485: 8.31600480033, 486: 1.02926818459, 487: 7.30736192273, 488: 2.18300879006, 489: 1.38404523525, 490: 18.2717206732, 491: -5.68525833296, 492: 15.4073209417, 493: 1.44441978221, 494: -5.43170448757, 495: 1.31441764734, 496: 1.26028614791, 497: 5.85824495808, 498: 1.03601144228, 499: 6.72146544412, 500: 4.98105546046, 501: 1.07173609546, 502: 1.12869596314, 503: 1.60167393206, 504: 1.30463120037, 505: 8.98356097592, 506: -20.8153072184, 507: 0.814999621507, 508: 2.33746371297, 509: 1.35802881395, 510: 0.257905554859, 511: -6.46909525284, 512: -16.8872769671, 513: -0.240255180422, 514: -12.3865286763, 515: -4.00732311919, 516: 1.73817337747, 517: 2.23167099801, 518: -1.03720921381, 519: -3.15644149352, 520: -4.61701327858, 521: 1.21475375265, 522: 23.2789414906, 523: 1.18048179838, 524: 15.7028429619, 525: 2.61272504652, 526: -18.8940750433, 527: 1.31802182104, 528: 0.670622518197, 529: 1.03481848461, 530: 0.859346561322, 531: 2.02827635082, 532: -1.41512685794, 533: -3.80969439677, 534: -7.84013877708, 535: 1.11299260143, 536: -11.253133253, 537: -0.329223253084, 538: 2.64076537581, 539: -1.14787789042, 540: -6.19187634169, 541: -11.5089462317, 542: 1.30766640588, 543: 11.2623905273, 544: 0.610612730399, 545: 2.33313181185, 546: 0.0604959614011, 547: 11.2186340487, 548: 1.07476440173, 549: 5.85440826885, 550: 1.14159202834, 551: 15.4310603863, 552: 1.03360380671, 553: 1.14237093213, 554: -13.7853529535, 555: 5.47559764023, 556: -22.3390932945, 557: 0.931160516602, 558: 1.08838891724, 559: 0.392429128666, 560: 1.13734284233, 561: 5.61631617294, 562: -5.42683921216, 563: -5.09980177953, 564: 1.06211338279, 565: 1.95291176841, 566: -7.99684556092, 567: 3.31952479275, 568: 3.90274990533, 569: 14.6210547593, 570: 1.10125712593, 571: 2.05013443149, 572: 2.25528925265, 573: -1.7701361432, 574: -2.95844098289, 575: 1.1032046402, 576: 21.5480607147, 577: -4.44847533625, 578: 1.08112614279, 579: 0.868688461252, 580: 14.6483229243, 581: 0.768492347036, 582: -14.9959954547, 583: 15.4072461087, 584: 4.94316323613, 585: -2.18169982797, 586: -8.48429189464, 587: 3.24278335028, 588: -5.30726092146, 589: 6.39661176042, 590: -19.3662436372, 591: 1.40055928795, 592: -39.0695274982, 593: -21.529181018, 594: 7.66925821616, 595: 13.8946739191, 596: 11.1834517277, 597: 8.80054122305, 598: -0.843238997165, 599: -15.5378223007, 600: -20.4468563722, 601: 0.107682412877, 602: -7.98359643009, 603: -32.9951019194, 604: -2.81465516727, 605: 2.72192077375, 606: -0.534470296905, 607: 6.17591589282, 608: 6.6832121674, 609: 7.85380236308, 610: -2.46489975934, 611: 3.54249489206, 612: 1.88222171851, 613: -0.426424524909, 614: 8.89918195432, 615: -24.5108496263, 616: 6.32290881139, 617: 1.53935556905, 618: -12.6700690528, 619: -2.02551031694, 620: -4.88590162887, 621: -27.5280034453, 622: -8.94120495937, 623: -1.10835457261, 624: -2.99570585269, 625: -16.8566599204, 626: -0.483484296428, 627: 5.57665527497, 628: -3.4072824517, 629: -1.24473919088, 630: 7.00417716614, 631: -22.3162564262, 632: 22.4519677543, 633: 0.652730753029, 634: -11.3370038275, 635: -5.7615239237, 636: 12.4682355274, 637: 20.5234565469, 638: -1.55680298243, 639: -16.0948438837, 640: -4.31710412381, 641: -9.82253453559, 642: -13.239907756, 643: 2.04126123424, 644: 14.4987223793, 645: -1.26280207318, 646: 29.576540385, 647: 0.778521557744, 648: -4.85946988442, 649: -0.843795330014, 650: 5.85640907751, 651: -6.35958190554, 652: -4.6207408176, 653: -22.9071386452, 654: -0.406982054936, 655: 2.48089375295, 656: 1.86219826893, 657: 5.95725020666, 658: 5.94906495261, 659: -2.1092920209, 660: -16.1509178017, 661: 4.84278284924, 662: 4.64660687117, 663: -25.7846508451, 664: 5.46907068441, 665: 6.40831685306, 666: 6.40445467491, 667: 5.1003762724, 668: 19.6368335811, 669: 9.2184339521, 670: 5.3083598741, 671: -6.55037213941, 672: 2.05554138435, 673: -2.4846062092, 674: 2.25192220966, 675: 4.05957908462, 676: -28.8233386576, 677: 32.2859153307, 678: -25.3399469669, 679: -23.0874428484, 680: 12.1414784378, 681: 18.3515967111, 682: -17.8007979292, 683: -2.29379078845, 684: -13.0388536056, 685: -5.4173013097, 686: -29.0537364304, 687: 3.85857642327, 688: 4.22853497266, 689: 14.2360489304, 690: -14.6202053744, 691: 2.49897318249, 692: -7.11741761382, 693: 20.2519599218, 694: 4.07012153722, 695: -2.52363833353, 696: -3.64068997079, 697: 8.59961448447, 698: 5.93443343568, 699: 0.227812112506, 700: 15.7956008616, 701: 36.505860696, 702: -7.54966366155, 703: -15.1917319639, 704: 22.659001603, 705: -13.9050153639, 706: -40.3342518003, 707: 21.4098753469, 708: 6.52605890572, 709: -8.93540395944, 710: -0.676442342821, 711: -3.5497882283, 712: -50.1955056138, 713: -2.24155135525, 714: 30.7169714738, 715: -57.124779607, 716: 15.8652937428, 717: -64.1016499262, 718: 34.2291520079, 719: 14.9118118033, 720: -100.0, 721: 3.91776007327, 722: -34.8328002145, 723: -3.1376000478, 724: 47.3833057592, 725: 19.0917192991, 726: 1.70754044498, 727: -7.80118318051, 728: -19.4170503777, 729: 27.5551681192, 730: 21.8081424493, 731: 7.37712392022, 732: 5.51188938445, 733: -30.4778962663, 734: -1.83892189137, 735: 17.541816503, 736: -3.71306578906, 737: 19.6226258973, 738: 21.5062184645, 739: 1.84256382644, 740: -8.11846777094, 741: 51.342249117, 742: -46.5420568318, 743: -5.92173795599, 744: 12.5919559884, 745: 2.44663594944, 746: 0.430343993211, 747: 15.5307030772, 748: 7.53626517709, 749: -23.0068963137, 750: 40.2360691692, 751: 1.7326777178, 752: 4.43974202498, 753: -26.5252558614, 754: -1.2779897575, 755: 4.92833952317, 756: 1.41378792408, 757: -18.2375697169, 758: 5.19485398854, 759: -74.5444647927, 760: -5.24012330252, 761: 32.2601683707, 762: 41.9893681231, 763: -24.4019046476, 764: -2.59055329048, 765: -33.2335653526, 766: 0.771427053258, 767: 1.99964780467, 768: 0.722180108339, 769: 0.618828122145, 770: -2.32311444351, 771: 1.15696653863, 772: -24.6652784381, 773: 12.6919253392, 774: 2.99671870979, 775: 4.50861292591, 776: 0.80280441123, 777: -16.8474166018, 778: 56.9902934329, 779: -0.11075519204, 780: 11.1269834459, 781: 1.18182986364, 782: -9.65131907342, 783: -4.78086666673, 784: 7.8197869958, 785: -26.845013054, 786: 17.5821595024, 787: 1.03043959426, 788: 7.34198084953, 789: -0.835099553451, 790: 2.54523248986, 791: -36.1877421533, 792: -13.5192512463, 793: 0.499364471189, 794: -5.55483371839, 795: -11.5012385522, 796: 0.768184463861, 797: 0.69111606247, 798: -0.913676424844, 799: 26.683740229, 800: 0.349793388839, 801: -13.9488979668, 802: 1.38242542632, 803: 15.9748696942, 804: -13.2877953331, 805: -0.241480144061, 806: 1.88313286585, 807: 0.863087143514, 808: 7.69932528683, 809: -14.5609215862, 810: 17.0463839696, 811: 25.2248403435, 812: -28.0410181693, 813: -67.9538348331, 814: -24.6013234564, 815: 12.7252116922, 816: -11.8193573828, 817: -44.6653268774, 818: -40.679919092, 819: 100.0, 820: 44.5661322123, 821: 27.3152205594, 822: -16.3196848465, 823: -10.7925785537, 824: 0.261409398217, 825: 1.35938888315, 826: 8.73511281289, 827: -5.43383087065, 828: 0.241940581788, 829: 0.314637121102, 830: -0.722165340965, 831: -20.9721790291, 832: 1.78402857537, 833: 9.6346106647, 834: 0.0259097547008, 835: -10.5623475373, 836: 21.9462318542, 837: 17.5393057271, 838: -0.294084981694, 839: -26.1591734838, 840: -20.7815652231, 841: -89.670602653, 842: -1.08773252058, 843: 26.7363580895, 844: 0.258333189092, 845: 0.332063211014, 846: -1.55586778306, 847: 0.16612343582, 848: 0.625145306819, 849: -20.8712055718, 850: 1.15695425334, 851: 0.00507511288914, 852: 0.342151484835, 853: 0.0791362985198, 854: 1.44234469566, 855: -61.6718613618, 856: -0.0550736886778, 857: 0.844153804377, 858: 4.28550386558, 859: -6.82881222503, 860: 1.21205397188, 861: 71.6830902683, 862: 0.420248809628, 863: -100.0, 864: 99.1718363869, 865: -7.59114574744, 866: 6.24735140988, 867: -23.2460331846, 868: 4.77396970142, 869: -2.95082405497, 870: 0.195033011056, 871: 2.73108689428, 872: 0.228366660914, 873: 75.3585889448, 874: 36.2052962063, 875: -2.09896673685, 876: -1.22084247165, 877: 0.201244979158, 878: 0.345216139771, 879: 0.152602278267, 880: -0.167866234505, 881: 7.89644833904, 882: 71.6886282207, 883: -3.1037276811, 884: 46.2788172169, 885: 34.9733076818, 886: -16.614818355, 887: 9.50089115977, 888: -29.299808275, 889: -5.54093927228, 890: 18.950732038, 891: 0.297347388373, 892: -29.0788640578, 893: -6.84303270227, 894: 1.11535586122, 895: 7.00570488211, 896: -44.7445789636, 897: 0.259686732984, 898: 1.01689212709, 899: 0.0543125240979, 900: -19.9849401244, 901: 0.181489399886, 902: 0.0784478823566, 903: -0.557697667027, 904: 1.20625533186, 905: -22.4032768072, 906: 0.293577020821, 907: 0.0458577562364, 908: 0.345704608868, 909: 0.729934381165, 910: -0.983692207536, 911: 8.59724165053, 912: 33.3441989307, 913: -1.48001277133, 914: 5.63949817937, 915: 25.4583137611, 916: -0.253549035006, 917: 34.5908028466, 918: 16.2297774946, 919: -16.345329528, 920: -2.02600398399, 921: 11.4204637495, 922: 8.99067190535, 923: -23.4300109274, 924: 0.354682320116, 925: -22.3495572311, 926: 14.3011423596, 927: 4.44728363507, 928: 8.65758569303, 929: 36.5909841551, 930: 0.490699699038, 931: 8.71918660178, 932: 11.6700714723, 933: 2.39175236805, 934: 31.8352940132, 935: -15.2623675285, 936: -62.3270884004, 937: 2.24456470791, 938: 0.460629647369, 939: 99.2506190304, 940: 0.302987156718, 941: -13.7842845656, 942: 5.02402389013, 943: 3.40959833356, 944: 34.2351005832, 945: -53.8727804827, 946: 56.3843796645, 947: 48.6939556492, 948: -13.8325340678, 949: -11.4022451505, 950: 34.4583017156, 951: -23.3178211536, 952: 22.1952192002, 953: 1.64749560632, 954: 1.05129152026, 955: -12.1412543998, 956: 4.11742827866, 957: 0.632889580185, 958: -14.843537009, 959: -36.2901167066, 960: 7.75810071483, 961: 8.93088276836, 962: 14.3614759467, 963: 1.2897130728, 964: 58.3934355994, 965: -64.6293995204, 966: -2.95327130144, 967: -56.8648871631, 968: -3.37046124586, 969: -0.648684018243, 970: -53.50278919, 971: 33.655262894, 972: -18.8953858664, 973: -60.520980225, 974: 48.4626951235, 975: 34.6735338995, 976: 45.3760292683, 977: 11.6202564989, 978: 41.1313578139, 979: -4.74935616211, 980: -5.74583729961, 981: 38.1399457735, 982: 10.5973380961, 983: -12.6530297089, 984: 38.1952934822, 985: 2.01960742583, 986: -9.13392225693, 987: 0.377617497541, 988: 17.4757771112, 989: -19.6776628179, 990: -48.3482568216, 991: 8.67588783822, 992: 4.09895918538, 993: -23.4052957901, 994: 6.95275096236, 995: -40.7905706205, 996: 10.5084283104, 997: -2.63814627357, 998: 16.6015182895, 999: -4.24182882388, 1000: -61.2041002706, 1001: -9.43193603663, 1002: -100.0, 1003: -43.3605757387, 1004: 0.501710276488, 1005: -28.696500384, 1006: 1.16988844314, 1007: 1.16308404097, 1008: 4.71595039448, 1009: 32.4813416508, 1010: 0.5754878705, 1011: 2.16751156789, 1012: -2.0637588156, 1013: 1.16763607335, 1014: -9.4515096737, 1015: 1.35091920229, 1016: 1.45800022914, 1017: -3.31732626248, 1018: 2.10260104104, 1019: -2.39119225061, 1020: -30.4418820766, 1021: -4.56282009202, 1022: 18.0686386116, 1023: 13.0640390069, 1024: -38.2670042048, 1025: -15.8176477801, 1026: -40.9252102435, 1027: 14.6059956997, 1028: -24.3221844524, 1029: 4.68650558698, 1030: -35.6537241658, 1031: 12.0401768013, 1032: -40.640053592, 1033: 0.940411495664, 1034: 86.914237158, 1035: -3.51329586008, 1036: 100.0, 1037: 1.1950213931, 1038: 15.2705087941, 1039: -1.14373776579, 1040: 20.2802907127, 1041: 15.9203832455, 1042: 41.1100583183, 1043: 20.0072627474, 1044: -56.8002822552, 1045: -6.50962931342, 1046: 34.9175338511, 1047: -32.964216329, 1048: -0.351116515195, 1049: -61.633114835, 1050: -17.5570617424, 1051: -5.71909277369, 1052: -54.8217484919, 1053: 18.06023918, 1054: 10.6130538635, 1055: 9.42634302371, 1056: 13.1848815228, 1057: -20.4256757326, 1058: 6.52293028996, 1059: -0.235896207517, 1060: -3.67041460003, 1061: -17.1071609954, 1062: 27.4542587279, 1063: -8.36707680029, 1064: 5.02581782369, 1065: 1.32662505212, 1066: 15.8366756389, 1067: -2.97730863775, 1068: -20.3387718961, 1069: -14.7919995107, 1070: -19.8639070342, 1071: 4.24563534432, 1072: 16.4342106759, 1073: -4.1315655066, 1074: 3.0535708223, 1075: 17.977430147, 1076: 14.9451666826, 1077: -12.7759638411, 1078: -2.24299011571, 1079: 27.9332486684, 1080: -22.8159970489, 1081: -9.60398320545, 1082: -2.25161663973, 1083: -10.8685293389, 1084: -9.03173423964, 1085: -38.9820862232, 1086: -7.34210448544, 1087: 16.5861277256, 1088: 15.9006851493, 1089: 1.13655511818, 1090: -10.4572170722, 1091: 7.64652706847, 1092: -11.8608198128, 1093: 7.56964784554, 1094: 6.97534519374, 1095: 0.216831406027, 1096: 4.71685129012, 1097: 9.66351300764, 1098: 5.10960008433, 1099: -35.8095388306, 1100: -1.42772681138, 1101: -7.56168045772, 1102: -2.95093720141, 1103: -1.24905768663, 1104: -34.8047202489, 1105: -6.99398738432, 1106: 18.2930292386, 1107: 28.5939117419, 1108: -43.3807588072, 1109: -23.8884980219, 1110: -23.09341291, 1111: -19.3081198836, 1112: 14.4678532006, 1113: 7.061340272, 1114: -4.51223754994, 1115: -0.819107730123, 1116: -34.0541596713, 1117: -0.379956714624, 1118: -0.845213680258, 1119: 7.90165061245, 1120: -1.24016436942, 1121: 4.09015557758, 1122: -5.31508459447, 1123: -1.77698976968, 1124: -2.0901612486, 1125: -6.80602034285, 1126: -39.7067072914, 1127: -25.6522718619, 1128: -0.841279896656, 1129: 33.3333509707, 1130: 0.671963104473, 1131: 13.6470684036, 1132: 12.0368060196, 1133: 2.50278754462, 1134: 11.5490589738, 1135: -15.4473074416, 1136: -1.1107677963, 1137: 9.74579884224, 1138: 9.15040792206, 1139: -0.936192390469, 1140: -7.32150948097, 1141: -0.737528867108, 1142: -0.681367896785, 1143: 3.24676546445, 1144: 24.5022113236, 1145: -0.645994867491, 1146: 2.99443033096, 1147: 5.44724512356, 1148: -4.60912703527, 1149: -1.18670337336, 1150: -1.35614626471, 1151: -0.698983586183, 1152: 5.35537427385, 1153: 7.46104199931, 1154: -1.00472583708, 1155: -1.15056760853, 1156: -0.57846329929, 1157: 22.8593236286, 1158: 10.8676178952, 1159: 11.808668725, 1160: -2.00151375742, 1161: 0.0443976032016, 1162: 26.3304542487, 1163: -3.47685798931, 1164: -3.47804784165, 1165: -5.71294696537, 1166: 16.2430807838, 1167: -23.0267753651, 1168: 10.2894655511, 1169: -26.1058731762, 1170: 13.8696290264, 1171: 4.04368887567, 1172: -10.5139416523, 1173: -0.357401426159, 1174: -1.10015006048, 1175: 16.7130859607, 1176: 1.55169499547, 1177: -0.234982928332, 1178: -0.599094556907, 1179: 0.278514038548, 1180: 2.53968855638, 1181: 0.794627021006, 1182: -5.5433743266, 1183: 26.5539077578, 1184: 0.42323621164, 1185: -16.1752548831, 1186: 0.502874641133, 1187: -1.49833651874, 1188: 15.0965758512, 1189: -15.7108735389, 1190: -25.1422439853, 1191: 0.503394667837, 1192: 19.6079216659, 1193: -0.587531831457, 1194: -0.196905987223, 1195: -1.71343012372, 1196: -0.248226530005, 1197: -0.815411960345, 1198: -2.16659633238, 1199: -0.142478323023, 1200: -0.196959307505, 1201: -0.342515912441, 1202: -0.306291094525, 1203: -4.30994837854, 1204: -16.5194493649, 1205: -0.119839384265, 1206: -1.59623423423, 1207: 3.18392351564, 1208: -0.170346408354, 1209: -3.93265489462, 1210: 2.74373632738, 1211: -2.02509420223, 1212: -34.8267743322, 1213: -13.081554228, 1214: 8.37148020723, 1215: -12.5542019999, 1216: -9.06264164585, 1217: 1.05000990128, 1218: 25.7512927775, 1219: -0.0194667419247, 1220: 9.18386414305, 1221: -0.296116392191, 1222: -13.1151645546, 1223: -24.8883369475, 1224: 6.68095945926, 1225: 2.07234017331, 1226: -0.195847517699, 1227: -0.288979779018, 1228: -1.09444084574, 1229: 0.497868001374, 1230: -11.9682334624, 1231: -46.6924602237, 1232: -7.52133563148, 1233: 52.6887465434, 1234: -26.719953619, 1235: 6.95453212621, 1236: 19.8653836069, 1237: 1.43633234733, 1238: 1.47548912386, 1239: 2.6525799567, 1240: -0.0910691644092, 1241: -61.5833574382, 1242: 4.68764666229, 1243: -4.483527873, 1244: -7.49594104911, 1245: 10.6168509434, 1246: -0.345917289061, 1247: -1.179683165, 1248: -0.403622646304, 1249: -8.82327542762, 1250: -0.310791513462, 1251: -0.258192551203, 1252: -10.9435414302, 1253: -1.59082560428, 1254: -27.5196537584, 1255: -0.288704782379, 1256: -0.196054589918, 1257: -2.01326874695, 1258: -0.0696193678967, 1259: -0.944460801512, 1260: -7.11928801851, 1261: 33.4628940784, 1262: 0.310667497107, 1263: -0.228368638366, 1264: 14.5044672666, 1265: 9.59246237952, 1266: -32.3020252328, 1267: -0.118988552468, 1268: -8.92699743413, 1269: -2.0795093982, 1270: -10.5201516698, 1271: -0.714914535296, 1272: -3.47472731194, 1273: -0.275384014115, 1274: 2.30494998996, 1275: -8.92297311074, 1276: 7.61223513911, 1277: 6.70044868212, 1278: 8.00694537735, 1279: -1.01612485714, 1280: -26.7554536984, 1281: 21.122188687, 1282: -2.10018029576, 1283: -21.3541554368, 1284: 6.95146929148, 1285: -4.20576773848, 1286: -0.476985034981, 1287: -0.728675974283, 1288: -25.7589221335, 1289: 9.59707158222, 1290: 19.759684694, 1291: 18.3371480951, 1292: -0.100873188433, 1293: -5.08321905681, 1294: 4.78363508068, 1295: 2.29737634358, 1296: -6.99404660072, 1297: -15.2410245171, 1298: 46.2565156293, 1299: 45.8999956685, 1300: -13.2987195427, 1301: -14.7293207488, 1302: 12.3092825176, 1303: -0.178352901822, 1304: 15.3419925984, 1305: -0.895480272828, 1306: -0.866772673167, 1307: -4.32077933352, 1308: 0.720426647357, 1309: -1.30170068681, 1310: -7.47254739187, 1311: -10.6594472822, 1312: -4.55009062653, 1313: -14.8587006513, 1314: 31.611374755, 1315: 13.6911245802, 1316: 10.9707576129, 1317: -0.993335300632, 1318: -2.2536088493, 1319: 9.59593919801, 1320: -11.7529991504, 1321: 20.4129474891, 1322: -41.1107403528, 1323: 10.7443305535, 1324: -7.96515202723, 1325: -5.4346330438, 1326: -2.97763530151, 1327: -0.277559435681, 1328: -6.16172316947, 1329: -4.18967919798, 1330: -13.2144469405, 1331: -10.6842502386, 1332: 17.3601739394, 1333: -38.5174217505, 1334: -0.749826825792, 1335: -2.02815941825, 1336: 9.15043990729, 1337: -18.0432639699, 1338: 14.6444110595, 1339: 18.8152899134, 1340: -0.600593361814, 1341: -17.0742142161, 1342: -25.9490474224, 1343: 25.4623048271, 1344: 28.1228328843, 1345: 1.76291691212, 1346: 1.1411683107, 1347: -6.89520841249, 1348: -1.96166049526, 1349: 43.7147125327, 1350: 9.0099770252, 1351: 51.5935949514, 1352: 4.09793745336, 1353: -0.79584388188, 1354: 29.2491202746, 1355: -1.02642976269, 1356: -1.16739942058, 1357: 7.25906342919, 1358: -13.5123195279, 1359: -0.783532346395, 1360: -1.27359891257, 1361: -16.4407749458, 1362: -1.41764467938, 1363: 26.5147292164, 1364: -1.15562323799, 1365: -0.979471297074, 1366: -5.76983032224, 1367: -8.63358992408, 1368: -1.01040542743, 1369: -32.2855069509, 1370: 40.9309468087, 1371: 28.8649562725, 1372: -30.9404153086, 1373: 33.4997589698, 1374: 8.39762181057, 1375: -0.499112559076, 1376: 5.57995623104, 1377: 2.88435662821, 1378: -15.4062797978, 1379: -20.2932854026, 1380: -0.475672819508, 1381: 3.4318052663, 1382: -1.83678329206, 1383: 37.3718637563, 1384: 10.619855388, 1385: 29.4781810451, 1386: -1.11051893727, 1387: 4.47630524261, 1388: -10.4683519673, 1389: 3.3605942525, 1390: 17.7519269409, 1391: -43.2284124586, 1392: 4.71897531941, 1393: 72.1487536906, 1394: -36.9185541693, 1395: -12.7998725321, 1396: 63.6685778649, 1397: 0.159621895109, 1398: -5.41332080071, 1399: -3.07772088332, 1400: -0.960970906966, 1401: 2.9582519914, 1402: -5.2168453554, 1403: 5.81600965595, 1404: 9.19882907712, 1405: -1.36988961721, 1406: -5.3716720828, 1407: -20.282042714, 1408: -10.1041213613, 1409: 21.5097162145, 1410: 2.16158563706, 1411: -3.21245773771, 1412: -8.17328378849, 1413: -7.67091660687, 1414: -3.26633336245, 1415: 18.9859246877, 1416: -3.38119909996, 1417: -0.0747679597872, 1418: -10.4856446738, 1419: -10.6980386841, 1420: -2.07475933126, 1421: -11.014855154, 1422: 0.743533212848, 1423: -12.6968827977, 1424: 2.56746839206, 1425: 0.476309816962, 1426: -1.39770101523, 1427: -9.33205795414, 1428: -0.631790682549, 1429: -1.40476966504, 1430: 2.49525543081, 1431: 1.83762297548, 1432: 2.46510323114, 1433: 0.0804232360245, 1434: -5.63758511771, 1435: -0.146065506365, 1436: -1.00896421721, 1437: 6.14059812622, 1438: 5.82534906982, 1439: -19.9131535102, 1440: -14.6095951635, 1441: 2.17864315603, 1442: -1.8725672815, 1443: 3.14766283311, 1444: 0.222734965836, 1445: -3.41033423424, 1446: 6.86767362951, 1447: 20.7429694226, 1448: -3.7903755427, 1449: 1.25688582136, 1450: -8.70367039608, 1451: 4.60732282214, 1452: 1.82997651244, 1453: 14.8725306123, 1454: -0.224427308632, 1455: -3.14218385166, 1456: 28.3747271314, 1457: -15.2096725672, 1458: 1.86896338519, 1459: 0.186906593477, 1460: -3.7852978225, 1461: 9.14865013491, 1462: 0.807983423632, 1463: -3.14197064166, 1464: 0.420786188105, 1465: 1.78420309165, 1466: 0.55493979201, 1467: 0.353483054566, 1468: 3.1836534027, 1469: 1.02136811895, 1470: 26.2359676007, 1471: -0.463591960041, 1472: 2.27260868357, 1473: -4.23299847971, 1474: -4.16441424759, 1475: 12.1979662715, 1476: 8.10568100684, 1477: 1.23590690169, 1478: 7.63995598863, 1479: 1.13559243159, 1480: -1.26294790968, 1481: 2.1607609487, 1482: 27.17372999, 1483: 13.7371314453, 1484: 21.8564628561, 1485: 3.17126736054, 1486: 5.60953134827, 1487: 1.43473673956, 1488: 3.39512498302, 1489: 14.5073906322, 1490: 1.02820264635, 1491: 0.671372854012, 1492: 0.681684300407, 1493: 7.99315913495, 1494: 0.782985776261, 1495: 10.3959960882, 1496: 0.57537731034, 1497: 14.1339238095, 1498: -4.3854105136, 1499: 9.53948543568, 1500: 1.36835540679, 1501: 0.952766382298, 1502: -14.9698810958, 1503: 1.07869617875, 1504: 1.53426832482, 1505: 2.86370580395, 1506: 3.61724256925, 1507: 0.657595479667, 1508: -1.26927537852, 1509: -24.2068148723, 1510: -9.37354253449, 1511: 3.9414081358, 1512: 1.38733035286, 1513: 2.81554056607, 1514: -3.03276464149, 1515: 0.0486629241019, 1516: -3.0354245588, 1517: -7.53564411383, 1518: -6.36057161431, 1519: -16.161298289, 1520: 6.57113485293, 1521: -14.2541651651, 1522: 0.242411145297, 1523: 0.863589490066, 1524: 7.09291267538, 1525: 5.42391350813, 1526: 0.0982287196178, 1527: 0.555840293219, 1528: -5.96461179686, 1529: 3.07465723032, 1530: -1.50724198326, 1531: -3.85581042097, 1532: -16.7801369326, 1533: 3.92688484753, 1534: 19.3099586526, 1535: 1.5368736285, 1536: 1.56851170468, 1537: -4.46886793413, 1538: 5.01175123925, 1539: -10.5160711696, 1540: 1.22417465897, 1541: -12.5944142613, 1542: -0.0886995231003, 1543: 0.113267978295, 1544: -4.46862849743, 1545: 0.192048031648, 1546: 1.35329069376, 1547: 5.80205981807, 1548: 1.80908472283, 1549: 0.133650429425, 1550: 0.372212010637, 1551: 1.02198719759, 1552: -1.31166345965, 1553: -11.0273413925, 1554: -4.43255243402, 1555: -0.570247818749, 1556: -2.56784840295, 1557: -6.93988978429, 1558: 10.8672792295, 1559: -9.6620205976, 1560: -0.959626759238, 1561: 32.5638127564, 1562: -1.3182967205, 1563: -9.4685943299, 1564: -5.13268960238, 1565: 0.490970367282, 1566: 4.05410018079, 1567: 4.25655406992, 1568: 0.120308954894, 1569: -16.6140127989, 1570: 0.249834873392, 1571: 0.232225212464, 1572: 13.0659504665, 1573: -43.389824877, 1574: 2.18552489746, 1575: 0.195324974797, 1576: 0.494172573519, 1577: 0.248656486899, 1578: -0.00046305387023, 1579: -1.8518749436, 1580: 11.7359963566, 1581: 3.1165490782, 1582: 9.53212804415, 1583: 7.60569784931, 1584: 1.35749558527, 1585: -0.939109914274, 1586: -1.20495108393, 1587: -12.8812305485, 1588: -7.37343714666, 1589: 0.076181451525, 1590: -2.82601633411, 1591: -0.628510514067, 1592: -6.96528321339, 1593: 1.55998556074, 1594: -1.0854336687, 1595: 0.13144876092, 1596: 1.15058407612, 1597: -0.040214409088, 1598: 31.6235498021, 1599: 0.146406024004, 1600: -0.00659246730748, 1601: -15.4317444836, 1602: 5.58845263371, 1603: 3.16909015262, 1604: 0.143327121806, 1605: 0.00458969987717, 1606: -0.341768928865, 1607: 0.193598036131, 1608: 0.984081668034, 1609: -4.37170748317, 1610: 2.06563593645, 1611: -0.231330788972, 1612: 0.352093872385, 1613: 13.3092535915, 1614: -2.52499489213, 1615: 21.3786061809, 1616: -26.0059394404, 1617: -2.81851481144, 1618: -3.26041610682, 1619: -14.4840045388, 1620: 9.50843366598, 1621: -20.5775245605, 1622: 0.0485068930206, 1623: 0.909323625203, 1624: 3.67627042759, 1625: -2.37808450353, 1626: -1.29064138181, 1627: -2.58070296494, 1628: 0.503625813499, 1629: 23.5839129847, 1630: -8.30322844269, 1631: 1.51519117004, 1632: -9.25761899557, 1633: -3.26501131307, 1634: 3.36189193909, 1635: 1.18355748907, 1636: 0.440085225584, 1637: -2.05736525379, 1638: -7.48236337897, 1639: -0.321198810539, 1640: -1.94868632547, 1641: 3.90164418369, 1642: 1.33974912505, 1643: -3.35141729412, 1644: 7.92785070634, 1645: -2.39390975015, 1646: 0.580135368616, 1647: -7.66601831011, 1648: 3.60708486507, 1649: -14.8778152295, 1650: -1.07511168007, 1651: 5.53732832267, 1652: 4.98991570287, 1653: 3.48656613165, 1654: 0.319695950995, 1655: 1.34261252069, 1656: -5.83434404937, 1657: -2.69728106789, 1658: 1.05340100834, 1659: -4.27044105142, 1660: -8.86390997978, 1661: -1.46367741251, 1662: -0.988145244903, 1663: 9.5363598764, 1664: 21.6208733122, 1665: -3.81655582526, 1666: -7.69879214161, 1667: 4.77409376104, 1668: 5.57641364857, 1669: -2.88598981988, 1670: -11.8381043261, 1671: 11.4703039443, 1672: -0.812774448348, 1673: -8.89175732709, 1674: 0.103021359629, 1675: -13.3031366015, 1676: -6.76350075587, 1677: 11.4073599302, 1678: 8.37956933726, 1679: -3.80064713817, 1680: -8.77091995229, 1681: 38.5560753902, 1682: 3.65060670009, 1683: 1.0072273433, 1684: -4.09148448757, 1685: 0.912925750331, 1686: 8.33244604, 1687: -1.18812516221, 1688: 2.42684409872, 1689: 2.58556572953, 1690: -1.64918859513, 1691: -1.86681577698, 1692: -5.75415350052, 1693: 0.697930483317, 1694: -9.17014390364, 1695: 7.34270878921, 1696: -9.24993456588, 1697: -10.295545781, 1698: 12.5702950833, 1699: 1.34877258104, 1700: 3.16632833765, 1701: 3.8171159163, 1702: 0.637520510691, 1703: -1.21376036703, 1704: 1.21180112758, 1705: 1.04093521846, 1706: 13.1579557205, 1707: -18.2932005764, 1708: 1.2353667566, 1709: 3.5183888448, 1710: 3.57583908322, 1711: 5.57970925522, 1712: -0.474400384547, 1713: 1.17791691749, 1714: 1.19293872524, 1715: 0.272203797977, 1716: 1.48927461079, 1717: 1.16301145671, 1718: 2.70671092257, 1719: 2.31391799067, 1720: 8.97832062563, 1721: 2.1697416096, 1722: -12.5805102112, 1723: 5.21040139583, 1724: 14.4406314764, 1725: 29.8049635228, 1726: -4.77382538557, 1727: -11.4432373234, 1728: -0.060642784872, 1729: 12.5494471146, 1730: 29.7187389409, 1731: 2.0685512418, 1732: 2.7632026236, 1733: 3.12675936714, 1734: -20.599680961, 1735: 1.12669080174, 1736: 3.97372224737, 1737: -4.84982979504, 1738: -0.281626284326, 1739: -12.9909879303, 1740: -4.41101236739, 1741: 7.14319932071, 1742: 30.4600550557, 1743: -5.18054923005, 1744: 35.7047192864, 1745: 2.85398412464, 1746: -0.392421937075, 1747: -1.70993457562, 1748: -18.8579064233, 1749: -19.0412345667, 1750: -5.41877494254, 1751: 12.9029086663, 1752: -3.13868374518, 1753: -14.7735809315, 1754: 9.7649047754, 1755: 21.1714915048, 1756: 10.342665837, 1757: 20.1989599812, 1758: 68.8797473979, 1759: 1.02297933817, 1760: -18.1572563666, 1761: 18.614310978, 1762: 30.5559902089, 1763: 1.63751640874, 1764: -5.23164922467, 1765: -16.7002919059, 1766: 13.4997016583, 1767: 2.26291776438, 1768: -24.2932816518, 1769: 3.73529040646, 1770: 5.15474122991, 1771: 6.26443507745, 1772: 8.48510325177, 1773: -14.4126899947, 1774: 7.98509161557, 1775: -16.0509864413, 1776: -10.5291774516, 1777: -8.7267907175, 1778: 5.64634945606, 1779: -10.6822979214, 1780: -8.31984074115, 1781: -7.52471275116, 1782: 0.99434890163, 1783: -3.9965907377, 1784: -14.6601558819, 1785: -10.5059722831, 1786: 21.5587590379, 1787: 2.84550010048, 1788: 1.27273735817, 1789: -17.408121493, 1790: 5.19847843018, 1791: 0.0980156429969, 1792: 2.25665150444, 1793: -0.266559156431, 1794: 7.63521780919, 1795: 3.85686715291, 1796: 11.7217595362, 1797: 7.65121197866, 1798: -0.332242617789, 1799: -11.4725538903, 1800: 4.08916771197, 1801: 2.72474466004, 1802: 8.65046600842, 1803: 13.2783482601, 1804: 10.714125119, 1805: 41.2739006326, 1806: 12.8431585229, 1807: 29.2123926591, 1808: 7.4977589513, 1809: 11.2854157381, 1810: -5.28677606881, 1811: 3.2008735219, 1812: -22.4076931037, 1813: -0.158859200374, 1814: -4.56152029999, 1815: 0.47669643886, 1816: -0.7261886783, 1817: 3.29603722558, 1818: 0.173817000398, 1819: -9.46062278507, 1820: 8.32165506706, 1821: -0.0577587155904, 1822: 7.99722771645, 1823: 8.15515709041, 1824: -12.9972919216, 1825: -23.6896306503, 1826: -4.71816533304, 1827: 20.5425761774, 1828: 3.34797944163, 1829: 7.0359675718, 1830: -4.79757008255, 1831: 1.95953575884, 1832: -0.534261616039, 1833: 3.83562089527, 1834: 0.290167903724, 1835: 1.48251945401, 1836: 0.562378627387, 1837: 0.095645660646, 1838: -1.80748726883, 1839: 0.300311478383, 1840: 1.0372133585, 1841: 0.40558206378, 1842: 3.1196273104, 1843: 0.753511411984, 1844: 10.5212177887, 1845: 0.257256953039, 1846: -1.28942919484, 1847: 19.9182536104, 1848: 11.4881228115, 1849: 3.40712611941, 1850: 14.8725576551, 1851: 36.8224967786, 1852: 0.147850584697, 1853: 4.21681390861, 1854: 0.267064064162, 1855: -1.91485602942, 1856: 0.637357320647, 1857: 12.9669033959, 1858: -3.46286794746, 1859: -7.15326919761, 1860: -16.6203158735, 1861: -5.99494707135, 1862: 14.8648615776, 1863: 7.21162205548, 1864: 1.64980296675, 1865: -5.24801620311, 1866: 36.5023839483, 1867: 1.33235961506, 1868: 10.842474847, 1869: -7.03616550692, 1870: 6.57320694418, 1871: -0.0842221318324, 1872: -0.181577829626, 1873: 2.27961819062, 1874: -17.3874682135, 1875: -0.328788867896, 1876: -0.110112021688, 1877: -3.41489541926, 1878: -17.510297414, 1879: -9.82464982313, 1880: -3.57472946526, 1881: -20.8648482497, 1882: -2.33461092113, 1883: 4.06893650944, 1884: -13.0265833414, 1885: 6.28817209583, 1886: -26.1601046639, 1887: 3.8149001319, 1888: -0.510247131816, 1889: -0.456668261443, 1890: 4.73094239051, 1891: -0.163060025838, 1892: -0.111046533449, 1893: 1.62603946455, 1894: -0.137550155615, 1895: 5.37821200339, 1896: 4.33580125049, 1897: -0.283642923404, 1898: -0.19665565, 1899: -0.389275601867, 1900: -0.166933210154, 1901: 0.416849615985, 1902: 29.6822618222, 1903: 0.654521890749, 1904: -0.3909534135, 1905: -0.0712795181558, 1906: -0.411810702074, 1907: 2.38040253934, 1908: -16.8167576266, 1909: -10.1879017129, 1910: 10.0171659653, 1911: 7.87249852054, 1912: -13.5944681965, 1913: -0.753516922822, 1914: 8.89379288668, 1915: -6.88601528099, 1916: 1.34487165843, 1917: -0.436620177079, 1918: -7.81809021389, 1919: -0.311038372224, 1920: 43.1030056678, 1921: 12.8852728238, 1922: 2.81660265887, 1923: 8.54726519699, 1924: -0.180801485119, 1925: -0.733129521335, 1926: -0.521705490744, 1927: 0.328996597882, 1928: 2.552311556, 1929: 75.7869623688, 1930: -32.178472015, 1931: -21.8715299938, 1932: 17.9565740107, 1933: -2.41114337892, 1934: 2.50595831182, 1935: 9.57434363461, 1936: 7.09233062705, 1937: 10.7857521069, 1938: -0.357518223292, 1939: -10.4616159032, 1940: 4.21685593616, 1941: -6.57354053313, 1942: -1.88379374386, 1943: 1.40623818536, 1944: -0.281328666752, 1945: 0.408830080477, 1946: -0.0787566316379, 1947: 1.50532587249, 1948: -0.220167288813, 1949: -0.0748604325084, 1950: 1.69233792521, 1951: 0.417374511189, 1952: 2.95968176384, 1953: -0.527124042971, 1954: -0.0773569114756, 1955: -0.788921264208, 1956: -0.24587384823, 1957: 2.04125851028, 1958: 8.84777593934, 1959: 1.25440637127, 1960: -0.241617350862, 1961: -1.07274890072, 1962: -16.4186974313, 1963: -2.65948292188, 1964: -14.4671268341, 1965: -6.53934596462, 1966: -17.5388189525, 1967: -10.4977698416, 1968: 3.98918883114, 1969: 5.04792017589, 1970: 12.3691429103, 1971: -0.429016663382, 1972: -0.524058400019, 1973: -6.24477330808, 1974: -12.5370247899, 1975: -0.230253311412, 1976: -23.0923663812, 1977: 0.374834751022, 1978: 15.223356711, 1979: 10.4011786377, 1980: -1.96967546725, 1981: 9.9161486921, 1982: 29.9931489486, 1983: -14.6797805346, 1984: 0.013918839789, 1985: -0.132046276384, 1986: 23.1973049251, 1987: 2.7066521353, 1988: -23.196337215, 1989: 5.53656944583, 1990: 17.8211062368, 1991: 34.9848992859, 1992: 16.6218470802, 1993: 7.8976417213, 1994: 14.2116858977, 1995: 9.37781161147, 1996: -0.790370754894, 1997: 0.321529368784, 1998: -28.9529354243, 1999: 21.8344214166, 2000: -0.605839164991, 2001: -2.99305803294, 2002: -0.744572865061, 2003: 0.464376302479, 2004: 5.36069434277, 2005: 4.88341108865, 2006: -0.333740833838, 2007: -0.31631945263, 2008: 1.89053079682, 2009: 1.35359462968, 2010: 0.274375829621, 2011: 24.3756331873, 2012: -11.1935634951, 2013: -11.8624026852, 2014: 5.99556378819, 2015: 1.91648341025, 2016: 1.58915917948, 2017: -2.85652464363, 2018: -47.3890563704, 2019: -30.349517212, 2020: -5.43892639816, 2021: 53.5207314543, 2022: 2.5512808782, 2023: -16.2300112703, 2024: -11.0405148443, 2025: 2.29116660296, 2026: 6.86443596216, 2027: 18.3427501033, 2028: 41.2234578308, 2029: -19.8627887685, 2030: 5.5564409436, 2031: -5.3460856093, 2032: 0.390361507309, 2033: -0.274174074026, 2034: 3.46211920847, 2035: -18.872897719, 2036: -18.2125629423, 2037: -3.33872339807, 2038: -14.6199623472, 2039: 1.06545704666, 2040: -9.20414499577, 2041: -16.6721806202, 2042: -2.0890872337, 2043: -3.10500495277, 2044: 7.33461320735, 2045: 5.68863997404, 2046: 5.44348112253, 2047: -8.47045566675, 2048: 3.43228689831, 2049: -5.73422595545, 2050: 7.35144934413, 2051: 1.07028602581, 2052: 2.34270589021, 2053: 0.268373539374, 2054: 0.255024934406, 2055: 0.351120489929, 2056: 0.780335071093, 2057: 0.402148366813, 2058: 0.158007356035, 2059: 3.07692017104, 2060: 0.173593181517, 2061: 2.5268079279, 2062: 12.2039875317, 2063: 0.0378181602229, 2064: 4.82715118797, 2065: 0.882212788311, 2066: 2.60250579664, 2067: -5.48770784518, 2068: 5.39913147132, 2069: 23.6363647086, 2070: 23.5122030483, 2071: -27.4898540928, 2072: 5.16959089062, 2073: -32.2838388774, 2074: -10.5923553249, 2075: 0.80248503636, 2076: -26.0498642968, 2077: 2.88246172539, 2078: 9.99144583768, 2079: 15.3860671607, 2080: -1.69002511825, 2081: 4.5730343119, 2082: 5.72728926585, 2083: -7.18135210612, 2084: -0.433638464198, 2085: 6.42607808313, 2086: 0.107506805511, 2087: 2.18963293342, 2088: 17.1920959116, 2089: -0.501045911317, 2090: -4.01325350826, 2091: 35.0302468753, 2092: 7.11892100022, 2093: 1.16149093535, 2094: 25.9748720917, 2095: -0.197775063491, 2096: 4.82712160631, 2097: -2.11291366031, 2098: -2.95016245315, 2099: 2.98480246277, 2100: 1.88415743496, 2101: -3.30651166161, 2102: -4.12722631676, 2103: 2.7394415456, 2104: 8.13948067663, 2105: 17.4289888946, 2106: 1.35754744709, 2107: 1.67115526053, 2108: 2.65468194877, 2109: 4.26199260204, 2110: -6.217393344, 2111: 17.4695505566, 2112: 3.15291922093, 2113: 2.45216170886, 2114: -6.17922973603, 2115: 4.23411979643, 2116: 12.6271575489, 2117: 2.80414429448, 2118: 0.0475924256473, 2119: 12.0491332641, 2120: -1.14523368992, 2121: -5.0742174784, 2122: -5.69344593283, 2123: -4.18055393072, 2124: -2.83869544538, 2125: 4.54192210533, 2126: 2.63317943603, 2127: 1.27494522229, 2128: -2.81009958039, 2129: -9.15739992363, 2130: -1.31468569431, 2131: -2.60689170742, 2132: -2.68434012153, 2133: -6.79598094338, 2134: -9.85876330446, 2135: 8.42228305305, 2136: -1.90903403262, 2137: -0.554560770263, 2138: 3.76716393819, 2139: -1.8752894631, 2140: 1.52217392638, 2141: 2.07814732943, 2142: 0.819906184354, 2143: 5.74188328003, 2144: 2.98724242631, 2145: -0.487940861631, 2146: 6.24700355604, 2147: -1.69444670525, 2148: -2.57007421351, 2149: -6.14443183935, 2150: 0.372246865833, 2151: -0.754666809448, 2152: -0.525257499149, 2153: -7.15407688926, 2154: 2.7811556307, 2155: 2.04082585621, 2156: -1.29080235926, 2157: 0.771382661552, 2158: -1.73798200355, 2159: 3.90858110553, 2160: -2.58974258079, 2161: 6.76930141472, 2162: -2.73159402478, 2163: -13.5349819519, 2164: -2.13360379698, 2165: -2.15798067194, 2166: -4.89317214581, 2167: -1.61760230284, 2168: -17.9123573829, 2169: -8.05358591306, 2170: 4.80359305183, 2171: -6.13261124333, 2172: -2.13441981494, 2173: -9.87013767165, 2174: -11.7538629616, 2175: 3.3716622521, 2176: 7.28238357158, 2177: -4.36276104228, 2178: 3.84826463862, 2179: 3.20767268112, 2180: -5.43383217908, 2181: 2.88695535528, 2182: 11.1984509802, 2183: -3.16558638163, 2184: -2.29830126693, 2185: 0.357982911816, 2186: -1.24113775872, 2187: 8.91719086232, 2188: -1.81860049462, 2189: -2.12933454165, 2190: -2.32819197325, 2191: -0.211107951235, 2192: -1.60215147221, 2193: 1.20562079791, 2194: -4.04392587029, 2195: 3.54454205279, 2196: 6.72832167679, 2197: 2.30426366196, 2198: -0.0808755209403, 2199: -6.2476800783, 2200: 16.1850468073, 2201: -0.112780592659, 2202: 0.466958342411, 2203: -3.2827398859, 2204: 14.0713172221, 2205: 5.46177125412, 2206: 0.0127776386996, 2207: 24.260598839, 2208: -5.47495554173, 2209: -11.5151572021, 2210: 1.67568738076, 2211: -0.772157905515, 2212: 1.19366134729, 2213: -17.5927641674, 2214: 10.9156373786, 2215: 1.21823820532, 2216: 2.02483680531, 2217: 2.88186529098, 2218: 2.02818281301, 2219: -0.680056418869, 2220: 0.863153325437, 2221: -0.439019152915, 2222: -12.067949723, 2223: -2.9590529442, 2224: -0.278178985065, 2225: -2.17332628941, 2226: 0.469254864704, 2227: 3.31762941988, 2228: 0.0124266441625, 2229: 0.826684405986, 2230: 4.64992700362, 2231: -3.02690055618, 2232: -19.4451687662, 2233: 7.2267533126, 2234: 8.0357201851, 2235: -5.40606928241, 2236: -17.8787867797, 2237: 14.2805070813, 2238: -0.786654906957, 2239: 6.46257227372, 2240: -0.337945951529, 2241: -0.24701761973, 2242: 7.89868572306, 2243: -0.155057167667, 2244: -1.80896119901, 2245: 1.88501582653, 2246: -0.403782965037, 2247: -0.152320076192, 2248: -0.449236414872, 2249: -0.15284564646, 2250: -2.10521402682, 2251: -34.5558982994, 2252: -1.43916560889, 2253: 0.0992042177555, 2254: 0.365275144878, 2255: -7.27766462381, 2256: -13.8360751018, 2257: 3.32448242362, 2258: -2.49097893245, 2259: -17.2098541706, 2260: 2.07149377439, 2261: 2.59974648208, 2262: -1.32548536506, 2263: 4.61794001068, 2264: -5.93654058353, 2265: -0.577881141863, 2266: -0.572747165733, 2267: 14.817040413, 2268: -0.351197509536, 2269: -8.57828785001, 2270: 4.0575785516, 2271: 3.1800755802, 2272: 0.472189214365, 2273: -0.291362392674, 2274: -0.0858306921553, 2275: 0.0600553287259, 2276: -0.331343710895, 2277: 2.34734468566, 2278: -21.7352888041, 2279: -2.05627344559, 2280: -9.73873988783, 2281: -10.1951946098, 2282: 13.1620410733, 2283: -1.07703673897, 2284: 2.83694219205, 2285: 7.67059679215, 2286: -6.52754512072, 2287: 0.331383680657, 2288: -6.05588776956, 2289: -1.14915186266, 2290: -0.341790368009, 2291: 3.27779366794, 2292: 3.60166634436, 2293: -0.398816803208, 2294: -0.637948042551, 2295: -0.318768006884, 2296: 13.6620347459, 2297: -0.148195850436, 2298: -0.290939725431, 2299: 23.6902742811, 2300: -1.46485481615, 2301: -9.0195552423, 2302: -0.27594267999, 2303: -0.199011958536, 2304: -0.679925671629, 2305: -0.420127854597, 2306: 0.840794210299, 2307: 16.382050433, 2308: 0.177898992811, 2309: -0.743324128569, 2310: 1.56257356432, 2311: -7.16062098305, 2312: 2.68953236216, 2313: -2.27842893113, 2314: 10.7431549353, 2315: -0.222830850496, 2316: -2.26211448409, 2317: -4.76104228246, 2318: -1.94294504626, 2319: 4.50814665945, 2320: -3.19409468168, 2321: 7.39092638271, 2322: -0.377327572786, 2323: -1.14638269218, 2324: -2.85682135241, 2325: -9.55960353427, 2326: -0.633165875957, 2327: -16.1103573768, 2328: 1.32554101261, 2329: -11.1417265324, 2330: -5.52225345391, 2331: 10.0811564045, 2332: -7.02188739663, 2333: 0.49268093677, 2334: -2.2281799715, 2335: -8.4201733316, 2336: -3.02956586781, 2337: 0.236489370657, 2338: 5.60542794558, 2339: -0.419108907113, 2340: 7.84170012126, 2341: 14.2184541103, 2342: -11.659875025, 2343: 14.2873954187, 2344: -3.13013760634, 2345: 18.3957559253, 2346: -2.20752175017, 2347: 11.8976856018, 2348: -2.87898374418, 2349: -1.12397050629, 2350: -2.83389788333, 2351: 3.80166215998, 2352: -1.45817450739, 2353: -1.6570924989, 2354: 1.67475875143, 2355: 11.6691319658, 2356: -1.62747866502, 2357: 4.05094712491, 2358: 3.96256496679, 2359: -2.17843591274, 2360: 31.9556066686, 2361: -3.17990236473, 2362: -13.1636613094, 2363: -4.13622922469, 2364: -3.28450361025, 2365: 5.31457400147, 2366: 1.70548379867, 2367: 4.02263981244, 2368: 11.1004243174, 2369: -8.77175495758, 2370: -7.32413052077, 2371: -7.46838357961, 2372: -9.5078283454, 2373: -9.20933444465, 2374: -9.55029215087, 2375: -0.979936752433, 2376: 2.79883203944, 2377: 3.16006265669, 2378: -1.8847992004, 2379: -8.63281384542, 2380: -4.35537915841, 2381: -2.0340849161, 2382: -1.9972584052, 2383: -4.24912249217, 2384: 5.74386330043, 2385: 16.6711327357, 2386: 0.0532618193932, 2387: 6.04284232669, 2388: -7.33626452041, 2389: 1.43194158792, 2390: 5.38628704916, 2391: -1.3752420519, 2392: 7.68311662694, 2393: 0.847881403532, 2394: 0.566084006991, 2395: -11.6427364844, 2396: 4.27023245786, 2397: 3.58438588839, 2398: -6.32250949244, 2399: 5.41823069166, 2400: -2.0929771306, 2401: 2.59810263103, 2402: -0.689280924905, 2403: -0.738655858898, 2404: -1.18645677045, 2405: -4.14921046225, 2406: -2.30068300152, 2407: -1.38360231685, 2408: -20.2858193111, 2409: -1.36758510733, 2410: 18.6604631524, 2411: -1.6469937876, 2412: -1.363600437, 2413: -7.43974760727, 2414: 8.78692002131, 2415: 0.631410942424, 2416: -6.57790881525, 2417: -16.3816220609, 2418: 5.33649272894, 2419: 2.88443787257, 2420: 0.189598718005, 2421: 5.13227026521, 2422: -19.0872410796, 2423: 1.71577845845, 2424: -1.04515959227, 2425: -3.49875925174, 2426: 2.84527984076, 2427: 0.85665010668, 2428: 1.36516418219, 2429: 7.86127820534, 2430: -5.19662341916, 2431: -3.80929503216, 2432: 6.85857068552, 2433: -0.694510074941, 2434: -0.661752889563, 2435: 2.83953687295, 2436: -1.35054508023, 2437: 0.145438532201, 2438: -1.1346428583, 2439: -3.48992388296, 2440: 16.3204905679, 2441: -3.85531936329, 2442: 0.537379345605, 2443: -0.573437631684, 2444: -1.5665053205, 2445: -30.6912586005, 2446: 8.30043323137, 2447: -25.6752908408, 2448: 80.5083432276, 2449: -27.8627926188, 2450: 83.7450118948, 2451: 66.5178621145, 2452: 20.3895468267, 2453: -54.1653763262, 2454: -5.59023489781, 2455: -26.4442543207, 2456: -47.3224980709, 2457: -3.14224730006, 2458: 14.6201506269, 2459: 4.48156575884, 2460: 26.0550974211, 2461: -0.13315170074, 2462: -74.2117269421, 2463: -18.0104714644, 2464: 14.9658851131, 2465: -33.1793383253, 2466: -24.5069613184, 2467: 25.0216018937, 2468: -34.4774041037, 2469: 55.8119006041, 2470: 65.2253280396, 2471: 32.8006803494, 2472: -40.9989903544, 2473: 4.1105786346, 2474: -58.7222153653, 2475: -84.9539832209, 2476: -30.9363624667, 2477: 46.9738579966, 2478: -14.7551482765, 2479: 34.5582807776, 2480: -26.2593807041, 2481: 52.6690917815, 2482: 75.8134480949, 2483: 67.291497597, 2484: -39.8206914703, 2485: -6.66364660841, 2486: -44.9538062479, 2487: -9.58214092723, 2488: 37.1235317789, 2489: -3.38092347027, 2490: 69.6274925061, 2491: -8.65787010715, 2492: -15.685936551, 2493: 17.6229492224, 2494: 7.78290447481, 2495: -11.2538515621, 2496: -10.9160099855, 2497: -36.3806020392, 2498: -9.23255160707, 2499: -7.0893191104, 2500: 15.7529424718, 2501: -8.91561486521, 2502: -55.809771429, 2503: 64.4441856547, 2504: -11.0842781531, 2505: 40.6003833304, 2506: 36.323862412, 2507: 21.6624619466, 2508: -15.725312212, 2509: 13.9764057146, 2510: 13.2010432163, 2511: -0.308232024586, 2512: -55.8344575108, 2513: -9.08992503057, 2514: -9.64973476734, 2515: 5.91886538498, 2516: -9.71627410906, 2517: 14.7731636398, 2518: 29.1353373084, 2519: -57.6415060735, 2520: -13.6307674335, 2521: 17.4044494837, 2522: -49.5858221768, 2523: -35.4244337991, 2524: -7.27490494168, 2525: 18.4105081421, 2526: -9.24847214911, 2527: -9.48430029703, 2528: -25.6267220089, 2529: -27.3715608764, 2530: 0.806884048037, 2531: -13.1899061843, 2532: -26.9921719182, 2533: -15.4928104011, 2534: 9.42201537808, 2535: -9.37967130961, 2536: 38.565224076, 2537: 1.63580256102, 2538: -6.49347917335, 2539: -6.25187889734, 2540: 14.6481392699, 2541: -8.52278179167, 2542: -9.44212163917, 2543: -16.420837023, 2544: 61.7591635423, 2545: 9.89734140997, 2546: 44.3578829988, 2547: -8.93862190859, 2548: -10.0955064136, 2549: 16.418706282, 2550: -9.50368296697, 2551: -9.56160666692, 2552: -6.1050957428, 2553: 34.79530837, 2554: -8.73084133833, 2555: 30.5670049101, 2556: -100.0, 2557: 5.27813363215, 2558: -40.8237962658, 2559: 25.691529028, 2560: 46.6957732589, 2561: -16.090161949, 2562: -5.88082932863, 2563: -40.3376915167, 2564: -3.92950163149, 2565: 99.9827488179, 2566: 8.65377021526, 2567: 52.2523842886, 2568: -71.1715873785, 2569: -1.77194586576, 2570: -2.17182373907, 2571: -0.897768957209, 2572: -1.90920946254, 2573: -1.81275158514, 2574: -8.8923506195, 2575: -7.59496614703, 2576: 23.30607719, 2577: -1.16686159129, 2578: -3.4188821054, 2579: -52.5374334213, 2580: -3.52574994747, 2581: -72.6495425622, 2582: -3.59376340655, 2583: -4.52458980048, 2584: 18.4688828479, 2585: -39.5856760423, 2586: 24.1534072458, 2587: -1.54189720697, 2588: -8.63327072727, 2589: -2.33221700871, 2590: -2.31721273425, 2591: -8.60215334464, 2592: -1.71648655438, 2593: -9.52108770668, 2594: -11.8093624927, 2595: -1.91920278899, 2596: -1.7454309699, 2597: -2.53015889965, 2598: -1.72929873299, 2599: -2.48037906017, 2600: -47.2179559907, 2601: -2.88856882621, 2602: -3.28315289975, 2603: -2.21319526791, 2604: -7.53473051079, 2605: -2.67506340069, 2606: 20.5340042712, 2607: 16.297109102, 2608: -32.7203046212, 2609: 10.4443852249, 2610: 3.55009780891, 2611: -4.37750113342, 2612: -2.8061650911, 2613: -33.6469647629, 2614: -11.1250706492, 2615: -1.8722828188, 2616: 19.2528222458, 2617: -1.92517378721, 2618: -36.0987205406, 2619: 66.3186558137, 2620: -8.61062920554, 2621: -4.04798640203, 2622: -1.81543899558, 2623: -3.61598030079, 2624: -2.68253562674, 2625: 1.27893809998, 2626: -1.35361098092, 2627: -35.1678305728, 2628: 8.2803570943, 2629: 22.7256936636, 2630: 61.9395459196, 2631: -0.615306318224, 2632: 0.199981429707, 2633: 2.94154425262, 2634: -2.65484386992, 2635: 50.8616881903, 2636: -1.83433240388, 2637: 28.5679707322, 2638: 22.7590826417, 2639: -3.95797586681, 2640: -4.4743377101, 2641: 15.1252004218, 2642: -1.83983991492, 2643: -9.47658598172, 2644: -1.23455585413, 2645: -35.5859764614, 2646: -1.73658724045, 2647: -1.7031899011, 2648: 79.6094536464, 2649: -9.30382995333, 2650: -82.5145283767, 2651: -1.69795876327, 2652: -1.72545269567, 2653: -3.82960205513, 2654: -1.92985210887, 2655: 1.72228416645, 2656: 5.33193315196, 2657: 39.7448805182, 2658: -1.32774317873, 2659: -3.04172786076, 2660: 13.4979056985, 2661: -6.96940162505, 2662: -17.0988308951, 2663: -47.1901603856, 2664: 21.430530369, 2665: -13.0941842907, 2666: -6.98913309953, 2667: -0.888722861254, 2668: 16.3394716167, 2669: -1.49970462942, 2670: -61.6098243342, 2671: -9.66715186621, 2672: 1.19253485276, 2673: -0.524894659898, 2674: 30.678366025, 2675: -1.23226103677, 2676: -11.3861243432, 2677: 27.0095371738, 2678: -9.28539121065, 2679: 44.7985250801, 2680: 71.3606630961, 2681: 32.0420044119, 2682: -6.87058162382, 2683: -8.88848793407, 2684: -22.6416842856, 2685: -10.371353486, 2686: 2.99356310396, 2687: -15.6682072533, 2688: 57.9709194243, 2689: 75.6454160597, 2690: 89.1603962942, 2691: 47.6799937809, 2692: -42.4798309813, 2693: -39.9467996575, 2694: 23.1301662654, 2695: -1.25409504995, 2696: -76.7057353219, 2697: -16.8662444194, 2698: 8.74000237361, 2699: 27.2828173435, 2700: 54.3870238168, 2701: -10.3369688485, 2702: -9.48696052647, 2703: -10.9089291892, 2704: 0.509917060361, 2705: -10.0156530164, 2706: -32.9670079518, 2707: -10.2316825728, 2708: -2.3600418879, 2709: 2.13016339327, 2710: 17.0407499736, 2711: 16.1161783482, 2712: -11.7181186461, 2713: -9.01391151524, 2714: 1.74083460721, 2715: -57.8732792405, 2716: -2.71265569496, 2717: 27.5534587415, 2718: 100.0, 2719: -38.6292060636, 2720: 2.23277848516, 2721: 44.8067724865, 2722: -54.6090020813, 2723: -38.3131538884, 2724: -16.979892957, 2725: 27.015880537, 2726: -39.1307112481, 2727: 99.4822068969, 2728: -5.3259380742, 2729: 55.5318517948, 2730: -10.0515519326, 2731: 40.2915309856, 2732: -1.69353375342, 2733: -29.3535161902, 2734: -0.484706315289, 2735: -8.270494822, 2736: -15.2325436419, 2737: -16.0783262382, 2738: 51.0577794899, 2739: 62.7016908419, 2740: -31.69929849, 2741: 91.3709686979, 2742: -4.67028847292, 2743: 33.2079763231, 2744: -14.8290532185, 2745: -5.67463959067, 2746: -6.6676886414, 2747: -3.91913835408, 2748: 27.9742520672, 2749: -6.32937064709, 2750: -15.0841394951, 2751: -9.38167915565, 2752: -9.50328597555, 2753: -6.95969855591, 2754: 21.6603292624, 2755: -9.24060975496, 2756: -9.51965385251, 2757: -99.3616987154, 2758: -9.13363129691, 2759: 1.97991705641, 2760: -8.66564021917, 2761: -9.58595274427, 2762: 66.1582796019, 2763: -14.4640259422, 2764: 5.20725379641, 2765: 2.02088847597, 2766: 11.2922600987, 2767: -13.4667042047, 2768: -17.4092856139, 2769: 100.0, 2770: 23.0537517163, 2771: 20.8405357979, 2772: 68.0628858075, 2773: 14.1095021108, 2774: -57.7313499552, 2775: -16.1875942237, 2776: 5.53522598993, 2777: 1.83503855447, 2778: 22.490297963, 2779: 14.3642735425, 2780: 15.4045905471, 2781: -3.09335683444, 2782: -9.54033808564, 2783: 27.3407374019, 2784: 25.807718312, 2785: -15.7248135145, 2786: 43.9257111125, 2787: 44.7949625094, 2788: 44.4713491449, 2789: -28.2401372955, 2790: -84.3334491474, 2791: 17.035072575, 2792: -10.7918392261, 2793: 0.145904794599, 2794: -1.98095166615, 2795: -0.509428093364, 2796: -5.80046977478, 2797: 4.66277597358, 2798: -1.62106962872, 2799: -0.466953986689, 2800: 3.95743894326, 2801: -7.20446323426, 2802: -7.31582089106, 2803: -17.2812759711, 2804: 4.14892632553, 2805: -5.35575347324, 2806: -3.76800806834, 2807: -3.57351119711, 2808: 9.90123660956, 2809: 6.41652926169, 2810: -3.62115044161, 2811: 12.6718767995, 2812: 2.41803323552, 2813: -8.44081890834, 2814: 4.42834860536, 2815: -9.3463210111, 2816: -2.16786669647, 2817: -13.0933420089, 2818: -0.664817477962, 2819: 11.0850746906, 2820: 7.60070447184, 2821: 6.75398614323, 2822: 3.55444042961, 2823: 5.76185093678, 2824: -9.89622668442, 2825: -2.80465079844, 2826: 0.115291012277, 2827: 10.5800759801, 2828: 1.0351616771, 2829: 4.27525327921, 2830: 12.3187009502, 2831: 7.68075289234, 2832: 7.55970198498, 2833: -12.097393433, 2834: 0.135709690417, 2835: 21.9880170511, 2836: 7.75191123442, 2837: 8.57747432572, 2838: -3.9763553361, 2839: 4.00107081919, 2840: 1.71804264243, 2841: -9.03419092954, 2842: -6.5129139148, 2843: -4.17845386976, 2844: 22.0473717293, 2845: 3.3499316725, 2846: 3.43980056988, 2847: -1.35651242007, 2848: 0.419949347945, 2849: 1.24330106435, 2850: -0.600251078251, 2851: 18.8132983744, 2852: -3.03776740786, 2853: 11.0794846401, 2854: -1.02767323896, 2855: -2.56447099255, 2856: -2.59808080536, 2857: -1.59908974576, 2858: -2.1577787523, 2859: -5.79142678538, 2860: 0.284390215682, 2861: 11.5766580429, 2862: -0.564136017155, 2863: 0.734314734178, 2864: -12.4398783522, 2865: 1.05220262496, 2866: 24.2204459247, 2867: 0.185467876941, 2868: 12.9150681397, 2869: 6.21928203286, 2870: 6.97885626126, 2871: 17.006165966, 2872: 15.5939556925, 2873: -2.3522808321, 2874: -15.2194443458, 2875: -0.537826045947, 2876: -4.97606846349, 2877: -3.23882984329, 2878: -6.84397442447, 2879: 2.48951578754, 2880: -13.8877250978, 2881: -0.325889835343, 2882: -10.2351454961, 2883: -6.23448522691, 2884: 0.920232122964, 2885: -0.731869724406, 2886: 0.258099944324, 2887: 0.697744411947, 2888: -0.146404982961, 2889: -9.26133337079, 2890: 0.0127692333761, 2891: -8.06664403244, 2892: 3.66558476734, 2893: -3.23167647657, 2894: -14.3474225578, 2895: -9.67723644284, 2896: 0.772573766641, 2897: -4.91946791687, 2898: -17.6989231719, 2899: 1.02015979089, 2900: 3.43781887012, 2901: -1.70882775005, 2902: -17.8130109948, 2903: -5.04833189946, 2904: -2.81702398968, 2905: -13.8500132071, 2906: 35.5941231844, 2907: -7.66357612739, 2908: -0.114957808983, 2909: 7.3338449139, 2910: 11.9116475648, 2911: 11.4283725874, 2912: -11.4303961301, 2913: -5.65304951161, 2914: 6.15934446641, 2915: 52.1287698487, 2916: -0.611139440186, 2917: 4.8332175962, 2918: -0.102694496408, 2919: -0.558935865419, 2920: -3.20671304597, 2921: 6.62488659217, 2922: 0.22882322813, 2923: 0.388884936502, 2924: -2.82676439923, 2925: -5.83920869745, 2926: 2.04507980461, 2927: 1.59142841168, 2928: 3.86028335921, 2929: -3.39574547923, 2930: 4.24757639101, 2931: 0.294152427928, 2932: -2.92901678761, 2933: 15.426439825, 2934: 6.7053225933, 2935: 7.56102229573, 2936: -0.59804041109, 2937: 0.975202981673, 2938: 0.0325923549026, 2939: 0.038360395018, 2940: -11.7032036934, 2941: 0.128716329447, 2942: 0.975492072997, 2943: -8.82775959025, 2944: 0.211989449785, 2945: 0.228690379735, 2946: 0.0974297193934, 2947: 0.28778485833, 2948: 5.79659352074, 2949: -10.5930323572, 2950: -0.747012064901, 2951: 0.624236910469, 2952: -0.783741348962, 2953: 1.7069337398, 2954: 4.44507900262, 2955: -4.74110769558, 2956: -3.40360564988, 2957: 5.99230796477, 2958: -11.2179685701, 2959: -0.0227694982996, 2960: -2.24088164902, 2961: 0.900240171535, 2962: 0.691821388997, 2963: 2.17910221606, 2964: 0.0228809194413, 2965: -9.991456896, 2966: 0.0759471059601, 2967: 8.10734217928, 2968: -10.9893769981, 2969: 21.5510238234, 2970: -0.313901335492, 2971: 0.0396765279435, 2972: 0.163425721829, 2973: 0.436134067046, 2974: -0.156474742628, 2975: 5.48435525927, 2976: 17.9460328204, 2977: 2.18710012548, 2978: 20.0101236955, 2979: 20.2227191797, 2980: -9.57298498644, 2981: -1.53106962955, 2982: 1.5657489159, 2983: 1.61352295378, 2984: 1.84606703972, 2985: -0.048036699489, 2986: 11.3379324439, 2987: 1.209456105, 2988: -3.01973882345, 2989: 0.805054458828, 2990: 0.772296996516, 2991: 0.0101755924945, 2992: 1.42396630015, 2993: 0.262252872453, 2994: -10.6931704427, 2995: 0.294968320897, 2996: 0.0970727150824, 2997: -0.128478413843, 2998: 1.47674183829, 2999: -17.4436638704, 3000: 0.0806922983241, 3001: 0.0272484456513, 3002: 0.0695375135993, 3003: -0.131830283175, 3004: -1.06380177079, 3005: -7.02282315234, 3006: 4.78195720198, 3007: -0.229694626369, 3008: -0.936730542188, 3009: 31.3397666773, 3010: -4.65482114137, 3011: -4.9830979403, 3012: -14.2625703624, 3013: 2.74320482658, 3014: -7.10011287552, 3015: -3.10037567535, 3016: -0.907416333239, 3017: -3.84753832327, 3018: 0.0917352674857, 3019: 4.22074509934, 3020: 4.20238253496, 3021: -3.22507267069, 3022: 0.43290008622, 3023: 17.9910690902, 3024: 0.911634309796, 3025: -0.731808144922, 3026: 4.97051018352, 3027: 25.2123682302, 3028: 16.1587767336, 3029: -4.68451567125, 3030: -3.78374290048, 3031: -0.49122747894, 3032: 0.453546379658, 3033: -4.69872725976, 3034: 4.70704670266, 3035: 0.418963769957, 3036: -0.381636854851, 3037: 1.68380263959, 3038: -12.1839352791, 3039: -6.7438829485, 3040: 21.791930373, 3041: -36.922653156, 3042: -9.03599045516, 3043: -9.33621732495, 3044: -14.9970676199, 3045: 5.58753504247, 3046: -0.0985375784423, 3047: 0.298570348516, 3048: 3.35000633507, 3049: -1.66727259976, 3050: 1.53894080952, 3051: 1.02685485102, 3052: -5.12894581015, 3053: -0.128192824251, 3054: -0.336782696656, 3055: 2.97777136983, 3056: -3.79486260411, 3057: 5.83616005521, 3058: 3.10838826925, 3059: 0.232586574697, 3060: 10.8423020873, 3061: -1.60100233264, 3062: 3.43933316658, 3063: 1.31988864999, 3064: 3.03051442646, 3065: -7.56251453519, 3066: -2.187119231, 3067: 20.5056308482, 3068: 12.9227642082, 3069: 2.38356831103, 3070: 13.3318555641, 3071: 2.975059328, 3072: 19.0196533603, 3073: 7.12608207762, 3074: -0.50783901865, 3075: 9.30021503489, 3076: -10.1468039693, 3077: -2.34492731933, 3078: 1.91962696687, 3079: 1.58911782194, 3080: 3.98710692784, 3081: -2.88590864698, 3082: 6.5611053057, 3083: -0.0560934759548, 3084: -17.8023606501, 3085: -6.40186587395, 3086: 5.9264885402, 3087: -13.0516829783, 3088: -5.14580516271, 3089: 2.98445642198, 3090: -13.9496084902, 3091: 0.0972637739165, 3092: -1.58119906702, 3093: 0.85045946856, 3094: -6.52058826533, 3095: 0.0208473875992, 3096: 1.60527303797, 3097: -4.57031402489, 3098: 0.673168903699, 3099: -8.48672064642, 3100: 1.3368260629, 3101: 1.23747678267, 3102: -1.47084908267, 3103: 5.26287304232, 3104: -1.26126106733, 3105: 0.190559920335, 3106: 6.83925146813, 3107: 1.53712322872, 3108: 1.43760955727, 3109: 1.04211375414, 3110: 2.28602706631, 3111: 13.8393636208, 3112: 10.7616890148, 3113: -2.81839781527, 3114: -1.32816681739, 3115: 2.22081692645, 3116: -3.15611034037, 3117: -20.9993434499, 3118: -2.61346686766, 3119: -4.53254166499, 3120: 21.5191031787, 3121: -9.39116525631, 3122: 0.345133589862, 3123: -0.227844404049, 3124: -5.2641988812, 3125: -0.0533078123392, 3126: 2.48590870728, 3127: -9.99520770087, 3128: -1.52820891736, 3129: -13.8012866, 3130: 2.24741011082, 3131: 0.970635079917, 3132: 4.52927177325, 3133: -4.44189996544, 3134: -17.2394773249, 3135: 6.4265704402, 3136: 1.09352908245, 3137: -0.56646730662, 3138: -17.5709078222, 3139: 6.49495397122, 3140: 5.05574411254, 3141: -7.47795991998, 3142: 1.90901196225, 3143: 9.88315054248, 3144: -19.8434661755, 3145: 7.15431308385, 3146: -31.3488294088, 3147: 3.30298791016, 3148: -0.767360050205, 3149: 34.4307363708, 3150: -27.609736847, 3151: 12.8258227153, 3152: -12.3487952889, 3153: -43.6440790053, 3154: 12.312501694, 3155: -26.2541502047, 3156: -41.8845410192, 3157: -8.99332043949, 3158: -68.8053752155, 3159: -21.9563362808, 3160: 46.1391547245, 3161: -8.7618968964, 3162: -10.9234385059, 3163: -1.9090884244, 3164: 11.2300312858, 3165: 18.0335885003, 3166: 0.733937691891, 3167: -53.8051280084, 3168: 11.2362504633, 3169: 7.67688625626, 3170: 21.9117675113, 3171: 11.4582904253, 3172: -2.23753866779, 3173: 9.30245022853, 3174: 15.3631102359, 3175: -4.06291246003, 3176: 9.83011902237, 3177: 1.13885827372, 3178: -22.0010716486, 3179: -10.034551114, 3180: 5.97653570656, 3181: -31.0377016949, 3182: -30.1925235949, 3183: 0.165831411012, 3184: -4.21560020156, 3185: 37.0799924814, 3186: 1.22658319611, 3187: -33.4832115132, 3188: -3.34338349726, 3189: 4.29178496541, 3190: 7.84122721401, 3191: 5.39252340963, 3192: 5.66234509862, 3193: -8.0502372569, 3194: 1.93017751997, 3195: -0.609087592852, 3196: -20.6036000283, 3197: 5.22250842614, 3198: 41.750222113, 3199: 6.09493772937, 3200: -38.13295285, 3201: -46.5708673916, 3202: 11.1385956037, 3203: -1.29571416603, 3204: 0.368966023857, 3205: 8.71469878959, 3206: -1.34269465716, 3207: 5.59676741363, 3208: 13.7386456867, 3209: 5.61880997212, 3210: 37.4322992277, 3211: 5.76101979245, 3212: 5.72635551996, 3213: 10.393868683, 3214: 5.47494812193, 3215: 33.9549700207, 3216: 2.65606073858, 3217: 4.27831306854, 3218: 3.73785069185, 3219: 24.6638962004, 3220: 7.0374528213, 3221: 4.13998887887, 3222: 4.3106757734, 3223: 5.28428119501, 3224: 5.31753883722, 3225: 5.44504146223, 3226: -12.8350286292, 3227: 3.96677269615, 3228: -20.9718217532, 3229: 20.1917356958, 3230: -5.27380284276, 3231: -14.7298010782, 3232: 5.35856975711, 3233: 5.1899483959, 3234: 14.5877123084, 3235: 6.91212740577, 3236: 3.66764528264, 3237: 7.93857743619, 3238: 2.44241126871, 3239: 5.7795573596, 3240: 3.69894801708, 3241: 5.06366206398, 3242: 1.98272400523, 3243: 5.5003036739, 3244: -16.6339067677, 3245: 5.69467402807, 3246: 5.4273666147, 3247: -47.0258945599, 3248: 5.47092011498, 3249: 5.37304132473, 3250: -4.4605156303, 3251: 0.516191717489, 3252: 6.95778857389, 3253: -11.269535059, 3254: -48.0877849604, 3255: 5.53933794839, 3256: -27.4861191216, 3257: -2.30591701674, 3258: 39.3208166828, 3259: 8.26064340427, 3260: -43.6283108874, 3261: 4.93619650326, 3262: -9.89729219387, 3263: 14.5681707897, 3264: 52.6988594581, 3265: 15.883244511, 3266: -23.5599976443, 3267: 1.46636379764, 3268: 2.33462314232, 3269: -3.99699412452, 3270: -41.090413385, 3271: -0.510875940669, 3272: 5.55297849705, 3273: 3.12429291994, 3274: 9.46068759534, 3275: 2.6176720992, 3276: 2.52007711704, 3277: 13.0638701049, 3278: 1.01625144949, 3279: -6.56830137706, 3280: -1.82648297912, 3281: 5.57913893789, 3282: -29.4482492669, 3283: -21.1930436788, 3284: 20.2828797338, 3285: 3.12945934306, 3286: 21.0659418469, 3287: 1.96840411109, 3288: 2.00094747232, 3289: 7.84450305553, 3290: 1.94863528129, 3291: 5.27520682111, 3292: 11.528938498, 3293: 2.21263877425, 3294: 1.8732159384, 3295: 1.73384157567, 3296: 2.0176690114, 3297: 7.89724393057, 3298: -0.144318391019, 3299: 4.25270352514, 3300: 2.722544195, 3301: 1.80540712013, 3302: -18.1691766875, 3303: 4.52154490313, 3304: 40.6424570956, 3305: 2.99808497169, 3306: 9.04285116556, 3307: 23.4632416902, 3308: 24.2120422144, 3309: 12.3890117078, 3310: 20.1989896343, 3311: 5.21715378919, 3312: 1.8970007021, 3313: 2.10267762352, 3314: -46.0434625518, 3315: 1.83227814516, 3316: 12.7770369508, 3317: 96.2865801995, 3318: -35.5699549761, 3319: 1.433086596, 3320: 1.94992057133, 3321: 1.94060156989, 3322: 1.49029172946, 3323: 1.98520855656, 3324: 3.02604339064, 3325: -4.31653377529, 3326: 2.36390971525, 3327: -42.2324282134, 3328: 41.168057253, 3329: -1.86770683049, 3330: 6.33651386151, 3331: -12.7725838606, 3332: -17.839158656, 3333: -16.9778004385, 3334: 2.32341534041, 3335: -23.8259797401, 3336: 6.07811121609, 3337: 3.75791085953, 3338: -1.50588263644, 3339: -5.35841677816, 3340: 1.84605243554, 3341: 5.61293568138, 3342: 2.45283395805, 3343: -21.0289661932, 3344: 2.010054816, 3345: 1.90401801657, 3346: 20.8918103714, 3347: 5.1143443557, 3348: -42.3168387928, 3349: 2.31301226737, 3350: 1.88644490664, 3351: -3.55187889987, 3352: 1.37356804012, 3353: 3.41714487317, 3354: 4.30277505847, 3355: 5.16178072291, 3356: 4.64852920134, 3357: 1.32818057541, 3358: 35.3176048373, 3359: -1.80475684485, 3360: -5.94503773987, 3361: -0.770141038047, 3362: 10.3887234718, 3363: 0.467313626999, 3364: -6.49321894481, 3365: 5.0383289241, 3366: 0.422977922863, 3367: 2.24636133938, 3368: 1.13734219594, 3369: 1.65781381907, 3370: -0.795227885928, 3371: -22.9574930461, 3372: -37.5259629859, 3373: 3.15227853665, 3374: -9.05149828113, 3375: -15.5276969507, 3376: 6.80409821416, 3377: -11.2227035527, 3378: -41.1953915383, 3379: -35.2493266575, 3380: 18.7870664463, 3381: 5.44367156163, 3382: 37.0904676769, 3383: -30.2177220567, 3384: 34.4168480894, 3385: 18.6226378123, 3386: 8.98190650007, 3387: 4.35496791695, 3388: 14.0027519899, 3389: 35.9906954122, 3390: -89.347079235, 3391: 4.18934185554, 3392: 2.6509082101, 3393: -17.3253675084, 3394: -37.1540026163, 3395: -26.1818112987, 3396: 5.12376317332, 3397: 4.62181671233, 3398: 33.4812815973, 3399: 5.62834219066, 3400: 5.31426574808, 3401: 11.6099500576, 3402: 30.2395786475, 3403: 5.32098292209, 3404: 5.69288519647, 3405: -9.65671730646, 3406: 7.98041489824, 3407: 2.68451846016, 3408: -6.33082444887, 3409: -11.817754155, 3410: -45.9077009354, 3411: -8.46543539441, 3412: 4.68845082183, 3413: -29.0243291195, 3414: 7.21335036542, 3415: 15.9325805155, 3416: -24.4504325056, 3417: 15.1318536626, 3418: -33.0680868535, 3419: 1.47703130995, 3420: 5.3066307353, 3421: -20.9548292444, 3422: -0.245984454849, 3423: 29.4690645412, 3424: 12.6797909135, 3425: -4.33777299954, 3426: -40.5683874287, 3427: -53.2385370223, 3428: 5.36796309777, 3429: 6.00787092122, 3430: 5.77404917499, 3431: 2.07410243636, 3432: 3.01634984075, 3433: -3.66127517034, 3434: 9.87677481242, 3435: 16.0379372285, 3436: -13.9970967344, 3437: 17.9363881562, 3438: 17.5692427497, 3439: -20.8408963085, 3440: -4.61829170267, 3441: -34.889233255, 3442: -0.191506758253, 3443: 32.8871029504, 3444: 1.43214893808, 3445: -23.5817389176, 3446: -19.089783715, 3447: 3.7567310147, 3448: 5.49536446806, 3449: 5.38502568142, 3450: 5.37737335316, 3451: 5.22433823537, 3452: 25.4034821753, 3453: 5.47545126698, 3454: 5.22508997068, 3455: -4.88340813216, 3456: 5.04581411457, 3457: 16.3047948338, 3458: 3.46941400073, 3459: 5.43062840219, 3460: 42.2421158458, 3461: 5.49124411247, 3462: 1.70698320279, 3463: 5.51414602444, 3464: -23.7464909569, 3465: -14.0482147422, 3466: -47.1965292501, 3467: -11.046070097, 3468: -1.33827673284, 3469: 14.1317898846, 3470: -5.89686972639, 3471: -7.61753430854, 3472: 18.7728901883, 3473: 17.5517885451, 3474: 4.656942066, 3475: -11.0989507099, 3476: -0.222112074793, 3477: 29.8895750311, 3478: -0.80058558248, 3479: -0.263139501193, 3480: 5.36905927301, 3481: 5.70980218079, 3482: 5.09178749118, 3483: 2.29997286981, 3484: -31.6329674429, 3485: -1.88262184943, 3486: -17.3731135355, 3487: -64.362571811, 3488: -99.5908129093, 3489: -36.858565828, 3490: -39.5501207895, 3491: -12.2219100829, 3492: 3.25871473157, 3493: 0.544792539592, 3494: -1.51501658392, 3495: 1.14631194415, 3496: -1.47328386329, 3497: 1.14357938377, 3498: -0.0998925612163, 3499: -0.928849919806, 3500: 0.998501554874, 3501: -1.13428832501}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: -0.306771865089, 2: -10.7005495762, 3: -18.7041591659, 4: -21.8219221902, 5: -18.8399080159, 6: 1.61509586535, 7: 10.3491872926, 8: 15.4039588708, 9: 10.8294239332, 10: -10.9326806831, 11: 18.3664866083, 12: -11.2589538227, 13: -2.73930035068, 14: -2.08728876503, 15: -13.298880722, 16: 8.07517213737, 17: 18.4369868411, 18: -23.6660267342, 19: 19.3243796998, 20: 17.4918363353, 21: -5.97833434446, 22: 25.958913305, 23: -5.95167116778, 24: -4.85306991669, 25: -16.6277411044, 26: -10.40314091, 27: 37.0309186736, 28: 12.9969226054, 29: -3.94347884145, 30: -13.0848716139, 31: -2.65327976837, 32: -13.5444949647, 33: -19.0074458603, 34: -18.5326551616, 35: -11.4262032844, 36: -2.40333700476, 37: 7.65851137302, 38: 6.43440196988, 39: -25.7443244456, 40: -10.6441025624, 41: -6.18753912728, 42: -6.88744593728, 43: -40.8978894491, 44: -44.1392088609, 45: 10.5822900294, 46: -21.4607378789, 47: 22.152337601, 48: 24.0193692219, 49: -3.40532399688, 50: 7.43393473839, 51: 3.54827521404, 52: -14.4329513351, 53: -1.28549850687, 54: 6.12917996721, 55: 1.73176971346, 56: -1.05882459313, 57: -3.18429142172, 58: -11.6665459165, 59: 31.4433713492, 60: -26.8489781081, 61: -44.8061278925, 62: -13.2242789926, 63: 3.2167134811, 64: -27.9883211643, 65: 14.5381946553, 66: 4.29267137391, 67: 26.498234843, 68: -0.0110050624182, 69: 5.8886512591, 70: -1.2215392959, 71: -0.397866099319, 72: 19.0581462266, 73: -1.07954692009, 74: 29.4558921345, 75: 5.81218896888, 76: -2.6554283955, 77: 6.7911159924, 78: -5.33104474636, 79: -35.1319871884, 80: -15.1043998756, 81: 9.72001317672, 82: 75.3957795649, 83: -1.00240425165, 84: -10.0634732192, 85: 6.54574651343, 86: -10.6713362616, 87: -1.67465256284, 88: -5.76640361313, 89: -1.13819633089, 90: 2.75507416592, 91: 0.497815071766, 92: -1.10409849978, 93: 9.03470837968, 94: -0.641849592041, 95: -0.161547770967, 96: -1.33868867928, 97: -10.0792772937, 98: -1.15173682242, 99: 0.661557742655, 100: -2.66652160283, 101: 97.370434976, 102: -1.33423817807, 103: 7.5767435412, 104: 3.42351011922, 105: -12.3514296613, 106: 4.98652816418, 107: 9.62695749278, 108: 1.8656288971, 109: -1.09966376871, 110: 13.2625884736, 111: -7.29437130232, 112: -2.49242994644, 113: -87.2133181749, 114: 5.68621000652, 115: -21.0249208563, 116: 8.3523546975, 117: -3.73911268738, 118: -0.731039881668, 119: -1.78366104725, 120: 36.3414942514, 121: -7.8884591844, 122: -3.43599078286, 123: 23.5222680278, 124: 7.41420130017, 125: -50.9920701369, 126: -0.455232489747, 127: 2.92381138467, 128: 11.0165223787, 129: -7.44702399119, 130: -0.29017399713, 131: -1.09887597278, 132: -3.36128333345, 133: 14.1547094578, 134: -3.13405512363, 135: 1.27208339065, 136: 5.90284381604, 137: 2.75512235447, 138: -23.7792331881, 139: -1.04076451861, 140: 5.18488924185, 141: -30.8527041266, 142: -15.3780272573, 143: 0.574218048552, 144: -0.0536723956197, 145: -1.97172913845, 146: -0.389531221663, 147: -0.327593722278, 148: -11.3980876306, 149: -0.311995503552, 150: -1.25946811284, 151: 7.23721544974, 152: -0.898235784687, 153: -0.211848126851, 154: -0.333145479646, 155: -0.261380661336, 156: -1.18419251151, 157: -64.8406562226, 158: -2.04308030951, 159: -0.97177700763, 160: -0.528630307879, 161: -0.823509486835, 162: 1.91883203391, 163: -29.1687137521, 164: -9.98021026191, 165: 10.4999704973, 166: -5.88377790116, 167: 6.2405114237, 168: -2.65229696339, 169: 1.79250237346, 170: 7.90553896957, 171: -29.1570161074, 172: 0.0495749447744, 173: -14.0442029742, 174: -0.155949121782, 175: -14.7276068383, 176: 22.0864402767, 177: -33.1665388709, 178: 1.82173099264, 179: -0.289895654478, 180: -0.104580506315, 181: -0.176033341505, 182: -0.229198008639, 183: 6.93960584955, 184: -2.54978819512, 185: 8.90821498428, 186: -15.7829643522, 187: 21.7963108253, 188: 2.93597949864, 189: 4.8402256748, 190: -17.1918554259, 191: 19.9546069082, 192: 20.3989297548, 193: -0.416592427045, 194: -23.8651629419, 195: -1.63818089894, 196: -2.97809210282, 197: 3.09942753729, 198: -24.9285435193, 199: -0.326042951416, 200: -1.35721907633, 201: -0.185468171883, 202: 10.9421619394, 203: -0.266517307529, 204: -0.216788323025, 205: 5.86771944158, 206: -1.23175258502, 207: -8.01288925181, 208: -0.281895148515, 209: -0.207481058419, 210: -1.55316663216, 211: -0.197224009856, 212: -1.50335940117, 213: -2.77428441142, 214: -8.95264754353, 215: -0.50103181594, 216: 0.447599460806, 217: 50.72256651, 218: 6.2348330779, 219: 10.1637211731, 220: 26.1007027463, 221: 0.640872838366, 222: 0.386211171721, 223: -6.61985757047, 224: 0.972201956404, 225: -2.52380763664, 226: 0.91497954937, 227: -12.3560928269, 228: -18.1488411217, 229: 1.44111751634, 230: 4.46484044078, 231: -20.4630853606, 232: -0.591743842733, 233: -0.126833432164, 234: 5.19068141186, 235: 3.4578035902, 236: -22.4383590127, 237: -0.527410596641, 238: -6.67733244935, 239: 0.287405174332, 240: -1.00757154924, 241: -17.5104316323, 242: 19.7794152761, 243: 59.7574384223, 244: -9.7958234443, 245: 2.39133244552, 246: 0.924560155138, 247: 9.94688775358, 248: 6.64175723772, 249: 17.9175916396, 250: 9.96285419045, 251: 3.80033789609, 252: -9.06678012616, 253: 9.63766642058, 254: -3.8854157629, 255: 2.64897890819, 256: -1.36156697443, 257: 45.0093436843, 258: 3.38517586518, 259: -1.07243850413, 260: 8.04060853683, 261: 8.05752435846, 262: 0.0849996994563, 263: -0.577977995836, 264: -2.04524184282, 265: -1.19199895582, 266: 71.0551624889, 267: 27.4144121114, 268: 23.8503629306, 269: 44.3117392829, 270: 1.10244618554, 271: -4.12485477764, 272: 30.4552948052, 273: -19.5403193332, 274: -1.98873148823, 275: -11.4272988987, 276: 2.84981865586, 277: 1.82142662888, 278: 36.5495149187, 279: 21.5116888418, 280: -22.5293316096, 281: -6.95238785539, 282: -5.37326971213, 283: -9.25365050837, 284: -45.7704006838, 285: 29.9586511053, 286: 20.7298937041, 287: -1.06421416832, 288: -10.7802946063, 289: -2.70109211724, 290: 4.14846013349, 291: 2.97981198787, 292: 26.2820043431, 293: 17.6140892561, 294: -9.46311988058, 295: -29.5929100361, 296: 12.9211364, 297: 20.5334655632, 298: -4.98316682843, 299: 2.53195462414, 300: -0.764420643933, 301: 0.875198804163, 302: 23.008574862, 303: 1.42222821514, 304: -11.67377078, 305: 22.3692903614, 306: -0.250604857237, 307: 8.24688435046, 308: -1.29212132558, 309: -1.41659353647, 310: 4.03418511149, 311: -22.382525022, 312: -1.28363395337, 313: -1.231245607, 314: -9.15752003797, 315: -1.08724076393, 316: -23.7738276203, 317: -1.2382052797, 318: -1.22235212789, 319: 3.63476021743, 320: -1.135089937, 321: -1.4865226237, 322: 22.8384140185, 323: -27.3303327013, 324: 12.6212750446, 325: -5.49974403844, 326: -15.8906153266, 327: -2.94609210782, 328: 53.6998495749, 329: 26.1302587706, 330: 12.6660991589, 331: 6.61145656979, 332: -10.5426494864, 333: 1.36584327529, 334: 9.32802915118, 335: -1.12878257178, 336: 0.716212871794, 337: 5.10236318907, 338: -11.5425066637, 339: -1.51928712523, 340: 37.2779488008, 341: -5.67279469191, 342: -5.93975127964, 343: -11.3624002521, 344: -31.1221352318, 345: 0.845807398431, 346: 48.5620127921, 347: -40.4174369775, 348: -3.30312984151, 349: -3.61392513001, 350: 0.418252193796, 351: 7.41041067812, 352: -1.97006331242, 353: -0.210525545887, 354: 2.48618369921, 355: -8.317800871, 356: -2.63413538689, 357: 14.5540492836, 358: -0.620733921329, 359: 2.06073821496, 360: -3.47946474555, 361: 3.83484964589, 362: 0.129150261281, 363: -7.99147605976, 364: -1.78117281318, 365: 1.44636045471, 366: 1.65834777215, 367: 5.05016342515, 368: 3.79235114125, 369: 5.24412272081, 370: -0.770209875057, 371: 5.97204275244, 372: 1.48670286243, 373: 8.08058657874, 374: 2.46372364253, 375: -5.52527248892, 376: 1.17904266345, 377: -0.819850415864, 378: -3.78418342135, 379: 3.20398927388, 380: 3.96154694003, 381: 1.42263105823, 382: -1.63350083905, 383: -0.998917718812, 384: -8.89631462244, 385: 2.32418632625, 386: 1.27292288011, 387: -4.86445969534, 388: 7.62801819311, 389: 0.201810501087, 390: 4.16816747518, 391: 8.19939942514, 392: -5.62287176768, 393: -1.34717175365, 394: 4.37809863349, 395: -7.63925368395, 396: 6.42607355785, 397: 0.0894026976388, 398: -0.986000263108, 399: -2.47055350413, 400: -0.744694717529, 401: -4.85857464818, 402: 4.18431460754, 403: 2.46880095205, 404: -3.47209102013, 405: 0.728692801798, 406: 3.88092473577, 407: 0.588961657153, 408: 2.92231723855, 409: -14.6339452266, 410: 4.2011611472, 411: -2.63880204114, 412: 1.63331670826, 413: -2.37648446947, 414: -6.29135520386, 415: 1.46900902969, 416: 10.5898735654, 417: 1.19908108841, 418: 6.24912286249, 419: 0.980078834436, 420: 0.483408765243, 421: -0.58908552068, 422: 0.573809320355, 423: -13.6110933742, 424: -3.70806098794, 425: 0.626868904432, 426: -2.00107535146, 427: 5.5231391164, 428: 5.45397364304, 429: 6.39762927846, 430: -3.20125872134, 431: 2.37866246089, 432: 0.546057599007, 433: 0.455986750236, 434: 0.441193209661, 435: -1.44470903222, 436: -2.18074121503, 437: 7.38353316702, 438: -0.29105501339, 439: -2.91626876317, 440: -0.338850123354, 441: 0.443255263631, 442: -13.6243738383, 443: 0.30375212251, 444: 0.429619936492, 445: -0.547402005361, 446: 5.01849456723, 447: 0.830738417278, 448: -3.82095429699, 449: 1.7443564744, 450: 6.16389260498, 451: -0.721650975248, 452: -1.75041608759, 453: 0.482324098788, 454: 3.00754617726, 455: -7.6209171123, 456: 0.138739667013, 457: 0.40079660341, 458: -0.53756003531, 459: -1.32416957099, 460: 0.232697110027, 461: -2.2795410807, 462: 12.7139679298, 463: -15.9737400278, 464: -15.4740814702, 465: -4.65405923419, 466: -1.31005923311, 467: 4.49769664945, 468: -6.73740014969, 469: -12.929761607, 470: 5.25673617132, 471: -7.43467648553, 472: -13.0634232814, 473: 1.6032024437, 474: -5.92148364722, 475: 0.37820973928, 476: 0.549445759677, 477: 1.89499100163, 478: 4.06868798486, 479: 0.439488706356, 480: 0.49694868434, 481: -0.624245850464, 482: 16.0131372242, 483: 0.964711830487, 484: 1.26975783563, 485: -14.8137543035, 486: -0.339899136378, 487: -0.446678097351, 488: 2.87761777682, 489: -1.5088187482, 490: 0.726026540882, 491: 4.51609423119, 492: -7.69774159071, 493: 0.417335295975, 494: -1.17641375595, 495: 0.336100386408, 496: 0.507327744602, 497: 2.32808931178, 498: 0.333636899842, 499: -4.29613485705, 500: 1.31518591651, 501: 0.184148769166, 502: 0.526197355391, 503: 0.514046758871, 504: 0.470388885783, 505: -4.20967745782, 506: -5.51716044042, 507: 0.732624400003, 508: 0.633838801493, 509: 0.159153489631, 510: -2.26156582717, 511: 0.265568310662, 512: -2.759876722, 513: -1.78367036433, 514: -8.03979372719, 515: 2.29265701906, 516: 6.27217819267, 517: -11.0823679105, 518: -0.428210235144, 519: 0.699887524395, 520: 2.69728030006, 521: 0.453902781726, 522: -10.8230228264, 523: 0.329994792451, 524: -3.38494818561, 525: -11.1211874835, 526: -12.2827767476, 527: 0.951866166081, 528: 0.255233419454, 529: 0.519631040126, 530: 0.230941256601, 531: 0.68375410884, 532: -2.76325925652, 533: 0.384402826003, 534: -0.0430106756458, 535: 0.699782160382, 536: 2.68200523269, 537: 2.36201587177, 538: -1.44055333031, 539: -1.8293005519, 540: -2.73907560424, 541: -4.61807647352, 542: 0.472901495, 543: 10.9140363283, 544: 3.32601676866, 545: 1.69655180382, 546: 0.260200855252, 547: -11.220478583, 548: 0.352273591878, 549: 0.714296973198, 550: 0.361294628375, 551: -3.48464610505, 552: 0.488892872795, 553: 0.49905856494, 554: -10.3751887209, 555: -0.0928938128656, 556: 16.196870435, 557: 0.427899241038, 558: 0.489568700267, 559: 2.45550279028, 560: 0.49227865456, 561: -0.0951139977703, 562: 0.959520071656, 563: 5.76002847961, 564: 0.176853857249, 565: 1.15172948745, 566: -8.65735920173, 567: 1.06216102914, 568: 6.32691138215, 569: -7.2024467741, 570: -0.597320012681, 571: -5.67165462123, 572: 3.11232584231, 573: -0.910124108826, 574: -1.35271491244, 575: 0.674843422548, 576: 14.3331547014, 577: 1.42782943027, 578: 1.08463362692, 579: 2.81783849151, 580: -7.38690118251, 581: 0.0659720267651, 582: -14.1344956701, 583: 4.98831990966, 584: -0.74983658537, 585: -7.24256539733, 586: 1.7219298029, 587: 15.6370431646, 588: 2.11604868888, 589: 0.540166315771, 590: -4.49697831458, 591: -15.1591744082, 592: 4.77853331975, 593: -13.8784024567, 594: 9.57619139183, 595: 2.78475477357, 596: 1.40464041281, 597: 8.02297517764, 598: 10.0000503174, 599: -0.965722459538, 600: 0.0422033425304, 601: 3.6401462463, 602: 5.05116014132, 603: 3.08127303521, 604: 0.0920227137486, 605: -0.0539587145864, 606: 1.34913734367, 607: 0.372173299011, 608: -4.31968139059, 609: 1.7811149655, 610: 2.07599877015, 611: -3.60709957576, 612: -1.83532757249, 613: 1.54305783001, 614: -4.2912460992, 615: -4.91699163128, 616: -14.5117880767, 617: -8.18716419682, 618: 2.6767484117, 619: -3.09694366527, 620: -2.11275714098, 621: -1.72735632693, 622: -2.68206652733, 623: 1.4616560533, 624: 6.80535689629, 625: 1.36160544389, 626: -11.1585879419, 627: 1.10400968448, 628: 0.590123180818, 629: -11.2893362798, 630: -0.879173071539, 631: 5.46686344096, 632: -1.34890621899, 633: -11.1326783751, 634: -11.3048763439, 635: 8.8496106835, 636: 0.893973074302, 637: 1.68152807335, 638: 2.00812468336, 639: 2.09805320035, 640: 1.91676104775, 641: -8.5660427116, 642: 3.54251025988, 643: -2.57940823946, 644: 11.861433648, 645: 4.76137678507, 646: -0.642652336476, 647: 8.93974508878, 648: -1.78859583506, 649: -0.486972310448, 650: -4.19161679933, 651: -9.97817677032, 652: -0.351819912605, 653: 14.013328087, 654: -8.26151018519, 655: 0.444542448858, 656: 9.7170835147, 657: 0.649365307065, 658: 0.709249076824, 659: 1.6239874597, 660: -5.64239918227, 661: -0.668375140619, 662: 0.483879328363, 663: 12.9895745959, 664: 0.0361162185492, 665: 1.33445660056, 666: -0.508313327407, 667: 0.643678331556, 668: -11.4733893479, 669: 0.52193670415, 670: 0.0444088300798, 671: -1.11016597527, 672: -7.02926402632, 673: 2.55810482123, 674: 7.96577988211, 675: 11.1589128044, 676: 2.11335537075, 677: -18.2525711912, 678: 6.94767099136, 679: 1.46505221759, 680: 5.44081346327, 681: 4.937675599, 682: -1.6531045696, 683: -1.6936589767, 684: -2.45746748373, 685: -12.1773682539, 686: 7.1821764256, 687: -1.05096512112, 688: 1.00941378888, 689: -6.87075571361, 690: -4.31079695058, 691: 0.730784864307, 692: 2.48126964053, 693: 1.77279693471, 694: -1.45557799456, 695: -21.2259427333, 696: 5.02031154162, 697: -4.56032192193, 698: -0.625891232868, 699: -1.6359135219, 700: -38.1341770374, 701: -54.7151132283, 702: -11.1534204142, 703: -74.4188629271, 704: 23.7492975674, 705: 6.20445148005, 706: 8.10988482899, 707: -8.22235150208, 708: -17.0194040753, 709: -8.21996683411, 710: -32.267261556, 711: 55.9101523642, 712: 27.2285961583, 713: -21.0138288548, 714: 44.3700156864, 715: 100.0, 716: 47.5265716717, 717: -7.51382128465, 718: -58.8746313573, 719: -3.82348328046, 720: -4.91819449233, 721: 7.3661307698, 722: 28.7752255751, 723: -57.1944255067, 724: -14.5986848193, 725: 2.36875640835, 726: 5.35959306984, 727: 58.7112982293, 728: -11.253803431, 729: -36.4071109184, 730: -40.2560188204, 731: 6.58274543218, 732: -0.779024065259, 733: 100.0, 734: -30.8173217079, 735: -2.58662553218, 736: 11.711782092, 737: -37.5952517023, 738: -6.83889276651, 739: -19.2261308609, 740: -10.2655011217, 741: -42.81899649, 742: -10.4635361011, 743: -38.1350113499, 744: -22.0269942087, 745: 1.04370195613, 746: -8.31029580647, 747: -7.87198037042, 748: -11.4737739603, 749: -11.7612611189, 750: 16.742827247, 751: -13.6716298029, 752: 5.85786872685, 753: 31.8775442528, 754: -7.09346561808, 755: -31.7600570994, 756: -10.9542755974, 757: 12.2548360984, 758: 0.401589161043, 759: 23.4075346299, 760: -15.9644907276, 761: -9.81900201186, 762: 73.5321633759, 763: -26.5139434918, 764: -4.58695138544, 765: -8.2759146993, 766: -12.2091699141, 767: -18.5537720941, 768: -13.9520814922, 769: -10.0526502926, 770: -6.84921292094, 771: -10.5524346289, 772: 72.8243963387, 773: -25.6174457875, 774: -2.18217275032, 775: -10.7203890737, 776: -21.5027436873, 777: 15.1448035844, 778: -34.5167492373, 779: 1.50274833263, 780: -26.4063323017, 781: -7.12587808751, 782: -4.20338224185, 783: 67.6746640282, 784: -9.48414905605, 785: -16.5551455314, 786: 11.6066243924, 787: -10.8873240462, 788: 7.04577794286, 789: -1.56842966575, 790: -10.4979465491, 791: -9.94399285662, 792: -9.89392599268, 793: -7.65987530082, 794: -8.90939889278, 795: -17.3255435474, 796: -6.41698423111, 797: -9.44289719326, 798: -28.9642773007, 799: -12.5991993447, 800: -10.0577704061, 801: 33.5007834597, 802: -4.16748937922, 803: 4.74469096312, 804: 73.3333416896, 805: -14.4308304867, 806: 6.71730250035, 807: -8.19815832693, 808: 41.4962680742, 809: -9.32562207201, 810: -66.6318377026, 811: -21.0880963331, 812: 97.2556900463, 813: -34.340181459, 814: 10.2570305908, 815: -24.0194318479, 816: 0.0560242834573, 817: 85.0733201443, 818: -53.7058539232, 819: 9.88421957292, 820: -13.1058398573, 821: 9.77392819119, 822: -19.4542052686, 823: 17.523368565, 824: -1.88411610203, 825: -0.702871829357, 826: -0.969257709185, 827: -7.23860680208, 828: -1.86337747039, 829: -13.0817315495, 830: -10.6574374785, 831: 3.41736198666, 832: -1.1051033325, 833: -9.00620308804, 834: 6.17997449912, 835: 8.50796659797, 836: 2.89063161401, 837: -0.243040413888, 838: -7.58695189456, 839: 4.12800458009, 840: -5.15719432888, 841: -6.54972964002, 842: -2.8905066395, 843: 5.98084901102, 844: -1.85459880449, 845: -1.85960688653, 846: -2.20120410978, 847: -1.95714050017, 848: -6.75153854853, 849: 26.6479361495, 850: -2.97926393167, 851: -1.92559712604, 852: -2.41325163309, 853: -2.2918583939, 854: -7.71015259656, 855: 20.1927896528, 856: -2.78477021077, 857: -2.75357698517, 858: -3.93461019187, 859: 0.446723552386, 860: -8.47779536713, 861: 0.832381542708, 862: -6.37246379108, 863: 42.3475457927, 864: -41.239629614, 865: -1.81948007888, 866: -1.60360531383, 867: -26.2480934948, 868: -5.3061323453, 869: -7.77048386409, 870: -2.63713207356, 871: -5.17552842552, 872: -1.91632005353, 873: 5.75060702254, 874: 26.6008963796, 875: 27.2649184548, 876: -7.53044542366, 877: -1.94481731779, 878: -1.94749372524, 879: -2.50299588806, 880: -2.22891691061, 881: -1.89115860088, 882: 73.715382324, 883: 14.4351951925, 884: 31.3888360396, 885: -15.9077258978, 886: 22.6030133203, 887: 9.09722786812, 888: 19.4024150518, 889: -16.9299866791, 890: 46.1201891844, 891: -1.7704119496, 892: -3.6023448413, 893: 1.00191477896, 894: 16.8463350846, 895: 8.8158042795, 896: 86.2162420278, 897: -1.89734050424, 898: -9.30643542184, 899: -1.93649530701, 900: 18.9835260342, 901: -1.83362654121, 902: -1.91011431693, 903: -8.0184334872, 904: -10.0565914202, 905: -25.2489290261, 906: -1.86125817558, 907: -2.03016971908, 908: -0.761511906971, 909: -1.82765350858, 910: -0.183150023635, 911: -24.6313249714, 912: 43.9096913291, 913: -1.22232181905, 914: -2.60355578655, 915: 35.7990666552, 916: 4.82551242332, 917: 10.9146942636, 918: -19.6282231022, 919: 15.9144172523, 920: 21.1972962603, 921: -6.74893249607, 922: -7.66704571849, 923: 30.3610478335, 924: -4.33528934437, 925: -67.6190138484, 926: -6.92829447421, 927: 9.86668078716, 928: -6.4803850338, 929: 12.5894417942, 930: -2.28863283255, 931: 26.3440166098, 932: 29.355241836, 933: -9.80507697983, 934: -25.7662714937, 935: 5.60530203993, 936: 36.2768437972, 937: -10.1788633841, 938: -13.0889535076, 939: -21.8016200848, 940: 46.8952008637, 941: -49.6642404103, 942: 9.91747252459, 943: -11.3792383831, 944: -58.8904536897, 945: 60.4386501569, 946: -44.2463919355, 947: 22.0781897881, 948: 66.4747867855, 949: -56.1261606591, 950: 46.3072152915, 951: 35.8703029862, 952: -27.1450955971, 953: -16.334362984, 954: -10.822563375, 955: 0.268748711745, 956: -8.02068765027, 957: -6.54196151675, 958: 27.740621834, 959: 8.68289072845, 960: -10.3837660166, 961: -2.66796760514, 962: 17.7511112153, 963: -7.71475336412, 964: 12.4698291347, 965: 89.6963393408, 966: -0.971966219802, 967: 51.2240503503, 968: -5.92886846438, 969: 6.81730138096, 970: 21.0144722393, 971: -36.6430220754, 972: 53.0987822363, 973: -52.7813781963, 974: 39.5301720686, 975: 82.6263723218, 976: -25.1737401013, 977: -3.50540161935, 978: 0.467711274727, 979: 15.9687236548, 980: -15.2847826094, 981: -38.3113096613, 982: 2.10546884615, 983: -44.5139487551, 984: 56.2495767062, 985: -10.3746271253, 986: -2.44474523181, 987: -2.53256538139, 988: -13.4652537714, 989: 20.8064707259, 990: 32.0960941937, 991: 59.2094263746, 992: 4.50391492947, 993: -64.6354523423, 994: 6.75813124979, 995: -47.3893768902, 996: 10.0058945349, 997: -10.169328757, 998: 29.261229277, 999: -11.0051028419, 1000: 17.0346158213, 1001: 9.09377639428, 1002: 57.1487099405, 1003: 30.5749661197, 1004: -7.48632474703, 1005: -47.3766501291, 1006: -9.26885340011, 1007: -9.31160917735, 1008: -9.9276857268, 1009: 4.16930398802, 1010: -11.2100600375, 1011: -10.5310801208, 1012: -6.09086567043, 1013: -10.046269469, 1014: -3.37630212639, 1015: -10.5742240172, 1016: -10.4230178847, 1017: -12.3783792331, 1018: -10.0107776577, 1019: -2.43418306534, 1020: -16.3551052626, 1021: 26.9211640839, 1022: -73.8859926777, 1023: -35.7710578934, 1024: 31.6269506189, 1025: -3.11321745267, 1026: -3.77249597832, 1027: 41.2531838512, 1028: 16.9178885763, 1029: -18.0004738498, 1030: -24.6657528892, 1031: -5.00790952646, 1032: 10.5209661268, 1033: -20.008099021, 1034: -22.9188395772, 1035: 17.9395743768, 1036: -3.42209541822, 1037: -8.78682495558, 1038: 53.8772528392, 1039: 21.2921409827, 1040: -54.2754909445, 1041: -13.985413831, 1042: -28.8689177806, 1043: -2.23408033136, 1044: 49.2751077375, 1045: -36.1322722463, 1046: -13.2796388788, 1047: 38.4092405379, 1048: 0.205256024768, 1049: 2.2611344646, 1050: 0.403524704995, 1051: 1.89915065184, 1052: -3.29577621754, 1053: -1.01676164984, 1054: -1.515825405, 1055: 0.652606965802, 1056: -1.34644460944, 1057: 3.11515676927, 1058: -3.62606148121, 1059: 3.01830575736, 1060: 0.372297465748, 1061: 1.61260197036, 1062: -0.206618324206, 1063: -1.08223687166, 1064: 0.985975255809, 1065: 0.554723017339, 1066: -0.519415085651, 1067: -0.174185895548, 1068: -0.815744956338, 1069: -0.778420596112, 1070: -0.143977884099, 1071: 3.18252792531, 1072: -0.335258769292, 1073: -0.927419854088, 1074: -2.68165908098, 1075: -2.13776610959, 1076: -2.87612164047, 1077: -0.770401529873, 1078: 1.39471464199, 1079: 4.97653782564, 1080: 1.38583293644, 1081: 1.24812573079, 1082: -1.32022726882, 1083: -0.115469842929, 1084: -1.11526068093, 1085: -0.565114298859, 1086: 1.7417593941, 1087: -0.530012839855, 1088: 1.08219363625, 1089: 2.99532020168, 1090: 6.04401934696, 1091: 5.79596628195, 1092: 2.38274254129, 1093: 2.6976779172, 1094: -0.937893810592, 1095: 1.22487142377, 1096: -2.54501238594, 1097: -2.1463145574, 1098: -3.43813337574, 1099: 0.261625510584, 1100: -1.57896957149, 1101: 0.690778025011, 1102: -3.56563195762, 1103: 0.319968443622, 1104: 3.85690203815, 1105: -2.97889969105, 1106: -0.765618957107, 1107: -2.59226506059, 1108: 2.49621207846, 1109: -3.6881190817, 1110: 5.25481368572, 1111: 3.91782992912, 1112: -2.44068465345, 1113: 0.284877119235, 1114: 4.78242971592, 1115: 0.0544985420649, 1116: 2.8266116635, 1117: 0.551430863374, 1118: 0.340916474328, 1119: -2.94340275275, 1120: 0.381219982322, 1121: 1.44524013489, 1122: -0.184458916107, 1123: 2.12390325601, 1124: 0.118002923495, 1125: -0.112290191531, 1126: 3.6871739448, 1127: 3.95020557455, 1128: -4.05215394231, 1129: -5.9719184851, 1130: -0.87943544438, 1131: -2.07059121181, 1132: 1.20542278258, 1133: -3.67431144767, 1134: -2.76464694678, 1135: 2.18664073475, 1136: -0.439648959417, 1137: -2.33587851388, 1138: 0.125632072141, 1139: 0.410684350525, 1140: 0.561966904229, 1141: 0.0041225977436, 1142: 0.250730278378, 1143: 0.726577418709, 1144: -4.82445376347, 1145: -0.850854142441, 1146: -2.3784875598, 1147: -1.60400720555, 1148: 0.852092195142, 1149: 0.425413375301, 1150: -3.60652094266, 1151: 0.124149525296, 1152: -1.80583181346, 1153: -4.96945602112, 1154: 0.552046184965, 1155: 1.05657356672, 1156: -0.615662551273, 1157: -4.48618977229, 1158: 0.759511811346, 1159: -0.80595633214, 1160: 7.79899794214, 1161: 1.85499589175, 1162: -3.0347173459, 1163: -0.660384137523, 1164: 2.24660384742, 1165: 1.33474958718, 1166: -1.10724707369, 1167: -4.77854215018, 1168: 5.23560905998, 1169: 3.35530872618, 1170: -2.78436990516, 1171: 0.52501060578, 1172: 5.57937280295, 1173: 0.178561960196, 1174: 0.1991513377, 1175: -1.65744754448, 1176: 7.51917338808, 1177: 0.349907608669, 1178: 0.369838542075, 1179: -0.185074704865, 1180: 5.0710191397, 1181: -0.104265974678, 1182: -0.283332740474, 1183: -4.58322130247, 1184: 0.906660028268, 1185: -1.69309669592, 1186: -1.7511547297, 1187: -2.93604372405, 1188: 2.23436255415, 1189: 0.545461585086, 1190: 1.94902235285, 1191: 0.346873931808, 1192: -1.44706466982, 1193: 0.32346874563, 1194: 0.271561157239, 1195: 0.782172046945, 1196: 0.121203008094, 1197: 0.562993304706, 1198: -0.0274560289657, 1199: 0.200091631733, 1200: 0.079203945774, 1201: 0.325848144855, 1202: 0.149390532908, 1203: 0.336020468167, 1204: -1.18170654348, 1205: 0.521704758157, 1206: 0.0847195580149, 1207: 0.605315514928, 1208: -0.403777982579, 1209: 1.74692753416, 1210: 0.601713532157, 1211: 0.655358769433, 1212: -13.5157663231, 1213: 3.95663412591, 1214: -1.77078969843, 1215: 0.435639438719, 1216: 3.41047003944, 1217: -3.31515839758, 1218: -2.55827771224, 1219: 0.183030868479, 1220: -3.94251329861, 1221: 0.0623650659824, 1222: -1.0137820345, 1223: -0.09689425209, 1224: 1.16863256839, 1225: 0.446004065434, 1226: 0.279487421972, 1227: 0.465523646501, 1228: 0.292968764451, 1229: 0.0579666546067, 1230: 1.44492943886, 1231: -0.131968232418, 1232: 1.47888218973, 1233: 8.44119805446, 1234: -0.608097592485, 1235: 0.137159807586, 1236: -1.08900235515, 1237: -0.420933929794, 1238: 3.30353377817, 1239: 1.88783045659, 1240: -0.00355319905911, 1241: 3.18975330041, 1242: -0.113148162191, 1243: -1.96377903687, 1244: 0.493320502229, 1245: 1.77823402872, 1246: 0.13909128044, 1247: -0.042088539127, 1248: 0.111266378484, 1249: -1.91228035656, 1250: 0.278500785294, 1251: 0.246018339265, 1252: -0.402441774041, 1253: 0.350308315473, 1254: -7.0449834063, 1255: 0.213779601508, 1256: 0.176173516849, 1257: -0.170234195247, 1258: 0.125839191537, 1259: 0.1367985393, 1260: -2.92403870259, 1261: 2.1601859384, 1262: 0.246116509201, 1263: 0.201134462884, 1264: -4.21865835173, 1265: 0.0949863257276, 1266: 5.76853925228, 1267: -6.12930511942, 1268: 0.485140206517, 1269: -0.740611571287, 1270: -0.00135156991069, 1271: -0.00227998746816, 1272: 2.85042448925, 1273: 0.399457555483, 1274: 1.05528938656, 1275: 0.639308665416, 1276: 0.740176211756, 1277: -0.155399339025, 1278: -0.774933753739, 1279: 0.279794203925, 1280: 1.02557710953, 1281: -0.346979811501, 1282: -0.404715451136, 1283: -2.95482032903, 1284: 4.22064014973, 1285: -4.04230041003, 1286: -1.23392518791, 1287: 0.448763871198, 1288: 2.3004940282, 1289: -4.0130178152, 1290: -6.43395141377, 1291: -1.88636972475, 1292: -0.269266542876, 1293: 2.19652818672, 1294: 1.0885236837, 1295: 2.04631569633, 1296: -4.2815076544, 1297: -0.411376244429, 1298: -3.17127047381, 1299: -3.25145088697, 1300: 4.19451074556, 1301: 0.88769923108, 1302: -0.0123869866119, 1303: -0.19245139861, 1304: -2.12916583254, 1305: 0.327768463894, 1306: 0.553315413023, 1307: 2.39026157635, 1308: 0.954942366508, 1309: 0.507546511272, 1310: -0.629826049987, 1311: 0.33829463089, 1312: 0.344589486728, 1313: 0.478599016814, 1314: -1.91077275998, 1315: -2.29244580292, 1316: -1.14353467444, 1317: 1.47249720895, 1318: -0.0694161112815, 1319: -2.66956652691, 1320: 1.68992629411, 1321: 9.74154589102, 1322: -0.86166360962, 1323: -0.0376247950674, 1324: -0.575734217559, 1325: 3.95430315612, 1326: -2.14699094895, 1327: 2.10594397493, 1328: 1.30949265677, 1329: 3.38991777567, 1330: -1.26577286643, 1331: -2.78317936997, 1332: 5.0154185923, 1333: -4.10445459769, 1334: 0.460134459049, 1335: -1.57838347488, 1336: -1.08193770653, 1337: 1.18508118235, 1338: 0.342740236367, 1339: -1.72520514509, 1340: 1.55520741887, 1341: -3.4522331584, 1342: 4.61089377143, 1343: -1.60998724285, 1344: -3.59875357856, 1345: -1.91157255668, 1346: -0.528648978891, 1347: -0.33764414443, 1348: -1.57172587309, 1349: -1.59180793471, 1350: -2.20521171605, 1351: 2.11819637703, 1352: -2.79647262158, 1353: 0.424549163216, 1354: -4.45915284019, 1355: 0.0050446365762, 1356: 0.023410188253, 1357: -1.41799241259, 1358: 2.98614223386, 1359: 0.0888480715156, 1360: -0.930107712427, 1361: 2.93783427496, 1362: 0.392506448116, 1363: 3.9243595506, 1364: -0.102319219613, 1365: -0.0239123723866, 1366: -4.83483690053, 1367: 0.379819630897, 1368: 0.179939608799, 1369: 2.54022171208, 1370: 3.21914102128, 1371: -4.5432856305, 1372: 0.69410013635, 1373: 4.64105317905, 1374: -0.920771425955, 1375: -17.7012643658, 1376: 5.26382601842, 1377: -4.67727322099, 1378: -0.60193742933, 1379: -2.39737530334, 1380: -2.7833291282, 1381: -0.89361726894, 1382: -2.74283614619, 1383: 0.949196849624, 1384: 0.0222412966846, 1385: 6.32263622007, 1386: -0.683700163296, 1387: -3.57507946447, 1388: 2.76680392993, 1389: 3.00357819479, 1390: 0.32137542781, 1391: -0.56538119226, 1392: -0.435066539747, 1393: -4.97329733864, 1394: 1.77809869038, 1395: 3.60815166479, 1396: -5.50841321666, 1397: -0.350806693766, 1398: 17.2432976505, 1399: 8.43022018272, 1400: -17.2666905449, 1401: 44.0833189765, 1402: 16.0767420458, 1403: 18.0071671864, 1404: 6.87229413432, 1405: 31.7837800943, 1406: -10.1936068546, 1407: 74.10957229, 1408: 3.56998400226, 1409: 17.1974987652, 1410: -2.69768542696, 1411: 0.439235396149, 1412: 27.6902141886, 1413: -12.2239583326, 1414: -9.96386905628, 1415: -42.6269416249, 1416: -43.5705060917, 1417: -12.8741991219, 1418: 2.35845352845, 1419: 24.4839636464, 1420: -4.57274643541, 1421: -20.7329567404, 1422: 1.08793226184, 1423: -9.06122990797, 1424: 28.2404218417, 1425: 24.0103376039, 1426: -13.8732307721, 1427: -30.0368506353, 1428: -29.5644576532, 1429: -0.296852162918, 1430: -0.663276582449, 1431: 21.1811005281, 1432: -17.8250613148, 1433: 7.94272717102, 1434: 18.1835602183, 1435: -2.96720497026, 1436: 36.0206442701, 1437: 16.6853256632, 1438: -1.54050504426, 1439: 30.3388069701, 1440: 35.4135869684, 1441: -24.9613213052, 1442: -29.5854694659, 1443: 84.3602969494, 1444: 9.54826283107, 1445: -5.86289935172, 1446: -6.26429173503, 1447: 6.19704075027, 1448: 57.7463881824, 1449: 15.4758399178, 1450: -8.69421537243, 1451: -35.9889124873, 1452: 3.70623164481, 1453: -99.2392120104, 1454: 4.02347534664, 1455: -37.6729218883, 1456: -34.0594334809, 1457: 58.997254491, 1458: 35.4464995648, 1459: -67.4711329848, 1460: 15.5431834883, 1461: 21.5275590528, 1462: 3.1923036534, 1463: 59.2292049615, 1464: -5.58323848745, 1465: 6.44952889201, 1466: 4.64206173609, 1467: 13.0371538647, 1468: 4.08043889292, 1469: 1.61320902332, 1470: -19.8097211781, 1471: -13.7298684987, 1472: 1.96460091714, 1473: 7.74490444342, 1474: -0.541743277442, 1475: -12.3798159144, 1476: -24.0386029774, 1477: 37.7634850428, 1478: 20.327388435, 1479: -4.84552493506, 1480: 7.49787525638, 1481: 8.06748383122, 1482: 8.09905426723, 1483: 24.070971061, 1484: -11.5118077582, 1485: -5.40327756114, 1486: 36.3089449241, 1487: 9.70179145974, 1488: -1.52339090077, 1489: -24.0153009634, 1490: -2.92889102915, 1491: -1.12578772, 1492: -2.54533490081, 1493: -30.0696641267, 1494: -5.3932166408, 1495: 6.5003005832, 1496: -10.363815194, 1497: -8.44147724686, 1498: -3.19532887515, 1499: 5.17325903183, 1500: 1.77439197799, 1501: -14.6578296386, 1502: 100.0, 1503: -0.413036905701, 1504: 15.4277856378, 1505: -5.48191549413, 1506: 51.5171021599, 1507: 8.36470567022, 1508: -4.15774096121, 1509: 65.6148196419, 1510: 37.9008831045, 1511: -61.527206652, 1512: -8.13332154703, 1513: -22.4974620123, 1514: 0.512996514523, 1515: 12.6040539603, 1516: 2.65977104808, 1517: -78.1229350216, 1518: -12.257208109, 1519: -36.2867456421, 1520: -1.92260884703, 1521: -21.7200835656, 1522: -0.310431237389, 1523: -0.566735030563, 1524: -19.5547171604, 1525: -31.9012130248, 1526: -0.522559069876, 1527: -1.36150002638, 1528: 5.33875224006, 1529: -46.3715312511, 1530: 1.41852006067, 1531: 10.0006104987, 1532: 98.4252953979, 1533: -5.6628282986, 1534: -23.7135941666, 1535: 8.47887776826, 1536: -10.2493503199, 1537: -46.4413305438, 1538: -22.7948287598, 1539: -35.7447970509, 1540: -6.27767557866, 1541: -18.6664326804, 1542: -0.428629055378, 1543: -0.522695622339, 1544: -4.86397719905, 1545: -0.414267532314, 1546: -1.93285136225, 1547: 4.42343046072, 1548: -1.30638973194, 1549: -0.454864807079, 1550: -0.881928461899, 1551: -0.240425070521, 1552: 0.874443223522, 1553: 46.156154247, 1554: 0.556639525317, 1555: -2.90760651252, 1556: -3.69615432024, 1557: 21.1456467318, 1558: 2.87400353106, 1559: 9.28337939578, 1560: -7.9685421721, 1561: -64.3079490232, 1562: -57.6465058943, 1563: 9.62278230555, 1564: -7.5456092077, 1565: 21.8812058007, 1566: 2.5108002383, 1567: 14.3317487491, 1568: -0.723009931802, 1569: -16.5442859443, 1570: -0.489926274958, 1571: 23.1966110925, 1572: 72.4364107354, 1573: -53.4272847538, 1574: 0.257914268761, 1575: -0.519483677168, 1576: 0.868790180006, 1577: 1.61692185776, 1578: -1.41605259051, 1579: -7.75383469615, 1580: -50.191244888, 1581: 23.2022647478, 1582: -70.4726730461, 1583: -1.27620961946, 1584: 43.9493520626, 1585: -28.9450724427, 1586: 15.2982967195, 1587: -9.11224038721, 1588: -59.9401212452, 1589: -1.19222430546, 1590: -100.0, 1591: -10.8247708531, 1592: -3.83783601952, 1593: 3.48914214426, 1594: 40.5692551195, 1595: -0.256733070283, 1596: -4.51758478753, 1597: -0.280841747788, 1598: -54.998087001, 1599: -0.59804320255, 1600: -0.373440052502, 1601: -56.4673790544, 1602: -1.6242103663, 1603: -41.7995189712, 1604: -0.813746708714, 1605: -0.457047731875, 1606: -4.16760753014, 1607: -0.033078203621, 1608: -0.220816267479, 1609: 4.71464992139, 1610: -25.6842277004, 1611: -0.859816777528, 1612: -1.39708396968, 1613: -12.4042549642, 1614: 20.1056155918, 1615: 21.6505622948, 1616: 77.4501393923, 1617: -15.3785582039, 1618: -3.79604630787, 1619: -17.2893753349, 1620: -0.131127528328, 1621: 17.072751783, 1622: -1.21078639851, 1623: -47.1013721033, 1624: -5.44336721945, 1625: -2.59268519846, 1626: -15.1705626641, 1627: 37.2966335596, 1628: -0.727480777648, 1629: -14.6651827865, 1630: 1.43890178941, 1631: -1.75904684006, 1632: -60.3315788836, 1633: -30.1090062758, 1634: -9.80249648636, 1635: 3.1761601148, 1636: -1.19129418213, 1637: -58.326691335, 1638: 44.0725311268, 1639: 27.6064283169, 1640: -56.0554392077, 1641: -36.8843821625, 1642: 5.41351631897, 1643: 17.2565206296, 1644: 15.2495218133, 1645: 37.2862525706, 1646: 77.8718952037, 1647: 22.3796045695, 1648: 62.147158685, 1649: 34.9140179376, 1650: 33.9583204969, 1651: 1.77879296894, 1652: 5.95976065319, 1653: -34.6868663772, 1654: -2.5983103172, 1655: -1.92674519033, 1656: 6.58314738458, 1657: 50.9748942621, 1658: -1.63206519954, 1659: 15.6672420027, 1660: -7.43426090772, 1661: 0.900889313555, 1662: 8.9891950348, 1663: -52.6650363057, 1664: 14.132596096, 1665: -4.01200840687, 1666: 16.3850146202, 1667: -4.86661117258, 1668: 3.66173013488, 1669: 10.7155402772, 1670: 17.2998752012, 1671: -65.700523671, 1672: -22.5870243894, 1673: 19.2949461996, 1674: -31.4770159297, 1675: -3.25924492313, 1676: -22.2071498033, 1677: 13.6162163501, 1678: 9.90642396038, 1679: 20.1217143853, 1680: -37.4894636081, 1681: -30.0311937743, 1682: -32.7775952081, 1683: 1.23919145253, 1684: -10.3323338348, 1685: -19.877763482, 1686: 3.16037736974, 1687: 18.938451583, 1688: 21.2622950781, 1689: 0.784131204378, 1690: 31.4580477757, 1691: 2.73827196939, 1692: 56.8324297421, 1693: -20.6352092374, 1694: 34.6024464001, 1695: -8.32895130866, 1696: -16.9303606259, 1697: -44.9521752678, 1698: 6.36758938799, 1699: 11.8538731726, 1700: 39.7204325305, 1701: 41.2419202112, 1702: -1.19153200629, 1703: 36.2266934241, 1704: -4.36299033067, 1705: -4.56560851493, 1706: -1.72737364201, 1707: 1.99727980391, 1708: 2.20105183925, 1709: -3.68998141885, 1710: 47.133729115, 1711: -1.75117632727, 1712: 99.861986465, 1713: 9.6626340362, 1714: -1.4556401009, 1715: 32.8202578531, 1716: -13.775968925, 1717: -13.5119698212, 1718: -28.6324705515, 1719: -1.73982680967, 1720: -3.67956592478, 1721: 5.54224572176, 1722: -61.2328466519, 1723: 3.75049675385, 1724: 39.7880278959, 1725: -76.4797389285, 1726: 21.761145824, 1727: 50.1904342116, 1728: 21.842731156, 1729: 2.23340449189, 1730: 39.3784109561, 1731: 8.758633235, 1732: 39.5166462102, 1733: -0.440867449775, 1734: -88.0685892388, 1735: -3.4820268808, 1736: 57.4508268343, 1737: -50.1390296404, 1738: 7.10718780307, 1739: 29.0107899705, 1740: -1.56575127301, 1741: -14.0026706995, 1742: 31.6606875207, 1743: -9.4695142694, 1744: -33.3045852962, 1745: -9.26117267186, 1746: -1.49827313764, 1747: -13.9319947312, 1748: 10.0610714273, 1749: 1.92313612327, 1750: -68.8627797704, 1751: 18.0003310722, 1752: -32.4663194284, 1753: 49.3379336788, 1754: 52.5359180412, 1755: 47.5957107261, 1756: 58.6275208435, 1757: 38.1499000821, 1758: -0.15185044049, 1759: 43.151038183, 1760: 67.6943502772, 1761: 1.12307840001, 1762: 1.42764005173, 1763: -12.6255055086, 1764: -0.348083051297, 1765: -4.0195728963, 1766: 30.5728946132, 1767: -30.4910522091, 1768: -21.1252621725, 1769: -14.8961150301, 1770: 99.9850556384, 1771: 54.7789064555, 1772: 21.6882186462, 1773: 11.1743537004, 1774: -85.3189981362, 1775: 42.8530515641, 1776: 46.0577569123, 1777: 27.8484924817, 1778: -28.1504952314, 1779: -11.8421781138, 1780: 92.3735082599, 1781: 32.0420986485, 1782: -14.6010068921, 1783: 3.29178748534, 1784: -16.5782967333, 1785: 62.9316479871, 1786: -20.3113818818, 1787: 29.8349620532, 1788: 23.7675977989, 1789: -16.2955177754, 1790: -8.38446055987, 1791: 70.1252189827, 1792: 6.06497400302, 1793: -6.97176542678, 1794: -3.92263147496, 1795: 23.8727734173, 1796: -1.01706647148, 1797: -15.9392532105, 1798: -6.80105711595, 1799: 10.6723888405, 1800: 3.22544387921, 1801: -1.13892177644, 1802: -81.1535845947, 1803: -6.5323109308, 1804: 2.5027132977, 1805: 64.6006827894, 1806: -32.6729518804, 1807: 35.784793113, 1808: -27.5985235166, 1809: -32.5243010313, 1810: -16.7989433267, 1811: 33.4293747921, 1812: 38.6129620626, 1813: -5.07853843789, 1814: 3.6193550431, 1815: -6.83228818685, 1816: -6.80198241944, 1817: 10.8153791183, 1818: -6.84270897477, 1819: 33.6185653639, 1820: -8.36237463967, 1821: -7.03419597322, 1822: -2.92249443132, 1823: -6.0477179054, 1824: 6.65007996337, 1825: -47.7474297273, 1826: -6.43517965705, 1827: 18.1698815369, 1828: -6.99154504569, 1829: -6.69637197127, 1830: -43.0791548994, 1831: -6.58711005186, 1832: 50.5684413725, 1833: 57.9106741569, 1834: 26.1486711634, 1835: -3.66233822862, 1836: -6.90726921963, 1837: -6.91251904435, 1838: -24.0572471694, 1839: -6.81972160635, 1840: -6.73689104772, 1841: -6.85951359392, 1842: -5.49942458941, 1843: -6.89877511885, 1844: -7.04137352804, 1845: -5.91524812867, 1846: 30.4533473985, 1847: -6.86892479852, 1848: 49.5672386074, 1849: -5.26098062594, 1850: -6.88164868879, 1851: -1.5672535803, 1852: -6.88211609623, 1853: -6.99018201421, 1854: 26.1970023846, 1855: 52.9921164984, 1856: -20.43359842, 1857: -8.91104739252, 1858: -64.9286713241, 1859: -24.7486389479, 1860: -30.3034926117, 1861: 33.2347052251, 1862: -3.8095699096, 1863: 1.5039247679, 1864: 39.9764069702, 1865: 10.4395023226, 1866: -16.8672875392, 1867: -12.694158459, 1868: -17.5022853901, 1869: -65.5658248075, 1870: 16.3694706967, 1871: -1.59262040334, 1872: -2.78199721203, 1873: -0.445414588633, 1874: 22.582468148, 1875: -1.59723358109, 1876: -6.8714729827, 1877: -5.39185963812, 1878: 0.450592497315, 1879: -0.898308579008, 1880: -2.21973014698, 1881: -10.2987065431, 1882: 2.68485682336, 1883: -29.0090858634, 1884: 0.562915378703, 1885: 1.44241382224, 1886: -21.2879408559, 1887: -27.9361879138, 1888: 5.27286269757, 1889: -1.24466199463, 1890: -12.7569648956, 1891: -1.65433730559, 1892: -1.51069471009, 1893: -16.7599331381, 1894: -1.39334264101, 1895: -7.01116141803, 1896: 2.97522853272, 1897: -2.3990411222, 1898: -1.67341645718, 1899: -2.11486695245, 1900: -1.59720815925, 1901: -3.81083569816, 1902: 64.0680526977, 1903: 3.62441120182, 1904: -2.23229185604, 1905: -2.72974841267, 1906: -4.16498807089, 1907: -2.18832194776, 1908: 55.0757665235, 1909: -11.2580035113, 1910: 32.4396920859, 1911: 3.78024207048, 1912: -0.706244445763, 1913: -2.87885474926, 1914: 13.1290548929, 1915: -26.2460972675, 1916: 38.3801645031, 1917: -2.26897970803, 1918: -30.8577657389, 1919: -1.57132647883, 1920: -2.03167951307, 1921: 7.09725563192, 1922: 38.0611221516, 1923: -4.53117563186, 1924: -1.66283445274, 1925: -1.4410931906, 1926: -1.6797488426, 1927: -1.58152752587, 1928: -2.96690716349, 1929: 23.3278689658, 1930: 1.51548557221, 1931: 44.6503658829, 1932: -15.3946526397, 1933: -10.4860467539, 1934: -4.85028634112, 1935: 3.31142625405, 1936: 28.6468426225, 1937: 46.191845708, 1938: -2.08318377883, 1939: -69.5365154201, 1940: -9.10204518422, 1941: -0.185318306572, 1942: -7.02142742226, 1943: 38.2953139015, 1944: -1.54415193717, 1945: -6.82898979458, 1946: -2.61143569412, 1947: 15.1365777464, 1948: -1.65098803983, 1949: -1.70302243832, 1950: 6.0015442285, 1951: -6.87993418286, 1952: 10.956967083, 1953: -1.66194511214, 1954: -1.61240208372, 1955: 0.938004518906, 1956: -1.53522188754, 1957: -10.6987769525, 1958: 15.8984685948, 1959: 44.5985797417, 1960: 0.992584989129, 1961: -0.688517346011, 1962: 16.1181299366, 1963: 1.93316073679, 1964: -25.0590996187, 1965: 55.1803413574, 1966: -5.69647154674, 1967: 16.9701993038, 1968: -0.985967277919, 1969: -5.86212527415, 1970: -4.02534364387, 1971: -2.09173096948, 1972: -23.1751932696, 1973: -5.74237825374, 1974: -5.46805200051, 1975: -6.06765931213, 1976: -12.9203490199, 1977: -1.68526699063, 1978: -23.5718945703, 1979: 35.5873687011, 1980: -5.67987241885, 1981: -51.9125863796, 1982: 34.9009553013, 1983: -43.5338513883, 1984: -6.99187518284, 1985: -6.82448112445, 1986: -21.2289150518, 1987: 19.7271667041, 1988: -16.2194365248, 1989: -39.6700163964, 1990: 10.5144052854, 1991: -31.4803427402, 1992: -23.7708755374, 1993: -28.8316914323, 1994: -26.1899603772, 1995: -1.1662800037, 1996: -8.19978487281, 1997: -0.198417706799, 1998: 18.8312457372, 1999: 23.2933409904, 2000: -6.94174144118, 2001: -7.04535184319, 2002: 33.5332558584, 2003: -10.8935709981, 2004: -7.02616144907, 2005: -14.8491784668, 2006: 25.3852721552, 2007: -6.9588284995, 2008: 9.04541144737, 2009: -6.79248731972, 2010: -3.77419249374, 2011: 35.3533389929, 2012: -40.444811787, 2013: 18.8519676932, 2014: 9.45638943689, 2015: -2.97036046176, 2016: -3.21541251579, 2017: -3.61908960685, 2018: 5.86596406956, 2019: 26.7095750814, 2020: 9.19093773865, 2021: 25.8544271151, 2022: -50.7587050695, 2023: 51.2710375762, 2024: -63.9328075266, 2025: 1.02144824377, 2026: -0.044355543657, 2027: 3.86531889057, 2028: 1.47768684584, 2029: -17.5951444962, 2030: -12.1267568814, 2031: -16.9200058439, 2032: -6.96365676703, 2033: -7.04555494981, 2034: -6.866422989, 2035: -7.45162298611, 2036: -7.07815237743, 2037: 8.40256222644, 2038: -11.791775055, 2039: 6.03655609024, 2040: -20.5176992806, 2041: 16.0715593413, 2042: 65.9446592439, 2043: -20.3672850441, 2044: 10.1084245259, 2045: -8.33559790957, 2046: -2.68950257064, 2047: 45.294366586, 2048: -14.4973711425, 2049: -31.5285856514, 2050: 14.4257606634, 2051: -6.4998706668, 2052: 25.4921454297, 2053: -7.00925410918, 2054: -6.81012749608, 2055: -6.83566703714, 2056: 20.0805260544, 2057: -6.95372014841, 2058: -7.03660566195, 2059: -19.8370555042, 2060: -6.99789324087, 2061: 16.5221542448, 2062: -7.0655387922, 2063: -6.84748767814, 2064: -13.6581266245, 2065: -6.85978832711, 2066: -14.3596833997, 2067: -9.413198713, 2068: -19.1466432311, 2069: 11.0885773325, 2070: -1.03582425305, 2071: 11.5434085305, 2072: -31.0280366347, 2073: -66.5200273379, 2074: 16.1900804496, 2075: 16.1715143112, 2076: -16.2628442666, 2077: 29.5443056349, 2078: -2.22481985061, 2079: -7.11412665616, 2080: -0.977112120322, 2081: -5.34048789555, 2082: -7.2619364206, 2083: 6.24098012539, 2084: -6.79149481052, 2085: 15.312415766, 2086: 13.8056576961, 2087: -7.73898330824, 2088: 96.8111843381, 2089: 20.0879138902, 2090: 18.0547964546, 2091: 3.5429568993, 2092: 22.1605875135, 2093: 32.8655611858, 2094: -36.5743065285, 2095: -0.266968986045, 2096: -5.4473364725, 2097: 2.80785461292, 2098: 4.34212860217, 2099: 4.30493244537, 2100: -2.19978745539, 2101: 10.4801621512, 2102: -12.5977255364, 2103: 10.1882403194, 2104: 2.28833831478, 2105: -9.6884152061, 2106: -8.9122301275, 2107: -4.88405666758, 2108: 8.23930151259, 2109: -1.33998059362, 2110: -2.98087482115, 2111: 24.1084354779, 2112: 6.14539523086, 2113: 0.423977701718, 2114: -4.35289810443, 2115: 25.1142980252, 2116: 2.16258287313, 2117: -12.9101718688, 2118: -8.65924544252, 2119: -6.1549885576, 2120: 34.9429416854, 2121: -13.3647487282, 2122: -4.50140550517, 2123: -24.0796515718, 2124: 0.128809954504, 2125: -13.2931884845, 2126: 67.0615999452, 2127: -3.40644052558, 2128: 3.97519851036, 2129: 10.6855496845, 2130: 10.3339423884, 2131: 8.66565327261, 2132: 34.494792622, 2133: 8.40563503997, 2134: 15.7040459549, 2135: -2.20986854384, 2136: -1.8118450872, 2137: -3.45126042585, 2138: -11.6151053741, 2139: -6.3953172303, 2140: -4.31053523713, 2141: -17.1885679173, 2142: -1.9395873514, 2143: -9.53500793968, 2144: -7.53923043704, 2145: 20.1964165895, 2146: 17.4880279548, 2147: -1.76335338978, 2148: 5.98167746992, 2149: 12.8349821921, 2150: -0.945681332999, 2151: -21.5600711288, 2152: -2.08059289244, 2153: 5.23678131923, 2154: 34.0524742248, 2155: -0.593139349929, 2156: 5.4902330345, 2157: -0.745890738858, 2158: 11.9045085444, 2159: 2.64902239688, 2160: -1.42690782082, 2161: 2.83504747005, 2162: -1.86560872266, 2163: -8.28399561687, 2164: -2.11899706769, 2165: -2.31339513544, 2166: 5.73942159067, 2167: -1.93764407807, 2168: -6.93192446455, 2169: 3.59754136178, 2170: -0.943349031275, 2171: -8.41114914527, 2172: 0.0510746121462, 2173: 8.24166232312, 2174: 19.4840877567, 2175: 1.71279179199, 2176: 4.07239474202, 2177: -2.89335654356, 2178: -3.5106742656, 2179: 1.27429109055, 2180: 3.87798655028, 2181: -1.39487378827, 2182: -8.75757316951, 2183: -1.07299166526, 2184: 0.205171417337, 2185: -2.17442412769, 2186: -1.73741249938, 2187: -17.3524756098, 2188: -2.10753815199, 2189: -0.882959919543, 2190: 4.10911154932, 2191: 10.1499605096, 2192: -2.21672084697, 2193: -0.573312479472, 2194: -3.13756532429, 2195: 3.85514040749, 2196: 1.45685394479, 2197: -14.3630063589, 2198: -1.6600481312, 2199: 4.9360240409, 2200: 15.496645228, 2201: -1.54938628155, 2202: -1.35147481028, 2203: -1.9679185828, 2204: -1.28908373569, 2205: -2.71794637266, 2206: -14.4139647731, 2207: 22.1437233229, 2208: -14.0353367514, 2209: 12.1667267628, 2210: -1.40540102746, 2211: -22.0235401016, 2212: 8.5641612852, 2213: 15.2171547952, 2214: 19.7714797493, 2215: 1.9432438367, 2216: -21.6289380614, 2217: -10.3591327571, 2218: -12.6908308607, 2219: -6.61183114725, 2220: -0.543482157109, 2221: 0.416053899141, 2222: -12.0022051574, 2223: -13.3237652702, 2224: -0.378263562094, 2225: -1.28739275727, 2226: -0.0346462536291, 2227: -40.6670584271, 2228: 0.77683111998, 2229: 2.29574717717, 2230: -14.4654111064, 2231: -0.396255665672, 2232: 12.8732950674, 2233: -2.80060337742, 2234: 3.14589860788, 2235: 14.8008970501, 2236: -8.36414341384, 2237: -11.5983769761, 2238: -1.29584377803, 2239: 4.30782928312, 2240: -0.183511118138, 2241: -0.382180710593, 2242: 28.4129757683, 2243: -0.253057332922, 2244: 0.903180531965, 2245: 6.89717430804, 2246: -0.376975415269, 2247: -0.208134923445, 2248: -0.439734077775, 2249: -0.250252020711, 2250: -2.4429009009, 2251: 36.7260668175, 2252: -0.863492873112, 2253: -0.842059671356, 2254: 1.87633531485, 2255: -2.44147626475, 2256: 0.479285324627, 2257: 23.828953409, 2258: 2.0875493482, 2259: -44.0244163237, 2260: -3.31515678613, 2261: -19.8144930126, 2262: 9.16364832866, 2263: -6.55335156335, 2264: -1.61339065885, 2265: -22.2959362658, 2266: -0.229480643767, 2267: -8.75133432808, 2268: -0.313499350335, 2269: 2.79984544771, 2270: 0.344261778141, 2271: -26.2369258332, 2272: 0.540321067109, 2273: -0.374507097032, 2274: -0.0482655987816, 2275: 0.416337982153, 2276: -0.0342838946889, 2277: -7.43459405171, 2278: -5.72744374996, 2279: -2.60260050415, 2280: 1.22750630057, 2281: 3.49397093942, 2282: -5.09234344557, 2283: 11.8171977538, 2284: 10.967877515, 2285: -9.58771675166, 2286: 4.45392902619, 2287: -0.254946723544, 2288: 5.80964741358, 2289: -2.62178015046, 2290: 0.446993497518, 2291: -0.174718851354, 2292: -3.73882133183, 2293: -0.293842804061, 2294: -2.1090160767, 2295: -0.245247741978, 2296: -7.36962619251, 2297: -0.245456166324, 2298: -0.246145208464, 2299: -13.694054108, 2300: -2.03054931315, 2301: -7.77333783682, 2302: -0.373360809041, 2303: -0.421708742063, 2304: -2.42664234831, 2305: -0.31469022064, 2306: -4.55618944309, 2307: 7.41410950126, 2308: -15.6935519698, 2309: -2.05654397164, 2310: -0.422412938952, 2311: 4.83360718872, 2312: -5.71186695402, 2313: 6.87988946013, 2314: -7.55071349399, 2315: -4.04329151681, 2316: 1.23067780529, 2317: -9.42934497833, 2318: -1.25641991218, 2319: -3.05172664625, 2320: -0.238051737876, 2321: 10.0814742994, 2322: -7.90251565484, 2323: -3.50440032085, 2324: -5.47742503011, 2325: 7.25518073637, 2326: 1.13607730397, 2327: -4.99782632846, 2328: 3.3157116004, 2329: 6.05254925101, 2330: 15.55959733, 2331: -8.82015904467, 2332: -2.58154087185, 2333: -1.43998954905, 2334: -1.38804611565, 2335: 7.13182549379, 2336: 31.5088743766, 2337: 17.0303974527, 2338: 12.3086684644, 2339: -6.50878026462, 2340: -5.85054105938, 2341: -12.6420459909, 2342: 4.97521313227, 2343: -27.0393841515, 2344: -13.1907578883, 2345: 10.3550053694, 2346: 18.7038741972, 2347: 28.8018313926, 2348: -4.23683436225, 2349: -2.00518519223, 2350: -1.43983362559, 2351: -16.3533316561, 2352: -2.31631860259, 2353: 0.996763541215, 2354: 6.27013073946, 2355: -0.276089615411, 2356: -1.96812486936, 2357: -2.93394479712, 2358: 3.20506862315, 2359: -2.60380501951, 2360: -9.72947680317, 2361: -10.5617988485, 2362: 6.90665201507, 2363: -6.68852157862, 2364: 1.92770743734, 2365: -0.438223501595, 2366: -12.1945331319, 2367: 8.72921570507, 2368: -24.6354092507, 2369: 18.9493268345, 2370: 18.4412105673, 2371: -1.77774737088, 2372: -10.7154298168, 2373: -1.4115970327, 2374: 3.06876865775, 2375: 1.37935648107, 2376: -6.76809971255, 2377: 0.602083894248, 2378: 6.7620296974, 2379: 26.3287673444, 2380: 3.55543059936, 2381: -1.95745185486, 2382: -9.90200790642, 2383: -27.5959615526, 2384: -6.84082091393, 2385: 8.75494618863, 2386: 0.286131731043, 2387: 18.6865557988, 2388: -12.9900573692, 2389: -26.5071205965, 2390: 0.814701206283, 2391: -2.4802804027, 2392: -8.01624441844, 2393: -8.35486579814, 2394: -8.20017104067, 2395: -0.356240478137, 2396: -3.73299264281, 2397: 10.9970672744, 2398: -9.70782854143, 2399: -4.71575248001, 2400: -0.955214217163, 2401: -0.76459432742, 2402: -2.31336942606, 2403: -2.19089836245, 2404: -2.31273531076, 2405: 7.72382794694, 2406: -1.63483105964, 2407: -1.62315616202, 2408: -1.54743446936, 2409: -1.95098671722, 2410: 7.386559891, 2411: -1.78370448472, 2412: -1.95058949494, 2413: 5.23102338363, 2414: 1.91619449734, 2415: -5.16516896518, 2416: -10.6304202783, 2417: 15.8004803261, 2418: 16.87559164, 2419: -7.63712456641, 2420: 30.8025274293, 2421: -0.541231782582, 2422: 7.60407898582, 2423: 22.4165850145, 2424: -0.40800462031, 2425: 27.161064909, 2426: -1.92740344829, 2427: 0.642296055061, 2428: 13.6516901824, 2429: -3.63406476572, 2430: 0.966226237873, 2431: -1.61535710177, 2432: -30.3715001685, 2433: -1.95292126343, 2434: 2.31007599367, 2435: -2.22559802672, 2436: -13.0907573356, 2437: 3.63406485941, 2438: 1.14912550823, 2439: 13.9924785304, 2440: 31.526575842, 2441: 51.4403782459, 2442: 14.5764600412, 2443: 19.5560037839, 2444: 0.00780486909031, 2445: -19.4510022294, 2446: 1.37893394244, 2447: 2.41080663236, 2448: -13.8284105931, 2449: 6.73294838317, 2450: 14.9620569443, 2451: -5.508567054, 2452: 14.2634646899, 2453: -4.6351877529, 2454: 7.00347933474, 2455: -2.45016801608, 2456: -0.474848563296, 2457: -4.69389328062, 2458: -3.21551680425, 2459: -2.13278479559, 2460: 5.30713878567, 2461: -1.73412968224, 2462: -18.5936091685, 2463: 1.43010205793, 2464: -6.68847814793, 2465: -5.18142118305, 2466: -4.44795032167, 2467: 11.2867509019, 2468: -3.05387109108, 2469: 14.808015334, 2470: -12.6284738118, 2471: 3.8972135955, 2472: -0.771442986496, 2473: -1.35418872116, 2474: -14.7258246868, 2475: -1.42729605613, 2476: -11.8694126057, 2477: -2.27686690283, 2478: 1.77944459685, 2479: -0.6442479197, 2480: -1.98691895632, 2481: 3.32765386154, 2482: -20.6787403828, 2483: 8.77009632779, 2484: -6.94079037861, 2485: -0.639596746532, 2486: -6.52183320681, 2487: -16.6074849661, 2488: -1.58391257455, 2489: -0.129655101156, 2490: 4.74475100346, 2491: -1.54882553886, 2492: -4.86099060819, 2493: 0.573436946348, 2494: 3.72317788172, 2495: -11.920417462, 2496: -0.533498475038, 2497: 0.0713179377962, 2498: -3.95057597658, 2499: 0.732105867725, 2500: -9.1785181144, 2501: 0.668474563857, 2502: -3.46917690391, 2503: 10.2956742407, 2504: -6.4568522149, 2505: 11.9163317061, 2506: -5.94174608666, 2507: 3.21326011048, 2508: 8.94164991678, 2509: 3.09722452737, 2510: -3.48289312272, 2511: -0.377418209655, 2512: 3.3043803571, 2513: -1.36159995021, 2514: 0.0977564948791, 2515: 9.97451791762, 2516: -0.576270828963, 2517: 4.96089641607, 2518: -5.39662306821, 2519: -8.40773214018, 2520: -3.24675313896, 2521: -2.77447614238, 2522: -19.1744544907, 2523: -2.7793234969, 2524: -0.212189383875, 2525: 6.45237703696, 2526: 0.0367494612688, 2527: 5.95084341416, 2528: 6.86446948042, 2529: 3.75927062765, 2530: 3.0192797388, 2531: -7.91384429253, 2532: -0.594073393844, 2533: 13.6681437455, 2534: -0.728355034118, 2535: -0.523942367525, 2536: -6.10915927448, 2537: 0.117505862344, 2538: -0.731222998659, 2539: -3.84202799983, 2540: 0.380755445681, 2541: -0.00816196943481, 2542: -2.70360753736, 2543: 4.09061283723, 2544: 3.69900175242, 2545: 3.47784140101, 2546: -2.3886914489, 2547: 1.65209339857, 2548: 8.85691019939, 2549: 1.82650821896, 2550: -0.717042237529, 2551: -0.554443417564, 2552: 2.02856502602, 2553: 6.37641947699, 2554: 0.610413492457, 2555: 0.325566311723, 2556: -17.1456401671, 2557: -34.3896792752, 2558: 22.1552338014, 2559: 3.49468976395, 2560: -1.73926552114, 2561: -1.32087453512, 2562: -7.39641125222, 2563: 3.68730787973, 2564: 4.92422472761, 2565: -16.3467935758, 2566: -10.5970237216, 2567: -4.32893248694, 2568: 4.48974304983, 2569: -0.154993749718, 2570: -0.0213768474595, 2571: 7.07419727936, 2572: 1.09808794985, 2573: -0.0815947208134, 2574: -0.373423363836, 2575: 0.43736775987, 2576: -9.9161439799, 2577: -1.81417448916, 2578: 1.27798031512, 2579: 1.30214131228, 2580: 0.572587054227, 2581: -5.5804767268, 2582: 2.00560403594, 2583: 0.0435860064981, 2584: -6.95624995827, 2585: -1.74284934512, 2586: -7.44174655561, 2587: 0.513843660282, 2588: 7.90700562523, 2589: -0.351848465475, 2590: -0.0588299213553, 2591: 0.450726711639, 2592: -0.101207558982, 2593: -0.665372992351, 2594: -1.88041008225, 2595: -0.116743874927, 2596: -0.114383520589, 2597: 0.0310777164819, 2598: 0.00697457560362, 2599: 1.26578081367, 2600: 0.33926844969, 2601: 0.65289768483, 2602: 0.0126410327308, 2603: -0.568692479384, 2604: 2.49384007704, 2605: -1.48970394488, 2606: -11.353770037, 2607: 0.231131695032, 2608: -8.60825664469, 2609: 4.63344044792, 2610: 6.66282472226, 2611: 3.99914394815, 2612: 3.66017414773, 2613: -1.28831887667, 2614: 0.946780703041, 2615: -0.0436460512384, 2616: -12.6586054552, 2617: -0.0832364441013, 2618: -4.92668045267, 2619: -5.9449773534, 2620: -12.3005511207, 2621: -0.140593546985, 2622: 0.0483343718064, 2623: -0.0285738660495, 2624: -0.457240485161, 2625: 0.496361972484, 2626: -15.7790391253, 2627: -7.26386013113, 2628: 3.2894299062, 2629: 15.2113763352, 2630: 15.5239823657, 2631: -6.29360773266, 2632: 2.05750708834, 2633: -1.35650142488, 2634: 0.228827815175, 2635: -6.47686063882, 2636: 0.0491907225314, 2637: -19.3124024142, 2638: 0.0755077443051, 2639: 2.20875831102, 2640: 1.8659672361, 2641: 1.56344589937, 2642: -0.153404339414, 2643: 1.08538597235, 2644: -0.0487965009046, 2645: -21.0724817785, 2646: 0.0897723832725, 2647: -0.072976345656, 2648: -0.415900387613, 2649: -0.60465992339, 2650: 8.52598866451, 2651: -0.06082440862, 2652: 0.0174780187124, 2653: 0.910523808895, 2654: -0.0758292580222, 2655: -0.699048793247, 2656: 0.30727339311, 2657: 9.21654846597, 2658: 0.132887322262, 2659: -0.496621054991, 2660: -2.82628909536, 2661: -6.5136506419, 2662: -20.4541210424, 2663: -3.80269181737, 2664: -3.86205318449, 2665: -7.28422615033, 2666: 3.22999447075, 2667: -0.131783341316, 2668: -2.56872775969, 2669: -0.164051097025, 2670: 4.10869675974, 2671: -9.3710822135, 2672: -2.7386825699, 2673: 3.29418152036, 2674: -8.75039251925, 2675: -0.0548237833666, 2676: -6.01708975125, 2677: 0.165883781117, 2678: 0.614165170753, 2679: 7.1045515352, 2680: -4.79648977094, 2681: 1.68030695016, 2682: -0.81865654839, 2683: -0.422444026279, 2684: 4.81778282608, 2685: 7.55274048983, 2686: 8.5249728408, 2687: -1.35313530037, 2688: -2.20584802805, 2689: 4.27724445737, 2690: 3.78026565528, 2691: -8.63234263033, 2692: -18.8930845629, 2693: 0.0785920772132, 2694: 8.53063519315, 2695: 16.3826890215, 2696: 4.56144234569, 2697: -6.53780576176, 2698: 7.95486338821, 2699: 0.383999869173, 2700: 0.0203165243939, 2701: -0.451705957222, 2702: -0.527940835074, 2703: -1.63928724646, 2704: 2.50522723596, 2705: -1.12782991401, 2706: -4.51898824073, 2707: -2.12263088017, 2708: 1.1924013231, 2709: -0.0337102674769, 2710: -3.89532233064, 2711: 8.11287939824, 2712: -1.31820928034, 2713: 2.23073887751, 2714: -0.953794706675, 2715: -12.1435217136, 2716: -2.09776526542, 2717: -3.38813173512, 2718: 6.17796652413, 2719: 0.234816100455, 2720: -1.81609481463, 2721: -2.24213743574, 2722: -3.47712753682, 2723: 6.72146483844, 2724: -4.09047436058, 2725: 2.39776088926, 2726: 12.104696709, 2727: -11.8591867918, 2728: -3.91107859777, 2729: 8.39491512638, 2730: -0.685648621835, 2731: -2.52972787037, 2732: 3.80004686921, 2733: -0.130305147733, 2734: 2.70988022817, 2735: 5.23312356642, 2736: 7.4789267374, 2737: -4.50071395768, 2738: -19.9191095049, 2739: 9.19258010372, 2740: -1.98571888489, 2741: 1.25921263616, 2742: 1.62597244374, 2743: 1.21696749567, 2744: 0.646506187292, 2745: 14.9316228747, 2746: 0.353659019056, 2747: 1.79995584526, 2748: -6.67015841955, 2749: -0.807909420923, 2750: 10.3993828014, 2751: 1.1399201697, 2752: 1.05951399386, 2753: 0.925524628379, 2754: 13.9516741172, 2755: -0.707025858343, 2756: -0.843704007793, 2757: 2.82429597729, 2758: -0.578763018944, 2759: -8.66428970097, 2760: -0.539880881281, 2761: -0.587274172179, 2762: 3.54092623536, 2763: -4.21287605423, 2764: -1.74854917125, 2765: -0.251335472674, 2766: 3.36317417641, 2767: 12.9247344104, 2768: 3.19660819854, 2769: 2.55574362242, 2770: 1.01236497818, 2771: 19.4152604554, 2772: 29.8675827848, 2773: -2.41277506967, 2774: -0.985337324685, 2775: -3.66013245676, 2776: 0.903409480317, 2777: 5.5365741438, 2778: -0.224518951392, 2779: 27.6615953414, 2780: 2.05867647302, 2781: -5.95592114522, 2782: -0.233251087856, 2783: -2.02244893736, 2784: 6.04098516322, 2785: 3.41544861454, 2786: 2.04254984402, 2787: -6.16633843364, 2788: 3.64051825366, 2789: 4.18319618326, 2790: 0.332465768137, 2791: -2.36576752007, 2792: 26.6474726505, 2793: 0.204575406166, 2794: 13.2935822702, 2795: 1.88483944525, 2796: 10.3731712611, 2797: -7.11189517535, 2798: -6.63136232395, 2799: -1.42138457033, 2800: 13.2159484213, 2801: 5.4203588012, 2802: 13.2143017182, 2803: -18.3114054931, 2804: -0.444492509723, 2805: -14.3925116789, 2806: 4.30306034553, 2807: -19.6117427313, 2808: 3.30255110023, 2809: 9.22068996463, 2810: 0.247714784265, 2811: -5.31986662263, 2812: -14.4856726882, 2813: 3.02235446399, 2814: 10.4460706841, 2815: -2.07961727707, 2816: 4.78921148993, 2817: 7.52940015518, 2818: -26.624871994, 2819: -8.49927369253, 2820: -9.56679065376, 2821: 5.93111878642, 2822: -1.57383218687, 2823: -12.978732673, 2824: -1.05607444467, 2825: 6.65026372847, 2826: 2.01861073016, 2827: 9.90806956554, 2828: 7.79381792668, 2829: -0.240649188083, 2830: -7.26265231592, 2831: 2.75519165407, 2832: -1.56868763486, 2833: 6.26154251162, 2834: 4.05790404357, 2835: -1.90543802536, 2836: 7.04789603678, 2837: 3.30865718645, 2838: 5.97773644841, 2839: -8.64779885134, 2840: 0.272765020626, 2841: -20.4391798197, 2842: 0.0834618868707, 2843: -10.0010272816, 2844: -2.7682162444, 2845: -0.0939907872727, 2846: 11.3462007016, 2847: 3.16284448783, 2848: 0.0413306165179, 2849: -11.0864967008, 2850: -1.72417704263, 2851: 6.187278596, 2852: 6.80821720286, 2853: 0.52579002256, 2854: -7.84314825825, 2855: 8.89763985083, 2856: -1.75537335095, 2857: -26.3707846408, 2858: -0.0676067567615, 2859: 3.86736459928, 2860: -0.104237681771, 2861: 12.2909749123, 2862: 0.0385669811753, 2863: -1.80201131958, 2864: 1.41736977591, 2865: 0.174639902108, 2866: 5.75055610794, 2867: 0.501855919604, 2868: -0.026880969658, 2869: 0.331786693723, 2870: 0.00447841026658, 2871: -13.6254128292, 2872: -3.31730762099, 2873: -0.185887163274, 2874: -11.1964033601, 2875: -0.0315730037493, 2876: -0.163701026624, 2877: -10.4346627484, 2878: -4.0392606048, 2879: 1.44211523448, 2880: -16.7577783817, 2881: -0.171868388197, 2882: -13.5399511218, 2883: -1.14824805334, 2884: 0.0108042574891, 2885: 17.4053276796, 2886: -0.207797957467, 2887: 0.0266412199876, 2888: 1.30089611766, 2889: 2.27538963822, 2890: 0.0412819425849, 2891: -0.0930500418951, 2892: -0.00743874697177, 2893: -10.658948602, 2894: -0.1725450896, 2895: 0.654514790553, 2896: -0.0417947406412, 2897: -0.114814483878, 2898: -12.5138849093, 2899: -0.0620903713466, 2900: -0.0362661629644, 2901: 0.395577345668, 2902: -4.79597870786, 2903: 0.486565709131, 2904: 2.3397479868, 2905: -15.8275319057, 2906: -13.8104417775, 2907: -17.6997260543, 2908: 3.08819409584, 2909: -3.74814247654, 2910: 3.04788681268, 2911: -7.45719844013, 2912: 0.610022759734, 2913: 0.463814496355, 2914: 11.7612239544, 2915: -25.250897074, 2916: 5.05674187899, 2917: 1.0228789212, 2918: 0.228517270678, 2919: 0.365204564966, 2920: -2.30173288133, 2921: 6.5690990338, 2922: 0.256616643101, 2923: 0.0336932466243, 2924: -0.442137412798, 2925: 4.63330250263, 2926: 3.07366173445, 2927: -2.41851447628, 2928: -14.2894950765, 2929: -0.119838221346, 2930: -0.961426294508, 2931: 1.7871492562, 2932: -6.7881959794, 2933: -11.9886880867, 2934: -1.1760946617, 2935: 8.32125740932, 2936: -0.147043965867, 2937: 1.54316785651, 2938: 0.331268143618, 2939: 0.21844142013, 2940: -5.561828175, 2941: 0.129676740046, 2942: 0.020808231773, 2943: 0.408272864678, 2944: 0.207230845409, 2945: 0.119532501473, 2946: 0.242757192869, 2947: 0.27123516728, 2948: -0.24755178479, 2949: -17.0797072522, 2950: -0.923242209755, 2951: 0.209071375845, 2952: 0.193956783094, 2953: 0.357307545052, 2954: 0.872881587669, 2955: 2.17077375038, 2956: -0.463395624825, 2957: -7.08020656218, 2958: -9.55753598078, 2959: 1.98563996725, 2960: 3.80841680156, 2961: -1.44674575001, 2962: -0.602628757787, 2963: -1.398893174, 2964: 0.230036210974, 2965: -7.3952708392, 2966: 0.149204820946, 2967: 5.21305937051, 2968: -12.1157691504, 2969: 17.12510778, 2970: 1.6373485592, 2971: 0.106936197454, 2972: 0.331670361336, 2973: 0.0854836177402, 2974: 0.12467671345, 2975: 0.504679283738, 2976: -7.28893678445, 2977: -3.06631883639, 2978: -2.37440842753, 2979: -6.15226071528, 2980: 14.0160489786, 2981: -1.43463340616, 2982: 2.09957741385, 2983: -1.23564519069, 2984: 2.58852079753, 2985: 0.263302702092, 2986: -1.54584611541, 2987: 0.305251362687, 2988: -0.164873446838, 2989: 1.15359822748, 2990: 0.883147663279, 2991: 0.159184363716, 2992: -0.0501718955409, 2993: 0.329572461378, 2994: -5.4748691282, 2995: 0.224830411351, 2996: 0.209335783013, 2997: 1.54181034076, 2998: -0.173350339495, 2999: -0.519217287804, 3000: 0.103812068784, 3001: 0.102915405525, 3002: -0.290189608297, 3003: 0.218174864675, 3004: -0.234787612426, 3005: 4.73453563364, 3006: -0.830037990389, 3007: 0.0680483902268, 3008: 0.153887579621, 3009: -9.62216120264, 3010: 2.43786889093, 3011: -7.36052382045, 3012: 7.54587257041, 3013: 2.57991219462, 3014: -0.183428455429, 3015: 5.44906928548, 3016: -0.00294435908928, 3017: -0.883103667716, 3018: 0.0129091770973, 3019: 17.4152500526, 3020: -0.550042241332, 3021: 1.91244065504, 3022: -1.23980703697, 3023: -4.45592138148, 3024: 0.510349369078, 3025: -3.66714428731, 3026: -6.61966796261, 3027: -0.140980651476, 3028: -4.41812956207, 3029: 2.29457587204, 3030: 6.53353052331, 3031: -0.840569751207, 3032: -0.0791517753035, 3033: 6.1125559009, 3034: 0.66965315771, 3035: -10.816563576, 3036: -10.7448542014, 3037: 5.04629786371, 3038: 10.0120310039, 3039: -0.534470201382, 3040: -6.784785839, 3041: -1.41218043313, 3042: 6.27317592233, 3043: 1.8820214514, 3044: -9.23402534022, 3045: 3.84509994579, 3046: -2.65490089816, 3047: -0.132547902873, 3048: -3.34390966794, 3049: 2.51086364731, 3050: -0.151600021983, 3051: -0.0709048083842, 3052: 0.818041927951, 3053: -3.98447374246, 3054: -0.00119454809776, 3055: -0.533760906946, 3056: 0.227746375565, 3057: -0.168306514267, 3058: 1.3382195103, 3059: 16.9559082474, 3060: -6.8478643613, 3061: -7.40758180758, 3062: 0.815921365508, 3063: 1.58456687672, 3064: -10.9591317653, 3065: -13.5809313796, 3066: 13.9045961086, 3067: 28.4641273728, 3068: 24.5006694892, 3069: -2.09641777295, 3070: -2.70560566911, 3071: -1.39774066834, 3072: -1.08625900733, 3073: 2.04740334137, 3074: 12.7527302244, 3075: 3.5108245152, 3076: 12.1754373821, 3077: -9.4006264504, 3078: -2.73217773236, 3079: -0.00168689619434, 3080: 0.410702829897, 3081: -0.0239993565819, 3082: -0.0451834627685, 3083: 0.647101974964, 3084: -16.9878349733, 3085: 12.071475026, 3086: 12.2916705843, 3087: 9.61015143958, 3088: -4.0320649744, 3089: -10.8320906835, 3090: 0.392377774326, 3091: -0.780610654145, 3092: 11.3920757407, 3093: -1.31799500891, 3094: -2.42991041734, 3095: -0.510690294385, 3096: -5.67851550672, 3097: 3.05031654937, 3098: 0.0986427580518, 3099: -0.923535200246, 3100: -0.155472560451, 3101: 0.0165590130055, 3102: 0.695647495455, 3103: -3.53894358279, 3104: 0.0304888582439, 3105: -0.131760773447, 3106: 4.73152214108, 3107: -0.0780632440782, 3108: -17.7250404125, 3109: -0.137830530103, 3110: -0.189945631259, 3111: -12.2854035506, 3112: -0.165363863171, 3113: -0.015911990119, 3114: 8.66261164264, 3115: 0.069166871006, 3116: 6.5080917872, 3117: -9.61945890688, 3118: 1.27868955928, 3119: -5.46245357204, 3120: 18.0940625435, 3121: 2.99852912988, 3122: -9.33333631682, 3123: 5.16058956036, 3124: 7.9199504346, 3125: -0.190814121696, 3126: -0.883765267575, 3127: -2.6173894414, 3128: -18.7990217574, 3129: -2.20155225379, 3130: 13.6961783124, 3131: -0.201981477559, 3132: 8.35506162764, 3133: 4.33635229731, 3134: -1.84758928755, 3135: -17.7444961622, 3136: -9.9260172963, 3137: -7.35586328449, 3138: 5.96594559777, 3139: 7.7542489619, 3140: 2.08129568077, 3141: -10.0594532727, 3142: 0.137409555699, 3143: 1.9214686098, 3144: -6.21579887433, 3145: -18.279777627, 3146: -12.2446309607, 3147: -2.96331826204, 3148: -14.2753147898, 3149: 17.01215531, 3150: -11.3896807871, 3151: 13.0727636544, 3152: -15.7739445862, 3153: 12.3674100679, 3154: -13.5152505959, 3155: 7.20087878733, 3156: 10.5548422419, 3157: 5.77006712566, 3158: -45.0939944477, 3159: -26.4617908059, 3160: -9.32574241244, 3161: 6.00076217093, 3162: -1.13107265299, 3163: 2.30878371041, 3164: 18.8911731307, 3165: -5.55012885419, 3166: -4.24333180337, 3167: -39.4505578307, 3168: 30.4233902872, 3169: -11.1247367398, 3170: -18.333099554, 3171: -27.8032725135, 3172: 13.802439593, 3173: 14.1391470625, 3174: 2.88963017665, 3175: 4.59960853759, 3176: -23.6227600072, 3177: -13.0500860308, 3178: -9.45051783306, 3179: 0.205881501884, 3180: -5.37980653629, 3181: -0.800052331473, 3182: 9.36074073802, 3183: -0.0250997312844, 3184: -6.29926219664, 3185: 20.6373528086, 3186: 13.2597045585, 3187: 0.296419765286, 3188: 18.400432615, 3189: 0.51109639955, 3190: -25.3132689016, 3191: 10.6283950909, 3192: -4.97684751743, 3193: -8.0751018613, 3194: 0.17529008698, 3195: 2.12475264026, 3196: -4.67622842386, 3197: -0.988127157266, 3198: 15.9001835582, 3199: 0.10197140446, 3200: -6.15411278686, 3201: -52.6888279916, 3202: -66.8173978076, 3203: -17.3175002866, 3204: 17.3162275979, 3205: 15.2323264353, 3206: -3.99464819007, 3207: -7.16554715169, 3208: 17.9876961373, 3209: 0.464102569131, 3210: 15.0220903748, 3211: -3.7702275507, 3212: -0.518108524712, 3213: -4.88314269555, 3214: 0.18276487068, 3215: -26.1232199982, 3216: 9.43461971029, 3217: -1.79430946362, 3218: 12.2208856895, 3219: -10.232808328, 3220: -3.87800289102, 3221: 28.3280372452, 3222: -6.41103449436, 3223: -32.0657481916, 3224: 1.30652560318, 3225: -8.26883482631, 3226: -9.03276358874, 3227: -5.73572671628, 3228: 5.07763694594, 3229: -28.871928702, 3230: -0.523986294891, 3231: -4.06226390797, 3232: 1.79922820743, 3233: -0.00197868757834, 3234: 3.21606551596, 3235: 0.598214770904, 3236: 0.588901122441, 3237: 11.0260979302, 3238: -15.416131913, 3239: 1.03255081165, 3240: -8.24072808674, 3241: 2.05177902325, 3242: 6.40586188778, 3243: 1.06797970544, 3244: -13.3591769672, 3245: -2.00855735095, 3246: 4.18160480887, 3247: 3.25960771798, 3248: 0.287856961227, 3249: -2.36571374906, 3250: -0.65471500831, 3251: -31.4660538347, 3252: 7.78716399703, 3253: 14.6358477726, 3254: -1.20560200334, 3255: 50.251258364, 3256: 4.74280339037, 3257: -7.23206210222, 3258: 7.6160269083, 3259: -0.666117150081, 3260: 42.1219820757, 3261: -5.34269236771, 3262: -9.8237203618, 3263: 27.4940892398, 3264: 17.9806708537, 3265: 18.4940728304, 3266: 43.0578974056, 3267: 0.250440732723, 3268: 1.15690552234, 3269: 17.3760902146, 3270: -6.64012714104, 3271: 0.0116730667338, 3272: 4.54618653383, 3273: -4.21584002897, 3274: 14.0937183614, 3275: -1.31512523434, 3276: 2.62083191757, 3277: 10.9357643225, 3278: -12.1388330225, 3279: -1.42478116437, 3280: 1.50320881958, 3281: 4.31710832255, 3282: 6.78266252257, 3283: -3.11505953229, 3284: -3.96437683076, 3285: -0.240961539778, 3286: -34.209169658, 3287: 0.0654412004683, 3288: 0.0131949275001, 3289: 20.2257277358, 3290: 0.0731737491991, 3291: -2.4922620911, 3292: -3.27459874024, 3293: -0.532039322759, 3294: 0.0808544031652, 3295: 0.192308250568, 3296: 0.124744244253, 3297: -1.84344121493, 3298: 5.59431314562, 3299: -2.9196194146, 3300: 1.10025430391, 3301: -0.763792783058, 3302: 20.1766236975, 3303: 1.8504981744, 3304: -27.9177709019, 3305: -2.6431233636, 3306: -44.4130823736, 3307: 2.54360238921, 3308: 9.24901056267, 3309: 13.5717255072, 3310: -8.45534512837, 3311: -1.28309922598, 3312: -3.75978956627, 3313: 0.0711198663455, 3314: -20.1242978446, 3315: 0.205631968607, 3316: -2.96472102942, 3317: 34.0728006782, 3318: 15.9775834682, 3319: -0.79915372341, 3320: 0.106804185181, 3321: -0.441835755405, 3322: -0.0102251451657, 3323: 0.155786456125, 3324: 13.1827218895, 3325: 40.1340621189, 3326: 6.28995012039, 3327: -12.2443259495, 3328: 20.9411281781, 3329: 6.01958584038, 3330: 7.57002809787, 3331: -0.674405894452, 3332: 4.69255093259, 3333: 30.0544702098, 3334: -0.0486994859126, 3335: 23.0174946104, 3336: 0.91831109629, 3337: 2.19181653958, 3338: -1.0057826983, 3339: 5.48167674619, 3340: 0.150314419134, 3341: 0.0984288444188, 3342: 0.085432761469, 3343: 22.0366397026, 3344: 0.179701746108, 3345: 0.21101679265, 3346: -15.3984412523, 3347: 0.158637537007, 3348: -21.7588645123, 3349: -0.0321363966186, 3350: -0.0306331808599, 3351: 0.755555334341, 3352: 0.054234594359, 3353: -6.63321793654, 3354: 28.235773974, 3355: -15.5262106131, 3356: -0.0111319882974, 3357: -1.04453032709, 3358: 49.3280390183, 3359: 17.4265297704, 3360: 14.2503491176, 3361: 9.30756998796, 3362: 2.30312273074, 3363: 11.8697969936, 3364: -3.43611378457, 3365: -4.00450387193, 3366: 11.1450892257, 3367: -0.101586371579, 3368: -52.534554867, 3369: -2.79646243163, 3370: -0.764615248957, 3371: -1.80844655143, 3372: -10.1050702374, 3373: 0.942693491902, 3374: 11.866757813, 3375: -20.8094394497, 3376: -6.26563073727, 3377: -12.7866168189, 3378: 7.23936186499, 3379: -0.901549585522, 3380: 0.0190588997471, 3381: 4.71428549093, 3382: -8.66791195037, 3383: 3.9898837306, 3384: 15.6620313621, 3385: -31.8686716084, 3386: -12.7611729327, 3387: -18.4109468614, 3388: 6.57756523886, 3389: 63.954438645, 3390: -5.25620128424, 3391: 61.9175197492, 3392: -12.5665637775, 3393: -22.0753589991, 3394: 11.252298885, 3395: 25.1726034145, 3396: -6.01537600766, 3397: -0.482463147673, 3398: -18.1538946104, 3399: -1.49288357407, 3400: -2.5650426746, 3401: -2.75688922896, 3402: 20.9224304957, 3403: -0.284444409747, 3404: -0.366373867825, 3405: -4.31589234165, 3406: -1.8449352686, 3407: -20.8414958531, 3408: -6.14891761435, 3409: -8.46007531673, 3410: 0.989614773502, 3411: 19.3970127216, 3412: 1.82322315542, 3413: -9.62895504173, 3414: -11.5268613267, 3415: 13.1414182346, 3416: -31.2745826824, 3417: -12.7480400293, 3418: 18.6270744931, 3419: 19.3974830242, 3420: -1.98358121013, 3421: -39.5863781806, 3422: -1.07771226479, 3423: 6.84705310713, 3424: -15.4109733506, 3425: -21.511002374, 3426: 32.1455111547, 3427: -49.8552557852, 3428: 0.215245233121, 3429: -3.71451648561, 3430: 1.96529187915, 3431: -8.49874044395, 3432: 18.5706512666, 3433: 9.24587505502, 3434: -19.0435811934, 3435: -14.002306225, 3436: 26.0895466012, 3437: 11.7709300155, 3438: 16.2350191747, 3439: -27.8925799565, 3440: 12.8888567389, 3441: -11.4710622684, 3442: 13.488296502, 3443: -11.1709845916, 3444: -5.07646223431, 3445: 44.5728284542, 3446: 27.3481150212, 3447: 0.581400581951, 3448: -23.6003061388, 3449: 0.119124771731, 3450: 0.139970644227, 3451: -3.39365288203, 3452: -34.0457501747, 3453: -0.457852407784, 3454: -0.36489879805, 3455: 22.8799972081, 3456: 0.106686172175, 3457: -42.2578723181, 3458: 0.331848324431, 3459: 0.309450821234, 3460: 18.1075126995, 3461: -1.01376923826, 3462: -12.203660724, 3463: -0.767688358993, 3464: -17.544565019, 3465: -10.8342246867, 3466: -15.7501299668, 3467: 24.8999979438, 3468: -8.83519485351, 3469: 16.4509836489, 3470: 23.7251028967, 3471: -2.9789803406, 3472: 17.3273100979, 3473: -11.6161842803, 3474: -6.62448899894, 3475: 9.21470475303, 3476: -28.3230413745, 3477: -6.7985482719, 3478: -16.8628399866, 3479: 12.5796323293, 3480: 0.712045899016, 3481: -2.44891544152, 3482: -6.57949168412, 3483: 8.57603772914, 3484: -8.00987146871, 3485: -1.2908760215, 3486: 13.0252993056, 3487: -27.4939198724, 3488: -32.3926898488, 3489: -0.193907128268, 3490: -32.0219222597, 3491: 0.672301326976, 3492: -1.4156222237, 3493: -0.601175461335, 3494: 1.52635244578, 3495: 0.442169085467, 3496: -1.49559241785, 3497: 1.24558290339, 3498: -0.0374408111281, 3499: -1.97786920398, 3500: -3.71443711321, 3501: -0.699825974491}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: -0.0230745518867, 2: -1.93697987772, 3: 1.45434014196, 4: 1.44745812227, 5: -0.58652906263, 6: 1.68933639722, 7: 2.42064229776, 8: -2.00261135478, 9: -2.97686818368, 10: -1.77645323582, 11: 4.80276104279, 12: -0.825710638728, 13: -1.67011392831, 14: 0.524807398996, 15: 1.54206225114, 16: -1.54819822532, 17: -5.6354749363, 18: -2.65776027047, 19: 6.15035750782, 20: 5.52716506541, 21: -1.9843960862, 22: 4.73905321667, 23: 1.28483869743, 24: 2.35941176041, 25: -7.11699419561, 26: -1.0696593424, 27: -0.738355397586, 28: 4.56131303202, 29: -8.21516579659, 30: 3.16605674058, 31: -0.692674804157, 32: 5.9190907368, 33: 5.35876011739, 34: 1.17155858607, 35: 2.24961592403, 36: -1.65643921963, 37: 2.38079160537, 38: -2.50871649428, 39: 3.23794682848, 40: -1.03843994553, 41: -2.8862179486, 42: 0.889701919622, 43: -0.160687596125, 44: 0.157648944836, 45: 1.8206738104, 46: -0.643897356999, 47: 1.74142327993, 48: -0.716244597985, 49: 1.6999151348, 50: -0.854837353354, 51: 3.42003270298, 52: 0.6577236707, 53: 0.591646173709, 54: -6.14056164213, 55: 0.00963540849267, 56: -0.378246101357, 57: 1.97170261434, 58: -0.720329904515, 59: 6.39150359028, 60: 4.90739803179, 61: -4.06039175475, 62: 0.0372406659373, 63: 2.94505856405, 64: -0.731192437327, 65: -3.08946083413, 66: -1.36263941767, 67: -2.19726602517, 68: 0.351256392318, 69: 2.97275171116, 70: 0.556535511187, 71: 0.593743141218, 72: -4.12614530242, 73: 0.525535519198, 74: 0.569368419821, 75: -0.298292850088, 76: 0.821571568972, 77: -1.55244721642, 78: 0.381567484481, 79: -2.15688754333, 80: -4.34258525666, 81: 0.177016424289, 82: -0.843421739147, 83: 0.163426336282, 84: -1.09307354498, 85: -2.70238999668, 86: -6.19325735966, 87: -0.420124786266, 88: 6.99259816245, 89: 0.0518640126436, 90: -1.28402942266, 91: -0.538403921816, 92: 0.444370347821, 93: -4.51024304855, 94: 0.542331348367, 95: 0.548254544945, 96: 0.425404555249, 97: -0.656585513583, 98: 0.640423292993, 99: -0.173778745518, 100: 0.191990996761, 101: 8.3189367036, 102: 0.426041623685, 103: 3.20750998489, 104: -0.554836899197, 105: 0.820337202999, 106: -2.23470805092, 107: 0.500842327736, 108: 0.574286675915, 109: 0.00930179136408, 110: -0.7450008856, 111: -0.613440120152, 112: 2.64232639683, 113: 4.60060965813, 114: 3.1331971195, 115: -0.973677955486, 116: -0.751728771713, 117: 0.802414573196, 118: -1.61352663875, 119: -1.83567542188, 120: -0.737389698248, 121: -0.87263402164, 122: 1.65434876146, 123: 5.44484069191, 124: 2.37044303062, 125: -4.021074808, 126: 0.0647226765735, 127: -0.189495162477, 128: 1.47169352044, 129: -0.874995115834, 130: -0.0135600331096, 131: 0.919182367009, 132: -0.464672099078, 133: 2.29664463401, 134: -1.64106120715, 135: -0.922150358179, 136: 1.97538664254, 137: -1.26631064171, 138: 0.182875810696, 139: 0.203929213478, 140: 3.70417045207, 141: -1.83783683866, 142: 1.32520617459, 143: 1.82897267346, 144: -0.461352451847, 145: -0.801244405003, 146: 0.0717022895092, 147: 0.058119625707, 148: -2.41980364164, 149: 0.0483481027348, 150: -0.242574517108, 151: -1.59150212836, 152: 0.0460523964043, 153: -0.0202001175704, 154: -0.0131614570859, 155: -0.00305516503124, 156: 0.0699466582041, 157: -1.34995933642, 158: 0.704749694323, 159: 0.21447304824, 160: -0.0541328711418, 161: -0.0222356159618, 162: -0.271614014827, 163: -4.94394555572, 164: -0.548701153421, 165: -6.19482935771, 166: 1.37268606743, 167: -1.8956551818, 168: -0.480662899707, 169: 1.98610159981, 170: 3.10365968842, 171: 2.20706188145, 172: -0.0028432976425, 173: -1.19311109471, 174: 0.00975574152015, 175: -3.1133949613, 176: 2.11844182581, 177: -3.33549917587, 178: 0.767480883453, 179: 0.0896039182483, 180: 0.0106783922435, 181: -0.0491927303018, 182: -0.122445086781, 183: -0.760004124675, 184: -6.71255419575, 185: -0.842609583244, 186: 2.16386993213, 187: -1.53750824277, 188: 2.18767553236, 189: -1.08580927617, 190: -0.570432788382, 191: -2.69105082328, 192: -4.01361985096, 193: -0.126122076844, 194: 2.75719777079, 195: -1.15043274146, 196: 0.554761061402, 197: -0.898758159286, 198: 0.538225644617, 199: 0.0500095712473, 200: 0.507514911854, 201: 0.0412048722839, 202: 2.16396721632, 203: 0.00990973745822, 204: -0.00625440788926, 205: -2.91320620184, 206: -0.44324056732, 207: -3.20206744606, 208: 0.13710335596, 209: 0.00646172036674, 210: -0.13171231457, 211: -0.101807980993, 212: 2.05165444981, 213: -2.203401728, 214: 2.14971766582, 215: 0.00129667192657, 216: -0.394609593027, 217: 1.45118699325, 218: 0.0483722105572, 219: 3.86787255509, 220: -2.72541567052, 221: 0.924777460162, 222: -1.34223455767, 223: -1.05645251517, 224: -0.4454481034, 225: 0.710359799674, 226: 0.0934097868284, 227: -7.38831713617, 228: 0.507353317377, 229: 0.372066599025, 230: 0.855572586436, 231: 0.475789875981, 232: -0.0140860737304, 233: 1.84044947709, 234: -4.21276360984, 235: -1.05170989795, 236: -1.52148387936, 237: 3.06700912256, 238: -2.43924111516, 239: 0.712256454552, 240: 0.81908864763, 241: -1.25645213889, 242: -4.27308659035, 243: 2.40336052104, 244: -0.0871875626658, 245: 0.231105400516, 246: 7.87653824151, 247: 0.744509713692, 248: -0.690423477957, 249: 7.56117433787, 250: -1.77378772283, 251: 0.409527557628, 252: -0.268917808661, 253: -3.6902646428, 254: 1.02384905943, 255: 1.79713783351, 256: 0.519395450064, 257: -0.720168404396, 258: 0.58049216114, 259: -0.217708221402, 260: 3.40610950593, 261: 0.257313066638, 262: 0.0326732708791, 263: -1.46513354992, 264: -0.366602849079, 265: 0.192807972684, 266: 2.26818871056, 267: -3.70423501617, 268: -5.95065258631, 269: -3.21521998898, 270: 1.62135203118, 271: -0.443159259706, 272: -1.76743181028, 273: 2.83577815952, 274: -4.64522034213, 275: -3.62758893844, 276: 1.28648839783, 277: -1.12336065045, 278: -1.08902904304, 279: 3.64836113626, 280: 2.46929274268, 281: -0.40799769584, 282: 4.31790264947, 283: 0.241200272195, 284: 2.69928617804, 285: -1.20142777608, 286: 3.16161416996, 287: -0.336552973684, 288: 0.587420367346, 289: 0.271678280046, 290: 2.17767661123, 291: -0.286301999285, 292: -2.44910037516, 293: 0.750650879953, 294: -1.44546690369, 295: 3.95715101395, 296: 3.47891386924, 297: 2.42517503002, 298: 0.366428935553, 299: 1.1151332172, 300: -0.932362466922, 301: 0.565775344087, 302: -0.836828477186, 303: 2.64585746732, 304: -0.141856241603, 305: -1.79802803645, 306: 0.391513240991, 307: 0.344101284217, 308: 0.620563649218, 309: 0.625703903078, 310: 0.0492546223791, 311: -0.287284878279, 312: 0.421497306835, 313: -0.183633890732, 314: 5.30092486996, 315: -0.662536910313, 316: -2.50867616205, 317: -0.337955859052, 318: 0.739784087407, 319: 7.11897444329, 320: 1.61375862097, 321: 3.51591266667, 322: -2.3113846966, 323: -1.51906867474, 324: -7.11182781569, 325: -1.37275764569, 326: -1.88134585766, 327: -3.50471835572, 328: -10.8743989868, 329: 0.292576438238, 330: 1.41750626209, 331: 5.11088054292, 332: -1.66933237832, 333: -1.89752746612, 334: 0.111948266598, 335: 0.763637640447, 336: 2.29419258544, 337: -2.16948482937, 338: 11.187613672, 339: 0.50783292573, 340: 2.95968804414, 341: -1.20507334314, 342: 0.284735585535, 343: -5.61799869059, 344: -0.36505555333, 345: -2.63026656204, 346: -13.5987055332, 347: -0.496933369173, 348: -0.48399235755, 349: -3.063109452, 350: 0.0899485411684, 351: -4.77722996582, 352: -8.61549602744, 353: 11.4089232848, 354: -5.50634956201, 355: -6.19237524491, 356: 2.04028898747, 357: 13.4168306966, 358: 6.18774405135, 359: -3.39751011353, 360: -6.37145469126, 361: 0.643006505482, 362: -7.21469915502, 363: -13.5042074544, 364: 2.64601608762, 365: -4.69441227158, 366: -8.399953559, 367: 21.4589867019, 368: 2.43811968093, 369: 0.237148276569, 370: 2.23742763196, 371: -2.01544052601, 372: 15.3929542794, 373: 8.95375145278, 374: 5.10722908781, 375: -0.209331071076, 376: -3.3338849992, 377: -3.01808975747, 378: 12.4265799312, 379: -1.81453670276, 380: -3.88579931249, 381: -4.96873704138, 382: -4.41026517195, 383: -2.47630886696, 384: -2.409997273, 385: 4.8900932983, 386: -1.66336594105, 387: 3.66735244948, 388: 0.893902930419, 389: 3.53731741933, 390: 3.71160249996, 391: -0.464410537301, 392: -8.28608720353, 393: -14.2706940747, 394: 2.96916759791, 395: -1.19133946672, 396: 20.4780484565, 397: 7.97080835961, 398: -1.24201736209, 399: 7.99396244903, 400: -12.965051143, 401: 7.8097925166, 402: -0.286290829452, 403: -8.83404046921, 404: 5.68634087278, 405: -0.369936534814, 406: -12.4003295217, 407: 0.486375956829, 408: -27.7485796866, 409: -16.1814052962, 410: 6.43976450526, 411: 3.12032534115, 412: -4.07037195276, 413: 3.85885596395, 414: 10.5415102627, 415: 9.23700249728, 416: 2.35198649467, 417: -0.437697438781, 418: 2.52360806898, 419: -0.424321225238, 420: 8.84254220697, 421: 10.5113049285, 422: -0.484362603794, 423: -1.68415570875, 424: 3.18851798886, 425: -0.435795271518, 426: 2.97711106674, 427: 1.1830966877, 428: -16.6505999194, 429: -15.6641930323, 430: 0.043060191179, 431: -9.7292307994, 432: -0.253285256164, 433: -6.51752400033, 434: 4.7616442443, 435: -29.3078071291, 436: 5.41032739953, 437: -13.1546162039, 438: -0.434624354692, 439: 10.464039743, 440: 1.47256694194, 441: -0.317521892395, 442: 8.64563302852, 443: 1.31976928856, 444: -0.386435951313, 445: 1.11833942599, 446: 2.43145364255, 447: -0.237429068279, 448: -0.449043950702, 449: -14.1321250324, 450: 26.6308398243, 451: 0.603203257142, 452: -0.763116246587, 453: 1.37069529247, 454: -0.59061456555, 455: 14.1905676832, 456: -0.226620921777, 457: -0.387736616897, 458: -0.386590257051, 459: 0.108258859046, 460: 2.48019160329, 461: -12.8243410718, 462: 48.8059246308, 463: 0.839021202139, 464: 3.57316323801, 465: 0.966457595916, 466: -0.630675969644, 467: 14.0018120876, 468: -3.1955260022, 469: -0.0512442238749, 470: -0.796787396075, 471: -2.53654820311, 472: 0.102219779416, 473: -3.35249914685, 474: 5.39496704824, 475: 0.0789298962591, 476: 2.65593154879, 477: 9.88320534069, 478: 8.95648859232, 479: 0.0449742198719, 480: -0.373468098728, 481: -1.35289007726, 482: -28.9967104253, 483: 2.07583570312, 484: -2.73683800487, 485: -10.184642715, 486: 2.77324389041, 487: -4.61705330597, 488: -0.587442907042, 489: -16.4860635703, 490: -20.803320182, 491: 12.1159728971, 492: -11.2626938782, 493: -0.117986568878, 494: 10.2702444564, 495: 0.136560565962, 496: 0.00817188098229, 497: 6.17714286147, 498: 0.0324111186389, 499: -0.47388760922, 500: -19.9688742557, 501: -0.534641722158, 502: -0.142035017127, 503: 0.0534316545516, 504: 0.0895267843506, 505: -0.477195096114, 506: -3.31096729706, 507: 3.39138787978, 508: -0.0313095163088, 509: -0.215066563659, 510: -13.736277792, 511: 2.15839444809, 512: -1.40049413647, 513: 8.56407956422, 514: 0.78355755568, 515: -14.1035696277, 516: 17.359496528, 517: -1.99420639634, 518: 2.20487964241, 519: -1.45190751035, 520: -4.56786820935, 521: 0.0396020433367, 522: -9.63880081862, 523: -0.0481883442009, 524: 3.47592883302, 525: 2.62080319006, 526: -15.862694276, 527: 2.11834693986, 528: -0.037884397244, 529: -0.0746422375975, 530: -0.0163355880026, 531: 0.00721195140837, 532: 2.3951738518, 533: 4.96210777512, 534: 4.32868616862, 535: 15.3907703891, 536: -0.171794393768, 537: -11.0705979436, 538: 7.54883816467, 539: -1.7807060969, 540: 1.35269550939, 541: 0.47041974552, 542: 0.110989518577, 543: -13.2929219568, 544: 0.574793440699, 545: 0.219443114619, 546: -2.6299340778, 547: 0.246424185565, 548: 0.0878427872598, 549: -0.336352955056, 550: -0.00553756012063, 551: 10.7503408772, 552: -0.0455355930791, 553: 0.0499582525992, 554: 22.2019818826, 555: -0.246881768273, 556: -8.29033349462, 557: 0.0970197652055, 558: 0.0975469158687, 559: -1.25111435056, 560: -0.0859312677462, 561: -0.296941423154, 562: 6.5866944024, 563: 7.05748401651, 564: -0.0838113980458, 565: 0.591869004525, 566: -5.18164977311, 567: 0.227847604901, 568: 12.9987456353, 569: -0.937444228954, 570: -3.64496019259, 571: -14.9612657966, 572: -4.68513941278, 573: 0.615501343, 574: 1.01791046744, 575: 0.48873863643, 576: 10.6644920855, 577: -0.358726476558, 578: 0.735802862123, 579: 3.84994430222, 580: -7.41779039376, 581: -0.610860309431, 582: -8.31232953297, 583: 7.2046779553, 584: -1.0992621031, 585: 14.0147908184, 586: -8.59229459377, 587: -9.1415960411, 588: 0.247012149173, 589: -0.170529067051, 590: -8.88431660656, 591: 27.2770717179, 592: 3.80046457683, 593: 12.9258024684, 594: -4.4613395347, 595: -5.09066572205, 596: -0.152503793, 597: -21.381551402, 598: -15.4026987418, 599: 26.2133977802, 600: -19.4266950573, 601: 15.21495998, 602: 7.31678233967, 603: 13.3694893825, 604: -0.299729329932, 605: 2.51504945842, 606: 1.8295643202, 607: 1.66960974539, 608: -0.292513470851, 609: 10.260745425, 610: 5.69626841572, 611: 0.903781804135, 612: 1.16222541384, 613: 1.33402952924, 614: -0.38125873368, 615: 29.6216104855, 616: -3.35938027134, 617: 5.43037398548, 618: 9.07562416829, 619: 1.49403605127, 620: 1.37575227953, 621: 5.57460734544, 622: 3.58022034374, 623: -1.40302815191, 624: -12.0594188405, 625: 8.22086212888, 626: 4.03290606568, 627: -5.8798347331, 628: -15.374385611, 629: 4.61641414131, 630: -0.471710321812, 631: 3.94834001181, 632: 0.173218868096, 633: 2.65282681879, 634: 9.95678474926, 635: 12.666238243, 636: -0.916845380294, 637: 1.71953185363, 638: 1.04765184224, 639: -0.49897097955, 640: 1.76334393851, 641: 15.3946864882, 642: 4.839334862, 643: 33.978240702, 644: 1.53085505477, 645: -20.4962982716, 646: -4.15041545147, 647: -2.35524305931, 648: 2.47980269041, 649: -3.17694869954, 650: -0.115534618746, 651: 8.92863319531, 652: 0.862369150389, 653: 1.84210163924, 654: -9.40234700294, 655: -0.413707149981, 656: 7.85563801434, 657: -0.447327727913, 658: -0.335073856057, 659: -0.319384175798, 660: 18.8667007964, 661: -0.284749394612, 662: -0.221624487873, 663: -16.4805532623, 664: -0.345740282971, 665: 1.79649629861, 666: -1.69089138578, 667: -0.460240800422, 668: -0.777207048121, 669: 0.136574959194, 670: -0.249285056392, 671: -6.61678022614, 672: 3.85350998376, 673: -14.8291891678, 674: -4.38986849195, 675: 34.7517260549, 676: -0.451580989219, 677: -36.4991994785, 678: 0.59747496773, 679: 1.07890189885, 680: -17.2116703794, 681: -3.23893620537, 682: 0.824900559021, 683: 3.00927332039, 684: -13.4318393606, 685: -5.54789242466, 686: -0.875822302999, 687: -14.9226702044, 688: -0.317194694881, 689: -16.5391847323, 690: 0.510167116668, 691: 7.56593370791, 692: 10.6087057193, 693: 0.33365197363, 694: 13.5356338753, 695: 12.6035555655, 696: -9.01229772411, 697: 6.76672677885, 698: -0.357749784174, 699: -0.223424878817, 700: 0.194155508948, 701: 0.545041603485, 702: -3.33924554831, 703: 0.0829345828749, 704: -1.98413602285, 705: -3.00636658486, 706: 2.33798836462, 707: 4.98343881854, 708: 3.40597360398, 709: -3.14966317638, 710: 1.65614336877, 711: 1.98623280371, 712: 0.890567339794, 713: 0.482337409877, 714: 2.36874159711, 715: 8.44177321211, 716: 2.01708259567, 717: -12.7388411924, 718: -5.90990499804, 719: 3.46070390334, 720: -4.35385926994, 721: -4.09998297784, 722: -2.48780917987, 723: 9.42157772597, 724: 1.33648333884, 725: -0.353416287813, 726: -5.34954242895, 727: 6.03004565255, 728: -4.39583316943, 729: 1.86108642429, 730: -6.50090186229, 731: -5.44102509325, 732: -0.880118757618, 733: -2.23940409867, 734: 2.00111147239, 735: -2.72447514252, 736: 1.49462550007, 737: -4.74745030082, 738: 0.627056395212, 739: 4.02030284101, 740: 0.911118649747, 741: 3.89142745968, 742: 0.354473335957, 743: -1.1050959186, 744: 0.728787365466, 745: -7.41583194971, 746: -1.06569708917, 747: -2.12676353233, 748: 2.94589999305, 749: -1.8044372513, 750: -3.13362164196, 751: -0.766258753077, 752: 4.31908629113, 753: -2.34344365364, 754: -0.233222253432, 755: 1.41395939409, 756: -0.563284136637, 757: 6.88197055386, 758: -3.753849619, 759: 0.273502934179, 760: -1.24227414807, 761: -3.58864892943, 762: -0.904343931062, 763: 2.75393209499, 764: -0.0233081567449, 765: 0.617220640231, 766: -2.14959440092, 767: -3.74250345275, 768: -0.535541622919, 769: -0.810202333108, 770: 3.58995056692, 771: -0.697505192256, 772: -3.62272654354, 773: 2.59818063143, 774: -0.0319493740902, 775: -2.13429781942, 776: -0.294340763796, 777: 9.28671676122, 778: -0.509699595809, 779: 0.374092915544, 780: 0.825575051075, 781: -0.765066093327, 782: -1.57281635203, 783: 3.33783846739, 784: -0.261751897033, 785: 6.39820692576, 786: -6.04737295983, 787: 0.133161359692, 788: 4.75864761428, 789: 4.86932170063, 790: -0.674026396924, 791: 0.495312995042, 792: -0.706649561092, 793: -0.929658023401, 794: -0.818460173815, 795: 0.740833368368, 796: -0.763344181762, 797: -1.02609107164, 798: -0.678260715247, 799: -9.22243838583, 800: 0.86478675423, 801: 4.41967863759, 802: 0.096574551048, 803: -0.228166767693, 804: 0.496700880021, 805: -0.880719011539, 806: -0.444726988099, 807: 0.278628595968, 808: -0.125903385955, 809: -1.56161071865, 810: -0.777484755002, 811: -3.44294362482, 812: -3.40490425179, 813: -2.36570157655, 814: 1.56435745919, 815: -2.08681765011, 816: 1.16380969621, 817: 0.566342573629, 818: 2.04503037542, 819: 0.975037656577, 820: 1.68255562108, 821: -10.6754219431, 822: -2.80322906033, 823: -7.51077353714, 824: -0.0567617875231, 825: -0.389117699849, 826: 1.28975823429, 827: 5.85288852122, 828: -0.227774889269, 829: 0.916853591749, 830: -0.563969646526, 831: -2.11994009149, 832: -1.01889631336, 833: -1.74776118098, 834: -1.82942210461, 835: -0.724256644086, 836: 1.97877025251, 837: 0.0848702374947, 838: 0.640058297645, 839: 1.92465173466, 840: -1.34603704671, 841: -1.18956196062, 842: -0.29174285047, 843: -0.685400252143, 844: -0.239983170432, 845: -0.0824422530541, 846: -0.440407629431, 847: -0.124481449353, 848: -0.895993108225, 849: -1.77894370475, 850: -0.107814243845, 851: -0.188935988332, 852: -0.273564614698, 853: 0.00831703666984, 854: -1.20309758131, 855: -4.90550838367, 856: 1.04446243818, 857: -0.173022351952, 858: 0.0162732020699, 859: -2.79748669594, 860: 2.46934396925, 861: -3.51806906565, 862: -1.13950518151, 863: 3.58317694227, 864: 3.81079826613, 865: -0.84221220056, 866: 0.581757806354, 867: -0.829950390575, 868: -0.100792526998, 869: 2.57574936517, 870: 0.0239893374995, 871: 2.02508448893, 872: -0.25105155939, 873: 3.74469747031, 874: -0.491728705996, 875: 4.79596017266, 876: 0.740648572168, 877: -0.110939858236, 878: -0.100647646322, 879: -0.243437702296, 880: -0.187266351477, 881: 0.72397735538, 882: 3.60108873777, 883: 1.50597568837, 884: -0.4669388913, 885: 2.07571068308, 886: 0.943557498733, 887: -0.259279085577, 888: -0.364013634582, 889: 2.55385985471, 890: 0.537152219782, 891: -0.232689261155, 892: -0.84652759321, 893: -2.22462135819, 894: -0.138422786774, 895: -1.69762764809, 896: -1.02409465664, 897: -0.127398751803, 898: -0.62949100003, 899: -0.15848114263, 900: 2.02492813519, 901: -0.0435394399541, 902: -0.0625116795676, 903: 2.69319238495, 904: -0.85135623565, 905: 1.01798662808, 906: -0.201690958822, 907: -0.187654449703, 908: -0.148507355755, 909: -0.26942141542, 910: -0.294463258028, 911: 3.14010744936, 912: -4.17185714889, 913: -0.871726140487, 914: -0.352634370963, 915: -8.28494951151, 916: 0.24969050553, 917: 6.86747923208, 918: 4.29079393268, 919: -0.327058330536, 920: 0.659594354794, 921: -2.71414595235, 922: -0.25472745085, 923: -0.849253688902, 924: 0.418975273724, 925: 6.71138298779, 926: 0.250295231713, 927: 0.385538459047, 928: 0.35959254765, 929: 0.35220255798, 930: -0.180082954223, 931: -3.48034150669, 932: 7.81063966607, 933: 0.173868937272, 934: 5.42768948784, 935: -4.81648784245, 936: 4.84287775221, 937: 0.908203457093, 938: 0.899658570493, 939: 4.17141546382, 940: 2.36698581236, 941: -4.15302845435, 942: -4.41283534119, 943: -0.206016610598, 944: -8.91933885295, 945: -2.84029877397, 946: 4.46211289214, 947: -5.17846859335, 948: 0.635201709534, 949: -0.218307307426, 950: -0.0857133666473, 951: 7.59005220762, 952: 0.323251551105, 953: 1.2788062409, 954: -0.399194115854, 955: 2.24065516594, 956: -0.342390512433, 957: -0.954889724085, 958: 2.02132548009, 959: -2.13351423318, 960: -1.02387080967, 961: 0.876196778304, 962: 0.0179309896379, 963: -1.18816428777, 964: 1.91635482035, 965: 5.11476074378, 966: 5.7597035106, 967: 3.86564869116, 968: 3.39035390005, 969: 0.082768629417, 970: 4.1494176619, 971: -1.58229998061, 972: -1.33088343131, 973: 4.79586132714, 974: -3.78984701797, 975: -0.944125130936, 976: 3.63226754162, 977: -2.04802336231, 978: -0.856091242227, 979: -2.15097160967, 980: -6.32198733453, 981: 2.4555844292, 982: -7.53646676967, 983: 2.55683119046, 984: -2.32906972473, 985: -0.630915801757, 986: -1.76551436303, 987: -0.915988340688, 988: 0.546667904839, 989: 0.489673305731, 990: 6.73001646815, 991: 1.30931120593, 992: -0.435339688863, 993: -5.95453653347, 994: 0.213220568655, 995: -2.92791455796, 996: -1.69960645326, 997: 4.21287533063, 998: 2.60455651556, 999: -0.933815320918, 1000: -0.359022010119, 1001: -0.303209087389, 1002: 0.0900656106313, 1003: 2.5820980031, 1004: -0.930444322355, 1005: -2.54857218345, 1006: -0.805480070439, 1007: -0.701656529438, 1008: -0.835113895639, 1009: -3.485758542, 1010: -0.630216615314, 1011: 0.818294893287, 1012: -5.61352697739, 1013: -1.06602729409, 1014: -2.01952090149, 1015: 0.333183435602, 1016: -0.657312846095, 1017: -7.12717938336, 1018: 1.01487052138, 1019: 0.072080860512, 1020: 6.91425220362, 1021: 7.98614586017, 1022: 9.15575624054, 1023: 2.2345868252, 1024: 5.02387913119, 1025: 4.24049203868, 1026: 5.81876179255, 1027: -0.654799655874, 1028: -2.64649295034, 1029: -5.48112180645, 1030: -1.09221972755, 1031: 0.114963598976, 1032: -0.09351172778, 1033: -3.32760646253, 1034: 1.37433908306, 1035: -4.36980678097, 1036: -9.1598533531, 1037: -0.772110962119, 1038: 2.00408925139, 1039: 2.72508746194, 1040: 3.66319684209, 1041: 6.28663116054, 1042: 3.35263010331, 1043: 3.4727280298, 1044: 12.2913368333, 1045: 3.24236540237, 1046: 3.34396864274, 1047: 7.83815630461, 1048: -0.143578702952, 1049: 5.55555948388, 1050: -3.73476966416, 1051: 7.79080804127, 1052: -5.70378801163, 1053: 6.37453495866, 1054: 7.54500778586, 1055: 1.69275183984, 1056: -1.5242351565, 1057: -37.5542313594, 1058: -60.1715018602, 1059: 35.8859519477, 1060: 16.7288548857, 1061: -7.21472649683, 1062: -37.184704052, 1063: -4.44967185948, 1064: 38.5921207122, 1065: 11.7790199145, 1066: 38.5805477308, 1067: 13.3574186876, 1068: -31.2232725168, 1069: -53.9347749359, 1070: 43.2102622184, 1071: -35.0061834128, 1072: -1.4119099835, 1073: 30.8440915829, 1074: 3.09053939866, 1075: 30.5969088608, 1076: 32.3978791398, 1077: -26.8968101459, 1078: -3.08531983709, 1079: -10.3440942188, 1080: -19.8376179049, 1081: 0.804427972225, 1082: -1.11884984329, 1083: -16.0017222046, 1084: -1.146114968, 1085: 14.112896461, 1086: -2.36649688827, 1087: 16.1332423666, 1088: -2.19781401068, 1089: -1.77498575369, 1090: 18.5095902533, 1091: -3.44745634903, 1092: -40.6473692061, 1093: 1.91570535504, 1094: -34.7456094807, 1095: 0.0435846757553, 1096: 3.48211471953, 1097: -24.9110057945, 1098: -12.2891011001, 1099: 23.3489317962, 1100: -0.59465113611, 1101: -3.0658517353, 1102: 20.9532198182, 1103: -1.50007769428, 1104: 20.5810908182, 1105: 8.9514887294, 1106: 6.03504709634, 1107: 30.646510875, 1108: -7.39590857009, 1109: -9.56772535349, 1110: 5.95658760519, 1111: -3.9998512149, 1112: 25.5646220346, 1113: -3.28832443523, 1114: -8.81577130669, 1115: -9.21128608089, 1116: 5.44444922306, 1117: -2.40260889421, 1118: -2.2130239275, 1119: -16.2948592529, 1120: -3.12423652386, 1121: 57.6456347144, 1122: 7.79820382472, 1123: 0.511351352725, 1124: 2.67654026325, 1125: -1.32183064584, 1126: -23.7651096633, 1127: -1.77625085061, 1128: -3.92324221186, 1129: -17.0556395334, 1130: -8.27678430514, 1131: -2.98450575621, 1132: 53.7021275584, 1133: 15.7928886166, 1134: -12.2470148113, 1135: -45.4871742373, 1136: -2.10571157578, 1137: 25.823143187, 1138: -13.9613016425, 1139: -1.55156336106, 1140: 26.2633711479, 1141: -2.98092484642, 1142: -2.90088415032, 1143: -2.37104420696, 1144: -4.57201923949, 1145: -3.75153369872, 1146: 5.75832884037, 1147: -5.11745943768, 1148: -14.3220570368, 1149: -2.73333359953, 1150: 5.42774170242, 1151: -1.26068162719, 1152: -5.37269466063, 1153: -2.77362502393, 1154: -1.97398200596, 1155: -1.56116170884, 1156: -2.68434540333, 1157: 23.4052113802, 1158: -2.99390324919, 1159: -37.9707990103, 1160: -46.4189202368, 1161: 0.498584605454, 1162: -25.8075889173, 1163: -7.47592358101, 1164: -26.9592516837, 1165: 5.4643564128, 1166: 62.6923982149, 1167: -12.4659105426, 1168: -30.373165906, 1169: -19.4697111555, 1170: -36.8354601259, 1171: 4.63180651775, 1172: -0.352477578725, 1173: -0.513415591533, 1174: 0.660420535688, 1175: 12.6498655794, 1176: -14.0661051929, 1177: 0.111541063254, 1178: -2.6581812159, 1179: 0.780875025263, 1180: -7.41165690479, 1181: 0.0370527994499, 1182: -0.376268300588, 1183: 3.41187864484, 1184: -1.78011728269, 1185: -16.464441586, 1186: 8.21919534691, 1187: -36.9231248335, 1188: 3.60475255058, 1189: -5.01748843271, 1190: -1.18308846513, 1191: -2.00691435323, 1192: 4.10760259123, 1193: 1.28639667773, 1194: -0.516328975101, 1195: -2.50260455353, 1196: -0.320320139145, 1197: -1.7471948582, 1198: -11.7369634756, 1199: 0.536356517258, 1200: -0.18111689728, 1201: -0.319358163389, 1202: 0.0647727276478, 1203: -3.64755731182, 1204: -8.30779556729, 1205: -4.62718638164, 1206: -1.72790333391, 1207: 0.0890213790756, 1208: 7.35439089387, 1209: -11.9847957883, 1210: 61.3238118445, 1211: -36.8691851188, 1212: -49.0555195061, 1213: -51.9046847176, 1214: -8.61781179352, 1215: 11.5128001265, 1216: -8.03560203195, 1217: 1.97739401202, 1218: 3.53453191689, 1219: -0.150001518377, 1220: -21.7405648341, 1221: -0.423586969814, 1222: 77.0686490453, 1223: 11.7586363124, 1224: 46.8027842583, 1225: -4.12191622642, 1226: -0.245158645074, 1227: -0.369251939653, 1228: -0.227272914127, 1229: -0.808885687704, 1230: 0.108797898883, 1231: -26.6226571538, 1232: 25.3275551496, 1233: -64.3176234768, 1234: -31.4064008983, 1235: -7.98171868263, 1236: 14.4302544355, 1237: 26.1255292156, 1238: -19.5179066016, 1239: -29.3215654924, 1240: -0.139739740365, 1241: -54.1503312929, 1242: -2.22909324204, 1243: -3.43525779001, 1244: 7.47334818508, 1245: 15.51718038, 1246: -0.272265173384, 1247: -1.87824057359, 1248: -0.280632830892, 1249: -12.3325614205, 1250: -0.2834621158, 1251: -0.433253214297, 1252: -22.9678882457, 1253: -2.32379001513, 1254: 58.3975732453, 1255: -0.385664574781, 1256: -0.381678194916, 1257: 0.515574937275, 1258: -0.374212482376, 1259: -6.45824973577, 1260: -30.3469329853, 1261: -65.8780656399, 1262: -1.87400063453, 1263: 0.396842655398, 1264: 8.98708363981, 1265: -1.1928124599, 1266: -23.2131149683, 1267: 13.1842402257, 1268: -2.23717686802, 1269: 22.908532621, 1270: 24.1358381893, 1271: -1.51325518478, 1272: -16.9377411952, 1273: -0.745679746871, 1274: 51.1778568924, 1275: 4.16651782969, 1276: 9.70732212849, 1277: 3.91653290654, 1278: 3.00316527079, 1279: -0.656715519863, 1280: 54.6859164802, 1281: -7.77536866522, 1282: 0.786642720854, 1283: -6.24565293803, 1284: -25.7053758107, 1285: -39.3726641147, 1286: -18.7658517755, 1287: -2.60836734127, 1288: -2.41007814552, 1289: 12.8172898565, 1290: 46.3138264975, 1291: -9.41397739897, 1292: 0.983694489194, 1293: -21.4336442981, 1294: 11.316297794, 1295: 17.9088951976, 1296: -39.8434252181, 1297: 27.7907672256, 1298: 35.2543937823, 1299: -4.14761303025, 1300: -44.6456713041, 1301: -19.1100646396, 1302: -40.5750540837, 1303: -0.642331806585, 1304: -5.57847809774, 1305: -1.81266910087, 1306: -1.88198496473, 1307: -11.0452339808, 1308: -2.25979419338, 1309: -4.15616130261, 1310: 1.93198249821, 1311: -2.57187212234, 1312: -3.80812603409, 1313: -59.5076929219, 1314: 25.1498538396, 1315: 6.106726432, 1316: 2.76232171671, 1317: 12.3535417572, 1318: -1.78349161638, 1319: 54.8386969197, 1320: -20.1726135239, 1321: 93.8028973122, 1322: 20.0820832526, 1323: 21.1325087433, 1324: 3.1071550938, 1325: -13.1908984845, 1326: 5.52513525205, 1327: -18.2003990239, 1328: -5.69582916948, 1329: 0.0243524014656, 1330: -43.2518332116, 1331: -25.6207706694, 1332: -62.7662767055, 1333: -3.19909188385, 1334: -4.3635050706, 1335: -0.393068489395, 1336: -3.16840347162, 1337: 11.7880278825, 1338: -17.493629704, 1339: -3.16654080802, 1340: -8.37786693655, 1341: 8.80523651693, 1342: 1.86488003529, 1343: -34.8702682461, 1344: 6.35446126955, 1345: 3.01889394128, 1346: -21.8589046297, 1347: -3.86681012117, 1348: -8.98289676667, 1349: 19.8318545004, 1350: 3.64531632831, 1351: -9.13936744074, 1352: -25.6885637465, 1353: -3.02434539444, 1354: -13.3775235855, 1355: -2.0174972524, 1356: -1.95879711868, 1357: -1.35481332189, 1358: -9.95771933499, 1359: -1.94024647761, 1360: -1.06719579476, 1361: 28.8447323023, 1362: -2.30098360603, 1363: 46.3026863417, 1364: 2.66044093088, 1365: -1.90132366865, 1366: 23.8850388283, 1367: -2.22558664251, 1368: -8.97237076912, 1369: -11.563366215, 1370: -43.177643018, 1371: 99.9018150591, 1372: 28.3484054596, 1373: -25.5488906093, 1374: 7.28315323899, 1375: 96.2292876654, 1376: 31.8116862018, 1377: 10.2166435732, 1378: 36.3009782719, 1379: 55.721405498, 1380: -0.145500165855, 1381: -22.3331876477, 1382: -1.66533847126, 1383: -95.2267271808, 1384: -10.4161874222, 1385: -44.9615834718, 1386: -2.03887456895, 1387: -5.13202358205, 1388: -4.42438939123, 1389: 1.35424434212, 1390: 13.5469566056, 1391: -9.68544627945, 1392: -8.17831615502, 1393: 100.0, 1394: 52.0973194876, 1395: 29.8657706514, 1396: 17.6661428792, 1397: 0.0645635951504, 1398: -19.9837896882, 1399: 1.9862143431, 1400: -16.3821433933, 1401: -34.3572338525, 1402: -7.11052785637, 1403: 7.55323388728, 1404: 23.3247796537, 1405: 15.9678717389, 1406: -5.26199674296, 1407: -12.1979735671, 1408: 15.4551031516, 1409: 6.03265409215, 1410: 18.5876553872, 1411: 4.20210756478, 1412: 1.96207460205, 1413: 13.464380938, 1414: 8.13170501012, 1415: -0.470850504271, 1416: 12.9401654271, 1417: 13.9858600264, 1418: 7.54318733434, 1419: -5.95693851211, 1420: 5.68582957742, 1421: 11.2522498088, 1422: 3.37843110868, 1423: 0.602577262063, 1424: -5.08772784202, 1425: 12.1563406752, 1426: -17.2864524207, 1427: 17.7487969941, 1428: 5.65200239524, 1429: -1.86025221307, 1430: 3.174050065, 1431: -10.9234863329, 1432: 6.564413807, 1433: -11.8244983712, 1434: -1.83432873881, 1435: -11.8640902679, 1436: 5.81386036959, 1437: 4.30323871992, 1438: -0.304261504432, 1439: -38.5953362024, 1440: -25.0082601399, 1441: 10.2813379024, 1442: -3.93786610952, 1443: 10.8220957659, 1444: 15.2861742586, 1445: 3.00543570815, 1446: -6.50805796165, 1447: -1.7349855682, 1448: 9.05997419186, 1449: -0.452816818442, 1450: -2.22629404139, 1451: 5.7409948777, 1452: -0.303349524325, 1453: -2.55211003728, 1454: -0.0235366525087, 1455: -20.2922872778, 1456: -25.2046770759, 1457: -10.5514974717, 1458: -14.4063844869, 1459: -7.3193940928, 1460: -15.2314363938, 1461: -7.60247473443, 1462: 4.2387759159, 1463: 6.93728666256, 1464: 0.668389471031, 1465: -19.562448812, 1466: 4.75059712235, 1467: 2.38811263183, 1468: 2.40910257864, 1469: 0.56928777343, 1470: 35.4415034938, 1471: -3.32100953824, 1472: -2.38179016582, 1473: -1.74738043579, 1474: 0.490049599396, 1475: -43.4606131909, 1476: -19.7113751627, 1477: 8.43510308363, 1478: 4.96580410436, 1479: 3.57064090429, 1480: 1.32375719952, 1481: 5.28562227938, 1482: 5.40937446454, 1483: 0.330727793528, 1484: -26.2212998911, 1485: 6.70036510868, 1486: 14.8476488781, 1487: 21.6740019795, 1488: -0.0317588280607, 1489: -12.3593495264, 1490: 1.98816233273, 1491: 0.66555590361, 1492: 1.17345268152, 1493: 14.6360947557, 1494: 0.722189523403, 1495: -1.36241691914, 1496: 3.69098329702, 1497: -8.01909616705, 1498: 0.988958529294, 1499: 24.1184975631, 1500: 2.05244645679, 1501: 0.380413469523, 1502: 1.7148193527, 1503: -0.305444984608, 1504: -0.442001974832, 1505: 7.27705853957, 1506: 11.4309099171, 1507: 2.22465062158, 1508: -4.47940028155, 1509: 7.09358513397, 1510: -32.8560317471, 1511: 14.9561141152, 1512: 9.3441765072, 1513: -9.58607109963, 1514: -20.1588317017, 1515: -7.18664630509, 1516: -51.4013400715, 1517: -3.13866642723, 1518: -1.44993359107, 1519: -5.03650399384, 1520: -10.4165155079, 1521: -39.735008387, 1522: -0.148410604318, 1523: 1.22046983762, 1524: 1.29064886568, 1525: -6.63116291744, 1526: -0.0153477133589, 1527: 0.265870064822, 1528: 0.134267170769, 1529: 16.5121696022, 1530: 0.538479119321, 1531: -4.85700495112, 1532: 3.60161250467, 1533: -6.22465353264, 1534: -7.43089428157, 1535: 4.99629952979, 1536: -0.188873631588, 1537: -20.03681335, 1538: 13.9303631083, 1539: -1.87237674733, 1540: 0.486105694985, 1541: 1.93840933371, 1542: -0.835026649258, 1543: -0.147814955629, 1544: -25.3246939181, 1545: -0.152988587812, 1546: -0.249750712493, 1547: 1.28679976496, 1548: -0.512659784601, 1549: -0.131304017184, 1550: -0.0392793907232, 1551: -0.0442620238204, 1552: 0.506999526205, 1553: 11.4766574037, 1554: 0.433690818209, 1555: -0.445687868765, 1556: -0.339485220355, 1557: 1.05609392235, 1558: -2.43917939276, 1559: 12.8816049693, 1560: -4.84307824959, 1561: -16.5763817648, 1562: -28.6920413446, 1563: 14.5454130567, 1564: -1.338284861, 1565: 6.03972596117, 1566: 2.07382827541, 1567: 28.2911658092, 1568: -0.0446500871586, 1569: 22.4796027814, 1570: -0.122947904993, 1571: 5.38912964972, 1572: 28.9571128059, 1573: -33.380660083, 1574: -2.44868195621, 1575: -0.0497498592474, 1576: -0.231787492763, 1577: -0.281131141948, 1578: 0.0486153575486, 1579: -7.79754457575, 1580: -28.6946939391, 1581: -0.261157555532, 1582: 45.2222710826, 1583: -23.7989486916, 1584: 5.20122175566, 1585: -2.32894426406, 1586: 5.64901684941, 1587: -12.3102324609, 1588: -2.31305163282, 1589: -0.0254372567918, 1590: -20.2987259383, 1591: 7.54630387999, 1592: -0.709399496148, 1593: -8.91625379521, 1594: -6.33758208564, 1595: -0.054607433172, 1596: -0.0764351914448, 1597: -0.375975757739, 1598: -21.10126798, 1599: 0.00731341199259, 1600: -0.0327057694403, 1601: -37.7229926665, 1602: -0.345251332554, 1603: 14.951743491, 1604: -0.0359709373075, 1605: -0.0853164370351, 1606: 2.96733452565, 1607: -0.212318484431, 1608: 5.1477005351, 1609: 19.9565204189, 1610: 2.74660167054, 1611: 1.24027563049, 1612: -0.985886182728, 1613: -20.5147513199, 1614: 6.45281903909, 1615: 38.5722885699, 1616: 9.54994645092, 1617: -2.41858311091, 1618: -17.5152360623, 1619: -12.4393037425, 1620: 0.808731501676, 1621: 3.63756396961, 1622: 0.36413308159, 1623: -10.5231191006, 1624: 1.11285749216, 1625: -6.24794573136, 1626: 0.633156687722, 1627: 2.62143608406, 1628: 1.99308154343, 1629: -17.5075935704, 1630: 9.18609051673, 1631: 0.510259096838, 1632: -2.71106321422, 1633: 7.50366645758, 1634: 7.29032995187, 1635: 3.93946381432, 1636: 0.151160917128, 1637: -5.43497674035, 1638: -16.8352837514, 1639: -0.750841106982, 1640: 1.99523936997, 1641: -6.24958471195, 1642: -19.0391176241, 1643: -5.06327866775, 1644: 3.70210509763, 1645: -28.2383925126, 1646: 4.18594696704, 1647: -22.931426705, 1648: 24.2158900637, 1649: -17.1050778439, 1650: 2.95389540309, 1651: 15.3085504244, 1652: 0.55091110722, 1653: 16.8392502745, 1654: 1.93410526714, 1655: -0.26268717953, 1656: -5.25491904587, 1657: 14.6361968794, 1658: -0.022169639146, 1659: -4.42843127982, 1660: 0.123282385916, 1661: 0.661448202871, 1662: 10.6479732164, 1663: -9.13313379692, 1664: 3.39455965509, 1665: 2.0800358462, 1666: 0.0515897079742, 1667: -0.772046478579, 1668: 32.3608856366, 1669: 4.7023609237, 1670: -4.17339754984, 1671: 1.26449165466, 1672: 4.66478653214, 1673: 9.77521206249, 1674: -6.03377054522, 1675: 1.62792473653, 1676: 22.8110586781, 1677: 0.309138987544, 1678: -13.2820804832, 1679: 8.00025096301, 1680: -18.127051717, 1681: 2.57259225706, 1682: 28.9694234868, 1683: 0.519811529601, 1684: -10.12139538, 1685: 2.60076621116, 1686: 0.354534605975, 1687: -2.22793556326, 1688: 5.02189360695, 1689: 21.0264504012, 1690: 10.2336753977, 1691: 6.70621094613, 1692: -2.08767841067, 1693: -9.81914127993, 1694: 18.8554912479, 1695: -12.3383652812, 1696: -1.66749514239, 1697: 3.28279959705, 1698: 10.6471891991, 1699: 1.65619941134, 1700: 35.5068343482, 1701: -14.9720960266, 1702: 0.699946415727, 1703: 22.1472081377, 1704: -0.102167308381, 1705: 0.124243587341, 1706: 0.66986166644, 1707: 17.5004889465, 1708: 2.42510416873, 1709: -0.35731154685, 1710: 12.5621595178, 1711: -0.327182726192, 1712: 5.59123946232, 1713: -0.506512425389, 1714: -0.146234394711, 1715: 29.5143539129, 1716: -0.263654182893, 1717: 6.26397462077, 1718: -10.3627418357, 1719: 17.4436924343, 1720: -2.42253148532, 1721: -14.9018981024, 1722: 0.110994885995, 1723: -8.80982556385, 1724: 2.55934784868, 1725: 37.6462887196, 1726: -8.86532554859, 1727: -18.2731646819, 1728: 2.40709543509, 1729: 1.73197341705, 1730: 4.74857515957, 1731: -23.2321597067, 1732: 23.1476780934, 1733: 2.0264813779, 1734: 4.31928156189, 1735: -0.0669497709314, 1736: 8.28953373527, 1737: 6.78417280274, 1738: 18.4576982481, 1739: -7.18506228315, 1740: -3.48182021931, 1741: -16.3113429147, 1742: 56.150305066, 1743: -32.4982109554, 1744: -3.94026486473, 1745: 26.2305726757, 1746: -0.167345984733, 1747: -20.2661078809, 1748: -5.19188738189, 1749: -17.8965774536, 1750: -11.6252467841, 1751: 13.6576622989, 1752: 19.6847515212, 1753: 4.17519262469, 1754: -28.7041911875, 1755: -10.3933216204, 1756: -20.2244505375, 1757: 6.95272849629, 1758: -0.925730759541, 1759: 3.06742154264, 1760: -11.6978245791, 1761: -1.53038059879, 1762: 17.201311593, 1763: 2.94969838081, 1764: -5.06684682286, 1765: -2.38000825749, 1766: -14.6633986214, 1767: -5.15477383521, 1768: 9.11780962162, 1769: -2.47314532889, 1770: -1.68753522112, 1771: 16.0149057599, 1772: 21.2888857648, 1773: -3.91935505517, 1774: 23.7084399397, 1775: -15.2183609955, 1776: -7.47052645279, 1777: 0.303649883501, 1778: -14.7061325288, 1779: -0.26004889275, 1780: -6.75605027396, 1781: 3.10713572078, 1782: -23.1921775317, 1783: -14.0032277319, 1784: -22.074467305, 1785: -23.6096673336, 1786: 15.3494584149, 1787: -17.8495040366, 1788: -12.3843191806, 1789: -17.1212104714, 1790: 6.49238565849, 1791: 1.48754658103, 1792: -14.4456018924, 1793: 3.02452057269, 1794: 13.0260174988, 1795: -18.9346071327, 1796: 9.1084939833, 1797: -16.5377892537, 1798: -2.34905490206, 1799: 19.3590251046, 1800: -4.41346414443, 1801: -2.00584462687, 1802: -10.8216174044, 1803: -3.8491351658, 1804: 9.87170890839, 1805: 12.54951635, 1806: -20.7581393771, 1807: 12.1282377608, 1808: -18.1692030228, 1809: -21.9798102074, 1810: 8.14540003555, 1811: 0.9346747833, 1812: -1.01135507437, 1813: 0.468251520703, 1814: 2.39437340808, 1815: -0.47514867841, 1816: -0.190932964457, 1817: 11.7137803612, 1818: -0.678961047121, 1819: 16.3135332906, 1820: -4.98106729311, 1821: -9.33448398604, 1822: -4.16920624785, 1823: -3.47913753975, 1824: -8.44187846018, 1825: 31.9607526594, 1826: 8.75007360682, 1827: 8.66999418394, 1828: -1.49297546154, 1829: -0.700403278532, 1830: 1.75555875128, 1831: 34.471168615, 1832: 4.85338002881, 1833: -0.514235553153, 1834: -1.84260111787, 1835: -0.932969813216, 1836: 4.02764835187, 1837: -1.79281219977, 1838: 10.6924253982, 1839: -1.29527136163, 1840: -0.950204772828, 1841: -1.28341103362, 1842: -7.88129977281, 1843: -0.404325852156, 1844: 8.38716033793, 1845: 1.58164845572, 1846: 18.971390686, 1847: -4.57208488279, 1848: -1.67330936134, 1849: 4.46600223769, 1850: -1.59171066112, 1851: 5.68315950035, 1852: -1.82601888836, 1853: -2.41759932513, 1854: -0.664949756659, 1855: 8.86620805957, 1856: 6.11186600846, 1857: 9.78049610644, 1858: 4.1881438604, 1859: 11.1301390889, 1860: -0.124610581605, 1861: 12.9749027935, 1862: 15.6420678304, 1863: -3.0436099381, 1864: -10.7963942465, 1865: -25.1009078586, 1866: 28.4261229533, 1867: 2.05031537979, 1868: 4.7523628652, 1869: 5.46002931872, 1870: 24.5487221367, 1871: -0.456391173866, 1872: -0.006672514741, 1873: -1.82574185245, 1874: -7.79639287278, 1875: -0.45617612562, 1876: -1.27308772635, 1877: -0.912807370976, 1878: 1.16544044364, 1879: -4.87762491701, 1880: -3.55509809278, 1881: -0.0395649163423, 1882: -0.654480342242, 1883: -7.65559578156, 1884: 2.25256237823, 1885: -0.0262387137852, 1886: 4.13379136692, 1887: -19.3993114938, 1888: -4.90386672949, 1889: 1.12109078436, 1890: 4.24947584577, 1891: -0.513564085358, 1892: -0.393983477542, 1893: 0.668074468599, 1894: -0.356263601684, 1895: -2.04779268523, 1896: -1.92313967324, 1897: -0.453578954214, 1898: -0.438954416449, 1899: -0.293172975218, 1900: -0.348665432106, 1901: -0.837239414692, 1902: -5.43709849892, 1903: 1.10549946705, 1904: -1.06149602888, 1905: -1.08016393884, 1906: 4.65297565741, 1907: 5.47142392317, 1908: 23.3127702504, 1909: 6.45449034092, 1910: -7.65139954819, 1911: -31.7954510045, 1912: 1.34578655652, 1913: -4.71689192546, 1914: 6.53080112174, 1915: -1.47271987377, 1916: 4.51662211839, 1917: -0.39617819246, 1918: 0.913651239141, 1919: -0.305409174776, 1920: 1.90093691332, 1921: 20.3177652963, 1922: -3.3926472586, 1923: 1.34329902209, 1924: -0.37611532371, 1925: -0.411392646569, 1926: -0.756573351743, 1927: -0.145498908675, 1928: -10.9338337782, 1929: -17.6128250424, 1930: 3.03794747451, 1931: -10.361545302, 1932: 16.1312838061, 1933: 10.7617777202, 1934: 2.78703083633, 1935: -0.327938021637, 1936: -2.42157772468, 1937: 5.6274862639, 1938: -0.463077005661, 1939: -25.1200464891, 1940: 4.23886872008, 1941: 2.24550226166, 1942: -5.54668566068, 1943: -7.2426245867, 1944: -0.458569402937, 1945: 1.0629572733, 1946: -0.333292227154, 1947: 1.63736662683, 1948: -0.428906181308, 1949: -0.34727909499, 1950: 15.5109675113, 1951: -1.86804676712, 1952: 10.0642877382, 1953: -0.162103719318, 1954: -0.217452495904, 1955: -0.827816350577, 1956: -0.438443864245, 1957: 0.278345539015, 1958: -4.23443735632, 1959: 12.6899405069, 1960: 2.32350864855, 1961: -1.52990565222, 1962: 3.49246223292, 1963: -7.83490927539, 1964: -12.579273855, 1965: -6.5852741714, 1966: -2.41373731192, 1967: 0.408615697747, 1968: 14.1002115264, 1969: -0.132706652046, 1970: 8.22266390931, 1971: -0.341580533643, 1972: 2.62888598346, 1973: -8.58976327204, 1974: -1.65291247727, 1975: 1.2270790475, 1976: -6.31135143654, 1977: -0.323491694722, 1978: -3.88007164235, 1979: -12.8082826449, 1980: 1.58033928952, 1981: 9.56055670866, 1982: 4.43120687541, 1983: 6.01037630707, 1984: -2.40609754477, 1985: -1.37405153049, 1986: 13.8603022605, 1987: -0.243812735521, 1988: -8.21723912681, 1989: 6.11575637489, 1990: 5.59870475218, 1991: 5.48720440956, 1992: -21.9063613129, 1993: -9.39983611048, 1994: -28.6915308688, 1995: 1.45924946754, 1996: 6.57619617861, 1997: 32.6642429481, 1998: -29.4107408369, 1999: -6.62520662303, 2000: -5.99373586552, 2001: -0.337643227194, 2002: 11.0773562353, 2003: -1.76693397493, 2004: -1.89491218707, 2005: -0.537025986533, 2006: 4.87491915224, 2007: 0.505032902231, 2008: -8.41400341545, 2009: -2.07322322802, 2010: -0.700764634282, 2011: -5.1688703412, 2012: 2.49794169487, 2013: 16.3837515777, 2014: 19.0948561831, 2015: 5.6878911862, 2016: 4.4862376396, 2017: -31.3117478381, 2018: 19.4165811134, 2019: -4.40197344543, 2020: -23.6591850002, 2021: -8.88337541254, 2022: 3.37213993918, 2023: 3.06521352154, 2024: -2.68829627957, 2025: -8.42364028555, 2026: 1.71385432432, 2027: -1.08394222136, 2028: 8.05146794216, 2029: 24.2883967147, 2030: -4.41511555447, 2031: -28.2731752725, 2032: -0.734157266198, 2033: -9.03167600409, 2034: 0.353097426461, 2035: 2.72761547972, 2036: 0.996961191256, 2037: -15.8788392935, 2038: 7.29329498092, 2039: 24.1224078062, 2040: 3.61249562008, 2041: 11.4233819826, 2042: -14.9897850759, 2043: 21.7590000758, 2044: -0.904888933945, 2045: 5.65262197875, 2046: 3.79427634929, 2047: 12.9251478906, 2048: -3.80349177462, 2049: 22.1311352697, 2050: -27.630205525, 2051: -0.964854242573, 2052: 13.8602788352, 2053: 1.07599307469, 2054: 1.06569041074, 2055: -1.05834427424, 2056: 6.08994968576, 2057: -1.34304468421, 2058: -2.08443548582, 2059: 2.6251945918, 2060: -1.71282284261, 2061: -18.1207740264, 2062: -5.4790871136, 2063: -1.88477305832, 2064: 10.3915766206, 2065: -1.76839610961, 2066: -0.227091528701, 2067: 24.1292550156, 2068: -12.9002897815, 2069: -8.83311954057, 2070: 0.614967030556, 2071: 12.1500433815, 2072: -7.8209759006, 2073: 25.6859271199, 2074: -0.324027067515, 2075: -3.12009995943, 2076: -32.5701314714, 2077: -3.52040945593, 2078: 2.3019749282, 2079: 4.80523426786, 2080: 0.282612776172, 2081: 21.3990490332, 2082: 0.295277649853, 2083: -13.5539916628, 2084: -1.83165865254, 2085: -4.22662339595, 2086: 9.63744849885, 2087: -1.99998357007, 2088: -3.57426363406, 2089: 16.4204977524, 2090: 0.265243036218, 2091: 19.1786244046, 2092: -46.5726094578, 2093: -0.478197895737, 2094: 20.7653085195, 2095: 0.134567429552, 2096: -13.5322367861, 2097: -0.929842867724, 2098: -19.5214423593, 2099: -12.8442300571, 2100: -3.16742982341, 2101: 20.0586699111, 2102: 11.7069219582, 2103: -18.7210282925, 2104: -8.3409068844, 2105: -17.0989407029, 2106: -0.483834220749, 2107: -4.1934130375, 2108: 2.7666415796, 2109: -7.19856162336, 2110: -2.17420888502, 2111: 29.7556782274, 2112: 5.92425771728, 2113: 10.5768766834, 2114: 0.395216527049, 2115: -10.0163606189, 2116: 7.9573953214, 2117: 3.23156991586, 2118: 6.21060397165, 2119: -1.41302390659, 2120: 7.51342340503, 2121: 17.045894994, 2122: -6.38302545332, 2123: 20.9113226938, 2124: -4.13851919858, 2125: -3.25393180734, 2126: 1.25895689665, 2127: 1.45046903706, 2128: 1.06230609644, 2129: -6.38936757781, 2130: 4.28325403478, 2131: -17.9145852583, 2132: -7.72802202255, 2133: -13.1005347798, 2134: -17.4232037163, 2135: 14.5586497396, 2136: -6.98187141025, 2137: -17.5769952323, 2138: -11.0234203539, 2139: 21.1041460211, 2140: -0.688946431508, 2141: 11.7892670983, 2142: 0.155772158629, 2143: 6.60159544899, 2144: -11.3388220953, 2145: 9.11827098515, 2146: -2.43820450633, 2147: 0.382667793493, 2148: 8.42924445589, 2149: -4.51217854452, 2150: 0.336381414408, 2151: 3.50807876644, 2152: 3.16125162156, 2153: -2.41207760712, 2154: 8.29983832356, 2155: -23.72310937, 2156: 6.75273702509, 2157: -8.74580538673, 2158: -19.7532794484, 2159: -3.98572367078, 2160: -0.853640147236, 2161: 22.0272613299, 2162: 1.50663026279, 2163: 0.583294606278, 2164: 0.0923176113862, 2165: -0.899086451302, 2166: -4.08012126019, 2167: 0.151966972642, 2168: 5.47694049142, 2169: -6.26851926279, 2170: 3.25217915278, 2171: -7.50282289337, 2172: 2.1274011313, 2173: -13.0358353802, 2174: 10.5872648236, 2175: 3.26058032867, 2176: 10.7205078449, 2177: -0.0552234949261, 2178: -2.17520195764, 2179: 0.81243791013, 2180: -0.956726893834, 2181: 9.63727438021, 2182: 8.78234910515, 2183: 0.364985574059, 2184: 0.283495925584, 2185: 0.211960207485, 2186: 0.164834362093, 2187: 14.4675389211, 2188: 0.126602607157, 2189: 1.44340851311, 2190: -0.0765217756647, 2191: 4.1291003181, 2192: -0.0554435106465, 2193: -0.688738469983, 2194: 3.17978861196, 2195: 3.38252940808, 2196: 3.47451755723, 2197: 3.74775735434, 2198: 1.4312746791, 2199: -0.641598828227, 2200: 1.34244894074, 2201: 0.268686840662, 2202: 0.648252431891, 2203: 0.168283138118, 2204: 6.00094671965, 2205: 4.32665320913, 2206: 6.77593460805, 2207: 10.0171287743, 2208: 8.62573691539, 2209: -13.4642518125, 2210: 10.5081053183, 2211: 9.11875305602, 2212: 3.14009936057, 2213: -7.94538629319, 2214: -14.2736775808, 2215: 12.230840649, 2216: -4.63293580432, 2217: 5.84863150887, 2218: 3.75142465163, 2219: 5.88525673823, 2220: 0.0485987054012, 2221: 0.812363687171, 2222: 0.613553832865, 2223: -6.48603349583, 2224: -0.021347973168, 2225: 0.398428374559, 2226: 1.6910380736, 2227: 6.71867377351, 2228: 3.42871237535, 2229: 0.385503787788, 2230: -6.6151257556, 2231: -2.08300433425, 2232: 0.578482091836, 2233: 0.673614570963, 2234: 1.15935101706, 2235: -2.03437828133, 2236: -14.4964395845, 2237: -7.5968285328, 2238: 0.10236321426, 2239: 5.35285719654, 2240: -0.0544603028272, 2241: 0.0230295310763, 2242: 3.84956937689, 2243: 0.188962007886, 2244: 0.227220680838, 2245: 2.6735091196, 2246: 0.29242219953, 2247: 0.0576524227545, 2248: 0.102021745463, 2249: 0.0666369685988, 2250: 0.212370287898, 2251: -8.42822055027, 2252: -0.614865197197, 2253: -0.304304520129, 2254: -0.250384979263, 2255: 1.67444749031, 2256: 5.34714575289, 2257: 6.81071889724, 2258: 3.06268891694, 2259: -10.660962138, 2260: -14.9887351786, 2261: -1.38347029517, 2262: 1.27913025213, 2263: -10.5875706205, 2264: 0.36503837403, 2265: -8.73057627135, 2266: 0.166065947333, 2267: -7.14875333704, 2268: 0.123359978156, 2269: 4.64686448933, 2270: 18.6350051231, 2271: -1.72954729634, 2272: -0.107802994009, 2273: 0.16629432742, 2274: 0.664057130101, 2275: 1.06570307305, 2276: -0.0876445531232, 2277: -2.68113753548, 2278: -10.9352378026, 2279: 3.08597501049, 2280: 9.75169391322, 2281: -0.323478555047, 2282: -3.00013872828, 2283: 7.84683319683, 2284: -0.153070830291, 2285: 0.197718584919, 2286: 6.18050563206, 2287: 0.0835611525106, 2288: -13.135739009, 2289: -0.48645029211, 2290: 0.0937688923939, 2291: -2.91978165853, 2292: -10.0187069136, 2293: -0.0418639785624, 2294: 0.328015959163, 2295: 0.126849608858, 2296: 2.01251493722, 2297: 0.15576164548, 2298: -0.0248105951954, 2299: 11.1209079484, 2300: 0.199702800374, 2301: 7.58825966163, 2302: 0.190476883973, 2303: 0.171026869669, 2304: -1.47573190872, 2305: 0.175903524335, 2306: -0.329149088847, 2307: 6.28729493733, 2308: 11.7015551348, 2309: -0.354683281513, 2310: 0.741846031519, 2311: -15.0977333682, 2312: 6.84212119941, 2313: 12.4619090926, 2314: -0.0577888796507, 2315: -8.13383391156, 2316: -7.03236588739, 2317: -0.456234283733, 2318: 0.27184059734, 2319: -1.96822274219, 2320: 0.0316787356437, 2321: -1.98751971472, 2322: -2.30176117049, 2323: -3.10847242732, 2324: -4.34615352697, 2325: 1.13089813584, 2326: -0.640373334997, 2327: -9.89882310015, 2328: 1.08770196547, 2329: 0.441925479321, 2330: -0.0879359512181, 2331: 8.19845136312, 2332: 0.840976722096, 2333: 0.251539198545, 2334: 0.495058585438, 2335: 6.88525711402, 2336: -4.03490436985, 2337: -8.40897979385, 2338: 9.21961221503, 2339: 5.54917889329, 2340: 6.12691345897, 2341: -16.4905990921, 2342: -7.66665870332, 2343: -15.8450514653, 2344: -12.6795332373, 2345: -8.95711376181, 2346: 29.3108822884, 2347: 0.166289604261, 2348: 1.22574468278, 2349: 7.92374167117, 2350: -0.165182852889, 2351: 5.25343627006, 2352: 0.134078666557, 2353: 0.235383819132, 2354: -0.880735647726, 2355: -7.77746003547, 2356: 0.349931131497, 2357: -0.0546672470648, 2358: -0.348510109563, 2359: 0.117163501343, 2360: -0.397122494387, 2361: -3.07601703897, 2362: 6.63224091448, 2363: 9.47509253375, 2364: 3.79196147332, 2365: 0.675828751491, 2366: -18.2784312783, 2367: 12.1114677769, 2368: -10.4321827837, 2369: -14.0179582824, 2370: -3.53990010654, 2371: 12.434105253, 2372: -10.5119773779, 2373: 1.12679230407, 2374: -8.61885944971, 2375: 5.94497102666, 2376: 7.81061312642, 2377: -1.2906352089, 2378: 8.79164325541, 2379: 0.409884966646, 2380: -15.6308603925, 2381: 0.133581406552, 2382: -9.8936569858, 2383: -2.69176600234, 2384: 2.02752357122, 2385: -4.18628732977, 2386: -18.2187490936, 2387: 5.89360224874, 2388: 18.0072925873, 2389: 4.72018541719, 2390: -1.09028785698, 2391: -7.26489565476, 2392: 20.6500929059, 2393: -0.411693044834, 2394: 2.23677095799, 2395: -8.19863756482, 2396: 11.8165045679, 2397: 1.7783192766, 2398: 10.100953628, 2399: -28.9615344843, 2400: 1.46740734537, 2401: 16.7614462045, 2402: 0.228588914534, 2403: 0.336236417092, 2404: 0.45481837776, 2405: 15.8437537637, 2406: -0.445526084795, 2407: 0.228537362663, 2408: 9.95857916329, 2409: 0.314192655776, 2410: 2.88628961494, 2411: 0.358719770842, 2412: 0.182310452318, 2413: 6.88661459177, 2414: 0.527978960438, 2415: -0.642775079497, 2416: 17.0751663364, 2417: -11.3945559232, 2418: -1.96118134912, 2419: -9.65010463748, 2420: 23.8525665813, 2421: -9.52157124368, 2422: 8.69811210297, 2423: 6.52676285703, 2424: -8.81561539282, 2425: -30.8369571167, 2426: -2.28634927276, 2427: 1.32405790896, 2428: 0.514089746531, 2429: 1.26032205447, 2430: 16.7879270062, 2431: 5.96776767251, 2432: 4.19021513297, 2433: 0.498886472853, 2434: -2.17514790798, 2435: 6.30490638125, 2436: -1.11372805319, 2437: -1.7744698288, 2438: -2.66087453359, 2439: -10.4377429402, 2440: 35.1170028434, 2441: -26.7984140281, 2442: -9.56250402075, 2443: 9.27277100347, 2444: -0.486318404408, 2445: -34.965974462, 2446: -10.3310458332, 2447: 7.07908047528, 2448: 5.63620151054, 2449: 10.6708781839, 2450: -0.0484449687487, 2451: 32.718426528, 2452: 18.0870456652, 2453: 4.47877569283, 2454: 3.96080921476, 2455: 4.76951801432, 2456: -10.5300338206, 2457: -10.8486530312, 2458: -8.51134198071, 2459: 1.54542050729, 2460: -9.62947156069, 2461: -3.25043677983, 2462: -35.8726783055, 2463: -4.42646821475, 2464: -7.3916910128, 2465: -12.594284029, 2466: -8.9643199692, 2467: 17.0773345852, 2468: 25.3212170871, 2469: -2.78799576955, 2470: 3.12451962447, 2471: -6.39568880115, 2472: -0.0394593414861, 2473: -14.7537715884, 2474: -15.163271431, 2475: -21.9367621914, 2476: -23.1352199202, 2477: -6.94010899964, 2478: -2.10776359141, 2479: -4.87150935703, 2480: 4.3899051623, 2481: -15.9643672661, 2482: -14.7872186798, 2483: 18.971025001, 2484: 4.61606394572, 2485: -4.72322035536, 2486: -18.4965439507, 2487: 1.49416303149, 2488: 8.28195870187, 2489: 1.32694233187, 2490: 7.3997784638, 2491: -2.59852554908, 2492: 8.21662304449, 2493: -1.86229889774, 2494: -3.38042465837, 2495: 3.508982318, 2496: -3.50584087116, 2497: 2.34537804869, 2498: -2.84903890021, 2499: -1.58725766866, 2500: -28.4175968332, 2501: -8.3383783228, 2502: -29.6809868056, 2503: 34.5848733673, 2504: -32.5386747649, 2505: -4.45991553322, 2506: -10.8525941981, 2507: -1.83374897935, 2508: 10.308196718, 2509: 0.86935570262, 2510: -5.68126588058, 2511: -3.20060160567, 2512: 10.1108113427, 2513: -3.2362023145, 2514: -2.28508729301, 2515: -1.23074662798, 2516: -3.32959063386, 2517: -15.0693696077, 2518: -8.35285194606, 2519: -14.8975244239, 2520: -13.2498944952, 2521: 3.44576147569, 2522: -34.4923948308, 2523: -17.2662879988, 2524: -0.0608118702235, 2525: 7.8578587176, 2526: -2.53436562737, 2527: 11.7110338954, 2528: 1.1851019262, 2529: -17.0337224606, 2530: 5.8278813315, 2531: -9.43393478539, 2532: -3.76025124817, 2533: 11.0909494141, 2534: 8.29397703086, 2535: -3.39478408061, 2536: 28.8567295434, 2537: -1.91916360688, 2538: -3.06094534352, 2539: -3.28640052188, 2540: 3.29220816467, 2541: -2.67986932658, 2542: 13.9143411442, 2543: -0.159076193261, 2544: -1.28157730067, 2545: -0.281568037858, 2546: -9.81470357786, 2547: 0.916228655198, 2548: 8.87190872949, 2549: 15.0491117963, 2550: -3.62211740768, 2551: -3.80157267617, 2552: -1.89610208632, 2553: 16.5906872143, 2554: 3.78827246827, 2555: 11.7177189702, 2556: -27.5207710779, 2557: -8.68923329403, 2558: 19.4156200926, 2559: 11.4130160508, 2560: -4.9003308143, 2561: -3.41909015676, 2562: 0.107076566109, 2563: 0.105900581625, 2564: 14.4453654891, 2565: -13.1043281542, 2566: 6.23499578487, 2567: 3.57332875336, 2568: 3.18830979998, 2569: -1.00038038839, 2570: -0.839336090918, 2571: 14.2566926394, 2572: -3.86734139236, 2573: -0.731434940143, 2574: -10.2589613427, 2575: -0.774798526868, 2576: -19.0667502501, 2577: -0.973018134967, 2578: -0.403868639128, 2579: 12.8005495986, 2580: -5.29059787057, 2581: 4.97342369674, 2582: 1.49783119489, 2583: -1.73602025555, 2584: -13.7652232601, 2585: -16.9566175554, 2586: -12.3846561895, 2587: -2.16594934235, 2588: 5.34696771593, 2589: -1.62551092146, 2590: -0.805245081008, 2591: -17.4366614842, 2592: -0.63100617336, 2593: -0.824546666347, 2594: -3.69710960087, 2595: -0.613581782652, 2596: -0.641802304118, 2597: -0.728174629244, 2598: -0.661030524729, 2599: -2.72936893147, 2600: -5.61189042126, 2601: 0.157810929257, 2602: -1.51060995617, 2603: -0.997141689716, 2604: -2.31344060258, 2605: -1.84295349102, 2606: -9.95019209391, 2607: -10.4574349358, 2608: -6.43394847575, 2609: 3.38140553712, 2610: 32.4510103721, 2611: -2.32775082666, 2612: 26.5897342179, 2613: 0.538101359346, 2614: 14.434522803, 2615: -0.602173436851, 2616: -26.3906860207, 2617: -0.6686750843, 2618: 2.19180068353, 2619: -42.6212354103, 2620: 8.02278761517, 2621: -0.605927060295, 2622: -0.690934558329, 2623: -0.535436454848, 2624: -2.54320838929, 2625: -0.192311186306, 2626: -17.0026037322, 2627: -10.9813829294, 2628: 2.25605412949, 2629: 27.8028632556, 2630: 20.0980418902, 2631: -1.58602676871, 2632: 1.07528950989, 2633: -8.96322135435, 2634: -3.78902407124, 2635: -0.462431716661, 2636: -0.617950163025, 2637: -37.7564071676, 2638: 0.618552918614, 2639: -7.1037996152, 2640: -2.66811323984, 2641: -4.71455741088, 2642: -0.777167520956, 2643: -3.12289659331, 2644: -0.782955454474, 2645: 9.64419750382, 2646: -0.62306553725, 2647: -0.565185963355, 2648: 6.59036067779, 2649: -3.29929529523, 2650: -14.096397135, 2651: -0.585511786412, 2652: -0.769040305506, 2653: 0.887823911737, 2654: -0.842128497482, 2655: -1.43850141031, 2656: 18.2950677412, 2657: 52.0128767998, 2658: 0.40570752524, 2659: -1.00517198313, 2660: -33.4319569872, 2661: 19.7196284579, 2662: -29.0772811218, 2663: 11.5337711041, 2664: -6.58550612097, 2665: -6.04982162544, 2666: -4.45058275662, 2667: -1.7663066461, 2668: 15.8851998424, 2669: -0.74445516736, 2670: 5.61967017035, 2671: -9.01689477133, 2672: 2.8569124995, 2673: 14.5822512534, 2674: 1.19968830012, 2675: -2.89500044728, 2676: -16.1168385014, 2677: 18.3737373977, 2678: 15.1139252843, 2679: -2.33251289494, 2680: 5.3607521201, 2681: -3.37029143153, 2682: -3.5408148574, 2683: -10.2912640051, 2684: -8.66177838793, 2685: 5.93674573854, 2686: 3.23707666976, 2687: 10.259765117, 2688: -0.381880707722, 2689: -5.97838607795, 2690: -6.81123318304, 2691: 23.6790445231, 2692: -1.10219985582, 2693: 5.63710500916, 2694: -5.4149081281, 2695: 26.7855792232, 2696: 2.19984072685, 2697: 8.3323603517, 2698: 5.41517050821, 2699: 0.475335145459, 2700: 11.5469033967, 2701: -1.07200407952, 2702: -0.685630158472, 2703: -4.38915125915, 2704: -2.45672857596, 2705: -2.94151491335, 2706: 1.73439448054, 2707: -10.4596021455, 2708: -2.60389792077, 2709: -5.04102423773, 2710: 11.5038959255, 2711: 9.77288670948, 2712: 2.14814565305, 2713: -2.38676799546, 2714: -2.07265976722, 2715: -16.0442433219, 2716: -0.997641247557, 2717: -42.603176709, 2718: -16.728755536, 2719: -6.56966963376, 2720: -6.93025021099, 2721: 13.4193525722, 2722: 1.32547258637, 2723: -5.67083961672, 2724: -0.340750107623, 2725: 24.6191567787, 2726: -25.6720332629, 2727: 5.30186877297, 2728: -36.869795734, 2729: 24.383628517, 2730: -2.95414418011, 2731: -10.8932499095, 2732: 16.162999511, 2733: -2.65442405475, 2734: -9.6905993597, 2735: 2.08103446026, 2736: 0.420338872091, 2737: -4.88638107881, 2738: -12.83472474, 2739: 29.6797397569, 2740: 9.50797457519, 2741: -1.90682993772, 2742: -2.3532662273, 2743: -9.19615712488, 2744: -0.677215207121, 2745: 24.9325206004, 2746: 9.91623063049, 2747: 21.6758468142, 2748: -8.79583072827, 2749: -3.23660396319, 2750: 24.6401600909, 2751: -3.2060171304, 2752: -3.07516765298, 2753: -2.49994427482, 2754: -18.2400670122, 2755: -3.106308745, 2756: -3.74987579813, 2757: -6.27957665533, 2758: -3.52440604764, 2759: -4.35831259206, 2760: -3.61221139808, 2761: -3.42058980634, 2762: -0.877730725031, 2763: -3.99932602232, 2764: -1.7171276191, 2765: -2.6145844889, 2766: -31.8563676528, 2767: 14.0815352375, 2768: -13.8661307501, 2769: 34.76225926, 2770: 2.92415446423, 2771: 36.7633075454, 2772: 48.7617097065, 2773: -5.20705634529, 2774: -0.451685615249, 2775: -13.0512712388, 2776: -0.845829813065, 2777: -5.1221727653, 2778: -5.81376004864, 2779: 39.1019955645, 2780: 24.9851200315, 2781: 16.7632475537, 2782: -3.39665034544, 2783: 6.34351614767, 2784: 9.55518310189, 2785: -4.62382916457, 2786: 10.3001050909, 2787: -6.99097924406, 2788: 2.591779852, 2789: 23.3633697062, 2790: -15.2826810432, 2791: -10.7344714251, 2792: 36.7396838046, 2793: 0.120840238114, 2794: 26.8400007334, 2795: 16.200696332, 2796: 10.8516911477, 2797: 16.6222429111, 2798: -0.738562608424, 2799: 2.83496005507, 2800: -12.2761991254, 2801: -24.2737038964, 2802: 1.78456276701, 2803: -6.89924981635, 2804: -18.7953354901, 2805: -1.72760779928, 2806: -2.92599047931, 2807: 9.80425177878, 2808: 1.26633672108, 2809: -22.784893399, 2810: 6.91737037944, 2811: -36.8469168529, 2812: -9.46705144499, 2813: -1.42707337155, 2814: -4.23711646091, 2815: -3.72959542378, 2816: -7.45253923162, 2817: 0.597052250427, 2818: -3.89631706147, 2819: 11.1251567659, 2820: 2.77073791356, 2821: 3.47381310468, 2822: 27.7775334396, 2823: -13.5961076772, 2824: 8.77038517944, 2825: 35.6724602259, 2826: 3.44494536251, 2827: 17.5936145085, 2828: -2.22408126112, 2829: 4.29357617777, 2830: 7.69661412422, 2831: 12.6009557038, 2832: -10.6812053726, 2833: -0.092933540544, 2834: -20.024430742, 2835: 10.61137725, 2836: -2.81205918468, 2837: 8.81768967253, 2838: -1.28171409878, 2839: -15.3812400257, 2840: 0.398115459792, 2841: -5.22292283508, 2842: 0.919436432691, 2843: 14.151737836, 2844: 19.9880959568, 2845: 0.170231108736, 2846: -7.05271462454, 2847: -2.26333215032, 2848: 0.733421428499, 2849: 2.69298113349, 2850: -1.26511001116, 2851: 24.8996494713, 2852: -6.12327788412, 2853: 0.166106581946, 2854: -9.35595306617, 2855: -5.45548575365, 2856: 3.86578198728, 2857: -15.1172272037, 2858: 0.346196034487, 2859: -7.26820509968, 2860: 1.37243503276, 2861: 16.0170743354, 2862: -1.68941595071, 2863: 1.86221945306, 2864: 3.74783985272, 2865: 0.755602675192, 2866: -20.3975645602, 2867: 6.36556059432, 2868: -1.89349670877, 2869: 6.83408361295, 2870: -10.82893406, 2871: 18.1860443821, 2872: -7.63552934185, 2873: -1.67970860223, 2874: -6.92047065141, 2875: 1.02695454146, 2876: 4.61867495941, 2877: 3.09174802584, 2878: -36.4046596561, 2879: -5.17800953541, 2880: -8.02497256569, 2881: 1.02553224602, 2882: -20.2282960897, 2883: 2.25116515191, 2884: 0.522621883043, 2885: 19.2474352883, 2886: 0.960261697698, 2887: 1.05804957471, 2888: 1.46728153611, 2889: 6.61285767605, 2890: 0.960690398385, 2891: 7.29658661729, 2892: -4.3623175655, 2893: -17.3330915272, 2894: 1.48213767165, 2895: 15.3316064895, 2896: 0.954963152753, 2897: -1.00202304938, 2898: -20.7247915729, 2899: 0.592665961777, 2900: -0.125944797433, 2901: 1.17517971755, 2902: -11.380557506, 2903: -12.7699747122, 2904: -9.10729429099, 2905: -14.5758027534, 2906: 4.03564687848, 2907: -24.1754263438, 2908: -9.06993549041, 2909: -2.94177989008, 2910: 5.47375325426, 2911: 18.5434724472, 2912: -1.93953598083, 2913: -20.0110957925, 2914: -2.22998627385, 2915: -17.227466605, 2916: 10.8889586294, 2917: 9.98975388786, 2918: 0.155458633343, 2919: 0.724550928319, 2920: 5.87809469433, 2921: 14.3444150865, 2922: 0.0608890296238, 2923: 0.708669691153, 2924: -1.17936378029, 2925: 0.336456437059, 2926: 6.93887520661, 2927: -0.86627499494, 2928: -13.9099777248, 2929: -4.15493884656, 2930: 18.0597165496, 2931: -2.91925377433, 2932: -6.52712746752, 2933: 17.095731942, 2934: 16.5752998554, 2935: 5.63385901386, 2936: 0.659993978724, 2937: -13.1239896671, 2938: -0.0234800198202, 2939: 0.0383608597838, 2940: -9.20159186014, 2941: 0.183894395301, 2942: 0.583151114793, 2943: 0.667083640139, 2944: 0.0879902965619, 2945: 0.227965740615, 2946: 0.235058644756, 2947: -0.0629084569412, 2948: -0.216523291445, 2949: 5.60488547481, 2950: -0.191769731722, 2951: 0.709675308682, 2952: 0.765267636253, 2953: 13.5862540756, 2954: 4.56219344566, 2955: -27.0256848877, 2956: 2.33524852441, 2957: 43.6601479827, 2958: -0.418542090059, 2959: 1.78843742128, 2960: -3.83355411262, 2961: -17.4828835044, 2962: -0.714558049006, 2963: 5.33605623231, 2964: 0.0562485793474, 2965: -23.0538359628, 2966: 0.209760913241, 2967: -9.72243442985, 2968: 47.1327853799, 2969: 23.2683220153, 2970: -1.68995024091, 2971: 0.168674531773, 2972: -0.158093087853, 2973: -0.0333215528942, 2974: 0.183898547633, 2975: 2.67581087974, 2976: 28.393021997, 2977: 6.63588311086, 2978: 2.02607233021, 2979: 24.0782335923, 2980: 8.3152673588, 2981: 13.3875909686, 2982: 4.39803310293, 2983: -1.76414739942, 2984: 0.154961024188, 2985: 0.0775103015591, 2986: 38.7604914196, 2987: 2.55861972605, 2988: 2.11980257603, 2989: 7.19641000805, 2990: 1.09767251925, 2991: 0.0234983162986, 2992: 0.651172364765, 2993: 0.134859849589, 2994: 45.4871369159, 2995: 0.033991854937, 2996: 0.232257603998, 2997: 19.465945675, 2998: -1.37730530576, 2999: -30.2106020247, 3000: 0.196953191948, 3001: 0.398080688437, 3002: 3.19907871885, 3003: 0.153375319073, 3004: -0.43674352204, 3005: 4.37239894729, 3006: 6.1689748328, 3007: 0.420976971127, 3008: 1.70192806457, 3009: -25.1744967798, 3010: 2.80040743243, 3011: 20.8599263973, 3012: 7.89537914481, 3013: -1.10194999838, 3014: 18.0536770175, 3015: -0.834608439047, 3016: -0.241823194026, 3017: 1.51265946518, 3018: -0.00323178594989, 3019: -18.8958007979, 3020: 0.413898297697, 3021: -5.25581101158, 3022: -7.58146887246, 3023: 1.88449811775, 3024: 0.502688545485, 3025: 11.4874834392, 3026: -19.9440118807, 3027: 1.28149198557, 3028: -4.43430116065, 3029: 2.31034287179, 3030: 17.383806701, 3031: 0.531218488432, 3032: 0.849656745756, 3033: -19.2499216323, 3034: -1.36944386976, 3035: 1.07309795764, 3036: -4.30528687223, 3037: 8.00446472282, 3038: 21.4054023471, 3039: 7.64022691307, 3040: -6.08661630274, 3041: 54.2152128086, 3042: -5.9651235115, 3043: 7.57525532964, 3044: -18.4421973102, 3045: -16.3234911757, 3046: 7.52311603229, 3047: -0.527106183477, 3048: 0.908822219675, 3049: 22.8370039965, 3050: 0.745863561989, 3051: 0.691510750496, 3052: 0.50883537347, 3053: -0.527842825257, 3054: 0.361972852123, 3055: -4.64677710376, 3056: -0.106219447604, 3057: -0.224385853233, 3058: 4.37929749323, 3059: 11.6138004258, 3060: -24.5367369734, 3061: 1.12934790014, 3062: 13.8413573554, 3063: 4.16332786362, 3064: 9.76936084363, 3065: -2.07823753368, 3066: -41.4026194668, 3067: -13.9500807289, 3068: 7.80512441976, 3069: -7.06531254167, 3070: -6.50152822356, 3071: -1.09352259683, 3072: 1.77444988361, 3073: 9.41625386463, 3074: -7.19529952329, 3075: 5.68821111043, 3076: -47.125385468, 3077: 9.95331417763, 3078: -30.0252049839, 3079: 1.00258303351, 3080: 4.54526042442, 3081: 2.44800208878, 3082: -4.44921414359, 3083: 6.50805543963, 3084: -13.1952835009, 3085: -12.9970720199, 3086: -16.7844220702, 3087: -27.4118012851, 3088: 6.11243661144, 3089: -13.6483305199, 3090: -4.09923933591, 3091: -0.551572704916, 3092: 6.32547455004, 3093: 10.2853174902, 3094: -13.7626041999, 3095: -1.42209177789, 3096: 10.1782028767, 3097: 14.9732337892, 3098: 1.06387892742, 3099: -21.5032849608, 3100: 0.552123106353, 3101: 0.722466545131, 3102: 1.2677342997, 3103: -39.7987540036, 3104: -0.428551853541, 3105: -1.45906959175, 3106: 13.4860916771, 3107: -1.40152300412, 3108: -16.7612619656, 3109: 0.681827863914, 3110: -6.92118450188, 3111: 7.95411049561, 3112: -0.3425848094, 3113: -1.91073880996, 3114: 3.0765391504, 3115: -4.7976692511, 3116: 3.99344999033, 3117: -20.1789359418, 3118: -6.61374676289, 3119: 3.7797252089, 3120: -22.2384174201, 3121: -26.3444319446, 3122: 0.000422809233148, 3123: -16.7876624636, 3124: 3.54113731206, 3125: -0.651901990263, 3126: -10.9237039877, 3127: -13.7511780856, 3128: -24.7604675331, 3129: -16.7536953929, 3130: 1.63216344009, 3131: 0.497692096414, 3132: -0.552476638972, 3133: 3.43786194071, 3134: -1.43528967346, 3135: 5.92244004853, 3136: 10.6433475734, 3137: -14.7322616177, 3138: -32.2113792279, 3139: 13.2948298314, 3140: -10.3440617897, 3141: -11.470063699, 3142: -0.233290869566, 3143: -0.532176301513, 3144: -28.9346081235, 3145: 9.76868104643, 3146: -53.8035664678, 3147: -5.27368228788, 3148: 12.0821595949, 3149: 7.28756731782, 3150: -10.7334182729, 3151: -21.1772326573, 3152: 3.17830868447, 3153: 14.480535208, 3154: 59.1584841625, 3155: -24.2980738971, 3156: 6.34700760992, 3157: 25.9445943617, 3158: 28.2384805254, 3159: 20.4046401856, 3160: 15.9144874324, 3161: -0.617514481596, 3162: -21.0350158633, 3163: 33.5888865074, 3164: 6.86272692237, 3165: -17.7294424385, 3166: 18.1593652856, 3167: 12.9518085854, 3168: -8.19483351219, 3169: 26.37364308, 3170: 32.1494160781, 3171: -19.8754828831, 3172: -22.80383016, 3173: 16.5421235881, 3174: 7.09684457922, 3175: -1.15881084812, 3176: 11.4145146293, 3177: -9.41826234141, 3178: -12.6020876544, 3179: -9.8044057989, 3180: -9.42518835103, 3181: -12.17155695, 3182: 6.5769719759, 3183: -2.55336767778, 3184: -2.44391287365, 3185: 18.7580743652, 3186: -3.40192659552, 3187: -8.70790535487, 3188: 1.51224634954, 3189: -2.93184746749, 3190: 7.4386342567, 3191: -3.33452601897, 3192: 6.02260767809, 3193: -9.15163946074, 3194: -2.41396929311, 3195: 20.0190033292, 3196: 4.02391718159, 3197: -0.72191644115, 3198: -18.5401910715, 3199: -2.21303462579, 3200: 5.1313737023, 3201: 8.58050017417, 3202: 17.8820539954, 3203: 6.61908041993, 3204: 4.57238822252, 3205: 6.10934322442, 3206: 9.40234856477, 3207: -0.0182930581693, 3208: -2.87471037583, 3209: -2.17326712995, 3210: -9.66047172205, 3211: -2.53021466918, 3212: -2.46010980577, 3213: -6.74420481936, 3214: -2.4437295453, 3215: -42.6028489523, 3216: -0.942189896675, 3217: -2.33580980322, 3218: -0.81537464564, 3219: 1.31081929133, 3220: -1.32249054873, 3221: 3.566729075, 3222: -2.3337144248, 3223: -2.2680906423, 3224: -2.37435674803, 3225: -1.7972923695, 3226: 13.2385079293, 3227: -2.49239218367, 3228: 8.80967171936, 3229: -9.15338444291, 3230: -2.29907836719, 3231: -0.845299416345, 3232: -2.4042312308, 3233: -2.31387966779, 3234: 24.4240167721, 3235: -1.89615141188, 3236: -2.54211189314, 3237: 5.28865840702, 3238: 3.17705981357, 3239: -2.2784739749, 3240: -2.64158456222, 3241: -2.09023636318, 3242: -2.73166089126, 3243: -2.37839190858, 3244: 19.6732432186, 3245: -0.646560373486, 3246: -2.36793418596, 3247: -0.550490431917, 3248: 0.285218496641, 3249: -2.31224334289, 3250: -2.18140628596, 3251: 15.4966493853, 3252: -1.73016497474, 3253: -6.67569645048, 3254: 8.93330896058, 3255: 26.3470504768, 3256: 6.62852941594, 3257: 10.0527306218, 3258: -2.83765998761, 3259: -0.190883201917, 3260: 5.13978578686, 3261: 2.59363705966, 3262: -50.3987926375, 3263: -4.02396655418, 3264: -7.57731702834, 3265: -9.14147607503, 3266: 2.78045210871, 3267: -0.369867482835, 3268: -0.397449453011, 3269: -1.19378186502, 3270: -43.386596838, 3271: -0.25015632822, 3272: -2.57996409976, 3273: -1.8927952405, 3274: -23.6860863911, 3275: -1.43658272346, 3276: -1.21576722748, 3277: -5.23965108666, 3278: -0.977354455159, 3279: 15.2482527291, 3280: -1.65916521503, 3281: 5.9006389244, 3282: 15.1440916414, 3283: -13.5400432637, 3284: -5.70571316663, 3285: -0.146381262306, 3286: -1.44474665552, 3287: -0.379618274529, 3288: -0.41674503257, 3289: -21.8655617842, 3290: -0.342489504684, 3291: -2.44384660625, 3292: 7.3621807687, 3293: -0.544388517144, 3294: -0.304707159217, 3295: -0.652962049517, 3296: -0.252045997355, 3297: -2.49076899995, 3298: 20.0638654883, 3299: -0.509513999708, 3300: -0.593242790069, 3301: -0.506221839236, 3302: 0.198030223486, 3303: -2.29376317844, 3304: 10.586319473, 3305: -5.71049388411, 3306: -1.08164498008, 3307: 15.5548522698, 3308: -0.10518974019, 3309: 1.61037012188, 3310: 3.28860974236, 3311: -2.77679819165, 3312: -1.54569956422, 3313: -0.272169633779, 3314: -5.12501046749, 3315: -0.227433184092, 3316: 13.872488749, 3317: 21.0551905541, 3318: 2.56414309218, 3319: 0.347294940532, 3320: -0.461776412224, 3321: -0.419950350127, 3322: -0.376950457435, 3323: -0.347154796621, 3324: 0.108885270318, 3325: 8.00143171423, 3326: -4.87444586637, 3327: 34.2873066698, 3328: -0.918826298654, 3329: 12.1227368328, 3330: -1.67693410069, 3331: -9.04884317556, 3332: -1.14784810317, 3333: 9.36364890022, 3334: -0.351334021928, 3335: 3.00728307593, 3336: -1.48434717318, 3337: -0.744516203395, 3338: -1.67310723934, 3339: -0.863382418774, 3340: -0.259297232281, 3341: -2.36600889675, 3342: -0.460205426977, 3343: -12.3134105426, 3344: -0.364442218502, 3345: -0.434681875791, 3346: 15.4361777424, 3347: -2.39512295239, 3348: 41.9649692663, 3349: -0.274806487187, 3350: -0.345275609987, 3351: 0.279200867883, 3352: -0.407176469401, 3353: -2.3741201779, 3354: -4.95350083066, 3355: -1.94427435663, 3356: -0.70013385298, 3357: -0.211754312587, 3358: 4.68097516341, 3359: -0.99458271996, 3360: -7.96411173179, 3361: 18.9997399613, 3362: 7.25634187584, 3363: -2.63894730762, 3364: -3.65978743384, 3365: -1.40031926488, 3366: -1.974775673, 3367: -0.542816981741, 3368: -35.4729215815, 3369: -1.6732738323, 3370: -0.619888087517, 3371: -0.896369541905, 3372: -10.0784398657, 3373: -0.42994973227, 3374: 48.5448725739, 3375: 5.50930354409, 3376: -2.34643652669, 3377: -3.09689840337, 3378: -0.869503213053, 3379: 4.08511575649, 3380: -2.47934632388, 3381: -2.47782470985, 3382: 10.0106175005, 3383: 14.6209930342, 3384: 24.5051264517, 3385: 15.4464295086, 3386: -4.38442354054, 3387: 8.38407423207, 3388: -19.7997629871, 3389: -25.2704183564, 3390: 17.2299973076, 3391: 0.0960141673708, 3392: 20.3485602967, 3393: -1.4990657694, 3394: 4.78579880976, 3395: 7.73398686707, 3396: -2.51284784566, 3397: -2.36652436666, 3398: 2.0816854516, 3399: -2.41627888137, 3400: -2.35646262059, 3401: 3.28551936674, 3402: 7.06917931952, 3403: -2.44694910992, 3404: 4.23046706687, 3405: -2.35418866675, 3406: -2.56968343243, 3407: 7.23966093022, 3408: 8.06455558086, 3409: 2.51107138381, 3410: -2.4222715102, 3411: -0.88756981246, 3412: -2.05889644822, 3413: 24.2728198222, 3414: 5.24141780788, 3415: 4.72691569583, 3416: -59.1513738554, 3417: -0.486781261955, 3418: 5.12833052882, 3419: -8.87280202307, 3420: -9.40637987616, 3421: 0.521159356095, 3422: -0.940430208144, 3423: 4.12599865819, 3424: 13.936042705, 3425: 24.6011642429, 3426: -10.1206474877, 3427: 16.6879604485, 3428: -2.39763463226, 3429: -1.70863406352, 3430: -2.33733354874, 3431: -2.69177445105, 3432: 2.73231777393, 3433: -18.2398764487, 3434: 6.18565884576, 3435: -17.6055272937, 3436: -19.7125984127, 3437: -0.582574919966, 3438: 4.66595350669, 3439: 1.84800803254, 3440: -1.812805493, 3441: 2.32906695232, 3442: -1.64476794383, 3443: -10.8548547235, 3444: -1.49622559293, 3445: -4.75084177076, 3446: -8.13643952272, 3447: -2.54048694075, 3448: 0.615470983514, 3449: -2.27764203053, 3450: -2.27916280571, 3451: -2.36945292323, 3452: 14.9740586453, 3453: -2.38614521421, 3454: -2.46448374296, 3455: 15.4932301585, 3456: -2.49364413236, 3457: -6.63889471022, 3458: -2.5022881037, 3459: -2.35864900572, 3460: -4.67674448468, 3461: -2.51947879346, 3462: -2.6106211812, 3463: -17.0924150246, 3464: -0.924962159712, 3465: 26.3615647848, 3466: -3.52654383421, 3467: 14.2600239156, 3468: -6.72877312432, 3469: 9.07215191692, 3470: 23.0142651991, 3471: 11.2429609456, 3472: 1.1103134743, 3473: 12.4089365893, 3474: -0.302943887695, 3475: -2.11518121197, 3476: 0.248283523278, 3477: -18.3352703191, 3478: 5.18897802681, 3479: -19.0163280013, 3480: -2.47500881821, 3481: 10.2458153392, 3482: -2.3298709178, 3483: 2.91908450224, 3484: 18.0928771988, 3485: -2.51117537408, 3486: 5.97766553151, 3487: 2.38095648125, 3488: 11.9685917007, 3489: -4.12136543448, 3490: 5.11264169891, 3491: 3.06053205848, 3492: 1.19602873212, 3493: 2.50062720371, 3494: 0.697918750685, 3495: -1.10196399675, 3496: -1.74102569167, 3497: 1.89613674398, 3498: -0.798380151918, 3499: -1.57967657577, 3500: -0.881207711615, 3501: -2.80190462798}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: -0.0940129054406, 2: -4.91770212305, 3: 1.06958724331, 4: -4.45040554708, 5: 5.39274038451, 6: -2.50658397108, 7: 1.60315239919, 8: 6.35440845165, 9: -0.271464728228, 10: -3.27290031411, 11: -18.3586299736, 12: 6.98459523842, 13: 9.10117661136, 14: -6.70162479624, 15: -4.23176683727, 16: -9.23256887542, 17: -2.59190759469, 18: -3.32794236634, 19: -18.084535113, 20: -1.81759205405, 21: -4.86341562816, 22: -16.1185604794, 23: 4.42182144938, 24: 0.420712388533, 25: 2.731955023, 26: -10.292237036, 27: 1.13105877707, 28: -2.05728914448, 29: 14.4570061828, 30: -2.98293856573, 31: -2.35551677683, 32: -14.6142858385, 33: -1.99189961804, 34: -1.15546375987, 35: 2.21635075321, 36: -5.14046202927, 37: 0.75155871402, 38: 0.635531609944, 39: -4.13716631817, 40: -0.757242009955, 41: 4.51693961102, 42: -10.8436352992, 43: -7.63367728088, 44: 3.07791704118, 45: -12.1658148127, 46: 4.51978216454, 47: -6.83406887909, 48: -5.95147047647, 49: 4.88351870625, 50: -4.80448802289, 51: -10.8385002061, 52: 2.54330695384, 53: -0.0799289746828, 54: -0.213695724706, 55: 5.68680891505, 56: 0.0994436253407, 57: -20.3737629807, 58: -5.34859372501, 59: -4.96482063337, 60: 5.9100696318, 61: 10.6725421929, 62: -2.66993852778, 63: -0.887134542908, 64: -5.77492113254, 65: 9.67732442609, 66: 2.54607326727, 67: 10.5836260802, 68: -5.23323307183, 69: -1.64509737761, 70: -3.02219391301, 71: -0.383604066993, 72: -7.74910398462, 73: -3.09156495651, 74: -5.70238204746, 75: 2.65804362233, 76: -1.94588752452, 77: -0.160997916329, 78: -4.200670972, 79: -12.4569908198, 80: 6.45954328993, 81: -1.4430035331, 82: -18.9363242333, 83: 0.56063280246, 84: -0.603588100175, 85: -2.66546226665, 86: -5.66528914354, 87: 2.57107044831, 88: 5.16438331997, 89: 6.45269680302, 90: 6.71670451786, 91: 6.6261478647, 92: 0.0857739168046, 93: -5.95519889622, 94: 8.51517749111, 95: 1.45124063154, 96: 1.93578076546, 97: -2.54040387083, 98: 0.921410108225, 99: -10.8695955142, 100: -0.0480467654385, 101: -18.8697423955, 102: -6.08666881438, 103: -7.79209364501, 104: -2.56138390581, 105: 0.120424329324, 106: 0.573699911886, 107: 0.44917494511, 108: 0.0770181739757, 109: 3.77665873162, 110: -8.3968351084, 111: -3.05983790249, 112: 1.53226738392, 113: -12.9815664291, 114: 2.79091633157, 115: -15.6522776513, 116: 1.5457311924, 117: -4.56695916448, 118: -6.82520482922, 119: 10.8028145648, 120: 2.0958725045, 121: 1.76797682526, 122: 8.50368304652, 123: -3.00682265944, 124: 2.57070667672, 125: 11.7570451715, 126: 0.0817185415681, 127: 0.155822589107, 128: -1.68009828028, 129: -0.0188676008807, 130: 0.00119074152624, 131: -0.418387490826, 132: 0.334725442998, 133: 4.88844887705, 134: 0.684874803472, 135: 0.138968825666, 136: 2.80942668071, 137: -0.498110460202, 138: 5.34153762484, 139: 1.42778399492, 140: 3.18447645873, 141: 8.08317087238, 142: -9.15085189717, 143: -0.437057134061, 144: 0.954270448, 145: -0.0866272361931, 146: -0.0561457434305, 147: -0.00688719736424, 148: 13.4340610867, 149: -0.0280016098604, 150: 0.44021169116, 151: -6.11880226322, 152: 0.0763359971855, 153: 0.113233027886, 154: 0.081116967962, 155: 0.00323465157589, 156: 0.330497665954, 157: 15.3977125541, 158: -0.609288031423, 159: 0.139805602396, 160: -0.561238593392, 161: 11.4610941611, 162: 3.71043379395, 163: -18.3224182384, 164: 1.0377638148, 165: 3.33041780859, 166: 8.4777095559, 167: 8.35129365332, 168: -2.04928679413, 169: 11.7127950224, 170: 2.53694387282, 171: 13.1706604214, 172: 0.00240256025425, 173: 0.795731888419, 174: 0.0826905107762, 175: -3.81210805423, 176: -0.610935971522, 177: -15.6583942116, 178: 0.432135104345, 179: 0.124289555139, 180: 0.139791833451, 181: -0.243932556419, 182: 0.0442887702563, 183: -4.19365747918, 184: 7.33730145258, 185: -1.35520761166, 186: -15.8737773019, 187: -23.9275822643, 188: 0.825291165863, 189: -1.16506725336, 190: 6.94692173788, 191: -4.27925905727, 192: 3.6472441827, 193: -0.111077930853, 194: 8.18875309088, 195: -0.776561629629, 196: -0.15169484801, 197: 3.14859851361, 198: 3.61424561133, 199: 0.103213394394, 200: 0.678624675692, 201: -0.269645478519, 202: 8.36546683666, 203: -0.085508162228, 204: -0.0613293243665, 205: -0.113472433037, 206: -0.652629995134, 207: 20.2644631095, 208: -0.0298542487475, 209: -0.012895696348, 210: 1.14078654795, 211: -0.144606905961, 212: 0.687685704008, 213: -2.58518708926, 214: -6.29056171019, 215: 0.123264414712, 216: -0.0300192572416, 217: 7.83945886356, 218: -0.569086100415, 219: -1.96480827253, 220: 6.24351333062, 221: 4.41703031023, 222: -1.75571271189, 223: -0.0972865454016, 224: -0.464400375727, 225: 8.37638729987, 226: -0.85721877043, 227: 11.5202490904, 228: 1.42350267464, 229: -2.14306513481, 230: 0.655034842856, 231: 11.2047291097, 232: 0.345196057946, 233: -5.20082338624, 234: 0.533991286518, 235: 0.905316492487, 236: 6.72785938195, 237: -0.658345702646, 238: -2.84549288689, 239: -5.16068424412, 240: -0.526687462004, 241: 4.1757936513, 242: 1.4973790926, 243: -4.11467806847, 244: 3.03067846093, 245: 1.53811733818, 246: 1.53073349685, 247: -5.3682122015, 248: -14.2633000528, 249: 10.649364176, 250: -6.51162543597, 251: 9.79670384252, 252: 1.26253307636, 253: 7.08342033965, 254: 3.477120501, 255: -1.87527226373, 256: -1.29312908112, 257: -7.62568058356, 258: -0.954900997399, 259: 0.359983275991, 260: -7.60242093719, 261: 4.09612727491, 262: -0.182174874995, 263: -6.47178862024, 264: 1.61143349543, 265: 0.431506695953, 266: 15.8743208687, 267: -29.7504900125, 268: -5.43598994894, 269: 1.5042317049, 270: 3.46869632975, 271: 1.67241273871, 272: 5.51062336129, 273: 4.15386430257, 274: -19.9727016509, 275: -0.31265150849, 276: 2.57355629144, 277: -1.74461778634, 278: 3.40784131377, 279: 3.8276013873, 280: -9.14255692183, 281: -1.78151486982, 282: 3.12826954624, 283: -5.17727937846, 284: 6.79562181775, 285: 4.41992128849, 286: 4.02114767259, 287: -0.495186941148, 288: -0.968173103582, 289: 1.86252756723, 290: -7.63747299159, 291: -6.89171428841, 292: -19.8051730803, 293: 1.18024615174, 294: -32.9251509544, 295: 5.30469590368, 296: 10.9315850628, 297: -2.39260569723, 298: 7.07500521075, 299: -1.33799028042, 300: -3.10591624441, 301: -4.72865076643, 302: -9.93725805885, 303: -1.63395949349, 304: 6.69526904569, 305: 7.14209333056, 306: 1.40249906884, 307: -3.54634209437, 308: 0.669548326917, 309: 0.762570570796, 310: 0.713412163835, 311: -0.816555078941, 312: 0.221250373545, 313: 1.04587424379, 314: -4.85235354596, 315: -0.587984967995, 316: -3.6375019154, 317: -0.683252269924, 318: -2.08413390019, 319: -6.02611547134, 320: -0.354912844727, 321: 0.610894941295, 322: -2.39537059906, 323: -9.98979892351, 324: 14.147701377, 325: 4.99144364429, 326: -1.50258786897, 327: -5.2928093874, 328: 27.5463750236, 329: -11.0315607789, 330: 1.11064076829, 331: 7.63052960158, 332: 4.22309094659, 333: -1.88129786485, 334: -14.5231477396, 335: 9.21544322959, 336: -11.9498995798, 337: 7.62595449, 338: 0.289614022021, 339: 0.583888188955, 340: -2.09783957739, 341: -0.700039991214, 342: -7.66758445046, 343: 2.63664599178, 344: 8.58620543422, 345: -5.12876681839, 346: -11.1035351418, 347: 9.34845026741, 348: -11.4299285702, 349: -1.7092874808, 350: 0.457219666886, 351: 45.5572994626, 352: -3.53745784388, 353: -12.6386966406, 354: 4.56000033162, 355: -3.86275775484, 356: -14.8846507158, 357: -14.7724813196, 358: 2.33039904288, 359: -1.22945659464, 360: -9.73034406444, 361: -6.68248707129, 362: 7.28059889338, 363: 6.00974179051, 364: -8.89582563821, 365: 0.395369393247, 366: -4.22261752904, 367: 0.221858169174, 368: 6.29929961865, 369: -6.01769737244, 370: 12.653937997, 371: -0.0922115569974, 372: 16.1006910526, 373: -12.9930511957, 374: 0.357579163216, 375: -4.92149921844, 376: -11.9449710485, 377: 1.57233739982, 378: -2.0420989086, 379: -1.21905666278, 380: 5.70648946774, 381: -10.7463840193, 382: 2.74403683033, 383: 2.26329606447, 384: 3.60914211743, 385: -1.58335147016, 386: 2.78565961716, 387: 11.7939282979, 388: 5.93533612377, 389: -6.18305908135, 390: 3.27288227073, 391: 3.47691655329, 392: 8.54840713325, 393: 9.46333193377, 394: -7.87335146499, 395: -1.89423140128, 396: 1.5145392183, 397: 1.86295619006, 398: -4.14143904997, 399: -3.19736551153, 400: -6.90644237001, 401: 14.6323730087, 402: 3.67077190067, 403: -3.26546004048, 404: 5.15919221426, 405: 3.68111539642, 406: 8.79256578632, 407: 0.979122933411, 408: 1.25328647087, 409: -7.29244039764, 410: 38.9534309614, 411: 8.20083454812, 412: -4.94968899715, 413: 6.02374994696, 414: -1.62188723397, 415: -1.05853293575, 416: 9.33225041049, 417: 2.95311387021, 418: -5.81321023103, 419: 1.59005569168, 420: 1.24667603229, 421: 2.07150625608, 422: 3.04223973443, 423: -5.73750651711, 424: 1.31887093395, 425: 4.59565418237, 426: 2.48272347305, 427: 2.20695328337, 428: 9.8849448196, 429: 5.40328879112, 430: -2.20575400119, 431: -5.48350861604, 432: 2.72474426347, 433: 0.860236935128, 434: -4.19329646902, 435: -2.99028475502, 436: -8.16139455201, 437: 8.68682200488, 438: 3.62407016171, 439: 4.34840590291, 440: -11.6289502122, 441: 3.63989310444, 442: 18.1161342316, 443: 3.18123927178, 444: 2.79749574162, 445: 2.06606113506, 446: -5.75128197519, 447: 1.8562404048, 448: 5.34205724471, 449: 4.5659292067, 450: -8.21719013905, 451: 1.02163934105, 452: -13.2285715676, 453: 1.05769139454, 454: 5.51237312035, 455: 0.740552760176, 456: 3.07788421137, 457: 3.66483845844, 458: 1.38647617317, 459: -13.838858334, 460: -4.91999324618, 461: -2.09845783672, 462: -1.08142496434, 463: 1.39750630996, 464: -21.5405739835, 465: -2.86171600276, 466: -2.53873632418, 467: -13.8006784396, 468: 2.38010275862, 469: -2.45617169586, 470: -5.46126689523, 471: 8.48657798953, 472: 16.758157818, 473: -8.12898616337, 474: 10.1923209937, 475: 0.745475773469, 476: 3.46734775313, 477: -5.87803474249, 478: 6.55755616198, 479: 0.575614818837, 480: -1.25167661258, 481: -1.28000066086, 482: 2.85187178133, 483: 0.337853696273, 484: 4.51458090323, 485: 4.86964147954, 486: 0.554197155369, 487: 18.6787174415, 488: -0.299694420796, 489: 3.68405525916, 490: -4.9391445732, 491: -11.0209495649, 492: 4.48449072274, 493: -1.81404530653, 494: -1.52439845114, 495: 0.694817150259, 496: 0.542811564224, 497: 11.6580363835, 498: 0.425541442048, 499: 2.66366414947, 500: 2.77970847218, 501: 0.617140556003, 502: 0.481509273512, 503: 0.446035763229, 504: 0.886690328336, 505: 0.554019443946, 506: 8.25123201904, 507: -0.553654365444, 508: 0.261475096183, 509: 4.65975140841, 510: 4.0726421898, 511: 3.61548526213, 512: 18.1443358429, 513: 2.27890381038, 514: -10.1207588651, 515: -0.763536132171, 516: -8.59752931401, 517: 9.40756660453, 518: -1.71929459532, 519: -1.10641603603, 520: -13.5118872759, 521: 0.485954482668, 522: 15.5274690297, 523: 0.539186693618, 524: -1.54333151759, 525: 5.74521987212, 526: 12.9722631456, 527: 1.12136358618, 528: 0.422770661906, 529: 0.51987261037, 530: 1.74408586187, 531: 0.255139474959, 532: 6.22327040386, 533: -0.229564968451, 534: -5.25821733983, 535: -30.2519370192, 536: -6.2303334757, 537: 2.59301024137, 538: -12.6241624639, 539: -6.82194759649, 540: -1.24400005291, 541: 7.17135433949, 542: 0.494533362025, 543: 20.4259979313, 544: -6.9447524697, 545: 2.25740979248, 546: 1.15446537755, 547: 4.45350932299, 548: 0.556540575796, 549: 1.65746281651, 550: 0.526391271165, 551: 22.1819673792, 552: 0.592425276119, 553: 0.646930514685, 554: -0.682711386118, 555: 2.45337004395, 556: -9.23709682521, 557: 0.642912043656, 558: 0.489832914637, 559: 0.315216511272, 560: 0.679508396505, 561: -0.906718111873, 562: -4.44520949986, 563: -0.713476041027, 564: -0.4157498913, 565: -1.15450971086, 566: 2.52780349198, 567: 2.09125137901, 568: -0.612861964869, 569: 1.95238589955, 570: 8.0548916091, 571: 1.20300116472, 572: -5.53982382092, 573: 2.60704291466, 574: -18.0223502449, 575: 0.782263673952, 576: -0.214550581469, 577: 15.5085108209, 578: 2.50538421696, 579: -1.81226545102, 580: 14.7464371385, 581: 0.940476714697, 582: 7.10412312943, 583: -8.22807933678, 584: -0.269596811111, 585: -1.86188196735, 586: -10.9340206892, 587: 6.93050928365, 588: 2.28819927318, 589: -1.39473680043, 590: 1.21945086504, 591: -2.37724064453, 592: -2.29174937969, 593: -3.23629792382, 594: -6.38510689575, 595: 8.44214456019, 596: -10.6730499649, 597: -2.85999489126, 598: -21.8912659987, 599: 1.89035713841, 600: -0.865141446572, 601: -18.8890436016, 602: 2.17209274234, 603: 1.46637445482, 604: 6.13139738874, 605: -0.870729339403, 606: -13.9348249117, 607: 2.9404734065, 608: 2.7346777907, 609: -0.451540450581, 610: -0.443674924689, 611: 3.10028289151, 612: 1.41133986598, 613: -0.179546781215, 614: 0.428655745221, 615: 6.39112175952, 616: -7.02122688434, 617: -5.13129562316, 618: 4.90201222916, 619: 4.77752183021, 620: 2.88921159338, 621: -24.2998393946, 622: -4.62066316626, 623: 30.5408891221, 624: 38.9147682057, 625: 11.2194879746, 626: 4.73872903002, 627: 2.03279072311, 628: -0.144217193721, 629: -8.26211707127, 630: 9.81215026004, 631: -7.47258140612, 632: 12.8811807123, 633: 17.6327088243, 634: -12.6411005298, 635: -6.91854198242, 636: 3.10673093621, 637: 0.0613822799937, 638: -9.44526870364, 639: 0.0309605423313, 640: -4.80974332526, 641: -5.08803385651, 642: -6.2944725889, 643: 1.36883965759, 644: 24.9953195451, 645: -27.7325112964, 646: -8.15585749423, 647: -3.68381772028, 648: -2.5674168778, 649: -2.5561599929, 650: 8.77364566382, 651: -19.8636317206, 652: -2.69004542407, 653: -12.3533171052, 654: 5.01881777058, 655: 2.77860350007, 656: -16.1564796463, 657: 1.55660934589, 658: 1.53459035915, 659: 0.0148960751138, 660: -10.9829206437, 661: 2.77222843617, 662: 3.75252735885, 663: -0.873034641929, 664: 2.41969067893, 665: 2.22867387191, 666: 3.3364135794, 667: 3.3674977204, 668: -4.70631416379, 669: -4.56911727804, 670: -0.7696597167, 671: 3.77900685684, 672: -6.48076252904, 673: -3.06489947623, 674: -3.62176717516, 675: -23.8768496689, 676: 0.749109651245, 677: -1.11513865365, 678: -6.37505448352, 679: -5.65735674826, 680: 3.78125271085, 681: 0.0847878276111, 682: 2.65664481663, 683: 5.6942299104, 684: -1.81742887338, 685: -15.7402689583, 686: 9.3133024621, 687: 2.2411506153, 688: 3.02726161024, 689: -0.459275719864, 690: -4.26502635292, 691: -0.142290090657, 692: -6.17799963195, 693: 15.0942237835, 694: -10.1242556759, 695: 1.89952974131, 696: 11.1010760366, 697: 10.5685315158, 698: -49.6759642034, 699: -1.02007395619, 700: 14.2966681096, 701: 18.3791846012, 702: 10.029589481, 703: 22.0804701451, 704: 8.96862548893, 705: 6.53336189885, 706: 17.4628352635, 707: 8.28634436084, 708: -19.94579689, 709: 21.8577772354, 710: -3.05183595012, 711: -1.65051101632, 712: -1.67753014669, 713: -4.80905782244, 714: 17.0865949846, 715: 5.71375754281, 716: 30.1411131608, 717: 3.87845169103, 718: 8.61767657365, 719: 1.64518815111, 720: -11.8928344195, 721: -20.7979057268, 722: -0.550643641261, 723: -0.280758210239, 724: 5.08401056622, 725: -15.3463100506, 726: 21.7463370484, 727: -2.57253277736, 728: -13.8789876819, 729: -12.8590861877, 730: 18.13442674, 731: -3.61404470994, 732: 8.98159187852, 733: 13.1780658421, 734: -5.03399763047, 735: -15.2826103139, 736: 8.45665995309, 737: -12.6328332075, 738: 18.7480487303, 739: -30.7365178269, 740: -8.08398272166, 741: -3.92977003059, 742: 49.0867934343, 743: 9.56223931573, 744: 13.9855223494, 745: -1.02489845805, 746: 4.21379601053, 747: -3.47755363798, 748: -5.21425797886, 749: 3.23173058535, 750: 16.8971395071, 751: -7.15561835545, 752: 6.21245930314, 753: -7.3094347004, 754: -3.16513791569, 755: 5.86867562391, 756: -7.10752013353, 757: -0.934471469896, 758: -17.623385858, 759: -12.670166404, 760: 16.1338576363, 761: -5.68198413362, 762: 13.9442336302, 763: -5.77386814874, 764: 0.503376182133, 765: -1.74979026298, 766: -7.39774032952, 767: -1.18425305877, 768: -6.85078084515, 769: -6.87786867324, 770: 7.99928400972, 771: -7.19631102696, 772: 28.644600966, 773: 10.4253288753, 774: -5.46327767856, 775: -4.39420046634, 776: -6.60893133268, 777: -18.3667013486, 778: -8.10435134191, 779: 0.159433487707, 780: 4.20318639663, 781: -7.22644243901, 782: -7.91164682115, 783: 1.4609903178, 784: -2.72175630299, 785: -19.3498871734, 786: 3.77116895662, 787: -2.20590486499, 788: 9.61617951034, 789: 1.01725266425, 790: -7.35430739576, 791: -5.38860008551, 792: -3.9898009005, 793: -2.51119346541, 794: -8.16184463497, 795: -1.71751868998, 796: -6.47954896948, 797: -4.17542253941, 798: -1.21008876001, 799: -1.45958094461, 800: -6.76635037972, 801: 10.0979525308, 802: -1.09215588492, 803: -9.36501777485, 804: 7.66851260278, 805: 0.234384165317, 806: -1.66696176864, 807: 11.6604331107, 808: 7.07114311769, 809: 3.88802703847, 810: 1.82281051095, 811: 51.9521243844, 812: 17.5227520165, 813: -28.1074389001, 814: -8.79751292327, 815: -10.9189935036, 816: 7.02257737658, 817: 26.2803854416, 818: -19.2426467988, 819: -0.750041605312, 820: 50.1629352445, 821: 19.4924000292, 822: 17.4292231876, 823: 2.10200518834, 824: -1.29331043318, 825: -1.70964729847, 826: 11.6058485534, 827: -2.20817777829, 828: -1.43790207432, 829: 3.84403028858, 830: -2.48012623441, 831: -24.2858269586, 832: 0.18412794353, 833: 0.609968646328, 834: -5.07964496483, 835: -1.3899313248, 836: -11.216215727, 837: -2.25902878107, 838: 4.61575259313, 839: -28.2821886134, 840: -11.3203060135, 841: 10.9081462687, 842: -2.07008269637, 843: 2.61966520943, 844: -1.61770736102, 845: -1.37112673285, 846: -8.06732089808, 847: -1.42270677924, 848: -7.05890122508, 849: -5.49050110274, 850: -1.43503485139, 851: -1.29959905503, 852: -1.62307220134, 853: -1.40860593436, 854: -1.64579962778, 855: -4.00454220581, 856: -1.61692297493, 857: -2.56503778409, 858: -2.01870016875, 859: 0.54786724998, 860: -3.01512433371, 861: 10.1137758835, 862: -1.8204136154, 863: -7.55764841459, 864: -20.7877352724, 865: 3.17847746154, 866: -0.253725118077, 867: -3.04247384054, 868: -0.846966051618, 869: -2.21356453663, 870: -1.28185151779, 871: -3.14955979871, 872: -1.38630287861, 873: 12.2410749311, 874: 63.3000156775, 875: 26.4302696259, 876: -4.77766507388, 877: -1.18916782287, 878: -1.72882688487, 879: -3.00129602065, 880: -1.63868036015, 881: 7.41012929838, 882: 4.50373359644, 883: 8.74442989216, 884: 11.8366747453, 885: -29.4087793724, 886: 32.3386601316, 887: -6.38979239684, 888: -27.2635451219, 889: -3.00232562046, 890: 2.68021706563, 891: -1.40593923483, 892: -6.12112275212, 893: -2.89338698771, 894: 1.27049041762, 895: 3.24725353662, 896: 5.85347812322, 897: -1.37546485566, 898: -6.87937514592, 899: -1.43974859407, 900: -20.7372627713, 901: -1.37505550441, 902: -1.33987692887, 903: -7.43425505353, 904: -6.13491833894, 905: -14.9157012666, 906: -1.32830738008, 907: -1.4031763145, 908: -0.569605052262, 909: -1.43014288, 910: -4.90466496796, 911: 2.1225587638, 912: 51.9280804969, 913: 1.32078677821, 914: -2.15718705911, 915: -15.5511321664, 916: 4.26890176004, 917: 13.5808682014, 918: -4.59759152763, 919: -0.4446474963, 920: 17.5922397219, 921: -9.50134953214, 922: -4.17779954345, 923: -3.97325740983, 924: -1.60540667772, 925: -17.9190150656, 926: -15.6719134306, 927: -5.24622372522, 928: 0.15068451578, 929: -1.35114987375, 930: -1.85821479386, 931: -14.7222921776, 932: 8.61236951897, 933: -2.37888004802, 934: -10.2578810668, 935: -13.1784020292, 936: -2.82626034005, 937: 1.64735764666, 938: 3.83186861903, 939: 1.58380598161, 940: 25.1290695035, 941: -14.219788613, 942: 5.14877226552, 943: 9.542744707, 944: 4.09121359466, 945: 18.987579867, 946: 29.9090961373, 947: -18.847115488, 948: 52.7215836631, 949: 2.78746892729, 950: 1.42912181689, 951: 2.37940182604, 952: 19.6224535576, 953: -0.505487527152, 954: -3.8868200336, 955: 12.5744851634, 956: -6.54468293786, 957: -6.98288135558, 958: -5.99235820305, 959: -2.31507214854, 960: -7.22448381726, 961: -5.58849223134, 962: -6.71345985824, 963: -1.47327935835, 964: -5.37702361233, 965: 3.59403202226, 966: 20.8065436097, 967: 5.6586270771, 968: 1.53570106588, 969: -2.07542553248, 970: -44.1086317322, 971: -16.1471084608, 972: 13.429080587, 973: 10.7649668441, 974: 17.2489411297, 975: 10.6836110459, 976: 4.72623423135, 977: -0.753411986601, 978: -5.00341532611, 979: -8.61600811176, 980: -11.1949098017, 981: 24.5158326553, 982: -23.5494924217, 983: 29.8127849579, 984: -28.6301093366, 985: -7.63069307252, 986: -10.9992179775, 987: 10.0862962297, 988: -1.47685258067, 989: 20.8688684546, 990: 11.0780235109, 991: 11.8172906373, 992: 17.7809529224, 993: 59.2224532535, 994: 16.4053837671, 995: 37.5076444883, 996: -10.370190712, 997: -6.02124963211, 998: 25.773909382, 999: -4.48244563858, 1000: 24.0775170979, 1001: 2.40209652703, 1002: -10.1399142057, 1003: 11.8477761679, 1004: -2.33701013513, 1005: -1.05040661413, 1006: -6.69463077977, 1007: -6.74184345852, 1008: -4.63184952228, 1009: 15.3840721061, 1010: -6.83228185826, 1011: -7.26825188632, 1012: 5.28039169298, 1013: -6.11129633015, 1014: -19.2927644013, 1015: -7.99176822453, 1016: -7.41255157356, 1017: 19.5792298022, 1018: -7.33703638424, 1019: -3.96961540949, 1020: 19.1711595249, 1021: -27.7598772378, 1022: -18.541480439, 1023: -19.0746694815, 1024: 14.1688210785, 1025: 15.4350148203, 1026: -47.237882298, 1027: 14.2911688702, 1028: -5.79561159018, 1029: 8.54424815437, 1030: 8.26923940395, 1031: -0.950088434881, 1032: 4.02936992162, 1033: 10.5611364725, 1034: -18.537299254, 1035: -15.5941657962, 1036: 13.032024037, 1037: -6.72236676744, 1038: -1.76191219501, 1039: -5.6984188036, 1040: 12.7514968788, 1041: 8.25403352676, 1042: -2.94557955263, 1043: 1.57068358081, 1044: 25.3319169446, 1045: -25.0927988185, 1046: -6.79909174139, 1047: -25.6788680683, 1048: 2.59288338394, 1049: 10.477018603, 1050: 21.2213832153, 1051: -8.55718611407, 1052: 46.3582773945, 1053: -29.8423041448, 1054: 16.4084043888, 1055: -60.3797648262, 1056: -49.9823152141, 1057: 23.3423242814, 1058: -29.706176846, 1059: 72.5918005719, 1060: 57.1371476293, 1061: 11.2627156444, 1062: 37.3626021894, 1063: 6.98453207959, 1064: 20.36709533, 1065: 60.1225661953, 1066: -85.112136214, 1067: -16.4163251906, 1068: 43.9840193304, 1069: -75.337160212, 1070: -100.0, 1071: -49.9040054325, 1072: -74.6954051959, 1073: -48.3545777124, 1074: 5.59457927107, 1075: 4.14875800253, 1076: 42.7832007374, 1077: -50.7067068878, 1078: 50.9955745657, 1079: -21.5581022527, 1080: -44.6151455562, 1081: -56.4279542951, 1082: 2.10228488564, 1083: -75.7385362678, 1084: -3.97015351809, 1085: 6.97370803031, 1086: 3.15655414947, 1087: -96.2280645411, 1088: -2.87375695788, 1089: 17.0725492785, 1090: 9.0282832146, 1091: -41.7164636703, 1092: 21.3454656763, 1093: 57.1983800744, 1094: 7.05142885591, 1095: 30.0127279139, 1096: -58.2750649216, 1097: -34.5271939296, 1098: -8.35993723661, 1099: -50.1666813783, 1100: 14.0917602479, 1101: 25.1521684176, 1102: 1.16211937165, 1103: 11.0478644704, 1104: -57.1235925517, 1105: 19.2812951987, 1106: -16.4354278205, 1107: 32.0456451271, 1108: -11.0113432226, 1109: -17.0348110294, 1110: -34.1131437695, 1111: -29.3181987655, 1112: 100.0, 1113: 0.0685502227246, 1114: 12.715551464, 1115: 17.5405595266, 1116: 40.5937135835, 1117: 16.302657479, 1118: 16.2883106594, 1119: -48.5631854616, 1120: 16.1866890153, 1121: -14.3480946753, 1122: 21.5543503512, 1123: 27.9209080029, 1124: 6.81642086635, 1125: 21.6820591671, 1126: 100.0, 1127: 66.1868923783, 1128: 14.2311723763, 1129: -89.2904209078, 1130: 16.4335795782, 1131: 13.3614936106, 1132: -9.56985718692, 1133: -12.5569658975, 1134: -40.4665272692, 1135: -48.1494794736, 1136: 7.23007480116, 1137: -28.2292604872, 1138: 69.8129216663, 1139: 11.7216260362, 1140: -26.2769612831, 1141: 16.674257844, 1142: 14.3855521623, 1143: 14.2536319231, 1144: 3.22540756735, 1145: 15.8427897315, 1146: 17.038897848, 1147: -8.67945638449, 1148: -71.6804957508, 1149: 14.5545854953, 1150: 11.8629067679, 1151: 9.8435343591, 1152: -0.373080331724, 1153: -4.96135296384, 1154: 14.7813510924, 1155: 17.4140688987, 1156: 4.34141709178, 1157: -41.0455760186, 1158: -11.0016217687, 1159: 49.1376246626, 1160: 46.105406872, 1161: 39.2794937839, 1162: -100.0, 1163: 35.3537848833, 1164: -18.8200829779, 1165: -5.57117368704, 1166: -81.8905953163, 1167: 33.3206360567, 1168: 36.3557055847, 1169: 13.9509490996, 1170: 20.252540112, 1171: -68.2092243547, 1172: -15.3840904661, 1173: 2.96044580851, 1174: 3.38028841096, 1175: 7.02654663517, 1176: 12.1022615449, 1177: 4.41730128229, 1178: 17.2545534252, 1179: 4.59562488318, 1180: -29.7861140006, 1181: 4.26337807953, 1182: 1.2370420193, 1183: -35.7617313384, 1184: -0.076278661451, 1185: -44.5629731224, 1186: 2.41352262572, 1187: -6.95819383661, 1188: 11.2182422858, 1189: 16.9682265602, 1190: 27.8057754485, 1191: 3.73442480045, 1192: 11.3094730446, 1193: 3.74984172102, 1194: 4.02868783976, 1195: 22.5752740488, 1196: 2.97385478732, 1197: 16.8868481476, 1198: 3.87826793771, 1199: 3.62958376813, 1200: 2.82292473526, 1201: 3.385340585, 1202: 2.99794454043, 1203: 12.6829002974, 1204: -15.5967252545, 1205: 5.21992752205, 1206: 5.06212718479, 1207: 3.14803634201, 1208: 5.75078433608, 1209: -6.20104896436, 1210: -39.0234273804, 1211: 4.89073004025, 1212: -92.314440992, 1213: 18.6486550466, 1214: -4.39742865214, 1215: 6.65105023803, 1216: 0.899807517193, 1217: 57.8674539449, 1218: -76.7673530958, 1219: 3.49509203709, 1220: 14.4634099571, 1221: 3.08551687159, 1222: -14.584714685, 1223: -91.2530336479, 1224: -33.4087515649, 1225: 4.20869691672, 1226: 2.69298820024, 1227: 3.42339366812, 1228: 3.6466607286, 1229: 3.14191794262, 1230: 14.2647586753, 1231: 11.4322835802, 1232: 2.43599612378, 1233: -0.0563904776466, 1234: 54.6464755677, 1235: -2.61512725985, 1236: -23.8869758968, 1237: 6.34101641603, 1238: 9.7913576058, 1239: 80.0803064611, 1240: 3.29869843145, 1241: 10.5731798653, 1242: 5.22595440964, 1243: 4.46240198274, 1244: -7.7559648409, 1245: 50.932343766, 1246: 2.97864599229, 1247: 16.6763900567, 1248: 3.34159313789, 1249: 82.6884416736, 1250: 2.95503342173, 1251: 3.28356063227, 1252: 41.6291253071, 1253: 15.6468708465, 1254: 24.4527563441, 1255: 3.00571409188, 1256: 3.83075922898, 1257: -0.786236385933, 1258: 3.01876644469, 1259: 13.3509913216, 1260: -30.6379186159, 1261: 15.2867616483, 1262: 3.12076044452, 1263: 0.955379343234, 1264: 18.7120921634, 1265: 4.02078510276, 1266: -3.62397489708, 1267: -17.2463138514, 1268: 8.8087554092, 1269: 2.96358337631, 1270: -7.29559249967, 1271: -26.8183044944, 1272: -4.96405114758, 1273: 4.2731837745, 1274: 12.2962725874, 1275: 14.8354401891, 1276: 3.08738214711, 1277: 1.38545557154, 1278: -56.1611285226, 1279: 3.24998443273, 1280: -33.4303244021, 1281: -22.7818029427, 1282: 21.9348068715, 1283: 21.7207240322, 1284: -84.1199161261, 1285: 11.7212464142, 1286: -7.06319850526, 1287: 17.1903762527, 1288: -57.8464757402, 1289: 85.4957785545, 1290: 1.1081861633, 1291: -33.3937869523, 1292: 78.0556115735, 1293: 68.1041968844, 1294: -0.646387318857, 1295: 17.1130880485, 1296: 75.6553517447, 1297: -54.4124475818, 1298: 54.7324261575, 1299: 6.61244507375, 1300: -37.3413488425, 1301: 42.4078503781, 1302: -2.55356976085, 1303: -41.9408322408, 1304: -27.5949688301, 1305: 17.6935964531, 1306: 16.798434828, 1307: 11.7283884157, 1308: -30.1726538289, 1309: 17.6550315748, 1310: 15.5174957838, 1311: 15.5810985943, 1312: 12.8046942528, 1313: -4.48720041659, 1314: -41.4590044454, 1315: -8.61290605399, 1316: 35.8051366093, 1317: 3.63368097518, 1318: 0.635627239203, 1319: 20.7892274755, 1320: -72.937683282, 1321: -37.4198696913, 1322: -65.0126896647, 1323: 49.2632664465, 1324: -15.1741521225, 1325: -54.0223098927, 1326: 89.6548882883, 1327: -64.5226283858, 1328: -8.7288320115, 1329: 72.1713462668, 1330: -16.3354533218, 1331: -100.0, 1332: -7.21848380507, 1333: 39.2763826266, 1334: 28.9712327015, 1335: -13.4424452875, 1336: -0.377502024424, 1337: 18.3205647502, 1338: -5.7518980045, 1339: -16.7768855382, 1340: -17.0861122826, 1341: -47.1072789858, 1342: -80.9845137099, 1343: -50.3546931063, 1344: -13.0090473896, 1345: 10.7665475182, 1346: 8.86732510927, 1347: 0.43649958556, 1348: 1.95881215111, 1349: -21.0840177035, 1350: -16.9336330281, 1351: -35.6839240625, 1352: 11.2911697559, 1353: 14.300145363, 1354: -35.8347939464, 1355: 16.5261162974, 1356: 16.5174512938, 1357: 11.3060230282, 1358: -40.8492167729, 1359: 15.9975616597, 1360: 5.30208694266, 1361: 21.1418690674, 1362: 15.8119779721, 1363: 53.8148747467, 1364: 11.1491226729, 1365: -14.9753374638, 1366: -42.8944172119, 1367: 13.1937365805, 1368: 14.6242317545, 1369: -43.0423197337, 1370: 24.35412868, 1371: -22.6574541892, 1372: 26.9812415934, 1373: 12.9365265162, 1374: -47.2726376336, 1375: -6.48092769358, 1376: -9.67059837184, 1377: -67.3966090591, 1378: -20.8272841735, 1379: -7.47015300106, 1380: -53.7507268745, 1381: -12.2083408804, 1382: 0.0270791025746, 1383: 12.9406711014, 1384: -35.1425622387, 1385: 70.9331362628, 1386: 16.8596301934, 1387: 51.8542323646, 1388: -32.8224359699, 1389: 53.5110713575, 1390: -99.9007754403, 1391: -62.8102680298, 1392: 14.362238465, 1393: -2.29698663315, 1394: 20.9690042657, 1395: 35.0159179295, 1396: -22.1356342659, 1397: 0.438511720336, 1398: 17.9978392006, 1399: -33.3738761095, 1400: -51.5168738398, 1401: -14.7727382989, 1402: 45.6143052002, 1403: -27.7096788611, 1404: -8.17039103456, 1405: -46.4979515823, 1406: -6.33283703946, 1407: -36.4770840951, 1408: -2.31619811041, 1409: -42.1882557152, 1410: -4.454001125, 1411: -30.6570495178, 1412: 8.09876338214, 1413: -39.1975915119, 1414: 17.9039632451, 1415: -8.22543541295, 1416: -39.9947361129, 1417: 14.6956192135, 1418: 3.08244187827, 1419: 39.0357086259, 1420: -18.3684364451, 1421: -6.07833105614, 1422: -2.82655227085, 1423: 8.53640066994, 1424: -1.31233532649, 1425: 25.5115005961, 1426: -8.91327668423, 1427: -8.80930045267, 1428: -17.308373016, 1429: 3.22234428537, 1430: 2.48181376896, 1431: -17.1234635926, 1432: -21.9597065052, 1433: 3.32787003759, 1434: -4.26153236387, 1435: 22.4859216921, 1436: -0.230486892785, 1437: 9.88667374335, 1438: -5.34162936379, 1439: -39.6912624817, 1440: -19.3884490334, 1441: 0.897167649104, 1442: 4.19996315132, 1443: -30.6886692592, 1444: 8.49883419024, 1445: -17.6045804109, 1446: 0.25715939466, 1447: -11.5794904943, 1448: -23.1251395218, 1449: 1.91624285728, 1450: -22.7194131134, 1451: 3.28331172916, 1452: 2.54319556555, 1453: 36.8580076517, 1454: -9.29645956752, 1455: 35.9784800471, 1456: 3.06321625273, 1457: -43.3607021644, 1458: -35.3085061474, 1459: 18.269333338, 1460: 66.7835969292, 1461: -7.97998923519, 1462: 4.09222178804, 1463: 22.1519693637, 1464: 2.190383776, 1465: -17.7244663843, 1466: 4.08194622018, 1467: 3.37916035478, 1468: 16.2722946472, 1469: 1.21601417149, 1470: 12.6420437551, 1471: 7.03950468148, 1472: -1.46079852208, 1473: 1.16738846341, 1474: 6.98798042142, 1475: 51.0305385096, 1476: 24.9840831444, 1477: 1.65253770371, 1478: -75.6812609066, 1479: 1.83328243583, 1480: -1.16864726128, 1481: -23.7482370071, 1482: 4.2467733102, 1483: -46.2675036959, 1484: 23.316613557, 1485: 1.77713117138, 1486: 2.65196463851, 1487: 1.8541922164, 1488: 2.16176111852, 1489: -7.86119084786, 1490: 2.28293668553, 1491: 2.79228743018, 1492: 2.43608864482, 1493: 4.31870810487, 1494: 2.88644064621, 1495: -4.24206730451, 1496: 4.32489264136, 1497: -19.9079807389, 1498: 3.78882991761, 1499: -16.6861979041, 1500: 3.31696215092, 1501: -15.3821554884, 1502: -55.0615826736, 1503: 1.91450326556, 1504: -0.376402713056, 1505: 2.2475329184, 1506: -17.5955133833, 1507: 1.14476791043, 1508: -8.45367752676, 1509: 7.92933920739, 1510: -9.67204658152, 1511: 68.4934904278, 1512: 5.40908668734, 1513: 5.60075877853, 1514: 10.1393224402, 1515: 80.0067616375, 1516: 19.4905843834, 1517: 0.636603862763, 1518: -48.5343653518, 1519: -38.7526569409, 1520: -17.6928181537, 1521: 67.8309973244, 1522: 1.47147785822, 1523: 1.96036908285, 1524: -10.6484404983, 1525: 53.4545699916, 1526: 0.480776976697, 1527: 1.85782893087, 1528: 1.3863962582, 1529: 2.364778518, 1530: 4.4463005744, 1531: -2.46907652499, 1532: -32.2246065457, 1533: -1.45129577441, 1534: 82.7163785786, 1535: 0.140719216023, 1536: -2.53229684843, 1537: -1.94892316571, 1538: 43.7933952023, 1539: 0.467892570922, 1540: 4.33047938158, 1541: 12.1663831844, 1542: 0.319697016867, 1543: 0.357937107842, 1544: -9.90947556077, 1545: 0.517129965648, 1546: 2.4120015099, 1547: 0.777665020208, 1548: 0.388922119346, 1549: 0.340851063098, 1550: 0.551297479841, 1551: 0.524830460267, 1552: -6.10392327278, 1553: 14.2153943337, 1554: 1.09430335517, 1555: 1.32645144721, 1556: 0.445081029287, 1557: 10.9196699802, 1558: 21.0484448002, 1559: 14.5914013333, 1560: 7.8992064381, 1561: 83.6340006764, 1562: 32.1174264168, 1563: -5.67992165824, 1564: 3.97100616295, 1565: 3.89318865394, 1566: 6.96369587285, 1567: 15.8594780145, 1568: 0.180059021765, 1569: -9.04962068084, 1570: 0.317385690701, 1571: -31.0573036103, 1572: -25.9627339097, 1573: -0.251039252466, 1574: 2.41648525576, 1575: 0.390670028102, 1576: 1.48738248452, 1577: 0.0750792921472, 1578: 0.725715265416, 1579: -3.17246918973, 1580: -9.42820497663, 1581: -12.5497275297, 1582: -38.8674195338, 1583: 57.0017615652, 1584: -16.7553524887, 1585: -2.3945498212, 1586: 0.310250875031, 1587: 0.306200881286, 1588: 18.1809021995, 1589: 0.250096676398, 1590: 49.3685518284, 1591: -3.12612137943, 1592: 7.75247103054, 1593: 7.83318548908, 1594: 0.43987573566, 1595: 0.41061236355, 1596: 3.07070543846, 1597: 0.553747758193, 1598: 30.8721777705, 1599: 0.351489970366, 1600: 0.512226350555, 1601: -41.0681893744, 1602: 2.17468335196, 1603: 23.73669709, 1604: 0.485757483888, 1605: 0.399553601913, 1606: 2.06148928193, 1607: 0.372714560413, 1608: 3.99188326161, 1609: -0.00621300884058, 1610: -2.88552710835, 1611: 0.544603033013, 1612: -1.7087962446, 1613: -34.0809722933, 1614: 9.64038971191, 1615: -31.9221735174, 1616: 37.4235505728, 1617: 10.2268046116, 1618: 11.4893488392, 1619: 10.9229732522, 1620: 3.29750370761, 1621: -6.71809133378, 1622: 0.0302243013469, 1623: -23.7100604819, 1624: 2.59508136952, 1625: 8.43024497489, 1626: -3.85780780353, 1627: -36.4064988627, 1628: 6.2019501712, 1629: 22.3429780015, 1630: -13.5921667719, 1631: -31.8920655407, 1632: 66.7225974997, 1633: -33.05603181, 1634: -14.3166416858, 1635: 1.53468067535, 1636: 1.75100371626, 1637: -42.161302708, 1638: 3.53708462811, 1639: -24.5994910462, 1640: -0.125663287006, 1641: 18.2920741601, 1642: -7.25580912347, 1643: -44.486832935, 1644: -45.3083640461, 1645: 43.5961036441, 1646: 64.6134901516, 1647: -59.3281260044, 1648: -20.5406652537, 1649: -18.4427032483, 1650: -6.6984880884, 1651: 2.36504860075, 1652: 2.94225991059, 1653: -3.61885720875, 1654: 2.08061954849, 1655: 2.31882806433, 1656: 1.51114695716, 1657: 8.05057115893, 1658: 1.89639861806, 1659: 2.65396801504, 1660: -3.29162356285, 1661: -6.10167638759, 1662: -40.5212318211, 1663: 20.3604434951, 1664: -24.7899887758, 1665: -7.9269087246, 1666: 8.07318858633, 1667: 21.4624661164, 1668: -29.1785654419, 1669: -35.4280490144, 1670: -10.5012321964, 1671: -37.5818727924, 1672: 37.1412381443, 1673: 4.36461935247, 1674: 10.4708746511, 1675: 6.68282423266, 1676: -36.5829209513, 1677: 9.02123132511, 1678: -16.3943611681, 1679: -16.7263311542, 1680: 42.4870671307, 1681: -5.33695056207, 1682: -1.18990160737, 1683: 2.38468844403, 1684: -9.5173977207, 1685: 4.48547883715, 1686: -21.5697582283, 1687: 4.39653390445, 1688: 23.6264360062, 1689: -17.515537975, 1690: 33.3860582183, 1691: -12.63687122, 1692: -15.8157681588, 1693: -0.952461393087, 1694: 15.5718290157, 1695: 1.28620400937, 1696: 16.6093281225, 1697: -7.17404412278, 1698: -24.6919802949, 1699: -12.6149449366, 1700: 7.21173058219, 1701: -23.122792961, 1702: 2.73017896675, 1703: 5.60045666691, 1704: 3.13373631415, 1705: 3.07867801952, 1706: 2.53986895528, 1707: -14.3291370492, 1708: 2.130095729, 1709: 2.18542026007, 1710: 55.3867025037, 1711: 2.14163876453, 1712: -9.83763054011, 1713: 1.99807161548, 1714: 2.37391792866, 1715: -24.570140667, 1716: 2.15799135986, 1717: 3.88229666242, 1718: 9.62108958539, 1719: 7.15586656263, 1720: 5.44224467466, 1721: -26.772092471, 1722: -10.5507565548, 1723: 11.8401024775, 1724: 23.0308138789, 1725: 16.0309972201, 1726: 5.12279157374, 1727: -33.6911642656, 1728: 5.41180968475, 1729: 0.985005882683, 1730: -16.3671241764, 1731: 16.824908813, 1732: 13.5955463428, 1733: 17.0754502533, 1734: -32.6983422593, 1735: 2.29462141312, 1736: -1.30355058011, 1737: -8.7423599981, 1738: -41.9793341079, 1739: -17.747679535, 1740: -7.53696114565, 1741: 8.94650968683, 1742: -49.9582047588, 1743: -40.7118940438, 1744: -14.6778171782, 1745: -9.24469199252, 1746: 0.00224384588489, 1747: -5.10908512501, 1748: -3.3989881892, 1749: 0.0772673144611, 1750: -1.41745875237, 1751: 4.33258337532, 1752: -1.49874227096, 1753: 0.976072185798, 1754: 13.5331125269, 1755: -1.41771950204, 1756: 7.29452513236, 1757: -1.52308309477, 1758: -6.64335108566, 1759: 21.3296683338, 1760: -9.815305925, 1761: 3.93668013769, 1762: 19.4272676357, 1763: 0.512439885166, 1764: 0.880303875576, 1765: 2.06081364668, 1766: -3.78087708314, 1767: -12.0875024222, 1768: -3.97867542212, 1769: -4.34853015625, 1770: 7.94891376551, 1771: 13.4313673913, 1772: 0.434747445487, 1773: -0.519436031357, 1774: -5.07824483351, 1775: 3.26702752444, 1776: -1.77713342019, 1777: -3.90302714681, 1778: 0.807910880376, 1779: 4.96441316347, 1780: 4.58310050218, 1781: 9.23165387138, 1782: 3.82961422225, 1783: 1.48626390006, 1784: -1.87918791214, 1785: 14.3698996118, 1786: -0.637983709777, 1787: -0.920379272206, 1788: -0.834043311082, 1789: 10.013360597, 1790: 0.350509835218, 1791: -5.24432299859, 1792: -7.50275805571, 1793: -5.12094693377, 1794: -10.0936958417, 1795: -2.38682915721, 1796: 7.57684727776, 1797: 30.0355977969, 1798: -1.31046015668, 1799: -1.7351508036, 1800: -0.553502554429, 1801: -0.801253613516, 1802: 6.01609883745, 1803: 1.14826616216, 1804: -1.01963305189, 1805: 1.42104139953, 1806: 6.23474178123, 1807: -9.3761698288, 1808: 17.6670741777, 1809: 3.85804849829, 1810: -17.8428053366, 1811: -1.31394590228, 1812: 2.02377648041, 1813: -1.01227112532, 1814: -8.1579456672, 1815: -2.91431859936, 1816: -3.11021506331, 1817: -7.13381835534, 1818: -1.13764673835, 1819: 3.8946507277, 1820: 6.7740138682, 1821: 1.25855838538, 1822: -3.75655114441, 1823: -3.87236780069, 1824: -10.7405577655, 1825: 8.66566392221, 1826: 2.19039650253, 1827: -0.851426666293, 1828: -0.543739166523, 1829: -3.57946496001, 1830: 5.12609479851, 1831: -11.8818715173, 1832: -16.8755297012, 1833: -1.18121923921, 1834: -0.934882980354, 1835: 18.6249003325, 1836: -0.321500953365, 1837: 1.13289028395, 1838: -1.06756914583, 1839: -1.94796839784, 1840: -1.8227475145, 1841: -2.96171092275, 1842: 6.38867492209, 1843: -2.21878711229, 1844: 0.464638824017, 1845: -2.4795407583, 1846: 14.2906905594, 1847: -0.0636743134051, 1848: 9.28222429738, 1849: -1.34164214729, 1850: -3.82262541914, 1851: 1.96414243857, 1852: -1.07942309274, 1853: -0.0313730666653, 1854: -3.62792413726, 1855: 5.93238343694, 1856: -3.15724523653, 1857: -6.97522824813, 1858: 7.6986522296, 1859: 6.62501740382, 1860: 24.2734076434, 1861: -3.30364503098, 1862: -9.04801297216, 1863: 10.9296264706, 1864: 11.6783659005, 1865: 3.3988863573, 1866: -4.72324541805, 1867: 0.154139868521, 1868: -7.32561595777, 1869: 3.20388699889, 1870: 6.6857422631, 1871: -0.103964586968, 1872: 2.16015063564, 1873: 1.12044392101, 1874: -7.82979795182, 1875: 0.141874984077, 1876: 2.25192530956, 1877: -0.972531291592, 1878: -8.84963076824, 1879: -0.87780987389, 1880: -0.678901939921, 1881: 6.0148615534, 1882: -0.215657200336, 1883: 2.52927442814, 1884: -2.02001636905, 1885: -1.10860936385, 1886: 4.13451249502, 1887: -2.25191853367, 1888: 6.32047286516, 1889: -1.09211480712, 1890: -6.56300975829, 1891: 0.372575191079, 1892: -0.321809613479, 1893: 0.579651546869, 1894: -0.0240893825324, 1895: -1.03639102912, 1896: -0.976416283207, 1897: -0.163563102085, 1898: -0.0250412982391, 1899: -0.0923457835603, 1900: -0.00351744579516, 1901: -2.08309478902, 1902: -0.313844863831, 1903: 0.524397567322, 1904: -1.22716860859, 1905: 1.65168154389, 1906: -0.459451503562, 1907: -0.206853831485, 1908: 10.0038182118, 1909: 3.58962599541, 1910: -0.941427307587, 1911: -16.1091340162, 1912: -3.32259646751, 1913: -3.65502846874, 1914: -5.7678389896, 1915: -1.31423180544, 1916: 3.26884294186, 1917: -0.389915369772, 1918: -1.1614702388, 1919: -0.0483541654636, 1920: 9.84103365596, 1921: 9.81804921881, 1922: 0.458643962992, 1923: 0.138564148651, 1924: -0.232356624372, 1925: -0.176889854342, 1926: 0.241568831219, 1927: -0.447895419937, 1928: 1.69924626251, 1929: -3.19990082524, 1930: -0.054423318483, 1931: 13.5434138573, 1932: -7.56234536518, 1933: 1.00591512372, 1934: -0.766711967421, 1935: 6.1805214566, 1936: 3.3471681029, 1937: -4.97690021939, 1938: -0.214611339183, 1939: -1.15730214607, 1940: -1.70207992086, 1941: -0.208585456004, 1942: 0.0385443033162, 1943: -1.61828015458, 1944: -0.199398830937, 1945: -1.53467946897, 1946: -0.179900070149, 1947: 1.01354183454, 1948: 0.00224202066247, 1949: -0.2511081534, 1950: -19.2434527916, 1951: -0.710472847011, 1952: 8.51270675265, 1953: -0.081709966903, 1954: -0.0704577051293, 1955: -0.627993904348, 1956: 0.030448270792, 1957: -2.40685279705, 1958: 4.9889316326, 1959: -0.68738992498, 1960: -0.644481607557, 1961: 0.184565779695, 1962: 6.24122302622, 1963: 2.51110160285, 1964: -26.1648605124, 1965: -4.03224224286, 1966: -7.92345400761, 1967: -4.52345819434, 1968: -6.94958742385, 1969: -1.52048823617, 1970: 2.6935000139, 1971: 0.0140371080477, 1972: -10.5912418039, 1973: 0.324895232138, 1974: 4.73295414666, 1975: -2.7542617608, 1976: -1.43554289083, 1977: -0.0893829007387, 1978: -2.10605562466, 1979: 0.774446970862, 1980: -1.12706919719, 1981: 4.9186621861, 1982: 3.84445467848, 1983: -4.74442103236, 1984: -8.6330034509, 1985: 2.27687729016, 1986: 7.21296116542, 1987: 6.62099264737, 1988: -6.04575637293, 1989: -0.610002771548, 1990: -8.11344207807, 1991: -9.66842419441, 1992: 1.54412865415, 1993: 12.0585046558, 1994: -12.5571015072, 1995: -4.88523181952, 1996: -13.4377893359, 1997: -0.262793894988, 1998: -12.946324007, 1999: -0.97871491417, 2000: -9.70618475636, 2001: 4.81846487337, 2002: -9.39464235218, 2003: -0.94267749334, 2004: -1.01118858956, 2005: -2.28996051053, 2006: 3.09423641723, 2007: -2.42719151138, 2008: -0.551826989658, 2009: -6.48514920498, 2010: -2.17051108213, 2011: 2.49225173498, 2012: 10.0993921165, 2013: 0.240874494876, 2014: -3.38460010017, 2015: 0.460088074887, 2016: -0.809351026248, 2017: -2.36504012146, 2018: 6.70068801369, 2019: -16.4344702299, 2020: 3.38799953274, 2021: 22.2882541934, 2022: 11.5629103592, 2023: 9.21048775761, 2024: -0.952404075422, 2025: 5.42671647657, 2026: 8.86987728868, 2027: -13.0626410116, 2028: 7.03264316412, 2029: -12.7886298721, 2030: 15.457466834, 2031: -2.54834869219, 2032: 0.988252183649, 2033: -10.0427185879, 2034: -4.43400162671, 2035: -5.27394971371, 2036: -2.53900129534, 2037: 0.0926606947006, 2038: -2.12652926816, 2039: -0.311291760062, 2040: 7.90732698524, 2041: -2.16674922138, 2042: 7.70165250161, 2043: -2.04664755434, 2044: 0.966169056603, 2045: -0.71793910836, 2046: 0.603149101284, 2047: 2.16575020023, 2048: -2.61539725586, 2049: -2.47897508289, 2050: 3.44087611569, 2051: -1.86139084942, 2052: 3.91437363419, 2053: -1.4341529679, 2054: -1.59732727708, 2055: -0.392591901648, 2056: 4.93355373957, 2057: -0.951830334029, 2058: 1.1478020849, 2059: 6.21188558364, 2060: -0.887760309849, 2061: -4.56438823508, 2062: -0.099836820279, 2063: -1.03805580242, 2064: 9.6086389321, 2065: -0.566288059814, 2066: -3.09209684363, 2067: -4.40573132397, 2068: 2.3116965509, 2069: 3.99755894132, 2070: -4.66720766536, 2071: 11.523168788, 2072: 3.79840789141, 2073: -0.714939005615, 2074: 4.32986617919, 2075: -2.52923304273, 2076: 13.1685321449, 2077: -3.19583058173, 2078: -0.637901571123, 2079: -9.5120863936, 2080: -4.20397469066, 2081: 2.96678657819, 2082: -8.13163409659, 2083: 0.767584605303, 2084: -1.14493803743, 2085: -6.02896460867, 2086: -1.74110489139, 2087: -6.91061245208, 2088: 12.3649127686, 2089: -15.0399195633, 2090: 7.23133220625, 2091: 24.853355399, 2092: 12.5652682089, 2093: 1.94898681162, 2094: 14.0342087378, 2095: -0.774717487765, 2096: -8.55738289727, 2097: 2.11380628824, 2098: 30.6653919974, 2099: -4.74569969601, 2100: 2.31180675194, 2101: -4.92749345666, 2102: -9.88022714325, 2103: 29.4472245174, 2104: 9.02646583071, 2105: 34.6378774483, 2106: -4.04862689812, 2107: -4.21656674997, 2108: 0.625964884192, 2109: 7.28902009289, 2110: 8.66116801256, 2111: 35.0388879338, 2112: 0.185623649253, 2113: 24.0948008906, 2114: -12.1331462143, 2115: -12.8927368547, 2116: -19.2716722868, 2117: -1.20239573782, 2118: 1.32272846632, 2119: 12.2676576866, 2120: -0.70342117044, 2121: -9.82233177018, 2122: -5.88052866752, 2123: -1.79858562938, 2124: 13.0138882793, 2125: 16.6951911574, 2126: -5.11473435855, 2127: 11.4499813683, 2128: -2.84562566184, 2129: -10.1715328536, 2130: 7.52968977757, 2131: -1.96735703279, 2132: -9.34439950947, 2133: -0.462630477471, 2134: 4.64695851463, 2135: 34.272864983, 2136: -7.27231028134, 2137: 14.9375484392, 2138: -1.12471158778, 2139: 7.59934610392, 2140: -8.70345381851, 2141: 16.811672109, 2142: -2.97444538905, 2143: 12.5668943178, 2144: 4.77382708349, 2145: 0.62134131493, 2146: 31.3305876866, 2147: -5.84722678957, 2148: -0.0394404040322, 2149: -6.74499910542, 2150: -3.40922819454, 2151: 2.49215436767, 2152: -2.49875925633, 2153: 34.6564358864, 2154: 10.3951116984, 2155: 32.2467235251, 2156: 9.07092272537, 2157: -2.1669397119, 2158: 6.78486404208, 2159: -24.317321694, 2160: -2.08958517147, 2161: -13.7525451426, 2162: -1.81748492433, 2163: 15.6197381802, 2164: -3.66478896886, 2165: -3.43093272685, 2166: 10.8999116225, 2167: -5.38438600786, 2168: -10.328794141, 2169: 0.774741862742, 2170: -0.500066832879, 2171: 5.73803592523, 2172: 1.89552313484, 2173: 0.466089821375, 2174: -5.54743676945, 2175: -3.33158518207, 2176: 3.14461197188, 2177: -4.87909105951, 2178: -4.86101890875, 2179: 10.8768725137, 2180: 2.13514621186, 2181: 2.67027208039, 2182: -3.40676776957, 2183: -0.485252829752, 2184: -4.34905773426, 2185: -0.957038551384, 2186: -4.96990916769, 2187: 14.3715415895, 2188: -3.47826690602, 2189: -3.22186914677, 2190: -3.3953291284, 2191: -0.0650411284724, 2192: -4.25672088293, 2193: 5.01268671264, 2194: -3.9189033015, 2195: 22.1987258825, 2196: -4.13337221725, 2197: 3.19452435091, 2198: -1.80893419185, 2199: -9.63776448753, 2200: 26.3930886473, 2201: -5.05896279296, 2202: -5.08416371186, 2203: 1.91999721984, 2204: -3.57201263792, 2205: -3.22937488925, 2206: -0.814423312434, 2207: -11.1295706326, 2208: 82.7597160161, 2209: 31.8088589334, 2210: -7.53612096698, 2211: -0.665988604458, 2212: -10.5625467372, 2213: -1.96229591275, 2214: 1.48738777636, 2215: -12.1807267458, 2216: 10.1261289774, 2217: 11.5012910281, 2218: 1.95826587423, 2219: 11.1568989598, 2220: -0.929263114639, 2221: -1.40115278068, 2222: -13.1806027099, 2223: -4.43570305131, 2224: -0.950557835301, 2225: -4.1003153623, 2226: -1.80431023672, 2227: -15.6127138265, 2228: 1.52441445207, 2229: -1.82719762799, 2230: 4.98678791283, 2231: -2.04056567613, 2232: 0.309086840688, 2233: 1.09704720934, 2234: -1.2880316933, 2235: -6.55936866897, 2236: 6.55867225463, 2237: 4.03463913842, 2238: 1.72452184996, 2239: -5.47222366552, 2240: -0.748020939707, 2241: -0.819875124052, 2242: -8.74395295065, 2243: -0.804452328196, 2244: -4.49644041996, 2245: -6.04070284283, 2246: -0.734618212415, 2247: -0.923841284385, 2248: -0.97417905876, 2249: -1.56260224281, 2250: 0.431031055381, 2251: -2.35467979379, 2252: -0.968573213273, 2253: -1.49651799956, 2254: -2.01880381994, 2255: -5.32770599371, 2256: 13.1167034419, 2257: 28.0032003265, 2258: -0.0626168653495, 2259: -27.1466829226, 2260: 3.72578507285, 2261: -5.38485279045, 2262: -15.7811262351, 2263: 9.39435290458, 2264: -2.17066112775, 2265: -2.31667188079, 2266: -1.34155811743, 2267: -7.60231218153, 2268: -1.00382000927, 2269: 13.7266822991, 2270: 30.8468733779, 2271: -0.0374175279209, 2272: -0.139794747459, 2273: -0.773789264514, 2274: -1.19806492978, 2275: -1.1869783223, 2276: -0.333295957979, 2277: -8.45990749949, 2278: 0.565767734089, 2279: 2.399380125, 2280: -6.93175976911, 2281: 4.0410527391, 2282: 4.97927138134, 2283: 3.29904081867, 2284: 10.3540830482, 2285: -5.20839593732, 2286: -5.08232412025, 2287: -0.857090222067, 2288: 38.5077998615, 2289: 1.87947674073, 2290: -1.73059273253, 2291: -0.740699428647, 2292: -20.1287447769, 2293: -0.759221772818, 2294: -4.7130087994, 2295: -1.12307138126, 2296: 12.6151403967, 2297: -1.00519228484, 2298: -0.838411186487, 2299: -6.6238353865, 2300: -4.83040197283, 2301: 39.0619093796, 2302: -0.889187483661, 2303: -0.915560385052, 2304: 2.88749221233, 2305: -0.977748658516, 2306: -1.93697100023, 2307: 7.06493562611, 2308: -1.30161054724, 2309: -1.31801713478, 2310: 0.960048419448, 2311: -10.3248873911, 2312: -3.45412209748, 2313: -16.9409632485, 2314: -0.146008026895, 2315: -5.48925733658, 2316: 2.28400632793, 2317: -1.48205806809, 2318: -3.11254929042, 2319: 4.31661489677, 2320: -3.27582443662, 2321: -14.8651988974, 2322: -4.41351923198, 2323: 2.17763172647, 2324: 6.2023890968, 2325: -21.6017411615, 2326: -1.10260121211, 2327: -9.4366569882, 2328: -28.5251210417, 2329: 1.30808760765, 2330: -10.5533644951, 2331: 3.78377880851, 2332: -3.33239581658, 2333: -2.21202467064, 2334: -3.89218841303, 2335: -12.4125833906, 2336: 0.0794496752581, 2337: -3.80030256989, 2338: -9.08980958525, 2339: 5.14708807071, 2340: -1.5855141702, 2341: -4.99776224799, 2342: 26.5403651358, 2343: 54.1497358365, 2344: -1.79977370519, 2345: -33.4176519671, 2346: -2.25887784202, 2347: -14.61719042, 2348: -4.37601057276, 2349: -18.307563645, 2350: -4.70992720845, 2351: -6.13655210959, 2352: -3.99147998286, 2353: -4.53218675569, 2354: -6.14126622692, 2355: -5.28816716002, 2356: -3.30819326238, 2357: -2.76628954797, 2358: -5.27997186174, 2359: 0.529187882631, 2360: -3.32595263408, 2361: -3.12649108757, 2362: -4.04868140235, 2363: 19.545285074, 2364: -1.26839677452, 2365: 18.0057193532, 2366: -11.9022048013, 2367: 1.97448690428, 2368: 11.1733629993, 2369: 8.81133883014, 2370: 11.5563659159, 2371: -0.508986522603, 2372: 4.80605887525, 2373: -1.99068700844, 2374: -7.15140222341, 2375: -1.92202783105, 2376: -13.621510137, 2377: 19.0245727283, 2378: -15.8610592984, 2379: 8.70528736685, 2380: -14.4319250632, 2381: -5.06437749587, 2382: 0.712672669825, 2383: -2.63298137944, 2384: -1.06661944345, 2385: 0.804865974253, 2386: 6.15067886633, 2387: -0.410227938436, 2388: -4.37704419314, 2389: 8.02462640736, 2390: -10.4403201826, 2391: 19.6345991535, 2392: 5.1055338936, 2393: -4.60712028128, 2394: 13.5237349979, 2395: -19.6023790436, 2396: -16.2128869374, 2397: -1.75441342339, 2398: -20.9481391156, 2399: 4.44098137514, 2400: -3.20503864667, 2401: 2.69789352036, 2402: -4.84066557727, 2403: -4.67765362884, 2404: -3.7483926125, 2405: 1.92041661976, 2406: -0.315648143795, 2407: -4.87472975727, 2408: 5.50750009768, 2409: -4.94504393871, 2410: -30.4699862994, 2411: -5.36903556214, 2412: -5.0048381525, 2413: 1.708488587, 2414: -4.62858348937, 2415: -0.00381470194191, 2416: -38.2681536603, 2417: 2.37182492303, 2418: 19.4461590919, 2419: 10.0880991584, 2420: -20.307574779, 2421: -5.22290258374, 2422: -6.75559130541, 2423: 1.921266903, 2424: -3.7499938044, 2425: 17.5736615827, 2426: -8.62892025784, 2427: -2.45798486301, 2428: 2.7083474652, 2429: -2.10952689566, 2430: 14.2058315304, 2431: 3.18776606288, 2432: 30.1483601127, 2433: -4.76845581111, 2434: -5.31430422657, 2435: 9.16871194766, 2436: 1.489360045, 2437: -5.40388894235, 2438: -3.57627185043, 2439: 10.5339370263, 2440: 33.9189680044, 2441: -32.7968169439, 2442: 15.3594394713, 2443: 9.87980563734, 2444: -0.493003236959, 2445: -3.32332298128, 2446: 2.66302094761, 2447: -37.5469251057, 2448: 0.327785811217, 2449: 36.1038957182, 2450: -19.6815530814, 2451: -87.6578536721, 2452: 24.8712093595, 2453: -46.385533353, 2454: -35.3536436315, 2455: 4.11066858152, 2456: -34.1195264142, 2457: 7.58594949445, 2458: 15.0153997389, 2459: 18.233756958, 2460: 47.9751860807, 2461: -3.89570557789, 2462: -40.9370395425, 2463: 30.5476855142, 2464: 31.4864845045, 2465: -100.0, 2466: 70.5937954393, 2467: 0.788090691124, 2468: -0.902998381808, 2469: -0.584993158443, 2470: 18.3391734014, 2471: 32.2255788095, 2472: -98.4046156338, 2473: 0.957219779281, 2474: 69.1989034655, 2475: 47.3506538433, 2476: -37.9768977428, 2477: 9.80580296671, 2478: -5.64772514667, 2479: -19.0463307036, 2480: 23.7471371077, 2481: 13.7187979338, 2482: 2.27247347427, 2483: 25.4063466739, 2484: 20.8828075371, 2485: -3.0848773225, 2486: 51.8795348948, 2487: 32.8527263505, 2488: -3.78614172603, 2489: 26.7499989402, 2490: -54.9850729625, 2491: -4.86519190651, 2492: -33.3298773239, 2493: -14.05315205, 2494: -29.7200557416, 2495: -21.8812091939, 2496: -5.03210494726, 2497: 4.22365406573, 2498: -8.41288791814, 2499: -13.8913847046, 2500: -44.1413758334, 2501: 0.837359771908, 2502: 43.1718535495, 2503: -13.0758553958, 2504: -7.75948747808, 2505: -71.7425484545, 2506: -19.8235678972, 2507: 53.3278096189, 2508: -4.2689517707, 2509: -17.76491011, 2510: -17.6616527804, 2511: -3.87316919377, 2512: -7.43773925795, 2513: -4.50537410391, 2514: -4.73163147154, 2515: 10.9626945231, 2516: -4.0612785219, 2517: -1.81674824636, 2518: -5.36668374002, 2519: -3.27545480167, 2520: -6.00697863584, 2521: 9.22683262586, 2522: 16.2578852911, 2523: -12.5341094675, 2524: -4.08300795106, 2525: -17.0465000073, 2526: -4.72883042592, 2527: -7.50287703537, 2528: -25.5752891912, 2529: -3.8885972622, 2530: -5.50273506342, 2531: 40.7762238945, 2532: -5.10415527831, 2533: -5.87000790341, 2534: -1.45124156949, 2535: -3.22312534774, 2536: -100.0, 2537: -4.13486206052, 2538: -4.77736815757, 2539: -5.15549745849, 2540: -24.3966177471, 2541: -4.83159665103, 2542: -7.56444543935, 2543: -4.99407640267, 2544: 16.1763207312, 2545: -4.58838012765, 2546: -6.45159201659, 2547: -16.6426992575, 2548: -7.14591616382, 2549: -77.2173977108, 2550: -3.48900141896, 2551: -2.81827659816, 2552: -5.26355021069, 2553: -3.70977064937, 2554: -3.11839577465, 2555: -11.9362149501, 2556: 100.0, 2557: -28.0064885461, 2558: 4.34645357685, 2559: 3.88464058076, 2560: 14.791731932, 2561: 9.12094445788, 2562: -2.30325779937, 2563: 33.5909352258, 2564: 14.8099593241, 2565: 53.2593245367, 2566: -6.55008547846, 2567: 5.60215433364, 2568: -22.1504921706, 2569: -0.714270980271, 2570: -2.00486087095, 2571: -1.56873750478, 2572: 38.561123817, 2573: -0.686379193397, 2574: -4.17318467591, 2575: -0.658599364615, 2576: 19.8650913005, 2577: -9.57643599784, 2578: 4.42542714216, 2579: 3.7575025907, 2580: -2.12765753929, 2581: 29.4430923834, 2582: 19.9985751001, 2583: 27.1558109229, 2584: 37.4569829144, 2585: 40.385651296, 2586: -12.1419405526, 2587: -1.28201827784, 2588: 28.7681764599, 2589: -0.58081579597, 2590: -0.74793624699, 2591: 11.6794615938, 2592: -0.598156736006, 2593: -4.59274509845, 2594: -18.6729067658, 2595: -0.712222519877, 2596: -0.752100066761, 2597: -0.296412099155, 2598: -0.775454355379, 2599: -3.93684084751, 2600: 41.5633940422, 2601: 4.95965569353, 2602: -1.17447933757, 2603: -0.788427796588, 2604: -2.39620114883, 2605: -3.44035212353, 2606: 33.4435289137, 2607: 11.6658330508, 2608: -100.0, 2609: 52.910450854, 2610: -15.4298213359, 2611: -1.75936915386, 2612: 2.71683826916, 2613: 27.3516616903, 2614: 29.4722430081, 2615: -1.03160390157, 2616: -13.4004207221, 2617: -0.664200116812, 2618: -13.9709274012, 2619: 60.0927220353, 2620: 4.5086038561, 2621: -3.05434064474, 2622: -0.812172917371, 2623: -0.654264853182, 2624: -0.624042212972, 2625: -0.899715305604, 2626: -5.17401032676, 2627: -40.9371447449, 2628: 6.83768743476, 2629: 34.7876003715, 2630: -34.5679836431, 2631: -49.0570013231, 2632: 7.44103161464, 2633: -9.45728038872, 2634: -19.2288911012, 2635: 34.9335170999, 2636: -0.681550559278, 2637: 41.9524133167, 2638: -4.47077231973, 2639: -3.77027387774, 2640: -8.96143233603, 2641: -19.994102529, 2642: -0.730754712482, 2643: -4.31957982428, 2644: -0.806396669277, 2645: 3.20839837973, 2646: -0.670443867486, 2647: -0.621523878016, 2648: 53.9016053918, 2649: -4.08738204285, 2650: 86.3478809296, 2651: -0.873174566666, 2652: -0.847851618375, 2653: 4.82691151131, 2654: -0.718386177524, 2655: -4.44430171231, 2656: 6.40560449063, 2657: 6.58739002784, 2658: -0.553060911194, 2659: -3.75471853538, 2660: 10.3833140349, 2661: 25.5353380845, 2662: -5.60918179254, 2663: -8.42382568009, 2664: -6.43289504204, 2665: 3.93498453527, 2666: 3.88681946989, 2667: -9.83461233395, 2668: -2.6575327099, 2669: -0.871124075751, 2670: 48.5664352923, 2671: -2.73088478915, 2672: -11.3987334682, 2673: 8.66868503743, 2674: 7.77964090992, 2675: -1.0719673109, 2676: -36.9257283322, 2677: -31.5454423833, 2678: -1.92891810409, 2679: 52.6124347342, 2680: 64.2268732963, 2681: 13.2841243841, 2682: -3.29430613942, 2683: -4.29290146168, 2684: 3.39247362144, 2685: -12.171948267, 2686: -7.65517699068, 2687: 25.4542743209, 2688: -2.73124213788, 2689: -0.882835678096, 2690: -94.289130227, 2691: 37.8597728883, 2692: 68.2615022014, 2693: -24.0484358625, 2694: -44.9522404159, 2695: 2.5207254036, 2696: 41.5480618257, 2697: 8.28554867156, 2698: -1.28627095898, 2699: -3.85562397303, 2700: 6.10071777976, 2701: -6.86409948754, 2702: -4.71428874044, 2703: -13.2991723383, 2704: 7.54733580428, 2705: -5.04440822051, 2706: -10.7419859403, 2707: -5.60699439627, 2708: -4.03852107957, 2709: 15.573273474, 2710: -35.8853028608, 2711: 24.1684986493, 2712: -4.8366423047, 2713: -0.896633684685, 2714: -3.22479252827, 2715: 30.2469003895, 2716: 99.1026397549, 2717: -30.2785249154, 2718: 13.5401684053, 2719: -29.9565404024, 2720: -28.0973262579, 2721: 13.6969012029, 2722: 52.3287138895, 2723: -6.61116321229, 2724: 10.2687845554, 2725: 15.4923815396, 2726: 28.9925200071, 2727: -47.7104689894, 2728: -51.0512435003, 2729: 13.0374663011, 2730: -4.5094936657, 2731: -6.32791249672, 2732: -4.88433104951, 2733: 0.15557265495, 2734: -6.12733021848, 2735: -99.9223162966, 2736: 32.9233375932, 2737: -0.624450015748, 2738: 1.4724700915, 2739: 49.7920308092, 2740: -55.8743726784, 2741: 7.03354604556, 2742: -14.7748233201, 2743: 5.06114185874, 2744: -2.73551204427, 2745: -30.8886722273, 2746: -4.35844709558, 2747: -18.2546724621, 2748: 11.8175003783, 2749: -4.72433676815, 2750: -28.7926812392, 2751: -4.41421925753, 2752: -4.17364999381, 2753: -4.68229855137, 2754: 10.414530785, 2755: -3.08627386727, 2756: -6.68674111674, 2757: -62.9513780499, 2758: -4.10315740042, 2759: -37.7444702388, 2760: -3.43419993916, 2761: -4.21928357117, 2762: -13.4871733893, 2763: -3.75442771388, 2764: -4.49091375993, 2765: 46.3298067213, 2766: -3.39126429206, 2767: -15.381247395, 2768: 17.237987076, 2769: -30.0697111886, 2770: -26.7554332151, 2771: -19.2131194222, 2772: 34.9085571064, 2773: -42.3839247827, 2774: -7.29546655357, 2775: -24.1798876215, 2776: -15.4375916531, 2777: -6.26683721304, 2778: -18.2340645631, 2779: 74.048091891, 2780: -15.7892498809, 2781: 9.59848632556, 2782: -4.23262514622, 2783: 19.8569766938, 2784: 12.8121514855, 2785: -3.19374952608, 2786: 10.9437333283, 2787: 9.90149572543, 2788: 12.1968400238, 2789: -26.6533802834, 2790: -40.6248086752, 2791: -0.690279716264, 2792: -3.16566341709, 2793: -1.62297952975, 2794: -9.44671595544, 2795: 9.02682876756, 2796: 34.2243175413, 2797: -9.249941526, 2798: 14.4598512671, 2799: -8.50272592063, 2800: -24.3930493768, 2801: 22.5876953613, 2802: 35.7443516589, 2803: 69.9097094726, 2804: -24.4972834116, 2805: 32.9623388815, 2806: 9.45687891437, 2807: 6.39212914274, 2808: -21.169544764, 2809: 0.970642835353, 2810: -9.0017704012, 2811: -10.0127400236, 2812: -0.92997023074, 2813: -21.0779286015, 2814: -4.02893481393, 2815: 6.59246863016, 2816: -5.48552238643, 2817: 11.8417205929, 2818: 34.4530210013, 2819: 25.3146515165, 2820: -26.6974544389, 2821: -14.9610126852, 2822: -8.23077037562, 2823: 26.3203906426, 2824: 28.6378108263, 2825: -14.6486495668, 2826: 21.5467814456, 2827: 22.9334687199, 2828: 0.948489134153, 2829: 8.77214493926, 2830: 26.2826271218, 2831: -9.33746510589, 2832: 28.7198215308, 2833: 27.9998938684, 2834: 6.77210538715, 2835: 7.48107942376, 2836: -3.71702400295, 2837: -9.20724521336, 2838: -7.70541689087, 2839: 48.4810209125, 2840: -6.67085690979, 2841: 29.0201517132, 2842: -17.5203595496, 2843: 1.75947744037, 2844: 4.3692419749, 2845: -10.5525165771, 2846: 7.02847341324, 2847: 13.9585762523, 2848: -9.7621271833, 2849: -86.2852879587, 2850: -5.8937264413, 2851: -14.378548313, 2852: 1.41057852849, 2853: 53.849751662, 2854: -14.5970627483, 2855: -12.2127555544, 2856: 20.8204697013, 2857: -9.1655327679, 2858: -8.13087326687, 2859: 20.9720280715, 2860: -8.92624292254, 2861: 6.80311618478, 2862: -9.8149471485, 2863: -9.80272112545, 2864: -6.60484734262, 2865: -9.67736550494, 2866: 32.4655173922, 2867: -1.2104210399, 2868: -9.07298094717, 2869: -5.6107876468, 2870: -29.6584357934, 2871: 3.67726978039, 2872: -43.2852991196, 2873: -8.77515661794, 2874: 26.3201768932, 2875: -7.45865853186, 2876: -5.78263603185, 2877: -15.446784977, 2878: -9.86622299318, 2879: -13.860740506, 2880: -41.6433177013, 2881: -6.70448355599, 2882: -46.3000220176, 2883: -7.80906189135, 2884: -9.82915925303, 2885: -0.293177019427, 2886: -9.38760285578, 2887: -6.40109567969, 2888: -6.94355159665, 2889: 10.0502528382, 2890: -9.46415392742, 2891: -10.0631237946, 2892: -2.32693300155, 2893: 19.1032445881, 2894: -9.47553206553, 2895: -5.43349240874, 2896: -9.61424250036, 2897: 14.1137827206, 2898: 22.3382335, 2899: -9.85227942478, 2900: -10.3256382473, 2901: -6.68739256643, 2902: 15.4359479154, 2903: -6.18663616341, 2904: -8.37516262566, 2905: 37.3133393997, 2906: 21.0934365473, 2907: 40.7756060908, 2908: 6.48398757212, 2909: 7.51894858429, 2910: 4.45949050608, 2911: -9.73221951023, 2912: -29.7598862819, 2913: -29.682763391, 2914: 53.6041419163, 2915: 12.0391934905, 2916: 9.43404634482, 2917: -20.5478473134, 2918: -1.89687847018, 2919: -1.92344474767, 2920: -1.77199176255, 2921: -57.3502715261, 2922: -1.82451422542, 2923: -8.00888583707, 2924: -1.00011625425, 2925: -4.5820641394, 2926: -2.58425056586, 2927: -0.917992328988, 2928: 18.9590217587, 2929: 2.13120468844, 2930: 4.6381868942, 2931: -2.26825448028, 2932: -10.0354257989, 2933: 18.5778897666, 2934: -27.4784748432, 2935: -32.2050373177, 2936: -1.8122571944, 2937: -7.32346147749, 2938: -1.74061151528, 2939: -1.84383843536, 2940: 17.8679995684, 2941: -1.80082686926, 2942: -10.0455960432, 2943: -19.4656922547, 2944: -2.83182404933, 2945: -1.90031163222, 2946: -1.9568945402, 2947: -1.7996929792, 2948: -4.41397346147, 2949: 50.1720393517, 2950: 1.11772275712, 2951: -2.61888555463, 2952: -1.8921127867, 2953: -2.78565340799, 2954: -3.72674141917, 2955: -40.4363730991, 2956: 8.12631995278, 2957: -79.5735363856, 2958: -94.3747614887, 2959: -14.2887309338, 2960: 6.46731707552, 2961: -14.9275356722, 2962: -6.55052198533, 2963: 28.8347335561, 2964: -2.09848863403, 2965: -53.1160506415, 2966: -1.84389102835, 2967: -2.62140727181, 2968: 3.04945397006, 2969: 16.5036716195, 2970: -4.64725875465, 2971: -1.800575054, 2972: -2.24424941536, 2973: -1.95820738115, 2974: -2.03680263719, 2975: -0.648335943306, 2976: 30.5634688124, 2977: 18.3295482444, 2978: -30.2088276119, 2979: 9.50292994238, 2980: 7.49942960906, 2981: 9.18038800117, 2982: 24.2100294959, 2983: 50.4290237912, 2984: 21.5134430081, 2985: -1.86652289398, 2986: -24.8173106882, 2987: -1.84177909635, 2988: -2.64718753863, 2989: 5.66483027818, 2990: -15.6893415099, 2991: -1.7270826358, 2992: -9.88553119738, 2993: -1.91004462931, 2994: 0.558825030843, 2995: -1.67063410474, 2996: -1.93208385321, 2997: -24.2073002858, 2998: -5.43339575531, 2999: 19.8839227778, 3000: -1.93322212526, 3001: -1.75215729592, 3002: -0.428799900039, 3003: -1.83415837506, 3004: -8.9381540524, 3005: 2.09515567695, 3006: 21.9343512972, 3007: 0.1404217898, 3008: -0.528612921093, 3009: 12.2629600401, 3010: 6.79511303728, 3011: -35.655640733, 3012: 48.1902447936, 3013: 17.0724611561, 3014: 0.268224819071, 3015: 19.8927456475, 3016: -5.3516417404, 3017: 22.8748001555, 3018: -2.66281519111, 3019: -21.6396129645, 3020: -10.0341605516, 3021: -5.25686197375, 3022: -26.6116495997, 3023: 38.4703797868, 3024: -2.8163445225, 3025: -3.69262648292, 3026: 25.7822858784, 3027: -8.42463746347, 3028: -15.7850406016, 3029: 1.41691216579, 3030: -28.8291518753, 3031: -6.41274808904, 3032: -7.86142195072, 3033: 17.9406431341, 3034: 26.3213542604, 3035: 41.9245831344, 3036: -22.6878327307, 3037: -8.89042103112, 3038: -43.4643044148, 3039: -3.15731281561, 3040: -0.299847219315, 3041: -2.820149529, 3042: -4.39538234529, 3043: 27.3631407022, 3044: 23.6683297737, 3045: -19.898053051, 3046: 14.85952159, 3047: -11.3232170996, 3048: -9.19401255161, 3049: -14.4425193255, 3050: -9.95687739348, 3051: -9.94321876048, 3052: -27.6579400346, 3053: 5.99646119471, 3054: -3.98769041465, 3055: -6.98845651884, 3056: -21.5327212595, 3057: -4.38326788894, 3058: -1.42700058822, 3059: 8.44597126435, 3060: -8.4749512496, 3061: 14.820021209, 3062: -2.10664636235, 3063: -3.85298817699, 3064: 23.4574409752, 3065: 16.2986059014, 3066: -21.8800211564, 3067: 60.5162530758, 3068: 17.9260636185, 3069: -11.129385965, 3070: -15.1319646905, 3071: -6.07334063318, 3072: 21.799354199, 3073: 2.38148415004, 3074: 29.6528915592, 3075: 1.03827664337, 3076: -16.3370031314, 3077: -13.6893379252, 3078: 6.35639060783, 3079: -7.56859276173, 3080: 7.15379008477, 3081: -8.47418665079, 3082: -9.14228613718, 3083: 11.4220911382, 3084: 67.9060488338, 3085: -3.19968085434, 3086: 21.9104238107, 3087: -60.6569201798, 3088: -50.1519772329, 3089: -6.0874210191, 3090: 5.79710898997, 3091: 12.6809201618, 3092: 3.57164204433, 3093: -5.43549105479, 3094: 24.5184504752, 3095: -1.71290091936, 3096: 35.7283172428, 3097: 84.5856786412, 3098: -6.38043077643, 3099: 50.7362794265, 3100: -9.8146206, 3101: -9.86927671671, 3102: -12.3078173093, 3103: 16.2696854496, 3104: -10.9586014718, 3105: -4.5773585441, 3106: -60.951190796, 3107: -5.49955947951, 3108: 47.9707942489, 3109: -7.84269036584, 3110: -9.96321574633, 3111: 5.54791127551, 3112: -10.6619273668, 3113: -8.64625179719, 3114: 28.7334057646, 3115: -31.173088331, 3116: -9.9085680521, 3117: -18.2951339604, 3118: 29.5770319369, 3119: -10.6536539016, 3120: 7.06154308329, 3121: -28.8662958245, 3122: 17.7401842712, 3123: 22.1228900063, 3124: 78.0689397555, 3125: -3.12393532925, 3126: 3.68507100995, 3127: 1.13260171541, 3128: -11.5853741465, 3129: -6.66182801628, 3130: -0.0419982133181, 3131: -9.97878463066, 3132: 13.7926965779, 3133: 0.371646127911, 3134: -12.2891571463, 3135: -0.337045838358, 3136: 9.16943960305, 3137: 21.7901332177, 3138: 24.7208065279, 3139: -22.4460681729, 3140: 15.7104784154, 3141: 11.1907906263, 3142: 0.471905837709, 3143: -37.2180334912, 3144: 8.30608866681, 3145: 5.76488182051, 3146: 37.3167645282, 3147: -6.68169475665, 3148: 11.2635556362, 3149: 53.1955533889, 3150: -34.2454276667, 3151: -8.07363247877, 3152: -26.509600088, 3153: -14.4426922861, 3154: -3.46418292897, 3155: -23.1299351337, 3156: -4.04802465992, 3157: -13.12001842, 3158: -5.75921848342, 3159: -6.871229251, 3160: 2.16626848072, 3161: -22.885900475, 3162: -32.4956814819, 3163: 19.9058376232, 3164: -12.0857976818, 3165: 25.4920388017, 3166: -10.7716859399, 3167: -28.6745432186, 3168: 6.37650534865, 3169: -10.8971819178, 3170: 28.7138154151, 3171: -13.2832602323, 3172: -28.0367711152, 3173: -28.0741515926, 3174: 1.39347864709, 3175: -4.93307633492, 3176: 3.76490859189, 3177: -6.72052602674, 3178: -8.69343029497, 3179: -16.7649144169, 3180: 7.20203852187, 3181: 11.9095378945, 3182: 2.11761979453, 3183: 1.27725559143, 3184: 41.3374953551, 3185: -13.0418690246, 3186: 7.26958717792, 3187: -6.05649105294, 3188: 11.0349170682, 3189: 1.93201209437, 3190: 31.3134704969, 3191: -1.82663563388, 3192: -15.2557580159, 3193: -1.48509037595, 3194: 3.32881969795, 3195: -6.56841981467, 3196: 2.90271966871, 3197: 2.8976633071, 3198: 2.03195340353, 3199: 3.29033695212, 3200: -20.316099381, 3201: -15.5048768983, 3202: -22.5846678235, 3203: -14.9086843802, 3204: 8.93530554084, 3205: 9.43939292713, 3206: 13.2127486376, 3207: 3.59054563769, 3208: 36.2236691075, 3209: 3.0251220408, 3210: -1.62728053925, 3211: 0.893912145012, 3212: 0.824197870328, 3213: -9.44140132652, 3214: 2.38308894789, 3215: 35.0396797953, 3216: -12.0662178841, 3217: 3.98681690219, 3218: -7.16441221561, 3219: 3.68910699347, 3220: -20.1547053219, 3221: -43.492283874, 3222: 5.67511369342, 3223: 46.8100435348, 3224: 2.03906009224, 3225: -11.9536179785, 3226: -2.50611268185, 3227: -4.08489647006, 3228: 4.7152365467, 3229: 22.6203230189, 3230: 2.38887627321, 3231: 4.16554033876, 3232: 7.07741633158, 3233: 2.65122384237, 3234: -7.96091512688, 3235: 0.917466005364, 3236: 1.00421432249, 3237: 2.44205641104, 3238: 7.46172129012, 3239: 1.02421930219, 3240: 4.15647572203, 3241: 3.1034322643, 3242: 44.5698784868, 3243: 6.17847452363, 3244: 6.44898043865, 3245: 1.66292037568, 3246: 5.19446675884, 3247: 10.9153393981, 3248: 2.80414995517, 3249: 3.3002590077, 3250: 0.090810350414, 3251: 12.9394107297, 3252: 5.99169929335, 3253: 4.72966966417, 3254: 3.35068542063, 3255: 7.54159571128, 3256: -19.9816040991, 3257: 9.97774615809, 3258: 28.5626965971, 3259: 26.8656565483, 3260: 27.4071950626, 3261: -16.1690383081, 3262: 1.30747698232, 3263: -3.64503287936, 3264: -37.3769331222, 3265: 16.7111414388, 3266: -2.63927860351, 3267: 0.4739961391, 3268: -0.557242826783, 3269: -8.33787258035, 3270: -52.7819597608, 3271: -0.118091865658, 3272: 3.04721722878, 3273: 6.30791017929, 3274: 23.3313365122, 3275: -3.73958791167, 3276: 14.8277600348, 3277: -8.27216897682, 3278: -2.04828996903, 3279: -28.5463153676, 3280: 0.980330607168, 3281: -6.80354054905, 3282: -18.9918947118, 3283: 2.35622936002, 3284: -19.618085053, 3285: -0.853441647517, 3286: -13.7358085464, 3287: 0.55632298492, 3288: 0.254626016278, 3289: -0.308406711229, 3290: 0.553560595012, 3291: 2.35559790685, 3292: 16.3178357393, 3293: 2.19549489893, 3294: 0.466749286062, 3295: 0.424608220706, 3296: 0.852748003326, 3297: -0.472516314349, 3298: 27.6820680788, 3299: -0.186303173353, 3300: 0.379406097248, 3301: 3.39423040924, 3302: 11.7353379134, 3303: 2.70159037622, 3304: 17.1854268036, 3305: -5.80035335383, 3306: -9.60358524026, 3307: -6.91741606384, 3308: 2.03404447779, 3309: 0.848200888923, 3310: -2.92548733053, 3311: -12.2936973858, 3312: -7.50936750127, 3313: 0.517708910866, 3314: -17.4676438445, 3315: 0.524729431071, 3316: 20.0754680266, 3317: 0.789944545869, 3318: -38.7558998616, 3319: 0.307930563207, 3320: 0.504031331766, 3321: 0.501286396897, 3322: 2.19198415595, 3323: 0.587550381395, 3324: -3.57441322743, 3325: -9.326614914, 3326: -2.39171350091, 3327: 10.5671688767, 3328: 3.14471119562, 3329: 4.68931550858, 3330: 11.227243829, 3331: -8.20983148466, 3332: -13.4635883866, 3333: -17.4396061435, 3334: 0.237327968734, 3335: -12.556847247, 3336: -1.53476274937, 3337: -1.47134055466, 3338: 1.93502003005, 3339: -12.2135016121, 3340: 0.412364510295, 3341: 2.07346375691, 3342: 0.405772035657, 3343: -4.25884657175, 3344: 0.420640633271, 3345: 0.546987136517, 3346: -6.51652502399, 3347: 2.52801491463, 3348: -13.9976830294, 3349: 0.29702836396, 3350: 0.415793606037, 3351: -2.42934530713, 3352: 0.41065026554, 3353: 0.829979780301, 3354: 18.6354227966, 3355: -14.012118342, 3356: -0.393795696286, 3357: -0.7231188009, 3358: -14.6842088979, 3359: -37.8501172151, 3360: 18.2575929263, 3361: 17.9091044739, 3362: 9.2805307657, 3363: -1.79234701394, 3364: -14.1588590165, 3365: 4.23746727539, 3366: 12.4369744468, 3367: 0.865000085816, 3368: -5.94988835513, 3369: 4.60944621125, 3370: -1.66729060707, 3371: -6.43780144234, 3372: -18.9851731251, 3373: 1.35229458702, 3374: 0.952486097526, 3375: 18.1428668895, 3376: 3.45897866484, 3377: -3.23353072968, 3378: -6.79171289278, 3379: 20.0865920871, 3380: 23.948066905, 3381: 2.84710709772, 3382: 6.33007320754, 3383: -17.8376098978, 3384: 5.27265662958, 3385: 4.17474766329, 3386: -7.85223827284, 3387: 22.5426265875, 3388: 17.7737854939, 3389: -13.8846829626, 3390: 5.58383731659, 3391: 6.7630637047, 3392: 4.32591348616, 3393: 16.9671278557, 3394: 35.5214359642, 3395: -1.07866443796, 3396: -0.405646266064, 3397: 3.54853425304, 3398: 12.2680299502, 3399: 2.40695935078, 3400: 2.48875082087, 3401: -1.54278725727, 3402: -20.7259386401, 3403: 4.47583134962, 3404: 3.80561949803, 3405: 1.00312540117, 3406: -0.705941228735, 3407: -19.0552487168, 3408: -33.9859190822, 3409: 15.8154865816, 3410: 1.1932915836, 3411: 17.4804847925, 3412: 1.56045116912, 3413: -37.5689656651, 3414: 4.50102683736, 3415: -6.4016100751, 3416: 5.47514544864, 3417: 18.4758517922, 3418: -35.5620631539, 3419: 0.866052662789, 3420: -12.8166954576, 3421: 12.957203387, 3422: 5.32143428837, 3423: 0.881842984011, 3424: 8.06699579345, 3425: 0.0327942080236, 3426: 4.41132961634, 3427: 38.8010513888, 3428: 2.47258900993, 3429: 18.5184670531, 3430: -15.4592052729, 3431: 4.4818093611, 3432: 7.92395391959, 3433: 22.1533232085, 3434: 4.9576177866, 3435: 14.0790594091, 3436: -31.0285719438, 3437: 24.3635932058, 3438: -5.99470799292, 3439: -29.5929854663, 3440: -7.48908338536, 3441: 12.3890133631, 3442: 17.5145525667, 3443: 17.0819011078, 3444: 2.43312595585, 3445: 20.7257981456, 3446: -28.2458225246, 3447: 1.01164791004, 3448: -1.0284477904, 3449: 1.89466070466, 3450: 2.08994473535, 3451: -1.6691282352, 3452: -1.14048942346, 3453: 2.23160418881, 3454: 1.90815313743, 3455: -19.1965797245, 3456: 2.58530516265, 3457: 22.4014615263, 3458: 2.64019596977, 3459: 2.87000920493, 3460: 7.72507143835, 3461: 4.76737104855, 3462: -0.802409228346, 3463: -20.2257830814, 3464: -14.2947598572, 3465: 35.8980263621, 3466: -17.8554203867, 3467: -6.88047475381, 3468: -1.09589281149, 3469: -37.5628928949, 3470: 10.6931193232, 3471: -2.96650194263, 3472: -3.72805426644, 3473: -3.65387373573, 3474: 10.0932516692, 3475: 7.89150570996, 3476: -8.73020775239, 3477: -10.1774622582, 3478: 6.29629612163, 3479: -23.6920448159, 3480: 2.24779907613, 3481: -39.8608027432, 3482: 10.8313705692, 3483: 15.2574165764, 3484: -13.1845074881, 3485: 11.2641711687, 3486: -6.63928719471, 3487: 43.6037987085, 3488: 14.8110219586, 3489: 51.2306592382, 3490: -2.12926846383, 3491: -2.49525945839, 3492: 1.83050346581, 3493: -1.70144707138, 3494: 0.687156757914, 3495: -0.478442894548, 3496: 1.62840561707, 3497: -1.21574332489, 3498: -0.0861166568639, 3499: 0.576779635148, 3500: -1.00611531744, 3501: 0.648446343048}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    nets.append({'outno': {1: 359}, 'eni': {1: {1: 0.0304046423545, 2: 0.651959362024}, 2: {1: 0.00261812943927, 2: 0.138937931856}, 3: {1: 0.000173034126841, 2: 0.137653838555}, 4: {1: 0.0223427311116, 2: 0.148292791916}, 5: {1: 0.00402370444764, 2: 0.15}, 6: {1: 0.0319499025528, 2: 0.15}, 7: {1: 0.00853658536585, 2: 0.15}, 8: {1: 0.07, 2: 0.15}, 9: {1: 0.00843373493976, 2: 0.15}, 10: {1: 0.04375, 2: 0.15}, 11: {1: 0.7, 2: 0.15}, 12: {1: 0.0318181818182, 2: 0.15}, 13: {1: 0.14, 2: 0.15}, 14: {1: 0.00666666666667, 2: 0.15}, 15: {1: 0.175, 2: 0.15}, 16: {1: 0.0875, 2: 0.15}, 17: {1: 0.0875, 2: 0.15}, 18: {1: 0.021875, 2: 0.15}, 19: {1: 0.025, 2: 0.15}, 20: {1: 0.0134615384615, 2: 0.15}, 21: {1: 0.14, 2: 0.15}, 22: {1: 0.1, 2: 0.15}, 23: {1: 0.07, 2: 0.15}, 24: {1: 0.0777777777778, 2: 0.15}, 25: {1: 0.07, 2: 0.15}, 26: {1: 0.14, 2: 0.15}, 27: {1: 0.14, 2: 0.15}, 28: {1: 0.175, 2: 0.15}, 29: {1: 0.35, 2: 0.15}, 30: {1: 0.7, 2: 0.15}, 31: {1: 0.0538461538462, 2: 0.15}, 32: {1: 0.7, 2: 0.15}, 33: {1: 0.7, 2: 0.15}, 34: {1: 0.0875, 2: 0.15}, 35: {1: 0.7, 2: 0.15}, 36: {1: 0.0777777777778, 2: 0.15}, 37: {1: 0.7, 2: 0.15}, 38: {1: 0.7, 2: 0.15}, 39: {1: 0.35, 2: 0.15}, 40: {1: 0.7, 2: 0.15}, 41: {1: 0.7, 2: 0.15}, 42: {1: 0.7, 2: 0.15}, 43: {1: 0.0102941176471, 2: 0.15}, 44: {1: 0.35, 2: 0.15}, 45: {1: 0.0466666666667, 2: 0.15}, 46: {1: 0.0291666666667, 2: 0.15}, 47: {1: 0.175, 2: 0.15}, 48: {1: 0.00945945945946, 2: 0.15}, 49: {1: 0.0538461538462, 2: 0.15}, 50: {1: 0.0777777777778, 2: 0.15}, 51: {1: 0.1, 2: 0.15}, 52: {1: 0.0205882352941, 2: 0.15}, 53: {1: 0.0304347826087, 2: 0.15}, 54: {1: 0.00804597701149, 2: 0.15}, 55: {1: 0.1, 2: 0.15}, 56: {1: 0.14, 2: 0.15}, 57: {1: 0.0583333333333, 2: 0.15}, 58: {1: 0.14, 2: 0.15}, 59: {1: 0.35, 2: 0.15}, 60: {1: 0.233333333333, 2: 0.15}, 61: {1: 0.233333333333, 2: 0.15}, 62: {1: 0.35, 2: 0.15}, 63: {1: 0.7, 2: 0.15}, 64: {1: 0.0583333333333, 2: 0.15}, 65: {1: 0.07, 2: 0.15}, 66: {1: 0.0636363636364, 2: 0.15}, 67: {1: 0.175, 2: 0.15}, 68: {1: 0.0259259259259, 2: 0.15}, 69: {1: 0.116666666667, 2: 0.15}, 70: {1: 0.175, 2: 0.15}, 71: {1: 0.35, 2: 0.15}, 72: {1: 0.0212121212121, 2: 0.15}, 73: {1: 0.0466666666667, 2: 0.15}, 74: {1: 0.00777777777778, 2: 0.15}, 75: {1: 0.0225806451613, 2: 0.15}, 76: {1: 0.0368421052632, 2: 0.15}, 77: {1: 0.0636363636364, 2: 0.15}, 78: {1: 0.175, 2: 0.15}, 79: {1: 0.14, 2: 0.15}, 80: {1: 0.14, 2: 0.15}, 81: {1: 0.35, 2: 0.15}, 82: {1: 0.7, 2: 0.15}, 83: {1: 0.116666666667, 2: 0.15}, 84: {1: 0.14, 2: 0.15}, 85: {1: 0.7, 2: 0.15}, 86: {1: 0.7, 2: 0.15}, 87: {1: 0.35, 2: 0.15}, 88: {1: 0.7, 2: 0.15}, 89: {1: 0.175, 2: 0.15}, 90: {1: 0.175, 2: 0.15}, 91: {1: 0.7, 2: 0.15}, 92: {1: 0.35, 2: 0.15}, 93: {1: 0.0583333333333, 2: 0.15}, 94: {1: 0.7, 2: 0.15}, 95: {1: 0.7, 2: 0.15}, 96: {1: 0.233333333333, 2: 0.15}, 97: {1: 0.233333333333, 2: 0.15}, 98: {1: 0.07, 2: 0.15}, 99: {1: 0.0583333333333, 2: 0.15}, 100: {1: 0.0179487179487, 2: 0.15}, 101: {1: 0.05, 2: 0.15}, 102: {1: 0.116666666667, 2: 0.15}, 103: {1: 0.1, 2: 0.15}, 104: {1: 0.35, 2: 0.15}, 105: {1: 0.0318181818182, 2: 0.15}, 106: {1: 0.116666666667, 2: 0.15}, 107: {1: 0.14, 2: 0.15}, 108: {1: 0.1, 2: 0.15}, 109: {1: 0.233333333333, 2: 0.15}, 110: {1: 0.7, 2: 0.15}, 111: {1: 0.7, 2: 0.15}, 112: {1: 0.233333333333, 2: 0.15}, 113: {1: 0.35, 2: 0.15}, 114: {1: 0.35, 2: 0.15}, 115: {1: 0.7, 2: 0.15}, 116: {1: 0.7, 2: 0.15}, 117: {1: 3.87374150895e-06, 2: 0.241484336616}, 118: {1: 9.1513333069e-05, 2: 0.766520574414}, 119: {1: 1.00992148454e-06, 2: 0.163933793932}, 120: {1: 2.48414365543e-05, 2: 0.789884848031}, 121: {1: 1.30975514685e-05, 2: 0.766361493713}, 122: {1: 6.59324162593e-07, 2: 0.17025699176}, 123: {1: 8.17739471218e-07, 2: 0.164133755398}, 124: {1: 0.000167593113975, 2: 0.737334549945}, 125: {1: 1.5464440133e-06, 2: 0.16902428405}, 126: {1: 1.22347176214e-06, 2: 0.15}, 127: {1: 7.30883962944e-07, 2: 0.829064171192}, 128: {1: 5.7291070724e-06, 2: 0.759815993108}, 129: {1: 7.49262071283e-07, 2: 0.831419138892}, 130: {1: 1.08182938119e-05, 2: 0.23912023872}, 131: {1: 1.15823776086e-05, 2: 0.207196926645}, 132: {1: 5.66538965622e-06, 2: 0.805092286718}, 133: {1: 7.25309313051e-07, 2: 0.169603659975}, 134: {1: 2.14071551021e-05, 2: 0.85}, 135: {1: 1.43789916949e-05, 2: 0.85}, 136: {1: 2.07879410441e-05, 2: 0.15}, 137: {1: 1.4712765834e-05, 2: 0.15}, 138: {1: 8.28474553007e-05, 2: 0.15}, 139: {1: 4.48132351188e-06, 2: 0.85}, 140: {1: 7.29335404606e-06, 2: 0.85}, 141: {1: 7.52267963728e-06, 2: 0.85}, 142: {1: 1.02205250245e-05, 2: 0.85}, 143: {1: 8.64871808319e-06, 2: 0.15}, 144: {1: 0.000105484564531, 2: 0.15}, 145: {1: 5.26783530509e-06, 2: 0.15}, 146: {1: 4.4794304865e-05, 2: 0.85}, 147: {1: 8.6071196746e-05, 2: 0.85}, 148: {1: 6.42726531599e-05, 2: 0.15}, 149: {1: 2.88337019365e-06, 2: 0.85}, 150: {1: 8.20115210765e-05, 2: 0.15}, 151: {1: 4.59702491179e-05, 2: 0.85}, 152: {1: 1.42834286818e-06, 2: 0.165217591298}, 153: {1: 1.08646375961e-05, 2: 0.85}, 154: {1: 2.67353702245e-06, 2: 0.849818558507}, 155: {1: 6.16660516646e-07, 2: 0.164406019442}, 156: {1: 6.23053341881e-07, 2: 0.155232189189}, 157: {1: 2.15322086711e-05, 2: 0.85}, 158: {1: 8.89598109933e-07, 2: 0.15}, 159: {1: 9.78923087766e-07, 2: 0.15}, 160: {1: 7.44721657761e-07, 2: 0.815304690842}, 161: {1: 2.5585212063e-06, 2: 0.714192567361}, 162: {1: 2.53560643555e-07, 2: 0.848300867199}, 163: {1: 3.03479876511e-06, 2: 0.156137054749}, 164: {1: 6.83582644776e-06, 2: 0.160017249654}, 165: {1: 7.46703541115e-06, 2: 0.772736409688}, 166: {1: 6.35602431782e-07, 2: 0.15}, 167: {1: 4.28132932403e-07, 2: 0.15}, 168: {1: 4.5619236151e-07, 2: 0.775861322815}, 169: {1: 2.26111797938e-06, 2: 0.85}, 170: {1: 3.54925419893e-07, 2: 0.85}, 171: {1: 5.28972169235e-06, 2: 0.85}, 172: {1: 3.39018182563e-06, 2: 0.85}, 173: {1: 3.51530418416e-06, 2: 0.159291858218}, 174: {1: 1.68375498935e-06, 2: 0.15}, 175: {1: 1.48535582186e-05, 2: 0.15}, 176: {1: 3.04981097616e-06, 2: 0.85}, 177: {1: 1.03224930002e-06, 2: 0.15}, 178: {1: 2.43009983658e-05, 2: 0.85}, 179: {1: 6.65575477243e-07, 2: 0.15}, 180: {1: 1.12256869854e-06, 2: 0.15}, 181: {1: 6.77344120409e-07, 2: 0.775348711427}, 182: {1: 1.91204751533e-06, 2: 0.726883199998}, 183: {1: 1.07997369738e-07, 2: 0.85}, 184: {1: 5.40332864562e-07, 2: 0.15}, 185: {1: 1.78351478793e-06, 2: 0.15}, 186: {1: 5.33651617685e-06, 2: 0.693993575673}, 187: {1: 3.75459032168e-07, 2: 0.15}, 188: {1: 3.74940334159e-05, 2: 0.15}, 189: {1: 2.43043514963e-05, 2: 0.15}, 190: {1: 9.87764471542e-07, 2: 0.72284175909}, 191: {1: 1.82151336559e-06, 2: 0.646662722318}, 192: {1: 1.33879211779e-07, 2: 0.85}, 193: {1: 4.80980529835e-07, 2: 0.15}, 194: {1: 2.92122211936e-06, 2: 0.15}, 195: {1: 6.39148628048e-06, 2: 0.85}, 196: {1: 7.10090127137e-07, 2: 0.77089749952}, 197: {1: 2.79244481728e-06, 2: 0.85}, 198: {1: 2.22838538385e-07, 2: 0.85}, 199: {1: 4.81155800939e-07, 2: 0.15}, 200: {1: 3.38210815284e-06, 2: 0.15}, 201: {1: 9.37452115359e-06, 2: 0.85}, 202: {1: 1.18318889482e-06, 2: 0.248599577106}, 203: {1: 1.74020480673e-05, 2: 0.85}, 204: {1: 2.97358178694e-05, 2: 0.85}, 205: {1: 2.04662535052e-05, 2: 0.334703335937}, 206: {1: 8.28082859147e-07, 2: 0.85}, 207: {1: 1.69218106344e-06, 2: 0.277363300137}, 208: {1: 2.09956887178e-06, 2: 0.285838747635}, 209: {1: 2.68256612731e-07, 2: 0.211890792482}, 210: {1: 1.47301043354e-06, 2: 0.802495129318}, 211: {1: 3.82691779191e-06, 2: 0.814028569912}, 212: {1: 6.71107109288e-06, 2: 0.275516715583}, 213: {1: 1.00596730095e-06, 2: 0.654788033668}, 214: {1: 3.0613085843e-07, 2: 0.15}, 215: {1: 1.66632022455e-06, 2: 0.85}, 216: {1: 3.08220797994e-06, 2: 0.85}, 217: {1: 3.33635498928e-06, 2: 0.29918211991}, 218: {1: 1.90008203665e-07, 2: 0.85}, 219: {1: 3.32049850362e-05, 2: 0.15}, 220: {1: 5.6308839472e-05, 2: 0.85}, 221: {1: 9.43695879039e-07, 2: 0.15}, 222: {1: 3.11730517483e-05, 2: 0.392700518419}, 223: {1: 1.5771216231e-06, 2: 0.85}, 224: {1: 4.19320874302e-05, 2: 0.85}, 225: {1: 1.9807710649e-06, 2: 0.15}, 226: {1: 0.0184210526316, 2: 0.15}, 227: {1: 0.14, 2: 0.15}, 228: {1: 0.0137254901961, 2: 0.15}, 229: {1: 0.175, 2: 0.15}, 230: {1: 0.0875, 2: 0.15}, 231: {1: 0.0291666666667, 2: 0.15}, 232: {1: 0.175, 2: 0.15}, 233: {1: 0.0636363636364, 2: 0.15}, 234: {1: 0.116666666667, 2: 0.15}, 235: {1: 0.0127272727273, 2: 0.15}, 236: {1: 0.116666666667, 2: 0.15}, 237: {1: 0.0777777777778, 2: 0.15}, 238: {1: 0.233333333333, 2: 0.15}, 239: {1: 0.7, 2: 0.15}, 240: {1: 0.35, 2: 0.15}, 241: {1: 0.7, 2: 0.15}, 242: {1: 0.7, 2: 0.15}, 243: {1: 0.7, 2: 0.15}, 244: {1: 0.0538461538462, 2: 0.15}, 245: {1: 0.35, 2: 0.15}, 246: {1: 0.7, 2: 0.15}, 247: {1: 0.35, 2: 0.15}, 248: {1: 0.7, 2: 0.15}, 249: {1: 0.175, 2: 0.15}, 250: {1: 0.35, 2: 0.15}, 251: {1: 0.7, 2: 0.15}, 252: {1: 0.7, 2: 0.15}, 253: {1: 0.35, 2: 0.15}, 254: {1: 0.7, 2: 0.15}, 255: {1: 0.7, 2: 0.15}, 256: {1: 0.7, 2: 0.15}, 257: {1: 0.175, 2: 0.15}, 258: {1: 0.7, 2: 0.15}, 259: {1: 0.35, 2: 0.15}, 260: {1: 0.35, 2: 0.15}, 261: {1: 0.35, 2: 0.15}, 262: {1: 0.35, 2: 0.15}, 263: {1: 0.35, 2: 0.15}, 264: {1: 0.233333333333, 2: 0.15}, 265: {1: 0.07, 2: 0.15}, 266: {1: 0.35, 2: 0.15}, 267: {1: 0.7, 2: 0.15}, 268: {1: 0.233333333333, 2: 0.15}, 269: {1: 0.7, 2: 0.15}, 270: {1: 0.175, 2: 0.15}, 271: {1: 0.35, 2: 0.15}, 272: {1: 0.7, 2: 0.15}, 273: {1: 0.35, 2: 0.15}, 274: {1: 0.35, 2: 0.15}, 275: {1: 0.35, 2: 0.15}, 276: {1: 0.116666666667, 2: 0.15}, 277: {1: 0.7, 2: 0.15}, 278: {1: 0.233333333333, 2: 0.15}, 279: {1: 0.7, 2: 0.15}, 280: {1: 0.7, 2: 0.15}, 281: {1: 0.0777777777778, 2: 0.15}, 282: {1: 0.7, 2: 0.15}, 283: {1: 0.35, 2: 0.15}, 284: {1: 0.7, 2: 0.15}, 285: {1: 0.0875, 2: 0.15}, 286: {1: 0.14, 2: 0.15}, 287: {1: 0.0291666666667, 2: 0.15}, 288: {1: 0.35, 2: 0.15}, 289: {1: 0.233333333333, 2: 0.15}, 290: {1: 0.233333333333, 2: 0.15}, 291: {1: 0.175, 2: 0.15}, 292: {1: 0.7, 2: 0.15}, 293: {1: 0.14, 2: 0.15}, 294: {1: 0.7, 2: 0.15}, 295: {1: 0.233333333333, 2: 0.15}, 296: {1: 0.7, 2: 0.15}, 297: {1: 0.7, 2: 0.15}, 298: {1: 0.35, 2: 0.15}, 299: {1: 0.7, 2: 0.15}, 300: {1: 0.7, 2: 0.15}, 301: {1: 0.175, 2: 0.15}, 302: {1: 0.7, 2: 0.15}, 303: {1: 0.7, 2: 0.15}, 304: {1: 0.233333333333, 2: 0.15}, 305: {1: 0.7, 2: 0.15}, 306: {1: 0.175, 2: 0.15}, 307: {1: 0.35, 2: 0.15}, 308: {1: 0.7, 2: 0.15}, 309: {1: 0.233333333333, 2: 0.15}, 310: {1: 0.14, 2: 0.15}, 311: {1: 0.14, 2: 0.15}, 312: {1: 0.1, 2: 0.15}, 313: {1: 0.14, 2: 0.15}, 314: {1: 0.116666666667, 2: 0.15}, 315: {1: 0.14, 2: 0.15}, 316: {1: 0.233333333333, 2: 0.15}, 317: {1: 0.0875, 2: 0.15}, 318: {1: 0.05, 2: 0.15}, 319: {1: 0.0636363636364, 2: 0.15}, 320: {1: 0.0636363636364, 2: 0.15}, 321: {1: 0.0291666666667, 2: 0.15}, 322: {1: 0.0388888888889, 2: 0.15}, 323: {1: 0.0175, 2: 0.15}, 324: {1: 0.00886075949367, 2: 0.15}, 325: {1: 0.00933333333333, 2: 0.15}, 326: {1: 0.00823529411765, 2: 0.15}, 327: {1: 0.116666666667, 2: 0.15}, 328: {1: 0.0777777777778, 2: 0.15}, 329: {1: 0.0777777777778, 2: 0.15}, 330: {1: 0.35, 2: 0.15}, 331: {1: 0.0636363636364, 2: 0.15}, 332: {1: 0.175, 2: 0.15}, 333: {1: 0.175, 2: 0.15}, 334: {1: 0.35, 2: 0.15}, 335: {1: 0.233333333333, 2: 0.15}, 336: {1: 0.175, 2: 0.15}, 337: {1: 0.175, 2: 0.15}, 338: {1: 0.116666666667, 2: 0.15}, 339: {1: 0.00933333333333, 2: 0.15}, 340: {1: 0.00729166666667, 2: 0.15}, 341: {1: 0.00546875, 2: 0.15}, 342: {1: 0.00269230769231, 2: 0.15}, 343: {1: 0.00333333333333, 2: 0.15}, 344: {1: 0.00263157894737, 2: 0.15}, 345: {1: 0.0538461538462, 2: 0.15}, 346: {1: 0.07, 2: 0.15}, 347: {1: 0.07, 2: 0.15}, 348: {1: 0.0155555555556, 2: 0.15}}, 'weights': {1: 0.372787572493, 2: 39.6491049566, 3: 24.3457064618, 4: 7.74672619262, 5: 47.1302586693, 6: -0.811080653887, 7: -8.35287489766, 8: -16.6498261938, 9: -5.5694029189, 10: 50.0542912379, 11: 27.9904995735, 12: 29.2616667362, 13: 5.63152167803, 14: 6.58512112819, 15: 1.49474302728, 16: -17.2997094408, 17: 37.799431957, 18: -36.815508153, 19: 22.4767169151, 20: 6.66366817094, 21: 2.4230494126, 22: 6.95225594101, 23: 5.37105185521, 24: -35.9324649864, 25: 73.0990594434, 26: 12.088102019, 27: 41.6356254966, 28: -31.0269547812, 29: 15.8296874463, 30: 17.725327324, 31: 52.2722264002, 32: 6.69398241178, 33: -48.2379438356, 34: 2.3420559966, 35: -51.0783422926, 36: 12.4125638422, 37: 22.7529823707, 38: 70.3630810304, 39: 35.8958917367, 40: 4.83729392555, 41: 3.70735359018, 42: 1.55954079629, 43: -0.791649241445, 44: 41.2943893361, 45: -13.0531548934, 46: 34.3373906421, 47: -12.5368910247, 48: 2.58107539331, 49: 9.72944903764, 50: 2.81649623733, 51: -2.07381055098, 52: 18.8314791207, 53: 2.85079148475, 54: 20.6793227948, 55: 3.34791788425, 56: 2.36073806401, 57: -20.0660656305, 58: 2.02632026839, 59: 42.4111513769, 60: -1.48331862535, 61: -2.31200802777, 62: 45.0662031786, 63: -23.47606746, 64: -24.7035176625, 65: -15.8905125999, 66: 2.06638030708, 67: -13.048552607, 68: -0.189707049464, 69: -12.8865422327, 70: 0.571597154288, 71: 0.695984173487, 72: 2.64639844178, 73: 2.63834135115, 74: 67.7969909395, 75: 16.3579221849, 76: 5.95565968814, 77: -1.0457451819, 78: -0.172672745233, 79: 11.9930312726, 80: -56.152028807, 81: -18.5313206237, 82: -0.857870193932, 83: 2.1088287201, 84: 0.557638539812, 85: -29.2627491148, 86: 1.97866800354, 87: -7.57381089454, 88: -92.8393257936, 89: 0.650099950464, 90: 13.4583534999, 91: -15.9999500957, 92: 2.28923528778, 93: 14.9575701879, 94: -4.22661397524, 95: -1.45992925468, 96: -13.6702416137, 97: 3.17345942477, 98: 1.38010735306, 99: 10.8089219468, 100: 5.150563515, 101: -9.42767080586, 102: -4.17816355988, 103: 22.8886875382, 104: 0.567657885613, 105: 0.680198074071, 106: -69.3174468134, 107: 2.28236876842, 108: 0.718107517945, 109: 17.0787549788, 110: -26.8934598127, 111: -4.00611673694, 112: -3.9043293874, 113: 5.69598486626, 114: 42.6182776668, 115: -100.0, 116: 4.67927419996, 117: 46.3165575976, 118: 17.6051857515, 119: -92.3089361089, 120: 32.2430327342, 121: 44.6614443362, 122: 100.0, 123: 23.0537705361, 124: 44.9628845751, 125: 37.3536040098, 126: 0.744903166244, 127: 3.23164966297, 128: 0.18214736417, 129: -6.28839518029, 130: 0.452106706431, 131: -3.98512540507, 132: 2.12482550277, 133: 3.61467999926, 134: -2.06020113462, 135: 5.85793220404, 136: 20.8957082404, 137: 2.73567568042, 138: -27.1077375782, 139: 0.2059030882, 140: -0.459995764105, 141: 64.125941923, 142: -27.5852035187, 143: 26.6746723294, 144: -0.893864239335, 145: 20.4348907674, 146: 0.54912194893, 147: 0.399892166035, 148: -29.8546858658, 149: 0.455318408222, 150: -6.8169988469, 151: 3.65576625127, 152: 0.277891527931, 153: 0.386484237693, 154: 0.407095748791, 155: 0.353513655098, 156: -0.802181391109, 157: -21.7356425659, 158: 0.671688380562, 159: 1.13703312071, 160: 1.28058640254, 161: 36.0073428835, 162: 19.3757808, 163: 57.7438946432, 164: -10.4404288519, 165: 55.9101209247, 166: -58.4654342247, 167: -24.9444922586, 168: -8.73596172606, 169: -1.89203118625, 170: -4.16677378993, 171: 3.77231434427, 172: 1.45755333506, 173: 4.34379011094, 174: 0.347059933056, 175: 6.83288022382, 176: -35.5797272802, 177: -25.5870625149, 178: -0.698377086395, 179: 0.475745498187, 180: 0.360856693655, 181: 0.732893593015, 182: -0.24560444371, 183: 5.50553095862, 184: -44.7876064946, 185: 9.66530558693, 186: -49.5954674537, 187: -59.4027568768, 188: -3.22948564867, 189: 17.7579464821, 190: -9.70806382433, 191: 0.314464106031, 192: -46.4312005814, 193: 0.859790588259, 194: 5.92040134487, 195: -4.22372780105, 196: 3.70298470096, 197: 4.81330404568, 198: 5.020999933, 199: 0.380991878766, 200: 1.11301416426, 201: 0.470427008808, 202: -89.6104961959, 203: 0.388754779246, 204: 0.480671490526, 205: 8.23122532221, 206: 0.421768655174, 207: -51.1462341131, 208: 0.501407223155, 209: 0.351528541306, 210: 0.089465333566, 211: 0.713558451958, 212: -5.23368602329, 213: -13.7377818432, 214: 22.9171138372, 215: 0.0212017305625, 216: 2.1794688797, 217: -63.8480808213, 218: 26.9594543141, 219: 7.00234497664, 220: -46.6446952083, 221: 7.39822622599, 222: -39.8465335502, 223: -20.0447026414, 224: 0.450917683278, 225: -9.93992008844, 226: 1.57290230232, 227: 50.2902658865, 228: 31.3699755592, 229: 0.152262590012, 230: -9.29858550063, 231: -6.11690807733, 232: -1.9128205814, 233: -0.940581209736, 234: 56.0794975906, 235: -9.41465151566, 236: 24.3654716448, 237: 33.3009503174, 238: -6.33826194602, 239: 2.08704491039, 240: -3.9386747862, 241: 0.586028934974, 242: 2.19098853468, 243: 71.115240805, 244: -16.8493401291, 245: -41.3970963441, 246: -27.8623556893, 247: 42.8646749693, 248: -100.0, 249: -23.6212088925, 250: -42.7914661033, 251: 30.6670674873, 252: -61.8030344474, 253: -24.4517851267, 254: -8.17930289056, 255: 4.49703257831, 256: 2.42079142636, 257: -1.3465198395, 258: 2.32719662118, 259: -6.74336469735, 260: 2.86459093225, 261: 8.87895157905, 262: 2.42721477067, 263: -2.20594846608, 264: -0.276774747573, 265: -0.938485375083, 266: -16.5715755745, 267: 25.254994673, 268: -23.003240902, 269: -27.8287462674, 270: 40.0419030767, 271: -12.419212139, 272: 38.08880439, 273: -13.0979309859, 274: 12.6941491451, 275: 28.0414130776, 276: -12.7850973309, 277: -38.619196745, 278: 44.0793130686, 279: -12.3520703784, 280: 15.0468257983, 281: -20.9001150467, 282: 21.7096583648, 283: 14.0402061003, 284: -57.9184680318, 285: 67.4624253732, 286: -100.0, 287: 1.68345674871, 288: 25.7141031191, 289: -0.651482052672, 290: 35.6341606545, 291: 33.5948626764, 292: 6.1038403364, 293: 14.2245536273, 294: 64.408520308, 295: 38.4982108535, 296: 20.0421018783, 297: 14.0731820348, 298: -34.5831681587, 299: -2.31324782026, 300: -38.3714831659, 301: -21.5172036814, 302: -3.62978598827, 303: -1.9258209496, 304: 21.6668655068, 305: 2.27332642961, 306: -1.37018116891, 307: 23.0074236227, 308: 1.12096524146, 309: 0.93511830413, 310: -3.79228155657, 311: 13.257535701, 312: 2.34821863923, 313: 2.41414861355, 314: -7.51601258165, 315: 0.387173007891, 316: 79.4280779912, 317: 2.16433714302, 318: 2.37801487347, 319: -29.191438094, 320: -2.20635658342, 321: -9.09884278236, 322: -15.6135710988, 323: 17.0037886379, 324: 4.46363081997, 325: -16.3905355774, 326: 8.90719883659, 327: 40.7040708814, 328: -70.5355666816, 329: -10.88659528, 330: -20.0752595164, 331: 20.4397029268, 332: -4.90111011498, 333: 0.772520244677, 334: -5.76324660945, 335: -14.2177111437, 336: 15.0117552702, 337: -59.9888373353, 338: -39.921058303, 339: 0.315545846894, 340: -26.1132739509, 341: -2.15237728235, 342: -15.3141797916, 343: 29.4691921722, 344: -27.5412650768, 345: -3.02073259599, 346: -60.1013800526, 347: 37.1284603981, 348: -50.4179316265, 349: -42.265223023, 350: -0.0120688125554, 351: -0.472606301018, 352: -0.647567583498, 353: 3.17411335251, 354: -9.83008782344, 355: 1.76918718251, 356: 3.68820668995, 357: -4.67023208809, 358: -3.65693903784, 359: 1.4220086917, 360: 8.82530363437, 361: -1.23435311069, 362: -6.35459402645, 363: 2.43892316181, 364: 0.375810358766, 365: -3.9367575386, 366: -10.9758351986, 367: 0.950891671329, 368: -1.19903436903, 369: -10.2498406597, 370: 2.794696636, 371: -0.86617609985, 372: -5.93672056279, 373: 3.56522570014, 374: -8.15719774807, 375: 0.214659641612, 376: 3.06262237244, 377: -0.00806953230368, 378: -2.13743390979, 379: -7.81860714048, 380: -5.92490864796, 381: 8.01271543385, 382: 5.0856153436, 383: 1.17147923457, 384: 3.66009398766, 385: -0.421708692692, 386: -5.64463895573, 387: 4.88140143081, 388: -3.30498655202, 389: -7.21199388319, 390: -3.35074305093, 391: 1.08760878403, 392: -0.152830744046, 393: -3.6520530766, 394: 5.53545192293, 395: -1.23318601323, 396: 0.333795554927, 397: -1.98001225516, 398: -1.52477644223, 399: 0.489809961998, 400: 1.91258129949, 401: -4.2557814537, 402: 1.82702499562, 403: -1.22591678181, 404: -7.01722011837, 405: 0.840887741506, 406: -2.71514029315, 407: 3.40534008257, 408: -6.80703468084, 409: -10.451488768, 410: -0.391702096454, 411: -4.01294783679, 412: 3.33606940504, 413: -0.939147575202, 414: 6.77793035383, 415: -3.95958929761, 416: -8.01540866566, 417: 0.116346607986, 418: -2.32554369301, 419: 0.421812358236, 420: -3.56073205058, 421: -12.29579147, 422: 3.8837196718, 423: -9.88829716268, 424: -6.55687798834, 425: 5.0762626317, 426: 0.994042476559, 427: 0.57312375361, 428: -13.0204475834, 429: -7.31485001069, 430: 4.48036086865, 431: 15.3800742886, 432: 1.4601612843, 433: -1.51756268405, 434: 6.9392423319, 435: -3.95109781947, 436: 17.2666742377, 437: 9.92393012364, 438: -7.78587249274, 439: -4.76625831797, 440: 5.09140531093, 441: 1.75445755099, 442: 0.250821570585, 443: -2.66057152028, 444: -1.03603106573, 445: 1.27434739026, 446: 8.64357898431, 447: 0.831583446858, 448: 5.5769659836, 449: 2.75281188062, 450: -1.03784952337, 451: 1.39785298934, 452: -10.6365780731, 453: 0.680352551352, 454: 1.43550868143, 455: 8.97141654647, 456: 1.42676009512, 457: 2.99665521424, 458: -6.80942502872, 459: 5.83425202615, 460: 0.388626873927, 461: -5.64932032782, 462: 10.0234107082, 463: 1.63765038195, 464: -1.99576143441, 465: 3.54842683808, 466: -0.974461012847, 467: 0.0108274125961, 468: -12.8660931549, 469: -10.3284306842, 470: -0.561006399269, 471: -3.94183516494, 472: 0.966950725893, 473: -4.79494100032, 474: -2.80132529768, 475: 0.124578484501, 476: 0.32173658596, 477: -4.6126077014, 478: -4.62630192896, 479: -0.133684944009, 480: 1.21655926611, 481: 0.639977697692, 482: -0.500238048061, 483: 2.92683492065, 484: -1.96336855138, 485: -4.69651537703, 486: -4.80912879093, 487: -9.34427907116, 488: -0.402472629035, 489: -0.069611343939, 490: -5.57427887207, 491: 0.153689122662, 492: 3.66833384931, 493: -0.182309907213, 494: -8.36446170434, 495: 0.350522244433, 496: 0.0310987442919, 497: -3.50199490031, 498: 0.0877205596232, 499: 1.61943452655, 500: -3.61487312679, 501: 0.524533296866, 502: 0.193878676411, 503: 0.232183225546, 504: 0.159476353693, 505: -0.438950228729, 506: 10.1655043122, 507: 0.965557557547, 508: 0.510112851718, 509: 0.464897884213, 510: -3.00138795673, 511: 0.206555773423, 512: 2.20557997257, 513: 1.05170872555, 514: -8.81684846471, 515: 9.87148559895, 516: -1.24857374247, 517: 2.25282394356, 518: 7.09700402522, 519: -1.93778190202, 520: 7.57185582685, 521: 0.291334054777, 522: 10.963835234, 523: 0.103363844713, 524: 1.29817287193, 525: 8.36426156481, 526: -12.726317775, 527: -0.976288706592, 528: 0.0684956746013, 529: 0.545400658329, 530: 0.482779494863, 531: -0.0204598755843, 532: 0.664242255768, 533: 2.06010024082, 534: 0.105650159355, 535: 15.6172695663, 536: 12.7254690372, 537: 7.57108467998, 538: -0.605547843217, 539: -2.03156547106, 540: 7.48932528702, 541: -6.32537510689, 542: 0.0818480953265, 543: -10.8697256453, 544: 0.180432895064, 545: -0.296006415438, 546: -0.510342052801, 547: 2.03547176573, 548: 0.0734999571206, 549: 1.76723016848, 550: 0.4868981007, 551: -8.02264437271, 552: 0.121630849611, 553: 0.200958662364, 554: 11.7269898765, 555: 1.14947468676, 556: -7.22010302658, 557: 0.268201295996, 558: 0.128843371923, 559: -1.50639886649, 560: 0.494833201972, 561: 0.00815222451478, 562: 11.0081861223, 563: -2.49660292146, 564: -0.519760233472, 565: 1.03423543691, 566: -2.27927552213, 567: 1.21858202844, 568: -2.36217397464, 569: 6.72447995893, 570: -2.38427936395, 571: -2.65898722853, 572: -6.34005071193, 573: 6.32034308617, 574: 2.53725887467, 575: 0.685247825129, 576: -3.41768321652, 577: -0.292661168469, 578: -2.71753987542, 579: -0.295089121218, 580: 0.713110470412, 581: -0.77107319801, 582: -4.52341209609, 583: -4.28647781186, 584: 3.10363695792, 585: -3.25624361625, 586: 3.30598805422, 587: 4.02038913151, 588: 5.73729795297, 589: 1.17443774716, 590: -5.37086282726, 591: 7.12865214596, 592: -3.37147516719, 593: -2.52661696746, 594: 8.82902729344, 595: -0.604328484161, 596: 1.13618950553, 597: 4.72615156081, 598: -4.02992329009, 599: -5.33634823256, 600: -4.7125920677, 601: -3.66150686422, 602: 6.41221717122, 603: 6.33214770636, 604: 1.2943202561, 605: 2.41839995496, 606: 1.77375998055, 607: 1.58555810419, 608: 1.86558749643, 609: 3.34313835416, 610: -0.220014524187, 611: 1.45543552319, 612: 2.01390812113, 613: 5.15280927245, 614: -0.541369518124, 615: -5.28863995703, 616: -8.73432710889, 617: 8.34285576306, 618: 13.6578640719, 619: -3.04765369466, 620: -0.712625327126, 621: -1.71130910723, 622: 10.9870644309, 623: -5.85801087064, 624: -13.1659060823, 625: 0.426228742295, 626: 4.87458136548, 627: 6.13517620653, 628: 1.07243786678, 629: -0.771988882467, 630: -0.218247142114, 631: -3.54902534465, 632: 1.11564271354, 633: 7.4875204654, 634: 8.26777038945, 635: 7.0613254197, 636: -0.779179516821, 637: -4.66603431636, 638: -0.958351538216, 639: 3.42642988335, 640: 1.68307314444, 641: 3.96484313698, 642: -9.75887070595, 643: -12.2371368002, 644: -7.09811867192, 645: 1.13732807015, 646: -5.25324982379, 647: 2.12583169471, 648: 9.66603394189, 649: 14.5038973774, 650: 3.73059693708, 651: 5.05117827219, 652: 4.89107702293, 653: 3.05607428707, 654: -6.29812524688, 655: -0.967295302553, 656: 1.13037199165, 657: 1.56047334092, 658: 1.70289587362, 659: -0.949985341129, 660: 7.81584046453, 661: 1.60595507376, 662: 2.10882471989, 663: -6.6826909672, 664: 1.28275323047, 665: 4.19885871286, 666: 1.34448279781, 667: 1.67499089354, 668: -3.48827755205, 669: -0.656062007815, 670: -0.23618245651, 671: 6.66933449596, 672: -15.015904403, 673: 12.1791817232, 674: -9.48358864988, 675: 10.5488532009, 676: -4.6747731748, 677: -8.46525001058, 678: -11.0301450884, 679: 4.83989744315, 680: -10.2255408524, 681: -2.16240804953, 682: 8.77752517341, 683: 7.93257092627, 684: -2.08048124754, 685: 4.4652996187, 686: -1.49498416578, 687: -15.0215402511, 688: 1.33794246496, 689: 2.30389344731, 690: -3.806785156, 691: 17.3763165376, 692: 2.06689867316, 693: -10.3521602085, 694: -3.37606653473, 695: 20.024864174, 696: 1.4121925027, 697: 3.32705886788, 698: 0.804203999036, 699: -1.38780096485, 700: -38.939921734, 701: 33.7950030561, 702: -1.80600221726, 703: -0.487387486405, 704: 12.1035886674, 705: -6.47891798079, 706: 36.939710885, 707: 9.94513969194, 708: -15.7213624434, 709: -8.92383652822, 710: 48.6892186434, 711: 92.9985171405, 712: 18.6000484684, 713: 15.9366017693, 714: 13.0158804885, 715: -36.3024737147, 716: -17.8019826849, 717: 2.78730545831, 718: 39.7320953127, 719: -1.22866236338, 720: 28.5777888124, 721: -0.674200819777, 722: 13.7686893617, 723: -16.1346595381, 724: -16.3622581629, 725: -24.0838612854, 726: -9.0595967844, 727: -21.3398226014, 728: -2.78507952587, 729: -2.84899232462, 730: 14.017773932, 731: 11.9977432928, 732: 0.0286466735669, 733: 9.17324721641, 734: 6.9450576235, 735: 6.11626172733, 736: 19.2650891897, 737: -3.8384746562, 738: -0.224051429164, 739: 43.5790709716, 740: -4.88516205385, 741: 9.23091529571, 742: 8.74266466768, 743: -0.919230907572, 744: -2.76503520556, 745: 4.37563820828, 746: -4.90640392884, 747: -1.04281711492, 748: -0.693523800785, 749: 8.76765102825, 750: 19.7611578501, 751: 5.86368382451, 752: 20.8044132973, 753: 3.84697102926, 754: -3.75844919524, 755: 1.54882995222, 756: -5.03925481908, 757: 15.93340903, 758: 49.780041811, 759: -6.9182574279, 760: -13.2428153387, 761: -3.5450534629, 762: 5.17547347543, 763: -27.953858007, 764: -3.38666497759, 765: 1.86757093033, 766: -4.78511215898, 767: 12.707297327, 768: -5.15533183114, 769: -5.27598736096, 770: -5.10662324568, 771: -5.0978264185, 772: -47.3027587887, 773: -4.75532152954, 774: 6.73966676333, 775: 7.03126368624, 776: -5.09240745901, 777: -12.4652097972, 778: -6.64340175517, 779: -0.680020991901, 780: 43.6802998069, 781: -5.21221112005, 782: -4.55843362725, 783: -26.5153632609, 784: -4.69774449828, 785: 4.30666076698, 786: 36.223679464, 787: -5.22809655422, 788: 0.20794517841, 789: -5.47355825545, 790: -5.10283088651, 791: 3.32909319686, 792: -5.13662444838, 793: -4.28725675343, 794: -5.04968088932, 795: -5.25125255809, 796: -5.18269934536, 797: -6.44647096893, 798: -5.05947407659, 799: 41.2838313567, 800: -5.57989198485, 801: 26.1015911138, 802: -3.25327049253, 803: -5.29803324176, 804: 47.9470211801, 805: -5.02721495664, 806: -3.35159769823, 807: -5.16262905307, 808: 20.138051524, 809: 32.8667132236, 810: 17.6189614647, 811: -10.4054582865, 812: -4.95785485338, 813: 1.52260962855, 814: 11.2505364154, 815: -29.1787396043, 816: -0.00619411972585, 817: 45.0511916667, 818: 80.2811277912, 819: 56.1349122962, 820: -7.20087420904, 821: -9.31625664722, 822: 12.0339801204, 823: -40.134895798, 824: -1.77053326641, 825: -1.98768555526, 826: 5.12976236105, 827: -10.795447305, 828: -1.48074101142, 829: -5.39904388324, 830: -1.68622225027, 831: -29.0165743791, 832: -4.30434327896, 833: 2.37242319191, 834: -2.19246560356, 835: -1.92834318597, 836: -4.9096209435, 837: -2.17928737633, 838: 5.39782117283, 839: -20.4931472703, 840: -42.219565581, 841: -31.8539527926, 842: -0.209558822152, 843: 5.71524182093, 844: -1.32287489083, 845: -1.39654399161, 846: -30.4826554094, 847: -1.57120179865, 848: -4.00059533191, 849: 0.982322184, 850: -1.07901021394, 851: -1.33703851425, 852: -1.4973923606, 853: -1.40371493164, 854: 0.499772073454, 855: -2.90891957305, 856: -2.65969852179, 857: -1.41259269033, 858: -2.11339476646, 859: -3.08972751476, 860: -5.69601284961, 861: -25.0003890163, 862: 0.159612946864, 863: 42.4777159486, 864: -67.8993595621, 865: 1.81973258175, 866: -17.0230098424, 867: 3.04200707761, 868: -4.79449680578, 869: -1.9813387104, 870: -1.29292676814, 871: -14.6941145643, 872: -1.33549073526, 873: -28.7522972815, 874: -45.0937680992, 875: -2.72501834371, 876: -1.05467414603, 877: -1.54037962819, 878: -1.40623535738, 879: -1.33127288416, 880: -1.77972199707, 881: 4.0149702119, 882: -2.03072022679, 883: 15.5135180459, 884: 4.05251797276, 885: 30.5126208373, 886: 3.4100148155, 887: 25.3315423527, 888: -1.53716436545, 889: -5.10579204765, 890: 4.96397797461, 891: -1.42027900341, 892: -26.1673636351, 893: 2.43512269167, 894: -0.997864769457, 895: -1.62146357415, 896: 23.36785547, 897: -1.33326929148, 898: -5.11701971533, 899: -1.54177148355, 900: -15.3064925846, 901: -1.43922795964, 902: -1.50687169171, 903: -1.17613260323, 904: -5.10742386618, 905: 9.35891938297, 906: -1.3517612441, 907: -1.50276025361, 908: -0.532864094248, 909: -1.49771839737, 910: -4.64586715262, 911: 2.99953017743, 912: 2.99863205905, 913: -1.18959676846, 914: -0.109095358292, 915: -7.36079483652, 916: 6.22402887127, 917: -13.7941518983, 918: 6.98058731202, 919: 8.46722114206, 920: -0.813218035479, 921: 8.06093925101, 922: -4.41924797068, 923: -3.87965002063, 924: -1.36428429519, 925: -25.2435735013, 926: -4.2449529028, 927: -1.66343284724, 928: 0.190089833796, 929: 22.5982146478, 930: -1.49182020101, 931: -11.6142876236, 932: -1.15879603554, 933: 5.96458989485, 934: 3.92257458049, 935: -16.0428930808, 936: 7.2654055936, 937: -5.25756613893, 938: -5.43141035501, 939: 7.63179916225, 940: 8.20581976466, 941: -20.6608692003, 942: -26.9855445825, 943: -7.52526611963, 944: 6.61243710821, 945: -7.00786847465, 946: 69.1671649659, 947: 3.07007987703, 948: 62.8750182452, 949: -24.9228727991, 950: 30.690785537, 951: 0.693122495255, 952: -7.69615408779, 953: -5.58768053643, 954: -5.14306502245, 955: 38.3486935966, 956: -4.55200169517, 957: -4.0106181205, 958: 1.04535475704, 959: -1.24604533132, 960: -5.07661892638, 961: -3.25436290066, 962: -3.75502573087, 963: 0.468914432559, 964: -2.31536229626, 965: -0.690095080682, 966: -4.41571394942, 967: -4.51876428105, 968: -2.9666502969, 969: -5.35845160931, 970: -17.2018763567, 971: -12.9167104688, 972: -19.8588285756, 973: -22.5743272117, 974: 67.4935765778, 975: 17.71503687, 976: -30.0069051107, 977: -5.4016695181, 978: -0.0646132992435, 979: -5.55422701099, 980: -2.64375105026, 981: 8.81805256972, 982: -22.9630100163, 983: 28.7087652856, 984: 3.93526535552, 985: -5.04981687729, 986: -5.29921446519, 987: -5.26338875925, 988: -4.50075866979, 989: -4.58787478485, 990: 0.314800791451, 991: -8.54476137565, 992: -31.3533492842, 993: 10.3625222988, 994: -6.32751200483, 995: 13.9208517676, 996: 14.3841948138, 997: -4.90827149289, 998: -0.782727133558, 999: -3.24500774996, 1000: 25.8809691405, 1001: 7.17364199047, 1002: -10.262729899, 1003: -41.4358129829, 1004: -4.35224356687, 1005: 11.0149929295, 1006: -5.04604228257, 1007: -5.03479001173, 1008: -5.15044876246, 1009: -5.00203697599, 1010: -5.13090157711, 1011: -5.22535369658, 1012: 11.5290985681, 1013: -5.17704545595, 1014: 8.25772106608, 1015: -5.13992990544, 1016: -5.23365377762, 1017: 11.9461743326, 1018: -5.11528499698, 1019: -4.16652475907, 1020: 21.3062996759, 1021: -8.11595521369, 1022: -14.2233533839, 1023: 16.9742431125, 1024: -51.8100701328, 1025: 7.74248878898, 1026: 11.3900542855, 1027: 14.5797899525, 1028: 26.3901350211, 1029: 7.81869024283, 1030: -1.69576638356, 1031: -3.79307585074, 1032: -3.62024211624, 1033: 0.913699205286, 1034: -21.5037902913, 1035: -4.64232706681, 1036: 10.1647028225, 1037: -2.55131565859, 1038: -12.7576869687, 1039: -14.0547568652, 1040: -5.52567710622, 1041: -14.6626233172, 1042: 4.55289842905, 1043: 3.44837453436, 1044: -10.4748520498, 1045: 0.449008679503, 1046: 14.5295409818, 1047: 32.5854789079, 1048: 0.672239396833, 1049: 35.1180297687, 1050: -6.87646628841, 1051: -3.14812617933, 1052: 3.83909518896, 1053: -10.1567260456, 1054: 13.6753267569, 1055: -8.35014971506, 1056: -16.8167224989, 1057: -26.1837310828, 1058: -37.0963935386, 1059: -7.89229419784, 1060: 20.3109897718, 1061: 18.7912462917, 1062: -0.726848777491, 1063: -4.48116464539, 1064: 5.33459534956, 1065: -9.05260048983, 1066: 36.9868065354, 1067: -8.30770206639, 1068: -5.62008713717, 1069: -12.1175386284, 1070: 9.3127501289, 1071: 0.466357527504, 1072: -16.3790553421, 1073: -18.4430303292, 1074: -8.20567672228, 1075: 7.99141487715, 1076: 32.830856168, 1077: -3.04125482842, 1078: -7.35191013408, 1079: -15.0068655946, 1080: 1.7092082822, 1081: -2.70756889789, 1082: 7.97409051097, 1083: 11.1113799972, 1084: 4.64044777613, 1085: 18.2214813414, 1086: 13.2419840757, 1087: -2.11565547151, 1088: 20.2146281322, 1089: -8.16102615344, 1090: -8.04188034032, 1091: 0.878775591146, 1092: -19.3578871501, 1093: 4.30547758495, 1094: -18.4479743191, 1095: 10.0377658323, 1096: 6.15349413117, 1097: 1.81831058903, 1098: -26.4972644317, 1099: -46.0441127704, 1100: 5.57776978298, 1101: 13.8862164067, 1102: -5.82344964452, 1103: 3.21908455814, 1104: 22.8370034768, 1105: 3.22073400979, 1106: 29.2273585888, 1107: 6.13429949791, 1108: 19.9490225542, 1109: -3.79889121445, 1110: 18.1450163431, 1111: -15.6328259046, 1112: 10.6168603307, 1113: -7.3184938507, 1114: -3.27998180892, 1115: 1.48339235033, 1116: -6.70745378397, 1117: 4.56066577555, 1118: 4.4553316524, 1119: 4.64423964486, 1120: 5.76652511078, 1121: -37.0558969583, 1122: 2.88613449954, 1123: 3.48299005063, 1124: -5.55070886589, 1125: 4.01160330356, 1126: 20.3425576768, 1127: -0.127237371341, 1128: 4.28004533394, 1129: 2.86368835155, 1130: 3.56714535715, 1131: -9.64975630256, 1132: 2.17698522732, 1133: -1.99629130502, 1134: -1.44964525436, 1135: 14.5775878433, 1136: 1.8639937118, 1137: 2.97446618325, 1138: -1.78842770243, 1139: 5.44043664989, 1140: -46.702529063, 1141: 3.47808163046, 1142: 0.716988864731, 1143: 3.04932843546, 1144: -5.75253647022, 1145: 4.58083415348, 1146: 4.75094677557, 1147: 2.81891172491, 1148: 1.02995422177, 1149: -1.34038652908, 1150: 14.1177729581, 1151: -4.51551182103, 1152: -6.66309529827, 1153: -6.00100269456, 1154: 1.5832027007, 1155: 5.85688768978, 1156: 0.785938316159, 1157: -26.9813373139, 1158: -16.5337925781, 1159: -2.26230623149, 1160: 72.5262062053, 1161: -6.84825336874, 1162: -25.9431386957, 1163: 2.83694526793, 1164: -14.3167743601, 1165: 28.3803228228, 1166: -17.1165026046, 1167: 18.0969608587, 1168: 23.4018836338, 1169: -25.9777180347, 1170: 3.65451519046, 1171: -18.7585761738, 1172: 20.8471346795, 1173: 0.942203446284, 1174: 2.70369645161, 1175: 0.515358741377, 1176: 27.7911309233, 1177: 1.48746750148, 1178: 3.07554197891, 1179: 1.08462336185, 1180: -21.3158474049, 1181: -0.0280790020955, 1182: 7.1395285722, 1183: -5.27548471898, 1184: 0.614675267091, 1185: 31.6286552711, 1186: 1.64511747615, 1187: -1.87830128189, 1188: 4.31208309453, 1189: 16.9179090717, 1190: 30.3253812332, 1191: 2.10338288547, 1192: -6.36460619444, 1193: 1.09272531641, 1194: 0.754028852966, 1195: 0.804265836969, 1196: 1.16663090602, 1197: 3.69956358814, 1198: 1.63922369007, 1199: 1.2039877066, 1200: 0.878921470325, 1201: 1.26157979599, 1202: 0.901610617094, 1203: 0.914106890392, 1204: -25.3150649563, 1205: 2.93286129792, 1206: 2.02281796112, 1207: 0.661691677504, 1208: -1.25195392816, 1209: -7.20315639636, 1210: -25.5733614062, 1211: -2.13788709377, 1212: -47.8137395463, 1213: -13.4232078867, 1214: -5.41112221924, 1215: 1.42632518509, 1216: 7.03238910759, 1217: 6.66217842038, 1218: -26.1975983632, 1219: 1.15121094851, 1220: 15.1264744989, 1221: 0.842025946483, 1222: -10.4633475491, 1223: -17.0460282818, 1224: -12.564681471, 1225: 0.868349170371, 1226: 0.934004067569, 1227: 1.31664048247, 1228: 1.25077300118, 1229: 1.19091327581, 1230: -7.4089684768, 1231: 11.8748664867, 1232: -6.47437772636, 1233: -39.8825909485, 1234: -9.06228145608, 1235: 12.5444269403, 1236: 9.56817914646, 1237: 7.83337088153, 1238: 1.86318560716, 1239: -12.556447823, 1240: 1.24960262995, 1241: 24.9927804701, 1242: -0.980493533377, 1243: 1.10568628955, 1244: -5.21464081422, 1245: 15.8186914545, 1246: 0.758326598333, 1247: 4.29769794703, 1248: 0.934704855203, 1249: -12.6460472688, 1250: 0.816672737917, 1251: 0.881010909806, 1252: -29.2404797206, 1253: 3.63681829234, 1254: 57.3273170328, 1255: 0.991821118798, 1256: 0.797186960384, 1257: -1.58210599857, 1258: 0.921263940993, 1259: 4.30321028653, 1260: -7.42238090988, 1261: -20.8832579717, 1262: -0.661911039301, 1263: 1.26608126112, 1264: -6.86769110121, 1265: -3.40463137947, 1266: 3.91259420893, 1267: -45.6117984545, 1268: -1.25371270321, 1269: -4.58146297868, 1270: 9.75544909105, 1271: -1.77349258992, 1272: -10.3081214749, 1273: 2.90367668855, 1274: 2.17095472384, 1275: 8.0803743379, 1276: 8.03542990549, 1277: 0.756228768517, 1278: -3.37462795856, 1279: 0.87071038533, 1280: -11.1299029393, 1281: -3.92810933065, 1282: -0.450104851333, 1283: 21.763901421, 1284: -0.226879139949, 1285: -9.29295682923, 1286: -13.6848397895, 1287: 3.05737059983, 1288: 8.77538478837, 1289: 10.0507156594, 1290: -27.2581834285, 1291: 8.36407827346, 1292: 9.33725773118, 1293: 0.76454167557, 1294: -7.05222770042, 1295: -11.3115856825, 1296: 0.0968568249199, 1297: 24.5828065977, 1298: 4.83907296912, 1299: 7.74793131382, 1300: -5.35912014615, 1301: -22.5076241255, 1302: -2.0364723578, 1303: 4.57045624173, 1304: 10.9981517202, 1305: -2.7471602516, 1306: 3.70199023191, 1307: 10.9646062521, 1308: 0.865819297213, 1309: 5.76375174367, 1310: 0.238824056457, 1311: 9.53000529273, 1312: 0.709450730783, 1313: 20.4302886349, 1314: 1.89893470705, 1315: -16.5659817223, 1316: -1.59557506776, 1317: -1.39404775549, 1318: 10.8873078745, 1319: 24.1176832197, 1320: -28.4321538824, 1321: -12.8675795145, 1322: 33.0404588795, 1323: -18.7329035721, 1324: -19.3923989515, 1325: -10.8319795112, 1326: 9.88097067523, 1327: -23.6108278581, 1328: -7.40074395988, 1329: -8.8788922687, 1330: -27.2534290737, 1331: -29.3190627276, 1332: 0.909302164945, 1333: 43.4776221213, 1334: 2.51109459467, 1335: -0.383515729873, 1336: 1.34924017604, 1337: 2.15248009847, 1338: 1.05345037694, 1339: -47.7493456871, 1340: 5.03810683579, 1341: -12.4941667199, 1342: 10.2825097167, 1343: -5.67195390112, 1344: -17.0685342454, 1345: -10.9313258167, 1346: 7.59467917191, 1347: 25.3276780347, 1348: -9.95744377266, 1349: -24.9426364657, 1350: 1.9512799192, 1351: -4.00715741833, 1352: -15.909342157, 1353: 0.547514001193, 1354: -34.1603519028, 1355: 4.32189013507, 1356: 4.40830446223, 1357: 3.98357162782, 1358: -30.750275409, 1359: 2.15325520815, 1360: 4.6454475852, 1361: 28.5295000028, 1362: 3.83161222198, 1363: -70.3527660587, 1364: 3.96688478489, 1365: 4.55927056565, 1366: 7.99835803026, 1367: 2.31659843038, 1368: 4.45427540665, 1369: -15.9244177484, 1370: 26.392905381, 1371: 3.63529156318, 1372: -17.4266103727, 1373: -12.963525394, 1374: -1.37921208413, 1375: 2.45005581795, 1376: 16.5165055969, 1377: -6.12383866367, 1378: -27.0309810529, 1379: -23.789864978, 1380: -0.701713460608, 1381: -4.54268262475, 1382: -15.5343325243, 1383: 22.155793087, 1384: -4.06912253738, 1385: 3.05012011934, 1386: 4.64175816145, 1387: -45.2716605387, 1388: 1.46371101501, 1389: 1.03626728208, 1390: -4.1225692372, 1391: -42.1777198928, 1392: -17.2180706874, 1393: 21.1406602027, 1394: -6.78647279489, 1395: 3.29053792529, 1396: 9.79299544213, 1397: 0.708985580112, 1398: 0.684545088415, 1399: 1.94325232876, 1400: -4.20112042308, 1401: 0.590363390297, 1402: -0.0672040505641, 1403: -8.57687960712, 1404: -7.9635140426, 1405: 7.6532236061, 1406: -3.18385760647, 1407: -0.680292395992, 1408: -1.04639324786, 1409: 7.89113116847, 1410: 0.246831856105, 1411: 1.37193461007, 1412: 0.981312504019, 1413: 8.59647867254, 1414: 0.822695422213, 1415: -8.99315302449, 1416: 0.795554690297, 1417: 5.06091761622, 1418: -2.77641811943, 1419: -0.0856823900035, 1420: 3.79448108617, 1421: -2.32515358943, 1422: 3.13814081877, 1423: -10.1642040812, 1424: 0.900258838176, 1425: 0.578188355437, 1426: -0.776356644819, 1427: 4.88785968809, 1428: 4.0170000138, 1429: -5.81563177003, 1430: 2.20452911302, 1431: 0.505146959393, 1432: -4.53290363324, 1433: 0.533707074101, 1434: -1.007670215, 1435: -2.82373642568, 1436: 1.03054601696, 1437: -0.0736426134874, 1438: -0.152293802082, 1439: 1.35473690326, 1440: 5.02614353727, 1441: -2.43634988351, 1442: -2.09813131603, 1443: -4.13334432374, 1444: 5.13784342513, 1445: 1.34879254466, 1446: 2.22405422912, 1447: 2.16585081497, 1448: 2.75310954231, 1449: -0.906950109679, 1450: -5.99307864053, 1451: -1.77737881446, 1452: -0.496250701619, 1453: -8.238933397, 1454: -1.4536404885, 1455: 5.18715949134, 1456: -2.09315894952, 1457: -0.757889461653, 1458: -1.55320790669, 1459: 1.35432538244, 1460: 0.916902359605, 1461: -17.4022278091, 1462: -0.424769411178, 1463: -0.122342769476, 1464: -4.49424080308, 1465: 1.6411762289, 1466: 5.53185831509, 1467: 4.48874921569, 1468: -2.15406423425, 1469: 5.06068399235, 1470: -26.0010548932, 1471: 3.71346658069, 1472: 0.844080448352, 1473: 0.28392201023, 1474: 1.76707604965, 1475: -7.65090962253, 1476: -5.09929795302, 1477: 1.84103127264, 1478: -6.53446964945, 1479: 5.33621588163, 1480: -0.654510663953, 1481: -0.0575404338558, 1482: 1.84755232856, 1483: -16.0152824228, 1484: -12.5200865276, 1485: 0.804259356006, 1486: -6.14792255891, 1487: 0.653499122464, 1488: 5.32474792327, 1489: 1.40842222408, 1490: -2.65527647899, 1491: 3.03465082645, 1492: -3.06644531817, 1493: -9.61287910834, 1494: 0.555747918747, 1495: -7.86129980718, 1496: 5.43766721587, 1497: -10.8016774844, 1498: 2.47302089227, 1499: 0.630093779856, 1500: -1.94572285564, 1501: 3.71811985024, 1502: -1.66654903585, 1503: 4.933196972, 1504: -3.68656365379, 1505: 0.156969253865, 1506: -3.63665933045, 1507: 2.1180883502, 1508: -3.30191237873, 1509: 4.66560689853, 1510: 4.64804030892, 1511: -4.16461044233, 1512: -0.411686020672, 1513: -1.90690655356, 1514: -2.33072829801, 1515: 12.3993440195, 1516: -1.39531536102, 1517: 11.4315749371, 1518: 1.45651645253, 1519: 8.34630965376, 1520: -0.885485692881, 1521: -12.39753938, 1522: 0.858079238104, 1523: 1.43693680688, 1524: 4.27285986546, 1525: 2.03265537008, 1526: 0.879398239549, 1527: 5.19801089522, 1528: 2.17549839477, 1529: -4.62384242176, 1530: -2.21628689873, 1531: 2.15620739529, 1532: -1.76236453736, 1533: 1.45703475832, 1534: 1.45951083789, 1535: -0.625297881737, 1536: 1.43581529148, 1537: 0.942343540062, 1538: -2.13088275382, 1539: -1.84834719956, 1540: -0.721601969714, 1541: 1.65112537333, 1542: 0.75290900017, 1543: 0.823104385388, 1544: 4.29182875106, 1545: 0.636966389672, 1546: 4.48712159508, 1547: 3.46296850733, 1548: 1.09481428044, 1549: 0.640799323235, 1550: 0.951756284354, 1551: 1.06792939624, 1552: 4.54851930867, 1553: 2.9122433644, 1554: 5.00482017179, 1555: 0.850394752951, 1556: 1.27046770545, 1557: -3.61387858441, 1558: 6.96640897945, 1559: 0.471988663308, 1560: 0.470168209542, 1561: 4.00318024486, 1562: 1.28842532753, 1563: -8.56155296421, 1564: 2.05962389017, 1565: -16.9499185877, 1566: 8.91236259468, 1567: 5.4156743794, 1568: 0.87794785845, 1569: 0.774263160901, 1570: 0.807490221206, 1571: 3.49435075785, 1572: -16.1449045833, 1573: 7.97548452856, 1574: 1.58306135372, 1575: 0.736494867127, 1576: 1.03323883768, 1577: 0.745404356288, 1578: 0.987464361441, 1579: -0.426114494311, 1580: -0.672887648957, 1581: 5.62314565843, 1582: 0.713600015653, 1583: -2.95757737592, 1584: -10.1030055035, 1585: 6.53769255924, 1586: -7.65664017731, 1587: -9.28233942554, 1588: 11.9005269605, 1589: 0.952235123313, 1590: 1.12936048007, 1591: -2.19963770465, 1592: -3.39476174812, 1593: 3.37788963831, 1594: 0.5667233938, 1595: 0.739739899799, 1596: 1.85278496907, 1597: 0.906984023923, 1598: 1.73876067677, 1599: 0.701819787247, 1600: 0.755794579483, 1601: -3.82907542897, 1602: 5.08399411199, 1603: 2.5684133823, 1604: 0.840947848286, 1605: 0.838680830791, 1606: -2.11218060704, 1607: 1.05593082913, 1608: 1.47610469162, 1609: -3.44736688974, 1610: 1.28259681827, 1611: 0.777297838661, 1612: 0.768434723485, 1613: 9.53929802183, 1614: -1.07633743219, 1615: 11.7725130556, 1616: -3.59814312969, 1617: -3.25458574109, 1618: -1.23379880406, 1619: 0.802326975522, 1620: 2.2843957988, 1621: 0.757088831843, 1622: 0.621852200324, 1623: -1.5493663245, 1624: 2.26829344265, 1625: 2.33663695538, 1626: -5.95171633467, 1627: 0.691782216841, 1628: 0.0972030941375, 1629: -8.03081781742, 1630: 4.52547193388, 1631: -0.0221809149553, 1632: -15.0723266315, 1633: 1.39000434937, 1634: -3.16963595994, 1635: 2.61845333464, 1636: 5.07410880436, 1637: 1.66421433771, 1638: -0.696750407426, 1639: -3.57103125906, 1640: -0.759480743804, 1641: -1.59128466889, 1642: -2.44047721121, 1643: 6.02636496167, 1644: 5.74018149596, 1645: -6.15118680392, 1646: -2.08635384367, 1647: 4.24414525491, 1648: 4.45945929395, 1649: -4.98792418908, 1650: 3.64343535718, 1651: 1.9125004094, 1652: -0.841710528427, 1653: 4.70821092071, 1654: 5.03409699714, 1655: 4.61247594967, 1656: 6.81380403371, 1657: -2.38013546367, 1658: 6.17466713236, 1659: -2.27930467216, 1660: -2.97422860496, 1661: 4.5218747699, 1662: -3.88925988558, 1663: 9.16100062564, 1664: 5.54786564491, 1665: -3.68817296215, 1666: -5.22873481207, 1667: 3.28053584717, 1668: -4.92860317311, 1669: 0.516484655327, 1670: -2.610015374, 1671: 2.64189298787, 1672: -0.667369536805, 1673: -4.94185133043, 1674: -11.9087638318, 1675: 0.853440064449, 1676: -10.1809043326, 1677: 8.41322269719, 1678: -2.10326940048, 1679: -11.9208673922, 1680: 10.7399320702, 1681: -3.18808020219, 1682: -5.46262055555, 1683: 3.76701359429, 1684: 0.796020278713, 1685: 4.85271796514, 1686: -2.63027848774, 1687: -0.923346131738, 1688: 3.43362189936, 1689: 3.54337672564, 1690: 2.09382203553, 1691: 6.65208157274, 1692: -4.17137411583, 1693: 1.23473536229, 1694: -5.17711331328, 1695: 1.06854695728, 1696: 0.317376753182, 1697: -6.79634660409, 1698: 1.91058258312, 1699: 1.77915172985, 1700: 0.433386983904, 1701: -1.64927448756, 1702: 2.98944035139, 1703: 0.754877396256, 1704: 1.8388464845, 1705: 1.71654493815, 1706: 1.02357416616, 1707: -12.0446232004, 1708: 4.9137744644, 1709: 5.09220764343, 1710: 3.14489594093, 1711: 5.04326661363, 1712: -4.14261443756, 1713: 4.6654439357, 1714: -3.68866381119, 1715: -5.12863395994, 1716: 0.873032407479, 1717: 1.58176221338, 1718: -2.43732528427, 1719: -1.62603715754, 1720: -1.05671644955, 1721: 5.14257463572, 1722: -8.35325331713, 1723: -1.6092694152, 1724: -3.91112912173, 1725: 1.81651360798, 1726: -1.79060297451, 1727: -6.71760951494, 1728: -3.70408589431, 1729: -2.94542743471, 1730: -1.16936235789, 1731: 0.585583018976, 1732: 13.8582863991, 1733: -10.7501147188, 1734: 3.27908488212, 1735: 3.81169532732, 1736: -4.58189131795, 1737: 1.51664584335, 1738: -5.48610232171, 1739: -6.19797399935, 1740: 3.4793279878, 1741: 6.61127465271, 1742: -13.9208879797, 1743: -7.78425680751, 1744: 7.4028187125, 1745: 6.83065662439, 1746: 0.755645375117, 1747: 7.47330029114, 1748: 9.51043983894, 1749: 7.63897440806, 1750: 26.9663597854, 1751: -13.0079009147, 1752: -23.5389603177, 1753: -23.0729415562, 1754: 4.87138671433, 1755: -4.40496633495, 1756: -6.29328176111, 1757: -7.68567098902, 1758: 1.42084596885, 1759: 9.77653392868, 1760: -8.31051707997, 1761: 1.94317173723, 1762: 10.2237274935, 1763: 3.61385080969, 1764: 3.54839698796, 1765: 5.48193333523, 1766: 8.99357114552, 1767: 3.20838888278, 1768: 18.8900772031, 1769: -17.7963853323, 1770: 0.948681290149, 1771: -1.20031642429, 1772: -11.9886607073, 1773: 4.45731012104, 1774: -14.8733506908, 1775: 18.4903396947, 1776: 22.7783669757, 1777: -4.74743468085, 1778: 6.00354025965, 1779: 2.27093604436, 1780: 2.38060760992, 1781: 3.46720854432, 1782: 18.1281783804, 1783: 18.8922598426, 1784: 23.0873116989, 1785: 4.17627821865, 1786: 0.452674787347, 1787: 4.30581113987, 1788: 11.2916715702, 1789: 9.17030347149, 1790: 4.12748582185, 1791: -0.819353523525, 1792: -11.1979410353, 1793: 2.97787794857, 1794: -1.43496996524, 1795: -12.2361226416, 1796: 0.17124044369, 1797: 25.7087691069, 1798: 4.06884447418, 1799: -9.44065555256, 1800: 3.14299841014, 1801: 3.51412635877, 1802: 12.560876959, 1803: 4.46637476444, 1804: 5.77835091025, 1805: -8.48723003022, 1806: 34.6019771377, 1807: -5.72976410707, 1808: 0.127353448148, 1809: -0.862874484975, 1810: -10.5756886046, 1811: 0.841793614736, 1812: -16.2883279968, 1813: 4.0950519981, 1814: -4.3668258788, 1815: 3.53968401424, 1816: -7.0124396094, 1817: -15.9732233674, 1818: 3.93266862842, 1819: -18.325565787, 1820: 8.25391830275, 1821: 9.50728116216, 1822: 11.9400218856, 1823: 6.64106027935, 1824: 16.4521899697, 1825: 9.84532072557, 1826: 0.829774322233, 1827: -2.0337047376, 1828: -0.846129001299, 1829: 4.2325011137, 1830: -5.27699784311, 1831: -0.331017918303, 1832: -21.7078453813, 1833: 2.15716035374, 1834: 3.86782614398, 1835: -23.5949689827, 1836: -28.2437073911, 1837: 3.97108516907, 1838: -16.2077632802, 1839: 2.91858148151, 1840: -2.183613706, 1841: 3.63563310393, 1842: 3.23222774731, 1843: -3.65430164387, 1844: 1.02675861135, 1845: 1.10316626991, 1846: 4.16843241933, 1847: -8.89014471575, 1848: 3.86825202941, 1849: 2.39330927966, 1850: 3.7226022175, 1851: -13.9271943688, 1852: 4.0280703662, 1853: 4.0948753334, 1854: 5.4692871501, 1855: -9.28895891857, 1856: 0.292804479361, 1857: -8.96806960959, 1858: 6.80314638617, 1859: -3.1597598794, 1860: -6.47936596858, 1861: -11.7208029221, 1862: -1.10960630841, 1863: 14.7940526548, 1864: 15.7350306881, 1865: 12.9259973711, 1866: -12.6650187127, 1867: 9.46647149265, 1868: -12.1628928707, 1869: 2.96510199815, 1870: 12.3154901011, 1871: 0.836411470332, 1872: 0.689146660495, 1873: -9.93350576174, 1874: -0.635834173718, 1875: 0.768242124309, 1876: 3.21965759677, 1877: 0.808303319166, 1878: -9.24653935549, 1879: 1.98052505139, 1880: 1.22585150636, 1881: -1.32246922979, 1882: 1.84146811964, 1883: 8.57680035289, 1884: 2.81401032667, 1885: -0.915912157527, 1886: 16.4108074398, 1887: -4.04019097819, 1888: 2.88449917802, 1889: -0.59572436811, 1890: -4.02523308378, 1891: 1.16934297395, 1892: 0.805980625647, 1893: 24.3895910046, 1894: 0.738518489113, 1895: 3.90689718598, 1896: 3.84342877962, 1897: 0.756474183306, 1898: 0.711215509964, 1899: 0.815825764137, 1900: 0.837072727047, 1901: 1.69745312491, 1902: 0.436912052873, 1903: -1.16174213735, 1904: 1.63836365185, 1905: 2.3824874295, 1906: 4.70300250324, 1907: 5.56201257451, 1908: 9.0192197761, 1909: -1.60162888056, 1910: 1.64903802151, 1911: -4.82628468343, 1912: -35.9730667082, 1913: 6.35374134792, 1914: -13.5251385845, 1915: 2.60040842316, 1916: 29.7272472494, 1917: 0.801852133391, 1918: 6.84210828867, 1919: 0.856963553614, 1920: -2.38616482334, 1921: -17.25674597, 1922: -19.1962509279, 1923: 0.0571570610392, 1924: 0.861117960976, 1925: 0.686287243047, 1926: 0.721759804873, 1927: 0.34104313503, 1928: 8.47996738984, 1929: 27.1359531614, 1930: -1.85263768401, 1931: -47.6405721522, 1932: -1.7288498752, 1933: -1.98701914142, 1934: -5.84277666956, 1935: 10.2441289006, 1936: 0.779729162488, 1937: 4.24096169287, 1938: 0.830384339042, 1939: 24.5036797632, 1940: -4.07229429663, 1941: -0.477113433889, 1942: 1.47348083654, 1943: 3.38982538544, 1944: 0.718703996503, 1945: 2.64375515262, 1946: 0.673679133746, 1947: 16.9687826229, 1948: 0.796276282798, 1949: 0.754740495835, 1950: 14.1802429038, 1951: 3.75639267164, 1952: 36.1840504201, 1953: 0.686128679297, 1954: 0.867343420575, 1955: -1.42163829014, 1956: 0.842305789213, 1957: 0.950871034495, 1958: -9.86891780194, 1959: -8.91435168754, 1960: 0.658852540648, 1961: -1.14483647128, 1962: -4.24597617268, 1963: -3.0356626583, 1964: 7.24908237818, 1965: -14.3992531835, 1966: 3.07485379012, 1967: 12.3914985677, 1968: -1.76859332685, 1969: 3.56962493799, 1970: -13.3549687438, 1971: 0.961882076064, 1972: 13.4846054709, 1973: 10.8498234465, 1974: 8.28763420852, 1975: -6.43388241383, 1976: 11.2877157505, 1977: 1.29165565194, 1978: 9.66549624794, 1979: 6.25254750331, 1980: 2.60993060616, 1981: 10.4066653628, 1982: -3.75817677309, 1983: -19.312218845, 1984: 4.13975036251, 1985: 3.12660363291, 1986: -3.5128991711, 1987: 1.40987929251, 1988: -1.62236949465, 1989: 15.6679640835, 1990: -1.29465075848, 1991: 8.90286227847, 1992: -1.10814918813, 1993: -14.7629952651, 1994: 54.9015293553, 1995: 0.138449483771, 1996: -0.598021166602, 1997: -12.6407307872, 1998: 42.2847066537, 1999: -0.906761776145, 2000: -4.85555666207, 2001: 3.41070905558, 2002: -23.1948495653, 2003: 3.4601137669, 2004: 3.88315365032, 2005: 3.83932478746, 2006: 0.402846088423, 2007: 4.01042717882, 2008: 3.96192378931, 2009: 2.13890513621, 2010: 1.59864112402, 2011: -0.0641474227862, 2012: -0.717028214844, 2013: -1.59088389307, 2014: -14.0049789757, 2015: 5.16388709941, 2016: 5.15795780137, 2017: -0.140069531486, 2018: -6.80627657647, 2019: 17.9137800846, 2020: 1.87680268644, 2021: -31.0228872381, 2022: 8.17737356817, 2023: -19.7831847667, 2024: 3.6496906904, 2025: -19.9874036146, 2026: 6.71049731692, 2027: 0.438705060932, 2028: 2.65505730375, 2029: 8.68931908255, 2030: 10.639033227, 2031: 25.1590713081, 2032: 3.79249582101, 2033: 3.35054514813, 2034: 4.67897488829, 2035: -6.85477074892, 2036: -0.193816764062, 2037: -21.8297171962, 2038: 0.650770053597, 2039: 13.8370022647, 2040: 29.1227139576, 2041: -40.2968919654, 2042: 6.35007623237, 2043: -4.73965142115, 2044: -2.37728365577, 2045: -55.696987183, 2046: 5.5143697364, 2047: -16.7954370368, 2048: -0.818999144364, 2049: 3.58341292446, 2050: 0.828751713498, 2051: -1.94529149019, 2052: -8.46543917226, 2053: 2.76074069482, 2054: 2.76462258861, 2055: 0.757432512128, 2056: -4.06759867057, 2057: -1.64618367518, 2058: 4.03680953808, 2059: -5.42175899022, 2060: 3.85333034252, 2061: -29.4834922391, 2062: 4.02637447397, 2063: 4.08063407145, 2064: 16.1705907247, 2065: 3.37302641049, 2066: 0.984789013651, 2067: -25.672137391, 2068: 14.1543713691, 2069: 5.68651391021, 2070: 0.809898984969, 2071: -26.2927143877, 2072: -3.76786576617, 2073: 0.9032692952, 2074: 1.06326980432, 2075: -7.24178136245, 2076: 20.4778817327, 2077: 16.3219379222, 2078: 3.282687967, 2079: -1.49563029221, 2080: 12.5657318325, 2081: -49.3590063017, 2082: -28.0616443123, 2083: -5.00629829193, 2084: 3.70495635144, 2085: -6.68313556377, 2086: -7.55384009104, 2087: -11.7010885533, 2088: -12.8466090262, 2089: 6.52796798292, 2090: -1.19588290382, 2091: -3.40869523603, 2092: 27.7270485186, 2093: 18.3798362133, 2094: -51.6814132748, 2095: -0.671910906735, 2096: 1.68049578597, 2097: 2.43185091927, 2098: 1.09809761145, 2099: 72.9416645463, 2100: 19.3393207484, 2101: -10.8408945638, 2102: 8.42948246099, 2103: 1.31999959678, 2104: -0.840327831843, 2105: -31.7585329335, 2106: -25.928691611, 2107: 18.9383522289, 2108: -1.56676165188, 2109: -12.3233629539, 2110: 1.53327812866, 2111: -67.8382891916, 2112: -22.2439451403, 2113: 94.7191884711, 2114: 37.8999604986, 2115: -27.5099423601, 2116: 58.970931087, 2117: 12.2420042053, 2118: 8.3007733932, 2119: 14.4857703566, 2120: -13.303163458, 2121: -25.1066833866, 2122: -9.94824126439, 2123: 11.7378082747, 2124: 12.4085519329, 2125: 10.1448595682, 2126: 0.142535881775, 2127: -17.855953908, 2128: -2.77025783381, 2129: -7.78176239461, 2130: -4.61366988945, 2131: -20.6478059464, 2132: -2.51395168274, 2133: -2.75999121018, 2134: -22.838861692, 2135: 37.2083706297, 2136: 5.94448958933, 2137: -46.1612202216, 2138: -8.64834229896, 2139: -4.00840750166, 2140: -4.24558745007, 2141: 2.03237929819, 2142: -7.76156646844, 2143: 23.4846110969, 2144: -6.7328766487, 2145: 7.75212165298, 2146: -11.2651330408, 2147: -5.44049547114, 2148: 33.6191913757, 2149: 1.42145943185, 2150: -4.68802337531, 2151: 8.69061201638, 2152: -9.25787989988, 2153: 0.45271210607, 2154: -5.76434283582, 2155: 12.0223979418, 2156: -32.1312882124, 2157: -4.91100080305, 2158: -0.213004361656, 2159: 6.38449312995, 2160: -3.50208188882, 2161: -2.28274003792, 2162: -3.95066581787, 2163: 4.80696611999, 2164: -4.27526372987, 2165: 0.000399258372208, 2166: -21.454780155, 2167: -4.70401697103, 2168: -5.69311415834, 2169: 2.61242718848, 2170: -5.19582418044, 2171: 1.466337104, 2172: -4.20873442499, 2173: 2.86748053904, 2174: 35.8170284421, 2175: -5.29706904368, 2176: 31.3620879983, 2177: -4.88345055041, 2178: -3.13016282268, 2179: 38.0005818974, 2180: 8.62648086133, 2181: 97.2840694013, 2182: 14.6381764526, 2183: -1.65975287817, 2184: 51.8974050585, 2185: 38.7751555703, 2186: -6.17314597486, 2187: -11.1687536542, 2188: 3.50268352262, 2189: -1.57193698464, 2190: -3.32836006539, 2191: 1.31409563682, 2192: -4.20257518175, 2193: 1.83054938596, 2194: -5.29367604711, 2195: 57.1387178461, 2196: 31.928907082, 2197: 17.9245601391, 2198: 2.09161343416, 2199: -4.34575696964, 2200: 34.884695962, 2201: 1.36548945193, 2202: -4.90733765809, 2203: -1.12363060301, 2204: 26.3500623316, 2205: 3.12629157752, 2206: -42.0481372878, 2207: -39.0550794314, 2208: 20.8232675883, 2209: 22.693256562, 2210: 4.11559035209, 2211: -28.3976692826, 2212: 1.98506552112, 2213: 100.0, 2214: -42.2988290135, 2215: 34.8482957454, 2216: -46.8148415797, 2217: 19.9265138321, 2218: -11.2260086847, 2219: -26.8185817167, 2220: -0.848687596998, 2221: 0.294676009188, 2222: 11.1983535148, 2223: 16.0773499889, 2224: -0.797999507789, 2225: 2.37744163978, 2226: -7.31178288626, 2227: -10.9218447214, 2228: 0.0637900041235, 2229: -10.4833988995, 2230: -12.5856991967, 2231: 1.38992687528, 2232: 8.39455363051, 2233: -4.7466989798, 2234: -18.3765333907, 2235: 6.33932651061, 2236: -0.407602972259, 2237: -18.3315913196, 2238: 1.58709668651, 2239: -13.1500685831, 2240: -0.848167947136, 2241: -0.738454504958, 2242: -7.28118587622, 2243: -0.967699940485, 2244: -4.2957485177, 2245: 1.20036885765, 2246: -0.906872239807, 2247: -0.725735988391, 2248: -0.995433238206, 2249: -0.902609878233, 2250: -3.74494118341, 2251: -17.0693076139, 2252: -2.77043048784, 2253: -1.19669835346, 2254: -1.33664760631, 2255: 12.1131325142, 2256: 4.75949477979, 2257: -15.293806946, 2258: 18.9223159341, 2259: 32.199897673, 2260: 47.573199869, 2261: 4.78347065589, 2262: -8.62356616322, 2263: -3.30945928814, 2264: -2.7111240216, 2265: 7.45537223149, 2266: -1.37774443559, 2267: -63.7657037274, 2268: -0.947249265174, 2269: 14.338516589, 2270: 3.40068917249, 2271: 48.9683925561, 2272: 0.676461436745, 2273: -0.93932801103, 2274: -0.899609514124, 2275: -0.825212943649, 2276: -1.14222784675, 2277: -1.31996826877, 2278: 4.15115035526, 2279: 6.63156621451, 2280: 3.11614608372, 2281: -2.32465366844, 2282: -1.6240665493, 2283: 12.8412114035, 2284: 0.307044671529, 2285: -4.84052108636, 2286: 0.588698600882, 2287: -0.941823880301, 2288: 26.7973281716, 2289: 2.93168980657, 2290: 4.03311629583, 2291: -6.10895296129, 2292: 40.9548830599, 2293: -0.901220937194, 2294: -4.59247673998, 2295: -1.09838980282, 2296: -6.77941650692, 2297: -0.927029264987, 2298: -0.815798350737, 2299: 33.1908295475, 2300: -4.5872432815, 2301: -60.275764476, 2302: -0.779851529743, 2303: -0.725328160569, 2304: 1.878982476, 2305: -1.09358217781, 2306: 0.906570970731, 2307: 4.23191632227, 2308: -19.6712097922, 2309: -0.724998966783, 2310: -1.00596197168, 2311: -17.6215705669, 2312: -0.108282423808, 2313: 26.5419525077, 2314: -38.3757400982, 2315: -4.12466563154, 2316: -4.36846773473, 2317: 8.13304737786, 2318: -3.59209585685, 2319: 2.89382676569, 2320: -1.61448421024, 2321: 18.4369571763, 2322: -9.64551258828, 2323: -5.09315482359, 2324: 32.4359816487, 2325: -82.3684703881, 2326: -0.891474084838, 2327: -14.3747265945, 2328: 6.86313919636, 2329: -3.9907308794, 2330: -5.55271381507, 2331: 30.8077334949, 2332: -24.6158377475, 2333: -4.47039880643, 2334: 2.40954327615, 2335: -16.2504051811, 2336: 11.9954708421, 2337: -21.9509861856, 2338: -6.62883318374, 2339: -13.6911177565, 2340: -14.5603076554, 2341: 50.7015297694, 2342: 4.83290216092, 2343: 2.73866052081, 2344: 12.4195069481, 2345: 21.5593303526, 2346: 24.224500437, 2347: 59.6697779446, 2348: 14.5994596042, 2349: -4.31788537427, 2350: -5.4031877399, 2351: 1.68436174677, 2352: -3.71515284408, 2353: -4.26571277749, 2354: 1.39965930844, 2355: -0.285740732945, 2356: -4.49294696732, 2357: -1.85374286523, 2358: -1.26134740123, 2359: -3.8398814931, 2360: -39.6301119387, 2361: 12.2056590246, 2362: 22.1565595108, 2363: -0.973085617672, 2364: 12.3916302586, 2365: 7.62790694967, 2366: 1.79685741453, 2367: 21.9100021355, 2368: -24.0552261599, 2369: 4.59394262865, 2370: -11.7919171303, 2371: -19.4256461531, 2372: 35.2273700143, 2373: -0.985620042271, 2374: 20.1040770692, 2375: 17.3054247809, 2376: -7.67979676401, 2377: -32.7733008245, 2378: 19.7457483914, 2379: 13.0092526748, 2380: -19.8691861897, 2381: -4.31106739765, 2382: -26.195131479, 2383: -3.83474280781, 2384: 5.21676113295, 2385: -5.60839745284, 2386: 49.3923269349, 2387: 43.312370098, 2388: 3.21795078465, 2389: -5.65607561569, 2390: -0.345091586042, 2391: 2.50682400746, 2392: -1.16627040465, 2393: -2.81492865431, 2394: -4.5473019976, 2395: 7.0423902832, 2396: 23.2733827106, 2397: 3.78170624831, 2398: 2.57482436958, 2399: 32.4602909943, 2400: -1.7120854454, 2401: 0.261122712344, 2402: -4.53466254322, 2403: -4.65102507693, 2404: 4.76037153085, 2405: -9.63827544214, 2406: 0.757081584832, 2407: -6.2734484024, 2408: -78.6835860044, 2409: -4.37673396959, 2410: 26.8851474566, 2411: -4.56040359899, 2412: -4.75263047352, 2413: -54.5216033598, 2414: -4.20848793599, 2415: 3.71664425125, 2416: 9.56228325134, 2417: -2.45688231642, 2418: 11.4295988035, 2419: 17.7094877692, 2420: -17.7868120055, 2421: 11.139320446, 2422: -21.5931187094, 2423: -18.0623889433, 2424: -13.5670477146, 2425: -4.00441300626, 2426: 47.9976209995, 2427: 2.73578565874, 2428: -6.46752692524, 2429: -71.7478621348, 2430: -11.641065871, 2431: 6.46111157172, 2432: -19.8906540227, 2433: -4.55353954174, 2434: -14.0202613434, 2435: -2.96309657048, 2436: -4.07113060252, 2437: 15.0150634277, 2438: -46.2792404103, 2439: -13.1558935335, 2440: 56.6839801543, 2441: 5.99391952854, 2442: -37.4799233199, 2443: -30.3642929397, 2444: -0.288796832917, 2445: -43.9372561111, 2446: 14.1868992831, 2447: 7.8985225414, 2448: 55.0506617571, 2449: -4.38576255358, 2450: 42.7152116465, 2451: 50.3762665275, 2452: -55.0966355861, 2453: 43.5846324413, 2454: 15.077286418, 2455: 23.8812545858, 2456: -33.5176028234, 2457: -27.107796421, 2458: 11.498141871, 2459: 39.1756332831, 2460: -60.9715073998, 2461: 7.82064410125, 2462: 4.01031360263, 2463: -5.15924989469, 2464: -18.2263749844, 2465: -10.5858834854, 2466: -7.94685498331, 2467: -9.02534248737, 2468: -22.2226572845, 2469: 23.9692612311, 2470: 40.6823162566, 2471: -10.8800693523, 2472: -29.6735403285, 2473: -7.86904180554, 2474: -6.79832250956, 2475: -17.8370122366, 2476: -19.1947651968, 2477: -73.7292588474, 2478: 9.77458517345, 2479: -1.61210589523, 2480: 4.68867827058, 2481: -19.7138323923, 2482: 18.3339598832, 2483: -59.7168961269, 2484: -21.0207309532, 2485: -2.00285707233, 2486: -30.7462700463, 2487: -37.5651694849, 2488: 3.34469008478, 2489: 1.07706029646, 2490: 55.0785761825, 2491: -2.21142588978, 2492: -4.18379534369, 2493: -1.63098415547, 2494: 22.9344663782, 2495: -42.5290044958, 2496: -2.037014165, 2497: 68.6359002548, 2498: -45.0167373017, 2499: -1.94481935027, 2500: 39.0481472843, 2501: -30.9156456108, 2502: -96.3755955216, 2503: 22.1269028618, 2504: -25.7918173355, 2505: -66.4513638618, 2506: -26.483451946, 2507: 1.78818878849, 2508: 44.919551031, 2509: -1.81921432228, 2510: 26.9775932912, 2511: -2.1210616762, 2512: 13.0183141551, 2513: -2.77850600152, 2514: -9.11814748154, 2515: -37.7980239188, 2516: -1.98703948067, 2517: 75.6443663263, 2518: -12.1820746485, 2519: -1.32455866738, 2520: -9.72839876654, 2521: -2.35890342125, 2522: 24.1631420481, 2523: 97.4022973948, 2524: -0.633312665878, 2525: 20.3977863606, 2526: -2.24589272015, 2527: -3.47147865834, 2528: 15.004708091, 2529: -3.37996973997, 2530: 11.3274216151, 2531: 3.95317309752, 2532: -2.66267659442, 2533: 12.9228117757, 2534: 12.2370116406, 2535: -2.01159234109, 2536: -17.1117788171, 2537: -2.72230901894, 2538: -3.27949551023, 2539: -2.19363336532, 2540: 2.8615495664, 2541: -2.12717203398, 2542: -2.31856119707, 2543: -2.36992381948, 2544: 37.1361953659, 2545: -1.84994845398, 2546: 1.752634983, 2547: -1.75282094344, 2548: 6.47684585933, 2549: 15.5813027488, 2550: -1.97665530009, 2551: -1.43465623846, 2552: -2.71703021671, 2553: 10.0722430804, 2554: -2.4263398807, 2555: -6.40200960383, 2556: -100.0, 2557: 5.04957047105, 2558: -23.2207285213, 2559: -21.008414968, 2560: -80.4882418123, 2561: 14.3031946779, 2562: 14.0101195931, 2563: -45.3343433538, 2564: -5.64117614137, 2565: -97.0227218706, 2566: 33.6088705549, 2567: -23.4275155057, 2568: -11.8118127764, 2569: -0.63838153058, 2570: 0.710021840383, 2571: 3.37291384284, 2572: 11.924954495, 2573: -0.277602701356, 2574: 6.6147530662, 2575: 3.39295324786, 2576: 11.9527286749, 2577: 0.25990809352, 2578: 12.6940201209, 2579: 4.06738467333, 2580: -10.2998442482, 2581: 8.25498500713, 2582: 3.08174384542, 2583: -10.8619896255, 2584: 28.1458372103, 2585: -8.14103533972, 2586: -13.9231342182, 2587: 0.657722241195, 2588: 0.966796123148, 2589: 0.14183516022, 2590: -0.489136714712, 2591: 29.296140914, 2592: -0.297577161559, 2593: -4.05961443649, 2594: -27.732440787, 2595: 1.18177544596, 2596: -0.397073811008, 2597: -0.381368494651, 2598: -0.376390079428, 2599: -2.40835303872, 2600: -39.8123762223, 2601: 2.53912302811, 2602: 1.14834315059, 2603: -1.12247261348, 2604: -0.0244800853665, 2605: -2.96744077022, 2606: -48.2406723231, 2607: 24.5226811076, 2608: -49.33765797, 2609: 30.6054882482, 2610: -35.3278346714, 2611: -18.6485582784, 2612: 1.60135710975, 2613: 9.10719760584, 2614: 46.3164857678, 2615: -0.270710975758, 2616: -32.8726364756, 2617: -0.212377200802, 2618: -13.8641726898, 2619: -8.79042338101, 2620: -6.90380190587, 2621: -1.59364127634, 2622: -0.37207899912, 2623: 0.0839165548247, 2624: -0.41143812977, 2625: -0.608152314941, 2626: -8.52870020841, 2627: -6.56054050887, 2628: 19.6074347203, 2629: 36.3648795915, 2630: 22.9694366343, 2631: 22.4130536691, 2632: 17.3441967191, 2633: 16.4677811079, 2634: -17.2099451023, 2635: -30.6832244501, 2636: -0.768078242003, 2637: -3.40555023904, 2638: -5.22401046688, 2639: -15.648648295, 2640: 7.14141206807, 2641: -50.5380286438, 2642: -0.342186648797, 2643: -1.89767147961, 2644: -0.409977544623, 2645: -4.01175217646, 2646: -0.344072479949, 2647: -0.502933283954, 2648: -42.6022104804, 2649: -1.82323158363, 2650: 3.23792586454, 2651: -0.399405111928, 2652: -0.301376044143, 2653: -2.98837662533, 2654: -0.375725502923, 2655: -4.65424459677, 2656: 16.3831306994, 2657: 6.40810506617, 2658: -2.36789563404, 2659: -0.593714037535, 2660: -16.0151168456, 2661: 4.09582726794, 2662: -43.6305512533, 2663: -24.7688839104, 2664: 14.8805521441, 2665: -9.93625078039, 2666: -26.2012744191, 2667: -1.77359062984, 2668: -20.768809584, 2669: -0.132029257891, 2670: 61.6310093159, 2671: -3.32710030081, 2672: -10.2687564087, 2673: 12.8904807108, 2674: -4.14548349451, 2675: -1.73818628491, 2676: -2.0674928066, 2677: -14.9870606712, 2678: 18.7837294283, 2679: -9.33226309298, 2680: -15.0803458802, 2681: -27.7282186049, 2682: -1.91590054832, 2683: 6.62164434915, 2684: 56.9891690399, 2685: 22.298058657, 2686: -46.8902362508, 2687: 12.6408103187, 2688: -14.9733166109, 2689: 32.9760156945, 2690: -13.7857065106, 2691: 32.8770024163, 2692: 15.6219813857, 2693: 14.8547770087, 2694: 6.58980774996, 2695: 12.0283993273, 2696: -10.3584878962, 2697: -31.4572382916, 2698: -4.90197612329, 2699: -2.00000517826, 2700: -50.5464114374, 2701: -2.07716537443, 2702: -4.15220761282, 2703: -30.7024459494, 2704: -7.9722016701, 2705: -1.68469431787, 2706: -1.18833074262, 2707: 7.602067587, 2708: -2.25656702519, 2709: 10.9722894238, 2710: -24.9817865181, 2711: -64.4204468486, 2712: 41.2957311635, 2713: 11.6319119634, 2714: -6.18989991899, 2715: -28.9808880693, 2716: 31.7037236433, 2717: -0.406573240455, 2718: -61.7197832804, 2719: -5.21296517515, 2720: 1.44028660641, 2721: -1.38285728537, 2722: 12.8245919847, 2723: 30.8022631981, 2724: -10.6153011881, 2725: 43.3007718363, 2726: 17.429139578, 2727: -35.2068129779, 2728: -56.4468381488, 2729: 36.6009475843, 2730: -2.02833792577, 2731: -2.03185599965, 2732: -2.27397937228, 2733: -6.81566219879, 2734: 9.55744205243, 2735: -6.34661401928, 2736: 61.8985708969, 2737: -38.5623135073, 2738: 19.9842777993, 2739: 41.9396701935, 2740: 16.445677438, 2741: 60.9307669725, 2742: -8.12570975296, 2743: 0.747180078536, 2744: 7.12678254773, 2745: -54.7336418802, 2746: 0.81642393479, 2747: -29.9903047606, 2748: 2.91970514899, 2749: -3.08933656064, 2750: 12.4975616831, 2751: -1.9394578625, 2752: -1.8040156123, 2753: 1.50022204113, 2754: 17.0994892728, 2755: -1.47316819818, 2756: -1.73259943086, 2757: 82.5375251608, 2758: -1.8774667786, 2759: -32.2631928562, 2760: -1.97504652904, 2761: -1.93499069134, 2762: -12.6132600856, 2763: -1.96528358943, 2764: -6.61997627644, 2765: -60.4246684164, 2766: 57.2254842111, 2767: -74.8586736683, 2768: 53.9146757093, 2769: -52.4447001161, 2770: 53.0101785658, 2771: -64.2016288319, 2772: 6.45828404091, 2773: 16.8786025156, 2774: 100.0, 2775: 8.97377414227, 2776: -1.23145889925, 2777: -3.24891784531, 2778: 2.27281915096, 2779: -60.7830242695, 2780: -17.4593822954, 2781: -0.174654307961, 2782: -1.5266827163, 2783: 7.52719134737, 2784: -6.69190389204, 2785: 20.064047885, 2786: -1.48071005884, 2787: -39.8301431977, 2788: 58.9347054453, 2789: 100.0, 2790: 100.0, 2791: 34.4460499761, 2792: 89.7474770911, 2793: -1.18601609691, 2794: 23.87389611, 2795: 2.36885086986, 2796: 30.7178199763, 2797: 69.6278309258, 2798: -26.6565936536, 2799: 2.71770889821, 2800: -19.8610257011, 2801: 40.5376935208, 2802: 34.6339732322, 2803: 81.8232740221, 2804: 18.1292159831, 2805: 4.73547739663, 2806: 29.4502313925, 2807: -2.97760800634, 2808: 18.5970134667, 2809: 16.2119874464, 2810: -8.35222347427, 2811: 50.0119691694, 2812: -41.9367577555, 2813: -11.1552544668, 2814: 9.15858383651, 2815: -20.31574633, 2816: 20.5524480112, 2817: -28.2302848473, 2818: 25.1465551213, 2819: -12.7045410399, 2820: -29.4384200885, 2821: 41.2315654555, 2822: -19.5687854126, 2823: -4.12153853144, 2824: -44.246467846, 2825: -7.59072720618, 2826: -13.6160167237, 2827: 47.5631870752, 2828: -20.7031279783, 2829: -22.1866387324, 2830: 13.0440425119, 2831: -30.8611048461, 2832: -15.9517145445, 2833: -16.1971541548, 2834: -8.34695153661, 2835: -7.14849700112, 2836: -7.75781633544, 2837: 10.053202415, 2838: -13.0696046526, 2839: 38.832374884, 2840: -8.15975752994, 2841: 40.7631618957, 2842: -20.2491110749, 2843: 32.2438949747, 2844: 0.623205173545, 2845: 9.22965734563, 2846: -1.09216136349, 2847: 6.31392797565, 2848: -7.7852698348, 2849: -54.7829592404, 2850: -4.8232202077, 2851: -23.6637356138, 2852: 0.249924691795, 2853: 67.9577821058, 2854: -6.03639677713, 2855: 2.2551708034, 2856: -14.6260244856, 2857: 15.7433039199, 2858: -1.24659100846, 2859: 12.231145966, 2860: -7.93539603737, 2861: -23.9934577558, 2862: -8.12340257119, 2863: -1.46550660584, 2864: 67.6932931363, 2865: -8.097866939, 2866: 38.4908888172, 2867: -5.1907378645, 2868: -8.32994859007, 2869: 5.0861790502, 2870: -3.82235687798, 2871: -4.65685317389, 2872: -5.55762167774, 2873: -3.47518204545, 2874: 30.4012013316, 2875: -7.96842817119, 2876: -3.84058764421, 2877: 9.61298111276, 2878: -39.9066156507, 2879: -7.37891032054, 2880: 32.3303236893, 2881: -7.94467753537, 2882: -6.84529184484, 2883: -8.80612942008, 2884: -8.2197075955, 2885: 16.7454888013, 2886: -8.25523441873, 2887: -8.02996573551, 2888: 15.584857727, 2889: 35.2867282422, 2890: -8.27997411384, 2891: 2.79277872524, 2892: -5.64479450703, 2893: -40.3495079776, 2894: -8.08676186238, 2895: -4.03806300163, 2896: -7.74095839416, 2897: -0.0381360659673, 2898: 100.0, 2899: -8.27249811235, 2900: -6.61213451966, 2901: -8.01371504952, 2902: 34.1123121726, 2903: -7.57409664863, 2904: -8.56462549231, 2905: 19.0016074275, 2906: 20.8812520844, 2907: 38.6998869438, 2908: -2.675136854, 2909: 13.9524590219, 2910: 5.50344706094, 2911: 12.8685693454, 2912: 21.8079253853, 2913: -25.2934238489, 2914: -6.92055484288, 2915: -4.97930186165, 2916: -30.6719721905, 2917: 13.9817751, 2918: -1.37849902878, 2919: -1.76543872948, 2920: 6.19944256809, 2921: -86.2241014618, 2922: -1.43707271915, 2923: -6.07749189848, 2924: 0.213937044157, 2925: -2.30607646346, 2926: -5.85468403352, 2927: -9.56954172674, 2928: 19.0919469507, 2929: 3.42533698561, 2930: -11.5237802217, 2931: 0.688597105441, 2932: -27.2532367148, 2933: 40.4568487712, 2934: 18.5788209701, 2935: -13.7997979909, 2936: -1.13810991602, 2937: -34.9339282577, 2938: -1.48437712751, 2939: -1.39917189347, 2940: -5.57342909483, 2941: -1.5336376251, 2942: 1.72385654112, 2943: -2.96561158191, 2944: -2.22689181976, 2945: -1.40204279045, 2946: -1.89059220359, 2947: -1.53161542194, 2948: -7.67580910133, 2949: -13.7759284137, 2950: -2.73411631985, 2951: -2.1889488576, 2952: -2.14589801202, 2953: -5.19948327897, 2954: -12.8532167245, 2955: 4.13504722894, 2956: 24.1823244399, 2957: -13.6128442696, 2958: -43.4489581086, 2959: -0.872941077824, 2960: 14.5825810876, 2961: 11.6556633368, 2962: 10.9882305841, 2963: 23.5593016196, 2964: -1.41539205669, 2965: -1.06196774584, 2966: -1.47211384447, 2967: -2.11437163077, 2968: 6.48278312426, 2969: -20.0350780529, 2970: -3.08929543801, 2971: -1.43015286115, 2972: -2.00626626175, 2973: -1.49145842807, 2974: -1.47615642339, 2975: 1.14671249733, 2976: 21.2391757488, 2977: 15.089402543, 2978: 3.08157882853, 2979: 12.674954221, 2980: 0.112901442657, 2981: 23.361781142, 2982: -0.713975331596, 2983: 9.63134268389, 2984: 0.209943690358, 2985: -1.51217107365, 2986: -15.1286867718, 2987: 2.21215836707, 2988: 0.402613831989, 2989: 3.0291617852, 2990: -13.8968935764, 2991: -1.51769858299, 2992: -8.184237768, 2993: -1.59783434605, 2994: 1.85747028749, 2995: -1.52612855517, 2996: -1.50441101115, 2997: 6.51096353459, 2998: -8.17164198088, 2999: -5.78552824784, 3000: -1.4651313542, 3001: -1.45209771648, 3002: -2.08525480877, 3003: -1.47872820168, 3004: -1.17751749499, 3005: 21.5266741024, 3006: -12.7722905531, 3007: -1.83081432388, 3008: 0.701074467781, 3009: -4.59174378521, 3010: -3.84896026882, 3011: -15.94522179, 3012: 45.6130984996, 3013: -11.5911904928, 3014: 4.16875648413, 3015: -1.72144019915, 3016: -9.13315759208, 3017: 1.97906770827, 3018: -2.08909120415, 3019: -55.5374878781, 3020: -22.4702761112, 3021: 5.83928779222, 3022: -4.24238321298, 3023: 21.5153257554, 3024: -3.4203985411, 3025: 22.4364075444, 3026: 21.9956971336, 3027: -8.91529006689, 3028: 10.7669148202, 3029: -9.3256030624, 3030: -28.4997640855, 3031: -8.29058692177, 3032: -6.07337358827, 3033: 36.3007314304, 3034: 14.7974061598, 3035: -35.2063134158, 3036: -20.069396148, 3037: 0.282436091242, 3038: -50.4660571599, 3039: 60.6475793864, 3040: -61.9511358114, 3041: 9.00363175423, 3042: -23.9400114011, 3043: 77.6411660284, 3044: 30.7601398385, 3045: -47.2729307817, 3046: 12.4168019901, 3047: -9.29941998439, 3048: -8.71625380994, 3049: 2.13317482933, 3050: -8.17721249796, 3051: 1.69752704995, 3052: -1.37773175817, 3053: 13.1213479067, 3054: -8.1230964895, 3055: -1.89202992151, 3056: -9.38368324563, 3057: -7.70652156655, 3058: 39.0578774186, 3059: -9.68776636354, 3060: 5.54349320484, 3061: 17.1405609463, 3062: -7.06775311309, 3063: -11.1547271985, 3064: 15.4939487798, 3065: 1.72277043808, 3066: 10.70830482, 3067: -67.3039471485, 3068: 13.2272434863, 3069: 22.4815945371, 3070: -30.7877549741, 3071: 13.8060292201, 3072: 32.338046045, 3073: -7.70834361069, 3074: 20.781771369, 3075: -38.1495575237, 3076: -5.57652175615, 3077: -22.9301013564, 3078: -9.07750053877, 3079: -8.08024025951, 3080: -2.51773712395, 3081: -7.98518166006, 3082: -22.3887060737, 3083: -13.1165251669, 3084: 41.6232710131, 3085: 1.09404959607, 3086: 10.5218114783, 3087: -46.112720055, 3088: -23.5496274766, 3089: 4.43363430574, 3090: -12.6821081941, 3091: 3.7060155744, 3092: -18.2998835131, 3093: -5.66473234364, 3094: -26.6471976432, 3095: 4.9757746743, 3096: 21.3298123917, 3097: 21.9850702357, 3098: -8.12372209081, 3099: 33.3417562839, 3100: -8.13718607068, 3101: -8.00719698546, 3102: -8.17581372116, 3103: 9.2203478047, 3104: -8.2298712809, 3105: -8.6359988925, 3106: 14.422706085, 3107: -8.26831179681, 3108: -15.0345289191, 3109: -8.05332345004, 3110: -8.23618383464, 3111: 15.7520639422, 3112: -2.38031650602, 3113: 3.81282703469, 3114: -44.563669206, 3115: 12.9208383186, 3116: -6.31606224647, 3117: -17.7022068442, 3118: 9.25079517823, 3119: -16.436154549, 3120: 30.9491848149, 3121: 75.797610918, 3122: 23.810746612, 3123: -1.53332633062, 3124: 15.4925319042, 3125: -9.41267054973, 3126: -4.85814744116, 3127: 19.3080257266, 3128: -28.5038266609, 3129: -17.3187012566, 3130: -13.8508774685, 3131: -8.0740795328, 3132: -0.59765037905, 3133: -1.15458228821, 3134: -7.97425658626, 3135: -14.8737506975, 3136: 10.7102807119, 3137: 21.1139937811, 3138: 8.44885033091, 3139: 25.4801344155, 3140: 4.11604774502, 3141: 33.4397869164, 3142: -0.189231076156, 3143: -1.11204980204, 3144: -0.191721138857, 3145: -0.147728241978, 3146: -12.7554408916, 3147: 6.91871564571, 3148: 0.905421938371, 3149: -11.8330644826, 3150: -6.8081712619, 3151: 1.44800962141, 3152: 17.1541150472, 3153: 3.51634942378, 3154: -7.16366722151, 3155: 2.17842372384, 3156: 2.2233468686, 3157: -2.49267095618, 3158: -6.09792722866, 3159: -0.603268342323, 3160: 2.74962002509, 3161: -4.87179582995, 3162: 9.5542231054, 3163: -9.58276591532, 3164: -4.11732650141, 3165: 1.1796973029, 3166: -2.65714313873, 3167: 6.19868039329, 3168: 7.02115253186, 3169: 4.54692168985, 3170: -3.80210569165, 3171: -5.09603145438, 3172: 0.00363699253217, 3173: 16.241486267, 3174: 0.311456263522, 3175: -1.69723670755, 3176: 3.62879250703, 3177: 0.273342883944, 3178: -0.417409912163, 3179: 12.8226368152, 3180: 1.93008286812, 3181: -7.2237442564, 3182: -8.77452863616, 3183: -7.49199289216, 3184: -17.8924063782, 3185: -4.7294971196, 3186: 8.2643142301, 3187: 0.652114795431, 3188: -1.7200685215, 3189: -1.42874902257, 3190: -0.780435512269, 3191: 3.86210666694, 3192: 4.09651090485, 3193: -6.30504166247, 3194: -2.25198839376, 3195: 0.265895164185, 3196: -7.84184146877, 3197: -1.13594806957, 3198: 1.63660707805, 3199: -0.427514299283, 3200: -3.95841595937, 3201: -12.7900898142, 3202: -4.22004782412, 3203: -4.95653096181, 3204: -0.0100125434859, 3205: -3.82879723681, 3206: -1.57297784309, 3207: 0.953658864199, 3208: 1.54851001274, 3209: -0.124148418987, 3210: 1.77672958055, 3211: 0.0264183082397, 3212: 0.379285808281, 3213: 13.2624403374, 3214: 0.66019662444, 3215: -10.6514982665, 3216: -3.84914014644, 3217: 0.460013879704, 3218: -4.95497016749, 3219: 5.3282585192, 3220: -22.2089870693, 3221: -8.24859801962, 3222: 3.01989204791, 3223: 11.5198759105, 3224: -1.05341650269, 3225: 0.52675886214, 3226: 2.9031257838, 3227: 4.62394976401, 3228: 11.1193707642, 3229: 2.52991949, 3230: -1.05212884355, 3231: -1.81890472323, 3232: 2.77874120173, 3233: -1.16747810275, 3234: -8.26825754377, 3235: -0.875919628094, 3236: -0.0572838034944, 3237: -0.806517214338, 3238: -2.55792710638, 3239: -0.312134463541, 3240: -5.71712155401, 3241: -0.504965783531, 3242: -1.44544746867, 3243: -1.04143959637, 3244: -12.8925097627, 3245: 1.11264280915, 3246: -1.34642224404, 3247: 8.25208040578, 3248: 0.476131530517, 3249: -1.54476683034, 3250: -0.963444392435, 3251: 15.0995503866, 3252: 2.7660401138, 3253: -7.41450288866, 3254: 2.88160055171, 3255: 6.23313947406, 3256: -9.47543788966, 3257: 0.313463840721, 3258: -4.56971713183, 3259: 0.0278641100793, 3260: -15.6871378452, 3261: -12.9248662944, 3262: 11.3667770121, 3263: 2.75751830221, 3264: 2.91931978755, 3265: -4.75665755406, 3266: 19.5612556309, 3267: -0.383978348879, 3268: -0.148457208628, 3269: -2.30997014, 3270: 14.8249767859, 3271: -0.236419698292, 3272: -0.899692126138, 3273: -1.58744385981, 3274: -4.52315624266, 3275: -1.48221402636, 3276: 3.34633777335, 3277: 2.97345897352, 3278: 2.73214780734, 3279: -9.26339485883, 3280: -1.05772163031, 3281: 0.763169324399, 3282: 2.27398688821, 3283: -0.334605138695, 3284: -2.51506344171, 3285: 1.59013096768, 3286: -4.54612370774, 3287: -0.361875036813, 3288: -0.192898400882, 3289: -1.63017061133, 3290: -0.11009234848, 3291: -1.09006262338, 3292: -3.72313772742, 3293: -0.359918024893, 3294: -0.361602929393, 3295: -0.291756128298, 3296: -0.465598639611, 3297: -0.184885252411, 3298: -19.4483119046, 3299: -0.0080167162645, 3300: -0.226874206387, 3301: -1.15704929095, 3302: 3.62243173405, 3303: -9.31996648039, 3304: -8.20080096552, 3305: 1.0646906736, 3306: -16.1043733176, 3307: -16.3369095933, 3308: 4.41272337631, 3309: 1.317609611, 3310: -1.20731399408, 3311: 0.245376039205, 3312: -12.6104722801, 3313: -0.326126979066, 3314: 17.3056565519, 3315: -0.309704163481, 3316: -5.68316548691, 3317: -4.32723054723, 3318: -3.34746446134, 3319: 1.6259454652, 3320: -0.349666786003, 3321: -0.0405325957982, 3322: -0.353316875866, 3323: -0.222712690403, 3324: -0.151731905407, 3325: -7.8252551574, 3326: -3.2890888801, 3327: 1.62402296278, 3328: 9.61220366686, 3329: 1.27781905041, 3330: 5.19460868646, 3331: 2.36362812072, 3332: 3.56173189902, 3333: 9.46890579968, 3334: -0.228925529575, 3335: -0.226705644021, 3336: 2.65745853404, 3337: 0.0149297690876, 3338: 0.142681563184, 3339: 0.282382547929, 3340: -0.295342377304, 3341: -0.512508187876, 3342: -0.310767181771, 3343: 8.54425596516, 3344: -0.310916792897, 3345: -0.338817846633, 3346: -3.15943968438, 3347: -0.00160549125309, 3348: 8.22041266319, 3349: -0.155640392303, 3350: -0.319675883603, 3351: 0.884222962689, 3352: -0.149496308069, 3353: -0.650901270193, 3354: 2.66751652596, 3355: -8.05714973345, 3356: -0.3342112309, 3357: -0.413555249297, 3358: 13.0581008062, 3359: -0.401131918004, 3360: -3.08832334965, 3361: -10.706403529, 3362: -1.35237991148, 3363: 2.01020949127, 3364: -0.48970907066, 3365: -0.427996776179, 3366: 5.10301865613, 3367: -0.789766431905, 3368: 2.56412998955, 3369: -1.87831790485, 3370: 0.849957960896, 3371: 1.631688984, 3372: -0.217993155498, 3373: 0.528163889867, 3374: -4.75456583795, 3375: -7.11174017861, 3376: 3.63175685786, 3377: -0.0926242108797, 3378: 8.37786749639, 3379: -9.31770049033, 3380: 4.58360122601, 3381: -0.984453549675, 3382: -7.19026152978, 3383: 9.6645898656, 3384: -1.40471279977, 3385: 4.72168649573, 3386: 5.70953374152, 3387: -1.6359392987, 3388: 5.3068539393, 3389: 0.238441714614, 3390: -10.2843566998, 3391: 0.892581561898, 3392: 0.529457548096, 3393: 2.2678225126, 3394: -0.889055724119, 3395: 0.575171906514, 3396: 1.57601410906, 3397: -0.695985893188, 3398: 3.4114025163, 3399: -0.398346574682, 3400: -1.01880895256, 3401: -7.99399843435, 3402: 2.24303909723, 3403: 0.275295824475, 3404: -1.45766745982, 3405: 2.04001623347, 3406: -0.132175305596, 3407: 20.0287055274, 3408: -2.36503494165, 3409: 4.91168981796, 3410: 10.6941592838, 3411: 2.59848553758, 3412: -4.72609886667, 3413: 13.2928394125, 3414: 3.84973562019, 3415: 7.54804088172, 3416: 8.69782409681, 3417: -4.47236526972, 3418: 1.69602763724, 3419: -5.42788235535, 3420: 0.329243228223, 3421: 10.0264469845, 3422: -4.52982936386, 3423: -2.2835737801, 3424: 14.2942797619, 3425: 10.4425164971, 3426: 9.5255342643, 3427: -3.7643290307, 3428: 1.76578319149, 3429: -0.554982631828, 3430: -0.363475397558, 3431: 0.290101679662, 3432: 2.42967585507, 3433: 7.28127739428, 3434: -15.7289975867, 3435: -1.52778141858, 3436: 8.35142618328, 3437: -5.04149077432, 3438: -2.60492979944, 3439: 12.7347228983, 3440: 0.250412786406, 3441: -7.90606304393, 3442: 3.61142364243, 3443: -7.95636669517, 3444: 3.37101470752, 3445: 7.37451153299, 3446: -18.9362005944, 3447: -0.0332548938609, 3448: 5.54030479178, 3449: -0.741321180175, 3450: -0.677758675519, 3451: 0.170094593285, 3452: -9.0739225965, 3453: -0.484813895982, 3454: -1.18791913864, 3455: 18.4309007378, 3456: -0.028118478852, 3457: -6.61940612362, 3458: -0.822278108486, 3459: -1.30878724342, 3460: -2.19989095513, 3461: -0.6988335142, 3462: -1.1583674543, 3463: -1.33109319691, 3464: 8.68042430069, 3465: 5.14093068052, 3466: -11.9979013154, 3467: -0.967561976218, 3468: -4.39997762129, 3469: -11.4833204212, 3470: -9.46153266453, 3471: 9.54113433324, 3472: -6.98206583813, 3473: -6.38837193357, 3474: 0.54613324574, 3475: 0.166099410992, 3476: -2.16884426499, 3477: 12.0911360533, 3478: -7.57361916464, 3479: -12.9332662617, 3480: -0.883081756649, 3481: -6.31540520588, 3482: -1.52074836323, 3483: 12.3532934627, 3484: 5.32833673425, 3485: -16.3078720116, 3486: -1.98803928125, 3487: 10.5907449958, 3488: 3.89592458306, 3489: -4.30076030139, 3490: 23.7012826441, 3491: 0.526696877588, 3492: 1.27026220716, 3493: -1.02244917665, 3494: 1.32748051149, 3495: 1.34647134218, 3496: 1.48944591944, 3497: -0.602322981958, 3498: 0.0378468276193, 3499: 0.464210200179, 3500: -0.880906327613, 3501: -1.00625061641}, 'conec': {1: {1: 0, 2: 352}, 2: {1: 1, 2: 352}, 3: {1: 2, 2: 352}, 4: {1: 3, 2: 352}, 5: {1: 4, 2: 352}, 6: {1: 5, 2: 352}, 7: {1: 6, 2: 352}, 8: {1: 7, 2: 352}, 9: {1: 8, 2: 352}, 10: {1: 9, 2: 352}, 11: {1: 10, 2: 352}, 12: {1: 11, 2: 352}, 13: {1: 12, 2: 352}, 14: {1: 13, 2: 352}, 15: {1: 14, 2: 352}, 16: {1: 15, 2: 352}, 17: {1: 16, 2: 352}, 18: {1: 17, 2: 352}, 19: {1: 18, 2: 352}, 20: {1: 19, 2: 352}, 21: {1: 20, 2: 352}, 22: {1: 21, 2: 352}, 23: {1: 22, 2: 352}, 24: {1: 23, 2: 352}, 25: {1: 24, 2: 352}, 26: {1: 25, 2: 352}, 27: {1: 26, 2: 352}, 28: {1: 27, 2: 352}, 29: {1: 28, 2: 352}, 30: {1: 29, 2: 352}, 31: {1: 30, 2: 352}, 32: {1: 31, 2: 352}, 33: {1: 32, 2: 352}, 34: {1: 33, 2: 352}, 35: {1: 34, 2: 352}, 36: {1: 35, 2: 352}, 37: {1: 36, 2: 352}, 38: {1: 37, 2: 352}, 39: {1: 38, 2: 352}, 40: {1: 39, 2: 352}, 41: {1: 40, 2: 352}, 42: {1: 41, 2: 352}, 43: {1: 42, 2: 352}, 44: {1: 43, 2: 352}, 45: {1: 44, 2: 352}, 46: {1: 45, 2: 352}, 47: {1: 46, 2: 352}, 48: {1: 47, 2: 352}, 49: {1: 48, 2: 352}, 50: {1: 49, 2: 352}, 51: {1: 50, 2: 352}, 52: {1: 51, 2: 352}, 53: {1: 52, 2: 352}, 54: {1: 53, 2: 352}, 55: {1: 54, 2: 352}, 56: {1: 55, 2: 352}, 57: {1: 56, 2: 352}, 58: {1: 57, 2: 352}, 59: {1: 58, 2: 352}, 60: {1: 59, 2: 352}, 61: {1: 60, 2: 352}, 62: {1: 61, 2: 352}, 63: {1: 62, 2: 352}, 64: {1: 63, 2: 352}, 65: {1: 64, 2: 352}, 66: {1: 65, 2: 352}, 67: {1: 66, 2: 352}, 68: {1: 67, 2: 352}, 69: {1: 68, 2: 352}, 70: {1: 69, 2: 352}, 71: {1: 70, 2: 352}, 72: {1: 71, 2: 352}, 73: {1: 72, 2: 352}, 74: {1: 73, 2: 352}, 75: {1: 74, 2: 352}, 76: {1: 75, 2: 352}, 77: {1: 76, 2: 352}, 78: {1: 77, 2: 352}, 79: {1: 78, 2: 352}, 80: {1: 79, 2: 352}, 81: {1: 80, 2: 352}, 82: {1: 81, 2: 352}, 83: {1: 82, 2: 352}, 84: {1: 83, 2: 352}, 85: {1: 84, 2: 352}, 86: {1: 85, 2: 352}, 87: {1: 86, 2: 352}, 88: {1: 87, 2: 352}, 89: {1: 88, 2: 352}, 90: {1: 89, 2: 352}, 91: {1: 90, 2: 352}, 92: {1: 91, 2: 352}, 93: {1: 92, 2: 352}, 94: {1: 93, 2: 352}, 95: {1: 94, 2: 352}, 96: {1: 95, 2: 352}, 97: {1: 96, 2: 352}, 98: {1: 97, 2: 352}, 99: {1: 98, 2: 352}, 100: {1: 99, 2: 352}, 101: {1: 100, 2: 352}, 102: {1: 101, 2: 352}, 103: {1: 102, 2: 352}, 104: {1: 103, 2: 352}, 105: {1: 104, 2: 352}, 106: {1: 105, 2: 352}, 107: {1: 106, 2: 352}, 108: {1: 107, 2: 352}, 109: {1: 108, 2: 352}, 110: {1: 109, 2: 352}, 111: {1: 110, 2: 352}, 112: {1: 111, 2: 352}, 113: {1: 112, 2: 352}, 114: {1: 113, 2: 352}, 115: {1: 114, 2: 352}, 116: {1: 115, 2: 352}, 117: {1: 116, 2: 352}, 118: {1: 117, 2: 352}, 119: {1: 118, 2: 352}, 120: {1: 119, 2: 352}, 121: {1: 120, 2: 352}, 122: {1: 121, 2: 352}, 123: {1: 122, 2: 352}, 124: {1: 123, 2: 352}, 125: {1: 124, 2: 352}, 126: {1: 125, 2: 352}, 127: {1: 126, 2: 352}, 128: {1: 127, 2: 352}, 129: {1: 128, 2: 352}, 130: {1: 129, 2: 352}, 131: {1: 130, 2: 352}, 132: {1: 131, 2: 352}, 133: {1: 132, 2: 352}, 134: {1: 133, 2: 352}, 135: {1: 134, 2: 352}, 136: {1: 135, 2: 352}, 137: {1: 136, 2: 352}, 138: {1: 137, 2: 352}, 139: {1: 138, 2: 352}, 140: {1: 139, 2: 352}, 141: {1: 140, 2: 352}, 142: {1: 141, 2: 352}, 143: {1: 142, 2: 352}, 144: {1: 143, 2: 352}, 145: {1: 144, 2: 352}, 146: {1: 145, 2: 352}, 147: {1: 146, 2: 352}, 148: {1: 147, 2: 352}, 149: {1: 148, 2: 352}, 150: {1: 149, 2: 352}, 151: {1: 150, 2: 352}, 152: {1: 151, 2: 352}, 153: {1: 152, 2: 352}, 154: {1: 153, 2: 352}, 155: {1: 154, 2: 352}, 156: {1: 155, 2: 352}, 157: {1: 156, 2: 352}, 158: {1: 157, 2: 352}, 159: {1: 158, 2: 352}, 160: {1: 159, 2: 352}, 161: {1: 160, 2: 352}, 162: {1: 161, 2: 352}, 163: {1: 162, 2: 352}, 164: {1: 163, 2: 352}, 165: {1: 164, 2: 352}, 166: {1: 165, 2: 352}, 167: {1: 166, 2: 352}, 168: {1: 167, 2: 352}, 169: {1: 168, 2: 352}, 170: {1: 169, 2: 352}, 171: {1: 170, 2: 352}, 172: {1: 171, 2: 352}, 173: {1: 172, 2: 352}, 174: {1: 173, 2: 352}, 175: {1: 174, 2: 352}, 176: {1: 175, 2: 352}, 177: {1: 176, 2: 352}, 178: {1: 177, 2: 352}, 179: {1: 178, 2: 352}, 180: {1: 179, 2: 352}, 181: {1: 180, 2: 352}, 182: {1: 181, 2: 352}, 183: {1: 182, 2: 352}, 184: {1: 183, 2: 352}, 185: {1: 184, 2: 352}, 186: {1: 185, 2: 352}, 187: {1: 186, 2: 352}, 188: {1: 187, 2: 352}, 189: {1: 188, 2: 352}, 190: {1: 189, 2: 352}, 191: {1: 190, 2: 352}, 192: {1: 191, 2: 352}, 193: {1: 192, 2: 352}, 194: {1: 193, 2: 352}, 195: {1: 194, 2: 352}, 196: {1: 195, 2: 352}, 197: {1: 196, 2: 352}, 198: {1: 197, 2: 352}, 199: {1: 198, 2: 352}, 200: {1: 199, 2: 352}, 201: {1: 200, 2: 352}, 202: {1: 201, 2: 352}, 203: {1: 202, 2: 352}, 204: {1: 203, 2: 352}, 205: {1: 204, 2: 352}, 206: {1: 205, 2: 352}, 207: {1: 206, 2: 352}, 208: {1: 207, 2: 352}, 209: {1: 208, 2: 352}, 210: {1: 209, 2: 352}, 211: {1: 210, 2: 352}, 212: {1: 211, 2: 352}, 213: {1: 212, 2: 352}, 214: {1: 213, 2: 352}, 215: {1: 214, 2: 352}, 216: {1: 215, 2: 352}, 217: {1: 216, 2: 352}, 218: {1: 217, 2: 352}, 219: {1: 218, 2: 352}, 220: {1: 219, 2: 352}, 221: {1: 220, 2: 352}, 222: {1: 221, 2: 352}, 223: {1: 222, 2: 352}, 224: {1: 223, 2: 352}, 225: {1: 224, 2: 352}, 226: {1: 225, 2: 352}, 227: {1: 226, 2: 352}, 228: {1: 227, 2: 352}, 229: {1: 228, 2: 352}, 230: {1: 229, 2: 352}, 231: {1: 230, 2: 352}, 232: {1: 231, 2: 352}, 233: {1: 232, 2: 352}, 234: {1: 233, 2: 352}, 235: {1: 234, 2: 352}, 236: {1: 235, 2: 352}, 237: {1: 236, 2: 352}, 238: {1: 237, 2: 352}, 239: {1: 238, 2: 352}, 240: {1: 239, 2: 352}, 241: {1: 240, 2: 352}, 242: {1: 241, 2: 352}, 243: {1: 242, 2: 352}, 244: {1: 243, 2: 352}, 245: {1: 244, 2: 352}, 246: {1: 245, 2: 352}, 247: {1: 246, 2: 352}, 248: {1: 247, 2: 352}, 249: {1: 248, 2: 352}, 250: {1: 249, 2: 352}, 251: {1: 250, 2: 352}, 252: {1: 251, 2: 352}, 253: {1: 252, 2: 352}, 254: {1: 253, 2: 352}, 255: {1: 254, 2: 352}, 256: {1: 255, 2: 352}, 257: {1: 256, 2: 352}, 258: {1: 257, 2: 352}, 259: {1: 258, 2: 352}, 260: {1: 259, 2: 352}, 261: {1: 260, 2: 352}, 262: {1: 261, 2: 352}, 263: {1: 262, 2: 352}, 264: {1: 263, 2: 352}, 265: {1: 264, 2: 352}, 266: {1: 265, 2: 352}, 267: {1: 266, 2: 352}, 268: {1: 267, 2: 352}, 269: {1: 268, 2: 352}, 270: {1: 269, 2: 352}, 271: {1: 270, 2: 352}, 272: {1: 271, 2: 352}, 273: {1: 272, 2: 352}, 274: {1: 273, 2: 352}, 275: {1: 274, 2: 352}, 276: {1: 275, 2: 352}, 277: {1: 276, 2: 352}, 278: {1: 277, 2: 352}, 279: {1: 278, 2: 352}, 280: {1: 279, 2: 352}, 281: {1: 280, 2: 352}, 282: {1: 281, 2: 352}, 283: {1: 282, 2: 352}, 284: {1: 283, 2: 352}, 285: {1: 284, 2: 352}, 286: {1: 285, 2: 352}, 287: {1: 286, 2: 352}, 288: {1: 287, 2: 352}, 289: {1: 288, 2: 352}, 290: {1: 289, 2: 352}, 291: {1: 290, 2: 352}, 292: {1: 291, 2: 352}, 293: {1: 292, 2: 352}, 294: {1: 293, 2: 352}, 295: {1: 294, 2: 352}, 296: {1: 295, 2: 352}, 297: {1: 296, 2: 352}, 298: {1: 297, 2: 352}, 299: {1: 298, 2: 352}, 300: {1: 299, 2: 352}, 301: {1: 300, 2: 352}, 302: {1: 301, 2: 352}, 303: {1: 302, 2: 352}, 304: {1: 303, 2: 352}, 305: {1: 304, 2: 352}, 306: {1: 305, 2: 352}, 307: {1: 306, 2: 352}, 308: {1: 307, 2: 352}, 309: {1: 308, 2: 352}, 310: {1: 309, 2: 352}, 311: {1: 310, 2: 352}, 312: {1: 311, 2: 352}, 313: {1: 312, 2: 352}, 314: {1: 313, 2: 352}, 315: {1: 314, 2: 352}, 316: {1: 315, 2: 352}, 317: {1: 316, 2: 352}, 318: {1: 317, 2: 352}, 319: {1: 318, 2: 352}, 320: {1: 319, 2: 352}, 321: {1: 320, 2: 352}, 322: {1: 321, 2: 352}, 323: {1: 322, 2: 352}, 324: {1: 323, 2: 352}, 325: {1: 324, 2: 352}, 326: {1: 325, 2: 352}, 327: {1: 326, 2: 352}, 328: {1: 327, 2: 352}, 329: {1: 328, 2: 352}, 330: {1: 329, 2: 352}, 331: {1: 330, 2: 352}, 332: {1: 331, 2: 352}, 333: {1: 332, 2: 352}, 334: {1: 333, 2: 352}, 335: {1: 334, 2: 352}, 336: {1: 335, 2: 352}, 337: {1: 336, 2: 352}, 338: {1: 337, 2: 352}, 339: {1: 338, 2: 352}, 340: {1: 339, 2: 352}, 341: {1: 340, 2: 352}, 342: {1: 341, 2: 352}, 343: {1: 342, 2: 352}, 344: {1: 343, 2: 352}, 345: {1: 344, 2: 352}, 346: {1: 345, 2: 352}, 347: {1: 346, 2: 352}, 348: {1: 347, 2: 352}, 349: {1: 348, 2: 352}, 350: {1: 0, 2: 353}, 351: {1: 1, 2: 353}, 352: {1: 2, 2: 353}, 353: {1: 3, 2: 353}, 354: {1: 4, 2: 353}, 355: {1: 5, 2: 353}, 356: {1: 6, 2: 353}, 357: {1: 7, 2: 353}, 358: {1: 8, 2: 353}, 359: {1: 9, 2: 353}, 360: {1: 10, 2: 353}, 361: {1: 11, 2: 353}, 362: {1: 12, 2: 353}, 363: {1: 13, 2: 353}, 364: {1: 14, 2: 353}, 365: {1: 15, 2: 353}, 366: {1: 16, 2: 353}, 367: {1: 17, 2: 353}, 368: {1: 18, 2: 353}, 369: {1: 19, 2: 353}, 370: {1: 20, 2: 353}, 371: {1: 21, 2: 353}, 372: {1: 22, 2: 353}, 373: {1: 23, 2: 353}, 374: {1: 24, 2: 353}, 375: {1: 25, 2: 353}, 376: {1: 26, 2: 353}, 377: {1: 27, 2: 353}, 378: {1: 28, 2: 353}, 379: {1: 29, 2: 353}, 380: {1: 30, 2: 353}, 381: {1: 31, 2: 353}, 382: {1: 32, 2: 353}, 383: {1: 33, 2: 353}, 384: {1: 34, 2: 353}, 385: {1: 35, 2: 353}, 386: {1: 36, 2: 353}, 387: {1: 37, 2: 353}, 388: {1: 38, 2: 353}, 389: {1: 39, 2: 353}, 390: {1: 40, 2: 353}, 391: {1: 41, 2: 353}, 392: {1: 42, 2: 353}, 393: {1: 43, 2: 353}, 394: {1: 44, 2: 353}, 395: {1: 45, 2: 353}, 396: {1: 46, 2: 353}, 397: {1: 47, 2: 353}, 398: {1: 48, 2: 353}, 399: {1: 49, 2: 353}, 400: {1: 50, 2: 353}, 401: {1: 51, 2: 353}, 402: {1: 52, 2: 353}, 403: {1: 53, 2: 353}, 404: {1: 54, 2: 353}, 405: {1: 55, 2: 353}, 406: {1: 56, 2: 353}, 407: {1: 57, 2: 353}, 408: {1: 58, 2: 353}, 409: {1: 59, 2: 353}, 410: {1: 60, 2: 353}, 411: {1: 61, 2: 353}, 412: {1: 62, 2: 353}, 413: {1: 63, 2: 353}, 414: {1: 64, 2: 353}, 415: {1: 65, 2: 353}, 416: {1: 66, 2: 353}, 417: {1: 67, 2: 353}, 418: {1: 68, 2: 353}, 419: {1: 69, 2: 353}, 420: {1: 70, 2: 353}, 421: {1: 71, 2: 353}, 422: {1: 72, 2: 353}, 423: {1: 73, 2: 353}, 424: {1: 74, 2: 353}, 425: {1: 75, 2: 353}, 426: {1: 76, 2: 353}, 427: {1: 77, 2: 353}, 428: {1: 78, 2: 353}, 429: {1: 79, 2: 353}, 430: {1: 80, 2: 353}, 431: {1: 81, 2: 353}, 432: {1: 82, 2: 353}, 433: {1: 83, 2: 353}, 434: {1: 84, 2: 353}, 435: {1: 85, 2: 353}, 436: {1: 86, 2: 353}, 437: {1: 87, 2: 353}, 438: {1: 88, 2: 353}, 439: {1: 89, 2: 353}, 440: {1: 90, 2: 353}, 441: {1: 91, 2: 353}, 442: {1: 92, 2: 353}, 443: {1: 93, 2: 353}, 444: {1: 94, 2: 353}, 445: {1: 95, 2: 353}, 446: {1: 96, 2: 353}, 447: {1: 97, 2: 353}, 448: {1: 98, 2: 353}, 449: {1: 99, 2: 353}, 450: {1: 100, 2: 353}, 451: {1: 101, 2: 353}, 452: {1: 102, 2: 353}, 453: {1: 103, 2: 353}, 454: {1: 104, 2: 353}, 455: {1: 105, 2: 353}, 456: {1: 106, 2: 353}, 457: {1: 107, 2: 353}, 458: {1: 108, 2: 353}, 459: {1: 109, 2: 353}, 460: {1: 110, 2: 353}, 461: {1: 111, 2: 353}, 462: {1: 112, 2: 353}, 463: {1: 113, 2: 353}, 464: {1: 114, 2: 353}, 465: {1: 115, 2: 353}, 466: {1: 116, 2: 353}, 467: {1: 117, 2: 353}, 468: {1: 118, 2: 353}, 469: {1: 119, 2: 353}, 470: {1: 120, 2: 353}, 471: {1: 121, 2: 353}, 472: {1: 122, 2: 353}, 473: {1: 123, 2: 353}, 474: {1: 124, 2: 353}, 475: {1: 125, 2: 353}, 476: {1: 126, 2: 353}, 477: {1: 127, 2: 353}, 478: {1: 128, 2: 353}, 479: {1: 129, 2: 353}, 480: {1: 130, 2: 353}, 481: {1: 131, 2: 353}, 482: {1: 132, 2: 353}, 483: {1: 133, 2: 353}, 484: {1: 134, 2: 353}, 485: {1: 135, 2: 353}, 486: {1: 136, 2: 353}, 487: {1: 137, 2: 353}, 488: {1: 138, 2: 353}, 489: {1: 139, 2: 353}, 490: {1: 140, 2: 353}, 491: {1: 141, 2: 353}, 492: {1: 142, 2: 353}, 493: {1: 143, 2: 353}, 494: {1: 144, 2: 353}, 495: {1: 145, 2: 353}, 496: {1: 146, 2: 353}, 497: {1: 147, 2: 353}, 498: {1: 148, 2: 353}, 499: {1: 149, 2: 353}, 500: {1: 150, 2: 353}, 501: {1: 151, 2: 353}, 502: {1: 152, 2: 353}, 503: {1: 153, 2: 353}, 504: {1: 154, 2: 353}, 505: {1: 155, 2: 353}, 506: {1: 156, 2: 353}, 507: {1: 157, 2: 353}, 508: {1: 158, 2: 353}, 509: {1: 159, 2: 353}, 510: {1: 160, 2: 353}, 511: {1: 161, 2: 353}, 512: {1: 162, 2: 353}, 513: {1: 163, 2: 353}, 514: {1: 164, 2: 353}, 515: {1: 165, 2: 353}, 516: {1: 166, 2: 353}, 517: {1: 167, 2: 353}, 518: {1: 168, 2: 353}, 519: {1: 169, 2: 353}, 520: {1: 170, 2: 353}, 521: {1: 171, 2: 353}, 522: {1: 172, 2: 353}, 523: {1: 173, 2: 353}, 524: {1: 174, 2: 353}, 525: {1: 175, 2: 353}, 526: {1: 176, 2: 353}, 527: {1: 177, 2: 353}, 528: {1: 178, 2: 353}, 529: {1: 179, 2: 353}, 530: {1: 180, 2: 353}, 531: {1: 181, 2: 353}, 532: {1: 182, 2: 353}, 533: {1: 183, 2: 353}, 534: {1: 184, 2: 353}, 535: {1: 185, 2: 353}, 536: {1: 186, 2: 353}, 537: {1: 187, 2: 353}, 538: {1: 188, 2: 353}, 539: {1: 189, 2: 353}, 540: {1: 190, 2: 353}, 541: {1: 191, 2: 353}, 542: {1: 192, 2: 353}, 543: {1: 193, 2: 353}, 544: {1: 194, 2: 353}, 545: {1: 195, 2: 353}, 546: {1: 196, 2: 353}, 547: {1: 197, 2: 353}, 548: {1: 198, 2: 353}, 549: {1: 199, 2: 353}, 550: {1: 200, 2: 353}, 551: {1: 201, 2: 353}, 552: {1: 202, 2: 353}, 553: {1: 203, 2: 353}, 554: {1: 204, 2: 353}, 555: {1: 205, 2: 353}, 556: {1: 206, 2: 353}, 557: {1: 207, 2: 353}, 558: {1: 208, 2: 353}, 559: {1: 209, 2: 353}, 560: {1: 210, 2: 353}, 561: {1: 211, 2: 353}, 562: {1: 212, 2: 353}, 563: {1: 213, 2: 353}, 564: {1: 214, 2: 353}, 565: {1: 215, 2: 353}, 566: {1: 216, 2: 353}, 567: {1: 217, 2: 353}, 568: {1: 218, 2: 353}, 569: {1: 219, 2: 353}, 570: {1: 220, 2: 353}, 571: {1: 221, 2: 353}, 572: {1: 222, 2: 353}, 573: {1: 223, 2: 353}, 574: {1: 224, 2: 353}, 575: {1: 225, 2: 353}, 576: {1: 226, 2: 353}, 577: {1: 227, 2: 353}, 578: {1: 228, 2: 353}, 579: {1: 229, 2: 353}, 580: {1: 230, 2: 353}, 581: {1: 231, 2: 353}, 582: {1: 232, 2: 353}, 583: {1: 233, 2: 353}, 584: {1: 234, 2: 353}, 585: {1: 235, 2: 353}, 586: {1: 236, 2: 353}, 587: {1: 237, 2: 353}, 588: {1: 238, 2: 353}, 589: {1: 239, 2: 353}, 590: {1: 240, 2: 353}, 591: {1: 241, 2: 353}, 592: {1: 242, 2: 353}, 593: {1: 243, 2: 353}, 594: {1: 244, 2: 353}, 595: {1: 245, 2: 353}, 596: {1: 246, 2: 353}, 597: {1: 247, 2: 353}, 598: {1: 248, 2: 353}, 599: {1: 249, 2: 353}, 600: {1: 250, 2: 353}, 601: {1: 251, 2: 353}, 602: {1: 252, 2: 353}, 603: {1: 253, 2: 353}, 604: {1: 254, 2: 353}, 605: {1: 255, 2: 353}, 606: {1: 256, 2: 353}, 607: {1: 257, 2: 353}, 608: {1: 258, 2: 353}, 609: {1: 259, 2: 353}, 610: {1: 260, 2: 353}, 611: {1: 261, 2: 353}, 612: {1: 262, 2: 353}, 613: {1: 263, 2: 353}, 614: {1: 264, 2: 353}, 615: {1: 265, 2: 353}, 616: {1: 266, 2: 353}, 617: {1: 267, 2: 353}, 618: {1: 268, 2: 353}, 619: {1: 269, 2: 353}, 620: {1: 270, 2: 353}, 621: {1: 271, 2: 353}, 622: {1: 272, 2: 353}, 623: {1: 273, 2: 353}, 624: {1: 274, 2: 353}, 625: {1: 275, 2: 353}, 626: {1: 276, 2: 353}, 627: {1: 277, 2: 353}, 628: {1: 278, 2: 353}, 629: {1: 279, 2: 353}, 630: {1: 280, 2: 353}, 631: {1: 281, 2: 353}, 632: {1: 282, 2: 353}, 633: {1: 283, 2: 353}, 634: {1: 284, 2: 353}, 635: {1: 285, 2: 353}, 636: {1: 286, 2: 353}, 637: {1: 287, 2: 353}, 638: {1: 288, 2: 353}, 639: {1: 289, 2: 353}, 640: {1: 290, 2: 353}, 641: {1: 291, 2: 353}, 642: {1: 292, 2: 353}, 643: {1: 293, 2: 353}, 644: {1: 294, 2: 353}, 645: {1: 295, 2: 353}, 646: {1: 296, 2: 353}, 647: {1: 297, 2: 353}, 648: {1: 298, 2: 353}, 649: {1: 299, 2: 353}, 650: {1: 300, 2: 353}, 651: {1: 301, 2: 353}, 652: {1: 302, 2: 353}, 653: {1: 303, 2: 353}, 654: {1: 304, 2: 353}, 655: {1: 305, 2: 353}, 656: {1: 306, 2: 353}, 657: {1: 307, 2: 353}, 658: {1: 308, 2: 353}, 659: {1: 309, 2: 353}, 660: {1: 310, 2: 353}, 661: {1: 311, 2: 353}, 662: {1: 312, 2: 353}, 663: {1: 313, 2: 353}, 664: {1: 314, 2: 353}, 665: {1: 315, 2: 353}, 666: {1: 316, 2: 353}, 667: {1: 317, 2: 353}, 668: {1: 318, 2: 353}, 669: {1: 319, 2: 353}, 670: {1: 320, 2: 353}, 671: {1: 321, 2: 353}, 672: {1: 322, 2: 353}, 673: {1: 323, 2: 353}, 674: {1: 324, 2: 353}, 675: {1: 325, 2: 353}, 676: {1: 326, 2: 353}, 677: {1: 327, 2: 353}, 678: {1: 328, 2: 353}, 679: {1: 329, 2: 353}, 680: {1: 330, 2: 353}, 681: {1: 331, 2: 353}, 682: {1: 332, 2: 353}, 683: {1: 333, 2: 353}, 684: {1: 334, 2: 353}, 685: {1: 335, 2: 353}, 686: {1: 336, 2: 353}, 687: {1: 337, 2: 353}, 688: {1: 338, 2: 353}, 689: {1: 339, 2: 353}, 690: {1: 340, 2: 353}, 691: {1: 341, 2: 353}, 692: {1: 342, 2: 353}, 693: {1: 343, 2: 353}, 694: {1: 344, 2: 353}, 695: {1: 345, 2: 353}, 696: {1: 346, 2: 353}, 697: {1: 347, 2: 353}, 698: {1: 348, 2: 353}, 699: {1: 0, 2: 354}, 700: {1: 1, 2: 354}, 701: {1: 2, 2: 354}, 702: {1: 3, 2: 354}, 703: {1: 4, 2: 354}, 704: {1: 5, 2: 354}, 705: {1: 6, 2: 354}, 706: {1: 7, 2: 354}, 707: {1: 8, 2: 354}, 708: {1: 9, 2: 354}, 709: {1: 10, 2: 354}, 710: {1: 11, 2: 354}, 711: {1: 12, 2: 354}, 712: {1: 13, 2: 354}, 713: {1: 14, 2: 354}, 714: {1: 15, 2: 354}, 715: {1: 16, 2: 354}, 716: {1: 17, 2: 354}, 717: {1: 18, 2: 354}, 718: {1: 19, 2: 354}, 719: {1: 20, 2: 354}, 720: {1: 21, 2: 354}, 721: {1: 22, 2: 354}, 722: {1: 23, 2: 354}, 723: {1: 24, 2: 354}, 724: {1: 25, 2: 354}, 725: {1: 26, 2: 354}, 726: {1: 27, 2: 354}, 727: {1: 28, 2: 354}, 728: {1: 29, 2: 354}, 729: {1: 30, 2: 354}, 730: {1: 31, 2: 354}, 731: {1: 32, 2: 354}, 732: {1: 33, 2: 354}, 733: {1: 34, 2: 354}, 734: {1: 35, 2: 354}, 735: {1: 36, 2: 354}, 736: {1: 37, 2: 354}, 737: {1: 38, 2: 354}, 738: {1: 39, 2: 354}, 739: {1: 40, 2: 354}, 740: {1: 41, 2: 354}, 741: {1: 42, 2: 354}, 742: {1: 43, 2: 354}, 743: {1: 44, 2: 354}, 744: {1: 45, 2: 354}, 745: {1: 46, 2: 354}, 746: {1: 47, 2: 354}, 747: {1: 48, 2: 354}, 748: {1: 49, 2: 354}, 749: {1: 50, 2: 354}, 750: {1: 51, 2: 354}, 751: {1: 52, 2: 354}, 752: {1: 53, 2: 354}, 753: {1: 54, 2: 354}, 754: {1: 55, 2: 354}, 755: {1: 56, 2: 354}, 756: {1: 57, 2: 354}, 757: {1: 58, 2: 354}, 758: {1: 59, 2: 354}, 759: {1: 60, 2: 354}, 760: {1: 61, 2: 354}, 761: {1: 62, 2: 354}, 762: {1: 63, 2: 354}, 763: {1: 64, 2: 354}, 764: {1: 65, 2: 354}, 765: {1: 66, 2: 354}, 766: {1: 67, 2: 354}, 767: {1: 68, 2: 354}, 768: {1: 69, 2: 354}, 769: {1: 70, 2: 354}, 770: {1: 71, 2: 354}, 771: {1: 72, 2: 354}, 772: {1: 73, 2: 354}, 773: {1: 74, 2: 354}, 774: {1: 75, 2: 354}, 775: {1: 76, 2: 354}, 776: {1: 77, 2: 354}, 777: {1: 78, 2: 354}, 778: {1: 79, 2: 354}, 779: {1: 80, 2: 354}, 780: {1: 81, 2: 354}, 781: {1: 82, 2: 354}, 782: {1: 83, 2: 354}, 783: {1: 84, 2: 354}, 784: {1: 85, 2: 354}, 785: {1: 86, 2: 354}, 786: {1: 87, 2: 354}, 787: {1: 88, 2: 354}, 788: {1: 89, 2: 354}, 789: {1: 90, 2: 354}, 790: {1: 91, 2: 354}, 791: {1: 92, 2: 354}, 792: {1: 93, 2: 354}, 793: {1: 94, 2: 354}, 794: {1: 95, 2: 354}, 795: {1: 96, 2: 354}, 796: {1: 97, 2: 354}, 797: {1: 98, 2: 354}, 798: {1: 99, 2: 354}, 799: {1: 100, 2: 354}, 800: {1: 101, 2: 354}, 801: {1: 102, 2: 354}, 802: {1: 103, 2: 354}, 803: {1: 104, 2: 354}, 804: {1: 105, 2: 354}, 805: {1: 106, 2: 354}, 806: {1: 107, 2: 354}, 807: {1: 108, 2: 354}, 808: {1: 109, 2: 354}, 809: {1: 110, 2: 354}, 810: {1: 111, 2: 354}, 811: {1: 112, 2: 354}, 812: {1: 113, 2: 354}, 813: {1: 114, 2: 354}, 814: {1: 115, 2: 354}, 815: {1: 116, 2: 354}, 816: {1: 117, 2: 354}, 817: {1: 118, 2: 354}, 818: {1: 119, 2: 354}, 819: {1: 120, 2: 354}, 820: {1: 121, 2: 354}, 821: {1: 122, 2: 354}, 822: {1: 123, 2: 354}, 823: {1: 124, 2: 354}, 824: {1: 125, 2: 354}, 825: {1: 126, 2: 354}, 826: {1: 127, 2: 354}, 827: {1: 128, 2: 354}, 828: {1: 129, 2: 354}, 829: {1: 130, 2: 354}, 830: {1: 131, 2: 354}, 831: {1: 132, 2: 354}, 832: {1: 133, 2: 354}, 833: {1: 134, 2: 354}, 834: {1: 135, 2: 354}, 835: {1: 136, 2: 354}, 836: {1: 137, 2: 354}, 837: {1: 138, 2: 354}, 838: {1: 139, 2: 354}, 839: {1: 140, 2: 354}, 840: {1: 141, 2: 354}, 841: {1: 142, 2: 354}, 842: {1: 143, 2: 354}, 843: {1: 144, 2: 354}, 844: {1: 145, 2: 354}, 845: {1: 146, 2: 354}, 846: {1: 147, 2: 354}, 847: {1: 148, 2: 354}, 848: {1: 149, 2: 354}, 849: {1: 150, 2: 354}, 850: {1: 151, 2: 354}, 851: {1: 152, 2: 354}, 852: {1: 153, 2: 354}, 853: {1: 154, 2: 354}, 854: {1: 155, 2: 354}, 855: {1: 156, 2: 354}, 856: {1: 157, 2: 354}, 857: {1: 158, 2: 354}, 858: {1: 159, 2: 354}, 859: {1: 160, 2: 354}, 860: {1: 161, 2: 354}, 861: {1: 162, 2: 354}, 862: {1: 163, 2: 354}, 863: {1: 164, 2: 354}, 864: {1: 165, 2: 354}, 865: {1: 166, 2: 354}, 866: {1: 167, 2: 354}, 867: {1: 168, 2: 354}, 868: {1: 169, 2: 354}, 869: {1: 170, 2: 354}, 870: {1: 171, 2: 354}, 871: {1: 172, 2: 354}, 872: {1: 173, 2: 354}, 873: {1: 174, 2: 354}, 874: {1: 175, 2: 354}, 875: {1: 176, 2: 354}, 876: {1: 177, 2: 354}, 877: {1: 178, 2: 354}, 878: {1: 179, 2: 354}, 879: {1: 180, 2: 354}, 880: {1: 181, 2: 354}, 881: {1: 182, 2: 354}, 882: {1: 183, 2: 354}, 883: {1: 184, 2: 354}, 884: {1: 185, 2: 354}, 885: {1: 186, 2: 354}, 886: {1: 187, 2: 354}, 887: {1: 188, 2: 354}, 888: {1: 189, 2: 354}, 889: {1: 190, 2: 354}, 890: {1: 191, 2: 354}, 891: {1: 192, 2: 354}, 892: {1: 193, 2: 354}, 893: {1: 194, 2: 354}, 894: {1: 195, 2: 354}, 895: {1: 196, 2: 354}, 896: {1: 197, 2: 354}, 897: {1: 198, 2: 354}, 898: {1: 199, 2: 354}, 899: {1: 200, 2: 354}, 900: {1: 201, 2: 354}, 901: {1: 202, 2: 354}, 902: {1: 203, 2: 354}, 903: {1: 204, 2: 354}, 904: {1: 205, 2: 354}, 905: {1: 206, 2: 354}, 906: {1: 207, 2: 354}, 907: {1: 208, 2: 354}, 908: {1: 209, 2: 354}, 909: {1: 210, 2: 354}, 910: {1: 211, 2: 354}, 911: {1: 212, 2: 354}, 912: {1: 213, 2: 354}, 913: {1: 214, 2: 354}, 914: {1: 215, 2: 354}, 915: {1: 216, 2: 354}, 916: {1: 217, 2: 354}, 917: {1: 218, 2: 354}, 918: {1: 219, 2: 354}, 919: {1: 220, 2: 354}, 920: {1: 221, 2: 354}, 921: {1: 222, 2: 354}, 922: {1: 223, 2: 354}, 923: {1: 224, 2: 354}, 924: {1: 225, 2: 354}, 925: {1: 226, 2: 354}, 926: {1: 227, 2: 354}, 927: {1: 228, 2: 354}, 928: {1: 229, 2: 354}, 929: {1: 230, 2: 354}, 930: {1: 231, 2: 354}, 931: {1: 232, 2: 354}, 932: {1: 233, 2: 354}, 933: {1: 234, 2: 354}, 934: {1: 235, 2: 354}, 935: {1: 236, 2: 354}, 936: {1: 237, 2: 354}, 937: {1: 238, 2: 354}, 938: {1: 239, 2: 354}, 939: {1: 240, 2: 354}, 940: {1: 241, 2: 354}, 941: {1: 242, 2: 354}, 942: {1: 243, 2: 354}, 943: {1: 244, 2: 354}, 944: {1: 245, 2: 354}, 945: {1: 246, 2: 354}, 946: {1: 247, 2: 354}, 947: {1: 248, 2: 354}, 948: {1: 249, 2: 354}, 949: {1: 250, 2: 354}, 950: {1: 251, 2: 354}, 951: {1: 252, 2: 354}, 952: {1: 253, 2: 354}, 953: {1: 254, 2: 354}, 954: {1: 255, 2: 354}, 955: {1: 256, 2: 354}, 956: {1: 257, 2: 354}, 957: {1: 258, 2: 354}, 958: {1: 259, 2: 354}, 959: {1: 260, 2: 354}, 960: {1: 261, 2: 354}, 961: {1: 262, 2: 354}, 962: {1: 263, 2: 354}, 963: {1: 264, 2: 354}, 964: {1: 265, 2: 354}, 965: {1: 266, 2: 354}, 966: {1: 267, 2: 354}, 967: {1: 268, 2: 354}, 968: {1: 269, 2: 354}, 969: {1: 270, 2: 354}, 970: {1: 271, 2: 354}, 971: {1: 272, 2: 354}, 972: {1: 273, 2: 354}, 973: {1: 274, 2: 354}, 974: {1: 275, 2: 354}, 975: {1: 276, 2: 354}, 976: {1: 277, 2: 354}, 977: {1: 278, 2: 354}, 978: {1: 279, 2: 354}, 979: {1: 280, 2: 354}, 980: {1: 281, 2: 354}, 981: {1: 282, 2: 354}, 982: {1: 283, 2: 354}, 983: {1: 284, 2: 354}, 984: {1: 285, 2: 354}, 985: {1: 286, 2: 354}, 986: {1: 287, 2: 354}, 987: {1: 288, 2: 354}, 988: {1: 289, 2: 354}, 989: {1: 290, 2: 354}, 990: {1: 291, 2: 354}, 991: {1: 292, 2: 354}, 992: {1: 293, 2: 354}, 993: {1: 294, 2: 354}, 994: {1: 295, 2: 354}, 995: {1: 296, 2: 354}, 996: {1: 297, 2: 354}, 997: {1: 298, 2: 354}, 998: {1: 299, 2: 354}, 999: {1: 300, 2: 354}, 1000: {1: 301, 2: 354}, 1001: {1: 302, 2: 354}, 1002: {1: 303, 2: 354}, 1003: {1: 304, 2: 354}, 1004: {1: 305, 2: 354}, 1005: {1: 306, 2: 354}, 1006: {1: 307, 2: 354}, 1007: {1: 308, 2: 354}, 1008: {1: 309, 2: 354}, 1009: {1: 310, 2: 354}, 1010: {1: 311, 2: 354}, 1011: {1: 312, 2: 354}, 1012: {1: 313, 2: 354}, 1013: {1: 314, 2: 354}, 1014: {1: 315, 2: 354}, 1015: {1: 316, 2: 354}, 1016: {1: 317, 2: 354}, 1017: {1: 318, 2: 354}, 1018: {1: 319, 2: 354}, 1019: {1: 320, 2: 354}, 1020: {1: 321, 2: 354}, 1021: {1: 322, 2: 354}, 1022: {1: 323, 2: 354}, 1023: {1: 324, 2: 354}, 1024: {1: 325, 2: 354}, 1025: {1: 326, 2: 354}, 1026: {1: 327, 2: 354}, 1027: {1: 328, 2: 354}, 1028: {1: 329, 2: 354}, 1029: {1: 330, 2: 354}, 1030: {1: 331, 2: 354}, 1031: {1: 332, 2: 354}, 1032: {1: 333, 2: 354}, 1033: {1: 334, 2: 354}, 1034: {1: 335, 2: 354}, 1035: {1: 336, 2: 354}, 1036: {1: 337, 2: 354}, 1037: {1: 338, 2: 354}, 1038: {1: 339, 2: 354}, 1039: {1: 340, 2: 354}, 1040: {1: 341, 2: 354}, 1041: {1: 342, 2: 354}, 1042: {1: 343, 2: 354}, 1043: {1: 344, 2: 354}, 1044: {1: 345, 2: 354}, 1045: {1: 346, 2: 354}, 1046: {1: 347, 2: 354}, 1047: {1: 348, 2: 354}, 1048: {1: 0, 2: 355}, 1049: {1: 1, 2: 355}, 1050: {1: 2, 2: 355}, 1051: {1: 3, 2: 355}, 1052: {1: 4, 2: 355}, 1053: {1: 5, 2: 355}, 1054: {1: 6, 2: 355}, 1055: {1: 7, 2: 355}, 1056: {1: 8, 2: 355}, 1057: {1: 9, 2: 355}, 1058: {1: 10, 2: 355}, 1059: {1: 11, 2: 355}, 1060: {1: 12, 2: 355}, 1061: {1: 13, 2: 355}, 1062: {1: 14, 2: 355}, 1063: {1: 15, 2: 355}, 1064: {1: 16, 2: 355}, 1065: {1: 17, 2: 355}, 1066: {1: 18, 2: 355}, 1067: {1: 19, 2: 355}, 1068: {1: 20, 2: 355}, 1069: {1: 21, 2: 355}, 1070: {1: 22, 2: 355}, 1071: {1: 23, 2: 355}, 1072: {1: 24, 2: 355}, 1073: {1: 25, 2: 355}, 1074: {1: 26, 2: 355}, 1075: {1: 27, 2: 355}, 1076: {1: 28, 2: 355}, 1077: {1: 29, 2: 355}, 1078: {1: 30, 2: 355}, 1079: {1: 31, 2: 355}, 1080: {1: 32, 2: 355}, 1081: {1: 33, 2: 355}, 1082: {1: 34, 2: 355}, 1083: {1: 35, 2: 355}, 1084: {1: 36, 2: 355}, 1085: {1: 37, 2: 355}, 1086: {1: 38, 2: 355}, 1087: {1: 39, 2: 355}, 1088: {1: 40, 2: 355}, 1089: {1: 41, 2: 355}, 1090: {1: 42, 2: 355}, 1091: {1: 43, 2: 355}, 1092: {1: 44, 2: 355}, 1093: {1: 45, 2: 355}, 1094: {1: 46, 2: 355}, 1095: {1: 47, 2: 355}, 1096: {1: 48, 2: 355}, 1097: {1: 49, 2: 355}, 1098: {1: 50, 2: 355}, 1099: {1: 51, 2: 355}, 1100: {1: 52, 2: 355}, 1101: {1: 53, 2: 355}, 1102: {1: 54, 2: 355}, 1103: {1: 55, 2: 355}, 1104: {1: 56, 2: 355}, 1105: {1: 57, 2: 355}, 1106: {1: 58, 2: 355}, 1107: {1: 59, 2: 355}, 1108: {1: 60, 2: 355}, 1109: {1: 61, 2: 355}, 1110: {1: 62, 2: 355}, 1111: {1: 63, 2: 355}, 1112: {1: 64, 2: 355}, 1113: {1: 65, 2: 355}, 1114: {1: 66, 2: 355}, 1115: {1: 67, 2: 355}, 1116: {1: 68, 2: 355}, 1117: {1: 69, 2: 355}, 1118: {1: 70, 2: 355}, 1119: {1: 71, 2: 355}, 1120: {1: 72, 2: 355}, 1121: {1: 73, 2: 355}, 1122: {1: 74, 2: 355}, 1123: {1: 75, 2: 355}, 1124: {1: 76, 2: 355}, 1125: {1: 77, 2: 355}, 1126: {1: 78, 2: 355}, 1127: {1: 79, 2: 355}, 1128: {1: 80, 2: 355}, 1129: {1: 81, 2: 355}, 1130: {1: 82, 2: 355}, 1131: {1: 83, 2: 355}, 1132: {1: 84, 2: 355}, 1133: {1: 85, 2: 355}, 1134: {1: 86, 2: 355}, 1135: {1: 87, 2: 355}, 1136: {1: 88, 2: 355}, 1137: {1: 89, 2: 355}, 1138: {1: 90, 2: 355}, 1139: {1: 91, 2: 355}, 1140: {1: 92, 2: 355}, 1141: {1: 93, 2: 355}, 1142: {1: 94, 2: 355}, 1143: {1: 95, 2: 355}, 1144: {1: 96, 2: 355}, 1145: {1: 97, 2: 355}, 1146: {1: 98, 2: 355}, 1147: {1: 99, 2: 355}, 1148: {1: 100, 2: 355}, 1149: {1: 101, 2: 355}, 1150: {1: 102, 2: 355}, 1151: {1: 103, 2: 355}, 1152: {1: 104, 2: 355}, 1153: {1: 105, 2: 355}, 1154: {1: 106, 2: 355}, 1155: {1: 107, 2: 355}, 1156: {1: 108, 2: 355}, 1157: {1: 109, 2: 355}, 1158: {1: 110, 2: 355}, 1159: {1: 111, 2: 355}, 1160: {1: 112, 2: 355}, 1161: {1: 113, 2: 355}, 1162: {1: 114, 2: 355}, 1163: {1: 115, 2: 355}, 1164: {1: 116, 2: 355}, 1165: {1: 117, 2: 355}, 1166: {1: 118, 2: 355}, 1167: {1: 119, 2: 355}, 1168: {1: 120, 2: 355}, 1169: {1: 121, 2: 355}, 1170: {1: 122, 2: 355}, 1171: {1: 123, 2: 355}, 1172: {1: 124, 2: 355}, 1173: {1: 125, 2: 355}, 1174: {1: 126, 2: 355}, 1175: {1: 127, 2: 355}, 1176: {1: 128, 2: 355}, 1177: {1: 129, 2: 355}, 1178: {1: 130, 2: 355}, 1179: {1: 131, 2: 355}, 1180: {1: 132, 2: 355}, 1181: {1: 133, 2: 355}, 1182: {1: 134, 2: 355}, 1183: {1: 135, 2: 355}, 1184: {1: 136, 2: 355}, 1185: {1: 137, 2: 355}, 1186: {1: 138, 2: 355}, 1187: {1: 139, 2: 355}, 1188: {1: 140, 2: 355}, 1189: {1: 141, 2: 355}, 1190: {1: 142, 2: 355}, 1191: {1: 143, 2: 355}, 1192: {1: 144, 2: 355}, 1193: {1: 145, 2: 355}, 1194: {1: 146, 2: 355}, 1195: {1: 147, 2: 355}, 1196: {1: 148, 2: 355}, 1197: {1: 149, 2: 355}, 1198: {1: 150, 2: 355}, 1199: {1: 151, 2: 355}, 1200: {1: 152, 2: 355}, 1201: {1: 153, 2: 355}, 1202: {1: 154, 2: 355}, 1203: {1: 155, 2: 355}, 1204: {1: 156, 2: 355}, 1205: {1: 157, 2: 355}, 1206: {1: 158, 2: 355}, 1207: {1: 159, 2: 355}, 1208: {1: 160, 2: 355}, 1209: {1: 161, 2: 355}, 1210: {1: 162, 2: 355}, 1211: {1: 163, 2: 355}, 1212: {1: 164, 2: 355}, 1213: {1: 165, 2: 355}, 1214: {1: 166, 2: 355}, 1215: {1: 167, 2: 355}, 1216: {1: 168, 2: 355}, 1217: {1: 169, 2: 355}, 1218: {1: 170, 2: 355}, 1219: {1: 171, 2: 355}, 1220: {1: 172, 2: 355}, 1221: {1: 173, 2: 355}, 1222: {1: 174, 2: 355}, 1223: {1: 175, 2: 355}, 1224: {1: 176, 2: 355}, 1225: {1: 177, 2: 355}, 1226: {1: 178, 2: 355}, 1227: {1: 179, 2: 355}, 1228: {1: 180, 2: 355}, 1229: {1: 181, 2: 355}, 1230: {1: 182, 2: 355}, 1231: {1: 183, 2: 355}, 1232: {1: 184, 2: 355}, 1233: {1: 185, 2: 355}, 1234: {1: 186, 2: 355}, 1235: {1: 187, 2: 355}, 1236: {1: 188, 2: 355}, 1237: {1: 189, 2: 355}, 1238: {1: 190, 2: 355}, 1239: {1: 191, 2: 355}, 1240: {1: 192, 2: 355}, 1241: {1: 193, 2: 355}, 1242: {1: 194, 2: 355}, 1243: {1: 195, 2: 355}, 1244: {1: 196, 2: 355}, 1245: {1: 197, 2: 355}, 1246: {1: 198, 2: 355}, 1247: {1: 199, 2: 355}, 1248: {1: 200, 2: 355}, 1249: {1: 201, 2: 355}, 1250: {1: 202, 2: 355}, 1251: {1: 203, 2: 355}, 1252: {1: 204, 2: 355}, 1253: {1: 205, 2: 355}, 1254: {1: 206, 2: 355}, 1255: {1: 207, 2: 355}, 1256: {1: 208, 2: 355}, 1257: {1: 209, 2: 355}, 1258: {1: 210, 2: 355}, 1259: {1: 211, 2: 355}, 1260: {1: 212, 2: 355}, 1261: {1: 213, 2: 355}, 1262: {1: 214, 2: 355}, 1263: {1: 215, 2: 355}, 1264: {1: 216, 2: 355}, 1265: {1: 217, 2: 355}, 1266: {1: 218, 2: 355}, 1267: {1: 219, 2: 355}, 1268: {1: 220, 2: 355}, 1269: {1: 221, 2: 355}, 1270: {1: 222, 2: 355}, 1271: {1: 223, 2: 355}, 1272: {1: 224, 2: 355}, 1273: {1: 225, 2: 355}, 1274: {1: 226, 2: 355}, 1275: {1: 227, 2: 355}, 1276: {1: 228, 2: 355}, 1277: {1: 229, 2: 355}, 1278: {1: 230, 2: 355}, 1279: {1: 231, 2: 355}, 1280: {1: 232, 2: 355}, 1281: {1: 233, 2: 355}, 1282: {1: 234, 2: 355}, 1283: {1: 235, 2: 355}, 1284: {1: 236, 2: 355}, 1285: {1: 237, 2: 355}, 1286: {1: 238, 2: 355}, 1287: {1: 239, 2: 355}, 1288: {1: 240, 2: 355}, 1289: {1: 241, 2: 355}, 1290: {1: 242, 2: 355}, 1291: {1: 243, 2: 355}, 1292: {1: 244, 2: 355}, 1293: {1: 245, 2: 355}, 1294: {1: 246, 2: 355}, 1295: {1: 247, 2: 355}, 1296: {1: 248, 2: 355}, 1297: {1: 249, 2: 355}, 1298: {1: 250, 2: 355}, 1299: {1: 251, 2: 355}, 1300: {1: 252, 2: 355}, 1301: {1: 253, 2: 355}, 1302: {1: 254, 2: 355}, 1303: {1: 255, 2: 355}, 1304: {1: 256, 2: 355}, 1305: {1: 257, 2: 355}, 1306: {1: 258, 2: 355}, 1307: {1: 259, 2: 355}, 1308: {1: 260, 2: 355}, 1309: {1: 261, 2: 355}, 1310: {1: 262, 2: 355}, 1311: {1: 263, 2: 355}, 1312: {1: 264, 2: 355}, 1313: {1: 265, 2: 355}, 1314: {1: 266, 2: 355}, 1315: {1: 267, 2: 355}, 1316: {1: 268, 2: 355}, 1317: {1: 269, 2: 355}, 1318: {1: 270, 2: 355}, 1319: {1: 271, 2: 355}, 1320: {1: 272, 2: 355}, 1321: {1: 273, 2: 355}, 1322: {1: 274, 2: 355}, 1323: {1: 275, 2: 355}, 1324: {1: 276, 2: 355}, 1325: {1: 277, 2: 355}, 1326: {1: 278, 2: 355}, 1327: {1: 279, 2: 355}, 1328: {1: 280, 2: 355}, 1329: {1: 281, 2: 355}, 1330: {1: 282, 2: 355}, 1331: {1: 283, 2: 355}, 1332: {1: 284, 2: 355}, 1333: {1: 285, 2: 355}, 1334: {1: 286, 2: 355}, 1335: {1: 287, 2: 355}, 1336: {1: 288, 2: 355}, 1337: {1: 289, 2: 355}, 1338: {1: 290, 2: 355}, 1339: {1: 291, 2: 355}, 1340: {1: 292, 2: 355}, 1341: {1: 293, 2: 355}, 1342: {1: 294, 2: 355}, 1343: {1: 295, 2: 355}, 1344: {1: 296, 2: 355}, 1345: {1: 297, 2: 355}, 1346: {1: 298, 2: 355}, 1347: {1: 299, 2: 355}, 1348: {1: 300, 2: 355}, 1349: {1: 301, 2: 355}, 1350: {1: 302, 2: 355}, 1351: {1: 303, 2: 355}, 1352: {1: 304, 2: 355}, 1353: {1: 305, 2: 355}, 1354: {1: 306, 2: 355}, 1355: {1: 307, 2: 355}, 1356: {1: 308, 2: 355}, 1357: {1: 309, 2: 355}, 1358: {1: 310, 2: 355}, 1359: {1: 311, 2: 355}, 1360: {1: 312, 2: 355}, 1361: {1: 313, 2: 355}, 1362: {1: 314, 2: 355}, 1363: {1: 315, 2: 355}, 1364: {1: 316, 2: 355}, 1365: {1: 317, 2: 355}, 1366: {1: 318, 2: 355}, 1367: {1: 319, 2: 355}, 1368: {1: 320, 2: 355}, 1369: {1: 321, 2: 355}, 1370: {1: 322, 2: 355}, 1371: {1: 323, 2: 355}, 1372: {1: 324, 2: 355}, 1373: {1: 325, 2: 355}, 1374: {1: 326, 2: 355}, 1375: {1: 327, 2: 355}, 1376: {1: 328, 2: 355}, 1377: {1: 329, 2: 355}, 1378: {1: 330, 2: 355}, 1379: {1: 331, 2: 355}, 1380: {1: 332, 2: 355}, 1381: {1: 333, 2: 355}, 1382: {1: 334, 2: 355}, 1383: {1: 335, 2: 355}, 1384: {1: 336, 2: 355}, 1385: {1: 337, 2: 355}, 1386: {1: 338, 2: 355}, 1387: {1: 339, 2: 355}, 1388: {1: 340, 2: 355}, 1389: {1: 341, 2: 355}, 1390: {1: 342, 2: 355}, 1391: {1: 343, 2: 355}, 1392: {1: 344, 2: 355}, 1393: {1: 345, 2: 355}, 1394: {1: 346, 2: 355}, 1395: {1: 347, 2: 355}, 1396: {1: 348, 2: 355}, 1397: {1: 0, 2: 356}, 1398: {1: 1, 2: 356}, 1399: {1: 2, 2: 356}, 1400: {1: 3, 2: 356}, 1401: {1: 4, 2: 356}, 1402: {1: 5, 2: 356}, 1403: {1: 6, 2: 356}, 1404: {1: 7, 2: 356}, 1405: {1: 8, 2: 356}, 1406: {1: 9, 2: 356}, 1407: {1: 10, 2: 356}, 1408: {1: 11, 2: 356}, 1409: {1: 12, 2: 356}, 1410: {1: 13, 2: 356}, 1411: {1: 14, 2: 356}, 1412: {1: 15, 2: 356}, 1413: {1: 16, 2: 356}, 1414: {1: 17, 2: 356}, 1415: {1: 18, 2: 356}, 1416: {1: 19, 2: 356}, 1417: {1: 20, 2: 356}, 1418: {1: 21, 2: 356}, 1419: {1: 22, 2: 356}, 1420: {1: 23, 2: 356}, 1421: {1: 24, 2: 356}, 1422: {1: 25, 2: 356}, 1423: {1: 26, 2: 356}, 1424: {1: 27, 2: 356}, 1425: {1: 28, 2: 356}, 1426: {1: 29, 2: 356}, 1427: {1: 30, 2: 356}, 1428: {1: 31, 2: 356}, 1429: {1: 32, 2: 356}, 1430: {1: 33, 2: 356}, 1431: {1: 34, 2: 356}, 1432: {1: 35, 2: 356}, 1433: {1: 36, 2: 356}, 1434: {1: 37, 2: 356}, 1435: {1: 38, 2: 356}, 1436: {1: 39, 2: 356}, 1437: {1: 40, 2: 356}, 1438: {1: 41, 2: 356}, 1439: {1: 42, 2: 356}, 1440: {1: 43, 2: 356}, 1441: {1: 44, 2: 356}, 1442: {1: 45, 2: 356}, 1443: {1: 46, 2: 356}, 1444: {1: 47, 2: 356}, 1445: {1: 48, 2: 356}, 1446: {1: 49, 2: 356}, 1447: {1: 50, 2: 356}, 1448: {1: 51, 2: 356}, 1449: {1: 52, 2: 356}, 1450: {1: 53, 2: 356}, 1451: {1: 54, 2: 356}, 1452: {1: 55, 2: 356}, 1453: {1: 56, 2: 356}, 1454: {1: 57, 2: 356}, 1455: {1: 58, 2: 356}, 1456: {1: 59, 2: 356}, 1457: {1: 60, 2: 356}, 1458: {1: 61, 2: 356}, 1459: {1: 62, 2: 356}, 1460: {1: 63, 2: 356}, 1461: {1: 64, 2: 356}, 1462: {1: 65, 2: 356}, 1463: {1: 66, 2: 356}, 1464: {1: 67, 2: 356}, 1465: {1: 68, 2: 356}, 1466: {1: 69, 2: 356}, 1467: {1: 70, 2: 356}, 1468: {1: 71, 2: 356}, 1469: {1: 72, 2: 356}, 1470: {1: 73, 2: 356}, 1471: {1: 74, 2: 356}, 1472: {1: 75, 2: 356}, 1473: {1: 76, 2: 356}, 1474: {1: 77, 2: 356}, 1475: {1: 78, 2: 356}, 1476: {1: 79, 2: 356}, 1477: {1: 80, 2: 356}, 1478: {1: 81, 2: 356}, 1479: {1: 82, 2: 356}, 1480: {1: 83, 2: 356}, 1481: {1: 84, 2: 356}, 1482: {1: 85, 2: 356}, 1483: {1: 86, 2: 356}, 1484: {1: 87, 2: 356}, 1485: {1: 88, 2: 356}, 1486: {1: 89, 2: 356}, 1487: {1: 90, 2: 356}, 1488: {1: 91, 2: 356}, 1489: {1: 92, 2: 356}, 1490: {1: 93, 2: 356}, 1491: {1: 94, 2: 356}, 1492: {1: 95, 2: 356}, 1493: {1: 96, 2: 356}, 1494: {1: 97, 2: 356}, 1495: {1: 98, 2: 356}, 1496: {1: 99, 2: 356}, 1497: {1: 100, 2: 356}, 1498: {1: 101, 2: 356}, 1499: {1: 102, 2: 356}, 1500: {1: 103, 2: 356}, 1501: {1: 104, 2: 356}, 1502: {1: 105, 2: 356}, 1503: {1: 106, 2: 356}, 1504: {1: 107, 2: 356}, 1505: {1: 108, 2: 356}, 1506: {1: 109, 2: 356}, 1507: {1: 110, 2: 356}, 1508: {1: 111, 2: 356}, 1509: {1: 112, 2: 356}, 1510: {1: 113, 2: 356}, 1511: {1: 114, 2: 356}, 1512: {1: 115, 2: 356}, 1513: {1: 116, 2: 356}, 1514: {1: 117, 2: 356}, 1515: {1: 118, 2: 356}, 1516: {1: 119, 2: 356}, 1517: {1: 120, 2: 356}, 1518: {1: 121, 2: 356}, 1519: {1: 122, 2: 356}, 1520: {1: 123, 2: 356}, 1521: {1: 124, 2: 356}, 1522: {1: 125, 2: 356}, 1523: {1: 126, 2: 356}, 1524: {1: 127, 2: 356}, 1525: {1: 128, 2: 356}, 1526: {1: 129, 2: 356}, 1527: {1: 130, 2: 356}, 1528: {1: 131, 2: 356}, 1529: {1: 132, 2: 356}, 1530: {1: 133, 2: 356}, 1531: {1: 134, 2: 356}, 1532: {1: 135, 2: 356}, 1533: {1: 136, 2: 356}, 1534: {1: 137, 2: 356}, 1535: {1: 138, 2: 356}, 1536: {1: 139, 2: 356}, 1537: {1: 140, 2: 356}, 1538: {1: 141, 2: 356}, 1539: {1: 142, 2: 356}, 1540: {1: 143, 2: 356}, 1541: {1: 144, 2: 356}, 1542: {1: 145, 2: 356}, 1543: {1: 146, 2: 356}, 1544: {1: 147, 2: 356}, 1545: {1: 148, 2: 356}, 1546: {1: 149, 2: 356}, 1547: {1: 150, 2: 356}, 1548: {1: 151, 2: 356}, 1549: {1: 152, 2: 356}, 1550: {1: 153, 2: 356}, 1551: {1: 154, 2: 356}, 1552: {1: 155, 2: 356}, 1553: {1: 156, 2: 356}, 1554: {1: 157, 2: 356}, 1555: {1: 158, 2: 356}, 1556: {1: 159, 2: 356}, 1557: {1: 160, 2: 356}, 1558: {1: 161, 2: 356}, 1559: {1: 162, 2: 356}, 1560: {1: 163, 2: 356}, 1561: {1: 164, 2: 356}, 1562: {1: 165, 2: 356}, 1563: {1: 166, 2: 356}, 1564: {1: 167, 2: 356}, 1565: {1: 168, 2: 356}, 1566: {1: 169, 2: 356}, 1567: {1: 170, 2: 356}, 1568: {1: 171, 2: 356}, 1569: {1: 172, 2: 356}, 1570: {1: 173, 2: 356}, 1571: {1: 174, 2: 356}, 1572: {1: 175, 2: 356}, 1573: {1: 176, 2: 356}, 1574: {1: 177, 2: 356}, 1575: {1: 178, 2: 356}, 1576: {1: 179, 2: 356}, 1577: {1: 180, 2: 356}, 1578: {1: 181, 2: 356}, 1579: {1: 182, 2: 356}, 1580: {1: 183, 2: 356}, 1581: {1: 184, 2: 356}, 1582: {1: 185, 2: 356}, 1583: {1: 186, 2: 356}, 1584: {1: 187, 2: 356}, 1585: {1: 188, 2: 356}, 1586: {1: 189, 2: 356}, 1587: {1: 190, 2: 356}, 1588: {1: 191, 2: 356}, 1589: {1: 192, 2: 356}, 1590: {1: 193, 2: 356}, 1591: {1: 194, 2: 356}, 1592: {1: 195, 2: 356}, 1593: {1: 196, 2: 356}, 1594: {1: 197, 2: 356}, 1595: {1: 198, 2: 356}, 1596: {1: 199, 2: 356}, 1597: {1: 200, 2: 356}, 1598: {1: 201, 2: 356}, 1599: {1: 202, 2: 356}, 1600: {1: 203, 2: 356}, 1601: {1: 204, 2: 356}, 1602: {1: 205, 2: 356}, 1603: {1: 206, 2: 356}, 1604: {1: 207, 2: 356}, 1605: {1: 208, 2: 356}, 1606: {1: 209, 2: 356}, 1607: {1: 210, 2: 356}, 1608: {1: 211, 2: 356}, 1609: {1: 212, 2: 356}, 1610: {1: 213, 2: 356}, 1611: {1: 214, 2: 356}, 1612: {1: 215, 2: 356}, 1613: {1: 216, 2: 356}, 1614: {1: 217, 2: 356}, 1615: {1: 218, 2: 356}, 1616: {1: 219, 2: 356}, 1617: {1: 220, 2: 356}, 1618: {1: 221, 2: 356}, 1619: {1: 222, 2: 356}, 1620: {1: 223, 2: 356}, 1621: {1: 224, 2: 356}, 1622: {1: 225, 2: 356}, 1623: {1: 226, 2: 356}, 1624: {1: 227, 2: 356}, 1625: {1: 228, 2: 356}, 1626: {1: 229, 2: 356}, 1627: {1: 230, 2: 356}, 1628: {1: 231, 2: 356}, 1629: {1: 232, 2: 356}, 1630: {1: 233, 2: 356}, 1631: {1: 234, 2: 356}, 1632: {1: 235, 2: 356}, 1633: {1: 236, 2: 356}, 1634: {1: 237, 2: 356}, 1635: {1: 238, 2: 356}, 1636: {1: 239, 2: 356}, 1637: {1: 240, 2: 356}, 1638: {1: 241, 2: 356}, 1639: {1: 242, 2: 356}, 1640: {1: 243, 2: 356}, 1641: {1: 244, 2: 356}, 1642: {1: 245, 2: 356}, 1643: {1: 246, 2: 356}, 1644: {1: 247, 2: 356}, 1645: {1: 248, 2: 356}, 1646: {1: 249, 2: 356}, 1647: {1: 250, 2: 356}, 1648: {1: 251, 2: 356}, 1649: {1: 252, 2: 356}, 1650: {1: 253, 2: 356}, 1651: {1: 254, 2: 356}, 1652: {1: 255, 2: 356}, 1653: {1: 256, 2: 356}, 1654: {1: 257, 2: 356}, 1655: {1: 258, 2: 356}, 1656: {1: 259, 2: 356}, 1657: {1: 260, 2: 356}, 1658: {1: 261, 2: 356}, 1659: {1: 262, 2: 356}, 1660: {1: 263, 2: 356}, 1661: {1: 264, 2: 356}, 1662: {1: 265, 2: 356}, 1663: {1: 266, 2: 356}, 1664: {1: 267, 2: 356}, 1665: {1: 268, 2: 356}, 1666: {1: 269, 2: 356}, 1667: {1: 270, 2: 356}, 1668: {1: 271, 2: 356}, 1669: {1: 272, 2: 356}, 1670: {1: 273, 2: 356}, 1671: {1: 274, 2: 356}, 1672: {1: 275, 2: 356}, 1673: {1: 276, 2: 356}, 1674: {1: 277, 2: 356}, 1675: {1: 278, 2: 356}, 1676: {1: 279, 2: 356}, 1677: {1: 280, 2: 356}, 1678: {1: 281, 2: 356}, 1679: {1: 282, 2: 356}, 1680: {1: 283, 2: 356}, 1681: {1: 284, 2: 356}, 1682: {1: 285, 2: 356}, 1683: {1: 286, 2: 356}, 1684: {1: 287, 2: 356}, 1685: {1: 288, 2: 356}, 1686: {1: 289, 2: 356}, 1687: {1: 290, 2: 356}, 1688: {1: 291, 2: 356}, 1689: {1: 292, 2: 356}, 1690: {1: 293, 2: 356}, 1691: {1: 294, 2: 356}, 1692: {1: 295, 2: 356}, 1693: {1: 296, 2: 356}, 1694: {1: 297, 2: 356}, 1695: {1: 298, 2: 356}, 1696: {1: 299, 2: 356}, 1697: {1: 300, 2: 356}, 1698: {1: 301, 2: 356}, 1699: {1: 302, 2: 356}, 1700: {1: 303, 2: 356}, 1701: {1: 304, 2: 356}, 1702: {1: 305, 2: 356}, 1703: {1: 306, 2: 356}, 1704: {1: 307, 2: 356}, 1705: {1: 308, 2: 356}, 1706: {1: 309, 2: 356}, 1707: {1: 310, 2: 356}, 1708: {1: 311, 2: 356}, 1709: {1: 312, 2: 356}, 1710: {1: 313, 2: 356}, 1711: {1: 314, 2: 356}, 1712: {1: 315, 2: 356}, 1713: {1: 316, 2: 356}, 1714: {1: 317, 2: 356}, 1715: {1: 318, 2: 356}, 1716: {1: 319, 2: 356}, 1717: {1: 320, 2: 356}, 1718: {1: 321, 2: 356}, 1719: {1: 322, 2: 356}, 1720: {1: 323, 2: 356}, 1721: {1: 324, 2: 356}, 1722: {1: 325, 2: 356}, 1723: {1: 326, 2: 356}, 1724: {1: 327, 2: 356}, 1725: {1: 328, 2: 356}, 1726: {1: 329, 2: 356}, 1727: {1: 330, 2: 356}, 1728: {1: 331, 2: 356}, 1729: {1: 332, 2: 356}, 1730: {1: 333, 2: 356}, 1731: {1: 334, 2: 356}, 1732: {1: 335, 2: 356}, 1733: {1: 336, 2: 356}, 1734: {1: 337, 2: 356}, 1735: {1: 338, 2: 356}, 1736: {1: 339, 2: 356}, 1737: {1: 340, 2: 356}, 1738: {1: 341, 2: 356}, 1739: {1: 342, 2: 356}, 1740: {1: 343, 2: 356}, 1741: {1: 344, 2: 356}, 1742: {1: 345, 2: 356}, 1743: {1: 346, 2: 356}, 1744: {1: 347, 2: 356}, 1745: {1: 348, 2: 356}, 1746: {1: 0, 2: 357}, 1747: {1: 1, 2: 357}, 1748: {1: 2, 2: 357}, 1749: {1: 3, 2: 357}, 1750: {1: 4, 2: 357}, 1751: {1: 5, 2: 357}, 1752: {1: 6, 2: 357}, 1753: {1: 7, 2: 357}, 1754: {1: 8, 2: 357}, 1755: {1: 9, 2: 357}, 1756: {1: 10, 2: 357}, 1757: {1: 11, 2: 357}, 1758: {1: 12, 2: 357}, 1759: {1: 13, 2: 357}, 1760: {1: 14, 2: 357}, 1761: {1: 15, 2: 357}, 1762: {1: 16, 2: 357}, 1763: {1: 17, 2: 357}, 1764: {1: 18, 2: 357}, 1765: {1: 19, 2: 357}, 1766: {1: 20, 2: 357}, 1767: {1: 21, 2: 357}, 1768: {1: 22, 2: 357}, 1769: {1: 23, 2: 357}, 1770: {1: 24, 2: 357}, 1771: {1: 25, 2: 357}, 1772: {1: 26, 2: 357}, 1773: {1: 27, 2: 357}, 1774: {1: 28, 2: 357}, 1775: {1: 29, 2: 357}, 1776: {1: 30, 2: 357}, 1777: {1: 31, 2: 357}, 1778: {1: 32, 2: 357}, 1779: {1: 33, 2: 357}, 1780: {1: 34, 2: 357}, 1781: {1: 35, 2: 357}, 1782: {1: 36, 2: 357}, 1783: {1: 37, 2: 357}, 1784: {1: 38, 2: 357}, 1785: {1: 39, 2: 357}, 1786: {1: 40, 2: 357}, 1787: {1: 41, 2: 357}, 1788: {1: 42, 2: 357}, 1789: {1: 43, 2: 357}, 1790: {1: 44, 2: 357}, 1791: {1: 45, 2: 357}, 1792: {1: 46, 2: 357}, 1793: {1: 47, 2: 357}, 1794: {1: 48, 2: 357}, 1795: {1: 49, 2: 357}, 1796: {1: 50, 2: 357}, 1797: {1: 51, 2: 357}, 1798: {1: 52, 2: 357}, 1799: {1: 53, 2: 357}, 1800: {1: 54, 2: 357}, 1801: {1: 55, 2: 357}, 1802: {1: 56, 2: 357}, 1803: {1: 57, 2: 357}, 1804: {1: 58, 2: 357}, 1805: {1: 59, 2: 357}, 1806: {1: 60, 2: 357}, 1807: {1: 61, 2: 357}, 1808: {1: 62, 2: 357}, 1809: {1: 63, 2: 357}, 1810: {1: 64, 2: 357}, 1811: {1: 65, 2: 357}, 1812: {1: 66, 2: 357}, 1813: {1: 67, 2: 357}, 1814: {1: 68, 2: 357}, 1815: {1: 69, 2: 357}, 1816: {1: 70, 2: 357}, 1817: {1: 71, 2: 357}, 1818: {1: 72, 2: 357}, 1819: {1: 73, 2: 357}, 1820: {1: 74, 2: 357}, 1821: {1: 75, 2: 357}, 1822: {1: 76, 2: 357}, 1823: {1: 77, 2: 357}, 1824: {1: 78, 2: 357}, 1825: {1: 79, 2: 357}, 1826: {1: 80, 2: 357}, 1827: {1: 81, 2: 357}, 1828: {1: 82, 2: 357}, 1829: {1: 83, 2: 357}, 1830: {1: 84, 2: 357}, 1831: {1: 85, 2: 357}, 1832: {1: 86, 2: 357}, 1833: {1: 87, 2: 357}, 1834: {1: 88, 2: 357}, 1835: {1: 89, 2: 357}, 1836: {1: 90, 2: 357}, 1837: {1: 91, 2: 357}, 1838: {1: 92, 2: 357}, 1839: {1: 93, 2: 357}, 1840: {1: 94, 2: 357}, 1841: {1: 95, 2: 357}, 1842: {1: 96, 2: 357}, 1843: {1: 97, 2: 357}, 1844: {1: 98, 2: 357}, 1845: {1: 99, 2: 357}, 1846: {1: 100, 2: 357}, 1847: {1: 101, 2: 357}, 1848: {1: 102, 2: 357}, 1849: {1: 103, 2: 357}, 1850: {1: 104, 2: 357}, 1851: {1: 105, 2: 357}, 1852: {1: 106, 2: 357}, 1853: {1: 107, 2: 357}, 1854: {1: 108, 2: 357}, 1855: {1: 109, 2: 357}, 1856: {1: 110, 2: 357}, 1857: {1: 111, 2: 357}, 1858: {1: 112, 2: 357}, 1859: {1: 113, 2: 357}, 1860: {1: 114, 2: 357}, 1861: {1: 115, 2: 357}, 1862: {1: 116, 2: 357}, 1863: {1: 117, 2: 357}, 1864: {1: 118, 2: 357}, 1865: {1: 119, 2: 357}, 1866: {1: 120, 2: 357}, 1867: {1: 121, 2: 357}, 1868: {1: 122, 2: 357}, 1869: {1: 123, 2: 357}, 1870: {1: 124, 2: 357}, 1871: {1: 125, 2: 357}, 1872: {1: 126, 2: 357}, 1873: {1: 127, 2: 357}, 1874: {1: 128, 2: 357}, 1875: {1: 129, 2: 357}, 1876: {1: 130, 2: 357}, 1877: {1: 131, 2: 357}, 1878: {1: 132, 2: 357}, 1879: {1: 133, 2: 357}, 1880: {1: 134, 2: 357}, 1881: {1: 135, 2: 357}, 1882: {1: 136, 2: 357}, 1883: {1: 137, 2: 357}, 1884: {1: 138, 2: 357}, 1885: {1: 139, 2: 357}, 1886: {1: 140, 2: 357}, 1887: {1: 141, 2: 357}, 1888: {1: 142, 2: 357}, 1889: {1: 143, 2: 357}, 1890: {1: 144, 2: 357}, 1891: {1: 145, 2: 357}, 1892: {1: 146, 2: 357}, 1893: {1: 147, 2: 357}, 1894: {1: 148, 2: 357}, 1895: {1: 149, 2: 357}, 1896: {1: 150, 2: 357}, 1897: {1: 151, 2: 357}, 1898: {1: 152, 2: 357}, 1899: {1: 153, 2: 357}, 1900: {1: 154, 2: 357}, 1901: {1: 155, 2: 357}, 1902: {1: 156, 2: 357}, 1903: {1: 157, 2: 357}, 1904: {1: 158, 2: 357}, 1905: {1: 159, 2: 357}, 1906: {1: 160, 2: 357}, 1907: {1: 161, 2: 357}, 1908: {1: 162, 2: 357}, 1909: {1: 163, 2: 357}, 1910: {1: 164, 2: 357}, 1911: {1: 165, 2: 357}, 1912: {1: 166, 2: 357}, 1913: {1: 167, 2: 357}, 1914: {1: 168, 2: 357}, 1915: {1: 169, 2: 357}, 1916: {1: 170, 2: 357}, 1917: {1: 171, 2: 357}, 1918: {1: 172, 2: 357}, 1919: {1: 173, 2: 357}, 1920: {1: 174, 2: 357}, 1921: {1: 175, 2: 357}, 1922: {1: 176, 2: 357}, 1923: {1: 177, 2: 357}, 1924: {1: 178, 2: 357}, 1925: {1: 179, 2: 357}, 1926: {1: 180, 2: 357}, 1927: {1: 181, 2: 357}, 1928: {1: 182, 2: 357}, 1929: {1: 183, 2: 357}, 1930: {1: 184, 2: 357}, 1931: {1: 185, 2: 357}, 1932: {1: 186, 2: 357}, 1933: {1: 187, 2: 357}, 1934: {1: 188, 2: 357}, 1935: {1: 189, 2: 357}, 1936: {1: 190, 2: 357}, 1937: {1: 191, 2: 357}, 1938: {1: 192, 2: 357}, 1939: {1: 193, 2: 357}, 1940: {1: 194, 2: 357}, 1941: {1: 195, 2: 357}, 1942: {1: 196, 2: 357}, 1943: {1: 197, 2: 357}, 1944: {1: 198, 2: 357}, 1945: {1: 199, 2: 357}, 1946: {1: 200, 2: 357}, 1947: {1: 201, 2: 357}, 1948: {1: 202, 2: 357}, 1949: {1: 203, 2: 357}, 1950: {1: 204, 2: 357}, 1951: {1: 205, 2: 357}, 1952: {1: 206, 2: 357}, 1953: {1: 207, 2: 357}, 1954: {1: 208, 2: 357}, 1955: {1: 209, 2: 357}, 1956: {1: 210, 2: 357}, 1957: {1: 211, 2: 357}, 1958: {1: 212, 2: 357}, 1959: {1: 213, 2: 357}, 1960: {1: 214, 2: 357}, 1961: {1: 215, 2: 357}, 1962: {1: 216, 2: 357}, 1963: {1: 217, 2: 357}, 1964: {1: 218, 2: 357}, 1965: {1: 219, 2: 357}, 1966: {1: 220, 2: 357}, 1967: {1: 221, 2: 357}, 1968: {1: 222, 2: 357}, 1969: {1: 223, 2: 357}, 1970: {1: 224, 2: 357}, 1971: {1: 225, 2: 357}, 1972: {1: 226, 2: 357}, 1973: {1: 227, 2: 357}, 1974: {1: 228, 2: 357}, 1975: {1: 229, 2: 357}, 1976: {1: 230, 2: 357}, 1977: {1: 231, 2: 357}, 1978: {1: 232, 2: 357}, 1979: {1: 233, 2: 357}, 1980: {1: 234, 2: 357}, 1981: {1: 235, 2: 357}, 1982: {1: 236, 2: 357}, 1983: {1: 237, 2: 357}, 1984: {1: 238, 2: 357}, 1985: {1: 239, 2: 357}, 1986: {1: 240, 2: 357}, 1987: {1: 241, 2: 357}, 1988: {1: 242, 2: 357}, 1989: {1: 243, 2: 357}, 1990: {1: 244, 2: 357}, 1991: {1: 245, 2: 357}, 1992: {1: 246, 2: 357}, 1993: {1: 247, 2: 357}, 1994: {1: 248, 2: 357}, 1995: {1: 249, 2: 357}, 1996: {1: 250, 2: 357}, 1997: {1: 251, 2: 357}, 1998: {1: 252, 2: 357}, 1999: {1: 253, 2: 357}, 2000: {1: 254, 2: 357}, 2001: {1: 255, 2: 357}, 2002: {1: 256, 2: 357}, 2003: {1: 257, 2: 357}, 2004: {1: 258, 2: 357}, 2005: {1: 259, 2: 357}, 2006: {1: 260, 2: 357}, 2007: {1: 261, 2: 357}, 2008: {1: 262, 2: 357}, 2009: {1: 263, 2: 357}, 2010: {1: 264, 2: 357}, 2011: {1: 265, 2: 357}, 2012: {1: 266, 2: 357}, 2013: {1: 267, 2: 357}, 2014: {1: 268, 2: 357}, 2015: {1: 269, 2: 357}, 2016: {1: 270, 2: 357}, 2017: {1: 271, 2: 357}, 2018: {1: 272, 2: 357}, 2019: {1: 273, 2: 357}, 2020: {1: 274, 2: 357}, 2021: {1: 275, 2: 357}, 2022: {1: 276, 2: 357}, 2023: {1: 277, 2: 357}, 2024: {1: 278, 2: 357}, 2025: {1: 279, 2: 357}, 2026: {1: 280, 2: 357}, 2027: {1: 281, 2: 357}, 2028: {1: 282, 2: 357}, 2029: {1: 283, 2: 357}, 2030: {1: 284, 2: 357}, 2031: {1: 285, 2: 357}, 2032: {1: 286, 2: 357}, 2033: {1: 287, 2: 357}, 2034: {1: 288, 2: 357}, 2035: {1: 289, 2: 357}, 2036: {1: 290, 2: 357}, 2037: {1: 291, 2: 357}, 2038: {1: 292, 2: 357}, 2039: {1: 293, 2: 357}, 2040: {1: 294, 2: 357}, 2041: {1: 295, 2: 357}, 2042: {1: 296, 2: 357}, 2043: {1: 297, 2: 357}, 2044: {1: 298, 2: 357}, 2045: {1: 299, 2: 357}, 2046: {1: 300, 2: 357}, 2047: {1: 301, 2: 357}, 2048: {1: 302, 2: 357}, 2049: {1: 303, 2: 357}, 2050: {1: 304, 2: 357}, 2051: {1: 305, 2: 357}, 2052: {1: 306, 2: 357}, 2053: {1: 307, 2: 357}, 2054: {1: 308, 2: 357}, 2055: {1: 309, 2: 357}, 2056: {1: 310, 2: 357}, 2057: {1: 311, 2: 357}, 2058: {1: 312, 2: 357}, 2059: {1: 313, 2: 357}, 2060: {1: 314, 2: 357}, 2061: {1: 315, 2: 357}, 2062: {1: 316, 2: 357}, 2063: {1: 317, 2: 357}, 2064: {1: 318, 2: 357}, 2065: {1: 319, 2: 357}, 2066: {1: 320, 2: 357}, 2067: {1: 321, 2: 357}, 2068: {1: 322, 2: 357}, 2069: {1: 323, 2: 357}, 2070: {1: 324, 2: 357}, 2071: {1: 325, 2: 357}, 2072: {1: 326, 2: 357}, 2073: {1: 327, 2: 357}, 2074: {1: 328, 2: 357}, 2075: {1: 329, 2: 357}, 2076: {1: 330, 2: 357}, 2077: {1: 331, 2: 357}, 2078: {1: 332, 2: 357}, 2079: {1: 333, 2: 357}, 2080: {1: 334, 2: 357}, 2081: {1: 335, 2: 357}, 2082: {1: 336, 2: 357}, 2083: {1: 337, 2: 357}, 2084: {1: 338, 2: 357}, 2085: {1: 339, 2: 357}, 2086: {1: 340, 2: 357}, 2087: {1: 341, 2: 357}, 2088: {1: 342, 2: 357}, 2089: {1: 343, 2: 357}, 2090: {1: 344, 2: 357}, 2091: {1: 345, 2: 357}, 2092: {1: 346, 2: 357}, 2093: {1: 347, 2: 357}, 2094: {1: 348, 2: 357}, 2095: {1: 0, 2: 358}, 2096: {1: 1, 2: 358}, 2097: {1: 2, 2: 358}, 2098: {1: 3, 2: 358}, 2099: {1: 4, 2: 358}, 2100: {1: 5, 2: 358}, 2101: {1: 6, 2: 358}, 2102: {1: 7, 2: 358}, 2103: {1: 8, 2: 358}, 2104: {1: 9, 2: 358}, 2105: {1: 10, 2: 358}, 2106: {1: 11, 2: 358}, 2107: {1: 12, 2: 358}, 2108: {1: 13, 2: 358}, 2109: {1: 14, 2: 358}, 2110: {1: 15, 2: 358}, 2111: {1: 16, 2: 358}, 2112: {1: 17, 2: 358}, 2113: {1: 18, 2: 358}, 2114: {1: 19, 2: 358}, 2115: {1: 20, 2: 358}, 2116: {1: 21, 2: 358}, 2117: {1: 22, 2: 358}, 2118: {1: 23, 2: 358}, 2119: {1: 24, 2: 358}, 2120: {1: 25, 2: 358}, 2121: {1: 26, 2: 358}, 2122: {1: 27, 2: 358}, 2123: {1: 28, 2: 358}, 2124: {1: 29, 2: 358}, 2125: {1: 30, 2: 358}, 2126: {1: 31, 2: 358}, 2127: {1: 32, 2: 358}, 2128: {1: 33, 2: 358}, 2129: {1: 34, 2: 358}, 2130: {1: 35, 2: 358}, 2131: {1: 36, 2: 358}, 2132: {1: 37, 2: 358}, 2133: {1: 38, 2: 358}, 2134: {1: 39, 2: 358}, 2135: {1: 40, 2: 358}, 2136: {1: 41, 2: 358}, 2137: {1: 42, 2: 358}, 2138: {1: 43, 2: 358}, 2139: {1: 44, 2: 358}, 2140: {1: 45, 2: 358}, 2141: {1: 46, 2: 358}, 2142: {1: 47, 2: 358}, 2143: {1: 48, 2: 358}, 2144: {1: 49, 2: 358}, 2145: {1: 50, 2: 358}, 2146: {1: 51, 2: 358}, 2147: {1: 52, 2: 358}, 2148: {1: 53, 2: 358}, 2149: {1: 54, 2: 358}, 2150: {1: 55, 2: 358}, 2151: {1: 56, 2: 358}, 2152: {1: 57, 2: 358}, 2153: {1: 58, 2: 358}, 2154: {1: 59, 2: 358}, 2155: {1: 60, 2: 358}, 2156: {1: 61, 2: 358}, 2157: {1: 62, 2: 358}, 2158: {1: 63, 2: 358}, 2159: {1: 64, 2: 358}, 2160: {1: 65, 2: 358}, 2161: {1: 66, 2: 358}, 2162: {1: 67, 2: 358}, 2163: {1: 68, 2: 358}, 2164: {1: 69, 2: 358}, 2165: {1: 70, 2: 358}, 2166: {1: 71, 2: 358}, 2167: {1: 72, 2: 358}, 2168: {1: 73, 2: 358}, 2169: {1: 74, 2: 358}, 2170: {1: 75, 2: 358}, 2171: {1: 76, 2: 358}, 2172: {1: 77, 2: 358}, 2173: {1: 78, 2: 358}, 2174: {1: 79, 2: 358}, 2175: {1: 80, 2: 358}, 2176: {1: 81, 2: 358}, 2177: {1: 82, 2: 358}, 2178: {1: 83, 2: 358}, 2179: {1: 84, 2: 358}, 2180: {1: 85, 2: 358}, 2181: {1: 86, 2: 358}, 2182: {1: 87, 2: 358}, 2183: {1: 88, 2: 358}, 2184: {1: 89, 2: 358}, 2185: {1: 90, 2: 358}, 2186: {1: 91, 2: 358}, 2187: {1: 92, 2: 358}, 2188: {1: 93, 2: 358}, 2189: {1: 94, 2: 358}, 2190: {1: 95, 2: 358}, 2191: {1: 96, 2: 358}, 2192: {1: 97, 2: 358}, 2193: {1: 98, 2: 358}, 2194: {1: 99, 2: 358}, 2195: {1: 100, 2: 358}, 2196: {1: 101, 2: 358}, 2197: {1: 102, 2: 358}, 2198: {1: 103, 2: 358}, 2199: {1: 104, 2: 358}, 2200: {1: 105, 2: 358}, 2201: {1: 106, 2: 358}, 2202: {1: 107, 2: 358}, 2203: {1: 108, 2: 358}, 2204: {1: 109, 2: 358}, 2205: {1: 110, 2: 358}, 2206: {1: 111, 2: 358}, 2207: {1: 112, 2: 358}, 2208: {1: 113, 2: 358}, 2209: {1: 114, 2: 358}, 2210: {1: 115, 2: 358}, 2211: {1: 116, 2: 358}, 2212: {1: 117, 2: 358}, 2213: {1: 118, 2: 358}, 2214: {1: 119, 2: 358}, 2215: {1: 120, 2: 358}, 2216: {1: 121, 2: 358}, 2217: {1: 122, 2: 358}, 2218: {1: 123, 2: 358}, 2219: {1: 124, 2: 358}, 2220: {1: 125, 2: 358}, 2221: {1: 126, 2: 358}, 2222: {1: 127, 2: 358}, 2223: {1: 128, 2: 358}, 2224: {1: 129, 2: 358}, 2225: {1: 130, 2: 358}, 2226: {1: 131, 2: 358}, 2227: {1: 132, 2: 358}, 2228: {1: 133, 2: 358}, 2229: {1: 134, 2: 358}, 2230: {1: 135, 2: 358}, 2231: {1: 136, 2: 358}, 2232: {1: 137, 2: 358}, 2233: {1: 138, 2: 358}, 2234: {1: 139, 2: 358}, 2235: {1: 140, 2: 358}, 2236: {1: 141, 2: 358}, 2237: {1: 142, 2: 358}, 2238: {1: 143, 2: 358}, 2239: {1: 144, 2: 358}, 2240: {1: 145, 2: 358}, 2241: {1: 146, 2: 358}, 2242: {1: 147, 2: 358}, 2243: {1: 148, 2: 358}, 2244: {1: 149, 2: 358}, 2245: {1: 150, 2: 358}, 2246: {1: 151, 2: 358}, 2247: {1: 152, 2: 358}, 2248: {1: 153, 2: 358}, 2249: {1: 154, 2: 358}, 2250: {1: 155, 2: 358}, 2251: {1: 156, 2: 358}, 2252: {1: 157, 2: 358}, 2253: {1: 158, 2: 358}, 2254: {1: 159, 2: 358}, 2255: {1: 160, 2: 358}, 2256: {1: 161, 2: 358}, 2257: {1: 162, 2: 358}, 2258: {1: 163, 2: 358}, 2259: {1: 164, 2: 358}, 2260: {1: 165, 2: 358}, 2261: {1: 166, 2: 358}, 2262: {1: 167, 2: 358}, 2263: {1: 168, 2: 358}, 2264: {1: 169, 2: 358}, 2265: {1: 170, 2: 358}, 2266: {1: 171, 2: 358}, 2267: {1: 172, 2: 358}, 2268: {1: 173, 2: 358}, 2269: {1: 174, 2: 358}, 2270: {1: 175, 2: 358}, 2271: {1: 176, 2: 358}, 2272: {1: 177, 2: 358}, 2273: {1: 178, 2: 358}, 2274: {1: 179, 2: 358}, 2275: {1: 180, 2: 358}, 2276: {1: 181, 2: 358}, 2277: {1: 182, 2: 358}, 2278: {1: 183, 2: 358}, 2279: {1: 184, 2: 358}, 2280: {1: 185, 2: 358}, 2281: {1: 186, 2: 358}, 2282: {1: 187, 2: 358}, 2283: {1: 188, 2: 358}, 2284: {1: 189, 2: 358}, 2285: {1: 190, 2: 358}, 2286: {1: 191, 2: 358}, 2287: {1: 192, 2: 358}, 2288: {1: 193, 2: 358}, 2289: {1: 194, 2: 358}, 2290: {1: 195, 2: 358}, 2291: {1: 196, 2: 358}, 2292: {1: 197, 2: 358}, 2293: {1: 198, 2: 358}, 2294: {1: 199, 2: 358}, 2295: {1: 200, 2: 358}, 2296: {1: 201, 2: 358}, 2297: {1: 202, 2: 358}, 2298: {1: 203, 2: 358}, 2299: {1: 204, 2: 358}, 2300: {1: 205, 2: 358}, 2301: {1: 206, 2: 358}, 2302: {1: 207, 2: 358}, 2303: {1: 208, 2: 358}, 2304: {1: 209, 2: 358}, 2305: {1: 210, 2: 358}, 2306: {1: 211, 2: 358}, 2307: {1: 212, 2: 358}, 2308: {1: 213, 2: 358}, 2309: {1: 214, 2: 358}, 2310: {1: 215, 2: 358}, 2311: {1: 216, 2: 358}, 2312: {1: 217, 2: 358}, 2313: {1: 218, 2: 358}, 2314: {1: 219, 2: 358}, 2315: {1: 220, 2: 358}, 2316: {1: 221, 2: 358}, 2317: {1: 222, 2: 358}, 2318: {1: 223, 2: 358}, 2319: {1: 224, 2: 358}, 2320: {1: 225, 2: 358}, 2321: {1: 226, 2: 358}, 2322: {1: 227, 2: 358}, 2323: {1: 228, 2: 358}, 2324: {1: 229, 2: 358}, 2325: {1: 230, 2: 358}, 2326: {1: 231, 2: 358}, 2327: {1: 232, 2: 358}, 2328: {1: 233, 2: 358}, 2329: {1: 234, 2: 358}, 2330: {1: 235, 2: 358}, 2331: {1: 236, 2: 358}, 2332: {1: 237, 2: 358}, 2333: {1: 238, 2: 358}, 2334: {1: 239, 2: 358}, 2335: {1: 240, 2: 358}, 2336: {1: 241, 2: 358}, 2337: {1: 242, 2: 358}, 2338: {1: 243, 2: 358}, 2339: {1: 244, 2: 358}, 2340: {1: 245, 2: 358}, 2341: {1: 246, 2: 358}, 2342: {1: 247, 2: 358}, 2343: {1: 248, 2: 358}, 2344: {1: 249, 2: 358}, 2345: {1: 250, 2: 358}, 2346: {1: 251, 2: 358}, 2347: {1: 252, 2: 358}, 2348: {1: 253, 2: 358}, 2349: {1: 254, 2: 358}, 2350: {1: 255, 2: 358}, 2351: {1: 256, 2: 358}, 2352: {1: 257, 2: 358}, 2353: {1: 258, 2: 358}, 2354: {1: 259, 2: 358}, 2355: {1: 260, 2: 358}, 2356: {1: 261, 2: 358}, 2357: {1: 262, 2: 358}, 2358: {1: 263, 2: 358}, 2359: {1: 264, 2: 358}, 2360: {1: 265, 2: 358}, 2361: {1: 266, 2: 358}, 2362: {1: 267, 2: 358}, 2363: {1: 268, 2: 358}, 2364: {1: 269, 2: 358}, 2365: {1: 270, 2: 358}, 2366: {1: 271, 2: 358}, 2367: {1: 272, 2: 358}, 2368: {1: 273, 2: 358}, 2369: {1: 274, 2: 358}, 2370: {1: 275, 2: 358}, 2371: {1: 276, 2: 358}, 2372: {1: 277, 2: 358}, 2373: {1: 278, 2: 358}, 2374: {1: 279, 2: 358}, 2375: {1: 280, 2: 358}, 2376: {1: 281, 2: 358}, 2377: {1: 282, 2: 358}, 2378: {1: 283, 2: 358}, 2379: {1: 284, 2: 358}, 2380: {1: 285, 2: 358}, 2381: {1: 286, 2: 358}, 2382: {1: 287, 2: 358}, 2383: {1: 288, 2: 358}, 2384: {1: 289, 2: 358}, 2385: {1: 290, 2: 358}, 2386: {1: 291, 2: 358}, 2387: {1: 292, 2: 358}, 2388: {1: 293, 2: 358}, 2389: {1: 294, 2: 358}, 2390: {1: 295, 2: 358}, 2391: {1: 296, 2: 358}, 2392: {1: 297, 2: 358}, 2393: {1: 298, 2: 358}, 2394: {1: 299, 2: 358}, 2395: {1: 300, 2: 358}, 2396: {1: 301, 2: 358}, 2397: {1: 302, 2: 358}, 2398: {1: 303, 2: 358}, 2399: {1: 304, 2: 358}, 2400: {1: 305, 2: 358}, 2401: {1: 306, 2: 358}, 2402: {1: 307, 2: 358}, 2403: {1: 308, 2: 358}, 2404: {1: 309, 2: 358}, 2405: {1: 310, 2: 358}, 2406: {1: 311, 2: 358}, 2407: {1: 312, 2: 358}, 2408: {1: 313, 2: 358}, 2409: {1: 314, 2: 358}, 2410: {1: 315, 2: 358}, 2411: {1: 316, 2: 358}, 2412: {1: 317, 2: 358}, 2413: {1: 318, 2: 358}, 2414: {1: 319, 2: 358}, 2415: {1: 320, 2: 358}, 2416: {1: 321, 2: 358}, 2417: {1: 322, 2: 358}, 2418: {1: 323, 2: 358}, 2419: {1: 324, 2: 358}, 2420: {1: 325, 2: 358}, 2421: {1: 326, 2: 358}, 2422: {1: 327, 2: 358}, 2423: {1: 328, 2: 358}, 2424: {1: 329, 2: 358}, 2425: {1: 330, 2: 358}, 2426: {1: 331, 2: 358}, 2427: {1: 332, 2: 358}, 2428: {1: 333, 2: 358}, 2429: {1: 334, 2: 358}, 2430: {1: 335, 2: 358}, 2431: {1: 336, 2: 358}, 2432: {1: 337, 2: 358}, 2433: {1: 338, 2: 358}, 2434: {1: 339, 2: 358}, 2435: {1: 340, 2: 358}, 2436: {1: 341, 2: 358}, 2437: {1: 342, 2: 358}, 2438: {1: 343, 2: 358}, 2439: {1: 344, 2: 358}, 2440: {1: 345, 2: 358}, 2441: {1: 346, 2: 358}, 2442: {1: 347, 2: 358}, 2443: {1: 348, 2: 358}, 2444: {1: 0, 2: 349}, 2445: {1: 1, 2: 349}, 2446: {1: 2, 2: 349}, 2447: {1: 3, 2: 349}, 2448: {1: 4, 2: 349}, 2449: {1: 5, 2: 349}, 2450: {1: 6, 2: 349}, 2451: {1: 7, 2: 349}, 2452: {1: 8, 2: 349}, 2453: {1: 9, 2: 349}, 2454: {1: 10, 2: 349}, 2455: {1: 11, 2: 349}, 2456: {1: 12, 2: 349}, 2457: {1: 13, 2: 349}, 2458: {1: 14, 2: 349}, 2459: {1: 15, 2: 349}, 2460: {1: 16, 2: 349}, 2461: {1: 17, 2: 349}, 2462: {1: 18, 2: 349}, 2463: {1: 19, 2: 349}, 2464: {1: 20, 2: 349}, 2465: {1: 21, 2: 349}, 2466: {1: 22, 2: 349}, 2467: {1: 23, 2: 349}, 2468: {1: 24, 2: 349}, 2469: {1: 25, 2: 349}, 2470: {1: 26, 2: 349}, 2471: {1: 27, 2: 349}, 2472: {1: 28, 2: 349}, 2473: {1: 29, 2: 349}, 2474: {1: 30, 2: 349}, 2475: {1: 31, 2: 349}, 2476: {1: 32, 2: 349}, 2477: {1: 33, 2: 349}, 2478: {1: 34, 2: 349}, 2479: {1: 35, 2: 349}, 2480: {1: 36, 2: 349}, 2481: {1: 37, 2: 349}, 2482: {1: 38, 2: 349}, 2483: {1: 39, 2: 349}, 2484: {1: 40, 2: 349}, 2485: {1: 41, 2: 349}, 2486: {1: 42, 2: 349}, 2487: {1: 43, 2: 349}, 2488: {1: 44, 2: 349}, 2489: {1: 45, 2: 349}, 2490: {1: 46, 2: 349}, 2491: {1: 47, 2: 349}, 2492: {1: 48, 2: 349}, 2493: {1: 49, 2: 349}, 2494: {1: 50, 2: 349}, 2495: {1: 51, 2: 349}, 2496: {1: 52, 2: 349}, 2497: {1: 53, 2: 349}, 2498: {1: 54, 2: 349}, 2499: {1: 55, 2: 349}, 2500: {1: 56, 2: 349}, 2501: {1: 57, 2: 349}, 2502: {1: 58, 2: 349}, 2503: {1: 59, 2: 349}, 2504: {1: 60, 2: 349}, 2505: {1: 61, 2: 349}, 2506: {1: 62, 2: 349}, 2507: {1: 63, 2: 349}, 2508: {1: 64, 2: 349}, 2509: {1: 65, 2: 349}, 2510: {1: 66, 2: 349}, 2511: {1: 67, 2: 349}, 2512: {1: 68, 2: 349}, 2513: {1: 69, 2: 349}, 2514: {1: 70, 2: 349}, 2515: {1: 71, 2: 349}, 2516: {1: 72, 2: 349}, 2517: {1: 73, 2: 349}, 2518: {1: 74, 2: 349}, 2519: {1: 75, 2: 349}, 2520: {1: 76, 2: 349}, 2521: {1: 77, 2: 349}, 2522: {1: 78, 2: 349}, 2523: {1: 79, 2: 349}, 2524: {1: 80, 2: 349}, 2525: {1: 81, 2: 349}, 2526: {1: 82, 2: 349}, 2527: {1: 83, 2: 349}, 2528: {1: 84, 2: 349}, 2529: {1: 85, 2: 349}, 2530: {1: 86, 2: 349}, 2531: {1: 87, 2: 349}, 2532: {1: 88, 2: 349}, 2533: {1: 89, 2: 349}, 2534: {1: 90, 2: 349}, 2535: {1: 91, 2: 349}, 2536: {1: 92, 2: 349}, 2537: {1: 93, 2: 349}, 2538: {1: 94, 2: 349}, 2539: {1: 95, 2: 349}, 2540: {1: 96, 2: 349}, 2541: {1: 97, 2: 349}, 2542: {1: 98, 2: 349}, 2543: {1: 99, 2: 349}, 2544: {1: 100, 2: 349}, 2545: {1: 101, 2: 349}, 2546: {1: 102, 2: 349}, 2547: {1: 103, 2: 349}, 2548: {1: 104, 2: 349}, 2549: {1: 105, 2: 349}, 2550: {1: 106, 2: 349}, 2551: {1: 107, 2: 349}, 2552: {1: 108, 2: 349}, 2553: {1: 109, 2: 349}, 2554: {1: 110, 2: 349}, 2555: {1: 111, 2: 349}, 2556: {1: 112, 2: 349}, 2557: {1: 113, 2: 349}, 2558: {1: 114, 2: 349}, 2559: {1: 115, 2: 349}, 2560: {1: 116, 2: 349}, 2561: {1: 117, 2: 349}, 2562: {1: 118, 2: 349}, 2563: {1: 119, 2: 349}, 2564: {1: 120, 2: 349}, 2565: {1: 121, 2: 349}, 2566: {1: 122, 2: 349}, 2567: {1: 123, 2: 349}, 2568: {1: 124, 2: 349}, 2569: {1: 125, 2: 349}, 2570: {1: 126, 2: 349}, 2571: {1: 127, 2: 349}, 2572: {1: 128, 2: 349}, 2573: {1: 129, 2: 349}, 2574: {1: 130, 2: 349}, 2575: {1: 131, 2: 349}, 2576: {1: 132, 2: 349}, 2577: {1: 133, 2: 349}, 2578: {1: 134, 2: 349}, 2579: {1: 135, 2: 349}, 2580: {1: 136, 2: 349}, 2581: {1: 137, 2: 349}, 2582: {1: 138, 2: 349}, 2583: {1: 139, 2: 349}, 2584: {1: 140, 2: 349}, 2585: {1: 141, 2: 349}, 2586: {1: 142, 2: 349}, 2587: {1: 143, 2: 349}, 2588: {1: 144, 2: 349}, 2589: {1: 145, 2: 349}, 2590: {1: 146, 2: 349}, 2591: {1: 147, 2: 349}, 2592: {1: 148, 2: 349}, 2593: {1: 149, 2: 349}, 2594: {1: 150, 2: 349}, 2595: {1: 151, 2: 349}, 2596: {1: 152, 2: 349}, 2597: {1: 153, 2: 349}, 2598: {1: 154, 2: 349}, 2599: {1: 155, 2: 349}, 2600: {1: 156, 2: 349}, 2601: {1: 157, 2: 349}, 2602: {1: 158, 2: 349}, 2603: {1: 159, 2: 349}, 2604: {1: 160, 2: 349}, 2605: {1: 161, 2: 349}, 2606: {1: 162, 2: 349}, 2607: {1: 163, 2: 349}, 2608: {1: 164, 2: 349}, 2609: {1: 165, 2: 349}, 2610: {1: 166, 2: 349}, 2611: {1: 167, 2: 349}, 2612: {1: 168, 2: 349}, 2613: {1: 169, 2: 349}, 2614: {1: 170, 2: 349}, 2615: {1: 171, 2: 349}, 2616: {1: 172, 2: 349}, 2617: {1: 173, 2: 349}, 2618: {1: 174, 2: 349}, 2619: {1: 175, 2: 349}, 2620: {1: 176, 2: 349}, 2621: {1: 177, 2: 349}, 2622: {1: 178, 2: 349}, 2623: {1: 179, 2: 349}, 2624: {1: 180, 2: 349}, 2625: {1: 181, 2: 349}, 2626: {1: 182, 2: 349}, 2627: {1: 183, 2: 349}, 2628: {1: 184, 2: 349}, 2629: {1: 185, 2: 349}, 2630: {1: 186, 2: 349}, 2631: {1: 187, 2: 349}, 2632: {1: 188, 2: 349}, 2633: {1: 189, 2: 349}, 2634: {1: 190, 2: 349}, 2635: {1: 191, 2: 349}, 2636: {1: 192, 2: 349}, 2637: {1: 193, 2: 349}, 2638: {1: 194, 2: 349}, 2639: {1: 195, 2: 349}, 2640: {1: 196, 2: 349}, 2641: {1: 197, 2: 349}, 2642: {1: 198, 2: 349}, 2643: {1: 199, 2: 349}, 2644: {1: 200, 2: 349}, 2645: {1: 201, 2: 349}, 2646: {1: 202, 2: 349}, 2647: {1: 203, 2: 349}, 2648: {1: 204, 2: 349}, 2649: {1: 205, 2: 349}, 2650: {1: 206, 2: 349}, 2651: {1: 207, 2: 349}, 2652: {1: 208, 2: 349}, 2653: {1: 209, 2: 349}, 2654: {1: 210, 2: 349}, 2655: {1: 211, 2: 349}, 2656: {1: 212, 2: 349}, 2657: {1: 213, 2: 349}, 2658: {1: 214, 2: 349}, 2659: {1: 215, 2: 349}, 2660: {1: 216, 2: 349}, 2661: {1: 217, 2: 349}, 2662: {1: 218, 2: 349}, 2663: {1: 219, 2: 349}, 2664: {1: 220, 2: 349}, 2665: {1: 221, 2: 349}, 2666: {1: 222, 2: 349}, 2667: {1: 223, 2: 349}, 2668: {1: 224, 2: 349}, 2669: {1: 225, 2: 349}, 2670: {1: 226, 2: 349}, 2671: {1: 227, 2: 349}, 2672: {1: 228, 2: 349}, 2673: {1: 229, 2: 349}, 2674: {1: 230, 2: 349}, 2675: {1: 231, 2: 349}, 2676: {1: 232, 2: 349}, 2677: {1: 233, 2: 349}, 2678: {1: 234, 2: 349}, 2679: {1: 235, 2: 349}, 2680: {1: 236, 2: 349}, 2681: {1: 237, 2: 349}, 2682: {1: 238, 2: 349}, 2683: {1: 239, 2: 349}, 2684: {1: 240, 2: 349}, 2685: {1: 241, 2: 349}, 2686: {1: 242, 2: 349}, 2687: {1: 243, 2: 349}, 2688: {1: 244, 2: 349}, 2689: {1: 245, 2: 349}, 2690: {1: 246, 2: 349}, 2691: {1: 247, 2: 349}, 2692: {1: 248, 2: 349}, 2693: {1: 249, 2: 349}, 2694: {1: 250, 2: 349}, 2695: {1: 251, 2: 349}, 2696: {1: 252, 2: 349}, 2697: {1: 253, 2: 349}, 2698: {1: 254, 2: 349}, 2699: {1: 255, 2: 349}, 2700: {1: 256, 2: 349}, 2701: {1: 257, 2: 349}, 2702: {1: 258, 2: 349}, 2703: {1: 259, 2: 349}, 2704: {1: 260, 2: 349}, 2705: {1: 261, 2: 349}, 2706: {1: 262, 2: 349}, 2707: {1: 263, 2: 349}, 2708: {1: 264, 2: 349}, 2709: {1: 265, 2: 349}, 2710: {1: 266, 2: 349}, 2711: {1: 267, 2: 349}, 2712: {1: 268, 2: 349}, 2713: {1: 269, 2: 349}, 2714: {1: 270, 2: 349}, 2715: {1: 271, 2: 349}, 2716: {1: 272, 2: 349}, 2717: {1: 273, 2: 349}, 2718: {1: 274, 2: 349}, 2719: {1: 275, 2: 349}, 2720: {1: 276, 2: 349}, 2721: {1: 277, 2: 349}, 2722: {1: 278, 2: 349}, 2723: {1: 279, 2: 349}, 2724: {1: 280, 2: 349}, 2725: {1: 281, 2: 349}, 2726: {1: 282, 2: 349}, 2727: {1: 283, 2: 349}, 2728: {1: 284, 2: 349}, 2729: {1: 285, 2: 349}, 2730: {1: 286, 2: 349}, 2731: {1: 287, 2: 349}, 2732: {1: 288, 2: 349}, 2733: {1: 289, 2: 349}, 2734: {1: 290, 2: 349}, 2735: {1: 291, 2: 349}, 2736: {1: 292, 2: 349}, 2737: {1: 293, 2: 349}, 2738: {1: 294, 2: 349}, 2739: {1: 295, 2: 349}, 2740: {1: 296, 2: 349}, 2741: {1: 297, 2: 349}, 2742: {1: 298, 2: 349}, 2743: {1: 299, 2: 349}, 2744: {1: 300, 2: 349}, 2745: {1: 301, 2: 349}, 2746: {1: 302, 2: 349}, 2747: {1: 303, 2: 349}, 2748: {1: 304, 2: 349}, 2749: {1: 305, 2: 349}, 2750: {1: 306, 2: 349}, 2751: {1: 307, 2: 349}, 2752: {1: 308, 2: 349}, 2753: {1: 309, 2: 349}, 2754: {1: 310, 2: 349}, 2755: {1: 311, 2: 349}, 2756: {1: 312, 2: 349}, 2757: {1: 313, 2: 349}, 2758: {1: 314, 2: 349}, 2759: {1: 315, 2: 349}, 2760: {1: 316, 2: 349}, 2761: {1: 317, 2: 349}, 2762: {1: 318, 2: 349}, 2763: {1: 319, 2: 349}, 2764: {1: 320, 2: 349}, 2765: {1: 321, 2: 349}, 2766: {1: 322, 2: 349}, 2767: {1: 323, 2: 349}, 2768: {1: 324, 2: 349}, 2769: {1: 325, 2: 349}, 2770: {1: 326, 2: 349}, 2771: {1: 327, 2: 349}, 2772: {1: 328, 2: 349}, 2773: {1: 329, 2: 349}, 2774: {1: 330, 2: 349}, 2775: {1: 331, 2: 349}, 2776: {1: 332, 2: 349}, 2777: {1: 333, 2: 349}, 2778: {1: 334, 2: 349}, 2779: {1: 335, 2: 349}, 2780: {1: 336, 2: 349}, 2781: {1: 337, 2: 349}, 2782: {1: 338, 2: 349}, 2783: {1: 339, 2: 349}, 2784: {1: 340, 2: 349}, 2785: {1: 341, 2: 349}, 2786: {1: 342, 2: 349}, 2787: {1: 343, 2: 349}, 2788: {1: 344, 2: 349}, 2789: {1: 345, 2: 349}, 2790: {1: 346, 2: 349}, 2791: {1: 347, 2: 349}, 2792: {1: 348, 2: 349}, 2793: {1: 0, 2: 350}, 2794: {1: 1, 2: 350}, 2795: {1: 2, 2: 350}, 2796: {1: 3, 2: 350}, 2797: {1: 4, 2: 350}, 2798: {1: 5, 2: 350}, 2799: {1: 6, 2: 350}, 2800: {1: 7, 2: 350}, 2801: {1: 8, 2: 350}, 2802: {1: 9, 2: 350}, 2803: {1: 10, 2: 350}, 2804: {1: 11, 2: 350}, 2805: {1: 12, 2: 350}, 2806: {1: 13, 2: 350}, 2807: {1: 14, 2: 350}, 2808: {1: 15, 2: 350}, 2809: {1: 16, 2: 350}, 2810: {1: 17, 2: 350}, 2811: {1: 18, 2: 350}, 2812: {1: 19, 2: 350}, 2813: {1: 20, 2: 350}, 2814: {1: 21, 2: 350}, 2815: {1: 22, 2: 350}, 2816: {1: 23, 2: 350}, 2817: {1: 24, 2: 350}, 2818: {1: 25, 2: 350}, 2819: {1: 26, 2: 350}, 2820: {1: 27, 2: 350}, 2821: {1: 28, 2: 350}, 2822: {1: 29, 2: 350}, 2823: {1: 30, 2: 350}, 2824: {1: 31, 2: 350}, 2825: {1: 32, 2: 350}, 2826: {1: 33, 2: 350}, 2827: {1: 34, 2: 350}, 2828: {1: 35, 2: 350}, 2829: {1: 36, 2: 350}, 2830: {1: 37, 2: 350}, 2831: {1: 38, 2: 350}, 2832: {1: 39, 2: 350}, 2833: {1: 40, 2: 350}, 2834: {1: 41, 2: 350}, 2835: {1: 42, 2: 350}, 2836: {1: 43, 2: 350}, 2837: {1: 44, 2: 350}, 2838: {1: 45, 2: 350}, 2839: {1: 46, 2: 350}, 2840: {1: 47, 2: 350}, 2841: {1: 48, 2: 350}, 2842: {1: 49, 2: 350}, 2843: {1: 50, 2: 350}, 2844: {1: 51, 2: 350}, 2845: {1: 52, 2: 350}, 2846: {1: 53, 2: 350}, 2847: {1: 54, 2: 350}, 2848: {1: 55, 2: 350}, 2849: {1: 56, 2: 350}, 2850: {1: 57, 2: 350}, 2851: {1: 58, 2: 350}, 2852: {1: 59, 2: 350}, 2853: {1: 60, 2: 350}, 2854: {1: 61, 2: 350}, 2855: {1: 62, 2: 350}, 2856: {1: 63, 2: 350}, 2857: {1: 64, 2: 350}, 2858: {1: 65, 2: 350}, 2859: {1: 66, 2: 350}, 2860: {1: 67, 2: 350}, 2861: {1: 68, 2: 350}, 2862: {1: 69, 2: 350}, 2863: {1: 70, 2: 350}, 2864: {1: 71, 2: 350}, 2865: {1: 72, 2: 350}, 2866: {1: 73, 2: 350}, 2867: {1: 74, 2: 350}, 2868: {1: 75, 2: 350}, 2869: {1: 76, 2: 350}, 2870: {1: 77, 2: 350}, 2871: {1: 78, 2: 350}, 2872: {1: 79, 2: 350}, 2873: {1: 80, 2: 350}, 2874: {1: 81, 2: 350}, 2875: {1: 82, 2: 350}, 2876: {1: 83, 2: 350}, 2877: {1: 84, 2: 350}, 2878: {1: 85, 2: 350}, 2879: {1: 86, 2: 350}, 2880: {1: 87, 2: 350}, 2881: {1: 88, 2: 350}, 2882: {1: 89, 2: 350}, 2883: {1: 90, 2: 350}, 2884: {1: 91, 2: 350}, 2885: {1: 92, 2: 350}, 2886: {1: 93, 2: 350}, 2887: {1: 94, 2: 350}, 2888: {1: 95, 2: 350}, 2889: {1: 96, 2: 350}, 2890: {1: 97, 2: 350}, 2891: {1: 98, 2: 350}, 2892: {1: 99, 2: 350}, 2893: {1: 100, 2: 350}, 2894: {1: 101, 2: 350}, 2895: {1: 102, 2: 350}, 2896: {1: 103, 2: 350}, 2897: {1: 104, 2: 350}, 2898: {1: 105, 2: 350}, 2899: {1: 106, 2: 350}, 2900: {1: 107, 2: 350}, 2901: {1: 108, 2: 350}, 2902: {1: 109, 2: 350}, 2903: {1: 110, 2: 350}, 2904: {1: 111, 2: 350}, 2905: {1: 112, 2: 350}, 2906: {1: 113, 2: 350}, 2907: {1: 114, 2: 350}, 2908: {1: 115, 2: 350}, 2909: {1: 116, 2: 350}, 2910: {1: 117, 2: 350}, 2911: {1: 118, 2: 350}, 2912: {1: 119, 2: 350}, 2913: {1: 120, 2: 350}, 2914: {1: 121, 2: 350}, 2915: {1: 122, 2: 350}, 2916: {1: 123, 2: 350}, 2917: {1: 124, 2: 350}, 2918: {1: 125, 2: 350}, 2919: {1: 126, 2: 350}, 2920: {1: 127, 2: 350}, 2921: {1: 128, 2: 350}, 2922: {1: 129, 2: 350}, 2923: {1: 130, 2: 350}, 2924: {1: 131, 2: 350}, 2925: {1: 132, 2: 350}, 2926: {1: 133, 2: 350}, 2927: {1: 134, 2: 350}, 2928: {1: 135, 2: 350}, 2929: {1: 136, 2: 350}, 2930: {1: 137, 2: 350}, 2931: {1: 138, 2: 350}, 2932: {1: 139, 2: 350}, 2933: {1: 140, 2: 350}, 2934: {1: 141, 2: 350}, 2935: {1: 142, 2: 350}, 2936: {1: 143, 2: 350}, 2937: {1: 144, 2: 350}, 2938: {1: 145, 2: 350}, 2939: {1: 146, 2: 350}, 2940: {1: 147, 2: 350}, 2941: {1: 148, 2: 350}, 2942: {1: 149, 2: 350}, 2943: {1: 150, 2: 350}, 2944: {1: 151, 2: 350}, 2945: {1: 152, 2: 350}, 2946: {1: 153, 2: 350}, 2947: {1: 154, 2: 350}, 2948: {1: 155, 2: 350}, 2949: {1: 156, 2: 350}, 2950: {1: 157, 2: 350}, 2951: {1: 158, 2: 350}, 2952: {1: 159, 2: 350}, 2953: {1: 160, 2: 350}, 2954: {1: 161, 2: 350}, 2955: {1: 162, 2: 350}, 2956: {1: 163, 2: 350}, 2957: {1: 164, 2: 350}, 2958: {1: 165, 2: 350}, 2959: {1: 166, 2: 350}, 2960: {1: 167, 2: 350}, 2961: {1: 168, 2: 350}, 2962: {1: 169, 2: 350}, 2963: {1: 170, 2: 350}, 2964: {1: 171, 2: 350}, 2965: {1: 172, 2: 350}, 2966: {1: 173, 2: 350}, 2967: {1: 174, 2: 350}, 2968: {1: 175, 2: 350}, 2969: {1: 176, 2: 350}, 2970: {1: 177, 2: 350}, 2971: {1: 178, 2: 350}, 2972: {1: 179, 2: 350}, 2973: {1: 180, 2: 350}, 2974: {1: 181, 2: 350}, 2975: {1: 182, 2: 350}, 2976: {1: 183, 2: 350}, 2977: {1: 184, 2: 350}, 2978: {1: 185, 2: 350}, 2979: {1: 186, 2: 350}, 2980: {1: 187, 2: 350}, 2981: {1: 188, 2: 350}, 2982: {1: 189, 2: 350}, 2983: {1: 190, 2: 350}, 2984: {1: 191, 2: 350}, 2985: {1: 192, 2: 350}, 2986: {1: 193, 2: 350}, 2987: {1: 194, 2: 350}, 2988: {1: 195, 2: 350}, 2989: {1: 196, 2: 350}, 2990: {1: 197, 2: 350}, 2991: {1: 198, 2: 350}, 2992: {1: 199, 2: 350}, 2993: {1: 200, 2: 350}, 2994: {1: 201, 2: 350}, 2995: {1: 202, 2: 350}, 2996: {1: 203, 2: 350}, 2997: {1: 204, 2: 350}, 2998: {1: 205, 2: 350}, 2999: {1: 206, 2: 350}, 3000: {1: 207, 2: 350}, 3001: {1: 208, 2: 350}, 3002: {1: 209, 2: 350}, 3003: {1: 210, 2: 350}, 3004: {1: 211, 2: 350}, 3005: {1: 212, 2: 350}, 3006: {1: 213, 2: 350}, 3007: {1: 214, 2: 350}, 3008: {1: 215, 2: 350}, 3009: {1: 216, 2: 350}, 3010: {1: 217, 2: 350}, 3011: {1: 218, 2: 350}, 3012: {1: 219, 2: 350}, 3013: {1: 220, 2: 350}, 3014: {1: 221, 2: 350}, 3015: {1: 222, 2: 350}, 3016: {1: 223, 2: 350}, 3017: {1: 224, 2: 350}, 3018: {1: 225, 2: 350}, 3019: {1: 226, 2: 350}, 3020: {1: 227, 2: 350}, 3021: {1: 228, 2: 350}, 3022: {1: 229, 2: 350}, 3023: {1: 230, 2: 350}, 3024: {1: 231, 2: 350}, 3025: {1: 232, 2: 350}, 3026: {1: 233, 2: 350}, 3027: {1: 234, 2: 350}, 3028: {1: 235, 2: 350}, 3029: {1: 236, 2: 350}, 3030: {1: 237, 2: 350}, 3031: {1: 238, 2: 350}, 3032: {1: 239, 2: 350}, 3033: {1: 240, 2: 350}, 3034: {1: 241, 2: 350}, 3035: {1: 242, 2: 350}, 3036: {1: 243, 2: 350}, 3037: {1: 244, 2: 350}, 3038: {1: 245, 2: 350}, 3039: {1: 246, 2: 350}, 3040: {1: 247, 2: 350}, 3041: {1: 248, 2: 350}, 3042: {1: 249, 2: 350}, 3043: {1: 250, 2: 350}, 3044: {1: 251, 2: 350}, 3045: {1: 252, 2: 350}, 3046: {1: 253, 2: 350}, 3047: {1: 254, 2: 350}, 3048: {1: 255, 2: 350}, 3049: {1: 256, 2: 350}, 3050: {1: 257, 2: 350}, 3051: {1: 258, 2: 350}, 3052: {1: 259, 2: 350}, 3053: {1: 260, 2: 350}, 3054: {1: 261, 2: 350}, 3055: {1: 262, 2: 350}, 3056: {1: 263, 2: 350}, 3057: {1: 264, 2: 350}, 3058: {1: 265, 2: 350}, 3059: {1: 266, 2: 350}, 3060: {1: 267, 2: 350}, 3061: {1: 268, 2: 350}, 3062: {1: 269, 2: 350}, 3063: {1: 270, 2: 350}, 3064: {1: 271, 2: 350}, 3065: {1: 272, 2: 350}, 3066: {1: 273, 2: 350}, 3067: {1: 274, 2: 350}, 3068: {1: 275, 2: 350}, 3069: {1: 276, 2: 350}, 3070: {1: 277, 2: 350}, 3071: {1: 278, 2: 350}, 3072: {1: 279, 2: 350}, 3073: {1: 280, 2: 350}, 3074: {1: 281, 2: 350}, 3075: {1: 282, 2: 350}, 3076: {1: 283, 2: 350}, 3077: {1: 284, 2: 350}, 3078: {1: 285, 2: 350}, 3079: {1: 286, 2: 350}, 3080: {1: 287, 2: 350}, 3081: {1: 288, 2: 350}, 3082: {1: 289, 2: 350}, 3083: {1: 290, 2: 350}, 3084: {1: 291, 2: 350}, 3085: {1: 292, 2: 350}, 3086: {1: 293, 2: 350}, 3087: {1: 294, 2: 350}, 3088: {1: 295, 2: 350}, 3089: {1: 296, 2: 350}, 3090: {1: 297, 2: 350}, 3091: {1: 298, 2: 350}, 3092: {1: 299, 2: 350}, 3093: {1: 300, 2: 350}, 3094: {1: 301, 2: 350}, 3095: {1: 302, 2: 350}, 3096: {1: 303, 2: 350}, 3097: {1: 304, 2: 350}, 3098: {1: 305, 2: 350}, 3099: {1: 306, 2: 350}, 3100: {1: 307, 2: 350}, 3101: {1: 308, 2: 350}, 3102: {1: 309, 2: 350}, 3103: {1: 310, 2: 350}, 3104: {1: 311, 2: 350}, 3105: {1: 312, 2: 350}, 3106: {1: 313, 2: 350}, 3107: {1: 314, 2: 350}, 3108: {1: 315, 2: 350}, 3109: {1: 316, 2: 350}, 3110: {1: 317, 2: 350}, 3111: {1: 318, 2: 350}, 3112: {1: 319, 2: 350}, 3113: {1: 320, 2: 350}, 3114: {1: 321, 2: 350}, 3115: {1: 322, 2: 350}, 3116: {1: 323, 2: 350}, 3117: {1: 324, 2: 350}, 3118: {1: 325, 2: 350}, 3119: {1: 326, 2: 350}, 3120: {1: 327, 2: 350}, 3121: {1: 328, 2: 350}, 3122: {1: 329, 2: 350}, 3123: {1: 330, 2: 350}, 3124: {1: 331, 2: 350}, 3125: {1: 332, 2: 350}, 3126: {1: 333, 2: 350}, 3127: {1: 334, 2: 350}, 3128: {1: 335, 2: 350}, 3129: {1: 336, 2: 350}, 3130: {1: 337, 2: 350}, 3131: {1: 338, 2: 350}, 3132: {1: 339, 2: 350}, 3133: {1: 340, 2: 350}, 3134: {1: 341, 2: 350}, 3135: {1: 342, 2: 350}, 3136: {1: 343, 2: 350}, 3137: {1: 344, 2: 350}, 3138: {1: 345, 2: 350}, 3139: {1: 346, 2: 350}, 3140: {1: 347, 2: 350}, 3141: {1: 348, 2: 350}, 3142: {1: 0, 2: 351}, 3143: {1: 1, 2: 351}, 3144: {1: 2, 2: 351}, 3145: {1: 3, 2: 351}, 3146: {1: 4, 2: 351}, 3147: {1: 5, 2: 351}, 3148: {1: 6, 2: 351}, 3149: {1: 7, 2: 351}, 3150: {1: 8, 2: 351}, 3151: {1: 9, 2: 351}, 3152: {1: 10, 2: 351}, 3153: {1: 11, 2: 351}, 3154: {1: 12, 2: 351}, 3155: {1: 13, 2: 351}, 3156: {1: 14, 2: 351}, 3157: {1: 15, 2: 351}, 3158: {1: 16, 2: 351}, 3159: {1: 17, 2: 351}, 3160: {1: 18, 2: 351}, 3161: {1: 19, 2: 351}, 3162: {1: 20, 2: 351}, 3163: {1: 21, 2: 351}, 3164: {1: 22, 2: 351}, 3165: {1: 23, 2: 351}, 3166: {1: 24, 2: 351}, 3167: {1: 25, 2: 351}, 3168: {1: 26, 2: 351}, 3169: {1: 27, 2: 351}, 3170: {1: 28, 2: 351}, 3171: {1: 29, 2: 351}, 3172: {1: 30, 2: 351}, 3173: {1: 31, 2: 351}, 3174: {1: 32, 2: 351}, 3175: {1: 33, 2: 351}, 3176: {1: 34, 2: 351}, 3177: {1: 35, 2: 351}, 3178: {1: 36, 2: 351}, 3179: {1: 37, 2: 351}, 3180: {1: 38, 2: 351}, 3181: {1: 39, 2: 351}, 3182: {1: 40, 2: 351}, 3183: {1: 41, 2: 351}, 3184: {1: 42, 2: 351}, 3185: {1: 43, 2: 351}, 3186: {1: 44, 2: 351}, 3187: {1: 45, 2: 351}, 3188: {1: 46, 2: 351}, 3189: {1: 47, 2: 351}, 3190: {1: 48, 2: 351}, 3191: {1: 49, 2: 351}, 3192: {1: 50, 2: 351}, 3193: {1: 51, 2: 351}, 3194: {1: 52, 2: 351}, 3195: {1: 53, 2: 351}, 3196: {1: 54, 2: 351}, 3197: {1: 55, 2: 351}, 3198: {1: 56, 2: 351}, 3199: {1: 57, 2: 351}, 3200: {1: 58, 2: 351}, 3201: {1: 59, 2: 351}, 3202: {1: 60, 2: 351}, 3203: {1: 61, 2: 351}, 3204: {1: 62, 2: 351}, 3205: {1: 63, 2: 351}, 3206: {1: 64, 2: 351}, 3207: {1: 65, 2: 351}, 3208: {1: 66, 2: 351}, 3209: {1: 67, 2: 351}, 3210: {1: 68, 2: 351}, 3211: {1: 69, 2: 351}, 3212: {1: 70, 2: 351}, 3213: {1: 71, 2: 351}, 3214: {1: 72, 2: 351}, 3215: {1: 73, 2: 351}, 3216: {1: 74, 2: 351}, 3217: {1: 75, 2: 351}, 3218: {1: 76, 2: 351}, 3219: {1: 77, 2: 351}, 3220: {1: 78, 2: 351}, 3221: {1: 79, 2: 351}, 3222: {1: 80, 2: 351}, 3223: {1: 81, 2: 351}, 3224: {1: 82, 2: 351}, 3225: {1: 83, 2: 351}, 3226: {1: 84, 2: 351}, 3227: {1: 85, 2: 351}, 3228: {1: 86, 2: 351}, 3229: {1: 87, 2: 351}, 3230: {1: 88, 2: 351}, 3231: {1: 89, 2: 351}, 3232: {1: 90, 2: 351}, 3233: {1: 91, 2: 351}, 3234: {1: 92, 2: 351}, 3235: {1: 93, 2: 351}, 3236: {1: 94, 2: 351}, 3237: {1: 95, 2: 351}, 3238: {1: 96, 2: 351}, 3239: {1: 97, 2: 351}, 3240: {1: 98, 2: 351}, 3241: {1: 99, 2: 351}, 3242: {1: 100, 2: 351}, 3243: {1: 101, 2: 351}, 3244: {1: 102, 2: 351}, 3245: {1: 103, 2: 351}, 3246: {1: 104, 2: 351}, 3247: {1: 105, 2: 351}, 3248: {1: 106, 2: 351}, 3249: {1: 107, 2: 351}, 3250: {1: 108, 2: 351}, 3251: {1: 109, 2: 351}, 3252: {1: 110, 2: 351}, 3253: {1: 111, 2: 351}, 3254: {1: 112, 2: 351}, 3255: {1: 113, 2: 351}, 3256: {1: 114, 2: 351}, 3257: {1: 115, 2: 351}, 3258: {1: 116, 2: 351}, 3259: {1: 117, 2: 351}, 3260: {1: 118, 2: 351}, 3261: {1: 119, 2: 351}, 3262: {1: 120, 2: 351}, 3263: {1: 121, 2: 351}, 3264: {1: 122, 2: 351}, 3265: {1: 123, 2: 351}, 3266: {1: 124, 2: 351}, 3267: {1: 125, 2: 351}, 3268: {1: 126, 2: 351}, 3269: {1: 127, 2: 351}, 3270: {1: 128, 2: 351}, 3271: {1: 129, 2: 351}, 3272: {1: 130, 2: 351}, 3273: {1: 131, 2: 351}, 3274: {1: 132, 2: 351}, 3275: {1: 133, 2: 351}, 3276: {1: 134, 2: 351}, 3277: {1: 135, 2: 351}, 3278: {1: 136, 2: 351}, 3279: {1: 137, 2: 351}, 3280: {1: 138, 2: 351}, 3281: {1: 139, 2: 351}, 3282: {1: 140, 2: 351}, 3283: {1: 141, 2: 351}, 3284: {1: 142, 2: 351}, 3285: {1: 143, 2: 351}, 3286: {1: 144, 2: 351}, 3287: {1: 145, 2: 351}, 3288: {1: 146, 2: 351}, 3289: {1: 147, 2: 351}, 3290: {1: 148, 2: 351}, 3291: {1: 149, 2: 351}, 3292: {1: 150, 2: 351}, 3293: {1: 151, 2: 351}, 3294: {1: 152, 2: 351}, 3295: {1: 153, 2: 351}, 3296: {1: 154, 2: 351}, 3297: {1: 155, 2: 351}, 3298: {1: 156, 2: 351}, 3299: {1: 157, 2: 351}, 3300: {1: 158, 2: 351}, 3301: {1: 159, 2: 351}, 3302: {1: 160, 2: 351}, 3303: {1: 161, 2: 351}, 3304: {1: 162, 2: 351}, 3305: {1: 163, 2: 351}, 3306: {1: 164, 2: 351}, 3307: {1: 165, 2: 351}, 3308: {1: 166, 2: 351}, 3309: {1: 167, 2: 351}, 3310: {1: 168, 2: 351}, 3311: {1: 169, 2: 351}, 3312: {1: 170, 2: 351}, 3313: {1: 171, 2: 351}, 3314: {1: 172, 2: 351}, 3315: {1: 173, 2: 351}, 3316: {1: 174, 2: 351}, 3317: {1: 175, 2: 351}, 3318: {1: 176, 2: 351}, 3319: {1: 177, 2: 351}, 3320: {1: 178, 2: 351}, 3321: {1: 179, 2: 351}, 3322: {1: 180, 2: 351}, 3323: {1: 181, 2: 351}, 3324: {1: 182, 2: 351}, 3325: {1: 183, 2: 351}, 3326: {1: 184, 2: 351}, 3327: {1: 185, 2: 351}, 3328: {1: 186, 2: 351}, 3329: {1: 187, 2: 351}, 3330: {1: 188, 2: 351}, 3331: {1: 189, 2: 351}, 3332: {1: 190, 2: 351}, 3333: {1: 191, 2: 351}, 3334: {1: 192, 2: 351}, 3335: {1: 193, 2: 351}, 3336: {1: 194, 2: 351}, 3337: {1: 195, 2: 351}, 3338: {1: 196, 2: 351}, 3339: {1: 197, 2: 351}, 3340: {1: 198, 2: 351}, 3341: {1: 199, 2: 351}, 3342: {1: 200, 2: 351}, 3343: {1: 201, 2: 351}, 3344: {1: 202, 2: 351}, 3345: {1: 203, 2: 351}, 3346: {1: 204, 2: 351}, 3347: {1: 205, 2: 351}, 3348: {1: 206, 2: 351}, 3349: {1: 207, 2: 351}, 3350: {1: 208, 2: 351}, 3351: {1: 209, 2: 351}, 3352: {1: 210, 2: 351}, 3353: {1: 211, 2: 351}, 3354: {1: 212, 2: 351}, 3355: {1: 213, 2: 351}, 3356: {1: 214, 2: 351}, 3357: {1: 215, 2: 351}, 3358: {1: 216, 2: 351}, 3359: {1: 217, 2: 351}, 3360: {1: 218, 2: 351}, 3361: {1: 219, 2: 351}, 3362: {1: 220, 2: 351}, 3363: {1: 221, 2: 351}, 3364: {1: 222, 2: 351}, 3365: {1: 223, 2: 351}, 3366: {1: 224, 2: 351}, 3367: {1: 225, 2: 351}, 3368: {1: 226, 2: 351}, 3369: {1: 227, 2: 351}, 3370: {1: 228, 2: 351}, 3371: {1: 229, 2: 351}, 3372: {1: 230, 2: 351}, 3373: {1: 231, 2: 351}, 3374: {1: 232, 2: 351}, 3375: {1: 233, 2: 351}, 3376: {1: 234, 2: 351}, 3377: {1: 235, 2: 351}, 3378: {1: 236, 2: 351}, 3379: {1: 237, 2: 351}, 3380: {1: 238, 2: 351}, 3381: {1: 239, 2: 351}, 3382: {1: 240, 2: 351}, 3383: {1: 241, 2: 351}, 3384: {1: 242, 2: 351}, 3385: {1: 243, 2: 351}, 3386: {1: 244, 2: 351}, 3387: {1: 245, 2: 351}, 3388: {1: 246, 2: 351}, 3389: {1: 247, 2: 351}, 3390: {1: 248, 2: 351}, 3391: {1: 249, 2: 351}, 3392: {1: 250, 2: 351}, 3393: {1: 251, 2: 351}, 3394: {1: 252, 2: 351}, 3395: {1: 253, 2: 351}, 3396: {1: 254, 2: 351}, 3397: {1: 255, 2: 351}, 3398: {1: 256, 2: 351}, 3399: {1: 257, 2: 351}, 3400: {1: 258, 2: 351}, 3401: {1: 259, 2: 351}, 3402: {1: 260, 2: 351}, 3403: {1: 261, 2: 351}, 3404: {1: 262, 2: 351}, 3405: {1: 263, 2: 351}, 3406: {1: 264, 2: 351}, 3407: {1: 265, 2: 351}, 3408: {1: 266, 2: 351}, 3409: {1: 267, 2: 351}, 3410: {1: 268, 2: 351}, 3411: {1: 269, 2: 351}, 3412: {1: 270, 2: 351}, 3413: {1: 271, 2: 351}, 3414: {1: 272, 2: 351}, 3415: {1: 273, 2: 351}, 3416: {1: 274, 2: 351}, 3417: {1: 275, 2: 351}, 3418: {1: 276, 2: 351}, 3419: {1: 277, 2: 351}, 3420: {1: 278, 2: 351}, 3421: {1: 279, 2: 351}, 3422: {1: 280, 2: 351}, 3423: {1: 281, 2: 351}, 3424: {1: 282, 2: 351}, 3425: {1: 283, 2: 351}, 3426: {1: 284, 2: 351}, 3427: {1: 285, 2: 351}, 3428: {1: 286, 2: 351}, 3429: {1: 287, 2: 351}, 3430: {1: 288, 2: 351}, 3431: {1: 289, 2: 351}, 3432: {1: 290, 2: 351}, 3433: {1: 291, 2: 351}, 3434: {1: 292, 2: 351}, 3435: {1: 293, 2: 351}, 3436: {1: 294, 2: 351}, 3437: {1: 295, 2: 351}, 3438: {1: 296, 2: 351}, 3439: {1: 297, 2: 351}, 3440: {1: 298, 2: 351}, 3441: {1: 299, 2: 351}, 3442: {1: 300, 2: 351}, 3443: {1: 301, 2: 351}, 3444: {1: 302, 2: 351}, 3445: {1: 303, 2: 351}, 3446: {1: 304, 2: 351}, 3447: {1: 305, 2: 351}, 3448: {1: 306, 2: 351}, 3449: {1: 307, 2: 351}, 3450: {1: 308, 2: 351}, 3451: {1: 309, 2: 351}, 3452: {1: 310, 2: 351}, 3453: {1: 311, 2: 351}, 3454: {1: 312, 2: 351}, 3455: {1: 313, 2: 351}, 3456: {1: 314, 2: 351}, 3457: {1: 315, 2: 351}, 3458: {1: 316, 2: 351}, 3459: {1: 317, 2: 351}, 3460: {1: 318, 2: 351}, 3461: {1: 319, 2: 351}, 3462: {1: 320, 2: 351}, 3463: {1: 321, 2: 351}, 3464: {1: 322, 2: 351}, 3465: {1: 323, 2: 351}, 3466: {1: 324, 2: 351}, 3467: {1: 325, 2: 351}, 3468: {1: 326, 2: 351}, 3469: {1: 327, 2: 351}, 3470: {1: 328, 2: 351}, 3471: {1: 329, 2: 351}, 3472: {1: 330, 2: 351}, 3473: {1: 331, 2: 351}, 3474: {1: 332, 2: 351}, 3475: {1: 333, 2: 351}, 3476: {1: 334, 2: 351}, 3477: {1: 335, 2: 351}, 3478: {1: 336, 2: 351}, 3479: {1: 337, 2: 351}, 3480: {1: 338, 2: 351}, 3481: {1: 339, 2: 351}, 3482: {1: 340, 2: 351}, 3483: {1: 341, 2: 351}, 3484: {1: 342, 2: 351}, 3485: {1: 343, 2: 351}, 3486: {1: 344, 2: 351}, 3487: {1: 345, 2: 351}, 3488: {1: 346, 2: 351}, 3489: {1: 347, 2: 351}, 3490: {1: 348, 2: 351}, 3491: {1: 352, 2: 359}, 3492: {1: 353, 2: 359}, 3493: {1: 354, 2: 359}, 3494: {1: 355, 2: 359}, 3495: {1: 356, 2: 359}, 3496: {1: 357, 2: 359}, 3497: {1: 358, 2: 359}, 3498: {1: 0, 2: 359}, 3499: {1: 349, 2: 359}, 3500: {1: 350, 2: 359}, 3501: {1: 351, 2: 359}}, 'deo': {1: {1: 21.7454982145, 2: -3.26182473217}}, 'inno': {1: 348, 2: 347, 3: 346, 4: 345, 5: 344, 6: 343, 7: 342, 8: 341, 9: 340, 10: 339, 11: 338, 12: 337, 13: 336, 14: 335, 15: 334, 16: 333, 17: 332, 18: 331, 19: 330, 20: 329, 21: 328, 22: 327, 23: 326, 24: 325, 25: 324, 26: 323, 27: 322, 28: 321, 29: 320, 30: 319, 31: 318, 32: 317, 33: 316, 34: 315, 35: 314, 36: 313, 37: 312, 38: 311, 39: 310, 40: 309, 41: 308, 42: 307, 43: 306, 44: 305, 45: 304, 46: 303, 47: 302, 48: 301, 49: 300, 50: 299, 51: 298, 52: 297, 53: 296, 54: 295, 55: 294, 56: 293, 57: 292, 58: 291, 59: 290, 60: 289, 61: 288, 62: 287, 63: 286, 64: 285, 65: 284, 66: 283, 67: 282, 68: 281, 69: 280, 70: 279, 71: 278, 72: 277, 73: 276, 74: 275, 75: 274, 76: 273, 77: 272, 78: 271, 79: 270, 80: 269, 81: 268, 82: 267, 83: 266, 84: 265, 85: 264, 86: 263, 87: 262, 88: 261, 89: 260, 90: 259, 91: 258, 92: 257, 93: 256, 94: 255, 95: 254, 96: 253, 97: 252, 98: 251, 99: 250, 100: 249, 101: 248, 102: 247, 103: 246, 104: 245, 105: 244, 106: 243, 107: 242, 108: 241, 109: 240, 110: 239, 111: 238, 112: 237, 113: 236, 114: 235, 115: 234, 116: 233, 117: 232, 118: 231, 119: 230, 120: 229, 121: 228, 122: 227, 123: 226, 124: 225, 125: 224, 126: 223, 127: 222, 128: 221, 129: 220, 130: 219, 131: 218, 132: 217, 133: 216, 134: 215, 135: 214, 136: 213, 137: 212, 138: 211, 139: 210, 140: 209, 141: 208, 142: 207, 143: 206, 144: 205, 145: 204, 146: 203, 147: 202, 148: 201, 149: 200, 150: 199, 151: 198, 152: 197, 153: 196, 154: 195, 155: 194, 156: 193, 157: 192, 158: 191, 159: 190, 160: 189, 161: 188, 162: 187, 163: 186, 164: 185, 165: 184, 166: 183, 167: 182, 168: 181, 169: 180, 170: 179, 171: 178, 172: 177, 173: 176, 174: 175, 175: 174, 176: 173, 177: 172, 178: 171, 179: 170, 180: 169, 181: 168, 182: 167, 183: 166, 184: 165, 185: 164, 186: 163, 187: 162, 188: 161, 189: 160, 190: 159, 191: 158, 192: 157, 193: 156, 194: 155, 195: 154, 196: 153, 197: 152, 198: 151, 199: 150, 200: 149, 201: 148, 202: 147, 203: 146, 204: 145, 205: 144, 206: 143, 207: 142, 208: 141, 209: 140, 210: 139, 211: 138, 212: 137, 213: 136, 214: 135, 215: 134, 216: 133, 217: 132, 218: 131, 219: 130, 220: 129, 221: 128, 222: 127, 223: 126, 224: 125, 225: 124, 226: 123, 227: 122, 228: 121, 229: 120, 230: 119, 231: 118, 232: 117, 233: 116, 234: 115, 235: 114, 236: 113, 237: 112, 238: 111, 239: 110, 240: 109, 241: 108, 242: 107, 243: 106, 244: 105, 245: 104, 246: 103, 247: 102, 248: 101, 249: 100, 250: 99, 251: 98, 252: 97, 253: 96, 254: 95, 255: 94, 256: 93, 257: 92, 258: 91, 259: 90, 260: 89, 261: 88, 262: 87, 263: 86, 264: 85, 265: 84, 266: 83, 267: 82, 268: 81, 269: 80, 270: 79, 271: 78, 272: 77, 273: 76, 274: 75, 275: 74, 276: 73, 277: 72, 278: 71, 279: 70, 280: 69, 281: 68, 282: 67, 283: 66, 284: 65, 285: 64, 286: 63, 287: 62, 288: 61, 289: 60, 290: 59, 291: 58, 292: 57, 293: 56, 294: 55, 295: 54, 296: 53, 297: 52, 298: 51, 299: 50, 300: 49, 301: 48, 302: 47, 303: 46, 304: 45, 305: 44, 306: 43, 307: 42, 308: 41, 309: 40, 310: 39, 311: 38, 312: 37, 313: 36, 314: 35, 315: 34, 316: 33, 317: 32, 318: 31, 319: 30, 320: 29, 321: 28, 322: 27, 323: 26, 324: 25, 325: 24, 326: 23, 327: 22, 328: 21, 329: 20, 330: 19, 331: 18, 332: 17, 333: 16, 334: 15, 335: 14, 336: 13, 337: 12, 338: 11, 339: 10, 340: 9, 341: 8, 342: 7, 343: 6, 344: 5, 345: 4, 346: 3, 347: 2, 348: 1}})
    return nets

print("\nNNScore 2.02")
print("============\n")

print((textwrap.fill("NNScore 2.02 is released under the GNU General Public License (see http://www.gnu.org/licenses/gpl.html). If you have any questions, comments, or suggestions, please contact the author, Jacob Durrant, at durrantj [at] pitt [dot] edu. If you use NNScore 2.02 in your work, please cite NNScore 2.0: A Neural-Network Receptor-Ligand Scoring Function. Jacob D. Durrant, Andrew McCammon. Journal of Chemical Information and Modeling, 2011, 51 (11), pp 2897-2903.")))

print("")

print((textwrap.fill("NNScore 2.02 is based in part on the ffnet python package developed by Marek Wojciechowski (http://ffnet.sourceforge.net/).")))
print("")

print("Use the -help command-line parameter for extended help.\n")
print("Example: python NNScore2.py -receptor myreceptor.pdbqt -ligand myligand.pdbqt -vina_executable /PATH/TO/VINA/1.1.2/vina\n")

cmd_params = command_line_parameters(sys.argv[:])

if cmd_params.okay_to_proceed() is False:
    print("ERROR: You need to specify the ligand and receptor PDBQT files, as\nwell as the full path the the AutoDock Vina 1.1.2 executable, using the\n-receptor, -ligand, and -vina_executable tags from the command line.\nThe -ligand tag can also specify an AutoDock Vina output file.\n")
    sys.exit(0)

lig = cmd_params.params['ligand']
rec = cmd_params.params['receptor']

def calculate_score(lig, rec, cmd_params, actual_filename_if_lig_is_list="", actual_filename_if_rec_is_list="", line_header = ""):

    d = binana(lig, rec, cmd_params, line_header, actual_filename_if_lig_is_list, actual_filename_if_rec_is_list)

    # now load in neural networks
    scores = []
    total = 0.0
    nets = networks()
    for net_array in nets:
        try:
            net = ffnet()
            net.load(net_array)

            val = net.normcall(d.input_vector)
            print((line_header + "Network #" + str(len(scores) + 1) + " gave a score of " + str( round(val,3)) + " (" + score_to_kd(val) + ")"))
            scores.append(val)
            total = total + val
        except OverflowError:
            print((line_header + "The output of network #" + str(len(scores) + 1) + " could not be determined because of an overflow error!"))

    if len(scores) != 0:
        average = total / len(scores)

        best_score = 0.0
        best_network = -1
        count = 1
        sum = 0.0
        for score in scores:
            if score > best_score:
                best_score = score
                best_network = count
            sum = sum + math.pow(score - average,2)
            count = count + 1
        stdev = math.pow(sum / (len(scores)-1), 0.5)

        print("")

        print((line_header + "Best Score:         ", round(best_score,3), "(" + score_to_kd(best_score) + ")"))
        print((line_header + "Average Score:      ", round(average,3), "(" + score_to_kd(average) + ")"))
        print((line_header + "Standard Deviation: ", round(stdev,3)))
        print("")

        return [average, stdev, best_score, best_network, scores]
    else:
        print((line_header + "Could not compute the score of this receptor-ligand complex because none of the networks returned a valid score."))
        return [0, 0, 0, 0, []]

# load the rec into an array so you only have to load it from the disk once
print("\nLOADING THE RECEPTOR")
print("====================\n")

#f2 = open(rec,'r')
#rec_array = f2.readlines()
#f2.close()
receptor = PDB()
receptor.LoadPDB_from_file(rec)
receptor.OrigFileName = rec

print("\nEVALUATING EACH OF THE POSES IN THE LIGAND FILE USING 20 TRAINED NEURAL NETWORKS")
print("================================================================================\n")

# determine if the ligand input file is a single pdbqt or an autodock vina output file. Both are acceptable inputs.

f = open(lig,'r')

lig_array = []
line = "NULL"

scores = []

model_id = 1
while len(line) != 0:
    line = f.readline()
    if line[:6] != "ENDMDL": lig_array.append(line)
    if line[:6] == "ENDMDL" or len(line) == 0:
        if len(lig_array) != 0 and lig_array != ['']:
            temp_filename = lig + ".MODEL_" + str(model_id) + ".pdbqt"

            temp_f = open(temp_filename, 'w')
            for ln in lig_array: temp_f.write(ln)
            temp_f.close()

            model_name = "MODEL " + str(model_id)

            print(model_name)

            score=calculate_score(lig_array, receptor, cmd_params, temp_filename, rec, "\t")
            scores.append([score[0], score[1], score[2], score[3], score[4], model_name])

            os.remove(temp_filename)

            lig_array = []
            model_id = model_id + 1

f.close()

# now find the best score across all models, for each network
print("\nRANKED POSES AND SCORES WHEN EACH OF THE 20 NETWORKS IS CONSIDERED SEPARATELY")
print("=============================================================================\n")

best_network_scores = []
for t in range(20):
    net_scores = []
    for score in scores:
        net_scores.append((score[4][t],score[5]))
    net_scores = sorted(net_scores,key=lambda net_score: net_score[0], reverse=True) # sort by score
    print(("USING NETWORK #" + str(t+1)).center(42," "))
    print(" Rank | Pose     | Score | Predicted Kd ")
    print("------+----------+-------+--------------")
    count = 1
    best_network_scores.append(net_scores[0])
    for net_score in net_scores:
        print((str(count).center(6," ") + "|" + net_score[1].center(10," ") + "|" + str(round(net_score[0],3)).center(7," ") + "|" + score_to_kd(net_score[0]).replace("Kd = ","").center(14," ")))
        count = count + 1
    print("")

print("\nRANKED POSES AND SCORES WHEN A CONSENSUS OF NETWORK OUTPUTS ARE CONSIDERED")
print("==========================================================================\n")

# now get the best score of all vina output poses
best_scores = sorted(scores,key=lambda score: score[2], reverse=True) # sort by score

best_best_score = best_scores[0]

print(("BEST SCORE OF ALL 20 NETWORKS, BY POSE".center(65," ")))
print(" Rank | Pose     | Average Score | Predicted Kd | Network")
print("------+----------+---------------+--------------+---------")

count = 1
for score in best_scores:
    print((str(count).center(6," ") + "|" + score[5].center(10," ") + "|" + str(round(score[2],3)).center(15," ") + "|" + score_to_kd(score[2]).replace("Kd = ","").center(14," ") + "|" + ("#" + str(score[3])).center(9," ")))
    count = count + 1

print("")

# now get the best average score of all vina output poses
average_scores = sorted(scores,key=lambda score: score[0], reverse=True) # sort by score

best_of_average_scores = average_scores[0]

print(("AVERAGE SCORE OF ALL 20 NETWORKS, BY POSE".center(70," ")))
print(" Rank | Pose     | Average Score | Standard Deviation | Predicted Kd")
print("------+----------+---------------+--------------------+--------------")

count = 1
for score in average_scores:
    print((str(count).center(6," ") + "|" + score[5].center(10," ") + "|" + str(round(score[0],3)).center(15," ") + "|" + str(round(score[1],3)).center(20," ") + "|" + score_to_kd(score[0]).replace("Kd = ","").center(15," ")))
    count = count + 1

print("")
print("\nSUMMARY")
print("=======\n")

count = 1
for best_network_score in best_network_scores:
    print((textwrap.fill("Best pose scored by network #" + str(count) + ": " + best_network_score[1] + " (Score = " + str(round(best_network_score[0],3)) + " = " + score_to_kd(best_network_score[0]).replace("Kd = ","") + ")")))
    count = count + 1

print("")

print((textwrap.fill("When the poses were ranked by the best of the 20 network scores associated with each pose, the best-scoring pose was " + best_best_score[5] + " (Score = " + str(round(best_best_score[2],3)) + " = " + score_to_kd(best_best_score[2]).replace("Kd = ","") + ")")))
print("")
print((textwrap.fill("When the poses were ranked by the average of the 20 network scores associated with each pose, the best-scoring pose was " + best_of_average_scores[5] + " (Score = " + str(round(best_of_average_scores[0],3)) + " +/- " + str(round(best_of_average_scores[1],3)) + " = " + score_to_kd(best_of_average_scores[0]).replace("Kd = ","") + "). This is the recommended ranking/scoring metric.")))

print("")
